#!/usr/bin/python
#
#   splbuild_cygwin_builder.pl
#
#   SPL component builder: A builder for Win32 x86 target that uses cygwin toolchain
#
#   $Date: 2008/08/06 12:06:40 $
#   $Revision: 1.13 $
#   $Author: akapusti $
#
#   Copyright (C) 2007-2008 Cybernetic Intelligence GmbH
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##########
import sys, os
from splbuild_builder import *
from splbuild_project import *
from splbuild_project_reader import *

class CygwinBuilder(Builder):
    """A builder for Win32 x86 target that uses cygwin toolchain"""
    
    __CommandLineLengthLimit = 8192
    """Do not emit command lines longer than that into a makefile"""
    
    def __init__(self):
        """The class constructor"""
        Builder.__init__(self, ".o")

    def canHandle(self, targetProcessor, targetOs, toolchain):
        """Returns True iff this builder can perform builds for the specified
        target, False otherwise"""
        if (targetProcessor != "80386" and
            targetProcessor != "80486" and
            targetProcessor.lower() != "pentium" and
            targetProcessor.lower() != "pentiumpro" and
            targetProcessor.lower() != "pentium2" and
            targetProcessor.lower() != "pentium3" and
            targetProcessor.lower() != "pentium4"):
            return False
        if targetOs.lower() != "windows":
            return False
        if toolchain.lower() != "cygwin":
            return False
        return True
    
    def setupDefaults(self, project):
        """Given a project, assigns the Builder-specific default values
        to those project properties that were not defined explicitly"""
        assert project.builder is self
        if project.cCompiler is None:
            project.cCompiler = "gcc"
        if project.cppCompiler is None:
            project.cppCompiler = "g++"
        if project.splCompiler is None:
            project.splCompiler = "spl"
        if project.assembler is None:
            project.assembler = "as"
        if project.linker is None:
            project.linker = "g++"
        if project.librarian is None:
            project.librarian = "ar"
        if project.make is None:
            project.make = "make"
        project.makefileName = project.projectFileName + ".mk"          #   The name of the temporary makefile to use
        project.splViaFileName = project.projectFileName + ".splvia"    #   The name of the temporary SPL -Via file to use

    def validate(self, project):
        """Given a project, checks it for validity"""
        #   The extension of the main output file must
        #   match that required by the project type
        projectOwnFileName = os.path.basename(project.output)
        if project.type == TargetType.Executable:
            extensionGood = (projectOwnFileName.lower().endswith(".exe"))
        elif project.type == TargetType.StaticLibrary:
            extensionGood = (projectOwnFileName.lower().endswith(".a"))
        elif project.type == TargetType.DynamicLibrary:
            extensionGood = (projectOwnFileName.lower().endswith(".dll"))
        else:
            extensionGood = False
        if not extensionGood:
            print "Invalid extension of the project '" + project.projectFileName + "' output"
            sys.exit(1)
        #   Library names must start with "lib"
        if (project.type == TargetType.StaticLibrary) or (project.type == TargetType.DynamicLibrary):
            if not projectOwnFileName.startswith("lib"):
                print "Output library name of project '" + project.projectFileName + "' must start with 'lib'"
                sys.exit(1)

    def getFileType(self, filename):
        """A helper service for deriving a file type from file name.
        Returns the file type or None if the file type cannot be determined"""
        if filename.lower().endswith(".c"):
            return FileType.CSource
        if filename.lower().endswith(".h"):
            return FileType.CHeader
        if filename.lower().endswith(".cpp") or filename.lower().endswith(".cxx"):
            return FileType.CppSource
        if filename.lower().endswith(".hpp") or filename.lower().endswith(".hxx"):
            return FileType.CppHeader
        if filename.lower().endswith(".spl"):
            return FileType.SplSource
        if filename.lower().endswith(".sph"):
            return FileType.SplHeader
        if filename.lower().endswith(".s"):
            return FileType.AssemblerSource
        if filename.lower().endswith(".inc"):
            return FileType.AssemblerHeader
        if filename.lower().endswith(".o"):
            return FileType.ObjectModule
        if filename.lower().endswith(".a"):
            return FileType.StaticLibrary
        if filename.lower().endswith(".dll"):
            return FileType.DynamicLibrary
        return None

    def writeMakefileToolsAndOptions(self, makefile, project, projectClosure):
        """Writes make definitions for tool and option strings"""
        makefile.write("CTOOL=" + self.translateFileName(project.cCompiler) + "\n")
        makefile.write("COPTS=" + self.__prepareStandardCCompilerOptions(project, projectClosure) + " " + project.cCompilerOptions + "\n")
        makefile.write("CPPTOOL=" + self.translateFileName(project.cppCompiler) + "\n")
        makefile.write("CPPOPTS=" + self.__prepareStandardCppCompilerOptions(project, projectClosure) + " " + project.cppCompilerOptions + "\n")
        makefile.write("SPLTOOL=" + self.translateFileName(project.splCompiler) + "\n")
        makefile.write("SPLOPTS=" + project.splCompilerOptions + "\n")
        makefile.write("ASMTOOL=" + self.translateFileName(project.assembler) + "\n")
        makefile.write("ASMOPTS=" + self.__prepareStandardAssemblerOptions(project, projectClosure) + " " + project.assemblerOptions + "\n")
        makefile.write("LINKTOOL=" + self.translateFileName(project.linker) + "\n")
        makefile.write("LINKOPTS=" + self.__prepareStandardLinkerOptions(project) + " " + project.linkerOptions + "\n")
        makefile.write("LIBTOOL=" + self.translateFileName(project.librarian) + "\n")
        makefile.write("LIBOPTS=" + self.__prepareStandardLibrarianOptions(project) + " " + project.librarianOptions + "\n")
        
    def prepareMakeCommandLine(self, project):
        """Prepares and returns a "make" command line for the project
        in the form of argument list (one element per argument)"""
        return [project.make, "-f", project.makefileName] + self.breakArgumentString(project.makeOptions)
    
    def translateFileName(self, fileName):
        """Given a name of a file, creates and returns the form in which
        this filename must be present in the makefile"""
        #   Translate for toolchain
        if (len(fileName) > 2) and (fileName[1] == ':'):
            fileName = "/cygdrive/" + fileName[0].lower() + fileName[2:]
            fileName = fileName.replace("\\", "/")
        #   Done
        return fileName

    def writeCCompilerRule(self, makefile, sourceFileName, outputFileName):
        """Writes a makefile rule for invoking a C compiler with specified
        source and output filenames"""
        makefile.write("\t$(CTOOL) $(COPTS) -o " + self.translateFileName(outputFileName) + " " + self.translateFileName(sourceFileName) + "\n")

    def writeCppCompilerRule(self, makefile, sourceFileName, outputFileName):
        """Writes a makefile rule for invoking a C++ compiler with specified
        source and output filenames"""
        makefile.write("\t$(CPPTOOL) $(CPPOPTS) -o " + self.translateFileName(outputFileName) + " " + self.translateFileName(sourceFileName) + "\n")

    def writeSplCompilerRule(self, makefile, sourceFileName, outputFileName, dependencyFileName):
        """Writes a makefile rule for invoking an SPL compiler with specified
        source and output filenames"""
        print "SPL compilation not yet implemented"
        sys.exit(1)
        makefile.write("\t$(SPLTOOL) $(SPLOPTS) -output:" + self.translateFileName(outputFileName) + " " + self.translateFileName(sourceFileName) + "\n")

    def writeAssemblerRule(self, makefile, sourceFileName, outputFileName):
        """Writes a makefile rule for invoking an assembler with specified
        source and output filenames"""
        makefile.write("\t$(ASMTOOL) $(ASMOPTS) -o " + self.translateFileName(outputFileName) + " " + self.translateFileName(sourceFileName) + "\n")

    def writeExeLinkerRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName):
        """Writes a makefile rule for invoking a linker with the specified
        object files and libraries to create an executable"""
        self.__writeLinkerRule(makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName, "")

    def writeDllLinkerRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName):
        """Writes a makefile rule for invoking a linker with the specified
        object files and libraries to create a dynamic library"""
        extraOptions = "-Wl,--out-implib=" + self.translateFileName(outputFileName) + ".a"
        self.__writeLinkerRule(makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName, extraOptions)

    def writeLibrarianRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName):
        """Writes a makefile rule for invoking a librarian with the specified
        object files and libraries to create a static library"""
        makefile.write("\trm -f " + self.translateFileName(outputFileName) + "\n")
        inputFiles = objectFileNames + staticLibraryFileNames
        #for systemLibraryFileName in systemLibraryFileNames:
        #    inputFiles.append("/usr/lib/" + systemLibraryFileName)
        #   Now add members by chunks not longer than the command line length limit
        i = 0
        while i < len(inputFiles):
            inputChunk = ""
            while (i < len(inputFiles)) and (len(inputChunk) < CygwinBuilder.__CommandLineLengthLimit):
                inputChunk = inputChunk + " " + self.translateFileName(inputFiles[i])
                i = i + 1
            makefile.write("\t$(LIBTOOL) -rc $(LIBOPTS) " + self.translateFileName(outputFileName) + inputChunk + "\n")

    def __prepareGccTargetProcessorOptions(self, project):
        """Given a project, prepares the G++ options for target processor specification"""
        if project.targetProcessor == "80386":
            return "-march=i386"
        elif project.targetProcessor == "80486":
            return "-march=i486"
        elif project.targetProcessor.lower() == "pentium":
            return "-march=pentium"
        elif project.targetProcessor.lower() == "pentiumpro":
            return "-march=pentiumpro"
        elif project.targetProcessor.lower() == "pentium2":
            return "-march=pentium2"
        elif project.targetProcessor.lower() == "pentium3":
            return "-march=pentium3"
        elif project.targetProcessor.lower() == "pentium4":
            return "-march=pentium4"
        else:
            print "Unsupported target processor " + project.targetProcessor
            sys.exit(1)
 
    def __prepareGccDefineOptions(self, project):
        """Given a project, prepares the cl.exe options for defines"""
        result = ""
        for define in project.defines:
            if len(result) != 0:
                result = result + " "
            result = result + "-D" + define
        return result

    def __prepareGccAsIncludePathOptions(self, project, projectClosure):
        """Given a project, prepares the g++.exe options for include paths"""
        result = ""
        writtenIncludePaths = []
        #   Explicit include paths
        for someProjectFile in projectClosure.keys():
            someProject = projectClosure[someProjectFile]
            if (someProject.type <> TargetType.Executable) or (someProject.projectFileName == project.projectFileName):
                #   Dependencies which are executables do NOT contribute headers
                for includePath in someProject.includePaths:
                    if includePath not in writtenIncludePaths:
                        writtenIncludePaths.append(includePath) 
                        if len(result) != 0:
                            result = result + " "
                        result = result + "-I\"" + self.translateFileName(includePath) + "\""
        #   Done
        return result

    def __prepareStandardCCompilerOptions(self, project, projectClosure):
        """Given a project, creates & returns the standard options string
        for a C compiler"""
        result = "-c -DWIN32 -D_WINDOWS -D_MBCS -std=c89 -Wall -mthreads -m32 -mcygwin -Wno-long-long"
        if project.mode == BuildMode.Debug:
            result = result + " -O0 -D_DEBUG -gcoff -g"
            if project.standardLibrary == StandardLibrary.Dynamic:
                result = result + " -D_DLL -mdll -D_USRDLL -D_WINDLL"
        else:
            result = result + " -O3 -momit-leaf-frame-pointer -DNDEBUG"
            if project.standardLibrary == StandardLibrary.Dynamic:
                result = result + " -D_DLL -mdll -D_USRDLL -D_WINDLL"
        return result + " " + self.__prepareGccTargetProcessorOptions(project) + " " + self.__prepareGccDefineOptions(project) + " " + self.__prepareGccAsIncludePathOptions(project, projectClosure)
    
    def __prepareStandardCppCompilerOptions(self, project, projectClosure):
        """Given a project, creates & returns the standard options string
        for a C++ compiler"""
        result = "-c -DWIN32 -D_WINDOWS -D_MBCS -std=c++98 -fexceptions -frtti -Wall -mthreads -m32 -mcygwin -Wno-long-long"
        if project.mode == BuildMode.Debug:
            result = result + " -O0 -D_DEBUG -gcoff -g"
            if project.standardLibrary == StandardLibrary.Dynamic:
                result = result + " -D_DLL -mdll -D_USRDLL -D_WINDLL"
        else:
            result = result + " -O3 -momit-leaf-frame-pointer -fno-enforce-eh-specs -DNDEBUG"
            if project.standardLibrary == StandardLibrary.Dynamic:
                result = result + " -D_DLL -mdll -D_USRDLL -D_WINDLL"
        return result + " " + self.__prepareGccTargetProcessorOptions(project) + " " + self.__prepareGccDefineOptions(project) + " " + self.__prepareGccAsIncludePathOptions(project, projectClosure)

    def __prepareStandardAssemblerOptions(self, project, projectClosure):
        """Given a project, creates & returns the standard options string
        for an assembler"""
        return self.__prepareGccAsIncludePathOptions(project, projectClosure)
    
    def __prepareStandardLinkerOptions(self, project):
        """Given a project, creates & returns the standard options string
        for a linker"""
        result = ""
        if project.mode == BuildMode.Debug:
            result = result + " -g"
        if project.type == TargetType.DynamicLibrary:
            result = result + " -shared -Wl,--enable-auto-import"
        if project.standardLibrary == StandardLibrary.Static:
            result = result + " -Wl,-Bstatic"
        else:
            result = result + " -Wl,-Bdynamic"
        return result
    
    def __prepareStandardLibrarianOptions(self, project):
        """Given a project, creates & returns the standard options string
        for a librarian"""
        return ""

    def __writeLinkerRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName, extraOptions):
        """Writes a makefile rule for invoking a linker with the specified
        object files and libraries to create a dynamic library or executable"""
        #   To avoid hitting the limit of the command line length, all object files
        #   are first assembled into a static library, which is then linked as-a-whole
        objectFileContainerFileName = outputFileName + ".tmp.a"
        self.writeLibrarianRule(makefile, objectFileNames, [], [], [], objectFileContainerFileName)
        #   Finally, assemble the link rule
        makefile.write("\t$(LINKTOOL) $(LINKOPTS) -o " + self.translateFileName(outputFileName) + " " + extraOptions)
        #   Library search paths...
        librarySearchPaths = []
        for staticLibraryFileName in staticLibraryFileNames:
            directory = os.path.dirname(staticLibraryFileName)
            if directory not in librarySearchPaths:
                librarySearchPaths.append(directory)
                makefile.write(" -L" + self.translateFileName(directory))
        for dynamicLibraryFileName in dynamicLibraryFileNames:
            directory = os.path.dirname(dynamicLibraryFileName)
            if directory not in librarySearchPaths:
                librarySearchPaths.append(directory)
                makefile.write(" -L" + self.translateFileName(directory))
        #   Object files...
        makefile.write(" -Wl,--whole-archive " + self.translateFileName(objectFileContainerFileName) + " -Wl,--no-whole-archive")
        #   Static libraries...
        makefile.write(" -Wl,--start-group")
        for staticLibraryFileName in staticLibraryFileNames:
            libraryName = os.path.basename(staticLibraryFileName)
            if libraryName.startswith("lib"):
                libraryName = libraryName[3:]
            if libraryName.endswith(".a"):
                libraryName = libraryName[:-2]
            makefile.write(" -l" + libraryName)
        #   Dynamic libraries
        for dynamicLibraryFileName in dynamicLibraryFileNames:
            libraryName = os.path.basename(dynamicLibraryFileName)
            if libraryName.startswith("lib"):
                libraryName = libraryName[3:]
            if libraryName.endswith(".dll"):
                libraryName = libraryName[:-4]
            makefile.write(" -l" + libraryName)
        #   System libraries
        for systemLibraryFileName in systemLibraryFileNames:
            libraryName = os.path.basename(systemLibraryFileName)
            if libraryName.startswith("lib"):
                libraryName = libraryName[3:]
            if libraryName.endswith(".a"):
                libraryName = libraryName[:-2]
            if libraryName.endswith(".dll"):
                libraryName = libraryName[:-4]
            makefile.write(" -l" + libraryName)
        #   Done
        makefile.write(" -Wl,--end-group\n")
        #   Remove the temporary object file container
        makefile.write("\trm -f " + self.translateFileName(objectFileContainerFileName) + "\n")
        
#   Plug-in initialization: register the builder
Builder.register(CygwinBuilder())

#   End of splbuild_cygwin_builder.pl
