#!/usr/bin/python
#
#   splbuild_msvc2003_builder.pl
#
#   SPL component builder: A builder for Win32 x86 target that uses MSVC2003 toolchain
#
#   $Date: 2008/08/06 12:06:40 $
#   $Revision: 1.15 $
#   $Author: akapusti $
#
#   Copyright (C) 2007-2008 Cybernetic Intelligence GmbH
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##########
import sys
from splbuild_builder import *
from splbuild_project import *
from splbuild_project_reader import *

class Msvc2003Builder(Builder):
    """A builder for Win32 x86 target that uses MSVC2003 toolchain"""
    
    def __init__(self):
        """The class constructor"""
        Builder.__init__(self, ".obj")

    def canHandle(self, targetProcessor, targetOs, toolchain):
        """Returns True iff this builder can perform builds for the specified
        target, False otherwise"""
        if (targetProcessor != "80386" and
            targetProcessor != "80486" and
            targetProcessor.lower() != "pentium" and
            targetProcessor.lower() != "pentiumpro" and
            targetProcessor.lower() != "pentium2" and
            targetProcessor.lower() != "pentium3" and
            targetProcessor.lower() != "pentium4"):
            return False
        if targetOs.lower() != "windows":
            return False
        if toolchain.lower() != "msvc2003":
            return False
        return True
    
    def setupDefaults(self, project):
        """Given a project, assigns the Builder-specific default values
        to those project properties that were not defined explicitly"""
        assert project.builder is self
        if project.cCompiler is None:
            project.cCompiler = "cl.exe"
        if project.cppCompiler is None:
            project.cppCompiler = "cl.exe"
        if project.splCompiler is None:
            project.splCompiler = "spl.exe"
        if project.assembler is None:
            project.assembler = "ml.exe"
        if project.linker is None:
            project.linker = "link.exe"
        if project.librarian is None:
            project.librarian = "lib.exe"
        if project.make is None:
            project.make = "nmake.exe"
        project.makefileName = project.projectFileName + ".mk"          #   The name of the temporary makefile to use
        project.splViaFileName = project.projectFileName + ".splvia"    #   The name of the temporary SPL -Via file to use

    def validate(self, project):
        """Given a project, checks it for validity"""
        #   The extension of the main output file must
        #   match that required by the project type
        projectOwnFileName = os.path.basename(project.output)
        if project.type == TargetType.Executable:
            extensionGood = (projectOwnFileName.lower().endswith(".exe"))
        elif project.type == TargetType.StaticLibrary:
            extensionGood = (projectOwnFileName.lower().endswith(".lib"))
        elif project.type == TargetType.DynamicLibrary:
            extensionGood = (projectOwnFileName.lower().endswith(".dll"))
        else:
            extensionGood = False
        if not extensionGood:
            print "Invalid extension of the project '" + project.projectFileName + "' output"
            sys.exit(1)

    def getFileType(self, filename):
        """A helper service for deriving a file type from file name.
        Returns the file type or None if the file type cannot be determined"""
        if filename.lower().endswith(".c"):
            return FileType.CSource
        if filename.lower().endswith(".h"):
            return FileType.CHeader
        if filename.lower().endswith(".cpp") or filename.lower().endswith(".cxx"):
            return FileType.CppSource
        if filename.lower().endswith(".hpp") or filename.lower().endswith(".hxx"):
            return FileType.CppHeader
        if filename.lower().endswith(".spl"):
            return FileType.SplSource
        if filename.lower().endswith(".sph"):
            return FileType.SplHeader
        if filename.lower().endswith(".asm"):
            return FileType.AssemblerSource
        if filename.lower().endswith(".inc"):
            return FileType.AssemblerHeader
        if filename.lower().endswith(".obj"):
            return FileType.ObjectModule
        if filename.lower().endswith(".lib"):
            return FileType.StaticLibrary
        if filename.lower().endswith(".dll"):
            return FileType.DynamicLibrary
        return None

    def writeMakefileToolsAndOptions(self, makefile, project, projectClosure):
        """Writes make definitions for tool and option strings"""
        makefile.write("CTOOL=" + self.translateFileName(project.cCompiler) + "\n")
        makefile.write("COPTS=" + self.__prepareStandardCCompilerOptions(project, projectClosure) + " " + project.cCompilerOptions + "\n")
        makefile.write("CPPTOOL=" + self.translateFileName(project.cppCompiler) + "\n")
        makefile.write("CPPOPTS=" + self.__prepareStandardCppCompilerOptions(project, projectClosure) + " " + project.cppCompilerOptions + "\n")
        makefile.write("SPLTOOL=" + self.translateFileName(project.splCompiler) + "\n")
        makefile.write("SPLOPTS=" + project.splCompilerOptions + " -via:" + self.translateFileName(project.splViaFileName) + "\n")
        makefile.write("ASMTOOL=" + self.translateFileName(project.assembler) + "\n")
        makefile.write("ASMOPTS=" + self.__prepareStandardAssemblerOptions(project, projectClosure) + " " + project.assemblerOptions + "\n")
        makefile.write("LINKTOOL=" + self.translateFileName(project.linker) + "\n")
        makefile.write("LINKOPTS=" + self.__prepareStandardLinkerOptions(project) + " " + project.linkerOptions + "\n")
        makefile.write("LIBTOOL=" + self.translateFileName(project.librarian) + "\n")
        makefile.write("LIBOPTS=" + self.__prepareStandardLibrarianOptions(project) + " " + project.librarianOptions + "\n")
        #   Must also write the VIA file for the SPL compiler
        try:
            viaFile = open(project.splViaFileName, "w")
            viaFile.write(self._prepareStandardSplCompilerOptions(project, projectClosure) + " " + self.__prepareSplTargetOptions(project) + "\n")
            viaFile.close() 
        except IOError:
            print "Error writing the SPL compiler VIA file '" + project.splViaFileName + "'"
            sys.exit(1)

    def prepareMakeCommandLine(self, project):
        """Prepares and returns a "make" command line for the project
        in the form of argument list (one element per argument)"""
        return [project.make, "/nologo", "/F", project.makefileName] + self.breakArgumentString(project.makeOptions)

    def translateFileName(self, fileName):
        """Given a name of a file, creates and returns the form in which
        this filename must be present in the makefile"""
        return fileName.replace("/", "\\")

    def writeCCompilerRule(self, makefile, sourceFileName, outputFileName):
        """Writes a makefile rule for invoking a C compiler with specified
        source and output filenames"""
        makefile.write("\t$(CTOOL) $(COPTS) /Fo" + self.translateFileName(outputFileName) + " " + self.translateFileName(sourceFileName) + "\n")

    def writeCppCompilerRule(self, makefile, sourceFileName, outputFileName):
        """Writes a makefile rule for invoking a C++ compiler with specified
        source and output filenames"""
        makefile.write("\t$(CPPTOOL) $(CPPOPTS) /Fo" + self.translateFileName(outputFileName) + " " + self.translateFileName(sourceFileName) + "\n")

    def writeSplCompilerRule(self, makefile, sourceFileName, outputFileName, dependencyFileName):
        """Writes a makefile rule for invoking an SPL compiler with specified
        source and output filenames"""
#        tempCppFileName = outputFileName + ".cpp"
#        makefile.write("\t$(SPLTOOL) $(SPLOPTS) -output:" + self.translateFileName(tempCppFileName) + " -dependencies:" + self.translateFileName(dependencyFileName) + " " + self.translateFileName(sourceFileName) + "\n")
#        self.writeCppCompilerRule(makefile, tempCppFileName, outputFileName)
        makefile.write("\t$(SPLTOOL) $(SPLOPTS) -output:" + self.translateFileName(outputFileName) + " -dependencies:" + self.translateFileName(dependencyFileName) + " " + self.translateFileName(sourceFileName) + "\n")

    def writeAssemblerRule(self, makefile, sourceFileName, outputFileName):
        """Writes a makefile rule for invoking an assembler with specified
        source and output filenames"""
        makefile.write("\t$(ASMTOOL) $(ASMOPTS) /Fo" + self.translateFileName(outputFileName) + " " + self.translateFileName(sourceFileName) + "\n")

    def writeExeLinkerRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName):
        """Writes a makefile rule for invoking a linker with the specified
        object files and libraries to create an executable"""
        self.__writeLinkerRule(makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName, "")

    def writeDllLinkerRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName):
        """Writes a makefile rule for invoking a linker with the specified
        object files and libraries to create a dynamic library"""
        importLibraryFileName = outputFileName
        if importLibraryFileName.lower().endswith(".dll"):
            importLibraryFileName = importLibraryFileName[:-4]
        importLibraryFileName = importLibraryFileName + ".ilb"
        self.__writeLinkerRule(makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName, "/dll /IMPLIB:\"" + self.translateFileName(importLibraryFileName) + "\"")

    def writeLibrarianRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName):
        """Writes a makefile rule for invoking a librarian with the specified
        object files and libraries to create a static library"""
        makefile.write("\tif exist \"" + self.translateFileName(outputFileName) + "\" del \"" + self.translateFileName(outputFileName) + "\"\n")
        inputFiles = objectFileNames + staticLibraryFileNames
        makefile.write("\t$(LIBTOOL) /OUT:\"" + self.translateFileName(outputFileName) + "\" $(LIBOPTS) @<<\n")
        for inputFile in inputFiles:
            makefile.write("  " + self.translateFileName(inputFile) + "\n")
        makefile.write("<<\n")

    def __prepareClTargetProcessorOptions(self, project):
        """Given a project, prepares the cl.exe options for target processor specification"""
        if project.targetProcessor == "80386":
            return "/G3"
        elif project.targetProcessor == "80486":
            return "/G4"
        elif project.targetProcessor.lower() == "pentium":
            return "/G5"
        elif project.targetProcessor.lower() == "pentiumpro":
            return "/G6"
        elif project.targetProcessor.lower() == "pentium2":
            return "/G6"
        elif project.targetProcessor.lower() == "pentium3":
            return "/G6"
        elif project.targetProcessor.lower() == "pentium4":
            return "/G6"
        else:
            print "Unsupported target processor " + project.targetProcessor
            sys.exit(1)
 
    def __prepareClMlDefineOptions(self, project):
        """Given a project, prepares the cl.exe options for defines"""
        result = ""
        for define in project.defines:
            if len(result) != 0:
                result = result + " "
            result = result + "/D" + define
        return result

    def __prepareClMlIncludePathOptions(self, project, projectClosure):
        """Given a project, prepares the cl.exe options for include paths"""
        result = ""
        writtenIncludePaths = []
        #   Explicit include paths
        for someProjectFile in projectClosure.keys():
            someProject = projectClosure[someProjectFile]
            if (someProject.type <> TargetType.Executable) or (someProject.projectFileName == project.projectFileName):
                #   Dependencies which are executables do NOT contribute headers
                for includePath in someProject.includePaths:
                    if includePath not in writtenIncludePaths:
                        writtenIncludePaths.append(includePath) 
                        if len(result) != 0:
                            result = result + " "
                        result = result + "/I\"" + includePath + "\""
        #   Done
        return result

    def __prepareStandardCCompilerOptions(self, project, projectClosure):
        """Given a project, creates & returns the standard options string
        for a C compiler"""
        result = "/c /nologo /Zm1500 /D WIN32 /D _WINDOWS /D _MBCS /W4 /TC /GF /Gy"
        if project.mode == BuildMode.Debug:
            result = result + " /Od  /D _DEBUG /GS /RTC1 /Zi"
            if project.standardLibrary == StandardLibrary.Static:
                result = result + " /MTd /D_LIB"
            else:
                result = result + " /MDd /D_USRDLL /D_WINDLL"
        else:
            result = result + " /O2 /D NDEBUG /Oi /Oy /Og"
            if project.standardLibrary == StandardLibrary.Static:
                result = result + " /MT /D_LIB"
            else:
                result = result + " /MD /D_USRDLL /D_WINDLL"
        return result + " " + self.__prepareClTargetProcessorOptions(project) + " " + self.__prepareClMlDefineOptions(project) + " " + self.__prepareClMlIncludePathOptions(project, projectClosure)
    
    def __prepareStandardCppCompilerOptions(self, project, projectClosure):
        """Given a project, creates & returns the standard options string
        for a C++ compiler"""
        result = "/c /nologo /Zm1500 /D WIN32 /D _WINDOWS /D _MBCS /EHsc /GR /W4 /TP /GF /Gy /Zc:wchar_t /Zc:forScope"
        if project.mode == BuildMode.Debug:
            result = result + " /Od  /D _DEBUG /GS /RTC1 /Zi"
            if project.standardLibrary == StandardLibrary.Static:
                result = result + " /MTd /D_LIB"
            else:
                result = result + " /MDd /D_USRDLL /D_WINDLL"
        else:
            result = result + " /O2 /D NDEBUG /Oi /Oy /Og"
            if project.standardLibrary == StandardLibrary.Static:
                result = result + " /MT /D_LIB"
            else:
                result = result + " /MD /D_USRDLL /D_WINDLL"
        return result + " " + self.__prepareClTargetProcessorOptions(project) + " " + self.__prepareClMlDefineOptions(project) + " " + self.__prepareClMlIncludePathOptions(project, projectClosure)

    def __prepareSplTargetOptions(self, project):
        """Given a project, prepares the spl.exe options for target specification"""
        result = "-Target:C++:Endian=Little,Flavour=MSVC2003,Format=C++.Obj,OS=Win32,Processor=" + project.targetProcessor + ",Stdlib="
        if project.standardLibrary == StandardLibrary.Static:
            result = result + "Static"
        else:
            result = result + "Dynamic"
        return result

    def __prepareStandardAssemblerOptions(self, project, projectClosure):
        """Given a project, creates & returns the standard options string
        for an assembler"""
        return "/c " + self.__prepareClMlDefineOptions(project) + " " + self.__prepareClMlIncludePathOptions(project, projectClosure)
    
    def __prepareStandardLinkerOptions(self, project):
        """Given a project, creates & returns the standard options string
        for a linker"""
        if project.mode == BuildMode.Debug:
            return "/nologo /debug /incremental:no"
        else:
            return "/nologo /incremental:no"

    def __prepareStandardLibrarianOptions(self, project):
        """Given a project, creates & returns the standard options string
        for a librarian"""
        return "/nologo"

    def __writeLinkerRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName, extraOptions):
        """Writes a makefile rule for invoking a linker with the specified
        object files and libraries to create a dynamic library or executable"""
        makefile.write("\t$(LINKTOOL) /OUT:\"" + self.translateFileName(outputFileName) + "\" $(LINKOPTS) " + extraOptions + " @<<\n")
        for objectFileName in objectFileNames:
            makefile.write("  " + self.translateFileName(objectFileName) + "\n")
        for staticLibraryFileName in staticLibraryFileNames:
            makefile.write("  " + self.translateFileName(staticLibraryFileName) + "\n")
        for dynamicLibraryFileName in dynamicLibraryFileNames:
            importLibraryFileName = dynamicLibraryFileName
            if importLibraryFileName.lower().endswith(".dll"):
                importLibraryFileName = importLibraryFileName[:-4]
            importLibraryFileName = importLibraryFileName + ".ilb"
            makefile.write("  " + self.translateFileName(importLibraryFileName) + "\n")
        for systemLibraryFileName in systemLibraryFileNames:
            makefile.write("  " + systemLibraryFileName + "\n")
        makefile.write("<<\n")

#   Plug-in initialization: register the builder
Builder.register(Msvc2003Builder())

#   End of splbuild_msvc2003_builder.pl
