#!/usr/bin/python
#
#   splbuild_builder.pl
#
#   SPL component builder: An abstract builder
#
#   $Date: 2008/08/06 12:06:41 $
#   $Revision: 1.14 $
#   $Author: akapusti $
#
#   Copyright (C) 2007-2008 Cybernetic Intelligence GmbH
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##########
import sys, subprocess, time
from splbuild_project import *
from splbuild_project_reader import *

class Builder(object):
    """A Builder is an agent that knows how to perform builds on a
    for a particular CPU & OS using a particular toolchain"""
    
    def __init__(self, objectFileExtension):
        """The class constructor"""
        self.objectFileExtension = objectFileExtension
        self.projectsAlreadyBuilt = []

    __knownBuilders = []
    """A list of all currently registered builders"""
    
    def register(builder):
        """Adds the specified builder to the registered Builders list"""
        Builder.__knownBuilders.append(builder)
    register = staticmethod(register)
    
    def findBuilder(targetProcessor, targetOs, toolchain):
        """Finds and returns a Builder for the specified platform, or None
        if there is no such Builder"""
        for builder in Builder.__knownBuilders:
            if builder.canHandle(targetProcessor, targetOs, toolchain):
                return builder
        return None
    findBuilder = staticmethod(findBuilder)
    
    def canHandle(self, targetProcessor, targetOs, toolchain):
        """Returns True iff this builder can perform builds for the specified
        target, False otherwise"""
        print "Abstract method Builder.canHandle called"
        sys.exit(1)
    
    def setupDefaults(self, project):
        """Given a project, assigns the Builder-specific default values
        to those project properties that were not defined explicitly"""
        print "Abstract method Builder.setupDefaults called"
        sys.exit(1)

    def validate(self, project):
        """Given a project, checks it for validity"""
        print "Abstract method Builder.validate called"
        sys.exit(1)
        
    def getFileType(self, fileName):
        """A helper service for deriving a file type from file name.
        Returns the file type or None if the file type cannot be determined"""
        print "Abstract method Builder.getFileType called"
        sys.exit(1)

    def writeMakefileToolsAndOptions(self, makefile, project, projectClosure):
        """Writes make definitions for tool and option strings"""
        print "Abstract method Builder.writeMakefileToolsAndOptions called"
        sys.exit(1)

    def prepareMakeCommandLine(self, project):
        """Prepares and returns a "make" command line for the project
        in the form of argument list (one element per argument)"""
        print "Abstract method Builder.prepareMakeCommandLine called"
        sys.exit(1)

    def translateFileName(self, fileName):
        """Given a name of a file, creates and returns the form in which
        this filename must be present in the makefile"""
        return fileName

    def writeCCompilerRule(self, makefile, sourceFileName, outputFileName):
        """Writes a makefile rule for invoking a C compiler with specified
        source and output filenames"""
        print "Abstract method Builder.writeCCompilerRule called"
        sys.exit(1)

    def writeCppCompilerRule(self, makefile, sourceFileName, outputFileName):
        """Writes a makefile rule for invoking a C++ compiler with specified
        source and output filenames"""
        print "Abstract method Builder.writeCppCompilerRule called"
        sys.exit(1)

    def writeSplCompilerRule(self, makefile, sourceFileName, outputFileName, dependencyFileName):
        """Writes a makefile rule for invoking an SPL compiler with specified
        source and output filenames"""
        print "Abstract method Builder.writeSplCompilerRule called"
        sys.exit(1)

    def writeAssemblerRule(self, makefile, sourceFileName, outputFileName):
        """Writes a makefile rule for invoking an assembler with specified
        source and output filenames"""
        print "Abstract method Builder.writeAssemblerRule called"
        sys.exit(1)

    def writeExeLinkerRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName):
        """Writes a makefile rule for invoking a linker with the specified
        object files and libraries to create an executable"""
        print "Abstract method Builder.writeExeLinkerRule called"
        sys.exit(1)

    def writeDllLinkerRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName):
        """Writes a makefile rule for invoking a linker with the specified
        object files and libraries to create a dynamic library"""
        print "Abstract method Builder.writeExeLinkerRule called"
        sys.exit(1)

    def writeLibrarianRule(self, makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, outputFileName):
        """Writes a makefile rule for invoking a librarian with the specified
        object files and libraries to create a static library"""
        print "Abstract method Builder.writeExeLinkerRule called"
        sys.exit(1)

    def build(self, project, projectClosure):
        """Builds the project"""
        assert project.builder is self
        #   Do not build the same project more than once
        if project.projectFileName in self.projectsAlreadyBuilt:
            return
        self.projectsAlreadyBuilt.append(project.projectFileName)
        print "Building " + project.projectFileName + "..."
        #   Go!
        self.validate(project)
        try:
            #   Create temp directory if not yet there
            if not os.path.isdir(project.tempDirectory):
                try:
                    os.makedirs(project.tempDirectory)
                except Exception:
                    print "Can't create temporary directory '" + project.tempDirectory + "'"
                    sys.exit(1)
            #   Create output directory if not yet there
            outputDirectory = os.path.dirname(project.output)
            if not os.path.isdir(outputDirectory):
                try:
                    os.makedirs(outputDirectory)
                except Exception:
                    print "Can't create output directory '" + outputDirectory + "'"
                    sys.exit(1)
            #   Write the makefile...
            self.__writeMakefile(project, projectClosure)
            #   ...then make it
            makeCommandLine = self.prepareMakeCommandLine(project)
            self.executeProgram(makeCommandLine)
        finally:
            #   ...do the cleanup
            if os.path.exists(project.makefileName):
                os.remove(project.makefileName)
            if os.path.exists(project.splViaFileName):
                os.remove(project.splViaFileName)

    def executeProgram(self, commandLineArguments):
        """Given a list of command line arguments, executes the
        corresponding program (the 1st element of the list is
        treated as the executable file name)"""
        try:
            #   Start the process
            childProcess = subprocess.Popen(commandLineArguments, universal_newlines = True)
            #   Now wait for the process to terminate
            exitCode = None
            while True:
                exitCode = childProcess.poll()
                if exitCode is not None:
                    break
                time.sleep(1)
            if exitCode != 0:
                sys.exit(exitCode)
        except OSError:
            commandLine = " ".join(commandLineArguments)
            print "Can't execute command '" + commandLine + "'"
            sys.exit(exitCode)

    def breakArgumentString(self, argumentString):
        """Given a string containing command-line arguments, breaks it
        into a list containing one argument per element. Takes quoting
        into account"""
        result = []
        scan = 0
        while scan < len(argumentString):
            #   Skip spaces
            while (scan < len(argumentString)) and (argumentString[scan].isspace()):
                scan = scan + 1
            if scan == len(argumentString):
                break
            #   Argument starts here
            qc = 0
            argument = ""
            while scan < len(argumentString):
                #   Quotes require special treatment
                if (argumentString[scan] == '"') and (qc != 1):
                    qc = 2 - qc
                    scan = scan + 1
                    if (qc == 0) and (scan < len(argumentString)) and (argumentString[scan] == '"'):
                        argument = argument + '"'
                        scan = scan + 1
                        qc = 2 - qc
                elif (argumentString[scan] == "'") and (qc != 2):
                    qc = 1 - qc
                    scan = scan + 1
                    if (qc == 0) and (scan < len(argumentString)) and (argumentString[scan] == "'"):
                        argument = argument + "'"
                        scan = scan + 1
                        qc = 1 - qc
                else:
                    argument = argument + argumentString[scan]
                    scan = scan + 1
            result.append(argument)
        return result

    def inferExtendedFileName(sourceFileName, project):
        """Creates returns a name of a temporary file for the specified source file.
        The temporary file name is formed by stripping common path prefix
        from the source file and replacing path separators with underscores"""
        #   Do the work
        projectDirectory = os.path.normpath(os.path.dirname(project.projectFileName))
        fullSourceFileName = os.path.normpath(os.path.join(projectDirectory, sourceFileName))
        sourceFileNameSuffix = fullSourceFileName
        if fullSourceFileName.startswith(projectDirectory):
            sourceFileNameSuffix = sourceFileNameSuffix[len(projectDirectory):]
        result = sourceFileNameSuffix
        result = result.replace("/", "_")
        result = result.replace("\\", "_")
        result = result.replace(":", "_")
        while result.startswith("_") and not result.startswith("_."):
            result = result[1:]
        return os.path.join(project.tempDirectory, result)
    inferExtendedFileName = staticmethod(inferExtendedFileName)
    
    def prepareObjectFileName(self, sourceFileName, project):
        """Given a name of a source file, generates and returns the name
        for the object file into which that source file is to be compiled"""
        objectFile = Builder.inferExtendedFileName(sourceFileName, project)
        head, tail = os.path.split(objectFile)
        root, ext = os.path.splitext(tail)
        return os.path.join(head, root) + self.objectFileExtension
        
    def prepareDependencyFileName(self, sourceFileName, project):
        """Given a name of a source file, generates and returns the name
        for the dependency file where source file's dependencies will live"""
        objectFile = Builder.inferExtendedFileName(sourceFileName, project)
        head, tail = os.path.split(objectFile)
        root, ext = os.path.splitext(tail)
        return os.path.join(head, root) + ".dep"

    def __writeSourceFileDependency(self, makefile, fullSourceFileName, fullObjectFileName, project):
        """Writes a dependency for the specified object file on a specified
        source file"""
        makefile.write(self.translateFileName(fullObjectFileName) + " : " + self.translateFileName(fullSourceFileName))
        #   Use dependencies file if one exists
        dependencyFileOk = False
        dependencyFileName = self.prepareDependencyFileName(fullSourceFileName, project)
        if os.path.isfile(dependencyFileName):
            try:
                dependencyFile = open(dependencyFileName, "r")
                dependencyFileOk = True
                dependencies = []
                while True:
                    line = dependencyFile.readline()
                    #   EOF ?
                    if len(line) == 0:
                        break
                    #   Chop off tailing line separator
                    line = line.strip(" \t\n\r")
                    #   Skip empty lines
                    if len(line) <> 0:
                        if os.path.isfile(line):
                            dependencies.append(line)
                        else:
                            dependencyFileOk = False
                dependencyFile.close()
                if dependencyFileOk:
                    for dependency in dependencies:
                        makefile.write("\\\n  " + dependency)
            except Exception:
                pass
        if not dependencyFileOk:
            makefile.write(" $(ALL_HEADERS)")
        #   Changes to a project file cause full rebuild
        for subprojectFileName in project.subprojectFileNames:
            makefile.write("\\\n  " + self.translateFileName(subprojectFileName))
        #   Done
        makefile.write("\n")

    def __writeMakefile(self, project, projectClosure):
        """A helper service to write a makefile for a project"""
        try:
            #   Open the makefile
            makefile = open(project.makefileName, "w")
            #   Write tool & option variables
            self.writeMakefileToolsAndOptions(makefile, project, projectClosure)
            #   Write headers
            makefile.write("ALL_HEADERS=")
            writtenHeaders = []
            for someProjectFileName in projectClosure.keys():
                someProject = projectClosure[someProjectFileName]
                if (someProject.type <> TargetType.Executable) or (someProject.projectFileName == project.projectFileName):
                    #   Dependencies which are executables do NOT contribute headers
                    for someHeaderFileName in someProject.headers:
                        if someHeaderFileName not in writtenHeaders: 
                            makefile.write(" \\\n    " + self.translateFileName(someHeaderFileName))
                            writtenHeaders.append(someHeaderFileName)
            makefile.write("\n\n")
            #   Prepare a list of all object files, static libraries and
            #   dynamic libraries that must be combined to form a target
            objectFileNames = []
            staticLibraryFileNames = []
            dynamicLibraryFileNames = []
            systemLibraryFileNames = []
            for source in project.sources:
                fullSourceFileName = os.path.join(os.path.dirname(project.projectFileName), source)
                fullObjectFileName = self.prepareObjectFileName(fullSourceFileName, project)
                fileType = self.getFileType(fullSourceFileName)
                if fileType == FileType.CSource:
                    objectFileNames.append(fullObjectFileName)
                elif fileType == FileType.CppSource:
                    objectFileNames.append(fullObjectFileName)
                elif fileType == FileType.SplSource:
                    objectFileNames.append(fullObjectFileName)
                elif fileType == FileType.AssemblerSource:
                    objectFileNames.append(fullObjectFileName)
                elif fileType == FileType.ObjectModule:
                    objectFileNames.append(fullSourceFileName)
                elif fileType == FileType.StaticLibrary:
                    staticLibraryFileNames.append(fullSourceFileName)
                elif fileType == FileType.DynamicLibrary:
                    print "Not yet implemented"
                    sys.exit(1)
                else:
                    print "Don't know how to compile '" + fullSourceFileName + "'"
                    sys.exit(1)
            #   Dependencies as well...
            for someProjectFileName in projectClosure.keys():
                someProject = projectClosure[someProjectFileName]
                for library in someProject.libraries:
                    if library not in systemLibraryFileNames:
                        systemLibraryFileNames.append(library)
                if someProjectFileName == project.projectFileName:
                    continue
                if project.type == TargetType.StaticLibrary:
                    continue
                if someProject.type == TargetType.StaticLibrary:
                    library = self.translateFileName(someProject.output)
                    if library not in staticLibraryFileNames:
                        staticLibraryFileNames.append(library)
                if someProject.type == TargetType.DynamicLibrary:
                    library = self.translateFileName(someProject.output)
                    if library not in dynamicLibraryFileNames: 
                        dynamicLibraryFileNames.append(library)
            #   Write a rule to build the main project's target
            makefile.write(self.translateFileName(project.output) + " : ")
            for objectFileName in objectFileNames:
                makefile.write("\\\n  " + self.translateFileName(objectFileName))
            for staticLibraryFileName in staticLibraryFileNames:
                makefile.write("\\\n  " + self.translateFileName(staticLibraryFileName))
            for dynamicLibraryFileName in dynamicLibraryFileNames:
                makefile.write("\\\n  " + self.translateFileName(dynamicLibraryFileName))
            #   Changes to a project file cause full rebuild
            for subprojectFileName in project.subprojectFileNames:
                makefile.write("\\\n  " + self.translateFileName(subprojectFileName))
            makefile.write("\n")
            if project.type == TargetType.Executable:
                self.writeExeLinkerRule(makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, project.output)
            elif project.type == TargetType.StaticLibrary:
                self.writeLibrarianRule(makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, project.output)
            elif project.type == TargetType.DynamicLibrary:
                self.writeDllLinkerRule(makefile, objectFileNames, staticLibraryFileNames, dynamicLibraryFileNames, systemLibraryFileNames, project.output)
            else:
                print "Invalid project type '" + project.type + "'"
                sys.exit(1)
            makefile.write("\n")
            #   Write rules for compiling source files
            for source in project.sources:
                fullSourceFileName = os.path.join(os.path.dirname(project.projectFileName), source)
                fullObjectFileName = self.prepareObjectFileName(fullSourceFileName, project)
                #   Write the build rule
                fileType = self.getFileType(fullSourceFileName)
                if fileType == FileType.CSource:
                    self.__writeSourceFileDependency(makefile, fullSourceFileName, fullObjectFileName, project)
                    self.writeCCompilerRule(makefile, fullSourceFileName, fullObjectFileName)
                    makefile.write("\n")
                elif fileType == FileType.CppSource:
                    self.__writeSourceFileDependency(makefile, fullSourceFileName, fullObjectFileName, project)
                    self.writeCppCompilerRule(makefile, fullSourceFileName, fullObjectFileName)
                    makefile.write("\n")
                elif fileType == FileType.SplSource:
                    self.__writeSourceFileDependency(makefile, fullSourceFileName, fullObjectFileName, project)
                    dependencyFileName = self.prepareDependencyFileName(fullSourceFileName, project)
                    self.writeSplCompilerRule(makefile, fullSourceFileName, fullObjectFileName, dependencyFileName)
                    makefile.write("\n")
                elif fileType == FileType.AssemblerSource:
                    self.__writeSourceFileDependency(makefile, fullSourceFileName, fullObjectFileName, project)
                    self.writeAssemblerRule(makefile, fullSourceFileName, fullObjectFileName)
                    makefile.write("\n")
                elif fileType == FileType.ObjectModule:
                    pass    #   Will be linked in
                elif fileType == FileType.StaticLibrary:
                    pass    #   Will be linked in
                elif fileType == FileType.DynamicLibrary:
                    pass    #   Will be linked in
                else:
                    print "Don't know how to compile '" + fullSourceFileName + "'"
                    sys.exit(1)
            #   Done
            makefile.close();
        except IOError:
            print "Error writing the makefile '" + project.makefileName + "'"
            sys.exit(1)

    #   Most portions of the SPL compiler command line are platform-independent
    def _prepareSplDefineOptions(self, project):
        """Given a project, prepares the spl.exe options for defines"""
        result = ""
        for define in project.defines:
            if len(result) != 0:
                result = result + " "
            result = result + "-Define:" + define
        return result

    def _prepareSplIncludePathOptions(self, project, projectClosure):
        """Given a project, prepares the spl.exe options for include paths"""
        result = ""
        writtenIncludePaths = []
        #   Explicit include paths
        for someProjectFile in projectClosure.keys():
            someProject = projectClosure[someProjectFile]
            if (someProject.type <> TargetType.Executable) or (someProject.projectFileName == project.projectFileName):
                #   Dependencies which are executables do NOT contribute headers
                for includePath in someProject.includePaths:
                    if includePath not in writtenIncludePaths:
                        writtenIncludePaths.append(includePath) 
                        if len(result) != 0:
                            result = result + " "
                        result = result + "-IncludePath:" + self.translateFileName(includePath) + ""
        #   Done
        return result

    def _prepareSplIncludeContextOptions(self, project, projectClosure):
        """Given a project, prepares the spl.exe options for include contexts"""
        result = ""
        allIncludeContexts = {}    #   Keys are include context names; values are directory lists for each include context
        for someProjectFile in projectClosure.keys():
            someProject = projectClosure[someProjectFile]
            if (someProject.type <> TargetType.Executable) or (someProject.projectFileName == project.projectFileName):
                #   Dependencies which are executables do NOT contribute headers
                for includeContext in someProject.includeContexts:
                    #   Get the current directiry list
                    if includeContext in allIncludeContexts.keys():
                        includeDirectories = allIncludeContexts[includeContext]
                    else:
                        includeDirectories = []
                    #   All directories containing this project's headers
                    #   contribute to the includeContext
                    for header in someProject.headers:
                        headerDirectory = os.path.dirname(header)
                        if headerDirectory not in includeDirectories:
                            includeDirectories.append(headerDirectory)
                    #   Update the map
                    allIncludeContexts[includeContext] = includeDirectories
        #   Now form the command line portion  
        for includeContext in allIncludeContexts.keys():
            result = result + " -IncludeContext:" + includeContext + "="
            pathList = ""
            for path in allIncludeContexts[includeContext]:
                if len(pathList) <> 0:
                    pathList = pathList + os.pathsep
                pathList = pathList + self.translateFileName(path)
            result = result + pathList
        #   Done
        return result

    def _prepareStandardSplCompilerOptions(self, project, projectClosure):
        """Given a project, creates & returns the standard options string
        for an SPL compiler"""
        result = "-Quiet"
        if project.mode == BuildMode.Debug:
            result = result + " -Optimize:Off -GeneratePreconditions:On -GeneratePostconditions:On -GenerateAssertions:On -CheckPointers:On -CheckSubscripts:On -CheckEnumerations:On -Debug:On"
        else:
            result = result + " -Optimize -GeneratePreconditions:Off -GeneratePostconditions:Off -GenerateAssertions:Off -CheckPointers:Off -CheckSubscripts:Off -CheckEnumerations:Off -Debug:Off"
        return result + " " + self._prepareSplDefineOptions(project) + " " + self._prepareSplIncludePathOptions(project, projectClosure) + " " + self._prepareSplIncludeContextOptions(project, projectClosure)
#   End of splbuild_builder.pl
