#!/usr/bin/python
#
#   splbuild_project.pl
#
#   SPL component builder: Project IR
#
#   $Date: 2008/03/02 22:41:06 $
#   $Revision: 1.12 $
#   $Author: akapusti $
#
#   Copyright (C) 2007-2008 Cybernetic Intelligence GmbH
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##########
import os

class FileType(object):
    """An utility class serving as namespace for constants representing
    different types of files"""
    CSource = "csource"
    CHeader = "cheader"
    CppSource = "cppsource"
    CppHeader = "cppheader"
    SplSource = "splsource"
    SplHeader = "splheader"
    AssemblerSource = "assemblersource"
    AssemblerHeader = "assemblerheader"
    ObjectModule = "objectmodule"
    StaticLibrary = "staticlibrary"
    DynamicLibrary = "dynamiclibrary"

class TargetType(object):
    """An utility class serving as namespace for constants representing
    supported project target types"""
    Executable = "exe"
    StaticLibrary = "lib"
    DynamicLibrary = "dll"

class BuildMode(object):
    """An utility class serving as namespace for constants representing
    supported project build modes"""
    Debug = "debug"
    Release = "release"

class StandardLibrary(object):
    """An utility class serving as namespace for constants representing
    supported standard library linkage modes"""
    Static = "static"
    Dynamic = "dynamic"

class Project(object):
    """An internal representation of a single project"""
    
    def __init__(self, projectFileName):
        """The class constructor"""
        assert os.path.isabs(projectFileName)
        self.projectFileName = projectFileName      #   The full path of the project file
        self.subprojectFileNames = []               #   The list of all project definition files (i.e. main project file + all files included by "use" directive)
        self.sources = []                           #   All source filenames are absolute
        self.headers = []                           #   All header filenames are absolute
        self.libraries = []                         #   All system library filenames are stored as-is
        self.type = None                            #   Project type (TargetType.???)
        self.output = None                          #   Absolute main output file name
        self.tempDirectory = None                   #   Absolute temp directory path
        self.targetProcessor = None                 #   Target processor (TargetProcessor.???)
        self.targetOs = None                        #   Target OS (TargetOs.???)
        self.toolchain = None                       #   Toolchain to use (Toolchain.???)
        self.defines = []                           #   A list of "<name>=<value>" strings
        self.includePaths = []                      #   A list of absolute include directory paths
        self.includeContexts = []                   #   A list of include context names defined by this project
        self.mode = None                            #   The build mode (BuildMode.???)
        self.standardLibrary = None                 #   Standard library linkage (StandardLibrary.???)
        self.dependencies = []                      #   A list of absolute filenames of all dependency project files
        self.cCompiler = None                       #   The C compiler executable (as it appears in project file)
        self.cppCompiler = None                     #   The C++ compiler executable (as it appears in project file)
        self.splCompiler = None                     #   The SPL compiler executable (as it appears in project file)
        self.assembler = None                       #   The assembler executable (as it appears in project file)
        self.linker = None                          #   The linker executable (as it appears in project file)
        self.librarian = None                       #   The librarian executable (as it appears in project file)
        self.make = None                            #   The make utility executable (as it appears in project file)
        self.cCompilerOptions = None                #   The C compiler extra options (as specified in project file)
        self.cppCompilerOptions = None              #   The C++ compiler extra options (as specified in project file)
        self.splCompilerOptions = None              #   The SPL compiler extra options (as specified in project file)
        self.assemblerOptions = None                #   The assembler extra options (as specified in project file)
        self.linkerOptions = None                   #   The linker extra options (as specified in project file)
        self.librarianOptions = None                #   The librarian extra options (as specified in project file)
        self.makeOptions = None                     #   The make utility extra options (as specified in project file)
        self.builder = None                         #   The builder to use for this project

    def getFileType(fileName):
        """A helper service for deriving a file type from file name.
        Returns the file type or None if the file type cannot be determined"""
        if filename.lower().endswith(".c"):
            return FileType.CSource
        if filename.lower().endswith(".h"):
            return FileType.CHeader
        if filename.lower().endswith(".cpp") or filename.lower().endswith(".cxx"):
            return FileType.CppSource
        if filename.lower().endswith(".hpp") or filename.lower().endswith(".hxx"):
            return FileType.CppHeader
        if filename.lower().endswith(".spl"):
            return FileType.SplSource
        if filename.lower().endswith(".sph"):
            return FileType.SplHeader
        if filename.lower().endswith(".asm") or filename.lower().endswith(".s"):
            return FileType.AssemblerSource
        if filename.lower().endswith(".inc"):
            return FileType.AssemblerHeader
        if filename.lower().endswith(".o") or filename.lower().endswith(".cbj"):
            return FileType.ObjectModule
        if filename.lower().endswith(".a") or filename.lower().endswith(".lib"):
            return FileType.StaticLibrary
        if filename.lower().endswith(".so") or filename.lower().endswith(".dll"):
            return FileType.DynamicLibrary
        return None
    getFileType = staticmethod(getFileType)

#   End of splbuild_project.pl
