#!/usr/bin/python
#
#   splbuild_project_reader.pl
#
#   SPL component builder: Project file reader
#
#   $Date: 2008/03/02 22:41:06 $
#   $Revision: 1.12 $
#   $Author: akapusti $
#
#   Copyright (C) 2007-2008 Cybernetic Intelligence GmbH
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##########
import os, sys
from splbuild_project import *
from splbuild_builder import *

class ProjectReader(object):
    """This utility class provides project file reading services"""
    
    def __init__(self):
        """Utility class cannot be instantiated"""
        print "Utility ProjectReader class cannot be instantiated"
        sys.exit(1)
        
    def readProjectFile(fileName):
        """Reads a project file with the specified name and returns
        the corresponding Project instance"""
        #   Prepare absolute project file name
        projectFileName = os.path.abspath(fileName)
        projectDirectory = os.path.dirname(projectFileName)
        #   Create a new, empty, project
        project = Project(projectFileName)
        #   Read!
        ProjectReader.__readProjectFileContents(projectFileName, project)
        #   Validate
        if project.type is None:
            print projectFileName + ": Project type not specified"
            sys.exit(1)
        if project.output is None:
            print projectFileName + ": Project output not specified"
            sys.exit(1)
        if project.targetProcessor is None:
            print projectFileName + ": Project target processor not specified"
            sys.exit(1)
        if project.targetOs is None:
            print projectFileName + ": Project target OS not specified"
            sys.exit(1)
        if project.toolchain is None:
            print projectFileName + ": Project toolchain not specified"
            sys.exit(1)
        #   Find the builder for this project
        project.builder = Builder.findBuilder(project.targetProcessor, project.targetOs, project.toolchain)
        if project.builder is None:
            print projectFileName + ": Don't know how to build"
            sys.exit(1)
        #   Let the Builder a chance to fill the missing properties of a project
        #   with Builder-specific defaults
        project.builder.setupDefaults(project)
        #   Make sure the Builder did its job and fill in the rest of the defaults
        if project.tempDirectory is None:
            project.tempDirectory = os.path.join(projectDirectory, "obj")
        if project.mode is None:
            project.mode = BuildMode.Debug
        if project.standardLibrary is None:
            if project.type == TargetType.StaticLibrary:
                project.standardLibrary = StandardLibrary.Static
            else:  
                project.standardLibrary = StandardLibrary.Dynamic
        if project.cCompiler is None:
            project.cCompiler = "Nonexistent_C_Compiler"
        if project.cppCompiler is None:
            project.cCompiler = "Nonexistent_Cpp_Compiler"
        if project.splCompiler is None:
            project.cCompiler = "Nonexistent_SPL_Compiler"
        if project.assembler is None:
            project.cCompiler = "Nonexistent_Assembler"
        if project.linker is None:
            project.cCompiler = "Nonexistent_Linker"
        if project.librarian is None:
            project.cCompiler = "Nonexistent_Librarian"
        if project.make is None:
            project.cCompiler = "Nonexistent_Make"
        if project.cCompilerOptions is None:
            project.cCompilerOptions = ""
        if project.cppCompilerOptions is None:
            project.cppCompilerOptions = ""
        if project.splCompilerOptions is None:
            project.splCompilerOptions = ""
        if project.assemblerOptions is None:
            project.assemblerOptions = ""
        if project.linkerOptions is None:
            project.linkerOptions = ""
        if project.librarianOptions is None:
            project.librarianOptions = ""
        if project.makeOptions is None:
            project.makeOptions = ""
        #   Done
        return project
    readProjectFile = staticmethod(readProjectFile)

    def readProjectClosure(projectFileName):
        """Reads a project file and all its direct or indirect dependencies; 
        creates and returns a dictionary where project file names are keys 
        and project instances are values"""
        #   Read the main project
        mainProject = ProjectReader.readProjectFile(projectFileName)
        projectClosure = { mainProject.projectFileName : mainProject }
        #   Read dependencies until no more
        while True:
            moreDependenciesRead = False
            for projectFileName in projectClosure.keys():
                project =  projectClosure[projectFileName]
                for dependencyFileName in project.dependencies:
                    dependency = ProjectReader.readProjectFile(dependencyFileName)
                    if dependency.projectFileName not in projectClosure.keys():
                        projectClosure[dependency.projectFileName] = dependency
                        moreDependenciesRead = True
            if not moreDependenciesRead:
                break
        #   Done
        return projectClosure
    readProjectClosure = staticmethod(readProjectClosure)
    
    def __readProjectFileContents(projectFileName, project):
        """A helper method for reading a project file"""
        if projectFileName not in project.subprojectFileNames:
            project.subprojectFileNames.append(projectFileName)
        projectDirectory = os.path.dirname(projectFileName)
        try:
            #   Open the project file
            projectFile = open(projectFileName, "r")
            #   Read!
            #   Read project properties line-by-line
            lineNumber = 0
            while True:
                #   Read next line
                line = projectFile.readline()
                lineNumber = lineNumber + 1
                #   EOF ?
                if len(line) == 0:
                    break
                #   Chop off tailing line separator
                line = line.strip(" \t\n\r")
                #   Remove comment, if any
                hashIndex = line.find("#")
                if hashIndex != -1:
                    line = line[0:hashIndex]
                line = line.strip(" \t")
                if len(line) == 0:
                    continue
                #   Perform environment variable replacement
                line = ProjectReader.__expandMacros(line, projectFileName + "(" + str(lineNumber) + ")")
                #   Split line into setting name and setting value
                colonIndex = line.find(":")
                if colonIndex == -1:
                    print projectFileName + "(" + str(lineNumber) + "): Syntax error"
                    sys.exit(1)
                settingName = line[0:colonIndex].strip(" \t").lower()
                settingValue = line[colonIndex + 1:].strip(" \t")
                #   Process the setting
                if settingName == "source":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Source file name is missing"
                        sys.exit(1)
                    sourceFileName = os.path.join(projectDirectory, settingValue)
                    if sourceFileName not in project.sources:
                        project.sources.append(sourceFileName)
                elif settingName == "header":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Header file name is missing"
                        sys.exit(1)
                    headerFileName = os.path.join(projectDirectory, settingValue)
                    if headerFileName not in project.headers:
                        project.headers.append(headerFileName)
                elif settingName == "library":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Library file name is missing"
                        sys.exit(1)
                    if settingValue not in project.libraries:
                        project.libraries.append(settingValue)
                elif settingName == "type":
                    if project.type is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Project type redefined"
                        sys.exit(1)
                    elif settingValue.lower() == TargetType.Executable.lower():
                        project.type = TargetType.Executable
                    elif settingValue.lower() == TargetType.StaticLibrary.lower():
                        project.type = TargetType.StaticLibrary
                    elif settingValue.lower() == TargetType.DynamicLibrary.lower():
                        project.type = TargetType.DynamicLibrary
                    else:
                        print projectFileName + "(" + str(lineNumber) + "): Invalid project type '" + settingValue + "'"
                        sys.exit(1)
                elif settingName == "output":
                    if project.output is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Project output file name redefined"
                        sys.exit(1)
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Output file name is missing"
                        sys.exit(1)
                    project.output = os.path.join(projectDirectory, settingValue)
                elif settingName == "temp-directory":
                    if project.tempDirectory is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Project temporary directory redefined"
                        sys.exit(1)
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Temporary directory name is missing"
                        sys.exit(1)
                    project.tempDirectory = os.path.join(projectDirectory, settingValue)
                elif settingName == "target-processor":
                    if project.targetProcessor is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Target processor redefined"
                        sys.exit(1)
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Target processor name is missing"
                        sys.exit(1)
                    project.targetProcessor = settingValue
                elif settingName == "target-os":
                    if project.targetOs is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Target OS redefined"
                        sys.exit(1)
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Target OS name is missing"
                        sys.exit(1)
                    project.targetOs = settingValue
                elif settingName == "toolchain":
                    if project.toolchain is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Toolchain redefined"
                        sys.exit(1)
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Toolchain name is missing"
                        sys.exit(1)
                    project.toolchain = settingValue
                elif settingName == "define":
                    eqIndex = settingValue.find("=")
                    if eqIndex == -1:
                        if not ProjectReader.__isValidIdentifier(settingValue):
                            print projectFileName + "(" + str(lineNumber) + "): Invalid defined variable name '" + settingValue + "'"
                            sys.exit(1)
                    else:
                        variableName = settingValue[0:eqIndex]
                        if not ProjectReader.__isValidIdentifier(variableName):
                            print projectFileName + "(" + str(lineNumber) + "): Invalid defined variable name '" + variableName + "'"
                            sys.exit(1)
                    if settingValue not in project.defines:
                        project.defines.append(settingValue)
                elif settingName == "include-path":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Include directory name is missing"
                        sys.exit(1)
                    includeDirectoryName = os.path.join(projectDirectory, settingValue)
                    if includeDirectoryName not in project.includePaths:
                        project.includePaths.append(includeDirectoryName)
                elif settingName == "include-context":
                    if not ProjectReader.__isValidIdentifier(settingValue):
                        print projectFileName + "(" + str(lineNumber) + "): Invalid include context name '" + settingValue + "'"
                        sys.exit(1)
                    if settingValue not in project.includeContexts:
                        project.includeContexts.append(settingValue)
                elif settingName == "mode":
                    if project.mode is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Project build mode redefined"
                        sys.exit(1)
                    elif settingValue.lower() == BuildMode.Debug.lower():
                        project.mode = BuildMode.Debug
                    elif settingValue.lower() == BuildMode.Release.lower():
                        project.mode = BuildMode.Release
                    else:
                        print projectFileName + "(" + str(lineNumber) + "): Invalid project build mode '" + settingValue + "'"
                        sys.exit(1)
                elif settingName == "standard-library":
                    if project.standardLibrary is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Project standard library linkage redefined"
                        sys.exit(1)
                    elif settingValue.lower() == StandardLibrary.Static.lower():
                        project.standardLibrary = StandardLibrary.Static
                    elif settingValue.lower() == StandardLibrary.Dynamic.lower():
                        project.standardLibrary = StandardLibrary.Dynamic
                    else:
                        print projectFileName + "(" + str(lineNumber) + "): Invalid standard library linkage '" + settingValue + "'"
                        sys.exit(1)
                elif settingName == "dependency":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Dependency file name is missing"
                        sys.exit(1)
                    dependencyFileName = os.path.join(projectDirectory, settingValue)
                    if dependencyFileName not in project.dependencies:
                        project.dependencies.append(dependencyFileName)
                elif settingName == "use":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Include file name is missing"
                        sys.exit(1)
                    includeFileName = os.path.join(projectDirectory, settingValue)
                    ProjectReader.__readProjectFileContents(includeFileName, project)
                elif settingName == "c-compiler":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): C compiler file name is missing"
                        sys.exit(1)
                    if project.cCompiler is not None:
                        print projectFileName + "(" + str(lineNumber) + "): C compiler location redefined"
                        sys.exit(1)
                    project.cCompiler = settingValue
                elif settingName == "cpp-compiler":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): C++ compiler file name is missing"
                        sys.exit(1)
                    if project.cppCompiler is not None:
                        print projectFileName + "(" + str(lineNumber) + "): C++ compiler location redefined"
                        sys.exit(1)
                    project.cppCompiler = settingValue
                elif settingName == "spl-compiler":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): SPL compiler file name is missing"
                        sys.exit(1)
                    if project.splCompiler is not None:
                        print projectFileName + "(" + str(lineNumber) + "): SPL compiler location redefined"
                        sys.exit(1)
                    project.splCompiler = settingValue
                elif settingName == "assembler":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Assembler file name is missing"
                        sys.exit(1)
                    if project.assembler is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Assembler location redefined"
                        sys.exit(1)
                    project.assembler = settingValue
                elif settingName == "linker":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Linker file name is missing"
                        sys.exit(1)
                    if project.linker is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Linker location redefined"
                        sys.exit(1)
                    project.linker = settingValue
                elif settingName == "librarian":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Librarian file name is missing"
                        sys.exit(1)
                    if project.librarian is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Librarian location redefined"
                        sys.exit(1)
                    project.librarian = settingValue
                elif settingName == "make":
                    if len(settingValue) == 0:
                        print projectFileName + "(" + str(lineNumber) + "): Make file name is missing"
                        sys.exit(1)
                    if project.make is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Make location redefined"
                        sys.exit(1)
                    project.make = settingValue
                elif settingName == "c-compiler-options":
                    if project.cCompilerOptions is not None:
                        print projectFileName + "(" + str(lineNumber) + "): C compiler custom options redefined"
                        sys.exit(1)
                    project.cCompilerOptions = settingValue
                elif settingName == "cpp-compiler-options":
                    if project.cppCompilerOptions is not None:
                        print projectFileName + "(" + str(lineNumber) + "): C++ compiler custom options redefined"
                        sys.exit(1)
                    project.cppCompilerOptions = settingValue
                elif settingName == "spl-compiler-options":
                    if project.splCompilerOptions is not None:
                        print projectFileName + "(" + str(lineNumber) + "): SPL compiler custom options redefined"
                        sys.exit(1)
                    project.splCompilerOptions = settingValue
                elif settingName == "assembler-options":
                    if project.assemblerOptions is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Assembler custom options redefined"
                        sys.exit(1)
                    project.assemblerOptions = settingValue
                elif settingName == "linker-options":
                    if project.linkerOptions is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Linker custom options redefined"
                        sys.exit(1)
                    project.linkerOptions = settingValue
                elif settingName == "librarian-options":
                    if project.librarianOptions is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Librarian custom options redefined"
                        sys.exit(1)
                    project.librarianOptions = settingValue
                elif settingName == "make-options":
                    if project.makeOptions is not None:
                        print projectFileName + "(" + str(lineNumber) + "): Make custom options redefined"
                        sys.exit(1)
                    project.makeOptions = settingValue
                else:
                    print projectFileName + "(" + str(lineNumber) + "): Invalid project setting '" + settingName + "'"
                    sys.exit(1)
            #   Done
            projectFile.close()
        except IOError:
            print "Can't read project file '" + projectFileName + "'"
            sys.exit(1)
    __readProjectFileContents = staticmethod(__readProjectFileContents)

    def __isValidIdentifier(s):
        """A helper method to check whether a string is a valid C/C++/SPL/Asm
        identidier"""
        if len(s) == 0:
            return False
        if not (s[0].islower() or s[0].isupper() or s[0] == "_"):
            return False
        for c in s[1:]:
            if not (c.islower() or c.isupper() or c.isdigit() or c == "_"):
                return False
        return True
    __isValidIdentifier = staticmethod(__isValidIdentifier)

    def __expandMacros(s, anchor):
        """Replaces all occurences of the form $(<name>) with the values of 
        corresponding environment variables. If a name does not refer to an 
        existing environment variable, dies with an error message."""
        while True:
            macroCallStartIndex = s.find("$(")
            if macroCallStartIndex == -1:
                break
            macroCallEndIndex = s.find(")", macroCallStartIndex)
            if macroCallEndIndex == -1:
                break
            variableName = s[macroCallStartIndex + 2:macroCallEndIndex]
            if not variableName in os.environ:
                print anchor + ": Environment variable '" + variableName + "' does not exist"
                sys.exit(1)
            variableValue = os.environ[variableName]
            s = s[0:macroCallStartIndex] + variableValue + s[macroCallEndIndex + 1:]
        return s
    __expandMacros = staticmethod(__expandMacros)

#   End of splbuild_project_reader.pl
