//
//  CARL_Application.hpp - Application object.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.15 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL application framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Application.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Application_hpp_included__
#define __CARL_Application_hpp_included__

namespace carl
{
    /**
     *  @class Application CARL.hpp
     *  A generic application.
     *
     *  Each CARL application must have exactly one statically allocated
     *  instance of some concrete subclass of this class. The CARL client
     *  is responsible for creating a concrete subclass of Application
     *  and a static instance of that concrete subclass.
     *
     *  Static and dynamic libraries that use CARL will not define the
     *  global static instance of Application - derived class.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Application
    {
        friend class Process;
        //////////
        //  Can't assign or copy construct
    private:
        Application(const Application &);
        void                        operator = (const Application &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes and registers a new instance of Application class.
         *  Exactly one instance of Application (created as a part of an
         *  instance of Application - derived concrete subclass) must be
         *  defined.
         *
         *  @since 1.0
         */
        Application()
        {
            require(_pApplication == NULL);
            _pApplication = this;
        }

        /**
         *  The class destructor.
         *
         *  Unregisters the destroyed instance of Application (registered
         *  earlier by Application constructor).
         *
         *  @since 1.0
         */
        virtual ~Application()
        {
            require(_pApplication == this);
            _pApplication = NULL;
        }

        //////////
        //  Operations
    public:
        /**
         *  Returns a pointer to the one and only static Application instance.
         *
         *  The CARL client application is responsible for defining a concrete
         *  Application subclass and a global static instance of that concrete
         *  subclass. During application execution, calls to this method will
         *  return a pointer to that global static instance.
         *
         *  Static and dynamic libraries that use CARL can use this method
         *  to access the Application - derived instance of an application
         *  which links to them. This allows a library to determine the
         *  properties of an application that uses it, such as startup path.
         *
         *  @return A pointer to the one and only global static Application
         *          instance.
         *  @since 1.0
         */
        static Application *        getApplication()
        {
            return _pApplication;
        }

        /**
         *  The application entry point.
         *
         *  An execution of CARL application constitutes of calling this
         *  method with an array of command-line arguments:
         *
         *  -   The 0th element of an array always exists and is a fully
         *      qualified path name of an executable file from which an
         *      application has been started.
         *  -   Remaining elements, if any, specify additional command-line
         *      arguments specified for the application.
         *
         *  @param args A list of command-line arguments. The 0th element of
         *              the list is always the application executable file name.
         *  @return     The application exit code, which is then used as
         *              a process exit code.
         *  @since 1.0
         */
        virtual int                 run(const Array<String> & args) = 0;

        //////////
        //  Implementation
    private:
        static Application *        _pApplication;
        static Int64                _startTimeNanosecondsFromReferencePoint;
    };
    
    #if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        #if !defined(__CARL_BUILD__)
            extern "C" CARL_PUBLIC int main(int argc, char ** argv);
            extern "C" CARL_PUBLIC int __stdcall WinMain(void *, void *, char *, int);
        #endif
    #elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
          (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        #if !defined(__CARL_BUILD__)
            extern "C" CARL_PUBLIC int main(int argc, char ** argv);
        #endif
    #else
        #error Unsupported platform
    #endif
}

#endif  //  __CARL_Application_hpp_included__
//  End of CARL_Application.hpp
