//
//  CARL_Array.hpp - A variable-length array.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.19 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines a variable-length dynamic array.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Array.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Array_hpp_included__
#define __CARL_Array_hpp_included__

namespace carl
{
    namespace _impl
    {
        //////////
        //  Variable-length array implementstion
        //
        //  The value type (T) must be:
        //  *   Default constructable (T::T() must be defined and public)
        //  *   Copy constructable (T::T(const T &) must be defined and public)
        //  *   Destructable (T::~T() must be defined and public)
        //  *   Assignable (T::operator=(const T &) must be defined and public)
        template<class T>
        class _ArrayImpl
        {
            //////////
            //  Can't assign
        private:
            void                    operator = (const _ArrayImpl &);

            //////////
            //  Construction/destruction
        public:
            _ArrayImpl();
            explicit _ArrayImpl(UIntSize initialCapacity);
            _ArrayImpl(const T * pElements, UIntSize elementCount);
            _ArrayImpl(const _ArrayImpl<T> & src);
            ~_ArrayImpl();

            //////////
            //  Operations
        public:
            bool                    isEmpty() const;
            UIntSize                size() const;
            void                    setSize(UIntSize size);
            T &                     operator [] (UIntSize index);
            void                    append(const T & element);
            void                    append(const T * pElements, UIntSize elementCount);
            void                    insertAt(UIntSize index, const T & element);
            void                    insertAt(UIntSize index, const T & element, UIntSize repeatCount);
            void                    insertAt(UIntSize index, const T * pElements, UIntSize elementCount);
            bool                    removeFirst(const T & element, const IComparator<T> & comparator);
            bool                    removeLast(const T & element, const IComparator<T> & comparator);
            UIntSize                removeAll(const T & element, const IComparator<T> & comparator);
            void                    removeAt(UIntSize index, UIntSize elementCount);
            void                    removeAll();
            bool                    contains(const T & element, const IComparator<T> & comparator) const;
            bool                    findFirst(const T & element, UIntSize & position, UIntSize startIndex, const IComparator<T> & comparator) const;
            bool                    findLast(const T & element, UIntSize & position, const IComparator<T> & comparator) const;
            T *                     getData();
            Iterator<T>             enumerateElementsStartToEnd() const;
            Iterator<T>             enumerateElementsEndToStart() const;
            UIntSize                getReferenceCount() const;
            void                    incrementReferenceCount();
            void                    decrementReferenceCount();

            //////////
            //  Implementation
        private:
            UIntSize                _allocated;
            UIntSize                _used;
            T *                     _pData;
            UIntSize                _referenceCount;

            //  Helpers
            void                    _ensureCapacity(UIntSize capacity);
            _impl::_SnapshotIteratorImpl<T> *   _createForwardIteratorImpl() const;
            _impl::_SnapshotIteratorImpl<T> *   _createReverseIteratorImpl() const;
        };

        template<class T>
        _ArrayImpl<T>::_ArrayImpl()
        {
            _allocated = 0;
            _used = 0;
            _pData = NULL;
            _referenceCount = 0;
        }

        template<class T>
        _ArrayImpl<T>::_ArrayImpl(UIntSize initialCapacity)
        {
            _allocated = initialCapacity;
            _used = 0;
            if (initialCapacity == 0)
            {
                _pData = NULL;
            }
            else
            {
                _pData = new T[initialCapacity];
            }
            _referenceCount = 0;
        }

        template<class T>
        _ArrayImpl<T>::_ArrayImpl(const T * pElements, UIntSize elementCount)
        {
            require(pElements != 0 || elementCount);

            _allocated = elementCount;
            _used = elementCount;
            if (elementCount == 0)
            {
                _pData = NULL;
            }
            else
            {
                _pData = new T[elementCount];
                for (UIntSize i = 0; i < elementCount; i++)
                {
                    _pData[i] = pElements[i];
                }
            }
            _referenceCount = 0;
        }

        template<class T>
        _ArrayImpl<T>::_ArrayImpl(const _ArrayImpl<T> & src)
        {
            _allocated = src._used;
            _used = src._used;
            if (_allocated == 0)
            {
                _pData = NULL;
            }
            else
            {
                _pData = new T[_allocated];
                for (UIntSize i = 0; i < _used; i++)
                {
                    _pData[i] = src._pData[i];
                }
            }
            _referenceCount = 0;
        }

        template<class T>
        _ArrayImpl<T>::~_ArrayImpl()
        {
            delete [] _pData;
        }

        template<class T>
        bool _ArrayImpl<T>::isEmpty() const
        {
            return _used == 0;
        }

        template<class T>
        UIntSize _ArrayImpl<T>::size() const
        {
            return _used;
        }

        template<class T>
        void _ArrayImpl<T>::setSize(UIntSize size)
        {
            _ensureCapacity(size);
            _used = size;
        }

        template<class T>
        T & _ArrayImpl<T>::operator [] (UIntSize index)
        {
            require(index < _used);

            return _pData[index];
        }

        template<class T>
        void _ArrayImpl<T>::append(const T & element)
        {
            _ensureCapacity(_used + 1);
            _pData[_used++] = element;
        }

        template<class T>
        void _ArrayImpl<T>::append(const T * pElements, UIntSize elementCount)
        {
            require(pElements != NULL || elementCount == 0);

            if (elementCount != 0)
            {
                _ensureCapacity(_used + elementCount);
                for (UIntSize i = 0; i < elementCount; i++)
                {
                    _pData[_used++] = pElements[i];
                }
            }
        }

        template<class T>
        void _ArrayImpl<T>::insertAt(UIntSize index, const T & element)
        {
            require(index <= _used);

            _ensureCapacity(_used + 1);
            for (UIntSize i = _used; i > index; i--)
            {
                _pData[i] = _pData[i - 1];
            }
            _pData[index] = element;
            _used++;
        }

        template<class T>
        void _ArrayImpl<T>::insertAt(UIntSize index, const T & element, UIntSize repeatCount)
        {
            require(index <= _used);

            if (repeatCount != 0)
            {
                _ensureCapacity(_used + repeatCount);
                for (UIntSize i = _used; i > index; i--)
                {
                    _pData[i + repeatCount - 1] = _pData[i - 1];
                }
                for (UIntSize i = 0; i < repeatCount; i++)
                {
                    _pData[index + i] = element;
                }
                _used += repeatCount;
            }
        }

        template<class T>
        void _ArrayImpl<T>::insertAt(UIntSize index, const T * pElements, UIntSize elementCount)
        {
            require(index <= _used);
            require(pElements != NULL || elementCount == 0);

            if (elementCount != 0)
            {
                _ensureCapacity(_used + elementCount);
                for (UIntSize i = _used; i > index; i--)
                {
                    _pData[i + elementCount - 1] = _pData[i - 1];
                }
                for (UIntSize i = 0; i < elementCount; i++)
                {
                    _pData[index + i] = pElements[i];
                }
                _used += elementCount;
            }
        }

        template<class T>
        bool _ArrayImpl<T>::removeFirst(const T & element, const IComparator<T> & comparator)
        {
            for (UIntSize i = 0; i < _used; i++)
            {
                if (comparator.equal(_pData[i], element))
                {
                    removeAt(i, 1);
                    return true;
                }
            }
            return false;
        }

        template<class T>
        bool _ArrayImpl<T>::removeLast(const T & element, const IComparator<T> & comparator)
        {
            for (UIntSize i = _used; i != 0; i--)
            {
                if (comparator.equal(_pData[i - 1], element))
                {
                    removeAt(i - 1, 1);
                    return true;
                }
            }
            return false;
        }

        template<class T>
        UIntSize _ArrayImpl<T>::removeAll(const T & element, const IComparator<T> & comparator)
        {
            UIntSize result = 0;
            for (UIntSize i = _used; i != 0; i--)
            {
                if (comparator.equal(_pData[i - 1], element))
                {
                    removeAt(i - 1, 1);
                    result++;
                }
            }
            return result;
        }

        template<class T>
        void _ArrayImpl<T>::removeAt(UIntSize index, UIntSize elementCount)
        {
            require(index <= _used);
            require(elementCount <= _used);
            require(index + elementCount >= index &&
                    index + elementCount >= elementCount &&
                    index + elementCount <= _used);

            for (UIntSize i = index + elementCount; i < _used; i++)
            {
                _pData[i - elementCount] = _pData[i];
            }
            _used -= elementCount;
        }

        template<class T>
        void _ArrayImpl<T>::removeAll()
        {
            if (_pData != NULL)
            {
                delete[] _pData;
                _allocated = 0;
                _used = 0;
                _pData = NULL;
            }
        }

        template<class T>
        bool _ArrayImpl<T>::contains(const T & element, const IComparator<T> & comparator) const
        {
            for (UIntSize i = 0; i < _used; i++)
            {
                if (comparator.equal(_pData[i], element))
                {
                    return true;
                }
            }
            return false;
        }

        template<class T>
        bool _ArrayImpl<T>::findFirst(const T & element, UIntSize & position, UIntSize startIndex /*= 0*/, const IComparator<T> & comparator) const
        {
            for (UIntSize i = startIndex; i < _used; i++)
            {
                if (comparator.equal(_pData[i], element))
                {
                    position = i;
                    return true;
                }
            }
            return false;
        }

        template<class T>
        bool _ArrayImpl<T>::findLast(const T & element, UIntSize & position, const IComparator<T> & comparator) const
        {
            for (UIntSize i = _used; i > 0; i--)
            {
                if (comparator.equal(_pData[i - 1], element))
                {
                    position = i - 1;
                    return true;
                }
            }
            return false;
        }

        template<class T>
        T * _ArrayImpl<T>::getData()
        {
            return _pData;
        }

        template<class T>
        Iterator<T> _ArrayImpl<T>::enumerateElementsStartToEnd() const
        {
            return Iterator<T>(_createForwardIteratorImpl());
        }

        template<class T>
        Iterator<T> _ArrayImpl<T>::enumerateElementsEndToStart() const
        {
            return Iterator<T>(_createReverseIteratorImpl());
        }

        template<class T>
        UIntSize _ArrayImpl<T>::getReferenceCount() const
        {
            return _referenceCount;
        }

        template<class T>
        void _ArrayImpl<T>::incrementReferenceCount()
        {
            _referenceCount++;
            require(_referenceCount != 0);
        }

        template<class T>
        void _ArrayImpl<T>::decrementReferenceCount()
        {
            require(_referenceCount != 0);
            _referenceCount--;
        }

        template<class T>
        void _ArrayImpl<T>::_ensureCapacity(UIntSize capacity)
        {
            if (capacity > _allocated)
            {   //  Must expand the array
                UIntSize newAllocated = _allocated;
                while (capacity > newAllocated)
                {
                    newAllocated = (3 * newAllocated / 2) + 1;
                }
                T * pNewData = new T[newAllocated];
                if (_pData != NULL)
                {
                    for (UIntSize i = 0; i < _used; i++)
                    {
                        pNewData[i] = _pData[i];
                    }
                    delete[] _pData;
                }
                _pData = pNewData;
                _allocated = newAllocated;
            }
        }

        template<class T>
        _impl::_SnapshotIteratorImpl<T> * _ArrayImpl<T>::_createForwardIteratorImpl() const
        {
            T * pElements = new T[_used];
            for (UIntSize i = 0; i < _used; i++)
            {
                pElements[i] = _pData[i];
            }
            return new _impl::_SnapshotIteratorImpl<T>(_used, pElements);
        }

        template<class T>
        _impl::_SnapshotIteratorImpl<T> * _ArrayImpl<T>::_createReverseIteratorImpl() const
        {
            T * pElements = new T[_used];
            for (UIntSize i = 0; i < _used; i++)
            {
                pElements[i] = _pData[_used - i - 1];
            }
            return new _impl::_SnapshotIteratorImpl<T>(_used, pElements);
        }
    }

    /**
     *  @class Array CARL.hpp
     *  A variable-length array.
     *
     *  CARL arrays use implicit sharing to reduce the copy construction and 
     *  assignment overhead. This means that it is feasible to both pass
     *  arrays to functions as value parameters and return arrays from functions.
     *
     *  The value type (T) must be:
     *
     *  -   Default constructable (T::T() must be defined and public).
     *  -   Copy constructable (T::T(const T &) must be defined and public).
     *  -   Destructable (T::~T() must be defined and public).
     *  -   Assignable (T::operator=(const T &) must be defined and public).
     *
     *  @param T                The type of array's elements.
     *  @since 1.0
     */
    template<class T>
    class Array
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default class constructor.
         *
         *  Creates an initially empty array.
         *
         *  @since 1.0
         */
        Array();

        /**
         *  The class constructor.
         *
         *  Creates an initially empty array that can accomodate up to the
         *  specified number of elements before it needs to grow.
         *
         *  @param initialCapacity  The initial array capacity. It is guaranteed
         *                          that the array will not have to grow until it
         *                          contains at least that many elements.
         *  @since 1.0
         */
        explicit Array(UIntSize initialCapacity);

        /**
         *  The class constructor.
         *
         *  Creates an array of the specified size and initializes it with
         *  the specified list of values.
         *
         *  @param pElements    A pointer to a list of values with which to
         *                      initialize elements of a newly created array.
         *                      The list must contain at least \a elementCount
         *                      elements; if it contains more then only the 
         *                      first \a elementCount values are used for array
         *                      initialization.
         *  @param elementCount The number of elements a newly created array
         *                      shall have.
         *  @since 1.0
         */
        Array(const T * pElements, UIntSize elementCount);

        /**
         *  The copy constructor.
         *
         *  Initializes a newly created array to be a copy of the specified
         *  array. Array copy construction uses lazy initialization.
         *
         *  @param src          The existing array that shall be copied into
         *                      the newly constructed array.
         *  @since 1.0
         */
        Array(const Array<T> & src);

        /**
         *  The class destructor.
         *
         *  Releases all memory used by the array.
         *
         *  @since 1.0
         */
        ~Array();

        /**
         *  The assignment operator.
         *
         *  Makes the assigned-to array an exact copy of the assigned-from
         *  array. Array assignment uses lazy initialization.
         *
         *  @param src          The existing array that shall be assigned 
         *                      to another array.
         *  @return             A reference to the assigned-to array.
         *  @since 1.0
         */
        Array<T> &              operator = (const Array & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns true iff this array has no elements, false otherwise.
         *
         *  @return             True iff this array has no elements, false otherwise.
         *  @since 1.0
         */
        bool                    isEmpty() const;

        /**
         *  Returns the number of elements in this array.
         *
         *  @return             The number of elements in this array.
         *  @since 1.0
         */
        UIntSize                size() const;

        /**
         *  Sets the number of elements in this array.
         *
         *  If an array size is increased, the values of newly introduced
         *  array elements are unpredictabe. Existing array elements
         *  retain their values up to the minimum of the old and new 
         *  array sizes.
         *
         *  @param size         The new array size.
         *  @since 1.0
         */
        void                    setSize(UIntSize size);

        /**
         *  Returns a copy of the array element at the specified position.
         *
         *  Array indexes always start from 0. The call is an error if the
         *  specified array index lies outside array boundaries.
         *
         *  @param index        The 0-based index of an array element to return.
         *  @return             The copy of the array element at the specified position.
         *  @since 1.0
         */
        T                       operator [] (UIntSize index) const;

        /**
         *  Returns a reference to the array element at the specified position.
         *
         *  Array indexes always start from 0. The call is an error if the
         *  specified array index lies outside array boundaries.
         *
         *  @param index        The 0-based index of an array element to return.
         *  @return             The reference to the array element at the 
         *                      specified position.
         *  @since 1.0
         */
        T &                     operator [] (UIntSize index);

        /**
         *  Appends one element to the end of the array.
         *
         *  The array size is increased by 1.
         *
         *  @param element  The value to append to the end of the array.
         *  @since 1.0
         */
        void                    append(const T & element);

        /**
         *  Appends a list of elements to the end of the array.
         *
         *  The array size is increased by the number of elements appended.
         *
         *  @param pElements    A pointer to the list of values to append
         *                      to the end of the array. The list must contain
         *                      at least \a elementCount elements; if it contains 
         *                      more then only the first \a elementCount values are 
         *                      appended.
         *  @param elementCount The number of elements to append to the array.
         *  @since 1.0
         */
        void                    append(const T * pElements, UIntSize elementCount);

        /**
         *  Appends elements of one array to the end of another array.
         *
         *  The source array (where appended elements come from) remains
         *  unchanged.
         *
         *  @param elements An array of elements to append to this array.
         *  @since 1.0
         */
        void                    append(const Array<T> & elements);

        /**
         *  Inserts one element at an arbitrary location within an array.
         *
         *  The array size is increased by 1.
         *
         *  @param index        The 0-based index at which the new element
         *                      shall be inserted. The index must either be
         *                      within the array boundaries or refer to a 
         *                      position immediately after the last array element.
         *  @param element      The value to insert into the array.
         *  @since 1.0
         */
        void                    insertAt(UIntSize index, const T & element);

        /**
         *  Inserts multiple consecutive copies of the same value at an 
         *  arbitrary location within an array.
         *
         *  The array size is increased by the number of value copies inserted.
         *
         *  @param index        The 0-based index at which the new element
         *                      shall be inserted. The index must either be
         *                      within the array boundaries or refer to a 
         *                      position immediately after the last array element.
         *  @param element      The value to insert into the array.
         *  @param repeatCount  The number of copies of element that shall
         *                      be inserted. If this value is 0, the array
         *                      remains unchanged.
         *  @since 1.0
         */
        void                    insertAt(UIntSize index, const T & element, UIntSize repeatCount);

        /**
         *  Inserts a list of elements at an arbitrary location within an array.
         *
         *  The array size is increased by the number of elements inserted.
         *
         *  @param index        The 0-based index at which the elements shall
         *                      shall be inserted. The index must either be
         *                      within the array boundaries or refer to a 
         *                      position immediately after the last array element.
         *  @param pElements    A pointer to the list of values to insert
         *                      into the array. The list must contain
         *                      at least elementCount elements; if it contains 
         *                      more then only the first elementCount values are 
         *                      inserted.
         *  @param elementCount The number of elements to insert into the array.
         *  @since 1.0
         */
        void                    insertAt(UIntSize index, const T * pElements, UIntSize elementCount);

        /**
         *  Inserts all elements of one array into another array at an 
         *  arbitrary location within an array.
         *
         *  The array size is increased by the number of elements inserted.
         *
         *  @param index        The 0-based index at which the elements shall
         *                      shall be inserted. The index must either be
         *                      within the array boundaries or refer to a 
         *                      position immediately after the last array element.
         *  @param elements     An array of elements to insert into this array.
         *  @since 1.0
         */
        void                    insertAt(UIntSize index, const Array<T> & elements);

        /**
         *  Removes the first occurrence of a value from an array.
         *
         *  The "first occurrence" of a value within an array is defined
         *  as the leftmost position at which an array element compares equal
         *  to the specified value.
         *
         *  If the specified value was not found in the array, the array
         *  remains unchanged; otherwise array size decreases by 1.
         *
         *  @param element      The value to remove from an array.
         *  @param comparator   The algorithm for deciding whether two values are
         *                      equal. If not specified, a natural equality is
         *                      used, where values are compared using the "=="
         *                      operator.
         *  @return             True if an element was actually removed from
         *                      an array; false otherwise.
         *  @since 1.0
         */
        bool                    removeFirst(const T & element, const IComparator<T> & comparator = NaturalComparator<T>());

        /**
         *  Removes the last occurrence of a value from an array.
         *
         *  The "last occurrence" of a value within an array is defined
         *  as the rightmost position at which an array element compares equal
         *  to the specified value.
         *
         *  If the specified value was not found in the array, the array
         *  remains unchanged; otherwise array size decreases by 1.
         *
         *  @param element      The value to remove from an array.
         *  @param comparator   The algorithm for deciding whether two values are
         *                      equal. If not specified, a natural equality is
         *                      used, where values are compared using the "=="
         *                      operator.
         *  @return             True if an element was actually removed from
         *                      an array; false otherwise.
         *  @since 1.0
         */
        bool                    removeLast(const T & element, const IComparator<T> & comparator = NaturalComparator<T>());

        /**
         *  Removes all occurrences of a value from an array.
         *
         *  The "occurrence" of a value within an array is defined as the
         *  position at which an array element compares equal to the 
         *  specified value.
         *
         *  If the specified value was not found in the array, the array
         *  remains unchanged; otherwise array size decreases by the number
         *  of elements removed.
         *
         *  @param element      The value to remove from an array.
         *  @param comparator   The algorithm for deciding whether two values are
         *                      equal. If not specified, a natural equality is
         *                      used, where values are compared using the "=="
         *                      operator.
         *  @return             The total number of values removed from the
         *                      array. This will be 0 if the value was not
         *                      found in the array.
         *  @since 1.0
         */
        UIntSize                removeAll(const T & element, const IComparator<T> & comparator = NaturalComparator<T>());

        /**
         *  Removes one or more array elements starting with the specified position.
         *
         *  All removed array elements must be within array boundaries. The
         *  array size decreases by the number of elements removed.
         *
         *  @param index        The index of the first element to remove.
         *  @param elementCount The number of elements to remove.
         *  @since 1.0
         */
        void                    removeAt(UIntSize index, UIntSize elementCount = 1);

        /**
         *  Removes all array elements, leaving the array empty.
         *
         *  @since 1.0
         */
        void                    removeAll();

        /**
         *  Determines if the array contains a specified element.
         *
         *  The "containemt" of a value within an array is defined as an
         *  existence of some position at which an array element compares 
         *  equal to the specified value.
         *
         *  @param element      The value to look for in the array.
         *  @param comparator   The algorithm for deciding whether two values are
         *                      equal. If not specified, a natural equality is
         *                      used, where values are compared using the "=="
         *                      operator.
         *  @return             True if the specified value exists in an array
         *                      at some position, false otherwise.
         *  @since 1.0
         */
        bool                    contains(const T & element, const IComparator<T> & comparator = NaturalComparator<T>()) const;

        /**
         *  Finds the leftmost occurrence of an element in an array at or
         *  after the specified index.
         *
         *  @param element      The value to look for in the array.
         *  @param position     The recipient for the index of a found element.
         *                      Upon successful search, the value of this parameter
         *                      is set to the leftmost position where a matching
         *                      array element has been found. If the search fails,
         *                      the value of this parameter remains unchanges.
         *  @param startIndex   The leftmost position at which to start the search.
         *                      If this value is 0 (which is the default), the
         *                      entire array is searched.
         *  @param comparator   The algorithm for deciding whether two values are
         *                      equal. If not specified, a natural equality is
         *                      used, where values are compared using the "=="
         *                      operator.
         *  @return             True if an element was found in an array, false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    findFirst(const T & element, UIntSize & position, UIntSize startIndex = 0, const IComparator<T> & comparator = NaturalComparator<T>()) const;

        /**
         *  Finds the rightmost occurrence of an element in an array.
         *
         *  @param element      The value to look for in the array.
         *  @param position     The recipient for the index of a found element.
         *                      Upon successful search, the value of this parameter
         *                      is set to the rightmost position where a matching
         *                      array element has been found. If the search fails,
         *                      the value of this parameter remains unchanges.
         *  @param comparator   The algorithm for deciding whether two values are
         *                      equal. If not specified, a natural equality is
         *                      used, where values are compared using the "=="
         *                      operator.
         *  @return             True if an element was found in an array, false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    findLast(const T & element, UIntSize & position, const IComparator<T> & comparator = NaturalComparator<T>()) const;

        /**
         *  Returns a pointer to array's data buffer.
         *
         *  The return value can be used to read, but not modify, the
         *  array elements directly.
         *
         *  @return             A pointer to the array's internal buffer where
         *                      array elements are stored.
         *  @since 1.0
         */
        const T *               getData() const;

        /**
         *  Returns a nonshared pointer to array's data buffer.
         *
         *  The return value can be used to both read and modify the
         *  array elements directly.
         *
         *  @return             A pointer to the array's internal buffer where
         *                      array elements are stored.
         *  @since 1.0
         */
        T *                     getData();

        /**
         *  Returns a forward iterator for array elements.
         *
         *  The returned iterator behaves as if a snapshot of the
         *  array has been taken. In particular, removing elements from
         *  the array after an iterator has been created has no effect
         *  on the iterator, which will still iterate over the entire
         *  list of array elements as they were at the time when the
         *  iterator was created.
         *
         *  @return             A forward iterator over all array elements.
         *  @since 1.0
         */
        Iterator<T>             enumerateElementsStartToEnd() const;

        /**
         *  Returns a reverse iterator for array elements.
         *
         *  The returned iterator behaves as if a snapshot of the
         *  array has been taken. In particular, removing elements from
         *  the array after an iterator has been created has no effect
         *  on the iterator, which will still iterate over the entire
         *  list of array elements as they were at the time when the
         *  iterator was created.
         *
         *  @return             A reverse iterator over all array elements.
         *  @since 1.0
         */
        Iterator<T>             enumerateElementsEndToStart() const;

        //////////
        //  Implementation
    private:
        _impl::_ArrayImpl<T> *  _pImpl;
        void                    _ensureNonsharedImplementation();
    };

    template<class T>
    Array<T>::Array()
    {
        _pImpl = new _impl::_ArrayImpl<T>();
        _pImpl->incrementReferenceCount();
    }

    template<class T>
    Array<T>::Array(UIntSize initialCapacity)
    {
        _pImpl = new _impl::_ArrayImpl<T>(initialCapacity);
        _pImpl->incrementReferenceCount();
    }

    template<class T>
    Array<T>::Array(const T * pElements, UIntSize elementCount)
    {
        _pImpl = new _impl::_ArrayImpl<T>(pElements, elementCount);
        _pImpl->incrementReferenceCount();
    }

    template<class T>
    Array<T>::Array(const Array<T> & src)
    {
        _pImpl = src._pImpl;
        _pImpl->incrementReferenceCount();
    }

    template<class T>
    Array<T>::~Array()
    {
        _pImpl->decrementReferenceCount();
        if (_pImpl->getReferenceCount() == 0)
        {
            delete _pImpl;
        }
    }

    template<class T>
    Array<T> & Array<T>::operator = (const Array & src)
    {
        if (this != (&src))
        {
            _pImpl->decrementReferenceCount();
            if (_pImpl->getReferenceCount() == 0)
            {
                delete _pImpl;
            }
            _pImpl = src._pImpl;
            _pImpl->incrementReferenceCount();
        }
        return *this;
    }

    template<class T>
    bool Array<T>::isEmpty() const
    {
        return _pImpl->isEmpty();
    }

    template<class T>
    UIntSize Array<T>::size() const
    {
        return _pImpl->size();
    }

    template<class T>
    void Array<T>::setSize(UIntSize size)
    {
        _ensureNonsharedImplementation();
        return _pImpl->setSize(size);
    }

    template<class T>
    T Array<T>::operator [] (UIntSize index) const
    {
        return (*_pImpl)[index];
    }

    template<class T>
    T & Array<T>::operator [] (UIntSize index)
    {
        _ensureNonsharedImplementation();
        return (*_pImpl)[index];
    }

    template<class T>
    void Array<T>::append(const T & element)
    {
        _ensureNonsharedImplementation();
        _pImpl->append(element);
    }

    template<class T>
    void Array<T>::append(const T * pElements, UIntSize elementCount)
    {
        _ensureNonsharedImplementation();
        _pImpl->append(pElements, elementCount);
    }

    template<class T>
    void Array<T>::append(const Array<T> & elements)
    {
        _ensureNonsharedImplementation();
        Array<T> elementsToAppend = elements;
        append(elementsToAppend.getData(), elementsToAppend.size());
    }

    template<class T>
    void Array<T>::insertAt(UIntSize index, const T & element)
    {
        _ensureNonsharedImplementation();
        _pImpl->insertAt(index, element);
    }

    template<class T>
    void Array<T>::insertAt(UIntSize index, const T & element, UIntSize repeatCount)
    {
        _ensureNonsharedImplementation();
        _pImpl->insertAt(index, element, repeatCount);
    }

    template<class T>
    void Array<T>::insertAt(UIntSize index, const T * pElements, UIntSize elementCount)
    {
        _ensureNonsharedImplementation();
        _pImpl->insertAt(index, pElements, elementCount);
    }

    template<class T>
    void Array<T>::insertAt(UIntSize index, const Array<T> & elements)
    {
        _ensureNonsharedImplementation();
        Array<T> elementsToInsert = elements;
        insertAt(index, elementsToInsert.getData(), elementsToInsert.size());
    }

    template<class T>
    bool Array<T>::removeFirst(const T & element, const IComparator<T> & comparator /*= NaturalComparator<T>()*/)
    {
        _ensureNonsharedImplementation();
        return _pImpl->removeFirst(element, comparator);
    }

    template<class T>
    bool Array<T>::removeLast(const T & element, const IComparator<T> & comparator /*= NaturalComparator<T>()*/)
    {
        _ensureNonsharedImplementation();
        return _pImpl->removeLast(element, comparator);
    }

    template<class T>
    UIntSize Array<T>::removeAll(const T & element, const IComparator<T> & comparator /*= NaturalComparator<T>()*/)
    {
        _ensureNonsharedImplementation();
        return _pImpl->removeAll(element, comparator);
    }

    template<class T>
    void Array<T>::removeAt(UIntSize index, UIntSize elementCount /*= 1*/)
    {
        _ensureNonsharedImplementation();
        _pImpl->removeAt(index, elementCount);
    }

    template<class T>
    void Array<T>::removeAll()
    {
        _ensureNonsharedImplementation();
        _pImpl->removeAll();
    }

    template<class T>
    bool Array<T>::contains(const T & element, const IComparator<T> & comparator /*= NaturalComparator<T>()*/) const
    {
        return _pImpl->contains(element, comparator);
    }

    template<class T>
    bool Array<T>::findFirst(const T & element, UIntSize & position, UIntSize startIndex /*= 0*/, const IComparator<T> & comparator /*= NaturalComparator<T>()*/) const
    {
        return _pImpl->findFirst(element, position, startIndex, comparator);
    }

    template<class T>
    bool Array<T>::findLast(const T & element, UIntSize & position, const IComparator<T> & comparator /*= NaturalComparator<T>()*/) const
    {
        return _pImpl->findLast(element, position, comparator);
    }

    template<class T>
    const T * Array<T>::getData() const
    {
        return _pImpl->getData();
    }

    template<class T>
    T * Array<T>::getData()
    {
        _ensureNonsharedImplementation();
        return _pImpl->getData();
    }

    template<class T>
    Iterator<T> Array<T>::enumerateElementsStartToEnd() const
    {
        return _pImpl->enumerateElementsStartToEnd();
    }

    template<class T>
    Iterator<T> Array<T>::enumerateElementsEndToStart() const
    {
        return _pImpl->enumerateElementsEndToStart();
    }

    template<class T>
    void Array<T>::_ensureNonsharedImplementation()
    {
        if (_pImpl->getReferenceCount() != 1)
        {
            _impl::_ArrayImpl<T> * pNewImpl = new _impl::_ArrayImpl<T>(*_pImpl);
            _pImpl->decrementReferenceCount();
            pNewImpl->incrementReferenceCount();
            _pImpl = pNewImpl;
        }
    }
}

#endif  //  __CARL_Array_hpp_included__
//  End of CARL_Array.hpp
