//
//  CARL_BasicTypes.hpp - Provides platform-independent basic types.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/25 12:11:26 $
//  $Revision: 1.18 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines basic types.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_BasicTypes.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_BasicTypes_hpp_included__
#define __CARL_BasicTypes_hpp_included__

namespace carl
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)

    //////////
    //  Integer types
    /**
     *  A 8-bit signed integer value.
     *
     *  @since 1.0
     */
    typedef signed char         Int8;

    /**
     *  A 16-bit signed integer value.
     *
     *  @since 1.0
     */
    typedef signed short        Int16;

    /**
     *  A 32-bit signed integer value.
     *
     *  @since 1.0
     */
    typedef signed int          Int32;

    /**
     *  A 64-bit signed integer value.
     *
     *  @since 1.0
     */
    typedef signed __int64      Int64;

    /**
     *  A 8-bit unsigned integer value.
     *
     *  @since 1.0
     */
    typedef unsigned char       UInt8;

    /**
     *  A 16-bit unsigned integer value.
     *
     *  @since 1.0
     */
    typedef unsigned short      UInt16;

    /**
     *  A 32-bit unsigned integer value.
     *
     *  @since 1.0
     */
    typedef unsigned int        UInt32;

    /**
     *  A 64-bit unsigned integer value.
     *
     *  @since 1.0
     */
    typedef unsigned __int64    UInt64;

    /**
     *  An unsigned integer value that can be used to express sizes of objects in memory.
     *
     *  @since 1.0
     */
    typedef unsigned int        UIntSize;

    /**
     *  An unsigned integer value that can be used to represent pointers numerically.
     *
     *  @since 1.0
     */
    typedef unsigned int        UIntPtr;

    /**
     *  A signed integer value that can be used to represent difference between two pointers.
     *
     *  @since 1.0
     */
    typedef signed int          IntPtrDiff;

    //////////
    //  Real types
    /**
     *  A 32-bit IEEE754 floating-point value.
     *
     *  @since 1.0
     */
    typedef float               Real32;

    /**
     *  A 64-bit IEEE754 floating-point value.
     *
     *  @since 1.0
     */
    typedef double              Real64;

    //////////
    //  Character type
    /**
     *  A 32-bit ISO-10646 character.
     *
     *  @since 1.0
     */
    typedef unsigned int        Char;

    //////////
    //  Macros for forming constants of basic types
    /**
     *  Expands into a valid syntactic form for a 8-bit signed integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_INT8_CONSTANT(n)       (static_cast<carl::Int8>(n))

    /**
     *  Expands into a valid syntactic form for a 16-bit signed integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_INT16_CONSTANT(n)      (static_cast<carl::Int16>(n))

    /**
     *  Expands into a valid syntactic form for a 32-bit signed integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_INT32_CONSTANT(n)      (static_cast<carl::Int32>(n))
    
    /**
     *  Expands into a valid syntactic form for a 64-bit signed integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_INT64_CONSTANT(n)      (static_cast<carl::Int64>(n##I64))
    
    /**
     *  Expands into a valid syntactic form for a 8-bit unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINT8_CONSTANT(n)      (static_cast<carl::UInt8>(n))

    /**
     *  Expands into a valid syntactic form for a 16-bit unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINT16_CONSTANT(n)     (static_cast<carl::UInt16>(n))

    /**
     *  Expands into a valid syntactic form for a 32-bit unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINT32_CONSTANT(n)     (static_cast<carl::UInt32>(n))

    /**
     *  Expands into a valid syntactic form for a 64-bit unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINT64_CONSTANT(n)     (static_cast<carl::UInt64>(n##I64))
    
    /**
     *  Expands into a valid syntactic form for a size unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINTSIZE_CONSTANT(n)   (static_cast<carl::UIntSize>(n))
    
    /**
     *  Expands into a valid syntactic form for a pointer unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINTPTR_CONSTANT(n)    (static_cast<carl::UIntPtr>(n))
    
    /**
     *  Expands into a valid syntactic form for a pointer difference signed integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_INTPTRDIFF_CONSTANT(n) (static_cast<carl::IntPtrDiff>(n))
    
    /**
     *  Expands into a valid syntactic form for a 32-bit IEEE754 real constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_REAL32_CONSTANT(n)     (static_cast<carl::Real32>(n))

    /**
     *  Expands into a valid syntactic form for a 64-bit IEEE754 real constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_REAL64_CONSTANT(n)     (static_cast<carl::Real64>(n))
    
    /**
     *  Expands into a valid syntactic form for an ISO-10646 character constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_CHAR_CONSTANT(n)       (static_cast<carl::Char>(n))

#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    //////////
    //  Integer types
    /**
     *  A 8-bit signed integer value.
     *
     *  @since 1.0
     */
    typedef signed char         Int8;

    /**
     *  A 16-bit signed integer value.
     *
     *  @since 1.0
     */
    typedef signed short        Int16;

    /**
     *  A 32-bit signed integer value.
     *
     *  @since 1.0
     */
    typedef signed int          Int32;

    /**
     *  A 64-bit signed integer value.
     *
     *  @since 1.0
     */
    typedef signed long long    Int64;

    /**
     *  A 8-bit unsigned integer value.
     *
     *  @since 1.0
     */
    typedef unsigned char       UInt8;

    /**
     *  A 16-bit unsigned integer value.
     *
     *  @since 1.0
     */
    typedef unsigned short      UInt16;

    /**
     *  A 32-bit unsigned integer value.
     *
     *  @since 1.0
     */
    typedef unsigned int        UInt32;

    /**
     *  A 64-bit unsigned integer value.
     *
     *  @since 1.0
     */
    typedef unsigned long long  UInt64;

    /**
     *  An unsigned integer value that can be used to express sizes of objects in memory.
     *
     *  @since 1.0
     */
    typedef unsigned int        UIntSize;

    /**
     *  An unsigned integer value that can be used to represent pointers numerically.
     *
     *  @since 1.0
     */
    typedef unsigned int        UIntPtr;

    /**
     *  A signed integer value that can be used to represent difference between two pointers.
     *
     *  @since 1.0
     */
    typedef signed int          IntPtrDiff;

    //////////
    //  Real types
    /**
     *  A 32-bit IEEE754 floating-point value.
     *
     *  @since 1.0
     */
    typedef float               Real32;

    /**
     *  A 64-bit IEEE754 floating-point value.
     *
     *  @since 1.0
     */
    typedef double              Real64;

    //////////
    //  Character type
    /**
     *  A 32-bit ISO-10646 character.
     *
     *  @since 1.0
     */
    typedef unsigned int        Char;

    //////////
    //  Macros for forming constants of basic types
    /**
     *  Expands into a valid syntactic form for a 8-bit signed integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_INT8_CONSTANT(n)       (static_cast<carl::Int8>(n))

    /**
     *  Expands into a valid syntactic form for a 16-bit signed integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_INT16_CONSTANT(n)      (static_cast<carl::Int16>(n))

    /**
     *  Expands into a valid syntactic form for a 32-bit signed integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_INT32_CONSTANT(n)      (static_cast<carl::Int32>(n))

    /**
     *  Expands into a valid syntactic form for a 64-bit signed integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_INT64_CONSTANT(n)      (static_cast<carl::Int64>(n##ll))

    /**
     *  Expands into a valid syntactic form for a 8-bit unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINT8_CONSTANT(n)      (static_cast<carl::UInt8>(n))

    /**
     *  Expands into a valid syntactic form for a 16-bit unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINT16_CONSTANT(n)     (static_cast<carl::UInt16>(n))

    /**
     *  Expands into a valid syntactic form for a 32-bit unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINT32_CONSTANT(n)     (static_cast<carl::UInt32>(n))

    /**
     *  Expands into a valid syntactic form for a 64-bit unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINT64_CONSTANT(n)     (static_cast<carl::UInt64>(n##ll))

    /**
     *  Expands into a valid syntactic form for a size unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINTSIZE_CONSTANT(n)   (static_cast<carl::UIntSize>(n))

    /**
     *  Expands into a valid syntactic form for a pointer unsigned integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_UINTPTR_CONSTANT(n)    (static_cast<carl::UIntPtr>(n))

    /**
     *  Expands into a valid syntactic form for a pointer difference signed integer constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_INTPTRDIFF_CONSTANT(n) (static_cast<carl::IntPtrDiff>(n))

    /**
     *  Expands into a valid syntactic form for a 32-bit IEEE754 real constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_REAL32_CONSTANT(n)     (static_cast<carl::Real32>(n))

    /**
     *  Expands into a valid syntactic form for a 64-bit IEEE754 real constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_REAL64_CONSTANT(n)     (static_cast<carl::Real64>(n))

    /**
     *  Expands into a valid syntactic form for an ISO-10646 character constant.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define CARL_CHAR_CONSTANT(n)       (static_cast<carl::Char>(n))

#else
    #error Unsupported platform
#endif

    /**
     *  @class Limits CARL.hpp
     *  Defines platform-specific limits for basic CARL types.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Limits
    {
        //////////
        //  Utility classes cannot be instantiated
    private:
        Limits();
        Limits(const Limits &);
        ~Limits();
        void                    operator = (const Limits &);

        //////////
        //  Limits
    public:
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        /**
         *  The smallest representable value of type Int8.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int8       Int8_Min = CARL_INT8_CONSTANT(-128);

        /**
         *  The largest representable value of type Int8.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int8       Int8_Max = CARL_INT8_CONSTANT(127);

        /**
         *  The smallest representable value of type Int16.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int16      Int16_Min = CARL_INT16_CONSTANT(-32768);

        /**
         *  The largest representable value of type Int16.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int16      Int16_Max = CARL_INT16_CONSTANT(32767);

        /**
         *  The smallest representable value of type Int32.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int32      Int32_Min = CARL_INT32_CONSTANT(-CARL_INT32_CONSTANT(2147483648));

        /**
         *  The largest representable value of type Int32.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int32      Int32_Max = CARL_INT32_CONSTANT(2147483647);

        /**
         *  The smallest representable value of type Int64.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int64      Int64_Min = CARL_INT64_CONSTANT(-9223372036854775808);

        /**
         *  The largest representable value of type Int64.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int64      Int64_Max = CARL_INT64_CONSTANT(9223372036854775807);

        /**
         *  The smallest representable value of type UInt8.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt8      UInt8_Min = CARL_UINT8_CONSTANT(0);

        /**
         *  The largest representable value of type UInt8.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt8      UInt8_Max = CARL_UINT8_CONSTANT(255);

        /**
         *  The smallest representable value of type UInt16.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt16     UInt16_Min = CARL_UINT16_CONSTANT(0);

        /**
         *  The largest representable value of type UInt16.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt16     UInt16_Max = CARL_UINT16_CONSTANT(65535);

        /**
         *  The smallest representable value of type UInt32.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt32     UInt32_Min = CARL_UINT32_CONSTANT(0);

        /**
         *  The largest representable value of type UInt32.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt32     UInt32_Max = CARL_UINT32_CONSTANT(4294967295);

        /**
         *  The smallest representable value of type UInt64.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt64     UInt64_Min = CARL_UINT64_CONSTANT(0);

        /**
         *  The largest representable value of type UInt64.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt64     UInt64_Max = CARL_UINT64_CONSTANT(18446744073709551615);

        /**
         *  The smallest representable value of type UIntSize.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntSize   UIntSize_Min = CARL_UINTSIZE_CONSTANT(0);

        /**
         *  The largest representable value of type UIntSize.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntSize   UIntSize_Max = CARL_UINTSIZE_CONSTANT(4294967295);

        /**
         *  The smallest representable value of type UIntPtr.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntPtr    UIntPtr_Min = CARL_UINTPTR_CONSTANT(0);

        /**
         *  The largest representable value of type UIntPtr.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntPtr    UIntPtr_Max = CARL_UINTPTR_CONSTANT(4294967295);

        /**
         *  The smallest representable value of type IntPtrDiff.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const IntPtrDiff IntPtrDiff_Min = CARL_INTPTRDIFF_CONSTANT(-CARL_INTPTRDIFF_CONSTANT(2147483648));

        /**
         *  The largest representable value of type IntPtrDiff.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const IntPtrDiff IntPtrDiff_Max = CARL_INTPTRDIFF_CONSTANT(2147483647);

        /**
         *  The smallest representable value of type Char.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char       Char_Min = CARL_CHAR_CONSTANT(0);

        /**
         *  The largest representable value of type Char.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char       Char_Max = CARL_CHAR_CONSTANT(4294967295);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        /**
         *  The smallest representable value of type Int8.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int8       Int8_Min = CARL_INT8_CONSTANT(-128);

        /**
         *  The largest representable value of type Int8.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int8       Int8_Max = CARL_INT8_CONSTANT(127);

        /**
         *  The smallest representable value of type Int16.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int16      Int16_Min = CARL_INT16_CONSTANT(-32768);

        /**
         *  The largest representable value of type Int16.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int16      Int16_Max = CARL_INT16_CONSTANT(32767);

        /**
         *  The smallest representable value of type Int32.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int32      Int32_Min = CARL_INT32_CONSTANT(-CARL_INT32_CONSTANT(2147483647) - 1);

        /**
         *  The largest representable value of type Int32.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int32      Int32_Max = CARL_INT32_CONSTANT(2147483647);

        /**
         *  The smallest representable value of type Int64.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int64      Int64_Min = CARL_INT64_CONSTANT(-9223372036854775807) - 1;

        /**
         *  The largest representable value of type Int64.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Int64      Int64_Max = CARL_INT64_CONSTANT(9223372036854775807);

        /**
         *  The smallest representable value of type UInt8.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt8      UInt8_Min = CARL_UINT8_CONSTANT(0);

        /**
         *  The largest representable value of type UInt8.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt8      UInt8_Max = CARL_UINT8_CONSTANT(255);

        /**
         *  The smallest representable value of type UInt16.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt16     UInt16_Min = CARL_UINT16_CONSTANT(0);

        /**
         *  The largest representable value of type UInt16.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt16     UInt16_Max = CARL_UINT16_CONSTANT(65535);

        /**
         *  The smallest representable value of type UInt32.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt32     UInt32_Min = CARL_UINT32_CONSTANT(0);

        /**
         *  The largest representable value of type UInt32.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt32     UInt32_Max = CARL_UINT32_CONSTANT(0xFFFFFFFF);

        /**
         *  The smallest representable value of type UInt64.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt64     UInt64_Min = CARL_UINT64_CONSTANT(0);

        /**
         *  The largest representable value of type UInt64.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UInt64     UInt64_Max = CARL_UINT64_CONSTANT(0xFFFFFFFFFFFFFFFF);

        /**
         *  The smallest representable value of type UIntSize.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntSize   UIntSize_Min = CARL_UINTSIZE_CONSTANT(0);

        /**
         *  The largest representable value of type UIntSize.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntSize   UIntSize_Max = CARL_UINTSIZE_CONSTANT(0xFFFFFFFF);

        /**
         *  The smallest representable value of type UIntPtr.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntPtr    UIntPtr_Min = CARL_UINTPTR_CONSTANT(0);

        /**
         *  The largest representable value of type UIntPtr.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntPtr    UIntPtr_Max = CARL_UINTPTR_CONSTANT(0xFFFFFFFF);

        /**
         *  The smallest representable value of type IntPtrDiff.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const IntPtrDiff IntPtrDiff_Min = CARL_INTPTRDIFF_CONSTANT(-CARL_INTPTRDIFF_CONSTANT(2147483647) - 1 );

        /**
         *  The largest representable value of type IntPtrDiff.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const IntPtrDiff IntPtrDiff_Max = CARL_INTPTRDIFF_CONSTANT(2147483647);

        /**
         *  The smallest representable value of type Char.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char       Char_Min = CARL_CHAR_CONSTANT(0);

        /**
         *  The largest representable value of type Char.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char       Char_Max = CARL_CHAR_CONSTANT(0xFFFFFFFF);
#else
    #error Unsupported platform
#endif
    };
}

#ifndef NULL
    /**
     *  A symbollic constant that can be used for nil pointers.
     *
     *  @hideinitializer
     *  @since 1.0
     */
    #define NULL 0
#endif

#endif  //  __CARL_BasicTypes_hpp_included__
//  End of CARL_BasicTypes.hpp
