//
//  CARL_Charset.hpp - Character sets.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/08/07 21:34:25 $
//  $Revision: 1.17 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL character set framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Charset.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Charset_hpp_included__
#define __CARL_Charset_hpp_included__

namespace carl
{
    class CARL_PUBLIC CharacterSetFamily;

    /**
     *  @class InconsistentCharacterSetException CARL.hpp
     *  An exception thrown when an attempt is made to register an 
     *  inconsistent character set.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC InconsistentCharacterSetException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        InconsistentCharacterSetException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        InconsistentCharacterSetException(const InconsistentCharacterSetException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~InconsistentCharacterSetException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        InconsistentCharacterSetException & operator = (const InconsistentCharacterSetException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class DuplicateCharacterSetException CARL.hpp
     *  An exception thrown when an attempt is made to register a 
     *  duplicate character set.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DuplicateCharacterSetException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        DuplicateCharacterSetException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        DuplicateCharacterSetException(const DuplicateCharacterSetException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~DuplicateCharacterSetException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        DuplicateCharacterSetException &    operator = (const DuplicateCharacterSetException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class InconsistentCharacterSetFamilyException CARL.hpp
     *  An exception thrown when an attempt is made to register an 
     *  inconsistent character set family.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC InconsistentCharacterSetFamilyException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        InconsistentCharacterSetFamilyException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        InconsistentCharacterSetFamilyException(const InconsistentCharacterSetFamilyException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~InconsistentCharacterSetFamilyException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        InconsistentCharacterSetFamilyException &   operator = (const InconsistentCharacterSetFamilyException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class DuplicateCharacterSetFamilyException CARL.hpp
     *  An exception thrown when an attempt is made to register a 
     *  duplicate character set family.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DuplicateCharacterSetFamilyException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        DuplicateCharacterSetFamilyException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        DuplicateCharacterSetFamilyException(const DuplicateCharacterSetFamilyException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~DuplicateCharacterSetFamilyException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        DuplicateCharacterSetFamilyException &  operator = (const DuplicateCharacterSetFamilyException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class CharacterSet CARL.hpp
     *  A generic character set.
     *
     *  A character set defines:
     *
     *  -   A subset of ISO-10646.
     *  -   A set of rules for converting between characters that belong
     *      to the above subset and sequence of bytes.
     *
     *  Currently, the following character sets are predefined by 
     *  CARL (the list contains character set and character set family
     *  names, with display names in brackets after the identifier):
     *
     *  -   ISO_8859 (ISO 8859 character sets)
     *      -   ISO_8859-1:1987 (ISO Latin 1)
     *      -   ISO_8859-2:1987 (ISO Latin 2)
     *      -   ISO_8859-3:1988 (ISO Latin 3)
     *      -   ISO_8859-4:1988 (ISO Latin 4)
     *      -   ISO_8859-5:1988 (ISO Latin Cyrillic)
     *      -   ISO_8859-6:1987 (ISO Latin Arabic)
     *      -   ISO_8859-7:1987 (ISO Latin Greek)
     *      -   ISO_8859-8:1988 (ISO Latin Hebrew)
     *      -   ISO_8859-9:1989 (ISO Latin 5)
     *      -   ISO_8859-10:1992 (ISO Latin 6)
     *      -   ISO_8859-11:1999 (ISO Thai)
     *      -   ISO_8859-13:1998 (ISO Latin 7)
     *      -   ISO_8859-14:1998 (ISO Latin 8)
     *      -   ISO_8859-15 (ISO Latin 9)
     *  -   ISO_10646 (ISO 10646 character sets)
     *      -   UTF-32BE (UTF-32 Big-Endian)
     *      -   UTF-32LE (UTF-32 Little-Endian)
     *      -   UTF-16BE (UTF-16 Big-Endian)
     *      -   UTF-16LE (UTF-16 Little-Endian)
     *      -   UTF-8 (UTF-8)
     *  -   Windows (MS Windows)
     *      -   Windows-1250 (Windows-1250 (Central Europe))
     *      -   Windows-1251 (Windows-1251 (Cyrillic))
     *      -   Windows-1252 (Windows-1252 (Latin1))
     *      -   Windows-1253 (Windows-1253 (Greek))
     *      -   Windows-1254 (Windows-1254 (Turkish))
     *      -   Windows-1255 (Windows-1255 (Hebrew))
     *      -   Windows-1256 (Windows-1256 (Arabic))
     *      -   Windows-1257 (Windows-1257 (Baltic))
     *      -   Windows-1258 (Windows-1258 (Viet Nam))
     *      -   cp874 (Cp874 (Thai))
     *      -   cp932 (Cp932 (Japanese Shift-JIS))
     *      -   cp936 (Cp936 (Simplified Chinese GBK))
     *      -   cp949 (Cp949 (Korean))
     *      -   cp950 (Cp950 (Traditional Chinese Big5))
     *  -   PC (IBM PC character sets)
     *      -   IBM437 (Cp437 (United States))
     *      -   IBM737 (Cp737 (Greek))
     *      -   IBM775 (Cp775 (Baltic))
     *      -   IBM850 (Cp850 (Multilingual Latin I))
     *      -   IBM852 (Cp852 (Latin II))
     *      -   IBM855 (Cp855 (Cyrillic))
     *      -   IBM856 (Cp855 (Hebrew))
     *      -   IBM857 (Cp857 (Turkish))
     *      -   IBM860 (Cp860 (Portuguese))
     *      -   IBM861 (Cp861 (Icelandic))
     *      -   IBM862 (Cp862 (Hebrew))
     *      -   IBM863 (Cp863 (Canadian French))
     *      -   IBM864 (Cp864 (Arabic))
     *      -   IBM865 (Cp865 (Nordic))
     *      -   IBM866 (Cp866 (Russian))
     *      -   IBM869 (Cp869 (Modern Greek))
     *  -   EBCDIC (EBCDIC character sets)
     *      -   IBM037 (Cp037 (IBM US & Canada))
     *      -   IBM424 (Cp424 (IBM Hebrew))
     *      -   IBM500 (Cp500 (IBM International))
     *      -   IBM875 (Cp500 (IBM Greek))
     *      -   IBM500 (Cp1026 (IBM Turkish))
     *  -   ADOBE (Adobe character sets)
     *      -   ADOBESTANDARD (Adobe Standard)
     *      -   ADOBESYMBOL (Adobe Symbol)
     *      -   ADOBEZAPFDINGBATS (Adobe Zapf Dingbats)
     *  -   KOI (KOI cyryllic character sets)
     *      -   GOST-19768-87
     *      -   KOI-7
     *      -   KOI-8A
     *      -   KOI-8B
     *      -   KOI-8E
     *      -   KOI-8F
     *      -   KOI-8R
     *      -   KOI-8RU
     *      -   KOI-8U
     *  -   MAC (Mac OS character sets)
     *      -   MacArabic (Mac OS Arabic)
     *      -   MacCeltic (Mac OS Celtic)
     *      -   MacCentEuro (Mac OS Central European)
     *      -   MacCroatian (Mac OS Croatian)
     *      -   MacCyrillic (Mac OS Cyrillic)
     *      -   MacDevanagari (Mac OS Devanagari)
     *      -   MacDingbats (Mac OS Dingbats)
     *      -   MacFarsi (Mac OS Farsi)
     *      -   MacGaelic (Mac OS Gaelic)
     *      -   MacGreek (Mac OS Greek)
     *      -   MacGujarati (Mac OS Gujarati)
     *      -   MacGurmukhi (Mac OS Gurmukhi)
     *      -   MacHebrew (Mac OS Hebrew)
     *      -   MacIcelandic (Mac OS Icelandic)
     *      -   MacInuit (Mac OS Inuit)
     *      -   MacRoman (Mac OS Roman)
     *      -   MacRomanian (Mac OS Romanian)
     *      -   MacSymbol (Mac OS Symbol)
     *      -   MacThai (Mac OS Thai)
     *      -   MacTurkish (Mac OS Turkish)
     *      -   MacUkrainian (Mac OS Ukrainian)
     *      -   MacChineseSimplified (Mac OS Chinese Simplified)
     *      -   MacChineseTraditional (Mac OS Chinese Traditional)
     *      -   MacKorean (Mac OS Korean)
     *      -   MacJapanese (Mac OS Japanese)
     *  -   MISCELLANEOUS (Miscellaneous character sets)
     *      -   ANSI_X3.4-1968 (US ASCII)
     *      -   NEXTSTEP (NextStep)
     *      -   ATARIST (Atari ST)
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC CharacterSet
    {
        //////////
        //  Can't assign or copy construct
    private:
        CharacterSet(const CharacterSet &);
        void                        operator = (const CharacterSet &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        CharacterSet() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~CharacterSet() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified character set with the CARL framework.
         *
         *  A registration of a character set informs the CARL framework that
         *  a specific character set exists. From this point on, the application
         *  can use the newly registers character set in exactly the same manner
         *  as a predefined character set. Once registered, a character set
         *  cannot be unregistered.
         *
         *  A character set family must be registered before any of its member
         *  character sets are registered.
         *
         *  A character set cannot be registered if one of its names clashes
         *  with one of the names of some character set already registered.
         *
         *  @param pCharacterSet    The character set to register.
         *  @exception InconsistentCharacterSetException    If the character set being
         *                          registered is inconsistent (for example, its
         *                          canonical name is not a member of its "all names" 
         *                          list).
         *  @exception DuplicateCharacterSetException   If one of the character
         *                          set names clashes with one of the names of another
         *                          character set registered earlier.
         *  @since 1.0
         */
        static void                 registerInstance(CharacterSet * pCharacterSet) throw (InconsistentCharacterSetException, DuplicateCharacterSetException);

        /**
         *  Returns a list of all registered character sets.
         *
         *  The returned list contains all character sets registered so far,
         *  including all predefined character sets. Elements in the list are
         *  not guaranteed to be in any specific order.
         *
         *  @return                 The list of all registered character sets.
         *  @since 1.0
         */
        static Array<CharacterSet*> getAllInstances();

        /**
         *  Finds and returns a character set with the specified name.
         *
         *  Character set names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the character set.
         *  @return                 The character set found.
         *  @exception NoSuchElementError   If a character set with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static CharacterSet *       getInstance(const char * pName);

        /**
         *  Finds and returns a character set with the specified name.
         *
         *  Character set names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ISO-10646 string
         *                          representing the name of the character set.
         *  @return                 The character set found.
         *  @exception NoSuchElementError   If a character set with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static CharacterSet *       getInstance(const Char * pName);

        /**
         *  Finds and returns a character set with the specified name.
         *
         *  Character set names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          character set.
         *  @return                 The character set found.
         *  @exception NoSuchElementError   If a character set with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static CharacterSet *       getInstance(const String & name);

        /**
         *  Returns the system character set.
         *
         *  A ayatem character set is the preferred character set for the
         *  platform where the CARL client runs.
         *
         *  @return                 The system character set.
         *  @since 1.0
         */
        static CharacterSet *       getSystem();

        /**
         *  Returns the current default character set.
         *
         *  When a CARL application starts, the current default character set
         *  is the same as the system character set. However, the application
         *  can change the current default character set if necessary.
         *
         *  @return                 The current default character set.
         *  @since 1.0
         */
        static CharacterSet *       getDefault();

        /**
         *  Changes the current default character set.
         *
         *  When a CARL application starts, the current default character set
         *  is the same as the system character set. However, the application
         *  can change the current default character set if necessary.
         *
         *  @param pCharacterSet    The character set to make a new current default.
         *  @since 1.0
         */
        static void                 setDefault(CharacterSet * pCharacterSet);

        /**
         *  Returns the family to which this character set belongs.
         *
         *  Character sets are organized into groups called "families", with
         *  each character set family representing a closely related group of
         *  character sets (such as a vendor or a platform). Each character set
         *  always belongs to exactly one character set family.
         *
         *  @return                 The family to which this character set belongs.
         *  @since 1.0
         */
        virtual CharacterSetFamily *    getFamily() const = 0;

        /**
         *  Returns a list of all names of this character set.
         *
         *  Elements in the list are not guaranteed to be in any particular order.
         *
         *  @return                 A list of all names of this character set.
         *  @since 1.0
         */
        virtual Array<String>       getAllNames() const = 0;

        /**
         *  Returns the canonical name of this character set.
         *
         *  A canonical name of a character set is a name under which the
         *  character set is known to standards committees.
         *
         *  @return                 The canonical name of this character set.
         *  @since 1.0
         */
        virtual String              getCanonicalName() const = 0;

        /**
         *  Returns the preferred name of this character set.
         *
         *  A preferred name of a character set is a name which is
         *  historically most frequently used to refer to that
         *  character set (in MIME headers, in HTML headers, etc.)
         *
         *  @return                 The preferred name of this character set.
         *  @since 1.0
         */
        virtual String              getPreferredName() const = 0;

        /**
         *  Returns the user-readable display name of this character set
         *  for the current default locale.
         *
         *  @return                 The display name of this character set.
         *  @since 1.0
         */
        virtual String              getDisplayName() const = 0;

        /**
         *  Returns the user-readable display name of this character set
         *  for the specified locale.
         *
         *  @param forLocale        The locale for which the display name of
         *                          the character set shall be tailored.
         *  @return                 The display name of this character set.
         *  @since 1.0
         */
        virtual String              getDisplayName(const Locale & forLocale) const = 0;

        /**
         *  Returns the maximum length, in bytes, of character encoding
         *  supported by this character set.
         *
         *  It is guaranteed that any character that is a member of this
         *  character set can be encoded using at most the specified
         *  number of bytes.
         *
         *  @return                 The maximum number of bytes required by
         *                          this character set to encode a character.
         *  @since 1.0
         */
        virtual int                 getMaximumEncodingLength() const = 0;

        /**
         *  Returns the length, in bytes, of an encoding for the specified
         *  character.
         *
         *  This method can also be used to determine if a specific character
         *  is a member of the character set (if it is not, the return value
         *  will be -1).
         *
         *  @param c                The character whose encoding is interrogated.
         *  @return                 The number of bytes required by this character
         *                          set to encode the specified character, or -1
         *                          if the specified character is not a member of
         *                          this character set.
         *  @since 1.0
         */
        virtual int                 getEncodingLengthForCharacter(Char c) const = 0;

        /**
         *  Converts character into a sequence of bytes.
         *
         *  The specified character is converted into a sequence of bytes
         *  which are stored in memory starting at the specified address.
         *
         *  @param c                The character to convert.
         *  @param buffer           The address at which the bytes resulting
         *                          from character conversion are stored.
         *  @return                 The number of bytes stored, or -1 if the
         *                          specified character is not a member of this character set.
         *  @since 1.0
         */
        virtual int                 encodeCharacter(Char c, void * buffer) const = 0;

        /**
         *  Determine the length of character encoding by the 1st byte of
         *  that encoding.
         *
         *  When a sequence of bytes is converted into a sequence of characters,
         *  the first byte of each encoding contains enough information to
         *  determine how long the encoding is (this 1st byte is also known as
         *  the lead byte).
         *
         *  @param leadByte         The first byte of a character encoding.
         *  @result                 The number of bytes in the encoding of a
         *                          single character starting with the specified
         *                          lead byte, or -1 if the specified byte is
         *                          not a valid lead byte for this character set.
         *  @since 1.0
         */
        virtual int                 getEncodingLengthForLeadByte(UInt8 leadByte) const = 0;

        /**
         *  Converts a sequence of bytes into a character.
         *
         *  One or more consecurive bytes from the specified buffer are analyzed
         *  and the corresponding ISO-10646 character is constructed.
         *
         *  @param buffer           The address of the 1st byte of character encoding.
         *  @param bytesAvailable   The number of available bytes. Note that the
         *                          actual character encoding may be found to constitute 
         *                          only a few leftmost bytes out of the specified
         *                          amount.
         *  @param c                Upon successful conversion, this parameter is 
         *                          assigned an ISO-10646 character that has just 
         *                          been decoded. If the conversion fails, this
         *                          parameter remains unchanged.
         *  @result                 The number of bytes that were decoded in
         *                          order to produce a single ISO-10646 character, or
         *                          -1 if the ISO-10646 character could not be produced.
         *  @since 1.0
         */
        virtual int                 decodeCharacter(const void * buffer, UIntSize bytesAvailable, Char & c) const = 0;

        /**
         *  Returns the default character for this character set.
         *
         *  A default character is a character that is used instead of
         *  all characters that cannot be encoded using this character
         *  set. By definition, the default character of a character
         *  set can always be encoded using the same character set.
         *
         *  @return                 The default character for this character set.
         *  @since 1.0
         */
        virtual Char                getDefaultCharacter() const = 0;

        /**
         *  Returns the signature of this character set.
         *
         *  A signature of a character set is a characteristic sequence of bytes
         *  that, when encountered in an input file, identifies that character
         *  set. Examples of signatures are {0xFE, 0xFF} and {0xFF, 0xFE}, which
         *  denote big- and little-endian UTF-16 character sets.
         *
         *  @return                 The array of bytes containing a signature
         *                          of this character set, or an empty array
         *                          if this character set has no associated
         *                          signature.
         *  @since 1.0
         */
        virtual Array<UInt8>        getSignature() const = 0;

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };

    /**
     *  @class CharacterSetFamily CARL.hpp
     *  A generic character set family.
     *
     *  Character sets are organized into groups called "families", with
     *  each character set family representing a closely related group of
     *  character sets (such as a vendor or a platform). Each character set
     *  always belongs to exactly one character set family.
     *
     *  Existing character set families can be extended with additional 
     *  member character sets; however, character sets cannot be reassigned
     *  from one character set family to another.
     *
     *  Currently, the following character set families are predefined by 
     *  CARL (the list contains character set and character set family
     *  names, with display names in brackets after the identifier):
     *
     *  -   ISO_8859 (ISO 8859 character sets)
     *      -   ISO_8859-1:1987 (ISO Latin 1)
     *      -   ISO_8859-2:1987 (ISO Latin 2)
     *      -   ISO_8859-3:1988 (ISO Latin 3)
     *      -   ISO_8859-4:1988 (ISO Latin 4)
     *      -   ISO_8859-5:1988 (ISO Latin Cyrillic)
     *      -   ISO_8859-6:1987 (ISO Latin Arabic)
     *      -   ISO_8859-7:1987 (ISO Latin Greek)
     *      -   ISO_8859-8:1988 (ISO Latin Hebrew)
     *      -   ISO_8859-9:1989 (ISO Latin 5)
     *      -   ISO_8859-10:1992 (ISO Latin 6)
     *      -   ISO_8859-11:1999 (ISO Thai)
     *      -   ISO_8859-13:1998 (ISO Latin 7)
     *      -   ISO_8859-14:1998 (ISO Latin 8)
     *      -   ISO_8859-15 (ISO Latin 9)
     *  -   ISO_10646 (ISO 10646 character sets)
     *      -   UTF-32BE (UTF-32 Big-Endian)
     *      -   UTF-32LE (UTF-32 Little-Endian)
     *      -   UTF-16BE (UTF-16 Big-Endian)
     *      -   UTF-16LE (UTF-16 Little-Endian)
     *      -   UTF-8 (UTF-8)
     *  -   Windows (MS Windows)
     *      -   Windows-1250 (Windows-1250 (Central Europe))
     *      -   Windows-1251 (Windows-1251 (Cyrillic))
     *      -   Windows-1252 (Windows-1252 (Latin1))
     *      -   Windows-1253 (Windows-1253 (Greek))
     *      -   Windows-1254 (Windows-1254 (Turkish))
     *      -   Windows-1255 (Windows-1255 (Hebrew))
     *      -   Windows-1256 (Windows-1256 (Arabic))
     *      -   Windows-1257 (Windows-1257 (Baltic))
     *      -   Windows-1258 (Windows-1258 (Viet Nam))
     *      -   cp874 (Cp874 (Thai))
     *      -   cp932 (Cp932 (Japanese Shift-JIS))
     *      -   cp936 (Cp936 (Simplified Chinese GBK))
     *      -   cp949 (Cp949 (Korean))
     *      -   cp950 (Cp950 (Traditional Chinese Big5))
     *  -   PC (IBM PC character sets)
     *      -   IBM437 (Cp437 (United States))
     *      -   IBM737 (Cp737 (Greek))
     *      -   IBM775 (Cp775 (Baltic))
     *      -   IBM850 (Cp850 (Multilingual Latin I))
     *      -   IBM852 (Cp852 (Latin II))
     *      -   IBM855 (Cp855 (Cyrillic))
     *      -   IBM856 (Cp855 (Hebrew))
     *      -   IBM857 (Cp857 (Turkish))
     *      -   IBM860 (Cp860 (Portuguese))
     *      -   IBM861 (Cp861 (Icelandic))
     *      -   IBM862 (Cp862 (Hebrew))
     *      -   IBM863 (Cp863 (Canadian French))
     *      -   IBM864 (Cp864 (Arabic))
     *      -   IBM865 (Cp865 (Nordic))
     *      -   IBM866 (Cp866 (Russian))
     *      -   IBM869 (Cp869 (Modern Greek))
     *  -   EBCDIC (EBCDIC character sets)
     *      -   IBM037 (Cp037 (IBM US & Canada))
     *      -   IBM424 (Cp424 (IBM Hebrew))
     *      -   IBM500 (Cp500 (IBM International))
     *      -   IBM875 (Cp500 (IBM Greek))
     *      -   IBM500 (Cp1026 (IBM Turkish))
     *  -   ADOBE (Adobe character sets)
     *      -   ADOBESTANDARD (Adobe Standard)
     *      -   ADOBESYMBOL (Adobe Symbol)
     *      -   ADOBEZAPFDINGBATS (Adobe Zapf Dingbats)
     *  -   KOI (KOI cyryllic character sets)
     *      -   GOST-19768-87
     *      -   KOI-7
     *      -   KOI-8A
     *      -   KOI-8B
     *      -   KOI-8E
     *      -   KOI-8F
     *      -   KOI-8R
     *      -   KOI-8RU
     *      -   KOI-8U
     *  -   MAC (Mac OS character sets)
     *      -   MacArabic (Mac OS Arabic)
     *      -   MacCeltic (Mac OS Celtic)
     *      -   MacCentEuro (Mac OS Central European)
     *      -   MacCroatian (Mac OS Croatian)
     *      -   MacCyrillic (Mac OS Cyrillic)
     *      -   MacDevanagari (Mac OS Devanagari)
     *      -   MacDingbats (Mac OS Dingbats)
     *      -   MacFarsi (Mac OS Farsi)
     *      -   MacGaelic (Mac OS Gaelic)
     *      -   MacGreek (Mac OS Greek)
     *      -   MacGujarati (Mac OS Gujarati)
     *      -   MacGurmukhi (Mac OS Gurmukhi)
     *      -   MacHebrew (Mac OS Hebrew)
     *      -   MacIcelandic (Mac OS Icelandic)
     *      -   MacInuit (Mac OS Inuit)
     *      -   MacRoman (Mac OS Roman)
     *      -   MacRomanian (Mac OS Romanian)
     *      -   MacSymbol (Mac OS Symbol)
     *      -   MacThai (Mac OS Thai)
     *      -   MacTurkish (Mac OS Turkish)
     *      -   MacUkrainian (Mac OS Ukrainian)
     *      -   MacChineseSimplified (Mac OS Chinese Simplified)
     *      -   MacChineseTraditional (Mac OS Chinese Traditional)
     *      -   MacKorean (Mac OS Korean)
     *      -   MacJapanese (Mac OS Japanese)
     *  -   MISCELLANEOUS (Miscellaneous character sets)
     *      -   ANSI_X3.4-1968 (US ASCII)
     *      -   NEXTSTEP (NextStep)
     *      -   ATARIST (Atari ST)
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC CharacterSetFamily
    {
        //////////
        //  Can't assign or copy construct
    private:
        CharacterSetFamily(const CharacterSetFamily &);
        void                        operator = (const CharacterSetFamily &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        CharacterSetFamily();
        virtual ~CharacterSetFamily();

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified character set family with the CARL framework.
         *
         *  A registration of a character set family informs the CARL framework that
         *  a specific character set family exists. From this point on, the application
         *  can use the newly registers character set family in exactly the same manner
         *  as a predefined character setfamily . Once registered, a character set family
         *  cannot be unregistered.
         *
         *  A character set family must be registered before any of its member
         *  character sets are registered.
         *
         *  A character set family cannot be registered if one of its names clashes
         *  with one of the names of some character set family already registered.
         *
         *  @param pCharacterSetFamily  The character set family to register.
         *  @exception InconsistentCharacterSetFamilyException If the character set 
         *                          family being registered is inconsistent (for 
         *                          example, its canonical name is not a member 
         *                          of its "all names" list).
         *  @exception DuplicateCharacterSetFamilyException If one of the character
         *                          set family names clashes with one of the names 
         *                          of another character set family registered earlier.
         *  @since 1.0
         */
        static void                 registerInstance(CharacterSetFamily * pCharacterSetFamily) throw (InconsistentCharacterSetFamilyException, DuplicateCharacterSetFamilyException);

        /**
         *  Returns a list of all registered character set families.
         *
         *  The returned list contains all character set families registered so far,
         *  including all predefined character set families. Elements in the list are
         *  not guaranteed to be in any specific order.
         *
         *  @return                 The list of all registered character set families.
         *  @since 1.0
         */
        static Array<CharacterSetFamily*>   getAllInstances();

        /**
         *  Finds and returns a character set family with the specified name.
         *
         *  Character set family names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the character set
         *                          family.
         *  @return                 The character set family found.
         *  @exception NoSuchElementError   If a character set family with the 
         *                          specified name does not exist.
         *  @since 1.0
         */
        static CharacterSetFamily * getInstance(const char * pName);

        /**
         *  Finds and returns a character set family with the specified name.
         *
         *  Character set family names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the character set
         *                          family.
         *  @return                 The character set family found.
         *  @exception NoSuchElementError   If a character set family with the 
         *                          specified name does not exist.
         *  @since 1.0
         */
        static CharacterSetFamily * getInstance(const Char * pName);

        /**
         *  Finds and returns a character set family with the specified name.
         *
         *  Character set family names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          character set family.
         *  @return                 The character set family found.
         *  @exception NoSuchElementError   If a character set family with the 
         *                          specified name does not exist.
         *  @since 1.0
         */
        static CharacterSetFamily * getInstance(const String & name);

        /**
         *  Returns a list of all members of this character set family.
         *
         *  Elements in the returned list are not guaranteed to be in
         *  any particular order.
         *
         *  @return                 A list of all members of this character set family.
         *  @since 1.0
         */
        Array<CharacterSet*>        getAllMembers() const;

        /**
         *  Finds and returns a member character set with the specified name.
         *
         *  Character set names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the character set.
         *  @return                 The character set found.
         *  @exception NoSuchElementError   If a character set with the specified
         *                          name does not exist or is not a member of
         *                          this character set family.
         *  @since 1.0
         */
        CharacterSet *              getMember(const char * pName) const;

        /**
         *  Finds and returns a member character set with the specified name.
         *
         *  Character set names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the character set.
         *  @return                 The character set found.
         *  @exception NoSuchElementError   If a character set with the specified
         *                          name does not exist or is not a member of
         *                          this character set family.
         *  @since 1.0
         */
        CharacterSet *              getMember(const Char * pName) const;

        /**
         *  Finds and returns a member character set with the specified name.
         *
         *  Character set names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          character set.
         *  @return                 The character set found.
         *  @exception NoSuchElementError   If a character set with the specified
         *                          name does not exist or is not a member of
         *                          this character set family.
         *  @since 1.0
         */
        CharacterSet *              getMember(const String & name) const;

        /**
         *  Returns a list of all names of this character set family.
         *
         *  Elements in the list are not guaranteed to be in any particular order.
         *
         *  @return                 A list of all names of this character set family.
         *  @since 1.0
         */
        virtual Array<String>       getAllNames() const = 0;

        /**
         *  Returns the canonical name of this character set family.
         *
         *  A canonical name of a character set family is a name under which the
         *  character set family is known to standards committees.
         *
         *  @return                 The canonical name of this character set family.
         *  @since 1.0
         */
        virtual String              getCanonicalName() const = 0;

        /**
         *  Returns the preferred name of this character set family.
         *
         *  A preferred name of a character set family is a name which is
         *  historically most frequently used to refer to that
         *  character set family (in MIME headers, in HTML headers, etc.)
         *
         *  @return                 The preferred name of this character set family.
         *  @since 1.0
         */
        virtual String              getPreferredName() const = 0;

        /**
         *  Returns the user-readable display name of this character set
         *  family for the current default locale.
         *
         *  @return                 The display name of this character set family.
         *  @since 1.0
         */
        virtual String              getDisplayName() const = 0;

        /**
         *  Returns the user-readable display name of this character set
         *  family for the specified default locale.
         *
         *  @param forLocale        The locale for which the display name of
         *                          the character set family shall be tailored.
         *  @return                 The display name of this character set family.
         *  @since 1.0
         */
        virtual String              getDisplayName(const Locale & forLocale) const = 0;

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };
}

#endif  //  __CARL_Charset_hpp_included__
//  End of CARL_Charset.hpp
