//
//  CARL_Collator.hpp - String collation services.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL string collation framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Collator.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Collator_hpp_included__
#define __CARL_Collator_hpp_included__

namespace carl
{
    /**
     *  @class Collator CARL.hpp
     *  A generic collator.
     *
     *  A collator defines a set of rules for sorting natural language strings
     *  for a given locale.
     *
     *  When obtaining a collator for a specified locale, it may well
     *  be the case that there is no collator registered for that locale.
     *  In this case the immediate parent locale is tried next, and the parent of 
     *  that if necessary, until either an invariant locale is reached or a
     *  collator is found. In the former case, a default collator is
     *  used. This behaviour allows specifying locale-generic collators
     *  (e.g. if an "German neutral" collator is specified, then it will
     *  be used for both "German/Germany" and "German/Switzerland" locales provided
     *  no separate collators are registered for either of those).
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Collator
    {
        //////////
        //  Can't assign or copy construct
    private:
        Collator(const Collator &);
        void                    operator = (const Collator &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        Collator() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~Collator() {}

        //////////
        //  Constants
    public:
        /**
         *  Specifies the collation strength.
         *
         *  The collation strength determines the level of difference considered
         *  significant in comparisons
         *
         *  @since 1.0
         */
        enum CollationStrength
        {
            /**
             *  Only primary differences are important during collation.
             *
             *  A primary differfence is that between different letters (e.g.
             *  a difference between 'a' and 'b').
             *
             *  @hideinitializer
             *  @since 1.0
             */
            CollationStrength_Primary   = 1,

            /**
             *  Only primary and secondary differences are important during collation.
             *
             *  A secondary differfence is that between same letters with different
             *  accents.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            CollationStrength_Secondary = 2,

            /**
             *  Primary, secondary and tertiary differences are important during collation.
             *
             *  A tertiary differfence is that between same letters with different
             *  case (e.g. a difference between 'a' and 'A').
             *
             *  @hideinitializer
             *  @since 1.0
             */
            CollationStrength_Tertiary  = 3,

            /**
             *  All differences are important during collation.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            CollationStrength_Identical = 4
        };

        /**
         *  Specifies how Unicode composed characters are handled.
         *
         *  @since 1.0
         */
        enum Decomposition
        {
            /**
             *  Accented characters are not decomposed for collation.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Decomposition_None          = 1,

            /**
             *  Characters that are canonical variants according to Unicode standard 
             *  will be decomposed for collation.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Decomposition_Canonical     = 2,

            /**
             *  Both Unicode canonical variants and Unicode compatibility variants 
             *  will be decomposed for collation.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Decomposition_Full          = 3
        };

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified collator with the CARL framework.
         *
         *  A registration of a collator informs the CARL framework that
         *  a specific collator exists. From this point on, the application
         *  can use the newly registers collator in exactly the same manner
         *  as a predefined collator. Once registered, a collator
         *  cannot be unregistered.
         *
         *  A collator is always registered for a specific locale. Any attempt
         *  to get a collator instance for that locale will always yield the
         *  collator registered for that locale. It is possible to register
         *  the same collator for more than one locale.
         *
         *  @param forLocale        The locale for which the newly registered
         *                          collator shall be used.
         *  @param pCollator        The collator to register.
         *  @since 1.0
         */
        static void                 registerInstance(const Locale & forLocale, Collator * pCollator);

        /**
         *  Finds and returns a collator for the current default locale.
         *
         *  @return                 The collator for the current default locale.
         *  @since 1.0
         */
        static Collator *           getInstance();

        /**
         *  Finds and returns a collator for the specified locale.
         *
         *  @param forLocale        The locale for which a collator is sought.
         *  @return                 The collator for the current default locale.
         *  @since 1.0
         */
        static Collator *           getInstance(const Locale & forLocale);

        /**
         *  Compares two strings accirding to their collation sequence.
         *
         *  @param s1               The 1st string to compare,
         *  @param s2               The 2nd string to compare,
         *  @param collationStrength    The collation strength specifies whether
         *                          primary, secondary and/or tertiary differences
         *                          are considered important or not.
         *  @param decomposition    The decomposition mode specifies whether composite
         *                          characters are collated correctly and, if yes,
         *                          whether non-canonical decompositions are allowed.
         *  @return                 
         *                      -   0 if the two strings are equal according to the 
         *                          collation sequence.
         *                      -   <0 if \a s1 comes before \a s2 in the collation sequence.
         *                      -   >0 if \a s1 comes after \a s2 in the collation sequence.
         *  @since 1.0
         */
        virtual int                 collate(const String & s1, const String & s2, CollationStrength collationStrength = CollationStrength_Primary, Decomposition decomposition = Decomposition_Full) = 0;

        /**
         *  Performs a decomposition of a string as required by the specified
         *  decomposition mode.
         *
         *  @param s            The string to perform the decomposition on..
         *  @param decomposition    The decomposition mode.
         *  @return             The decomposition result.
         *  @since 1.0
         */
        static String               performDecomposition(const String & s, Decomposition decomposition);

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };

    /**
     *  The string comparator that compares string according to a collation sequence.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC CollatorBasedStringComparator : public IOrderedComparator<String>
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  @param pCollator        The collator to use.
         *  @param collationStrength    The collation strength to use for string collation.
         *  @param decomposition        The decomposition mode to use for string collation.
         *  @since 1.0
         */
        CollatorBasedStringComparator(Collator * pCollator, Collator::CollationStrength collationStrength, Collator::Decomposition decomposition);

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        CollatorBasedStringComparator(const CollatorBasedStringComparator & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~CollatorBasedStringComparator();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        CollatorBasedStringComparator & operator = (const CollatorBasedStringComparator & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if the two values are equal.
         *
         *  @param value1           The 1st value to compare for equality.
         *  @param value2           The 2nd value to compare for equality.
         *  @return                 True if the two specified values are equal
         *                          under the equality rules defined by this
         *                          comparator, false otherwise.
         *  @since 1.0
         */
        virtual bool                equal(const String & value1, const String & value2) const;

        /**
         *  Checks if one value is less than another.
         *
         *  @param value1           The 1st value to compare for order.
         *  @param value2           The 2nd value to compare for order.
         *  @return                 True if the value1 is "less than" value2
         *                          under the order rules defined by this
         *                          comparator, false otherwise.
         *  @since 1.0
         */
        virtual bool                less(const String & value1, const String & value2) const;

        /**
         *  Checks if one value is greater than another.
         *
         *  @param value1           The 1st value to compare for order.
         *  @param value2           The 2nd value to compare for order.
         *  @return                 True if the value1 is "greater than" value2
         *                          under the order rules defined by this
         *                          comparator, false otherwise.
         *  @since 1.0
         */
        virtual bool                greater(const String & value1, const String & value2) const;

        //////////
        //  Implementation
    private:
        Collator *                  _pCollator;
        Collator::CollationStrength _collationStrength;
        Collator::Decomposition     _decomposition;
    };

    /**
     *  The sort order that arranges strings according to their collation sequence.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC ForwardCollatorBasedStringSortOrder : public ISortOrder<String>
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  @param pCollator        The collator to use.
         *  @param collationStrength    The collation strength to use for string collation.
         *  @param decomposition        The decomposition mode to use for string collation.
         *  @since 1.0
         */
        ForwardCollatorBasedStringSortOrder(Collator * pCollator, Collator::CollationStrength collationStrength, Collator::Decomposition decomposition);

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        ForwardCollatorBasedStringSortOrder(const ForwardCollatorBasedStringSortOrder & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~ForwardCollatorBasedStringSortOrder();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        ForwardCollatorBasedStringSortOrder & operator = (const ForwardCollatorBasedStringSortOrder & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a value shall appear before another value in a sorted sequence.
         *
         *  @param value1           The 1st value to compare.
         *  @param value2           The 2nd value to compare.
         *  @return                 True if \a value1 shall appear before \a value2
         *                          in the sorted sequence; false otherwise.
         *  @since 1.0
         */
        virtual bool                shallComeBefore(const String & value1, const String & value2) const;

        //////////
        //  Implementation
    private:
        Collator *                  _pCollator;
        Collator::CollationStrength _collationStrength;
        Collator::Decomposition     _decomposition;
    };

    /**
     *  The sort order that arranges strings according to their reverses collation sequence.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC ReversedCollatorBasedStringSortOrder : public ISortOrder<String>
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  @param pCollator        The collator to use.
         *  @param collationStrength    The collation strength to use for string collation.
         *  @param decomposition        The decomposition mode to use for string collation.
         *  @since 1.0
         */
        ReversedCollatorBasedStringSortOrder(Collator * pCollator, Collator::CollationStrength collationStrength, Collator::Decomposition decomposition);

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        ReversedCollatorBasedStringSortOrder(const ReversedCollatorBasedStringSortOrder & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~ReversedCollatorBasedStringSortOrder();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        ReversedCollatorBasedStringSortOrder & operator = (const ReversedCollatorBasedStringSortOrder & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a value shall appear before another value in a sorted sequence.
         *
         *  @param value1           The 1st value to compare.
         *  @param value2           The 2nd value to compare.
         *  @return                 True if \a value1 shall appear before \a value2
         *                          in the sorted sequence; false otherwise.
         *  @since 1.0
         */
        virtual bool                shallComeBefore(const String & value1, const String & value2) const;

        //////////
        //  Implementation
    private:
        Collator *                  _pCollator;
        Collator::CollationStrength _collationStrength;
        Collator::Decomposition     _decomposition;
    };
}

#endif  //  __CARL_Collator_hpp_included__
//  End of CARL_Collator.hpp
