//
//  CARL_Comparators.hpp - Cmparators.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/07/29 11:19:44 $
//  $Revision: 1.10 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL value comparision interfaces.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Comparators.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Comparators_hpp_included__
#define __CARL_Comparators_hpp_included__

namespace carl
{
    /**
     *  @interface IComparator CARL.hpp
     *  An abstract interface to a comparator.
     *
     *  A comparator is an object that establishes an equivalence relation
     *  over a specific type.
     *
     *  @param T                The type of compared values.
     *  @since 1.0
     */
    template <class T>
    class IComparator
    {
        //////////
        //  This is an interface
    public:
        IComparator() {}
        virtual ~IComparator() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if the two values are equal.
         *
         *  @param value1           The 1st value to compare for equality.
         *  @param value2           The 2nd value to compare for equality.
         *  @return                 True if the two specified values are equal
         *                          under the equality rules defined by this
         *                          comparator, false otherwise.
         *  @since 1.0
         */
        virtual bool                equal(const T & value1, const T & value2) const = 0;
    };

    /**
     *  @class NaturalComparator CARL.hpp
     *  A comparator that uses the operator '==' to determine whether two values
     *  are equal or not.
     *
     *  @param T                The type of compared values.
     *  @since 1.0
     */
    template <class T>
    class NaturalComparator : public virtual IComparator<T>
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        NaturalComparator();

        /**
         *  The copy constructor.
         *
         *  @since 1.0
         */
        NaturalComparator(const NaturalComparator<T> &);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~NaturalComparator();

        /**
         *  The assignment operator.
         *
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        NaturalComparator<T> &      operator = (const NaturalComparator<T> &);

        //////////
        //  Operations
    public:
        /**
         *  Returns true if value1 == value2, false otherwise.
         *
         *  @param value1           The 1st value to compare for equality.
         *  @param value2           The 2nd value to compare for equality.
         *  @return                 True if value1 == value2, false otherwise.
         *  @since 1.0
         */
        virtual bool                equal(const T & value1, const T & value2) const;
    };

    template <class T>
    NaturalComparator<T>::NaturalComparator()
    {
    }

    template <class T>
    NaturalComparator<T>::NaturalComparator(const NaturalComparator<T> &)
    {
    }

    template <class T>
    NaturalComparator<T>::~NaturalComparator()
    {
    }

    template <class T>
    NaturalComparator<T> & NaturalComparator<T>::operator = (const NaturalComparator<T> &)
    {
        return *this;
    }

    template <class T>
    bool NaturalComparator<T>::equal(const T & value1, const T & value2) const
    {
        return value1 == value2;
    }

    /**
     *  @interface IOrderedComparator CARL.hpp
     *  An abstract interface to ab ordered comparator.
     *
     *  An "ordered comparator" is an object that establishes equivalence and
     *  weak partial order relations over a specific type.
     *
     *  @param T                The type of compared values.
     *  @since 1.0
     */
    template <class T>
    class IOrderedComparator : public virtual IComparator<T>
    {
        //////////
        //  This is an interface
    public:
        IOrderedComparator() {}
        virtual ~IOrderedComparator() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if the two values are equal.
         *
         *  @param value1           The 1st value to compare for equality.
         *  @param value2           The 2nd value to compare for equality.
         *  @return                 True if the two specified values are equal
         *                          under the equality rules defined by this
         *                          comparator, false otherwise.
         *  @since 1.0
         */
        virtual bool                equal(const T & value1, const T & value2) const = 0;

        /**
         *  Checks if one value is less than another.
         *
         *  @param value1           The 1st value to compare for order.
         *  @param value2           The 2nd value to compare for order.
         *  @return                 True if the value1 is "less than" value2
         *                          under the order rules defined by this
         *                          comparator, false otherwise.
         *  @since 1.0
         */
        virtual bool                less(const T & value1, const T & value2) const = 0;

        /**
         *  Checks if one value is greater than another.
         *
         *  @param value1           The 1st value to compare for order.
         *  @param value2           The 2nd value to compare for order.
         *  @return                 True if the value1 is "greater than" value2
         *                          under the order rules defined by this
         *                          comparator, false otherwise.
         *  @since 1.0
         */
        virtual bool                greater(const T & value1, const T & value2) const = 0;
    };

    /**
     *  @class NaturalOrderedComparator CARL.hpp
     *
     *  A comparator that uses the operators '==', '<' and '>' to determine whether 
     *  two values are equal, or else which one is greater.
     *
     *  @param T                The type of compared values.
     *  @since 1.0
     */
    template <class T>
    class NaturalOrderedComparator : public virtual IOrderedComparator<T>
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  @since 1.0
         */
        NaturalOrderedComparator();

        /**
         *  The copy constructor.
         *
         *  @since 1.0
         */
        NaturalOrderedComparator(const NaturalOrderedComparator<T> &);

        /**
         *  The class destructor
         *
         *  @since 1.0
         */
        virtual ~NaturalOrderedComparator();

        /**
         *  The assignment operator.
         *
         *  @return                 A reference to the assigned-to object.
         *
         *  @since 1.0
         */
        NaturalOrderedComparator<T>&operator = (const NaturalOrderedComparator<T> &);

        //////////
        //  Operations
    public:
        /**
         *  Checks if the two values are equal.
         *
         *  @param value1           The 1st value to compare for equality.
         *  @param value2           The 2nd value to compare for equality.
         *  @return                 True if value1 == value2, false otherwise.
         *  @since 1.0
         */
        virtual bool                equal(const T & value1, const T & value2) const;

        /**
         *  Checks if one value is less than another.
         *
         *  @param value1           The 1st value to compare for order.
         *  @param value2           The 2nd value to compare for order.
         *  @return                 True if value1 < value2, false otherwise.
         *  @since 1.0
         */
        virtual bool                less(const T & value1, const T & value2) const;

        /**
         *  Checks if one value is greater than another.
         *
         *  @param value1           The 1st value to compare for order.
         *  @param value2           The 2nd value to compare for order.
         *  @return                 True if value1 > value2, false otherwise.
         *  @since 1.0
         */
        virtual bool                greater(const T & value1, const T & value2) const;
    };

    template <class T>
    NaturalOrderedComparator<T>::NaturalOrderedComparator()
    {
    }

    template <class T>
    NaturalOrderedComparator<T>::NaturalOrderedComparator(const NaturalOrderedComparator<T> &)
    {
    }

    template <class T>
    NaturalOrderedComparator<T>::~NaturalOrderedComparator()
    {
    }

    template <class T>
    NaturalOrderedComparator<T> & NaturalOrderedComparator<T>::operator = (const NaturalOrderedComparator<T> &)
    {
        return *this;
    }

    template <class T>
    bool NaturalOrderedComparator<T>::equal(const T & value1, const T & value2) const
    {
        return value1 == value2;
    }

    template <class T>
    bool NaturalOrderedComparator<T>::less(const T & value1, const T & value2) const
    {
        return value1 < value2;
    }

    template <class T>
    bool NaturalOrderedComparator<T>::greater(const T & value1, const T & value2) const
    {
        return value1 > value2;
    }

    /**
     *  @class CaseInsensitiveStringComparator CARL.hpp
     *  A comparator that compares strings ignoring case. Main use is for
     *  constructing hash sets & hash maps where case-insensitive strings
     *  are keys.
     *
     *  @since 1.0
     */
    class CaseInsensitiveStringComparator : public virtual IComparator<String>
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        CaseInsensitiveStringComparator() {}

        /**
         *  The copy constructor.
         *
         *  @since 1.0
         */
        CaseInsensitiveStringComparator(const CaseInsensitiveStringComparator &) {}

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~CaseInsensitiveStringComparator() {}

        /**
         *  The assignment operator.
         *
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        CaseInsensitiveStringComparator &   operator = (const CaseInsensitiveStringComparator &) { return *this; }

        //////////
        //  Operations
    public:
        /**
         *  Returns true if value1 == value2, false otherwise.
         *
         *  @param value1           The 1st value to compare for equality.
         *  @param value2           The 2nd value to compare for equality.
         *  @return                 True if value1 == value2, false otherwise.
         *  @since 1.0
         */
        virtual bool                equal(const String & value1, const String & value2) const { return value1.compareIgnoreCase(value2) == 0; }
    };
}

#endif  //  __CARL_Comparators_hpp_included__
//  End of CARL_Comparators.hpp
