//
//  CARL_Compression.hpp - Compression services.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.18 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL compression framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Compression.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Compression_hpp_included__
#define __CARL_Compression_hpp_included__

namespace carl
{
    class CARL_PUBLIC CompressionAlgorithm;
    class CARL_PUBLIC Compressor;
    class CARL_PUBLIC Decompressor;

    /**
     *  @class InconsistentCompressionAlgorithmException CARL.hpp
     *  An exception thrown when an attempt is made to register an inconsistent 
     *  compression algorithm.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC InconsistentCompressionAlgorithmException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        InconsistentCompressionAlgorithmException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        InconsistentCompressionAlgorithmException(const InconsistentCompressionAlgorithmException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~InconsistentCompressionAlgorithmException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        InconsistentCompressionAlgorithmException & operator = (const InconsistentCompressionAlgorithmException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class DuplicateCompressionAlgorithmException CARL.hpp
     *  An exception thrown when an attempt is made to register a duplicate 
     *  compression algorithm.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DuplicateCompressionAlgorithmException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        DuplicateCompressionAlgorithmException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        DuplicateCompressionAlgorithmException(const DuplicateCompressionAlgorithmException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~DuplicateCompressionAlgorithmException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        DuplicateCompressionAlgorithmException &    operator = (const DuplicateCompressionAlgorithmException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class CompressedDataCorruptException CARL.hpp
     *  An exception thrown when compressed data is corrupt and cannot be decompressed.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC CompressedDataCorruptException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        CompressedDataCorruptException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        CompressedDataCorruptException(const CompressedDataCorruptException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~CompressedDataCorruptException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        CompressedDataCorruptException &    operator = (const CompressedDataCorruptException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class CompressionAlgorithm CARL.hpp
     *  An abstract compression algorithm.
     *
     *  A compression algorithm specifies how a block of data can be compressed
     *  in order to reduce its size and later decompressed to retrieve the
     *  original data. Compression algorithms are uniquely identified by
     *  their case-insensitive names.
     *
     *  Currently, the following compression algorithms are predefined by 
     *  CARL (the list contains compression algorithm names, with display 
     *  names in brackets after the identifier):
     *
     *  -   NULL (No compression)
     *  -   RunLength (Run-length compression)
     *  -   LZW (LZW-based dictionary compression)
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC CompressionAlgorithm
    {
        //////////
        //  Can't assign or copy construct
    private:
        CompressionAlgorithm(const CompressionAlgorithm &);
        void                        operator = (const CompressionAlgorithm &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        CompressionAlgorithm() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~CompressionAlgorithm() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified compression algorithm with the CARL framework.
         *
         *  A registration of a compression algorithm informs the CARL framework that
         *  a specific compression algorithm exists. From this point on, the application
         *  can use the newly registers compression algorithm in exactly the same manner
         *  as a predefined compression algorithm. Once registered, a compression algorithm
         *  cannot be unregistered.
         *
         *  A compression algorithm cannot be registered if one of its names clashes
         *  with one of the names of some compression algorithm already registered.
         *
         *  @param pCompressionAlgorithm    The compression algorithm to register.
         *  @exception InconsistentCompressionAlgorithmException    If the compression 
         *                          algorithm being registered is inconsistent (for example, 
         *                          its canonical name is not a member of its "all names" 
         *                          list).
         *  @exception DuplicateCompressionAlgorithmException   If one of the compression 
         *                          algorithm names clashes with one of the names of another
         *                          compression algorithm registered earlier.
         *  @since 1.0
         */
        static void                         registerInstance(CompressionAlgorithm * pCompressionAlgorithm) throw (InconsistentCompressionAlgorithmException, DuplicateCompressionAlgorithmException);

        /**
         *  Returns a list of all registered compression algorithms.
         *
         *  The returned list contains all compression algorithms registered so far,
         *  including all predefined compression algorithms. Elements in the list are
         *  not guaranteed to be in any specific order.
         *
         *  @return                 The list of all registered compression algorithms.
         *  @since 1.0
         */
        static Array<CompressionAlgorithm*> getAllInstances();

        /**
         *  Finds and returns a compression algorithm with the specified name.
         *
         *  Compression algorithm names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the compression algorithm.
         *  @return                 The compression algorithm found.
         *  @exception NoSuchElementError   If a compression algorithm with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static CompressionAlgorithm *       getInstance(const char * pName);

        /**
         *  Finds and returns a compression algorithm with the specified name.
         *
         *  Compression algorithm names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ISO-10646 string
         *                          representing the name of the compression algorithm.
         *  @return                 The compression algorithm found.
         *  @exception NoSuchElementError   If a compression algorithm with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static CompressionAlgorithm *       getInstance(const Char * pName);

        /**
         *  Finds and returns a compression algorithm with the specified name.
         *
         *  Compression algorithm names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          compression algorithm.
         *  @return                 The compression algorithm found.
         *  @exception NoSuchElementError   If a compression algorithm with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static CompressionAlgorithm *       getInstance(const String & name);

        /**
         *  Returns a list of all names of this compression algorithm.
         *
         *  Elements in the list are not guaranteed to be in any particular order.
         *
         *  @return                 A list of all names of this compression algorithm.
         *  @since 1.0
         */
        virtual Array<String>               getAllNames() const = 0;

        /**
         *  Returns the canonical name of this compression algorithm.
         *
         *  A canonical name of a compression algorithm is a name under which the
         *  compression algorithm is known to standards committees.
         *
         *  @return                 The canonical name of this compression algorithm.
         *  @since 1.0
         */
        virtual String                      getCanonicalName() const = 0;

        /**
         *  Returns the preferred name of this compression algorithm.
         *
         *  A preferred name of a compression algorithm is a name which is
         *  historically most frequently used to refer to that
         *  compression algorithm (in MIME headers, in HTML headers, etc.)
         *
         *  @return                 The preferred name of this compression algorithm.
         *  @since 1.0
         */
        virtual String                      getPreferredName() const = 0;

        /**
         *  Returns the user-readable display name of this compression algorithm
         *  for the current default locale.
         *
         *  @return                 The display name of this compression algorithm.
         *  @since 1.0
         */
        virtual String                      getDisplayName() const = 0;

        /**
         *  Returns the user-readable display name of this compression algorithm
         *  for the specified default locale.
         *
         *  @param forLocale        The locale for which the display name of
         *                          the compression algorithm shall be tailored.
         *  @return                 The display name of this compression algorithm.
         *  @since 1.0
         */
        virtual String                      getDisplayName(const Locale & forLocale) const = 0;

        /**
         *  Returns the minimum size of the block that can be compressed by this algorithm.
         *
         *  @return                 The minimum size of the block that can be compressed.
         *  @since 1.0
         */
        virtual UIntSize                    getMinimumBlockSize() const = 0;

        /**
         *  Returns the preferred size of the block that can be compressed by this algorithm.
         *  Compressing blocks of the preferred size offers the best balance of
         *  performance and compression ratio.
         *
         *  @return                 The preferred of the size block that can be compressed.
         *  @since 1.0
         */
        virtual UIntSize                    getPreferredBlockSize() const = 0;

        /**
         *  Returns the maximum size of the block that can be compressed by this algorithm.
         *
         *  @return                 The maximum size of the block that can be compressed.
         *  @since 1.0
         */
        virtual UIntSize                    getMaximumBlockSize() const = 0;

        /**
         *  Returns a compressor for this compression algorithm.
         *
         *  A compressor is a part of compression algorithm that knows how
         *  to compress data. Compressors are stock objects; client shall
         *  not delete them when they are no longer needed.
         *
         *  @return                 A pointer to a stock compressor object.
         *  @since 1.0
         */
        virtual Compressor *                getCompressor() const = 0;

        /**
         *  Returns a decompressor for this compression algorithm.
         *
         *  A decompressor is a part of compression algorithm that knows how
         *  to decompress previously compressed data. Decompressors are stock 
         *  objects; client shall not delete them when they are no longer needed.
         *
         *  @return                 A pointer to a stock decompressor object.
         *  @since 1.0
         */
        virtual Decompressor *              getDecompressor() const = 0;

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };

    /**
     *  @class Compressor CARL.hpp
     *  An abstract compressor.
     *
     *  A compressor is a part of compression algorithm that knows how
     *  to compress data. Compressors are stock objects; client shall
     *  not delete them when they are no longer needed.
     *  
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Compressor
    {
        //////////
        //  Can't assign or copy construct
    private:
        Compressor(const Compressor &);
        void                        operator = (const Compressor &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        Compressor() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~Compressor() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns the compression algorithm to which this compressor belongs.
         *
         *  @return                 The compression algorithm to which this compressor belongs.
         *  @since 1.0
         */
        virtual CompressionAlgorithm *      getCompressionAlgorithm() = 0;

        /**
         *  Returns the required output buffer size for given input.
         *
         *  The size of the input must be within the limits imposed by the
         *  compression algorithm.
         *
         *  @param pUncompressedData    The pointer to the first byte of the data
         *                          block that is about to be compressed. The data
         *                          block must be at least \a uncompressedDataSize
         *                          bytes long; if it is longer then only the first
         *                          \a uncompressedDataSize bytes are taken into
         *                          consideration.
         *  @param uncompressedDataSize The size of the data block about to be
         *                          compressed.
         *  @return                 The number of bytes required to store the compressed
         *                          form of the specified data block.
         *  @since 1.0
         */
        virtual UIntSize                    getRequiredOutputBufferSize(const void * pUncompressedData, UIntSize uncompressedDataSize) = 0;

        /**
         *  Comrpesses the block of data into the specified output buffer.
         *
         *  The size of the input must be within the limits imposed by the
         *  compression algorithm.
         *
         *  @param pUncompressedData    The pointer to the first byte of the data
         *                          block to compress. The data block must be at 
         *                          least \a uncompressedDataSize bytes long; if 
         *                          it is longer then only the first \a 
         *                          uncompressedDataSize bytes are taken into
         *                          consideration.
         *  @param uncompressedDataSize The size of the data block to compress.
         *  @param pCompressedData  The buffer where the compressed data is placed.
         *                          The buffer is assumed to be large enough to
         *                          hold the entire compressed block.
         *  @return                 The number of bytes stored into the output buffer.
         *  @since 1.0
         */
        virtual UIntSize                    compress(const void * pUncompressedData, UIntSize uncompressedDataSize, void * pCompressedData) = 0;
    };

    /**
     *  @class Decompressor CARL.hpp
     *  An abstract decompressor.
     *
     *  A decompressor is a part of compression algorithm that knows how
     *  to decompress previously compressed data. Decompressors are stock 
     *  objects; client shall not delete them when they are no longer needed.
     *  
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Decompressor
    {
        //////////
        //  Can't assign or copy construct
    private:
        Decompressor(const Decompressor &);
        void                        operator = (const Decompressor &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        Decompressor() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~Decompressor() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns the compression algorithm to which this decompressor belongs.
         *
         *  @return                 The compression algorithm to which this decompressor belongs.
         *  @since 1.0
         */
        virtual CompressionAlgorithm *      getCompressionAlgorithm() = 0;

        /**
         *  Returns the required output buffer size for given input.
         *
         *  @param pCompressedData  The pointer to the first byte of the data
         *                          block that is about to be decompressed. The data
         *                          block must be at least \a compressedDataSize
         *                          bytes long; if it is longer then only the first
         *                          \a compressedDataSize bytes are taken into
         *                          consideration.
         *  @param compressedDataSize   The size of the data block about to be
         *                          decompressed.
         *  @return                 The number of bytes required to store the decompressed
         *                          form of the specified data block.
         *  @exception CompressedDataCorruptException   Thrown if the specified compressed
         *                          data is not recognized by the decompressor.
         *  @since 1.0
         */
        virtual UIntSize                    getRequiredOutputBufferSize(const void * pCompressedData, UIntSize compressedDataSize) throw(CompressedDataCorruptException) = 0;

        /**
         *  Decomrpesses the block of data into the specified output buffer.
         *
         *  @param pCompressedData  The pointer to the first byte of the data
         *                          block to decompress. The data block must be 
         *                          at least \a compressedDataSize bytes long; 
         *                          if it is longer then only the first 
         *                          \a compressedDataSize bytes are taken into
         *                          consideration.
         *  @param compressedDataSize   The size of the data block about to  decompress.
         *  @param pUncompressedData    The buffer where the decompressed data is placed.
         *                          The buffer is assumed to be large enough to
         *                          hold the entire decompressed block.
         *  @return                 The number of bytes stored into the output buffer.
         *  @exception CompressedDataCorruptException   Thrown if the specified compressed
         *                          data is not recognized by the decompressor.
         *  @since 1.0
         */
        virtual UIntSize                    decompress(const void * pCompressedData, UIntSize compressedDataSize, void * pUncompressedData) throw(CompressedDataCorruptException) = 0;
    };

    /**
     *  @class CompressOutputStream CARL.hpp
     *  An output stream that performs on-the-fly compression.
     *
     *  A compress output stream does its work in collaboration with two
     *  other objects: a compressor that performs the actual data compression
     *  and a target stream, where all compressed data is forwarded.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DecompressInputStream;
    class CARL_PUBLIC CompressOutputStream : public IByteOutput
    {
        friend class DecompressInputStream;
        //////////
        //  Can't assign or copy construct
    private:
        CompressOutputStream(const CompressOutputStream &);
        void                    operator = (const CompressOutputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes the newly created output stream by associating it with the
         *  specified target stream and compressor. The newly constructed stream
         *  is initially open.
         *
         *  @param pTarget          The stream where compressed data are forwarded.
         *  @param pCompressor      The compressor to use for data compression.
         *  @since 1.0
         */
        CompressOutputStream(IByteOutput * pTarget, Compressor * pCompressor);

        /**
         *  The class destructor.
         *
         *  Closes the output stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~CompressOutputStream() { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this output stream is still open.
         *
         *  @return                 True if this output stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _pBuffer != NULL; }

        /**
         *  Closes this output stream.
         *
         *  Any data written to the output stream but not yet compressed is
         *  compressed and forwarded to the associated target stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteOutput
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  If there is any data that has been written to this output stream
         *  and not yet compressed, this data is compressed immediately and
         *  forwarded to the associated target stream, then the target
         *  stream is flushed as well.
         *
         *  This method can only be called on open streams.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException);

        /**
         *  Writes a single byte to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param byte             The byte to write to this stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(UInt8 byte) throw(IOException) { write(&byte, 1); }

        /**
         *  Writes a sequence of bytes to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The address of the first byte of the sequence
         *                          that shall be written to the stream. The
         *                          sequence must be at least \a bytesToWrite
         *                          bytes long; if it is longer then only the
         *                          forst \a bytesToWrite bytes are written.
         *  @param bytesToWrite     The number of bytes to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException);

        //////////
        //  Implementation
    private:
        static const UIntSize   _MaximumBufferSize = 32768;
        IByteOutput *           _pTarget;
        UIntSize                _bufferSize;
        Compressor *            _pCompressor;
        UInt8 *                 _pBuffer;
        UIntSize                _bytesInBuffer;
        Array<UInt8>            _compressedData;

        //  Block types
        static const UInt8      _BlockType_Uncompressed = 1;
        static const UInt8      _BlockType_CompressedWithoutPadding = 2;
        static const UInt8      _BlockType_CompressedWithPadding = 3;

        //  Helpers
        void                    _flushBuffer();
    };

    /**
     *  @class DecompressInputStream CARL.hpp
     *  An input stream that performs on-the-fly decompression.
     *
     *  A decompress input stream does its work in collaboration with two
     *  other objects: a source stream, where compressed data is obtained
     *  from and a decompressor that performs the actual data decompression.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DecompressInputStream : public IByteInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        DecompressInputStream(const DecompressInputStream &);
        void                    operator = (const DecompressInputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes the newly created input stream by associating it with the
         *  specified source stream and decompressor. The newly constructed stream
         *  is initially open.
         *
         *  @param pSource          The stream where compressed data is obtained from.
         *  @param pDecompressor    The decompressor to use for data decompression.
         *  @since 1.0
         */
        DecompressInputStream(IByteInput * pSource, Decompressor * pDecompressor);

        /**
         *  The class destructor.
         *
         *  Closes the input stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~DecompressInputStream();

        //////////
        //  Operations
    public:
        /**
         *  Checks if this input stream is still open.
         *
         *  @return                 True if this input stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _pBuffer != NULL; }

        /**
         *  Closes this input stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteInput
    public:
        /**
         *  Returns the number of bytes that can be read from this input 
         *  stream without blocking.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The number of bytes that can be read from this
         *                          input stream without blocking.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        available() throw(IOException);

        /**
         *  Reads and returns the next byte or -1 if an end of stream is reached.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The next input byte, or -1 if an end of stream is reached.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual int             read() throw(IOException);

        /**
         *  Reads a sequence of bytes into the specified buffer.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is read, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The buffer where bytes read from the stream
         *                          shall be placed.
         *  @param bytesToRead      The maximum number of bytes to read.
         *  @return                 The number of bytes actually read. This can
         *                          be smaller than \a bytesToRead if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException);

        /**
         *  Skips a sequence of input bytes.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is skipped, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  This method can only be called on open streams.
         *
         *  @param bytesToSkip      The maximum number of bytes to skip.
         *  @return                 The number of bytes actually skipped. This can
         *                          be smaller than \a bytesToSkip if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException);

        //////////
        //  Implementation
    private:
        IByteInput *            _pSource;
        UIntSize                _bufferSize;
        Decompressor *          _pDecompressor;
        UInt8 *                 _pBuffer;
        UIntSize                _bytesInBuffer;
        UIntSize                _currentPosition;
        Array<UInt8>            _compressedData;

        //  Helpers
        void                    _refillBuffer();
    };
}

#endif  //  __CARL_Compression_hpp_included__
//  End of CARL_Compression.hpp
