//
//  CARL_DateTime.hpp - Date and time handling.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.18 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines date & time services.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_DateTime.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_DateTime_hpp_included__
#define __CARL_DateTime_hpp_included__

namespace carl
{
    /**
     *  @class InconsistentTimeZoneException CARL.hpp
     *  An exception thrown when an attempt is made to register an inconsistent time zone.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC InconsistentTimeZoneException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        InconsistentTimeZoneException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        InconsistentTimeZoneException(const InconsistentTimeZoneException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~InconsistentTimeZoneException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        InconsistentTimeZoneException &     operator = (const InconsistentTimeZoneException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class DuplicateTimeZoneException CARL.hpp
     *  An exception thrown when an attempt is made to register a duplicate time zone.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DuplicateTimeZoneException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        DuplicateTimeZoneException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        DuplicateTimeZoneException(const DuplicateTimeZoneException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~DuplicateTimeZoneException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        DuplicateTimeZoneException &        operator = (const DuplicateTimeZoneException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  Represents a month.
     *
     *  @since 1.0
     */
    enum Month
    {
        /**
         *  January.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_January           = 1,

        /**
         *  February.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_February          = 2,

        /**
         *  March.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_March             = 3,

        /**
         *  April.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_April             = 4,

        /**
         *  May.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_May               = 5,

        /**
         *  June.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_June              = 6,

        /**
         *  July.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_July              = 7,

        /**
         *  August.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_August            = 8,

        /**
         *  September.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_September         = 9,

        /**
         *  October.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_October           = 10,

        /**
         *  November.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_November          = 11,

        /**
         *  December.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        Month_December          = 12
    };

    /**
     *  Represents a day of the week.
     *
     *  @since 1.0
     */
    enum DayOfWeek
    {

        /**
         *  Monday.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        DayOfWeek_Monday        = 0,

        /**
         *  Tuesday.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        DayOfWeek_Tuesday       = 1,

        /**
         *  Wednesday.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        DayOfWeek_Wednesday     = 2,

        /**
         *  Thursday.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        DayOfWeek_Thursday      = 3,

        /**
         *  Friday.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        DayOfWeek_Friday        = 4,

        /**
         *  Saturday.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        DayOfWeek_Saturday      = 5,

        /**
         *  Sunday.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        DayOfWeek_Sunday        = 6
    };

    /**
     *  @class Calendar CARL.hpp
     *  This utility class provides miscellaneous date/time - related services.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Calendar
    {
        //////////
        //  Utility classes cannot be instantiated
    private:
        Calendar();
        Calendar(const Calendar &);
        ~Calendar();
        void                    operator = (const Calendar &);

        //////////
        //  Date helpers
    public:
        /**
         *  Returns true if the specified year is a leap year, false otherwise.
         *
         *  By definition  a year is a leap year iff:
         *
         *  @code
         *  (year % 4 == 0 && year % 100 != 0) || (year % 400 == 0)
         *  @endcode
         *
         *  @param year         The year to check (\a year >= 1).
         *  @return             True if \a year refers to a leap year, false otherwise.
         *  @since 1.0
         */
        static bool             isLeapYear(int year) { require(year != 0); return (year % 4 == 0 && year % 100 != 0) || (year % 400 == 0); }

        /**
         *  Returns a number of days in the specified year.
         *
         *  @param year         The year to check (\a year >= 1).
         *  @return             The number of days in the specified year.
         *  @since 1.0
         */
        static int              daysInYear(int year) { return isLeapYear(year) ? 366 : 365; }

        /**
         *  Returns a number of days in the specified month of the specified year.
         *
         *  @param year         The year to check (\a year >= 1).
         *  @param month        The month to check (a <= \a month <= 12).
         *  @return             The number of days in the specified month of the 
         *                      specified year.
         *  @since 1.0
         */
        static int              daysInMonth(int year, int month);

        /**
         *  Returns the day-of-week of this date.
         *
         *  @param year         The year to check (\a year >= 1).
         *  @param month        The month to check (a <= \a month <= 12).
         *  @param day          The day to check (1 <= \a day <= 31).
         *  @return             The day-of-week of the specified date.
         *  @since 1.0
         */
        static DayOfWeek        getDayOfWeek(int year, int month, int day);

        /**
         *  Returns the 1-based ordinal day number within the year.
         *
         *  @param year         The year to check (\a year >= 1).
         *  @param month        The month to check (a <= \a month <= 12).
         *  @param day          The day to check (1 <= \a day <= 31).
         *  @return             The 1-based ordinal day number within the year.
         *  @since 1.0
         */
        static int              getDayOfYear(int year, int month, int day);

        //////////
        //  Display names
    public:
        /**
         *  Returns a short display name for the specified day-of-week using
         *  the current default locale.
         *
         *  @param dayOfWeek    The day-of-week to return a name for 
         *                      (0 <= \a dayOfWeek <= 6, 0 == Monday).
         *  @return             The short display name of the specified day-of-week.
         *  @since 1.0
         */
        static String           getDayOfWeekShortDisplayName(int dayOfWeek) { return getDayOfWeekShortDisplayName(dayOfWeek, Locale::getDefault()); }

        /**
         *  Returns a short display name for the specified day-of-week using
         *  the specified locale.
         *
         *  @param dayOfWeek    The day-of-week to return a name for 
         *                      (0 <= \a dayOfWeek <= 6, 0 == Monday).
         *  @param forLocale    The locale for which a display name shall be tailored.
         *  @return             The short display name of the specified day-of-week.
         *  @since 1.0
         */
        static String           getDayOfWeekShortDisplayName(int dayOfWeek, const Locale & forLocale);

        /**
         *  Returns a full display name for the specified day-of-week using
         *  the current default locale.
         *
         *  @param dayOfWeek    The day-of-week to return a name for 
         *                      (0 <= \a dayOfWeek <= 6, 0 == Monday).
         *  @return             The full display name of the specified day-of-week.
         *  @since 1.0
         */
        static String           getDayOfWeekFullDisplayName(int dayOfWeek) { return getDayOfWeekFullDisplayName(dayOfWeek, Locale::getDefault()); }

        /**
         *  Returns a full display name for the specified day-of-week using
         *  the specified locale.
         *
         *  @param dayOfWeek    The day-of-week to return a name for 
         *                      (0 <= \a dayOfWeek <= 6, 0 == Monday).
         *  @param forLocale    The locale for which a display name shall be tailored.
         *  @return             The full display name of the specified day-of-week.
         *  @since 1.0
         */
        static String           getDayOfWeekFullDisplayName(int dayOfWeek, const Locale & forLocale);

        /**
         *  Returns a short display name for the specified month using
         *  the current default locale.
         *
         *  @param month        The month to return a name for (1 <= \a month <= 12).
         *  @return             The short display name of the specified month.
         *  @since 1.0
         */
        static String           getMonthShortDisplayName(int month) { return getMonthShortDisplayName(month, Locale::getDefault()); }

        /**
         *  Returns a short display name for the specified month using
         *  the specified locale.
         *
         *  @param month        The month to return a name for (1 <= \a month <= 12).
         *  @param forLocale    The locale for which a display name shall be tailored.
         *  @return             The short display name of the specified month.
         *  @since 1.0
         */
        static String           getMonthShortDisplayName(int month, const Locale & forLocale);

        /**
         *  Returns a full display name for the specified month using
         *  the current default locale.
         *
         *  @param month        The month to return a name for (1 <= \a month <= 12).
         *  @return             The full display name of the specified month.
         *  @since 1.0
         */
        static String           getMonthFullDisplayName(int month) { return getMonthFullDisplayName(month, Locale::getDefault()); }

        /**
         *  Returns a full display name for the specified month using
         *  the specified locale.
         *
         *  @param month        The month to return a name for (1 <= \a month <= 12).
         *  @param forLocale    The locale for which a display name shall be tailored.
         *  @return             The full display name of the specified month.
         *  @since 1.0
         */
        static String           getMonthFullDisplayName(int month, const Locale & forLocale);
    };

    /**
     *  @class Date CARL.hpp
     *  Represents a date as a (year, month, day) collection.
     *
     *  A special "invalid date" value can also be represented, which is
     *  different from any valid date. When comparing dates, an invalid date
     *  is considered "less" than any valid date.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Date
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  Initializes a newly created object to an invalid date.
         *
         *  @since 1.0
         */
        Date();

        /**
         *  The class constructor.
         *
         *  Initializes a newly created object to the specified date. The \a year,
         *  \a month and \a day parameters must represent a valid date.
         *
         *  @param year         The year portion of the date (\a year >= 1);
         *  @param month        The month portion of the date (1 <= \a month >= 12);
         *  @param day          The day portion of the date (1 <= \a month >= 31);
         *  @since 1.0
         */
        Date(int year, int month, int day);

        /**
         *  The copy constructor.
         *
         *  @param src          The date to copy prroperties from.
         *  @since 1.0
         */
        Date(const Date & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~Date() {}

        /**
         *  The assignment operator.
         *
         *  @param src          The date to copy prroperties from.
         *  @return             A reference to an assigned-to object.
         *  @since 1.0
         */
        Date &              operator = (const Date & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a Date instance represents a valid date.
         *
         *  @return             True if a Date instance represents a valid date, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                isValid() const { return _year != 0 && _month != 0 && _day != 0; }

        /**
         *  Returns the year portion of the date.
         *
         *  @return             The year portion of the date, or 0 if the date is invalid.
         *  @since 1.0
         */
        int                 getYear() const { return _year; }

        /**
         *  Returns the month portion of the date.
         *
         *  @return             The month portion of the date, or 0 if the date is invalid.
         *  @since 1.0
         */
        int                 getMonth() const { return _month; }

        /**
         *  Returns the day portion of the date.
         *
         *  @return             The day portion of the date, or 0 if the date is invalid.
         *  @since 1.0
         */
        int                 getDay() const { return _day; }

        /**
         *  Returns the day-of-week that corresponds to this date.
         *
         *  Calling this method on an invalid date is an error.
         *
         *  @return             The day-of-week that corresponds to this date.
         *  @since 1.0
         */
        DayOfWeek           getDayOfWeek() const;

        /**
         *  Returns the 1-based day number within a year for this date.
         *
         *  Calling this method on an invalid date is an error.
         *
         *  @return             The 1-based day number within a year for this date.
         *  @since 1.0
         */
        int                 getDayOfYear() const;

        /**
         *  Compares two dates chronologically.
         *
         *  An invalid date is considered to be equal to an invalid date
         *  and less than any valid date.
         *
         *  @param date         The date to compare this date with.
         *  @return             
         *                      -   0 if the two dates are equal.
         *                      -   <0 if this date comes before \a date.
         *                      -   >0 if this date comes after \a date.
         *  @since 1.0
         */
        int                 compare(const Date & date) const;

        /**
         *  Compares two dates for equality.
         *
         *  An invalid date is considered to be equal to an invalid date
         *  and less than any valid date.
         *
         *  @param date         The date to compare this date with.
         *  @return             True if this date is equal to \a date, false otherwise.
         *  @since 1.0
         */
        bool                operator == (const Date & date) const { return compare(date) == 0; }

        /**
         *  Compares two dates for inequality.
         *
         *  An invalid date is considered to be equal to an invalid date
         *  and less than any valid date.
         *
         *  @param date         The date to compare this date with.
         *  @return             True if this date is not equal to \a date, false otherwise.
         *  @since 1.0
         */
        bool                operator != (const Date & date) const { return compare(date) != 0; }

        /**
         *  Checks if one date comes before another date.
         *
         *  An invalid date is considered to be equal to an invalid date
         *  and less than any valid date.
         *
         *  @param date         The date to compare this date with.
         *  @return             True if this date comes before \a date, false otherwise.
         *  @since 1.0
         */
        bool                operator <  (const Date & date) const { return compare(date) < 0; }

        /**
         *  Checks if one date comes before or at another date.
         *
         *  An invalid date is considered to be equal to an invalid date
         *  and less than any valid date.
         *
         *  @param date         The date to compare this date with.
         *  @return             True if this date comes before \a date or is equal
         *                      to \a date, false otherwise.
         *  @since 1.0
         */
        bool                operator <= (const Date & date) const { return compare(date) <= 0; }

        /**
         *  Checks if one date comes after another date.
         *
         *  An invalid date is considered to be equal to an invalid date
         *  and less than any valid date.
         *
         *  @param date         The date to compare this date with.
         *  @return             True if this date comes after \a date, false otherwise.
         *  @since 1.0
         */
        bool                operator >  (const Date & date) const { return compare(date) > 0; }

        /**
         *  Checks if one date comes at or after another date.
         *
         *  An invalid date is considered to be equal to an invalid date
         *  and less than any valid date.
         *
         *  @param date         The date to compare this date with.
         *  @return             True if this date comes after \a date or is equal
         *                      to \a date, false otherwise.
         *  @since 1.0
         */
        bool                operator >= (const Date & date) const { return compare(date) >= 0; }

        //////////
        //  Implementation
    private:
        UInt16              _year;      //  1..  , 0 == invalid date
        UInt8               _month;     //  1..12, 0 == invalid date
        UInt8               _day;       //  1..31, 0 == invalid date
    };

    /**
     *  @class Time CARL.hpp
     *  Represents a time-of-day as a (hour, minute, second, nanosecond) collection.
     *
     *  A special "invalid time" value can also be represented, which is
     *  different from any valid time-of-day. When comparing times, an invalid time
     *  is considered "less" than any valid time.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Time
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  Initializes a newly created object to an invalid time-of-day.
         *
         *  @since 1.0
         */
        Time();

        /**
         *  The class constructor.
         *
         *  Initializes a newly created object to the specified time-of-day. The \a hour,
         *  \a minute, \a second and \a nanosecond parameters must represent a valid 
         *  time-of-day.
         *
         *  @param hour         The hour portion of the time-of-day (0 <= \a hour <= 23);
         *  @param minute       The minute portion of the time-of-day (0 <= \a minute <= 59);
         *  @param second       The second portion of the time-of-day (0 <= \a second <= 59);
         *  @param nanosecond   The nanosecond portion of the time-of-day (0 <= 
         *                      \a nanosecond <= 999999999);
         *  @since 1.0
         */
        Time(int hour, int minute, int second = 0, UInt32 nanosecond = 0);

        /**
         *  The copy constructor.
         *
         *  @param src          The time to copy prroperties from.
         *  @since 1.0
         */
        Time(const Time & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~Time() {}

        /**
         *  The assignment operator.
         *
         *  @param src          The time to copy prroperties from.
         *  @return             A reference to an assigned-to object.
         *  @since 1.0
         */
        Time &              operator = (const Time & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a Time instance represents a valid time-of-day.
         *
         *  @return             True if a Time instance represents a valid time-of-day, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                isValid() const { return _isValid; }

        /**
         *  Returns the hour portion of the time-of-day.
         *
         *  @return             The hour portion of the time-of-day, or 
         *                      0 if the time-of-day is invalid.
         *  @since 1.0
         */
        int                 getHour() const { return _hour; }

        /**
         *  Returns the minute portion of the time-of-day.
         *
         *  @return             The minute portion of the time-of-day, or 
         *                      0 if the time-of-day is invalid.
         *  @since 1.0
         */
        int                 getMinute() const { return _minute; }

        /**
         *  Returns the second portion of the time-of-day.
         *
         *  @return             The second portion of the time-of-day, or 
         *                      0 if the time-of-day is invalid.
         *  @since 1.0
         */
        int                 getSecond() const { return _second; }

        /**
         *  Returns the nanosecond portion of the time-of-day.
         *
         *  @return             The nanosecond portion of the time-of-day, or 
         *                      0 if the time-of-day is invalid.
         *  @since 1.0
         */
        UInt32              getNanosecond() const { return _nanosecond; }

        /**
         *  Compares two times-of-day chronologically.
         *
         *  An invalid time-of-day is considered to be equal to an invalid time-of-day
         *  and less than any valid time-of-day.
         *
         *  @param time         The time to compare this span to.
         *  @return             
         *                      -   0 if the two times-of-day are equal.
         *                      -   <0 if this time-of-day comes before \a time.
         *                      -   >0 if this time-of-day comes after \a time.
         *  @since 1.0
         */
        int                 compare(const Time & time) const;

        /**
         *  Compares two times-of-day for equality.
         *
         *  An invalid time-of-day is considered to be equal to an invalid time-of-day
         *  and less than any valid time-of-day.
         *
         *  @param time         The time to compare this span to.
         *  @return             True if this time-of-day is equal to \a time, false otherwise.
         *  @since 1.0
         */
        bool                operator == (const Time & time) const { return compare(time) == 0; }

        /**
         *  Compares two times-of-day for inequality.
         *
         *  An invalid time-of-day is considered to be equal to an invalid time-of-day
         *  and less than any valid time-of-day.
         *
         *  @param time         The time to compare this span to.
         *  @return             True if this time-of-day is not equal to \a time, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator != (const Time & time) const { return compare(time) != 0; }

        /**
         *  Checks if one time-of-day comes before another time-of-day.
         *
         *  An invalid time-of-day is considered to be equal to an invalid time-of-day
         *  and less than any valid time-of-day.
         *
         *  @param time         The time to compare this span to.
         *  @return             True if this time-of-day comes before \a time, false otherwise.
         *  @since 1.0
         */
        bool                operator <  (const Time & time) const { return compare(time) < 0; }

        /**
         *  Checks if one time-of-day comes before or at another time-of-day.
         *
         *  An invalid time-of-day is considered to be equal to an invalid time-of-day
         *  and less than any valid time-of-day.
         *
         *  @param time         The time to compare this span to.
         *  @return             True if this time-of-day comes before \a time or is equal
         *                      to \a time, false otherwise.
         *  @since 1.0
         */
        bool                operator <= (const Time & time) const { return compare(time) <= 0; }

        /**
         *  Checks if one time-of-day comes after another time-of-day.
         *
         *  An invalid time-of-day is considered to be equal to an invalid time-of-day
         *  and less than any valid time-of-day.
         *
         *  @param time         The time to compare this span to.
         *  @return             True if this time-of-day comes after \a time, false otherwise.
         *  @since 1.0
         */
        bool                operator >  (const Time & time) const { return compare(time) > 0; }

        /**
         *  Checks if one time-of-day comes at or after another time-of-day.
         *
         *  An invalid time-of-day is considered to be equal to an invalid time-of-day
         *  and less than any valid time-of-day.
         *
         *  @param time         The time to compare this span to.
         *  @return             True if this time-of-day comes after \a time or is equal
         *                      to \a time, false otherwise.
         *  @since 1.0
         */
        bool                operator >= (const Time & time) const { return compare(time) >= 0; }

        //////////
        //  Implementation
    private:
        bool                _isValid;       //  false == invalid time
        UInt8               _hour;          //  0..23
        UInt8               _minute;        //  0..59
        UInt8               _second;        //  0..59
        UInt32              _nanosecond;    //  0..999999999
    };

    /**
     *  @class DateTime CARL.hpp
     *  Represents a date and time-of-day as a (year, month, day, hour, minute, 
     *  second, nanosecond) collection.
     *
     *  A special "invalid date+time" value can also be represented, which is
     *  different from any valid date+time. When comparing dates+times, an invalid date+time
     *  is considered "less" than any valid date+time.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DateTime
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  Initializes a newly created object to an invalid date+time.
         *
         *  @since 1.0
         */
        DateTime();

        /**
         *  The class constructor.
         *
         *  Initializes a newly created object to the specified date+time. The \a year,
         *  \a month and \a day parameters must represent a valid date. The \a hour,
         *  \a minute, \a second and \a nanosecond parameters must represent a valid 
         *  time-of-day.
         *
         *  @param year         The year portion of the date (\a year >= 1);
         *  @param month        The month portion of the date (1 <= \a month >= 12);
         *  @param day          The day portion of the date (1 <= \a month >= 31);
         *  @param hour         The hour portion of the time-of-day (0 <= \a hour <= 23);
         *  @param minute       The minute portion of the time-of-day (0 <= \a minute <= 59);
         *  @param second       The second portion of the time-of-day (0 <= \a second <= 59);
         *  @param nanosecond   The nanosecond portion of the time-of-day (0 <= 
         *                      \a nanosecond <= 999999999);
         *  @since 1.0
         */
        DateTime(int year, int month, int day, int hour, int minute, int second = 0, UInt32 nanosecond = 0);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created object to the specified date+time.
         *
         *  @param date         The date portion of the date+time.
         *  @param time         The time-of-day portion of the date+time.
         *  @since 1.0
         */
        DateTime(const Date & date, const Time & time);

        /**
         *  The copy constructor.
         *
         *  @param src          The date+time to copy prroperties from.
         *  @since 1.0
         */
        DateTime(const DateTime & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~DateTime() {}

        /**
         *  The assignment operator.
         *
         *  @param src          The date+time to copy prroperties from.
         *  @return             A reference to an assigned-to object.
         *  @since 1.0
         */
        DateTime &          operator = (const DateTime & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a DateTime instance represents a valid date+time.
         *
         *  @return             True if a DateTime instance represents a valid date+time, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                isValid() const { return _year != 0 && _month != 0 && _day != 0; }

        /**
         *  Returns the date portion of the date+time.
         *
         *  @return             The date portion of the date+time.
         *  @since 1.0
         */
        Date                getDate() const;

        /**
         *  Returns the time portion of the date+time.
         *
         *  @return             The time portion of the date+time.
         *  @since 1.0
         */
        Time                getTime() const;

        /**
         *  Returns the year portion of the date.
         *
         *  @return             The year portion of the date+time, or 0 if 
         *                      the date+time is invalid.
         *  @since 1.0
         */
        int                 getYear() const { return _year; }

        /**
         *  Returns the month portion of the date.
         *
         *  @return             The month portion of the date+time, or 0 if 
         *                      the date+time is invalid.
         *  @since 1.0
         */
        int                 getMonth() const { return _month; }

        /**
         *  Returns the day portion of the date.
         *
         *  @return             The day portion of the date+time, or 0 if 
         *                      the date+time is invalid.
         *  @since 1.0
         */
        int                 getDay() const { return _day; }

        /**
         *  Returns the hour portion of the date.
         *
         *  @return             The hour portion of the date+time, or 0 if 
         *                      the date+time is invalid.
         *  @since 1.0
         */
        int                 getHour() const { return _hour; }

        /**
         *  Returns the minute portion of the date.
         *
         *  @return             The minute portion of the date+time, or 0 if 
         *                      the date+time is invalid.
         *  @since 1.0
         */
        int                 getMinute() const { return _minute; }

        /**
         *  Returns the second portion of the date.
         *
         *  @return             The second portion of the date+time, or 0 if 
         *                      the date+time is invalid.
         *  @since 1.0
         */
        int                 getSecond() const { return _second; }

        /**
         *  Returns the nanosecond portion of the date.
         *
         *  @return             The nanosecond portion of the date+time, or 0 if 
         *                      the date+time is invalid.
         *  @since 1.0
         */
        UInt32              getNanosecond() const { return _nanosecond; }

        /**
         *  Returns the day-of-week that corresponds to this date+time.
         *
         *  Calling this method on an invalid date+time is an error.
         *
         *  @return             The day-of-week that corresponds to this date+time.
         *  @since 1.0
         */
        DayOfWeek           getDayOfWeek() const;

        /**
         *  Returns the 1-based day number within a year for this date+time.
         *
         *  Calling this method on an invalid date+time is an error.
         *
         *  @return             The 1-based day number within a year for this date+time.
         *  @since 1.0
         */
        int                 getDayOfYear() const;

        /**
         *  Compares two dates+times chronologically.
         *
         *  An invalid date+time is considered to be equal to an invalid date+time
         *  and less than any valid date+time.
         *
         *  @param dateTime     The date+time to compare this date+time with.
         *  @return             
         *                      -   0 if the two dates+times are equal.
         *                      -   <0 if this date+time comes before \a dateTime.
         *                      -   >0 if this date+time comes after \a dateTime.
         *  @since 1.0
         */
        int                 compare(const DateTime & dateTime) const;

        /**
         *  Compares two dates+times for equality.
         *
         *  An invalid date+time is considered to be equal to an invalid date+time
         *  and less than any valid date+time.
         *
         *  @param dateTime     The date+time to compare this date+time with.
         *  @return             True if this date+time is equal to \a dateTime, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator == (const DateTime & dateTime) const { return compare(dateTime) == 0; }

        /**
         *  Compares two dates+times for inequality.
         *
         *  An invalid date+time is considered to be equal to an invalid date+time
         *  and less than any valid date+time.
         *
         *  @param dateTime     The date+time to compare this date+time with.
         *  @return             True if this date+time is not equal to \a dateTime, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator != (const DateTime & dateTime) const { return compare(dateTime) != 0; }

        /**
         *  Checks if one date+time comes before another date+time.
         *
         *  An invalid date+time is considered to be equal to an invalid date+time
         *  and less than any valid date+time.
         *
         *  @param dateTime     The date+time to compare this date+time with.
         *  @return             True if this date+time comes before \a dateTime, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator <  (const DateTime & dateTime) const { return compare(dateTime) < 0; }

        /**
         *  Checks if one date+time comes before or at another date+time.
         *
         *  An invalid date+time is considered to be equal to an invalid date+time
         *  and less than any valid date+time.
         *
         *  @param dateTime     The date+time to compare this date+time with.
         *  @return             True if this date+time comes before \a dateTime or is equal
         *                      to \a dateTime, false otherwise.
         *  @since 1.0
         */
        bool                operator <= (const DateTime & dateTime) const { return compare(dateTime) <= 0; }

        /**
         *  Checks if one date+time comes after another date+time.
         *
         *  An invalid date+time is considered to be equal to an invalid date+time
         *  and less than any valid date+time.
         *
         *  @param dateTime     The date+time to compare this date+time with.
         *  @return             True if this date+time comes after \a dateTime, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator >  (const DateTime & dateTime) const { return compare(dateTime) > 0; }

        /**
         *  Checks if one date+time comes at or after another date+time.
         *
         *  An invalid date+time is considered to be equal to an invalid date+time
         *  and less than any valid date+time.
         *
         *  @param dateTime     The date+time to compare this date+time with.
         *  @return             True if this date+time comes after \a dateTime or is equal
         *                      to \a dateTime, false otherwise.
         *  @since 1.0
         */
        bool                operator >= (const DateTime & dateTime) const { return compare(dateTime) >= 0; }

        //////////
        //  Implementation
    private:
        UInt16              _year;          //  1..  , 0 == invalid date/time
        UInt8               _month;         //  1..12, 0 == invalid date/time
        UInt8               _day;           //  1..31, 0 == invalid date/time
        UInt8               _hour;          //  0..23
        UInt8               _minute;        //  0..59
        UInt8               _second;        //  0..59
        UInt32              _nanosecond;    //  0..999999999
    };

    /**
     *  @class TimeSpan CARL.hpp
     *  Represents an interval between two moments in time with nanosecond precision.
     *
     *  A special "invalid time span" value can also be represented, which is
     *  different from any valid time span. When comparing time spans, an invalid time span
     *  is considered "less" than any valid time span.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC TimeSpan
    {
        friend class Application;

        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  Initializes a newly created object to an invalid time span.
         *
         *  @since 1.0
         */
        TimeSpan() { _nanoseconds = _InvalidValue; }

        /**
         *  The class constructor.
         *
         *  Initializes a newly created object to the specified time span measured
         *  in nanoseconds.
         *
         *  @param nanoseconds  The time interval in nanoseconds.
         *  @since 1.0
         */
        explicit TimeSpan(Int64 nanoseconds) { _nanoseconds = nanoseconds; }

        /**
         *  The copy constructor.
         *
         *  @param src          The time span to copy prroperties from.
         *  @since 1.0
         */
        TimeSpan(const TimeSpan & src) { _nanoseconds = src._nanoseconds; }

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~TimeSpan() {}

        /**
         *  The assignment operator.
         *
         *  @param src          The time span to copy prroperties from.
         *  @return             A reference to an assigned-to object.
         *  @since 1.0
         */
        TimeSpan &          operator = (const TimeSpan & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns an elapsed time since a well-defined reference point in the past.
         *
         *  This method can be used for measuring elapsed time intervals
         *  with the highest precision supported by the target platform.
         *  The exact definition of a "well-defined reference point in
         *  the past" may vary between platforms or even between different
         *  launches of the same application; however, it is guaranteed
         *  that the same reference point will be used for the entire duration
         *  of an application execution.
         *
         *  If the requested information is not available, returns an
         *  invalid time span.
         *
         *  @return         An elapsed time since a well-defined reference point.
         *  @since 1.0
         */
        static TimeSpan     getTimeSpanSinceReferencePoint();

        /**
         *  Checks if a DateTime instance represents a valid date+time.
         *
         *  @return             True if a DateTime instance represents a valid date+time, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                isValid() const { return _nanoseconds != _InvalidValue; }

        /**
         *  Returns a total number of nanoseconds in this time span.
         *
         *  The return value can be negative for inverse time spans.
         *
         *  @return             A total number of nanoseconds in this time span.
         *                      If the time span is invalid, the return value
         *                      is undefined.
         *  @since 1.0
         */
        Int64               getNanoseconds() const { return _nanoseconds; }

        /**
         *  Compares two time spans chronologically.
         *
         *  An invalid time span is considered to be equal to an invalid time span
         *  and less than any valid time span.
         *
         *  @param timeSpan     The time span to compare this time span to.
         *  @return             
         *                      -   0 if the two time spans are equal.
         *                      -   <0 if this time span is smaller than \a timeSpan.
         *                      -   >0 if this time span is greater than \a timeSpan.
         *  @since 1.0
         */
        int                 compare(const TimeSpan & timeSpan) const;

        /**
         *  Compares two time spans for equality.
         *
         *  An invalid time span is considered to be equal to an invalid time span
         *  and less than any valid time span.
         *
         *  @param timeSpan     The time span to compare this time span to.
         *  @return             True if this time span is equal to \a timeSpan, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator == (const TimeSpan & timeSpan) const { return compare(timeSpan) == 0; }

        /**
         *  Compares two time spans for inequality.
         *
         *  An invalid time span is considered to be equal to an invalid time span
         *  and less than any valid time span.
         *
         *  @param timeSpan     The time span to compare this time span to.
         *  @return             True if this time span is not equal to \a timeSpan, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator != (const TimeSpan & timeSpan) const { return compare(timeSpan) != 0; }

        /**
         *  Checks if one time span is smaller than another time span.
         *
         *  An invalid time span is considered to be equal to an invalid time span
         *  and less than any valid time span.
         *
         *  @param timeSpan     The time span to compare this time span to.
         *  @return             True if this time span is smaller than \a timeSpan, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator <  (const TimeSpan & timeSpan) const { return compare(timeSpan) < 0; }

        /**
         *  Checks if one time span is smaller than or equal to another time span.
         *
         *  An invalid time span is considered to be equal to an invalid time span
         *  and less than any valid time span.
         *
         *  @param timeSpan     The time span to compare this time span to.
         *  @return             True if this time span is smaller than or equal to
         *                      \a timeSpan, false otherwise.
         *  @since 1.0
         */
        bool                operator <= (const TimeSpan & timeSpan) const { return compare(timeSpan) <= 0; }

        /**
         *  Checks if one time span is greater than another time span.
         *
         *  An invalid time span is considered to be equal to an invalid time span
         *  and less than any valid time span.
         *
         *  @param timeSpan     The time span to compare this time span to.
         *  @return             True if this time span is greater than \a timeSpan, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator >  (const TimeSpan & timeSpan) const { return compare(timeSpan) > 0; }

        /**
         *  Checks if one time span is greater than or equal to another time span.
         *
         *  An invalid time span is considered to be equal to an invalid time span
         *  and less than any valid time span.
         *
         *  @param timeSpan     The time span to compare this time span to.
         *  @return             True if this time span is greater than or equal to
         *                      \a timeSpan, false otherwise.
         *  @since 1.0
         */
        bool                operator >= (const TimeSpan & timeSpan) const { return compare(timeSpan) >= 0; }

        /**
         *  Returns the value of the time span itself.
         *
         *  @return             A time span that is equal to this time span.
         *  @since 1.0
         */
        TimeSpan            operator + () const { return TimeSpan(_nanoseconds); }

        /**
         *  Returns an inverse of the time span.
         *
         *  An attempt to invert an invalid time span always results in an invalid 
         *  time span.
         *
         *  @return             A time span that is equal in magnitude to this time 
         *                      span but has an opposite sign.
         *  @since 1.0
         */
        TimeSpan            operator - () const;

        /**
         *  Returns a time span obtained by adding the specified number of
         *  nanoseconds to this time span.
         *
         *  An attempt to add anything to an invalid time span always results 
         *  in an invalid time span.
         *  
         *  @param nanoseconds  A number of nanoseconds to add to the time span.
         *  @return             A time span obtained by adding the specified number of
         *                      nanoseconds to this time span.
         *  @since 1.0
         */
        TimeSpan            operator + (Int64 nanoseconds) const;

        /**
         *  Returns a sum of the two time spans.
         *
         *  If either operand is an invalid time span, the result is always
         *  an invalid time span.
         *  
         *  @param timeSpan     A time span to add to the time span.
         *  @return             A sum of the two time spans.
         *  @since 1.0
         */
        TimeSpan            operator + (const TimeSpan & timeSpan) const;

        /**
         *  Returns a time span obtained by subtracting the specified number of
         *  nanoseconds tfrom this time span.
         *
         *  An attempt to subtract anything from an invalid time span always results 
         *  in an invalid time span.
         *  
         *  @param nanoseconds  A number of nanoseconds to add to this time span.
         *  @return             A time span obtained by subtracting the specified 
         *                      number of nanoseconds from this time span.
         *  @since 1.0
         */
        TimeSpan            operator - (Int64 nanoseconds) const;

        /**
         *  Returns a difference of the two time spans.
         *
         *  If either operand is an invalid time span, the result is always
         *  an invalid time span.
         *  
         *  @param timeSpan     A time span to subtract from the time span.
         *  @return             A difference of the two time spans.
         *  @since 1.0
         */
        TimeSpan            operator - (const TimeSpan & timeSpan) const;

        /**
         *  Returns a time span multiplied by the specified factor.
         *
         *  An attempt to multiply an invalid time span by any factor
         *  always results in an invalid time span.
         *  
         *  @param factor       The factor to multiply the time span by.
         *  @return             A time span multiplied by the specified factor.
         *  @since 1.0
         */
        TimeSpan            operator * (Int64 factor) const;

        /**
         *  Returns a time span divided by the specified factor.
         *
         *  An attempt to divide an invalid time span by any factor
         *  other than 0 always results in an invalid time span.
         *
         *  An attempt to divide a time span (whether valid or invalid)
         *  by a factor of 0 is an error.
         *  
         *  @param factor       The factor to divide the time span by.
         *  @return             A time span multiplied by the specified factor.
         *  @since 1.0
         */
        TimeSpan            operator / (Int64 factor) const;

        /**
         *  Adds the specified number of nanoseconds to this time span.
         *
         *  An attempt to add anything to an invalid time span always results 
         *  in an invalid time span.
         *
         *  @param nanoseconds  A number of nanoseconds to add to this time span.
         *  @return             A reference to an added-to value.
         *  @since 1.0
         */
        TimeSpan &          operator += (Int64 nanoseconds);

        /**
         *  Adds the specified time span of nanoseconds to this time span.
         *
         *  An attempt to add anything to an invalid time span always results 
         *  in an invalid time span. Similarly, an attempt to add an invalid 
         *  time span to a valid time span always results in an invalid time span.
         *
         *  @param timeSpan     A time span to add to this time span.
         *  @return             A reference to an added-to value.
         *  @since 1.0
         */
        TimeSpan &          operator += (const TimeSpan & timeSpan);

        /**
         *  Subtracts the specified number of nanoseconds from this time span.
         *
         *  An attempt to subtract anything from an invalid time span always results 
         *  in an invalid time span.
         *
         *  @param nanoseconds  A number of nanoseconds to subtract from this time span.
         *  @return             A reference to an subtracted-from to value.
         *  @since 1.0
         */
        TimeSpan &          operator -= (Int64 nanoseconds);

        /**
         *  Subtracts the specified time span of nanoseconds from this time span.
         *
         *  An attempt to subtract anything from an invalid time span always results 
         *  in an invalid time span. Similarly, an attempt to subtract an invalid 
         *  time span from a valid time span always results in an invalid time span.
         *
         *  @param timeSpan     A time span to dubtract from this time span.
         *  @return             A reference to an subtracted-from value.
         *  @since 1.0
         */
        TimeSpan &          operator -= (const TimeSpan & timeSpan);

        /**
         *  Multiplies this time span by the specified factor.
         *
         *  An attempt to multiply an invalid time span by any factor
         *  always results in an invalid time span.
         *  
         *  @param factor       The factor to multiply this time span by.
         *  @return             A reference to an scaled time span value.
         *  @since 1.0
         */
        TimeSpan &          operator *= (Int64 factor);

        /**
         *  Divides this time span by the specified factor.
         *
         *  An attempt to divide an invalid time span by any factor other
         *  than 0 always results in an invalid time span.
         *  
         *  An attempt to divide a time span (whether valid or invalid)
         *  by a factor of 0 is an error.
         *
         *  @param factor       The factor to divide this time span by.
         *  @return             A reference to an scaled time span value.
         *  @since 1.0
         */
        TimeSpan &          operator /= (Int64 factor);

        //////////
        //  Implementation
    private:
        static const Int64  _InvalidValue = static_cast<Int64>(1) << 63;
        Int64               _nanoseconds;
    };

    /**
     *  @class TimeStamp CARL.hpp
     *  Represents elapsed time since midnight January 1 1970 UTC with
     *  nanosecond precision.
     *
     *  A special "invalid time stamp" value can also be represented, which is
     *  different from any valid time stamp. When comparing time stamps, an invalid 
     *  time stamp is considered "less" than any valid time stamp.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC TimeStamp
    {
        //////////
        //  Construction/destruction/assignment
    public:
        TimeStamp() { _nanosecondsSinceEpoch = _InvalidValue; }
        explicit TimeStamp(UInt64 nanosecondsSinceEpoch) { _nanosecondsSinceEpoch = nanosecondsSinceEpoch; }
        TimeStamp(const TimeStamp & src) { _nanosecondsSinceEpoch = src._nanosecondsSinceEpoch; }
        ~TimeStamp() {}

        TimeStamp &         operator = (const TimeStamp & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the current UTC time.
         *
         *  If it is not possible to determine the current UTC time, returns
         *  an invalid time stamp.
         *
         *  @return             The current UTC time.
         *  @since 1.0
         */
        static TimeStamp    now();

        /**
         *  Checks if a TimeStamp instance represents a valid time stamp.
         *
         *  @return             True if a TimeStamp instance represents a valid time
         *                      stamp, false otherwise.
         *  @since 1.0
         */
        bool                isValid() const { return _nanosecondsSinceEpoch != _InvalidValue;}

        /**
         *  Returns the total number of elapsed nanoseconds since midnight 
         *  January 1 1970 UTC represented by this time stamp.
         *
         *  The return value is undefined for an invalid time stamp.
         *
         *  @return             Elapsed nanoseconds since midnight January 1 1970 UTC.
         *  @since 1.0
         */
        UInt64              getNanosecondsSinceEpoch() const { return _nanosecondsSinceEpoch; }

        /**
         *  Compares two time stamps chronologically.
         *
         *  An invalid time stamp is considered to be equal to an invalid time stamp
         *  and less than any valid time stamp.
         *
         *  @param timeStamp    The time stamp to compare this time stamp to.
         *  @return             
         *                      -   0 if the two time stamps are equal.
         *                      -   <0 if this time stamp comes before \a timeStamp.
         *                      -   >0 if this time stamp comes after \a timeStamp.
         *  @since 1.0
         */
        int                 compare(const TimeStamp & timeStamp) const;

        /**
         *  Compares two time stamps for equality.
         *
         *  An invalid time stamp is considered to be equal to an invalid time stamp
         *  and less than any valid time stamp.
         *
         *  @param timeStamp    The time stamp to compare this time stamp to.
         *  @return             True if this time stamp is equal to \a timeStasmp, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator == (const TimeStamp & timeStamp) const { return compare(timeStamp) == 0; }

        /**
         *  Compares two time stamps for inequality.
         *
         *  An invalid time stamp is considered to be equal to an invalid time stamp
         *  and less than any valid time stamp.
         *
         *  @param timeStamp    The time stamp to compare this time stamp to.
         *  @return             True if this time stamp is not equal to \a timeStasmp, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator != (const TimeStamp & timeStamp) const { return compare(timeStamp) != 0; }

        /**
         *  Checks if one time stamp comes before another time stamp.
         *
         *  An invalid time stamp is considered to be equal to an invalid time stamp
         *  and less than any valid time stamp.
         *
         *  @param timeStamp    The time stamp to compare this time stamp to.
         *  @return             True if this time stamp comes before \a timeStamp, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator <  (const TimeStamp & timeStamp) const { return compare(timeStamp) < 0; }

        /**
         *  Checks if one time stamp comes before or at another time stamp.
         *
         *  An invalid time stamp is considered to be equal to an invalid time stamp
         *  and less than any valid time stamp.
         *
         *  @param timeStamp    The time stamp to compare this time stamp to.
         *  @return             True if this time stamp comes before or at \a timeStamp, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator <= (const TimeStamp & timeStamp) const { return compare(timeStamp) <= 0; }

        /**
         *  Checks if one time stamp comes after another time stamp.
         *
         *  An invalid time stamp is considered to be equal to an invalid time stamp
         *  and less than any valid time stamp.
         *
         *  @param timeStamp    The time stamp to compare this time stamp to.
         *  @return             True if this time stamp comes after \a timeStamp, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator >  (const TimeStamp & timeStamp) const { return compare(timeStamp) > 0; }

        /**
         *  Checks if one time stamp comes at or after another time stamp.
         *
         *  An invalid time stamp is considered to be equal to an invalid time stamp
         *  and less than any valid time stamp.
         *
         *  @param timeStamp    The time stamp to compare this time stamp to.
         *  @return             True if this time stamp comes at or after \a timeStamp, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                operator >= (const TimeStamp & timeStamp) const { return compare(timeStamp) >= 0; }

        /**
         *  Advances a time stamp forward by the specified time interval.
         *
         *  An attempt to advance an invalid time stamp by any time interval
         *  always results in an invalid time stamp. Similarly, an attempt to
         *  advance any time stamp (whether valid or not) by an invalid
         *  time interval always results in an invalid time stamp.
         *
         *  @param timeSpan     The time interval to advance this time stamp by.
         *  @return             A reference to an added-to to value.
         *  @since 1.0
         */
        TimeStamp &         operator += (const TimeSpan & timeSpan);

        /**
         *  Advances a time stamp backward by the specified time interval.
         *
         *  An attempt to advance an invalid time stamp by any time interval
         *  always results in an invalid time stamp. Similarly, an attempt to
         *  advance any time stamp (whether valid or not) by an invalid
         *  time interval always results in an invalid time stamp.
         *
         *  @param timeSpan     The time interval to advance this time stamp by.
         *  @return             A reference to an subtracted-from to value.
         *  @since 1.0
         */
        TimeStamp &         operator -= (const TimeSpan & timeSpan);

        /**
         *  Returns a time stamp that is ahead of this time stamp by a 
         *  specified time interval.
         *
         *  An attempt to advance an invalid time stamp by any time interval
         *  always results in an invalid time stamp. Similarly, an attempt to
         *  advance any time stamp (whether valid or not) by an invalid
         *  time interval always results in an invalid time stamp.
         *
         *  @param timeSpan     The required time interval between thw two time stamps.
         *  @return             A time stamp that is ahead of this time stamp by a 
         *                      specified time interval.
         *  @since 1.0
         */
        TimeStamp           operator + (const TimeSpan & timeSpan) const;

        /**
         *  Returns a time stamp that is behind of this time stamp by a 
         *  specified time interval.
         *
         *  An attempt to advance an invalid time stamp by any time interval
         *  always results in an invalid time stamp. Similarly, an attempt to
         *  advance any time stamp (whether valid or not) by an invalid
         *  time interval always results in an invalid time stamp.
         *
         *  @param timeSpan     The required time interval between thw two time stamps.
         *  @return             A time stamp that is behind of this time stamp by a 
         *                      specified time interval.
         *  @since 1.0
         */
        TimeStamp           operator - (const TimeSpan & timeSpan) const;

        /**
         *  Returns a time interval between two time stamps.
         *
         *  If either operand is an invalid time stamp, the result is 
         *  always an invalid time span.
         *
         *  @param timeStamp    The second operand of the time stamp difference.
         *  @return             A time interval between two time stamps.
         *  @since 1.0
         */
        TimeSpan            operator - (const TimeStamp & timeStamp) const;

        //////////
        //  Implementation
    private:
        static const UInt64 _InvalidValue = ~static_cast<UInt64>(0);
        UInt64              _nanosecondsSinceEpoch;
    };

    /**
     *  @class TimeZone CARL.hpp
     *  Specifies a set of rules for converting between local times and UTC.
     *
     *  When convertins between UTC and local time, the time zone takes into
     *  consideration the time difference between UTC and the time zone's
     *  local area as well as daylyght saving time effects if the time
     *  zone uses DST.
     *
     *  Currently, the following time zones are predefined by 
     *  CARL (the list contains time zone names, with display 
     *  names in brackets after the identifier):
     *
     *  -   Afghanistan Standard Time ((GMT+04:30) Kabul)
     *  -   Alaskan Standard Time ((GMT-09:00) Alaska)
     *  -   Arab Standard Time ((GMT+03:00) Kuwait)
     *  -   Arabian Standard Time ((GMT+04:00) Abu Dhabi, Muscat)
     *  -   Arabic Standard Time ((GMT+03:00) Baghdad)
     *  -   Atlantic Standard Time ((GMT-04:00) Atlantic Time (Canada))
     *  -   AUS Central Standard Time ((GMT+09:30) Darwin)
     *  -   AUS Eastern Standard Time ((GMT+10:00) Canberra, Melbourne, Sydney)
     *  -   Azores Standard Time ((GMT-01:00) Azores)
     *  -   Canada Central Standard Time ((GMT-06:00) Saskatchewan)
     *  -   Cape Verde Standard Time ((GMT-01:00) Cape Verde Is.)
     *  -   Caucasus Standard Time ((GMT+04:00) Baku, Tbilisi, Yerevan)
     *  -   Cen. Australia Standard Time ((GMT+09:30) Adelaide)
     *  -   Central America Standard Time ((GMT-06:00) Central America)
     *  -   Central Asia Standard Time ((GMT+06:00) Astana, Dhaka)
     *  -   Central Europe Time ((GMT+01:00) Belgrade, Bratislava, Budapest, Ljubljana, Prague)
     *  -   Central European Time ((GMT+01:00) Sarajevo, Skopje, Warsaw, Zagreb)
     *  -   Central Pacific Standard Time ((GMT+11:00) Magadan, Solomon Is.)
     *  -   Central Standard Time ((GMT-06:00) Central Time (US & Canada))
     *  -   China Standard Time ((GMT+08:00) Beijing, Chongqing, Hong Kong, Urumqi)
     *  -   Dateline Standard Time ((GMT-12:00) International Date Line West)
     *  -   E. Africa Standard Time ((GMT+03:00) Nairobi)
     *  -   E. Australia Standard Time ((GMT+10:00) Brisbane)
     *  -   E. Europe Standard Time ((GMT+02:00) Bucharest)
     *  -   E. South America Standard Time ((GMT-03:00) Brasilia)
     *  -   Eastern Standard Time ((GMT-05:00) Eastern Time (US & Canada))
     *  -   Egypt Standard Time ((GMT+02:00) Cairo)
     *  -   Ekaterinburg Standard Time ((GMT+05:00) Ekaterinburg)
     *  -   Fiji Standard Time ((GMT+12:00) Fiji, Kamchatka, Marshall Is.)
     *  -   FLE Standard Time ((GMT+02:00) Helsinki, Kyiv, Riga, Sofia, Tallinn, Vilnius)
     *  -   GMT Standard Time ((GMT) Greenwich Mean Time : Dublin, Edinburgh, Lisbon, London)
     *  -   Greenland Standard Time ((GMT-03:00) Greenland)
     *  -   Greenwich Standard Time ((GMT) Casablanca)
     *  -   GTB Standard Time ((GMT+02:00) Athens, Beirut, Istanbul, Minsk)
     *  -   Hawaiian Standard Time ((GMT-10:00) Hawaii)
     *  -   India Standard Time ((GMT+05:30) Chennai, Kolkata, Mumbai, New Delhi)
     *  -   Iran Standard Time ((GMT+03:30) Tehran)
     *  -   Israel Standard Time ((GMT+02:00) Jerusalem)
     *  -   Korea Standard Time ((GMT+09:00) Seoul)
     *  -   Mexico Standard Time ((GMT-06:00) Guadalajara, Mexico City, Monterrey)
     *  -   Mexico Standard Time 2 ((GMT-07:00) Chihuahua, La Paz, Mazatlan)
     *  -   Mid-Atlantic Standard Time ((GMT-02:00) Mid-Atlantic)
     *  -   Mountain Standard Time ((GMT-07:00) Mountain Time (US & Canada))
     *  -   Myanmar Standard Time ((GMT+06:30) Rangoon)
     *  -   N. Central Asia Standard Time ((GMT+06:00) Almaty, Novosibirsk)
     *  -   Nepal Standard Time ((GMT+05:45) Kathmandu)
     *  -   New Zealand Standard Time ((GMT+12:00) Auckland, Wellington)
     *  -   Newfoundland Standard Time ((GMT-03:30) Newfoundland)
     *  -   North Asia East Standard Time ((GMT+08:00) Irkutsk, Ulaan Bataar)
     *  -   North Asia Standard Time ((GMT+07:00) Krasnoyarsk)
     *  -   Pacific SA Standard Time ((GMT-04:00) Santiago)
     *  -   Pacific Standard Time ((GMT-08:00) Pacific Time (US & Canada); Tijuana)
     *  -   Romance Standard Time ((GMT+01:00) Brussels, Copenhagen, Madrid, Paris)
     *  -   Russian Standard Time ((GMT+03:00) Moscow, St. Petersburg, Volgograd)
     *  -   SA Eastern Standard Time ((GMT-03:00) Buenos Aires, Georgetown)
     *  -   SA Pacific Standard Time ((GMT-05:00) Bogota, Lima, Quito)
     *  -   SA Western Standard Time ((GMT-04:00) Caracas, La Paz)
     *  -   Samoa Standard Time ((GMT-11:00) Midway Island, Samoa)
     *  -   SE Asia Standard Time ((GMT+07:00) Bangkok, Hanoi, Jakarta)
     *  -   Singapore Standard Time ((GMT+08:00) Kuala Lumpur, Singapore)
     *  -   South Africa Standard Time ((GMT+02:00) Harare, Pretoria)
     *  -   Sri Lanka Standard Time ((GMT+06:00) Sri Jayawardenepura)
     *  -   Taipei Standard Time ((GMT+08:00) Taipei)
     *  -   Tasmania Standard Time ((GMT+10:00) Hobart)
     *  -   Tokyo Standard Time ((GMT+09:00) Osaka, Sapporo, Tokyo)
     *  -   Tonga Standard Time ((GMT+13:00) Nuku'alofa)
     *  -   US Eastern Standard Time ((GMT-05:00) Indiana (East))
     *  -   US Mountain Standard Time ((GMT-07:00) Arizona)
     *  -   Vladivostok Standard Time ((GMT+10:00) Vladivostok)
     *  -   W. Australia Standard Time ((GMT+08:00) Perth)
     *  -   W. Central Africa Standard Time ((GMT+01:00) West Central Africa)
     *  -   W. Europe Standard Time ((GMT+01:00) Amsterdam, Berlin, Bern, Rome, Stockholm, Vienna)
     *  -   West Asia Standard Time ((GMT+05:00) Islamabad, Karachi, Tashkent)
     *  -   West Pacific Standard Time ((GMT+10:00) Guam, Port Moresby)
     *  -   Yakutsk Standard Time ((GMT+09:00) Yakutsk)
     *  -   UTC (Universal Coordinated Time)
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC CriticalSection;
    class CARL_PUBLIC TimeZone
    {
        //////////
        //  Can't assign or copy construct
    private:
        TimeZone(const TimeZone &);
        void                        operator = (const TimeZone &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        TimeZone() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~TimeZone() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified time zone with the CARL framework.
         *
         *  A registration of a time zone informs the CARL framework that
         *  a specific time zone exists. From this point on, the application
         *  can use the newly registers time zone in exactly the same manner
         *  as a predefined time zone. Once registered, a time zone
         *  cannot be unregistered.
         *
         *  A time zone cannot be registered if one of its names clashes
         *  with one of the names of some time zone already registered.
         *
         *  @param pTimeZone        The time zone to register.
         *  @exception InconsistentTimeZoneException    If the time zone being
         *                          registered is inconsistent (for example, its
         *                          canonical name is not a member of its "all names" 
         *                          list).
         *  @exception DuplicateTimeZoneException   If one of the time zone
         *                          names clashes with one of the names of another
         *                          time zone registered earlier.
         *  @since 1.0
         */
        static void                 registerInstance(TimeZone * pTimeZone) throw (InconsistentTimeZoneException, DuplicateTimeZoneException);

        /**
         *  Returns a list of all registered time zones.
         *
         *  The returned list contains all time zones registered so far,
         *  including all predefined time zones. Elements in the list are
         *  not guaranteed to be in any specific order.
         *
         *  @return                 The list of all registered time zones.
         *  @since 1.0
         */
        static Array<TimeZone*>     getAllInstances();

        /**
         *  Finds and returns a time zone with the specified name.
         *
         *  Time zone names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the time zone.
         *  @return                 The time zone found.
         *  @exception NoSuchElementError   If a time zone with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static TimeZone *           getInstance(const char * pName);

        /**
         *  Finds and returns a time zone with the specified name.
         *
         *  Time zone names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the time zone.
         *  @return                 The time zone found.
         *  @exception NoSuchElementError   If a time zone with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static TimeZone *           getInstance(const Char * pName);

        /**
         *  Finds and returns a time zone with the specified name.
         *
         *  Time zone names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          time zone.
         *  @return                 The time zone found.
         *  @exception NoSuchElementError   If a time zone with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static TimeZone *           getInstance(const String & name);

        /**
         *  Returns the system time zone.
         *
         *  A ayatem time zone is the preferred time zone for the
         *  platform where the CARL client runs.
         *
         *  @return                 The system time zone.
         *  @since 1.0
         */
        static TimeZone *           getSystem();

        /**
         *  Returns the current default time zone.
         *
         *  When a CARL application starts, the current default time zone
         *  is the same as the system time zone. However, the application
         *  can change the current default time zone if necessary.
         *
         *  @return                 The current default time zone.
         *  @since 1.0
         */
        static TimeZone *           getDefault();

        /**
         *  Changes the current default time zone.
         *
         *  When a CARL application starts, the current default time zone
         *  is the same as the system time zone. However, the application
         *  can change the current default time zone if necessary.
         *
         *  @param pCharacterSet    The time zone to make a new current default.
         *  @since 1.0
         */
        static void                 setDefault(TimeZone * pCharacterSet);

        /**
         *  Returns a list of all names of this time zone.
         *
         *  Elements in the list are not guaranteed to be in any particular order.
         *
         *  @return                 A list of all names of this time zone.
         *  @since 1.0
         */
        virtual Array<String>       getAllNames() const = 0;

        /**
         *  Returns the canonical name of this time zone.
         *
         *  A canonical name of a time zone is a name under which the
         *  time zone is known to standards committees.
         *
         *  @return                 The canonical name of this time zone.
         *  @since 1.0
         */
        virtual String              getCanonicalName() const = 0;

        /**
         *  Returns the preferred name of this time zone.
         *
         *  A preferred name of a time zone is a name which is
         *  historically most frequently used to refer to that
         *  time zone (in MIME headers, in HTML headers, etc.)
         *
         *  @return                 The preferred name of this time zone.
         *  @since 1.0
         */
        virtual String              getPreferredName() const = 0;

        /**
         *  Returns the user-readable display name of this time zone
         *  for the current default locale.
         *
         *  The display name of the time zone does not specify the DST
         *  settings even if the time zone uses DST.
         *
         *  @return                 The display name of this time zone.
         *  @since 1.0
         */
        virtual String              getDisplayName() const = 0;

        /**
         *  Returns the user-readable display name of this time zone
         *  for the specified locale.
         *
         *  The display name of the time zone does not specify the DST
         *  settings even if the time zone uses DST.
         *
         *  @param forLocale        The locale for which the display name of
         *                          the time zone shall be tailored.
         *  @return                 The display name of this time zone.
         *  @since 1.0
         */
        virtual String              getDisplayName(const Locale & forLocale) const = 0;

        /**
         *  Returns the user-readable display name of this time zone
         *  for the current default locale.
         *
         *  The display name of the time zone does reflects the DST settings
         *  appropriate for the specified local date+time.
         *
         *  @param dateTime         The local date+time that shall be used
         *                          to determine the DST portion of the time
         *                          zone's display name.
         *  @return                 The display name of this time zone.
         *  @since 1.0
         */
        virtual String              getDisplayName(const DateTime & dateTime) const = 0;

        /**
         *  Returns the user-readable display name of this time zone
         *  for the specified locale.
         *
         *  The display name of the time zone does reflects the DST settings
         *  appropriate for the specified local date+time.
         *
         *  @param dateTime         The local date+time that shall be used
         *                          to determine the DST portion of the time
         *                          zone's display name.
         *  @param forLocale        The locale for which the display name of
         *                          the time zone shall be tailored.
         *  @return                 The display name of this time zone.
         *  @since 1.0
         */
        virtual String              getDisplayName(const DateTime & dateTime, const Locale & forLocale) const = 0;

        /**
         *  Converts UTC time stamp to local date+time.
         *
         *  An attempt to convert an invalid time stamp to a local date+time
         *  always results in an invalid date+time.
         *
         *  @param timeStamp        The UTC time stamp to convert.
         *  @return                 The local date+time.
         *  @since 1.0
         */
        virtual DateTime            timeToLocal(const TimeStamp & timeStamp) const = 0;

        /**
         *  Converts local date+time + UTC time stamp.
         *
         *  An attempt to convert an invalid local date+time to a UTC time
         *  stamp always results in an invalid time stamp.
         *
         *  @param dateTime         The local date+time to convert.
         *  @return                 The UTC time stamp.
         *  @since 1.0
         */
        virtual TimeStamp           timeFromLocal(const DateTime & dateTime) const = 0;

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };

    //////////
    //  Hashing dates and times
    namespace StandardHashFunctions
    {
        /**
         *  Calculates a hash code for a Date instance.
         *
         *  @param key              The Date instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashDate(const Date & key);

        /**
         *  Calculates a hash code for a Time instance.
         *
         *  @param key              The Time instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashTime(const Time & key);

        /**
         *  Calculates a hash code for a DateTime instance.
         *
         *  @param key              The DateTime instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashDateTime(const DateTime & key);

        /**
         *  Calculates a hash code for a TimeSpan instance.
         *
         *  @param key              The TimeSpan instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashTimeSpan(const TimeSpan & key);

        /**
         *  Calculates a hash code for a TimeStamp instance.
         *
         *  @param key              The TimeStamp instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashTimeStamp(const TimeStamp & key);
    }
}

#endif  //  __CARL_DateTime_hpp_included__
//  End of CARL_DateTime.hpp
