//
//  CARL_Encryption.hpp - Encryption services.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.16 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL encryption framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Encryption.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Encryption_hpp_included__
#define __CARL_Encryption_hpp_included__

namespace carl
{
    /**
     *  @class InconsistentCipherException CARL.hpp
     *  An exception thrown when an attempt is made to register an inconsistent cipher.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC InconsistentCipherException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        InconsistentCipherException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        InconsistentCipherException(const InconsistentCipherException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~InconsistentCipherException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        InconsistentCipherException &   operator = (const InconsistentCipherException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class DuplicateCipherException CARL.hpp
     *  An exception thrown when an attempt is made to register a duplicate cipher.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DuplicateCipherException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        DuplicateCipherException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        DuplicateCipherException(const DuplicateCipherException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~DuplicateCipherException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        DuplicateCipherException &      operator = (const DuplicateCipherException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class InconsistentPaddingAlgorithmException CARL.hpp
     *  An exception thrown when an attempt is made to register an 
     *  inconsistent padding algorithm.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC InconsistentPaddingAlgorithmException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        InconsistentPaddingAlgorithmException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        InconsistentPaddingAlgorithmException(const InconsistentPaddingAlgorithmException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~InconsistentPaddingAlgorithmException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        InconsistentPaddingAlgorithmException & operator = (const InconsistentPaddingAlgorithmException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class DuplicatePaddingAlgorithmException CARL.hpp
     *  An exception thrown when an attempt is made to register a duplicate 
     *  padding algorithm.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DuplicatePaddingAlgorithmException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        DuplicatePaddingAlgorithmException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        DuplicatePaddingAlgorithmException(const DuplicatePaddingAlgorithmException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~DuplicatePaddingAlgorithmException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        DuplicatePaddingAlgorithmException &    operator = (const DuplicatePaddingAlgorithmException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class EncryptedDataCorruptException CARL.hpp
     *  An exception thrown when encrypted data is corrupt and cannot be decrypted.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC EncryptedDataCorruptException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        EncryptedDataCorruptException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        EncryptedDataCorruptException(const EncryptedDataCorruptException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~EncryptedDataCorruptException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        EncryptedDataCorruptException &     operator = (const EncryptedDataCorruptException & src);

        //////////
        //  Operations
    public:
        virtual String          toString() const;
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class InvalidCleartextBlockSizeError CARL.hpp
     *  An exception thrown when a block of an invalid length is presented to encryptor.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC InvalidCleartextBlockSizeError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        InvalidCleartextBlockSizeError();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        InvalidCleartextBlockSizeError(const InvalidCleartextBlockSizeError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~InvalidCleartextBlockSizeError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        InvalidCleartextBlockSizeError &    operator = (const InvalidCleartextBlockSizeError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class InvalidCiphertextBlockSizeError CARL.hpp
     *  An exception thrown when a block of an invalid length is presented to decryptor.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC InvalidCiphertextBlockSizeError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        InvalidCiphertextBlockSizeError();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        InvalidCiphertextBlockSizeError(const InvalidCiphertextBlockSizeError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~InvalidCiphertextBlockSizeError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        InvalidCiphertextBlockSizeError &   operator = (const InvalidCiphertextBlockSizeError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class InvalidKeyLengthError CARL.hpp
     *  An exception thrown when a key of an invalid length is offered to a cipher.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC InvalidKeyLengthError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        InvalidKeyLengthError();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        InvalidKeyLengthError(const InvalidKeyLengthError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~InvalidKeyLengthError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        InvalidKeyLengthError & operator = (const InvalidKeyLengthError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class PaddedDataCorruptException CARL.hpp
     *  An exception thrown when padding data is corrupt and cannot be removed.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC PaddedDataCorruptException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        PaddedDataCorruptException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        PaddedDataCorruptException(const PaddedDataCorruptException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~PaddedDataCorruptException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        PaddedDataCorruptException &        operator = (const PaddedDataCorruptException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class KeyCorruptException CARL.hpp
     *  An exception thrown when the key does not obey required conatraints.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC KeyCorruptException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        KeyCorruptException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        KeyCorruptException(const KeyCorruptException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~KeyCorruptException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        KeyCorruptException &           operator = (const KeyCorruptException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class CryptoKey CARL.hpp
     *  A generic key used by cryptography services.
     *
     *  A cryptographic key is a sequence of bits of a known size that is
     *  used for encrypting and desrypting data.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC CryptoKey
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  Initializes the newly created cryptographic key with a sequence of
         *  bits of the specified length.
         *
         *  The constructor makes a private copy of the key bits, so it
         *  is not necessary to maintain key bits externally after a
         *  key has been created.
         *
         *  @param pBits            An address of the first byte of the sequence
         *                          of bytes that specify the key bits. The sequence
         *                          of bytes must contain enough bytes to provide
         *                          at least \a keyLengthBits bits of the key;
         *                          if it is longer only the leftmost
         *                          \a keyLengthBits bits are used.
         *  @param keyLengthBits    The number of bits in the key.
         *  @since 1.0
         */
        CryptoKey(const void *pBits, UIntSize keyLengthBits);

        /**
         *  The copy constructor.
         *
         *  Initializes a newly created key to be an exact copy of the specified key.
         *
         *  @param src          The key to copy bits from.
         *  @since 1.0
         */
        CryptoKey(const CryptoKey & src);

        /**
         *  The class destructor.
         *
         *  Releases all memory used by the key.
         *
         *  @since 1.0
         */
        virtual ~CryptoKey() { delete [] _pBits; }

        /**
         *  The assignment operator.
         *
         *  Makes a key an exact copy of the specified key.
         *
         *  @param src          The key to copy bits from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        CryptoKey &             operator = (const CryptoKey & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a pointer to the read-only key data.
         *
         *  @return             A pointer to the read-only key data.
         *  @since 1.0
         */
        const void *            getKeyBits() const { return _pBits; }

        /**
         *  Returns the length, in bits, of this key.
         *
         *  @return             The length, in bits, of this key.
         *  @since 1.0
         */
        UIntSize                getKeyLengthBits() const { return _keyLengthBits; }

        //////////
        //  Implementation
    private:
        UInt8 *                 _pBits;
        UIntSize                _keyLengthBits;
    };

    /**
     *  @class SecretKey CARL.hpp
     *  A secret key used by symmetric ciphers.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC SecretKey : public CryptoKey
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  Initializes the newly created secret key with a sequence of
         *  bits of the specified length.
         *
         *  The constructor makes a private copy of the key bits, so it
         *  is not necessary to maintain key bits externally after a
         *  key has been created.
         *
         *  @param pBits            An address of the first byte of the sequence
         *                          of bytes that specify the key bits. The sequence
         *                          of bytes must contain enough bytes to provide
         *                          at least \a keyLengthBits bits of the key;
         *                          if it is longer only the leftmost
         *                          \a keyLengthBits bits are used.
         *  @param keyLengthBits    The number of bits in the key.
         *  @since 1.0
         */
        SecretKey(const void * pBits, UIntSize keyLengthBits) : CryptoKey(pBits, keyLengthBits) {}

        /**
         *  The copy constructor.
         *
         *  Initializes a newly created key to be an exact copy of the specified key.
         *
         *  @param src          The key to copy bits from.
         *  @since 1.0
         */
        SecretKey(const SecretKey & src) : CryptoKey(src) {}

        /**
         *  The class destructor.
         *
         *  Releases all memory used by the key.
         *
         *  @since 1.0
         */
        virtual ~SecretKey() {}

        /**
         *  The assignment operator.
         *
         *  Makes a key an exact copy of the specified key.
         *
         *  @param src          The key to copy bits from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        SecretKey &             operator = (const SecretKey & src);
    };

    /**
     *  @class PublicKey CARL.hpp
     *  A public key used by asymmetric ciphers.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC PublicKey : public CryptoKey
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  Initializes the newly created public key with a sequence of
         *  bits of the specified length.
         *
         *  The constructor makes a private copy of the key bits, so it
         *  is not necessary to maintain key bits externally after a
         *  key has been created.
         *
         *  @param pBits            An address of the first byte of the sequence
         *                          of bytes that specify the key bits. The sequence
         *                          of bytes must contain enough bytes to provide
         *                          at least \a keyLengthBits bits of the key;
         *                          if it is longer only the leftmost
         *                          \a keyLengthBits bits are used.
         *  @param keyLengthBits    The number of bits in the key.
         *  @since 1.0
         */
        PublicKey(const void *pBits, UIntSize keyLengthBits) : CryptoKey(pBits, keyLengthBits) {}

        /**
         *  The copy constructor.
         *
         *  Initializes a newly created key to be an exact copy of the specified key.
         *
         *  @param src          The key to copy bits from.
         *  @since 1.0
         */
        PublicKey(const PublicKey & src) : CryptoKey(src) {}

        /**
         *  The class destructor.
         *
         *  Releases all memory used by the key.
         *
         *  @since 1.0
         */
        virtual ~PublicKey() {}

        /**
         *  The assignment operator.
         *
         *  Makes a key an exact copy of the specified key.
         *
         *  @param src          The key to copy bits from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        PublicKey &             operator = (const PublicKey & src);
    };

    /**
     *  @class PrivateKey CARL.hpp
     *  A public key used by asymmetric ciphers.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC PrivateKey : public CryptoKey
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  Initializes the newly created public key with a sequence of
         *  bits of the specified length.
         *
         *  The constructor makes a private copy of the key bits, so it
         *  is not necessary to maintain key bits externally after a
         *  key has been created.
         *
         *  @param pBits            An address of the first byte of the sequence
         *                          of bytes that specify the key bits. The sequence
         *                          of bytes must contain enough bytes to provide
         *                          at least \a keyLengthBits bits of the key;
         *                          if it is longer only the leftmost
         *                          \a keyLengthBits bits are used.
         *  @param keyLengthBits    The number of bits in the key.
         *  @since 1.0
         */
        PrivateKey(const void *pBits, UIntSize keyLengthBits) : CryptoKey(pBits, keyLengthBits) {}

        /**
         *  The copy constructor.
         *
         *  Initializes a newly created key to be an exact copy of the specified key.
         *
         *  @param src          The key to copy bits from.
         *  @since 1.0
         */
        PrivateKey(const PrivateKey & src) : CryptoKey(src) {}

        /**
         *  The class destructor.
         *
         *  Releases all memory used by the key.
         *
         *  @since 1.0
         */
        virtual ~PrivateKey() {}

        /**
         *  The assignment operator.
         *
         *  Makes a key an exact copy of the specified key.
         *
         *  @param src          The key to copy bits from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        PrivateKey &            operator = (const PrivateKey & src);
    };

    /**
     *  @class KeyPair CARL.hpp
     *  A public/private key pair used by asymmetric ciphers.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC KeyPair
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  Initializes the newly created key pair with the specified public
         *  and private keys.
         *
         *  @param publicKey        The public key of the key pair.
         *  @param privateKey       The private key of the key pair.
         *  @since 1.0
         */
        KeyPair(const PublicKey & publicKey, const PrivateKey & privateKey);

        /**
         *  The copy constructor.
         *
         *  Initializes a newly created key pair to be an exact copy of the specified key pair.
         *
         *  @param src          The key pair to copy public and private keys from.
         *  @since 1.0
         */
        KeyPair(const KeyPair & src);

        /**
         *  The class destructor.
         *
         *  Releases all memory used by the key pair.
         *
         *  @since 1.0
         */
        ~KeyPair() {}

        /**
         *  The assignment operator.
         *
         *  Makes a key pair an exact copy of the specified key pair.
         *
         *  @param src          The key pair to copy public and private keys from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        KeyPair &               operator = (const KeyPair & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a public key of this key pair.
         *
         *  @return             A public key of this key pair.
         *  @since 1.0
         */
        PublicKey               getPublicKey() const { return _publicKey; }

        /**
         *  Returns a private key of this key pair.
         *
         *  @return             A private key of this key pair.
         *  @since 1.0
         */
        PrivateKey              getPrivateKey() const { return _privateKey; }

        //////////
        //  Implementation
    private:
        PublicKey               _publicKey;
        PrivateKey              _privateKey;
    };

    /**
     *  @class Cipher CARL.hpp
     *  A generic cipher.
     *
     *  A cipher is an algorithm for encryption and decryption of data.
     *
     *  Currently, the following ciphers are predefined by 
     *  CARL (the list contains cipher names, with display 
     *  names in brackets after the identifier):
     *
     *  -   NULL (No encryption)
     *  -   Blowfish (Blowfish)
     *  -   RSA (RSA)
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Encryptor;
    class CARL_PUBLIC Decryptor;

    class CARL_PUBLIC Cipher
    {
        //////////
        //  Can't assign or copy construct
    private:
        Cipher(const Cipher & src);
        Cipher &                operator = (const Cipher & src);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        Cipher() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~Cipher() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified cipher with the CARL framework.
         *
         *  A registration of a cipher informs the CARL framework that
         *  a specific cipher exists. From this point on, the application
         *  can use the newly registers cipher in exactly the same manner
         *  as a predefined cipher. Once registered, a cipher
         *  cannot be unregistered.
         *
         *  A cipher cannot be registered if one of its names clashes
         *  with one of the names of some cipher already registered.
         *
         *  @param pCipher          The cipher to register.
         *  @exception InconsistentCipherException  If the cipher
         *                          being registered is inconsistent (for example, 
         *                          its canonical name is not a member of its "all names" 
         *                          list).
         *  @exception DuplicateCipherException If one of the cipher
         *                          names clashes with one of the names of another
         *                          cipher registered earlier.
         *  @since 1.0
         */
        static void                         registerInstance(Cipher * pCipher) throw (InconsistentCipherException, DuplicateCipherException);

        /**
         *  Returns a list of all registered ciphers.
         *
         *  The returned list contains all ciphers registered so far,
         *  including all predefined ciphers. Elements in the list are
         *  not guaranteed to be in any specific order.
         *
         *  @return                 The list of all registered ciphers.
         *  @since 1.0
         */
        static Array<Cipher*>               getAllInstances();

        /**
         *  Finds and returns a cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the cipher.
         *  @return                 The cipher found.
         *  @exception NoSuchElementError   If a cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static Cipher *                     getInstance(const char * pName);

        /**
         *  Finds and returns a cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the cipher.
         *  @return                 The cipher found.
         *  @exception NoSuchElementError   If a cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static Cipher *                     getInstance(const Char * pName);

        /**
         *  Finds and returns a cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          cipher.
         *  @return                 The cipher found.
         *  @exception NoSuchElementError   If a cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static Cipher *                     getInstance(const String & name);

        /**
         *  Returns a list of all names of this cipher.
         *
         *  Elements in the list are not guaranteed to be in any particular order.
         *
         *  @return                 A list of all names of this cipher.
         *  @since 1.0
         */
        virtual Array<String>               getAllNames() const = 0;

        /**
         *  Returns the canonical name of this cipher.
         *
         *  A canonical name of a cipher is a name under which the
         *  cipher is known to standards committees.
         *
         *  @return                 The canonical name of this cipher.
         *  @since 1.0
         */
        virtual String                      getCanonicalName() const = 0;

        /**
         *  Returns the preferred name of this cipher.
         *
         *  A preferred name of a cipher is a name which is
         *  historically most frequently used to refer to that
         *  cipher (in MIME headers, in HTML headers, etc.)
         *
         *  @return                 The preferred name of this cipher.
         *  @since 1.0
         */
        virtual String                      getPreferredName() const = 0;

        /**
         *  Returns the user-readable display name of this cipher
         *  for the current default locale.
         *
         *  @return                 The display name of this cipher.
         *  @since 1.0
         */
        virtual String                      getDisplayName() const = 0;

        /**
         *  Returns the user-readable display name of this cipher
         *  for the specified default locale.
         *
         *  @param forLocale        The locale for which the display name of
         *                          the cipher shall be tailored.
         *  @return                 The display name of this cipher.
         *  @since 1.0
         */
        virtual String                      getDisplayName(const Locale & forLocale) const = 0;

        /**
         *  Creates and returns a new encryptor for this cipher that
         *  uses the specified key.
         *
         *  Unlike ciphers, which are stock objects, encryptors are created
         *  as necessary. The client is responsible for deleting the
         *  encryptor when it is no longer needed.
         *
         *  @param key              The key that shall be used for encryption.
         *  @return                 A new encryptor for the specified key.
         *  @since 1.0
         */
        virtual Encryptor *                 createEncryptor(const CryptoKey & key) const throw(KeyCorruptException) = 0;

        /**
         *  Creates and returns a new decryptor for this cipher that
         *  uses the specified key.
         *
         *  Unlike ciphers, which are stock objects, decryptors are created
         *  as necessary. The client is responsible for deleting the
         *  decryptor when it is no longer needed.
         *
         *  @param key              The key that shall be used for decryption.
         *  @return                 A new decryptor for the specified key.
         *  @since 1.0
         */
        virtual Decryptor *                 createDecryptor(const CryptoKey & key) const throw(KeyCorruptException) = 0;

        //////////
        //  Implementation
    private:
        static CriticalSection *            _getStaticGuard();
    };

    /**
     *  @class SymmetricCipher CARL.hpp
     *  A generic symmetric cipher.
     *
     *  A symmetric cipher is a cypher that uses the same key, known as a
     *  "secret key", for both encryption and decryption of data.
     *
     *  Currently, the following symmetric ciphers are predefined by 
     *  CARL (the list contains symmetric cipher names, with display 
     *  names in brackets after the identifier):
     *
     *  -   NULL (No encryption)
     *  -   Blowfish (Blowfish)
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC SymmetricCipher : public virtual Cipher
    {
        //////////
        //  Can't assign or copy construct
    private:
        SymmetricCipher(const SymmetricCipher & src);
        SymmetricCipher &                operator = (const SymmetricCipher & src);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        SymmetricCipher() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~SymmetricCipher() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns a list of all registered symmetric ciphers.
         *
         *  The returned list contains all symmetric ciphers registered so far,
         *  including all predefined symmetric ciphers. Elements in the list are
         *  not guaranteed to be in any specific order.
         *
         *  @return                 The list of all registered symmetric ciphers.
         *  @since 1.0
         */
        static Array<SymmetricCipher*>  getAllInstances();

        /**
         *  Finds and returns a symmetric cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the symmetric cipher.
         *  @return                 The symmetric cipher found.
         *  @exception NoSuchElementError   If a symmetric cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static SymmetricCipher *        getInstance(const char * pName) { return getInstance(String(pName)); }

        /**
         *  Finds and returns a symmetric cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the symmetric cipher.
         *  @return                 The symmetric cipher found.
         *  @exception NoSuchElementError   If a symmetric cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static SymmetricCipher *        getInstance(const Char * pName) { return getInstance(String(pName)); }

        /**
         *  Finds and returns a symmetric cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          symmetric cipher.
         *  @return                 The symmetric cipher found.
         *  @exception NoSuchElementError   If a symmetric cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static SymmetricCipher *        getInstance(const String & name);

        /**
         *  Returns a list of key lengths supported by this symmetric cipher.
         *
         *  Returned key lengths are always expressed in bits. Elements in the 
         *  returned list are not guaranteed to be in any particular order.
         *
         *  @return                 A list of key lengths supported by this symmetric cipher.
         *  @since 1.0
         */
        virtual Array<UIntSize>         getSupportedKeyLenghts() const = 0;

        /**
         *  Generates a random secret key of the specified length from the specified seed.
         *
         *  This method allows generating random secret keys in a repeatable manner.
         *
         *  @param seed             The seed for the random number generator.
         *  @param keyLengthBits    The size, in bits, of the key to generate.
         *  @return                 The generated secret key.
         *  @since 1.0
         */
        virtual SecretKey               generateKey(UInt64 seed, UIntSize keyLengthBits) const = 0;

        /**
         *  Generates a random secret key of the specified length.
         *
         *  @param keyLengthBits    The size, in bits, of the key to generate.
         *  @return                 The generated secret key.
         *  @since 1.0
         */
        virtual SecretKey               generateRandomKey(UIntSize keyLengthBits) const = 0;
    };

    /**
     *  @class AsymmetricCipher CARL.hpp
     *  A generic asymmetric cipher.
     *
     *  An asymmetric cipher is a cypher that uses two different keys, known as a
     *  "public/private key pair", for encryption and decryption of data.
     *
     *  Currently, the following asymmetric ciphers are predefined by 
     *  CARL (the list contains asymmetric cipher names, with display 
     *  names in brackets after the identifier):
     *
     *  -   RSA (RSA)
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC AsymmetricCipher : public virtual Cipher
    {
        //////////
        //  Can't assign or copy construct
    private:
        AsymmetricCipher(const AsymmetricCipher & src);
        AsymmetricCipher &              operator = (const AsymmetricCipher & src);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        AsymmetricCipher() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~AsymmetricCipher() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns a list of all registered asymmetric ciphers.
         *
         *  The returned list contains all asymmetric ciphers registered so far,
         *  including all predefined asymmetric ciphers. Elements in the list are
         *  not guaranteed to be in any specific order.
         *
         *  @return                 The list of all registered asymmetric ciphers.
         *  @since 1.0
         */
        static Array<AsymmetricCipher*> getAllInstances();

        /**
         *  Finds and returns an asymmetric cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the asymmetric cipher.
         *  @return                 The asymmetric cipher found.
         *  @exception NoSuchElementError   If an asymmetric cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static AsymmetricCipher *       getInstance(const char * pName) { return getInstance(String(pName)); }

        /**
         *  Finds and returns an asymmetric cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the asymmetric cipher.
         *  @return                 The asymmetric cipher found.
         *  @exception NoSuchElementError   If an asymmetric cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static AsymmetricCipher *       getInstance(const Char * pName) { return getInstance(String(pName)); }

        /**
         *  Finds and returns an asymmetric cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          asymmetric cipher.
         *  @return                 The asymmetric cipher found.
         *  @exception NoSuchElementError   If an asymmetric cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static AsymmetricCipher *       getInstance(const String & name);

        /**
         *  Returns a list of key pair lengths supported by this asymmetric cipher.
         *
         *  Returned key lengths are always expressed in bits. Elements in the returned 
         *  list are not guaranteed to be in any particular order.
         *  For each element, the first item of a pair constitutes the public key
         *  length and the second item of a pair constitutes the private key length.
         *
         *  @return                 A list of key pair lengths supported by 
         *                          this symmetric cipher.
         *  @since 1.0
         */
        virtual Array<Pair<UIntSize,UIntSize> > getSupportedKeyPairLenghts() const = 0;

        /**
         *  Generates a random key pair of the specified lengths from the specified seed.
         *
         *  This method allows generating random secret key pairs in a repeatable manner.
         *
         *  @param seed             The seed for the random number generator.
         *  @param publicKeyLengthBits  The size, in bits, of the public key to generate.
         *  @param privateKeyLengthBits The size, in bits, of the private key to generate.
         *  @return                 The generated public/private key pair.
         *  @since 1.0
         */
        virtual KeyPair                 generateKeyPair(UInt64 seed, UIntSize publicKeyLengthBits, UIntSize privateKeyLengthBits) const = 0;

        /**
         *  Generates a random key pair of the specified lengths.
         *
         *  @param publicKeyLengthBits  The size, in bits, of the public key to generate.
         *  @param privateKeyLengthBits The size, in bits, of the private key to generate.
         *  @return                 The generated public/private key pair.
         *  @since 1.0
         */
        virtual KeyPair                 generateRandomKeyPair(UIntSize publicKeyLengthBits, UIntSize privateKeyLengthBits) const = 0;
    };

    /**
     *  @class StreamCipher CARL.hpp
     *  A generic stream cipher.
     *
     *  A stream cipher encrypts and decrypts data bit-by-bit; for each
     *  input bit exactly one output bit is produced.
     *
     *  Currently, the following sream ciphers are predefined by 
     *  CARL (the list contains sream cipher names, with display 
     *  names in brackets after the identifier):
     *
     *  -   NULL (No encryption)
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC StreamCipher : public virtual Cipher
    {
        //////////
        //  Can't assign or copy construct
    private:
        StreamCipher(const StreamCipher & src);
        StreamCipher &                  operator = (const StreamCipher & src);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        StreamCipher() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~StreamCipher() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns a list of all registered stream ciphers.
         *
         *  The returned list contains all stream ciphers registered so far,
         *  including all predefined stream ciphers. Elements in the list are
         *  not guaranteed to be in any specific order.
         *
         *  @return                 The list of all registered stream ciphers.
         *  @since 1.0
         */
        static Array<StreamCipher*>     getAllInstances();

        /**
         *  Finds and returns a stream cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the stream cipher.
         *  @return                 The stream cipher found.
         *  @exception NoSuchElementError   If a stream cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static StreamCipher *           getInstance(const char * pName) { return getInstance(String(pName)); }

        /**
         *  Finds and returns a stream cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the stream cipher.
         *  @return                 The stream cipher found.
         *  @exception NoSuchElementError   If a stream cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static StreamCipher *           getInstance(const Char * pName) { return getInstance(String(pName)); }

        /**
         *  Finds and returns a stream cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          stream cipher.
         *  @return                 The stream cipher found.
         *  @return                 The streamcipher found.
         *  @exception NoSuchElementError   If a stream cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static StreamCipher *           getInstance(const String & name);
    };

    /**
     *  @class BlockCipher CARL.hpp
     *  A generic block cipher.
     *
     *  A block cipher encrypts and decrypts blocks of bits, for each
     *  input block exactly one output block is produced, possibly of
     *  different length.
     *
     *  Currently, the following block ciphers are predefined by 
     *  CARL (the list contains block cipher names, with display 
     *  names in brackets after the identifier):
     *
     *  -   Blowfish (Blowfish)
     *  -   RSA (RSA)
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC BlockCipher : public virtual Cipher
    {
        //////////
        //  Can't assign or copy construct
    private:
        BlockCipher(const BlockCipher & src);
        BlockCipher &                   operator = (const BlockCipher & src);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        BlockCipher() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~BlockCipher() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns a list of all registered block ciphers.
         *
         *  The returned list contains all block ciphers registered so far,
         *  including all predefined block ciphers. Elements in the list are
         *  not guaranteed to be in any specific order.
         *
         *  @return                 The list of all registered block ciphers.
         *  @since 1.0
         */
        static Array<BlockCipher*>      getAllInstances();

        /**
         *  Finds and returns a block cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the block cipher.
         *  @return                 The block cipher found.
         *  @exception NoSuchElementError   If a block cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static BlockCipher *            getInstance(const char * pName) { return getInstance(String(pName)); }

        /**
         *  Finds and returns a block cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the block cipher.
         *  @return                 The block cipher found.
         *  @exception NoSuchElementError   If a block cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static BlockCipher *            getInstance(const Char * pName) { return getInstance(String(pName)); }

        /**
         *  Finds and returns a block cipher with the specified name.
         *
         *  Cipher names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          block cipher.
         *  @return                 The block cipher found.
         *  @return                 The block cipher found.
         *  @exception NoSuchElementError   If a block cipher with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static BlockCipher *            getInstance(const String & name);

        /**
         *  Returns a list of block sizes supported by this block cipher.
         *
         *  Returned block sizes are always expressed in bits. Elements in the 
         *  returned list are not guaranteed to be in any particular order.
         *
         *  @return                 A list of block sizes supported by this block cipher.
         *  @since 1.0
         */
        virtual Array<UIntSize>         getSupportedBlockLenghts() const = 0;
    };

    /**
     *  @class Encryptor CARL.hpp
     *  A generic encryptor.
     *
     *  An encryptor is a part of a cipher that can convert plain
     *  data (also known as cleartext) into encrypted form (also
     *  known as ciphertext).
     *
     *  Unlike ciphers, which are stock objects, encryptors are
     *  created when a client needs to encrypt something and destroyed
     *  when no longer needed. An Encryptor instance, once created,
     *  uses a specific key for all encryption activities.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Encryptor
    {
        //////////
        //  Can't assign or copy construct
    private:
        Encryptor(const Encryptor & src);
        Encryptor &                     operator = (const Encryptor & src);

        //////////
        //  Construction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        Encryptor() {}

        //////////
        //  Destruction - public, as the client will have to destroy instances
    public:
        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Encryptor() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns a cipher to which this encryptor belongs.
         *
         *  @return                 A cipher to which this encryptor belongs.
         *  @since 1.0
         */
        virtual Cipher *                getCipher() const = 0;
    };

    /**
     *  @class Decryptor CARL.hpp
     *  A generic decryptor.
     *
     *  A decryptor is a part of a cipher that can recover plain
     *  data (also known as cleartext) from encrypted form (also
     *  known as ciphertext).
     *
     *  Unlike ciphers, which are stock objects, decryptors are
     *  created when a client needs to decrypt something and destroyed
     *  when no longer needed. A Decryptor instance, once created,
     *  uses a specific key for all decryption activities.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Decryptor
    {
        //////////
        //  Can't assign or copy construct
    private:
        Decryptor(const Decryptor & src);
        Decryptor &                     operator = (const Decryptor & src);

        //////////
        //  Construction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        Decryptor() {}

        //////////
        //  Destruction - public, as the client will have to destroy instances
    public:
        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Decryptor() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns a cipher to which this decryptor belongs.
         *
         *  @return                 A cipher to which this decryptor belongs.
         *  @since 1.0
         */
        virtual Cipher *                getCipher() const = 0;
    };

    /**
     *  @class StreamEncryptor CARL.hpp
     *  A generic stream encryptor.
     *
     *  A stream encryptor yields one ciphertext bit for one cleartext bit.
     *
     *  Unlike ciphers, which are stock objects, encryptors are
     *  created when a client needs to encrypt something and destroyed
     *  when no longer needed. An Encryptor instance, once created,
     *  uses a specific key for all encryption activities.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC StreamEncryptor : public Encryptor
    {
        //////////
        //  Can't assign or copy construct
    private:
        StreamEncryptor(const StreamEncryptor & src);
        StreamEncryptor &               operator = (const StreamEncryptor & src);

        //////////
        //  Construction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        StreamEncryptor() {}

        //////////
        //  Destruction - public, as the client will have to destroy instances
    public:
        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~StreamEncryptor() {}

        //////////
        //  Operations
    public:
        /**
         *  Encrypts one bit and yields the encryption result.
         *
         *  @param cleartextBit     Contains the bit to encrypt in the
         *                          lowest bit position.
         *  @return                 The encrypted bit (0 or 1).
         *  @since 1.0
         */
        virtual int                     encryptBit(int cleartextBit) = 0;
    };

    /**
     *  @class StreamDecryptor CARL.hpp
     *  A generic stream decryptor.
     *
     *  A stream decryptor yields one cleartext bit for one ciphertext bit.
     *
     *  Unlike ciphers, which are stock objects, decryptors are
     *  created when a client needs to decrypt something and destroyed
     *  when no longer needed. A Decryptor instance, once created,
     *  uses a specific key for all decryption activities.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC StreamDecryptor : public Decryptor
    {
        //////////
        //  Can't assign or copy construct
    private:
        StreamDecryptor(const StreamDecryptor & src);
        StreamDecryptor &               operator = (const StreamDecryptor & src);

        //////////
        //  Construction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        StreamDecryptor() {}

        //////////
        //  Destruction - public, as the client will have to destroy instances
    public:
        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~StreamDecryptor() {}

        //////////
        //  Operations
    public:
        /**
         *  Decrypts one bit and yields the decryption result.
         *
         *  @param ciphertextBit    Contains the bit to decrypt in the
         *                          lowest bit position.
         *  @return                 The encrypted bit (0 or 1).
         *  @exception EncryptedDataCorruptException    If the specified
         *                          ciphertext bit breaks the integrity
         *                          constraints imposed by the cipher
         *                          upon the ciphertext data.
         *  @since 1.0
         */
        virtual int                     decryptBit(int ciphertextBit) throw(EncryptedDataCorruptException) = 0;
    };

    /**
     *  @class BlockEncryptor CARL.hpp
     *  A generic block encryptor.
     *
     *  A block encryptor yields one ciphertext block of bits for one 
     *  cleartext block of bits, the result possibly being different 
     *  size from the original.
     *
     *  Unlike ciphers, which are stock objects, encryptors are
     *  created when a client needs to encrypt something and destroyed
     *  when no longer needed. An Encryptor instance, once created,
     *  uses a specific key for all encryption activities.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC BlockEncryptor : public Encryptor
    {
        //////////
        //  Can't assign or copy construct
    private:
        BlockEncryptor(const BlockEncryptor & src);
        BlockEncryptor &                operator = (const BlockEncryptor & src);

        //////////
        //  Construction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        BlockEncryptor() {}

        //////////
        //  Destruction - public, as the client will have to destroy instances
    public:
        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~BlockEncryptor() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns the size of the ciphertext block for the 
         *  specified cleartext block.
         *
         *  @param pCleartextBits   A pointer to the start of the cleartext 
         *                          block. The cleartext block must be at least 
         *                          \a cleartextBitsCount bits long; if it is 
         *                          longer then only the first \a cleartextBitsCount
         *                          bits are considered.
         *  @param cleartextBitsCount   The size, in bits, of the cleartext block.
         *  @return                 The size, in bits, of the ciphertext block
         *                          that would result from encrypting the
         *                          specified cleartext block.
         *  @since 1.0
         */
        virtual UIntSize                getCiphertextSizeForCleartext(const void * pCleartextBits, UIntSize cleartextBitsCount) = 0;

        /**
         *  Encrypts a block of data.
         *
         *  @param pCleartextBits   A pointer to the start of the cleartext 
         *                          block. The cleartext block must be at least 
         *                          \a cleartextBitsCount bits long; if it is 
         *                          longer then only the first \a cleartextBitsCount
         *                          bits are considered.
         *  @param cleartextBitsCount   The size, in bits, of the cleartext block.
         *  @param pCiphertextBits  The address of the buffer where the encrypted
         *                          ciphertext block is placed.
         *  @return                 The size, in bits, of the ciphertext block
         *                          resulting from encrypting the specified 
         *                          cleartext block.
         *  @since 1.0
         */
        virtual UIntSize                encryptBlock(const void * pCleartextBits, UIntSize cleartextBitsCount, void * pCiphertextBits) = 0;
    };

    /**
     *  @class BlockDecryptor CARL.hpp
     *  A generic block decryptor.
     *
     *  A block decryptor yields one cleartext block of bits for one 
     *  ciphertext block of bits, the result possibly being different 
     *  size from the original.
     *
     *  Unlike ciphers, which are stock objects, decryptors are
     *  created when a client needs to decrypt something and destroyed
     *  when no longer needed. A Decryptor instance, once created,
     *  uses a specific key for all decryption activities.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC BlockDecryptor : public Decryptor
    {
        //////////
        //  Can't assign or copy construct
    private:
        BlockDecryptor(const BlockDecryptor & src);
        BlockDecryptor &                operator = (const BlockDecryptor & src);

        //////////
        //  Construction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        BlockDecryptor() {}

        //////////
        //  Destruction - public, as the client will have to destroy instances
    public:
        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~BlockDecryptor() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns the size of the cleartext block for the 
         *  specified ciphertext block.
         *
         *  @param pCiphertextBits  A pointer to the start of the ciphertext 
         *                          block. The ciphertext block must be at least 
         *                          \a ciphertextBitsCount bits long; if it is 
         *                          longer then only the first \a ciphertextBitsCount
         *                          bits are considered.
         *  @param ciphertextBitsCount  The size, in bits, of the ciphertext block.
         *  @return                 The size, in bits, of the cleartext block
         *                          that would result from decrypting the
         *                          specified ciphertext block.
         *  @since 1.0
         */
        virtual UIntSize                getCleartextSizeForCiphertext(const void * pCiphertextBits, UIntSize ciphertextBitsCount) = 0;

        /**
         *  Decrypts a block of data.
         *
         *  @param pCiphertextBits  A pointer to the start of the ciphertext
         *                          block. The ciphertext block must be at least 
         *                          \a ciphertextBitsCount bits long; if it is 
         *                          longer then only the first \a ciphertextBitsCount
         *                          bits are considered.
         *  @param ciphertextBitsCount  The size, in bits, of the ciphertext block.
         *  @param pCleartextBits   The address of the buffer where the decrypted
         *                          cleartext block is placed.
         *  @return                 The size, in bits, of the cleartext block
         *                          resulting from decrypting the specified 
         *                          ciphertext block.
         *  @exception EncryptedDataCorruptException    If the specified
         *                          ciphertext block breaks the integrity
         *                          constraints imposed by the cipher
         *                          upon the ciphertext data.
         *  @since 1.0
         */
        virtual UIntSize                decryptBlock(const void * pCiphertextBits, UIntSize ciphertextBitsCount, void * pCleartextBits) throw(EncryptedDataCorruptException) = 0;
    };

    /**
     *  @class PaddingAlgorithm CARL.hpp
     *  An abstract padding algorithm.
     *
     *  A padding algorithm provides a way to pad variable-length blocks to a
     *  specified fixed length. It is used with block ciphers when the length
     *  of an encrypted data is not a proper multiple of the block size
     *  supported by the block cipher
     *
     *  Currently, the following padding algorithms are predefined by 
     *  CARL (the list contains padding algorithm names, with display 
     *  names in brackets after the identifier):
     *
     *  -   PKCS5 (PKCS5)
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC PaddingAlgorithm
    {
        //////////
        //  Can't assign or copy construct
    private:
        PaddingAlgorithm(const PaddingAlgorithm & src);
        PaddingAlgorithm &                operator = (const PaddingAlgorithm & src);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        PaddingAlgorithm() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~PaddingAlgorithm() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified padding algorithm with the CARL framework.
         *
         *  A registration of a padding algorithm informs the CARL framework that
         *  a specific padding algorithm exists. From this point on, the application
         *  can use the newly registers padding algorithm in exactly the same manner
         *  as a predefined padding algorithm. Once registered, a padding algorithm
         *  cannot be unregistered.
         *
         *  A padding algorithm cannot be registered if one of its names clashes
         *  with one of the names of some padding algorithm already registered.
         *
         *  @param pPaddingAlgorithm    The padding algorithm to register.
         *  @exception InconsistentPaddingAlgorithmException    If the padding 
         *                          algorithm being registered is inconsistent (for example, 
         *                          its canonical name is not a member of its "all names" 
         *                          list).
         *  @exception DuplicatePaddingAlgorithmException   If one of the padding 
         *                          algorithm names clashes with one of the names of another
         *                          padding algorithm registered earlier.
         *  @since 1.0
         */
        static void                         registerInstance(PaddingAlgorithm * pPaddingAlgorithm) throw (InconsistentPaddingAlgorithmException, DuplicatePaddingAlgorithmException);

        /**
         *  Returns a list of all registered padding algorithms.
         *
         *  The returned list contains all padding algorithms registered so far,
         *  including all predefined padding algorithms. Elements in the list are
         *  not guaranteed to be in any specific order.
         *
         *  @return                 The list of all registered padding algorithms.
         *  @since 1.0
         */
        static Array<PaddingAlgorithm*>     getAllInstances();

        /**
         *  Finds and returns a padding algorithm with the specified name.
         *
         *  Padding algorithm names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ASCII string
         *                          representing the name of the padding algorithm.
         *  @return                 The padding algorithm found.
         *  @exception NoSuchElementError   If a padding algorithm with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static PaddingAlgorithm *           getInstance(const char * pName);

        /**
         *  Finds and returns a padding algorithm with the specified name.
         *
         *  Padding algorithm names are case-insensitive.
         *
         *  @param pName            A pointer to nul-terminated ISO-10646 string
         *                          representing the name of the padding algorithm.
         *  @return                 The padding algorithm found.
         *  @exception NoSuchElementError   If a padding algorithm with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static PaddingAlgorithm *           getInstance(const Char * pName);

        /**
         *  Finds and returns a padding algorithm with the specified name.
         *
         *  Padding algorithm names are case-insensitive.
         *
         *  @param name             A string representing the name of the 
         *                          padding algorithm.
         *  @return                 The padding algorithm found.
         *  @exception NoSuchElementError   If a padding algorithm with the specified
         *                          name does not exist.
         *  @since 1.0
         */
        static PaddingAlgorithm *           getInstance(const String & name);

        /**
         *  Returns a list of all names of this padding algorithm.
         *
         *  Elements in the list are not guaranteed to be in any particular order.
         *
         *  @return                 A list of all names of this padding algorithm.
         *  @since 1.0
         */
        virtual Array<String>               getAllNames() const = 0;

        /**
         *  Returns the canonical name of this padding algorithm.
         *
         *  A canonical name of a padding algorithm is a name under which the
         *  padding algorithm is known to standards committees.
         *
         *  @return                 The canonical name of this padding algorithm.
         *  @since 1.0
         */
        virtual String                      getCanonicalName() const = 0;

        /**
         *  Returns the preferred name of this padding algorithm.
         *
         *  A preferred name of a padding algorithm is a name which is
         *  historically most frequently used to refer to that
         *  padding algorithm (in MIME headers, in HTML headers, etc.)
         *
         *  @return                 The preferred name of this padding algorithm.
         *  @since 1.0
         */
        virtual String                      getPreferredName() const = 0;

        /**
         *  Returns the user-readable display name of this padding algorithm
         *  for the current default locale.
         *
         *  @return                 The display name of this padding algorithm.
         *  @since 1.0
         */
        virtual String                      getDisplayName() const = 0;

        /**
         *  Returns the user-readable display name of this padding algorithm
         *  for the specified default locale.
         *
         *  @param forLocale        The locale for which the display name of
         *                          the padding algorithm shall be tailored.
         *  @return                 The display name of this padding algorithm.
         *  @since 1.0
         */
        virtual String                      getDisplayName(const Locale & forLocale) const = 0;

        /**
         *  Determines the number of padding bits required for a given data.
         *
         *  @param pDataBits        The data that may need to be padded. The
         *                          data must be at least \a dataBitsCount bits
         *                          long; if it is longer then only the first
         *                          \a dataBitsCount bits are considered.
         *  @param dataBitsCount    The size, in bits, of the data being padded.
         *  @param blockSizeBits    The size, in bits, to which the data shall
         *                          be padded.
         *  @returns                The number of padding bits that must be
         *                          appended to the data in order to make its
         *                          size a multiple of \a blockSizeBits. If
         *                          the return value is 0, no padding is
         *                          necessary for the specified data.
         *  @since 1.0
         */
        virtual UIntSize                    getPaddingBitsRequiredForData(const void * pDataBits, UIntSize dataBitsCount, UIntSize blockSizeBits) const = 0;

        /**
         *  Generates and stores necessary padding bits for the specified data.
         *
         *  @param pDataBits        The data that may need to be padded. The
         *                          data must be at least \a dataBitsCount bits
         *                          long; if it is longer then only the first
         *                          \a dataBitsCount bits are considered.
         *  @param dataBitsCount    The size, in bits, of the data being padded.
         *  @param blockSizeBits    The size, in bits, to which the data shall
         *                          be padded.
         *  @param pPaddingBits     The output buffer where generated padding bits 
         *                          are stored. The padding algorithm always
         *                          assumes that there is enough space in the
         *                          output buffer.
         *  @returns                The number of padding bits stored to the
         *                          output buffer.
         *  @since 1.0
         */
        virtual UIntSize                    createPaddingBitsForData(const void * pDataBits, UIntSize dataBitsCount, UIntSize blockSizeBits, void * pPaddingBits) const = 0;

        /**
         *  Determine how much padding a block of padded data has.
         *
         *  @param pDataBits        The data that may need to be un-padded. The
         *                          data must be at least \a dataBitsCount bits
         *                          long; if it is longer then only the first
         *                          \a dataBitsCount bits are considered.
         *  @param dataBitsCount    The size, in bits, of the data being un-padded.
         *  @param blockSizeBits    The size, in bits, to which the data has been
         *                          padded.
         *  @return                 The number of tailing bits in the specified
         *                          block that constitute padding.
         *  @exception PaddedDataCorruptException   If the amount of padding in
         *                          the specified data block cannot be determined.
         *  @since 1.0
         */
        virtual UIntSize                    getPaddingBitsStoredInData(const void * pDataBits, UIntSize dataBitsCount, UIntSize blockSizeBits) const throw(PaddedDataCorruptException) = 0;

        //////////
        //  Implementation
    private:
        static CriticalSection *            _getStaticGuard();
    };

    /**
     *  @class EncryptOutputStream CARL.hpp
     *  An output stream that performs on-the-fly encryption.
     *
     *  An encrypt output stream does its work in collaboration with two
     *  other objects: an encryptor that performs the actual data encryption
     *  and a target stream, where all encrypted data is forwarded. If the
     *  stream uses block encryption, then it also needs a padding algorithm
     *  that will be used if the amount of transferred data is not an exact
     *  multiple of the cipher's supported block size.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC EncryptOutputStream : public IByteOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        EncryptOutputStream(const EncryptOutputStream &);
        void                    operator = (const EncryptOutputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes the newly created output stream for stream encryption
         *  by associating it with the specified target stream and stream
         *  encryptor. The newly constructed stream is initially open.
         *
         *  @param pTarget          The stream where encrypted data are forwarded.
         *  @param pStreamEncryptor The stream encryptor to use for data encryption.
         *  @param autoDeleteStreamEncryptor    True iff this stream shall take 
         *                          ownership of the encryptor, in which case the
         *                          encryptor is destroyed automatically when
         *                          the stream is destroyed.
         *  @since 1.0
         */
        EncryptOutputStream(IByteOutput * pTarget, StreamEncryptor * pStreamEncryptor, bool autoDeleteStreamEncryptor = false);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created output stream for block encryption
         *  by associating it with the specified target stream, block
         *  encryptor and padding algorithm. The newly constructed stream 
         *  is initially open.
         *
         *  @param pTarget          The stream where encrypted data are forwarded.
         *  @param pBlockEncryptor  The block encryptor to use for data encryption.
         *  @param pPaddingAlgorithm    The padding algorithm to use for partial blocks.
         *  @param autoDeleteBlockEncryptor True iff this stream shall take 
         *                          ownership of the encryptor, in which case the
         *                          encryptor is destroyed automatically when
         *                          the stream is destroyed.
         *  @since 1.0
         */
        EncryptOutputStream(IByteOutput * pTarget, BlockEncryptor * pBlockEncryptor, PaddingAlgorithm * pPaddingAlgorithm, bool autoDeleteBlockEncryptor = false);

        /**
         *  The class destructor.
         *
         *  Closes the output stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~EncryptOutputStream() { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this output stream is still open.
         *
         *  @return                 True if this output stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _pTarget != NULL; }

        /**
         *  Closes this output stream.
         *
         *  Any data written to the output stream but not yet encrypted is
         *  encrypted and forwarded to the associated target stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteOutput
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  If there is any data that has been written to this output stream
         *  and not yet encrypted, this data is encrypted immediately and
         *  forwarded to the associated target stream, then the target
         *  stream is flushed as well.
         *
         *  This method can only be called on open streams.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException);

        /**
         *  Writes a single byte to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param byte             The byte to write to this stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(UInt8 byte) throw(IOException);

        /**
         *  Writes a sequence of bytes to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The address of the first byte of the sequence
         *                          that shall be written to the stream. The
         *                          sequence must be at least \a bytesToWrite
         *                          bytes long; if it is longer then only the
         *                          forst \a bytesToWrite bytes are written.
         *  @param bytesToWrite     The number of bytes to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException);

        //////////
        //  Implementation
    private:
        IByteOutput *           _pTarget;

        //  If a stream encryptor is used...
        StreamEncryptor *       _pStreamEncryptor;
        bool                    _autoDeleteStreamEncryptor;

        //  If a block encryptor is used...
        BlockEncryptor *        _pBlockEncryptor;
        PaddingAlgorithm *      _pPaddingAlgorithm;
        bool                    _autoDeleteBlockEncryptor;
        Array<UInt8>            _inputBuffer;
        UIntSize                _bitsReadyInInputBuffer;
        Array<UInt8>            _outputBuffer;
        UIntSize                _bitsReadyInOutputBuffer;
        UIntSize                _largestSupportedInputBlockSize;

        void                    _addBitToInputBuffer(int bit);
        void                    _padLastBlockAndFlush();
        void                    _writeUleb(UIntSize value);
    };

    /**
     *  @class DecryptInputStream CARL.hpp
     *  An input stream that performs on-the-fly decryption.
     *
     *  A decrypt input stream does its work in collaboration with two
     *  other objects: a source stream, where encrypted data is obtained
     *  from and a decryptor that performs the actual data decryption. If the
     *  stream uses block decryption, then it also needs a padding algorithm
     *  that will be used to remove any padding from the decrypted data.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DecryptInputStream : public IByteInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        DecryptInputStream(const DecryptInputStream &);
        void                    operator = (const DecryptInputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes the newly created input stream for stream decryption
         *  by associating it with the specified source stream and decryptor. 
         *  The newly constructed stream is initially open.
         *
         *  @param pSource          The stream where encrypted data is obtained from.
         *  @param pStreamDecryptor The decryptor to use for data decryption.
         *  @param autoDeleteStreamDecryptor    True iff this stream shall take 
         *                          ownership of the decryptor, in which case the
         *                          decryptor is destroyed automatically when
         *                          the stream is destroyed.
         *  @since 1.0
         */
        DecryptInputStream(IByteInput * pSource, StreamDecryptor * pStreamDecryptor, bool autoDeleteStreamDecryptor = false);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created input stream for block decryption
         *  by associating it with the specified source stream, decryptor and
         *  padding algorithm. The newly constructed stream is initially open.
         *
         *  @param pSource          The stream where encrypted data is obtained from.
         *  @param pBlockDecryptor  The decryptor to use for data decryption.
         *  @param pPaddingAlgorithm    The padding alrithm to use for removing
         *                          padding from decrypted data.
         *  @param autoDeleteBlockDecryptor True iff this stream shall take 
         *                          ownership of the decryptor, in which case the
         *                          decryptor is destroyed automatically when
         *                          the stream is destroyed.
         *  @since 1.0
         */
        DecryptInputStream(IByteInput * pSource, BlockDecryptor * pBlockDecryptor, PaddingAlgorithm * pPaddingAlgorithm, bool autoDeleteBlockDecryptor = false);

        /**
         *  The class destructor.
         *
         *  Closes the input stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~DecryptInputStream() { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this input stream is still open.
         *
         *  @return                 True if this input stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _pSource != NULL; }

        /**
         *  Closes this input stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteInput
    public:
        /**
         *  Returns the number of bytes that can be read from this input 
         *  stream without blocking.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The number of bytes that can be read from this
         *                          input stream without blocking.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        available() throw(IOException);

        /**
         *  Reads and returns the next byte or -1 if an end of stream is reached.
         *
         *  @return                 The next input byte, or -1 if an end of stream is reached.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual int             read() throw(IOException);

        /**
         *  Reads a sequence of bytes into the specified buffer.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is read, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  @param pBuffer          The buffer where bytes read from the stream
         *                          shall be placed.
         *  @param bytesToRead      The maximum number of bytes to read.
         *  @return                 The number of bytes actually read. This can
         *                          be smaller than \a bytesToRead if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException);

        /**
         *  Skips a sequence of input bytes.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is skipped, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  @param bytesToSkip      The maximum number of bytes to skip.
         *  @return                 The number of bytes actually skipped. This can
         *                          be smaller than \a bytesToSkip if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException);

        //////////
        //  Implementation
    private:
        IByteInput *            _pSource;

        //  If a stream decryptor is used...
        StreamDecryptor *       _pStreamDecryptor;
        bool                    _autoDeleteStreamDecryptor;

        //  If a block decryptor is used...
        BlockDecryptor *        _pBlockDecryptor;
        PaddingAlgorithm *      _pPaddingAlgorithm;
        bool                    _autoDeleteBlockDecryptor;
        Array<UInt8>            _inputBuffer;
        UIntSize                _bitsReadyInInputBuffer;
        UIntSize                _inputBufferScan;
        int                     _prereadByte;

        UIntSize                _readUleb() throw(IOException);
        void                    _readInputBlock() throw(IOException);
    };
}

#endif  //  __CARL_Encryption_hpp_included__
//  End of CARL_Encryption.hpp
