//
//  CARL_Environment.hpp - CARL process environment.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.12 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines process environments.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Environment.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Environment_hpp_included__
#define __CARL_Environment_hpp_included__

namespace carl
{
    /**
     *  @class Environment CARL.hpp
     *  An encapsulated environment.
     *
     *  An environment is a collection of (name, value) pairs. On the
     *  operating system level, environment provide portable means of
     *  specifying various global settings for processes.
     *
     *  A CARL application can create and manipulate any number of
     *  different environments internally. In addition, there always
     *  exists a predefined "process environment", that reflects upon
     *  the environment passed to the CARL application by its caller.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Environment
    {
        friend class Process;

        //////////
        //  Can't assign or copy construct
    private:
        Environment(const Environment & src);
        void                    operator = (const Environment & src);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  The newly created environment is always empty.
         *
         *  @since 1.0
         */
        Environment() : _mapNamesToValues(StandardHashFunctions::hashString) {}

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Environment() {}

        //////////
        //  Constants
    public:
        /**
         *  Specifies whether environment variable names on the host platform 
         *  are case-sensitive or not.
         *
         *  @hideinitializer
         *  @since 1.0
         */
#if defined(CARL_OS_WIN32)
        static const bool       CaseSensitiveVariableNames = false;
#elif defined(CARL_OS_CYGWIN32)
        static const bool       CaseSensitiveVariableNames = false;
#elif defined(CARL_OS_LINUX)
        static const bool       CaseSensitiveVariableNames = true;
#else
    #error Unsupported platform
#endif
        //////////
        //  Operations
    public:
        /**
         *  Returns a list of names of all variables defined in this environment.
         *
         *  elements in the returned list are not guaranteed to be in
         *  any particular order.
         *
         *  @return             A list of names of all variables defined 
         *                      in this environment.
         *  @since 1.0
         */
        Array<String>           getAllVariableNames() const;

        /**
         *  Checks if an environment variable with the specified name exists
         *  in this environment.
         *
         *  @param pName        A pointer to nul-terminated ASCII string
         *                      representing the name of the environment
         *                      variable.
         *  @return             True if the environment variable with the
         *                      specified name exists in this environment, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                    hasVariable(const char * pName) const;

        /**
         *  Checks if an environment variable with the specified name exists
         *  in this environment.
         *
         *  @param pName        A pointer to nul-terminated ISO-10646 string
         *                      representing the name of the environment
         *                      variable.
         *  @return             True if the environment variable with the
         *                      specified name exists in this environment, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                    hasVariable(const Char * pName) const;

        /**
         *  Checks if an environment variable with the specified name exists
         *  in this environment.
         *
         *  @param name         A string representing the name of the 
         *                      environment variable.
         *  @return             True if the environment variable with the
         *                      specified name exists in this environment, 
         *                      false otherwise.
         *  @since 1.0
         */
        bool                    hasVariable(const String & name) const;

        /**
         *  Returns a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the specified default value is
         *  returned instead.
         *
         *  @param pName        A pointer to nul-terminated ASCII string
         *                      representing the name of the environment
         *                      variable.
         *  @param pDefaultValue    A pointer to nul-terminated ASCII string
         *                      providing the default value to be returned
         *                      in case the required environment variable
         *                      does not exist.
         *  @return             The value of the specified environment variable,
         *                      or a default value if the environment variable
         *                      does not exist.
         *  @since 1.0
         */
        String                  getVariableValue(const char * pName, const char * pDefaultValue) const;

        /**
         *  Returns a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the specified default value is
         *  returned instead.
         *
         *  @param pName        A pointer to nul-terminated ASCII string
         *                      representing the name of the environment
         *                      variable.
         *  @param pDefaultValue    A pointer to nul-terminated ISO-10646 string
         *                      providing the default value to be returned
         *                      in case the required environment variable
         *                      does not exist.
         *  @return             The value of the specified environment variable,
         *                      or a default value if the environment variable
         *                      does not exist.
         *  @since 1.0
         */
        String                  getVariableValue(const char * pName, const Char * pDefaultValue) const;

        /**
         *  Returns a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the specified default value is
         *  returned instead.
         *
         *  @param pName        A pointer to nul-terminated ASCII string
         *                      representing the name of the environment
         *                      variable.
         *  @param defaultValue A default value to be returned
         *                      in case the required environment variable
         *                      does not exist.
         *  @return             The value of the specified environment variable,
         *                      or a default value if the environment variable
         *                      does not exist.
         *  @since 1.0
         */
        String                  getVariableValue(const char * pName, const String & defaultValue) const;

        /**
         *  Returns a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the specified default value is
         *  returned instead.
         *
         *  @param pName        A pointer to nul-terminated ISO-10646 string
         *                      representing the name of the environment
         *                      variable.
         *  @param pDefaultValue    A pointer to nul-terminated ASCII string
         *                      providing the default value to be returned
         *                      in case the required environment variable
         *                      does not exist.
         *  @return             The value of the specified environment variable,
         *                      or a default value if the environment variable
         *                      does not exist.
         *  @since 1.0
         */
        String                  getVariableValue(const Char * pName, const char * pDefaultValue) const;

        /**
         *  Returns a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the specified default value is
         *  returned instead.
         *
         *  @param pName        A pointer to nul-terminated ISO-10646 string
         *                      representing the name of the environment
         *                      variable.
         *  @param pDefaultValue    A pointer to nul-terminated ISO-10646 string
         *                      providing the default value to be returned
         *                      in case the required environment variable
         *                      does not exist.
         *  @return             The value of the specified environment variable,
         *                      or a default value if the environment variable
         *                      does not exist.
         *  @since 1.0
         */
        String                  getVariableValue(const Char * pName, const Char * pDefaultValue) const;

        /**
         *  Returns a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the specified default value is
         *  returned instead.
         *
         *  @param pName        A pointer to nul-terminated ISO-10646 string
         *                      representing the name of the environment
         *                      variable.
         *  @param defaultValue A default value to be returned
         *                      in case the required environment variable
         *                      does not exist.
         *  @return             The value of the specified environment variable,
         *                      or a default value if the environment variable
         *                      does not exist.
         *  @since 1.0
         */
        String                  getVariableValue(const Char * pName, const String & defaultValue) const;

        /**
         *  Returns a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the specified default value is
         *  returned instead.
         *
         *  @param name         A string representing the name of the environment
         *                      variable.
         *  @param pDefaultValue    A pointer to nul-terminated ASCII string
         *                      providing the default value to be returned
         *                      in case the required environment variable
         *                      does not exist.
         *  @return             The value of the specified environment variable,
         *                      or a default value if the environment variable
         *                      does not exist.
         *  @since 1.0
         */
        String                  getVariableValue(const String & name, const char * pDefaultValue) const;

        /**
         *  Returns a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the specified default value is
         *  returned instead.
         *
         *  @param name         A string representing the name of the environment
         *                      variable.
         *  @param pDefaultValue    A pointer to nul-terminated ISO-10646 string
         *                      providing the default value to be returned
         *                      in case the required environment variable
         *                      does not exist.
         *  @return             The value of the specified environment variable,
         *                      or a default value if the environment variable
         *                      does not exist.
         *  @since 1.0
         */
        String                  getVariableValue(const String & name, const Char * pDefaultValue) const;

        /**
         *  Returns a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the specified default value is
         *  returned instead.
         *
         *  @param name         A string representing the name of the environment
         *                      variable.
         *  @param defaultValue A default value to be returned
         *                      in case the required environment variable
         *                      does not exist.
         *  @return             The value of the specified environment variable,
         *                      or a default value if the environment variable
         *                      does not exist.
         *  @since 1.0
         */
        String                  getVariableValue(const String & name, const String & defaultValue) const;

        /**
         *  Sets a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the variable is created with the
         *  specified value; otherwise the previous value of the variable
         *  is replaced with the new one.
         *
         *  @param pName        A pointer to nul-terminated ASCII string
         *                      representing the name of the environment
         *                      variable.
         *  @param pValue       A pointer to nul-terminated ASCII string
         *                      to assign to the environment variable as a value.
         *  @since 1.0
         */
        void                    setVariableValue(const char * pName, const char * pValue);

        /**
         *  Sets a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the variable is created with the
         *  specified value; otherwise the previous value of the variable
         *  is replaced with the new one.
         *
         *  @param pName        A pointer to nul-terminated ASCII string
         *                      representing the name of the environment
         *                      variable.
         *  @param pValue       A pointer to nul-terminated ISO-10646 string
         *                      to assign to the environment variable as a value.
         *  @since 1.0
         */
        void                    setVariableValue(const char * pName, const Char * pValue);

        /**
         *  Sets a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the variable is created with the
         *  specified value; otherwise the previous value of the variable
         *  is replaced with the new one.
         *
         *  @param pName        A pointer to nul-terminated ASCII string
         *                      representing the name of the environment
         *                      variable.
         *  @param value        A string to assign to the environment 
         *                      variable as a value.
         *  @since 1.0
         */
        void                    setVariableValue(const char * pName, const String & value);

        /**
         *  Sets a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the variable is created with the
         *  specified value; otherwise the previous value of the variable
         *  is replaced with the new one.
         *
         *  @param pName        A pointer to nul-terminated ISO-10646 string
         *                      representing the name of the environment
         *                      variable.
         *  @param pValue       A pointer to nul-terminated ASCII string
         *                      to assign to the environment variable as a value.
         *  @since 1.0
         */
        void                    setVariableValue(const Char * pName, const char * pValue);

        /**
         *  Sets a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the variable is created with the
         *  specified value; otherwise the previous value of the variable
         *  is replaced with the new one.
         *
         *  @param pName        A pointer to nul-terminated ISO-10646 string
         *                      representing the name of the environment
         *                      variable.
         *  @param pValue       A pointer to nul-terminated ISo-10646 string
         *                      to assign to the environment variable as a value.
         *  @since 1.0
         */
        void                    setVariableValue(const Char * pName, const Char * pValue);

        /**
         *  Sets a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the variable is created with the
         *  specified value; otherwise the previous value of the variable
         *  is replaced with the new one.
         *
         *  @param pName        A pointer to nul-terminated ISO-10646 string
         *                      representing the name of the environment
         *                      variable.
         *  @param value        A string to assign to the environment variable 
         *                      as a value.
         *  @since 1.0
         */
        void                    setVariableValue(const Char * pName, const String & value);

        /**
         *  Sets a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the variable is created with the
         *  specified value; otherwise the previous value of the variable
         *  is replaced with the new one.
         *
         *  @param name         A string representing the name of the environment
         *                      variable.
         *  @param pValue       A pointer to nul-terminated ASCII string
         *                      to assign to the environment variable as a value.
         *  @since 1.0
         */
        void                    setVariableValue(const String & name, const char * pValue);

        /**
         *  Sets a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the variable is created with the
         *  specified value; otherwise the previous value of the variable
         *  is replaced with the new one.
         *
         *  @param name         A string representing the name of the environment
         *                      variable.
         *  @param pValue       A pointer to nul-terminated ISO-10646 string
         *                      to assign to the environment variable as a value.
         *  @since 1.0
         */
        void                    setVariableValue(const String & name, const Char * pValue);

        /**
         *  Sets a value of the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the variable is created with the
         *  specified value; otherwise the previous value of the variable
         *  is replaced with the new one.
         *
         *  @param name         A string representing the name of the environment
         *                      variable.
         *  @param value        A string to assign to the environment variable 
         *                      as a value.
         *  @since 1.0
         */
        void                    setVariableValue(const String & name, const String & value);

        /**
         *  Destroys the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the call has no effect.
         *
         *  @param pName        A pointer to nul-terminated ASCII string
         *                      representing the name of the environment
         *                      variable to be destroyed.
         *  @since 1.0
         */
        void                    removeVariable(const char * pName);

        /**
         *  Destroys the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the call has no effect.
         *
         *  @param pName        A pointer to nul-terminated ISo-10646 string
         *                      representing the name of the environment
         *                      variable to be destroyed.
         *  @since 1.0
         */
        void                    removeVariable(const Char * pName);

        /**
         *  Destroys the specified environment variable.
         *
         *  If this environment does not have an environment variable
         *  with the specified name, the call has no effect.
         *
         *  @param name         A string representing the name of the 
         *                      environment variable to be destroyed.
         *  @since 1.0
         */
        void                    removeVariable(const String & name);

        //////////
        //  Implementation
    private:
        HashMap<String, String> _mapNamesToValues;
    };
}

#endif  //  __CARL_Environment_hpp_included__
//  End of CARL_Environment.hpp
