//
//  CARL_Exceptions.hpp - CARL exceptions framework.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.10 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL exception framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Exceptions.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Exceptions_hpp_included__
#define __CARL_Exceptions_hpp_included__

namespace carl
{
    class CARL_PUBLIC Locale;

    /**
     *  @class Throwable CARL.hpp
     *  A common base for all CARL exceptions.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Throwable
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        Throwable();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        Throwable(const Throwable & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Throwable();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        Throwable &             operator = (const Throwable & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class Exception CARL.hpp
     *  A common base for all CARL recoverable exceptions.
     *
     *  An Exception is a throwable object representing a recoverable
     *  unusual condition. It is intended to be caught by application.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Exception : public Throwable
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        Exception();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        Exception(const Exception & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Exception();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        Exception &             operator = (const Exception & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class Error CARL.hpp
     *  A common base for all CARL unrecoverable exceptions.
     *
     *  An Error is a throwable object representing an unrecoverable
     *  unusual condition. It is not intended to be caught by application
     *  and proper recovery mey be impossible even if it is caught.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Error : public Throwable
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        Error();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        Error(const Error & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Error();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        Error &                 operator = (const Error & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class ProgramError CARL.hpp
     *  A common base for all CARL program errors.
     *
     *  A program error indicates a bug somewhere in the program.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC ProgramError : public Error
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        ProgramError();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        ProgramError(const ProgramError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~ProgramError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        ProgramError &          operator = (const ProgramError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class AssertionError CARL.hpp
     *  An exception thrown when an assertion fault is detected.
     *
     *  An assertion fault is a program error caused by an asserted
     *  condition failing to hold.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC AssertionError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  The newly created AssertionError instance is associated with an
         *  assertion string,m source file name string and line number; the
         *  former two being pointers to statically allocated nul-terminated
         *  ASCII strings.
         *
         *  @param pAssertionText   The assertion text.
         *  @param pSourceFileName  The name of the source file where an assertion
         *                          fault has been detected.
         *  @param sourceLineNumber The source line number where an assertion
         *                          fault has been detected.
         *  @since 1.0
         */
        AssertionError(const char * pAssertionText, const char * pSourceFileName, UIntSize sourceLineNumber);

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        AssertionError(const AssertionError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~AssertionError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        AssertionError &        operator = (const AssertionError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the pointer to a statically allocated text message
         *  associated with an assertion error.
         *
         *  @return             The text message associated with an assertion.
         *  @since 1.0
         */
        const char *            getAssertionText() const;

        /**
         *  Returns the pointer to a statically allocated source file name
         *  associated with an assertion error.
         *
         *  @return             The source file name associated with an assertion.
         *  @since 1.0
         */
        const char *            getSourceFileName() const;

        /**
         *  Returns the source line number associated with an assertion error.
         *
         *  @return             The source line number associated with an assertion.
         *  @since 1.0
         */
        UIntSize                getSourceLineNumber() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;

        //////////
        //  Implementation
    private:
        const char *            _pAssertionText;
        const char *            _pSourceFileName;
        UIntSize                _sourceLineNumber;
    };

    /**
     *  @class NoSuchElementError CARL.hpp
     *  An error throws when a nonexistent member of a data structure 
     *  is being accessed.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC NoSuchElementError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        NoSuchElementError();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        NoSuchElementError(const NoSuchElementError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~NoSuchElementError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        NoSuchElementError &    operator = (const NoSuchElementError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class SystemError CARL.hpp
     *  A system error indicates a fatal system failure for which a program 
     *  is not responsible.
     *
     *  Clean recovery from a system error is, in general case, not possible.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC SystemError : public Error
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        SystemError();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        SystemError(const SystemError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~SystemError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        SystemError &           operator = (const SystemError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class VersionCompatibilityError CARL.hpp
     *  Thrown when a CARL client attempts to dynamically link against
     *  an incompatible CARL version.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC VersionCompatibilityError : public SystemError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        VersionCompatibilityError();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        VersionCompatibilityError(const VersionCompatibilityError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~VersionCompatibilityError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        VersionCompatibilityError & operator = (const VersionCompatibilityError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class ParseException CARL.hpp
     *  An exception thrown when parsing a string fails.
     *
     *  @since 1.0
     */
    //////////
    //  Defines here instead of CARL_Parse.hpp because:
    //  1)  defining the exception in CARL_Parse.hpp would introduce a circular 
    //      dependency between CARL_Parse.hpp and CARL_Math.hpp and 
    //  2)  user application may want to parse something else besides numbers, 
    //      dates & times, so ParseException shound not be tied too closely to 
    //      number, date & time parsing.
    class CARL_PUBLIC ParseException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  @param parsedString     The string that was being parsed.
         *  @param parsePosition    The position within the parsed string
         *                          at which a parse error has been detected.
         *  @since 1.0
         */
        ParseException(const String & parsedString, UIntSize parsePosition);

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        ParseException(const ParseException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~ParseException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        ParseException &        operator = (const ParseException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the string that was being parsed.
         *
         *  @return             The string that was being parsed.
         *  @since 1.0
         */
        String                  getParsedString() const;

        /**
         *  Returns the position within the parsed string at which a parse 
         *  error has been detected.
         *
         *  @return             The position within the parsed string 
         *                      at which a parse error has been detected.
         *  @since 1.0
         */
        UIntSize                getParsePosition() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;

        //////////
        //  Implementation
    private:
        String                  _parsedString;
        UIntSize                _parsePosition;
    };
}

#endif  //  __CARL_Exceptions_hpp_included__
//  End of CARL_Exceptions.hpp
