//
//  CARL_Extension.hpp - Extension framework.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.12 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL extension framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Extension.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Extension_hpp_included__
#define __CARL_Extension_hpp_included__

#if defined(CARL_LINKAGE_DYNAMIC)

namespace carl
{
    /**
     *  @class Extension CARL.hpp
     *  A common base class for all CARL extensions.
     *
     *  A functionality of a dynamic CARL library can be extended
     *  post-deployment by installing additional CARL extension DLLs
     *  (also known as shared objects in Unix world). These extension
     *  DLLs can provide additional instances of standard CARL
     *  concepts (e.g. additional compression and encryption
     *  algorithms, additional character sets, CARL localization
     *  for additional locales, etc.)
     *
     *  Each CARL extension DLL must:
     *
     *  -   Link against the dynamic CARL library.
     *  -   Define one or more concrete subclasses of Extension.
     *      Each of these concrete subclasses represents an
     *      independent CARL extension.
     *  -   Create and export an instance of that subclass using
     *      special macros defined by CARL for that purpose.
     *
     *  If two concrete subclasses ExtensionA and ExtensionB of the
     *  Extension class are defined, the following code must then
     *  be written to export these extensions from the CARL
     *  extension DLL:
     *
     *  @code
     *  CARL_BEGIN_EXTENSION_EXPORT
     *  CARL_EXPORT_EXTENSION(ExtensionA)
     *  CARL_EXPORT_EXTENSION(ExtensionB)
     *  CARL_END_EXTENSION_EXPORT
     *  @endcode
     *
     *  All concrete Extension subclasses must define a public default
     *  constructor and a public destructor.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Extension
    {
        //////////
        //  Can't assign or copy construct
    private:
        Extension(const Extension &);
        void                    operator = (const Extension &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  All concrete derived classes must provide a public default constructor.
         *
         *  @since 1.0
         */
        Extension();

        /**
         *  The class destructor.
         *
         *  All concrete derived classes must provide a public destructor.
         *
         *  @since 1.0
         */
        virtual ~Extension();

        //////////
        //  Operations
    public:
        /**
         *  Returns the CARL build key associated with this extension.
         *
         *  A CARL build key is a string that encapsulates processor, toolchain,
         *  OS, mode and CARL version information. A client that uses a dynamic
         *  CARL library is considered to be compatible with that library if
         *  and only if the two have the same build keys.
         *
         *  @return             The build key used by the extension.
         *  @since 1.0
         */
        virtual const char *    getBuildKey() const = 0;

        /**
         *  Applies the extension.
         *
         *  The CARL extension framework calls this method to apply an extension.
         *  The implementstion provided by a concrete Extension subclass will
         *  typically use CARL stock object registration services to add to
         *  the set of CARL stock objects.
         *  @since 1.0
         */
        virtual void            apply() = 0;
    };

    /**
     *  @class ExtensionLoader CARL.hpp
     *  A utility class providing extension loading services.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC ExtensionLoader
    {
        //////////
        //  Utility classes cannot be instantiated
    private:
        ExtensionLoader();
        ExtensionLoader(const ExtensionLoader &);
        ~ExtensionLoader();
        void                    operator = (const ExtensionLoader &);

        //////////
        //  Operations
    public:
        /**
         *  Loads and applies all extensions defined in a specified DLL.
         *
         *  @param pFileName    A pointer to nul-terminated ASCII string
         *                      representing the file name of the CARL
         *                      extension DLL.
         *  @return             True if the CARL extension DLL was successfully
         *                      located and at least one CARL extension from
         *                      that DLL has been loaded and applied; false
         *                      otherwise.
         *  @since 1.0
         */
        static bool             loadFromFile(const char * pFileName);

        /**
         *  Loads and applies all extensions defined in a specified DLL.
         *
         *  @param pFileName    A pointer to nul-terminated ASCII string
         *                      representing the file name of the CARL
         *                      extension DLL.
         *  @return             True if the CARL extension DLL was successfully
         *                      located and at least one CARL extension from
         *                      that DLL has been loaded and applied; false
         *                      otherwise.
         *  @since 1.0
         */
        static bool             loadFromFile(const Char * pFileName);

        /**
         *  Loads and applies all extensions defined in a specified DLL.
         *
         *  @param fileName     A string representing the file name of the CARL
         *                      extension DLL.
         *  @return             True if the CARL extension DLL was successfully
         *                      located and at least one CARL extension from
         *                      that DLL has been loaded and applied; false
         *                      otherwise.
         *  @since 1.0
         */
        static bool             loadFromFile(const String & fileName);

        /**
         *  Loads and applies all extensions defined in a specified DLL.
         *
         *  @param fileName     A path name to the CARL extension DLL.
         *  @return             True if the CARL extension DLL was successfully
         *                      located and at least one CARL extension from
         *                      that DLL has been loaded and applied; false
         *                      otherwise.
         *  @since 1.0
         */
        static bool             loadFromFile(const PathName & fileName);

        /**
         *  Loads all CARL extension DLLs located in the specified directory
         *  and applies CARL extensions defined therein.
         *
         *  When the directory is scanned for CARL extension DLLs, only
         *  those files are considered that:
         *
         *  -   Have the proper filename extension (.dll for Windows, .so
         *      for Unix, etc.)
         *  -   Contain, somewhere within the file name, the substring
         *      that matches the platform for which the CARL dynamic
         *      library has been built. This substring is:
         *      -   "x86-win32-MSVC2003" for x86 Win32 MSVC .NET 2003
         *          build of the CARL dynamic library.
         *      -   "x86-win32-cygwin" for x86 Win32 cygwin
         *          build of the CARL dynamic library.
         *  -   The presense of the platform identification substring in
         *      an extension DLL file name allows extension DLLs built by
         *      several different toolchains to coexist in the same directory
         *      without interfering with each other.
         *
         *  @param pDirectoryName   A pointer to nul-terminated ASCII string
         *                      representing the name of the directory which
         *                      is searched for CARL extension DLL.
         *  @param processSubdirectories    If true, subdirectories of the
         *                      specified directory are recursively scanned for
         *                      CARL extension DLLs. Otherwise, only the
         *                      specified directory is scanned.
         *  @return             True if at least one CARL extension DLL was 
         *                      successfully located and at least one CARL 
         *                      extension from that DLL has been loaded and 
         *                      applied; false otherwise.
         *  @since 1.0
         */
        static bool             loadFromDirectory(const char * pDirectoryName, bool processSubdirectories = false);

        /**
         *  Loads all CARL extension DLLs located in the specified directory
         *  and applies CARL extensions defined therein.
         *
         *  When the directory is scanned for CARL extension DLLs, only
         *  those files are considered that:
         *
         *  -   Have the proper filename extension (.dll for Windows, .so
         *      for Unix, etc.)
         *  -   Contain, somewhere within the file name, the substring
         *      that matches the platform for which the CARL dynamic
         *      library has been built. This substring is:
         *      -   "x86-win32-MSVC2003" for x86 Win32 MSVC .NET 2003
         *          build of the CARL dynamic library.
         *      -   "x86-win32-cygwin" for x86 Win32 cygwin
         *          build of the CARL dynamic library.
         *  -   The presense of the platform identification substring in
         *      an extension DLL file name allows extension DLLs built by
         *      several different toolchains to coexist in the same directory
         *      without interfering with each other.
         *
         *  @param pDirectoryName   A pointer to nul-terminated ASCII string
         *                      representing the name of the directory which
         *                      is searched for CARL extension DLL.
         *  @param processSubdirectories    If true, subdirectories of the
         *                      specified directory are recursively scanned for
         *                      CARL extension DLLs. Otherwise, only the
         *                      specified directory is scanned.
         *  @return             True if at least one CARL extension DLL was 
         *                      successfully located and at least one CARL 
         *                      extension from that DLL has been loaded and 
         *                      applied; false otherwise.
         *  @since 1.0
         */
        static bool             loadFromDirectory(const Char * pDirectoryName, bool processSubdirectories = false);

        /**
         *  Loads all CARL extension DLLs located in the specified directory
         *  and applies CARL extensions defined therein.
         *
         *  When the directory is scanned for CARL extension DLLs, only
         *  those files are considered that:
         *
         *  -   Have the proper filename extension (.dll for Windows, .so
         *      for Unix, etc.)
         *  -   Contain, somewhere within the file name, the substring
         *      that matches the platform for which the CARL dynamic
         *      library has been built. This substring is:
         *      -   "x86-win32-MSVC2003" for x86 Win32 MSVC .NET 2003
         *          build of the CARL dynamic library.
         *      -   "x86-win32-cygwin" for x86 Win32 cygwin
         *          build of the CARL dynamic library.
         *  -   The presense of the platform identification substring in
         *      an extension DLL file name allows extension DLLs built by
         *      several different toolchains to coexist in the same directory
         *      without interfering with each other.
         *
         *  @param directoryName    A string representing the name of the 
         *                      directory which is searched for CARL extension DLL.
         *  @param processSubdirectories    If true, subdirectories of the
         *                      specified directory are recursively scanned for
         *                      CARL extension DLLs. Otherwise, only the
         *                      specified directory is scanned.
         *  @return             True if at least one CARL extension DLL was 
         *                      successfully located and at least one CARL 
         *                      extension from that DLL has been loaded and 
         *                      applied; false otherwise.
         *  @since 1.0
         */
        static bool             loadFromDirectory(const String & directoryName, bool processSubdirectories = false);

        /**
         *  Loads all CARL extension DLLs located in the specified directory
         *  and applies CARL extensions defined therein.
         *
         *  When the directory is scanned for CARL extension DLLs, only
         *  those files are considered that:
         *
         *  -   Have the proper filename extension (.dll for Windows, .so
         *      for Unix, etc.)
         *  -   Contain, somewhere within the file name, the substring
         *      that matches the platform for which the CARL dynamic
         *      library has been built. This substring is:
         *      -   "x86-win32-MSVC2003" for x86 Win32 MSVC .NET 2003
         *          build of the CARL dynamic library.
         *      -   "x86-win32-cygwin" for x86 Win32 cygwin
         *          build of the CARL dynamic library.
         *  -   The presense of the platform identification substring in
         *      an extension DLL file name allows extension DLLs built by
         *      several different toolchains to coexist in the same directory
         *      without interfering with each other.
         *
         *  @param directoryName    A path name to the directory which
         *                      is searched for CARL extension DLL.
         *  @param processSubdirectories    If true, subdirectories of the
         *                      specified directory are recursively scanned for
         *                      CARL extension DLLs. Otherwise, only the
         *                      specified directory is scanned.
         *  @return             True if at least one CARL extension DLL was 
         *                      successfully located and at least one CARL 
         *                      extension from that DLL has been loaded and 
         *                      applied; false otherwise.
         *  @since 1.0
         */
        static bool             loadFromDirectory(const PathName & directoryName, bool processSubdirectories = false);

        /**
         *  Loads and applies CARL extensions defined in "standard" places.
         *
         *  The "standard" places to look for CARL extensions are, on order:
         *
         *  -   The CARL application itself. Just like a CARL extension DLL,
         *      a CARL application can define and export CARL extensions.
         *  -   Directories referred to by the environment variable
         *      CARL_1_0_EXTENSIONS. These directories are searched one after
         *      another in the order in which they appear in the environment
         *      variable's value. Each directory is searched non-recursively
         *      for CARL extension DLLs.
         *
         *  @return             True if at least one CARL extension DLL was 
         *                      successfully located and at least one CARL 
         *                      extension from that DLL has been loaded and 
         *                      applied; false otherwise.
         *  @since 1.0
         */
        static bool             loadInstalledExtensions();

        //////////
        //  Implementation
    private:
        static HashMap<String, Library*> *  getMapFileNamesToLibraries();
        static CriticalSection *_getStaticGuard();
    };

    extern "C"
    {
        typedef void (CARL_CALLBACK * ExtensionProviderProc)(Array<Extension *> & extensions);
    }
}

#define CARL_BEGIN_EXTENSION_EXPORT                         \
extern "C" CARL_EXPORT                                      \
void CARL_GetExtensions(Array<Extension *> & extensions)    \
{                                                           \
    require(extensions.size() == 0);

#define CARL_EXPORT_EXTENSION(ExtensionType)    \
    static ExtensionType extension;             \
    extensions.append(&extension);

#define CARL_END_EXTENSION_EXPORT               \
}

#endif  //  defined(CARL_LINKAGE_DYNAMIC)

#endif  //  __CARL_Extension_hpp_included__
//  End of CARL_Extension.hpp
