//
//  CARL_FileSystem.hpp - File system API
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/07/28 10:40:28 $
//  $Revision: 1.25 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL file system API.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_FileSystem.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_FileSystem_hpp_included__
#define __CARL_FileSystem_hpp_included__

namespace carl
{
    class CARL_PUBLIC FileSystem;

    /**
     *  @enum FileSystemPermissions CARL.hpp
     *  Represents granted access permissions for file system entities.
     *
     *  @since 1.0
     */
    enum FileSystemPermissions
    {
        /**
         *  A permission to read the file is granted to the file's owner.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        FileSystemPermissions_OwnerRead = 0x0001,

        /**
         *  A permission to write the file is granted to the file's owner.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        FileSystemPermissions_OwnerWrite = 0x0002,

        /**
         *  A permission to execute the file is granted to the file's owner.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        FileSystemPermissions_OwnerExecute = 0x0004,

        /**
         *  A permission to read the file is granted to the file's group.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        FileSystemPermissions_GroupRead = 0x0008,

        /**
         *  A permission to write the file is granted to the file's group.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        FileSystemPermissions_GroupWrite = 0x0010,

        /**
         *  A permission to execute the file is granted to the file's group.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        FileSystemPermissions_GroupExecute = 0x0020,

        /**
         *  A permission to read the file is granted to the everybody.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        FileSystemPermissions_OtherRead = 0x0040,

        /**
         *  A permission to write the file is granted to the everybody.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        FileSystemPermissions_OtherWrite = 0x0080,

        /**
         *  A permission to execute the file is granted to the everybody.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        FileSystemPermissions_OtherExecute = 0x0100
    };

    /**
     *  @class PathName CARL.hpp
     *  A path name.
     *
     *  Represents either absolute or relative path name of a file system entity.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC PathName
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  Iniitializes a newly created path name to an empty path.
         *
         *  @since 1.0
         */
        PathName() {}

        /**
         *  The class constructor.
         *
         *  Iniitializes a newly created path name with the specified path string.
         *
         *  @param pPathNameString  A pointer to nul-terminated ASCII string
         *                          representing the path.
         *  @since 1.0
         */
        explicit PathName(const char * pPathNameString);

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        /**
         *  The class constructor.
         *
         *  Iniitializes a newly created path name with the specified path string.
         *
         *  @param pPathNameString  A pointer to nul-terminated Unicode string
         *                          representing the path.
         *  @since 1.0
         */
        explicit PathName(const wchar_t * pPathNameString);
#endif

        /**
         *  The class constructor.
         *
         *  Iniitializes a newly created path name with the specified path string.
         *
         *  @param pPathNameString  A pointer to nul-terminated ISO-10646 string
         *                          representing the path.
         *  @since 1.0
         */
        explicit PathName(const Char * pPathNameString);

        /**
         *  The class constructor.
         *
         *  Iniitializes a newly created path name with the specified path string.
         *
         *  @param pathNameString   A string representing the path.
         *  @since 1.0
         */
        explicit PathName(const String & pathNameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param pDirectoryString A pointer to nul-terminated ASCII string
         *                          representing the directory path.
         *  @param pFilenameString  A pointer to nul-terminated ASCII string
         *                          representing the file name within the 
         *                          directory.
         *  @since 1.0
         */
        PathName(const char * pDirectoryString, const char * pFilenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param pDirectoryString A pointer to nul-terminated ASCII string
         *                          representing the directory path.
         *  @param pFilenameString  A pointer to nul-terminated ISO-10646 string
         *                          representing the file name within the 
         *                          directory.
         *  @since 1.0
         */
        PathName(const char * pDirectoryString, const Char * pFilenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param pDirectoryString A pointer to nul-terminated ASCII string
         *                          representing the directory path.
         *  @param filenameString   A string representing the file name 
         *                          within the directory.
         *  @since 1.0
         */
        PathName(const char * pDirectoryString, const String & filenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param pDirectoryString A pointer to nul-terminated ISO-10646 string
         *                          representing the directory path.
         *  @param pFilenameString  A pointer to nul-terminated ASCII string
         *                          representing the file name within the 
         *                          directory.
         *  @since 1.0
         */
        PathName(const Char * pDirectoryString, const char * pFilenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param pDirectoryString A pointer to nul-terminated ISO-10646 string
         *                          representing the directory path.
         *  @param pFilenameString  A pointer to nul-terminated ISO-10646 string
         *                          representing the file name within the 
         *                          directory.
         *  @since 1.0
         */
        PathName(const Char * pDirectoryString, const Char * pFilenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param pDirectoryString A pointer to nul-terminated ISO-10646 string
         *                          representing the directory path.
         *  @param filenameString   A string representing the file name 
         *                          within the directory.
         *  @since 1.0
         */
        PathName(const Char * pDirectoryString, const String & filenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param directoryString  A string representing the directory path.
         *  @param pFilenameString  A pointer to nul-terminated ASCII string
         *                          representing the file name within the 
         *                          directory.
         *  @since 1.0
         */
        PathName(const String & directoryString, const char * pFilenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param directoryString  A string representing the directory path.
         *  @param pFilenameString  A pointer to nul-terminated ISo-10646 string
         *                          representing the file name within the 
         *                          directory.
         *  @since 1.0
         */
        PathName(const String & directoryString, const Char * pFilenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param directoryString  A string representing the directory path.
         *  @param filenameString   A string representing the file name 
         *                          within the directory.
         *  @since 1.0
         */
        PathName(const String & directoryString, const String & filenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param directory        A path to the directory.
         *  @param pFilenameString  A pointer to nul-terminated ASCII string
         *                          representing the file name within the 
         *                          directory.
         *  @since 1.0
         */
        PathName(const PathName & directory, const char * pFilenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param directory        A path to the directory.
         *  @param pFilenameString  A pointer to nul-terminated ISO-10646 string
         *                          representing the file name within the 
         *                          directory.
         *  @since 1.0
         */
        PathName(const PathName & directory, const Char * pFilenameString);

        /**
         *  The class constructor.
         *
         *  Combines a directory path with a file name and initializes a
         *  newly created PathName with the resulting path.
         *
         *  @param directory        A path to the directory.
         *  @param filenameString   A string representing the file name 
         *                          within the directory.
         *  @since 1.0
         */
        PathName(const PathName & directory, const String & filenameString);

        /**
         *  The copy constructor.
         *
         *  The newly created path name is initialized to be an exact copy
         *  of an existing path name.
         *
         *  @param src              The path name to copy properties from.
         *  @since 1.0
         */
        PathName(const PathName & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~PathName() {}

        /**
         *  The assigment operator.
         *
         *  Makes a path name an exact copy of an existing path name.
         *
         *  @param src              The path name to copy properties from.
         *  @return                 A reference to an assigned-to object.
         *  @since 1.0
         */
        PathName &          operator = (const PathName & src);

        /**
         *  The assigment operator.
         *
         *  Assigns the specified nul-terminated ASCII path string 
         *  to this path name.
         *
         *  @param pPathNameString  The path name string to assign to 
         *                          this path name.
         *  @return                 A reference to an assigned-to object.
         *  @since 1.0
         */
        PathName &          operator = (const char * pPathNameString);

        /**
         *  The assigment operator.
         *
         *  Assigns the specified nul-terminated ISo-10646 path string 
         *  to this path name.
         *
         *  @param pPathNameString  The path name string to assign to 
         *                          this path name.
         *  @return                 A reference to an assigned-to object.
         *  @since 1.0
         */
        PathName &          operator = (const Char * pPathNameString);

        /**
         *  The assigment operator.
         *
         *  Assigns the specified path string to this path name.
         *
         *  @param pathNameString   The path name string to assign to 
         *                          this path name.
         *  @return                 A reference to an assigned-to object.
         *  @since 1.0
         */
        PathName &          operator = (const String & pathNameString);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a path name is empty.
         *
         *  An empty path name is guaranteed not to refer to any
         *  file system entity.
         *
         *  @return                 True if this path name represents an
         *                          empty path, false otherwise.
         *  @since 1.0
         */
        bool                isEmpty() const { return _pathNameString.length() == 0; }

        /**
         *  Returns a string representation of this path name.
         *
         *  @return                 A string representation of this path name.
         *  @since 1.0
         */
        String              toString() const { return _pathNameString; }

        /**
         *  Checks if this path name is absolute.
         *
         *  An absolute path name specifies a full path from a file
         *  system root to a specific file system entity.
         *
         *  @return                 True if this is an absolute path name,
         *                          false otherwise.
         *  @since 1.0
         */
        bool                isAbsolute() const;
        
        /**
         *  Checks if this path name has a directory component
         *
         *  @return                 True if this path name had a directory
         *                          component, false if this path name
         *                          specifies only a file name.
         *  @since 1.0
         */
        bool                hasDirectory() const;

        /**
         *  Checks if the file name portion of this path name
         *  has an extension.
         *
         *  @return                 True if the filename portion of this path
         *                          name has an extension; false otherwise.
         *  @since 1.0
         */
        bool                hasExtension() const;

        /**
         *  Checks if the file name portion of this path name
         *  has the specified extension.
         *
         *  Depending on whether the host platform has case-sensitive
         *  or case-insensitive file names, the specified extension
         *  is compared to this path name's extension either
         *  case-sensitively or case-insensitively.
         *
         *  @param pExtension       A pointer to nul-terminated ASCII string
         *                          representing the extension to look for.
         *                          The string can start with an extension
         *                          separator, or the extension separator
         *                          can be omitted.
         *  @return                 True if the filename portion of this path
         *                          name has the specified extension; false 
         *                          otherwise.
         *  @since 1.0
         */
        bool                hasExtension(const char * pExtension) const;

        /**
         *  Checks if the file name portion of this path name
         *  has the specified extension.
         *
         *  Depending on whether the host platform has case-sensitive
         *  or case-insensitive file names, the specified extension
         *  is compared to this path name's extension either
         *  case-sensitively or case-insensitively.
         *
         *  @param pExtension       A pointer to nul-terminated ISo-10646 string
         *                          representing the extension to look for.
         *                          The string can start with an extension
         *                          separator, or the extension separator
         *                          can be omitted.
         *  @return                 True if the filename portion of this path
         *                          name has the specified extension; false 
         *                          otherwise.
         *  @since 1.0
         */
        bool                hasExtension(const Char * pExtension) const;

        /**
         *  Checks if the file name portion of this path name
         *  has the specified extension.
         *
         *  Depending on whether the host platform has case-sensitive
         *  or case-insensitive file names, the specified extension
         *  is compared to this path name's extension either
         *  case-sensitively or case-insensitively.
         *
         *  @param extension        A string representing the extension to 
         *                          look for. The string can start with 
         *                          an extension separator, or the extension 
         *                          separator can be omitted.
         *  @return                 True if the filename portion of this path
         *                          name has the specified extension; false 
         *                          otherwise.
         *  @since 1.0
         */
        bool                hasExtension(const String & extension) const;

        /**
         *  Returns the directory portion of this path name.
         *
         *  If this path name has no directory portion, returns an
         *  empty string.
         *
         *  The returned directory portion of the path name does not
         *  have a tailing path component separator unless it refers
         *  to a file system root.
         *
         *  @return                 The directory portion of this path name.
         *  @since 1.0
         */
        String              getDirectory() const;

        /**
         *  Returns the filename portion of this path name.
         *
         *  If this path name has no filename portion (i.e. this
         *  if an empty path name of a path name to a file system
         *  root), returns an empty string.
         *
         *  @return                 The filename portion of this path name.
         *  @since 1.0
         */
        String              getFilename() const;

        /**
         *  Returns the name (I.e. filename minus extension)
         *  portion of this path name.
         *
         *  If this path name has no filename portion (i.e. this
         *  if an empty path name of a path name to a file system
         *  root), returns an empty string.
         *
         *  @return                 The name portion of this path name.
         *  @since 1.0
         */
        String              getName() const;

        /**
         *  Returns the extension portion of this path name.
         *
         *  If this path name has no filename portion (i.e. this
         *  if an empty path name of a path name to a file system
         *  root), of the filename portion has no extension,
         *  returns an empty string.
         *
         *  @return                 The extension portion of this 
         *                          path name. It always starts with an
         *                          extension separator unless empty.
         *  @since 1.0
         */
        String              getExtension() const;

        /**
         *  Returns a canonical form of this path name.
         *
         *  A canonical form of a path name is a form that uses
         *  canonical device, path component and extension
         *  separators.
         *
         *  @return                 A canonical form of this path name.
         *  @since 1.0
         */
        PathName            toCanonical() const;

        /**
         *  Returns an absolute form of this path name.
         *
         *  An absolute path name specifies a full path from
         *  a file system root to a file system entity. The resulting
         *  absolute path is always represented in a canonical form.
         *
         *  @return                 An absolute form of this path name.
         *  @since 1.0
         */
        PathName            toAbsolute() const;

        /**
         *  Ensures that this path name has the specified extension.
         *
         *  If the file name portion of this path name already has an
         *  extension, it is replaced, otherwise it is added. The call
         *  has no effect on empty path names.
         *
         *  @param pExtension       A pointer to nul-terminated ASCII string
         *                          representing the extension to enforce.
         *                          The string can start with an extension
         *                          separator, or the extension separator
         *                          can be omitted.
         *  @return                 The path name with the required extension.
         *  @since 1.0
         */
        PathName            withExtension(const char * pExtension);

        /**
         *  Ensures that this path name has the specified extension.
         *
         *  If the file name portion of this path name already has an
         *  extension, it is replaced, otherwise it is added. The call
         *  has no effect on empty path names.
         *
         *  @param pExtension       A pointer to nul-terminated ISO-10646 string
         *                          representing the extension to enforce.
         *                          The string can start with an extension
         *                          separator, or the extension separator
         *                          can be omitted.
         *  @return                 The path name with the required extension.
         *  @since 1.0
         */
        PathName            withExtension(const Char * pExtension);

        /**
         *  Ensures that this path name has the specified extension.
         *
         *  If the file name portion of this path name already has an
         *  extension, it is replaced, otherwise it is added. The call
         *  has no effect on empty path names.
         *
         *  @param extension        A string representing the extension to enforce.
         *                          The string can start with an extension
         *                          separator, or the extension separator
         *                          can be omitted.
         *  @return                 The path name with the required extension.
         *  @since 1.0
         */
        PathName            withExtension(const String & extension);

        /**
         *  Returns a path name to a parent directory of a file system
         *  entity referred to by this path name.
         *
         *  If this path name refers to a file system root or is empty,
         *  returns an empty path name.
         *
         *  @return                 A path name to a parent directory.
         *  @since 1.0
         */
        PathName            getParent() const;

        /**
         *  Converts this path name to a canonical form.
         *
         *  A canonical form of a path name is a form that uses
         *  canonical device, path component and extension
         *  separators.
         *
         *  @since 1.0
         */
        void                makeCanonical();

        /**
         *  Converts this path name to an absolute form.
         *
         *  An absolute path name specifies a full path from
         *  a file system root to a file system entity. The resulting
         *  absolute path is always represented in a canonical form.
         *
         *  @since 1.0
         */
        void                makeAbsolute();

        /**
         *  Ensures that this path name has the specified extension.
         *
         *  If the file name portion of this path name already has an
         *  extension, it is replaced, otherwise it is added. The call
         *  has no effect on empty path names.
         *
         *  @param pExtension       A pointer to nul-terminated ASCII string
         *                          representing the extension to enforce.
         *                          The string can start with an extension
         *                          separator, or the extension separator
         *                          can be omitted.
         *  @return                 True if the extension was successfully
         *                          applied to this path name; false otherwise.
         *  @since 1.0
         */
        bool                ensureExtension(const char * pExtension);

        /**
         *  Ensures that this path name has the specified extension.
         *
         *  If the file name portion of this path name already has an
         *  extension, it is replaced, otherwise it is added. The call
         *  has no effect on empty path names.
         *
         *  @param pExtension       A pointer to nul-terminated ISO-10646 string
         *                          representing the extension to enforce.
         *                          The string can start with an extension
         *                          separator, or the extension separator
         *                          can be omitted.
         *  @return                 True if the extension was successfully
         *                          applied to this path name; false otherwise.
         *  @since 1.0
         */
        bool                ensureExtension(const Char * pExtension);

        /**
         *  Ensures that this path name has the specified extension.
         *
         *  If the file name portion of this path name already has an
         *  extension, it is replaced, otherwise it is added. The call
         *  has no effect on empty path names.
         *
         *  @param extension        A string representing the extension to enforce.
         *                          The string can start with an extension
         *                          separator, or the extension separator
         *                          can be omitted.
         *  @return                 True if the extension was successfully
         *                          applied to this path name; false otherwise.
         *  @since 1.0
         */
        bool                ensureExtension(const String & extension);

        /**
         *  Checks if a file system entity referred to by this path 
         *  name exists.
         *
         *  @return             True if a file system entity exists,
         *                      false otherwise.
         *  @since 1.0
         */
        bool                exists() const;

        /**
         *  Checks if a file system entity referred to by this path 
         *  name exists and is a plain file.
         *
         *  @return             True if a file system entity exists
         *                      and is a file, false otherwise.
         *  @since 1.0
         */
        bool                isFile() const;

        /**
         *  Checks if a file system entity referred to by this path 
         *  name exists and is a directory.
         *
         *  @return             True if a file system entity exists
         *                      and is a directory, false otherwise.
         *  @since 1.0
         */
        bool                isDirectory() const;

        /**
         *  Checks if a file system entity referred to by this path 
         *  name exists and can be read by the caller.
         *
         *  @return             True if a file system entity exists
         *                      and can be read by the caller, false 
         *                      otherwise.
         *  @since 1.0
         */
        bool                canRead() const;

        /**
         *  Sets the "read" permission of a file system entity referred to by 
         *  this path name.
         *
         *  @param yes          \a true to make the file system entity readable,
         *                      \a false to make it non-readable.
         *  @return             True if the "read" permission of a file system 
         *                      entity has been successfully set, false otherwise.
         *  @since 1.0
         */
        bool                setReadable(bool yes = true);

        /**
         *  Checks if a file system entity referred to by this path 
         *  name exists and can be written by the caller.
         *
         *  @return             True if a file system entity exists
         *                      and can be written by the caller, false 
         *                      otherwise.
         *  @since 1.0
         */
        bool                canWrite() const;

        /**
         *  Sets the "write" permission of a file system entity referred to by 
         *  this path name.
         *
         *  @param yes          \a true to make the file system entity writeable,
         *                      \a false to make it non-writeable.
         *  @return             True if the "write" permission of a file system 
         *                      entity has been successfully set, false otherwise.
         *  @since 1.0
         */
        bool                setWriteable(bool yes = true);

        /**
         *  Checks if a file system entity referred to by this path 
         *  name exists and can be executed by the caller.
         *
         *  @return             True if a file system entity exists
         *                      and can be executed by the caller, false 
         *                      otherwise.
         *  @since 1.0
         */
        bool                canExecute() const;

        /**
         *  Sets the "execute" permission of a file system entity referred to by 
         *  this path name.
         *
         *  @param yes          \a true to make the file system entity executable,
         *                      \a false to make it non-executable.
         *  @return             True if the "execute" permission of a file system 
         *                      entity has been successfully set, false otherwise.
         *  @since 1.0
         */
        bool                setExecutable(bool yes = true);

        /**
         *  Checks if a file system entity referred to by this path 
         *  name exists and is hidden.
         *
         *  The exact definition of what constitutes a "hidden" file
         *  system entity if platform specific. For example, on Win32
         *  platform there is a special "hidden" flag for each file
         *  system entity, whereas on Unix a file system entity is
         *  considered hidden if its name starts with a dot '.'.
         *
         *  @return             True if a file system entity exists
         *                      and is hidden, false otherwise.
         *  @since 1.0
         */
        bool                isHidden() const;

        /**
         *  Checks if the two path names refer to the same file system 
         *  entity.
         *
         *  @param pathName     The path name to compare this path name with.
         *  @return             True if both operands refer to the same
         *                      file system entity, false otherwise.
         *  @since 1.0
         */
        bool                operator == (const PathName & pathName) const;

        /**
         *  Checks if the two path names refer to the same file system 
         *  entity.
         *
         *  @param pathName     The path name to compare this path name with.
         *  @return             False if both operands refer to the same
         *                      file system entity, true otherwise.
         *  @since 1.0
         */
        bool                operator != (const PathName & pathName) const;

        /**
         *  Queries the access permissions assigned to the file system entity.
         *
         *  @param permissions  Upon successful query, this parameter is assigned
         *                      the permissions for the file system entity referred
         *                      to by this path name.
         *  @return             True if the permissions of a file system entity have
         *                      been queried successfully, false otherwise.
         *  @since 1.0
         */
        bool                getPermissions(int & permissions);

        /**
         *  Sets the access permissions assigned to the file system entity.
         *
         *  @param permissions  The permissions to assign to for the file system 
         *                      entity referred to by this path name.
         *  @return             True if the permissions of a file system entity have
         *                      been assigned successfully, false otherwise.
         *  @since 1.0
         */
        bool                setPermissions(int permissions);

        //////////
        //  Implementation
    private:
        String              _pathNameString;
        static String       _combineDirectoryAndFilename(const String & directory, const String & filename);
        void                _adjust();
        static int          _findRightmostPathComponentSeparatorIndex(const String & pathNameString);
        static int          _findRightmostExtensionSeparatorIndex(const String & pathNameString, UIntSize startAtIndex);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        friend class FileSystem;
        bool                _store(wchar_t * pBuffer, UIntSize bufferSize) const;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        friend class FileSystem;
        bool                _store(char * pBuffer, UIntSize bufferSize) const;
#endif
    };

    /**
     *  @class DirectoryEntry CARL.hpp
     *  A descriptor of a single directory entry.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DirectoryEntry
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  Initializes a newly created directory entry with all properties undefined.
         *
         *  @since 1.0
         */
        DirectoryEntry();

        /**
         *  The class constructor.
         *
         *  Initializes a newly created directory entry with the specified properties.
         *
         *  @param isDirectory      True iff this directory entry is itself a 
         *                          directory, false otherwise.
         *  @param shortName        The short (without device or directory) name
         *                          of this directory entry.
         *  @param fullName         The full path to the file system entity
         *                          that corresponds to this directory entry.
         *  @param size             The size, in bytes, of the file system entity
         *                          that corresponds to this directory entry.
         *  @param creationTime     The creation time of this directory entry,
         *                          or an invalid time stamp if this information
         *                          is not available.
         *  @param lastAccessTime   The last access time of this directory entry,
         *                          or an invalid time stamp if this information
         *                          is not available.
         *  @param lastWriteTime    The last modification time of this directory entry,
         *                          or an invalid time stamp if this information
         *                          is not available.
         *  @since 1.0
         */
        DirectoryEntry(bool isDirectory, const String & shortName, const String & fullName, UInt64 size, const TimeStamp & creationTime, const TimeStamp & lastAccessTime, const TimeStamp & lastWriteTime);

        /**
         *  The copy constructor.
         *
         *  The newly created directory entry is initialized to be an exact copy
         *  of an existing directory entry.
         *
         *  @param src              The directory entry to copy properties from.
         *  @since 1.0
         */
        DirectoryEntry(const DirectoryEntry & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~DirectoryEntry() {}

        /**
         *  The assigment operator.
         *
         *  Makes a directory entry an exact copy of an existing directory entry.
         *
         *  @param src              The directory entry to copy properties from.
         *  @return                 A reference to an assigned-to object.
         *  @since 1.0
         */
        DirectoryEntry &            operator = (const DirectoryEntry & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if this directory entry is itself a directory.
         *
         *  @return                 True if this directory entry is itself 
         *                          a directory; false otherwise.
         *  @since 1.0
         */
        bool                        isDirectory() const { return _isDirectory; }

        /**
         *  Returns a short name of this directory entry.
         *
         *  @return                 A short (without device or directory) 
         *                          name of this directory entry.
         *  @since 1.0
         */
        String                      getShortName() const { return _shortName; }

        /**
         *  Returns a full name of this directory entry.
         *
         *  @return                 A short (with device and full path) 
         *                          name of this directory entry.
         *  @since 1.0
         */
        String                      getFullName() const { return _fullName; }

        /**
         *  Returns a size of this directory entry.
         *
         *  @return                 A size of this directory entry.
         *  @since 1.0
         */
        UInt64                      getSize() const { return _size; }

        /**
         *  Returns a creation time of this directory entry.
         *
         *  @return                 A creation time of this directory entry,
         *                          or an invalid time stamp if this information
         *                          is not available.
         *  @since 1.0
         */
        TimeStamp                   getCreationTime() const { return _creationTime; }

        /**
         *  Returns a last access time of this directory entry.
         *
         *  @return                 A last access time of this directory entry,
         *                          or an invalid time stamp if this information
         *                          is not available.
         *  @since 1.0
         */
        TimeStamp                   getLastAccessTime() const { return _lastAccessTime; }

        /**
         *  Returns a last write time of this directory entry.
         *
         *  @return                 A last write time of this directory entry,
         *                          or an invalid time stamp if this information
         *                          is not available.
         *  @since 1.0
         */
        TimeStamp                   getLastWriteTime() const { return _lastWriteTime; }

        //////////
        //  Implementation
    private:
        bool                        _isDirectory;
        String                      _shortName;
        String                      _fullName;
        UInt64                      _size;
        TimeStamp                   _creationTime;
        TimeStamp                   _lastAccessTime;
        TimeStamp                   _lastWriteTime;
    };

    /**
     *  @interface IDirectoryEntryFilter CARL.hpp
     *  An abstract interface to a filter used during directory entry enumeration.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IDirectoryEntryFilter
    {
        //////////
        //  Construction/destruction
    public:
        IDirectoryEntryFilter() {}
        virtual ~IDirectoryEntryFilter() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if the specified directory entry matches this filter.
         *
         *  @param directoryEntry   The directory entry to check.
         *  @return                 True if the specified directory entry matches
         *                          this filter (and shall, therefore, be included
         *                          in the directory contents enumeration result);
         *                          false otherwise.
         *  @since 1.0
         */
        virtual bool                match(const DirectoryEntry & directoryEntry) = 0;
    };

    /**
     *  @class FileSystem CARL.hpp
     *  An abstract view of a host file system.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC FileSystem
    {
        //////////
        //  Utility classes cannot be instantiated
    private:
        FileSystem();
        FileSystem(const FileSystem &);
        ~FileSystem();
        void                operator = (const FileSystem &);

        //////////
        //  File system properties
    public:
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        /**
         *  The maximum path length that the host file system is guaranteed
         *  to support.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntSize       MaximumPathLength = 260;

        /**
         *  The maximum length of a single path component that the host file system 
         *  is guaranteed to support.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntSize       MaximumFilenameLength = 260;

        /**
         *  The canonical separator of components within a path.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           PathComponentSeparator = '\\';

        /**
         *  The canonical separator of paths in a path list string.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           PathListSeparator = ';';

        /**
         *  True if the host file system uses named devices (file system 
         *  roots), false otherwise.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const bool           UsesNamedDevices = true;

        /**
         *  The canonical device name separator for absolute paths.
         *
         *  If the host file system does not support the concept of
         *  named file system, roots (Unix is one such example), then
         *  device separators are not used.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           DeviceSeparator = ':';

        /**
         *  The canonical filename extension separator.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           FilenameExtensionSeparator = '.';

        /**
         *  The character that, when used in a file name pattern, means
         *  "match any sequence of characters, inclusing an empty one".
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           AnyStringMetacharacter = '*';

        /**
         *  The character that, when used in a file name pattern, means
         *  "match any single character".
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           AnyCharacterMetacharacter = '?';

        /**
         *  True if the host file system uses case-sensitive file
         *  names, false otherwise.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const bool           CaseSensitiveFileNames = false;

        /**
         *  A nul-terminated string that, when used as a path component,
         *  refers to the "current directory".
         *
         *  @since 1.0
         */
        static const char *         ThisDirectoryName;

        /**
         *  A nul-terminated string that, when used as a path component,
         *  refers to the "parent directory".
         *
         *  @since 1.0
         */
        static const char *         ParentDirectoryName;

        /**
         *  A nul-terminated sequence of characters that is used as line
         *  separator in host file system's text files.
         *
         *  @since 1.0
         */
        static const char *         LineSeparator;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        /**
         *  The maximum path length that the host file system is guaranteed
         *  to support.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntSize       MaximumPathLength = 260;

        /**
         *  The maximum length of a single path component that the host file system 
         *  is guaranteed to support.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntSize       MaximumFilenameLength = 260;

        /**
         *  The canonical separator of components within a path.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           PathComponentSeparator = '/';

        /**
         *  The canonical separator of paths in a path list string.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           PathListSeparator = ':';

        /**
         *  True if the host file system uses named devices (file system 
         *  roots), false otherwise.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const bool           UsesNamedDevices = false;

        /**
         *  The canonical device name separator for absolute paths.
         *
         *  If the host file system does not support the concept of
         *  named file system, roots (Unix is one such example), then
         *  device separators are not used.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           DeviceSeparator = ' ';

        /**
         *  The canonical filename extension separator.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           FilenameExtensionSeparator = '.';

        /**
         *  The character that, when used in a file name pattern, means
         *  "match any sequence of characters, inclusing an empty one".
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           AnyStringMetacharacter = '*';

        /**
         *  The character that, when used in a file name pattern, means
         *  "match any single character".
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const Char           AnyCharacterMetacharacter = '?';

        /**
         *  True if the host file system uses case-sensitive file
         *  names, false otherwise.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const bool           CaseSensitiveFileNames = false;

        /**
         *  A nul-terminated string that, when used as a path component,
         *  refers to the "current directory".
         *
         *  @since 1.0
         */
        static const char *         ThisDirectoryName;

        /**
         *  A nul-terminated string that, when used as a path component,
         *  refers to the "parent directory".
         *
         *  @since 1.0
         */
        static const char *         ParentDirectoryName;

        /**
         *  A nul-terminated sequence of characters that is used as line
         *  separator in host file system's text files.
         *
         *  @since 1.0
         */
        static const char *         LineSeparator;
#else
    #error Unsupported platform
#endif
        /**
         *  Checks if a character is a valid path component separator.
         *
         *  Each platform always has a canonical path component separator,
         *  which is the preferred character to be used in compound paths.
         *  In addition, some platforms (e.g. Win32) allow some other 
         *  characters to be used as path component separators.
         *
         *  @param c                The character to check.
         *  @return                 True if \a c is a valid path component
         *                          separator for the host platform, false
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 isPathComponentSeparator(Char c);

        /**
         *  Checks if a character is a valid path list separator.
         *
         *  Each platform always has a canonical path list separator,
         *  which is the preferred character to be used in strings
         *  that represent ordered lists of paths.
         *  In addition, some platforms may allow some other characters 
         *  to be used as path list separators.
         *
         *  @param c                The character to check.
         *  @return                 True if \a c is a valid path list
         *                          separator for the host platform, false
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 isPathListSeparator(Char c);

        /**
         *  Checks if a character is a valid filename extension separator.
         *
         *  Each platform always has a canonical filename extension
         *  separator, which is the preferred character to be used 
         *  for separatuing a name and an extension of a file system
         *  entity. In addition, some platforms may allow some other 
         *  characters to be used as filename extension separators.
         *
         *  @param c                The character to check.
         *  @return                 True if \a c is a valid filename extension
         *                          separator for the host platform, false
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 isFilenameExtensionSeparator(Char c);

        /**
         *  Checks if a character is a valid device separator.
         *
         *  Each platform always has a canonical device separator, 
         *  which is the preferred character to be used for 
         *  separatuing a file system root name (also known as a
         *  device name) from a path. In addition, some platforms 
         *  may allow some other characters to be used as device
         *  separators.
         *
         *  If the host file system does not support the concept of
         *  named file system, roots (Unix is one such example), then
         *  device separators are not used.
         *
         *  @param c                The character to check.
         *  @return                 True if \a c is a valid filename extension
         *                          separator for the host platform, false
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 isDeviceSeparator(Char c);

        //////////
        //  File system services
    public:
        /**
         *  Represents various media types used by file system.
         *
         *  @since 1.0
         */
        enum MediaType
        {
            /**
             *  The media type is unknown.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            MediaType_Unknown       = 0,

            /**
             *  Fixed media (nonremovable hard drive).
             *
             *  @hideinitializer
             *  @since 1.0
             */
            MediaType_Fixed         = 1,

            /**
             *  Removable media (floppy drive or removable hard drive).
             *
             *  @hideinitializer
             *  @since 1.0
             */
            MediaType_Removable     = 2,

            /**
             *  CD-ROM or DVD-ROM drive.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            MediaType_CdRom         = 3,

            /**
             *  RAM drive.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            MediaType_RamDisk       = 4,

            /**
             *  Remote (network) drive.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            MediaType_Remote        = 5
        };

        /**
         *  Returns the current working directory of the calling process.
         *
         *  @return                 The current working directory.
         *  @since 1.0
         */
        static PathName             getCurrentDirectory();

        /**
         *  Sets the current working directory for the calling process.
         *
         *  @param pCurrentDirectory    A pointer to a nul-terminated ASCII
         *                          string specifying the new current working
         *                          directory.
         *  @return                 True if the current working directory has
         *                          been changed successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 setCurrentDirectory(const char * pCurrentDirectory);

        /**
         *  Sets the current working directory for the calling process.
         *
         *  @param pCurrentDirectory    A pointer to a nul-terminated ISO-10646
         *                          string specifying the new current working
         *                          directory.
         *  @return                 True if the current working directory has
         *                          been changed successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 setCurrentDirectory(const Char * pCurrentDirectory);

        /**
         *  Sets the current working directory for the calling process.
         *
         *  @param currentDirectory A string specifying the new current 
         *                          working directory.
         *  @return                 True if the current working directory has
         *                          been changed successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 setCurrentDirectory(const String & currentDirectory);

        /**
         *  Sets the current working directory for the calling process.
         *
         *  @param currentDirectory A path to the new current working 
         *                          directory.
         *  @return                 True if the current working directory has
         *                          been changed successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 setCurrentDirectory(const PathName & currentDirectory);

        /**
         *  Returns a path to the temporary directory.
         *
         *  A temporary directory is a preferred location to create
         *  temporary files.
         *
         *  @return                 A path to the temporary directory, or
         *                          an empty path if the temporary directory
         *                          could not be determined.
         *  @since 1.0
         */
        static PathName             getTempDirectory();

        /**
         *  Returns a path to the user home directory.
         *
         *  A user home directory is a directory where the user
         *  keeps his personal files.
         *
         *  @return                 A path to the user home directory, or
         *                          an empty path if the user home directory
         *                          could not be determined.
         *  @since 1.0
         */
        static PathName             getHomeDirectory();

        /**
         *  Creates a single directory.
         *
         *  This method cannot be used to create several levels of a
         *  directory hierarchy (use createDirectories instead).
         *
         *  @param pDirectory       A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          directory to be created.
         *  @return                 True if the directory has been created
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 createDirectory(const char * pDirectory);

        /**
         *  Creates a single directory.
         *
         *  This method cannot be used to create several levels of a
         *  directory hierarchy (use createDirectories instead).
         *
         *  @param pDirectory       A pointer to a nul-terminated ISO-10646 
         *                          string specifying the name of the
         *                          directory to be created.
         *  @return                 True if the directory has been created
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 createDirectory(const Char * pDirectory);

        /**
         *  Creates a single directory.
         *
         *  This method cannot be used to create several levels of a
         *  directory hierarchy (use createDirectories instead).
         *
         *  @param directory        A string specifying the name of the
         *                          directory to be created.
         *  @return                 True if the directory has been created
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 createDirectory(const String & directory);

        /**
         *  Creates a single directory.
         *
         *  This method cannot be used to create several levels of a
         *  directory hierarchy (use createDirectories instead).
         *
         *  @param directory        A path to the directory to be created.
         *  @return                 True if the directory has been created
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 createDirectory(const PathName & directory);

        /**
         *  Creates multiple directories.
         *
         *  This method can be used to create several levels of the
         *  directory hierarchy all in one go.
         *
         *  @param pDirectory       A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          directory to be created.
         *  @return                 True if the specified directory exists
         *                          after the call; false otherwise.
         *  @since 1.0
         */
        static bool                 createDirectories(const char * pDirectory);

        /**
         *  Creates multiple directories.
         *
         *  This method can be used to create several levels of the
         *  directory hierarchy all in one go.
         *
         *  @param pDirectory       A pointer to a nul-terminated ISO-10646
         *                          string specifying the name of the
         *                          directory to be created.
         *  @return                 True if the specified directory exists
         *                          after the call; false otherwise.
         *  @since 1.0
         */
        static bool                 createDirectories(const Char * pDirectory);

        /**
         *  Creates multiple directories.
         *
         *  This method can be used to create several levels of the
         *  directory hierarchy all in one go.
         *
         *  @param directory        A string specifying the name of the
         *                          directory to be created.
         *  @return                 True if the specified directory exists
         *                          after the call; false otherwise.
         *  @since 1.0
         */
        static bool                 createDirectories(const String & directory);

        /**
         *  Creates multiple directories.
         *
         *  This method can be used to create several levels of the
         *  directory hierarchy all in one go.
         *
         *  @param directory        A path to the directory to be created.
         *  @return                 True if the specified directory exists
         *                          after the call; false otherwise.
         *  @since 1.0
         */
        static bool                 createDirectories(const PathName & directory);

        /**
         *  Deletes a file.
         *
         *  This method cannot be used to delete directories.
         *
         *  @param pFileName        A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          file to be deleted.
         *  @return                 True if the file has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 deleteFile(const char * pFileName);

        /**
         *  Deletes a file.
         *
         *  This method cannot be used to delete directories.
         *
         *  @param pFileName        A pointer to a nul-terminated ISO-10646
         *                          string specifying the name of the
         *                          file to be deleted.
         *  @return                 True if the file has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 deleteFile(const Char * pFileName);

        /**
         *  Deletes a file.
         *
         *  This method cannot be used to delete directories.
         *
         *  @param fileName         A string specifying the name of the
         *                          file to be deleted.
         *  @return                 True if the file has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 deleteFile(const String & fileName);

        /**
         *  Deletes a file.
         *
         *  This method cannot be used to delete directories.
         *
         *  @param fileName         A path to the file to be deleted.
         *  @return                 True if the file has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 deleteFile(const PathName & fileName);

        /**
         *  Deletes a directory.
         *
         *  A directory must be empty before it can be deleted.
         *
         *  This method cannot be used to delete files.
         *
         *  @param pDirectory       A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          directory to be deleted.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 deleteDirectory(const char * pDirectory);

        /**
         *  Deletes a directory.
         *
         *  A directory must be empty before it can be deleted.
         *
         *  This method cannot be used to delete files.
         *
         *  @param pDirectory       A pointer to a nul-terminated ISO-10646
         *                          string specifying the name of the
         *                          directory to be deleted.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 deleteDirectory(const Char * pDirectory);

        /**
         *  Deletes a directory.
         *
         *  A directory must be empty before it can be deleted.
         *
         *  This method cannot be used to delete files.
         *
         *  @param directory        A string specifying the name of the
         *                          directory to be deleted.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 deleteDirectory(const String & directory);

        /**
         *  Deletes a directory.
         *
         *  A directory must be empty before it can be deleted.
         *
         *  This method cannot be used to delete files.
         *
         *  @param directory        A path to the directory to be deleted.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 deleteDirectory(const PathName & directory);

        /**
         *  Renames a file or a directory.
         *
         *  @param pOldFileName     A pointer to a nul-terminated ASCII
         *                          string specifying the name of the file or
         *                          directory to be renamed.
         *  @param pNewFileName     A pointer to a nul-terminated ASCII
         *                          string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const char * pOldFileName, const char * pNewFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param pOldFileName     A pointer to a nul-terminated ISO-10646
         *                          string specifying the name of the file or
         *                          directory to be renamed.
         *  @param pNewFileName     A pointer to a nul-terminated ASCII
         *                          string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const Char * pOldFileName, const char * pNewFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param oldFileName      A string specifying the name of the file or
         *                          directory to be renamed.
         *  @param pNewFileName     A pointer to a nul-terminated ASCII
         *                          string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const String & oldFileName, const char * pNewFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param oldFileName      A path to the file or directory to be 
         *                          renamed.
         *  @param pNewFileName     A pointer to a nul-terminated ASCII
         *                          string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const PathName & oldFileName, const char * pNewFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param pOldFileName     A pointer to a nul-terminated ASCII
         *                          string specifying the name of the file or
         *                          directory to be renamed.
         *  @param pNewFileName     A pointer to a nul-terminated ISO-10646
         *                          string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const char * pOldFileName, const Char * pNewFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param pOldFileName     A pointer to a nul-terminated ISO-10646
         *                          string specifying the name of the file or
         *                          directory to be renamed.
         *  @param pNewFileName     A pointer to a nul-terminated ISO-10646
         *                          string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const Char * pOldFileName, const Char * pNewFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param oldFileName      A string specifying the name of the file or
         *                          directory to be renamed.
         *  @param pNewFileName     A pointer to a nul-terminated ISO-10646
         *                          string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const String & oldFileName, const Char * pNewFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param oldFileName      A path to the file or directory to be 
         *                          renamed.
         *  @param pNewFileName     A pointer to a nul-terminated ISO-10646
         *                          string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const PathName & oldFileName, const Char * pNewFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param pOldFileName     A pointer to a nul-terminated ASCII
         *                          string specifying the name of the file or
         *                          directory to be renamed.
         *  @param newFileName      A string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const char * pOldFileName, const String & newFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param pOldFileName     A pointer to a nul-terminated ISO-10646
         *                          string specifying the name of the file or
         *                          directory to be renamed.
         *  @param newFileName      A string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const Char * pOldFileName, const String & newFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param oldFileName      A string specifying the name of the file or
         *                          directory to be renamed.
         *  @param newFileName      A string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const String & oldFileName, const String & newFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param oldFileName      A path to the file or directory to be 
         *                          renamed.
         *  @param newFileName      A string specifying the new name for the
         *                          file or directory being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const PathName & oldFileName, const String & newFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param pOldFileName     A pointer to a nul-terminated ASCII
         *                          string specifying the name of the file or
         *                          directory to be renamed.
         *  @param newFileName      A new path for the file or directory 
         *                          being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const char * pOldFileName, const PathName & newFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param pOldFileName     A pointer to a nul-terminated ISo-10646
         *                          string specifying the name of the file or
         *                          directory to be renamed.
         *  @param newFileName      A new path for the file or directory 
         *                          being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const Char * pOldFileName, const PathName & newFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param oldFileName      A string specifying the name of the file or
         *                          directory to be renamed.
         *  @param newFileName      A new path for the file or directory 
         *                          being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const String & oldFileName, const PathName & newFileName);

        /**
         *  Renames a file or a directory.
         *
         *  @param oldFileName      A path to the file or directory to be 
         *                          renamed.
         *  @param newFileName      A new path for the file or directory 
         *                          being renamed.
         *  @return                 True if the directory has been deleted
         *                          successfully; false otherwise.
         *  @since 1.0
         */
        static bool                 renameFile(const PathName & oldFileName, const PathName & newFileName);

        /**
         *  Queries time stamps associated with a file or directory.
         *
         *  @param pFileOrDirectoryName A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          file or directory to be queried.
         *  @param creationTime     Upon successful query, this parameter is
         *                          assigned the creation time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastAccessTime   Upon successful query, this parameter is
         *                          assigned the last access time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastWriteTime    Upon successful query, this parameter is
         *                          assigned the last write time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @return                 True if the file or directory time stamps
         *                          have been queried successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 getFileTimes(const char * pFileOrDirectoryName,
                                            TimeStamp & creationTime, 
                                            TimeStamp & lastAccessTime, 
                                            TimeStamp & lastWriteTime);

        /**
         *  Queries time stamps associated with a file or directory.
         *
         *  @param pFileOrDirectoryName A pointer to a nul-terminated ISO-10646
         *                          string specifying the name of the
         *                          file or directory to be queried.
         *  @param creationTime     Upon successful query, this parameter is
         *                          assigned the creation time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastAccessTime   Upon successful query, this parameter is
         *                          assigned the last access time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastWriteTime    Upon successful query, this parameter is
         *                          assigned the last write time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @return                 True if the file or directory time stamps
         *                          have been queried successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 getFileTimes(const Char * pFileOrDirectoryName,
                                            TimeStamp & creationTime, 
                                            TimeStamp & lastAccessTime, 
                                            TimeStamp & lastWriteTime);

        /**
         *  Queries time stamps associated with a file or directory.
         *
         *  @param fileOrDirectoryName  A string specifying the name of the
         *                          file or directory to be queried.
         *  @param creationTime     Upon successful query, this parameter is
         *                          assigned the creation time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastAccessTime   Upon successful query, this parameter is
         *                          assigned the last access time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastWriteTime    Upon successful query, this parameter is
         *                          assigned the last write time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @return                 True if the file or directory time stamps
         *                          have been queried successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 getFileTimes(const String & fileOrDirectoryName,
                                            TimeStamp & creationTime, 
                                            TimeStamp & lastAccessTime, 
                                            TimeStamp & lastWriteTime);

        /**
         *  Queries time stamps associated with a file or directory.
         *
         *  @param fileOrDirectoryName  A path to the file or directory to be 
         *                          queried.
         *  @param creationTime     Upon successful query, this parameter is
         *                          assigned the creation time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastAccessTime   Upon successful query, this parameter is
         *                          assigned the last access time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastWriteTime    Upon successful query, this parameter is
         *                          assigned the last write time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @return                 True if the file or directory time stamps
         *                          have been queried successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 getFileTimes(const PathName & fileOrDirectoryName,
                                            TimeStamp & creationTime, 
                                            TimeStamp & lastAccessTime, 
                                            TimeStamp & lastWriteTime);

        /**
         *  Sets time stamps associated with a file or directory.
         *
         *  @param pFileOrDirectoryName A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          file or directory whose time stamps
         *                          are to be modified.
         *  @param creationTime     The new creation time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastAccessTime   The new last access time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastWriteTime    The new last write time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @return                 True if the file or directory time stamps
         *                          have been modified successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 setFileTimes(const char * pFileOrDirectoryName,
                                            const TimeStamp & creationTime, 
                                            const TimeStamp & lastAccessTime, 
                                            const TimeStamp & lastWriteTime);

        /**
         *  Sets time stamps associated with a file or directory.
         *
         *  @param pFileOrDirectoryName A pointer to a nul-terminated ISo-10646
         *                          string specifying the name of the
         *                          file or directory whose time stamps
         *                          are to be modified.
         *  @param creationTime     The new creation time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastAccessTime   The new last access time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastWriteTime    The new last write time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @return                 True if the file or directory time stamps
         *                          have been modified successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 setFileTimes(const Char * pFileOrDirectoryName,
                                            const TimeStamp & creationTime, 
                                            const TimeStamp & lastAccessTime, 
                                            const TimeStamp & lastWriteTime);

        /**
         *  Sets time stamps associated with a file or directory.
         *
         *  @param fileOrDirectoryName  A string specifying the name of the
         *                          file or directory whose time stamps
         *                          are to be modified.
         *  @param creationTime     The new creation time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastAccessTime   The new last access time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastWriteTime    The new last write time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @return                 True if the file or directory time stamps
         *                          have been modified successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 setFileTimes(const String & fileOrDirectoryName,
                                            const TimeStamp & creationTime, 
                                            const TimeStamp & lastAccessTime, 
                                            const TimeStamp & lastWriteTime);

        /**
         *  Sets time stamps associated with a file or directory.
         *
         *  @param fileOrDirectoryName  A path to the file or directory whose 
         *                          time stamps are to be modified.
         *  @param creationTime     The new creation time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastAccessTime   The new last access time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @param lastWriteTime    The new last write time of the specified
         *                          file or directory, or an invalid time if
         *                          this information is not available.
         *  @return                 True if the file or directory time stamps
         *                          have been modified successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 setFileTimes(const PathName & fileOrDirectoryName,
                                            const TimeStamp & creationTime, 
                                            const TimeStamp & lastAccessTime, 
                                            const TimeStamp & lastWriteTime);

        /**
         *  Queries the size of a file or directory.
         *
         *  @param pFileOrDirectoryName A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          file or directory to be queried.
         *  @param size             Upon successful query, this parameter is
         *                          assigned the size, in bytes, of the
         *                          specified file or directory.
         *  @return                 True if the file or directory size
         *                          has been queried successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 getFileSize(const char * pFileOrDirectoryName, UInt64 & size);

        /**
         *  Queries the size of a file or directory.
         *
         *  @param pFileOrDirectoryName A pointer to a nul-terminated ISO-10646
         *                          string specifying the name of the
         *                          file or directory to be queried.
         *  @param size             Upon successful query, this parameter is
         *                          assigned the size, in bytes, of the
         *                          specified file or directory.
         *  @return                 True if the file or directory size
         *                          has been queried successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 getFileSize(const Char * pFileOrDirectoryName, UInt64 & size);

        /**
         *  Queries the size of a file or directory.
         *
         *  @param fileOrDirectoryName  A string specifying the name of the
         *                          file or directory to be queried.
         *  @param size             Upon successful query, this parameter is
         *                          assigned the size, in bytes, of the
         *                          specified file or directory.
         *  @return                 True if the file or directory size
         *                          has been queried successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 getFileSize(const String & fileOrDirectoryName, UInt64 & size);

        /**
         *  Queries the size of a file or directory.
         *
         *  @param fileOrDirectoryName  A path to the file or directory to 
         *                          be queried.
         *  @param size             Upon successful query, this parameter is
         *                          assigned the size, in bytes, of the
         *                          specified file or directory.
         *  @return                 True if the file or directory size
         *                          has been queried successfully; false 
         *                          otherwise.
         *  @since 1.0
         */
        static bool                 getFileSize(const PathName & fileOrDirectoryName, UInt64 & size);

        /**
         *  Generates a temporary file name in default temporary directory.
         *
         *  The generated path is guaranteed not to refer to any file system
         *  entity that exists at the moment of temporary file name generation.
         *
         *  @return                 A path to a temporary file, or an empty
         *                          path name if a temporary file name could
         *                          not be generated.
         *  @since 1.0
         */
        static PathName             getTempFileName();

        /**
         *  Generates a temporary file name in the specified directory.
         *
         *  The generated path is guaranteed not to refer to any file system
         *  entity that exists at the moment of temporary file name generation.
         *
         *  @param pDirectory       A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          directory where the temporary file
         *                          shall be placed. This directory must
         *                          already exist.
         *  @return                 A path to a temporary file, or an empty
         *                          path name if a temporary file name could
         *                          not be generated.
         *  @since 1.0
         */
        static PathName             getTempFileName(const char * pDirectory);

        /**
         *  Generates a temporary file name in the specified directory.
         *
         *  The generated path is guaranteed not to refer to any file system
         *  entity that exists at the moment of temporary file name generation.
         *
         *  @param pDirectory       A pointer to a nul-terminated ISO-10646
         *                          string specifying the name of the
         *                          directory where the temporary file
         *                          shall be placed. This directory must
         *                          already exist.
         *  @return                 A path to a temporary file, or an empty
         *                          path name if a temporary file name could
         *                          not be generated.
         *  @since 1.0
         */
        static PathName             getTempFileName(const Char * pDirectory);

        /**
         *  Generates a temporary file name in the specified directory.
         *
         *  The generated path is guaranteed not to refer to any file system
         *  entity that exists at the moment of temporary file name generation.
         *
         *  @param directory        A string specifying the name of the
         *                          directory where the temporary file
         *                          shall be placed. This directory must
         *                          already exist.
         *  @return                 A path to a temporary file, or an empty
         *                          path name if a temporary file name could
         *                          not be generated.
         *  @since 1.0
         */
        static PathName             getTempFileName(const String & directory);

        /**
         *  Generates a temporary file name in the specified directory.
         *
         *  The generated path is guaranteed not to refer to any file system
         *  entity that exists at the moment of temporary file name generation.
         *
         *  @param directory        A path to the directory where the temporary 
         *                          file shall be placed. This directory must
         *                          already exist.
         *  @return                 A path to a temporary file, or an empty
         *                          path name if a temporary file name could
         *                          not be generated.
         *  @since 1.0
         */
        static PathName             getTempFileName(const PathName & directory);

        /**
         *  Returns a list of all file system roots.
         *
         *  If a host file system does not support the concept of named
         *  file system roots (devices) the returned list contains only
         *  one element (e.g '/' on Unix).
         *
         *  Elements in the returned list are not guarannteed to be in any
         *  particular order.
         *
         *  @return                 A list of all file system roots.
         *  @since 1.0
         */
        static Array<PathName>      getFileSystemRoots();

        /**
         *  Queries the total space of a file system under a specified directory.
         *
         *  @param pDirectory       A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          directory residing in the file system
         *                          whose total space is to be queried.
         *  @param totalSpace       Upon successful query, this parameter is
         *                          assigned a total size, in bytes, of the
         *                          file system where the specified directory
         *                          resides.
         *  @return                 True if the total file system size has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getTotalSpace(const char * pDirectory, UInt64 & totalSpace);

        /**
         *  Queries the total space of a file system under a specified directory.
         *
         *  @param pDirectory       A pointer to a nul-terminated ISO-10646
         *                          string specifying the name of the
         *                          directory residing in the file system
         *                          whose total space is to be queried.
         *  @param totalSpace       Upon successful query, this parameter is
         *                          assigned a total size, in bytes, of the
         *                          file system where the specified directory
         *                          resides.
         *  @return                 True if the total file system size has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getTotalSpace(const Char * pDirectory, UInt64 & totalSpace);

        /**
         *  Queries the total space of a file system under a specified directory.
         *
         *  @param directory        A string specifying the name of the
         *                          directory residing in the file system
         *                          whose total space is to be queried.
         *  @param totalSpace       Upon successful query, this parameter is
         *                          assigned a total size, in bytes, of the
         *                          file system where the specified directory
         *                          resides.
         *  @return                 True if the total file system size has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getTotalSpace(const String & directory, UInt64 & totalSpace);

        /**
         *  Queries the total space of a file system under a specified directory.
         *
         *  @param directory        A path to the directory residing in the file 
         *                          system whose total space is to be queried.
         *  @param totalSpace       Upon successful query, this parameter is
         *                          assigned a total size, in bytes, of the
         *                          file system where the specified directory
         *                          resides.
         *  @return                 True if the total file system size has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getTotalSpace(const PathName & directory, UInt64 & totalSpace);

        /**
         *  Queries the free space of a file system under a specified directory.
         *
         *  The "free space of a file system" is defined as an amount of space
         *  that can be used by the calling process. Specifically, if file system
         *  space quotas are in use, the returned value may be less than the
         *  total available free space.
         *
         *  @param pDirectory       A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          directory residing in the file system
         *                          whose free space is to be queried.
         *  @param freeSpace        Upon successful query, this parameter is
         *                          assigned a free space size, in bytes, of the
         *                          file system where the specified directory
         *                          resides.
         *  @return                 True if the free file system size has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getFreeSpace(const char * pDirectory, UInt64 & freeSpace);

        /**
         *  Queries the free space of a file system under a specified directory.
         *
         *  The "free space of a file system" is defined as an amount of space
         *  that can be used by the calling process. Specifically, if file system
         *  space quotas are in use, the returned value may be less than the
         *  total available free space.
         *
         *  @param pDirectory       A pointer to a nul-terminated ISo-10646
         *                          string specifying the name of the
         *                          directory residing in the file system
         *                          whose free space is to be queried.
         *  @param freeSpace        Upon successful query, this parameter is
         *                          assigned a free space size, in bytes, of the
         *                          file system where the specified directory
         *                          resides.
         *  @return                 True if the free file system size has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getFreeSpace(const Char * pDirectory, UInt64 & freeSpace);

        /**
         *  Queries the free space of a file system under a specified directory.
         *
         *  The "free space of a file system" is defined as an amount of space
         *  that can be used by the calling process. Specifically, if file system
         *  space quotas are in use, the returned value may be less than the
         *  total available free space.
         *
         *  @param directory        A string specifying the name of the
         *                          directory residing in the file system
         *                          whose free space is to be queried.
         *  @param freeSpace        Upon successful query, this parameter is
         *                          assigned a free space size, in bytes, of the
         *                          file system where the specified directory
         *                          resides.
         *  @return                 True if the free file system size has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getFreeSpace(const String & directory, UInt64 & freeSpace);

        /**
         *  Queries the free space of a file system under a specified directory.
         *
         *  The "free space of a file system" is defined as an amount of space
         *  that can be used by the calling process. Specifically, if file system
         *  space quotas are in use, the returned value may be less than the
         *  total available free space.
         *
         *  @param directory        A path to the directory residing in the file 
         *                          system whose free space is to be queried.
         *  @param freeSpace        Upon successful query, this parameter is
         *                          assigned a free space size, in bytes, of the
         *                          file system where the specified directory
         *                          resides.
         *  @return                 True if the free file system size has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getFreeSpace(const PathName & directory, UInt64 & freeSpace);

        /**
         *  Queries the media type of a file system for the specified directory.
         *
         *  @param pDirectory       A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          directory residing in the file system
         *                          whose media type is to be queried.
         *  @param mediaType        Upon successful query, this parameter is
         *                          assigned a media type of the file system 
         *                          where the specified directory resides.
         *  @return                 True if the file system media type has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getMediaType(const char * pDirectory, MediaType & mediaType);

        /**
         *  Queries the media type of a file system for the specified directory.
         *
         *  @param pDirectory       A pointer to a nul-terminated ISo-10646
         *                          string specifying the name of the
         *                          directory residing in the file system
         *                          whose media type is to be queried.
         *  @param mediaType        Upon successful query, this parameter is
         *                          assigned a media type of the file system 
         *                          where the specified directory resides.
         *  @return                 True if the file system media type has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getMediaType(const Char * pDirectory, MediaType & mediaType);

        /**
         *  Queries the media type of a file system for the specified directory.
         *
         *  @param directory        A string specifying the name of the
         *                          directory residing in the file system
         *                          whose media type is to be queried.
         *  @param mediaType        Upon successful query, this parameter is
         *                          assigned a media type of the file system 
         *                          where the specified directory resides.
         *  @return                 True if the file system media type has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getMediaType(const String & directory, MediaType & mediaType);

        /**
         *  Queries the media type of a file system for the specified directory.
         *
         *  @param directory        A path to the directory residing in the file 
         *                          system whose media type is to be queried.
         *  @param mediaType        Upon successful query, this parameter is
         *                          assigned a media type of the file system 
         *                          where the specified directory resides.
         *  @return                 True if the file system media type has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getMediaType(const PathName & directory, MediaType & mediaType);

        /**
         *  Queries the contents of the specified directory.
         *
         *  @param pDirectory       A pointer to a nul-terminated ASCII
         *                          string specifying the name of the
         *                          directory whose contents is queried.
         *  @param directoryEntries Upon successful query, this parameter is
         *                          assigned a list of all entries of the
         *                          specified directory that are accepted by
         *                          the specified filter.
         *  @param pFilter          The filter to use for deciding which directory
         *                          entries to include in the result and which
         *                          to ignore. If a filter is not specified, all
         *                          directory entries are included.
         *  @return                 True if the directory contents has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getDirectoryContents(const char * pDirectory, Array<DirectoryEntry> & directoryEntries, IDirectoryEntryFilter * pFilter = NULL);

        /**
         *  Queries the contents of the specified directory.
         *
         *  @param pDirectory       A pointer to a nul-terminated ISo-10646
         *                          string specifying the name of the
         *                          directory whose contents is queried.
         *  @param directoryEntries Upon successful query, this parameter is
         *                          assigned a list of all entries of the
         *                          specified directory that are accepted by
         *                          the specified filter.
         *  @param pFilter          The filter to use for deciding which directory
         *                          entries to include in the result and which
         *                          to ignore. If a filter is not specified, all
         *                          directory entries are included.
         *  @return                 True if the directory contents has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getDirectoryContents(const Char * pDirectory, Array<DirectoryEntry> & directoryEntries, IDirectoryEntryFilter * pFilter = NULL);

        /**
         *  Queries the contents of the specified directory.
         *
         *  @param directory        A string specifying the name of the
         *                          directory whose contents is queried.
         *  @param directoryEntries Upon successful query, this parameter is
         *                          assigned a list of all entries of the
         *                          specified directory that are accepted by
         *                          the specified filter.
         *  @param pFilter          The filter to use for deciding which directory
         *                          entries to include in the result and which
         *                          to ignore. If a filter is not specified, all
         *                          directory entries are included.
         *  @return                 True if the directory contents has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getDirectoryContents(const String & directory, Array<DirectoryEntry> & directoryEntries, IDirectoryEntryFilter * pFilter = NULL);

        /**
         *  Queries the contents of the specified directory.
         *
         *  @param directory        A path to the directory whose contents is 
         *                          queried.
         *  @param directoryEntries Upon successful query, this parameter is
         *                          assigned a list of all entries of the
         *                          specified directory that are accepted by
         *                          the specified filter.
         *  @param pFilter          The filter to use for deciding which directory
         *                          entries to include in the result and which
         *                          to ignore. If a filter is not specified, all
         *                          directory entries are included.
         *  @return                 True if the directory contents has been
         *                          queried successfully, false otherwise.
         *  @since 1.0
         */
        static bool                 getDirectoryContents(const PathName & directory, Array<DirectoryEntry> & directoryEntries, IDirectoryEntryFilter * pFilter = NULL);

        /**
         *  Splits a path list string into a list of path names.
         *
         *  Individual paths in a path list string are assumed to be separated
         *  by a platform-specific path component separator. Empty paths are not
         *  included into the list.
         *
         *  @param pPathListString  A pointer to a nul-terminated ASCII
         *                          path list string.
         *  @return                 A list of path names. Elements in the list
         *                          are always in the same order in which they
         *                          were mentioned in a path list string.
         *  @since 1.0
         */
        static Array<PathName>      splitPathListString(const char * pPathListString);

        /**
         *  Splits a path list string into a list of path names.
         *
         *  Individual paths in a path list string are assumed to be separated
         *  by a platform-specific path component separator. Empty paths are not
         *  included into the list.
         *
         *  @param pPathListString  A pointer to a nul-terminated ISO-10646
         *                          path list string.
         *  @return                 A list of path names. Elements in the list
         *                          are always in the same order in which they
         *                          were mentioned in a path list string.
         *  @since 1.0
         */
        static Array<PathName>      splitPathListString(const Char * pPathListString);

        /**
         *  Splits a path list string into a list of path names.
         *
         *  Individual paths in a path list string are assumed to be separated
         *  by a platform-specific path component separator. Empty paths are not
         *  included into the list.
         *
         *  @param pathListString   A path list string.
         *  @return                 A list of path names. Elements in the list
         *                          are always in the same order in which they
         *                          were mentioned in a path list string.
         *  @since 1.0
         */
        static Array<PathName>      splitPathListString(const String & pathListString);

        /**
         *  Builds a path list string from a list of path names.
         *
         *  Elements of a produced path list string are separated by the
         *  platform-specific canonical path list separator. All empty
         *  path names in the specified list are ignored. Paths appear
         *  in the generated path list string in the same order they are
         *  in the specified path names list.
         *
         *  @param pathNames        A list of path names.
         *  @return                 A generated path list string containing all
         *                          non-empty path names from the specified list.
         *  @since 1.0
         */
        static String               buildPathListString(const Array<PathName> & pathNames);
    };
}

#endif  //  __CARL_FileSystem_hpp_included__
//  End of CARL_FileSystem.hpp
