//
//  CARL_Format.hpp - Value formatting
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.22 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL value formatting framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Format.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Format_hpp_included__
#define __CARL_Format_hpp_included__

namespace carl
{
    namespace _impl
    {
        //////////
        //  Format managers
        class CARL_PUBLIC _NumberFormatManager;
        class CARL_PUBLIC _DateTimeFormatManager;
    }

    /**
     *  @class IntegerFormatProperties CARL.hpp
     *  Format properties governing formatiting of integer values.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IntegerFormatProperties
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  Represents various alignments of formatted integer value.
         *
         *  @since 1.0
         */
        enum Alignment
        {
            /**
             *  Align the formatted integer value to the left.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Alignment_Left      = 1,

            /**
             *  Align the formatted integer value to the right.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Alignment_Right     = 2
        };

        /**
         *  Represents letter cases to use in a formatted integer value.
         *
         *  @since 1.0
         */
        enum LetterCase
        {
            /**
             *  All letters in a formatted integer value shall be lowercase.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            LetterCase_Lower    = 1,

            /**
             *  All letters in a formatted integer value shall be uppercase.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            LetterCase_Upper    = 2,

            /**
             *  All letters in a formatted integer value shall be titlecase.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            LetterCase_Title    = 3
        };

        /**
         *  The default constructor.
         *
         *  Iniitializes a newly created integer format properties to default
         *  values:
         *
         *  -   Radix = 10.
         *  -   No minimum or maximum limits on the formatted value length
         *      or a number of significant digits to format.
         *  -   Use space as a filler.
         *  -   Use asterisk (*) as an error indicator.
         *  -   Formatted value is aligned to the right.
         *  -   Use lowercase letters if letters are necessary.
         *  -   Do not use prefix or thousand separator.
         *  -   Omit the '+' sign for positive values.
         *
         *  @since 1.0
         */
        IntegerFormatProperties();

        /**
         *  The class constructor.
         *
         *  Initializes a newly created integer format properties instance to
         *  specify the same formatting rules as would be effected by the
         *  specified string in CRT printf() family of functions.
         *
         *  @param pCrtFormatString A pointer to nul-terminated ASCII string
         *                          representing the CRT printf()-style
         *                          integer format specification.
         *  @since 1.0
         */
        explicit IntegerFormatProperties(const char * pCrtFormatString);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created integer format properties instance to
         *  specify the same formatting rules as would be effected by the
         *  specified string in CRT printf() family of functions.
         *
         *  @param pCrtFormatString A pointer to nul-terminated ISO-10646 string
         *                          representing the CRT printf()-style
         *                          integer format specification.
         *  @since 1.0
         */
        explicit IntegerFormatProperties(const Char * pCrtFormatString);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created integer format properties instance to
         *  specify the same formatting rules as would be effected by the
         *  specified string in CRT printf() family of functions.
         *
         *  @param crtFormatString  A string representing the CRT printf()-style
         *                          integer format specification.
         *  @since 1.0
         */
        explicit IntegerFormatProperties(const String & crtFormatString);

        /**
         *  The copy constructor.
         *
         *  The newly created integer format properties instance is initialized to 
         *  be an exact copy of an existing integer format properties.
         *
         *  @param src              The integer format properties name to copy 
         *                          properties from.
         *  @since 1.0
         */
        IntegerFormatProperties(const IntegerFormatProperties & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~IntegerFormatProperties() {}

        /**
         *  The assigment operator.
         *
         *  Makes an integer format properties an exact copy of an existing 
         *  integer format properties.
         *
         *  @param src              The integer format properties to copy properties from.
         *  @return                 A reference to an assigned-to object.
         *  @since 1.0
         */
        IntegerFormatProperties &   operator = (const IntegerFormatProperties & src);

        //////////
        //  Operations
    public:
        /**
         *  Changes the radix of the integer format specification.
         *
         *  The new radix must be in range 2..36.
         *
         *  @param radix            The new radix.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setRadix(UIntSize radix);

        /**
         *  Changes the minimum number of significant digits of the integer 
         *  format specification.
         *
         *  If the minimum number of significant digits of the integer 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the integer value being formatted has fewer significant digits
         *  than required by integer format properties that control the
         *  formatting, the value is padded with leading zeroes until the
         *  required minimum is met.
         *
         *  @param digits           The new minimum number of significant digits.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setMinimumDigits(UIntSize digits);

        /**
         *  Changes the maximum number of significant digits of the integer 
         *  format specification.
         *
         *  If the maximum number of significant digits of the integer 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the integer value being formatted has more significant digits
         *  than required by integer format properties that control the
         *  formatting, all value digits are replaced with an error
         *  indicator characters.
         *
         *  @param digits           The new maximum number of significant digits.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setMaximumDigits(UIntSize digits);

        /**
         *  Changes the minimum width of the formatted value.
         *
         *  If the minimum width of the formatted value of the integer 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the integer value being formatted results in a shorter string
         *  than required by integer format properties that control the
         *  formatting, the value is padded with a filler character at either
         *  left or right end, depending on the alignment specified by the
         *  integer format specification.
         *
         *  @param width            The new minimum width of the formatted value.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setMinimumWidth(UIntSize width);

        /**
         *  Changes the maximum width of the formatted value.
         *
         *  If the maximum width of the formatted value of the integer 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the integer value being formatted results in a longer string
         *  than required by integer format properties that control the
         *  formatting, the entire formatted string is filled with error
         *  indicator characters.
         *
         *  @param width            The new maximum width of the formatted value.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setMaximumWidth(UIntSize width);

        /**
         *  Changes the filler character of the formatted value.
         *
         *  A filler character is used to pad a formatted value from
         *  either left or right end, depending on the required alignment,
         *  if the formatted string is shorter than the required minimum
         *  length.
         *
         *  @param ch               The new filler character.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setFillerChar(Char ch);

        /**
         *  Changes the error indicator character of the formatted value.
         *
         *  An error indicator character is used to replace a formatted value
         *  when the value is too large to be formatted within the specified
         *  maximum width.
         *
         *  @param ch               The new error indicator character.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setErrorChar(Char ch);

        /**
         *  Changes the alignment of the formatted value.
         *
         *  An alignment is used when the result of formatting is shorter
         *  than the required minimum width, to decide whether the formatted
         *  value shall be padded at the left or right end to reach the
         *  required minimum width.
         *
         *  @param alignment        The new alignment.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setAlignment(Alignment alignment);

        /**
         *  Changes the letter case of the formatted value.
         *
         *  A letter case is used when the radix is greater than 10 to
         *  determine whether lower- upper- or title-case letters are
         *  used to represent digits in range 10..35.
         *
         *  @param letterCase       The new letter case.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setLetterCase(LetterCase letterCase);

        /**
         *  Enables or disables the use of prefixes in the formatted value.
         *
         *  When prefixes are enabled:
         *
         *  -   All values formatted with radix 8 are prefixed with '0'.
         *  -   All values formatted with radix 16 are prefixed with '0x' or
         *      '0X', depending on the letter case in effect.
         *
         *  @param usePrefix        Enables (true) or disables (false) the use
         *                          of prefixes in the formatted value.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setUsePrefix(bool usePrefix);

        /**
         *  Enables or disables the use of separators in the formatted value.
         *
         *  When separators are enabled, each N digits of the formatted value
         *  (typically N=3, but in general case the value is locale-specific)
         *  are followed by a locale-specific separator character.
         *
         *  @param useSeparator     Enables (true) or disables (false) the use
         *                          of separators in the formatted value.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setUseSeparator(bool useSeparator);

        /**
         *  Enables or disables the use of mandatory sign in the formatted value.
         *
         *  When mandatory sign is used, positive formatted values are preceded
         *  with a '+' sign and negative formatted values are preceded with a '-'
         *  sign. Otherwise, no sign is emitted for positive values.
         *
         *  @param mandatorySign    Enables (true) or disables (false) the use
         *                          of mandatory sign in the formatted value.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerFormatProperties &   setMandatorySign(bool mandatorySign);

        /**
         *  Returns the radix of this integer format properties.
         *
         *  The returned value is always in range 2..36.
         *
         *  @return                 The radix of this integer format properties.
         */
        UIntSize                    getRadix() const { return _radix; }

        /**
         *  Returns the minimum number of significant digits of the integer 
         *  format specification.
         *
         *  If the minimum number of significant digits of the integer
         *  format specification is not specified, the return value is 0.
         *
         *  When the integer value being formatted has fewer significant digits
         *  than required by integer format properties that control the
         *  formatting, the value is padded with leading zeroes until the
         *  required minimum is met.
         *
         *  @return                 The minimum number of significant digits.
         *  @since 1.0
         */
        UIntSize                    getMinimumDigits() const { return _minimumDigits; }

        /**
         *  Returns the maximum number of significant digits of the integer 
         *  format specification.
         *
         *  If the maximum number of significant digits of the integer 
         *  format specification is not specified, the return value is 0.
         *
         *  When the integer value being formatted has more significant digits
         *  than required by integer format properties that control the
         *  formatting, all value digits are replaced with an error
         *  indicator characters.
         *
         *  @return                 The maximum number of significant digits.
         *  @since 1.0
         */
        UIntSize                    getMaximumDigits() const { return _maximumDigits; }

        /**
         *  Returns the minimum width of the formatted value.
         *
         *  If the minimum width of the formatted value of the integer 
         *  format specification is not specified, the return value is 0.
         *
         *  When the integer value being formatted results in a shorter string
         *  than required by integer format properties that control the
         *  formatting, the value is padded with a filler character at either
         *  left or right end, depending on the alignment specified by the
         *  integer format specification.
         *
         *  @return                 The minimum width of the formatted value.
         *  @since 1.0
         */
        UIntSize                    getMinimumWidth() const { return _minimumWidth; }

        /**
         *  Returns the maximum width of the formatted value.
         *
         *  If the maximum width of the formatted value of the integer 
         *  format specification is not specified, the return value is 0.
         *
         *  When the integer value being formatted results in a longer string
         *  than required by integer format properties that control the
         *  formatting, the entire formatted string is filled with error
         *  indicator characters.
         *
         *  @return                 The maximum width of the formatted value.
         *  @since 1.0
         */
        UIntSize                    getMaximumWidth() const { return _maximumWidth; }

        /**
         *  Returns the filler character of the formatted value.
         *
         *  A filler character is used to pad a formatted value from
         *  either left or right end, depending on the required alignment,
         *  if the formatted string is shorter than the required minimum
         *  length.
         *
         *  @return                 The filler character.
         *  @since 1.0
         */
        Char                        getFillerChar() const { return _fillerChar; }

        /**
         *  Returns the error indicator character of the formatted value.
         *
         *  An error indicator character is used to replace a formatted value
         *  when the value is too large to be formatted within the specified
         *  maximum width.
         *
         *  @return                 The error indicator character.
         *  @since 1.0
         */
        Char                        getErrorChar() const { return _errorChar; }

        /**
         *  Returns the alignment of the formatted value.
         *
         *  An alignment is used when the result of formatting is shorter
         *  than the required minimum width, to decide whether the formatted
         *  value shall be padded at the left or right end to reach the
         *  required minimum width.
         *
         *  @return                 The alignment of the formatted value.
         *  @since 1.0
         */
        Alignment                   getAlignment() const { return _alignment; }

        /**
         *  Returns the letter case of the formatted value.
         *
         *  A letter case is used when the radix is greater than 10 to
         *  determine whether lower- upper- or title-case letters are
         *  used to represent digits in range 10..35.
         *
         *  @return                 The letter case of the formatted value.
         *  @since 1.0
         */
        LetterCase                  getLetterCase() const { return _letterCase; }

        /**
         *  Returns the prefix usage mode for the formatted values.
         *
         *  When prefixes are enabled:
         *
         *  -   All values formatted with radix 8 are prefixed with '0'.
         *  -   All values formatted with radix 16 are prefixed with '0x' or
         *      '0X', depending on the letter case in effect.
         *
         *  @return                 True if prefixes are enabled; false otherwise.
         *  @since 1.0
         */
        bool                        getUsePrefix() const { return _usePrefix; }

        /**
         *  Returns the separator usage mode for the formatted value.
         *
         *  When separators are enabled, each N digits of the formatted value
         *  (typically N=3, but in general case the value is locale-specific)
         *  are followed by a locale-specific separator character.
         *
         *  @return                 True if separators are enabled; false otherwise.
         *  @since 1.0
         */
        bool                        getUseSeparator() const { return _useSeparator; }

        /**
         *  Returns the mandatory sign mode for the formatted value.
         *
         *  When mandatory sign is used, positive formatted values are preceded
         *  with a '+' sign and negative formatted values are preceded with a '-'
         *  sign. Otherwise, no sign is emitted for positive values.
         *
         *  @return                 True if the '+' sign is mandatory for positive
         *                          values; false if the '+' sign is omitted for
         *                          positive values. The '-' sign is always
         *                          mandatory for negative values.
         *  @since 1.0
         */
        bool                        getMandatorySign() const { return _mandatorySign; }

        //////////
        //  Implementation
    private:
        UIntSize                    _radix;         //  2..36
        UIntSize                    _minimumDigits; //  0 == unspecified
        UIntSize                    _maximumDigits; //  0 == unspecified
        UIntSize                    _minimumWidth;  //  0 == unspecified
        UIntSize                    _maximumWidth;  //  0 == unspecified
        Char                        _fillerChar;
        Char                        _errorChar;
        Alignment                   _alignment;
        LetterCase                  _letterCase;
        bool                        _usePrefix;
        bool                        _useSeparator;
        bool                        _mandatorySign;

        //  Helpers
        void                        _setStandardProperties();
        void                        _parseCrtFormatString(const String & crtFormatString);
    };

    /**
     *  @class RealFormatProperties CARL.hpp
     *  Format properties governing formatiting of real values.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC RealFormatProperties
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  Represents various alignments of formatted real value.
         *
         *  @since 1.0
         */
        enum Alignment
        {
            /**
             *  Align the formatted real value to the left.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Alignment_Left      = 1,

            /**
             *  Align the formatted real value to the right.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Alignment_Right     = 2
        };

        /**
         *  Represents letter cases to use in a formatted real value.
         *
         *  @since 1.0
         */
        enum LetterCase
        {
            /**
             *  All letters in a formatted real value shall be lowercase.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            LetterCase_Lower    = 1,

            /**
             *  All letters in a formatted real value shall be uppercase.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            LetterCase_Upper    = 2,

            /**
             *  All letters in a formatted real value shall be titlecase.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            LetterCase_Title    = 3
        };

        /**
         *  Represents the representation of a formatted real value.
         *
         *  @since 1.0
         */
        enum Format
        {
            /**
             *  Represent formatted real value as a fixed-point number.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Format_FixedPoint       = 1,

            /**
             *  Represent formatted real value as a floating-point number.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Format_FloatingPoint    = 2,

            /**
             *  Automaticall choose between fixed-point and floating-point
             *  representation of the formatted real.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Format_AutoSelect       = 3
        };

        /**
         *  The default constructor.
         *
         *  Iniitializes a newly created real format properties to default
         *  values:
         *
         *  -   No minimum or maximum limits on the formatted value length
         *      or a number of fratcional or exponent digits to format.
         *  -   Use space as a filler.
         *  -   Use asterisk (*) as an error indicator.
         *  -   Formatted value is aligned to the right.
         *  -   Use lowercase letters if letters are necessary.
         *  -   Do not use prefix or thousand separator.
         *  -   Omit the '+' sign for positive values.
         *  -   Automatically choose between fixed-point or floating-point representation.
         *
         *  @since 1.0
         */
        RealFormatProperties();

        /**
         *  The class constructor.
         *
         *  Initializes a newly created real format properties instance to
         *  specify the same formatting rules as would be effected by the
         *  specified string in CRT printf() family of functions.
         *
         *  @param pCrtFormatString A pointer to nul-terminated ASCII string
         *                          representing the CRT printf()-style
         *                          real format specification.
         *  @since 1.0
         */
        explicit RealFormatProperties(const char * pCrtFormatString);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created real format properties instance to
         *  specify the same formatting rules as would be effected by the
         *  specified string in CRT printf() family of functions.
         *
         *  @param pCrtFormatString A pointer to nul-terminated ISO-10646 string
         *                          representing the CRT printf()-style
         *                          real format specification.
         *  @since 1.0
         */
        explicit RealFormatProperties(const Char * pCrtFormatString);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created real format properties instance to
         *  specify the same formatting rules as would be effected by the
         *  specified string in CRT printf() family of functions.
         *
         *  @param crtFormatString  A string representing the CRT printf()-style
         *                          real format specification.
         *  @since 1.0
         */
        explicit RealFormatProperties(const String & crtFormatString);

        /**
         *  The copy constructor.
         *
         *  The newly created real format properties instance is initialized to 
         *  be an exact copy of an existing real format properties.
         *
         *  @param src              The real format properties name to copy 
         *                          properties from.
         *  @since 1.0
         */
        RealFormatProperties(const RealFormatProperties & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~RealFormatProperties() {}

        /**
         *  The assigment operator.
         *
         *  Makes an real format properties an exact copy of an existing 
         *  real format properties.
         *
         *  @param src              The real format properties to copy properties from.
         *  @return                 A reference to an assigned-to object.
         *  @since 1.0
         */
        RealFormatProperties &      operator = (const RealFormatProperties & src);

        //////////
        //  Operations
    public:
        /**
         *  Changes the radix of the real format specification.
         *
         *  The new radix must be in range 2..36.
         *
         *  @param radix            The new radix.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setRadix(UIntSize radix);

        /**
         *  Changes the minimum number of significant digits of the real
         *  format specification.
         *
         *  If the minimum number of significant digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has fewer significant digits
         *  than required by real format properties that control the
         *  formatting, the value is padded with leading zeroes until the
         *  required minimum is met.
         *
         *  @param digits           The new minimum number of significant digits.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setMinimumDigits(UIntSize digits);

        /**
         *  Changes the maximum number of significant digits of the real 
         *  format specification.
         *
         *  If the maximum number of significant digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has more significant digits
         *  than required by real format properties that control the
         *  formatting, all value digits are replaced with an error
         *  indicator characters.
         *
         *  @param digits           The new maximum number of significant digits.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setMaximumDigits(UIntSize digits);

        /**
         *  Changes the minimum number of fractional digits of the real
         *  format specification.
         *
         *  If the minimum number of fractional digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has fewer fractional digits
         *  than required by real format properties that control the
         *  formatting, the value is padded with tailing zeroes until the
         *  required minimum is met.
         *
         *  @param digits           The new minimum number of fractional digits.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setMinimumFractionalDigits(UIntSize digits);

        /**
         *  Changes the maximum number of fractional digits of the real 
         *  format specification.
         *
         *  If the maximum number of fractional digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has more fractional digits
         *  than required by real format properties that control the
         *  formatting, all value digits are replaced with an error
         *  indicator characters.
         *
         *  @param digits           The new maximum number of fractional digits.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setMaximumFractionalDigits(UIntSize digits);

        /**
         *  Changes the minimum number of exponent digits of the real
         *  format specification.
         *
         *  If the minimum number of exponent digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has fewer exponent digits
         *  than required by real format properties that control the
         *  formatting, the exponent is padded with leading zeroes until the
         *  required minimum is met.
         *
         *  This setting has no effect if the real value is formatted
         *  using a fixed-point representation.
         *
         *  @param digits           The new minimum number of exponent digits.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setMinimumExponentDigits(UIntSize digits);

        /**
         *  Changes the maximum number of exponent digits of the real 
         *  format specification.
         *
         *  If the maximum number of exponent digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has more exponent digits
         *  than required by real format properties that control the
         *  formatting, all value digits are replaced with an error
         *  indicator characters.
         *
         *  This setting has no effect if the real value is formatted
         *  using a fixed-point representation.
         *
         *  @param digits           The new maximum number of exponent digits.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setMaximumExponentDigits(UIntSize digits);

        /**
         *  Changes the minimum width of the formatted value.
         *
         *  If the minimum width of the formatted value of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted results in a shorter string
         *  than required by real format properties that control the
         *  formatting, the value is padded with a filler character at either
         *  left or right end, depending on the alignment specified by the
         *  real format specification.
         *
         *  @param width            The new minimum width of the formatted value.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setMinimumWidth(UIntSize width);

        /**
         *  Changes the maximum width of the formatted value.
         *
         *  If the maximum width of the formatted value of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted results in a longer string
         *  than required by real format properties that control the
         *  formatting, the entire formatted string is filled with error
         *  indicator characters.
         *
         *  @param width            The new maximum width of the formatted value.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setMaximumWidth(UIntSize width);

        /**
         *  Changes the filler character of the formatted value.
         *
         *  A filler character is used to pad a formatted value from
         *  either left or right end, depending on the required alignment,
         *  if the formatted string is shorter than the required minimum
         *  length.
         *
         *  @param ch               The new filler character.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setFillerChar(Char ch);

        /**
         *  Changes the error indicator character of the formatted value.
         *
         *  An error indicator character is used to replace a formatted value
         *  when the value is too large to be formatted within the specified
         *  maximum width.
         *
         *  @param ch               The new error indicator character.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setErrorChar(Char ch);

        /**
         *  Changes the alignment of the formatted value.
         *
         *  An alignment is used when the result of formatting is shorter
         *  than the required minimum width, to decide whether the formatted
         *  value shall be padded at the left or right end to reach the
         *  required minimum width.
         *
         *  @param alignment        The new alignment.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setAlignment(Alignment alignment);

        /**
         *  Changes the letter case of the formatted value.
         *
         *  A letter case is used for an exponent mark 'E' when a real
         *  value is formatted to a floating-point representation.
         *
         *  @param letterCase       The new letter case.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setLetterCase(LetterCase letterCase);

        /**
         *  Enables or disables the use of separators in the formatted value.
         *
         *  When separators are enabled, each N digits of the formatted value
         *  (typically N=3, but in general case the value is locale-specific)
         *  are followed by a locale-specific separator character.
         *
         *  @param useSeparator     Enables (true) or disables (false) the use
         *                          of separators in the formatted value.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setUseSeparator(bool useSeparator);

        /**
         *  Enables or disables the use of mandatory sign in the formatted value.
         *
         *  When mandatory sign is used, positive formatted values are preceded
         *  with a '+' sign and negative formatted values are preceded with a '-'
         *  sign. Otherwise, no sign is emitted for positive values.
         *
         *  When a real value is formatted to a floating-point representation,
         *  this setting affects both mantissa and exponent of the formatted value.
         *
         *  @param mandatorySign    Enables (true) or disables (false) the use
         *                          of mandatory sign in the formatted value.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setMandatorySign(bool mandatorySign);

        /**
         *  Sets a preferred representation for formatted real values.
         *
         *  When a fixed-point representation is chosen, the formatted real
         *  value includes a decimal point but no exponent. As many significant
         *  digits are formatted as necessary to ensure that the decimal
         *  point appears within the formatted value.
         *
         *  When a floating-point representation is chosen, the formatted
         *  real value is presented in an exponential notation, with
         *  mantissa normalized so that there is exactly one significant
         *  digit before the decimal point.
         *
         *  When instructed to automatically select between fixed-point
         *  and floating-point representations, the one is always selected
         *  that results in a shorted formatted string.
         *
         *  @param format           The preferred representation for formatted 
         *                          values.
         *  @return                 A reference to the RealFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealFormatProperties &      setFormat(Format format);

        /**
         *  Returns the radix of this real format properties.
         *
         *  The returned value is always in range 2..36.
         *
         *  @return                 The radix of this real format properties.
         */
        UIntSize                    getRadix() const { return _radix; }

        /**
         *  Returns the minimum number of significant digits of the real
         *  format specification.
         *
         *  If the minimum number of significant digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has fewer significant digits
         *  than required by real format properties that control the
         *  formatting, the value is padded with leading zeroes until the
         *  required minimum is met.
         *
         *  @return                 The minimum number of significant digits.
         *  @since 1.0
         */
        UIntSize                    getMinimumDigits() const { return _minimumDigits; }

        /**
         *  Returns the maximum number of significant digits of the real 
         *  format specification.
         *
         *  If the maximum number of significant digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has more significant digits
         *  than required by real format properties that control the
         *  formatting, all value digits are replaced with an error
         *  indicator characters.
         *
         *  @return                 The maximum number of significant digits.
         *  @since 1.0
         */
        UIntSize                    getMaximumDigits() const { return _maximumDigits; }

        /**
         *  Returns the minimum number of fractional digits of the real
         *  format specification.
         *
         *  If the minimum number of fractional digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has fewer fractional digits
         *  than required by real format properties that control the
         *  formatting, the value is padded with tailing zeroes until the
         *  required minimum is met.
         *
         *  @return                 The minimum number of fractional digits.
         *  @since 1.0
         */
        UIntSize                    getMinimumFractionalDigits() const { return _minimumFractionalDigits; }

        /**
         *  Returns the maximum number of fractional digits of the real 
         *  format specification.
         *
         *  If the maximum number of fractional digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has more fractional digits
         *  than required by real format properties that control the
         *  formatting, all value digits are replaced with an error
         *  indicator characters.
         *
         *  @return                 The maximum number of fractional digits.
         *  @since 1.0
         */
        UIntSize                    getMaximumFractionalDigits() const { return _maximumFractionalDigits; }

        /**
         *  Returns the minimum number of exponent digits of the real
         *  format specification.
         *
         *  If the minimum number of exponent digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has fewer exponent digits
         *  than required by real format properties that control the
         *  formatting, the exponent is padded with leading zeroes until the
         *  required minimum is met.
         *
         *  This setting has no effect if the real value is formatted
         *  using a fixed-point representation.
         *
         *  @return                 The minimum number of exponent digits.
         *  @since 1.0
         */
        UIntSize                    getMinimumExponentDigits() const { return _minimumExponentDigits; }

        /**
         *  Returns the maximum number of exponent digits of the real 
         *  format specification.
         *
         *  If the maximum number of exponent digits of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted has more exponent digits
         *  than required by real format properties that control the
         *  formatting, all value digits are replaced with an error
         *  indicator characters.
         *
         *  This setting has no effect if the real value is formatted
         *  using a fixed-point representation.
         *
         *  @return                 The maximum number of exponent digits.
         *  @since 1.0
         */
        UIntSize                    getMaximumExponentDigits() const { return _maximumExponentDigits; }

        /**
         *  Returns the minimum width of the formatted value.
         *
         *  If the minimum width of the formatted value of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted results in a shorter string
         *  than required by real format properties that control the
         *  formatting, the value is padded with a filler character at either
         *  left or right end, depending on the alignment specified by the
         *  real format specification.
         *
         *  @return                 The minimum width of the formatted value.
         *  @since 1.0
         */
        UIntSize                    getMinimumWidth() const { return _minimumWidth;}

        /**
         *  Returns the maximum width of the formatted value.
         *
         *  If the maximum width of the formatted value of the real 
         *  format specification is set to 0, the limit is removed.
         *
         *  If the real value being formatted results in a longer string
         *  than required by real format properties that control the
         *  formatting, the entire formatted string is filled with error
         *  indicator characters.
         *
         *  @return                 The maximum width of the formatted value.
         *  @since 1.0
         */
        UIntSize                    getMaximumWidth() const { return _maximumWidth; }

        /**
         *  Returns the filler character of the formatted value.
         *
         *  A filler character is used to pad a formatted value from
         *  either left or right end, depending on the required alignment,
         *  if the formatted string is shorter than the required minimum
         *  length.
         *
         *  @return                 The filler character.
         *  @since 1.0
         */
        Char                        getFillerChar() const { return _fillerChar; }

        /**
         *  Retruns the error indicator character of the formatted value.
         *
         *  An error indicator character is used to replace a formatted value
         *  when the value is too large to be formatted within the specified
         *  maximum width.
         *
         *  @return                 The error indicator character.
         *  @since 1.0
         */
        Char                        getErrorChar() const { return _errorChar;}

        /**
         *  Returns the alignment of the formatted value.
         *
         *  An alignment is used when the result of formatting is shorter
         *  than the required minimum width, to decide whether the formatted
         *  value shall be padded at the left or right end to reach the
         *  required minimum width.
         *
         *  @return                 The alignment of the formatted value.
         *  @since 1.0
         */
        Alignment                   getAlignment() const { return _alignment; }

        /**
         *  Returns the letter case of the formatted value.
         *
         *  A letter case is used for an exponent mark 'E' when a real
         *  value is formatted to a floating-point representation.
         *
         *  @return                 The letter case of the formatted value.
         *  @since 1.0
         */
        LetterCase                  getLetterCase() const { return _letterCase; }

        /**
         *  Returns the separator usage mode for the formatted value.
         *
         *  When separators are enabled, each N digits of the formatted value
         *  (typically N=3, but in general case the value is locale-specific)
         *  are followed by a locale-specific separator character.
         *
         *  @return                 True if separators are enabled; false otherwise.
         *  @since 1.0
         */
        bool                        getUseSeparator() const { return _useSeparator; }

        /**
         *  Returns the mandatory sign mode for the formatted value.
         *
         *  When mandatory sign is used, positive formatted values are preceded
         *  with a '+' sign and negative formatted values are preceded with a '-'
         *  sign. Otherwise, no sign is emitted for positive values.
         *
         *  When a real value is formatted to a floating-point representation,
         *  this setting affects both mantissa and exponent of the formatted value.
         *
         *  @return                 True if the '+' sign is mandatory for positive
         *                          values; false if the '+' sign is omitted for
         *                          positive values. The '-' sign is always
         *                          mandatory for negative values.
         *  @since 1.0
         */
        bool                        getMandatorySign() const { return _mandatorySign; }

        /**
         *  Returns a preferred representation for formatted real values.
         *
         *  When a fixed-point representation is chosen, the formatted real
         *  value includes a decimal point but no exponent. As many significant
         *  digits are formatted as necessary to ensure that the decimal
         *  point appears within the formatted value.
         *
         *  When a floating-point representation is chosen, the formatted
         *  real value is presented in an exponential notation, with
         *  mantissa normalized so that there is exactly one significant
         *  digit before the decimal point.
         *
         *  When instructed to automatically select between fixed-point
         *  and floating-point representations, the one is always selected
         *  that results in a shorted formatted string.
         *
         *  @return                 The preferred representation for formatted 
         *                          values.
         *  @since 1.0
         */
        Format                      getFormat() const { return _format; }

        //////////
        //  Implementation
    private:
        UIntSize                    _radix;
        UIntSize                    _minimumDigits; //  0 == unspecified
        UIntSize                    _maximumDigits; //  0 == unspecified
        UIntSize                    _minimumFractionalDigits; //  0 == unspecified
        UIntSize                    _maximumFractionalDigits; //  0 == unspecified
        UIntSize                    _minimumExponentDigits; //  0 == unspecified
        UIntSize                    _maximumExponentDigits; //  0 == unspecified
        UIntSize                    _minimumWidth;  //  0 == unspecified
        UIntSize                    _maximumWidth;  //  0 == unspecified
        Char                        _fillerChar;
        Char                        _errorChar;
        Alignment                   _alignment;
        LetterCase                  _letterCase;
        bool                        _useSeparator;
        bool                        _mandatorySign;
        Format                      _format;

        //  Helpers
        void                        _setStandardProperties();
        void                        _parseCrtFormatString(const String & crtFormatString);
    };

    /**
     *  @class NumberFormat CARL.hpp
     *  A generic formattor for numeric values.
     *
     *  Formatting numeric values (i.e. converting them to string representation)
     *  is performed by locale-specific instances of concrete NumberFormat-derived
     *  classes. A NumberFormat subclass for a specific locale "knows" about rules
     *  for representing numbers that are specific to the appropriate locale. The
     *  representation of the formatted value can be varied to a certain degree
     *  by specifying custom "format properties" when formatting a value.
     *
     *  When obtaining a number formattor for a specified locale, it may well
     *  be the case that there is no number formattor registered for that locale.
     *  In this case the immediate parent locale is tried next, and the parent of 
     *  that if necessary, until either an invariant locale is reached or a number
     *  formattor is found. In the former case, a default number formattor is
     *  used. This behaviour allows specifying locale-generic number formattors
     *  (e.g. if an "German neutral" number formattor is specified, then it will
     *  be used for both "German/Germany" and "German/Switzerland" locales provided
     *  no separate number formattors are registered for either of those).
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC NumberFormat
    {
        friend class _impl::_NumberFormatManager;
        //////////
        //  Can't assign or copy construct
    private:
        NumberFormat(const NumberFormat &);
        void                        operator = (const NumberFormat &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        NumberFormat() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~NumberFormat() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified number format with the CARL framework.
         *
         *  A registration of a number format informs the CARL framework that
         *  a specific number format exists. From this point on, the application
         *  can use the newly registers number format in exactly the same manner
         *  as a predefined number format. Once registered, a number format
         *  cannot be unregistered.
         *
         *  A number format is always registered for a specific locale. Any attempt
         *  to get a number format instance for that locale will always yield the
         *  number format registered for that locale. It is possible to register
         *  the same number format for more than one locale.
         *
         *  @param forLocale        The locale for which the newly registered
         *                          number format shall be used.
         *  @param pNumberFormat    The number format to register.
         *  @since 1.0
         */
        static void                 registerInstance(const Locale & forLocale, NumberFormat * pNumberFormat);

        /**
         *  Returns a number formattor for the current default locale.
         *
         *  @return                 A number formattor for the current default locale.
         *  @since 1.0
         */
        static NumberFormat *       getInstance();

        /**
         *  Returns a number formattor for the specified locale.
         *
         *  @param forLocale        The locale for which a number formattor is sought.
         *  @return                 A number formattor for the specified locale.
         *  @since 1.0
         */
        static NumberFormat *       getInstance(const Locale & forLocale);

        //  Default formatting
        /**
         *  Formats a value using default integer format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(Int8 value) const { return format(value, IntegerFormatProperties()); }

        /**
         *  Formats a value using default integer format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(Int16 value) const { return format(value, IntegerFormatProperties()); }

        /**
         *  Formats a value using default integer format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(Int32 value) const { return format(value, IntegerFormatProperties()); }

        /**
         *  Formats a value using default integer format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(Int64 value) const { return format(value, IntegerFormatProperties()); }

        /**
         *  Formats a value using default integer format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(UInt8 value) const { return format(value, IntegerFormatProperties()); }

        /**
         *  Formats a value using default integer format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(UInt16 value) const { return format(value, IntegerFormatProperties()); }

        /**
         *  Formats a value using default integer format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(UInt32 value) const { return format(value, IntegerFormatProperties()); }

        /**
         *  Formats a value using default integer format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(UInt64 value) const { return format(value, IntegerFormatProperties()); }

        /**
         *  Formats a value using default integer format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(const BinaryInteger & value) const { return format(value, IntegerFormatProperties()); }

        /**
         *  Formats a value using default integer format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(const DecimalInteger & value) const { return format(value, IntegerFormatProperties()); }

        /**
         *  Formats a value using default real format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(Real32 value) const { return format(value, RealFormatProperties()); }

        /**
         *  Formats a value using default real format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(Real64 value) const { return format(value, RealFormatProperties()); }

        /**
         *  Formats a value using default real format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(const BinaryFixed & value) const { return format(value, RealFormatProperties()); }

        /**
         *  Formats a value using default real format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(const BinaryFloat & value) const { return format(value, RealFormatProperties()); }

        /**
         *  Formats a value using default real format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(const DecimalFixed & value) const { return format(value, RealFormatProperties()); }

        /**
         *  Formats a value using default real format properties.
         *
         *  @param value            A value to format.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        String                      format(const DecimalFloat & value) const { return format(value, RealFormatProperties()); }

        //  Custom formatting
        /**
         *  Formats a value using specified integer format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(Int8 value, const IntegerFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified integer format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(Int16 value, const IntegerFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified integer format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(Int32 value, const IntegerFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified integer format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(Int64 value, const IntegerFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified integer format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(UInt8 value, const IntegerFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified integer format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(UInt16 value, const IntegerFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified integer format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(UInt32 value, const IntegerFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified integer format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(UInt64 value, const IntegerFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified integer format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const BinaryInteger & value, const IntegerFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified integer format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const DecimalInteger & value, const IntegerFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified real format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(Real32 value, const RealFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified real format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(Real64 value, const RealFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified real format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const BinaryFixed & value, const RealFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified real format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const BinaryFloat & value, const RealFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified real format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const DecimalFixed & value, const RealFormatProperties & formatProperties) const = 0;

        /**
         *  Formats a value using specified real format properties.
         *
         *  @param value            A value to format.
         *  @param formatProperties The custom format properties governing the formatting.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const DecimalFloat & value, const RealFormatProperties & formatProperties) const = 0;

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };

    /**
     *  @class DateTimeFormat CARL.hpp
     *  A generic formattor for date+time values.
     *
     *  Formatting date+time values (i.e. converting them to string representation)
     *  is performed by locale-specific instances of concrete DateTimeFormat-derived
     *  classes. A DateTimeFormat subclass for a specific locale "knows" about rules
     *  for representing dates+times that are specific to the appropriate locale. The
     *  representation of the formatted value can be valied to a certain degree
     *  by specifying custom "format properties" when formatting a value.
     *
     *  When obtaining a date+time formattor for a specified locale, it may well
     *  be the case that there is no date+time formattor registered for that locale.
     *  In this case the immediate parent locale is tried next, and the parent of 
     *  that if necessary, until either an invariant locale is reached or a date+time
     *  formattor is found. In the former case, a default date+time formattor is
     *  used. This behaviour allows specifying locale-generic date+time formattors
     *  (e.g. if an "German neutral" date+time formattor is specified, then it will
     *  be used for both "German/Germany" and "German/Switzerland" locales provided
     *  no separate date+time formattors are registered for either of those).
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DateTimeFormat
    {
        friend class _impl::_DateTimeFormatManager;
        //////////
        //  Can't assign or copy construct
    private:
        DateTimeFormat(const DateTimeFormat &);
        void                        operator = (const DateTimeFormat &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        DateTimeFormat() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~DateTimeFormat() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified date+time format with the CARL framework.
         *
         *  A registration of a date+time format informs the CARL framework that
         *  a specific date+time format exists. From this point on, the application
         *  can use the newly registers date+time format in exactly the same manner
         *  as a predefined date+time format. Once registered, a date+time format
         *  cannot be unregistered.
         *
         *  A date+time format is always registered for a specific locale. Any attempt
         *  to get a date+time format instance for that locale will always yield the
         *  date+time format registered for that locale. It is possible to register
         *  the same date+time format for more than one locale.
         *
         *  @param forLocale        The locale for which the newly registered
         *                          date+time format shall be used.
         *  @param pDateTimeFormat  The date+time format to register.
         *  @since 1.0
         */
        static void                 registerInstance(const Locale & forLocale, DateTimeFormat * pDateTimeFormat);

        /**
         *  Returns a date+time formattor for the current default locale.
         *
         *  @return                 A date+time formattor for the current default locale.
         *  @since 1.0
         */
        static DateTimeFormat *     getInstance();

        /**
         *  Returns a date+time formattor for the specified locale.
         *
         *  @param forLocale        The locale for which a date+time formattor is sought.
         *  @return                 A date+time formattor for the specified locale.
         *  @since 1.0
         */
        static DateTimeFormat *     getInstance(const Locale & forLocale);

        //  Formatting
        /**
         *  Represents various date+time representation styles.
         *
         *  @since 1.0
         */
        enum Style
        {
            /**
             *  The shortest possible representation for formatted date+time.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Short             = 1,

            /**
             *  The somewhat longer representation for formatted date+time.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Medium            = 2,

            /**
             *  The longer yet representation for formatted date+time.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Long              = 3,

            /**
             *  The longest possible representation for formatted date+time.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Full              = 4
        };

        /**
         *  Represents a precision with which the time portion of date+time
         *  is represented.
         *
         *  @since 1.0
         */
        enum Precision
        {
            /**
             *  The time is represented with a minute precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Minute        = 1,

            /**
             *  The time is represented with a second precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Second        = 2,

            /**
             *  The time is represented with a millisecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Millisecond   = 3,

            /**
             *  The time is represented with a microsecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Microsecond   = 4,

            /**
             *  The time is represented with a nanosecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Nanosecond    = 5
        };

        /**
         *  Formats the specified date+time using the specified style and precision.
         *
         *  The exact representation of the formatted value depends on the requested
         *  style and precision, but is always correct for the locale for which the
         *  formattor has been requested.
         *
         *  @param dateTime         The value to format.
         *  @param style            The style of the date+time formatted value.
         *  @param precision        The time precision of the date+time formatted value.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const DateTime & dateTime, Style style, Precision precision) const = 0;

        /**
         *  Formats the specified date+time using the custom format specification.
         *
         *  The custom format specification is a string that defines how the result
         *  of formatting will look. This string is converted to a formatted value by
         *  replacing predefined placeholders defined within by various portions of
         *  the formatted date+time value.
         *
         *  The following placeholders can be used:
         *
         *  -   {year:full} - A full 4-digit year.
         *  -   {year:short} - The last 2 digits of the year.
         *  -   {month:full} - A 2-digit month number (01..12).
         *  -   {month:short} - A 1- or 2-digit month number (1..12).
         *  -   {month_name:full} - A full month name.
         *  -   {month_name:short} - An abbreviated month name.
         *  -   {day:full} - A 2-digit day number (01..31).
         *  -   {day:short} - A 1- or 2-digit day number (1..31).
         *  -   {day_of_week_name:full} - A full day-of-week name.
         *  -   {day_of_week_name:short} - An abbreviated day-of-week name.
         *  -   {hour12:full} - A 2-digit hour on a 12-hour scale (01..12).
         *  -   {hour12:short} - A 1- or 2-digit hour on a 12-hour scale (1..12).
         *  -   {hour24:full} - A 2-digit hour on a 24-hour scale (00..23).
         *  -   {hour24:short} - A 1- or 2-digit hour on a 24-hour scale (0..23).
         *  -   {minute:full} - A 2-digit minute (00..59).
         *  -   {minute:short} - A 1- or 2-digit minute (0..59).
         *  -   {second:full} - A 2-digit second (00..59).
         *  -   {second:short} - A 1- or 2-digit second (0..59).
         *  -   {millisecond:full} - A 3-digit millisecond (000..999).
         *  -   {millisecond:short} - A 1- to 3-digit millisecond (0..999).
         *  -   {microsecond:full} - A 6-digit microsecond (000000..999999).
         *  -   {microsecond:short} - A 1- to 6-digit microsecond (0..999999).
         *  -   {nanosecond:full} - A 9-digit vsecond (000000000..999999999).
         *  -   {nanosecond:short} - A 1- to 9-digit nanosecond (0..999999999).
         *  -   {ampm} - An AM/PM indicator.
         *
         *  Any character in a format specification string prefixed by a backslash '\' 
         *  is interpreted literally and not as a part of a placeholder. Specifically, 
         *  if a formatted value must contain a baskclash, the backslash must be written 
         *  twice in a format specification string.
         *
         *  @param dateTime         The value to format.
         *  @param format           The custom format specification.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const DateTime & dateTime, const String & format) const = 0;

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };

    /**
     *  @class DateFormat CARL.hpp
     *  A generic formattor for date values.
     *
     *  Formatting date values (i.e. converting them to string representation)
     *  is performed by locale-specific instances of concrete DateFormat-derived
     *  classes. A DateFormat subclass for a specific locale "knows" about rules
     *  for representing datess that are specific to the appropriate locale. The
     *  representation of the formatted value can be valied to a certain degree
     *  by specifying custom "format properties" when formatting a value.
     *
     *  When obtaining a date formattor for a specified locale, it may well
     *  be the case that there is no date formattor registered for that locale.
     *  In this case the immediate parent locale is tried next, and the parent of 
     *  that if necessary, until either an invariant locale is reached or a date
     *  formattor is found. In the former case, a default date formattor is
     *  used. This behaviour allows specifying locale-generic date formattors
     *  (e.g. if an "German neutral" date formattor is specified, then it will
     *  be used for both "German/Germany" and "German/Switzerland" locales provided
     *  no separate date formattors are registered for either of those).
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DateFormat
    {
        friend class _impl::_DateTimeFormatManager;
        //////////
        //  Can't assign or copy construct
    private:
        DateFormat(const DateFormat &);
        void                        operator = (const DateFormat &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        DateFormat() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~DateFormat() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified date format with the CARL framework.
         *
         *  A registration of a date format informs the CARL framework that
         *  a specific date format exists. From this point on, the application
         *  can use the newly registers date format in exactly the same manner
         *  as a predefined date format. Once registered, a date format
         *  cannot be unregistered.
         *
         *  A date format is always registered for a specific locale. Any attempt
         *  to get a date format instance for that locale will always yield the
         *  date format registered for that locale. It is possible to register
         *  the same date format for more than one locale.
         *
         *  @param forLocale        The locale for which the newly registered
         *                          date format shall be used.
         *  @param pDateFormat      The date format to register.
         *  @since 1.0
         */
        static void                 registerInstance(const Locale & forLocale, DateFormat * pDateFormat);

        /**
         *  Returns a date formattor for the current default locale.
         *
         *  @return                 A date formattor for the current default locale.
         *  @since 1.0
         */
        static DateFormat *         getInstance();

        /**
         *  Returns a date formattor for the specified locale.
         *
         *  @param forLocale        The locale for which a date formattor is sought.
         *  @return                 A date formattor for the specified locale.
         *  @since 1.0
         */
        static DateFormat *         getInstance(const Locale & forLocale);
        
        //  Formatting
        /**
         *  Represents various date representation styles.
         *
         *  @since 1.0
         */
        enum Style
        {
            /**
             *  The shortest possible representation for formatted date.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Short             = 1,

            /**
             *  The somewhat longer representation for formatted date.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Medium            = 2,

            /**
             *  The longer yet representation for formatted date.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Long              = 3,

            /**
             *  The longest possible representation for formatted date+time.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Full              = 4
        };

        /**
         *  Formats the specified date using the specified style and precision.
         *
         *  The exact representation of the formatted value depends on the requested
         *  style and precision, but is always correct for the locale for which the
         *  formattor has been requested.
         *
         *  @param date             The value to format.
         *  @param style            The style of the date formatted value.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const Date & date, Style style) const = 0;

        /**
         *  Formats the specified date using the custom format specification.
         *
         *  The custom format specification is a string that defines how the result
         *  of formatting will look. This string is converted to a formatted value by
         *  replacing predefined placeholders defined within by various portions of
         *  the formatted date value.
         *
         *  The following placeholders can be used:
         *
         *  -   {year:full} - A full 4-digit year.
         *  -   {year:short} - The last 2 digits of the year.
         *  -   {month:full} - A 2-digit month number (01..12).
         *  -   {month:short} - A 1- or 2-digit month number (1..12).
         *  -   {month_name:full} - A full month name.
         *  -   {month_name:short} - An abbreviated month name.
         *  -   {day:full} - A 2-digit day number (01..31).
         *  -   {day:short} - A 1- or 2-digit day number (1..31).
         *  -   {day_of_week_name:full} - A full day-of-week name.
         *  -   {day_of_week_name:short} - An abbreviated day-of-week name.
         *
         *  Any character in a format specification string prefixed by a backslash '\' 
         *  is interpreted literally and not as a part of a placeholder. Specifically, 
         *  if a formatted value must contain a baskclash, the backslash must be written 
         *  twice in a format specification string.
         *
         *  @param date             The value to format.
         *  @param format           The custom format specification.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const Date & date, const String & format) const = 0;

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };

    /**
     *  @class TimeFormat CARL.hpp
     *  A generic formattor for time values.
     *
     *  Formatting time values (i.e. converting them to string representation)
     *  is performed by locale-specific instances of concrete TimeFormat-derived
     *  classes. A TimeFormat subclass for a specific locale "knows" about rules
     *  for representing times that are specific to the appropriate locale. The
     *  representation of the formatted value can be valied to a certain degree
     *  by specifying custom "format properties" when formatting a value.
     *
     *  When obtaining a time formattor for a specified locale, it may well
     *  be the case that there is no time formattor registered for that locale.
     *  In this case the immediate parent locale is tried next, and the parent of 
     *  that if necessary, until either an invariant locale is reached or a time
     *  formattor is found. In the former case, a default time formattor is
     *  used. This behaviour allows specifying locale-generic time formattors
     *  (e.g. if an "German neutral" time formattor is specified, then it will
     *  be used for both "German/Germany" and "German/Switzerland" locales provided
     *  no separate time formattors are registered for either of those).
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC TimeFormat
    {
        friend class _impl::_DateTimeFormatManager;
        //////////
        //  Can't assign or copy construct
    private:
        TimeFormat(const TimeFormat &);
        void                        operator = (const TimeFormat &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        TimeFormat() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~TimeFormat() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified time format with the CARL framework.
         *
         *  A registration of a time format informs the CARL framework that
         *  a specific time format exists. From this point on, the application
         *  can use the newly registers time format in exactly the same manner
         *  as a predefined time format. Once registered, a time format
         *  cannot be unregistered.
         *
         *  A time format is always registered for a specific locale. Any attempt
         *  to get a time format instance for that locale will always yield the
         *  time format registered for that locale. It is possible to register
         *  the same time format for more than one locale.
         *
         *  @param forLocale        The locale for which the newly registered
         *                          time format shall be used.
         *  @param pTimeFormat      The time format to register.
         *  @since 1.0
         */
        static void                 registerInstance(const Locale & forLocale, TimeFormat * pTimeFormat);

        /**
         *  Returns a time formattor for the current default locale.
         *
         *  @return                 A time formattor for the current default locale.
         *  @since 1.0
         */
        static TimeFormat *         getInstance();

        /**
         *  Returns a time formattor for the specified locale.
         *
         *  @param forLocale        The locale for which a time formattor is sought.
         *  @return                 A time formattor for the specified locale.
         *  @since 1.0
         */
        static TimeFormat *         getInstance(const Locale & forLocale);

        //  Formatting
        /**
         *  Represents a precision with which the time is represented.
         *
         *  @since 1.0
         */
        enum Precision
        {
            /**
             *  The time is represented with a minute precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Minute        = 1,

            /**
             *  The time is represented with a second precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Second        = 2,

            /**
             *  The time is represented with a millisecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Millisecond   = 3,

            /**
             *  The time is represented with a microsecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Microsecond   = 4,

            /**
             *  The time is represented with a nanosecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Nanosecond    = 5
        };

        /**
         *  Formats the specified time using the specified style and precision.
         *
         *  The exact representation of the formatted value depends on the requested
         *  style and precision, but is always correct for the locale for which the
         *  formattor has been requested.
         *
         *  @param time             The value to format.
         *  @param precision        The time precision of the time formatted value.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const Time & time, Precision precision) const = 0;

        /**
         *  Formats the specified time using the custom format specification.
         *
         *  The custom format specification is a string that defines how the result
         *  of formatting will look. This string is converted to a formatted value by
         *  replacing predefined placeholders defined within by various portions of
         *  the formatted time value.
         *
         *  The following placeholders can be used:
         *
         *  -   {hour12:full} - A 2-digit hour on a 12-hour scale (01..12).
         *  -   {hour12:short} - A 1- or 2-digit hour on a 12-hour scale (1..12).
         *  -   {hour24:full} - A 2-digit hour on a 24-hour scale (00..23).
         *  -   {hour24:short} - A 1- or 2-digit hour on a 24-hour scale (0..23).
         *  -   {minute:full} - A 2-digit minute (00..59).
         *  -   {minute:short} - A 1- or 2-digit minute (0..59).
         *  -   {second:full} - A 2-digit second (00..59).
         *  -   {second:short} - A 1- or 2-digit second (0..59).
         *  -   {millisecond:full} - A 3-digit millisecond (000..999).
         *  -   {millisecond:short} - A 1- to 3-digit millisecond (0..999).
         *  -   {microsecond:full} - A 6-digit microsecond (000000..999999).
         *  -   {microsecond:short} - A 1- to 6-digit microsecond (0..999999).
         *  -   {nanosecond:full} - A 9-digit vsecond (000000000..999999999).
         *  -   {nanosecond:short} - A 1- to 9-digit nanosecond (0..999999999).
         *  -   {ampm} - An AM/PM indicator.
         *
         *  Any character in a format specification string prefixed by a backslash '\' 
         *  is interpreted literally and not as a part of a placeholder. Specifically, 
         *  if a formatted value must contain a baskclash, the backslash must be written 
         *  twice in a format specification string.
         *
         *  @param time             The value to format.
         *  @param format           The custom format specification.
         *  @return                 The string representing the formatted value.
         *  @since 1.0
         */
        virtual String              format(const Time & time, const String & format) const = 0;

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };
}

#endif  //  __CARL_Format_hpp_included__
//  End of CARL_Format.hpp
