//
//  CARL_Heap.hpp - Provides C++ heap routines.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/25 12:11:26 $
//  $Revision: 1.13 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL heap API.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Heap.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Heap_hpp_included__
#define __CARL_Heap_hpp_included__

namespace carl
{
    /**
     *  @interface IGarbageCollector CARL.hpp
     *  An abstract interface to a garbage collector.
     *
     *  A garbage collector is an agent that can take control when the heap
     *  is exhausted and attempt to free some of the memory.
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IGarbageCollector
    {
        //////////
        //  Can't assign or copy construct
    private:
        IGarbageCollector(const IGarbageCollector &);
        void                    operator = (const IGarbageCollector &);

        //////////
        //  Construction/destruction
    public:
        IGarbageCollector() {}
        virtual ~IGarbageCollector() {}

        //////////
        //  Operations
    public:
        /**
         *  Applies the garbage collector.
         *
         *  This method is called by the heap manager when an installed garbage
         *  collector must be applied.
         *
         *  @return             An approximate number of bytes that has been
         *                      freed in the heap. This value does not have to
         *                      be exact; however, a return value of 0 always
         *                      indicated that the garbage collector was unable
         *                      to free any memory.
         *  @since 1.0
         */
        virtual UIntSize        run() = 0;  //  Returns approximate number of freed bytes
    };

    /**
     *  @interface HeapManager CARL.hpp
     *  An encapsulated heap manager.
     *
     *  This singleton agent is responsible for all dynamic memory allocations.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC HeapManager
    {
        //////////
        //  Can't assign or copy construct
    private:
        HeapManager(const HeapManager &);
        void                        operator = (const HeapManager &);

        //////////
        //  Singleton
    private:
        HeapManager();
        virtual ~HeapManager();
        void *                      operator new(unsigned int size);
    public:
        /**
         *  Returns the one and only instance of this class.
         *
         *  @return                 A pointer to a singleton HeapManager instance.
         *  @since 1.0
         */
        static HeapManager *        getInstance();

        //////////
        //  Operations
    public:
        /**
         *  Installs a garbage collector.
         *
         *  The specified garbage collector is added to the set of garbage
         *  collectors registered with the heap manager. If the specified
         *  garbage collector is already installed, installing it again
         *  has no effect.
         *
         *  @param pGarbageCollector    The garbage collector to install.
         *  @return                 True if the garbage collector has been 
         *                          successfully installed; false if it was
         *                          already installed before the call.
         *  @since 1.0
         */
        bool                        installGarbageCollector(IGarbageCollector * pGarbageCollector);

        /**
         *  Installs a garbage collector.
         *
         *  The specified garbage collector is removed from the set of garbage
         *  collectors registered with the heap manager. If the specified
         *  garbage collector is not installed, uninstalling it has no effect.
         *
         *  @param pGarbageCollector    The garbage collector to uninstall.
         *  @return                 True if the garbage collector has been 
         *                          successfully uninstalled; false if it was
         *                          already not installed before the call.
         *  @since 1.0
         */
        bool                        uninstallGarbageCollector(IGarbageCollector * pGarbageCollector);

        /**
         *  Allocates a memory block from the heap and returns it.
         *
         *  @param size             The size, in bytes, to allocate.
         *  @return                 A pointer to the 1st byte of the allocated block.
         *  @exception OutOfMemoryError Thrown if there is insufficient memory.
         *  @since 1.0
         */
        void *                      allocate(UIntSize size) { return allocateFrom(size, NULL, 0); }

        /**
         *  Allocates a memory block from the heap and returns it.
         *
         *  @param size             The size, in bytes, to allocate.
         *  @param pFileName        The name of the source file from which the
         *                          allocation request was made, or NULL if this
         *                          information is not available.
         *  @param lineNumber       The number of the source line from which the
         *                          allocation request was made, or 0 if this
         *                          information is not available. 
         *  @return                 A pointer to the 1st byte of the allocated block.
         *  @exception OutOfMemoryError Thrown if there is insufficient memory.
         *  @since 1.0
         */
        void *                      allocateFrom(UIntSize size, const char * pFileName, UIntSize lineNumber);

        /**
         *  Frees a previously allocated heap block.
         *
         *  @param pBlock           A pointer to the 1st byte of a block to free.
         *  @since 1.0
         */
        void                        free(void * pBlock) { freeFrom(pBlock, NULL, 0); }

        /**
         *  Frees a previously allocated heap block.
         *
         *  @param pBlock           A pointer to the 1st byte of a block to free.
         *  @param pFileName        The name of the source file from which the
         *                          deallocation request was made, or NULL if this
         *                          information is not available.
         *  @param lineNumber       The number of the source line from which the
         *                          deallocation request was made, or 0 if this
         *                          information is not available. 
         *  @since 1.0
         */
        void                        freeFrom(void * pBlock, const char * pFileName, UIntSize lineNumber);

        /**
         *  Returns the total number of allocate() calls issued so far.
         *
         *  @return                 The total number of allocate() calls 
         *                          issued so far.
         *  @since 1.0
         */
        UInt64                      getAllocateCalls() const;

        /**
         *  Returns the total number of free() calls issued so far.
         *
         *  @return                 The total number of free() calls 
         *                          issued so far.
         *  @since 1.0
         */
        UInt64                      getFreeCalls() const;

        /**
         *  Returns the total size of all blocks currently allocated on the heap.
         *
         *  @return                 The total size of all blocks currently
         *                          allocated on the heap.
         *  @since 1.0
         */
        UIntSize                    getCurrentAllocation() const;

        /**
         *  Returns the largest size of all blocks allocated on the heap during
         *  program lifetime.
         *
         *  @return                 The largest size of all blocks allocated on 
         *                          the heap during program lifetime.
         *  @since 1.0
         */
        UIntSize                    getMaximumAllocation() const;

        /**
         *  Returns the current size of the heap (used blocks + free blocks).
         *
         *  @return                 The current size of the heap.
         *  @since 1.0
         */
        UIntSize                    getCurrentHeapSize() const;

        /**
         *  Returns the largest size of the heap (used blocks + free blocks)
         *  registered during program lifetime.
         *
         *  @return                 The largest size of the heap registered 
         *                          during program lifetime.
         *  @since 1.0
         */
        UIntSize                    getMaximumHeapSize() const;

        /**
         *  Returns the current number of used heap blocks.
         *
         *  @return                 The current number of used heap blocks.
         *  @since 1.0
         */
        UIntSize                    getCurrentUsedBlocks() const;

        /**
         *  Returns the largest number of used heap blocks registered during 
         *  program lifetime.
         *
         *  @return                 The largest number of used heap blocks 
         *                          registered during program lifetime.
         *  @since 1.0
         */
        UIntSize                    getMaximumUsedBlocks() const;

        //////////
        //  Implementation
    private:
        mutable CriticalSection     _guard;
        IGarbageCollector **        _pInstalledGarbageCollectors;
        UIntSize                    _installedGarbageCollectorsAllocated;
        UIntSize                    _installedGarbageCollectorsUsed;
        //  Live heap statistics
        UInt64                      _allocateCalls;
        UInt64                      _freeCalls;
        UIntSize                    _currentAllocation;
        UIntSize                    _maximumAllocation;
        UIntSize                    _currentHeapSize;
        UIntSize                    _maximumHeapSize;
        UIntSize                    _currentUsedBlocks;
        UIntSize                    _maximumUsedBlocks;
        //  Helpers
        void *                      _tryAllocateFrom(UIntSize size, const char * pFileName, UIntSize lineNumber);
        void                        _tryFreeFrom(void * pBlock, const char * pFileName, UIntSize lineNumber);
        bool                        _applyGarbageCollectors();

        //  Debug heap
#if defined(CARL_MODE_DEBUG)
        //  In debug mode, each block allocated on the heap has an associated heap
        //  block descriptor used for debug purposes. All heap block descriptors for
        //  currently allocated heap blocks are kept in a doubly-linked list.
        //  When an allocation of X bytes is asked for, the actual allocated heap cell
        //  is X + (2 * _HeapCellPadding) bytes large, with X bytes in the middle being
        //  used for storing user data and _HeapCellPadding bytes at the beginning and
        //  the end of the cell serving as guard bytes. These guard bytes are initialized
        //  to _Filler when a heap cell is allocated; if they are different at some later
        //  point in the future, then the heap cell has been corrupted by an invalid program.
        //  In addition, each heap cell is always immediately preceeded by a pointer DP
        //  to its corresponding heap block descriptor. When a heap block is being
        //  deallocated, the check is made to see if the heap block descriptor at the
        //  address DP points back to the heap block; if not, then we can assume that
        //  an attempt is made to free a nonexistent block.
        static const UIntSize       _HeapCellPadding = 16;
        static const UInt8          _Filler = 0xCD;
        struct _HeapBlockDescriptor
        {
            UIntSize                _cellSize;                  //  The size of the allocated cell
            UIntSize                _dataSize;                  //  The size of the data portion in the allocated cell
            const char *            _pAllocatedFromSourceFile;  //  NULL == unknown
            UIntSize                _allocatedFromSourceLine;   //  NULL == unknown
            void *                  _pCell;                     //  The start of the allocated cell
            void *                  _pData;                     //  The start of the data portion in the allocated cell
            _HeapBlockDescriptor *  _pPrevBlockDescriptor;      //  NULL == this is the 1st block
            _HeapBlockDescriptor *  _pNextBlockDescriptor;      //  NULL == this is the last block
        };
        _HeapBlockDescriptor *      _pFirstBlockDescriptor;
        _HeapBlockDescriptor *      _pLastBlockDescriptor;
#endif  //  defined(CARL_MODE_DEBUG)
    };

    /**
     *  @class OutOfMemoryError CARL.hpp
     *  An exception thrown when an attempt to allocate memory from the heap fails.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC OutOfMemoryError : public SystemError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  @param allocationRequestSize    The size, in bytes, of the block
         *                          that could not be allocated due to
         *                          insufficient memory.
         *  @since 1.0
         */
        OutOfMemoryError(UIntSize allocationRequestSize);

        /**
         *  The copy constructor.
         *
         *  @param src              The exception to copy properties from.
         *  @since 1.0
         */
        OutOfMemoryError(const OutOfMemoryError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~OutOfMemoryError();

        /**
         *  The assignment operator.
         *
         *  @param src              The exception to copy properties from.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        OutOfMemoryError &          operator = (const OutOfMemoryError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the size, in bytes, of the block that could not be 
         *  allocated due to insufficient memory.
         *
         *  @return                 The size, in bytes, of the block
         *                          that could not be allocated due to
         *                          insufficient memory.
         *  @since 1.0
         */
        UIntSize                    getAllocationRequestSize() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return                 The user-readable error message describing
         *                          this exception.
         *  @since 1.0
         */
        virtual String              toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale        The locale for which an error message shall
         *                          be tailored.
         *  @return                 The user-readable error message describing
         *                          this exception.
         *  @since 1.0
         */
        virtual String              toString(const Locale & forLocale) const;

        //////////
        //  Implementation
    private:
        UIntSize                    _allocationRequestSize;

        //  Helpers
        static String *             _getErrorMessageString();
    };

    /**
     *  @class HeapBlockCorruptError CARL.hpp
     *  An exception thrown when a heap block corruption is discovered.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC HeapBlockCorruptError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  @param pBlock           An address of the heap block that was
         *                          discovered to be corrupt.
         *  @param pSourceFileName  The name of the source file where the
         *                          heap block corruption was discoveered,
         *                          or NULL if this information is not
         *                          available.
         *  @param sourceLineNumber The number of the source line in which the
         *                          heap block corruption was discoveered, 
         *                          or 0 if this information is not available. 
         *  @since 1.0
         */
        HeapBlockCorruptError(const void * pBlock, const char * pSourceFileName, UIntSize sourceLineNumber);

        /**
         *  The copy constructor.
         *
         *  @param src              The exception to copy properties from.
         *  @since 1.0
         */
        HeapBlockCorruptError(const HeapBlockCorruptError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~HeapBlockCorruptError();

        /**
         *  The assignment operator.
         *
         *  @param src              The exception to copy properties from.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        HeapBlockCorruptError & operator = (const HeapBlockCorruptError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns an address of the heap block that was discovered to be corrupt.
         *
         *  @return                 An address of the heap block that was
         *                          discovered to be corrupt.
         *  @since 1.0
         */
        const void *                getBlock() const;

        /**
         *  Returns the name of the source file where the heap block corruption 
         *  was discoveered, or NULL if this information is not available.
         *
         *  @return                 The name of the source file where the heap 
         *                          block corruption was discoveered.
         *  @since 1.0
         */
        const char *                getSourceFileName() const;

        /**
         *  Returns the number of the source line in which the heap block corruption 
         *  was discoveered,  or 0 if this information is not available. 
         *
         *  @return                 The number of the source line in which the heap 
         *                          block corruption was discoveered.
         *  @since 1.0
         */
        UIntSize                    getSourceLineNumber() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return                 The user-readable error message describing
         *                          this exception.
         *  @since 1.0
         */
        virtual String              toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale        The locale for which an error message shall
         *                          be tailored.
         *  @return                 The user-readable error message describing
         *                          this exception.
         *  @since 1.0
         */
        virtual String              toString(const Locale & forLocale) const;

        //////////
        //  Implementation
    private:
        const void *                _pBlock;
        const char *                _pSourceFileName;
        UIntSize                    _sourceLineNumber;
    };


    /**
     *  @class NoSuchHeapBlockError CARL.hpp
     *  An exception thrown when an attempt is made to use or delete 
     *  a nonexistent heap block.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC NoSuchHeapBlockError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  @param pBlock           An address of the heap block that was
         *                          discovered to be nonexistent.
         *  @param pSourceFileName  The name of the source file where the
         *                          heap block nonexistence was discoveered,
         *                          or NULL if this information is not
         *                          available.
         *  @param sourceLineNumber The number of the source line in which the
         *                          heap block nonexistence was discoveered, 
         *                          or 0 if this information is not available. 
         *  @since 1.0
         */
        NoSuchHeapBlockError(const void * pBlock, const char * pSourceFileName, UIntSize sourceLineNumber);

        /**
         *  The copy constructor.
         *
         *  @param src              The exception to copy properties from.
         *  @since 1.0
         */
        NoSuchHeapBlockError(const NoSuchHeapBlockError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~NoSuchHeapBlockError();

        /**
         *  The assignment operator.
         *
         *  @param src              The exception to copy properties from.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        NoSuchHeapBlockError & operator = (const NoSuchHeapBlockError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns an address of the heap block that was discovered to be nonexistent.
         *
         *  @return                 An address of the heap block that was
         *                          discovered to be nonexistent.
         *  @since 1.0
         */
        const void *                getBlock() const;

        /**
         *  Returns the name of the source file where the heap block nonexistence 
         *  was discoveered, or NULL if this information is not available.
         *
         *  @return                 The name of the source file where the heap 
         *                          block nonexistence was discoveered.
         *  @since 1.0
         */
        const char *                getSourceFileName() const;

        /**
         *  Returns the name of the source file where the heap block nonexistence 
         *  was discoveered, or NULL if this information is not available.
         *
         *  @return                 The name of the source file where the heap 
         *                          block nonexistence was discoveered.
         *  @since 1.0
         */
        UIntSize                    getSourceLineNumber() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return                 The user-readable error message describing
         *                          this exception.
         *  @since 1.0
         */
        virtual String              toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale        The locale for which an error message shall
         *                          be tailored.
         *  @return                 The user-readable error message describing
         *                          this exception.
         *  @since 1.0
         */
        virtual String              toString(const Locale & forLocale) const;

        //////////
        //  Implementation
    private:
        const void *                _pBlock;
        const char *                _pSourceFileName;
        UIntSize                    _sourceLineNumber;
    };
}

//////////
//  Heap routines
/**
 *  Allocates a memory block for an object from the heap and returns it.
 *
 *  @param size             The size, in bytes, to allocate.
 *  @return                 A pointer to the 1st byte of the allocated block.
 *  @exception OutOfMemoryError Thrown if there is insufficient memory.
 *  @since 1.0
 */
inline void * CARL_API operator new(unsigned int size)
{
    return ::carl::HeapManager::getInstance()->allocate(size);
}

/**
 *  Allocates a memory block for an object from the heap and returns it.
 *
 *  @param size             The size, in bytes, to allocate.
 *  @param pFileName        The name of the source file from which the
 *                          allocation request was made, or NULL if this
 *                          information is not available.
 *  @param lineNumber       The number of the source line from which the
 *                          allocation request was made, or 0 if this
 *                          information is not available. 
 *  @return                 A pointer to the 1st byte of the allocated block.
 *  @exception OutOfMemoryError Thrown if there is insufficient memory.
 *  @since 1.0
 */
inline void * CARL_API operator new(unsigned int size, const char * pFileName, unsigned int lineNumber)
{
    return ::carl::HeapManager::getInstance()->allocateFrom(size, pFileName, lineNumber);
}

/**
 *  Allocates a memory block for an object array from the heap and returns it.
 *
 *  @param size             The size, in bytes, to allocate.
 *  @return                 A pointer to the 1st byte of the allocated block.
 *  @exception OutOfMemoryError Thrown if there is insufficient memory.
 *  @since 1.0
 */
inline void * CARL_API operator new[](unsigned int size)
{
    return ::carl::HeapManager::getInstance()->allocate(size);
}

/**
 *  Allocates a memory block for an object array from the heap and returns it.
 *
 *  @param size             The size, in bytes, to allocate.
 *  @param pFileName        The name of the source file from which the
 *                          allocation request was made, or NULL if this
 *                          information is not available.
 *  @param lineNumber       The number of the source line from which the
 *                          allocation request was made, or 0 if this
 *                          information is not available. 
 *  @return                 A pointer to the 1st byte of the allocated block.
 *  @exception OutOfMemoryError Thrown if there is insufficient memory.
 *  @since 1.0
 */
inline void * CARL_API operator new[](unsigned int size, const char * pFileName, unsigned int lineNumber)
{
    return ::carl::HeapManager::getInstance()->allocateFrom(size, pFileName, lineNumber);
}

/**
 *  Frees a previously allocated object heap block.
 *
 *  @param pBlock           A pointer to the 1st byte of a block to free.
 *  @since 1.0
 */
inline void CARL_API operator delete(void * pBlock)
{
    return ::carl::HeapManager::getInstance()->free(pBlock);
}

/**
 *  Frees a previously allocated object heap block.
 *
 *  @param pBlock           A pointer to the 1st byte of a block to free.
 *  @param pFileName        The name of the source file from which the
 *                          deallocation request was made, or NULL if this
 *                          information is not available.
 *  @param lineNumber       The number of the source line from which the
 *                          deallocation request was made, or 0 if this
 *                          information is not available. 
 *  @since 1.0
 */
inline void CARL_API operator delete(void * pBlock, const char * pFileName, unsigned int lineNumber)
{
    return ::carl::HeapManager::getInstance()->freeFrom(pBlock, pFileName, lineNumber);
}

/**
 *  Frees a previously allocated object array heap block.
 *
 *  @param pBlock           A pointer to the 1st byte of a block to free.
 *  @since 1.0
 */
inline void CARL_API operator delete[](void * pBlock)
{
    return ::carl::HeapManager::getInstance()->free(pBlock);
}

/**
 *  Frees a previously allocated object array heap block.
 *
 *  @param pBlock           A pointer to the 1st byte of a block to free.
 *  @param pFileName        The name of the source file from which the
 *                          deallocation request was made, or NULL if this
 *                          information is not available.
 *  @param lineNumber       The number of the source line from which the
 *                          deallocation request was made, or 0 if this
 *                          information is not available. 
 *  @since 1.0
 */
inline void CARL_API operator delete[](void * pBlock, const char * pFileName, unsigned int lineNumber)
{
    return ::carl::HeapManager::getInstance()->freeFrom(pBlock, pFileName, lineNumber);
}

//////////
//  Debug heap routines are tracking
#ifdef _DEBUG
    #define new new(__FILE__, __LINE__)
#endif

#endif  //  __CARL_Heap_hpp_included__
//  End of CARL_Heap.hpp
