//
//  CARL_Iso10646.hpp - ISO 10646 API
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/29 12:14:39 $
//  $Revision: 1.11 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL ISO 10646 API.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Iso10646.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Iso10646_hpp_included__
#define __CARL_Iso10646_hpp_included__

namespace carl
{
    /**
     *  @class Iso10646 CARL.hpp
     *  Implements the ISO-10646 / Unicode specification.
     *
     *  The Unicode version currently implemented is 5.0.0.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Iso10646
    {
        //////////
        //  This is a utility class that cannot be instantiated
    private:
        Iso10646();
        Iso10646(const Iso10646 &);
        ~Iso10646();

        void                    operator = (const Iso10646 &);

        //////////
        //  Types & constants
    public:
        /**
         *  Represents the general categiry of a character.
         *
         *  @since 1.0
         */
        enum GeneralCategory
        {
            /**
             *  Letter, Uppercase.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Lu  = 1,

            /**
             *  Letter, Lowercase.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Ll  = 2,

            /**
             *  Letter, Titlecase.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Lt  = 3,

            /**
             *  Letter, Modifier.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Lm  = 4,

            /**
             *  Letter, Other.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Lo  = 5,

            /**
             *  Mark, Non-Spacing.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Mn  = 6,

            /**
             *  Mark, Spacing Combining.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Mc  = 7,

            /**
             *  Mark, Enclosing.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Me  = 8,

            /**
             *  Number, Decimal Digit.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Nd  = 9,

            /**
             *  Number, Letter.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Nl  = 10,

            /**
             *  Number, Other.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_No  = 11,

            /**
             *  Punctuation, Connector.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Pc  = 12,

            /**
             *  Punctuation, Dash.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Pd  = 13,

            /**
             *  Punctuation, Open.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Ps  = 14,

            /**
             *  Punctuation, Close.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Pe  = 15,

            /**
             *  Punctuation, Initial quote (may behave like Ps or Pe depending on usage).
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Pi  = 16,

            /**
             *  Punctuation, Final quote (may behave like Ps or Pe depending on usage).
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Pf  = 17,

            /**
             *  Punctuation, Other.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Po  = 18,

            /**
             *  Symbol, Math.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Sm  = 19,

            /**
             *  Symbol, Currency.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Sc  = 20,

            /**
             *  Symbol, Modifier.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Sk  = 21,

            /**
             *  Symbol, Other.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_So  = 22,

            /**
             *  Separator, Space.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Zs  = 23,

            /**
             *  Separator, Line.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Zl  = 24,

            /**
             *  Separator, Paragraph.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Zp  = 25,

            /**
             *  Other, Control.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Cc  = 26,

            /**
             *  Other, Format.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Cf  = 27,

            /**
             *  Other, Surrogate.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Cs  = 28,

            /**
             *  Other, Private Use.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Co  = 29,

            /**
             *  Other, Not Assigned.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GeneralCategory_Cn  = 30
        };

        /**
         *  Represents the bidirectional categiry of a character.
         *
         *  @since 1.0
         */
        enum BidirectionalCategory
        {
            /**
             *  Left-to-Right.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_L     = 1,

            /**
             *  Left-to-Right Embedding.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_LRE   = 2,

            /**
             *  Left-to-Right Override.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_LRO   = 3,

            /**
             *  Right-to-Left.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_R     = 4,

            /**
             *  Right-to-Left Arabic.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_AL    = 5,

            /**
             *  Right-to-Left Embedding.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_RLE   = 6,

            /**
             *  Right-to-Left Override.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_RLO   = 7,

            /**
             *  Pop Directional Format.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_PDF   = 8,

            /**
             *  European Number.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_EN    = 9,

            /**
             *  European Number Separator.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_ES    = 10,

            /**
             *  European Number Terminator.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_ET    = 11,

            /**
             *  Arabic Number.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_AN    = 12,

            /**
             *  Common Number Separator.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_CS    = 13,

            /**
             *  Non-Spacing Mark.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_NSM   = 14,

            /**
             *  Boundary Neutral.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_BN    = 16,

            /**
             *  Paragraph Separator.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_B     = 17,

            /**
             *  Segment Separator.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_S     = 18,

            /**
             *  Whitespace.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_WS    = 19,

            /**
             *  Other Neutrals.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            BidirectionalCategory_ON    = 20
        };

        /**
         *  Represents the decomposition mapping tag of a character.
         *
         *  @since 1.0
         */
        enum DecompositionMappingTag
        {
            /**
             *  Decomposition mapping tag not available.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_none        = 1,

            /**
             *  Canonical decomposition mapping.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_canonical   = 2,

            /**
             *  A font variant (e.g. a blackletter form).
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_font        = 3,

            /**
             *  A no-break version of a space or hyphen.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_noBreak     = 4,

            /**
             *  An initial presentation form (Arabic).
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_initial     = 5,

            /**
             *  A medial presentation form (Arabic).
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_medial      = 6,

            /**
             *  A final presentation form (Arabic).
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_final       = 7,

            /**
             *  An isolated presentation form (Arabic).
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_isolated    = 8,

            /**
             *  An encircled form.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_circle      = 9,

            /**
             *  A superscript form.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_super       = 10,

            /**
             *  A subscript form.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_sub         = 11,

            /**
             *  A vertical layout presentation form.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_vertical    = 12,

            /**
             *  A wide (or zenkaku) compatibility character.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_wide        = 13,

            /**
             *  A narrow (or hankaku) compatibility character.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_narrow      = 14,

            /**
             *  A small variant form (CNS compatibility).
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_small       = 15,

            /**
             *  A CJK squared font variant.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_square      = 16,

            /**
             *  A vulgar fraction form.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_fraction    = 17,

            /**
             *  Otherwise unspecified compatibility character.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            DecompositionMappingTag_compat      = 18
        };

        //////////
        //  Operations
    public:
        //  Character classification

        /**
         *  Checks if a character is an alphanumeric character.
         *
         *  A character is an alphanumeric character if it belongs to
         *  one of the following general categories:
         *
         *  -   Lu
         *  -   Ll
         *  -   Lt
         *  -   Lm
         *  -   Lo
         *  -   Nd
         *  -   Nl
         *  -   No
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is an alphanumeric 
         *                      character; false otherwise.
         *  @since 1.0
         */
        static bool             isAlNum(Char c);

        /**
         *  Checks if a character is an alphabetic character.
         *
         *  A character is an alphabetic character if it belongs to
         *  one of the following general categories:
         *
         *  -   Lu
         *  -   Ll
         *  -   Lt
         *  -   Lm
         *  -   Lo
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is an alphabetic 
         *                      character; false otherwise.
         *  @since 1.0
         */
        static bool             isAlpha(Char c);

        /**
         *  Checks if a character is an ANSI character.
         *
         *  A character is an ANSI character if it lies in range [0..255].
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is an ANSI character;
         *                      false otherwise.
         *  @since 1.0
         */
        static bool             isAnsi(Char c) { return c < 0x00000100; }

        /**
         *  Checks if a character is an ASCII character.
         *
         *  A character is an ASCII character if it lies in range [0..127].
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is an ASCII character; 
         *                      false otherwise.
         *  @since 1.0
         */
        static bool             isAscii(Char c) { return c < 0x00000080; }

        /**
         *  Checks if a character is a control character.
         *
         *  A character is a control character if it belongs to
         *  one of the following general categories:
         *
         *  -   Cc
         *  -   Cf
         *  -   Cs
         *  -   Co
         *  -   Cn
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is a control character; 
         *                      false otherwise.
         *  @since 1.0
         */
        static bool             isControl(Char c);

        /**
         *  Checks if a character is valid within an identifier.
         *
         *  A character is considered valid within an identifier if
         *  it is an alphanumeric character or an underscore '_'.
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is valid within an 
         *                      identifier; false otherwise.
         *  @since 1.0
         */
        static bool             isCSym(Char c) { return (c == '_') || isAlNum(c); }

        /**
         *  Checks if a character is valid at the start of an identifier.
         *
         *  A character is considered valid at the start of an identifier if
         *  it is an alphabetic character or an underscore '_'.
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is valid at the start of 
         *                      an identifier; false otherwise.
         *  @since 1.0
         */
        static bool             isCSymF(Char c) { return (c == '_') || isAlpha(c); }

        /**
         *  Checks if a character is a digit character.
         *
         *  A character is a digit character if it belongs to
         *  one of the following general categories:
         *
         *  -   Nd
         *  -   Nl
         *  -   No
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is a digit character; 
         *                      false otherwise.
         *  @since 1.0
         */
        static bool             isDigit(Char c);

        /**
         *  Checks if a character is a decimal digit character.
         *
         *  A character is a decimal digit character if it belongs to
         *  the range ['0'..'9'].
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is a decimal digit 
         *                      character; false otherwise.
         *  @since 1.0
         */
        static bool             isDecimalDigit(Char c) { return (c >= '0' && c <= '9'); }

        /**
         *  Checks if a character has graphical representation.
         *
         *  A character has a graphical representation if it does not to
         *  one of the following general categories:
         *
         *  -   Cc
         *  -   Cf
         *  -   Cs
         *  -   Co
         *  -   Cn
         *  -   Zs
         *  -   Zl
         *  -   Zp
         *
         *  @param c            The character to test.
         *  @returns            True if \a c  has a graphical representation;
         *                      false otherwise.
         *  @since 1.0
         */
        static bool             isGraph(Char c);

        /**
         *  Checks if a character is a hexadecimal digit character.
         *
         *  A character is a hexadecimal digit character if it belongs to
         *  one of the ranges ['0'..'9'], ['a'..'f'] or ['A'..'F'].
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is a hexadecimal digit 
         *                      character; false otherwise.
         *  @since 1.0
         */
        static bool             isHexadecimalDigit(Char c) { return (c >= '0' && c <= '9') || (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F'); }

        /**
         *  Checks if a character is a valid ISO-10646 character.
         *
         *  A character is considered to be a valid ISO-10646 character
         *  if there exists an entry for that character in the CARL
         *  character database.
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is a valid ISO-10646 
         *                      character; false otherwise.
         *  @since 1.0
         */
        static bool             isIso10646(Char c);

        /**
         *  Checks if a character is a lowercase letter.
         *
         *  A character is a lowercase letter if it belongs to
         *  one of the following general categories:
         *
         *  -   Ll
         *
         *  @param c            The character to test.
         *  @returns            True if \a c  is a lowercase 
         *                      letter; false otherwise.
         *  @since 1.0
         */
        static bool             isLower(Char c);

        /**
         *  Checks if a character is an octal digit character.
         *
         *  A character is an octal digit character if it belongs to
         *  the range ['0'..'7'].
         *
         *  @param c            The character to test.
         *  @returns            True if \a c  is an octal digit 
         *                      character; false otherwise.
         *  @since 1.0
         */
        static bool             isOctalDigit(Char c) { return (c >= '0' && c <= '7'); }

        /**
         *  Checks if a character is printable.
         *
         *  A character is printable if it does not to
         *  one of the following general categories:
         *
         *  -   Cc
         *  -   Cf
         *  -   Cs
         *  -   Co
         *  -   Cn
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is printable; 
         *                      false otherwise.
         *  @since 1.0
         */
        static bool             isPrint(Char c);

        /**
         *  Checks if a character is a punctuation character.
         *
         *  A character is a punctuation character if it belongs to
         *  one of the following general categories:
         *
         *  -   Pc
         *  -   Pd
         *  -   Ps
         *  -   Pe
         *  -   Pi
         *  -   Pf
         *  -   Po
         *  -   Sm
         *  -   Sc
         *  -   Sk
         *  -   So
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is a punctuation 
         *                      character; false otherwise.
         *  @since 1.0
         */
        static bool             isPunct(Char c);

        /**
         *  Checks if a character is a space character.
         *
         *  A character is a space character if it belongs to
         *  one of the following general categories:
         *
         *  -   Zs
         *  -   Zl
         *  -   Zp
         *
         *  A character is also considered a space if it lies within
         *  the range return [9..13] or it is 32.
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is a space character; 
         *                      false otherwise.
         *  @since 1.0
         */
        static bool             isSpace(Char c);

        /**
         *  Checks if a character is a titlecase letter.
         *
         *  A character is a titlecase letter if it belongs to
         *  one of the following general categories:
         *
         *  -   Lt
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is a titlecase letter; 
         *                      false otherwise.
         *  @since 1.0
         */
        static bool             isTitle(Char c);

        /**
         *  Checks if a character is a Unicode character.
         *
         *  A character is an octal digit character if it belongs to
         *  the range [0..2097152] and the corresponding entry exists
         *  within the CARL character database.
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is a Unicode character; 
         *                      false otherwise.
         *  @since 1.0
         */
        static bool             isUnicode(Char c);

        /**
         *  Checks if a character is an uppercase letter.
         *
         *  A character is an uppercase letter if it belongs to
         *  one of the following general categories:
         *
         *  -   Lu
         *
         *  @param c            The character to test.
         *  @returns            True if \a c is an uppercase letter; 
         *                      false otherwise.
         *  @since 1.0
         */
        static bool             isUpper(Char c);

        //  Character case conversion
        /**
         *  Returns the lowercase equivalent of a character.
         *
         *  If the specified character has no lowercase equivalent,
         *  the returned value is the character itself.
         *
         *  @param c            The character to convert.
         *  @return             The lowercase equivalent of a character.
         *  @since 1.0
         */
        static Char             toLower(Char c);

        /**
         *  Returns the uppercase equivalent of a character.
         *
         *  If the specified character has no uppercase equivalent,
         *  the returned value is the character itself.
         *
         *  @param c            The character to convert.
         *  @return             The uppercase equivalent of a character.
         *  @since 1.0
         */
        static Char             toUpper(Char c);

        /**
         *  Returns the titlecase equivalent of a character.
         *
         *  If the specified character has no titlecase equivalent,
         *  the returned value is the character itself.
         *
         *  @param c            The character to convert.
         *  @return             The titlecase equivalent of a character.
         *  @since 1.0
         */
        static Char             toTitle(Char c);

        //  Direct access to ISO-10646 data
        /**
         *  Queries the general category of a character.
         *
         *  @param c            The character to query.
         *  @param generalCategory  Upon successful query, this parameter
         *                      is assigned the general category of \a c.
         *  @return             True if the query was successful; false
         *                      if the specified character does not exist
         *                      in the CARL character database.
         *  @since 1.0
         */
        static bool             getGeneralCategory(Char c, GeneralCategory & generalCategory);

        /**
         *  Queries the canonical combining class of a character.
         *
         *  @param c            The character to query.
         *  @param canonicalCombiningClass  Upon successful query, this parameter
         *                      is assigned the canonical combining class of \a c.
         *  @return             True if the query was successful; false
         *                      if the specified character does not exist
         *                      in the CARL character database.
         *  @since 1.0
         */
        static bool             getCanonicalCombiningClass(Char c, UInt8 & canonicalCombiningClass);

        /**
         *  Queries the bidirectional category of a character.
         *
         *  @param c            The character to query.
         *  @param bidirectionalCategory    Upon successful query, this parameter
         *                      is assigned the bidirectional category of \a c.
         *  @return             True if the query was successful; false
         *                      if the specified character does not exist
         *                      in the CARL character database.
         *  @since 1.0
         */
        static bool             getBidirectionalCategory(Char c, BidirectionalCategory & bidirectionalCategory);

        /**
         *  Queries the decomposition mapping tag of a character.
         *
         *  @param c            The character to query.
         *  @param decompositionMappingTag  Upon successful query, this parameter
         *                      is assigned the decomposition mapping tag of \a c.
         *  @return             True if the query was successful; false
         *                      if the specified character does not exist
         *                      in the CARL character database.
         *  @since 1.0
         */
        static bool             getDecompositionMappingTag(Char c, DecompositionMappingTag & decompositionMappingTag);

        /**
         *  Queries the decomposition mapping length of a character.
         *
         *  A decomposition mapping length of a character is defined
         *  as a number of other characters that the original character
         *  immediately decomposes to.
         *
         *  @param c            The character to query.
         *  @param decompositionMappingSize Upon successful query, this parameter
         *                      is assigned the decomposition mapping length of \a c.
         *  @return             True if the query was successful; false
         *                      if the specified character does not exist
         *                      in the CARL character database or does not
         *                      have a decomposition mapping.
         *  @since 1.0
         */
        static bool             getDecompositionMappingSize(Char c, UIntSize & decompositionMappingSize);

        /**
         *  Queries the decomposition mapping of a character.
         *
         *  A decomposition mapping of a character is defined as a number 
         *  of other characters that the original character immediately 
         *  decomposes to.
         *
         *  @param c            The character to query.
         *  @param pDecompositionMapping    Upon successful query, the decomposition
         *                      mapping of \a c is stored in the buffer pointer to
         *                      by this parameter. The buffer is assumed to be of
         *                      sufficient size to accomodate the entire decomposition
         *                      mapping.
         *  @return             True if the query was successful; false
         *                      if the specified character does not exist
         *                      in the CARL character database or does not
         *                      have a decomposition mapping.
         *  @since 1.0
         */
        static bool             getDecompositionMapping(Char c, Char * pDecompositionMapping);

        /**
         *  Queries the decimal digit value of a character.
         *
         *  @param c            The character to query.
         *  @param decimalDigitValue    Upon successful query, this parameter
         *                      is assigned the decimal digit value of \a c.
         *  @return             True if the query was successful; false
         *                      if the specified character does not exist
         *                      in the CARL character database or is not
         *                      a decimal digit.
         *  @since 1.0
         */
        static bool             getDecimalDigitValue(Char c, UInt8 & decimalDigitValue);

        /**
         *  Queries the digit value of a character.
         *
         *  @param c            The character to query.
         *  @param digitValue   Upon successful query, this parameter
         *                      is assigned the digit value of \a c.
         *  @return             True if the query was successful; false
         *                      if the specified character does not exist
         *                      in the CARL character database or is not
         *                      a digit.
         *  @since 1.0
         */
        static bool             getDigitValue(Char c, Int64 & digitValue);

        /**
         *  Queries the numeric value of a character.
         *
         *  A numeric value of a character can be integer or rational.
         *  In both cases, a pair (numerator, denominator) is returned.
         *  The rational numeric value of such a pair is
         *  numerator / denominator. For integer numeric values,
         *  denominator is always 1.
         *
         *  @param c            The character to query.
         *  @param numerator    Upon successful query, this parameter
         *                      is assigned the numerator component of
         *                      the numeric value of \a c.
         *  @param denominator  Upon successful query, this parameter
         *                      is assigned the denominator component of
         *                      the numeric value of \a c.
         *  @return             True if the query was successful; false
         *                      if the specified character does not exist
         *                      in the CARL character database or does not
         *                      have a numeric value.
         *  @since 1.0
         */
        static bool             getNumericValue(Char c, Int64 & numerator, Int64 & denominator);

        /**
         *  Queries the mirroring property of a character.
         *
         *  @param c            The character to query.
         *  @return             True if the specified character is mirrored;
         *                      false if the character does not appear in the 
         *                      CARL character database or is mirrored.
         *  @since 1.0
         */
        static bool             isMirrored(Char c);

        //  Character <-> digit conversion
        /**
         *  Converts a character to a digit.
         *
         *  Any character that has a digit value converts to that
         *  value. In addition, the following nondigit characters
         *  are converted to digits:
         *
         *  -   Characters in range ['a'..'z'] are converted to digits [10..35].
         *  -   Characters in range ['A'..'Z'] are converted to digits [10..35].
         *
         *  @param c            The character to convert.
         *  @param radix        The radix to use for conversion (2 <= \a radix <= 36).
         *  @param digit        Upon successful conversion, this parameter is
         *                      assigned the value of the digit that is represented
         *                      by \a c.
         *  @return             True if the character-to-digit conversion was
         *                      successful, false otherwise. The latter can happed
         *                      when \a c does not represent a digit, or when the
         *                      digit represented by \a c is too large for the
         *                      specified radix.
         *  @since 1.0
         */
        static bool             charToDigit(Char c, int radix, int & digit);

        /**
         *  Converts a digit to a character.
         *
         *  The following conversion rules are used:
         *
         *  -   Digits in range [0..9] are converted to characters 
         *      in range ['0'..'9'].
         *  -   Digits in range [10..35] are converted to characters 
         *      in range ['A'..'Z'].
         *
         *  @param digit        The digit to convert.
         *  @param radix        The radix to use for conversion (2 <= \a radix <= 36).
         *  @param c            Upon successful conversion, this parameter is
         *                      assigned the value of the character that represents
         *                      the specified digit.
         *  @return             True if the digit-to-character conversion was
         *                      successful, false otherwise. The latter can happed
         *                      digit represented by \a digit is too large for the
         *                      specified radix.
         *  @since 1.0
         */
        static bool             digitToChar(int digit, int radix, Char & c);

        //  Decomposition helpers
        /**
         *  Performs a canonical ordering of a string.
         *
         *  @param s            The string to perform canonical ordering on.
         *  @return             The canonical ordering of \a s.
         *  @since 1.0
         */
        static String           performCanonicalOrdering(const String & s);

        /**
         *  Performs a canonical decomposition of a string.
         *
         *  @param s            The string to perform canonical decomposition on.
         *  @return             The canonical decomposition of \a s.
         *  @since 1.0
         */
        static String           performCanonicalDecomposition(const String & s);

        /**
         *  Performs a compatibility decomposition of a string.
         *
         *  @param s            The string to perform compatibility decomposition on.
         *  @return             The compatibility decomposition of \a s.
         *  @since 1.0
         */
        static String           performCompatibilityDecomposition(const String & s);

        /**
         *  Performs a canonical composition of a string.
         *
         *  @param s            The string to perform canonical composition on.
         *  @return             The canonical composition of \a s.
         *  @since 1.0
         */
        static String           performCanonicalComposition(const String & s);

        /**
         *  Converts a string to Normal Form D.
         *
         *  @param s            The string to convert to normal form D.
         *  @return             The normal form D of \a s.
         *  @since 1.0
         */
        static String           toNormalFormD(const String & s) { return performCanonicalDecomposition(s); }

        /**
         *  Converts a string to Normal Form C.
         *
         *  @param s            The string to convert to normal form C.
         *  @return             The normal form C of \a s.
         *  @since 1.0
         */
        static String           toNormalFormC(const String & s) { return performCanonicalComposition(performCanonicalDecomposition(s)); }

        /**
         *  Converts a string to Normal Form KD.
         *
         *  @param s            The string to convert to normal form KD.
         *  @return             The normal form KD of \a s.
         *  @since 1.0
         */
        static String           toNormalFormKD(const String & s) { return performCompatibilityDecomposition(s); }

        /**
         *  Converts a string to Normal Form KC.
         *
         *  @param s            The string to convert to normal form KC.
         *  @return             The normal form KC of \a s.
         *  @since 1.0
         */
        static String           toNormalFormKC(const String & s) { return performCanonicalComposition(performCompatibilityDecomposition(s)); }

        //////////
        //  Implementation
    private:

        //  Helpers
        static String           _performCanonicalDecompositionWithoutOrdering(const String & s);
        static String           _performCompatibilityDecompositionWithoutOrdering(const String & s);
        static bool             _isAscii7String(const String & s);

        //  Composition tree
        class _CompositionTree
        {
            //////////
            //  Can't assign or copy construct
        private:
            _CompositionTree(const _CompositionTree &);
            void                operator = (const _CompositionTree &);

            //////////
            //  Construction/destruction
        public:
            _CompositionTree();
            explicit _CompositionTree(Char c);
            ~_CompositionTree();

            //////////
            //  Operations
        public:
            bool                hasCharacter() const;
            Char                getCharacter() const;
            _CompositionTree *  findSubtree(Char c);
            _CompositionTree *  createSubtree(Char c);
            _CompositionTree *  createSubtree(Char c, Char composition);

            //////////
            //  Implementation
        private:
            bool                _hasCharacter;
            Char                _character;
            HashMap<Char,_CompositionTree*> _subtrees;
        };

        class _CompositionTreeRoot
        {
            //////////
            //  Can't assign or copy construct
        private:
            _CompositionTreeRoot(const _CompositionTreeRoot &);
            void                operator = (const _CompositionTreeRoot &);

            //////////
            //  Singleton
        private:
            _CompositionTreeRoot();
            ~_CompositionTreeRoot();
        public:
            static _CompositionTreeRoot *   getInstance();
            bool                findComposition(const String & s, UIntSize startPosition, Char & composition, UIntSize & compositionLength);

            //////////
            //  Implementation
        private:
            _CompositionTree *  _pRoot;
        };
    };
}

#endif  //  __CARL_Iso10646_hpp_included__
//  End of CARL_Iso10646.hpp
