//
//  CARL_Iso3166.hpp - ISO 3166 API
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL ISO 3166 API.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Iso3166.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Iso3166_hpp_included__
#define __CARL_Iso3166_hpp_included__

namespace carl
{
    namespace _impl
    {
        //////////
        //  An implementation of an ISO 639 lookup table
        class CARL_PUBLIC _Iso3166Impl;
    }

    /**
     *  @class Iso3166 CARL.hpp
     *  Implements the ISO-3166 specification.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Iso3166
    {
        //////////
        //  This is a utility class that cannot be instantiated
    private:
        Iso3166();
        Iso3166(const Iso3166 &);
        ~Iso3166();

        void                            operator = (const Iso3166 &);

        //////////
        //  Types
    public:
        /**
         *  @class Country CARL.hpp
         *  A country included into the ISO-3166 standard.
         *
         *  @since 1.0
         */
        class CARL_PUBLIC Country
        {
            friend class _impl::_Iso3166Impl;
            //////////
            //  Can't assign or copy construct
        private:
            Country(const Country &);
            void                        operator = (const Country &);

            //////////
            //  Construction/destruction - from friend classes only
        private:
            Country(const char * pCode2, const char * pCode3, int codeN);
            ~Country() {}

            //////////
            //  Operations
        public:
            /**
             *  Returns the 2-letter country code of this country.
             *
             *  The returned value always consists of uppercase letters.
             *
             *  @return                 The 2-letter country code.
             *  @since 1.0
             */
            String                      getCode2() const { return String(_code2, 2); }

            /**
             *  Returns the 3-letter country code of this country.
             *
             *  The returned value always consists of uppercase letters.
             *
             *  @return                 The 3-letter country code.
             *  @since 1.0
             */
            String                      getCode3() const { return String(_code3, 3);}

            /**
             *  Returns the numeric country code of this country.
             *
             *  @return                 The numeric country code.
             *  @since 1.0
             */
            int                         getCodeN() const { return _codeN; }

            /**
             *  Returns the user-readable display name of this country
             *  for the current default locale.
             *
             *  @return                 The display name of this country.
             *  @since 1.0
             */
            String                      getDisplayName() const { return getDisplayName(Locale::getDefault()); }

            /**
             *  Returns the user-readable display name of this country
             *  for the specified locale.
             *
             *  @param forLocale        The locale for which the display name of
             *                          the country shall be tailored.
             *  @return                 The display name of this country.
             *  @since 1.0
             */
            String                      getDisplayName(const Locale & forLocale) const;

            //////////
            //  Implementation
        private:
            Char                        _code2[2];  //  ISO-3166 2-character country code
            Char                        _code3[3];  //  ISO-3166 3-character country code
            int                         _codeN;     //  ISO-3166 numeric country code
        };

        //////////
        //  Operations
    public:
        /**
         *  Checks if a string is a valid 2-letter country code.
         *
         *  A country code is valid if it appears in the ISO-3166 country
         *  database. The string is treated case-insensitively.
         *
         *  @param pCode2               A pointer to a nul-terminated ASCII
         *                              string to check.
         *  @return                     True if the specified string is a valid
         *                              2-letter ISO-3166 country code; false otherwise.
         *  @since 1.0
         */
        static bool                     isValidCode2(const char * pCode2);

        /**
         *  Checks if a string is a valid 2-letter country code.
         *
         *  A country code is valid if it appears in the ISO-3166 country
         *  database. The string is treated case-insensitively.
         *
         *  @param pCode2               A pointer to a nul-terminated ISO-10646
         *                              string to check.
         *  @return                     True if the specified string is a valid
         *                              2-letter ISO-3166 country code; false otherwise.
         *  @since 1.0
         */
        static bool                     isValidCode2(const Char * pCode2);

        /**
         *  Checks if a string is a valid 2-letter country code.
         *
         *  A country code is valid if it appears in the ISO-3166 country
         *  database. The string is treated case-insensitively.
         *
         *  @param code2                A string to check.
         *  @return                     True if the specified string is a valid
         *                              2-letter ISO-3166 country code; false otherwise.
         *  @since 1.0
         */
        static bool                     isValidCode2(const String & code2);

        /**
         *  Checks if a string is a valid 3-letter country code.
         *
         *  A country code is valid if it appears in the ISO-3166 country
         *  database. The string is treated case-insensitively.
         *
         *  @param pCode3               A pointer to a nul-terminated ASCII
         *                              string to check.
         *  @return                     True if the specified string is a valid
         *                              3-letter ISO-3166 country code; false otherwise.
         *  @since 1.0
         */
        static bool                     isValidCode3(const char * pCode3);

        /**
         *  Checks if a string is a valid 3-letter country code.
         *
         *  A country code is valid if it appears in the ISO-3166 country
         *  database. The string is treated case-insensitively.
         *
         *  @param pCode3               A pointer to a nul-terminated ISO-10646
         *                              string to check.
         *  @return                     True if the specified string is a valid
         *                              3-letter ISO-3166 country code; false otherwise.
         *  @since 1.0
         */
        static bool                     isValidCode3(const Char * pCode3);

        /**
         *  Checks if a string is a valid 3-letter country code.
         *
         *  A country code is valid if it appears in the ISO-3166 country
         *  database. The string is treated case-insensitively.
         *
         *  @param code3                A string to check.
         *  @return                     True if the specified string is a valid
         *                              3-letter ISO-3166 country code; false otherwise.
         *  @since 1.0
         */
        static bool                     isValidCode3(const String & code3);

        /**
         *  Checks if a number is a valid numeric country code.
         *
         *  A country code is valid if it appears in the ISO-3166 country
         *  database.
         *
         *  @param codeN                A numeric value to check.
         *  @return                     True if the specified number is a valid
         *                              numeric ISO-3166 country code; false otherwise.
         *  @since 1.0
         */
        static bool                     isValidCodeN(int codeN);

        /**
         *  Returns a list of all countries registered in the ISO-3166
         *  country database.
         *
         *  Elements in the returned list are not guaranteed to be in
         *  any particular order.
         *
         *  @return                     A list of all countries registered 
         *                              in the ISO-3166 country database.
         *  @since 1.0
         */
        static Array<const Country*>    getAllCountries();

        /**
         *  Finds a country by its 2-letter country code.
         *
         *  The country code string is treated case-insensitively.
         *
         *  @param pCode2               A pointer to a nul-terminated ASCII string
         *                              representing the 2-letter country code.
         *  @return                     The country with the specified code, or
         *                              NULL if there is no such.
         *  @since 1.0
         */
        static const Country *          findCountryByCode2(const char * pCode2);

        /**
         *  Finds a country by its 2-letter country code.
         *
         *  The country code string is treated case-insensitively.
         *
         *  @param pCode2               A pointer to a nul-terminated ISO-10646 string
         *                              representing the 2-letter country code.
         *  @return                     The country with the specified code, or
         *                              NULL if there is no such.
         *  @since 1.0
         */
        static const Country *          findCountryByCode2(const Char * pCode2);

        /**
         *  Finds a country by its 2-letter country code.
         *
         *  The country code string is treated case-insensitively.
         *
         *  @param code2                A string representing the 2-letter 
         *                              country code.
         *  @return                     The country with the specified code, or
         *                              NULL if there is no such.
         *  @since 1.0
         */
        static const Country *          findCountryByCode2(const String & code2);

        /**
         *  Finds a country by its 3-letter country code.
         *
         *  The country code string is treated case-insensitively.
         *
         *  @param pCode3               A pointer to a nul-terminated ASCII string
         *                              representing the 3-letter country code.
         *  @return                     The country with the specified code, or
         *                              NULL if there is no such.
         *  @since 1.0
         */
        static const Country *          findCountryByCode3(const char * pCode3);

        /**
         *  Finds a country by its 3-letter country code.
         *
         *  The country code string is treated case-insensitively.
         *
         *  @param pCode3               A pointer to a nul-terminated ISO-10646 string
         *                              representing the 3-letter country code.
         *  @return                     The country with the specified code, or
         *                              NULL if there is no such.
         *  @since 1.0
         */
        static const Country *          findCountryByCode3(const Char * pCode3);

        /**
         *  Finds a country by its 3-letter country code.
         *
         *  The country code string is treated case-insensitively.
         *
         *  @param code3                A string representing the 3-letter 
         *                              country code.
         *  @return                     The country with the specified code, or
         *                              NULL if there is no such.
         *  @since 1.0
         */
        static const Country *          findCountryByCode3(const String & code3);

        /**
         *  Finds a country by its numeric country code.
         *
         *  @param codeN                A numeric country code.
         *  @return                     The country with the specified code, or
         *                              NULL if there is no such.
         *  @since 1.0
         */
        static const Country *          findCountryByCodeN(int codeN);
    };
}

#endif  //  __CARL_Iso3166_hpp_included__
//  End of CARL_Iso3166.hpp
