//
//  CARL_Iso639.hpp - ISO 639 API
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL ISO 639 API.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Iso639.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Iso639_hpp_included__
#define __CARL_Iso639_hpp_included__

namespace carl
{
    namespace _impl
    {
        //////////
        //  An implementation of an ISO 639 lookup table
        class CARL_PUBLIC _Iso639Impl;
    }

    /**
     *  @class Iso639 CARL.hpp
     *  Implements the ISO-639 specification.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Iso639
    {
        //////////
        //  This is a utility class that cannot be instantiated
    private:
        Iso639();
        Iso639(const Iso639 &);
        ~Iso639();

        void                            operator = (const Iso639 &);

        //////////
        //  Types
    public:
        /**
         *  @class Language CARL.hpp
         *  A language included into the ISO-639 standard.
         *
         *  @since 1.0
         */
        class CARL_PUBLIC Language
        {
            friend class _impl::_Iso639Impl;
            //////////
            //  Can't assign or copy construct
        private:
            Language(const Language &);
            void                        operator = (const Language &);

            //////////
            //  Construction/destruction - from friend classes only
        private:
            Language(const char * pCode2);
            ~Language() {}

            //////////
            //  Operations
        public:
            /**
             *  Returns the 2-letter language code of this language.
             *
             *  The returned value always consists of lowercase letters.
             *
             *  @return                 The 2-letter language code.
             *  @since 1.0
             */
            String                      getCode2() const { return String(_code2, 2); }

            /**
             *  Returns the user-readable display name of this language
             *  for the current default locale.
             *
             *  @return                 The display name of this language.
             *  @since 1.0
             */
            String                      getDisplayName() const { return getDisplayName(Locale::getDefault()); }

            /**
             *  Returns the user-readable display name of this language
             *  for the specified locale.
             *
             *  @param forLocale        The locale for which the display name of
             *                          the language shall be tailored.
             *  @return                 The display name of this language.
             *  @since 1.0
             */
            String                      getDisplayName(const Locale & forLocale) const;

            //////////
            //  Implementation
        private:
            Char                        _code2[2];  //  ISO-639 language code
        };

        //////////
        //  Operations
    public:
        /**
         *  Checks if a string is a valid 2-letter language code.
         *
         *  A language code is valid if it appears in the ISO-639 language
         *  database. The string is treated case-insensitively.
         *
         *  @param pCode2               A pointer to a nul-terminated ASCII
         *                              string to check.
         *  @return                     True if the specified string is a valid
         *                              2-letter ISO-639 language code; false otherwise.
         *  @since 1.0
         */
        static bool                     isValidCode2(const char * pCode2);

        /**
         *  Checks if a string is a valid 2-letter language code.
         *
         *  A language code is valid if it appears in the ISO-639 language
         *  database. The string is treated case-insensitively.
         *
         *  @param pCode2               A pointer to a nul-terminated ISO-10646
         *                              string to check.
         *  @return                     True if the specified string is a valid
         *                              2-letter ISO-639 language code; false otherwise.
         *  @since 1.0
         */
        static bool                     isValidCode2(const Char * pCode2);

        /**
         *  Checks if a string is a valid 2-letter language code.
         *
         *  A language code is valid if it appears in the ISO-639 language
         *  database. The string is treated case-insensitively.
         *
         *  @param code2                A string to check.
         *  @return                     True if the specified string is a valid
         *                              2-letter ISO-639 language code; false otherwise.
         *  @since 1.0
         */
        static bool                     isValidCode2(const String & code2);

        /**
         *  Returns a list of all languages registered in the ISO-639
         *  language database.
         *
         *  Elements in the returned list are not guaranteed to be in
         *  any particular order.
         *
         *  @return                     A list of all languages registered 
         *                              in the ISO-639 language database.
         *  @since 1.0
         */
        static Array<const Language*>   getAllLanguages();

        /**
         *  Finds a language by its 2-letter language code.
         *
         *  The language code string is treated case-insensitively.
         *
         *  @param pCode2               A pointer to a nul-terminated ASCII string
         *                              representing the 2-letter language code.
         *  @return                     The language with the specified code, or
         *                              NULL if there is no such.
         *  @since 1.0
         */
        static const Language *         findLanguageByCode2(const char * pCode2);

        /**
         *  Finds a language by its 2-letter language code.
         *
         *  The language code string is treated case-insensitively.
         *
         *  @param pCode2               A pointer to a nul-terminated ISO-10646 string
         *                              representing the 2-letter language code.
         *  @return                     The language with the specified code, or
         *                              NULL if there is no such.
         *  @since 1.0
         */
        static const Language *         findLanguageByCode2(const Char * pCode2);

        /**
         *  Finds a language by its 2-letter language code.
         *
         *  The language code string is treated case-insensitively.
         *
         *  @param code2                A string representing the 2-letter 
         *                              language code.
         *  @return                     The language with the specified code, or
         *                              NULL if there is no such.
         *  @since 1.0
         */
        static const Language *         findLanguageByCode2(const String & code2);
    };
}

#endif  //  __CARL_Iso639_hpp_included__
//  End of CARL_Iso639.hpp
