//
//  CARL_Iterator.hpp - Iteratorss.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.10 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines iterator framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Iterator.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Iterator_hpp_included__
#define __CARL_Iterator_hpp_included__

namespace carl
{
    template<class T> class Iterator;

    namespace _impl
    {
        //////////
        //  An implementation of an iterator over a set of elements of type T
        //
        //  The value type (T) must be:
        //  *   Copy constructable (T::T(const T &) must be defined and public)
        //  *   Destructable (T::~T() must be defined and public)
        //  *   Assignable (T::operator=(const T &) must be defined and public)
        template<class T>
        class _IteratorImpl
        {
            friend class Iterator<T>;

            //////////
            //  Can't assign or copy construct
        private:
            _IteratorImpl(_IteratorImpl<T> & );
            void                    operator = (const _IteratorImpl<T> & );

            //////////
            //  Construction/destruction
        public:
            _IteratorImpl();
            virtual ~_IteratorImpl();

            //////////
            //  Operations
        public:
            virtual bool            finished() const = 0;
            virtual T               item() const = 0;
            virtual void            advance() = 0;

            //////////
            //  Implementation
        private:
            UIntSize                _referenceCount;
        };

        template<class T>
        _IteratorImpl<T>::_IteratorImpl() 
        { 
            _referenceCount = 0; 
        }

        template<class T>
        _IteratorImpl<T>::~_IteratorImpl() 
        {
        }

        //////////
        //  An implementation of a snapshot iterator over a set of elements of type T
        //
        //  The value type (T) must be:
        //  *   Default constructable (T::T() must be defined and public)
        //  *   Copy constructable (T::T(const T &) must be defined and public)
        //  *   Destructable (T::~T() must be defined and public)
        //  *   Assignable (T::operator=(const T &) must be defined and public)
        template<class T>
        class _SnapshotIteratorImpl : public _IteratorImpl<T>
        {
            //////////
            //  Can't assign or copy construct
        private:
            _SnapshotIteratorImpl(_SnapshotIteratorImpl<T> & );
            void                    operator = (const _SnapshotIteratorImpl<T> & );

            //////////
            //  Construction/destruction
        public:
            _SnapshotIteratorImpl(UIntSize elementCount, T * pElements);
            virtual ~_SnapshotIteratorImpl();

            //////////
            //  Operations
        public:
            virtual bool            finished() const;
            virtual T               item() const;
            virtual void            advance();

            //////////
            //  Implementation
        private:
            UIntSize                _elementCount;
            T *                     _pElements;
            UIntSize                _currentPosition;
        };

        template<class T>
        _SnapshotIteratorImpl<T>::_SnapshotIteratorImpl(UIntSize elementCount, T * pElements)
        {
            _elementCount = elementCount;
            _pElements = pElements;
            _currentPosition = 0;
        }

        template<class T>
        _SnapshotIteratorImpl<T>::~_SnapshotIteratorImpl()
        {
            delete [] _pElements;
        }

        template<class T>
        bool _SnapshotIteratorImpl<T>::finished() const 
        { 
            return _currentPosition >= _elementCount; 
        }

        template<class T>
        T _SnapshotIteratorImpl<T>::item() const 
        { 
            require(!finished());
            return _pElements[_currentPosition];
        }

        template<class T>
        void _SnapshotIteratorImpl<T>::advance()
        {
            require(!finished());
            _currentPosition++;
        }
    }

    /**
     *  @class Iterator CARL.hpp
     *  An iterator over a set of elements of type T.
     *
     *  To a programmer, an iterator consists of:
     *
     *  -   An ordered set of elements to iterate over, and
     *  -   A pointer to the current element.
     *
     *  When an iterator is created, the "current element" pointer refers
     *  to the first element of the set. The "current element" pointer can
     *  then be advanced until it reaches the end of the set; each time
     *  it is advanced the next element of the set can be accessed.
     *
     *  The value type (T) must be:
     *
     *  -   Default constructable (T::T() must be defined and public)
     *  -   Copy constructable (T::T(const T &) must be defined and public).
     *  -   Destructable (T::~T() must be defined and public).
     *  -   Assignable (T::operator=(const T &) must be defined and public).
     *
     *  Iterators are independent of the data structures that create them (i.e.
     *  changing a contents of a data structure does not affect the set of
     *  elements that any existing iterator is iterating over).
     *
     *  @param T                The type of the value iterated over.
     *  @since 1.0
     */
    template<class T>
    class Iterator
    {
        //////////
        //  Construction/destruction/assignment
    public:
        Iterator(_impl::_IteratorImpl<T> * pIteratorImpl);
        /**
         *  The copy constructor.
         *
         *  @param src          The iterator to copy properties from.
         *  @since 1.0
         */
        Iterator(const Iterator<T> & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~Iterator();

        /**
         *  The assignment operator.
         *
         *  @param src          The iterator to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        Iterator &                  operator = (const Iterator<T> & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if there are more elements to iterate over.
         *
         *  @return                 True if all elements have been iterated 
         *                          over; false otherwise.
         *  @since 1.0
         */
        bool                        finished() const;

        /**
         *  Returns the current element.
         *
         *  If the iterator has already iterated over all necessary
         *  elements, the call is an error.
         *
         *  @return                 The current element.
         *  @since 1.0
         */
        T                           item() const;

        /**
         *  Advances the iterator to the next element.
         *
         *  If the iterator has already iterated over all necessary
         *  elements, the call is an error.
         *
         *  @since 1.0
         */
        void                        advance();

        //////////
        //  Implementation
    private:
        _impl::_IteratorImpl<T> *   _pIteratorImpl;
    };

    template<class T>
    Iterator<T>::Iterator(_impl::_IteratorImpl<T> * pIteratorImpl)
    {
        require(pIteratorImpl != 0);
        _pIteratorImpl = pIteratorImpl;
        _pIteratorImpl->_referenceCount++;
    }

    template<class T>
    Iterator<T>::Iterator(const Iterator<T> & src)
    {
        _pIteratorImpl = src._pIteratorImpl;
        _pIteratorImpl->_referenceCount++;
    }

    template<class T>
    Iterator<T>::~Iterator()
    {
        _pIteratorImpl->_referenceCount--;
        if (_pIteratorImpl->_referenceCount == 0)
        {
            delete _pIteratorImpl;
        }
    }

    template<class T>
    Iterator<T> & Iterator<T>::operator = (const Iterator<T> & src)
    {
        if (this != (&src) && _pIteratorImpl != src._pIteratorImpl)
        {
            _pIteratorImpl->_referenceCount--;
            if (_pIteratorImpl->_referenceCount == 0)
            {
                delete _pIteratorImpl;
            }
            _pIteratorImpl = src._pIteratorImpl;
            _pIteratorImpl->_referenceCount++;
        }
        return *this;
    }

    template<class T>
    bool Iterator<T>::finished() const 
    { 
        return _pIteratorImpl->finished(); 
    }

    template<class T>
    T Iterator<T>::item() const 
    { 
        return _pIteratorImpl->item(); 
    }

    template<class T>
    void Iterator<T>::advance() 
    { 
        _pIteratorImpl->advance(); 
    }
}

#endif  //  __CARL_Iterator_hpp_included__
//  End of CARL_Iterator.hpp
