//
//  CARL_Library.hpp - Support for dynamically loading shared libraries (DLLs)
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.9 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL dynamic library loading services.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Library.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Library_hpp_included__
#define __CARL_Library_hpp_included__

namespace carl
{
    /**
     *  @class Library CARL.hpp
     *  A dynamically loadable library.
     *
     *  On modern platforms, libraries required by an application can
     *  be loaded dynamically at runtime (DLLs on Win32, shared objects
     *  on Unix, etc.)
     *
     *  Each instance of this class corresponds to one such library.
     *  It can be in one of the following state:
     *
     *  -   Unbound - a Library instance is not bound to any dynamically
     *      loaded library.
     *  -   Bound - a Library instance is bound to a specific dynamically
     *      loaded library, but the said library is not yet loaded.
     *  -   Loaded - a Library instance is bound to a specific dynamically
     *      loaded library, and the said library is loaded and can be used
     *      by the application.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Library
    {
        //////////
        //  Can't assign or copy construct
    private:
        Library(const Library &);
        void                operator = (const Library &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  The newly created Library instance is in an Unbund state.
         *
         *  @since 1.0
         */
        Library();

        /**
         *  The class constructor.
         *
         *  The newly created Library instance is bound to the specified
         *  library file name. If the \a load parameter is true, an attempt
         *  is made to load the library; if this attempt succeeds then the
         *  library is in the Loaded state.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing the file name of the library.
         *  @param load         If true, the library is loaded; otherwise
         *                      the library is constructed in Bound state.
         *  @exception IOException  Thrown if the \a load parameter is true
         *                      but the library could not be loaded for some
         *                      reason.
         *  @since 1.0
         */
        Library(const char * pFileName, bool load = true) throw(IOException);

        /**
         *  The class constructor.
         *
         *  The newly created Library instance is bound to the specified
         *  library file name. If the \a load parameter is true, an attempt
         *  is made to load the library; if this attempt succeeds then the
         *  library is in the Loaded state.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing the file name of the library.
         *  @param load         If true, the library is loaded; otherwise
         *                      the library is constructed in Bound state.
         *  @exception IOException  Thrown if the \a load parameter is true
         *                      but the library could not be loaded for some
         *                      reason.
         *  @since 1.0
         */
        Library(const Char * pFileName, bool load = true) throw(IOException);

        /**
         *  The class constructor.
         *
         *  The newly created Library instance is bound to the specified
         *  library file name. If the \a load parameter is true, an attempt
         *  is made to load the library; if this attempt succeeds then the
         *  library is in the Loaded state.
         *
         *  @param fileName     A string representing the file name of the library.
         *  @param load         If true, the library is loaded; otherwise
         *                      the library is constructed in Bound state.
         *  @exception IOException  Thrown if the \a load parameter is true
         *                      but the library could not be loaded for some
         *                      reason.
         *  @since 1.0
         */
        Library(const String & fileName, bool load = true) throw(IOException);

        /**
         *  The class constructor.
         *
         *  The newly created Library instance is bound to the specified
         *  library file name. If the \a load parameter is true, an attempt
         *  is made to load the library; if this attempt succeeds then the
         *  library is in the Loaded state.
         *
         *  @param pathName     A path to the library.
         *  @param load         If true, the library is loaded; otherwise
         *                      the library is constructed in Bound state.
         *  @exception IOException  Thrown if the \a load parameter is true
         *                      but the library could not be loaded for some
         *                      reason.
         *  @since 1.0
         */
        Library(const PathName & pathName, bool load = true) throw(IOException);

        /**
         *  The class destructor.
         *
         *  If the library is in the Loaded state, unloads it.
         *
         *  @since 1.0
         */
        ~Library();

        //////////
        //  Library properties
    public:
        /**
         *  A pointer to a nul-terminated ASCII string containing the
         *  platform-specific filename extension used by dynamically
         *  loadable libraries.
         *
         *  The string always starts with a filename extension separator
         *  (e.g. ".dll" on Win32, ".so" on Unix, etc.)
         *
         *  @since 1.0
         */
        static const char * PreferredExtension;

        //////////
        //  Operations
    public:
        /**
         *  Returns the file name to which this library is bound.
         *
         *  If this Library instance is not bound to a file name,
         *  the returned value is an empty string.
         *
         *  @return             The file name to which this library is bound.
         *  @since 1.0
         */
        String              getFileName() const;

        /**
         *  Binds the Library instance to a specified file name.
         *
         *  If the library is already bound to a file name, the old
         *  binding is replaced with a new one. The binding of a
         *  library cannot be changed if it is already loaded.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing the file name to bind the library to.
         *  @since 1.0
         */
        void                setFileName(const char * pFileName);

        /**
         *  Binds the Library instance to a specified file name.
         *
         *  If the library is already bound to a file name, the old
         *  binding is replaced with a new one. The binding of a
         *  library cannot be changed if it is already loaded.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing the file name to bind the library to.
         *  @since 1.0
         */
        void                setFileName(const Char * pFileName);

        /**
         *  Binds the Library instance to a specified file name.
         *
         *  If the library is already bound to a file name, the old
         *  binding is replaced with a new one. The binding of a
         *  library cannot be changed if it is already loaded.
         *
         *  @param fileName     A string representing the file name 
         *                      to bind the library to.
         *  @since 1.0
         */
        void                setFileName(const String & fileName);

        /**
         *  Binds the Library instance to a specified file name.
         *
         *  If the library is already bound to a file name, the old
         *  binding is replaced with a new one. The binding of a
         *  library cannot be changed if it is already loaded.
         *
         *  @param pathName     A path name to bind the library to.
         *  @since 1.0
         */
        void                setFileName(const PathName & pathName);

        /**
         *  Checks if this library is bound.
         *
         *  @return             True if this library is bound to a 
         *                      specific file name; false otherwise.
         *  @since 1.0
         */
        bool                isBound() { return _fileName.length() != 0; }

        /**
         *  Checks if this library is loaded.
         *
         *  @return             True if this library is bound to a 
         *                      specific file name and loaded;
         *                      false otherwise.
         *  @since 1.0
         */
        bool                isLoaded() { return _isLoaded; }

        /**
         *  Loads the library.
         *
         *  The library must be bound to a file name for this
         *  call to succeed. An attempt to load a library that
         *  is already loaded is an error.
         *
         *  @returns            True if the library has been loaded
         *                      successfully; false otherwise.
         *  @since 1.0
         */
        bool                load();

        /**
         *  Binds and loads the library.
         *
         *  The library is bound to the specified file name and
         *  then an attempt is made to load it. An attempt to load 
         *  a library that is already loaded is an error.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing the file name to bind the library to.
         *  @returns            True if the library has been loaded
         *                      successfully; false otherwise.
         *  @since 1.0
         */
        bool                load(const char * pFileName);

        /**
         *  Binds and loads the library.
         *
         *  The library is bound to the specified file name and
         *  then an attempt is made to load it. An attempt to load 
         *  a library that is already loaded is an error.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing the file name to bind the library to.
         *  @returns            True if the library has been loaded
         *                      successfully; false otherwise.
         *  @since 1.0
         */
        bool                load(const Char * pFileName);

        /**
         *  Binds and loads the library.
         *
         *  The library is bound to the specified file name and
         *  then an attempt is made to load it. An attempt to load 
         *  a library that is already loaded is an error.
         *
         *  @param fileName     A string representing the file name to 
         *                      bind the library to.
         *  @returns            True if the library has been loaded
         *                      successfully; false otherwise.
         *  @since 1.0
         */
        bool                load(const String & fileName);

        /**
         *  Binds and loads the library.
         *
         *  The library is bound to the specified file name and
         *  then an attempt is made to load it. An attempt to load 
         *  a library that is already loaded is an error.
         *
         *  @param pathName     A path name to bind the library to.
         *  @returns            True if the library has been loaded
         *                      successfully; false otherwise.
         *  @since 1.0
         */
        bool                load(const PathName & pathName);

        /**
         *  Unloads the library.
         *
         *  An attempt to call this method on a Library instance that
         *  is not currently loaded is an error. Otherwise, the library
         *  is unloaded and the Library instance reverts to the Bound
         *  state.
         *
         *  @return             True if the library has been unloaded
         *                      successfully; false otherwise.
         *  @since 1.0
         */
        bool                unload();

        /**
         *  Locates a public symbol defined by the library.
         *
         *  An attempt to call this method on a Library instance that is
         *  not loaded is an error.
         *
         *  @param pSymbolName  A pointer to a nul-terminated ASCII string
         *                      representing the name of the global symbol
         *                      to search for.
         *  @return             An address of the global symbol or NULL if
         *                      this library does not define a global symbol
         *                      with the specified name.
         *  @since 1.0
         */
        void *              resolve(const char * pSymbolName);

        /**
         *  Locates a public symbol defined by the library.
         *
         *  An attempt to call this method on a Library instance that is
         *  not loaded is an error.
         *
         *  @param pSymbolName  A pointer to a nul-terminated ISO-10646 string
         *                      representing the name of the global symbol
         *                      to search for.
         *  @return             An address of the global symbol or NULL if
         *                      this library does not define a global symbol
         *                      with the specified name.
         *  @since 1.0
         */
        void *              resolve(const Char * pSymbolName);

        /**
         *  Locates a public symbol defined by the library.
         *
         *  An attempt to call this method on a Library instance that is
         *  not loaded is an error.
         *
         *  @param symbolName   A string representing the name of the global 
         *                      symbol to search for.
         *  @return             An address of the global symbol or NULL if
         *                      this library does not define a global symbol
         *                      with the specified name.
         *  @since 1.0
         */
        void *              resolve(const String & symbolName);

        //////////
        //  Implementation
    private:
        String              _fileName;
        bool                _isLoaded;
        void *              _pHandle;

        //  Helpers
        bool                _load(const String & fileName);
    };
}

#endif  //  __CARL_Library_hpp_included__
//  End of CARL_Library.hpp
