//
//  CARL_Locale.hpp - Locale API
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.14 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines locales.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Locale.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Locale_hpp_included__
#define __CARL_Locale_hpp_included__

namespace carl
{
    class CARL_PUBLIC CriticalSection;

    /**
     *  @class Locale CARL.hpp
     *  An encapsulated locale.
     *
     *  A locale is a (language, country, variant) collection where:
     *
     *  -   \a language is a 2-letter ISO-639 language code, or an
     *      empty string if the locale does not specify a language.
     *  -   \a country is a 2-letter ISO-3166 country code, or an
     *      empty string if the locale does not specify a country.
     *  -   \a variant is an arbitrary string specifying a locale variant,
     *      or an empty string if the locale does not specify a variant.
     *
     *  Locales do not by themselves provide means of localization, instead
     *  they serve as containers for descriptions of various cultures. All
     *  services that behave differently depending on the culture for which
     *  they are invoked accept locales as one of the arguments.
     *
     *  Depending in what locale properties are specified, locales
     *  are categorized as:
     *
     *  -   Invariant locale (none of the locale properties are specified).
     *  -   Neutral locale (only language is specified).
     *  -   Culture locale (language and country are specified).
     *  -   Full locale (language, country and variant are all specified).
     *
     *  Locales are logically organized into a hierarchy, where:
     *
     *  -   The hierarchy root is an invariant locale.
     *  -   A parent locale has one less defined property than each of
     *      its immediate children.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Locale
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructors.
         *
         *  The newly created locale is initialized to an invariant locale.
         *
         *  @since 1.0
         */
        Locale();

        /**
         *  The class constructor.
         *
         *  The newly created locale is initialized to a neutral locale
         *  for the specified language.
         *
         *  @param pLanguage        A pointer to a nul-terminated ASCII string
         *                          representing the 2-letter ISO-639 language
         *                          code.
         *  @since 1.0
         */
        explicit Locale(const char * pLanguage);

        /**
         *  The class constructor.
         *
         *  The newly created locale is initialized to a culture locale
         *  for the specified language and country.
         *
         *  @param pLanguage        A pointer to a nul-terminated ASCII string
         *                          representing the 2-letter ISO-639 language
         *                          code.
         *  @param pCountry         A pointer to a nul-terminated ASCII string
         *                          representing the 2-letter ISO-3166 country
         *                          code.
         *  @since 1.0
         */
        Locale(const char * pLanguage, const char * pCountry);

        /**
         *  The class constructor.
         *
         *  The newly created locale is initialized to a full locale
         *  for the specified language, country and variant.
         *
         *  @param pLanguage        A pointer to a nul-terminated ASCII string
         *                          representing the 2-letter ISO-639 language
         *                          code.
         *  @param pCountry         A pointer to a nul-terminated ASCII string
         *                          representing the 2-letter ISO-3166 country
         *                          code.
         *  @param pVariant         A pointer to a nul-terminated ASCII string
         *                          representing the locale variant.
         *  @since 1.0
         */
        Locale(const char * pLanguage, const char * pCountry, const char * pVariant);

        /**
         *  The class constructor.
         *
         *  The newly created locale is initialized to a neutral locale
         *  for the specified language.
         *
         *  @param pLanguage        A pointer to a nul-terminated ISO-10646 string
         *                          representing the 2-letter ISO-639 language
         *                          code.
         *  @since 1.0
         */
        explicit Locale(const Char * pLanguage);

        /**
         *  The class constructor.
         *
         *  The newly created locale is initialized to a culture locale
         *  for the specified language and country.
         *
         *  @param pLanguage        A pointer to a nul-terminated ISO-10646 string
         *                          representing the 2-letter ISO-639 language
         *                          code.
         *  @param pCountry         A pointer to a nul-terminated ISO-10646 string
         *                          representing the 2-letter ISO-3166 country
         *                          code.
         *  @since 1.0
         */
        Locale(const Char * pLanguage, const Char * pCountry);

        /**
         *  The class constructor.
         *
         *  The newly created locale is initialized to a full locale
         *  for the specified language, country and variant.
         *
         *  @param pLanguage        A pointer to a nul-terminated ISO-10646 string
         *                          representing the 2-letter ISO-639 language
         *                          code.
         *  @param pCountry         A pointer to a nul-terminated ISO-10646 string
         *                          representing the 2-letter ISO-3166 country
         *                          code.
         *  @param pVariant         A pointer to a nul-terminated ISO-10646 string
         *                          representing the locale variant.
         *  @since 1.0
         */
        Locale(const Char * pLanguage, const Char * pCountry, const Char * pVariant);

        /**
         *  The class constructor.
         *
         *  The newly created locale is initialized to a neutral locale
         *  for the specified language.
         *
         *  @param language         A string representing the 2-letter 
         *                          ISO-639 language code.
         *  @since 1.0
         */
        explicit Locale(const String & language);

        /**
         *  The class constructor.
         *
         *  The newly created locale is initialized to a culture locale
         *  for the specified language and country.
         *
         *  @param language         A string representing the 2-letter ISO-639
         *                          language code.
         *  @param country          A string representing the 2-letter ISO-3166 
         *                          country code.
         *  @since 1.0
         */
        Locale(const String & language, const String & country);

        /**
         *  The class constructor.
         *
         *  The newly created locale is initialized to a full locale
         *  for the specified language, country and variant.
         *
         *  @param language         A string representing the 2-letter ISO-639
         *                          language code.
         *  @param country          A string representing the 2-letter ISO-3166 
         *                          country code.
         *  @param variant          A string representing the locale variant.
         *  @since 1.0
         */
        Locale(const String & language, const String & country, const String & variant);

        /**
         *  The copy constructor.
         *
         *  @param src          The locale to copy properties from.
         *  @since 1.0
         */
        Locale(const Locale & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~Locale() {}

        /**
         *  The assignment operator.
         *
         *  @param src          The locale to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        Locale &                operator = (const Locale & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the system locale.
         *
         *  A system locale is determined by the current localization
         *  settings of the host operating system.
         *
         *  @return             The system locale.
         *  @since 1.0
         */
        static Locale           getSystem();

        /**
         *  Returns the current default locale.
         *
         *  When a CARL application starts, the current default locale
         *  is the same as the system locale. However, while system locale
         *  remains constant throughout application execution, the current
         *  default locale can be freely changed if necessary.
         *
         *  @return             The current default locale.
         *  @since 1.0
         */
        static Locale           getDefault();

        /**
         *  Sets the current default locale.
         *
         *  When a CARL application starts, the current default locale
         *  is the same as the system locale. However, while system locale
         *  remains constant throughout application execution, the current
         *  default locale can be freely changed if necessary.
         *
         *  @param locale       The locale to set up as the current default locale.
         *  @since 1.0
         */
        static void             setDefault(const Locale & locale);

        /**
         *  Checks if a locale is an invariant locale.
         *
         *  An invariant locale does not specify language, country
         *  or variant.
         *
         *  @return             True if this is an invariant locale;
         *                      false otherwise.
         *  @since 1.0
         */
        bool                    isInvariant() const { return _language[0] == 0 && _country[0] == 0 && _variant.length() == 0; }

        /**
         *  Checks if a locale is a neutral locale.
         *
         *  A neutral locale specifies language, but not country
         *  or variant.
         *
         *  @return             True if this is a neutral locale;
         *                      false otherwise.
         *  @since 1.0
         */
        bool                    isNeutral() const { return _language[0] != 0 && _country[0] == 0 && _variant.length() == 0; }

        /**
         *  Checks if a locale is a culture locale.
         *
         *  A neutral locale specifies language and country, but not
         *  variant.
         *
         *  @return             True if this is a culture locale;
         *                      false otherwise.
         *  @since 1.0
         */
        bool                    isCulture() const { return _language[0] != 0 && _country[0] != 0 && _variant.length() == 0; }

        /**
         *  Checks if a locale is a full locale.
         *
         *  A neutral locale specifies all three of language, country
         *  and variant.
         *
         *  @return             True if this is a full locale;
         *                      false otherwise.
         *  @since 1.0
         */
        bool                    isFull() const { return _language[0] != 0 && _country[0] != 0 && _variant.length() != 0; }

        /**
         *  Returns the language of this locale.
         *
         *  If the locale does not specify a language, the returned
         *  value is an empty string.
         *
         *  @return             The language of this locale.
         *  @since 1.0
         */
        String                  getLanguage() const { return (_language[0] == 0) ? String() : String(_language, 2); }

        /**
         *  Returns the country of this locale.
         *
         *  If the locale does not specify a country, the returned
         *  value is an empty string.
         *
         *  @return             The country of this locale.
         *  @since 1.0
         */
        String                  getCountry() const { return (_country[0] == 0) ? String() : String(_country, 2); }

        /**
         *  Returns the variant of this locale.
         *
         *  If the locale does not specify a variant, the returned
         *  value is an empty string.
         *
         *  @return             The variant of this locale.
         *  @since 1.0
         */
        String                  getVariant() const { return _variant; }

        /**
         *  Returns the display name of the language of this locale.
         *
         *  If the locale does not specify a language, the returned
         *  value is an empty string. Otherwise the display name of the
         *  locale's language is tailored for the current default locale.
         *
         *  @return             The display name of the language of this locale.
         *  @since 1.0
         */
        String                  getDisplayLanguage() const { return getDisplayLanguage(getDefault()); }

        /**
         *  Returns the display name of the country of this locale.
         *
         *  If the locale does not specify a country, the returned
         *  value is an empty string. Otherwise the display name of the
         *  locale's country is tailored for the current default locale.
         *
         *  @return             The display name of the country of this locale.
         *  @since 1.0
         */
        String                  getDisplayCountry() const { return getDisplayCountry(getDefault()); }

        /**
         *  Returns the display name of the variant of this locale.
         *
         *  If the locale does not specify a variant, the returned
         *  value is an empty string. Otherwise the display name of the
         *  locale's variant is tailored for the current default locale.
         *
         *  @return             The display name of the variant of this locale.
         *  @since 1.0
         */
        String                  getDisplayVariant() const { return getDisplayVariant(getDefault()); }

        /**
         *  Returns the display name of this locale.
         *
         *  If the locale does not specify a language, the returned
         *  value is an empty string. Otherwise the display name of the
         *  locale is tailored for the current default locale.
         *
         *  Depending on what properties of a locale are specified, the
         *  display name of the locale has one of the following forms:
         *
         *  -   &lt;invariant locale name&gt; if the locale is an invariant locale.
         *  -   &lt;language> if the locale is a neutral locale.
         *  -   &lt;language>(&lt;country&gt;) if the locale is a culture locale.
         *  -   &lt;language>(&lt;country&gt;) &lt;variant&gt; if the locale is a full locale.
         *
         *  @return             The display name of this locale.
         *  @since 1.0
         */
        String                  getDisplayName() const { return getDisplayName(getDefault()); }

        /**
         *  Returns the display name of the language of this locale.
         *
         *  If the locale does not specify a language, the returned
         *  value is an empty string. Otherwise the display name of the
         *  locale's language is tailored for the specified locale.
         *
         *  @param forLocale    The locale for which the display name should
         *                      be tailored.
         *  @return             The display name of the language of this locale.
         *  @since 1.0
         */
        String                  getDisplayLanguage(const Locale & forLocale) const;

        /**
         *  Returns the display name of the country of this locale.
         *
         *  If the locale does not specify a country, the returned
         *  value is an empty string. Otherwise the display name of the
         *  locale's country is tailored for the specified locale.
         *
         *  @param forLocale    The locale for which the display name should
         *                      be tailored.
         *  @return             The display name of the country of this locale.
         *  @since 1.0
         */
        String                  getDisplayCountry(const Locale & forLocale) const;

        /**
         *  Returns the display name of the variant of this locale.
         *
         *  If the locale does not specify a variant, the returned
         *  value is an empty string. Otherwise the display name of the
         *  locale's variant is tailored for the specified locale.
         *
         *  @param forLocale    The locale for which the display name should
         *                      be tailored.
         *  @return             The display name of the variant of this locale.
         *  @since 1.0
         */
        String                  getDisplayVariant(const Locale & /*forLocale*/) const { return _variant; }

        /**
         *  Returns the display name of this locale.
         *
         *  If the locale does not specify a language, the returned
         *  value is an empty string. Otherwise the display name of the
         *  locale is tailored for the specified locale.
         *
         *  Depending on what properties of a locale are specified, the
         *  display name of the locale has one of the following forms:
         *
         *  -   &lt;invariant locale name&gt; if the locale is an invariant locale.
         *  -   &lt;language> if the locale is a neutral locale.
         *  -   &lt;language>(&lt;country&gt;) if the locale is a culture locale.
         *  -   &lt;language>(&lt;country&gt;) &lt;variant&gt; if the locale is a full locale.
         *
         *  @param forLocale    The locale for which the display name should
         *                      be tailored.
         *  @return             The display name of this locale.
         *  @since 1.0
         */
        String                  getDisplayName(const Locale & forLocale) const;

        /**
         *  Returns an immediate parent locale of this locale.
         *
         *  The following rules are used to determine parent locales:
         *
         *  -   A parent of a full locale is a culture locale with the same
         *      language and country as the original locale.
         *  -   A parent of a culture locale is a neutral locale with the same
         *      language as the original locale.
         *  -   A parent of a neutral or invariant locale is an invariant locale.
         *
         *  @return             An immediate parent locale of this locale.
         *  @since 1.0
         */
        Locale                  getParent() const;

        /**
         *  Checks if two locales are equal.
         *
         *  Two locales are equal iff all their properties are equal.
         *
         *  @param locale       The locale to compare this locale with.
         *  @return             True if both locales represent the same
         *                      culture; false otherwise.
         *  @since 1.0
         */
        bool                    operator == (const Locale & locale) const;

        /**
         *  Checks if two locales are different.
         *
         *  Two locales are different iff at least one of their properties is different.
         *
         *  @param locale       The locale to compare this locale with.
         *  @return             True if both locales represent different
         *                      cultures; false otherwise.
         *  @since 1.0
         */
        bool                    operator != (const Locale & locale) const;

        /**
         *  Returns a string representing this locale.
         *
         *  The returned string is formatted according to the following rules:
         *
         *  -   &lt;empty string&gt; for an invariant locale.
         *  -   &lt;language&gt; for a neutral locale, where &lt;language&gt; is a
         *      lowercase 2-letter ISO-639 language code.
         *  -   &lt;language&gt;_&lt;country&gt; for a culture locale, where &lt;language&gt;
         *      is a lowercase 2-letter ISO-639 language code and &lt;country&gt;
         *      is an uppercase 2-letter ISO-3166 country code.
         *  -   &lt;language&gt;_&lt;country&gt;_&lt;variant&gt; for a full locale, where &lt;language&gt;
         *      is a lowercase 2-letter ISO-639 language code, &lt;country&gt;
         *      is an uppercase 2-letter ISO-3166 country code and &lt;variant&gt;
         *      is a variant string.
         *
         *  @return             The string representing this locale.
         *  @since 1.0
         */
        String                  toString() const;

        /**
         *  Returns a locale represented by the specified string.
         *
         *  The locale specification string can be in several of the
         *  following forms:
         *
         *  -   &lt;empty string&gt; for an invariant locale.
         *  -   &lt;language&gt; for a neutral locale, where &lt;language&gt; is a
         *      case-insensitive 2-letter ISO-639 language code.
         *  -   &lt;language&gt;_&lt;country&gt; for a culture locale, where &lt;language&gt;
         *      is a lowercase 2-letter ISO-639 language code and &lt;country&gt;
         *      is a case-insensitive 2-letter ISO-3166 country code.
         *  -   &lt;language&gt;_&lt;country&gt;_&lt;variant&gt; for a full locale, where &lt;language&gt;
         *      is a case-insensitive 2-letter ISO-639 language code, &lt;country&gt;
         *      is acase-insensitive 2-letter ISO-3166 country code and &lt;variant&gt;
         *      is a variant string.
         *
         *  If the string does not obey the well-formedness rules defined above,
         *  the current deault locale is returned.
         *
         *  @param pLocaleString    The pointer to a nul-terminated
         *                      ASCII locale specification string.
         *  @return             The locale defined by \a localeString.
         *  @since 1.0
         */
        static Locale           fromString(const char * pLocaleString);

        /**
         *  Returns a locale represented by the specified string.
         *
         *  The locale specification string can be in several of the
         *  following forms:
         *
         *  -   &lt;empty string&gt; for an invariant locale.
         *  -   &lt;language&gt; for a neutral locale, where &lt;language&gt; is a
         *      case-insensitive 2-letter ISO-639 language code.
         *  -   &lt;language&gt;_&lt;country&gt; for a culture locale, where &lt;language&gt;
         *      is a lowercase 2-letter ISO-639 language code and &lt;country&gt;
         *      is a case-insensitive 2-letter ISO-3166 country code.
         *  -   &lt;language&gt;_&lt;country&gt;_&lt;variant&gt; for a full locale, where &lt;language&gt;
         *      is a case-insensitive 2-letter ISO-639 language code, &lt;country&gt;
         *      is acase-insensitive 2-letter ISO-3166 country code and &lt;variant&gt;
         *      is a variant string.
         *
         *  If the string does not obey the well-formedness rules defined above,
         *  the current deault locale is returned.
         *
         *  @param pLocaleString    The pointer to a nul-terminated
         *                      ISO-10646 locale specification string.
         *  @return             The locale defined by \a localeString.
         *  @since 1.0
         */
        static Locale           fromString(const Char * pLocaleString);

        /**
         *  Returns a locale represented by the specified string.
         *
         *  The locale specification string can be in several of the
         *  following forms:
         *
         *  -   &lt;empty string&gt; for an invariant locale.
         *  -   &lt;language&gt; for a neutral locale, where &lt;language&gt; is a
         *      case-insensitive 2-letter ISO-639 language code.
         *  -   &lt;language&gt;_&lt;country&gt; for a culture locale, where &lt;language&gt;
         *      is a lowercase 2-letter ISO-639 language code and &lt;country&gt;
         *      is a case-insensitive 2-letter ISO-3166 country code.
         *  -   &lt;language&gt;_&lt;country&gt;_&lt;variant&gt; for a full locale, where &lt;language&gt;
         *      is a case-insensitive 2-letter ISO-639 language code, &lt;country&gt;
         *      is acase-insensitive 2-letter ISO-3166 country code and &lt;variant&gt;
         *      is a variant string.
         *
         *  If the string does not obey the well-formedness rules defined above,
         *  the current deault locale is returned.
         *
         *  @param localeString The locale specification string.
         *  @return             The locale defined by \a localeString.
         *  @since 1.0
         */
        static Locale           fromString(const String & localeString);

        //////////
        //  Implementation
    private:
        Char                    _language[2];   //  {0, 0} is undefined; otherwise ISO-639 code2
        Char                    _country[2];    //  {0, 0} is undefined; otherwise ISO-3166 code2
        String                  _variant;
        static CriticalSection *_getStaticGuard();
    };

    //////////
    //  Hashing locales
    namespace StandardHashFunctions
    {
        /**
         *  Calculates a hash code for a Locale instance.
         *
         *  @param key              The Locale instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashLocale(const Locale & key);
    }
}

#endif  //  __CARL_Locale_hpp_included__
//  End of CARL_Locale.hpp
