//
//  CARL_Map.hpp - Key -> value maps.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.12 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines associative maps.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Map.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Map_hpp_included__
#define __CARL_Map_hpp_included__

namespace carl
{
    /**
     *  @class Map CARL.hpp
     *  A generic key -&gt; value map, where each key can be mapped to only one value.
     *
     *  CARL maps use implicit sharing to reduce the copy construction and 
     *  assignment overhead. This means that it is feasible to both pass
     *  maps to functions as value parameters and return maps from functions.
     *
     *   The key type (K) must be:
     *
     *   -   Default constructable (K::K() must be defined and public)
     *   -   Copy constructable (K::K(const K &) must be defined and public)
     *   -   Assignable (K::operator=(const K &) must be defined and public)
     *   -   Destructable (K::~K() must be defined and public)
     *
     *   The value type (V) must be:
     *
     *   -   Default constructable (V::V() must be defined and public)
     *   -   Copy constructable (V::V(const V &) must be defined and public)
     *   -   Assignable (V::operator=(const V &) must be defined and public)
     *   -   Destructable (V::~V() must be defined and public)
     *
     *  @param K                The type of map keys.
     *  @param V                The type of values associated with map keys.
     *  @since 1.0
     */
    template<class K, class V>
    class Map
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        Map();

        /**
         *  The copy constructor.
         *
         *  @since 1.0
         */
        Map(const Map<K,V> & );

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Map();

        /**
         *  The assignment operator.
         *
         *  @return             A reference to the assigned-to map.
         *  @since 1.0
         */
        Map<K,V> &              operator = (const Map<K,V> & );

        //////////
        //  Operations
    public:
        /**
         *  Checks if a map is empty.
         *
         *  @return             True if this map is empty; false otherwise.
         *  @since 1.0
         */
        virtual bool            isEmpty() const = 0;

        /**
         *  Returns the number of (key, value) pairs in this map.
         *
         *  @return             The number of (key, value) pairs in this map.
         *  @since 1.0
         */
        virtual UIntSize        size() const = 0;

        /**
         *  Checks if the map contains the specified key.
         *
         *  @param key          The key to check for existense in the map.
         *  @return             True if the map contains the specified key;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            containsKey(const K & key) const = 0;

        /**
         *  Returns the value associated with the specified key.
         *
         *  @param key          The key to look for.
         *  @return             The value associated with the specified key.
         *  @exception NoSuchElementError   If the map does not contain the
         *                      specified key.
         *  @since 1.0
         */
        virtual V               get(const K & key) const = 0;

        /**
         *  Inserts a (key, value) pair into the map.
         *
         *  If the map already contains the specified key, its old
         *  associated value is replaced with a new one.
         *
         *  @param key          The key.
         *  @param value        The value to associate with the key.
         *  @return             True if the (key, value) pair was actually
         *                      inserted into the map; false if the key
         *                      was already present in the map and its
         *                      associated value was replaced.
         *  @since 1.0
         */
        virtual bool            insert(const K & key, const V & value) = 0;

        /**
         *  Injects a map into this map.
         *
         *  All (key, value) pairs contained in the specified map
         *  are injected into this map.
         *
         *  @param map          The map to inject from.
         *  @return             The total number of new entries that were
         *                      created by injection.
         *  @since 1.0
         */
        virtual UIntSize        insert(const Map<K,V> & map) = 0;

        /**
         *  Removed a key from this map.
         *
         *  @param key          The key to remove.
         *  @return             True if the specified key actually existed
         *                      in the map and was removed; false if the
         *                      specified key did not exist in the map
         *                      before the call.
         *  @since 1.0
         */
        virtual bool            removeKey(const K & key) = 0;

        /**
         *  Clears out the map.
         *
         *  @since 1.0
         */
        virtual void            removeAll() = 0;

        /**
         *  Returns an iterator over a set of all keys currently in the map.
         *
         *  Keys are not guaranteed to be iterated over in any
         *  particular order.
         *
         *  @return             An iterator over a set of all keys.
         *  @since 1.0
         */
        virtual Iterator<K>     enumerateKeys() const = 0;

        /**
         *  Returns an iterator over a set of all (key, value) pairs
         *  currently in the map.
         *
         *  The (key, value) pairs are not guaranteed to be iterated over
         *  in any particular order.
         *
         *  @return             An iterator over a set of all (key, value) pairs.
         *  @since 1.0
         */
        virtual Iterator< Pair<K, V> >  enumerateElements() const = 0;
    };

    template<class K, class V>
    Map<K,V>::Map() 
    {
    }

    template<class K, class V>
    Map<K,V>::Map(const Map<K,V> & ) 
    {
    }

    template<class K, class V>
    Map<K,V>::~Map() 
    {
    }

    template<class K, class V>
    Map<K,V> & Map<K,V>::operator = (const Map<K,V> & ) 
    { 
        return *this; 
    }

    namespace _impl
    {
        //////////
        //  A key -> value map implementation via a hash table.
        //
        //  The key type (K) must be:
        //  *   Default constructable (K::K() must be defined and public)
        //  *   Copy constructable (K::K(const K &) must be defined and public)
        //  *   Assignable (K::operator=(const K &) must be defined and public)
        //  *   Destructable (K::~K() must be defined and public)
        //  The value type (V) must be:
        //  *   Default constructable (V::V() must be defined and public)
        //  *   Copy constructable (V::V(const V &) must be defined and public)
        //  *   Assignable (V::operator=(const V &) must be defined and public)
        //  *   Destructable (V::~V() must be defined and public)
        template<class K, class V, class C>
        class _HashMapImpl
        {
            //////////
            //  Types
        public:
            typedef UIntSize (CARL_CALLBACK * HashFunction)(const K & key);

            //////////
            //  Can't assign
        private:
            void                    operator = (const _HashMapImpl<K,V,C> &);

            //////////
            //  Construction/destruction
        public:
            _HashMapImpl(HashFunction hashFunction, C comparator);
            _HashMapImpl(HashFunction hashFunction, UIntSize initialHashTableCapacity, C comparator);
            _HashMapImpl(const _HashMapImpl<K,V,C> & src);
            ~_HashMapImpl();

            //////////
            //  Operations
        public:
            bool                    isEmpty() const;
            UIntSize                size() const;
            bool                    containsKey(const K & key) const;
            V                       get(const K & key) const;
            bool                    insert(const K & key, const V & value);
            UIntSize                insert(const Map<K,V> & map);
            bool                    removeKey(const K & key);
            void                    removeAll();
            Iterator< Pair<K, V> >  enumerateElements() const;
            Iterator<K>             enumerateKeys() const;
            UIntSize                getReferenceCount() const;
            void                    incrementReferenceCount();
            void                    decrementReferenceCount();

            //////////
            //  Implementation
        private:
            struct _HashTableEntry
            {
                _HashTableEntry(const K & key, const V & value)
                    :   _key(key),
                        _value(value)
                {
                    _pNextInChain = NULL;
                }

                K                   _key;
                V                   _value;
                _HashTableEntry *   _pNextInChain;
            };
            UIntSize                _hashTableSize;
            _HashTableEntry **      _hashTable;
            HashFunction            _hashFunction;
            C                       _comparator;
            UIntSize                _elementCount;
            UIntSize                _referenceCount;

            UIntSize                _hash(const K & key, UIntSize hashTableSize) const;
        };

        template<class K, class V, class C>
        _HashMapImpl<K,V,C>::_HashMapImpl(HashFunction hashFunction, C comparator)
            :   _comparator(comparator)
        {
            require(hashFunction != NULL);

            _hashTableSize = 0;
            _hashTable = NULL;
            _hashFunction = hashFunction;
            _elementCount = 0;
            _referenceCount = 0;
        }

        template<class K, class V, class C>
        _HashMapImpl<K,V,C>::_HashMapImpl(HashFunction hashFunction, UIntSize initialHashTableCapacity, C comparator)
            :   _comparator(comparator)
        {
            _hashTableSize = initialHashTableCapacity;
            if (initialHashTableCapacity == 0)
            {
                _hashTable = NULL;
            }
            else
            {
                _hashTable = new _HashTableEntry*[initialHashTableCapacity];
                for (UIntSize i = 0; i < initialHashTableCapacity; i++)
                {
                    _hashTable[i] = NULL;
                }
            }
            _hashFunction = hashFunction;
            _elementCount = 0;
            _referenceCount = 0;
        }

        template<class K, class V, class C>
        _HashMapImpl<K,V,C>::_HashMapImpl(const _HashMapImpl<K,V,C> & src)
            :   _comparator(src._comparator)
        {
            _hashTableSize = src.size() / 3 * 4;
            if (_hashTableSize == 0)
            {
                _hashTable = NULL;
            }
            else
            {
                _hashTable = new _HashTableEntry*[_hashTableSize];
                for (UIntSize i = 0; i < _hashTableSize; i++)
                {
                    _hashTable[i] = NULL;
                }
            }
            _hashFunction = src._hashFunction;
            _elementCount = 0;
            _referenceCount = 0;
            for (Iterator< Pair<K, V> > it = src.enumerateElements(); 
                !it.finished(); 
                it.advance())
            {
                insert(it.item().first(), it.item().second());
            }
        }

        template<class K, class V, class C>
        _HashMapImpl<K,V,C>::~_HashMapImpl()
        {
            if (_hashTable != NULL)
            {
                for (UIntSize i = 0; i < _hashTableSize; i++)
                {
                    while (_hashTable[i] != NULL)
                    {
                        _HashTableEntry * pHashTableEntry = _hashTable[i];
                        _hashTable[i] = pHashTableEntry->_pNextInChain;
                        delete pHashTableEntry;
                    }
                }
                delete [] _hashTable;
            }
        }

        template<class K, class V, class C>
        bool _HashMapImpl<K,V,C>::isEmpty() const
        {
            return _elementCount == 0;
        }

        template<class K, class V, class C>
        UIntSize _HashMapImpl<K,V,C>::size() const
        {
            return _elementCount;
        }

        template<class K, class V, class C>
        bool _HashMapImpl<K,V,C>::containsKey(const K & key) const
        {
            if (_elementCount == 0)
            {
                return false;
            }
            else
            {
                UIntSize hashCode = _hash(key, _hashTableSize);
                for (_HashTableEntry * pHashTableEntry = _hashTable[hashCode];
                     pHashTableEntry != NULL;
                     pHashTableEntry = pHashTableEntry->_pNextInChain)
                {
                    if (_comparator.equal(pHashTableEntry->_key, key))
                    {
                        return true;
                    }
                }
                return false;
            }
        }

        template<class K, class V, class C>
        V _HashMapImpl<K,V,C>::get(const K & key) const
        {
            if (_hashTable != NULL)
            {
                UIntSize hashCode = _hash(key, _hashTableSize);
                for (_HashTableEntry * pHashTableEntry = _hashTable[hashCode];
                    pHashTableEntry != NULL;
                    pHashTableEntry = pHashTableEntry->_pNextInChain)
                {
                    if (_comparator.equal(pHashTableEntry->_key, key))
                    {
                        return pHashTableEntry->_value;
                    }
                }
                throw NoSuchElementError();
            }
            else
            {
                throw NoSuchElementError();
            }
        }

        template<class K, class V, class C>
        bool _HashMapImpl<K,V,C>::insert(const K & key, const V & value)
        {
            //  If the key is already in the hash table, just replace the associated value
            if (_hashTable != NULL)
            {
                UIntSize hashCode = _hash(key, _hashTableSize);
                for (_HashTableEntry * pHashTableEntry = _hashTable[hashCode];
                    pHashTableEntry != NULL;
                    pHashTableEntry = pHashTableEntry->_pNextInChain)
                {
                    if (_comparator.equal(pHashTableEntry->_key, key))
                    {
                        pHashTableEntry->_value = value;
                        return false;
                    }
                }
            }

            //  The key is not currently in the hash table
            if (_hashTable == NULL ||
                _elementCount * 4 > _hashTableSize * 3)
            {   //  Hash table is either not allocated at all, or
                //  is 75% full - need a larger one.
                //  Allocate & initialize a new hash table...
                UIntSize newHashTableSize = (3 * _hashTableSize / 2) + 1;
                _HashTableEntry ** newHashTable = new _HashTableEntry*[newHashTableSize];
                for (UIntSize i = 0; i < newHashTableSize; i++)
                {
                    newHashTable[i] = NULL;
                }
                //  Copy data from old to new hash table
                if (_hashTable != NULL)
                {
                    for (UIntSize i = 0; i < _hashTableSize; i++)
                    {
                        while (_hashTable[i] != NULL)
                        {
                            //  Move one entry from old to new hash table
                            _HashTableEntry * pHashTableEntry = _hashTable[i];
                            _hashTable[i] = pHashTableEntry->_pNextInChain;
                            UIntSize newHashCode = _hash(pHashTableEntry->_key, newHashTableSize);
                            pHashTableEntry->_pNextInChain = newHashTable[newHashCode];
                            newHashTable[newHashCode] = pHashTableEntry;
                        }
                    }
                    delete [] _hashTable;
                }
                //  Finish replacing the hash table
                _hashTable = newHashTable;
                _hashTableSize = newHashTableSize;
            }

            //  Insert the new element into the hash table
            UIntSize hashCode = _hash(key, _hashTableSize);
            _HashTableEntry * pNewHashTableEntry = new _HashTableEntry(key, value);
            pNewHashTableEntry->_pNextInChain = _hashTable[hashCode];
            _hashTable[hashCode] = pNewHashTableEntry;
            _elementCount++;
            return true;
        }

        template<class K, class V, class C>
        UIntSize _HashMapImpl<K,V,C>::insert(const Map<K,V> & map)
        {
            UIntSize insertedItemCount = 0;
            for (Iterator< Pair<K, V> > it = map.enumerateElements(); 
                 !it.finished(); 
                 it.advance())
            {
                if (insert(it.item().first(), it.item().second()))
                {
                    insertedItemCount++;
                }
            }
            return insertedItemCount;
        }

        template<class K, class V, class C>
        bool _HashMapImpl<K,V,C>::removeKey(const K & key)
        {
            if (_hashTable != NULL)
            {
                UIntSize hashCode = _hash(key, _hashTableSize);
                for (_HashTableEntry ** ppHashTableEntry = &(_hashTable[hashCode]);
                    (*ppHashTableEntry) != NULL;
                    ppHashTableEntry = &((*ppHashTableEntry)->_pNextInChain))
                {
                    if (_comparator.equal((*ppHashTableEntry)->_key, key))
                    {
                        _HashTableEntry * pHashTableEntry = *ppHashTableEntry;
                        *ppHashTableEntry = pHashTableEntry->_pNextInChain;
                        delete pHashTableEntry;
                        _elementCount--;
                        return true;
                    }
                }
                return false;
            }
            else
            {
                return false;
            }
        }

        template<class K, class V, class C>
        void _HashMapImpl<K,V,C>::removeAll()
        {
            if (_hashTable != NULL)
            {
                for (UIntSize i = 0; i < _hashTableSize; i++)
                {
                    while (_hashTable[i] != NULL)
                    {
                        _HashTableEntry * pHashTableEntry = _hashTable[i];
                        _hashTable[i] = pHashTableEntry->_pNextInChain;
                        delete pHashTableEntry;
                    }
                }
                _elementCount = 0;
            }
        }

        template<class K, class V, class C>
        Iterator< Pair<K, V> > _HashMapImpl<K,V,C>::enumerateElements() const
        {
            Pair<K, V> * pElements = new Pair<K, V>[_elementCount];
            UIntSize elementIndex = 0;
            for (UIntSize i = 0; i < _hashTableSize; i++)
            {
                for (_HashTableEntry * pHashTableEntry = _hashTable[i];
                     pHashTableEntry != NULL;
                     pHashTableEntry = pHashTableEntry->_pNextInChain)
                {
                    pElements[elementIndex++] = Pair<K, V>(pHashTableEntry->_key, pHashTableEntry->_value);
                }
            }
            require(elementIndex == _elementCount);
            return Iterator< Pair<K, V> >(new _impl::_SnapshotIteratorImpl< Pair<K, V> >(_elementCount, pElements));
        }

        template<class K, class V, class C>
        Iterator<K> _HashMapImpl<K,V,C>::enumerateKeys() const
        {
            K * pElements = new K[_elementCount];
            UIntSize elementIndex = 0;
            for (UIntSize i = 0; i < _hashTableSize; i++)
            {
                for (_HashTableEntry * pHashTableEntry = _hashTable[i];
                     pHashTableEntry != NULL;
                     pHashTableEntry = pHashTableEntry->_pNextInChain)
                {
                    pElements[elementIndex++] = pHashTableEntry->_key;
                }
            }
            require(elementIndex == _elementCount);
            return Iterator<K>(new _impl::_SnapshotIteratorImpl<K>(_elementCount, pElements));
        }

        template<class K, class V, class C>
        UIntSize _HashMapImpl<K,V,C>::getReferenceCount() const
        {
            return _referenceCount;
        }

        template<class K, class V, class C>
        void _HashMapImpl<K,V,C>::incrementReferenceCount()
        {
            _referenceCount++;
            require(_referenceCount != 0);
        }

        template<class K, class V, class C>
        void _HashMapImpl<K,V,C>::decrementReferenceCount()
        {
            require(_referenceCount != 0);
            _referenceCount--;
        }

        template<class K, class V, class C>
        UIntSize _HashMapImpl<K,V,C>::_hash(const K & key, UIntSize hashTableSize) const
        {
            return _hashFunction(key) % hashTableSize;
        }
    }

    /**
     *  @class HashMap CARL.hpp
     *  A key -> value map implemented via a hash table.
     *
     *  The key type (K) must be:
     *
     *  -   Default constructable (K::K() must be defined and public)
     *  -   Copy constructable (K::K(const K &) must be defined and public)
     *  -   Assignable (K::operator=(const K &) must be defined and public)
     *  -   Destructable (K::~K() must be defined and public)
     *
     *  The value type (V) must be:
     *
     *  -   Default constructable (V::V() must be defined and public)
     *  -   Copy constructable (V::V(const V &) must be defined and public)
     *  -   Assignable (V::operator=(const V &) must be defined and public)
     *  -   Destructable (V::~V() must be defined and public)
     *
     *  @param K                The type of map keys.
     *  @param V                The type of values associated with map keys.
     *  @param C                The comparator to use for deciding whether
     *                          two keys of type \a K are equal or not.
     *  @since 1.0
     */
    template<class K, class V, class C = NaturalComparator<K> >
    class HashMap : public Map<K,V>
    {
        //////////
        //  Types
    public:
        /**
         *  The callback function for calculating hash codes for keys.
         *
         *  @param key          The key to calculate a hash code for.
         *  @return             The calculated hash code for the \a key.
         *  @since 1.0
         */
        typedef UIntSize (CARL_CALLBACK * HashFunction)(const K & key);

        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  The newly constructed hash map is associated with the 
         *  specified hash function and comparator.
         *
         *  @param hashFunction The callback function to use for calculating
         *                      hash codes for map keys.
         *  @param comparator   The comparator to use for deciding when two
         *                      keys are equal.
         *  @since 1.0
         */
        HashMap(HashFunction hashFunction, C comparator = C());

        /**
         *  The class constructor.
         *
         *  The newly constructed hash map is associated with the 
         *  specified hash function and comparator.
         *
         *  @param hashFunction The callback function to use for calculating
         *                      hash codes for map keys.
         *  @param initialHashTableCapacity The initial hash table capacity.
         *                      The hash table will not need to grow until
         *                      the hash map contains at least this number
         *                      of (key, value) pairs.
         *  @param comparator   The comparator to use for deciding when two
         *                      keys are equal.
         *  @since 1.0
         */
        HashMap(HashFunction hashFunction, UIntSize initialHashTableCapacity, C comparator = C());

        /**
         *  The copy constructor.
         *
         *  @param src          The hash map to copy contents from.
         *  @since 1.0
         */
        HashMap(const HashMap<K,V,C> & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~HashMap();

        /**
         *  The assignment operator.
         *
         *  @param src          The hash map to copy contents from.
         *  @return             A reference to the assigned-to map.
         *  @since 1.0
         */
        HashMap<K,V,C> &        operator = (const HashMap<K,V,C> & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a map is empty.
         *
         *  @return             True if this map is empty; false otherwise.
         *  @since 1.0
         */
        virtual bool            isEmpty() const;

        /**
         *  Returns the number of (key, value) pairs in this map.
         *
         *  @return             The number of (key, value) pairs in this map.
         *  @since 1.0
         */
        virtual UIntSize        size() const;

        /**
         *  Checks if the map contains the specified key.
         *
         *  @param key          The key to check for existense in the map.
         *  @return             True if the map contains the specified key;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            containsKey(const K & key) const;

        /**
         *  Returns the value associated with the specified key.
         *
         *  @param key          The key to look for.
         *  @return             The value associated with the specified key.
         *  @exception NoSuchElementError   If the map does not contain the
         *                      specified key.
         *  @since 1.0
         */
        virtual V               get(const K & key) const;

        /**
         *  Inserts a (key, value) pair into the map.
         *
         *  If the map already contains the specified key, its old
         *  associated value is replaced with a new one.
         *
         *  @param key          The key.
         *  @param value        The value to associate with the key.
         *  @return             True if the (key, value) pair was actually
         *                      inserted into the map; false if the key
         *                      was already present in the map and its
         *                      associated value was replaced.
         *  @since 1.0
         */
        virtual bool            insert(const K & key, const V & value);

        /**
         *  Injects a map into this map.
         *
         *  All (key, value) pairs contained in the specified map
         *  are injected into this map.
         *
         *  @param map          The map to inject from.
         *  @return             The total number of new entries that were
         *                      created by injection.
         *  @since 1.0
         */
        virtual UIntSize        insert(const Map<K,V> & map);

        /**
         *  Removed a key from this map.
         *
         *  @param key          The key to remove.
         *  @return             True if the specified key actually existed
         *                      in the map and was removed; false if the
         *                      specified key did not exist in the map
         *                      before the call.
         *  @since 1.0
         */
        virtual bool            removeKey(const K & key);

        /**
         *  Clears out the map.
         *
         *  @since 1.0
         */
        virtual void            removeAll();

        /**
         *  Returns an iterator over a set of all keys currently in the map.
         *
         *  Keys are not guaranteed to be iterated over in any
         *  particular order.
         *
         *  @return             An iterator over a set of all keys.
         *  @since 1.0
         */
        virtual Iterator<K>     enumerateKeys() const;

        /**
         *  Returns an iterator over a set of all (key, value) pairs
         *  currently in the map.
         *
         *  The (key, value) pairs are not guaranteed to be iterated over
         *  in any particular order.
         *
         *  @return             An iterator over a set of all (key, value) pairs.
         *  @since 1.0
         */
        virtual Iterator< Pair<K, V> >  enumerateElements() const;

        //////////
        //  Implementation
    private:
        _impl::_HashMapImpl<K,V,C> *    _pImpl;
        void                    _ensureNonsharedImplementation();
    };

    template<class K, class V, class C>
    HashMap<K,V,C>::HashMap(HashFunction hashFunction, C comparator /*= C()*/)
    {
        _pImpl = new _impl::_HashMapImpl<K,V,C>(hashFunction, comparator);
        _pImpl->incrementReferenceCount();
    }

    template<class K, class V, class C>
    HashMap<K,V,C>::HashMap(HashFunction hashFunction, UIntSize initialHashTableCapacity, C comparator /*= C()*/)
    {
        _pImpl = new _impl::_HashMapImpl<K,V,C>(hashFunction, initialHashTableCapacity, comparator);
        _pImpl->incrementReferenceCount();
    }

    template<class K, class V, class C>
    HashMap<K,V,C>::HashMap(const HashMap<K,V,C> & src)
    {
        _pImpl = src._pImpl;
        _pImpl->incrementReferenceCount();
    }

    template<class K, class V, class C>
    HashMap<K,V,C>::~HashMap()
    {
        _pImpl->decrementReferenceCount();
        if (_pImpl->getReferenceCount() == 0)
        {
            delete _pImpl;
        }
    }

    template<class K, class V, class C>
    HashMap<K,V,C> & HashMap<K,V,C>::operator = (const HashMap<K,V,C> & src)
    { 
        if (this != (&src))
        {
            _pImpl->decrementReferenceCount();
            if (_pImpl->getReferenceCount() == 0)
            {
                delete _pImpl;
            }
            _pImpl = src._pImpl;
            _pImpl->incrementReferenceCount();
        }
        return *this;
    }

    template<class K, class V, class C>
    bool HashMap<K,V,C>::isEmpty() const
    {
        return _pImpl->isEmpty();
    }

    template<class K, class V, class C>
    UIntSize HashMap<K,V,C>::size() const
    {
        return _pImpl->size();
    }

    template<class K, class V, class C>
    bool HashMap<K,V,C>::containsKey(const K & key) const
    {
        return _pImpl->containsKey(key);
    }

    template<class K, class V, class C>
    V HashMap<K,V,C>::get(const K & key) const
    {
        return _pImpl->get(key);
    }

    template<class K, class V, class C>
    bool HashMap<K,V,C>::insert(const K & key, const V & value)
    {
        _ensureNonsharedImplementation();
        return _pImpl->insert(key, value);
    }

    template<class K, class V, class C>
    UIntSize HashMap<K,V,C>::insert(const Map<K,V> & map)
    {
        _ensureNonsharedImplementation();
        return _pImpl->insert(map);
    }

    template<class K, class V, class C>
    bool HashMap<K,V,C>::removeKey(const K & key)
    {
        _ensureNonsharedImplementation();
        return _pImpl->removeKey(key);
    }

    template<class K, class V, class C>
    void HashMap<K,V,C>::removeAll()
    {
        _ensureNonsharedImplementation();
        return _pImpl->removeAll();
    }

    template<class K, class V, class C>
    Iterator<K> HashMap<K,V,C>::enumerateKeys() const
    {
        return _pImpl->enumerateKeys();
    }

    template<class K, class V, class C>
    Iterator< Pair<K, V> > HashMap<K,V,C>::enumerateElements() const
    {
        return _pImpl->enumerateElements();
    }

    template<class K, class V, class C>
    void HashMap<K,V,C>::_ensureNonsharedImplementation()
    {
        if (_pImpl->getReferenceCount() != 1)
        {
            _impl::_HashMapImpl<K,V,C> * pNewImpl = new _impl::_HashMapImpl<K,V,C>(*_pImpl);
            _pImpl->decrementReferenceCount();
            pNewImpl->incrementReferenceCount();
            _pImpl = pNewImpl;
        }
    }
}

#endif  //  __CARL_Map_hpp_included__
//  End of CARL_Map.hpp
