//
//  CARL_Memory.hpp - Raw memory services object.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL services for raw memory access.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Memory.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Memory_hpp_included__
#define __CARL_Memory_hpp_included__

namespace carl
{
    /**
     *  @class Memory CARL.hpp
     *  Provides low-level access to raw memory.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Memory
    {
        //////////
        //  This is a utility class
    private:
        Memory();
        Memory(const Memory &);
        ~Memory();
        void                operator = (const Memory &);

        //////////
        //  Operations
    public:
        //////////
        //  Byte order services - aligned access
        /**
         *  Loads the naturally aligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static UInt16       loadUInt16(const void * pRawData, ByteOrder rawDataByteOrder)
        {
            require(pRawData != NULL);

            UInt16 result = *static_cast<const UInt16*>(pRawData);
            if (rawDataByteOrder != System::HostByteOrder)
            {
                return _byteswapUInt16(result);
            }
            return result;
        }

        /**
         *  Loads the naturally aligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static UInt32       loadUInt32(const void * pRawData, ByteOrder rawDataByteOrder)
        {
            require(pRawData != NULL);

            UInt32 result = *static_cast<const UInt32*>(pRawData);
            if (rawDataByteOrder != System::HostByteOrder)
            {
                return _byteswapUInt32(result);
            }
            return result;
        }

        /**
         *  Loads the naturally aligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static UInt64       loadUInt64(const void * pRawData, ByteOrder rawDataByteOrder)
        {
            require(pRawData != NULL);

            UInt64 result = *static_cast<const UInt64*>(pRawData);
            if (rawDataByteOrder != System::HostByteOrder)
            {
                return _byteswapUInt64(result);
            }
            return result;
        }

        /**
         *  Loads the naturally aligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static Int16        loadInt16(const void * pRawData, ByteOrder rawDataByteOrder)
        {
            return static_cast<Int16>(loadUInt16(pRawData, rawDataByteOrder));
        }

        /**
         *  Loads the naturally aligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static Int32        loadInt32(const void * pRawData, ByteOrder rawDataByteOrder)
        {
            return static_cast<Int32>(loadUInt32(pRawData, rawDataByteOrder));
        }

        /**
         *  Loads the naturally aligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static Int64        loadInt64(const void * pRawData, ByteOrder rawDataByteOrder)
        {
            return static_cast<Int64>(loadUInt64(pRawData, rawDataByteOrder));
        }

        /**
         *  Loads the naturally aligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static Real32       loadReal32(const void * pRawData, ByteOrder rawDataByteOrder)
        {
            require(pRawData != NULL);

            carl::UInt32 bits = loadUInt32(pRawData, rawDataByteOrder);
            return *(carl::Real32*)(&bits);
        }

        /**
         *  Loads the naturally aligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static Real64       loadReal64(const void * pRawData, ByteOrder rawDataByteOrder)
        {
            require(pRawData != NULL);

            carl::UInt64 bits = loadUInt64(pRawData, rawDataByteOrder);
            return *(carl::Real64*)(&bits);
        }

        /**
         *  Stores the naturally aligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUInt16(UInt16 value, void * pRawData, ByteOrder rawDataByteOrder)
        {
            require(pRawData != NULL);

            if (rawDataByteOrder != System::HostByteOrder)
            {
                *static_cast<UInt16*>(pRawData) = _byteswapUInt16(value);
            }
            else
            {
                *static_cast<UInt16*>(pRawData) = value;
            }
        }

        /**
         *  Stores the naturally aligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUInt32(UInt32 value, void * pRawData, ByteOrder rawDataByteOrder)
        {
            require(pRawData != NULL);

            if (rawDataByteOrder != System::HostByteOrder)
            {
                *static_cast<UInt32*>(pRawData) = _byteswapUInt32(value);
            }
            else
            {
                *static_cast<UInt32*>(pRawData) = value;
            }
        }

        /**
         *  Stores the naturally aligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUInt64(UInt64 value, void * pRawData, ByteOrder rawDataByteOrder)
        {
            require(pRawData != NULL);

            if (rawDataByteOrder != System::HostByteOrder)
            {
                *static_cast<UInt64*>(pRawData) = _byteswapUInt64(value);
            }
            else
            {
                *static_cast<UInt64*>(pRawData) = value;
            }
        }

        /**
         *  Stores the naturally aligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeInt16(Int16 value, void * pRawData, ByteOrder rawDataByteOrder)
        {
            storeUInt16(static_cast<UInt16>(value), pRawData, rawDataByteOrder);
        }

        /**
         *  Stores the naturally aligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeInt32(Int32 value, void * pRawData, ByteOrder rawDataByteOrder)
        {
            storeUInt32(static_cast<UInt32>(value), pRawData, rawDataByteOrder);
        }

        /**
         *  Stores the naturally aligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeInt64(Int64 value, void * pRawData, ByteOrder rawDataByteOrder)
        {
            storeUInt64(static_cast<UInt64>(value), pRawData, rawDataByteOrder);
        }

        /**
         *  Stores the naturally aligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeReal32(Real32 value, void * pRawData, ByteOrder rawDataByteOrder)
        {
            require(pRawData != NULL);

            carl::UInt32 bits;
            *(carl::Real32*)(&bits) = value;
            storeUInt32(bits, pRawData, rawDataByteOrder);
        }

        /**
         *  Stores the naturally aligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeReal64(Real64 value, void * pRawData, ByteOrder rawDataByteOrder)
        {
            require(pRawData != NULL);

            carl::UInt64 bits;
            *(carl::Real64*)(&bits) = value;
            storeUInt64(bits, pRawData, rawDataByteOrder);
        }

        //////////
        //  Byte order services - unaligned access
        /**
         *  Loads the unaligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static UInt16       loadUnalignedUInt16(const void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Loads the unaligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static UInt32       loadUnalignedUInt32(const void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Loads the unaligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static UInt64       loadUnalignedUInt64(const void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Loads the unaligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static Int16        loadUnalignedInt16(const void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Loads the unaligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static Int32        loadUnalignedInt32(const void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Loads the unaligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static Int64        loadUnalignedInt64(const void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Loads the unaligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static Real32       loadUnalignedReal32(const void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Loads the unaligned value from memory using the specified byte order.
         *
         *  @param pRawData         The start address of the raw data to load.
         *  @param rawDataByteOrder The byte order to use for loading.
         *  @return                 The value loaded ftom the specified address.
         *  @since 1.0
         */
        static Real64       loadUnalignedReal64(const void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Stores the unaligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUnalignedUInt16(UInt16 value, void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Stores the unaligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUnalignedUInt32(UInt32 value, void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Stores the unaligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUnalignedUInt64(UInt64 value, void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Stores the unaligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUnalignedInt16(Int16 value, void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Stores the unaligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUnalignedInt32(Int32 value, void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Stores the unaligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUnalignedInt64(Int64 value, void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Stores the unaligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUnalignedReal32(Real32 value, void * pRawData, ByteOrder rawDataByteOrder);

        /**
         *  Stores the unaligned value to memory using the specified byte order.
         *
         *  @param value            The value to store.
         *  @param pRawData         The start address of the memory area where to store.
         *  @param rawDataByteOrder The byte order to use for storing.
         *  @since 1.0
         */
        static void         storeUnalignedReal64(Real64 value, void * pRawData, ByteOrder rawDataByteOrder);

        //////////
        //  pointer <-> integer translations
        /**
         *  Converts a pointer to numeric memory address.
         *
         *  @param pointer          The pointer to convert.
         *  @return                 The memory address that corresponds to the \a pointer.
         *  @since 1.0
         */
        static UIntPtr      addressOf(const void * pointer);

        /**
         *  Converts a numeric memory address to pointer.
         *
         *  @param address          The numeric memory address to convert.
         *  @return                 The pointer to the memory area at the specified address.
         *  @since 1.0
         */
        static void *       pointerTo(UIntPtr address);

        //////////
        //  Operations on raw memory blocks
        /**
         *  Copies raw bytes from one mrmory area to another memory area.
         *
         *  If some regions of the source area and the destination area overlap, the
         *  method ensures that the original source bytes in the overlapping region 
         *  are copied before being overwritten.
         *
         *  @param pDestination     The destination memory area.
         *  @param pSource          The source memory area.
         *  @param byteCount        The number of bytes to copy.
         *  @since 1.0
         */
        static void         copy(void * pDestination, const void * pSource, UIntSize byteCount);

        /**
         *  Fills a memory area with copies of the specified byte.
         *
         *  @param pDestination     The memory area to fill.
         *  @param byte             The byte to fill the memory area with.
         *  @param byteCount        The size of the memory area to fill.
         *  @since 1.0
         */
        static void         set(void * pDestination, UInt8 byte, UIntSize byteCount);

        /**
         *  Compares two memory areas lexicographically.
         *
         *  The memory areas are compared byte-by-byte until a difference is
         *  found or until the end of the emmory areas has been reached. In
         *  the former case the larger of two bytes signifies the lexicographically
         *  larger memory area; in the latter case memory areas are considered equal.
         *
         *  @param pBlock1          The 1st memory area to compare.
         *  @param pBlock2          The 2nd memory area to compare.
         *  @param byteCount        The number of bytes to compare.
         *  @return
         *                          -   0 if the two memory areas are equal.
         *                          -   <0 if the 1st memory area is lexicographically
         *                              smaller than the 2nd memory area.
         *                          -   >0 if the 1st memory area is lexicographically
         *                              larger than the 2nd memory area.
         *  @since 1.0
         */
        static int          compare(const void * pBlock1, const void * pBlock2, UIntSize byteCount);

        //////////
        //  Implementation helpers
    private:
        static UInt16 _byteswapUInt16(UInt16 value)
        {
            return (value << 8) | (value >> 8);
        }

        static UInt32 _byteswapUInt32(UInt32 value)
        {
            return (value << 24) |
                   ((value <<  8) & 0xFF0000) |
                   ((value >>  8) & 0xFF00) |
                   (value >> 24);
        }

        static UInt64 _byteswapUInt64(UInt64 value)
        {
            return (value << 56) |
                   ((value << 40) & (static_cast<UInt64>(0xFF) << 48)) |
                   ((value << 24) & (static_cast<UInt64>(0xFF) << 40)) |
                   ((value <<  8) & (static_cast<UInt64>(0xFF) << 32)) |
                   ((value >>  8) & (static_cast<UInt64>(0xFF) << 24)) |
                   ((value >> 24) & (static_cast<UInt64>(0xFF) << 16)) |
                   ((value >> 40) & (static_cast<UInt64>(0xFF) <<  8)) |
                   (value >> 56);
        }
    };
}

#endif  //  __CARL_Memory_hpp_included__
//  End of CARL_Memory.hpp
