//
//  CARL_Pair.hpp - Pairs.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.7 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines a Pair helper class.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Pair.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Pair_hpp_included__
#define __CARL_Pair_hpp_included__

namespace carl
{
    /**
     *  @class Pair CARL.hpp
     *  Represents a pair of values.
     *
     *  The value types of both elements (T1 and T2) must be:
     *
     *  -   Default constructable (Tn::Tn() must be defined and public)
     *  -   Copy constructable (Tn::Tn(const Tn &) must be defined and public)
     *  -   Destructable (Tn::~Tn() must be defined and public)
     *  -   Assignable (Tn::operator=(const Tn &) must be defined and public)
     *
     *  @param T1                   The type of the 1st element of a pair.
     *  @param T2                   The type of the 2nd element of a pair.
     *  @since 1.0
     */
    template<class T1, class T2>
    class Pair
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  Initializes a newly created pair with default values for both elements.
         *
         *  @since 1.0
         */
        Pair();

        /**
         *  The class constructor.
         *
         *  Initializes a newly created pair with specified values for both elements.
         *
         *  @param first            The value to assign to the 1st element of the pair.
         *  @param second           The value to assign to the 2nd element of the pair.
         *  @since 1.0
         */
        Pair(const T1 & first, const T2 & second);

        /**
         *  The copy constructor.
         *
         *  @param src              The pair to copy elements from.
         *  @since 1.0
         */
        Pair(const Pair<T1,T2> & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~Pair();

        /**
         *  The assignment operator.
         *
         *  @param src          The pair to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        Pair<T1,T2> &               operator = (const Pair<T1,T2> & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the 1st element of this pair.
         *
         *  @return                 The 1st element of this pair.
         *  @since 1.0
         */
        T1                          first() const;

        /**
         *  Returns the 2nd element of this pair.
         *
         *  @return                 The 2nd element of this pair.
         *  @since 1.0
         */
        T2                          second() const;

        //////////
        //  Implementation
    private:
        T1                          _first;
        T2                          _second;
    };

    template<class T1, class T2>
    Pair<T1,T2>::Pair()
    {
    }

    template<class T1, class T2>
    Pair<T1,T2>::Pair(const T1 & first, const T2 & second)
        :   _first(first), _second(second)
    {
    }

    template<class T1, class T2>
    Pair<T1,T2>::Pair(const Pair<T1,T2> & src)
        :   _first(src._first), _second(src._second)
    {
    }

    template<class T1, class T2>
    Pair<T1,T2>::~Pair() 
    {
    }

    template<class T1, class T2>
    Pair<T1,T2> & Pair<T1,T2>::operator = (const Pair<T1,T2> & src)
    {
        if (this != (&src))
        {
            _first = src._first;
            _second = src._second;
        }
        return *this;
    }

    template<class T1, class T2>
    T1 Pair<T1,T2>::first() const 
    { 
        return _first; 
    }

    template<class T1, class T2>
    T2 Pair<T1,T2>::second() const 
    { 
        return _second; 
    }
}

#endif  //  __CARL_Pair_hpp_included__
//  End of CARL_Pair.hpp
