//
//  CARL_Plugin.hpp - Plug-in framework.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.11 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL plugin framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Plugin.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Plugin_hpp_included__
#define __CARL_Plugin_hpp_included__

#if defined(CARL_LINKAGE_DYNAMIC)

namespace carl
{
    /**
     *  @class Plugin CARL.hpp
     *  A common base class for all CARL plugins.
     *
     *  A functionality of a CARL application or DLL that links against
     *  dynamic CARL library can be extended post-deployment by installing 
     *  additional plugin DLLs (also known as shared objects in Unix world).
     *  These plugin DLLs can provide additional functionality for a CARL
     *  application or DLL.
     *
     *  Each CARL plugin DLL must:
     *
     *  -   Link against the dynamic CARL library.
     *  -   Define an "application key", which uniquely identifier
     *      the CARL application or DLL for which the plugin is written.
     *      Application keys can be arbitrary ASCII strings.
     *  -   Define one or more concrete subclasses of Plugin.
     *      Each of these concrete subclasses represents an
     *      independent plugin.
     *  -   Create and export an instance of that subclass using
     *      special macros defined by CARL for that purpose.
     *
     *  If two concrete subclasses PluginA and PluginB of the
     *  Plugin class are defined, the following code must then
     *  be written to export these plugins from the CARL
     *  plugin DLL:
     *
     *  @code
     *  CARL_BEGIN_PLUGIN_EXPORT
     *  CARL_EXPORT_PLUGIN(PluginA)
     *  CARL_EXPORT_PLUGIN(PluginB)
     *  CARL_END_PLUGIN_EXPORT
     *  @endcode
     *
     *  All concrete Plugin subclasses must define a public default
     *  constructor and a public destructor.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Plugin
    {
        //////////
        //  Can't assign or copy construct
    private:
        Plugin(const Plugin &);
        void                    operator = (const Plugin &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  The newly created Plugin instance is associated with the specified
         *  application key, which is expected to be statically allocated (typically
         *  the derived class constructor will call this constructor with a character
         *  string constant as a parameter).
         *
         *  All concrete derived classes must provide a public default constructor.
         *
         *  @param pApplicationKey  The application key identifying the application
         *                      for which the plugin is designed.
         *  @since 1.0
         */
        explicit Plugin(const char * pApplicationKey);

        /**
         *  The class destructor.
         *
         *  All concrete derived classes must provide a public destructor.
         *
         *  @since 1.0
         */
        virtual ~Plugin() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns the CARL build key associated with this plugin.
         *
         *  A CARL build key is a string that encapsulates processor, toolchain,
         *  OS, mode and CARL version information. A client that uses a dynamic
         *  CARL library is considered to be compatible with that library if
         *  and only if the two have the same build keys.
         *
         *  @return             The build key used by the plugin.
         *  @since 1.0
         */
        virtual const char *    getBuildKey() const = 0;

        /**
         *  Returns the application key key associated with this plugin.
         *
         *  An application key is a string that uniquely identifier an
         *  application or DLL for which the plugin is designated.
         *
         *  @return             The build key used by the plugin.
         *  @since 1.0
         */
        const char *            getApplicationKey() const { return _pApplicationKey; }

        //////////
        //  Implementation
    private:
        const char *            _pApplicationKey;
    };

    /**
     *  @class PluginLoader CARL.hpp
     *  A utility class providing plugin loading services.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC PluginLoader
    {
        //////////
        //  Utility classes cannot be instantiated
    private:
        PluginLoader();
        PluginLoader(const PluginLoader &);
        ~PluginLoader();
        void                    operator = (const PluginLoader &);

        //////////
        //  Operations
    public:
        /**
         *  Loads all plugins defined in a specified DLL.
         *
         *  @param pFileName    A pointer to nul-terminated ASCII string
         *                      representing the file name of the plugin DLL.
         *  @param pApplicationKey  The required application key. All plugins
         *                      with different application keys defined in the 
         *                      specified plugin DLL are ignored.
         *  @return             A list of all plugin instances defined and
         *                      exported by the plugin DLL in question. Elements
         *                      in the list are not guaranteed to be in any
         *                      specific order.
         *  @since 1.0
         */
        static Array<Plugin*>   loadFromFile(const char * pFileName, const char * pApplicationKey);

        /**
         *  Loads all plugins defined in a specified DLL.
         *
         *  @param pFileName    A pointer to nul-terminated ISo-10646 string
         *                      representing the file name of the plugin DLL.
         *  @param pApplicationKey  The required application key. All plugins
         *                      with different application keys defined in the 
         *                      specified plugin DLL are ignored.
         *  @return             A list of all plugin instances defined and
         *                      exported by the plugin DLL in question. Elements
         *                      in the list are not guaranteed to be in any
         *                      specific order.
         *  @since 1.0
         */
        static Array<Plugin*>   loadFromFile(const Char * pFileName, const char * pApplicationKey);

        /**
         *  Loads all plugins defined in a specified DLL.
         *
         *  @param fileName     A string representing the file name of the plugin DLL.
         *  @param pApplicationKey  The required application key. All plugins
         *                      with different application keys defined in the 
         *                      specified plugin DLL are ignored.
         *  @return             A list of all plugin instances defined and
         *                      exported by the plugin DLL in question. Elements
         *                      in the list are not guaranteed to be in any
         *                      specific order.
         *  @since 1.0
         */
        static Array<Plugin*>   loadFromFile(const String & fileName, const char * pApplicationKey);

        /**
         *  Loads all plugins defined in a specified DLL.
         *
         *  @param fileName     A path to the plugin DLL.
         *  @param pApplicationKey  The required application key. All plugins
         *                      with different application keys defined in the 
         *                      specified plugin DLL are ignored.
         *  @return             A list of all plugin instances defined and
         *                      exported by the plugin DLL in question. Elements
         *                      in the list are not guaranteed to be in any
         *                      specific order.
         *  @since 1.0
         */
        static Array<Plugin*>   loadFromFile(const PathName & fileName, const char * pApplicationKey);

        /**
         *  Loads all plugin DLLs located in the specified directory.
         *
         *  When the directory is scanned for CARL plugin DLLs, only
         *  those files are considered that:
         *
         *  -   Have the proper filename extension (.dll for Windows, .so
         *      for Unix, etc.)
         *  -   Have the same build key as the application or DLL that does
         *      the loading.
         *  -   Have the required application key.
         *
         *  @param pDirectoryName   A pointer to nul-terminated ASCII string
         *                      representing the name of the directory which
         *                      is searched for plugin DLL.
         *  @param processSubdirectories    If true, subdirectories of the
         *                      specified directory are recursively scanned for
         *                      plugin DLLs. Otherwise, only the specified 
         *                      directory is scanned.
         *  @param pApplicationKey  The required application key. All plugins
         *                      with different application keys defined in the 
         *                      specified plugin DLL are ignored.
         *  @return             A list of all plugin instances defined and
         *                      exported by the plugin DLLs in question. Elements
         *                      in the list are not guaranteed to be in any
         *                      specific order.
         *  @since 1.0
         */
        static Array<Plugin*>   loadFromDirectory(const char * pDirectoryName, const char * pApplicationKey, bool processSubdirectories = false);

        /**
         *  Loads all plugin DLLs located in the specified directory.
         *
         *  When the directory is scanned for CARL plugin DLLs, only
         *  those files are considered that:
         *
         *  -   Have the proper filename extension (.dll for Windows, .so
         *      for Unix, etc.)
         *  -   Have the same build key as the application or DLL that does
         *      the loading.
         *  -   Have the required application key.
         *
         *  @param pDirectoryName   A pointer to nul-terminated ISO-10646 string
         *                      representing the name of the directory which
         *                      is searched for plugin DLL.
         *  @param processSubdirectories    If true, subdirectories of the
         *                      specified directory are recursively scanned for
         *                      plugin DLLs. Otherwise, only the specified 
         *                      directory is scanned.
         *  @param pApplicationKey  The required application key. All plugins
         *                      with different application keys defined in the 
         *                      specified plugin DLL are ignored.
         *  @return             A list of all plugin instances defined and
         *                      exported by the plugin DLLs in question. Elements
         *                      in the list are not guaranteed to be in any
         *                      specific order.
         *  @since 1.0
         */
        static Array<Plugin*>   loadFromDirectory(const Char * pDirectoryName, const char * pApplicationKey, bool processSubdirectories = false);

        /**
         *  Loads all plugin DLLs located in the specified directory.
         *
         *  When the directory is scanned for CARL plugin DLLs, only
         *  those files are considered that:
         *
         *  -   Have the proper filename extension (.dll for Windows, .so
         *      for Unix, etc.)
         *  -   Have the same build key as the application or DLL that does
         *      the loading.
         *  -   Have the required application key.
         *
         *  @param directoryName    A string representing the name of the 
         *                      directory which is searched for plugin DLL.
         *  @param processSubdirectories    If true, subdirectories of the
         *                      specified directory are recursively scanned for
         *                      plugin DLLs. Otherwise, only the specified 
         *                      directory is scanned.
         *  @param pApplicationKey  The required application key. All plugins
         *                      with different application keys defined in the 
         *                      specified plugin DLL are ignored.
         *  @return             A list of all plugin instances defined and
         *                      exported by the plugin DLLs in question. Elements
         *                      in the list are not guaranteed to be in any
         *                      specific order.
         *  @since 1.0
         */
        static Array<Plugin*>   loadFromDirectory(const String & directoryName, const char * pApplicationKey, bool processSubdirectories = false);

        /**
         *  Loads all plugin DLLs located in the specified directory.
         *
         *  When the directory is scanned for CARL plugin DLLs, only
         *  those files are considered that:
         *
         *  -   Have the proper filename extension (.dll for Windows, .so
         *      for Unix, etc.)
         *  -   Have the same build key as the application or DLL that does
         *      the loading.
         *  -   Have the required application key.
         *
         *  @param directoryName    A path to the directory which is searched 
         *                      for plugin DLL.
         *  @param processSubdirectories    If true, subdirectories of the
         *                      specified directory are recursively scanned for
         *                      plugin DLLs. Otherwise, only the specified 
         *                      directory is scanned.
         *  @param pApplicationKey  The required application key. All plugins
         *                      with different application keys defined in the 
         *                      specified plugin DLL are ignored.
         *  @return             A list of all plugin instances defined and
         *                      exported by the plugin DLLs in question. Elements
         *                      in the list are not guaranteed to be in any
         *                      specific order.
         *  @since 1.0
         */
        static Array<Plugin*>   loadFromDirectory(const PathName & directoryName, const char * pApplicationKey, bool processSubdirectories = false);

        /**
         *  Loads all application plugins available in standard locations.
         *
         *  The following locations are searched for plugins:
         *
         *  -   The application itself may export some plugins. These are found first.
         *  -   The subdirectory "plugins" of a directory where the main application
         *      executable resides is searched non-recursively for plugin DLLs next.
         *  -   Finally, the directory where the main application executable resides 
         *      is searched non-recursively for plugin DLLs next.
         *
         *  @param pApplicationKey  An application key identifying the application
         *                      for which plugins are sought.
         *  @return             A list of all plugin instances defined and
         *                      exported by the application or found plugin DLLs. 
         *                      Elements in the list are not guaranteed to be in any
         *                      specific order.
         *  @since 1.0
         */
        static Array<Plugin*>   loadInstalledPlugins(const char * pApplicationKey);

        //////////
        //  Implementation
    private:
        static HashMap<String, Library*> *  getMapFileNamesToLibraries();
        static CriticalSection *_getStaticGuard();
    };

    extern "C"
    {
        typedef void (CARL_CALLBACK * PluginProviderProc)(Array<Plugin *> & plugins);
    }
}

#define CARL_BEGIN_PLUGIN_EXPORT                    \
extern "C" CARL_EXPORT                              \
void CARL_GetPlugins(Array<Plugin *> & plugins)     \
{                                                   \
    require(plugins.size() == 0);

#define CARL_EXPORT_PLUGIN(PluginType)              \
    static PluginType plugin;                       \
    plugins.append(&plugin);

#define CARL_END_PLUGIN_EXPORT                      \
}

#endif  //  defined(CARL_LINKAGE_DYNAMIC)

#endif  //  __CARL_Plugin_hpp_included__
//  End of CARL_Plugin.hpp
