//
//  CARL_Process.hpp - A process abstraction.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.23 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL process API.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Process.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Process_hpp_included__
#define __CARL_Process_hpp_included__

namespace carl
{
    class CARL_PUBLIC Process;

    /**
     *  @interface IProcessOutput CARL.hpp
     *  An input stream used for capturing output from a sub-process.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IProcessOutput : public ITextInput
    {
        friend class Process;

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        IProcessOutput() {}
        virtual ~IProcessOutput() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns the character set used by this stream.
         *
         *  @return             The character set used by this stream.
         *  @since 1.0
         */
        virtual CharacterSet *  getCharacterSet() const = 0;

        /**
         *  Sets the character set used by this stream.
         *
         *  @param pCharacterSet    The character set to use for 
         *                      byte-to-character translation.
         *  @since 1.0
         */
        virtual void            setCharacterSet(CharacterSet * pCharacterSet) = 0;
    };

    /**
     *  @interface IProcessInput CARL.hpp
     *  An output stream used for providing input to a sub-process.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IProcessInput : public ITextOutput
    {
        friend class Process;

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        IProcessInput() {}
        virtual ~IProcessInput() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns the character set used by this stream.
         *
         *  @return             The character set used by this stream.
         *  @since 1.0
         */
        virtual CharacterSet *  getCharacterSet() const = 0;

        /**
         *  Sets the character set used by this stream.
         *
         *  @param pCharacterSet    The character set to use for 
         *                      character-to-byte translation.
         *  @since 1.0
         */
        virtual void            setCharacterSet(CharacterSet * pCharacterSet) = 0;
    };

    /**
     *  @class Process CARL.hpp
     *  Represents a process in the host OS sense.
     *
     *  An instance of this class can be in either of the two states:
     *
     *  -   Unbound - the Process instance does not correspond to any
     *      host OS process.
     *  -   Bound - the Process instance correspond to a specific
     *      host OS process.
     *
     *  Normally Process instances are created in an unbound state and used to
     *  start child processes, becoming bound to these child processes for the
     *  remainder of their life.
     *
     *  In addition, each CARL application has a predefined Process instance that
     *  corresponds to the "current process" and, as a consequence, is always bound.
     *
     *  Destroying a bound Process instance has no effect on the host OS process
     *  that the instance is bound to (except that the application no longer has
     *  access to that host OS process).
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Process
    {
        //////////
        //  Can't assign or copy construct
    private:
        Process(const Process &);
        void                        operator = (const Process &);

        //////////
        //  Construction/destruction
    private:
        Process(bool);
    public:
        /**
         *  The default constructor.
         *
         *  Creates a new, unbound, Process instance.
         *
         *  @since 1.0
         */
        Process();

        /**
         *  The class destructor.
         *
         *  Destroying a bound Process instance has no effect on the host OS process
         *  that the instance is bound to (except that the application no longer has
         *  access to that host OS process).
         *
         *  @since 1.0
         */
        ~Process();

        //////////
        //  Operations
    public:
        /**
         *  Possible process states.
         *
         *  @since 1.0
         */
        enum State
        {
            /**
             *  The Process instance has been constructed, but has not yet
             *  bound to any process.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            State_Initializing = 0,

            /**
             *  The Process instance is bound to a currently running process.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            State_Active = 1,

            /**
             *  The Process instace is bound to a process that has finished.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            State_Finished = 2
        };

        /**
         *  Standard exit codes for processes.
         *
         *  @since 1.0
         */
        enum ExitCode
        {
            /**
             *  The process has finished successfully.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ExitCode_Success    = 0,

            /**
             *  The process has encountered an error and ended prematurely.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ExitCode_Failure    = 1,

            /**
             *  The process has been aborted by the system instead of
             *  terminating (normally or abnormally) on its own accord.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ExitCode_Panic      = 127
        };

        /**
         *  Returns a pointer to the "current process" instance.
         *
         *  The returned Process instance is always bound to the current
         *  process and cannot be destroyed by the application.
         *
         *  @return                 A pointer to the "current process" singleton.
         *  @since 1.0
         */
        static Process *            getCurrentProcess();

        /**
         *  Queries the state of the Process instance.
         *
         *  Cients can call this method to, for example, determine when
         *  a child process finishes.
         *
         *  @return                 The current state of the Process instance.
         *  @since 1.0
         */
        State                       getState() const;

        /**
         *  Exits this process with the specified exit code.
         *
         *  If this method is called on the "current process" instance, then
         *  the current process exits with the specified error code. Otherwise,
         *  the host OS process to which the Process instance if bound is forced
         *  to terminate with the specified exit code.
         *
         *  Calling this method on an unbound Process instance is an error.
         *
         *  @param exitCode         The exit code for the terminated process.
         *  @since 1.0
         */
        void                        exit(ExitCode exitCode);

        /**
         *  Returns the absolute (with device and full path) name of the
         *  executable file that was used to start the host OS process
         *  bound to this Process instance.
         *
         *  Calling this method on an unbound Process instance is an error.
         *
         *  @return                 The full file name of the executable file
         *                          that was used to start the host OS process
         *                          bound to this Process instance.
         *  @since 1.0
         */
        String                      getApplicationFileName() const;

        /**
         *  Returns the environment associated with this process.
         *
         *  For the "current process" instance, the initial contents of the
         *  returned environment reflects on the environment passed to the
         *  current process by its creator when the current process was started.
         *
         *  A newly created unbound process inherits the current environment from
         *  the current process. This inherited environment can be modified before
         *  the process is started (and bound) if a child process needs an environment
         *  that is different from that of its parent.
         *
         *  @return                 A pointer to a private Environment instance for
         *                          this process.
         *  @since 1.0
         */
        Environment *               getEnvironment() const { return _pEnvironment; }

        /**
         *  Possible outcomes of an attempt to start a child process.
         *
         *  @since 1.0
         */
        enum StartOutcome
        {
            /**
             *  The child process was started successfully.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            StartOutcome_Success            = 0,

            /**
             *  The command line is invalid for the host OS.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            StartOutcome_InvalidCommandLine = 1,

            /**
             *  The initial environment is invalid for the host OS.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            StartOutcome_InvalidEnvironment = 2,

            /**
             *  The executable file for the child process does not exist.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            StartOutcome_ImageNotFound      = 3,

            /**
             *  There is not enough memory to start a child process.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            StartOutcome_InsufficientMemory = 4,

            /**
             *  The caller has no permission to start the required child process.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            StartOutcome_AccessDenied       = 5,

            /**
             *  A child process cannot be started for an unknown reason.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            StartOutcome_UnknownError       = 6
        };

         /**
         *  The flags specifying what standard streams of a child process are
         *  captured by the parent process.
         *
         *  If more than one standard stream for a child process needs to be
         *  captured, several of the flags can be combined together using the
         *  "|" operator.
         *  @since 1.0
         */
        enum CaptureStandardStream
        {
            /**
             *  The parent process captures the standard input stream of the
             *  child process.
             *
             *  If this flag is not specified, the child process uses the same
             *  standard input stream as a parent process who started it.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            CaptureStandardInputStream      = 0x01,

            /**
             *  The parent process captures the standard outnput stream of the
             *  child process.
             *
             *  If this flag is not specified, the child process uses the same
             *  standard output stream as a parent process who started it.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            CaptureStandardOutputStream     = 0x02,

            /**
             *  The parent process captures the standard error stream of the
             *  child process.
             *
             *  If this flag is not specified, the child process uses the same
             *  standard error stream as a parent process who started it.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            CaptureStandardErrorStream      = 0x04
        };

        /**
         *  Possible outcomes of an attempt to obtain an exit code of a process.
         *
         *  @since 1.0
         */
        enum GetExitCodeOutcome
        {
            /**
             *  The exit code of a process was obtained successfully.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GetExitCodeOutcome_Success            = 0,

            /**
             *  The Process instance does not correspond to an OS process.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GetExitCodeOutcome_InvalidState = 1,

            /**
             *  The process is still running.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GetExitCodeOutcome_StillRunning = 2,

            /**
             *  The exit code could not be retrieved for an unknown reason.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            GetExitCodeOutcome_NotAvailable = 3
        };

        /**
         *  Start a child process without capturing any of its standard streams.
         *
         *  The specified array of command-line arguments must contain at least
         *  1 element. The 0th element of an array always specifies the name of the
         *  executable file that shall be started as a child process. Any additional
         *  elements specify command line parameters for the child process, with
         *  each array element specifying one command line parameter.
         *
         *  The child process uses the same standard input, output and error streams
         *  as the parent process who started it.
         *
         *  @param commandNameAndArguments  Specifies the executable file to start
         *                          as a child process and (optionally) command line
         *                          parameters for the child process.
         *  @return                 The status code that tells whether the child
         *                          process was started sucecssfully and, if not,
         *                          the reason why starting a shild process failed.
         *  @since 1.0
         */
        StartOutcome                start(const Array<String> & commandNameAndArguments) { return start(commandNameAndArguments, 0); }

        /**
         *  Start a child process without, capturing one or more of its standard
         *  streams if necessary.
         *
         *  The specified array of command-line arguments must contain at least
         *  1 element. The 0th element of an array always specifies the name of the
         *  executable file that shall be started as a child process. Any additional
         *  elements specify command line parameters for the child process, with
         *  each array element specifying one command line parameter.
         *
         *  The \a captureStandardStreams parameter is a combination of
         *  CaptureStandardStream flags that specifies which of the standard stream
         *  of a child process shall be captured by the parent process.
         *
         *  @param commandNameAndArguments  Specifies the executable file to start
         *                          as a child process and (optionally) command line
         *                          parameters for the child process.
         *  @param captureStandardStreams   specifies which of the standard stream
         *                          of a child process shall be captured by the 
         *                          parent process.
         *  @return                 The status code that tells whether the child
         *                          process was started sucecssfully and, if not,
         *                          the reason why starting a shild process failed.
         *  @since 1.0
         */
        StartOutcome                start(const Array<String> & commandNameAndArguments, int captureStandardStreams);

        /**
         *  Stores the exit code of a process.
         *
         *  @param exitCode         The recipient for an exit code of a process.
         *  @return                 The status code specifying whether the exit
         *                          code of a process has been successfully stored
         *                          and, if not, what went wrong.
         *  @since 1.0
         */
        GetExitCodeOutcome          getExitCode(int & exitCode);

        /**
         *  Returns the captuted standard input stream of a bound child process.
         *
         *  If the standard input stream of a child process was not captured when
         *  the child process was started, returns NULL.
         *
         *  It this service is called on current process or in an unbound Process
         *  instance, returns NULL.
         *
         *  @return                 The captuted standard input stream of a bound 
         *                          child process.
         *  @since 1.0
         */
        IProcessInput *             getStdinStream();

        /**
         *  Returns the captuted standard output stream of a bound child process.
         *
         *  If the standard output stream of a child process was not captured when
         *  the child process was started, returns NULL.
         *
         *  It this service is called on current process or in an unbound Process
         *  instance, returns NULL.
         *
         *  @return                 The captuted standard output stream of a bound 
         *                          child process.
         *  @since 1.0
         */
        IProcessOutput *            getStdoutStream();

        /**
         *  Returns the captuted standard error stream of a bound child process.
         *
         *  If the standard error stream of a child process was not captured when
         *  the child process was started, returns NULL.
         *
         *  It this service is called on current process or in an unbound Process
         *  instance, returns NULL.
         *
         *  @return                 The captuted standard error stream of a bound 
         *                          child process.
         *  @since 1.0
         */
        IProcessOutput *            getStderrStream();

        /**
         *  Queries the total elapsed time for which this process has 
         *  been running so far.
         *
         *  For the current process, the return value is guaranteed to be
         *  accurate within platform's precision. For a child process, the
         *  return vaue is guaranteed to be accurate only after the child
         *  process finishes.
         *
         *  @return                 The elapsed time of the process.
         *  @since 1.0
         */
        TimeSpan                    getElapsedTime() const;

        /**
         *  Queries the total time used by the process in system mode so far.
         *
         *  For the current process, the return value is guaranteed to be
         *  accurate within platform's precision. For a child process, the
         *  return vaue is guaranteed to be accurate only after the child
         *  process finishes.
         *
         *  @return                 The system time of the process.
         *  @since 1.0
         */
        TimeSpan                    getSystemTime() const;

        /**
         *  Queries the total time used by the process in user mode so far.
         *
         *  For the current process, the return value is guaranteed to be
         *  accurate within platform's precision. For a child process, the
         *  return vaue is guaranteed to be accurate only after the child
         *  process finishes.
         *
         *  @return                 The user time of the process.
         *  @since 1.0
         */
        TimeSpan                    getUserTime() const;

        //////////
        //  Implementation
    private:
        static Process *            _pCurrentProcess;
        static CriticalSection *    _getCurrentProcessGuard();
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        void *                      _hProcess;
        UInt32                      _processId;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        int                         _pid;
        //  Timing data
        mutable TimeSpan            _elapsedTimeStart;
        mutable TimeSpan            _elapsedTimeEnd;
        mutable TimeSpan            _userTime;
        mutable TimeSpan            _systemTime;

        // status saved from getState() for getExitCode()
        mutable int                 _exitCode;
        mutable GetExitCodeOutcome  _exitState;
#else
    #error Unsupported platform
#endif
        mutable State               _state;
        bool                        _isCurrent;
        mutable Environment *       _pEnvironment;
        IProcessInput *             _pStdin;
        IProcessOutput *            _pStdout;
        IProcessOutput *            _pStderr;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        void *                      _hStdinRead;
        void *                      _hStdoutWrite;
        void *                      _hStderrWrite;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        int                         _hStdinRead;    //  -1 == not redirected
        int                         _hStdoutWrite;  //  -1 == not redirected
        int                         _hStderrWrite;  //  -1 == not redirected
#else
    #error Unsupported platform
#endif
    };
}

#endif  //  __CARL_Process_hpp_included__
//  End of CARL_Process.hpp
