//
//  CARL_Queue.hpp - A variable-length queue.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.7 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines a queue data type.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Queue.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Queue_hpp_included__
#define __CARL_Queue_hpp_included__

namespace carl
{
    namespace _impl
    {
        //////////
        //  Variable-length queue implementation
        //
        //  The value type (T) must be:
        //  *   Copy constructable (T::T(const T &) must be defined and public)
        //  *   Destructable (T::~T() must be defined and public)
        //  *   Assignable (T::operator=(const T &) must be defined and public)
        template<class T>
        class _QueueImpl
        {
            //////////
            //  Can't assign
        private:
            void                    operator = (const _QueueImpl<T> &);

            //////////
            //  Construction/destruction
        public:
            _QueueImpl();
            _QueueImpl(const _QueueImpl<T> & src);
            ~_QueueImpl();

            //////////
            //  Operations
        public:
            bool                    isEmpty() const;
            UIntSize                size() const;
            void                    enqueue(const T & value);
            T                       dequeue();
            T &                     front();
            void                    removeAll();
            Iterator<T>             enumerateElementsHeadToTail() const;
            Iterator<T>             enumerateElementsTailToHead() const;
            UIntSize                getReferenceCount() const;
            void                    incrementReferenceCount();
            void                    decrementReferenceCount();

            //////////
            //  Implementation
        private:
            struct _Element
            {
                _Element(const T & value) 
                    :   _value(value), _pNextInChain(NULL)
                {
                }

                T                   _value;
                _Element *          _pNextInChain;
            };
            _Element *              _pHead;
            _Element *              _pTail;
            UIntSize                _elementCount;
            UIntSize                _referenceCount;

            //  Helpers
            _impl::_SnapshotIteratorImpl<T> *   _createHeadToTailIteratorImpl() const;
            _impl::_SnapshotIteratorImpl<T> *   _createTailToHeadIteratorImpl() const;
        };

        template<class T>
        _QueueImpl<T>::_QueueImpl()
        {
            _pHead = NULL;
            _pTail = NULL;
            _elementCount = 0;
            _referenceCount = 0;
        }

        template<class T>
        _QueueImpl<T>::_QueueImpl(const _QueueImpl<T> & src)
        {
            _pHead = NULL;
            _pTail = NULL;
            _elementCount = 0;
            //  Copy elements from src
            for (Iterator<T> it = src.enumerateElementsHeadToTail(); !it.finished(); it.advance())
            {
                enqueue(it.item());
            }
            _referenceCount = 0;
        }

        template<class T>
        _QueueImpl<T>::~_QueueImpl()
        {
            removeAll();
        }

        template<class T>
        bool _QueueImpl<T>::isEmpty() const
        {
            return _elementCount == 0;
        }

        template<class T>
        UIntSize _QueueImpl<T>::size() const
        {
            return _elementCount;
        }

        template<class T>
        void _QueueImpl<T>::enqueue(const T & value)
        {
            _Element * pElement = new _Element(value);
            if (_pHead == NULL)
            {
                _pHead = _pTail = pElement;
            }
            else
            {
                _pTail->_pNextInChain = pElement;
                _pTail = pElement;
            }
            _elementCount++;
        }

        template<class T>
        T _QueueImpl<T>::dequeue()
        {
            require(_elementCount != 0);
            _Element * pElement = _pHead;
            _pHead = pElement->_pNextInChain;
            if (_pHead == NULL)
            {
                _pTail = NULL;
            }
            T result = pElement->_value;
            delete pElement;
            _elementCount--;
            return result;
        }

        template<class T>
        T & _QueueImpl<T>::front()
        {
            require(_elementCount != 0);
            return _pHead->_value;
        }

        template<class T>
        void _QueueImpl<T>::removeAll()
        {
            while (_elementCount != 0)
            {
                dequeue();
            }
        }

        template<class T>
        Iterator<T> _QueueImpl<T>::enumerateElementsHeadToTail() const
        {
            return Iterator<T>(_createHeadToTailIteratorImpl());
        }

        template<class T>
        Iterator<T> _QueueImpl<T>::enumerateElementsTailToHead() const
        {
            return Iterator<T>(_createTailToHeadIteratorImpl());
        }

        template<class T>
        UIntSize _QueueImpl<T>::getReferenceCount() const
        {
            return _referenceCount;
        }

        template<class T>
        void _QueueImpl<T>::incrementReferenceCount()
        {
            _referenceCount++;
            require(_referenceCount != 0);
        }

        template<class T>
        void _QueueImpl<T>::decrementReferenceCount()
        {
            require(_referenceCount != 0);
            _referenceCount--;
        }

        template<class T>
        _impl::_SnapshotIteratorImpl<T> * _QueueImpl<T>::_createHeadToTailIteratorImpl() const
        {
            T * pElements = new T[_elementCount];
            UIntSize elementIndex = 0;
            for (_Element * pElement = _pHead; pElement != NULL; pElement = pElement->_pNextInChain)
            {
                pElements[elementIndex++] = pElement->_value;
            }
            return new _impl::_SnapshotIteratorImpl<T>(_elementCount, pElements);
        }

        template<class T>
        _impl::_SnapshotIteratorImpl<T> * _QueueImpl<T>::_createTailToHeadIteratorImpl() const
        {
            T * pElements = new T[_elementCount];
            UIntSize elementIndex = _elementCount - 1;
            for (_Element * pElement = _pHead; pElement != NULL; pElement = pElement->_pNextInChain)
            {
                pElements[elementIndex--] = pElement->_value;
            }
            return new _impl::_SnapshotIteratorImpl<T>(_elementCount, pElements);
        }
    }

    /**
     *  @class Queue CARL.hpp
     *  Variable-length queue.
     *
     *  CARL queues use implicit sharing to reduce the copy construction and 
     *  assignment overhead. This means that it is feasible to both pass
     *  queues to functions as value parameters and return queues from functions.
     *
     *  The value type (T) must be:
     *
     *  -   Copy constructable (T::T(const T &) must be defined and public).
     *  -   Destructable (T::~T() must be defined and public).
     *  -   Assignable (T::operator=(const T &) must be defined and public).
     *
     *  @param T                The type of queue's elements.
     *  @since 1.0
     */
    template<class T>
    class Queue
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default class constructor.
         *
         *  Creates an initially empty queue.
         *
         *  @since 1.0
         */
        Queue();

        /**
         *  The copy constructor.
         *
         *  Initializes a newly created queue to be a copy of the specified
         *  queue. queue copy construction uses lazy initialization.
         *
         *  @param src          The existing queue that shall be copied into
         *                      the newly constructed queue.
         *  @since 1.0
         */
        Queue(const Queue<T> & src);

        /**
         *  The class destructor.
         *
         *  Releases all memory used by the queue.
         *
         *  @since 1.0
         */
        ~Queue();

        /**
         *  The assignment operator.
         *
         *  Makes the assigned-to queue an exact copy of the assigned-from
         *  queue. Queue assignment uses lazy initialization.
         *
         *  @param src          The existing queue that shall be assigned 
         *                      to another queue.
         *  @return             A reference to the assigned-to queue.
         *  @since 1.0
         */
        Queue<T> &              operator = (const Queue & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns true iff this queue has no elements, false otherwise.
         *
         *  @return             True iff this queue has no elements, false otherwise.
         *  @since 1.0
         */
        bool                    isEmpty() const;

        /**
         *  Returns the number of elements in this queue.
         *
         *  @return             The number of elements in this queue.
         *  @since 1.0
         */
        UIntSize                size() const;

        /**
         *  Adds an item to the tail of the queue.
         *
         *  @param value        The item to add to the tail of the queue.
         *  @since 1.0
         */
        void                    enqueue(const T & value);

        /**
         *  Removes an element at the head of the queue and returns it.
         *
         *  The call is an error if the queue is empty.
         *
         *  @return             The element that was at the head of the 
         *                      queue befiore the call.
         *  @since 1.0
         */
        T                       dequeue();

        /**
         *  Returns an element at the head of the queue.
         *
         *  The element remains at the head of the queue. The call is an 
         *  error if the queue is empty.
         *
         *  @return             The element at the head of the queue.
         *  @since 1.0
         */
        T                       front() const;

        /**
         *  Returns an element at the head of the queue.
         *
         *  The element remains at the head of the queue. The call is an 
         *  error if the queue is empty.
         *
         *  @return             The element at the head of the queue.
         *  @since 1.0
         */
        T &                     front();

        /**
         *  Removes all queue elements, leaving the queue empty.
         *
         *  @since 1.0
         */
        void                    removeAll();

        /**
         *  Returns a forward iterator for queue elements.
         *
         *  The returned iterator behaves as if a snapshot of the
         *  queue has been taken. In particular, removing elements from
         *  the queue after an iterator has been created has no effect
         *  on the iterator, which will still iterate over the entire
         *  list of queue elements as they were at the time when the
         *  iterator was created.
         *
         *  @return             A forward iterator over all queue elements.
         *  @since 1.0
         */
        Iterator<T>             enumerateElementsHeadToTail() const;

        /**
         *  Returns a backward iterator for queue elements.
         *
         *  The returned iterator behaves as if a snapshot of the
         *  queue has been taken. In particular, removing elements from
         *  the queue after an iterator has been created has no effect
         *  on the iterator, which will still iterate over the entire
         *  list of queue elements as they were at the time when the
         *  iterator was created.
         *
         *  @return             A backward iterator over all queue elements.
         *  @since 1.0
         */
        Iterator<T>             enumerateElementsTailToHead() const;

        //////////
        //  Implementation
    private:
        _impl::_QueueImpl<T> *  _pImpl;
        void                    _ensureNonsharedImplementation();
    };

    template<class T>
    Queue<T>::Queue()
    {
        _pImpl = new _impl::_QueueImpl<T>;
        _pImpl->incrementReferenceCount();
    }

    template<class T>
    Queue<T>::Queue(const Queue<T> & src)
    {
        _pImpl = src._pImpl;
        _pImpl->incrementReferenceCount();
    }

    template<class T>
    Queue<T>::~Queue()
    {
        _pImpl->decrementReferenceCount();
        if (_pImpl->getReferenceCount() == 0)
        {
            delete _pImpl;
        }
    }

    template<class T>
    Queue<T> & Queue<T>::operator = (const Queue & src)
    {
        if (this != (&src))
        {
            _pImpl->decrementReferenceCount();
            if (_pImpl->getReferenceCount() == 0)
            {
                delete _pImpl;
            }
            _pImpl = src._pImpl;
            _pImpl->incrementReferenceCount();
        }
        return *this;
    }

    template<class T>
    bool Queue<T>::isEmpty() const
    {
        return _pImpl->isEmpty();
    }

    template<class T>
    UIntSize Queue<T>::size() const
    {
        return _pImpl->size();
    }

    template<class T>
    void Queue<T>::enqueue(const T & value)
    {
        _ensureNonsharedImplementation();
        return _pImpl->enqueue(value);
    }

    template<class T>
    T Queue<T>::dequeue()
    {
        _ensureNonsharedImplementation();
        return _pImpl->dequeue();
    }

    template<class T>
    T Queue<T>::front() const
    {
        return _pImpl->front();
    }

    template<class T>
    T & Queue<T>::front()
    {
        _ensureNonsharedImplementation();
        return _pImpl->front();
    }

    template<class T>
    void Queue<T>::removeAll()
    {
        _ensureNonsharedImplementation();
        _pImpl->removeAll();
    }

    template<class T>
    Iterator<T> Queue<T>::enumerateElementsHeadToTail() const
    {
        return _pImpl->enumerateElementsHeadToTail();
    }

    template<class T>
    Iterator<T> Queue<T>::enumerateElementsTailToHead() const
    {
        return _pImpl->enumerateElementsTailToHead();
    }

    template<class T>
    void Queue<T>::_ensureNonsharedImplementation()
    {
        if (_pImpl->getReferenceCount() != 1)
        {
            _impl::_QueueImpl<T> * pNewImpl = new _impl::_QueueImpl<T>(*_pImpl);
            _pImpl->decrementReferenceCount();
            pNewImpl->incrementReferenceCount();
            _pImpl = pNewImpl;
        }
    }
}

#endif  //  __CARL_Queue_hpp_included__
//  End of CARL_Queue.hpp
