//
//  CARL_Random.hpp - Random number generator.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.9 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines a random number generator.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Random.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Random_hpp_included__
#define __CARL_Random_hpp_included__

namespace carl
{
    /**
     *  @class Random CARL.hpp
     *  The random number generator.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Random
    {
        //////////
        //  Can't assign or copy construct
    private:
        Random(const Random &);
        void                        operator = (const Random &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  The newly constructed random number generator is initialized
         *  with a seed dependent on exact current time.
         *
         *  @since 1.0
         */
        Random() { _seed = static_cast<UInt32>(TimeStamp::now().getNanosecondsSinceEpoch() / 1000000); }

        /**
         *  The class constructor.
         *
         *  The newly constructed random number generator is initialized
         *  with the specified seed. Two random number generators initialized
         *  with the same seed are guaranteed to produce the same sequences
         *  of pseudorandom numbers.
         *
         *  @param seed             The seed to intiialize a random number
         *                          generator with.
         *  @since 1.0
         */
        explicit Random(UInt32 seed) { _seed = seed; }
        ~Random() {}

        //////////
        //  Operations
    public:
        /**
         *  Reinitializes the random number generator with the specified seed.
         *  
         *  Two random number generators initialized with the same seed are 
         *  guaranteed to produce the same sequences of pseudorandom numbers.
         *
         *  @param seed             The seed to reintiialize a random number
         *                          generator with.
         *  @since 1.0
         */
        void                        seed(UInt32 seed) { _seed = seed; }

        /**
         *  Returns a randomly generated value.
         *
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        Int8                        nextInt8();

        /**
         *  Returns a randomly generated value.
         *
         *  The returned value is guaranteed to be within the specified
         *  boundaries (i.e. lowerLimit <= generated value <= upperLimit).
         *
         *  @param lowerLimit       The lower limit of the generated value range.
         *  @param upperLimit       The upper limit of the generated value range.
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        Int8                        nextInt8(Int8 lowerLimit, Int8 upperLimit);

        /**
         *  Returns a randomly generated value.
         *
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        Int16                       nextInt16();

        /**
         *  Returns a randomly generated value.
         *
         *  The returned value is guaranteed to be within the specified
         *  boundaries (i.e. lowerLimit <= generated value <= upperLimit).
         *
         *  @param lowerLimit       The lower limit of the generated value range.
         *  @param upperLimit       The upper limit of the generated value range.
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        Int16                       nextInt16(Int16 lowerLimit, Int16 upperLimit);

        /**
         *  Returns a randomly generated value.
         *
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        Int32                       nextInt32();

        /**
         *  Returns a randomly generated value.
         *
         *  The returned value is guaranteed to be within the specified
         *  boundaries (i.e. lowerLimit <= generated value <= upperLimit).
         *
         *  @param lowerLimit       The lower limit of the generated value range.
         *  @param upperLimit       The upper limit of the generated value range.
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        Int32                       nextInt32(Int32 lowerLimit, Int32 upperLimit);

        /**
         *  Returns a randomly generated value.
         *
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        Int64                       nextInt64();

        /**
         *  Returns a randomly generated value.
         *
         *  The returned value is guaranteed to be within the specified
         *  boundaries (i.e. lowerLimit <= generated value <= upperLimit).
         *
         *  @param lowerLimit       The lower limit of the generated value range.
         *  @param upperLimit       The upper limit of the generated value range.
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        Int64                       nextInt64(Int64 lowerLimit, Int64 upperLimit);

        /**
         *  Returns a randomly generated value.
         *
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        UInt8                       nextUInt8();

        /**
         *  Returns a randomly generated value.
         *
         *  The returned value is guaranteed to be within the specified
         *  boundaries (i.e. lowerLimit <= generated value <= upperLimit).
         *
         *  @param lowerLimit       The lower limit of the generated value range.
         *  @param upperLimit       The upper limit of the generated value range.
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        UInt8                       nextUInt8(UInt8 lowerLimit, UInt8 upperLimit);

        /**
         *  Returns a randomly generated value.
         *
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        UInt16                      nextUInt16();

        /**
         *  Returns a randomly generated value.
         *
         *  The returned value is guaranteed to be within the specified
         *  boundaries (i.e. lowerLimit <= generated value <= upperLimit).
         *
         *  @param lowerLimit       The lower limit of the generated value range.
         *  @param upperLimit       The upper limit of the generated value range.
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        UInt16                      nextUInt16(UInt16 lowerLimit, UInt16 upperLimit);

        /**
         *  Returns a randomly generated value.
         *
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        UInt32                      nextUInt32();

        /**
         *  Returns a randomly generated value.
         *
         *  The returned value is guaranteed to be within the specified
         *  boundaries (i.e. lowerLimit <= generated value <= upperLimit).
         *
         *  @param lowerLimit       The lower limit of the generated value range.
         *  @param upperLimit       The upper limit of the generated value range.
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        UInt32                      nextUInt32(UInt32 lowerLimit, UInt32 upperLimit);

        /**
         *  Returns a randomly generated value.
         *
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        UInt64                      nextUInt64();

        /**
         *  Returns a randomly generated value.
         *
         *  The returned value is guaranteed to be within the specified
         *  boundaries (i.e. lowerLimit <= generated value <= upperLimit).
         *
         *  @param lowerLimit       The lower limit of the generated value range.
         *  @param upperLimit       The upper limit of the generated value range.
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        UInt64                      nextUInt64(UInt64 lowerLimit, UInt64 upperLimit);

        /**
         *  Returns a randomly generated value.
         *
         *  @param numberOfBits     The size, in bits, of the random number to generate.
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        BinaryInteger               nextBinaryInteger(UIntSize numberOfBits);

        /**
         *  Returns a randomly generated value.
         *
         *  The returned value is guaranteed to be within the specified
         *  boundaries (i.e. lowerLimit <= generated value <= upperLimit).
         *
         *  @param lowerLimit       The lower limit of the generated value range.
         *  @param upperLimit       The upper limit of the generated value range.
         *  @return                 The randomly generated value.
         *  @since 1.0
         */
        BinaryInteger               nextBinaryInteger(const BinaryInteger & lowerLimit, const BinaryInteger & upperLimit);

        /**
         *  Generates a random sequence of bytes of the specified size.
         *
         *  @param pBuffer          A pointer to the buffer where randomly generated
         *                          bytes will be placed. The buffer is assumed to be
         *                          at least \a bytesToStore bytes long; if it is longer
         *                          then only the first \a bytesToStore bytes are used.
         *  @param bytesToStore     The number of random bytes to generate and store.
         *  @since 1.0
         */
        void                        nextBytes(void * pBuffer, UIntSize bytesToStore);

        //////////
        //  Implementation
    private:
        UInt32                      _seed;

        //  Helpers
        UInt32                      _next();
    };
}

#endif  //  __CARL_Random_hpp_included__
//  End of CARL_Random.hpp
