//
//  CARL_Resource.hpp - Resource framework
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/08/06 16:15:54 $
//  $Revision: 1.21 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL resource framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Resource.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Resource_hpp_included__
#define __CARL_Resource_hpp_included__

namespace carl
{
    /**
     *  @class ResourceFactory CARL.hpp
     *  An abstract resource factory.
     *
     *  A resource factory is a strategy object that can create locale-sensitive
     *  data on demand. Currently these locale-sensitive data can be:
     *
     *  -   Strings.
     *
     *  Each resource that can be created by a resource factory is uniquely
     *  identified by a string key.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC ResourceFactory
    {
        //////////
        //  Can't assign or copy construct
    private:
        ResourceFactory(const ResourceFactory &);
        void                        operator = (const ResourceFactory &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        ResourceFactory() {}

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~ResourceFactory() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns a list of all locales supported by this resource factory.
         *
         *  A resource factory keeps track of the locales it supports. The set of
         *  locales supported by a resource factory can be interrogated by client.
         *
         *  The elements in the returned list of locales are not guaranteed
         *  to be in any particular order.
         *
         *  @return                 A list of all locales supported by this 
         *                          resource factory.
         *  @since 1.0
         */
        virtual Array<Locale>       getSupportedLocales() = 0;

        /**
         *  Creates and return a string resource for the current default locale.
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const char * pKey) = 0;

        /**
         *  Creates and return a string resource for the current default locale.
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const Char * pKey) = 0;

        /**
         *  Creates and return a string resource for the current default locale.
         *
         *  @param key              The resource key.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const String & key) = 0;

        /**
         *  Creates and return a string resource for the specified locale.
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const char * pKey, const Locale & forLocale) = 0;

        /**
         *  Creates and return a string resource for the specified locale.
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const Char * pKey, const Locale & forLocale) = 0;

        /**
         *  Creates and return a string resource for the specified locale.
         *
         *  @param key              The resource key.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const String & key, const Locale & forLocale) = 0;

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param params           A list of parameters that shall be used to replace
         *                          parameter placeholders within the string resource.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const Array<String> & params);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param params           A list of parameters that shall be used to replace
         *                          parameter placeholders within the string resource.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const Array<String> & params, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @param param9           The 9th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8, const String & param9);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @param param9           The 9th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8, const String & param9, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param params           A list of parameters that shall be used to replace
         *                          parameter placeholders within the string resource.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const Array<String> & params);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param params           A list of parameters that shall be used to replace
         *                          parameter placeholders within the string resource.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const Array<String> & params, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @param param9           The 9th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8, const String & param9);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @param param9           The 9th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const Char * pKey, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8, const String & param9, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param params           A list of parameters that shall be used to replace
         *                          parameter placeholders within the string resource.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const Array<String> & params);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param params           A list of parameters that shall be used to replace
         *                          parameter placeholders within the string resource.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const Array<String> & params, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8, const Locale & forLocale);

        /**
         *  Creates and return a parameterized string resource for the current default locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @param param9           The 9th parameter.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8, const String & param9);

        /**
         *  Creates and return a parameterized string resource for the specified locale.
         *
         *  The string resource identified by the \a key can have within it placeholders
         *  that identify places where parameters shall be inserted. These placeholders
         *  are written as $0 .. $9, allowing a string resource to have up to 10 different
         *  parameters. If a dollar character $ must appear literally within the
         *  string resource, it shall be written twice $$.
         *
         *  When parameter injection is performed, all occurrences of $0 within the
         *  string resource are replaced by the 0th parameter, all occurences of $1
         *  with 1st parameter, and so on. If a placeholder refers to a nonexistent
         *  parameter (e.g. placeholder $9 when there is only 1 parameter), it is
         *  replaced with an empty string.
         *
         *  All parameter replacement is performed simultaneously (i.e. if a parameter
         *  has a substring $0 within itself, that substring is not replaced with the
         *  0th parameter).
         *
         *  @param key              The resource key.
         *  @param param0           The 0th parameter.
         *  @param param1           The 1st parameter.
         *  @param param2           The 2nd parameter.
         *  @param param3           The 3rd parameter.
         *  @param param4           The 4th parameter.
         *  @param param5           The 5th parameter.
         *  @param param6           The 6th parameter.
         *  @param param7           The 7th parameter.
         *  @param param8           The 8th parameter.
         *  @param param9           The 9th parameter.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale, with
         *                          parameter placeholders replaced by corresponding
         *                          parameters.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        String                      getParameterizedString(const String & key, const String & param0, const String & param1, const String & param2, const String & param3, const String & param4, const String & param5, const String & param6, const String & param7, const String & param8, const String & param9, const Locale & forLocale);
    };

    /**
     *  @class MissingResourceError CARL.hpp
     *  An exception thrown when a nonexistent resource is being accessed.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC MissingResourceError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  @param resourceKey      The resource key referring to a nonexistent resource.
         *  @since 1.0
         */
        explicit MissingResourceError(const String & resourceKey);

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        MissingResourceError(const MissingResourceError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~MissingResourceError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        MissingResourceError &      operator = (const MissingResourceError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the resource key referring to a nonexistent resource.
         *
         *  @return                 The resource key referring to a nonexistent resource.
         *  @since 1.0
         */
        String                      getResourceKey() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String              toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String              toString(const Locale & forLocale) const;

        //////////
        //  Implementation
    private:
        String                      _resourceKey;
    };

    /**
     *  @class ResourceMap CARL.hpp
     *  Provides a key to value mapping for a single locale.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC ResourceMap
    {
        //////////
        //  Can't assign or copy construct
    private:
        ResourceMap(const ResourceMap &);
        void                        operator = (const ResourceMap &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        ResourceMap() {}

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~ResourceMap() {}

        //////////
        //  Operations
    public:
        /**
         *  Creates and return a string resource.
         *
         *  @param key              The resource key.
         *  @return                 A string resource.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const String & key) = 0;
    };

    /**
     *  @class MapResourceFactory CARL.hpp
     *  A resource factory that aggregates one or more resource maps.
     *
     *  A map resource factory aggregates one or more resource maps, each
     *  of which is tagged with a locale. When a resource is sought for a
     *  specific locale, the resource map that most closely matches that
     *  locale is used to locate the resource.
     *
     *  One of the aggregated resource maps is tagged as a default resource
     *  map. It is used as a last-resort resource map, to provide resources
     *  for all locales that are not directly supported.
     *
     *  To locate a resource map for a givel locale L the following
     *  algorithm is used:
     *
     *  -   If the resource map for locale L exists, it is used.
     *  -   Otherwise resource maps for L's parents are examined, in
     *      the bottom-up order, up to but not including the invariant
     *      locale. If one of these resource maps exists, it is used.
     *  -   Otherwise, a default resource map is used.
     *
     *  The map resource factory has full ownership of all aggregated
     *  resource map, which are destroyed when a map resource factory
     *  is destroyed.
     *
     *  It is permitted to associate the same resource map with several
     *  different locales under the same resource factory.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC MapResourceFactory : public ResourceFactory
    {
        //////////
        //  Can't assign or copy construct
    private:
        MapResourceFactory(const MapResourceFactory &);
        void                        operator = (const MapResourceFactory &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  The newly constructed map resource factory initially aggregated
         *  only the specified resource map, which is set up as a default
         *  resource map and tagged with the current default locale.
         *
         *  @param pDefaultResourceMap  The resource map to use for the current
         *                          default locale and as a default resource map.
         *  @since 1.0
         */
        explicit MapResourceFactory(ResourceMap * pDefaultResourceMap);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~MapResourceFactory();

        //////////
        //  Operations
    public:
        /**
         *  Returns a list of all locales supported by this resource factory.
         *
         *  A resource factory keeps track of the locales it supports. The set of
         *  locales supported by a resource factory can be interrogated by client.
         *
         *  The elements in the returned list of locales are not guaranteed
         *  to be in any particular order.
         *
         *  @return                 A list of all locales supported by this 
         *                          resource factory.
         *  @since 1.0
         */
        virtual Array<Locale>       getSupportedLocales();

        /**
         *  Creates and return a string resource for the current default locale.
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const char * pKey);

        /**
         *  Creates and return a string resource for the current default locale.
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const Char * pKey);

        /**
         *  Creates and return a string resource for the current default locale.
         *
         *  @param key              The resource key.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const String & key);

        /**
         *  Creates and return a string resource for the specified locale.
         *
         *  @param pKey             The pointer to a nul-terminated ASCII resource key.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const char * pKey, const Locale & forLocale);

        /**
         *  Creates and return a string resource for the specified locale.
         *
         *  @param pKey             The pointer to a nul-terminated ISO-10646 resource key.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const Char * pKey, const Locale & forLocale);

        /**
         *  Creates and return a string resource for the specified locale.
         *
         *  @param key              The resource key.
         *  @param forLocale        The locale for which a string resource is sought.
         *  @return                 A string resource for the current default locale.
         *  @exception MissingResourceError If the specified key is not recognized
         *                          by the resource factory.
         *  @since 1.0
         */
        virtual String              getString(const String & key, const Locale & forLocale);

        /**
         *  Adds a locale-specific resource map to the map resource factory.
         *
         *  The specified resource map is aggregated under this map resource
         *  factory and tagged with the specified locale. If an aggregated
         *  resource map for the specified locale already existed before the call,
         *  it is dis-associated from the specified locale; if that means that
         *  the existing resource map loses all its associated locales it is
         *  destroyed. If the existing resource map was the default resource
         *  map, the newly added resource map becomes a new default.
         *
         *  @param forLocale        The locale for which the newly added resource
         *                          map will provide resources.
         *  @param pResourceMap     The resource map to add.
         *  @since 1.0
         */
        void                        addResourceMap(const Locale & forLocale, ResourceMap * pResourceMap);

        //////////
        //  Overridables
    public:
        /**
         *  Represents an outcome of an attempt to handle a missing locale.
         *
         *  @since 1.0
         */
        enum HandleMissingLocaleOutcome
        {
            /**
             *  The resource factory shall re-try locating a resource.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            HandleMissingLocaleOutcome_Retry        = 1,

            /**
             *  The resource factory shall stop trying to locate a resource and fail.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            HandleMissingLocaleOutcome_Abort        = 2,

            /**
             *  Use default behaviour to locate a resource.
             *
             *  The default behaviour is:
             *
             *  -   Check tha parent locale of the one that is missing,
             *      and the parent of that, up to but not including the
             *      invariant locale. If one of these resource maps exists, 
             *      it is used.
             *  -   Check the current default locale and its parents up to 
             *      but not including the invariant locale. If one of these 
             *      resource maps exists,  it is used.
             *  -   Otherwise use the default resource map.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            HandleMissingLocaleOutcome_UseDefault   = 3
        };

    protected:
        /**
         *  Called by framework when a resource is sought for locale for
         *  which no aggregated resource map exists.
         *
         *  @param forLocale        The locale for which the newly added resource
         *                          map will provide resources.
         *  @return                 The code specifying how the resource location
         *                          shall proceed.
         *  @since 1.0
         */
        virtual HandleMissingLocaleOutcome  handleMissingLocale(const Locale & forLocale);

        //////////
        //  Implementation
    private:
        ResourceMap *               _pDefaultResourceMap;
        HashMap<Locale, ResourceMap*>   _mapLocalesToResourceMaps;
    };

    /**
     *  @class FileResourceFactory CARL.hpp
     *  A map resource factory that loads missing resources from disk files on demand.
     *
     *  A file resource factory is backed up by zero or more resource files, each
     *  resource file having the name of the form:
     *
     *  &lt;base&gt;[_&lt;language&gt;[_&lt;country&gt;[_&lt;variant&gt;]]].&lt;extension&gt;
     *
     *  where:
     *
     *  -   &lt;base&gt; and &lt;extension&gt; specify the filename portions common
     *      for all resource files that are used by this file resource factory.
     *  -   &lt;language&gt; is a lowercase 2-letter ISO-639 language code.
     *  -   &lt;country&gt; is an uppercase 2-letter ISO-3166 country code.
     *  -   &lt;variant&gt; is an arbitrary string representing the locale variant.
     *
     *  The &lt;base&gt; and &lt;extension&gt; are specified when the file resource
     *  factory is being constructed. A list of search paths is also designated
     *  at the moment of construction.
     *
     *  All resource files are expected to be UTF-8 - encoded text files containing
     *  zero or more lines of the form:
     *
     *  &lt;key&gt; = &lt;value&gt; [ ;&lt;comment&gt; ]
     *
     *  where both &lt;key&gt; and &lt;value&gt; must be string constants. These use the same
     *  syntax as C constants, except:
     *
     *  -   2- and 4-byte escape sequences are allowed to define nonprintable 16- and 
     *      32-bit characters: "\u1234" and "\i12345678".
     *  -   The "\e" escape sequence means 0x00000027 (escape character).
     *
     *  A line in the resource file can have any number of continuation lines.
     *  In this case the last character of the line must be a backslash '\', and
     *  the continuation line must also have a backslash in its leftmost non-space
     *  position. Both backslashes are ignored, as well as all spaces preceeding the
     *  backslash in the continuation line.
     *  
     *  The default resources must still be provided for the file resource factory,
     *  because it may well happen that none of the resource files exist.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC FileResourceFactory : public MapResourceFactory
    {
        //////////
        //  Can't assign or copy construct
    private:
        FileResourceFactory(const FileResourceFactory &);
        void                        operator = (const FileResourceFactory &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  @param pDefaultResourceMap  The resource map to use for the current
         *                          default locale and as a default resource map.
         *  @param resourceFileNameBase The &lt;base&gt; portion of resource file names.
         *  @param resourceFileNameExtension    The &lt;extension&gt; portion of resource 
         *                          file names.
         *  @param searchPathList   A list of directories where resource files
         *                          are searched for.
         *  @since 1.0
         */
        FileResourceFactory(ResourceMap * pDefaultResourceMap, const String & resourceFileNameBase, const String & resourceFileNameExtension, const Array<String> & searchPathList);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~FileResourceFactory() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns a list of all locales supported by this resource factory.
         *
         *  A resource factory keeps track of the locales it supports. The set of
         *  locales supported by a resource factory can be interrogated by client.
         *
         *  The elements in the returned list of locales are not guaranteed
         *  to be in any particular order.
         *
         *  If this file resource factory had not yet located and loaded all
         *  existing resource files, these are located and loaded. Therefore,
         *  the returned list of locales reflects on which resource files
         *  could be found.
         *
         *  @return                 A list of all locales supported by this 
         *                          resource factory.
         *  @since 1.0
         */
        virtual Array<Locale>       getSupportedLocales();

        //////////
        //  Overridables
    protected:
        /**
         *  Called by framework when a resource is sought for locale for
         *  which no aggregated resource map exists.
         *
         *  If the resource file for the specified locale exists but has not
         *  yet been loaded, it is loaded and resource search re-tried.
         *  Otherwise the default behaviour is used.
         *
         *  @param forLocale        The locale for which the newly added resource
         *                          map will provide resources.
         *  @return                 The code specifying how the resource location
         *                          shall proceed.
         *  @since 1.0
         */
        virtual HandleMissingLocaleOutcome  handleMissingLocale(const Locale & forLocale);

        //////////
        //  Implementation
    private:
        bool                        _resourcesPreloaded;
        String                      _requiredResourceFileNamePrefix;
        String                      _requiredResourceFileNameSuffix;
        Array<String>               _searchPathList;
        HashSet<Locale>             _supportedLocales;
        HashMap<Locale, String>     _mapSupportedLocalesToDeferredResourceFileNames;

        //  Resource map loadable from file
        class CARL_PUBLIC _ResourceMap : public ResourceMap
        {
            //////////
            //  Can't assign or copy construct
        private:
            _ResourceMap(const _ResourceMap &);
            void                    operator = (const _ResourceMap &);

            //////////
            //  Construction/destruction
        public:
            explicit _ResourceMap(const HashMap<String, String> & mapKeysToStrings);
            virtual ~_ResourceMap() {}

            //////////
            //  Operations
        public:
            virtual String          getString(const String & key);

            //////////
            //  Implementation
        private:
            HashMap<String, String> _mapKeysToStrings;
        };

        //  Helpers
        void                        _ensureResourcesPreloaded();
        _ResourceMap *              _loadResourceFile(const String & fileName);
        bool                        _readLine(ITextInput & ti, String & line) throw(IOException);
        void                        _skipSpaces(const String & line, UIntSize & scan);
        bool                        _fetchStringConstant(const String & line, UIntSize & scan, String & value);
        bool                        _fetchCharacterLiteral(const String & line, UIntSize & scan, Char & value);
    };

    /**
     *  @class FrameworkResourceFactory CARL.hpp
     *  A file resource factory used for locating internal CARL resources.
     *
     *  Framework resource factories are used to provide all resources used internally
     *  by CARL library. They behave like "normal" file resource factories, with the
     *  following restrictions:
     *
     *  -   The &lt;extension&gt; portion of all resource file names is always ".res".
     *  -   The list of directories where resource files are searched for is
     *      obtained from the CARL_1_0_RESOURCES environment variable.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC FrameworkResourceFactory : public FileResourceFactory
    {
        //////////
        //  Can't assign or copy construct
    private:
        FrameworkResourceFactory(const FrameworkResourceFactory &);
        void                        operator = (const FrameworkResourceFactory &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  @param pDefaultResourceMap  The resource map to use for the current
         *                          default locale and as a default resource map.
         *  @param resourceFileNameBase The &lt;base&gt; portion of resource file names.
         *  @since 1.0
         */
        FrameworkResourceFactory(ResourceMap * pDefaultResourceMap, const String & resourceFileNameBase);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~FrameworkResourceFactory() {}

        //////////
        //  Implementation
    private:
        static Array<String>        _determineResourceSearchPaths();
    };
}

#endif  //  __CARL_Resource_hpp_included__
//  End of CARL_Resource.hpp
