//
//  CARL_Search.hpp - Searching framework.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/08/03 01:15:33 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL searching framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Search.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Search_hpp_included__
#define __CARL_Search_hpp_included__

namespace carl
{
    /**
     *  @class LinearSearch CARL.hpp
     *  A generic algorithm for searching an unordered sequence.
     *
     *  @param T                    A type of an element of a searched sequence.
     *  @since 1.0
     */
    template <class T>
    class LinearSearch
    {
        //////////
        //  Utility classes cannot be instantiated
    private:
        LinearSearch();
        LinearSearch(const LinearSearch<T> &);
        ~LinearSearch();
        void                        operator = (const LinearSearch<T> &);

        //////////
        //  Operations
    public:
        /**
         *  Finds the first matching element within a sequence.
         *
         *  @param pItems           A pointer to the first element of the
         *                          sequence to search. The sequence is expected
         *                          to be at least \a itemCount elements long; if
         *                          it is longer than only the first \a itemCount
         *                          elements are searched.
         *  @param itemCount        The number of element in a searched sequence.
         *  @param value            The value to search for,
         *  @param comparator       The comparator to use for establishing whether
         *                          two values are equal or not.
         *  @return                 A pointer to the first matching element within the
         *                          searched sequence, or NULL if there is no match.
         *  @since 1.0
         */
        static const T *            findFirst(const T * pItems, UIntSize itemCount, const T & value, const IComparator<T> & comparator = NaturalComparator<T>());

        /**
         *  Finds the last matching element within a sequence.
         *
         *  @param pItems           A pointer to the first element of the
         *                          sequence to search. The sequence is expected
         *                          to be at least \a itemCount elements long; if
         *                          it is longer than only the first \a itemCount
         *                          elements are searched.
         *  @param itemCount        The number of element in a searched sequence.
         *  @param value            The value to search for,
         *  @param comparator       The comparator to use for establishing whether
         *                          two values are equal or not.
         *  @return                 A pointer to the last matching element within the
         *                          searched sequence, or NULL if there is no match.
         *  @since 1.0
         */
        static const T *            findLast(const T * pItems, UIntSize itemCount, const T & value, const IComparator<T> & comparator = NaturalComparator<T>());

        /**
         *  Finds all matching elements within a sequence.
         *
         *  @param pItems           A pointer to the first element of the
         *                          sequence to search. The sequence is expected
         *                          to be at least \a itemCount elements long; if
         *                          it is longer than only the first \a itemCount
         *                          elements are searched.
         *  @param itemCount        The number of element in a searched sequence.
         *  @param value            The value to search for,
         *  @param comparator       The comparator to use for establishing whether
         *                          two values are equal or not.
         *  @return                 A list of pointers to all matching elements within
         *                          the searched sequence. If there is no match, the
         *                          return value is an empty array; otherwise elements
         *                          of the returned array appear in the same relative
         *                          order in which they appear in the searched sequence.
         *  @since 1.0
         */
        static Array<const T*>      findAll(const T * pItems, UIntSize itemCount, const T & value, const IComparator<T> & comparator = NaturalComparator<T>());
    };

    template <class T>
    const T * LinearSearch<T>::findFirst(const T * pItems, UIntSize itemCount, const T & value, const IComparator<T> & comparator /*= NaturalComparator<T>()*/)
    {
        require(pItems != NULL || itemCount == 0);

        for (UIntSize i = 0; i < itemCount; i++)
        {
            if (comparator.equal(pItems[i], value))
            {
                return pItems + i;
            }
        }
        return NULL;
    }

    template <class T>
    const T * LinearSearch<T>::findLast(const T * pItems, UIntSize itemCount, const T & value, const IComparator<T> & comparator /*= NaturalComparator<T>()*/)
    {
        require(pItems != NULL || itemCount == 0);

        for (UIntSize i = itemCount; i != 0; i--)
        {
            if (comparator.equal(pItems[i - 1], value))
            {
                return pItems + i - 1;
            }
        }
        return NULL;
    }

    template <class T>
    Array<const T*> LinearSearch<T>::findAll(const T * pItems, UIntSize itemCount, const T & value, const IComparator<T> & comparator /*= NaturalComparator<T>()*/)
    {
        require(pItems != NULL || itemCount == 0);

        Array<const T*> result;
        for (UIntSize i = 0; i < itemCount; i++)
        {
            if (comparator.equal(pItems[i], value))
            {
                result.append(pItems + i);
            }
        }
        return result;
    }

    /**
     *  @class BinarySearch CARL.hpp
     *  A generic algorithm for searching an ordered sequence.
     *
     *  For all services provided by this utility class, a sequence being 
     *  searched must be arranged so that each element of a sequence is
     *  not "greater" that any of the subsequent elements according to
     *  the comparator being used.
     *
     *  @param T                    A type of an element of a searched sequence.
     *  @since 1.0
     */
    template <class T>
    class BinarySearch
    {
        //////////
        //  Utility classes cannot be instantiated
    private:
        BinarySearch();
        BinarySearch(const BinarySearch<T> &);
        ~BinarySearch();
        void                        operator = (const BinarySearch<T> &);

        //////////
        //  Operations
    public:
        /**
         *  Finds the first matching element within a sequence.
         *
         *  @param pItems           A pointer to the first element of the
         *                          sequence to search. The sequence is expected
         *                          to be at least \a itemCount elements long; if
         *                          it is longer than only the first \a itemCount
         *                          elements are searched.
         *  @param itemCount        The number of element in a searched sequence.
         *  @param value            The value to search for,
         *  @param comparator       The comparator to use for establishing whether
         *                          two values are equal or not.
         *  @return                 A pointer to the first matching element within the
         *                          searched sequence, or NULL if there is no match.
         *  @since 1.0
         */
        static const T *            findFirst(const T * pItems, UIntSize itemCount, const T & value, const IOrderedComparator<T> & comparator = NaturalOrderedComparator<T>());

        /**
         *  Finds the last matching element within a sequence.
         *
         *  @param pItems           A pointer to the first element of the
         *                          sequence to search. The sequence is expected
         *                          to be at least \a itemCount elements long; if
         *                          it is longer than only the first \a itemCount
         *                          elements are searched.
         *  @param itemCount        The number of element in a searched sequence.
         *  @param value            The value to search for,
         *  @param comparator       The comparator to use for establishing whether
         *                          two values are equal or not.
         *  @return                 A pointer to the last matching element within the
         *                          searched sequence, or NULL if there is no match.
         *  @since 1.0
         */
        static const T *            findLast(const T * pItems, UIntSize itemCount, const T & value, const IOrderedComparator<T> & comparator = NaturalOrderedComparator<T>());

        /**
         *  Finds all matching elements within a sequence.
         *
         *  @param pItems           A pointer to the first element of the
         *                          sequence to search. The sequence is expected
         *                          to be at least \a itemCount elements long; if
         *                          it is longer than only the first \a itemCount
         *                          elements are searched.
         *  @param itemCount        The number of element in a searched sequence.
         *  @param value            The value to search for,
         *  @param comparator       The comparator to use for establishing whether
         *                          two values are equal or not.
         *  @return                 A list of pointers to all matching elements within
         *                          the searched sequence. If there is no match, the
         *                          return value is an empty array; otherwise elements
         *                          of the returned array appear in the same relative
         *                          order in which they appear in the searched sequence.
         *  @since 1.0
         */
        static Array<const T*>      findAll(const T * pItems, UIntSize itemCount, const T & value, const IOrderedComparator<T> & comparator = NaturalOrderedComparator<T>());

        //////////
        //  Implementation
    private:
        static bool                 _isOrdered(const T * pItems, UIntSize itemCount, const IOrderedComparator<T> & comparator);
        static const T *            _findAny(const T * pItems, UIntSize itemCount, const T & value, const IOrderedComparator<T> & comparator);
    };

    template <class T>
    BinarySearch<T>::BinarySearch()
    {
    }

    template <class T>
    BinarySearch<T>::~BinarySearch()
    {
    }

    template <class T>
    const T * BinarySearch<T>::findFirst(const T * pItems, UIntSize itemCount, const T & value, const IOrderedComparator<T> & comparator /*= NaturalOrderedComparator<T>()*/)
    {
        require(pItems != NULL || itemCount == 0);
        require(_isOrdered(pItems, itemCount, comparator));

        const T * pSomeItem = _findAny(pItems, itemCount, value, comparator);
        if (pSomeItem != NULL)
        {
            UIntSize firstItemIndex = static_cast<UIntSize>(pSomeItem - pItems);
            while (firstItemIndex != 0 && comparator.equal(pItems[firstItemIndex - 1], *pSomeItem))
            {
                firstItemIndex--;
            }
            pSomeItem = pItems + firstItemIndex;
        }
        return pSomeItem;
    }

    template <class T>
    const T * BinarySearch<T>::findLast(const T * pItems, UIntSize itemCount, const T & value, const IOrderedComparator<T> & comparator /*= NaturalOrderedComparator<T>()*/)
    {
        require(pItems != NULL || itemCount == 0);
        require(_isOrdered(pItems, itemCount, comparator));

        const T * pSomeItem = _findAny(pItems, itemCount, value, comparator);
        if (pSomeItem != NULL)
        {
            UIntSize lastItemIndex = static_cast<UIntSize>(pSomeItem - pItems);
            while (lastItemIndex != itemCount - 1 && comparator.equal(pItems[lastItemIndex + 1], *pSomeItem))
            {
                lastItemIndex++;
            }
            pSomeItem = pItems + lastItemIndex;
        }
        return pSomeItem;
    }

    template <class T>
    Array<const T*> BinarySearch<T>::findAll(const T * pItems, UIntSize itemCount, const T & value, const IOrderedComparator<T> & comparator /*= NaturalOrderedComparator<T>()*/)
    {
        require(pItems != NULL || itemCount == 0);
        require(_isOrdered(pItems, itemCount, comparator));

        Array<const T*> result;
        const T * pSomeItem = _findAny(pItems, itemCount, value, comparator);
        if (pSomeItem != NULL)
        {
            UIntSize firstItemIndex = static_cast<UIntSize>(pSomeItem - pItems);
            while (firstItemIndex != 0 && comparator.equal(pItems[firstItemIndex - 1], *pSomeItem))
            {
                firstItemIndex--;
            }
            UIntSize lastItemIndex = static_cast<UIntSize>(pSomeItem - pItems);
            while (lastItemIndex != itemCount - 1 && comparator.equal(pItems[lastItemIndex + 1], *pSomeItem))
            {
                lastItemIndex++;
            }
            for (UIntSize i = firstItemIndex; i <= lastItemIndex; i++)
            {
                result.append(pItems + i);
            }
        }
        return result;
    }

    template <class T>
    bool BinarySearch<T>::_isOrdered(const T * pItems, UIntSize itemCount, const IOrderedComparator<T> & comparator)
    {
        for (UIntSize i = 0; i + 1 < itemCount; i++)
        {
            if (comparator.greater(pItems[i], pItems[i + 1]))
            {
                return false;
            }
        }
        return true;
    }

    template <class T>
    const T * BinarySearch<T>::_findAny(const T * pItems, UIntSize itemCount, const T & value, const IOrderedComparator<T> & comparator)
    {
        if (itemCount == 0)
        {
            return NULL;
        }
        else
        {
            UIntSize startIndex = 0;
            UIntSize endIndex = itemCount - 1;
            for (; ; )
            {
                UIntSize midIndex = (startIndex + endIndex) / 2;
                if (comparator.equal(pItems[midIndex], value))
                {   //  item == value
                    return pItems + midIndex;
                }
                else if (comparator.less(pItems[midIndex], value))
                {   //  item < value; search upper half
                    if (midIndex + 1 > endIndex)
                    {
                        return NULL;
                    }
                    else
                    {
                        return _findAny(pItems + midIndex + 1, itemCount - midIndex - 1, value, comparator);
                    }
                }
                else
                {   //  item > value; search lower half
                    if (midIndex < startIndex + 1)
                    {
                        return NULL;
                    }
                    else
                    {
                        return _findAny(pItems, midIndex - 1, value, comparator);
                    }
                }
            }
        }
    }
}

#endif  //  __CARL_Search_hpp_included__
//  End of CARL_Search.hpp
