//
//  CARL_Set.hpp - Value sets.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:49 $
//  $Revision: 1.11 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines a set data type.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Set.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Set_hpp_included__
#define __CARL_Set_hpp_included__

namespace carl
{
    /**
     *  @class Set CARL.hpp
     *  A generic set.
     *
     *  CARL sets use implicit sharing to reduce the copy construction and 
     *  assignment overhead. This means that it is feasible to both pass
     *  sets to functions as value parameters and return sets from functions.
     *
     *   The element type (T) must be:
     *
     *   -   Default constructable (T::T() must be defined and public)
     *   -   Copy constructable (T::T(const T &) must be defined and public)
     *   -   Assignable (T::operator=(const T &) must be defined and public)
     *   -   Destructable (T::~T() must be defined and public)
     *
     *  @param T                The type of set elements.
     *  @since 1.0
     */
    template<class T>
    class Set
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        Set();

        /**
         *  The copy constructor.
         *
         *  @since 1.0
         */
        Set(const Set<T> & );

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Set();

        /**
         *  The assignment operator.
         *
         *  @return             A reference to the assigned-to set.
         *  @since 1.0
         */
        Set<T> &                operator = (const Set<T> & );

        //////////
        //  Operations
    public:
        /**
         *  Checks if a set is empty.
         *
         *  @return             True if this set is empty; false otherwise.
         *  @since 1.0
         */
        virtual bool            isEmpty() const = 0;

        /**
         *  Returns the number of elements in this set.
         *
         *  @return             The number of elements in this set.
         *  @since 1.0
         */
        virtual UIntSize        size() const = 0;

        /**
         *  Checks if the set contains the specified element.
         *
         *  @param value        The element to check for being a member of the set.
         *  @return             True if the set contains the specified element;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            contains(const T & value) const = 0;

        /**
         *  Inserts an element into the set.
         *
         *  If the element is already a member of the set, the call
         *  has no effect.
         *
         *  @param value        The element to insert.
         *  @return             True if the element has been actually inserted;
         *                      false if the element was already a member of the
         *                      set before the call.
         *  @since 1.0
         */
        virtual bool            insert(const T & value) = 0;

        /**
         *  Inserts all elements of a specified set into this set.
         *
         *  @param set          The set of elements to insert.
         *  @return             The actual number of new elements inserted.
         *  @since 1.0
         */
        virtual UIntSize        insert(const Set<T> & set) = 0;

        /**
         *  Removes an element from the set.
         *
         *  If the element is already not a member of the set, the call
         *  has no effect.
         *
         *  @param value        The element to remove.
         *  @return             True if the element has been actually removed;
         *                      false if the element was already not a member
         *                      of the set before the call.
         *  @since 1.0
         */
        virtual bool            remove(const T & value) = 0;

        /**
         *  Removes all elements of a specified set from this set.
         *
         *  @param set          The set of elements to remove.
         *  @return             The actual number of removed elements.
         *  @since 1.0
         */
        virtual UIntSize        remove(const Set<T> & set) = 0;

        /**
         *  Clears out the set.
         *
         *  @since 1.0
         */
        virtual void            removeAll() = 0;

        /**
         *  Returns an iterator over all elements of the set.
         *
         *  Elements are not guaranteed to be iterated over in any
         *  particular order.
         *
         *  @return             An iterator over all elements of the set.
         *  @since 1.0
         */
        virtual Iterator<T>     enumerateElements() const = 0;

        /**
         *  Checks it two sets are equal.
         *
         *  @param set          The set to compare to this set.
         *  @return             True if both seth contain exactly the same
         *                      elements; false otherwise.
         *  @since 1.0
         */
        virtual bool            operator == (const Set<T> & set) const = 0;

        /**
         *  Checks it two sets are unequal.
         *
         *  @param set          The set to compare to this set.
         *  @return             False if both seth contain exactly the same
         *                      elements; true otherwise.
         *  @since 1.0
         */
        virtual bool            operator != (const Set<T> & set) const = 0;

        /**
         *  Checks this set is a proper subset of another set.
         *
         *  @param set          The set to compare to this set.
         *  @return             True if this set is a proper subset of \a set;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            operator <  (const Set<T> & set) const = 0;

        /**
         *  Checks this set is a subset of another set.
         *
         *  @param set          The set to compare to this set.
         *  @return             True if this set is a subset of \a set;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            operator <= (const Set<T> & set) const = 0;

        /**
         *  Checks this set is a proper superset of another set.
         *
         *  @param set          The set to compare to this set.
         *  @return             True if this set is a proper superset of \a set;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            operator >  (const Set<T> & set) const = 0;

        /**
         *  Checks this set is a superset of another set.
         *
         *  @param set          The set to compare to this set.
         *  @return             True if this set is a superset of \a set;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            operator >= (const Set<T> & set) const = 0;

        /**
         *  Inserts an element into the set.
         *
         *  If the element is already a member of the set, the call
         *  has no effect.
         *
         *  @param value        The element to insert.
         *  @return             The reference to the inserted-to set.
         *  @since 1.0
         */
        virtual Set<T> &        operator += (const T & value) = 0;

        /**
         *  Inserts all elements of a specified set into this set.
         *
         *  @param set          The set of elements to insert.
         *  @return             The reference to the inserted-to set.
         *  @since 1.0
         */
        virtual Set<T> &        operator += (const Set<T> & set) = 0;

        /**
         *  Removes an element from the set.
         *
         *  If the element is already not a member of the set, the call
         *  has no effect.
         *
         *  @param value        The element to remove.
         *  @return             The reference to the removed-from set.
         *  @since 1.0
         */
        virtual Set<T> &        operator -= (const T & value) = 0;

        /**
         *  Removes all elements of a specified set from this set.
         *
         *  @param set          The set of elements to remove.
         *  @return             The reference to the removed-from set.
         *  @since 1.0
         */
        virtual Set<T> &        operator -= (const Set<T> & set) = 0;

        /**
         *  Intersects this set with the specified value.
         *
         *  The result of such an intersection is either an empty set (if
         *  the specified value was not a member of this set) or a set containing
         *  only one value (if the specified value was a member of this set).
         *
         *  @param value        The value to intersec tis set with.
         *  @return             The reference to the intersected set.
         *  @since 1.0
         */
        virtual Set<T> &        operator *= (const T & value) = 0;

        /**
         *  Intersects this set with the specified set.
         *
         *  All elements that are not present in the specified \a set
         *  are removed from this set.
         *
         *  @param set          The set to intersect this set with.
         *  @return             The reference to the intersected set.
         *  @since 1.0
         */
        virtual Set<T> &        operator *= (const Set<T> & set) = 0;
    };

    template<class T>
    Set<T>::Set() 
    {
    }

    template<class T>
    Set<T>::Set(const Set<T> & ) 
    {
    }

    template<class T>
    Set<T>::~Set() 
    {
    }

    template<class T>
    Set<T> & Set<T>::operator = (const Set<T> & ) 
    { 
        return *this; 
    }

    template<class T, class C>
    class HashSet;

    namespace _impl
    {
        //////////
        //  A value set implementation via a hash table.
        //
        //  The value type (T) must be:
        //  *   Default constructable (T::T() must be defined and public)
        //  *   Copy constructable (T::T(const T &) must be defined and public)
        //  *   Comparable (T::operator==(const T &) must be defined and public)
        //  *   Assignable (T::operator=(const T &) must be defined and public)
        //  *   Destructable (T::~T() must be defined and public)
        template<class T, class C>
        class _HashSetImpl
        {
            friend class HashSet<T,C>;
            //////////
            //  Types
        public:
            typedef UIntSize (CARL_CALLBACK * HashFunction)(const T & value);

            //////////
            //  Can't assign
        private:
            void                    operator = (const Set<T> & src);

            //////////
            //  Construction/destruction
        public:
            _HashSetImpl(HashFunction hashFunction, C comparator);
            _HashSetImpl(HashFunction hashFunction, UIntSize initialHashTableCapacity, C comparator);
            _HashSetImpl(const _HashSetImpl<T,C> & src);
            ~_HashSetImpl();

            //////////
            //  Operations
        public:
            bool                    isEmpty() const;
            UIntSize                size() const;
            bool                    contains(const T & value) const;
            bool                    insert(const T & value);
            UIntSize                insert(const Set<T> & set);
            bool                    remove(const T & value);
            UIntSize                remove(const Set<T> & set);
            void                    removeAll();
            Iterator<T>             enumerateElements() const;
            bool                    operator == (const Set<T> & set) const;
            bool                    operator != (const Set<T> & set) const;
            bool                    operator <  (const Set<T> & set) const;
            bool                    operator <= (const Set<T> & set) const;
            bool                    operator >  (const Set<T> & set) const;
            bool                    operator >= (const Set<T> & set) const;
            UIntSize                getReferenceCount() const;
            void                    incrementReferenceCount();
            void                    decrementReferenceCount();

            //////////
            //  Implementation
        private:
            struct _HashTableEntry
            {
                explicit _HashTableEntry(const T & value)
                    :   _value(value)
                {
                    _pNextInChain = NULL;
                }

                T                   _value;
                _HashTableEntry *   _pNextInChain;
            };
            UIntSize                _hashTableSize;
            _HashTableEntry **      _hashTable;
            HashFunction            _hashFunction;
            C                       _comparator;
            UIntSize                _elementCount;
            UIntSize                _referenceCount;

            UIntSize                _hash(const T & value, UIntSize hashTableSize) const;
        };

        template<class T, class C>
        _HashSetImpl<T,C>::_HashSetImpl(HashFunction hashFunction, C comparator)
            :   _comparator(comparator)
        {
            require(hashFunction != NULL);

            _hashTableSize = 0;
            _hashTable = NULL;
            _hashFunction = hashFunction;
            _elementCount = 0;
            _referenceCount = 0;
        }

        template<class T, class C>
        _HashSetImpl<T,C>::_HashSetImpl(HashFunction hashFunction, UIntSize initialHashTableCapacity, C comparator)
            :   _comparator(comparator)
        {
            _hashTableSize = initialHashTableCapacity;
            if (initialHashTableCapacity == 0)
            {
                _hashTable = NULL;
            }
            else
            {
                _hashTable = new _HashTableEntry*[initialHashTableCapacity];
                for (UIntSize i = 0; i < initialHashTableCapacity; i++)
                {
                    _hashTable[i] = NULL;
                }
            }
            _hashFunction = hashFunction;
            _elementCount = 0;
            _referenceCount = 0;
        }

        template<class T, class C>
        _HashSetImpl<T,C>::_HashSetImpl(const _HashSetImpl<T,C> & src)
            :   _comparator(src._comparator)
        {
            _hashTableSize = src.size() / 3 * 4;
            if (_hashTableSize == 0)
            {
                _hashTable = NULL;
            }
            else
            {
                _hashTable = new _HashTableEntry*[_hashTableSize];
                for (UIntSize i = 0; i < _hashTableSize; i++)
                {
                    _hashTable[i] = NULL;
                }
            }
            _hashFunction = src._hashFunction;
            _elementCount = 0;
            for (Iterator<T> it = src.enumerateElements(); !it.finished(); it.advance())
            {
                insert(it.item());
            }
            _referenceCount = 0;
        }

        template<class T, class C>
        _HashSetImpl<T,C>::~_HashSetImpl()
        {
            if (_hashTable != NULL)
            {
                for (UIntSize i = 0; i < _hashTableSize; i++)
                {
                    while (_hashTable[i] != NULL)
                    {
                        _HashTableEntry * pHashTableEntry = _hashTable[i];
                        _hashTable[i] = pHashTableEntry->_pNextInChain;
                        delete pHashTableEntry;
                    }
                }
                delete [] _hashTable;
            }
        }

        template<class T, class C>
        bool _HashSetImpl<T,C>::isEmpty() const
        {
            return _elementCount == 0;
        }

        template<class T, class C>
        UIntSize _HashSetImpl<T,C>::size() const
        {
            return _elementCount;
        }

        template<class T, class C>
        bool _HashSetImpl<T,C>::contains(const T & value) const
        {
            if (_elementCount == 0)
            {
                return false;
            }
            else
            {
                UIntSize hashCode = _hash(value, _hashTableSize);
                for (_HashTableEntry * pHashTableEntry = _hashTable[hashCode];
                     pHashTableEntry != NULL;
                     pHashTableEntry = pHashTableEntry->_pNextInChain)
                {
                    if (_comparator.equal(pHashTableEntry->_value, value))
                    {
                        return true;
                    }
                }
                return false;
            }
        }

        template<class T, class C>
        bool _HashSetImpl<T,C>::insert(const T & value)
        {
            //  If the value is already in the hash table, there's nothing to do
            if (contains(value))
            {
                return false;
            }

            //  The value is not currently in the hash table
            if (_hashTable == NULL ||
                _elementCount * 4 > _hashTableSize * 3)
            {   //  Hash table is either not allocated at all, or
                //  is 75% full - need a larger one.
                //  Allocate & initialize a new hash table...
                UIntSize newHashTableSize = (3 * _hashTableSize / 2) + 1;
                _HashTableEntry ** newHashTable = new _HashTableEntry*[newHashTableSize];
                for (UIntSize i = 0; i < newHashTableSize; i++)
                {
                    newHashTable[i] = NULL;
                }
                //  Copy data from old to new hash table
                if (_hashTable != NULL)
                {
                    for (UIntSize i = 0; i < _hashTableSize; i++)
                    {
                        while (_hashTable[i] != NULL)
                        {
                            //  Move one entry from old to new hash table
                            _HashTableEntry * pHashTableEntry = _hashTable[i];
                            _hashTable[i] = pHashTableEntry->_pNextInChain;
                            UIntSize newHashCode = _hash(pHashTableEntry->_value, newHashTableSize);
                            pHashTableEntry->_pNextInChain = newHashTable[newHashCode];
                            newHashTable[newHashCode] = pHashTableEntry;
                        }
                    }
                    delete [] _hashTable;
                }
                //  Finish replacing the hash table
                _hashTable = newHashTable;
                _hashTableSize = newHashTableSize;
            }

            //  Insert the new element into the hash table
            UIntSize hashCode = _hash(value, _hashTableSize);
            _HashTableEntry * pNewHashTableEntry = new _HashTableEntry(value);
            pNewHashTableEntry->_pNextInChain = _hashTable[hashCode];
            _hashTable[hashCode] = pNewHashTableEntry;
            _elementCount++;
            return true;
        }

        template<class T, class C>
        UIntSize _HashSetImpl<T,C>::insert(const Set<T> & set)
        {
            UIntSize insertedItemCount = 0;
            for (Iterator<T> it = set.enumerateElements(); !it.finished(); it.advance())
            {
                if (insert(it.item()))
                {
                    insertedItemCount++;
                }
            }
            return insertedItemCount;
        }

        template<class T, class C>
        bool _HashSetImpl<T,C>::remove(const T & value)
        {
            if (_hashTable != NULL)
            {
                UIntSize hashCode = _hash(value, _hashTableSize);
                for (_HashTableEntry ** ppHashTableEntry = &(_hashTable[hashCode]);
                    (*ppHashTableEntry) != NULL;
                    ppHashTableEntry = &((*ppHashTableEntry)->_pNextInChain))
                {
                    if (_comparator.equal((*ppHashTableEntry)->_value, value))
                    {
                        _HashTableEntry * pHashTableEntry = *ppHashTableEntry;
                        *ppHashTableEntry = pHashTableEntry->_pNextInChain;
                        delete pHashTableEntry;
                        _elementCount--;
                        return true;
                    }
                }
                return false;
            }
            else
            {
                return false;
            }
        }

        template<class T, class C>
        UIntSize _HashSetImpl<T,C>::remove(const Set<T> & set)
        {
            UIntSize removedItemCount = 0;
            for (Iterator<T> it = set.enumerateElements(); !it.finished(); it.advance())
            {
                if (remove(it.item()))
                {
                    removedItemCount++;
                }
            }
            return removedItemCount;
        }

        template<class T, class C>
        void _HashSetImpl<T,C>::removeAll()
        {
            if (_hashTable != NULL)
            {
                for (UIntSize i = 0; i < _hashTableSize; i++)
                {
                    while (_hashTable[i] != NULL)
                    {
                        _HashTableEntry * pHashTableEntry = _hashTable[i];
                        _hashTable[i] = pHashTableEntry->_pNextInChain;
                        delete pHashTableEntry;
                    }
                }
                _elementCount = 0;
            }
        }

        template<class T, class C>
        Iterator<T> _HashSetImpl<T,C>::enumerateElements() const
        {
            T * pElements = new T[_elementCount];
            UIntSize elementIndex = 0;
            for (UIntSize i = 0; i < _hashTableSize; i++)
            {
                for (_HashTableEntry * pHashTableEntry = _hashTable[i];
                     pHashTableEntry != NULL;
                     pHashTableEntry = pHashTableEntry->_pNextInChain)
                {
                    pElements[elementIndex++] = pHashTableEntry->_value;
                }
            }
            require(elementIndex == _elementCount);
            return Iterator<T>(new _impl::_SnapshotIteratorImpl<T>(_elementCount, pElements));
        }

        template<class T, class C>
        bool _HashSetImpl<T,C>::operator == (const Set<T> & set) const
        {
            if (_elementCount != set.size())
            {
                return false;
            }
            for (Iterator<T> it = enumerateElements(); !it.finished(); it.advance())
            {
                if (!set.contains(it.item()))
                {
                    return false;
                }
            }
            return true;
        }

        template<class T, class C>
        bool _HashSetImpl<T,C>::operator != (const Set<T> & set) const
        {
            return !((*this) == set);
        }

        template<class T, class C>
        bool _HashSetImpl<T,C>::operator <  (const Set<T> & set) const
        {
            if (_elementCount >= set.size())
            {
                return false;
            }
            for (Iterator<T> it = enumerateElements(); !it.finished(); it.advance())
            {
                if (!set.contains(it.item()))
                {
                    return false;
                }
            }
            return true;
        }

        template<class T, class C>
        bool _HashSetImpl<T,C>::operator <= (const Set<T> & set) const
        {
            if (_elementCount > set.size())
            {
                return false;
            }
            for (Iterator<T> it = enumerateElements(); !it.finished(); it.advance())
            {
                if (!set.contains(it.item()))
                {
                    return false;
                }
            }
            return true;
        }

        template<class T, class C>
        bool _HashSetImpl<T,C>::operator >  (const Set<T> & set) const
        {
            if (_elementCount <= set.size())
            {
                return false;
            }
            for (Iterator<T> it = set.enumerateElements(); !it.finished(); it.advance())
            {
                if (!contains(it.item()))
                {
                    return false;
                }
            }
            return true;
        }

        template<class T, class C>
        bool _HashSetImpl<T,C>::operator >= (const Set<T> & set) const
        {
            if (_elementCount < set.size())
            {
                return false;
            }
            for (Iterator<T> it = set.enumerateElements(); !it.finished(); it.advance())
            {
                if (!contains(it.item()))
                {
                    return false;
                }
            }
            return true;
        }

        template<class T, class C>
        UIntSize _HashSetImpl<T,C>::getReferenceCount() const
        {
            return _referenceCount;
        }

        template<class T, class C>
        void _HashSetImpl<T,C>::incrementReferenceCount()
        {
            _referenceCount++;
            require(_referenceCount != 0);
        }

        template<class T, class C>
        void _HashSetImpl<T,C>::decrementReferenceCount()
        {
            require(_referenceCount != 0);
            _referenceCount--;
        }

        template<class T, class C>
        UIntSize _HashSetImpl<T,C>::_hash(const T & value, UIntSize hashTableSize) const
        {
            return _hashFunction(value) % hashTableSize;
        }
    }

    /**
     *  @class HashSet CARL.hpp
     *  A set implemented via a hash table.
     *
     *   The element type (T) must be:
     *
     *   -   Default constructable (T::T() must be defined and public)
     *   -   Copy constructable (T::T(const T &) must be defined and public)
     *   -   Assignable (T::operator=(const T &) must be defined and public)
     *   -   Destructable (T::~T() must be defined and public)
     *
     *  @param T                The type of set elements.
     *  @param C                The comparator to use for deciding whether
     *                          two values of type \a T are equal or not.
     *  @since 1.0
     */
    template<class T, class C = NaturalComparator<T> >
    class HashSet : public Set<T>
    {
        //////////
        //  Types
    public:
        /**
         *  The callback function for calculating hash codes for values.
         *
         *  @param value        The value to calculate a hash code for.
         *  @return             The calculated hash code for the \a value.
         *  @since 1.0
         */
        typedef UIntSize (CARL_CALLBACK * HashFunction)(const T & value);

        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  The newly constructed hash set is associated with the 
         *  specified hash function and comparator.
         *
         *  @param hashFunction The callback function to use for calculating
         *                      hash codes for set elements.
         *  @param comparator   The comparator to use for deciding when two
         *                      values are equal.
         *  @since 1.0
         */
        HashSet(HashFunction hashFunction, C comparator = C());

        /**
         *  The class constructor.
         *
         *  The newly constructed hash set is associated with the 
         *  specified hash function and comparator.
         *
         *  @param hashFunction The callback function to use for calculating
         *                      hash codes for set elements.
         *  @param initialHashTableCapacity The initial hash table capacity.
         *                      The hash table will not need to grow until
         *                      the hash set contains at least this number
         *                      of elements.
         *  @param comparator   The comparator to use for deciding when two
         *                      values are equal.
         *  @since 1.0
         */
        HashSet(HashFunction hashFunction, UIntSize initialHashTableCapacity, C comparator = C());

        /**
         *  The copy constructor.
         *
         *  @param src          The hash set to copy contents from.
         *  @since 1.0
         */
        HashSet(const HashSet<T,C> & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~HashSet();

        /**
         *  The assignment operator.
         *
         *  @param src          The hash set to copy contents from.
         *  @return             A reference to the assigned-to set.
         *  @since 1.0
         */
        HashSet<T,C> &            operator = (const HashSet<T,C> & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a set is empty.
         *
         *  @return             True if this set is empty; false otherwise.
         *  @since 1.0
         */
        virtual bool            isEmpty() const;

        /**
         *  Returns the number of elements in this set.
         *
         *  @return             The number of elements in this set.
         *  @since 1.0
         */
        virtual UIntSize        size() const;

        /**
         *  Checks if the set contains the specified element.
         *
         *  @param value        The value to check for membership in a set.
         *  @return             True if the set contains the specified element;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            contains(const T & value) const;

        /**
         *  Inserts an element into the set.
         *
         *  If the element is already a member of the set, the call
         *  has no effect.
         *
         *  @param value        The element to insert.
         *  @return             True if the element has been actually inserted;
         *                      false if the element was already a member of the
         *                      set before the call.
         *  @since 1.0
         */
        virtual bool            insert(const T & value);

        /**
         *  Inserts all elements of a specified set into this set.
         *
         *  @param set          The set of elements to insert.
         *  @return             The actual number of new elements inserted.
         *  @since 1.0
         */
        virtual UIntSize        insert(const Set<T> & set);

        /**
         *  Removes an element from the set.
         *
         *  If the element is already not a member of the set, the call
         *  has no effect.
         *
         *  @param value        The element to remove.
         *  @return             True if the element has been actually removed;
         *                      false if the element was already not a member
         *                      of the set before the call.
         *  @since 1.0
         */
        virtual bool            remove(const T & value);

        /**
         *  Removes all elements of a specified set from this set.
         *
         *  @param set          The set of elements to remove.
         *  @return             The actual number of removed elements.
         *  @since 1.0
         */
        virtual UIntSize        remove(const Set<T> & set);

        /**
         *  Clears out the set.
         *
         *  @since 1.0
         */
        virtual void            removeAll();

        /**
         *  Returns an iterator over all elements of the set.
         *
         *  Elements are not guaranteed to be iterated over in any
         *  particular order.
         *
         *  @return             An iterator over all elements of the set.
         *  @since 1.0
         */
        virtual Iterator<T>     enumerateElements() const;

        /**
         *  Checks it two sets are equal.
         *
         *  @param set          The set to compare to this set.
         *  @return             True if both seth contain exactly the same
         *                      elements; false otherwise.
         *  @since 1.0
         */
        virtual bool            operator == (const Set<T> & set) const;

        /**
         *  Checks it two sets are unequal.
         *
         *  @param set          The set to compare to this set.
         *  @return             False if both seth contain exactly the same
         *                      elements; true otherwise.
         *  @since 1.0
         */
        virtual bool            operator != (const Set<T> & set) const;

        /**
         *  Checks this set is a proper subset of another set.
         *
         *  @param set          The set to compare to this set.
         *  @return             True if this set is a proper subset of \a set;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            operator <  (const Set<T> & set) const;

        /**
         *  Checks this set is a subset of another set.
         *
         *  @param set          The set to compare to this set.
         *  @return             True if this set is a subset of \a set;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            operator <= (const Set<T> & set) const;

        /**
         *  Checks this set is a proper superset of another set.
         *
         *  @param set          The set to compare to this set.
         *  @return             True if this set is a proper superset of \a set;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            operator >  (const Set<T> & set) const;

        /**
         *  Checks this set is a superset of another set.
         *
         *  @param set          The set to compare to this set.
         *  @return             True if this set is a superset of \a set;
         *                      false otherwise.
         *  @since 1.0
         */
        virtual bool            operator >= (const Set<T> & set) const;

        /**
         *  Inserts an element into the set.
         *
         *  If the element is already a member of the set, the call
         *  has no effect.
         *
         *  @param value        The element to insert.
         *  @return             The reference to the inserted-to set.
         *  @since 1.0
         */
        virtual Set<T> &        operator += (const T & value);

        /**
         *  Inserts all elements of a specified set into this set.
         *
         *  @param set          The set of elements to insert.
         *  @return             The reference to the inserted-to set.
         *  @since 1.0
         */
        virtual Set<T> &        operator += (const Set<T> & set);

        /**
         *  Removes an element from the set.
         *
         *  If the element is already not a member of the set, the call
         *  has no effect.
         *
         *  @param value        The element to remove.
         *  @return             The reference to the removed-from set.
         *  @since 1.0
         */
        virtual Set<T> &        operator -= (const T & value);

        /**
         *  Removes all elements of a specified set from this set.
         *
         *  @param set          The set of elements to remove.
         *  @return             The reference to the removed-from set.
         *  @since 1.0
         */
        virtual Set<T> &        operator -= (const Set<T> & set);

        /**
         *  Intersects this set with the specified value.
         *
         *  The result of such an intersection is either an empty set (if
         *  the specified value was not a member of this set) or a set containing
         *  only one value (if the specified value was a member of this set).
         *
         *  @param value        The value to intersec tis set with.
         *  @return             The reference to the intersected set.
         *  @since 1.0
         */
        virtual Set<T> &        operator *= (const T & value);

        /**
         *  Intersects this set with the specified set.
         *
         *  All elements that are not present in the specified \a set
         *  are removed from this set.
         *
         *  @param set          The set to intersect this set with.
         *  @return             The reference to the intersected set.
         *  @since 1.0
         */
        virtual Set<T> &        operator *= (const Set<T> & set);

        /**
         *  Calculates a union of this set and the specified value.
         *
         *  @param value        The element to include into the set.
         *  @return             The union of this set and the specified value.
         *  @since 1.0
         */
        virtual HashSet<T,C>    operator + (const T & value) const;

        /**
         *  Calculates a union of two sets.
         *
         *  @param set          The 2nd set to include into the result.
         *  @return             The union of this set and the specified set.
         *  @since 1.0
         */
        virtual HashSet<T,C>    operator + (const Set<T> & set) const;

        /**
         *  Calculates a difference of this set and the specified value.
         *
         *  @param value        The element to remove from the set.
         *  @return             The difference of this set and the specified value.
         *  @since 1.0
         */
        virtual HashSet<T,C>    operator - (const T & value) const;

        /**
         *  Calculates a difference of two sets.
         *
         *  @param set          The set to exclude from the result.
         *  @return             The difference of this set and the specified set.
         *  @since 1.0
         */
        virtual HashSet<T,C>    operator - (const Set<T> & set) const;

        /**
         *  Calculates an intersection of this set and the specified value.
         *
         *  @param value        The element to intersect with the set.
         *  @return             The intersection of this set and the specified value.
         *  @since 1.0
         */
        virtual HashSet<T,C>    operator * (const T & value) const;

        /**
         *  Calculates an intersection of two sets.
         *
         *  @param set          The set to intersection with.
         *  @return             The intersection of this set and the specified set.
         *  @since 1.0
         */
        virtual HashSet<T,C>    operator * (const Set<T> & set) const;

        //////////
        //  Implementation
    private:
        _impl::_HashSetImpl<T,C> *  _pImpl;
        void                    _ensureNonsharedImplementation();
    };

    template<class T, class C>
    HashSet<T,C>::HashSet(HashFunction hashFunction, C comparator /*= C()*/)
    {
        _pImpl = new _impl::_HashSetImpl<T,C>(hashFunction, comparator);
        _pImpl->incrementReferenceCount();
    }

    template<class T, class C>
    HashSet<T,C>::HashSet(HashFunction hashFunction, UIntSize initialHashTableCapacity, C comparator /*= C()*/)
    {
        _pImpl = new _impl::_HashSetImpl<T,C>(hashFunction, initialHashTableCapacity, comparator);
        _pImpl->incrementReferenceCount();
    }

    template<class T, class C>
    HashSet<T,C>::HashSet(const HashSet<T,C> & src)
    {
        _pImpl = src._pImpl;
        _pImpl->incrementReferenceCount();
    }

    template<class T, class C>
    HashSet<T,C>::~HashSet()
    {
        _pImpl->decrementReferenceCount();
        if (_pImpl->getReferenceCount() == 0)
        {
            delete _pImpl;
        }
    }

    template<class T, class C>
    HashSet<T,C> & HashSet<T,C>::operator = (const HashSet<T,C> & src)
    { 
        if (this != (&src))
        {
            _pImpl->decrementReferenceCount();
            if (_pImpl->getReferenceCount() == 0)
            {
                delete _pImpl;
            }
            _pImpl = src._pImpl;
            _pImpl->incrementReferenceCount();
        }
        return *this;
    }

    template<class T, class C>
    bool HashSet<T,C>::isEmpty() const
    {
        return _pImpl->isEmpty();
    }

    template<class T, class C>
    UIntSize HashSet<T,C>::size() const
    {
        return _pImpl->size();
    }

    template<class T, class C>
    bool HashSet<T,C>::contains(const T & value) const
    {
        return _pImpl->contains(value);
    }

    template<class T, class C>
    bool HashSet<T,C>::insert(const T & value)
    {
        _ensureNonsharedImplementation();
        return _pImpl->insert(value);
    }

    template<class T, class C>
    UIntSize HashSet<T,C>::insert(const Set<T> & set)
    {
        _ensureNonsharedImplementation();
        return _pImpl->insert(set);
    }

    template<class T, class C>
    bool HashSet<T,C>::remove(const T & value)
    {
        _ensureNonsharedImplementation();
        return _pImpl->remove(value);
    }

    template<class T, class C>
    UIntSize HashSet<T,C>::remove(const Set<T> & set)
    {
        _ensureNonsharedImplementation();
        return _pImpl->remove(set);
    }

    template<class T, class C>
    void HashSet<T,C>::removeAll()
    {
        _ensureNonsharedImplementation();
        return _pImpl->removeAll();
    }

    template<class T, class C>
    Iterator<T> HashSet<T,C>::enumerateElements() const
    {
        return _pImpl->enumerateElements();
    }

    template<class T, class C>
    bool HashSet<T,C>::operator == (const Set<T> & set) const
    {
        return (*_pImpl) == set;
    }

    template<class T, class C>
    bool HashSet<T,C>::operator != (const Set<T> & set) const
    {
        return (*_pImpl) != set;
    }

    template<class T, class C>
    bool HashSet<T,C>::operator <  (const Set<T> & set) const
    {
        return (*_pImpl) < set;
    }

    template<class T, class C>
    bool HashSet<T,C>::operator <= (const Set<T> & set) const
    {
        return (*_pImpl) <= set;
    }

    template<class T, class C>
    bool HashSet<T,C>::operator >  (const Set<T> & set) const
    {
        return (*_pImpl) > set;
    }

    template<class T, class C>
    bool HashSet<T,C>::operator >= (const Set<T> & set) const
    {
        return (*_pImpl) >= set;
    }

    template<class T, class C>
    Set<T> & HashSet<T,C>::operator += (const T & value)
    {
        _ensureNonsharedImplementation();
        _pImpl->insert(value);
        return *this;
    }

    template<class T, class C>
    Set<T> & HashSet<T,C>::operator += (const Set<T> & set)
    {
        _ensureNonsharedImplementation();
        _pImpl->insert(set);
        return *this;
    }

    template<class T, class C>
    Set<T> & HashSet<T,C>::operator -= (const T & value)
    {
        _ensureNonsharedImplementation();
        _pImpl->remove(value);
        return *this;
    }

    template<class T, class C>
    Set<T> & HashSet<T,C>::operator -= (const Set<T> & set)
    {
        _ensureNonsharedImplementation();
        _pImpl->remove(set);
        return *this;
    }

    template<class T, class C>
    Set<T> & HashSet<T,C>::operator *= (const T & value)
    {
        _ensureNonsharedImplementation();
        if (_pImpl->contains(value))
        {
            _pImpl->removeAll();
            _pImpl->insert(value);
        }
        else
        {
            _pImpl->removeAll();
        }
        return *this;
    }

    template<class T, class C>
    Set<T> & HashSet<T,C>::operator *= (const Set<T> & set)
    {
        HashSet<T,C> commonValues(_pImpl->_hashFunction, _pImpl->size());
        for (Iterator<T> it = enumerateElements(); !it.finished(); it.advance())
        {
            if (set.contains(it.item()))
            {
                commonValues.insert(it.item());
            }
        }
        removeAll();
        insert(commonValues);
        return *this;
    }

    template<class T, class C>
    HashSet<T,C> HashSet<T,C>::operator + (const T & value) const
    {
        HashSet<T,C> result = (*this);
        result += value;
        return result;
    }

    template<class T, class C>
    HashSet<T,C> HashSet<T,C>::operator + (const Set<T> & set) const
    {
        HashSet<T,C> result = (*this);
        result += set;
        return result;
    }

    template<class T, class C>
    HashSet<T,C> HashSet<T,C>::operator - (const T & value) const
    {
        HashSet<T,C> result = (*this);
        result -= value;
        return result;
    }

    template<class T, class C>
    HashSet<T,C> HashSet<T,C>::operator - (const Set<T> & set) const
    {
        HashSet<T,C> result = (*this);
        result -= set;
        return result;
    }

    template<class T, class C>
    HashSet<T,C> HashSet<T,C>::operator * (const T & value) const
    {
        HashSet<T,C> result = (*this);
        result *= value;
        return result;
    }

    template<class T, class C>
    HashSet<T,C> HashSet<T,C>::operator * (const Set<T> & set) const
    {
        HashSet<T,C> result = (*this);
        result *= set;
        return result;
    }

    template<class T, class C>
    void HashSet<T,C>::_ensureNonsharedImplementation()
    {
        if (_pImpl->getReferenceCount() != 1)
        {
            _impl::_HashSetImpl<T,C> * pNewImpl = new _impl::_HashSetImpl<T,C>(*_pImpl);
            _pImpl->decrementReferenceCount();
            pNewImpl->incrementReferenceCount();
            _pImpl = pNewImpl;
        }
    }
}

#endif  //  __CARL_Set_hpp_included__
//  End of CARL_Set.hpp
