//
//  CARL_Settings.hpp - Persisent settings.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:49 $
//  $Revision: 1.19 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL persistent settings framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Settings.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Settings_hpp_included__
#define __CARL_Settings_hpp_included__

namespace carl
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    namespace _impl
    {
        class CARL_PUBLIC _EmulatedSettings;
    }
#else
    #error Unsupported platform
#endif

    /**
     *  @class SettingsNode CARL.hpp
     *  A path identifying a single node within a settings tree.
     *
     *  A path to a specific node within a settings tree consists of
     *  one or more path components separated by forward slash characters /.
     *
     *  There are three types of settings node paths:
     *
     *  -   An absolute path always starts with two forward slashes
     *      followed by zero or more path components. Such path always
     *      specifies a complete path from the root of the settings tree
     *      to a settings node.
     *  -   A virtual absolute path starts with one forward slash
     *      followed by zero or more path components. Such path always
     *      specifies a path from the current virtual root of the settings
     *      tree to a settings node.
     *  -   A relative path does not start with a slash. It specifies a
     *      path to a settings node from another settings node.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC SettingsNode
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  The newly created settings node is initialized to be an
         *  absolute path to the root settings node (//).
         *
         *  @since 1.0
         */
        SettingsNode();

        /**
         *  The class constructor.
         *
         *  The newly created settings node is initialized with the
         *  specified string.
         *
         *  @param pPath        A pointer to a nul-terminated ASCII string
         *                      that specifies the settings node path.
         *  @since 1.0
         */
        explicit SettingsNode(const char * pPath);

        /**
         *  The class constructor.
         *
         *  The newly created settings node is initialized with the
         *  specified string.
         *
         *  @param pPath        A pointer to a nul-terminated ISO-10646 string
         *                      that specifies the settings node path.
         *  @since 1.0
         */
        explicit SettingsNode(const Char * pPath);

        /**
         *  The class constructor.
         *
         *  The newly created settings node is initialized with the
         *  specified string.
         *
         *  @param path         A string that specifies the settings node path.
         *  @since 1.0
         */
        explicit SettingsNode(const String & path);

        /**
         *  The class constructor.
         *
         *  The newly created settings node is initialized with a path to
         *  the specified subnode of an existing settings node
         *
         *  @param parent       The parent settings node.
         *  @param pSubnode     A pointer to a nul-terminated ASCII string
         *                      specifying relative path from the \a parent.
         *  @since 1.0
         */
        SettingsNode(const SettingsNode & parent, const char * pSubnode);

        /**
         *  The class constructor.
         *
         *  The newly created settings node is initialized with a path to
         *  the specified subnode of an existing settings node
         *
         *  @param parent       The parent settings node.
         *  @param pSubnode     A pointer to a nul-terminated ISO-10646 string
         *                      specifying relative path from the \a parent.
         *  @since 1.0
         */
        SettingsNode(const SettingsNode & parent, const Char * pSubnode);

        /**
         *  The class constructor.
         *
         *  The newly created settings node is initialized with a path to
         *  the specified subnode of an existing settings node
         *
         *  @param parent       The parent settings node.
         *  @param subnode      A string specifying relative path from 
         *                      the \a parent.
         *  @since 1.0
         */
        SettingsNode(const SettingsNode & parent, const String & subnode);

        /**
         *  The copy constructor.
         *
         *  @param src          The settings node to copy path from.
         *  @since 1.0
         */
        SettingsNode(const SettingsNode & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~SettingsNode() {}

        /**
         *  The assignment operator.
         *
         *  @param src          The settings node to copy path from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        SettingsNode &          operator = (const SettingsNode & src);

        //////////
        //  Operations
    public:
        /**
         *  Checks if the settings node represents an absolute path.
         *
         *  @return             True if this settings node represents
         *                      an absolute path; false otherwise.
         *  @since 1.0
         */
        bool                    isAbsolute() const { return _path.startsWith("//"); }
        /**
         *  Checks if the settings node represents a virtual absolute path.
         *
         *  @return             True if this settings node represents
         *                      a virtual absolute path; false otherwise.
         *  @since 1.0
         */
        bool                    isVirtualAbsolute() const { return _path.startsWith("/") && !_path.startsWith("//"); }

        /**
         *  Checks if the settings node represents a root.
         *
         *  @return             True if this settings node represents
         *                      an absolute or virtual absolute root;
         *                      false otherwise.
         *  @since 1.0
         */
        bool                    isRoot() const { return _path == '/' || _path == "//"; }

        /**
         *  Returns an immediate parent of this settings node.
         *
         *  If this settings node is a root, the returned value is
         *  the settings node itself.
         *
         *  @return             An immediate parent of this settings node.
         *  @since 1.0
         */
        SettingsNode            getParent() const;

        /**
         *  Returns the string representation of this settings node's path.
         *
         *  @return             The string representation of this settings node's path.
         *  @since 1.0
         */
        String                  toString() const { return _path; }

        //////////
        //  Implementation
    private:
        String                  _path;
    };

    /**
     *  @class Settings CARL.hpp
     *  The persistent settings repository.
     *
     *  CARL settings allow applications and DLLs to store user- or system-level
     *  prefrences and settings persistently. Settings are organized in a tree
     *  where:
     *
     *  -   Each node is identified by a name, which must be unique among
     *      its immediate siblings.
     *  -   Each node can have an arbitrary number of (key, value) pairs
     *      associated with it. All values associated with the same settings
     *      node must have different keys.
     *  -   All node and value names are case-sensitive.
     *
     *  A special node, known as the settings root, is the ultimate parent node
     *  of a settings tree. Each node within the tree can be uniquely identified 
     *  by a path from the settings root node to the node in question.
     *
     *  In addition to the settings root node, an arbitrary node within the
     *  settings tree can be designated as a virtual root node. Initially the
     *  virtual root node is the same as the settings root node, but it can be
     *  reassigned to an arbitrary node as many times as needed by the application.
     *  A special type of settings node paths - known as virtual absolute paths -
     *  refer to a settings node using the virtual root node as an origin.
     *
     *  Each CARL application has access to two settings trees:
     *
     *  -   A shared settings tree, which contains settings common for all users.
     *  -   A private settings tree, of which there is a separate copy for each user.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Settings
    {
        //////////
        //  Can't assign or copy construct
    private:
        Settings(const Settings &);
        void                operator = (const Settings &);

        //////////
        //  Construction/destruction - from CARL only
    private:
        Settings();
        ~Settings();

        //////////
        //  Operations
    public:
        /**
         *  Returns a pointer to the shared settings tree.
         *
         *  @return             A pointer to the shared settings tree.
         *  @since 1.0
         */
        static Settings *       getShared();

        /**
         *  Returns a pointer to the private settings tree.
         *
         *  @return             A pointer to the private settings tree.
         *  @since 1.0
         */
        static Settings *       getPrivate();

        /**
         *  Checks if the specified settings node exists.
         *
         *  @param node         The node to check.
         *  @return             True if the specified settings node exists;
         *                      false otherwise.
         *  @since 1.0
         */
        bool                    nodeExists(const SettingsNode & node);

        /**
         *  Creates the settings node.
         *
         *  If the specified settings node already exists, the service succeeds.
         *
         *  @param node         The settings node to create.
         *  @return             True if the \a node exists after the call (i.e.
         *                      the node was successfully created or already existed
         *                      before the call); false otherwise.
         *  @since 1.0
         */
        bool                    createNode(const SettingsNode & node);

        /**
         *  Destroys the settings node.
         *
         *  If the specified settings node already does not exist, the service succeeds.
         *  Otherwise, the settings node is destroyed along with all associated
         *  values and descendants.
         *
         *  @param node         The settings node to destroy.
         *  @return             True if the \a node does not exist after the call (i.e.
         *                      the node was successfully destroyed or already did not
         *                      exist before the call); false otherwise.
         *  @since 1.0
         */
        bool                    destroyNode(const SettingsNode & node);

        /**
         *  Returns a list of all direct children of the specified node.
         *
         *  Elements in the returned list are not guaranteed to be in any
         *  particular order.
         *
         *  @param node         The settings node whose children are queried.
         *  @return             The list of all direct children of the specified node.
         *  @since 1.0
         */
        Array<SettingsNode>     getChildNodes(const SettingsNode & node);

        /**
         *  Checks if a node has an associated value.
         *
         *  @param node         The settings node being checked.
         *  @param pValueName   A pointer to a nul-terminated ASCII string
         *                      representing the value key.
         *  @return             True if the specified settings node has an associated
         *                      value with the specified key; false otherwise.
         *  @since 1.0
         */
        bool                    valueExists(const SettingsNode & node, const char * pValueName) { return valueExists(node, String(pValueName)); }

        /**
         *  Checks if a node has an associated value.
         *
         *  @param node         The settings node being checked.
         *  @param pValueName   A pointer to a nul-terminated ISO-10646 string
         *                      representing the value key.
         *  @return             True if the specified settings node has an associated
         *                      value with the specified key; false otherwise.
         *  @since 1.0
         */
        bool                    valueExists(const SettingsNode & node, const Char * pValueName) { return valueExists(node, String(pValueName)); }

        /**
         *  Checks if a node has an associated value.
         *
         *  @param node         The settings node being checked.
         *  @param valueName    A string representing the value key.
         *  @return             True if the specified settings node has an associated
         *                      value with the specified key; false otherwise.
         *  @since 1.0
         */
        bool                    valueExists(const SettingsNode & node, const String & valueName);

        /**
         *  Returns a list of all value keys for the specified node.
         *
         *  Elements in the returned list are not guaranteed to be in any
         *  particular order.
         *
         *  @param node         The settings node whose values are queried.
         *  @return             The list of all value keys for the specified node.
         *  @since 1.0
         */
        Array<String>           getValueNames(const SettingsNode & node);

        /**
         *  Destroys a value associated with the specified node.
         *
         *  If the specified value already does not exist in the specified
         *  settings node, the service succeeds.
         *
         *  @param node         The settings node whose value is destroyed.
         *  @param pValueName   A pointer to a nul-terminated ASCII string
         *                      representing the value key.
         *  @return             True if the specified value does not exist in
         *                      the specified settings node after the call (i.e.
         *                      the value was successfully destroyed, or not there
         *                      to begin with); false otherwise.
         *  @since 1.0
         */
        bool                    destroyValue(const SettingsNode & node, const char * pValueName) { return destroyValue(node, String(pValueName)); }

        /**
         *  Destroys a value associated with the specified node.
         *
         *  If the specified value already does not exist in the specified
         *  settings node, the service succeeds.
         *
         *  @param node         The settings node whose value is destroyed.
         *  @param pValueName   A pointer to a nul-terminated ISO-10646 string
         *                      representing the value key.
         *  @return             True if the specified value does not exist in
         *                      the specified settings node after the call (i.e.
         *                      the value was successfully destroyed, or not there
         *                      to begin with); false otherwise.
         *  @since 1.0
         */
        bool                    destroyValue(const SettingsNode & node, const Char * pValueName) { return destroyValue(node, String(pValueName)); }

        /**
         *  Destroys a value associated with the specified node.
         *
         *  If the specified value already does not exist in the specified
         *  settings node, the service succeeds.
         *
         *  @param node         The settings node whose value is destroyed.
         *  @param valueName    A string representing the value key.
         *  @return             True if the specified value does not exist in
         *                      the specified settings node after the call (i.e.
         *                      the value was successfully destroyed, or not there
         *                      to begin with); false otherwise.
         *  @since 1.0
         */
        bool                    destroyValue(const SettingsNode & node, const String & valueName);

        /**
         *  Queries a value associated with the specified node.
         *
         *  If the specified value does not exist in the specified
         *  settings node, the default value is returned.
         *
         *  @param node         The settings node whose value is queried.
         *  @param pValueName   A pointer to a nul-terminated ASCII string
         *                      representing the value key.
         *  @param pDefaultValue    A pointer to a nul-terminated ASCII string
         *                      representing the default value to be used when
         *                      the settings node does not exist or does not
         *                      have an associated value with the required key.
         *  @return             The queried value, or the default value if the
         *                      query could not be made.
         *  @since 1.0
         */
        String                  getValue(const SettingsNode & node, const char * pValueName, const char * pDefaultValue) { return getValue(node, String(pValueName), String(pDefaultValue)); }

        /**
         *  Queries a value associated with the specified node.
         *
         *  If the specified value does not exist in the specified
         *  settings node, the default value is returned.
         *
         *  @param node         The settings node whose value is queried.
         *  @param pValueName   A pointer to a nul-terminated ASCII string
         *                      representing the value key.
         *  @param pDefaultValue    A pointer to a nul-terminated ISO-10646 string
         *                      representing the default value to be used when
         *                      the settings node does not exist or does not
         *                      have an associated value with the required key.
         *  @return             The queried value, or the default value if the
         *                      query could not be made.
         *  @since 1.0
         */
        String                  getValue(const SettingsNode & node, const char * pValueName, const Char * pDefaultValue) { return getValue(node, String(pValueName), String(pDefaultValue)); }

        /**
         *  Queries a value associated with the specified node.
         *
         *  If the specified value does not exist in the specified
         *  settings node, the default value is returned.
         *
         *  @param node         The settings node whose value is queried.
         *  @param pValueName   A pointer to a nul-terminated ASCII string
         *                      representing the value key.
         *  @param defaultValue A string representing the default value to be 
         *                      used when the settings node does not exist or 
         *                      does not have an associated value with the 
         *                      required key.
         *  @return             The queried value, or the default value if the
         *                      query could not be made.
         *  @since 1.0
         */
        String                  getValue(const SettingsNode & node, const char * pValueName, const String & defaultValue) { return getValue(node, String(pValueName), defaultValue); }

        /**
         *  Queries a value associated with the specified node.
         *
         *  If the specified value does not exist in the specified
         *  settings node, the default value is returned.
         *
         *  @param node         The settings node whose value is queried.
         *  @param pValueName   A pointer to a nul-terminated ISO-10646 string
         *                      representing the value key.
         *  @param pDefaultValue    A pointer to a nul-terminated ASCII string
         *                      representing the default value to be used when
         *                      the settings node does not exist or does not
         *                      have an associated value with the required key.
         *  @return             The queried value, or the default value if the
         *                      query could not be made.
         *  @since 1.0
         */
        String                  getValue(const SettingsNode & node, const Char * pValueName, const char * pDefaultValue) { return getValue(node, String(pValueName), String(pDefaultValue)); }

        /**
         *  Queries a value associated with the specified node.
         *
         *  If the specified value does not exist in the specified
         *  settings node, the default value is returned.
         *
         *  @param node         The settings node whose value is queried.
         *  @param pValueName   A pointer to a nul-terminated ISO-10646 string
         *                      representing the value key.
         *  @param pDefaultValue    A pointer to a nul-terminated ISO-10646 string
         *                      representing the default value to be used when
         *                      the settings node does not exist or does not
         *                      have an associated value with the required key.
         *  @return             The queried value, or the default value if the
         *                      query could not be made.
         *  @since 1.0
         */
        String                  getValue(const SettingsNode & node, const Char * pValueName, const Char * pDefaultValue) { return getValue(node, String(pValueName), String(pDefaultValue)); }

        /**
         *  Queries a value associated with the specified node.
         *
         *  If the specified value does not exist in the specified
         *  settings node, the default value is returned.
         *
         *  @param node         The settings node whose value is queried.
         *  @param pValueName   A pointer to a nul-terminated ISO-10646 string
         *                      representing the value key.
         *  @param defaultValue A string representing the default value to be 
         *                      used when the settings node does not exist or 
         *                      does not have an associated value with the 
         *                      required key.
         *  @return             The queried value, or the default value if the
         *                      query could not be made.
         *  @since 1.0
         */
        String                  getValue(const SettingsNode & node, const Char * pValueName, const String & defaultValue) { return getValue(node, String(pValueName), defaultValue); }

        /**
         *  Queries a value associated with the specified node.
         *
         *  If the specified value does not exist in the specified
         *  settings node, the default value is returned.
         *
         *  @param node         The settings node whose value is queried.
         *  @param valueName    A string representing the value key.
         *  @param pDefaultValue    A pointer to a nul-terminated ASCII string
         *                      representing the default value to be used when
         *                      the settings node does not exist or does not
         *                      have an associated value with the required key.
         *  @return             The queried value, or the default value if the
         *                      query could not be made.
         *  @since 1.0
         */
        String                  getValue(const SettingsNode & node, const String & valueName, const char * pDefaultValue) { return getValue(node, valueName, String(pDefaultValue)); }

        /**
         *  Queries a value associated with the specified node.
         *
         *  If the specified value does not exist in the specified
         *  settings node, the default value is returned.
         *
         *  @param node         The settings node whose value is queried.
         *  @param valueName    A string representing the value key.
         *  @param pDefaultValue    A pointer to a nul-terminated ISo-10646 string
         *                      representing the default value to be used when
         *                      the settings node does not exist or does not
         *                      have an associated value with the required key.
         *  @return             The queried value, or the default value if the
         *                      query could not be made.
         *  @since 1.0
         */
        String                  getValue(const SettingsNode & node, const String & valueName, const Char * pDefaultValue) { return getValue(node, valueName, String(pDefaultValue)); }

        /**
         *  Queries a value associated with the specified node.
         *
         *  If the specified value does not exist in the specified
         *  settings node, the default value is returned.
         *
         *  @param node         The settings node whose value is queried.
         *  @param valueName    A string representing the value key.
         *  @param defaultValue A string representing the default value to be 
         *                      used when the settings node does not exist or 
         *                      does not have an associated value with the 
         *                      required key.
         *  @return             The queried value, or the default value if the
         *                      query could not be made.
         *  @since 1.0
         */
        String                  getValue(const SettingsNode & node, const String & valueName, const String & defaultValue);

        /**
         *  Sets a value associated with the specified node.
         *
         *  If the specified settings node does not exist, it is created.
         *
         *  If the specified value already exists in the specified
         *  settings node, it is replaced with the new one.
         *
         *  @param node         The settings node whose value is changed.
         *  @param pValueName   A pointer to a nul-terminated ASCII string
         *                      representing the value key.
         *  @param pValue       A pointer to a nul-terminated ASCII string
         *                      representing the value associate with the
         *                      specified key.
         *  @return             True if the value was set successfully; false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    setValue(const SettingsNode & node, const char * pValueName, const char * pValue) { return setValue(node, String(pValueName), String(pValue)); }

        /**
         *  Sets a value associated with the specified node.
         *
         *  If the specified settings node does not exist, it is created.
         *
         *  If the specified value already exists in the specified
         *  settings node, it is replaced with the new one.
         *
         *  @param node         The settings node whose value is changed.
         *  @param pValueName   A pointer to a nul-terminated ASCII string
         *                      representing the value key.
         *  @param pValue       A pointer to a nul-terminated ISO-10646 string
         *                      representing the value associate with the
         *                      specified key.
         *  @return             True if the value was set successfully; false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    setValue(const SettingsNode & node, const char * pValueName, const Char * pValue) { return setValue(node, String(pValueName), String(pValue)); }

        /**
         *  Sets a value associated with the specified node.
         *
         *  If the specified settings node does not exist, it is created.
         *
         *  If the specified value already exists in the specified
         *  settings node, it is replaced with the new one.
         *
         *  @param node         The settings node whose value is changed.
         *  @param pValueName   A pointer to a nul-terminated ASCII string
         *                      representing the value key.
         *  @param value        A string representing the value associate with the
         *                      specified key.
         *  @return             True if the value was set successfully; false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    setValue(const SettingsNode & node, const char * pValueName, const String & value) { return setValue(node, String(pValueName), value); }

        /**
         *  Sets a value associated with the specified node.
         *
         *  If the specified settings node does not exist, it is created.
         *
         *  If the specified value already exists in the specified
         *  settings node, it is replaced with the new one.
         *
         *  @param node         The settings node whose value is changed.
         *  @param pValueName   A pointer to a nul-terminated ISO-10646 string
         *                      representing the value key.
         *  @param pValue       A pointer to a nul-terminated ASCII string
         *                      representing the value associate with the
         *                      specified key.
         *  @return             True if the value was set successfully; false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    setValue(const SettingsNode & node, const Char * pValueName, const char * pValue) { return setValue(node, String(pValueName), String(pValue)); }

        /**
         *  Sets a value associated with the specified node.
         *
         *  If the specified settings node does not exist, it is created.
         *
         *  If the specified value already exists in the specified
         *  settings node, it is replaced with the new one.
         *
         *  @param node         The settings node whose value is changed.
         *  @param pValueName   A pointer to a nul-terminated ISO-10646 string
         *                      representing the value key.
         *  @param pValue       A pointer to a nul-terminated ISO-10646 string
         *                      representing the value associate with the
         *                      specified key.
         *  @return             True if the value was set successfully; false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    setValue(const SettingsNode & node, const Char * pValueName, const Char * pValue) { return setValue(node, String(pValueName), String(pValue)); }

        /**
         *  Sets a value associated with the specified node.
         *
         *  If the specified settings node does not exist, it is created.
         *
         *  If the specified value already exists in the specified
         *  settings node, it is replaced with the new one.
         *
         *  @param node         The settings node whose value is changed.
         *  @param pValueName   A pointer to a nul-terminated ISO-10646 string
         *                      representing the value key.
         *  @param value        A string representing the value associate with the
         *                      specified key.
         *  @return             True if the value was set successfully; false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    setValue(const SettingsNode & node, const Char * pValueName, const String & value) { return setValue(node, String(pValueName), value); }

        /**
         *  Sets a value associated with the specified node.
         *
         *  If the specified settings node does not exist, it is created.
         *
         *  If the specified value already exists in the specified
         *  settings node, it is replaced with the new one.
         *
         *  @param node         The settings node whose value is changed.
         *  @param valueName    A string representing the value key.
         *  @param pValue       A pointer to a nul-terminated ASCII string
         *                      representing the value associate with the
         *                      specified key.
         *  @return             True if the value was set successfully; false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    setValue(const SettingsNode & node, const String & valueName, const char * pValue) { return setValue(node, valueName, String(pValue)); }

        /**
         *  Sets a value associated with the specified node.
         *
         *  If the specified settings node does not exist, it is created.
         *
         *  If the specified value already exists in the specified
         *  settings node, it is replaced with the new one.
         *
         *  @param node         The settings node whose value is changed.
         *  @param valueName    A string representing the value key.
         *  @param pValue       A pointer to a nul-terminated ISO-10646 string
         *                      representing the value associate with the
         *                      specified key.
         *  @return             True if the value was set successfully; false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    setValue(const SettingsNode & node, const String & valueName, const Char * pValue) { return setValue(node, valueName, String(pValue)); }

        /**
         *  Sets a value associated with the specified node.
         *
         *  If the specified settings node does not exist, it is created.
         *
         *  If the specified value already exists in the specified
         *  settings node, it is replaced with the new one.
         *
         *  @param node         The settings node whose value is changed.
         *  @param valueName    A string representing the value key.
         *  @param value        A string representing the value associate with the
         *                      specified key.
         *  @return             True if the value was set successfully; false
         *                      otherwise.
         *  @since 1.0
         */
        bool                    setValue(const SettingsNode & node, const String & valueName, const String & value);

        /**
         *  Queries the current virtual root of this settings tree.
         *
         *  @return             The current virtual root of this settings tree.
         *  @since 1.0
         */
        SettingsNode            getVirtualRoot() const { return _virtualRoot; }

        /**
         *  Changes the current virtual root of this settings tree.
         *
         *  @param virtualRoot  The new virtual root for this settings tree.
         *  @since 1.0
         */
        void                    setVirtualRoot(const SettingsNode & virtualRoot) { _virtualRoot = _constructAbsolutePath(virtualRoot); }

        //////////
        //  Implementation
    private:
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        _impl::_EmulatedSettings *  _pEmulatedSettings;
#else
    #error Unsupported platform
#endif
        SettingsNode            _virtualRoot;
        SettingsNode            _constructAbsolutePath(const SettingsNode & relativePath);
        SettingsNode            _constructRelativePath(const SettingsNode & absolutePath);
    };
}

#endif  //  __CARL_Settings_hpp_included__
//  End of CARL_Settings.hpp
