//
//  CARL_Sort.hpp - Sorting framework.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:49 $
//  $Revision: 1.11 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL sorting framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Sort.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Sort_hpp_included__
#define __CARL_Sort_hpp_included__

namespace carl
{
    /**
     *  @interface ISortOrder CARL.hpp
     *  An abstract interface to a sort order.
     *
     *  A "sort order" is an abstract concept that establishes the order in which
     *  values shall occur in a sorted sequence.
     *
     *  A concrete instance of the class ISortOrder<T> must impose a weak partial
     *  order on the domain of T values. This is implemented by an Order instance
     *  being able to tell "A shall come before B", "A shall come after B" or
     *  "I don't care in which order A and B occur".
     *
     *  @param T                The type of sorted values.
     *  @since 1.0
     */
    template <class T>
    class ISortOrder
    {
        //////////
        //  This is an interface
    public:
        ISortOrder() {}
        virtual ~ISortOrder() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if a value shall appear before another value in a sorted sequence.
         *
         *  @param value1           The 1st value to compare.
         *  @param value2           The 2nd value to compare.
         *  @return                 True if \a value1 shall appear before \a value2
         *                          in the sorted sequence; false otherwise.
         *  @since 1.0
         */
        virtual bool                shallComeBefore(const T & value1, const T & value2) const = 0;
    };

    /**
     *  @class NaturalIncreasingSortOrder CARL.hpp
     *
     *  A sort order that sorts values in increasing sequence using the
     *  operators '&lt;' and '&gt;' to determine which value is larger.
     *
     *  @param T                The type of compared values.
     *  @since 1.0
     */
    template <class T>
    class NaturalIncreasingSortOrder : public virtual ISortOrder<T>
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        NaturalIncreasingSortOrder();

        /**
         *  The copy constructor.
         *
         *  @since 1.0
         */
        NaturalIncreasingSortOrder(const NaturalIncreasingSortOrder<T> &);

        /**
         *  The class destructor
         *
         *  @since 1.0
         */
        virtual ~NaturalIncreasingSortOrder();

        /**
         *  The assignment operator.
         *
         *  @return                 A reference to the assigned-to object.
         *
         *  @since 1.0
         */
        NaturalIncreasingSortOrder<T>   operator = (const NaturalIncreasingSortOrder<T> &);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a value shall appear before another value in a sorted sequence.
         *
         *  @param value1           The 1st value to compare.
         *  @param value2           The 2nd value to compare.
         *  @return                 True if \a value1 &lt; \a value2; false otherwise.
         *  @since 1.0
         */
        virtual bool                shallComeBefore(const T & value1, const T & value2) const;
    };

    template <class T>
    NaturalIncreasingSortOrder<T>::NaturalIncreasingSortOrder()
    {
    }

    template <class T>
    NaturalIncreasingSortOrder<T>::NaturalIncreasingSortOrder(const NaturalIncreasingSortOrder<T> &)
    {
    }

    template <class T>
    NaturalIncreasingSortOrder<T>::~NaturalIncreasingSortOrder()
    {
    }

    template <class T>
    NaturalIncreasingSortOrder<T> NaturalIncreasingSortOrder<T>::operator = (const NaturalIncreasingSortOrder<T> &)
    {
        return *this;
    }

    template <class T>
    bool NaturalIncreasingSortOrder<T>::shallComeBefore(const T & value1, const T & value2) const
    {
        return value1 < value2;
    }

    /**
     *  @class NaturalDecreasingSortOrder CARL.hpp
     *
     *  A sort order that sorts values in decreasing sequence using the
     *  operators '&lt;' and '&gt;' to determine which value is larger.
     *
     *  @param T                The type of compared values.
     *  @since 1.0
     */
    template <class T>
    class NaturalDecreasingSortOrder : public virtual ISortOrder<T>
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        NaturalDecreasingSortOrder();

        /**
         *  The copy constructor.
         *
         *  @since 1.0
         */
        NaturalDecreasingSortOrder(const NaturalDecreasingSortOrder<T> &);

        /**
         *  The class destructor
         *
         *  @since 1.0
         */
        virtual ~NaturalDecreasingSortOrder();

        /**
         *  The assignment operator.
         *
         *  @return                 A reference to the assigned-to object.
         *
         *  @since 1.0
         */
        NaturalDecreasingSortOrder<T>   operator = (const NaturalDecreasingSortOrder<T> &);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a value shall appear before another value in a sorted sequence.
         *
         *  @param value1           The 1st value to compare.
         *  @param value2           The 2nd value to compare.
         *  @return                 True if \a value1 &gt; \a value2; false otherwise.
         *  @since 1.0
         */
        virtual bool                shallComeBefore(const T & value1, const T & value2) const;
    };

    template <class T>
    NaturalDecreasingSortOrder<T>::NaturalDecreasingSortOrder()
    {
    }

    template <class T>
    NaturalDecreasingSortOrder<T>::NaturalDecreasingSortOrder(const NaturalDecreasingSortOrder<T> &)
    {
    }

    template <class T>
    NaturalDecreasingSortOrder<T>::~NaturalDecreasingSortOrder()
    {
    }

    template <class T>
    NaturalDecreasingSortOrder<T> NaturalDecreasingSortOrder<T>::operator = (const NaturalDecreasingSortOrder<T> &)
    {
        return *this;
    }

    template <class T>
    bool NaturalDecreasingSortOrder<T>::shallComeBefore(const T & value1, const T & value2) const
    {
        return value1 > value2;
    }

    /**
     *  @class Sort CARL.hpp
     *  A generic sorting algorithm for sorting an array of items of type T. 
     *
     *  The sorted type (T) must be:
     *
     *  -   Copy constructable (T::T(const T &) must be defined and public).
     *  -   Assignable (T::operator=(const T &) must be defined and public).
     *
     *  Specific sorting algorithms may impose additional limitations in
     *  the required interface of T.
     *
     *  @param T                The type of sorted values.
     *  @since 1.0
     */
    template <class T>
    class Sort
    {
        //////////
        //  Can't assign or copy construct
    private:
        Sort(const Sort<T> &);
        void                        operator = (const Sort<T> &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        Sort();

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Sort();

        //////////
        //  Operations
    public:
        /**
         *  Sorts a sequence of items.
         *
         *  @param pItems           A pointer to the start of the sequence to be
         *                          sorted. The sequence must be at least \a itemCount
         *                          elements long; if it is longer then only the first
         *                          \a itemCount elements are sorted.
         *  @param itemCount        The number of items in the sorted sequence.
         *  @param order            The required sort order.
         *  @since 1.0
         */
        virtual void                sort(T * pItems, UIntSize itemCount, const ISortOrder<T> & order = NaturalIncreasingSortOrder<T>()) = 0;
    };

    template <class T>
    Sort<T>::Sort()
    {
    }

    template <class T>
    Sort<T>::~Sort()
    {
    }

    /**
     *  @class QuickSort CARL.hpp
     *  Implements the quicksort sorting algorithm.
     *
     *  Quicksort is an unstable sort with the following characteristics:
     *  
     *  -   O(N*log(N)) average time complexity.
     *  -   O(N**2) worst-case time complexity.
     *  -   O(0) worst-case space requirement (i.e. in-place sorting).
     *
     *  @param T                The type of sorted values.
     *  @since 1.0
     */
    template <class T>
    class QuickSort : public Sort<T>
    {
        //////////
        //  Can't assign or copy construct
    private:
        QuickSort(const QuickSort<T> &);
        void                        operator = (const QuickSort<T> &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        QuickSort();

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~QuickSort();

        //////////
        //  Operations
    public:
        /**
         *  Sorts a sequence of items.
         *
         *  @param pItems           A pointer to the start of the sequence to be
         *                          sorted. The sequence must be at least \a itemCount
         *                          elements long; if it is longer then only the first
         *                          \a itemCount elements are sorted.
         *  @param itemCount        The number of items in the sorted sequence.
         *  @param order            The required sort order.
         *  @since 1.0
         */
        virtual void                sort(T * pItems, UIntSize itemCount, const ISortOrder<T> & order = NaturalIncreasingSortOrder<T>());

        //////////
        //  Implementation
    private:
        static void                 _swap(T & x, T & y);
    };

    template <class T>
    QuickSort<T>::QuickSort()
    {
    }

    template <class T>
    QuickSort<T>::~QuickSort()
    {
    }

    template <class T>
    void QuickSort<T>::sort(T * pItems, UIntSize itemCount, const ISortOrder<T> & order)
    {
        require(pItems != NULL || itemCount == 0);

        if (itemCount > 1) 
        {
            UIntSize n1 = 0, n2 = itemCount - 1;
            for (; ; )
            {
                while (n1 < n2 && !order.shallComeBefore(pItems[n2], pItems[n1]))
                {
                    n1++;
                }
                if (n1 == n2)
                {
                    break;
                }
                _swap(pItems[n1], pItems[n2]);
                while (n1 < n2 && !order.shallComeBefore(pItems[n2], pItems[n1]))
                {
                    n2--;
                }
                if (n1 == n2)
                {
                    break;
                }
                _swap(pItems[n1], pItems[n2]);
            }
            if (n1 > 1)
            {
                sort(pItems, n1, order);
            }
            if (n2 < itemCount - 2)
            {
                sort(pItems + n2 + 1, itemCount - n2 - 1, order);
            }
        }
    }

    template<class T>
    void QuickSort<T>::_swap(T & x, T & y)
    {
        T temp = x;
        x = y;
        y = temp;
    }

    /**
     *  @class HeapSort CARL.hpp
     *  Implemets the heap sorting algorithm.
     *
     *  Heapsort is an unstable sort with the following characteristics:
     *  
     *  -   O(N*log(N)) worst-case time complexity.
     *  -   O(0) worst-case space requirement (i.e. in-place sorting).
     *
     *  @param T                The type of sorted values.
     *  @since 1.0
     */
    template <class T>
    class HeapSort : public Sort<T>
    {
        //////////
        //  Can't assign or copy construct
    private:
        HeapSort(const HeapSort<T> &);
        void                        operator = (const HeapSort<T> &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        HeapSort();

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~HeapSort();

        //////////
        //  Operations
    public:
        /**
         *  Sorts a sequence of items.
         *
         *  @param pItems           A pointer to the start of the sequence to be
         *                          sorted. The sequence must be at least \a itemCount
         *                          elements long; if it is longer then only the first
         *                          \a itemCount elements are sorted.
         *  @param itemCount        The number of items in the sorted sequence.
         *  @param order            The required sort order.
         *  @since 1.0
         */
        virtual void                sort(T * pItems, UIntSize itemCount, const ISortOrder<T> & order = NaturalIncreasingSortOrder<T>());

        //////////
        //  Implementation
    private:
        static void                 _sift(T * pItems, UIntSize start, UIntSize count, const ISortOrder<T> & order);
        static void                 _swap(T & x, T & y);
    };

    template <class T>
    HeapSort<T>::HeapSort()
    {
    }

    template <class T>
    HeapSort<T>::~HeapSort()
    {
    }

    template <class T>
    void HeapSort<T>::sort(T * pItems, UIntSize itemCount, const ISortOrder<T> & order)
    {
        require(pItems != NULL || itemCount == 0);

        if (itemCount > 1) 
        {
            UIntSize start = itemCount / 2 - 1;
            UIntSize end = itemCount - 1;

            for (; ; )
            {
                _sift(pItems, start, itemCount, order);
                if (start == 0)
                {
                    break;
                }
                start--;
            }

            while (end > 0)
            {
                _swap(pItems[end], pItems[0]);
                _sift(pItems, 0, end, order);
                end--;
            }
        }
    }

    template<class T>
    void HeapSort<T>::_sift(T * pItems, UIntSize start, UIntSize count, const ISortOrder<T> & order)
    {
        UIntSize root = start, child;

        while (root * 2 + 1 < count)
        {
            child = root * 2 + 1;
            if (child + 1 < count && order.shallComeBefore(pItems[child], pItems[child + 1]))
            {
                child++;
            }
            if (order.shallComeBefore(pItems[root], pItems[child]))
            {
                _swap(pItems[root], pItems[child]);
                root = child;
            }
            else
            {
                return;
            }
        }
    }

    template<class T>
    void HeapSort<T>::_swap(T & x, T & y)
    {
        T temp = x;
        x = y;
        y = temp;
    }

    /**
     *  @class MergeSort CARL.hpp
     *  Implemets the merge sorting algorithm.
     *
     *  Mergesort is a stable sort with the following characteristics:
     *  
     *  -   O(N*log(N)) worst-case time complexity.
     *  -   O(N) worst-case space requirement.
     *
     *  In addition to standard requirements imposed by Sort<T>, the value type (T) must be:
     *
     *  -   Default constructable (T::T() must be defined and public).
     *  -   Destructable (T::~T() must be defined and public).
     *
     *  @param T                The type of sorted values.
     *  @since 1.0
     */
    template <class T>
    class MergeSort : public Sort<T>
    {
        //////////
        //  Can't assign or copy construct
    private:
        MergeSort(const MergeSort<T> &);
        void                        operator = (const MergeSort<T> &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        MergeSort();

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~MergeSort();

        //////////
        //  Operations
    public:
        /**
         *  Sorts a sequence of items.
         *
         *  @param pItems           A pointer to the start of the sequence to be
         *                          sorted. The sequence must be at least \a itemCount
         *                          elements long; if it is longer then only the first
         *                          \a itemCount elements are sorted.
         *  @param itemCount        The number of items in the sorted sequence.
         *  @param order            The required sort order.
         *  @since 1.0
         */
        virtual void                sort(T * pItems, UIntSize itemCount, const ISortOrder<T> & order = NaturalIncreasingSortOrder<T>());

        //////////
        //  Implementation
    private:
        static void                 _sort(T * pItems, UIntSize itemCount, const ISortOrder<T> & order, T * pAuxiliaryMemory);
    };

    template <class T>
    MergeSort<T>::MergeSort()
    {
    }

    template <class T>
    MergeSort<T>::~MergeSort()
    {
    }

    template <class T>
    void MergeSort<T>::sort(T * pItems, UIntSize itemCount, const ISortOrder<T> & order)
    {
        require(pItems != NULL || itemCount == 0);

        if (itemCount > 1)
        {
            T * pAuxiliaryMemory = new T[itemCount];
            _sort(pItems, itemCount, order, pAuxiliaryMemory);
            delete [] pAuxiliaryMemory;
        }
    }

    template <class T>
    void MergeSort<T>::_sort(T * pItems, UIntSize itemCount, const ISortOrder<T> & order, T * pAuxiliaryMemory)
    {
        require(pItems != NULL);

        if (itemCount < 2)
        {
            return;
        }
        //  Split the original list of values into two sub-lists residing in auxiliary memory
        UIntSize subList1Size = itemCount / 2;
        UIntSize subList2Size = itemCount - subList1Size;
        T * pBaseSubList1 = pItems;
        T * pBaseSubList2 = pItems + subList1Size;
        T * pAuxSubList1 = pAuxiliaryMemory;
        T * pAuxSubList2 = pAuxiliaryMemory + subList1Size;
        for (UIntSize i = 0; i < itemCount; i++)
        {
            pAuxiliaryMemory[i] = pItems[i];
        }
        //  Sort the sublists, using original array as an auxiliary memory
        _sort(pAuxSubList1, subList1Size, order, pBaseSubList1);
        _sort(pAuxSubList2, subList2Size, order, pBaseSubList2);
        //  Re-combine the sorted sublists into the original list
        UIntSize subList1Index = 0;
        UIntSize subList2Index = 0;
        UIntSize baseIndex = 0;
        while (subList1Index != subList1Size && subList2Index != subList2Size)
        {
            if (order.shallComeBefore(pAuxSubList2[subList2Index], pAuxSubList1[subList1Index]))
            {   //  Take one element from sublist 2
                pItems[baseIndex++] = pAuxSubList2[subList2Index++];
            }
            else
            {   //  Take one element from sublist 1
                pItems[baseIndex++] = pAuxSubList1[subList1Index++];
            }
        }
        //  Copy remaining elements
        for (;subList1Index < subList1Size; subList1Index++)
        {
            pItems[baseIndex++] = pAuxSubList1[subList1Index];
        }
        for (;subList2Index < subList2Size; subList2Index++)
        {
            pItems[baseIndex++] = pAuxSubList2[subList2Index];
        }
    }
}

#endif  //  __CARL_Sort_hpp_included__
//  End of CARL_Sort.hpp
