//
//  CARL_Stack.hpp - A variable-length stack.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:49 $
//  $Revision: 1.7 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines a stack data type.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Stack.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Stack_hpp_included__
#define __CARL_Stack_hpp_included__

namespace carl
{
    namespace _impl
    {
        //////////
        //  Variable-length stack implementation
        //
        //  The value type (T) must be:
        //  *   Copy constructable (T::T(const T &) must be defined and public)
        //  *   Destructable (T::~T() must be defined and public)
        //  *   Assignable (T::operator=(const T &) must be defined and public)
        template<class T>
        class _StackImpl
        {
            //////////
            //  Can't assign
        private:
            void                    operator = (const _StackImpl<T> &);

            //////////
            //  Construction/destruction
        public:
            _StackImpl();
            _StackImpl(const _StackImpl<T> & src);
            ~_StackImpl();

            //////////
            //  Operations
        public:
            bool                    isEmpty() const;
            UIntSize                size() const;
            void                    push(const T & value);
            T                       pop();
            T &                     top();
            void                    removeAll();
            Iterator<T>             enumerateElementsTopToBottom() const;
            Iterator<T>             enumerateElementsBottomToTop() const;
            UIntSize                getReferenceCount() const;
            void                    incrementReferenceCount();
            void                    decrementReferenceCount();

            //////////
            //  Implementation
        private:
            struct _Element
            {
                _Element(const T & value, _Element * pNextInChain) 
                    :   _value(value), _pNextInChain(pNextInChain)
                {
                }

                T                   _value;
                _Element *          _pNextInChain;
            };
            _Element *              _pElements;
            UIntSize                _elementCount;
            UIntSize                _referenceCount;

            //  Helpers
            _impl::_SnapshotIteratorImpl<T> *   _createTopToBottomIteratorImpl() const;
            _impl::_SnapshotIteratorImpl<T> *   _createBottomToTopIteratorImpl() const;
        };

        template<class T>
        _StackImpl<T>::_StackImpl()
        {
            _pElements = NULL;
            _elementCount = 0;
            _referenceCount = 0;
        }

        template<class T>
        _StackImpl<T>::_StackImpl(const _StackImpl<T> & src)
        {
            _pElements = NULL;
            _elementCount = 0;
            //  Copy elements from src
            for (Iterator<T> it = src.enumerateElementsBottomToTop(); !it.finished(); it.advance())
            {
                push(it.item());
            }
            _referenceCount = 0;
        }

        template<class T>
        _StackImpl<T>::~_StackImpl()
        {
            removeAll();
        }

        template<class T>
        bool _StackImpl<T>::isEmpty() const
        {
            return _elementCount == 0;
        }

        template<class T>
        UIntSize _StackImpl<T>::size() const
        {
            return _elementCount;
        }

        template<class T>
        void _StackImpl<T>::push(const T & value)
        {
            _Element * pElement = new _Element(value, _pElements);
            _pElements = pElement;
            _elementCount++;
        }

        template<class T>
        T _StackImpl<T>::pop()
        {
            require(_elementCount != 0);
            _Element * pElement = _pElements;
            _pElements = pElement->_pNextInChain;
            T result = pElement->_value;
            delete pElement;
            _elementCount--;
            return result;
        }

        template<class T>
        T & _StackImpl<T>::top()
        {
            require(_elementCount != 0);
            return _pElements->_value;
        }

        template<class T>
        void _StackImpl<T>::removeAll()
        {
            while (_elementCount != 0)
            {
                pop();
            }
        }

        template<class T>
        Iterator<T> _StackImpl<T>::enumerateElementsTopToBottom() const
        {
            return Iterator<T>(_createTopToBottomIteratorImpl());
        }

        template<class T>
        Iterator<T> _StackImpl<T>::enumerateElementsBottomToTop() const
        {
            return Iterator<T>(_createBottomToTopIteratorImpl());
        }

        template<class T>
        UIntSize _StackImpl<T>::getReferenceCount() const
        {
            return _referenceCount;
        }

        template<class T>
        void _StackImpl<T>::incrementReferenceCount()
        {
            _referenceCount++;
            require(_referenceCount != 0);
        }

        template<class T>
        void _StackImpl<T>::decrementReferenceCount()
        {
            require(_referenceCount != 0);
            _referenceCount--;
        }

        template<class T>
        _impl::_SnapshotIteratorImpl<T> * _StackImpl<T>::_createTopToBottomIteratorImpl() const
        {
            T * pElements = new T[_elementCount];
            UIntSize elementIndex = 0;
            for (_Element * pElement = _pElements; pElement != NULL; pElement = pElement->_pNextInChain)
            {
                pElements[elementIndex++] = pElement->_value;
            }
            return new _impl::_SnapshotIteratorImpl<T>(_elementCount, pElements);
        }

        template<class T>
        _impl::_SnapshotIteratorImpl<T> * _StackImpl<T>::_createBottomToTopIteratorImpl() const
        {
            T * pElements = new T[_elementCount];
            UIntSize elementIndex = _elementCount - 1;
            for (_Element * pElement = _pElements; pElement != NULL; pElement = pElement->_pNextInChain)
            {
                pElements[elementIndex--] = pElement->_value;
            }
            return new _impl::_SnapshotIteratorImpl<T>(_elementCount, pElements);
        }
    }

    /**
     *  @class Stack CARL.hpp
     *  A variable-size stack.
     *
     *  CARL stacks use implicit sharing to reduce the copy construction and 
     *  assignment overhead. This means that it is feasible to both pass
     *  stacks to functions as value parameters and return stacks from functions.
     *
     *  The value type (T) must be:
     *
     *  -   Copy constructable (T::T(const T &) must be defined and public).
     *  -   Destructable (T::~T() must be defined and public).
     *  -   Assignable (T::operator=(const T &) must be defined and public).
     *
     *  @param T                The type of stack's elements.
     *  @since 1.0
     */
    template<class T>
    class Stack
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default class constructor.
         *
         *  Creates an initially empty stack.
         *
         *  @since 1.0
         */
        Stack();

        /**
         *  The copy constructor.
         *
         *  Initializes a newly created stack to be a copy of the specified
         *  stack. stack copy construction uses lazy initialization.
         *
         *  @param src          The existing stack that shall be copied into
         *                      the newly constructed stack.
         *  @since 1.0
         */
        Stack(const Stack<T> & src);

        /**
         *  The class destructor.
         *
         *  Releases all memory used by the stack.
         *
         *  @since 1.0
         */
        ~Stack();

        /**
         *  The assignment operator.
         *
         *  Makes the assigned-to stack an exact copy of the assigned-from
         *  stack. Stack assignment uses lazy initialization.
         *
         *  @param src          The existing stack that shall be assigned 
         *                      to another stack.
         *  @return             A reference to the assigned-to stack.
         *  @since 1.0
         */
        Stack<T> &              operator = (const Stack<T> & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns true iff this stack has no elements, false otherwise.
         *
         *  @return             True iff this stack has no elements, false otherwise.
         *  @since 1.0
         */
        bool                    isEmpty() const;

        /**
         *  Returns the number of elements in this stack.
         *
         *  @return             The number of elements in this stack.
         *  @since 1.0
         */
        UIntSize                size() const;

        /**
         *  Pushes the specified value on top of the stack.
         *
         *  @param value        The value to push.
         *  @since 1.0
         */
        void                    push(const T & value);

        /**
         *  Pops the value from the top of the stack and returns it.
         *
         *  The call is an error if the stack is empty.
         *
         *  @return             The element popped from the top of the stack.
         *  @since 1.0
         */
        T                       pop();

        /**
         *  Returns the value at the top of the stack.
         *
         *  The stack remains unchanged.
         *  The call is an error if the stack is empty.
         *
         *  @return             The element at the top of the stack.
         *  @since 1.0
         */
        T                       top() const;

        /**
         *  Returns the value at the top of the stack.
         *
         *  The stack remains unchanged.
         *  The call is an error if the stack is empty.
         *
         *  @return             The element at the top of the stack.
         *  @since 1.0
         */
        T &                     top();

        /**
         *  Removes all stack elements, leaving the stack empty.
         *
         *  @since 1.0
         */
        void                    removeAll();

        /**
         *  Returns a top-down iterator for stack elements.
         *
         *  The returned iterator behaves as if a snapshot of the
         *  stack has been taken. In particular, removing elements from
         *  the stack after an iterator has been created has no effect
         *  on the iterator, which will still iterate over the entire
         *  list of stack elements as they were at the time when the
         *  iterator was created.
         *
         *  @return             A top-down iterator over all stack elements.
         *  @since 1.0
         */
        Iterator<T>             enumerateElementsTopToBottom() const;

        /**
         *  Returns a bottom-up iterator for stack elements.
         *
         *  The returned iterator behaves as if a snapshot of the
         *  stack has been taken. In particular, removing elements from
         *  the stack after an iterator has been created has no effect
         *  on the iterator, which will still iterate over the entire
         *  list of stack elements as they were at the time when the
         *  iterator was created.
         *
         *  @return             A bottom-up iterator over all stack elements.
         *  @since 1.0
         */
        Iterator<T>             enumerateElementsBottomToTop() const;

        //////////
        //  Implementation
    private:
        _impl::_StackImpl<T> *  _pImpl;
        void                    _ensureNonsharedImplementation();
    };

    template<class T>
    Stack<T>::Stack()
    {
        _pImpl = new _impl::_StackImpl<T>;
        _pImpl->incrementReferenceCount();
    }

    template<class T>
    Stack<T>::Stack(const Stack<T> & src)
    {
        _pImpl = src._pImpl;
        _pImpl->incrementReferenceCount();
    }

    template<class T>
    Stack<T>::~Stack()
    {
        _pImpl->decrementReferenceCount();
        if (_pImpl->getReferenceCount() == 0)
        {
            delete _pImpl;
        }
    }

    template<class T>
    Stack<T> & Stack<T>::operator = (const Stack<T> & src)
    {
        if (this != (&src))
        {
            _pImpl->decrementReferenceCount();
            if (_pImpl->getReferenceCount() == 0)
            {
                delete _pImpl;
            }
            _pImpl = src._pImpl;
            _pImpl->incrementReferenceCount();
        }
        return *this;
    }

    template<class T>
    bool Stack<T>::isEmpty() const
    {
        return _pImpl->isEmpty();
    }

    template<class T>
    UIntSize Stack<T>::size() const
    {
        return _pImpl->size();
    }

    template<class T>
    void Stack<T>::push(const T & value)
    {
        _ensureNonsharedImplementation();
        _pImpl->push(value);
    }

    template<class T>
    T Stack<T>::pop()
    {
        _ensureNonsharedImplementation();
        return _pImpl->pop();
    }

    template<class T>
    T Stack<T>::top() const
    {
        return _pImpl->top();
    }

    template<class T>
    T & Stack<T>::top()
    {
        _ensureNonsharedImplementation();
        return _pImpl->top();
    }

    template<class T>
    void Stack<T>::removeAll()
    {
        _ensureNonsharedImplementation();
        _pImpl->removeAll();
    }

    template<class T>
    Iterator<T> Stack<T>::enumerateElementsTopToBottom() const
    {
        return _pImpl->enumerateElementsTopToBottom();
    }

    template<class T>
    Iterator<T> Stack<T>::enumerateElementsBottomToTop() const
    {
        return _pImpl->enumerateElementsBottomToTop();
    }

    template<class T>
    void Stack<T>::_ensureNonsharedImplementation()
    {
        if (_pImpl->getReferenceCount() != 1)
        {
            _impl::_StackImpl<T> * pNewImpl = new _impl::_StackImpl<T>(*_pImpl);
            _pImpl->decrementReferenceCount();
            pNewImpl->incrementReferenceCount();
            _pImpl = pNewImpl;
        }
    }
}

#endif  //  __CARL_Stack_hpp_included__
//  End of CARL_Stack.hpp
