//
//  CARL_Sync.hpp - Thread synchronization.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:49 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL thread synchronization services.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Sync.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Sync_hpp_included__
#define __CARL_Sync_hpp_included__

namespace carl
{
    namespace _impl
    {
        class CARL_PUBLIC _CriticalSectionImpl;
        class CARL_PUBLIC _SynchronizationObjectImpl;
        class CARL_PUBLIC _MutexImpl;
        class CARL_PUBLIC _SemaphoreImpl;
        class CARL_PUBLIC _EventImpl;
    }

    /**
     *  @class CriticalSection CARL.hpp
     *  A program critical section.
     *
     *  CARL critical sections provide fast thread synchronization facilities.
     *  A critical section can be entered by at most one thread at any given
     *  time.
     *
     *  A thread is allowed to enter a critical section repeatedly. In this
     *  case it will have to leave the critical section an equal number of times
     *  before another thread can enter it.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC CriticalSection
    {
        //////////
        //  Can't assign or copy construct
    private:
        CriticalSection(const CriticalSection &);
        void                    operator = (const CriticalSection &);

        //////////
        //  Construction
    public:
        /**
         *  The default constructor.
         *
         *  The newly created critical section is not initially entered
         *  by any thread.
         *
         *  @since 1.0
         */
        CriticalSection();

        /**
         *  The class destructor.
         *
         *  An attempt to destroy a critical section while it is entered
         *  by some thread is an error.
         *
         *  @since 1.0
         */
        ~CriticalSection();

        //////////
        //  Operations
    public:
        /**
         *  Enters the critical section on behalf of the current thread.
         *
         *  If another thread is currently entered in this critical section,
         *  the call will wait until that thread leaves the critical section
         *  before returning.
         *
         *  A thread is allowed to enter a critical section repeatedly. In this
         *  case it will have to leave the critical section an equal number of times
         *  before another thread can enter it.
         *
         *  @since 1.0
         */
        void                    enter();

        /**
         *  Leaves the critical section on behalf of the current thread.
         *
         *  It is an error for a thread to leave critical section without
         *  having entered it first.
         *
         *  @since 1.0
         */
        void                    leave();

        //////////
        //  Implementation
    private:
        _impl::_CriticalSectionImpl *   _pImpl;
    };

    /**
     *  @class CriticalSectionLock CARL.hpp
     *  A helper class for entering/leaving critical sections.
     *
     *  A critical section lock is always associated with an existing
     *  critical section at the moment of construction. The critical
     *  section is entered by \a CriticalSectionLock constructor and
     *  left by \a CriticalSectionLock destructor. This allows using
     *  local \a CriticalSectionLock instances to enter critical
     *  sections for the duration of a block.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC CriticalSectionLock
    {
        //////////
        //  Can't assign or copy construct
    private:
        CriticalSectionLock(const CriticalSectionLock &);
        void                    operator = (const CriticalSectionLock &);

        //////////
        //  Construction
    public:
        /**
         *  The class constructor.
         *
         *  Associates this critical section lock with the specified critical
         *  section and enters the critical section.
         *
         *  @param criticalSection  The critical section to associate with
         *                      this critical section lock.
         *  @since 1.0
         */
        explicit CriticalSectionLock(CriticalSection & criticalSection);

        /**
         *  The class destructor.
         *
         *  Leaves the associated critical section.
         *
         *  @since 1.0
         */
        ~CriticalSectionLock();

        //////////
        //  Implementation
    private:
        CriticalSection &       _criticalSection;
    };

    /**
     *  @class SynchronizationObject CARL.hpp
     *  A generic synchronization object.
     *
     *  A synchronization object is an object on which a CARL thread can
     *  wait. Once the waiting succeeds for some thread, the thread is
     *  granted the ownership of a synchrionization object. All other
     *  threads waiting for the same synchronization object will be
     *  stalled until the owning thread releases the ownership of a
     *  synchronization object, at which time one or more of other waiting
     *  threads will awake.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC SynchronizationObject
    {
        //////////
        //  Construction
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        SynchronizationObject();

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~SynchronizationObject();

        //////////
        //  Operations
    public:
        /**
         *  Waits for this synchronization object on behalf of the
         *  calling thread.
         *
         *  The method does not return until the calling thread is
         *  granted ownership of the synchronization object.
         *
         *  A thread is allowed to wait for a synchronization object that it
         *  already has an ownership of; such wait always succeeds immediately. In this
         *  case the thread will have to release the synchronization object an equal 
         *  number of times before another thread can be granted ownwership.
         *
         *  @since 1.0
         */
        virtual void            waitFor() = 0;

        /**
         *  Waits for this synchronization object on behalf of the
         *  calling thread.
         *
         *  The method does not return until either the calling thread is
         *  granted ownership of the synchronization object or the specified
         *  timeout expires.
         *
         *  @param timeout      The maximum time interval to spend in waiting.
         *  @return             True if the calling thread has been granted
         *                      ownership of the synchronization object; false
         *                      if the method returned because the specified
         *                      timeout has expired.
         *  @since 1.0
         */
        virtual bool            waitFor(const TimeSpan & timeout) = 0;

        /**
         *  Releases this synchronization object on behalf of the
         *  calling thread.
         *
         *  It is an error for a thread to release a synchronization object
         *  that is not owned by that thread.
         *
         *  @since 1.0
         */
        virtual void            release() = 0;

        //////////
        //  Generic wait services
    public:
        /**
         *  Waits for all synchronization objects in a list.
         *
         *  The wait suceeds when the calling thread is successfully
         *  granted ownership of all synchronization objects from
         *  the specified list.
         *
         *  @param pSynchronizationObjects  The pointer to a list of synchronization
         *                      objects to wait for. The list must be at least
         *                      \a synchronizationObjectCount elements long; if it 
         *                      is longer then only the first \a synchronizationObjectCount
         *                      elements are considered.
         *  @param synchronizationObjectCount   The size of the synchronization
         *                      object list.
         *  @since 1.0
         */
        static void             waitForAll(SynchronizationObject * const pSynchronizationObjects[], UIntSize synchronizationObjectCount);

        /**
         *  Waits for all synchronization objects in a list.
         *
         *  The wait suceeds when the calling thread is successfully
         *  granted ownership of all synchronization objects from
         *  the specified list.
         *
         *  @param synchronizationObjects   The list of synchronization objects 
         *                      to wait for.
         *  @since 1.0
         */
        static void             waitForAll(const Array<SynchronizationObject*> & synchronizationObjects);

        /**
         *  Waits for all synchronization objects in a list.
         *
         *  The wait suceeds when the calling thread is successfully
         *  granted ownership of all synchronization objects from
         *  the specified list. The wait fails if by the end of the
         *  specified timeout one or more of the specified synchronization
         *  objects is still unavailable. In the latter case, no ownership
         *  is granted to the calling thread over any of the specified
         *  synchronization objects.
         *
         *  @param pSynchronizationObjects  The pointer to a list of synchronization
         *                      objects to wait for. The list must be at least
         *                      \a synchronizationObjectCount elements long; if it 
         *                      is longer then only the first \a synchronizationObjectCount
         *                      elements are considered.
         *  @param synchronizationObjectCount   The size of the synchronization
         *                      object list.
         *  @param timeout      The maximum time interval to spend in waiting.
         *  @return             True if the calling thread has been granted
         *                      ownership of all specified synchronization objects;
         *                      false if the method returned because the specified
         *                      timeout has expired.
         *  @since 1.0
         */
        static bool             waitForAll(SynchronizationObject * const pSynchronizationObjects[], UIntSize synchronizationObjectCount, const TimeSpan & timeout);

        /**
         *  Waits for all synchronization objects in a list.
         *
         *  The wait suceeds when the calling thread is successfully
         *  granted ownership of all synchronization objects from
         *  the specified list. The wait fails if by the end of the
         *  specified timeout one or more of the specified synchronization
         *  objects is still unavailable. In the latter case, no ownership
         *  is granted to the calling thread over any of the specified
         *  synchronization objects.
         *
         *  @param synchronizationObjects   The list of synchronization objects 
         *                      to wait for.
         *  @param timeout      The maximum time interval to spend in waiting.
         *  @return             True if the calling thread has been granted
         *                      ownership of all specified synchronization objects;
         *                      false if the method returned because the specified
         *                      timeout has expired.
         *  @since 1.0
         */
        static bool             waitForAll(const Array<SynchronizationObject*> & synchronizationObjects, const TimeSpan & timeout);

        /**
         *  Waits for any synchronization object from a list.
         *
         *  The wait suceeds when the calling thread is successfully
         *  granted ownership of some synchronization object from
         *  the specified list.
         *
         *  @param pSynchronizationObjects  The pointer to a list of synchronization
         *                      objects to wait for. The list must be at least
         *                      \a synchronizationObjectCount elements long; if it 
         *                      is longer then only the first \a synchronizationObjectCount
         *                      elements are considered.
         *  @param synchronizationObjectCount   The size of the synchronization
         *                      object list.
         *  @since 1.0
         */
        static SynchronizationObject *  waitForAny(SynchronizationObject * const pSynchronizationObjects[], UIntSize synchronizationObjectCount);

        /**
         *  Waits for any synchronization object from a list.
         *
         *  The wait suceeds when the calling thread is successfully
         *  granted ownership of some synchronization object from
         *  the specified list.
         *
         *  @param synchronizationObjects   The list of synchronization objects 
         *                      to wait for.
         *  @since 1.0
         */
        static SynchronizationObject *  waitForAny(const Array<SynchronizationObject*> & synchronizationObjects);

        /**
         *  Waits for any synchronization object from a list.
         *
         *  The wait suceeds when the calling thread is successfully
         *  granted ownership of some synchronization object from
         *  the specified list. The wait fails if by the end of the
         *  specified timeout all of the specified synchronization
         *  objects are still unavailable.
         *
         *  @param pSynchronizationObjects  The pointer to a list of synchronization
         *                      objects to wait for. The list must be at least
         *                      \a synchronizationObjectCount elements long; if it 
         *                      is longer then only the first \a synchronizationObjectCount
         *                      elements are considered.
         *  @param synchronizationObjectCount   The size of the synchronization
         *                      object list.
         *  @param timeout      The maximum time interval to spend in waiting.
         *  @return             True if the calling thread has been granted
         *                      ownership of one of the specified synchronization objects;
         *                      false if the method returned because the specified
         *                      timeout has expired.
         *  @since 1.0
         */
        static SynchronizationObject *  waitForAny(SynchronizationObject * const pSynchronizationObjects[], UIntSize synchronizationObjectCount, const TimeSpan & timeout);

        /**
         *  Waits for any synchronization object from a list.
         *
         *  The wait suceeds when the calling thread is successfully
         *  granted ownership of some synchronization object from
         *  the specified list. The wait fails if by the end of the
         *  specified timeout all of the specified synchronization
         *  objects are still unavailable.
         *
         *  @param synchronizationObjects   The list of synchronization objects 
         *                      to wait for.
         *  @param timeout      The maximum time interval to spend in waiting.
         *  @return             True if the calling thread has been granted
         *                      ownership of one of the specified synchronization objects;
         *                      false if the method returned because the specified
         *                      timeout has expired.
         *  @since 1.0
         */
        static SynchronizationObject *  waitForAny(const Array<SynchronizationObject*> & synchronizationObjects, const TimeSpan & timeout);
    };

    /**
     *  @class Lock CARL.hpp
     *  A helper class for acquiring/releasing synchrinization objects.
     *
     *  A synchrinization object lock is always associated with an existing
     *  synchrinization object at the moment of construction. The synchrinization
     *  object is acquired by \a Lock constructor and released by \a Lock 
     *  destructor. This allows using local \a Lock instances to acquire
     *  synchrinization objects for the duration of a block.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Lock
    {
        //////////
        //  Can't assign or copy construct
    private:
        Lock(const Lock &);
        void                    operator = (const Lock &);

        //////////
        //  Construction
    public:
        /**
         *  The class constructor.
         *
         *  Associates this synchrinization object lock with the specified
         *  synchrinization object and acquires the synchrinization object.
         *
         *  @param synchronizationObject    The synchrinization object to associate with
         *                      this synchrinization object lock.
         *  @since 1.0
         */
        explicit Lock(SynchronizationObject & synchronizationObject);

        /**
         *  The class destructor.
         *
         *  Releases the associated synchrinization object.
         *
         *  @since 1.0
         */
        ~Lock();

        //////////
        //  Implementation
    private:
        SynchronizationObject & _synchronizationObject;
    };

    /**
     *  @class Mutex CARL.hpp
     *  A mutex synchrinization object.
     *
     *  A mutex is a synchronization object that can be owned by at most one
     *  thread at a time.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Mutex : public SynchronizationObject
    {
        friend class _impl::_SynchronizationObjectImpl;

        //////////
        //  Can't assign or copy construct
    private:
        Mutex(const Mutex &);
        void                    operator = (const Mutex &);

        //////////
        //  Construction
    public:
        /**
         *  The default constructor.
         *
         *  The newly created mutex is not owned by any thread.
         *
         *  @since 1.0
         */
        Mutex();

        /**
         *  The class destructor.
         *
         *  It is an error to destroy a mutex that is currently owned
         *  by some thread.
         *
         *  @since 1.0
         */
        virtual ~Mutex();

        //////////
        //  SynchronizationObject
    public:
        /**
         *  A synonym for acquire().
         *
         *  @since 1.0
         */
        virtual void            waitFor();

        /**
         *  A synonym for acquire(const TimeSpan & timeout).
         *
         *  @param timeout      The maximum time interval to spend in waiting.
         *  @return             True if the calling thread has been granted
         *                      ownership of the synchronization object; false
         *                      if the method returned because the specified
         *                      timeout has expired.
         *  @since 1.0
         */
        virtual bool            waitFor(const TimeSpan & timeout);

        /**
         *  Releases this mutex on behalf of the calling thread.
         *
         *  It is an error for a thread to release a synchronization object
         *  that is not owned by that thread.
         *
         *  @since 1.0
         */
        virtual void            release();

        //////////
        //  Mutex
    public:
        /**
         *  Acquires this synchronization object on behalf of the
         *  calling thread.
         *
         *  The method does not return until the calling thread is
         *  granted ownership of the synchronization object.
         *
         *  A thread is allowed to wait for a synchronization object that it
         *  already has an ownership of; such wait always succeeds immediately. In this
         *  case the thread will have to release the synchronization object an equal 
         *  number of times before another thread can be granted ownwership.
         *
         *  @since 1.0
         */
        void                    acquire();

        /**
         *  Acquires this synchronization object on behalf of the
         *  calling thread.
         *
         *  The method does not return until either the calling thread is
         *  granted ownership of the synchronization object or the specified
         *  timeout expires.
         *
         *  @param timeout      The maximum time interval to spend in waiting.
         *  @return             True if the calling thread has been granted
         *                      ownership of the synchronization object; false
         *                      if the method returned because the specified
         *                      timeout has expired.
         *  @since 1.0
         */
        bool                    acquire(const TimeSpan & timeout);

        //////////
        //  Implementation
    private:
        _impl::_MutexImpl *     _pImpl;
    };

    /**
     *  @class Semaphore CARL.hpp
     *  A semaphore synchrinization object.
     *
     *  A semaphore synchronization object is an atomic integer
     *  counter that supports two operations:
     *
     *  -   down - decrements the counter by 1. If the counter was >0
     *      before the call, the operation succeeds immediately; otherwise
     *      the caller must wait for the counter to become >0 before it
     *      can be decremented.
     *  -   up - increments the counter by 1.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Semaphore : public SynchronizationObject
    {
        friend class _impl::_SynchronizationObjectImpl;

        //////////
        //  Can't assign or copy construct
    private:
        Semaphore(const Semaphore &);
        void                    operator = (const Semaphore &);

        //////////
        //  Construction
    public:
        /**
         *  The class constructor,
         *
         *  The newly creates semaphore initially has the specified
         *  counter value. The counter is not allowed to grow above
         *  this value.
         *
         *  @param initialCount The initial value for the semaphore counter.
         *  @since 1.0
         */
        explicit Semaphore(UIntSize initialCount);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Semaphore();

        //////////
        //  SynchronizationObject
    public:
        /**
         *  A synonym for down().
         *
         *  @since 1.0
         */
        virtual void            waitFor();

        /**
         *  A synonym for down(const TimeSpan & timeout).
         *
         *  @param timeout      The maximum time interval to spend in waiting.
         *  @return             True if the calling thread has been granted
         *                      ownership of the synchronization object; false
         *                      if the method returned because the specified
         *                      timeout has expired.
         *  @since 1.0
         */
        virtual bool            waitFor(const TimeSpan & timeout);

        /**
         *  A synonym for up().
         *
         *  @since 1.0
         */
        virtual void            release();

        //////////
        //  Semaphore
    public:
        /**
         *  Decrements the counter by 1.
         *
         *  The method does not return until the calling thread has
         *  suceeded in decrementing the counter.
         *
         *  @since 1.0
         */
        void                    down();

        /**
         *  Decrements the counter by 1.
         *
         *  The method does not return until the calling thread has
         *  suceeded in decrementing the counter or the specified
         *  timeout has expired.
         *
         *  @param timeout      The maximum time interval to spend in waiting.
         *  @return             True if the calling thread has succeeded in
         *                      decrementing the semaphore's counter; false
         *                      if the method returned because the specified
         *                      timeout has expired.
         *  @since 1.0
         */
        bool                    down(const TimeSpan & timeout);

        /**
         *  Increments the counter by 1.
         *
         *  It is an error to increment the counter beyond the initial
         *  value specified when a semaphore is constructed.
         *
         *  @since 1.0
         */
        void                    up();

        //////////
        //  Implementation
    private:
        _impl::_SemaphoreImpl * _pImpl;
    };

    /**
     *  @class Event CARL.hpp
     *  An event synchrinization object.
     *
     *  An event is a synchronization object that can be in either signaled
     *  or nonsignaled state. Events can be created with either automatic
     *  or manual reset mode:
     *
     *  -   An automatic-reset event, when being set to a signaled state,
     *      causes one thread waiting on the event to be released, after
     *      which the event becomes nonsignaled again.
     *  -   A manual-reset event, once set to a signaled state, remains in
     *      that state until explicitly set to a non-signaled state.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Event : public SynchronizationObject
    {
        friend class _impl::_SynchronizationObjectImpl;

        //////////
        //  Can't assign or copy construct
    private:
        Event(const Event &);
        void                    operator = (const Event &);

        //////////
        //  Construction
    public:
        /**
         *  The class constructor.
         *
         *  @param manualReset  True if the newly created event shall use manual
         *                      reset; false if it shall use automatic reset.
         *  @param initiallySet True if the newly created event shall be initially
         *                      signaled; false otherwise.
         *  @since 1.0
         */
        Event(bool manualReset, bool initiallySet);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~Event();

        //////////
        //  SynchronizationObject
    public:
        /**
         *  Waits for this synchronization object on behalf of the
         *  calling thread.
         *
         *  The method does not return until the calling thread is
         *  granted ownership of the synchronization object.
         *
         *  A thread is allowed to wait for a synchronization object that it
         *  already has an ownership of; such wait always succeeds immediately. In this
         *  case the thread will have to release the synchronization object an equal 
         *  number of times before another thread can be granted ownwership.
         *
         *  @since 1.0
         */
        virtual void            waitFor();

        /**
         *  Waits for this synchronization object on behalf of the
         *  calling thread.
         *
         *  The method does not return until either the calling thread is
         *  granted ownership of the synchronization object or the specified
         *  timeout expires.
         *
         *  @param timeout      The maximum time interval to spend in waiting.
         *  @return             True if the calling thread has been granted
         *                      ownership of the synchronization object; false
         *                      if the method returned because the specified
         *                      timeout has expired.
         *  @since 1.0
         */
        virtual bool            waitFor(const TimeSpan & timeout);

        /**
         *  Releases this synchronization object on behalf of the
         *  calling thread.
         *
         *  It is an error for a thread to release a synchronization object
         *  that is not owned by that thread.
         *
         *  @since 1.0
         */
        virtual void            release();

        //////////
        //  Event
    public:
        /**
         *  Sets this event to a signaled state.
         *
         *  For a manual-reset event, the event remains signaled until
         *  reset() is called. For an automatic-reset event, the event
         *  remains set until a single thread waiting for that event is
         *  released, at which time the event is automatically set to
         *  the nonsignaled state.
         *
         *  @since 1.0
         */
        void                    set();

        /**
         *  Sets this event to a nonsignaled state.
         *
         *  @since 1.0
         */
        void                    reset();

        //////////
        //  Implementation
    private:
        _impl::_EventImpl *     _pImpl;
    };
}

#endif  //  __CARL_Sync_hpp_included__
//  End of CARL_Sync.hpp
