//
//  CARL_System.hpp - System-specific constants and limits
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/07/28 10:40:28 $
//  $Revision: 1.12 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines system properties API.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_System.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_System_hpp_included__
#define __CARL_System_hpp_included__

namespace carl
{
    /**
     *  @class System CARL.hpp
     *  Provides general information about target platform and CARL library.
     *
     *  @since 1.0
     */
    //////////
    //  Encapsulated constants & limits
    class CARL_PUBLIC System
    {
        //////////
        //  Utility class cannot be instantiated
    private:
        System();
        System(const System &);
        ~System();
        void                        operator = (const System &);

        //////////
        //  Host byte order
    public:
#if defined(CARL_CPU_IA32)
        /**
         *  The byte order used by the host processor.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const ByteOrder      HostByteOrder   = ByteOrder_LittleEndian;
#else
    #error Unsupported CPU
#endif

        //////////
        //  Byte & native word sizes, each expressed in bits
    public:
#if defined(CARL_CPU_IA32)
        /**
         *  The number of bits per byte for the host processor.
         *
         *  The "byte" is defined as the smallest addressable unit of memory
         *  that can be accessed by the hosr processor.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntSize       BitsPerByte     = 8;

        /**
         *  The number of bits per word for the host processor.
         *
         *  The "word" is defined as the preferred size of an integer
         *  data item for the host processor.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const UIntSize       BitsPerWord     = 32;
#else
    #error Unsupported CPU
#endif

        //////////
        //  Library identification
    public:
        /**
         *  The major version number of the CARL library.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const int            VersionMajor    = 1;

        /**
         *  The minor version number of the CARL library.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const int            VersionMinor    = 0;

        /**
         *  The patch number of the CARL library.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const int            PatchLevel      = 0;

        /**
         *  The build key associated with the CARL library.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        static const char *         getBuildKey();
    };
}

//////////
//  The configuration-specific build key. Two binaries with the
//  same build key are guaranteed to be binary-compatible.
#if defined(CARL_TOOLCHAIN_MSVC6)
    #define CARL_BUILD_KEY_TOOLCHAIN_PORTION    "MSVC6"
#elif defined(CARL_TOOLCHAIN_MSVCNET2003)
    #define CARL_BUILD_KEY_TOOLCHAIN_PORTION    "MSVCNET2003"
#elif defined(CARL_TOOLCHAIN_MSVCNET2005)
    #define CARL_BUILD_KEY_TOOLCHAIN_PORTION    "MSVCNET2005"
#elif defined(CARL_TOOLCHAIN_GCC) 
    #define CARL_BUILD_KEY_TOOLCHAIN_PORTION    "GCC"
#else
    #error Unsupported platform
#endif

#if defined(CARL_CPU_IA32)
    #define CARL_BUILD_KEY_CPU_PORTION          "IA32"
#else
    #error Unsupported CPU
#endif

#if defined(CARL_OS_WIN32)
    #define CARL_BUILD_KEY_OS_PORTION           "WIN32"
#elif defined(CARL_OS_CYGWIN32)
    #define CARL_BUILD_KEY_OS_PORTION           "CYGWIN32"
#elif defined(CARL_OS_LINUX32)
    #define CARL_BUILD_KEY_OS_PORTION           "LINUX32"
#else
    #error Unsupported platform
#endif

#if defined(CARL_MODE_DEBUG)
    #define CARL_BUILD_KEY_MODE_PORTION         "DEBUG"
#elif defined(CARL_MODE_RELEASE)
    #define CARL_BUILD_KEY_MODE_PORTION         "RELEASE"
#else
    #error Unsupported mode
#endif

#if defined(CARL_LINKAGE_STATIC)
    #define CARL_BUILD_KEY_LINKAGE_PORTION      "STATIC"
#elif defined(CARL_LINKAGE_DYNAMIC)
    #define CARL_BUILD_KEY_LINKAGE_PORTION      "DYNAMIC"
#else
    #error Unsupported linkage
#endif

/**
 *  The build key associated with the CARL client.
 *
 *  @hideinitializer
 *  @since 1.0
 */
#define CARL_BUILD_KEY  CARL_BUILD_KEY_TOOLCHAIN_PORTION "_"        \
                        CARL_BUILD_KEY_CPU_PORTION "_"              \
                        CARL_BUILD_KEY_OS_PORTION "_"               \
                        CARL_BUILD_KEY_MODE_PORTION "_"             \
                        CARL_BUILD_KEY_LINKAGE_PORTION "_"          \
                        "1_0"

#endif  //  __CARL_System_hpp_included__
//  End of CARL_System.hpp
