//
//  CARL_Thread.hpp - A thread abstraction.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:49 $
//  $Revision: 1.12 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL threading framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Thread.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Thread_hpp_included__
#define __CARL_Thread_hpp_included__

namespace carl
{
    /**
     *  @class Thread CARL.hpp
     *  A thread of control within a CARL application.
     *
     *  Each Thread instance corresponds to a potential thread of control.
     *  The actual thread is not created until Thread::start() is called,
     *  at which time the new thread starts to run by executing the Thread's
     *  run() method. Once the method exits, the thread finishes. A thread
     *  that has finished cannot be re-started.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Thread
    {
        //////////
        //  Can't assign or copy construct
    private:
        Thread(const Thread &);
        void                                operator = (const Thread &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  Creates, but does not start, a new thread.
         *
         *  @since 1.0
         */
        Thread();
    private:
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        Thread(UIntSize hcrtThread, UIntSize threadId);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        Thread(UInt32 pthreadId);
#else
    #error Unsupported platform
#endif
    public:
        /**
         *  The class destructor.
         *
         *  If the thread is still running when a Thread instance is
         *  destroyed, it is forced to terminate.
         *
         *  @since 1.0
         */
        virtual ~Thread();

        //////////
        //  Operations
    public:
        /**
         *  Represents possible states of a thread.
         *
         *  @since 1.0
         */
        enum State
        {
            /**
             *  The Thread instance has been initialized, but the corresponding
             *  thread has not yet started.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            State_Initialized   = 1,

            /**
             *  The Thread instance has been initialized, and the corresponding
             *  thread is currently running.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            State_Running       = 2,

            /**
             *  The Thread instance has been initialized, and the corresponding
             *  thread has already finished running.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            State_Finished      = 3
        };

        /**
         *  Represents possible priorities of a thread.
         *
         *  @since 1.0
         */
        enum Priority
        {
            /**
             *  Idle threads run only when there are no other threads to run.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Priority_Idle           = 1,

            /**
             *  The lowest supported thread priority that is still not Idle.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Priority_Lowest         = 2,

            /**
             *  Threads with thhis priority offer a slightly degrated performance
             *  as compared to normal threads.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Priority_BelowNormal    = 3,

            /**
             *  Normal thread priority, to be used by majority of threads.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Priority_Normal         = 4,

            /**
             *  Threads with this priority receive better treatment than normal threads.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Priority_AboveNormal    = 5,

            /**
             *  The highest possible thread priority that still does not hog CPU resources.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Priority_Highest        = 6,

            /**
             *  Time critical threads run until they voluntarily give up CPU resources.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Priority_TimeCritical   = 7
        };

        /**
         *  Returns a pointer to the Thread instance for the calling thread.
         *
         *  @return                 A pointer to the Thread instance for 
         *                          the calling thread.
         *  @since 1.0
         */
        static Thread *                     getCurrentThread();

        /**
         *  Returns the current state of this thread.
         *
         *  @return                 The current state of this thread.
         *  @since 1.0
         */
        State                               getState() const;

        /**
         *  Returns the current priority of this thread.
         *
         *  @return                 The current priority of this thread.
         *  @since 1.0
         */
        Priority                            getPriority() const;

        /**
         *  Changes the current priority of this thread.
         *
         *  Changing a priority of a thread that is initialized but not yet
         *  started has no immediate effect; the new priority will be effected
         *  when the thread is actually started.
         *
         *  Changing a proproty if a thread that has already finished
         *  has no effect.
         *
         *  @param priority         The new priority for this thread.
         *  @since 1.0
         */
        void                                setPriority(Priority priority);

        /**
         *  Starts the thread.
         *
         *  The call is an error if the thread is currently running or
         *  has already finished.
         *
         *  @since 1.0
         */
        void                                start();

        /**
         *  Called by framework to run the thread.
         *
         *  The thread finishes when this method returns. The default
         *  implementstion does nothing and returns immediately.
         *
         *  @since 1.0
         */
        virtual void                        run();

        /**
         *  Puts the calling thread to an idle sleep for the specified
         *  duration of time.
         *
         *  @param timeSpan         The time interval to sleep for.
         *  @since 1.0
         */
        static void                         sleep(const TimeSpan & timeSpan);

        //////////
        //  Implementation
    private:
        mutable CriticalSection             _guard;
        State                               _state;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        UIntSize                            _hcrtThread;
        UIntSize                            _threadId;
        static UIntSize __stdcall           _threadProc(void * pParam);
        static HashMap<UIntSize, Thread*> * _getMapThreadIdsToThreads();
        static CriticalSection *            _getMapThreadIdsToThreadsGuard();
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        Priority                            _currentPriority;
        UInt32                              _pthreadId; //  Only valid while state == Running
        static void *                       _threadProc(void * pParam);
        static HashMap<UInt32, Thread*> *   _getMapThreadIdsToThreads();
        static CriticalSection *            _getMapThreadIdsToThreadsGuard();
#else
    #error Unsupported platform
#endif
    };
}

#endif  //  __CARL_Thread_hpp_included__
//  End of CARL_Thread.hpp
