//
//  BinaryFixed.cpp - Implementation of the carl::BinaryFixed class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/08/08 14:42:28 $
//  $Revision: 1.10 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <float.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <math.h>
#else
    #error Unsupported platform
#endif
using namespace carl;

UIntSize BinaryFixed::_defaultDivisionPrecision = 0;

//////////
//  Construction/destruction/assignment
BinaryFixed::BinaryFixed(bool isNegative, const BinaryInteger & magnitude, UInt32 scale)
    :   _isNegative(isNegative),
        _magnitude(magnitude),
        _scale(scale)
{
    _normalize();
}

BinaryFixed::BinaryFixed()
    :   _isNegative(false),
        _magnitude(),
        _scale()
{
}

BinaryFixed::BinaryFixed(Int8 value)
    :   _isNegative(false),
        _magnitude(value),
        _scale(0)
{
    if (value < 0)
    {
        _magnitude = -_magnitude;
        _isNegative = true;
    }
    _normalize();
}

BinaryFixed::BinaryFixed(Int16 value)
    :   _isNegative(false),
        _magnitude(value),
        _scale(0)
{
    if (value < 0)
    {
        _magnitude = -_magnitude;
        _isNegative = true;
    }
    _normalize();
}

BinaryFixed::BinaryFixed(Int32 value)
    :   _isNegative(false),
        _magnitude(value),
        _scale(0)
{
    if (value < 0)
    {
        _magnitude = -_magnitude;
        _isNegative = true;
    }
    _normalize();
}

BinaryFixed::BinaryFixed(Int64 value)
    :   _isNegative(false),
        _magnitude(value),
        _scale(0)
{
    if (value < 0)
    {
        _magnitude = -_magnitude;
        _isNegative = true;
    }
    _normalize();
}

BinaryFixed::BinaryFixed(UInt8 value)
    :   _isNegative(false),
        _magnitude(value),
        _scale(0)
{
    _normalize();
}

BinaryFixed::BinaryFixed(UInt16 value)
    :   _isNegative(false),
        _magnitude(value),
        _scale(0)
{
    _normalize();
}

BinaryFixed::BinaryFixed(UInt32 value)
    :   _isNegative(false),
        _magnitude(value),
        _scale(0)
{
    _normalize();
}

BinaryFixed::BinaryFixed(UInt64 value)
    :   _isNegative(false),
        _magnitude(value),
        _scale(0)
{
    _normalize();
}

BinaryFixed::BinaryFixed(Real32 value)
{
    _fromReal(value);
}

BinaryFixed::BinaryFixed(Real64 value)
{
    _fromReal(value);
}

BinaryFixed::BinaryFixed(const BinaryInteger & value)
    :   _isNegative(false),
        _magnitude(value),
        _scale(0)
{
    if (value < 0)
    {
        _magnitude = -_magnitude;
        _isNegative = true;
    }
    _normalize();
}

BinaryFixed::BinaryFixed(const BinaryInteger & magnitude, UInt32 scale)
    :   _isNegative(false),
        _magnitude(magnitude),
        _scale(scale)
{
    if (magnitude < 0)
    {
        _magnitude = -_magnitude;
        _isNegative = true;
    }
    _normalize();
}

BinaryFixed::BinaryFixed(const char * pValueString) throw (ParseException)
{
    require(pValueString != NULL);

    _fromString(String(pValueString), 10);
}

BinaryFixed::BinaryFixed(const Char * pValueString) throw (ParseException)
{
    require(pValueString != NULL);

    _fromString(String(pValueString), 10);
}

BinaryFixed::BinaryFixed(const String & valueString) throw (ParseException)
{
    _fromString(valueString, 10);
}

BinaryFixed::BinaryFixed(const char * pValueString, UIntSize radix) throw (ParseException)
{
    require(pValueString != NULL);
    require(radix >= 2 && radix <= 36);

    _fromString(String(pValueString), radix);
}

BinaryFixed::BinaryFixed(const Char * pValueString, UIntSize radix) throw (ParseException)
{
    require(pValueString != NULL);
    require(radix >= 2 && radix <= 36);

    _fromString(String(pValueString), radix);
}

BinaryFixed::BinaryFixed(const String & valueString, UIntSize radix) throw (ParseException)
{
    require(radix >= 2 && radix <= 36);

    _fromString(valueString, radix);
}

BinaryFixed::BinaryFixed(const BinaryFixed & src)
    :   _isNegative(src._isNegative),
        _magnitude(src._magnitude),
        _scale(src._scale)
{
}

BinaryFixed::~BinaryFixed()
{
}

BinaryFixed & BinaryFixed::operator = (const BinaryFixed & src)
{
    if (this != (&src))
    {
        _isNegative = src._isNegative;
        _magnitude = src._magnitude;
        _scale = src._scale;
    }
    return *this;
}

//////////
//  Operations
UIntSize BinaryFixed::getDefaultDivisionPrecision()
{
    CriticalSectionLock lock(*_getStaticGuard());
    return _defaultDivisionPrecision;
}

void BinaryFixed::setDefaultDivisionPrecision(UIntSize precision)
{
    CriticalSectionLock lock(*_getStaticGuard());
    _defaultDivisionPrecision = precision;
}

BinaryFixed BinaryFixed::operator + () const
{
    return *this;
}

BinaryFixed BinaryFixed::operator - () const
{
    return BinaryFixed(!_isNegative, _magnitude, _scale);
}

BinaryFixed BinaryFixed::operator + (const BinaryFixed & op2) const
{
    if (_magnitude == 0)
    {
        return op2;
    }
    else if (op2._magnitude == 0)
    {
        return (*this);
    }
    else
    {   //  Both operands are finite non-zeroes
        //  Bring both operands to the same scale
        BinaryInteger magnitude1 = _magnitude;
        UInt32 scale1 = _scale;
        BinaryInteger magnitude2 = op2._magnitude;
        UInt32 scale2 = op2._scale;
        if (scale1 < scale2)
        {
            magnitude1 <<= (scale2 - scale1);
            scale1 = scale2;
        }
        if (scale1 > scale2)
        {
            magnitude2 <<= (scale1 - scale2);
            scale2 = scale1;
        }
        //  Perform the addition
        if (_isNegative == op2._isNegative)
        {   //  Both operands have the same sign
            return BinaryFixed(_isNegative, magnitude1 + magnitude2, scale1);
        }
        else if (magnitude1 >= magnitude2)
        {   //  Result has the sign of 1st operand
            return BinaryFixed(_isNegative, magnitude1 - magnitude2, scale1);
        }
        else
        {   //  Result has the sign of 2nd operand
            return BinaryFixed(op2._isNegative, magnitude2 - magnitude1, scale1);
        }
    }
}

BinaryFixed BinaryFixed::operator - (const BinaryFixed & op2) const
{
    if (_magnitude == 0)
    {
        return -op2;
    }
    else if (op2._magnitude == 0)
    {
        return (*this);
    }
    else
    {   //  Both operands are finite non-zeroes
        //  Bring both operands to the same scale
        BinaryInteger magnitude1 = _magnitude;
        UInt32 scale1 = _scale;
        BinaryInteger magnitude2 = op2._magnitude;
        UInt32 scale2 = op2._scale;
        if (scale1 < scale2)
        {
            magnitude1 <<= (scale2 - scale1);
            scale1 = scale2;
        }
        if (scale1 > scale2)
        {
            magnitude2 <<= (scale1 - scale2);
            scale2 = scale1;
        }
        //  Perform the subtraction
        if (_isNegative != op2._isNegative)
        {   //  Operands have different signs
            return BinaryFixed(_isNegative, magnitude1 + magnitude2, scale1);
        }
        else if (magnitude1 >= magnitude2)
        {   //  Result has the sign of 1st operand
            return BinaryFixed(_isNegative, magnitude1 - magnitude2, scale1);
        }
        else
        {   //  Result has the inverse sign of 1st operand
            return BinaryFixed(!_isNegative, magnitude2 - magnitude1, scale1);
        }
    }
}

BinaryFixed BinaryFixed::operator * (const BinaryFixed & op2) const
{
    if (_magnitude == 0 || op2._magnitude == 0)
    {
        return 0;
    }
    else if (_isNegative && op2._isNegative)
    {
        return (-(*this)) * (-op2);
    }
    else if (_isNegative && !op2._isNegative)
    {
        return -((-(*this)) * op2);
    }
    else if (!_isNegative && op2._isNegative)
    {
        return -((*this) * (-op2));
    }
    else
    {   //  Both operands are non-zero positive values
        return BinaryFixed(_isNegative != op2._isNegative, _magnitude * op2._magnitude, _scale + op2._scale);
    }
}

BinaryFixed BinaryFixed::operator / (const BinaryFixed & op2) const
{
    if (_magnitude == 0)
    {
        return 0;
    }
    else if (op2._magnitude == 0)
    {
        throw FixedPointDivisionByZeroError();
    }
    else if (_isNegative && op2._isNegative)
    {
        return (-(*this)) / (-op2);
    }
    else if (_isNegative && !op2._isNegative)
    {
        return -((-(*this)) / op2);
    }
    else if (!_isNegative && op2._isNegative)
    {
        return -((*this) / (-op2));
    }
    else
    {   //  Both operands are non-zero positive values
        CriticalSectionLock lock(*_getStaticGuard());
        /*
        //  Decide on the required scale of the result
        UInt32 op2DigitsBeforeDecimalPoint = op2._magnitude.getRequiredBits() - 1;
        if (op2DigitsBeforeDecimalPoint > _scale)
        {
            op2DigitsBeforeDecimalPoint -= _scale;
        }
        else
        {
            op2DigitsBeforeDecimalPoint = 1;
        }
        UInt32 resultScale = Math::max(_defaultDivisionPrecision, Math::max(_scale + op2DigitsBeforeDecimalPoint, op2._scale));
        BinaryInteger resultMagnitude = (_magnitude << (op2._scale + resultScale - _scale)) / op2._magnitude;
        return BinaryFixed(false, resultMagnitude, resultScale);
        */
        //  Decide on the required precision of the result
        UInt32 resultBits = _magnitude.getRequiredBits() + op2._magnitude.getRequiredBits();
        //  Divide
        BinaryInteger resultMagnitude = (_magnitude << resultBits) / op2._magnitude;
        UInt32 deltaScale = 0;
        if (resultBits + _scale < op2._scale)
        {
            deltaScale = op2._scale - (resultBits + _scale);
            resultMagnitude <<= deltaScale;
        }
        return BinaryFixed(resultMagnitude, resultBits + deltaScale + _scale - op2._scale);
    }
}

BinaryFixed & BinaryFixed::operator += (const BinaryFixed & op2)
{
    (*this) = (*this) + op2;
    return *this;
}

BinaryFixed & BinaryFixed::operator -= (const BinaryFixed & op2)
{
    (*this) = (*this) - op2;
    return *this;
}

BinaryFixed & BinaryFixed::operator *= (const BinaryFixed & op2)
{
    (*this) = (*this) * op2;
    return *this;
}

BinaryFixed & BinaryFixed::operator /= (const BinaryFixed & op2)
{
    (*this) = (*this) / op2;
    return *this;
}

BinaryFixed BinaryFixed::operator ++()
{
    (*this) += 1;
    return *this;
}

BinaryFixed BinaryFixed::operator --()
{
    (*this) -= 1;
    return *this;
}

BinaryFixed BinaryFixed::operator ++(int)
{
    BinaryFixed result = (*this);
    (*this) += 1;
    return result;
}

BinaryFixed BinaryFixed::operator --(int)
{
    BinaryFixed result = (*this);
    (*this) -= 1;
    return result;
}

int BinaryFixed::compare(const BinaryFixed & op2) const
{
    BinaryFixed diff = (*this) - op2;
    if (diff._isNegative)
    {
        return -1;
    }
    else if (diff._magnitude != 0)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

bool BinaryFixed::operator == (const BinaryFixed & op2) const
{
    return compare(op2) == 0;
}

bool BinaryFixed::operator != (const BinaryFixed & op2) const
{
    return compare(op2) != 0;
}

bool BinaryFixed::operator < (const BinaryFixed & op2) const
{
    return compare(op2) < 0;
}

bool BinaryFixed::operator <= (const BinaryFixed & op2) const
{
    return compare(op2) <= 0;
}

bool BinaryFixed::operator > (const BinaryFixed & op2) const
{
    return compare(op2) > 0;
}

bool BinaryFixed::operator >= (const BinaryFixed & op2) const
{
    return compare(op2) >= 0;
}

Int8 BinaryFixed::toInt8(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int8_Max)
    {
        return Limits::Int8_Max;
    }
    else if (value64 < Limits::Int8_Min)
    {
        return Limits::Int8_Min;
    }
    else
    {
        return static_cast<Int8>(value64);
    }
}

Int16 BinaryFixed::toInt16(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int16_Max)
    {
        return Limits::Int16_Max;
    }
    else if (value64 < Limits::Int16_Min)
    {
        return Limits::Int16_Min;
    }
    else
    {
        return static_cast<Int16>(value64);
    }
}

Int32 BinaryFixed::toInt32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int32_Max)
    {
        return Limits::Int32_Max;
    }
    else if (value64 < Limits::Int32_Min)
    {
        return Limits::Int32_Min;
    }
    else
    {
        return static_cast<Int32>(value64);
    }
}

Int64 BinaryFixed::toInt64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    BinaryInteger value = toBinaryInteger(roundingMode);
    if (value < Limits::Int64_Min)
    {
        return Limits::Int64_Min;
    }
    else if (value > Limits::Int64_Max)
    {
        return Limits::Int64_Max;
    }
    else
    {
        return value.toInt64();
    }
}

UInt8 BinaryFixed::toUInt8(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt8_Max)
    {
        return Limits::UInt8_Max;
    }
    else
    {
        return static_cast<UInt8>(value64);
    }
}

UInt16 BinaryFixed::toUInt16(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt16_Max)
    {
        return Limits::UInt16_Max;
    }
    else
    {
        return static_cast<UInt16>(value64);
    }
}

UInt32 BinaryFixed::toUInt32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt32_Max)
    {
        return Limits::UInt32_Max;
    }
    else
    {
        return static_cast<UInt32>(value64);
    }
}

UInt64 BinaryFixed::toUInt64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    BinaryInteger value = toBinaryInteger(roundingMode);
    if (value < Limits::UInt64_Min)
    {
        return Limits::UInt64_Min;
    }
    else if (value > Limits::UInt64_Max)
    {
        return Limits::UInt64_Max;
    }
    else
    {
        return value.toUInt64();
    }
}

UIntSize BinaryFixed::toUIntSize(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UIntSize_Max)
    {
        return Limits::UIntSize_Max;
    }
    else
    {
        return static_cast<UIntSize>(value64);
    }
}

UIntPtr BinaryFixed::toUIntPtr(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UIntPtr_Max)
    {
        return Limits::UIntPtr_Max;
    }
    else
    {
        return static_cast<UIntPtr>(value64);
    }
}

IntPtrDiff BinaryFixed::toIntPtrDiff(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::IntPtrDiff_Max)
    {
        return Limits::IntPtrDiff_Max;
    }
    else if (value64 < Limits::IntPtrDiff_Min)
    {
        return Limits::IntPtrDiff_Min;
    }
    else
    {
        return static_cast<IntPtrDiff>(value64);
    }
}

Real32 BinaryFixed::toReal32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    return static_cast<Real32>(toReal64(roundingMode));
}

Real64 BinaryFixed::toReal64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    if (_isNegative)
    {   //  Make sure the translated value is positive
        BinaryFixed minusThis = -(*this);
        switch (roundingMode)
        {
            case RoundingMode_TowardsNearest:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsZero:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsPositiveInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsNegativeInfinity));
            case RoundingMode_TowardsNegativeInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsPositiveInfinity));
            default:
                require(false);
                return 0;
        }
    }
    else
    {
        Real64 addend = 1.0;
        for (UInt32 scale = 0; scale < _scale; scale++)
        {
            addend /= 2.0;
        }
        Real64 result = 0;
        for (UIntSize bitNo = 0; bitNo < _magnitude.getRequiredBits(); bitNo++)
        {
            if (_magnitude.getBit(bitNo) != 0)
            {
                result += addend;
            }
            addend *= 2.0;
        }
        return result;
    }
}

BinaryInteger BinaryFixed::toBinaryInteger(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    if (_isNegative)
    {   //  Make sure the translated value is positive
        BinaryFixed minusThis = -(*this);
        switch (roundingMode)
        {
            case RoundingMode_TowardsNearest:
                return -(minusThis.toBinaryInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsZero:
                return -(minusThis.toBinaryInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsInfinity:
                return -(minusThis.toBinaryInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsPositiveInfinity:
                return -(minusThis.toBinaryInteger(RoundingMode_TowardsNegativeInfinity));
            case RoundingMode_TowardsNegativeInfinity:
                return -(minusThis.toBinaryInteger(RoundingMode_TowardsPositiveInfinity));
            default:
                require(false);
                return 0;
        }
    }
    else if (_scale == 0)
    {   //  No rounding is necessary
        return _magnitude;
    }
    else
    {   //  Rounding may be necessary
        BinaryInteger result = _magnitude;
        UIntSize nonzeroBitsShiftedOut = 0;
        int lastBitShiftedOut = 0;
        for (UInt32 bitNo = 0; bitNo > _scale; bitNo--)
        {
            if ((lastBitShiftedOut = result.getBit(bitNo)) != 0)
            {
                nonzeroBitsShiftedOut++;
            }
        }
        result >>= _scale;
        if (nonzeroBitsShiftedOut != 0)
        {   //  Rounding is necessary
            switch (roundingMode)
            {
                case RoundingMode_TowardsNearest:
                    if (lastBitShiftedOut != 0)
                    {
                        if (nonzeroBitsShiftedOut == 1)
                        {   //  Only one '1' bit was shifted out, and that was the last bit.
                            //  The value is exactly between two integers; chose an odd one
                            if (result.getBit(0) != 0)
                            {
                                result++;
                            }
                        }
                        else
                        {   //  More than one '1' bit was shifted out
                            result++;
                        }
                    }
                    break;
                case RoundingMode_TowardsZero:
                    break;
                case RoundingMode_TowardsInfinity:
                    result++;
                    break;
                case RoundingMode_TowardsPositiveInfinity:
                    result++;
                    break;
                case RoundingMode_TowardsNegativeInfinity:
                    break;
                default:
                    require(false);
            }
        }
        return result;
    }
}

String BinaryFixed::toString(UIntSize radix /*= 10*/) const
{
    RealFormatProperties formatProperties;
    formatProperties.setRadix(radix);
    return NumberFormat::getInstance(Locale())->format(*this, formatProperties);
}

int BinaryFixed::getSign() const
{
    if (_isNegative)
    {
        return -1;
    }
    else if (_magnitude != 0)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

BinaryInteger BinaryFixed::getMagnitude() const
{
    return _magnitude;
}

UInt32 BinaryFixed::getScale() const
{
    return _scale;
}

void BinaryFixed::setScale(UInt32 scale, RoundingMode roundingMode /*= RoundingMode_TowardsZero*/)
{
    if (scale == _scale)
    {
        return;
    }
    else if (scale > _scale)
    {   //  Scale increased, no rounding needed
        _magnitude <<= (scale - _scale);
        _scale = scale;
    }
    else
    {   //  Scale decreased, rounding in effect
        UIntSize bitsToRemove = (_scale - scale);
        UIntSize nonzeroBitsShiftedOut = 0;
        int lastBitShiftedOut = 0;
        for (UIntSize i = 0; i < bitsToRemove; i++)
        {
            if ((lastBitShiftedOut = _magnitude.getBit(i)) != 0)
            {
                nonzeroBitsShiftedOut++;
            }
        }
        _magnitude >>= bitsToRemove;
        _scale = scale;
        _normalize();
        if (nonzeroBitsShiftedOut != 0)
        {   //  Rounding is necessary
            switch (roundingMode)
            {
                case RoundingMode_TowardsNearest:
                    if (lastBitShiftedOut != 0)
                    {
                        if (nonzeroBitsShiftedOut == 1)
                        {   //  Only one '1' bit was shifted out, and that was the last bit.
                            //  The value is exactly between two integers; chose an odd one
                            if (_magnitude.getBit(0) != 0)
                            {
                                _magnitude++;
                            }
                        }
                        else
                        {   //  More than one '1' bit was shifted out
                            _magnitude++;
                        }
                    }
                    break;
                case RoundingMode_TowardsZero:
                    break;
                case RoundingMode_TowardsInfinity:
                    _magnitude++;
                    break;
                case RoundingMode_TowardsPositiveInfinity:
                    if (!_isNegative)
                    {
                        _magnitude++;
                    }
                    break;
                case RoundingMode_TowardsNegativeInfinity:
                    if (_isNegative)
                    {
                        _magnitude++;
                    }
                    break;
                default:
                    require(false);
            }
        }
    }
}

//////////
//  Implementation helpers
CriticalSection * BinaryFixed::_getStaticGuard()
{
    static CriticalSection instance;
    return &instance;
}

void BinaryFixed::_normalize()
{
    //  Count the number of tailing 0's
    UIntSize maxBitsToPrune = Math::min(_scale, _magnitude.getRequiredBits() - 1);
    UIntSize bitsToPrune;
    for (bitsToPrune = 0; bitsToPrune < maxBitsToPrune && _magnitude.getBit(bitsToPrune) == 0; bitsToPrune++)
    {   //  All work is done in the loop header
    }
    if (bitsToPrune != 0)
    {
        _magnitude >>= bitsToPrune;
        _scale -= bitsToPrune;
    }
    if (_magnitude == 0)
    {
        _scale = 0;
        _isNegative = false;
    }
}

void BinaryFixed::_fromReal(Real64 value)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    switch (_fpclass(value))
    {
        case _FPCLASS_SNAN:
        case _FPCLASS_QNAN:
        case _FPCLASS_NINF:
        case _FPCLASS_PINF:
            throw ValueNotRepresentableAsFixedPointError();
        default:
            break;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    int isInf = isinf(value);
    if (isInf == -1 || isInf == 1)
    {
        throw ValueNotRepresentableAsFixedPointError();
    }
    else if (isnan(value))
    {
        throw ValueNotRepresentableAsFixedPointError();
    }
#else
    #error Unsupported platform
#endif
    UInt64 valueBits = *((UInt64*)(&value));
    _isNegative = ((valueBits >> 63) & 0x01) != 0;
    Int16 exponent = static_cast<Int16>(((valueBits >> 52) & 0x07FF));
    UInt64 mantissa = (valueBits & CARL_UINT64_CONSTANT(0xFFFFFFFFFFFFF));
    if (exponent != 0)
    {   //  Normalized values have an implicit '1' bit
        mantissa |= CARL_UINT64_CONSTANT(0x10000000000000);
    }
    _magnitude = mantissa;
    exponent -= 1023;
    if (exponent <= 52)
    {   //  Just shift decimal point to the right
        _scale = 52 - exponent;
    }
    else
    {   //  Must pad
        _scale = 0;
        _magnitude <<= (exponent - 52);
    }
    _normalize();
}

void BinaryFixed::_fromString(const String & s, UIntSize radix) throw (ParseException)
{
    RealParseProperties parseProperties;
    parseProperties.setRadix(radix);
    BinaryFixed bf = NumberParse::getInstance(Locale())->parseBinaryFixed(s, parseProperties);
    _isNegative = bf._isNegative;
    _magnitude = bf._magnitude;
    _scale = bf._scale;
}

void BinaryFixed::_multiplyBinaryIntegerByRadix(BinaryInteger & bi, UInt8 radix, UIntSize repeatCount)
{
    require(radix >= 2 && radix <= 36);

    switch (radix)
    {
        case 2:
            bi <<= repeatCount;
            return;
        case 4:
            if (((repeatCount << 1) >> 1) == repeatCount)
            {
                bi <<= 2 * repeatCount;
                return;
            }
            break;
        case 8:
            if (((repeatCount << 2) >> 2) == repeatCount)
            {
                bi <<= 3 * repeatCount;
                return;
            }
            break;
        case 16:
            if (((repeatCount << 3) >> 3) == repeatCount)
            {
                bi <<= 4 * repeatCount;
                return;
            }
            break;
        case 32:
            if (((repeatCount << 4) >> 4) == repeatCount)
            {
                bi <<= 5 * repeatCount;
                return;
            }
            break;
        default:
            break;
    }
    for (UIntSize i = 0; i < repeatCount; i++)
    {
        bi *= radix;
    }
}

//////////
//  Additional arbitrary-precision operators
BinaryFixed CARL_API carl::operator + (Int8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) + op2;
}

BinaryFixed CARL_API carl::operator + (Int16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) + op2;
}

BinaryFixed CARL_API carl::operator + (Int32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) + op2;
}

BinaryFixed CARL_API carl::operator + (Int64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) + op2;
}

BinaryFixed CARL_API carl::operator + (UInt8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) + op2;
}

BinaryFixed CARL_API carl::operator + (UInt16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) + op2;
}

BinaryFixed CARL_API carl::operator + (UInt32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) + op2;
}

BinaryFixed CARL_API carl::operator + (UInt64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) + op2;
}

BinaryFixed CARL_API carl::operator + (Real32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) + op2;
}

BinaryFixed CARL_API carl::operator + (Real64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) + op2;
}

BinaryFixed CARL_API carl::operator - (Int8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) - op2;
}

BinaryFixed CARL_API carl::operator - (Int16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) - op2;
}

BinaryFixed CARL_API carl::operator - (Int32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) - op2;
}

BinaryFixed CARL_API carl::operator - (Int64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) - op2;
}

BinaryFixed CARL_API carl::operator - (UInt8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) - op2;
}

BinaryFixed CARL_API carl::operator - (UInt16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) - op2;
}

BinaryFixed CARL_API carl::operator - (UInt32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) - op2;
}

BinaryFixed CARL_API carl::operator - (UInt64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) - op2;
}

BinaryFixed CARL_API carl::operator - (Real32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) - op2;
}

BinaryFixed CARL_API carl::operator - (Real64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) - op2;
}

BinaryFixed CARL_API carl::operator * (Int8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) * op2;
}

BinaryFixed CARL_API carl::operator * (Int16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) * op2;
}

BinaryFixed CARL_API carl::operator * (Int32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) * op2;
}

BinaryFixed CARL_API carl::operator * (Int64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) * op2;
}

BinaryFixed CARL_API carl::operator * (UInt8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) * op2;
}

BinaryFixed CARL_API carl::operator * (UInt16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) * op2;
}

BinaryFixed CARL_API carl::operator * (UInt32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) * op2;
}

BinaryFixed CARL_API carl::operator * (UInt64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) * op2;
}

BinaryFixed CARL_API carl::operator * (Real32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) * op2;
}

BinaryFixed CARL_API carl::operator * (Real64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) * op2;
}

BinaryFixed CARL_API carl::operator / (Int8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) / op2;
}

BinaryFixed CARL_API carl::operator / (Int16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) / op2;
}

BinaryFixed CARL_API carl::operator / (Int32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) / op2;
}

BinaryFixed CARL_API carl::operator / (Int64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) / op2;
}

BinaryFixed CARL_API carl::operator / (UInt8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) / op2;
}

BinaryFixed CARL_API carl::operator / (UInt16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) / op2;
}

BinaryFixed CARL_API carl::operator / (UInt32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) / op2;
}

BinaryFixed CARL_API carl::operator / (UInt64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) / op2;
}

BinaryFixed CARL_API carl::operator / (Real32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) / op2;
}

BinaryFixed CARL_API carl::operator / (Real64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) / op2;
}

bool CARL_API carl::operator == (Int8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) == op2;
}

bool CARL_API carl::operator == (Int16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) == op2;
}

bool CARL_API carl::operator == (Int32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) == op2;
}

bool CARL_API carl::operator == (Int64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) == op2;
}

bool CARL_API carl::operator == (UInt8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) == op2;
}

bool CARL_API carl::operator == (UInt16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) == op2;
}

bool CARL_API carl::operator == (UInt32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) == op2;
}

bool CARL_API carl::operator == (UInt64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) == op2;
}

bool CARL_API carl::operator == (Real32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) == op2;
}

bool CARL_API carl::operator == (Real64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) == op2;
}

bool CARL_API carl::operator != (Int8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) != op2;
}

bool CARL_API carl::operator != (Int16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) != op2;
}

bool CARL_API carl::operator != (Int32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) != op2;
}

bool CARL_API carl::operator != (Int64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) != op2;
}

bool CARL_API carl::operator != (UInt8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) != op2;
}

bool CARL_API carl::operator != (UInt16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) != op2;
}

bool CARL_API carl::operator != (UInt32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) != op2;
}

bool CARL_API carl::operator != (UInt64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) != op2;
}

bool CARL_API carl::operator != (Real32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) != op2;
}

bool CARL_API carl::operator != (Real64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) != op2;
}

bool CARL_API carl::operator < (Int8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) < op2;
}

bool CARL_API carl::operator < (Int16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) < op2;
}

bool CARL_API carl::operator < (Int32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) < op2;
}

bool CARL_API carl::operator < (Int64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) < op2;
}

bool CARL_API carl::operator < (UInt8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) < op2;
}

bool CARL_API carl::operator < (UInt16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) < op2;
}

bool CARL_API carl::operator < (UInt32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) < op2;
}

bool CARL_API carl::operator < (UInt64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) < op2;
}

bool CARL_API carl::operator < (Real32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) < op2;
}

bool CARL_API carl::operator < (Real64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) < op2;
}

bool CARL_API carl::operator <= (Int8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (Int16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (Int32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (Int64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (Real32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (Real64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) <= op2;
}

bool CARL_API carl::operator > (Int8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) > op2;
}

bool CARL_API carl::operator > (Int16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) > op2;
}

bool CARL_API carl::operator > (Int32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) > op2;
}

bool CARL_API carl::operator > (Int64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) > op2;
}

bool CARL_API carl::operator > (UInt8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) > op2;
}

bool CARL_API carl::operator > (UInt16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) > op2;
}

bool CARL_API carl::operator > (UInt32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) > op2;
}

bool CARL_API carl::operator > (UInt64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) > op2;
}

bool CARL_API carl::operator > (Real32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) > op2;
}

bool CARL_API carl::operator > (Real64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) > op2;
}

bool CARL_API carl::operator >= (Int8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (Int16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (Int32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (Int64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt8 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt16 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (Real32 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (Real64 op1, const BinaryFixed & op2)
{
    return BinaryFixed(op1) >= op2;
}

//////////
//  Hashing arbitrary-precision values
UIntSize CARL_CALLBACK StandardHashFunctions::hashBinaryFixed(const BinaryFixed & key)
{
    return hashBinaryInteger(key.getMagnitude());
}

//  End of BinaryFixed.cpp
