//
//  BinaryFloat.cpp - Implementation of the carl::BinaryFloat class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:49 $
//  $Revision: 1.20 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <float.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <math.h>
#else
    #error Unsupported platform
#endif

UIntSize BinaryFloat::_defaultDivisionPrecision = 0;

BinaryFloat::BinaryFloat(bool isNegative, const BinaryInteger & mantissa, const BinaryInteger & scale)
    :   _category(ValueCategory_Normal),
        _isNegative(isNegative),
        _mantissa(mantissa),
        _scale(scale)
{
    _normalize();
}

BinaryFloat::BinaryFloat()
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(0),
        _scale(0)
{
}

BinaryFloat::BinaryFloat(ValueCategory category)
    :   _category(category),
        _isNegative(false),
        _mantissa(0),
        _scale(0)
{
    require(category != ValueCategory_Normal);
}

BinaryFloat::BinaryFloat(Int8 value)
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(value),
        _scale(0)
{
    if (value < 0)
    {
        _mantissa = -_mantissa;
        _isNegative = true;
    }
    _normalize();
}

BinaryFloat::BinaryFloat(Int16 value)
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(value),
        _scale(0)
{
    if (value < 0)
    {
        _mantissa = -_mantissa;
        _isNegative = true;
    }
    _normalize();
}

BinaryFloat::BinaryFloat(Int32 value)
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(value),
        _scale(0)
{
    if (value < 0)
    {
        _mantissa = -_mantissa;
        _isNegative = true;
    }
    _normalize();
}

BinaryFloat::BinaryFloat(Int64 value)
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(value),
        _scale(0)
{
    if (value < 0)
    {
        _mantissa = -_mantissa;
        _isNegative = true;
    }
    _normalize();
}

BinaryFloat::BinaryFloat(UInt8 value)
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

BinaryFloat::BinaryFloat(UInt16 value)
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

BinaryFloat::BinaryFloat(UInt32 value)
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

BinaryFloat::BinaryFloat(UInt64 value)
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

BinaryFloat::BinaryFloat(Real32 value)
{
    _fromReal(value);
}

BinaryFloat::BinaryFloat(Real64 value)
{
    _fromReal(value);
}

BinaryFloat::BinaryFloat(const BinaryInteger & value)
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(value),
        _scale(0)
{
    if (value < 0)
    {
        _mantissa = -_mantissa;
        _isNegative = true;
    }
    _normalize();
}

BinaryFloat::BinaryFloat(const BinaryInteger & mantissa, const BinaryInteger & scale)
    :   _category(ValueCategory_Normal),
        _isNegative(false),
        _mantissa(mantissa),
        _scale(scale)
{
    if (mantissa < 0)
    {
        _mantissa = -_mantissa;
        _isNegative = true;
    }
    _normalize();
}

BinaryFloat::BinaryFloat(const char * pValueString) throw (ParseException)
{
    require(pValueString != NULL);

    _fromString(String(pValueString), 10);
}

BinaryFloat::BinaryFloat(const Char * pValueString) throw (ParseException)
{
    require(pValueString != NULL);

    _fromString(String(pValueString), 10);
}

BinaryFloat::BinaryFloat(const String & valueString) throw (ParseException)
{
    _fromString(valueString, 10);
}

BinaryFloat::BinaryFloat(const char * pValueString, UIntSize radix) throw (ParseException)
{
    require(pValueString != NULL);
    require(radix >= 2 && radix <= 36);

    _fromString(String(pValueString), radix);
}

BinaryFloat::BinaryFloat(const Char * pValueString, UIntSize radix) throw (ParseException)
{
    require(pValueString != NULL);
    require(radix >= 2 && radix <= 36);

    _fromString(String(pValueString), radix);
}

BinaryFloat::BinaryFloat(const String & valueString, UIntSize radix) throw (ParseException)
{
    require(radix >= 2 && radix <= 36);

    _fromString(valueString, radix);
}

BinaryFloat::BinaryFloat(const BinaryFloat & src)
    :   _category(src._category),
        _isNegative(src._isNegative),
        _mantissa(src._mantissa),
        _scale(src._scale)
{
}

BinaryFloat::~BinaryFloat()
{
}

BinaryFloat & BinaryFloat::operator = (const BinaryFloat & src)
{
    if (this != (&src))
    {
        _category = src._category;
        _isNegative = src._isNegative;
        _mantissa = src._mantissa;
        _scale = src._scale;
    }
    return *this;
}

//////////
//  Operations
UIntSize BinaryFloat::getDefaultDivisionPrecision()
{
    CriticalSectionLock lock(*_getStaticGuard());
    return _defaultDivisionPrecision;
}

void BinaryFloat::setDefaultDivisionPrecision(UIntSize precision)
{
    CriticalSectionLock lock(*_getStaticGuard());
    _defaultDivisionPrecision = precision;
}

BinaryFloat BinaryFloat::operator + () const
{
    if (_category == ValueCategory_SignalingNan)
    {
        throw SignalingNanOperandError();
    }
    return *this;
}

BinaryFloat BinaryFloat::operator - () const
{
    switch (_category)
    {
        case ValueCategory_Normal:
            return BinaryFloat(!_isNegative, _mantissa, _scale);
        case ValueCategory_PositiveInfinity:
            return BinaryFloat(ValueCategory_NegativeInfinity);
        case ValueCategory_NegativeInfinity:
            return BinaryFloat(ValueCategory_PositiveInfinity);
        case ValueCategory_SignalingNan:
            throw SignalingNanOperandError();
        case ValueCategory_QuietNan:
            return BinaryFloat(ValueCategory_QuietNan);
        default:
            require(false);
            return BinaryFloat(ValueCategory_SignalingNan);
    }
}

BinaryFloat BinaryFloat::operator + (const BinaryFloat & op2) const
{
    if (_category == ValueCategory_SignalingNan || op2._category == ValueCategory_SignalingNan)
    {
        throw SignalingNanOperandError();
    }
    else if (_category == ValueCategory_QuietNan || op2._category == ValueCategory_QuietNan)
    {
        return BinaryFloat(ValueCategory_QuietNan);
    }
    else if (_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_PositiveInfinity)
    {
        return BinaryFloat(ValueCategory_PositiveInfinity);
    }
    else if (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_NegativeInfinity)
    {
        return BinaryFloat(ValueCategory_NegativeInfinity);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_NegativeInfinity) ||
             (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_PositiveInfinity))
    {
        return BinaryFloat(ValueCategory_QuietNan);
    }
    else if (_category == ValueCategory_PositiveInfinity || op2._category == ValueCategory_PositiveInfinity)
    {
        return BinaryFloat(ValueCategory_PositiveInfinity);
    }
    else if (_category == ValueCategory_NegativeInfinity || op2._category == ValueCategory_NegativeInfinity)
    {
        return BinaryFloat(ValueCategory_NegativeInfinity);
    }
    else if (_mantissa == 0)
    {
        return op2;
    }
    else if (op2._mantissa == 0)
    {
        return (*this);
    }
    else
    {   //  Both operands are finite non-zeroes
        //  Bring both operands to the same scale
        BinaryInteger mantissa1 = _mantissa;
        BinaryInteger scale1 = _scale;
        BinaryInteger mantissa2 = op2._mantissa;
        BinaryInteger scale2 = op2._scale;
        if (scale1 < scale2)
        {
            require(scale2 - scale1 <= Limits::UIntSize_Max);
            mantissa2 <<= (scale2 - scale1).toUIntSize();
            scale2 = scale1;
        }
        if (scale1 > scale2)
        {
            require(scale1 - scale2 <= Limits::UIntSize_Max);
            mantissa1 <<= (scale1 - scale2).toUIntSize();
            scale1 = scale2;
        }
        //  Perform the addition
        if (_isNegative == op2._isNegative)
        {   //  Both operands have the same sign
            return BinaryFloat(_isNegative, mantissa1 + mantissa2, scale1);
        }
        else if (mantissa1 >= mantissa2)
        {   //  Result has the sign of 1st operand
            return BinaryFloat(_isNegative, mantissa1 - mantissa2, scale1);
        }
        else
        {   //  Result has the sign of 2nd operand
            return BinaryFloat(op2._isNegative, mantissa2 - mantissa1, scale1);
        }
    }
}

BinaryFloat BinaryFloat::operator - (const BinaryFloat & op2) const
{
    if (_category == ValueCategory_SignalingNan || op2._category == ValueCategory_SignalingNan)
    {
        throw SignalingNanOperandError();
    }
    else if (_category == ValueCategory_QuietNan || op2._category == ValueCategory_QuietNan)
    {
        return BinaryFloat(ValueCategory_QuietNan);
    }
    else if (_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_NegativeInfinity)
    {
        return BinaryFloat(ValueCategory_PositiveInfinity);
    }
    else if (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_PositiveInfinity)
    {
        return BinaryFloat(ValueCategory_NegativeInfinity);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_PositiveInfinity) ||
             (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_NegativeInfinity))
    {
        return BinaryFloat(ValueCategory_QuietNan);
    }
    else if (_category == ValueCategory_PositiveInfinity)
    {
        return BinaryFloat(ValueCategory_PositiveInfinity);
    }
    else if (op2._category == ValueCategory_PositiveInfinity)
    {
        return BinaryFloat(ValueCategory_NegativeInfinity);
    }
    else if (_category == ValueCategory_NegativeInfinity)
    {
        return BinaryFloat(ValueCategory_NegativeInfinity);
    }
    else if (op2._category == ValueCategory_NegativeInfinity)
    {
        return BinaryFloat(ValueCategory_PositiveInfinity);
    }
    else if (_mantissa == 0)
    {
        return -op2;
    }
    else if (op2._mantissa == 0)
    {
        return (*this);
    }
    else
    {   //  Both operands are finite non-zeroes
        //  Bring both operands to the same scale
        BinaryInteger mantissa1 = _mantissa;
        BinaryInteger scale1 = _scale;
        BinaryInteger mantissa2 = op2._mantissa;
        BinaryInteger scale2 = op2._scale;
        if (scale1 < scale2)
        {
            require(scale2 - scale1 <= Limits::UIntSize_Max);
            mantissa2 <<= (scale2 - scale1).toUIntSize();
            scale2 = scale1;
        }
        if (scale1 > scale2)
        {
            require(scale1 - scale2 <= Limits::UIntSize_Max);
            mantissa1 <<= (scale1 - scale2).toUIntSize();
            scale1 = scale2;
        }
        //  Perform the subtraction
        if (_isNegative != op2._isNegative)
        {   //  Operands have different signs
            return BinaryFloat(_isNegative, mantissa1 + mantissa2, scale1);
        }
        else if (mantissa1 >= mantissa2)
        {   //  Result has the sign of 1st operand
            return BinaryFloat(_isNegative, mantissa1 - mantissa2, scale1);
        }
        else
        {   //  Result has the inverse sign of 1st operand
            return BinaryFloat(!_isNegative, mantissa2 - mantissa1, scale1);
        }
    }
}

BinaryFloat BinaryFloat::operator * (const BinaryFloat & op2) const
{
    if (_category == ValueCategory_SignalingNan || op2._category == ValueCategory_SignalingNan)
    {
        throw SignalingNanOperandError();
    }
    else if (_category == ValueCategory_QuietNan || op2._category == ValueCategory_QuietNan)
    {
        return BinaryFloat(ValueCategory_QuietNan);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_PositiveInfinity) ||
             (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_NegativeInfinity))
    {
        return BinaryFloat(ValueCategory_PositiveInfinity);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_NegativeInfinity) ||
             (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_PositiveInfinity))
    {
        return BinaryFloat(ValueCategory_NegativeInfinity);
    }
    else if (_category == ValueCategory_PositiveInfinity || op2._category == ValueCategory_PositiveInfinity)
    {
        return BinaryFloat(ValueCategory_PositiveInfinity);
    }
    else if (_category == ValueCategory_NegativeInfinity || op2._category == ValueCategory_NegativeInfinity)
    {
        return BinaryFloat(ValueCategory_NegativeInfinity);
    }
    else if (_mantissa == 0 || op2._mantissa == 0)
    {
        return 0;
    }
    else if (_isNegative && op2._isNegative)
    {
        return (-(*this)) * (-op2);
    }
    else if (_isNegative && !op2._isNegative)
    {
        return -((-(*this)) * op2);
    }
    else if (!_isNegative && op2._isNegative)
    {
        return -((*this) * (-op2));
    }
    else
    {   //  Both operands are non-zero positive values
        return BinaryFloat(_isNegative != op2._isNegative, _mantissa * op2._mantissa, _scale + op2._scale);
    }
}

BinaryFloat BinaryFloat::operator / (const BinaryFloat & op2) const
{
    if (_category == ValueCategory_SignalingNan || op2._category == ValueCategory_SignalingNan)
    {
        throw SignalingNanOperandError();
    }
    else if (_category == ValueCategory_QuietNan || op2._category == ValueCategory_QuietNan)
    {
        return BinaryFloat(ValueCategory_QuietNan);
    }
    else if ((_category == ValueCategory_PositiveInfinity || _category == ValueCategory_NegativeInfinity) &&
             (op2._category == ValueCategory_PositiveInfinity || op2._category == ValueCategory_NegativeInfinity))
    {
        return BinaryFloat(ValueCategory_QuietNan);
    }
    else if ((_category == ValueCategory_PositiveInfinity && !op2._isNegative) ||
             (_category == ValueCategory_NegativeInfinity && op2._isNegative))
    {
        return BinaryFloat(ValueCategory_PositiveInfinity);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2._isNegative) ||
             (_category == ValueCategory_NegativeInfinity && !op2._isNegative))
    {
        return BinaryFloat(ValueCategory_NegativeInfinity);
    }
    else if (op2._category == ValueCategory_PositiveInfinity || op2._category == ValueCategory_NegativeInfinity)
    {
        return 0;
    }
    else if (_mantissa == 0)
    {
        return 0;
    }
    else if (op2._mantissa == 0)
    {
        if (!_isNegative && _mantissa != 0)
        {
            return BinaryFloat(ValueCategory_PositiveInfinity);
        }
        else if (_isNegative)
        {
            return BinaryFloat(ValueCategory_NegativeInfinity);
        }
        else
        {
            return BinaryFloat(ValueCategory_QuietNan);
        }
    }
    else if (_isNegative && op2._isNegative)
    {
        return (-(*this)) / (-op2);
    }
    else if (_isNegative && !op2._isNegative)
    {
        return -((-(*this)) / op2);
    }
    else if (!_isNegative && op2._isNegative)
    {
        return -((*this) / (-op2));
    }
    else
    {   //  Both operands are non-zero positive values
        CriticalSectionLock lock(*_getStaticGuard());
        BinaryInteger mantissa1 = _mantissa;
        BinaryInteger scale1 = _scale;
        BinaryInteger mantissa2 = op2._mantissa;
        BinaryInteger scale2 = op2._scale;
        UIntSize precision1 = mantissa1.getRequiredBits() - 1;
        UIntSize precision2 = mantissa2.getRequiredBits() - 1;
        UIntSize resultPrecision = Math::max(Math::max(precision1, precision2), _defaultDivisionPrecision);
        UIntSize requiredPresision1 = precision2 + resultPrecision - 1;
        UIntSize op1ExtraScale = requiredPresision1 - precision1;
        BinaryInteger resultMantissa = (mantissa1 << op1ExtraScale) / mantissa2;
        BinaryInteger resultScale = scale1 - scale2 - op1ExtraScale;
        return BinaryFloat(resultMantissa, resultScale);
    }
}

BinaryFloat & BinaryFloat::operator += (const BinaryFloat & op2)
{
    (*this) = (*this) + op2;
    return *this;
}

BinaryFloat & BinaryFloat::operator -= (const BinaryFloat & op2)
{
    (*this) = (*this) - op2;
    return *this;
}

BinaryFloat & BinaryFloat::operator *= (const BinaryFloat & op2)
{
    (*this) = (*this) * op2;
    return *this;
}

BinaryFloat & BinaryFloat::operator /= (const BinaryFloat & op2)
{
    (*this) = (*this) / op2;
    return *this;
}

BinaryFloat BinaryFloat::operator ++()
{
    (*this) += 1;
    return *this;
}

BinaryFloat BinaryFloat::operator --()
{
    (*this) -= 1;
    return *this;
}

BinaryFloat BinaryFloat::operator ++(int)
{
    BinaryFloat result = (*this);
    (*this) += 1;
    return result;
}

BinaryFloat BinaryFloat::operator --(int)
{
    BinaryFloat result = (*this);
    (*this) -= 1;
    return result;
}

int BinaryFloat::compare(const BinaryFloat & op2) const
{
    BinaryFloat diff = (*this) - op2;
    if (diff._isNegative)
    {
        return -1;
    }
    else if (diff._mantissa != 0)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

bool BinaryFloat::operator == (const BinaryFloat & op2) const
{
    return compare(op2) == 0;
}

bool BinaryFloat::operator != (const BinaryFloat & op2) const
{
    return compare(op2) != 0;
}

bool BinaryFloat::operator < (const BinaryFloat & op2) const
{
    return compare(op2) < 0;
}

bool BinaryFloat::operator <= (const BinaryFloat & op2) const
{
    return compare(op2) <= 0;
}

bool BinaryFloat::operator > (const BinaryFloat & op2) const
{
    return compare(op2) > 0;
}

bool BinaryFloat::operator >= (const BinaryFloat & op2) const
{
    return compare(op2) >= 0;
}

Int8 BinaryFloat::toInt8(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int8_Max)
    {
        return Limits::Int8_Max;
    }
    else if (value64 < Limits::Int8_Min)
    {
        return Limits::Int8_Min;
    }
    else
    {
        return static_cast<Int8>(value64);
    }
}

Int16 BinaryFloat::toInt16(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int16_Max)
    {
        return Limits::Int16_Max;
    }
    else if (value64 < Limits::Int16_Min)
    {
        return Limits::Int16_Min;
    }
    else
    {
        return static_cast<Int16>(value64);
    }
}

Int32 BinaryFloat::toInt32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int32_Max)
    {
        return Limits::Int32_Max;
    }
    else if (value64 < Limits::Int32_Min)
    {
        return Limits::Int32_Min;
    }
    else
    {
        return static_cast<Int32>(value64);
    }
}

Int64 BinaryFloat::toInt64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    BinaryInteger value = toBinaryInteger(roundingMode);
    if (value < Limits::Int64_Min)
    {
        return Limits::Int64_Min;
    }
    else if (value > Limits::Int64_Max)
    {
        return Limits::Int64_Max;
    }
    else
    {
        return value.toInt64();
    }
}

UInt8 BinaryFloat::toUInt8(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt8_Max)
    {
        return Limits::UInt8_Max;
    }
    else
    {
        return static_cast<UInt8>(value64);
    }
}

UInt16 BinaryFloat::toUInt16(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt16_Max)
    {
        return Limits::UInt16_Max;
    }
    else
    {
        return static_cast<UInt16>(value64);
    }
}

UInt32 BinaryFloat::toUInt32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt32_Max)
    {
        return Limits::UInt32_Max;
    }
    else
    {
        return static_cast<UInt32>(value64);
    }
}

UInt64 BinaryFloat::toUInt64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    BinaryInteger value = toBinaryInteger(roundingMode);
    if (value < Limits::UInt64_Min)
    {
        return Limits::UInt64_Min;
    }
    else if (value > Limits::UInt64_Max)
    {
        return Limits::UInt64_Max;
    }
    else
    {
        return value.toUInt64();
    }
}

UIntSize BinaryFloat::toUIntSize(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UIntSize_Max)
    {
        return Limits::UIntSize_Max;
    }
    else
    {
        return static_cast<UIntSize>(value64);
    }
}

UIntPtr BinaryFloat::toUIntPtr(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UIntPtr_Max)
    {
        return Limits::UIntPtr_Max;
    }
    else
    {
        return static_cast<UIntPtr>(value64);
    }
}

IntPtrDiff BinaryFloat::toIntPtrDiff(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::IntPtrDiff_Max)
    {
        return Limits::IntPtrDiff_Max;
    }
    else if (value64 < Limits::IntPtrDiff_Min)
    {
        return Limits::IntPtrDiff_Min;
    }
    else
    {
        return static_cast<IntPtrDiff>(value64);
    }
}

Real32 BinaryFloat::toReal32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    return static_cast<Real32>(toReal64(roundingMode));
}

Real64 BinaryFloat::toReal64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    switch (_category)
    {
        case ValueCategory_Normal:
            break;
        case ValueCategory_PositiveInfinity:
            {
                Real64 value;
                *(UInt64*)(&value) = CARL_UINT64_CONSTANT(0x7FF0000000000000);
                return value;
            }
        case ValueCategory_NegativeInfinity:
            {
                Real64 value;
                *(UInt64*)(&value) = CARL_UINT64_CONSTANT(0xFFF0000000000000);
                return value;
            }
        case ValueCategory_SignalingNan:
            throw SignalingNanOperandError();
        case ValueCategory_QuietNan:
            {
                Real64 value;
                *(UInt64*)(&value) = CARL_UINT64_CONSTANT(0x7FF7FFFFFFFFFFFF);
                return value;
            }
        default:
            require(false);
            return 0;
    }
    if (_isNegative)
    {   //  Make sure the translated value is positive
        BinaryFloat minusThis = -(*this);
        switch (roundingMode)
        {
            case RoundingMode_TowardsNearest:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsZero:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsPositiveInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsNegativeInfinity));
            case RoundingMode_TowardsNegativeInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsPositiveInfinity));
            default:
                require(false);
                return 0;
        }
    }
    else
    {
        Real64 addend = 1.0;
        for (BinaryInteger scale = 0; scale < _scale; scale++)
        {
            addend *= 2.0;
        }
        for (BinaryInteger scale = 0; scale > _scale; scale--)
        {
            addend /= 2.0;
        }
        Real64 result = 0;
        for (UIntSize bitNo = 0; bitNo < _mantissa.getRequiredBits(); bitNo++)
        {
            if (_mantissa.getBit(bitNo) != 0)
            {
                result += addend;
            }
            addend *= 2.0;
        }
        return result;
    }
}

BinaryInteger BinaryFloat::toBinaryInteger(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    switch (_category)
    {
        case ValueCategory_Normal:
            break;
        case ValueCategory_PositiveInfinity:
        case ValueCategory_NegativeInfinity:
            throw ValueNotRepresentableAsFixedPointError();
        case ValueCategory_SignalingNan:
            throw SignalingNanOperandError();
        case ValueCategory_QuietNan:
            throw ValueNotRepresentableAsFixedPointError();
        default:
            require(false);
            return 0;
    }
    if (_isNegative)
    {   //  Make sure the translated value is positive
        BinaryFloat minusThis = -(*this);
        switch (roundingMode)
        {
            case RoundingMode_TowardsNearest:
                return -(minusThis.toBinaryInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsZero:
                return -(minusThis.toBinaryInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsInfinity:
                return -(minusThis.toBinaryInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsPositiveInfinity:
                return -(minusThis.toBinaryInteger(RoundingMode_TowardsNegativeInfinity));
            case RoundingMode_TowardsNegativeInfinity:
                return -(minusThis.toBinaryInteger(RoundingMode_TowardsPositiveInfinity));
            default:
                require(false);
                return 0;
        }
    }
    else if (_scale == 0)
    {   //  No rounding is necessary
        return _mantissa;
    }
    else if (_scale > 0)
    {   //  No rounding is necessary, but padding IS
        BinaryInteger result = _mantissa;
        require(_scale <= Limits::UIntSize_Max);
        result <<= _scale.toUIntSize();
        return result;
    }
    else
    {   //  Rounding may be necessary
        BinaryInteger result = _mantissa;
        UIntSize nonzeroBitsShiftedOut = 0;
        int lastBitShiftedOut = 0;
        for (BinaryInteger scale = 0; scale > _scale; scale--)
        {
            if ((lastBitShiftedOut = result.getBit(0)) != 0)
            {
                nonzeroBitsShiftedOut++;
            }
            result >>= 1;
        }
        if (nonzeroBitsShiftedOut != 0)
        {   //  Rounding is necessary
            switch (roundingMode)
            {
                case RoundingMode_TowardsNearest:
                    if (lastBitShiftedOut != 0)
                    {
                        if (nonzeroBitsShiftedOut == 1)
                        {   //  Only one '1' bit was shifted out, and that was the last bit.
                            //  The value is exactly between two integers; chose an odd one
                            if (result.getBit(0) != 0)
                            {
                                result++;
                            }
                        }
                        else
                        {   //  More than one '1' bit was shifted out
                            result++;
                        }
                    }
                    break;
                case RoundingMode_TowardsZero:
                    break;
                case RoundingMode_TowardsInfinity:
                    result++;
                    break;
                case RoundingMode_TowardsPositiveInfinity:
                    result++;
                    break;
                case RoundingMode_TowardsNegativeInfinity:
                    break;
                default:
                    require(false);
            }
        }
        return result;
    }
}

String BinaryFloat::toString(UIntSize radix /*= 10*/) const
{
    RealFormatProperties formatProperties;
    formatProperties.setRadix(radix);
    return NumberFormat::getInstance(Locale())->format(*this, formatProperties);
}

int BinaryFloat::getSign() const
{
    require(_category == ValueCategory_Normal);

    if (_isNegative)
    {
        return -1;
    }
    else if (_mantissa != 0)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

BinaryInteger BinaryFloat::getMantissa() const
{
    require(_category == ValueCategory_Normal);

    return _mantissa;
}

BinaryInteger BinaryFloat::getScale() const
{
    require(_category == ValueCategory_Normal);

    return _scale;
}

void BinaryFloat::setPrecision(UIntSize precision, RoundingMode roundingMode /*= RoundingMode_TowardsZero*/)
{
    require(_category == ValueCategory_Normal);

    UIntSize currentPrecision = _mantissa.getRequiredBits() - 1;
    if (precision >= currentPrecision)
    {   //  Scaling increases precision; no need for rounding
        UIntSize bitsToInsert = (precision - currentPrecision);
        _mantissa <<= bitsToInsert;
        _scale -= bitsToInsert;
        _normalize();
    }
    else
    {
        UIntSize bitsToRemove = (currentPrecision - precision);
        UIntSize nonzeroBitsShiftedOut = 0;
        int lastBitShiftedOut = 0;
        for (UIntSize i = 0; i < bitsToRemove; i++)
        {
            if ((lastBitShiftedOut = _mantissa.getBit(i)) != 0)
            {
                nonzeroBitsShiftedOut++;
            }
        }
        _mantissa >>= bitsToRemove;
        _scale += bitsToRemove;
        _normalize();
        if (nonzeroBitsShiftedOut != 0)
        {   //  Rounding is necessary
            switch (roundingMode)
            {
                case RoundingMode_TowardsNearest:
                    if (lastBitShiftedOut != 0)
                    {
                        if (nonzeroBitsShiftedOut == 1)
                        {   //  Only one '1' bit was shifted out, and that was the last bit.
                            //  The value is exactly between two integers; chose an odd one
                            if (_mantissa.getBit(0) != 0)
                            {
                                _mantissa++;
                            }
                        }
                        else
                        {   //  More than one '1' bit was shifted out
                            _mantissa++;
                        }
                    }
                    break;
                case RoundingMode_TowardsZero:
                    break;
                case RoundingMode_TowardsInfinity:
                    _mantissa++;
                    break;
                case RoundingMode_TowardsPositiveInfinity:
                    if (!_isNegative)
                    {
                        _mantissa++;
                    }
                    break;
                case RoundingMode_TowardsNegativeInfinity:
                    if (_isNegative)
                    {
                        _mantissa++;
                    }
                    break;
                default:
                    require(false);
            }
        }
    }
}

BinaryFloat::ValueCategory BinaryFloat::getCategory() const
{
    return _category;
}

bool BinaryFloat::isNormal() const
{
    return _category == ValueCategory_Normal;
}

bool BinaryFloat::isPositiveInfinity() const
{
    return _category == ValueCategory_PositiveInfinity;
}

bool BinaryFloat::isNegativeInfinity() const
{
    return _category == ValueCategory_NegativeInfinity;
}

bool BinaryFloat::isSignalingNan() const
{
    return _category == ValueCategory_SignalingNan;
}

bool BinaryFloat::isQuietNan() const
{
    return _category == ValueCategory_QuietNan;
}

//////////
//  Implementation helpers
CriticalSection * BinaryFloat::_getStaticGuard()
{
    static CriticalSection instance;
    return &instance;
}

void BinaryFloat::_normalize()
{
    if (_category == ValueCategory_Normal)
    {
        //  Count the number of tailing 0's
        UIntSize maxBitsToPrune = _mantissa.getRequiredBits() - 1;
        UIntSize bitsToPrune;
        for (bitsToPrune = 0; bitsToPrune < maxBitsToPrune && _mantissa.getBit(bitsToPrune) == 0; bitsToPrune++)
        {   //  All work is done in the loop header
        }
        if (bitsToPrune != 0)
        {
            _mantissa >>= bitsToPrune;
            _scale += bitsToPrune;
        }
        if (_mantissa == 0)
        {
            _scale = 0;
            _isNegative = false;
        }
    }
    else
    {
        _mantissa = _scale = 0;
    }
}

void BinaryFloat::_fromReal(Real64 value)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    switch (_fpclass(value))
    {
        case _FPCLASS_SNAN:
            _category = ValueCategory_SignalingNan;
            break;
        case _FPCLASS_QNAN:
            _category = ValueCategory_QuietNan;
            break;
        case _FPCLASS_NINF:
            _category = ValueCategory_NegativeInfinity;
            break;
        case _FPCLASS_PINF:
            _category = ValueCategory_PositiveInfinity;
            break;
        default:
            _category = ValueCategory_Normal;
            break;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    int isInf = isinf(value);
    if (isInf == -1)
    {
        _category = ValueCategory_NegativeInfinity;
    }
    else if (isInf == 1)
    {
        _category = ValueCategory_PositiveInfinity;
    }
    else if (isnan(value))
    {
        _category = ValueCategory_QuietNan;
    }
    else
    {
        _category = ValueCategory_Normal;
    }
#else
    #error Unsupported platform
#endif
    if (_category == ValueCategory_Normal)
    {
        UInt64 valueBits = *((UInt64*)(&value));
        _isNegative = ((valueBits >> 63) & 0x01) != 0;
        Int16 exponent = static_cast<Int16>(((valueBits >> 52) & 0x07FF));
        UInt64 mantissa = (valueBits & CARL_UINT64_CONSTANT(0xFFFFFFFFFFFFF));
        if (exponent != 0)
        {   //  Normalized values have an implicit '1' bit
            mantissa |= CARL_UINT64_CONSTANT(0x10000000000000);
        }
        _mantissa = mantissa;
        _scale = exponent - 1023 - 52;
        _normalize();
    }
}

void BinaryFloat::_fromString(const String & s, UIntSize radix) throw (ParseException)
{
    RealParseProperties parseProperties;
    parseProperties.setRadix(radix);
    BinaryFloat bf = NumberParse::getInstance(Locale())->parseBinaryFloat(s, parseProperties);
    _category = bf._category;
    _isNegative = bf._isNegative;
    _mantissa = bf._mantissa;
    _scale = bf._scale;
}

//////////
//  Additional arbitrary-precision operators
BinaryFloat CARL_API carl::operator + (Int8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) + op2;
}

BinaryFloat CARL_API carl::operator + (Int16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) + op2;
}

BinaryFloat CARL_API carl::operator + (Int32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) + op2;
}

BinaryFloat CARL_API carl::operator + (Int64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) + op2;
}

BinaryFloat CARL_API carl::operator + (UInt8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) + op2;
}

BinaryFloat CARL_API carl::operator + (UInt16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) + op2;
}

BinaryFloat CARL_API carl::operator + (UInt32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) + op2;
}

BinaryFloat CARL_API carl::operator + (UInt64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) + op2;
}

BinaryFloat CARL_API carl::operator + (Real32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) + op2;
}

BinaryFloat CARL_API carl::operator + (Real64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) + op2;
}

BinaryFloat CARL_API carl::operator - (Int8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) - op2;
}

BinaryFloat CARL_API carl::operator - (Int16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) - op2;
}

BinaryFloat CARL_API carl::operator - (Int32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) - op2;
}

BinaryFloat CARL_API carl::operator - (Int64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) - op2;
}

BinaryFloat CARL_API carl::operator - (UInt8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) - op2;
}

BinaryFloat CARL_API carl::operator - (UInt16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) - op2;
}

BinaryFloat CARL_API carl::operator - (UInt32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) - op2;
}

BinaryFloat CARL_API carl::operator - (UInt64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) - op2;
}

BinaryFloat CARL_API carl::operator - (Real32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) - op2;
}

BinaryFloat CARL_API carl::operator - (Real64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) - op2;
}

BinaryFloat CARL_API carl::operator * (Int8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) * op2;
}

BinaryFloat CARL_API carl::operator * (Int16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) * op2;
}

BinaryFloat CARL_API carl::operator * (Int32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) * op2;
}

BinaryFloat CARL_API carl::operator * (Int64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) * op2;
}

BinaryFloat CARL_API carl::operator * (UInt8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) * op2;
}

BinaryFloat CARL_API carl::operator * (UInt16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) * op2;
}

BinaryFloat CARL_API carl::operator * (UInt32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) * op2;
}

BinaryFloat CARL_API carl::operator * (UInt64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) * op2;
}

BinaryFloat CARL_API carl::operator * (Real32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) * op2;
}

BinaryFloat CARL_API carl::operator * (Real64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) * op2;
}

BinaryFloat CARL_API carl::operator / (Int8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) / op2;
}

BinaryFloat CARL_API carl::operator / (Int16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) / op2;
}

BinaryFloat CARL_API carl::operator / (Int32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) / op2;
}

BinaryFloat CARL_API carl::operator / (Int64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) / op2;
}

BinaryFloat CARL_API carl::operator / (UInt8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) / op2;
}

BinaryFloat CARL_API carl::operator / (UInt16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) / op2;
}

BinaryFloat CARL_API carl::operator / (UInt32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) / op2;
}

BinaryFloat CARL_API carl::operator / (UInt64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) / op2;
}

BinaryFloat CARL_API carl::operator / (Real32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) / op2;
}

BinaryFloat CARL_API carl::operator / (Real64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) / op2;
}

bool CARL_API carl::operator == (Int8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) == op2;
}

bool CARL_API carl::operator == (Int16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) == op2;
}

bool CARL_API carl::operator == (Int32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) == op2;
}

bool CARL_API carl::operator == (Int64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) == op2;
}

bool CARL_API carl::operator == (UInt8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) == op2;
}

bool CARL_API carl::operator == (UInt16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) == op2;
}

bool CARL_API carl::operator == (UInt32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) == op2;
}

bool CARL_API carl::operator == (UInt64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) == op2;
}

bool CARL_API carl::operator == (Real32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) == op2;
}

bool CARL_API carl::operator == (Real64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) == op2;
}

bool CARL_API carl::operator != (Int8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) != op2;
}

bool CARL_API carl::operator != (Int16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) != op2;
}

bool CARL_API carl::operator != (Int32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) != op2;
}

bool CARL_API carl::operator != (Int64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) != op2;
}

bool CARL_API carl::operator != (UInt8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) != op2;
}

bool CARL_API carl::operator != (UInt16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) != op2;
}

bool CARL_API carl::operator != (UInt32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) != op2;
}

bool CARL_API carl::operator != (UInt64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) != op2;
}

bool CARL_API carl::operator != (Real32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) != op2;
}

bool CARL_API carl::operator != (Real64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) != op2;
}

bool CARL_API carl::operator < (Int8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) < op2;
}

bool CARL_API carl::operator < (Int16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) < op2;
}

bool CARL_API carl::operator < (Int32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) < op2;
}

bool CARL_API carl::operator < (Int64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) < op2;
}

bool CARL_API carl::operator < (UInt8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) < op2;
}

bool CARL_API carl::operator < (UInt16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) < op2;
}

bool CARL_API carl::operator < (UInt32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) < op2;
}

bool CARL_API carl::operator < (UInt64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) < op2;
}

bool CARL_API carl::operator < (Real32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) < op2;
}

bool CARL_API carl::operator < (Real64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) < op2;
}

bool CARL_API carl::operator <= (Int8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (Int16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (Int32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (Int64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (Real32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (Real64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) <= op2;
}

bool CARL_API carl::operator > (Int8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) > op2;
}

bool CARL_API carl::operator > (Int16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) > op2;
}

bool CARL_API carl::operator > (Int32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) > op2;
}

bool CARL_API carl::operator > (Int64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) > op2;
}

bool CARL_API carl::operator > (UInt8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) > op2;
}

bool CARL_API carl::operator > (UInt16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) > op2;
}

bool CARL_API carl::operator > (UInt32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) > op2;
}

bool CARL_API carl::operator > (UInt64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) > op2;
}

bool CARL_API carl::operator > (Real32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) > op2;
}

bool CARL_API carl::operator > (Real64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) > op2;
}

bool CARL_API carl::operator >= (Int8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (Int16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (Int32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (Int64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt8 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt16 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (Real32 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (Real64 op1, const BinaryFloat & op2)
{
    return BinaryFloat(op1) >= op2;
}

//////////
//  Hashing arbitrary-precision values
UIntSize CARL_CALLBACK StandardHashFunctions::hashBinaryFloat(const BinaryFloat & key)
{
    if (key.isNormal())
    {
        return hashBinaryInteger(key.getMantissa());
    }
    else
    {
        return 0;
    }
}

//  End of BinaryFloat.cpp
