//
//  CompressOutputStream.cpp - Implementation of the carl::CompressOutputStream class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:49 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

//////////
//  Construction/destruction
CompressOutputStream::CompressOutputStream(IByteOutput * pTarget, Compressor * pCompressor)
{
    require(pTarget != NULL);
    require(pCompressor != NULL && pCompressor->getCompressionAlgorithm()->getMinimumBlockSize() <= _MaximumBufferSize);

    _pTarget = pTarget;
    _bufferSize = pCompressor->getCompressionAlgorithm()->getPreferredBlockSize();
    _pCompressor = pCompressor;
    if (_bufferSize > _MaximumBufferSize)
    {
        _bufferSize = _MaximumBufferSize;
    }
    require(_bufferSize >= pCompressor->getCompressionAlgorithm()->getMinimumBlockSize() &&
            _bufferSize <= pCompressor->getCompressionAlgorithm()->getMaximumBlockSize());
    _pBuffer = new UInt8[_bufferSize];
    _bytesInBuffer = 0;
}

//////////
//  Operations
void CompressOutputStream::close()
{
    if (_pBuffer != NULL)
    {
        _flushBuffer();
        delete [] _pBuffer;
        _pBuffer = NULL;
    }
}

//////////
//  IByteOutput
void CompressOutputStream::flush() throw(IOException)
{
    require(isOpen());

    _flushBuffer();
    _pTarget->flush();
}

void CompressOutputStream::write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException)
{
    require(isOpen());

    const UInt8 * pBytes = static_cast<const UInt8*>(pBuffer);
    UIntSize bytesFreeInBuffer = _bufferSize - _bytesInBuffer;
    if (bytesToWrite <= bytesFreeInBuffer)
    {   //  Can write everything to buffer; no need to flush
        Memory::copy(_pBuffer + _bytesInBuffer, pBytes, bytesToWrite);
        _bytesInBuffer += bytesToWrite;
    }
    else
    {   //  Must perform write in chunks
        while (bytesToWrite != 0)
        {
            //  Determine the size of the next chunk
            UIntSize chunkSize = (bytesToWrite > bytesFreeInBuffer) ? bytesFreeInBuffer : bytesToWrite;
            if (chunkSize == 0)
            {   //  There is no space in buffer
                _flushBuffer();
                bytesFreeInBuffer = _bufferSize;
            }
            else
            {   //  Write the next chunk to the buffer
                Memory::copy(_pBuffer + _bytesInBuffer, pBytes, chunkSize);
                _bytesInBuffer += chunkSize;
                pBytes += chunkSize;
                bytesToWrite -= chunkSize;
                bytesFreeInBuffer -= chunkSize;
            }
        }
    }
}

//////////
//  Implementation helpers
void CompressOutputStream::_flushBuffer()
{
    if (_bytesInBuffer != 0)
    {
        //  Is padding required ?
        UIntSize paddingBytes = 0;
        if (_bytesInBuffer < _pCompressor->getCompressionAlgorithm()->getMinimumBlockSize())
        {
            paddingBytes = _pCompressor->getCompressionAlgorithm()->getMinimumBlockSize() - _bytesInBuffer;
            _bytesInBuffer += paddingBytes;
            require(_bytesInBuffer <= _bufferSize);
        }
        //  Prepare compressed data
        _compressedData.setSize(_pCompressor->getRequiredOutputBufferSize(_pBuffer, _bytesInBuffer));
        UIntSize compressedDataSize = _pCompressor->compress(_pBuffer, _bytesInBuffer, _compressedData.getData());
        require(compressedDataSize == _compressedData.size());
        //  Write compressed or uncompressed data, whichever is smaller
        if (compressedDataSize + 3 >= _bytesInBuffer)
        {   //  There is no benefit in compression
            _pTarget->write(_BlockType_Uncompressed);
            _pTarget->write(static_cast<UInt8>(_bytesInBuffer >> 8));
            _pTarget->write(static_cast<UInt8>(_bytesInBuffer));
            _pTarget->write(_pBuffer, _bytesInBuffer);
        }
        else if (paddingBytes == 0)
        {   //  Compressed data, no padding required
            _pTarget->write(_BlockType_CompressedWithoutPadding);
            _pTarget->write(static_cast<UInt8>(compressedDataSize >> 8));
            _pTarget->write(static_cast<UInt8>(compressedDataSize));
            _pTarget->write(_compressedData.getData(), compressedDataSize);
        }
        else
        {   //  Compressed data, padding required
            _pTarget->write(_BlockType_CompressedWithPadding);
            _pTarget->write(static_cast<UInt8>(compressedDataSize >> 8));
            _pTarget->write(static_cast<UInt8>(compressedDataSize));
            _pTarget->write(static_cast<UInt8>(paddingBytes >> 8));
            _pTarget->write(static_cast<UInt8>(paddingBytes));
            _pTarget->write(_compressedData.getData(), compressedDataSize);
        }
        _bytesInBuffer = 0;
    }
}

//  End of CompressOutputStream.cpp
