//
//  CompressionImpl.hpp - carl::_impl::_CompressionImpl interface
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:49 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#ifndef __CompressionImpl_hpp_included__
#define __CompressionImpl_hpp_included__

namespace carl
{
    namespace _impl
    {
        //////////
        //  Compression resources
        class CARL_PUBLIC _CompressionResourceMap_en : public ResourceMap
        {
            //////////
            //  Can't assign or copy construct
        private:
            _CompressionResourceMap_en(const _CompressionResourceMap_en &);
            void                        operator = (const _CompressionResourceMap_en &);

            //////////
            //  Construction/destruction
        public:
            _CompressionResourceMap_en();
            virtual ~_CompressionResourceMap_en() {}

            //////////
            //  Operations
        public:
            virtual String              getString(const String & key);

            //////////
            //  Implementation
        private:
            HashMap<String, String>     _mapKeysToValues;
        };

        //////////
        //  Null compression algorithm performs no compression
        class CARL_PUBLIC _NullCompressor : public Compressor
        {
            //////////
            //  Can't assign or copy construct
        private:
            _NullCompressor(const _NullCompressor &);
            void                        operator = (const _NullCompressor &);

            //////////
            //  Singleton
        private:
            _NullCompressor();
            virtual ~_NullCompressor();
        public:
            static _NullCompressor *            getInstance();

            //////////
            //  Operations
        public:
            virtual CompressionAlgorithm *      getCompressionAlgorithm();
            virtual UIntSize                    getRequiredOutputBufferSize(const void * pUncompressedData, UIntSize uncompressedDataSize);
            virtual UIntSize                    compress(const void * pUncompressedData, UIntSize uncompressedDataSize, void * pCompressedData);
        };

        class CARL_PUBLIC _NullDecompressor : public Decompressor
        {
            //////////
            //  Can't assign or copy construct
        private:
            _NullDecompressor(const _NullDecompressor &);
            void                        operator = (const _NullDecompressor &);

            //////////
            //  Singleton
        private:
            _NullDecompressor();
            virtual ~_NullDecompressor();
        public:
            static _NullDecompressor *      getInstance();

            //////////
            //  Operations
        public:
            virtual CompressionAlgorithm *      getCompressionAlgorithm();
            virtual UIntSize                    getRequiredOutputBufferSize(const void * pCompressedData, UIntSize compressedDataSize) throw(CompressedDataCorruptException);
            virtual UIntSize                    decompress(const void * pCompressedData, UIntSize compressedDataSize, void * pUncompressedData) throw(CompressedDataCorruptException);
        };

        class CARL_PUBLIC _NullCompressionAlgorithm : public CompressionAlgorithm
        {
            //////////
            //  Can't assign or copy construct
        private:
            _NullCompressionAlgorithm(const _NullCompressionAlgorithm &);
            void                        operator = (const _NullCompressionAlgorithm &);

            //////////
            //  Singleton
        private:
            _NullCompressionAlgorithm();
            virtual ~_NullCompressionAlgorithm();
        public:
            static _NullCompressionAlgorithm *  getInstance();

            //////////
            //  Operations
        public:
            virtual Array<String>               getAllNames() const;
            virtual String                      getCanonicalName() const;
            virtual String                      getPreferredName() const;
            virtual String                      getDisplayName() const;
            virtual String                      getDisplayName(const Locale & forLocale) const;
            virtual UIntSize                    getMinimumBlockSize() const;
            virtual UIntSize                    getPreferredBlockSize() const;
            virtual UIntSize                    getMaximumBlockSize() const;
            virtual Compressor *                getCompressor() const;
            virtual Decompressor *              getDecompressor() const;
        };

        //////////
        //  Run-length compression algorithm
        class CARL_PUBLIC _RunLengthCompressor : public Compressor
        {
            //////////
            //  Can't assign or copy construct
        private:
            _RunLengthCompressor(const _RunLengthCompressor &);
            void                        operator = (const _RunLengthCompressor &);

            //////////
            //  Singleton
        private:
            _RunLengthCompressor();
            virtual ~_RunLengthCompressor();
        public:
            static _RunLengthCompressor *       getInstance();

            //////////
            //  Operations
        public:
            virtual CompressionAlgorithm *      getCompressionAlgorithm();
            virtual UIntSize                    getRequiredOutputBufferSize(const void * pUncompressedData, UIntSize uncompressedDataSize);
            virtual UIntSize                    compress(const void * pUncompressedData, UIntSize uncompressedDataSize, void * pCompressedData);

            //////////
            //  Implementation helpers
        private:
            bool                                _repeatedSequenceStartsAt(const UInt8 * pUncompressedBytes,
                                                        UIntSize uncompressedDataSize, 
                                                        UIntSize offset);
        };

        class CARL_PUBLIC _RunLengthDecompressor : public Decompressor
        {
            //////////
            //  Can't assign or copy construct
        private:
            _RunLengthDecompressor(const _RunLengthDecompressor &);
            void                        operator = (const _RunLengthDecompressor &);

            //////////
            //  Singleton
        private:
            _RunLengthDecompressor();
            virtual ~_RunLengthDecompressor();
        public:
            static _RunLengthDecompressor * getInstance();

            //////////
            //  Operations
        public:
            virtual CompressionAlgorithm *      getCompressionAlgorithm();
            virtual UIntSize                    getRequiredOutputBufferSize(const void * pCompressedData, UIntSize compressedDataSize) throw(CompressedDataCorruptException);
            virtual UIntSize                    decompress(const void * pCompressedData, UIntSize compressedDataSize, void * pUncompressedData) throw(CompressedDataCorruptException);
        };

        class CARL_PUBLIC _RunLengthCompressionAlgorithm : public CompressionAlgorithm
        {
            //////////
            //  Can't assign or copy construct
        private:
            _RunLengthCompressionAlgorithm(const _RunLengthCompressionAlgorithm &);
            void                        operator = (const _RunLengthCompressionAlgorithm &);

            //////////
            //  Singleton
        private:
            _RunLengthCompressionAlgorithm();
            virtual ~_RunLengthCompressionAlgorithm();
        public:
            static _RunLengthCompressionAlgorithm * getInstance();

            //////////
            //  Operations
        public:
            virtual Array<String>               getAllNames() const;
            virtual String                      getCanonicalName() const;
            virtual String                      getPreferredName() const;
            virtual String                      getDisplayName() const;
            virtual String                      getDisplayName(const Locale & forLocale) const;
            virtual UIntSize                    getMinimumBlockSize() const;
            virtual UIntSize                    getPreferredBlockSize() const;
            virtual UIntSize                    getMaximumBlockSize() const;
            virtual Compressor *                getCompressor() const;
            virtual Decompressor *              getDecompressor() const;
        };

        //////////
        //  LZW compression algorithm performs no compression
        class CARL_PUBLIC _LzwCompressor : public Compressor
        {
            //////////
            //  Can't assign or copy construct
        private:
            _LzwCompressor(const _LzwCompressor &);
            void                        operator = (const _LzwCompressor &);

            //////////
            //  Singleton
        private:
            _LzwCompressor();
            virtual ~_LzwCompressor();
        public:
            static _LzwCompressor *             getInstance();

            //////////
            //  Operations
        public:
            virtual CompressionAlgorithm *      getCompressionAlgorithm();
            virtual UIntSize                    getRequiredOutputBufferSize(const void * pUncompressedData, UIntSize uncompressedDataSize);
            virtual UIntSize                    compress(const void * pUncompressedData, UIntSize uncompressedDataSize, void * pCompressedData);

            //////////
            //  Implementation
        private:
            //  Hash table constants & macros
            static const UIntSize               _HT_BITS_PER_PIXEL  = 8;
            static const UIntSize               _HT_SIZE            = 8192;
            static const UIntSize               _HT_KEY_MASK        = 0x1FFF;
            static const UIntSize               _HT_KEY_NUM_BITS    = 13;
            static const UIntSize               _HT_MAX_KEY         = 8191;
            static const UIntSize               _HT_MAX_CODE        = 4095;

            template <class T>
            T min(T a, T b)
            {
                return (a < b) ? a : b;
            }

            UIntSize                            _HT_GET_KEY(UIntSize l)
            {
                return l >> 12;
            }

            UIntSize                            _HT_GET_CODE(UIntSize l)
            {
                return l & 0x0FFF;
            }

            UIntSize                            _HT_PUT_KEY(UIntSize l)
            {
                return l << 12;
            }

            UIntSize                            _HT_PUT_CODE(UIntSize l)
            {
                return l & 0x0FFF;
            }

            //  LZW macros
            static const UIntSize               _LZ_MAX_CODE        = 4095;
            static const UIntSize               _LZ_BUFFER_SIZE     = 4096;
            static const UIntSize               _FLUSH_OUT          = 4096;
            static const UIntSize               _FIRST_CODE         = 4097;
            static const UIntSize               _NOT_CODE           = 4098;
            static const UIntSize               _MAX_BIT_PER_CODE   = 12;

            //  Data used by compressor
            UInt8                               _buf[2 * _LZ_BUFFER_SIZE];
            Int16                               _curBufferShift;
            Int16                               _bitsPerCode;
            Int16                               _clearCode;
            Int16                               _EOFCode;
            Int16                               _pixel;
            Int16                               _maxCodeMult2;
            Int16                               _runningBits;
            Int16                               _runCode;
            Int16                               _curCode;
            Int16                               _newCode;
            Int16                               _nextByte;
            Int16                               _oldCode;
            Int16                               _curPrefix;

            UInt32                              _newKey;
            UInt32                              _codeBuffer;
            UInt32                              _pixelCount;
            UInt32                              _RunCode;

            UInt32                              _hashTable[_HT_SIZE];
            Int16 *                             _HTADR_INT()
            {
                return ((Int16 *)_hashTable);
            }

            UInt8                               _buffer[_LZ_BUFFER_SIZE * 2];
            Int16                               _bufferLength;

            Int16                               _stackCount;

            Int16 *                             _prefix()
            {
                return _HTADR_INT();
            }

            Int16 *                             _sufix()
            {
                return _HTADR_INT() + _LZ_MAX_CODE + 1;
            }

            Int16 *                             _stack()
            {
                return _HTADR_INT() + 2 * _LZ_MAX_CODE + 2;
            }

            //  Hash table functions
            void                                _clearHashTable(UInt32 *);
            void                                _initHashTable();
            Int16                               _KeyItem(UInt32);
            void                                _insertHashTable(UInt32 *,UInt32,Int16);
            Int16                               _existHashTable(UInt32 *,UInt32);

            //  Compress functions
            Int16                               _InitCompress();
            Int16                               _CompressLine(const UInt8 *,Int16,UInt8 *);
            Int16                               _CompressCode(Int16);
            Int16                               _OutByte(UInt8 *,Int16);
        };

        class CARL_PUBLIC _LzwDecompressor : public Decompressor
        {
            //////////
            //  Can't assign or copy construct
        private:
            _LzwDecompressor(const _LzwDecompressor &);
            void                        operator = (const _LzwDecompressor &);

            //////////
            //  Singleton
        private:
            _LzwDecompressor();
            virtual ~_LzwDecompressor();
        public:
            static _LzwDecompressor *       getInstance();

            //////////
            //  Operations
        public:
            virtual CompressionAlgorithm *      getCompressionAlgorithm();
            virtual UIntSize                    getRequiredOutputBufferSize(const void * pCompressedData, UIntSize compressedDataSize) throw(CompressedDataCorruptException);
            virtual UIntSize                    decompress(const void * pCompressedData, UIntSize compressedDataSize, void * pUncompressedData) throw(CompressedDataCorruptException);

            //////////
            //  Implementation
        private:
            //  Hash table constants & macros
            static const UIntSize               _HT_BITS_PER_PIXEL  = 8;
            static const UIntSize               _HT_SIZE            = 8192;
            static const UIntSize               _HT_KEY_MASK        = 0x1FFF;
            static const UIntSize               _HT_KEY_NUM_BITS    = 13;
            static const UIntSize               _HT_MAX_KEY         = 8191;
            static const UIntSize               _HT_MAX_CODE        = 4095;

            UIntSize                            _HT_GET_KEY(UIntSize l)
            {
                return l >> 12;
            }

            UIntSize                            _HT_GET_CODE(UIntSize l)
            {
                return l & 0x0FFF;
            }

            UIntSize                            _HT_PUT_KEY(UIntSize l)
            {
                return l << 12;
            }

            UIntSize                            _HT_PUT_CODE(UIntSize l)
            {
                return l & 0x0FFF;
            }

            //  LZW macros
            static const UIntSize               _LZ_MAX_CODE        = 4095;
            static const UIntSize               _LZ_BUFFER_SIZE     = 4096;
            static const UIntSize               _FLUSH_OUT          = 4096;
            static const UIntSize               _FIRST_CODE         = 4097;
            static const UIntSize               _NOT_CODE           = 4098;
            static const UIntSize               _MAX_BIT_PER_CODE   = 12;

            //  Data used by compressor
            UInt8                               _buf[2 * _LZ_BUFFER_SIZE];
            Int16                               _curBufferShift;
            Int16                               _bitsPerCode;
            Int16                               _clearCode;
            Int16                               _EOFCode;
            Int16                               _pixel;
            Int16                               _maxCodeMult2;
            Int16                               _runningBits;
            Int16                               _runCode;
            Int16                               _curCode;
            Int16                               _newCode;
            Int16                               _nextByte;
            Int16                               _oldCode;
            Int16                               _curPrefix;

            UInt32                              _newKey;
            UInt32                              _codeBuffer;
            UInt32                              _pixelCount;
            UInt32                              _RunCode;

            UInt32                              _hashTable[_HT_SIZE];
            Int16 *                             _HTADR_INT()
            {
                return ((Int16 *)_hashTable);
            }

            UInt8                               _buffer[_LZ_BUFFER_SIZE * 2];
            Int16                               _bufferLength;
            UInt8                               _inb[_LZ_BUFFER_SIZE];

            Int16                               _stackCount;

            Int16 *                             _prefix()
            {
                return _HTADR_INT();
            }

            Int16 *                             _sufix()
            {
                return _HTADR_INT() + _LZ_MAX_CODE + 1;
            }

            Int16 *                             _stack()
            {
                return _HTADR_INT() + 2 * _LZ_MAX_CODE + 2;
            }

            //  Hash table functions
            void                                _clearHashTable(UInt32 *);
            void                                _initHashTable();
            Int16                               _KeyItem(UInt32);
            void                                _insertHashTable(UInt32 *,UInt32,Int16);
            Int16                               _existHashTable(UInt32 *,UInt32);

            //  Decompress functions
            Int16                               _InitDecompress();
            Int16                               _DecompressLine(const UInt8 *, Int16,UInt8 *, Int16);
            void                                _getByte(UInt8 *,UInt8 *);
            Int16                               _getPrefixChar(UInt16 *,Int16,Int16);
            void                                _DecompressCode(Int16 *);
        };

        class CARL_PUBLIC _LzwCompressionAlgorithm : public CompressionAlgorithm
        {
            //////////
            //  Can't assign or copy construct
        private:
            _LzwCompressionAlgorithm(const _LzwCompressionAlgorithm &);
            void                        operator = (const _LzwCompressionAlgorithm &);

            //////////
            //  Singleton
        private:
            _LzwCompressionAlgorithm();
            virtual ~_LzwCompressionAlgorithm();
        public:
            static _LzwCompressionAlgorithm *   getInstance();

            //////////
            //  Operations
        public:
            virtual Array<String>               getAllNames() const;
            virtual String                      getCanonicalName() const;
            virtual String                      getPreferredName() const;
            virtual String                      getDisplayName() const;
            virtual String                      getDisplayName(const Locale & forLocale) const;
            virtual UIntSize                    getMinimumBlockSize() const;
            virtual UIntSize                    getPreferredBlockSize() const;
            virtual UIntSize                    getMaximumBlockSize() const;
            virtual Compressor *                getCompressor() const;
            virtual Decompressor *              getDecompressor() const;
        };

        //////////
        //  A platform-specific implementation of a compression manager
        class CARL_PUBLIC _CompressionManagerImpl
        {
            //////////
            //  Can't assign or copy construct
        private:
            _CompressionManagerImpl(const _CompressionManagerImpl &);
            void                        operator = (const _CompressionManagerImpl &);

            //////////
            //  Singleton
        private:
            _CompressionManagerImpl();
            ~_CompressionManagerImpl() {}
        public:
            static _CompressionManagerImpl *    getInstance();

            //////////
            //  Operations
        public:
            MapResourceFactory *            getResourceFactory() const;
            Array<CompressionAlgorithm*>    getAllCompressionAlgorithms();
            CompressionAlgorithm *          findCompressionAlgorithm(const String & name);
            void                            registerCompressionAlgorithm(CompressionAlgorithm * pCompressionAlgorithm) throw (InconsistentCompressionAlgorithmException, DuplicateCompressionAlgorithmException);

            //////////
            //  Implementation
        private:
            mutable FrameworkResourceFactory    _resourceFactory;
            CriticalSection                 _guard;
            Array<CompressionAlgorithm*>    _compressionAlgorithms;
            HashMap<String, CompressionAlgorithm*>  _mapNamesToCompressionAlgorithms;   //  Names are UPPERCASED
        };
    }
}

#endif  //  __CompressionImpl_hpp_included__
//  End of CompressionImpl.hpp
