//
//  DateTime.cpp - Implementation of the carl::DateTime class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:51 $
//  $Revision: 1.5 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

DateTime::DateTime()
{
    _year = 0;
    _month = 0;
    _day = 0;
    _hour = 0;
    _minute = 0;
    _second = 0;
    _nanosecond = 0;
}

DateTime::DateTime(int year, int month, int day, int hour, int minute, int second, UInt32 nanosecond)
{
    require(year > 0);
    require(month >= 1 && month <= 12);
    require(day >= 1 && day <= Calendar::daysInMonth(year, month));
    require(hour <= 23);
    require(minute <= 59);
    require(second <= 59);
    require(nanosecond <= 999999999);

    _year = static_cast<UInt16>(year);
    _month = static_cast<UInt8>(month);
    _day = static_cast<UInt8>(day);
    _hour = static_cast<UInt8>(hour);
    _minute = static_cast<UInt8>(minute);
    _second = static_cast<UInt8>(second);
    _nanosecond = nanosecond;
}

DateTime::DateTime(const Date & date, const Time & time)
{
    if (date.isValid() && time.isValid())
    {
        _year = static_cast<UInt16>(date.getYear());
        _month = static_cast<UInt8>(date.getMonth());
        _day = static_cast<UInt8>(date.getDay());
        _hour = static_cast<UInt8>(time.getHour());
        _minute = static_cast<UInt8>(time.getMinute());
        _second = static_cast<UInt8>(time.getSecond());
        _nanosecond = time.getNanosecond();
    }
    else
    {
        _year = 0;
        _month = 0;
        _day = 0;
        _hour = 0;
        _minute = 0;
        _second = 0;
        _nanosecond = 0;
    }
}

DateTime::DateTime(const DateTime & src)
{
    _year = src._year;
    _month = src._month;
    _day = src._day;
    _hour = src._hour;
    _minute = src._minute;
    _second = src._second;
    _nanosecond = src._nanosecond;
}

DateTime & DateTime::operator = (const DateTime & src)
{
    if (this != (&src))
    {
        _year = src._year;
        _month = src._month;
        _day = src._day;
        _hour = src._hour;
        _minute = src._minute;
        _second = src._second;
        _nanosecond = src._nanosecond;
    }
    return *this;
}

//////////
//  Operations
Date DateTime::getDate() const
{
    if (isValid())
    {
        return Date(_year, _month, _day);
    }
    return Date();
}

Time DateTime::getTime() const
{
    if (isValid())
    {
        return Time(_hour, _minute, _second, _nanosecond);
    }
    return Time();
}

DayOfWeek DateTime::getDayOfWeek() const
{
    require(isValid());

    return Calendar::getDayOfWeek(_year, _month, _day);
}

int DateTime::getDayOfYear() const
{
    require(isValid());

    return Calendar::getDayOfYear(_year, _month, _day);
}

int DateTime::compare(const DateTime & dateTime) const
{
    if (isValid())
    {
        if (dateTime.isValid())
        {   //  <valid datetime> <compare to> <valid datetime>
            if (_year > dateTime._year)
            {
                return 1;
            }
            if (_year < dateTime._year)
            {
                return -1;
            }
            if (_month > dateTime._month)
            {
                return 1;
            }
            if (_month < dateTime._month)
            {
                return -1;
            }
            if (_day > dateTime._day)
            {
                return 1;
            }
            if (_day < dateTime._day)
            {
                return -1;
            }
            if (_hour < dateTime._hour)
            {
                return -1;
            }
            if (_hour > dateTime._hour)
            {
                return 1;
            }
            if (_minute < dateTime._minute)
            {
                return -1;
            }
            if (_minute > dateTime._minute)
            {
                return 1;
            }
            if (_second < dateTime._second)
            {
                return -1;
            }
            if (_second > dateTime._second)
            {
                return 1;
            }
            if (_nanosecond < dateTime._nanosecond)
            {
                return -1;
            }
            if (_nanosecond > dateTime._nanosecond)
            {
                return 1;
            }
            return 0;
        }
        else
        {   //  <valid datetime> <compare to> <invalid datetime>
            return 1;
        }
    }
    else
    {
        if (dateTime.isValid())
        {   //  <invalid datetime> <compare to> <valid datetime>
            return -1;
        }
        else
        {   //  <invalid datetime> <compare to> <invalid datetime>
            return 0;
        }
    }
}

//////////
//  Hashing
UIntSize CARL_CALLBACK StandardHashFunctions::hashDateTime(const DateTime & key)
{
    return key.getSecond() * 60 * (key.getMinute() + 60 * (key.getHour() + 24 * (key.getDay() + 31 * (key.getMonth() + 12 * key.getYear()))));
}

//  End of DateTime.cpp
