//
//  DecimalFixed.cpp - Implementation of the carl::DecimalFixed class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/08/07 13:51:37 $
//  $Revision: 1.7 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <float.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <math.h>
#else
    #error Unsupported platform
#endif
using namespace carl;

UIntSize DecimalFixed::_defaultDivisionPrecision = 0;

//////////
//  Construction/destruction/assignment
DecimalFixed::DecimalFixed()
    :   _magnitude(),
        _scale()
{
}

DecimalFixed::DecimalFixed(Int8 value)
    :   _magnitude(value),
        _scale(0)
{
    _normalize();
}

DecimalFixed::DecimalFixed(Int16 value)
    :   _magnitude(value),
        _scale(0)
{
    _normalize();
}

DecimalFixed::DecimalFixed(Int32 value)
    :   _magnitude(value),
        _scale(0)
{
    _normalize();
}

DecimalFixed::DecimalFixed(Int64 value)
    :   _magnitude(value),
        _scale(0)
{
    _normalize();
}

DecimalFixed::DecimalFixed(UInt8 value)
    :   _magnitude(value),
        _scale(0)
{
    _normalize();
}

DecimalFixed::DecimalFixed(UInt16 value)
    :   _magnitude(value),
        _scale(0)
{
    _normalize();
}

DecimalFixed::DecimalFixed(UInt32 value)
    :   _magnitude(value),
        _scale(0)
{
    _normalize();
}

DecimalFixed::DecimalFixed(UInt64 value)
    :   _magnitude(value),
        _scale(0)
{
    _normalize();
}

DecimalFixed::DecimalFixed(Real32 value)
{
    _fromReal(value);
}

DecimalFixed::DecimalFixed(Real64 value)
{
    _fromReal(value);
}

DecimalFixed::DecimalFixed(const DecimalInteger & value)
    :   _magnitude(value),
        _scale(0)
{
    _normalize();
}

DecimalFixed::DecimalFixed(const DecimalInteger & magnitude, UInt32 scale)
    :   _magnitude(magnitude),
        _scale(scale)
{
    _normalize();
}

DecimalFixed::DecimalFixed(const char * pValueString) throw (ParseException)
{
    require(pValueString != NULL);

    _fromString(String(pValueString), 10);
}

DecimalFixed::DecimalFixed(const Char * pValueString) throw (ParseException)
{
    require(pValueString != NULL);

    _fromString(String(pValueString), 10);
}

DecimalFixed::DecimalFixed(const String & valueString) throw (ParseException)
{
    _fromString(valueString, 10);
}

DecimalFixed::DecimalFixed(const char * pValueString, UIntSize radix) throw (ParseException)
{
    require(pValueString != NULL);
    require(radix >= 2 && radix <= 36);

    _fromString(String(pValueString), radix);
}

DecimalFixed::DecimalFixed(const Char * pValueString, UIntSize radix) throw (ParseException)
{
    require(pValueString != NULL);
    require(radix >= 2 && radix <= 36);

    _fromString(String(pValueString), radix);
}

DecimalFixed::DecimalFixed(const String & valueString, UIntSize radix) throw (ParseException)
{
    require(radix >= 2 && radix <= 36);

    _fromString(valueString, radix);
}

DecimalFixed::DecimalFixed(const DecimalFixed & src)
    :   _magnitude(src._magnitude),
        _scale(src._scale)
{
}

DecimalFixed::~DecimalFixed()
{
}

DecimalFixed & DecimalFixed::operator = (const DecimalFixed & src)
{
    if (this != (&src))
    {
        _magnitude = src._magnitude;
        _scale = src._scale;
    }
    return *this;
}

//////////
//  Operations
UIntSize DecimalFixed::getDefaultDivisionPrecision()
{
    CriticalSectionLock lock(*_getStaticGuard());
    return _defaultDivisionPrecision;
}

void DecimalFixed::setDefaultDivisionPrecision(UIntSize precision)
{
    CriticalSectionLock lock(*_getStaticGuard());
    _defaultDivisionPrecision = precision;
}

DecimalFixed DecimalFixed::operator + () const
{
    return *this;
}

DecimalFixed DecimalFixed::operator - () const
{
    return DecimalFixed(-_magnitude, _scale);
}

DecimalFixed DecimalFixed::operator + (const DecimalFixed & op2) const
{
    if (_magnitude == 0)
    {
        return op2;
    }
    else if (op2._magnitude == 0)
    {
        return (*this);
    }
    else
    {   //  Both operands are finite non-zeroes
        //  Bring both operands to the same scale
        DecimalInteger magnitude1 = _magnitude;
        UInt32 scale1 = _scale;
        DecimalInteger magnitude2 = op2._magnitude;
        UInt32 scale2 = op2._scale;
        if (scale1 < scale2)
        {
            magnitude1.multiplyByPowerOf10(scale2 - scale1);
            scale1 = scale2;
        }
        if (scale1 > scale2)
        {
            magnitude2.multiplyByPowerOf10(scale1 - scale2);
            scale2 = scale1;
        }
        //  Perform the addition
        return DecimalFixed(magnitude1 + magnitude2, scale1);
    }
}

DecimalFixed DecimalFixed::operator - (const DecimalFixed & op2) const
{
    if (_magnitude == 0)
    {
        return -op2;
    }
    else if (op2._magnitude == 0)
    {
        return (*this);
    }
    else
    {   //  Both operands are finite non-zeroes
        //  Bring both operands to the same scale
        DecimalInteger magnitude1 = _magnitude;
        UInt32 scale1 = _scale;
        DecimalInteger magnitude2 = op2._magnitude;
        UInt32 scale2 = op2._scale;
        if (scale1 < scale2)
        {
            magnitude1.multiplyByPowerOf10(scale2 - scale1);
            scale1 = scale2;
        }
        if (scale1 > scale2)
        {
            magnitude2.multiplyByPowerOf10(scale1 - scale2);
            scale2 = scale1;
        }
        //  Perform the subtraction
        return DecimalFixed(magnitude1 - magnitude2, scale1);
    }
}

DecimalFixed DecimalFixed::operator * (const DecimalFixed & op2) const
{
    if (_magnitude == 0 || op2._magnitude == 0)
    {
        return 0;
    }
    else
    {
        return DecimalFixed(_magnitude * op2._magnitude, _scale + op2._scale);
    }
}

DecimalFixed DecimalFixed::operator / (const DecimalFixed & op2) const
{
    if (_magnitude == 0)
    {
        return 0;
    }
    else if (op2._magnitude == 0)
    {
        throw FixedPointDivisionByZeroError();
    }
    else
    {   //  Both operands are non-zero positive values
        CriticalSectionLock lock(*_getStaticGuard());
        //  Decide on the required precision of the result
        UInt32 resultDigits = _magnitude.getRequiredDigits() + op2._magnitude.getRequiredDigits();
        //  Divide
        DecimalInteger resultMagnitude = (_magnitude.multipliedByPowerOf10(resultDigits)) / op2._magnitude;
        UInt32 deltaScale = 0;
        if (resultDigits + _scale < op2._scale)
        {
            deltaScale = op2._scale - (resultDigits + _scale);
            resultMagnitude.multiplyByPowerOf10(deltaScale);
        }
        return DecimalFixed(resultMagnitude, resultDigits + deltaScale + _scale - op2._scale);
    }
}

DecimalFixed & DecimalFixed::operator += (const DecimalFixed & op2)
{
    (*this) = (*this) + op2;
    return *this;
}

DecimalFixed & DecimalFixed::operator -= (const DecimalFixed & op2)
{
    (*this) = (*this) - op2;
    return *this;
}

DecimalFixed & DecimalFixed::operator *= (const DecimalFixed & op2)
{
    (*this) = (*this) * op2;
    return *this;
}

DecimalFixed & DecimalFixed::operator /= (const DecimalFixed & op2)
{
    (*this) = (*this) / op2;
    return *this;
}

DecimalFixed DecimalFixed::operator ++()
{
    (*this) += 1;
    return *this;
}

DecimalFixed DecimalFixed::operator --()
{
    (*this) -= 1;
    return *this;
}

DecimalFixed DecimalFixed::operator ++(int)
{
    DecimalFixed result = (*this);
    (*this) += 1;
    return result;
}

DecimalFixed DecimalFixed::operator --(int)
{
    DecimalFixed result = (*this);
    (*this) -= 1;
    return result;
}

int DecimalFixed::compare(const DecimalFixed & op2) const
{
    return ((*this) - op2).getSign();
}

bool DecimalFixed::operator == (const DecimalFixed & op2) const
{
    return compare(op2) == 0;
}

bool DecimalFixed::operator != (const DecimalFixed & op2) const
{
    return compare(op2) != 0;
}

bool DecimalFixed::operator < (const DecimalFixed & op2) const
{
    return compare(op2) < 0;
}

bool DecimalFixed::operator <= (const DecimalFixed & op2) const
{
    return compare(op2) <= 0;
}

bool DecimalFixed::operator > (const DecimalFixed & op2) const
{
    return compare(op2) > 0;
}

bool DecimalFixed::operator >= (const DecimalFixed & op2) const
{
    return compare(op2) >= 0;
}

Int8 DecimalFixed::toInt8(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int8_Max)
    {
        return Limits::Int8_Max;
    }
    else if (value64 < Limits::Int8_Min)
    {
        return Limits::Int8_Min;
    }
    else
    {
        return static_cast<Int8>(value64);
    }
}

Int16 DecimalFixed::toInt16(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int16_Max)
    {
        return Limits::Int16_Max;
    }
    else if (value64 < Limits::Int16_Min)
    {
        return Limits::Int16_Min;
    }
    else
    {
        return static_cast<Int16>(value64);
    }
}

Int32 DecimalFixed::toInt32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int32_Max)
    {
        return Limits::Int32_Max;
    }
    else if (value64 < Limits::Int32_Min)
    {
        return Limits::Int32_Min;
    }
    else
    {
        return static_cast<Int32>(value64);
    }
}

Int64 DecimalFixed::toInt64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    DecimalInteger value = toDecimalInteger(roundingMode);
    if (value < Limits::Int64_Min)
    {
        return Limits::Int64_Min;
    }
    else if (value > Limits::Int64_Max)
    {
        return Limits::Int64_Max;
    }
    else
    {
        return value.toInt64();
    }
}

UInt8 DecimalFixed::toUInt8(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt8_Max)
    {
        return Limits::UInt8_Max;
    }
    else
    {
        return static_cast<UInt8>(value64);
    }
}

UInt16 DecimalFixed::toUInt16(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt16_Max)
    {
        return Limits::UInt16_Max;
    }
    else
    {
        return static_cast<UInt16>(value64);
    }
}

UInt32 DecimalFixed::toUInt32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt32_Max)
    {
        return Limits::UInt32_Max;
    }
    else
    {
        return static_cast<UInt32>(value64);
    }
}

UInt64 DecimalFixed::toUInt64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    DecimalInteger value = toDecimalInteger(roundingMode);
    if (value < Limits::UInt64_Min)
    {
        return Limits::UInt64_Min;
    }
    else if (value > Limits::UInt64_Max)
    {
        return Limits::UInt64_Max;
    }
    else
    {
        return value.toUInt64();
    }
}

UIntSize DecimalFixed::toUIntSize(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UIntSize_Max)
    {
        return Limits::UIntSize_Max;
    }
    else
    {
        return static_cast<UIntSize>(value64);
    }
}

UIntPtr DecimalFixed::toUIntPtr(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UIntPtr_Max)
    {
        return Limits::UIntPtr_Max;
    }
    else
    {
        return static_cast<UIntPtr>(value64);
    }
}

IntPtrDiff DecimalFixed::toIntPtrDiff(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::IntPtrDiff_Max)
    {
        return Limits::IntPtrDiff_Max;
    }
    else if (value64 < Limits::IntPtrDiff_Min)
    {
        return Limits::IntPtrDiff_Min;
    }
    else
    {
        return static_cast<IntPtrDiff>(value64);
    }
}

Real32 DecimalFixed::toReal32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    return static_cast<Real32>(toReal64(roundingMode));
}

Real64 DecimalFixed::toReal64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    if (_magnitude < 0)
    {   //  Make sure the translated value is positive
        DecimalFixed minusThis = -(*this);
        switch (roundingMode)
        {
            case RoundingMode_TowardsNearest:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsZero:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsPositiveInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsNegativeInfinity));
            case RoundingMode_TowardsNegativeInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsPositiveInfinity));
            default:
                require(false);
                return 0;
        }
    }
    else
    {
        Real64 addend = 1.0;
        for (UInt32 scale = 0; scale < _scale; scale++)
        {
            addend /= 10.0;
        }
        Real64 result = 0;
        for (UIntSize digitNo = 0; digitNo < _magnitude.getRequiredDigits(); digitNo++)
        {
            result += addend * _magnitude.getDigit(digitNo);
            addend *= 10.0;
        }
        return result;
    }
}

DecimalInteger DecimalFixed::toDecimalInteger(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    if (_magnitude.getSign() == -1)
    {   //  Make sure the translated value is positive
        DecimalFixed minusThis = -(*this);
        switch (roundingMode)
        {
            case RoundingMode_TowardsNearest:
                return -(minusThis.toDecimalInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsZero:
                return -(minusThis.toDecimalInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsInfinity:
                return -(minusThis.toDecimalInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsPositiveInfinity:
                return -(minusThis.toDecimalInteger(RoundingMode_TowardsNegativeInfinity));
            case RoundingMode_TowardsNegativeInfinity:
                return -(minusThis.toDecimalInteger(RoundingMode_TowardsPositiveInfinity));
            default:
                require(false);
                return 0;
        }
    }
    else if (_scale == 0)
    {   //  No rounding is necessary
        return _magnitude;
    }
    else
    {   //  Rounding may be necessary
        DecimalInteger result = _magnitude;
        UIntSize nonzeroDigitsShiftedOut = 0;
        int lastDigitShiftedOut = 0;
        for (UInt32 digitNo = 0; digitNo > _scale; digitNo--)
        {
            if ((lastDigitShiftedOut = result.getDigit(digitNo)) != 0)
            {
                nonzeroDigitsShiftedOut++;
            }
        }
        result.divideByPowerOf10(_scale);
        if (nonzeroDigitsShiftedOut != 0)
        {   //  Rounding is necessary
            switch (roundingMode)
            {
                case RoundingMode_TowardsNearest:
                    if (lastDigitShiftedOut >= 5)
                    {
                        if (nonzeroDigitsShiftedOut == 1)
                        {   //  Only one '5' digit was shifted out
                            if (lastDigitShiftedOut == 5)
                            {   //  The value is exactly between two integers; chose an odd one
                                if (result.getDigit(0) % 2 != 0)
                                {
                                    result++;
                                }
                            }
                            else
                            {
                                result++;
                            }
                        }
                        else
                        {   //  More than one digit was shifted out, and last of then was >=5
                            result++;
                        }
                    }
                    break;
                case RoundingMode_TowardsZero:
                    break;
                case RoundingMode_TowardsInfinity:
                    result++;
                    break;
                case RoundingMode_TowardsPositiveInfinity:
                    result++;
                    break;
                case RoundingMode_TowardsNegativeInfinity:
                    break;
                default:
                    require(false);
            }
        }
        return result;
    }
}

String DecimalFixed::toString(UIntSize radix /*= 10*/) const
{
    RealFormatProperties formatProperties;
    formatProperties.setRadix(radix);
    return NumberFormat::getInstance(Locale())->format(*this, formatProperties);
}

int DecimalFixed::getSign() const
{
    return _magnitude.getSign();
}

DecimalInteger DecimalFixed::getMagnitude() const
{
    return (_magnitude.getSign() == -1) ? (-_magnitude) : (_magnitude);
}

UInt32 DecimalFixed::getScale() const
{
    return _scale;
}

void DecimalFixed::setScale(UInt32 scale, RoundingMode roundingMode /*= RoundingMode_TowardsZero*/)
{
    if (scale == _scale)
    {
        return;
    }
    else if (scale > _scale)
    {   //  Scale increased, no rounding needed
        _magnitude.multiplyByPowerOf10(scale - _scale);
        _scale = scale;
    }
    else
    {   //  Scale decreased, rounding in effect
        UIntSize digitsToRemove = (_scale - scale);
        UIntSize nonzeroDigitsShiftedOut = 0;
        int lastDigitShiftedOut = 0;
        for (UIntSize i = 0; i < digitsToRemove; i++)
        {
            if ((lastDigitShiftedOut = _magnitude.getDigit(i)) != 0)
            {
                nonzeroDigitsShiftedOut++;
            }
        }
        _magnitude.divideByPowerOf10(digitsToRemove);
        _scale = scale;
        _normalize();
        if (nonzeroDigitsShiftedOut != 0)
        {   //  Rounding is necessary
            switch (roundingMode)
            {
                case RoundingMode_TowardsNearest:
                    if (lastDigitShiftedOut >= 5)
                    {
                        if (nonzeroDigitsShiftedOut == 1)
                        {   //  Only one '5' digit was shifted out
                            if (lastDigitShiftedOut == 5)
                            {   //  The value is exactly between two integers; chose an odd one
                                if (_magnitude.getDigit(0) % 2 != 0)
                                {
                                    _magnitude++;
                                }
                            }
                            else
                            {
                                _magnitude++;
                            }
                        }
                        else
                        {   //  More than one digit was shifted out, and last of then was >=5
                            _magnitude++;
                        }
                    }
                    break;
                case RoundingMode_TowardsZero:
                    break;
                case RoundingMode_TowardsInfinity:
                    _magnitude++;
                    break;
                case RoundingMode_TowardsPositiveInfinity:
                    if (_magnitude.getSign() != -1)
                    {
                        _magnitude++;
                    }
                    break;
                case RoundingMode_TowardsNegativeInfinity:
                    if (_magnitude.getSign() == -1)
                    {
                        _magnitude++;
                    }
                    break;
                default:
                    require(false);
            }
        }
    }
}

//////////
//  Implementation helpers
CriticalSection * DecimalFixed::_getStaticGuard()
{
    static CriticalSection instance;
    return &instance;
}

void DecimalFixed::_normalize()
{
    //  Count the number of tailing 0's
    UIntSize maxDigitsToPrune = Math::min(_scale, _magnitude.getRequiredDigits());
    UIntSize digitsToPrune;
    for (digitsToPrune = 0; digitsToPrune < maxDigitsToPrune && _magnitude.getDigit(digitsToPrune) == 0; digitsToPrune++)
    {   //  All work is done in the loop header
    }
    if (digitsToPrune != 0)
    {
        _magnitude.divideByPowerOf10(digitsToPrune);
        _scale -= digitsToPrune;
    }
    if (_magnitude == 0)
    {
        _scale = 0;
    }
}

void DecimalFixed::_fromReal(Real64 value)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    switch (_fpclass(value))
    {
        case _FPCLASS_SNAN:
        case _FPCLASS_QNAN:
        case _FPCLASS_NINF:
        case _FPCLASS_PINF:
            throw ValueNotRepresentableAsFixedPointError();
        default:
            break;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    int isInf = isinf(value);
    if (isInf == -1 || isInf == 1)
    {
        throw ValueNotRepresentableAsFixedPointError();
    }
    else if (isnan(value))
    {
        throw ValueNotRepresentableAsFixedPointError();
    }
#else
    #error Unsupported platform
#endif
    Int64 valueBits = *((Int64*)(&value));
    bool isNegative = ((valueBits >> 63) & 0x01) != 0;
    Int16 exponent = static_cast<Int16>(((valueBits >> 52) & 0x07FF));
    Int64 mantissa = (valueBits & CARL_INT64_CONSTANT(0xFFFFFFFFFFFFF));
    if (exponent != 0)
    {   //  Normalized values have an implicit '1' bit
        mantissa |= CARL_INT64_CONSTANT(0x10000000000000);
    }
    _magnitude = (isNegative) ? (-mantissa) : (mantissa);
    _scale = 0;
    exponent -= 1023 + 52;
    if (exponent > 0)
    {
        DecimalInteger multiplier = 1;
        for (Int16 i = 0; i < exponent; i++)
        {
            multiplier *= 2;
        }
        (*this) *= multiplier;
    }
    else if (exponent < 0)
    {
        DecimalInteger divisor = 1;
        for (Int16 i = 0; i > exponent; i--)
        {
            divisor *= 2;
        }
        (*this) /= divisor;
    }
    _normalize();
}

void DecimalFixed::_fromString(const String & s, UIntSize radix) throw (ParseException)
{
    RealParseProperties parseProperties;
    parseProperties.setRadix(radix);
    DecimalFixed bf = NumberParse::getInstance(Locale())->parseDecimalFixed(s, parseProperties);
    _magnitude = bf._magnitude;
    _scale = bf._scale;
}

void DecimalFixed::_multiplyDecimalIntegerByRadix(DecimalInteger & bi, UInt8 radix, UIntSize repeatCount)
{
    require(radix >= 2 && radix <= 36);

    switch (radix)
    {
        case 10:
            bi.multiplyByPowerOf10(repeatCount);
            return;
        default:
            break;
    }
    for (UIntSize i = 0; i < repeatCount; i++)
    {
        bi *= radix;
    }
}

//////////
//  Additional arbitrary-precision operators
DecimalFixed CARL_API carl::operator + (Int8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) + op2;
}

DecimalFixed CARL_API carl::operator + (Int16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) + op2;
}

DecimalFixed CARL_API carl::operator + (Int32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) + op2;
}

DecimalFixed CARL_API carl::operator + (Int64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) + op2;
}

DecimalFixed CARL_API carl::operator + (UInt8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) + op2;
}

DecimalFixed CARL_API carl::operator + (UInt16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) + op2;
}

DecimalFixed CARL_API carl::operator + (UInt32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) + op2;
}

DecimalFixed CARL_API carl::operator + (UInt64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) + op2;
}

DecimalFixed CARL_API carl::operator + (Real32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) + op2;
}

DecimalFixed CARL_API carl::operator + (Real64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) + op2;
}

DecimalFixed CARL_API carl::operator - (Int8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) - op2;
}

DecimalFixed CARL_API carl::operator - (Int16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) - op2;
}

DecimalFixed CARL_API carl::operator - (Int32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) - op2;
}

DecimalFixed CARL_API carl::operator - (Int64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) - op2;
}

DecimalFixed CARL_API carl::operator - (UInt8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) - op2;
}

DecimalFixed CARL_API carl::operator - (UInt16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) - op2;
}

DecimalFixed CARL_API carl::operator - (UInt32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) - op2;
}

DecimalFixed CARL_API carl::operator - (UInt64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) - op2;
}

DecimalFixed CARL_API carl::operator - (Real32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) - op2;
}

DecimalFixed CARL_API carl::operator - (Real64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) - op2;
}

DecimalFixed CARL_API carl::operator * (Int8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) * op2;
}

DecimalFixed CARL_API carl::operator * (Int16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) * op2;
}

DecimalFixed CARL_API carl::operator * (Int32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) * op2;
}

DecimalFixed CARL_API carl::operator * (Int64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) * op2;
}

DecimalFixed CARL_API carl::operator * (UInt8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) * op2;
}

DecimalFixed CARL_API carl::operator * (UInt16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) * op2;
}

DecimalFixed CARL_API carl::operator * (UInt32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) * op2;
}

DecimalFixed CARL_API carl::operator * (UInt64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) * op2;
}

DecimalFixed CARL_API carl::operator * (Real32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) * op2;
}

DecimalFixed CARL_API carl::operator * (Real64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) * op2;
}

DecimalFixed CARL_API carl::operator / (Int8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) / op2;
}

DecimalFixed CARL_API carl::operator / (Int16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) / op2;
}

DecimalFixed CARL_API carl::operator / (Int32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) / op2;
}

DecimalFixed CARL_API carl::operator / (Int64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) / op2;
}

DecimalFixed CARL_API carl::operator / (UInt8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) / op2;
}

DecimalFixed CARL_API carl::operator / (UInt16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) / op2;
}

DecimalFixed CARL_API carl::operator / (UInt32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) / op2;
}

DecimalFixed CARL_API carl::operator / (UInt64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) / op2;
}

DecimalFixed CARL_API carl::operator / (Real32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) / op2;
}

DecimalFixed CARL_API carl::operator / (Real64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) / op2;
}

bool CARL_API carl::operator == (Int8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) == op2;
}

bool CARL_API carl::operator == (Int16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) == op2;
}

bool CARL_API carl::operator == (Int32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) == op2;
}

bool CARL_API carl::operator == (Int64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) == op2;
}

bool CARL_API carl::operator == (UInt8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) == op2;
}

bool CARL_API carl::operator == (UInt16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) == op2;
}

bool CARL_API carl::operator == (UInt32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) == op2;
}

bool CARL_API carl::operator == (UInt64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) == op2;
}

bool CARL_API carl::operator == (Real32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) == op2;
}

bool CARL_API carl::operator == (Real64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) == op2;
}

bool CARL_API carl::operator != (Int8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) != op2;
}

bool CARL_API carl::operator != (Int16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) != op2;
}

bool CARL_API carl::operator != (Int32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) != op2;
}

bool CARL_API carl::operator != (Int64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) != op2;
}

bool CARL_API carl::operator != (UInt8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) != op2;
}

bool CARL_API carl::operator != (UInt16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) != op2;
}

bool CARL_API carl::operator != (UInt32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) != op2;
}

bool CARL_API carl::operator != (UInt64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) != op2;
}

bool CARL_API carl::operator != (Real32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) != op2;
}

bool CARL_API carl::operator != (Real64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) != op2;
}

bool CARL_API carl::operator < (Int8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) < op2;
}

bool CARL_API carl::operator < (Int16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) < op2;
}

bool CARL_API carl::operator < (Int32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) < op2;
}

bool CARL_API carl::operator < (Int64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) < op2;
}

bool CARL_API carl::operator < (UInt8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) < op2;
}

bool CARL_API carl::operator < (UInt16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) < op2;
}

bool CARL_API carl::operator < (UInt32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) < op2;
}

bool CARL_API carl::operator < (UInt64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) < op2;
}

bool CARL_API carl::operator < (Real32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) < op2;
}

bool CARL_API carl::operator < (Real64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) < op2;
}

bool CARL_API carl::operator <= (Int8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (Int16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (Int32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (Int64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (Real32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) <= op2;
}

bool CARL_API carl::operator <= (Real64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) <= op2;
}

bool CARL_API carl::operator > (Int8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) > op2;
}

bool CARL_API carl::operator > (Int16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) > op2;
}

bool CARL_API carl::operator > (Int32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) > op2;
}

bool CARL_API carl::operator > (Int64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) > op2;
}

bool CARL_API carl::operator > (UInt8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) > op2;
}

bool CARL_API carl::operator > (UInt16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) > op2;
}

bool CARL_API carl::operator > (UInt32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) > op2;
}

bool CARL_API carl::operator > (UInt64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) > op2;
}

bool CARL_API carl::operator > (Real32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) > op2;
}

bool CARL_API carl::operator > (Real64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) > op2;
}

bool CARL_API carl::operator >= (Int8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (Int16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (Int32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (Int64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt8 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt16 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (Real32 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) >= op2;
}

bool CARL_API carl::operator >= (Real64 op1, const DecimalFixed & op2)
{
    return DecimalFixed(op1) >= op2;
}

//////////
//  Hashing arbitrary-precision values
UIntSize CARL_CALLBACK StandardHashFunctions::hashDecimalFixed(const DecimalFixed & key)
{
    return hashDecimalInteger(key.getMagnitude());
}

//  End of DecimalFixed.cpp
