//
//  DecimalFloat.cpp - Implementation of the carl::DecimalFloat class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:52 $
//  $Revision: 1.7 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <float.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <math.h>
#else
    #error Unsupported platform
#endif

UIntSize DecimalFloat::_defaultDivisionPrecision = 0;

DecimalFloat::DecimalFloat()
    :   _category(ValueCategory_Normal),
        _mantissa(0),
        _scale(0)
{
}

DecimalFloat::DecimalFloat(ValueCategory category)
    :   _category(category),
        _mantissa(0),
        _scale(0)
{
    require(category != ValueCategory_Normal);
}

DecimalFloat::DecimalFloat(Int8 value)
    :   _category(ValueCategory_Normal),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

DecimalFloat::DecimalFloat(Int16 value)
    :   _category(ValueCategory_Normal),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

DecimalFloat::DecimalFloat(Int32 value)
    :   _category(ValueCategory_Normal),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

DecimalFloat::DecimalFloat(Int64 value)
    :   _category(ValueCategory_Normal),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

DecimalFloat::DecimalFloat(UInt8 value)
    :   _category(ValueCategory_Normal),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

DecimalFloat::DecimalFloat(UInt16 value)
    :   _category(ValueCategory_Normal),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

DecimalFloat::DecimalFloat(UInt32 value)
    :   _category(ValueCategory_Normal),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

DecimalFloat::DecimalFloat(UInt64 value)
    :   _category(ValueCategory_Normal),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

DecimalFloat::DecimalFloat(Real32 value)
{
    _fromReal(value);
}

DecimalFloat::DecimalFloat(Real64 value)
{
    _fromReal(value);
}

DecimalFloat::DecimalFloat(const DecimalInteger & value)
    :   _category(ValueCategory_Normal),
        _mantissa(value),
        _scale(0)
{
    _normalize();
}

DecimalFloat::DecimalFloat(const DecimalInteger & mantissa, const DecimalInteger & scale)
    :   _category(ValueCategory_Normal),
        _mantissa(mantissa),
        _scale(scale)
{
    _normalize();
}

DecimalFloat::DecimalFloat(const char * pValueString) throw (ParseException)
{
    require(pValueString != NULL);

    _fromString(String(pValueString), 10);
}

DecimalFloat::DecimalFloat(const Char * pValueString) throw (ParseException)
{
    require(pValueString != NULL);

    _fromString(String(pValueString), 10);
}

DecimalFloat::DecimalFloat(const String & valueString) throw (ParseException)
{
    _fromString(valueString, 10);
}

DecimalFloat::DecimalFloat(const char * pValueString, UIntSize radix) throw (ParseException)
{
    require(pValueString != NULL);
    require(radix >= 2 && radix <= 36);

    _fromString(String(pValueString), radix);
}

DecimalFloat::DecimalFloat(const Char * pValueString, UIntSize radix) throw (ParseException)
{
    require(pValueString != NULL);
    require(radix >= 2 && radix <= 36);

    _fromString(String(pValueString), radix);
}

DecimalFloat::DecimalFloat(const String & valueString, UIntSize radix) throw (ParseException)
{
    require(radix >= 2 && radix <= 36);

    _fromString(valueString, radix);
}

DecimalFloat::DecimalFloat(const DecimalFloat & src)
    :   _category(src._category),
        _mantissa(src._mantissa),
        _scale(src._scale)
{
}

DecimalFloat::~DecimalFloat()
{
}

DecimalFloat & DecimalFloat::operator = (const DecimalFloat & src)
{
    if (this != (&src))
    {
        _category = src._category;
        _mantissa = src._mantissa;
        _scale = src._scale;
    }
    return *this;
}

//////////
//  Operations
UIntSize DecimalFloat::getDefaultDivisionPrecision()
{
    CriticalSectionLock lock(*_getStaticGuard());
    return _defaultDivisionPrecision;
}

void DecimalFloat::setDefaultDivisionPrecision(UIntSize precision)
{
    CriticalSectionLock lock(*_getStaticGuard());
    _defaultDivisionPrecision = precision;
}

DecimalFloat DecimalFloat::operator + () const
{
    if (_category == ValueCategory_SignalingNan)
    {
        throw SignalingNanOperandError();
    }
    return *this;
}

DecimalFloat DecimalFloat::operator - () const
{
    switch (_category)
    {
        case ValueCategory_Normal:
            return DecimalFloat(-_mantissa, _scale);
        case ValueCategory_PositiveInfinity:
            return DecimalFloat(ValueCategory_NegativeInfinity);
        case ValueCategory_NegativeInfinity:
            return DecimalFloat(ValueCategory_PositiveInfinity);
        case ValueCategory_SignalingNan:
            throw SignalingNanOperandError();
        case ValueCategory_QuietNan:
            return DecimalFloat(ValueCategory_QuietNan);
        default:
            require(false);
            return DecimalFloat(ValueCategory_SignalingNan);
    }
}

DecimalFloat DecimalFloat::operator + (const DecimalFloat & op2) const
{
    if (_category == ValueCategory_SignalingNan || op2._category == ValueCategory_SignalingNan)
    {
        throw SignalingNanOperandError();
    }
    else if (_category == ValueCategory_QuietNan || op2._category == ValueCategory_QuietNan)
    {
        return DecimalFloat(ValueCategory_QuietNan);
    }
    else if (_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_PositiveInfinity)
    {
        return DecimalFloat(ValueCategory_PositiveInfinity);
    }
    else if (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_NegativeInfinity)
    {
        return DecimalFloat(ValueCategory_NegativeInfinity);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_NegativeInfinity) ||
             (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_PositiveInfinity))
    {
        return DecimalFloat(ValueCategory_QuietNan);
    }
    else if (_category == ValueCategory_PositiveInfinity || op2._category == ValueCategory_PositiveInfinity)
    {
        return DecimalFloat(ValueCategory_PositiveInfinity);
    }
    else if (_category == ValueCategory_NegativeInfinity || op2._category == ValueCategory_NegativeInfinity)
    {
        return DecimalFloat(ValueCategory_NegativeInfinity);
    }
    else if (_mantissa == 0)
    {
        return op2;
    }
    else if (op2._mantissa == 0)
    {
        return (*this);
    }
    else
    {   //  Both operands are finite non-zeroes
        //  Bring both operands to the same scale
        DecimalInteger mantissa1 = _mantissa;
        DecimalInteger scale1 = _scale;
        DecimalInteger mantissa2 = op2._mantissa;
        DecimalInteger scale2 = op2._scale;
        if (scale1 < scale2)
        {
            require(scale2 - scale1 <= Limits::UIntSize_Max);
            mantissa2.multiplyByPowerOf10((scale2 - scale1).toUIntSize());
            scale2 = scale1;
        }
        if (scale1 > scale2)
        {
            require(scale1 - scale2 <= Limits::UIntSize_Max);
            mantissa1.multiplyByPowerOf10((scale1 - scale2).toUIntSize());
            scale1 = scale2;
        }
        //  Perform the addition
        return DecimalFloat(mantissa1 + mantissa2, scale1);
    }
}

DecimalFloat DecimalFloat::operator - (const DecimalFloat & op2) const
{
    if (_category == ValueCategory_SignalingNan || op2._category == ValueCategory_SignalingNan)
    {
        throw SignalingNanOperandError();
    }
    else if (_category == ValueCategory_QuietNan || op2._category == ValueCategory_QuietNan)
    {
        return DecimalFloat(ValueCategory_QuietNan);
    }
    else if (_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_NegativeInfinity)
    {
        return DecimalFloat(ValueCategory_PositiveInfinity);
    }
    else if (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_PositiveInfinity)
    {
        return DecimalFloat(ValueCategory_NegativeInfinity);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_PositiveInfinity) ||
             (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_NegativeInfinity))
    {
        return DecimalFloat(ValueCategory_QuietNan);
    }
    else if (_category == ValueCategory_PositiveInfinity)
    {
        return DecimalFloat(ValueCategory_PositiveInfinity);
    }
    else if (op2._category == ValueCategory_PositiveInfinity)
    {
        return DecimalFloat(ValueCategory_NegativeInfinity);
    }
    else if (_category == ValueCategory_NegativeInfinity)
    {
        return DecimalFloat(ValueCategory_NegativeInfinity);
    }
    else if (op2._category == ValueCategory_NegativeInfinity)
    {
        return DecimalFloat(ValueCategory_PositiveInfinity);
    }
    else if (_mantissa == 0)
    {
        return -op2;
    }
    else if (op2._mantissa == 0)
    {
        return (*this);
    }
    else
    {   //  Both operands are finite non-zeroes
        //  Bring both operands to the same scale
        DecimalInteger mantissa1 = _mantissa;
        DecimalInteger scale1 = _scale;
        DecimalInteger mantissa2 = op2._mantissa;
        DecimalInteger scale2 = op2._scale;
        if (scale1 < scale2)
        {
            require(scale2 - scale1 <= Limits::UIntSize_Max);
            mantissa2.multiplyByPowerOf10((scale2 - scale1).toUIntSize());
            scale2 = scale1;
        }
        if (scale1 > scale2)
        {
            require(scale1 - scale2 <= Limits::UIntSize_Max);
            mantissa1.multiplyByPowerOf10((scale1 - scale2).toUIntSize());
            scale1 = scale2;
        }
        //  Perform the subtraction
        return DecimalFloat(mantissa1 - mantissa2, scale1);
    }
}

DecimalFloat DecimalFloat::operator * (const DecimalFloat & op2) const
{
    if (_category == ValueCategory_SignalingNan || op2._category == ValueCategory_SignalingNan)
    {
        throw SignalingNanOperandError();
    }
    else if (_category == ValueCategory_QuietNan || op2._category == ValueCategory_QuietNan)
    {
        return DecimalFloat(ValueCategory_QuietNan);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_PositiveInfinity) ||
             (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_NegativeInfinity))
    {
        return DecimalFloat(ValueCategory_PositiveInfinity);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2._category == ValueCategory_NegativeInfinity) ||
             (_category == ValueCategory_NegativeInfinity && op2._category == ValueCategory_PositiveInfinity))
    {
        return DecimalFloat(ValueCategory_NegativeInfinity);
    }
    else if (_category == ValueCategory_PositiveInfinity || op2._category == ValueCategory_PositiveInfinity)
    {
        return DecimalFloat(ValueCategory_PositiveInfinity);
    }
    else if (_category == ValueCategory_NegativeInfinity || op2._category == ValueCategory_NegativeInfinity)
    {
        return DecimalFloat(ValueCategory_NegativeInfinity);
    }
    else if (_mantissa == 0 || op2._mantissa == 0)
    {
        return 0;
    }
    else
    {   //  Both operands are non-zero positive values
        return DecimalFloat(_mantissa * op2._mantissa, _scale + op2._scale);
    }
}

DecimalFloat DecimalFloat::operator / (const DecimalFloat & op2) const
{
    if (_category == ValueCategory_SignalingNan || op2._category == ValueCategory_SignalingNan)
    {
        throw SignalingNanOperandError();
    }
    else if (_category == ValueCategory_QuietNan || op2._category == ValueCategory_QuietNan)
    {
        return DecimalFloat(ValueCategory_QuietNan);
    }
    else if ((_category == ValueCategory_PositiveInfinity || _category == ValueCategory_NegativeInfinity) &&
             (op2._category == ValueCategory_PositiveInfinity || op2._category == ValueCategory_NegativeInfinity))
    {
        return DecimalFloat(ValueCategory_QuietNan);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2.getSign() != -1) ||
             (_category == ValueCategory_NegativeInfinity && op2.getSign() == -1))
    {
        return DecimalFloat(ValueCategory_PositiveInfinity);
    }
    else if ((_category == ValueCategory_PositiveInfinity && op2.getSign() == -1) ||
             (_category == ValueCategory_NegativeInfinity && op2.getSign() != -1))
    {
        return DecimalFloat(ValueCategory_NegativeInfinity);
    }
    else if (op2._category == ValueCategory_PositiveInfinity || op2._category == ValueCategory_NegativeInfinity)
    {
        return 0;
    }
    else if (_mantissa == 0)
    {
        return 0;
    }
    else if (op2._mantissa == 0)
    {
        if (_mantissa.getSign() == 1)
        {
            return DecimalFloat(ValueCategory_PositiveInfinity);
        }
        else if (_mantissa.getSign() == -1)
        {
            return DecimalFloat(ValueCategory_NegativeInfinity);
        }
        else
        {
            return DecimalFloat(ValueCategory_QuietNan);
        }
    }
    else
    {   //  Both operands are non-zero values
        CriticalSectionLock lock(*_getStaticGuard());
        DecimalInteger mantissa1 = _mantissa;
        DecimalInteger scale1 = _scale;
        DecimalInteger mantissa2 = op2._mantissa;
        DecimalInteger scale2 = op2._scale;
        UIntSize precision1 = mantissa1.getRequiredDigits();
        UIntSize precision2 = mantissa2.getRequiredDigits();
        UIntSize resultPrecision = Math::max(Math::max(precision1, precision2), _defaultDivisionPrecision);
        UIntSize requiredPresision1 = precision2 + resultPrecision - 1;
        UIntSize op1ExtraScale = requiredPresision1 - precision1;
        DecimalInteger resultMantissa = (mantissa1.multipliedByPowerOf10(op1ExtraScale)) / mantissa2;
        DecimalInteger resultScale = scale1 - scale2 - op1ExtraScale;
        return DecimalFloat(resultMantissa, resultScale);
    }
}

DecimalFloat & DecimalFloat::operator += (const DecimalFloat & op2)
{
    (*this) = (*this) + op2;
    return *this;
}

DecimalFloat & DecimalFloat::operator -= (const DecimalFloat & op2)
{
    (*this) = (*this) - op2;
    return *this;
}

DecimalFloat & DecimalFloat::operator *= (const DecimalFloat & op2)
{
    (*this) = (*this) * op2;
    return *this;
}

DecimalFloat & DecimalFloat::operator /= (const DecimalFloat & op2)
{
    (*this) = (*this) / op2;
    return *this;
}

DecimalFloat DecimalFloat::operator ++()
{
    (*this) += 1;
    return *this;
}

DecimalFloat DecimalFloat::operator --()
{
    (*this) -= 1;
    return *this;
}

DecimalFloat DecimalFloat::operator ++(int)
{
    DecimalFloat result = (*this);
    (*this) += 1;
    return result;
}

DecimalFloat DecimalFloat::operator --(int)
{
    DecimalFloat result = (*this);
    (*this) -= 1;
    return result;
}

int DecimalFloat::compare(const DecimalFloat & op2) const
{
    DecimalFloat diff = (*this) - op2;
    return diff._mantissa.getSign();
}

bool DecimalFloat::operator == (const DecimalFloat & op2) const
{
    return compare(op2) == 0;
}

bool DecimalFloat::operator != (const DecimalFloat & op2) const
{
    return compare(op2) != 0;
}

bool DecimalFloat::operator < (const DecimalFloat & op2) const
{
    return compare(op2) < 0;
}

bool DecimalFloat::operator <= (const DecimalFloat & op2) const
{
    return compare(op2) <= 0;
}

bool DecimalFloat::operator > (const DecimalFloat & op2) const
{
    return compare(op2) > 0;
}

bool DecimalFloat::operator >= (const DecimalFloat & op2) const
{
    return compare(op2) >= 0;
}

Int8 DecimalFloat::toInt8(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int8_Max)
    {
        return Limits::Int8_Max;
    }
    else if (value64 < Limits::Int8_Min)
    {
        return Limits::Int8_Min;
    }
    else
    {
        return static_cast<Int8>(value64);
    }
}

Int16 DecimalFloat::toInt16(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int16_Max)
    {
        return Limits::Int16_Max;
    }
    else if (value64 < Limits::Int16_Min)
    {
        return Limits::Int16_Min;
    }
    else
    {
        return static_cast<Int16>(value64);
    }
}

Int32 DecimalFloat::toInt32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::Int32_Max)
    {
        return Limits::Int32_Max;
    }
    else if (value64 < Limits::Int32_Min)
    {
        return Limits::Int32_Min;
    }
    else
    {
        return static_cast<Int32>(value64);
    }
}

Int64 DecimalFloat::toInt64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    DecimalInteger value = toDecimalInteger(roundingMode);
    if (value < Limits::Int64_Min)
    {
        return Limits::Int64_Min;
    }
    else if (value > Limits::Int64_Max)
    {
        return Limits::Int64_Max;
    }
    else
    {
        return value.toInt64();
    }
}

UInt8 DecimalFloat::toUInt8(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt8_Max)
    {
        return Limits::UInt8_Max;
    }
    else
    {
        return static_cast<UInt8>(value64);
    }
}

UInt16 DecimalFloat::toUInt16(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt16_Max)
    {
        return Limits::UInt16_Max;
    }
    else
    {
        return static_cast<UInt16>(value64);
    }
}

UInt32 DecimalFloat::toUInt32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UInt32_Max)
    {
        return Limits::UInt32_Max;
    }
    else
    {
        return static_cast<UInt32>(value64);
    }
}

UInt64 DecimalFloat::toUInt64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    DecimalInteger value = toDecimalInteger(roundingMode);
    if (value < Limits::UInt64_Min)
    {
        return Limits::UInt64_Min;
    }
    else if (value > Limits::UInt64_Max)
    {
        return Limits::UInt64_Max;
    }
    else
    {
        return value.toUInt64();
    }
}

UIntSize DecimalFloat::toUIntSize(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UIntSize_Max)
    {
        return Limits::UIntSize_Max;
    }
    else
    {
        return static_cast<UIntSize>(value64);
    }
}

UIntPtr DecimalFloat::toUIntPtr(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    UInt64 value64 = toUInt64(roundingMode);
    if (value64 > Limits::UIntPtr_Max)
    {
        return Limits::UIntPtr_Max;
    }
    else
    {
        return static_cast<UIntPtr>(value64);
    }
}

IntPtrDiff DecimalFloat::toIntPtrDiff(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    Int64 value64 = toInt64(roundingMode);
    if (value64 > Limits::IntPtrDiff_Max)
    {
        return Limits::IntPtrDiff_Max;
    }
    else if (value64 < Limits::IntPtrDiff_Min)
    {
        return Limits::IntPtrDiff_Min;
    }
    else
    {
        return static_cast<IntPtrDiff>(value64);
    }
}

Real32 DecimalFloat::toReal32(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    return static_cast<Real32>(toReal64(roundingMode));
}

Real64 DecimalFloat::toReal64(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    switch (_category)
    {
        case ValueCategory_Normal:
            break;
        case ValueCategory_PositiveInfinity:
            {
                Real64 value;
                *(UInt64*)(&value) = CARL_UINT64_CONSTANT(0x7FF0000000000000);
                return value;
            }
        case ValueCategory_NegativeInfinity:
            {
                Real64 value;
                *(UInt64*)(&value) = CARL_UINT64_CONSTANT(0xFFF0000000000000);
                return value;
            }
        case ValueCategory_SignalingNan:
            throw SignalingNanOperandError();
        case ValueCategory_QuietNan:
            {
                Real64 value;
                *(UInt64*)(&value) = CARL_UINT64_CONSTANT(0x7FF7FFFFFFFFFFFF);
                return value;
            }
        default:
            require(false);
            return 0;
    }
    if (_mantissa.getSign() == -1)
    {   //  Make sure the translated value is positive
        DecimalFloat minusThis = -(*this);
        switch (roundingMode)
        {
            case RoundingMode_TowardsNearest:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsZero:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsPositiveInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsNegativeInfinity));
            case RoundingMode_TowardsNegativeInfinity:
                return -(minusThis.toReal64(RoundingMode_TowardsPositiveInfinity));
            default:
                require(false);
                return 0;
        }
    }
    else
    {
        Real64 addend = 1.0;
        for (DecimalInteger scale = 0; scale < _scale; scale++)
        {
            addend *= 10.0;
        }
        for (DecimalInteger scale = 0; scale > _scale; scale--)
        {
            addend /= 10.0;
        }
        Real64 result = 0;
        for (UIntSize digitNo = 0; digitNo < _mantissa.getRequiredDigits(); digitNo++)
        {
            result += addend * _mantissa.getDigit(digitNo);
            addend *= 10.0;
        }
        return result;
    }
}

DecimalInteger DecimalFloat::toDecimalInteger(RoundingMode roundingMode /*= RoundingMode_TowardsZero*/) const
{
    switch (_category)
    {
        case ValueCategory_Normal:
            break;
        case ValueCategory_PositiveInfinity:
        case ValueCategory_NegativeInfinity:
            throw ValueNotRepresentableAsFixedPointError();
        case ValueCategory_SignalingNan:
            throw SignalingNanOperandError();
        case ValueCategory_QuietNan:
            throw ValueNotRepresentableAsFixedPointError();
        default:
            require(false);
            return 0;
    }
    if (_mantissa.getSign() == -1)
    {   //  Make sure the translated value is positive
        DecimalFloat minusThis = -(*this);
        switch (roundingMode)
        {
            case RoundingMode_TowardsNearest:
                return -(minusThis.toDecimalInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsZero:
                return -(minusThis.toDecimalInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsInfinity:
                return -(minusThis.toDecimalInteger(RoundingMode_TowardsNearest));
            case RoundingMode_TowardsPositiveInfinity:
                return -(minusThis.toDecimalInteger(RoundingMode_TowardsNegativeInfinity));
            case RoundingMode_TowardsNegativeInfinity:
                return -(minusThis.toDecimalInteger(RoundingMode_TowardsPositiveInfinity));
            default:
                require(false);
                return 0;
        }
    }
    else if (_scale == 0)
    {   //  No rounding is necessary
        return _mantissa;
    }
    else if (_scale > 0)
    {   //  No rounding is necessary, but padding IS
        DecimalInteger result = _mantissa;
        require(_scale <= Limits::UIntSize_Max);
        result.multiplyByPowerOf10(_scale.toUIntSize());
        return result;
    }
    else
    {   //  Rounding may be necessary
        DecimalInteger result = _mantissa;
        UIntSize nonzeroDigitsShiftedOut = 0;
        int lastDigitShiftedOut = 0;
        for (DecimalInteger scale = 0; scale > _scale; scale--)
        {
            if ((lastDigitShiftedOut = result.getDigit(0)) != 0)
            {
                nonzeroDigitsShiftedOut++;
            }
            result.divideByPowerOf10(1);
        }
        if (nonzeroDigitsShiftedOut != 0)
        {   //  Rounding is necessary
            switch (roundingMode)
            {
                case RoundingMode_TowardsNearest:
                    if (lastDigitShiftedOut >= 5)
                    {
                        if (nonzeroDigitsShiftedOut == 1)
                        {   //  Only one '5' digit was shifted out
                            if (lastDigitShiftedOut == 5)
                            {   //  The value is exactly between two integers; chose an odd one
                                if (result.getDigit(0) % 2 != 0)
                                {
                                    result++;
                                }
                            }
                            else
                            {
                                result++;
                            }
                        }
                        else
                        {   //  More than one digit was shifted out, and last of then was >=5
                            result++;
                        }
                    }
                    break;
                case RoundingMode_TowardsZero:
                    break;
                case RoundingMode_TowardsInfinity:
                    result++;
                    break;
                case RoundingMode_TowardsPositiveInfinity:
                    result++;
                    break;
                case RoundingMode_TowardsNegativeInfinity:
                    break;
                default:
                    require(false);
            }
        }
        return result;
    }
}

String DecimalFloat::toString(UIntSize radix /*= 10*/) const
{
    RealFormatProperties formatProperties;
    formatProperties.setRadix(radix);
    return NumberFormat::getInstance(Locale())->format(*this, formatProperties);
}

int DecimalFloat::getSign() const
{
    require(_category == ValueCategory_Normal);
    return _mantissa.getSign();
}

DecimalInteger DecimalFloat::getMantissa() const
{
    require(_category == ValueCategory_Normal);

    return (_mantissa.getSign() == -1) ? (-_mantissa) : (_mantissa);
}

DecimalInteger DecimalFloat::getScale() const
{
    require(_category == ValueCategory_Normal);

    return _scale;
}

void DecimalFloat::setPrecision(UIntSize precision, RoundingMode roundingMode /*= RoundingMode_TowardsZero*/)
{
    require(_category == ValueCategory_Normal);

    UIntSize currentPrecision = _mantissa.getRequiredDigits();
    if (precision >= currentPrecision)
    {   //  Scaling increases precision; no need for rounding
        UIntSize digitsToInsert = (precision - currentPrecision);
        _mantissa.multiplyByPowerOf10(digitsToInsert);
        _scale -= digitsToInsert;
        _normalize();
    }
    else
    {
        UIntSize digitsToRemove = (currentPrecision - precision);
        UIntSize nonzeroDigitsShiftedOut = 0;
        int lastDigitShiftedOut = 0;
        for (UIntSize i = 0; i < digitsToRemove; i++)
        {
            if ((lastDigitShiftedOut = _mantissa.getDigit(i)) != 0)
            {
                nonzeroDigitsShiftedOut++;
            }
        }
        _mantissa.divideByPowerOf10(digitsToRemove);
        _scale += digitsToRemove;
        _normalize();
        if (nonzeroDigitsShiftedOut != 0)
        {   //  Rounding is necessary
            switch (roundingMode)
            {
                case RoundingMode_TowardsNearest:
                    if (lastDigitShiftedOut >= 5)
                    {
                        if (nonzeroDigitsShiftedOut == 1)
                        {   //  Only one '5' digit was shifted out
                            if (lastDigitShiftedOut == 5)
                            {   //  The value is exactly between two integers; chose an odd one
                                if (_mantissa.getDigit(0) % 2 != 0)
                                {
                                    _mantissa++;
                                }
                            }
                            else
                            {
                                _mantissa++;
                            }
                        }
                        else
                        {   //  More than one digit was shifted out, and last of then was >=5
                            _mantissa++;
                        }
                    }
                    break;
                case RoundingMode_TowardsZero:
                    break;
                case RoundingMode_TowardsInfinity:
                    _mantissa++;
                    break;
                case RoundingMode_TowardsPositiveInfinity:
                    if (_mantissa.getSign() != -1)
                    {
                        _mantissa++;
                    }
                    break;
                case RoundingMode_TowardsNegativeInfinity:
                    if (_mantissa.getSign() == -1)
                    {
                        _mantissa++;
                    }
                    break;
                default:
                    require(false);
            }
        }
    }
}

DecimalFloat::ValueCategory DecimalFloat::getCategory() const
{
    return _category;
}

bool DecimalFloat::isNormal() const
{
    return _category == ValueCategory_Normal;
}

bool DecimalFloat::isPositiveInfinity() const
{
    return _category == ValueCategory_PositiveInfinity;
}

bool DecimalFloat::isNegativeInfinity() const
{
    return _category == ValueCategory_NegativeInfinity;
}

bool DecimalFloat::isSignalingNan() const
{
    return _category == ValueCategory_SignalingNan;
}

bool DecimalFloat::isQuietNan() const
{
    return _category == ValueCategory_QuietNan;
}

//////////
//  Implementation helpers
CriticalSection * DecimalFloat::_getStaticGuard()
{
    static CriticalSection instance;
    return &instance;
}

void DecimalFloat::_normalize()
{
    if (_category == ValueCategory_Normal)
    {
        //  Count the number of tailing 0's
        UIntSize maxDigitsToPrune = _mantissa.getRequiredDigits();
        UIntSize digitsToPrune;
        for (digitsToPrune = 0; digitsToPrune < maxDigitsToPrune && _mantissa.getDigit(digitsToPrune) == 0; digitsToPrune++)
        {   //  All work is done in the loop header
        }
        if (digitsToPrune != 0)
        {
            _mantissa.divideByPowerOf10(digitsToPrune);
            _scale += digitsToPrune;
        }
        if (_mantissa == 0)
        {
            _scale = 0;
        }
    }
    else
    {
        _mantissa = _scale = 0;
    }
}

void DecimalFloat::_fromReal(Real64 value)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    switch (_fpclass(value))
    {
        case _FPCLASS_SNAN:
            _category = ValueCategory_SignalingNan;
            break;
        case _FPCLASS_QNAN:
            _category = ValueCategory_QuietNan;
            break;
        case _FPCLASS_NINF:
            _category = ValueCategory_NegativeInfinity;
            break;
        case _FPCLASS_PINF:
            _category = ValueCategory_PositiveInfinity;
            break;
        default:
            _category = ValueCategory_Normal;
            break;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    int isInf = isinf(value);
    if (isInf == -1)
    {
        _category = ValueCategory_NegativeInfinity;
    }
    else if (isInf == 1)
    {
        _category = ValueCategory_PositiveInfinity;
    }
    else if (isnan(value))
    {
        _category = ValueCategory_QuietNan;
    }
    else
    {
        _category = ValueCategory_Normal;
    }
#else
    #error Unsupported platform
#endif
    if (_category == ValueCategory_Normal)
    {
        Int64 valueBits = *((Int64*)(&value));
        bool isNegative = ((valueBits >> 63) & 0x01) != 0;
        Int16 exponent = static_cast<Int16>(((valueBits >> 52) & 0x07FF));
        Int64 mantissa = (valueBits & CARL_INT64_CONSTANT(0xFFFFFFFFFFFFF));
        if (exponent != 0)
        {   //  Normalized values have an implicit '1' bit
            mantissa |= CARL_INT64_CONSTANT(0x10000000000000);
        }
        _mantissa = (isNegative) ? (-mantissa) : (mantissa);
        _scale = 0;
        exponent -= 1023 + 52;
        if (exponent > 0)
        {
            DecimalInteger multiplier = 1;
            for (Int16 i = 0; i < exponent; i++)
            {
                multiplier *= 2;
            }
            (*this) *= multiplier;
        }
        else if (exponent < 0)
        {
            DecimalInteger divisor = 1;
            for (Int16 i = 0; i > exponent; i--)
            {
                divisor *= 2;
            }
            (*this) /= divisor;
        }
        _normalize();
    }
}

void DecimalFloat::_fromString(const String & s, UIntSize radix) throw (ParseException)
{
    RealParseProperties parseProperties;
    parseProperties.setRadix(radix);
    DecimalFloat bf = NumberParse::getInstance(Locale())->parseDecimalFloat(s, parseProperties);
    _category = bf._category;
    _mantissa = bf._mantissa;
    _scale = bf._scale;
}

//////////
//  Additional arbitrary-precision operators
DecimalFloat CARL_API carl::operator + (Int8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) + op2;
}

DecimalFloat CARL_API carl::operator + (Int16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) + op2;
}

DecimalFloat CARL_API carl::operator + (Int32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) + op2;
}

DecimalFloat CARL_API carl::operator + (Int64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) + op2;
}

DecimalFloat CARL_API carl::operator + (UInt8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) + op2;
}

DecimalFloat CARL_API carl::operator + (UInt16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) + op2;
}

DecimalFloat CARL_API carl::operator + (UInt32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) + op2;
}

DecimalFloat CARL_API carl::operator + (UInt64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) + op2;
}

DecimalFloat CARL_API carl::operator + (Real32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) + op2;
}

DecimalFloat CARL_API carl::operator + (Real64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) + op2;
}

DecimalFloat CARL_API carl::operator - (Int8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) - op2;
}

DecimalFloat CARL_API carl::operator - (Int16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) - op2;
}

DecimalFloat CARL_API carl::operator - (Int32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) - op2;
}

DecimalFloat CARL_API carl::operator - (Int64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) - op2;
}

DecimalFloat CARL_API carl::operator - (UInt8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) - op2;
}

DecimalFloat CARL_API carl::operator - (UInt16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) - op2;
}

DecimalFloat CARL_API carl::operator - (UInt32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) - op2;
}

DecimalFloat CARL_API carl::operator - (UInt64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) - op2;
}

DecimalFloat CARL_API carl::operator - (Real32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) - op2;
}

DecimalFloat CARL_API carl::operator - (Real64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) - op2;
}

DecimalFloat CARL_API carl::operator * (Int8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) * op2;
}

DecimalFloat CARL_API carl::operator * (Int16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) * op2;
}

DecimalFloat CARL_API carl::operator * (Int32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) * op2;
}

DecimalFloat CARL_API carl::operator * (Int64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) * op2;
}

DecimalFloat CARL_API carl::operator * (UInt8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) * op2;
}

DecimalFloat CARL_API carl::operator * (UInt16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) * op2;
}

DecimalFloat CARL_API carl::operator * (UInt32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) * op2;
}

DecimalFloat CARL_API carl::operator * (UInt64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) * op2;
}

DecimalFloat CARL_API carl::operator * (Real32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) * op2;
}

DecimalFloat CARL_API carl::operator * (Real64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) * op2;
}

DecimalFloat CARL_API carl::operator / (Int8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) / op2;
}

DecimalFloat CARL_API carl::operator / (Int16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) / op2;
}

DecimalFloat CARL_API carl::operator / (Int32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) / op2;
}

DecimalFloat CARL_API carl::operator / (Int64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) / op2;
}

DecimalFloat CARL_API carl::operator / (UInt8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) / op2;
}

DecimalFloat CARL_API carl::operator / (UInt16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) / op2;
}

DecimalFloat CARL_API carl::operator / (UInt32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) / op2;
}

DecimalFloat CARL_API carl::operator / (UInt64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) / op2;
}

DecimalFloat CARL_API carl::operator / (Real32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) / op2;
}

DecimalFloat CARL_API carl::operator / (Real64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) / op2;
}

bool CARL_API carl::operator == (Int8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) == op2;
}

bool CARL_API carl::operator == (Int16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) == op2;
}

bool CARL_API carl::operator == (Int32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) == op2;
}

bool CARL_API carl::operator == (Int64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) == op2;
}

bool CARL_API carl::operator == (UInt8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) == op2;
}

bool CARL_API carl::operator == (UInt16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) == op2;
}

bool CARL_API carl::operator == (UInt32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) == op2;
}

bool CARL_API carl::operator == (UInt64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) == op2;
}

bool CARL_API carl::operator == (Real32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) == op2;
}

bool CARL_API carl::operator == (Real64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) == op2;
}

bool CARL_API carl::operator != (Int8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) != op2;
}

bool CARL_API carl::operator != (Int16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) != op2;
}

bool CARL_API carl::operator != (Int32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) != op2;
}

bool CARL_API carl::operator != (Int64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) != op2;
}

bool CARL_API carl::operator != (UInt8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) != op2;
}

bool CARL_API carl::operator != (UInt16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) != op2;
}

bool CARL_API carl::operator != (UInt32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) != op2;
}

bool CARL_API carl::operator != (UInt64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) != op2;
}

bool CARL_API carl::operator != (Real32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) != op2;
}

bool CARL_API carl::operator != (Real64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) != op2;
}

bool CARL_API carl::operator < (Int8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) < op2;
}

bool CARL_API carl::operator < (Int16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) < op2;
}

bool CARL_API carl::operator < (Int32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) < op2;
}

bool CARL_API carl::operator < (Int64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) < op2;
}

bool CARL_API carl::operator < (UInt8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) < op2;
}

bool CARL_API carl::operator < (UInt16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) < op2;
}

bool CARL_API carl::operator < (UInt32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) < op2;
}

bool CARL_API carl::operator < (UInt64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) < op2;
}

bool CARL_API carl::operator < (Real32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) < op2;
}

bool CARL_API carl::operator < (Real64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) < op2;
}

bool CARL_API carl::operator <= (Int8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (Int16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (Int32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (Int64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (Real32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) <= op2;
}

bool CARL_API carl::operator <= (Real64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) <= op2;
}

bool CARL_API carl::operator > (Int8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) > op2;
}

bool CARL_API carl::operator > (Int16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) > op2;
}

bool CARL_API carl::operator > (Int32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) > op2;
}

bool CARL_API carl::operator > (Int64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) > op2;
}

bool CARL_API carl::operator > (UInt8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) > op2;
}

bool CARL_API carl::operator > (UInt16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) > op2;
}

bool CARL_API carl::operator > (UInt32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) > op2;
}

bool CARL_API carl::operator > (UInt64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) > op2;
}

bool CARL_API carl::operator > (Real32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) > op2;
}

bool CARL_API carl::operator > (Real64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) > op2;
}

bool CARL_API carl::operator >= (Int8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (Int16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (Int32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (Int64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt8 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt16 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (Real32 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) >= op2;
}

bool CARL_API carl::operator >= (Real64 op1, const DecimalFloat & op2)
{
    return DecimalFloat(op1) >= op2;
}

//////////
//  Hashing arbitrary-precision values
UIntSize CARL_CALLBACK StandardHashFunctions::hashDecimalFloat(const DecimalFloat & key)
{
    if (key.isNormal())
    {
        return hashDecimalInteger(key.getMantissa());
    }
    else
    {
        return 0;
    }
}

//  End of DecimalFloat.cpp
