//
//  DecimalInteger.cpp - Implementation of the carl::DecimalInteger class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/08/07 13:51:37 $
//  $Revision: 1.7 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

//////////
//  Construction/destruction/assignment
DecimalInteger::DecimalInteger(bool isNegative, UInt32 * pDigits, UIntSize wordsAllocated, UIntSize wordsUsed, bool takeOwnershipOfDigits)
{
    require(pDigits != NULL);
    require(wordsAllocated > 0);
    require(wordsUsed > 0);
    require(wordsAllocated >= wordsUsed);

    _isNegative = isNegative;
    if (takeOwnershipOfDigits)
    {
        _pDigits = pDigits;
    }
    else
    {
        _pDigits = new UInt32[wordsAllocated];
        for (UIntSize i = 0; i < wordsUsed; i++)
        {
            _pDigits[i] = pDigits[i];
        }
    }
    _wordsAllocated = wordsAllocated;
    _wordsUsed = wordsUsed;
    _normalize();
}

DecimalInteger::DecimalInteger()
{
    _isNegative = false;
    _pDigits = new UInt32[1];
    _pDigits[0] = 0;
    _wordsAllocated = 1;
    _wordsUsed = 1;
}

DecimalInteger::DecimalInteger(Int8 value)
{
    if (value >= 0)
    {
        _fromUInt64(static_cast<UInt8>(value));
    }
    else
    {
        _fromUInt64(static_cast<UInt8>(-value));
        _isNegative = true;
    }
}

DecimalInteger::DecimalInteger(Int16 value)
{
    if (value >= 0)
    {
        _fromUInt64(static_cast<UInt16>(value));
    }
    else
    {
        _fromUInt64(static_cast<UInt16>(-value));
        _isNegative = true;
    }
}

DecimalInteger::DecimalInteger(Int32 value)
{
    if (value >= 0)
    {
        _fromUInt64(static_cast<UInt32>(value));
    }
    else
    {
        _fromUInt64(static_cast<UInt32>(-value));
        _isNegative = true;
    }
}

DecimalInteger::DecimalInteger(Int64 value)
{
    if (value >= 0)
    {
        _fromUInt64(static_cast<UInt64>(value));
    }
    else
    {
        _fromUInt64(static_cast<UInt64>(-value));
        _isNegative = true;
    }
}

DecimalInteger::DecimalInteger(UInt8 value)
{
    _fromUInt64(value);
}

DecimalInteger::DecimalInteger(UInt16 value)
{
    _fromUInt64(value);
}

DecimalInteger::DecimalInteger(UInt32 value)
{
    _fromUInt64(value);
}

DecimalInteger::DecimalInteger(UInt64 value)
{
    _fromUInt64(value);
}

DecimalInteger::DecimalInteger(const char * pValueString) throw (ParseException)
{
    require(pValueString != NULL);

    _fromString(String(pValueString), 10);
}

DecimalInteger::DecimalInteger(const Char * pValueString) throw (ParseException)
{
    require(pValueString != NULL);

    _fromString(String(pValueString), 10);
}

DecimalInteger::DecimalInteger(const String & valueString) throw (ParseException)
{
    _fromString(valueString, 10);
}

DecimalInteger::DecimalInteger(const char * pValueString, UIntSize radix) throw (ParseException)
{
    require(pValueString != NULL);
    require(radix >= 2 && radix <= 36);

    _fromString(String(pValueString), radix);
}

DecimalInteger::DecimalInteger(const Char * pValueString, UIntSize radix) throw (ParseException)
{
    require(pValueString != NULL);
    require(radix >= 2 && radix <= 36);

    _fromString(String(pValueString), radix);
}

DecimalInteger::DecimalInteger(const String & valueString, UIntSize radix) throw (ParseException)
{
    require(radix >= 2 && radix <= 36);

    _fromString(valueString, radix);
}

DecimalInteger::DecimalInteger(const DecimalInteger & src)
{
    _isNegative = src._isNegative;
    _wordsAllocated = _wordsUsed = src._wordsUsed;
    _pDigits = new UInt32[_wordsUsed];
    for (UIntSize i = 0; i < _wordsUsed; i++)
    {
        _pDigits[i] = src._pDigits[i];
    }
}

DecimalInteger::~DecimalInteger()
{
    delete [] _pDigits;
}

DecimalInteger & DecimalInteger::operator = (const DecimalInteger & src)
{
    if (this != (&src))
    {
        _isNegative = src._isNegative;
        _ensureAllocation(src._wordsUsed);
        _wordsUsed = src._wordsUsed;
        for (UIntSize i = 0; i < _wordsUsed; i++)
        {
            _pDigits[i] = src._pDigits[i];
        }
    }
    return *this;
}

//////////
//  Operations
DecimalInteger DecimalInteger::operator + () const
{
    return *this;
}

DecimalInteger DecimalInteger::operator - () const
{
    DecimalInteger result = (*this);
    result._isNegative = !result._isNegative;
    result._normalize();
    return result;
}

DecimalInteger DecimalInteger::operator + (const DecimalInteger & op2) const
{
    if (!_isNegative && _wordsUsed == 0 && _pDigits[0] == 0)
    {   //  0 + op2 == op2
        return op2;
    }
    else if (!op2._isNegative && op2._wordsUsed == 0 && op2._pDigits[0] == 0)
    {   //  (*this) + 0 == (*this)
        return *this;
    }
    else if (_isNegative == op2._isNegative)
    {   //  Both operands have the same sign - add magnitudes
        UIntSize resultLength = Math::max(_wordsUsed, op2._wordsUsed) + 1;
        UInt32 * pResultDigits = new UInt32[resultLength];
        _addWords(pResultDigits, resultLength, _pDigits, _wordsUsed, op2._pDigits, op2._wordsUsed);
        return DecimalInteger(_isNegative, pResultDigits, resultLength, resultLength, true);
    }
    else
    {   //  Operands have different signs - subtract magnitudes
        int magnitudeOrder = _compareWords(_pDigits, _wordsUsed, op2._pDigits, op2._wordsUsed);
        if (magnitudeOrder == 0)
        {   //  A + (-A) == 0
            return DecimalInteger();
        }
        else if (magnitudeOrder > 0)
        {   //  Magnitude of (*this) > magnitude of op2; result has the sign of (*this)
            UIntSize resultLength = Math::max(_wordsUsed, op2._wordsUsed);
            UInt32 * pResultDigits = new UInt32[resultLength];
            _subtractWords(pResultDigits, resultLength, _pDigits, _wordsUsed, op2._pDigits, op2._wordsUsed);
            return DecimalInteger(_isNegative, pResultDigits, resultLength, resultLength, true);
        }
        else
        {   //  Magnitude of (*this) < magnitude of op2; result has the sign of op2
            UIntSize resultLength = Math::max(_wordsUsed, op2._wordsUsed);
            UInt32 * pResultDigits = new UInt32[resultLength];
            _subtractWords(pResultDigits, resultLength, op2._pDigits, op2._wordsUsed, _pDigits, _wordsUsed);
            return DecimalInteger(op2._isNegative, pResultDigits, resultLength, resultLength, true);
        }
    }
}

DecimalInteger DecimalInteger::operator - (const DecimalInteger & op2) const
{
    if (!_isNegative && _wordsUsed == 0 && _pDigits[0] == 0)
    {   //  0 - op2 == -op2
        return -op2;
    }
    else if (!op2._isNegative && op2._wordsUsed == 0 && op2._pDigits[0] == 0)
    {   //  (*this) - 0 == (*this)
        return *this;
    }
    else if (_isNegative != op2._isNegative)
    {   //  Operands have different signs - add magnitudes; result has 1st operand's sign
        UIntSize resultLength = Math::max(_wordsUsed, op2._wordsUsed) + 1;
        UInt32 * pResultDigits = new UInt32[resultLength];
        _addWords(pResultDigits, resultLength, _pDigits, _wordsUsed, op2._pDigits, op2._wordsUsed);
        return DecimalInteger(_isNegative, pResultDigits, resultLength, resultLength, true);
    }
    else
    {   //  Operands have the same sign - subtract magnitudes
        int magnitudeOrder = _compareWords(_pDigits, _wordsUsed, op2._pDigits, op2._wordsUsed);
        if (magnitudeOrder == 0)
        {   //  A - A == 0
            return DecimalInteger();
        }
        else if (magnitudeOrder > 0)
        {   //  Magnitude of (*this) > magnitude of op2; result has the sign of (*this)
            UIntSize resultLength = Math::max(_wordsUsed, op2._wordsUsed);
            UInt32 * pResultDigits = new UInt32[resultLength];
            _subtractWords(pResultDigits, resultLength, _pDigits, _wordsUsed, op2._pDigits, op2._wordsUsed);
            return DecimalInteger(_isNegative, pResultDigits, resultLength, resultLength, true);
        }
        else
        {   //  Magnitude of (*this) < magnitude of op2; result has the reverse sign of (*this)
            UIntSize resultLength = Math::max(_wordsUsed, op2._wordsUsed);
            UInt32 * pResultDigits = new UInt32[resultLength];
            _subtractWords(pResultDigits, resultLength, op2._pDigits, op2._wordsUsed, _pDigits, _wordsUsed);
            return DecimalInteger(!_isNegative, pResultDigits, resultLength, resultLength, true);
        }
    }
}

DecimalInteger DecimalInteger::operator * (const DecimalInteger & op2) const
{
    UInt32 resultWordsUsed = _wordsUsed + op2._wordsUsed;
    UInt32 * pResultWords = new UInt32[resultWordsUsed];
    _clearWords(pResultWords, resultWordsUsed);
    UIntSize numDigits1 = getRequiredDigits();
    UIntSize numDigits2 = op2.getRequiredDigits();
    for (UInt32 i = 0; i < numDigits1; i++)
    {
        for (UInt32 j = 0; j < numDigits2; j++)
        {
            int digit1 = getDigit(i);
            int digit2 = op2.getDigit(j);
            int partialProduct = digit1 * digit2;
            _addDigitInPlace(pResultWords, resultWordsUsed, partialProduct % 10, i + j);
            _addDigitInPlace(pResultWords, resultWordsUsed, partialProduct / 10, i + j + 1);
        }
    }
    return DecimalInteger(_isNegative != op2._isNegative, pResultWords, resultWordsUsed, resultWordsUsed, true);
}

DecimalInteger DecimalInteger::operator / (const DecimalInteger & op2) const
{
    if (op2._wordsUsed == 1 && op2._pDigits[0] == 0)
    {
        throw FixedPointDivisionByZeroError();
    }
    else if (_wordsUsed == 1 && _pDigits[0] == 0)
    {
        return DecimalInteger();
    }
    else if (_isNegative && op2._isNegative)
    {
        return (-(*this)) / (-op2);
    }
    else if (_isNegative && !op2._isNegative)
    {
        return -((-(*this)) / (op2));
    }
    else if (!_isNegative && op2._isNegative)
    {
        return -((*this) / (-op2));
    }
    else
    {   //  Both operands positive and non-zero
        UIntSize op1SignificantDigits = getRequiredDigits();
        UIntSize op2SignificantDigits = op2.getRequiredDigits();
        if (op1SignificantDigits < op2SignificantDigits)
        {   //  1st operand is at least an order of magnitude smaller than 2nd
            return DecimalInteger();
        }
        UIntSize resultSignificantDigits = op1SignificantDigits - op2SignificantDigits + 1;
        DecimalInteger divisor = (*this);
        DecimalInteger dividend = op2.multipliedByPowerOf10(resultSignificantDigits - 1);
        DecimalInteger result = 0;
        for (UIntSize i = 0; i < resultSignificantDigits; i++)
        {   //  Obtain one more digit of the result, which is [0..9]
            int digit = 0;
            for (; digit < 10; digit++)
            {
                if (divisor < dividend)
                {
                    break;
                }
                divisor -= dividend;
            }
            require(digit != 10);
            //  Store that one digit
            result = result.multipliedByPowerOf10(1) + digit;
            dividend.divideByPowerOf10(1);
        }
        return result;
    }
}

DecimalInteger DecimalInteger::operator % (const DecimalInteger & op2) const
{
    return (*this) - ((*this) / op2) * op2;
}

DecimalInteger & DecimalInteger::operator += (const DecimalInteger & op2)
{
    (*this) = (*this) + op2;
    return *this;
}

DecimalInteger & DecimalInteger::operator -= (const DecimalInteger & op2)
{
    (*this) = (*this) - op2;
    return *this;
}

DecimalInteger & DecimalInteger::operator *= (const DecimalInteger & op2)
{
    (*this) = (*this) * op2;
    return *this;
}

DecimalInteger & DecimalInteger::operator /= (const DecimalInteger & op2)
{
    (*this) = (*this) / op2;
    return *this;
}

DecimalInteger & DecimalInteger::operator %= (const DecimalInteger & op2)
{
    (*this) = (*this) % op2;
    return *this;
}

DecimalInteger DecimalInteger::operator ++()
{
    (*this) = (*this) + 1;
    return *this;
}

DecimalInteger DecimalInteger::operator --()
{
    (*this) = (*this) - 1;
    return *this;
}

DecimalInteger DecimalInteger::operator ++(int)
{
    DecimalInteger result = (*this);
    (*this) = (*this) + 1;
    return result;
}

DecimalInteger DecimalInteger::operator --(int)
{
    DecimalInteger result = (*this);
    (*this) = (*this) - 1;
    return result;
}

int DecimalInteger::compare(const DecimalInteger & op2) const
{
    return ((*this) - op2).getSign();
}

bool DecimalInteger::operator == (const DecimalInteger & op2) const
{
    return compare(op2) == 0;
}

bool DecimalInteger::operator != (const DecimalInteger & op2) const
{
    return compare(op2) != 0;
}

bool DecimalInteger::operator < (const DecimalInteger & op2) const
{
    return compare(op2) < 0;
}

bool DecimalInteger::operator <= (const DecimalInteger & op2) const
{
    return compare(op2) <= 0;
}

bool DecimalInteger::operator > (const DecimalInteger & op2) const
{
    return compare(op2) > 0;
}

bool DecimalInteger::operator >= (const DecimalInteger & op2) const
{
    return compare(op2) >= 0;
}

int DecimalInteger::getSign() const
{
    if (_isNegative)
    {
        return -1;
    }
    else if (_wordsUsed == 1 && _pDigits[0] == 0)
    {
        return 0;
    }
    else
    {
        return 1;
    }
}

UIntSize DecimalInteger::getRequiredDigits() const
{
    UIntSize result = _wordsUsed * 8;
    for (int i = 0; i < 7; i++)
    {
        if (getDigit(result - 1) == 0)
        {
            result--;
        }
    }
    return (result < 1) ? 1 : result;
}

int DecimalInteger::operator [] (UIntSize position) const
{
    return getDigit(position);
}

int DecimalInteger::getDigit(UIntSize position) const
{
    UIntSize wordIndex = position / 8;
    if (wordIndex >= _wordsUsed)
    {
        return 0;
    }
    else
    {
        UIntSize bitOffset = (position % 8) * 4;
        return (_pDigits[wordIndex] >> bitOffset) & 0x0F;
    }
}

void DecimalInteger::setDigit(UIntSize position, int digit)
{
    require(digit >= 0 && digit <= 9);

    UIntSize wordIndex = position / 8;
    UIntSize bitOffset = (position % 8) * 4;
    _ensureUsage(wordIndex + 1);
    _pDigits[wordIndex] = (_pDigits[wordIndex] & (~(0x0F << bitOffset))) | (digit << bitOffset);
    _normalize();
}

Int8 DecimalInteger::toInt8() const
{
    return static_cast<Int8>(toInt16());
}

Int16 DecimalInteger::toInt16() const
{
    return static_cast<Int16>(toInt32());
}

Int32 DecimalInteger::toInt32() const
{
    return static_cast<Int32>(toInt64());
}

Int64 DecimalInteger::toInt64() const
{
    return _toBinaryInteger().toInt64();
}

UInt8 DecimalInteger::toUInt8() const
{
    return static_cast<UInt8>(toUInt16());
}

UInt16 DecimalInteger::toUInt16() const
{
    return static_cast<UInt16>(toUInt32());
}

UInt32 DecimalInteger::toUInt32() const
{
    return static_cast<UInt32>(toUInt64());
}

UInt64 DecimalInteger::toUInt64() const
{
    return _toBinaryInteger().toUInt64();
}

UIntSize DecimalInteger::toUIntSize() const
{
    return static_cast<UIntSize>(toUInt64());
}

UIntPtr DecimalInteger::toUIntPtr() const
{
    return static_cast<UIntPtr>(toUInt64());
}

IntPtrDiff DecimalInteger::toIntPtrDiff() const
{
    return static_cast<IntPtrDiff>(toInt64());
}

String DecimalInteger::toString(UIntSize radix /*= 10*/) const
{
    IntegerFormatProperties formatProperties;
    formatProperties.setRadix(radix);
    return NumberFormat::getInstance(Locale())->format(*this, formatProperties);
}

DecimalInteger DecimalInteger::multipliedByPowerOf10(UIntSize powerOf10) const
{
    UIntSize significantDigits = getRequiredDigits();
    DecimalInteger result;
    for (UIntSize i = 0; i < significantDigits; i++)
    {
        result.setDigit(i + powerOf10, getDigit(i));
    }
    result._isNegative = _isNegative;
    return result;
}

DecimalInteger DecimalInteger::dividedByPowerOf10(UIntSize powerOf10) const
{
    UIntSize significantDigits = getRequiredDigits();
    if (powerOf10 >= significantDigits)
    {
        return DecimalInteger();
    }
    //  General case
    DecimalInteger result;
    for (UIntSize i = 0; i < significantDigits - powerOf10; i++)
    {
        result.setDigit(i, getDigit(i + powerOf10));
    }
    result._isNegative = _isNegative;
    return result;
}

DecimalInteger & DecimalInteger::multiplyByPowerOf10(UIntSize powerOf10)
{
    (*this) = this->multipliedByPowerOf10(powerOf10);
    return *this;
}

DecimalInteger & DecimalInteger::divideByPowerOf10(UIntSize powerOf10)
{
    (*this) = this->dividedByPowerOf10(powerOf10);
    return *this;
}

//////////
//  Implementation helpers
void DecimalInteger::_ensureAllocation(UIntSize allocation)
{
    if (allocation > _wordsAllocated)
    {
        UInt32 * pNewDigits = new UInt32[allocation];
        for (UIntSize i = 0; i < _wordsUsed; i++)
        {
            pNewDigits[i] = _pDigits[i];
        }
        delete [] _pDigits;
        _pDigits = pNewDigits;
        _wordsAllocated = allocation;
    }
}

void DecimalInteger::_ensureUsage(UIntSize usage)
{
    _ensureAllocation(usage);
    while (_wordsUsed < usage)
    {
        _pDigits[_wordsUsed++] = 0;
    }
}

void DecimalInteger::_normalize()
{
    while (_wordsUsed > 1 && _pDigits[_wordsUsed - 1] == 0)
    {
        _wordsUsed--;
    }
    if (_wordsUsed == 1 && _pDigits[0] == 0)
    {
        _isNegative = false;
    }
}

void DecimalInteger::_fromUInt64(UInt64 value)
{
    _isNegative = false;
    _pDigits = new UInt32[3];
    _pDigits[0] = _pDigits[1] = _pDigits[2] = 0;
    _wordsAllocated = 3;
    _wordsUsed = 3;
    for (int i = 0; i < 24; i++)
    {
        UInt32 digit = static_cast<UInt32>(value % 10);
        value /= 10;
        int wordOffset = (i / 8);
        int bitOffset = (i % 8) * 4;
        _pDigits[wordOffset] |= (digit << bitOffset);
    }
    _normalize();
}

void DecimalInteger::_fromString(const String & s, UIntSize radix) throw (ParseException)
{
    IntegerParseProperties parseProperties;
    parseProperties.setRadix(radix);
    DecimalInteger di = NumberParse::getInstance(Locale())->parseDecimalInteger(s, parseProperties);
    _isNegative = di._isNegative;
    _pDigits = new UInt32[di._wordsUsed];
    _wordsAllocated = _wordsUsed = di._wordsUsed;
    for (UIntSize i = 0; i < di._wordsUsed; i++)
    {
        _pDigits[i] = di._pDigits[i];
    }
}

BinaryInteger DecimalInteger::_toBinaryInteger() const
{
    BinaryInteger result;
    BinaryInteger scale(1);
    UIntSize significantDigits = getRequiredDigits();
    for (UIntSize i = 0; i < significantDigits; i++)
    {
        int digit = getDigit(i);
        result += digit * scale;
        scale *= 10;
        if (scale > Limits::UInt64_Max)
        {
            break;
        }
    }
    return (_isNegative) ? (-result) : (result);
}

int DecimalInteger::_compareWords(const UInt32 * pOp1, UIntSize op1Length, const UInt32 * pOp2, UIntSize op2Length)
{
    require(pOp1 != NULL);
    require(op1Length != 0);
    require(pOp2 != NULL);
    require(op2Length != 0);

    for (UIntSize i = Math::max(op1Length, op2Length); i != 0; i--)
    {
        UInt32 op1Word = (i <= op1Length) ? pOp1[i - 1] : 0;
        UInt32 op2Word = (i <= op2Length) ? pOp2[i - 1] : 0;
        if (op1Word < op2Word)
        {
            return -1;
        }
        if (op1Word > op2Word)
        {
            return 1;
        }
    }
    return 0;
}

UInt32 DecimalInteger::_addBcdWords(UInt32 w1, UInt32 w2, UInt32 & carry)
{
    require(carry == 0 || carry == 1);

    UInt32 result = 0;
    for (int bitOffset = 0; bitOffset < 32; bitOffset += 4)
    {
        UInt32 digit1 = (w1 >> bitOffset) & 0x0F;
        UInt32 digit2 = (w2 >> bitOffset) & 0x0F;
        UInt32 resultDigits = digit1 + digit2 + carry;
        carry = resultDigits / 10;
        result |= ((resultDigits % 10) << bitOffset);
    }
    return result;
}

void DecimalInteger::_addWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, const UInt32 * pOp2, UIntSize op2Length)
{
    require(pDest != NULL);
    require(destLength != 0 && destLength >= op1Length && destLength >= op2Length);
    require(pOp1 != NULL);
    require(op1Length != 0);
    require(pOp2 != NULL);
    require(op2Length != 0);

    UInt32 carry = 0;
    for (UIntSize i = 0; i < destLength; i++)
    {
        UInt32 op1Word = (i < op1Length) ? pOp1[i] : 0;
        UInt32 op2Word = (i < op2Length) ? pOp2[i] : 0;
        pDest[i] = _addBcdWords(op1Word, op2Word, carry);
    }
}

UInt32 DecimalInteger::_subtractBcdWords(UInt32 w1, UInt32 w2, UInt32 & borrow)
{
    require(borrow == 0 || borrow == 1);

    UInt32 result = 0;
    for (int bitOffset = 0; bitOffset < 32; bitOffset += 4)
    {
        Int32 digit1 = (w1 >> bitOffset) & 0x0F;
        Int32 digit2 = (w2 >> bitOffset) & 0x0F;
        Int32 resultDigits = digit1 - digit2 - borrow;
        if (resultDigits < 0)
        {
            borrow = 1;
            result |= ((resultDigits + 10) << bitOffset);
        }
        else
        {
            borrow = 0;
            result |= ((resultDigits) << bitOffset);
        }
    }
    return result;
}

void DecimalInteger::_subtractWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, const UInt32 * pOp2, UIntSize op2Length)
{
    require(pDest != NULL);
    require(destLength != 0 && destLength >= op1Length && destLength >= op2Length);
    require(pOp1 != NULL);
    require(op1Length != 0);
    require(pOp2 != NULL);
    require(op2Length != 0);

    UInt32 borrow = 0;
    for (UIntSize i = 0; i < destLength; i++)
    {
        UInt32 op1Word = (i < op1Length) ? pOp1[i] : 0;
        UInt32 op2Word = (i < op2Length) ? pOp2[i] : 0;
        pDest[i] = _subtractBcdWords(op1Word, op2Word, borrow);
    }
}

void DecimalInteger::_clearWords(UInt32 * pDest, UIntSize destLength)
{
    for (UIntSize i = 0; i < destLength; i++)
    {
        pDest[i] = 0;
    }
}

void DecimalInteger::_addDigitInPlace(UInt32 * pDest, UIntSize destLength, int digit, UIntSize digitPosition)
{
    require(pDest != NULL);
    require(destLength != 0);
    require(digitPosition < destLength * 8);
    require(digit >= 0 && digit <= 9);

    if (digit == 0)
    {
        return;
    }
    UInt32 carry = 0;
    for (; digitPosition < destLength * 8; digitPosition++)
    {
        UIntSize wordIndex = digitPosition / 8;
        UIntSize bitOffset = (digitPosition % 8) * 4;
        UInt32 oldDigit = (pDest[wordIndex] >> bitOffset) & 0x0F;
        UInt32 newDigit = oldDigit + digit + carry;
        carry = (newDigit / 10);
        pDest[wordIndex] = (pDest[wordIndex] & (~(0x0F << bitOffset))) | ((newDigit % 10) << bitOffset);
        if (carry == 0)
        {
            break;
        }
        digit = 0;
    }
}

//////////
//  Additional arbitrary-precision operators
DecimalInteger CARL_API carl::operator + (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) + op2;
}

DecimalInteger CARL_API carl::operator + (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) + op2;
}

DecimalInteger CARL_API carl::operator + (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) + op2;
}

DecimalInteger CARL_API carl::operator + (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) + op2;
}

DecimalInteger CARL_API carl::operator + (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) + op2;
}

DecimalInteger CARL_API carl::operator + (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) + op2;
}

DecimalInteger CARL_API carl::operator + (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) + op2;
}

DecimalInteger CARL_API carl::operator + (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) + op2;
}

DecimalInteger CARL_API carl::operator - (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) - op2;
}

DecimalInteger CARL_API carl::operator - (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) - op2;
}

DecimalInteger CARL_API carl::operator - (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) - op2;
}

DecimalInteger CARL_API carl::operator - (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) - op2;
}

DecimalInteger CARL_API carl::operator - (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) - op2;
}

DecimalInteger CARL_API carl::operator - (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) - op2;
}

DecimalInteger CARL_API carl::operator - (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) - op2;
}

DecimalInteger CARL_API carl::operator - (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) - op2;
}

DecimalInteger CARL_API carl::operator * (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) * op2;
}

DecimalInteger CARL_API carl::operator * (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) * op2;
}

DecimalInteger CARL_API carl::operator * (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) * op2;
}

DecimalInteger CARL_API carl::operator * (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) * op2;
}

DecimalInteger CARL_API carl::operator * (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) * op2;
}

DecimalInteger CARL_API carl::operator * (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) * op2;
}

DecimalInteger CARL_API carl::operator * (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) * op2;
}

DecimalInteger CARL_API carl::operator * (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) * op2;
}

DecimalInteger CARL_API carl::operator / (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) / op2;
}

DecimalInteger CARL_API carl::operator / (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) / op2;
}

DecimalInteger CARL_API carl::operator / (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) / op2;
}

DecimalInteger CARL_API carl::operator / (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) / op2;
}

DecimalInteger CARL_API carl::operator / (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) / op2;
}

DecimalInteger CARL_API carl::operator / (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) / op2;
}

DecimalInteger CARL_API carl::operator / (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) / op2;
}

DecimalInteger CARL_API carl::operator / (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) / op2;
}

DecimalInteger CARL_API carl::operator % (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) % op2;
}

DecimalInteger CARL_API carl::operator % (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) % op2;
}

DecimalInteger CARL_API carl::operator % (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) % op2;
}

DecimalInteger CARL_API carl::operator % (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) % op2;
}

DecimalInteger CARL_API carl::operator % (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) % op2;
}

DecimalInteger CARL_API carl::operator % (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) % op2;
}

DecimalInteger CARL_API carl::operator % (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) % op2;
}

DecimalInteger CARL_API carl::operator % (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) % op2;
}

bool CARL_API carl::operator == (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) == op2;
}

bool CARL_API carl::operator == (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) == op2;
}

bool CARL_API carl::operator == (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) == op2;
}

bool CARL_API carl::operator == (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) == op2;
}

bool CARL_API carl::operator == (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) == op2;
}

bool CARL_API carl::operator == (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) == op2;
}

bool CARL_API carl::operator == (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) == op2;
}

bool CARL_API carl::operator == (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) == op2;
}

bool CARL_API carl::operator != (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) != op2;
}

bool CARL_API carl::operator != (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) != op2;
}

bool CARL_API carl::operator != (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) != op2;
}

bool CARL_API carl::operator != (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) != op2;
}

bool CARL_API carl::operator != (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) != op2;
}

bool CARL_API carl::operator != (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) != op2;
}

bool CARL_API carl::operator != (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) != op2;
}

bool CARL_API carl::operator != (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) != op2;
}

bool CARL_API carl::operator < (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) < op2;
}

bool CARL_API carl::operator < (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) < op2;
}

bool CARL_API carl::operator < (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) < op2;
}

bool CARL_API carl::operator < (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) < op2;
}

bool CARL_API carl::operator < (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) < op2;
}

bool CARL_API carl::operator < (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) < op2;
}

bool CARL_API carl::operator < (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) < op2;
}

bool CARL_API carl::operator < (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) < op2;
}

bool CARL_API carl::operator <= (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) <= op2;
}

bool CARL_API carl::operator <= (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) <= op2;
}

bool CARL_API carl::operator <= (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) <= op2;
}

bool CARL_API carl::operator <= (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) <= op2;
}

bool CARL_API carl::operator <= (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) <= op2;
}

bool CARL_API carl::operator > (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) > op2;
}

bool CARL_API carl::operator > (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) > op2;
}

bool CARL_API carl::operator > (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) > op2;
}

bool CARL_API carl::operator > (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) > op2;
}

bool CARL_API carl::operator > (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) > op2;
}

bool CARL_API carl::operator > (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) > op2;
}

bool CARL_API carl::operator > (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) > op2;
}

bool CARL_API carl::operator > (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) > op2;
}

bool CARL_API carl::operator >= (Int8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) >= op2;
}

bool CARL_API carl::operator >= (Int16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) >= op2;
}

bool CARL_API carl::operator >= (Int32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) >= op2;
}

bool CARL_API carl::operator >= (Int64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt8 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt16 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt32 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) >= op2;
}

bool CARL_API carl::operator >= (UInt64 op1, const DecimalInteger & op2)
{
    return DecimalInteger(op1) >= op2;
}

//////////
//  Hashing arbitrary-precision values
UIntSize CARL_CALLBACK StandardHashFunctions::hashDecimalInteger(const DecimalInteger & key)
{
    UIntSize result = 0;
    for (UIntSize k = 0; k < key.getRequiredDigits(); k++)
    {
        result = 5 * result + key.getDigit(k);
    }
    return result;
}

//  End of DecimalInteger.cpp
