//
//  DecompressInputStream.cpp - Implementation of the carl::DecompressInputStream namespace
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:52 $
//  $Revision: 1.6 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

//////////
//  Construction/destruction
DecompressInputStream::DecompressInputStream(IByteInput * pSource, Decompressor * pDecompressor)
{
    require(pSource != NULL);
    require(pDecompressor != NULL && pDecompressor->getCompressionAlgorithm()->getMinimumBlockSize() <= CompressOutputStream::_MaximumBufferSize);

    _pSource = pSource;
    _bufferSize = pDecompressor->getCompressionAlgorithm()->getPreferredBlockSize();
    _pDecompressor = pDecompressor;
    if (_bufferSize > CompressOutputStream::_MaximumBufferSize)
    {
        _bufferSize = CompressOutputStream::_MaximumBufferSize;
    }
    require(_bufferSize >= pDecompressor->getCompressionAlgorithm()->getMinimumBlockSize() &&
            _bufferSize <= pDecompressor->getCompressionAlgorithm()->getMaximumBlockSize());
    _pBuffer = new UInt8[_bufferSize];
    _bytesInBuffer = 0;
    _currentPosition = 0;
}

DecompressInputStream::~DecompressInputStream()
{
    close();
}

//////////
//  Operations
void DecompressInputStream::close()
{
    delete [] _pBuffer;
    _pBuffer = NULL;
}

//////////
//  IByteInput
UIntSize DecompressInputStream::available() throw(IOException)
{
    require(isOpen());

    UIntSize avail1 = _bytesInBuffer - _currentPosition;
    if (avail1 != 0)
    {
        return avail1;
    }
    else if (_pSource->available() != 0)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

int DecompressInputStream::read() throw(IOException)
{
    require(isOpen());

    if (_currentPosition == _bytesInBuffer)
    {
        _refillBuffer();
        if (_currentPosition == _bytesInBuffer)
        {
            return -1;
        }
    }
    return _pBuffer[_currentPosition++];
}

UIntSize DecompressInputStream::read(void * pBuffer, UIntSize bytesToRead) throw(IOException)
{
    UInt8 * pBytes = static_cast<UInt8*>(pBuffer);
    for (UIntSize i = 0; i < bytesToRead; i++)
    {
        int bb = read();
        if (bb == -1)
        {
            return i;
        }
        pBytes[i] = static_cast<UInt8>(bb);
    }
    return bytesToRead;
}

UIntSize DecompressInputStream::skip(UIntSize bytesToSkip) throw(IOException)
{
    for (UIntSize i = 0; i < bytesToSkip; i++)
    {
        if (read() == -1)
        {
            return i;
        }
    }
    return bytesToSkip;
}

//////////
//  Helpers
void DecompressInputStream::_refillBuffer()
{
    //  Read the 1st byte which contains the block type
    int blockType = _pSource->read();
    if (blockType == -1)
    {   //  EOF
        _bytesInBuffer = 0;
        _currentPosition = 0;
    }
    else if (blockType == CompressOutputStream::_BlockType_Uncompressed)
    {   //  Uncompressed block
        int uncompressedSizeHi = _pSource->read();
        int uncompressedSizeLo = _pSource->read();
        if (uncompressedSizeHi == -1 || uncompressedSizeLo == -1)
        {
            throw InvalidDataException();
        }
        UIntSize uncompressedSize = (uncompressedSizeHi << 8) | (uncompressedSizeLo);
        if (uncompressedSize > _bufferSize)
        {
            throw InvalidDataException();
        }
        UIntSize bytesRead = _pSource->read(_pBuffer, uncompressedSize);
        if (bytesRead != uncompressedSize)
        {
            throw InvalidDataException();
        }
        _bytesInBuffer = bytesRead;
        _currentPosition = 0;
    }
    else if (blockType == CompressOutputStream::_BlockType_CompressedWithoutPadding)
    {   //  Compressed block without padding
        int compressedSizeHi = _pSource->read();
        int compressedSizeLo = _pSource->read();
        if (compressedSizeHi == -1 || compressedSizeLo == -1)
        {
            throw InvalidDataException();
        }
        UIntSize compressedSize = (compressedSizeHi << 8) | (compressedSizeLo);
        try
        {
            //  Read compressed data
            _compressedData.setSize(compressedSize);
            UIntSize bytesRead = _pSource->read(_compressedData.getData(), compressedSize);
            if (bytesRead != compressedSize)
            {
                throw InvalidDataException();
            }
            //  Decompress data
            UIntSize decompressedDataSize = _pDecompressor->getRequiredOutputBufferSize(_compressedData.getData(), _compressedData.size());
            if (decompressedDataSize > _bufferSize)
            {
                throw InvalidDataException();
            }
            UIntSize decompressedBytes = _pDecompressor->decompress(_compressedData.getData(), _compressedData.size(), _pBuffer);
            if (decompressedBytes != decompressedDataSize)
            {
                throw InvalidDataException();
            }
            _bytesInBuffer = decompressedBytes;
            _currentPosition = 0;
        }
        catch (CompressedDataCorruptException)
        {
            throw InvalidDataException();
        }
    }
    else if (blockType == CompressOutputStream::_BlockType_CompressedWithPadding)
    {   //  Compressed block without padding
        int compressedSizeHi = _pSource->read();
        int compressedSizeLo = _pSource->read();
        if (compressedSizeHi == -1 || compressedSizeLo == -1)
        {
            throw InvalidDataException();
        }
        UIntSize compressedSize = (compressedSizeHi << 8) | (compressedSizeLo);
        int paddingSizeHi = _pSource->read();
        int paddingSizeLo = _pSource->read();
        if (paddingSizeHi == -1 || paddingSizeLo == -1)
        {
            throw InvalidDataException();
        }
        UIntSize paddingSize = (paddingSizeHi << 8) | (paddingSizeLo);
        try
        {
            //  Read compressed data
            _compressedData.setSize(compressedSize);
            UIntSize bytesRead = _pSource->read(_compressedData.getData(), compressedSize);
            if (bytesRead != compressedSize)
            {
                throw InvalidDataException();
            }
            //  Decompress data
            UIntSize decompressedDataSize = _pDecompressor->getRequiredOutputBufferSize(_compressedData.getData(), _compressedData.size());
            if (decompressedDataSize > _bufferSize)
            {
                throw InvalidDataException();
            }
            UIntSize decompressedBytes = _pDecompressor->decompress(_compressedData.getData(), _compressedData.size(), _pBuffer);
            if (decompressedBytes != decompressedDataSize)
            {
                throw InvalidDataException();
            }
            _bytesInBuffer = decompressedBytes;
            _currentPosition = 0;
            //  Remove padding
            if (paddingSize > _bytesInBuffer)
            {
                throw InvalidDataException();
            }
            _bytesInBuffer -= paddingSize;
        }
        catch (CompressedDataCorruptException)
        {
            throw InvalidDataException();
        }
    }
    else
    {   //  Invalid block type
        throw InvalidDataException();
    }
}

//  End of DecompressInputStream.cpp
