//
//  DecryptInputStream.cpp - Implementation of the carl::DecryptInputStream class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:52 $
//  $Revision: 1.6 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

//////////
//  Construction/destruction
DecryptInputStream::DecryptInputStream(IByteInput * pSource, StreamDecryptor * pStreamDecryptor, bool autoDeleteStreamDecryptor /*= false*/)
    :   _pSource(pSource),
        _pStreamDecryptor(pStreamDecryptor),
        _autoDeleteStreamDecryptor(autoDeleteStreamDecryptor),
        _pBlockDecryptor(NULL),
        _pPaddingAlgorithm(NULL),
        _autoDeleteBlockDecryptor(false),
        _bitsReadyInInputBuffer(0),
        _inputBufferScan(0),
        _prereadByte(0)
{
    require(pSource);
    require(pStreamDecryptor);
}

DecryptInputStream::DecryptInputStream(IByteInput * pSource, BlockDecryptor * pBlockDecryptor, PaddingAlgorithm * pPaddingAlgorithm, bool autoDeleteBlockDecryptor /*= false*/)
    :   _pSource(pSource),
        _pStreamDecryptor(NULL),
        _autoDeleteStreamDecryptor(false),
        _pBlockDecryptor(pBlockDecryptor),
        _pPaddingAlgorithm(pPaddingAlgorithm),
        _autoDeleteBlockDecryptor(autoDeleteBlockDecryptor),
        _bitsReadyInInputBuffer(0),
        _inputBufferScan(0),
        _prereadByte(0)
{
    require(pSource);
    require(pBlockDecryptor);
    require(pPaddingAlgorithm);

    _prereadByte = _pSource->read();
}

//////////
//  Operations
void DecryptInputStream::close()
{
    if (_pSource != NULL)
    {
        if (_autoDeleteStreamDecryptor)
        {
            delete _pStreamDecryptor;
        }
        if (_autoDeleteBlockDecryptor)
        {
            delete _pBlockDecryptor;
        }
        _pSource = NULL;
    }
}

//////////
//  IByteInput
UIntSize DecryptInputStream::available() throw(IOException)
{
    require(_pSource != NULL);

    if (_bitsReadyInInputBuffer != 0)
    {
        return 1;
    }
    else
    {
        return (_pSource->available() == 0) ? 0 : 1;
    }
}

int DecryptInputStream::read() throw(IOException)
{
    require(_pSource != NULL);

    if (_pStreamDecryptor != NULL)
    {
        int ciphertextByte = _pSource->read();
        if (ciphertextByte == -1)
        {
            return -1;
        }
        UInt8 cleartextByte = 0;
        for (UIntSize i = 0; i < System::BitsPerByte; i++)
        {
            int ciphertextBit = (ciphertextByte >> i) & 0x01;
            int cleartextBit = _pStreamDecryptor->decryptBit(ciphertextBit);
            cleartextByte = static_cast<UInt8>(cleartextByte | (cleartextBit << i));
        }
        return cleartextByte;
    }
    else if (_pBlockDecryptor != NULL)
    {
        for (; ; )
        {
            if (_bitsReadyInInputBuffer >= System::BitsPerByte)
            {
                UInt8 b = _inputBuffer[0];
                _inputBuffer.removeAt(0);
                _bitsReadyInInputBuffer -= Math::min(System::BitsPerByte, _bitsReadyInInputBuffer);
                return b;
            }
            else if (_prereadByte == -1)
            {
                return -1;
            }
            else
            {
                _readInputBlock();
                if (_bitsReadyInInputBuffer == 0)
                {
                    return -1;
                }
                else if (_bitsReadyInInputBuffer < System::BitsPerByte)
                {
                    if (_inputBuffer.size() == 0)
                    {
                        _inputBuffer.append(0);
                    }
                    return _inputBuffer[0];
                }
            }
        }
    }
    else
    {
        require(false);
        return -1;
    }
}

UIntSize DecryptInputStream::read(void * pBuffer, UIntSize bytesToRead) throw(IOException)
{
    require(_pSource != NULL);

    UInt8 * pBytes = static_cast<UInt8*>(pBuffer);
    for (UIntSize i = 0; i < bytesToRead; i++)
    {
        int bb = read();
        if (bb == -1)
        {
            return i;
        }
        pBytes[i] = static_cast<UInt8>(bb);
    }
    return bytesToRead;
}

UIntSize DecryptInputStream::skip(UIntSize bytesToSkip) throw(IOException)
{
    require(_pSource != NULL);

    for (UIntSize i = 0; i < bytesToSkip; i++)
    {
        if (read() == -1)
        {
            return i;
        }
    }
    return bytesToSkip;
}

//////////
//  Helpers
UIntSize DecryptInputStream::_readUleb() throw(IOException)
{
    UIntSize result = 0;
    UIntSize shiftCount = 0;

    for (int byteNo = 0; ; byteNo++)
    {
        int b = (byteNo == 0) ? (_prereadByte) : (_pSource->read());
        if (b == -1)
        {
            throw InvalidDataException();
        }
        result |= ((b & 0x7F) << shiftCount);
        shiftCount += 7;
        if ((b & 0x80) == 0)
        {
            break;
        }
    }
    return result;
}

void DecryptInputStream::_readInputBlock() throw(IOException)
{
    //  Read ciphertext block size
    UIntSize ciphertextBlockSize = _readUleb();
    //  Read ciphertext block
    Array<UInt8> ciphertextBits;
    ciphertextBits.setSize(ciphertextBlockSize / System::BitsPerByte + 1);
    for (UIntSize i = 0; i < ciphertextBlockSize; i += System::BitsPerByte)
    {
        int b = _pSource->read();
        if (b == -1)
        {
            throw InvalidDataException();
        }
        ciphertextBits[i / System::BitsPerByte] = static_cast<UInt8>(b);
    }
    //  Decrypt
    UIntSize cleartextBlockSize = _pBlockDecryptor->getCleartextSizeForCiphertext(ciphertextBits.getData(), ciphertextBlockSize);
    while (_inputBuffer.size() < cleartextBlockSize / System::BitsPerByte + 1)
    {
        _inputBuffer.append(0);
    }
    Array<UInt8> cleartextBits;
    cleartextBits.setSize(cleartextBlockSize / System::BitsPerByte + 1);
    UIntSize cleartextBitCount = _pBlockDecryptor->decryptBlock(ciphertextBits.getData(), ciphertextBlockSize, cleartextBits.getData());
    require(cleartextBitCount == cleartextBlockSize);
    for (UIntSize i = 0; i < cleartextBitCount; i++)
    {
        UIntSize byteIndex = i / System::BitsPerByte;
        UIntSize bitIndex = i % System::BitsPerByte;
        int bit = (cleartextBits[byteIndex] >> (System::BitsPerByte - bitIndex - 1)) & 0x01;
        byteIndex = _bitsReadyInInputBuffer / System::BitsPerByte;
        bitIndex = _bitsReadyInInputBuffer % System::BitsPerByte;
        _inputBuffer[byteIndex] = static_cast<UInt8>(_inputBuffer[byteIndex] | (bit << (System::BitsPerByte - bitIndex - 1)));
        _bitsReadyInInputBuffer++;
    }
    //  If this is the last block, remove the padding
    _prereadByte = _pSource->read();
    if (_prereadByte == -1)
    {
        UIntSize paddingBits = _pPaddingAlgorithm->getPaddingBitsStoredInData(cleartextBits.getData(), cleartextBlockSize, cleartextBlockSize);
        for (UIntSize i = paddingBits; i < ciphertextBlockSize; i++)
        {
            UIntSize byteIndex = i / System::BitsPerByte;
            UIntSize bitIndex = i % System::BitsPerByte;
            ciphertextBits[byteIndex] = static_cast<UInt8>(ciphertextBits[byteIndex] & (~(1 << (System::BitsPerByte - bitIndex - 1))));
        }
        _bitsReadyInInputBuffer -= paddingBits;
    }
}

//  End of DecryptInputStream.cpp
