//
//  DefaultCollator.cpp - Implementation of the carl::_impl::_DefaultCollator class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/26 11:08:27 $
//  $Revision: 1.5 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "CollatorImpl.hpp"

//////////
//  Singleton
_impl::_DefaultCollator::_DefaultCollator()
{
}

_impl::_DefaultCollator::~_DefaultCollator()
{
}

_impl::_DefaultCollator * _impl::_DefaultCollator::getInstance()
{
    static _DefaultCollator instance;
    return &instance;
}

//////////
//  Operations
int _impl::_DefaultCollator::collate(const String & s1, const String & s2, CollationStrength collationStrength /*= CollationStrength_Primary*/, Decomposition decomposition /*= Decomposition_Full*/)
{
    //  Convert both strings to canonical forms
    String cs1 = Iso10646::performCanonicalComposition(_performDecomposition(s1, decomposition));
    String cs2 = Iso10646::performCanonicalComposition(_performDecomposition(s2, decomposition));
    //  Collate until one of the strings runs out
    for (UIntSize i = 0; i < cs1.length() && i < cs2.length(); i++)
    {
        Char c1 = cs1[i];
        Char c2 = cs2[i];
        int difference;
        switch (collationStrength)
        {
            case Collator::CollationStrength_Primary:
                difference = _calculatePrimaryDifference(c1, c2);
                break;
            case Collator::CollationStrength_Secondary:
                difference = _calculateSecondaryDifference(c1, c2);
                break;
            case Collator::CollationStrength_Tertiary:
                difference = _calculateTertiaryDifference(c1, c2);
                break;
            case Collator::CollationStrength_Identical:
                if (c1 < c2)
                {
                    difference = -1;
                }
                else if (c1 > c2)
                {
                    difference = 1;
                }
                else
                {
                    difference = 0;
                }
                break;
            default:
                require(false);
                difference = 0;
                break;
        }
        if (difference != 0)
        {
            return difference;
        }
    }
    //  One string starts with another
    if (cs1.length() < cs2.length())
    {
        return -1;
    }
    else if (cs1.length() > cs2.length())
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

//////////
//  Implementation helpers
String _impl::_DefaultCollator::_performDecomposition(const String & s, Decomposition decomposition)
{
    switch (decomposition)
    {
        case Decomposition_None:
            return s;
        case Decomposition_Canonical:
            return Iso10646::performCanonicalDecomposition(s);
        case Decomposition_Full:
            return Iso10646::performCompatibilityDecomposition(s);
        default:
            require(false);
            return s;
    }
}

int _impl::_DefaultCollator::_calculatePrimaryDifference(Char c1, Char c2)
{
    //  All control characters are equivalent at the primary level
    if (Iso10646::isControl(c1) && Iso10646::isControl(c2))
    {
        return 0;
    }

    //  Only the base characters shall be tested
    Char c1baseu = Iso10646::toUpper(_getBaseCharacter(c1));
    Char c2baseu = Iso10646::toUpper(_getBaseCharacter(c2));
    if (c1baseu < c2baseu)
    {
        return -1;
    }
    else if (c1baseu > c2baseu)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

int _impl::_DefaultCollator::_calculateSecondaryDifference(Char c1, Char c2)
{
    //  All control characters are equivalent at the secondary level
    if (Iso10646::isControl(c1) && Iso10646::isControl(c2))
    {
        return 0;
    }

    //  Same character?
    Char c1u = Iso10646::toUpper(c1);
    Char c2u = Iso10646::toUpper(c2);
    if (c1u < c2u)
    {
        return -1;
    }
    else if (c1u > c2u)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

int _impl::_DefaultCollator::_calculateTertiaryDifference(Char c1, Char c2)
{
    //  All control characters are equivalent at the tertiary level
    if (Iso10646::isControl(c1) && Iso10646::isControl(c2))
    {
        return 0;
    }

    //  Same character?
    Char c1u = Iso10646::toUpper(c1);
    Char c2u = Iso10646::toUpper(c2);
    if (c1u < c2u)
    {
        return -1;
    }
    else if (c1u > c2u)
    {
        return 1;
    }

    //  Same character, but, maybe, different letter case?
    if (Iso10646::isTitle(c1) && Iso10646::isTitle(c2))
    {
        return 0;
    }
    else if (Iso10646::isTitle(c1) && !Iso10646::isTitle(c2))
    {
        return -1;
    }
    else if (!Iso10646::isTitle(c1) && Iso10646::isTitle(c2))
    {
        return 1;
    }
    else if (Iso10646::isUpper(c1) && Iso10646::isUpper(c2))
    {
        return 0;
    }
    else if (Iso10646::isUpper(c1) && !Iso10646::isUpper(c2))
    {
        return -1;
    }
    else if (!Iso10646::isUpper(c1) && Iso10646::isUpper(c2))
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

Char _impl::_DefaultCollator::_getBaseCharacter(Char c)
{
    UIntSize decompositionLength;
    if (Iso10646::getDecompositionMappingSize(c, decompositionLength))
    {
        Array<Char> decompositionData;
        decompositionData.setSize(decompositionLength);
        if (Iso10646::getDecompositionMapping(c, decompositionData.getData()))
        {
            return decompositionData[0];
        }
        else
        {
            return c;
        }
    }
    else
    {
        return c;
    }
}

//  End of DefaultCollator.cpp
