//
//  DefaultNumberFormat.cpp - Implementation of the carl::_impl::_DefaultNumberFormat class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/09/12 11:36:17 $
//  $Revision: 1.18 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "NumberFormatImpl.hpp"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <float.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <math.h>
#else
    #error Unsupported platform
#endif

//////////
//  Helpers
namespace
{
    const char Digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";

    String createReplicatedString(Char ch, UIntSize count)
    {
        String result;
        for (UIntSize i = 0; i < count; i++)
        {
            result += ch;
        }
        return result;
    }

    template <class T>
    UInt8 integerToUInt8(const T & value)
    {
        return static_cast<UInt8>(value);
    }

    template <>
    UInt8 integerToUInt8<BinaryInteger>(const BinaryInteger & value)
    {
        return value.toUInt8();
    }

    template <>
    UInt8 integerToUInt8<DecimalInteger>(const DecimalInteger & value)
    {
        return value.toUInt8();
    }

    template <class T>
    String integerToDigits(T value, UIntSize radix)
    {
        require(value >= 0);

        if (value == 0)
        {
            return String("0");
        }
        else
        {
            String digitString;
            for (; value != 0; value = static_cast<T>(value / radix))
            {
                UInt8 digit = integerToUInt8(value % radix);
                digitString = Digits[digit] + digitString;
            }
            return digitString;
        }
    }

    template <class T>
    UInt8 realToUInt8(const T & value)
    {
        return static_cast<UInt8>(value);
    }

    template <>
    UInt8 realToUInt8<BinaryFixed>(const BinaryFixed & value)
    {
        return value.toUInt8();
    }

    template <>
    UInt8 realToUInt8<BinaryFloat>(const BinaryFloat & value)
    {
        return value.toUInt8();
    }

    template <>
    UInt8 realToUInt8<DecimalFixed>(const DecimalFixed & value)
    {
        return value.toUInt8();
    }

    template <>
    UInt8 realToUInt8<DecimalFloat>(const DecimalFloat & value)
    {
        return value.toUInt8();
    }

    template <class T, class D>
    String realToFloatingPointDigits(T value, UIntSize radix, UIntSize significantDigits, D & decimapPointPosition)
    {
        require(value >= 0);
        require(significantDigits >= 1);

        D exponent = 0;
        T rangeLo = 1;
        T rangeHi = radix;
        while (value < rangeLo)
        {
            exponent--;
            value *= radix;
        }
        while (value >= rangeHi)
        {
            exponent++;
            rangeLo *= radix;
            rangeHi *= radix;
        }
        //  Prepare the mantissa
        T mantissa = value / rangeLo;
        //  Prepare the required number of digits + 1 more for rounding
        Array<UInt8> digits;
        for (UIntSize i = 0; i < significantDigits + 1; i++)
        {
            UInt8 digit = realToUInt8(mantissa);
            digits.append(digit);
            mantissa = (mantissa - digit) * radix;
        }
        //  Perform rounding
        if (digits[digits.size() - 1] >= radix / 2)
        {
            UInt8 carry = 1;
            UIntSize position;
            for (position = digits.size() - 1; position != 0; position--)
            {
                UInt8 digitPlusCarry = static_cast<UInt8>(digits[position - 1] + carry);
                carry = static_cast<UInt8>(digitPlusCarry / radix);
                digits[position - 1] = static_cast<UInt8>(digitPlusCarry % radix);
                if (carry == 0)
                {
                    break;
                }
            }
            if (carry != 0 && position == 0)
            {   //  Carry out of the highest digit
                digits.insertAt(0, 1);
            }
        }
        //  Form the result
        String result;
        for (UIntSize i = 0; i < significantDigits; i++)
        {
            result += Digits[digits[i]];
        }
        //  Prepare decimap point position & return
        decimapPointPosition = exponent + 1;
        return result;
    }

    template <class T, class D>
    String realToFixedPointDigits(T value, UIntSize radix, UIntSize fractionalDigits, D & decimapPointPosition)
    {
        require(value >= 0);

        //  Format just 2 digits to find out where the decimal point is placed
        D ecvtDecimalPointPosition;
        realToFloatingPointDigits(value, radix, 2, ecvtDecimalPointPosition);
        //  Decide how many significant digits must be produced to match the
        //  required number of fractional digits
        UIntSize significantDigits;
        if (ecvtDecimalPointPosition >= 0)
        {
            significantDigits = ecvtDecimalPointPosition + fractionalDigits;
        }
        else if (static_cast<UIntSize>(-ecvtDecimalPointPosition) < fractionalDigits)
        {
            significantDigits = fractionalDigits + ecvtDecimalPointPosition;
        }
        else
        {
            significantDigits = 1;
        }
        //  Go!
        return realToFloatingPointDigits(value, radix, significantDigits, decimapPointPosition);
    }

    void inferFloatingPointFormattingProperties(const RealFormatProperties & formatProperties, UIntSize significantDigitCount, UIntSize defaultDigitsToFormat, UIntSize & digitsToFormat, UIntSize & insertDecimalPointAt)
    {
        require(defaultDigitsToFormat >= 1);

        if (formatProperties.getMinimumDigits() != 0 && formatProperties.getMaximumDigits() != 0)
        {   //  Both minimum & maximum number of digits to format are specified
            require(formatProperties.getMinimumDigits() <= formatProperties.getMaximumDigits());
            if (formatProperties.getMinimumFractionalDigits() != 0 && formatProperties.getMaximumFractionalDigits() != 0)
            {   //  Both minimum & maximum number of fractional digits are specified
                require(formatProperties.getMinimumFractionalDigits() <= formatProperties.getMaximumFractionalDigits());
                //
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt < formatProperties.getMinimumFractionalDigits())
                {
                    UIntSize deltaDigits = formatProperties.getMinimumFractionalDigits() - (digitsToFormat - insertDecimalPointAt);
                    digitsToFormat += deltaDigits;
                }
                if (digitsToFormat - insertDecimalPointAt > formatProperties.getMaximumFractionalDigits())
                {
                    insertDecimalPointAt = digitsToFormat - formatProperties.getMaximumFractionalDigits();
                }
            }
            else if (formatProperties.getMinimumFractionalDigits() != 0)
            {   //  The minimum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt < formatProperties.getMinimumFractionalDigits())
                {
                    UIntSize deltaDigits = formatProperties.getMinimumFractionalDigits() - (digitsToFormat - insertDecimalPointAt);
                    digitsToFormat += deltaDigits;
                }
            }
            else if (formatProperties.getMaximumFractionalDigits() != 0)
            {   //  The maximum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt > formatProperties.getMaximumFractionalDigits())
                {
                    insertDecimalPointAt = digitsToFormat - formatProperties.getMaximumFractionalDigits();
                }
            }
            else
            {   //  The number of fractional digits is not specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                insertDecimalPointAt = 1;
            }
        }
        else if (formatProperties.getMinimumDigits() != 0)
        {   //  A minimum number of digits to format is specified
            if (formatProperties.getMinimumFractionalDigits() != 0 && formatProperties.getMaximumFractionalDigits() != 0)
            {   //  Both minimum & maximum number of fractional digits are specified
                require(formatProperties.getMinimumFractionalDigits() <= formatProperties.getMaximumFractionalDigits());
                //
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt < formatProperties.getMinimumFractionalDigits())
                {
                    UIntSize deltaDigits = formatProperties.getMinimumFractionalDigits() - (digitsToFormat - insertDecimalPointAt);
                    digitsToFormat += deltaDigits;
                }
                if (digitsToFormat - insertDecimalPointAt > formatProperties.getMaximumFractionalDigits())
                {
                    insertDecimalPointAt = digitsToFormat - formatProperties.getMaximumFractionalDigits();
                }
            }
            else if (formatProperties.getMinimumFractionalDigits() != 0)
            {   //  The minimum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt < formatProperties.getMinimumFractionalDigits())
                {
                    UIntSize deltaDigits = formatProperties.getMinimumFractionalDigits() - (digitsToFormat - insertDecimalPointAt);
                    digitsToFormat += deltaDigits;
                }
            }
            else if (formatProperties.getMaximumFractionalDigits() != 0)
            {   //  The maximum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt > formatProperties.getMaximumFractionalDigits())
                {
                    insertDecimalPointAt = digitsToFormat - formatProperties.getMaximumFractionalDigits();
                }
            }
            else
            {   //  The number of fractional digits is not specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                insertDecimalPointAt = 1;
            }
        }
        else if (formatProperties.getMaximumDigits() != 0)
        {   //  A maximum number of digits to format is specified
            if (formatProperties.getMinimumFractionalDigits() != 0 && formatProperties.getMaximumFractionalDigits() != 0)
            {   //  Both minimum & maximum number of fractional digits are specified
                require(formatProperties.getMinimumFractionalDigits() <= formatProperties.getMaximumFractionalDigits());
                //
                digitsToFormat = significantDigitCount;
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt < formatProperties.getMinimumFractionalDigits())
                {
                    UIntSize deltaDigits = formatProperties.getMinimumFractionalDigits() - (digitsToFormat - insertDecimalPointAt);
                    digitsToFormat += deltaDigits;
                }
                if (digitsToFormat - insertDecimalPointAt > formatProperties.getMaximumFractionalDigits())
                {
                    insertDecimalPointAt = digitsToFormat - formatProperties.getMaximumFractionalDigits();
                }
            }
            else if (formatProperties.getMinimumFractionalDigits() != 0)
            {   //  The minimum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt < formatProperties.getMinimumFractionalDigits())
                {
                    UIntSize deltaDigits = formatProperties.getMinimumFractionalDigits() - (digitsToFormat - insertDecimalPointAt);
                    digitsToFormat += deltaDigits;
                }
            }
            else if (formatProperties.getMaximumFractionalDigits() != 0)
            {   //  The maximum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt > formatProperties.getMaximumFractionalDigits())
                {
                    insertDecimalPointAt = digitsToFormat - formatProperties.getMaximumFractionalDigits();
                }
            }
            else
            {   //  The number of fractional digits is not specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                insertDecimalPointAt = 1;
            }
        }
        else
        {   //  The number of digits to format is not specified
            if (formatProperties.getMinimumFractionalDigits() != 0 && formatProperties.getMaximumFractionalDigits() != 0)
            {   //  Both minimum & maximum number of fractional digits are specified
                require(formatProperties.getMinimumFractionalDigits() <= formatProperties.getMaximumFractionalDigits());
                //
                digitsToFormat = significantDigitCount;
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt < formatProperties.getMinimumFractionalDigits())
                {
                    UIntSize deltaDigits = formatProperties.getMinimumFractionalDigits() - (digitsToFormat - insertDecimalPointAt);
                    digitsToFormat += deltaDigits;
                }
                if (digitsToFormat - insertDecimalPointAt > formatProperties.getMaximumFractionalDigits())
                {
                    insertDecimalPointAt = digitsToFormat - formatProperties.getMaximumFractionalDigits();
                }
            }
            else if (formatProperties.getMinimumFractionalDigits() != 0)
            {   //  The minimum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt < formatProperties.getMinimumFractionalDigits())
                {
                    UIntSize deltaDigits = formatProperties.getMinimumFractionalDigits() - (digitsToFormat - insertDecimalPointAt);
                    digitsToFormat += deltaDigits;
                }
            }
            else if (formatProperties.getMaximumFractionalDigits() != 0)
            {   //  The maximum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                insertDecimalPointAt = 1;
                if (digitsToFormat - insertDecimalPointAt > formatProperties.getMaximumFractionalDigits())
                {
                    insertDecimalPointAt = digitsToFormat - formatProperties.getMaximumFractionalDigits();
                }
            }
            else
            {   //  The number of fractional digits is not specified
                //  Format defaultDigitsToFormat digits, with 1 digit before the decimal point
                digitsToFormat = defaultDigitsToFormat;
                insertDecimalPointAt = 1;
            }
        }
    }

    String inferFixedPointFormattingProperties(const RealFormatProperties & formatProperties, UIntSize significantDigitCount, UIntSize significantDigitsBeforeDecimalPoint, UIntSize & digitsToFormat)
    {
        if (formatProperties.getMinimumDigits() != 0 && formatProperties.getMaximumDigits() != 0)
        {   //  Both minimum & maximum number of digits to format are specified
            require(formatProperties.getMinimumDigits() <= formatProperties.getMaximumDigits());
            if (formatProperties.getMinimumFractionalDigits() != 0 && formatProperties.getMaximumFractionalDigits() != 0)
            {   //  Both minimum & maximum number of fractional digits are specified
                require(formatProperties.getMinimumFractionalDigits() <= formatProperties.getMaximumFractionalDigits());
                //
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint < formatProperties.getMinimumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMinimumFractionalDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint > formatProperties.getMaximumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMaximumFractionalDigits();
                }
                if (digitsToFormat < significantDigitsBeforeDecimalPoint || 
                    digitsToFormat < formatProperties.getMinimumDigits() ||
                    digitsToFormat - significantDigitsBeforeDecimalPoint < formatProperties.getMinimumFractionalDigits())
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMinimumDigits());
                }
                if (digitsToFormat > formatProperties.getMaximumDigits() ||
                    digitsToFormat - significantDigitsBeforeDecimalPoint > formatProperties.getMaximumFractionalDigits())
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMaximumDigits());
                }
            }
            else if (formatProperties.getMinimumFractionalDigits() != 0)
            {   //  The minimum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint < formatProperties.getMinimumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMinimumFractionalDigits();
                }
                if (digitsToFormat < significantDigitsBeforeDecimalPoint ||
                    digitsToFormat < formatProperties.getMinimumDigits() || digitsToFormat > formatProperties.getMaximumDigits())
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMinimumDigits());
                }
            }
            else if (formatProperties.getMaximumFractionalDigits() != 0)
            {   //  The maximum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint > formatProperties.getMaximumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMaximumFractionalDigits();
                }
                if (digitsToFormat < significantDigitsBeforeDecimalPoint || digitsToFormat < formatProperties.getMinimumDigits())
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMinimumDigits());
                }
            }
            else
            {   //  The number of fractional digits is not specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                if (digitsToFormat < significantDigitsBeforeDecimalPoint)
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMinimumDigits());
                }
            }
        }
        else if (formatProperties.getMinimumDigits() != 0)
        {   //  A minimum number of digits to format is specified
            if (formatProperties.getMinimumFractionalDigits() != 0 && formatProperties.getMaximumFractionalDigits() != 0)
            {   //  Both minimum & maximum number of fractional digits are specified
                require(formatProperties.getMinimumFractionalDigits() <= formatProperties.getMaximumFractionalDigits());
                //
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint < formatProperties.getMinimumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMinimumFractionalDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint > formatProperties.getMaximumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMaximumFractionalDigits();
                }
                if (digitsToFormat < significantDigitsBeforeDecimalPoint || digitsToFormat < formatProperties.getMinimumDigits() ||
                    digitsToFormat - significantDigitsBeforeDecimalPoint > formatProperties.getMaximumFractionalDigits())
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMinimumDigits());
                }
            }
            else if (formatProperties.getMinimumFractionalDigits() != 0)
            {   //  The minimum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint < formatProperties.getMinimumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMinimumFractionalDigits();
                }
            }
            else if (formatProperties.getMaximumFractionalDigits() != 0)
            {   //  The maximum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint > formatProperties.getMaximumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMaximumFractionalDigits();
                }
                if (digitsToFormat < significantDigitsBeforeDecimalPoint || digitsToFormat < formatProperties.getMinimumDigits())
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMinimumDigits());
                }
            }
            else
            {   //  The number of fractional digits is not specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat < formatProperties.getMinimumDigits())
                {
                    digitsToFormat = formatProperties.getMinimumDigits();
                }
            }
        }
        else if (formatProperties.getMaximumDigits() != 0)
        {   //  A maximum number of digits to format is specified
            if (formatProperties.getMinimumFractionalDigits() != 0 && formatProperties.getMaximumFractionalDigits() != 0)
            {   //  Both minimum & maximum number of fractional digits are specified
                require(formatProperties.getMinimumFractionalDigits() <= formatProperties.getMaximumFractionalDigits());
                //
                digitsToFormat = significantDigitCount;
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint < formatProperties.getMinimumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMinimumFractionalDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint > formatProperties.getMaximumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMaximumFractionalDigits();
                }
                if (digitsToFormat < significantDigitsBeforeDecimalPoint || digitsToFormat > formatProperties.getMaximumDigits())
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMaximumDigits());
                }
            }
            else if (formatProperties.getMinimumFractionalDigits() != 0)
            {   //  The minimum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint < formatProperties.getMinimumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMinimumFractionalDigits();
                }
                if (digitsToFormat < significantDigitsBeforeDecimalPoint || digitsToFormat > formatProperties.getMaximumDigits())
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMaximumDigits());
                }
            }
            else if (formatProperties.getMaximumFractionalDigits() != 0)
            {   //  The maximum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint > formatProperties.getMaximumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMaximumFractionalDigits();
                }
                if (digitsToFormat < significantDigitsBeforeDecimalPoint)
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMaximumDigits());
                }
            }
            else
            {   //  The number of fractional digits is not specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat > formatProperties.getMaximumDigits())
                {
                    digitsToFormat = formatProperties.getMaximumDigits();
                }
                if (digitsToFormat < significantDigitsBeforeDecimalPoint)
                {
                    return createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMaximumDigits());
                }
            }
        }
        else
        {   //  The number of digits to format is not specified
            if (formatProperties.getMinimumFractionalDigits() != 0 && formatProperties.getMaximumFractionalDigits() != 0)
            {   //  Both minimum & maximum number of fractional digits are specified
                require(formatProperties.getMinimumFractionalDigits() <= formatProperties.getMaximumFractionalDigits());
                //
                digitsToFormat = significantDigitCount;
                if (digitsToFormat - significantDigitsBeforeDecimalPoint < formatProperties.getMinimumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMinimumFractionalDigits();
                }
                if (digitsToFormat - significantDigitsBeforeDecimalPoint > formatProperties.getMaximumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMaximumFractionalDigits();
                }
            }
            else if (formatProperties.getMinimumFractionalDigits() != 0)
            {   //  The minimum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat - significantDigitsBeforeDecimalPoint < formatProperties.getMinimumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMinimumFractionalDigits();
                }
            }
            else if (formatProperties.getMaximumFractionalDigits() != 0)
            {   //  The maximum number of fractional digits is specified
                digitsToFormat = significantDigitCount;
                if (digitsToFormat - significantDigitsBeforeDecimalPoint > formatProperties.getMaximumFractionalDigits())
                {
                    digitsToFormat = significantDigitsBeforeDecimalPoint + formatProperties.getMaximumFractionalDigits();
                }
            }
            else
            {   //  The number of fractional digits is not specified
                digitsToFormat = significantDigitCount;
            }
        }
        return String();
    }

    String insertDecimalPointAndSeparators(String valueString, UIntSize insertDecimalPointAt, const RealFormatProperties & formatProperties, Char decimalPointChar, Char separatorChar)
    {
        require(insertDecimalPointAt >= 0 && insertDecimalPointAt <= valueString.length());

        //  Insert decimal point
        valueString.insert(decimalPointChar, insertDecimalPointAt);
        //  Insert separators
        if (formatProperties.getUseSeparator())
        {
            int separatorPosition = insertDecimalPointAt + (valueString.length() / 3 * 3) + 3;
            for (; separatorPosition > 0; separatorPosition -= 3)
            {
                if (static_cast<UIntSize>(separatorPosition) < valueString.length() && 
                    static_cast<UIntSize>(separatorPosition) != insertDecimalPointAt)
                {
                    valueString.insert(separatorChar, separatorPosition);
                }
            }
        }
        //  Done
        return valueString;
    }

    template <class T, class D>
    String formatFloatingPointRealValue(const T & value, UIntSize significantDigitCount, const RealFormatProperties & formatProperties, Char decimalPointChar, Char separatorChar)
    {
        String valueString;

        UIntSize digitsToFormat, insertDecimalPointAt;
        inferFloatingPointFormattingProperties(formatProperties, significantDigitCount, significantDigitCount, digitsToFormat, insertDecimalPointAt);
        //  Prepare the digits
        D dec;
        valueString = realToFloatingPointDigits(Math::abs(value), formatProperties.getRadix(), digitsToFormat, dec);
        //  Insert decimal point & separators
        valueString = insertDecimalPointAndSeparators(valueString, insertDecimalPointAt, formatProperties, decimalPointChar, separatorChar);
        //  Prepare exponent
        D exponentValue = (value != 0) ? (dec - insertDecimalPointAt) : (0);
        String exponentString = integerToDigits(Math::abs(exponentValue), formatProperties.getRadix());
        if (formatProperties.getMinimumExponentDigits() != 0)
        {
            while (exponentString.length() < formatProperties.getMinimumExponentDigits())
            {
                exponentString = String('0') + exponentString;
            }
        }
        if (formatProperties.getMaximumExponentDigits() != 0)
        {
            if (exponentString.length() > formatProperties.getMaximumExponentDigits())
            {
                exponentString = createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMaximumExponentDigits());
            }
        }
        if (exponentValue < 0)
        {
            exponentString = String('-') + exponentString;
        }
        else
        {
            exponentString = String('+') + exponentString;
        }
        valueString += String('e') + exponentString;
        //  Done
        return valueString;
    }

    template <class T, class D>
    String formatFixedPointRealValue(T value, UIntSize significantDigitCount, UIntSize significantDigitsBeforeDecimalPoint, const RealFormatProperties & formatProperties, Char decimalPointChar, Char separatorChar)
    {
        String valueString;

        require(significantDigitsBeforeDecimalPoint >= 1 && significantDigitCount >= significantDigitsBeforeDecimalPoint);
        UIntSize digitsToFormat;

        String errorString = inferFixedPointFormattingProperties(formatProperties, significantDigitCount, significantDigitsBeforeDecimalPoint, digitsToFormat);
        if (errorString.length() != 0)
        {
            return errorString;
        }
        //  Prepare the digits
        D dec;
        require(digitsToFormat >= significantDigitsBeforeDecimalPoint);
        valueString = realToFixedPointDigits(Math::abs(value), formatProperties.getRadix(), digitsToFormat - significantDigitsBeforeDecimalPoint, dec);
        while (dec < 1)
        {
            dec++;
            valueString = String('0') + valueString;
        }
        while (valueString.length() < digitsToFormat)
        {
            valueString = String('0') + valueString;
        }
        if (valueString.length() > digitsToFormat)
        {
            valueString.setLength(digitsToFormat);
        }
        //  Insert decimal point & separators
        valueString = insertDecimalPointAndSeparators(valueString, significantDigitsBeforeDecimalPoint, formatProperties, decimalPointChar, separatorChar);
        //  Done
        return valueString;
    }
}

//////////
//  Construction/destruction
_impl::_DefaultNumberFormat::_DefaultNumberFormat(Char decimalPointChar, Char separatorChar)
{
    _decimalPointChar = decimalPointChar;
    _separatorChar = separatorChar;
}

_impl::_DefaultNumberFormat::~_DefaultNumberFormat()
{
}

//////////
//  Operations
String _impl::_DefaultNumberFormat::format(Int8 value, const IntegerFormatProperties & formatProperties) const
{
    return _postFormatIntegerValue(integerToDigits(static_cast<UInt8>(Math::abs(value)), formatProperties.getRadix()), value < 0, formatProperties);
}

String _impl::_DefaultNumberFormat::format(Int16 value, const IntegerFormatProperties & formatProperties) const
{
    return _postFormatIntegerValue(integerToDigits(static_cast<UInt16>(Math::abs(value)), formatProperties.getRadix()), value < 0, formatProperties);
}

String _impl::_DefaultNumberFormat::format(Int32 value, const IntegerFormatProperties & formatProperties) const
{
    return _postFormatIntegerValue(integerToDigits(static_cast<UInt32>(Math::abs(value)), formatProperties.getRadix()), value < 0, formatProperties);
}

String _impl::_DefaultNumberFormat::format(Int64 value, const IntegerFormatProperties & formatProperties) const
{
    return _postFormatIntegerValue(integerToDigits(static_cast<UInt64>(Math::abs(value)), formatProperties.getRadix()), value < 0, formatProperties);
}

String _impl::_DefaultNumberFormat::format(UInt8 value, const IntegerFormatProperties & formatProperties) const
{
    return _postFormatIntegerValue(integerToDigits(value, formatProperties.getRadix()), false, formatProperties);
}

String _impl::_DefaultNumberFormat::format(UInt16 value, const IntegerFormatProperties & formatProperties) const
{
    return _postFormatIntegerValue(integerToDigits(value, formatProperties.getRadix()), false, formatProperties);
}

String _impl::_DefaultNumberFormat::format(UInt32 value, const IntegerFormatProperties & formatProperties) const
{
    return _postFormatIntegerValue(integerToDigits(value, formatProperties.getRadix()), false, formatProperties);
}

String _impl::_DefaultNumberFormat::format(UInt64 value, const IntegerFormatProperties & formatProperties) const
{
    return _postFormatIntegerValue(integerToDigits(value, formatProperties.getRadix()), false, formatProperties);
}

String _impl::_DefaultNumberFormat::format(const BinaryInteger & value, const IntegerFormatProperties & formatProperties) const
{
    return _postFormatIntegerValue(integerToDigits(Math::abs(value), formatProperties.getRadix()), value.getSign() == -1, formatProperties);
}

String _impl::_DefaultNumberFormat::format(const DecimalInteger & value, const IntegerFormatProperties & formatProperties) const
{
    return _postFormatIntegerValue(integerToDigits(Math::abs(value), formatProperties.getRadix()), value.getSign() == -1, formatProperties);
}

String _impl::_DefaultNumberFormat::format(Real32 value, const RealFormatProperties & formatProperties) const
{
    return format(static_cast<Real64>(value), formatProperties);
}

String _impl::_DefaultNumberFormat::format(Real64 value, const RealFormatProperties & formatProperties) const
{
    if (Math::isSignalingNan(value))
    {
        return _postFormatRealValue(String("#SNAN"), formatProperties);
    }
    else if (Math::isQuietNan(value))
    {
        return _postFormatRealValue(String("#QNAN"), formatProperties);
    }
    else if (Math::isNegativeInfinite(value))
    {
        return _postFormatRealValue(String("#-INF"), formatProperties);
    }
    else if (Math::isPositiveInfinite(value))
    {
        return _postFormatRealValue(String("#+INF"), formatProperties);
    }
    //  Store significant digits - 16 max
    int decimalPointPosition;
    String significantDigits = realToFloatingPointDigits(Math::abs(value), formatProperties.getRadix(), 16, decimalPointPosition);
    //  Count how many digits are actually significant
    while (significantDigits.endsWith('0'))
    {
        significantDigits = significantDigits.left(significantDigits.length() - 1);
    }
    int significantDigitCount = Math::max(static_cast<int>(significantDigits.length()), 1);
    //  Determine which format to use
    RealFormatProperties::Format format = formatProperties.getFormat();
    if (format == RealFormatProperties::Format_AutoSelect)
    {   //  Auto select format
        if (decimalPointPosition <= 0 || decimalPointPosition > significantDigitCount)
        {
            format = RealFormatProperties::Format_FloatingPoint;
        }
        else
        {
            format = RealFormatProperties::Format_FixedPoint;
        }
    }
    //  Prepare the value string
    String valueString;
    if (format == RealFormatProperties::Format_FloatingPoint)
    {   //  Floating-point
        valueString = formatFloatingPointRealValue<Real64, int>(value, significantDigitCount, formatProperties, _decimalPointChar, _separatorChar);
    }
    else
    {   //  Fixed-point
        if (decimalPointPosition > significantDigitCount)
        {
            significantDigitCount = decimalPointPosition;
        }
        UIntSize significantDigitsBeforeDecimalPoint;
        if (decimalPointPosition > 0)
        {
            significantDigitsBeforeDecimalPoint = decimalPointPosition;
        }
        else
        {
            significantDigitsBeforeDecimalPoint = 1;
            significantDigitCount += (-decimalPointPosition + 1);
        }
        valueString = formatFixedPointRealValue<Real64, int>(value, significantDigitCount, significantDigitsBeforeDecimalPoint, formatProperties, _decimalPointChar, _separatorChar);
    }
    //  Prepend the sign if necessary
    if (value < 0)
    {
        valueString = String("-") + valueString;
    }
    else if (formatProperties.getMandatorySign())
    {
        valueString = String("+") + valueString;
    }
    //  Post-format
    return _postFormatRealValue(valueString, formatProperties);
}

String _impl::_DefaultNumberFormat::format(const BinaryFixed & value, const RealFormatProperties & formatProperties) const
{
    //  Determine the number of significant digits to show in the output value
    IntPtrDiff significantDigitCount;
    if (formatProperties.getRadix() == 2)
    {
        significantDigitCount = value.getMagnitude().getRequiredBits() - 1;
    }
    else if (formatProperties.getRadix() <= 4)
    {
        significantDigitCount = (value.getMagnitude().getRequiredBits()) / 2;
    }
    else if (formatProperties.getRadix() <= 8)
    {
        significantDigitCount = (value.getMagnitude().getRequiredBits() + 1) / 3;
    }
    else if (formatProperties.getRadix() <= 16)
    {
        significantDigitCount = (value.getMagnitude().getRequiredBits() + 2) / 4;
    }
    else
    {
        significantDigitCount = (value.getMagnitude().getRequiredBits() + 3) / 5;
    }
    if (significantDigitCount == 0)
    {
        significantDigitCount = 1;
    }
    //  Store significant digits
    IntPtrDiff decimalPointPosition;
    String significantDigits = realToFloatingPointDigits(Math::abs(value), formatProperties.getRadix(), significantDigitCount, decimalPointPosition);
    //  Count how many digits are actually significant
    while (significantDigits.endsWith('0'))
    {
        significantDigits = significantDigits.left(significantDigits.length() - 1);
    }
    significantDigitCount = Math::max(static_cast<IntPtrDiff>(significantDigits.length()), 1);
    //  Determine which format to use
    RealFormatProperties::Format format = formatProperties.getFormat();
    if (format == RealFormatProperties::Format_AutoSelect)
    {   //  Auto select format
        if (decimalPointPosition <= 0 || decimalPointPosition > significantDigitCount)
        {
            format = RealFormatProperties::Format_FloatingPoint;
        }
        else
        {
            format = RealFormatProperties::Format_FixedPoint;
        }
    }
    //  Prepare the value string
    String valueString;
    if (format == RealFormatProperties::Format_FloatingPoint)
    {   //  Floating-point
        valueString = formatFloatingPointRealValue<BinaryFixed, IntPtrDiff>(value, significantDigitCount, formatProperties, _decimalPointChar, _separatorChar);
    }
    else
    {   //  Fixed-point
        if (decimalPointPosition > significantDigitCount)
        {
            significantDigitCount = decimalPointPosition;
        }
        UIntSize significantDigitsBeforeDecimalPoint;
        if (decimalPointPosition > 0)
        {
            significantDigitsBeforeDecimalPoint = decimalPointPosition;
        }
        else
        {
            significantDigitsBeforeDecimalPoint = 1;
            significantDigitCount += (-decimalPointPosition + 1);
        }
        valueString = formatFixedPointRealValue<BinaryFixed, IntPtrDiff>(value, significantDigitCount, significantDigitsBeforeDecimalPoint, formatProperties, _decimalPointChar, _separatorChar);
    }
    //  Prepend the sign if necessary
    if (value < 0)
    {
        valueString = String("-") + valueString;
    }
    else if (formatProperties.getMandatorySign())
    {
        valueString = String("+") + valueString;
    }
    //  Post-format
    return _postFormatRealValue(valueString, formatProperties);
}

String _impl::_DefaultNumberFormat::format(const BinaryFloat & value, const RealFormatProperties & formatProperties) const
{
    switch (value.getCategory())
    {
        case BinaryFloat::ValueCategory_SignalingNan:
            return _postFormatRealValue(String("#SNAN"), formatProperties);
        case BinaryFloat::ValueCategory_QuietNan:
            return _postFormatRealValue(String("#QNAN"), formatProperties);
        case BinaryFloat::ValueCategory_NegativeInfinity:
            return _postFormatRealValue(String("#-INF"), formatProperties);
        case BinaryFloat::ValueCategory_PositiveInfinity:
            return _postFormatRealValue(String("#+INF"), formatProperties);
        default:
            break;
    }
    //  Store significant digits - 16 max
    int decimalPointPosition;
    String significantDigits = realToFloatingPointDigits(Math::abs(value), formatProperties.getRadix(), 16, decimalPointPosition);
    //  Count how many digits are actually significant
    while (significantDigits.endsWith('0'))
    {
        significantDigits = significantDigits.left(significantDigits.length() - 1);
    }
    int significantDigitCount = Math::max(static_cast<int>(significantDigits.length()), 1);
    //  Determine which format to use
    RealFormatProperties::Format format = formatProperties.getFormat();
    if (format == RealFormatProperties::Format_AutoSelect)
    {   //  Auto select format
        if (decimalPointPosition <= 0 || decimalPointPosition > significantDigitCount)
        {
            format = RealFormatProperties::Format_FloatingPoint;
        }
        else
        {
            format = RealFormatProperties::Format_FixedPoint;
        }
    }
    //  Prepare the value string
    String valueString;
    if (format == RealFormatProperties::Format_FloatingPoint)
    {   //  Floating-point
        valueString = formatFloatingPointRealValue<BinaryFloat, IntPtrDiff>(value, significantDigitCount, formatProperties, _decimalPointChar, _separatorChar);
    }
    else
    {   //  Fixed-point
        if (decimalPointPosition > significantDigitCount)
        {
            significantDigitCount = decimalPointPosition;
        }
        UIntSize significantDigitsBeforeDecimalPoint;
        if (decimalPointPosition > 0)
        {
            significantDigitsBeforeDecimalPoint = decimalPointPosition;
        }
        else
        {
            significantDigitsBeforeDecimalPoint = 1;
            significantDigitCount += (-decimalPointPosition + 1);
        }
        valueString = formatFixedPointRealValue<BinaryFloat, IntPtrDiff>(value, significantDigitCount, significantDigitsBeforeDecimalPoint, formatProperties, _decimalPointChar, _separatorChar);
    }
    //  Prepend the sign if necessary
    if (value < 0)
    {
        valueString = String("-") + valueString;
    }
    else if (formatProperties.getMandatorySign())
    {
        valueString = String("+") + valueString;
    }
    //  Post-format
    return _postFormatRealValue(valueString, formatProperties);
}

String _impl::_DefaultNumberFormat::format(const DecimalFixed & value, const RealFormatProperties & formatProperties) const
{
    //  Determine the number of significant digits to show in the output value
    IntPtrDiff significantDigitCount;
    if (formatProperties.getRadix() >= 10)
    {
        significantDigitCount = value.getMagnitude().getRequiredDigits();
    }
    else if (formatProperties.getRadix() >= 8)
    {
        significantDigitCount = value.getMagnitude().getRequiredDigits() * 2;
    }
    else if (formatProperties.getRadix() >= 4)
    {
        significantDigitCount = value.getMagnitude().getRequiredDigits() * 3;
    }
    else
    {
        significantDigitCount = value.getMagnitude().getRequiredDigits() * 4;
    }
    if (significantDigitCount == 0)
    {
        significantDigitCount = 1;
    }
    //  Store significant digits
    IntPtrDiff decimalPointPosition;
    String significantDigits = realToFloatingPointDigits(Math::abs(value), formatProperties.getRadix(), significantDigitCount, decimalPointPosition);
    //  Count how many digits are actually significant
    while (significantDigits.endsWith('0'))
    {
        significantDigits = significantDigits.left(significantDigits.length() - 1);
    }
    significantDigitCount = Math::max(static_cast<IntPtrDiff>(significantDigits.length()), 1);
    //  Determine which format to use
    RealFormatProperties::Format format = formatProperties.getFormat();
    if (format == RealFormatProperties::Format_AutoSelect)
    {   //  Auto select format
        if (decimalPointPosition <= 0 || decimalPointPosition > significantDigitCount)
        {
            format = RealFormatProperties::Format_FloatingPoint;
        }
        else
        {
            format = RealFormatProperties::Format_FixedPoint;
        }
    }
    //  Prepare the value string
    String valueString;
    if (format == RealFormatProperties::Format_FloatingPoint)
    {   //  Floating-point
        valueString = formatFloatingPointRealValue<DecimalFixed, IntPtrDiff>(value, significantDigitCount, formatProperties, _decimalPointChar, _separatorChar);
    }
    else
    {   //  Fixed-point
        if (decimalPointPosition > significantDigitCount)
        {
            significantDigitCount = decimalPointPosition;
        }
        UIntSize significantDigitsBeforeDecimalPoint;
        if (decimalPointPosition > 0)
        {
            significantDigitsBeforeDecimalPoint = decimalPointPosition;
        }
        else
        {
            significantDigitsBeforeDecimalPoint = 1;
            significantDigitCount += (-decimalPointPosition + 1);
        }
        valueString = formatFixedPointRealValue<DecimalFixed, IntPtrDiff>(value, significantDigitCount, significantDigitsBeforeDecimalPoint, formatProperties, _decimalPointChar, _separatorChar);
    }
    //  Prepend the sign if necessary
    if (value < 0)
    {
        valueString = String("-") + valueString;
    }
    else if (formatProperties.getMandatorySign())
    {
        valueString = String("+") + valueString;
    }
    //  Post-format
    return _postFormatRealValue(valueString, formatProperties);
}

String _impl::_DefaultNumberFormat::format(const DecimalFloat & value, const RealFormatProperties & formatProperties) const
{
    switch (value.getCategory())
    {
        case DecimalFloat::ValueCategory_SignalingNan:
            return _postFormatRealValue(String("#SNAN"), formatProperties);
        case DecimalFloat::ValueCategory_QuietNan:
            return _postFormatRealValue(String("#QNAN"), formatProperties);
        case DecimalFloat::ValueCategory_NegativeInfinity:
            return _postFormatRealValue(String("#-INF"), formatProperties);
        case DecimalFloat::ValueCategory_PositiveInfinity:
            return _postFormatRealValue(String("#+INF"), formatProperties);
        default:
            break;
    }
    //  Store significant digits - 16 max
    int decimalPointPosition;
    String significantDigits = realToFloatingPointDigits(Math::abs(value), formatProperties.getRadix(), 16, decimalPointPosition);
    //  Count how many digits are actually significant
    while (significantDigits.endsWith('0'))
    {
        significantDigits = significantDigits.left(significantDigits.length() - 1);
    }
    int significantDigitCount = Math::max(static_cast<int>(significantDigits.length()), 1);
    //  Determine which format to use
    RealFormatProperties::Format format = formatProperties.getFormat();
    if (format == RealFormatProperties::Format_AutoSelect)
    {   //  Auto select format
        if (decimalPointPosition <= 0 || decimalPointPosition > significantDigitCount)
        {
            format = RealFormatProperties::Format_FloatingPoint;
        }
        else
        {
            format = RealFormatProperties::Format_FixedPoint;
        }
    }
    //  Prepare the value string
    String valueString;
    if (format == RealFormatProperties::Format_FloatingPoint)
    {   //  Floating-point
        valueString = formatFloatingPointRealValue<DecimalFloat, IntPtrDiff>(value, significantDigitCount, formatProperties, _decimalPointChar, _separatorChar);
    }
    else
    {   //  Fixed-point
        if (decimalPointPosition > significantDigitCount)
        {
            significantDigitCount = decimalPointPosition;
        }
        UIntSize significantDigitsBeforeDecimalPoint;
        if (decimalPointPosition > 0)
        {
            significantDigitsBeforeDecimalPoint = decimalPointPosition;
        }
        else
        {
            significantDigitsBeforeDecimalPoint = 1;
            significantDigitCount += (-decimalPointPosition + 1);
        }
        valueString = formatFixedPointRealValue<DecimalFloat, IntPtrDiff>(value, significantDigitCount, significantDigitsBeforeDecimalPoint, formatProperties, _decimalPointChar, _separatorChar);
    }
    //  Prepend the sign if necessary
    if (value < 0)
    {
        valueString = String("-") + valueString;
    }
    else if (formatProperties.getMandatorySign())
    {
        valueString = String("+") + valueString;
    }
    //  Post-format
    return _postFormatRealValue(valueString, formatProperties);
}

//////////
//  Implementation helpers
String _impl::_DefaultNumberFormat::_postFormatIntegerValue(String valueString, bool valueIsNegative, const IntegerFormatProperties & formatProperties) const
{
    //  Make sure at least the required minimum number of digits is formatted
    if (formatProperties.getMinimumDigits() != 0)
    {
        while (valueString.length() < formatProperties.getMinimumDigits())
        {
            valueString = String("0") + valueString;
        }
    }
    //  Make sure the required maximum number of digits is not exceeded
    if (formatProperties.getMaximumDigits() != 0 &&
        valueString.length() > formatProperties.getMaximumDigits())
    {   //  Too many digits !
        valueString = createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMaximumDigits());
    }
    //  Insert separators if necessary
    if (formatProperties.getUseSeparator())
    {
        for (UIntSize i = valueString.length(); i > 3; i -= 3)
        {
            valueString.insert(_separatorChar, i - 3);
        }
    }
    //  Prepend prefix if necessary
    if (formatProperties.getUsePrefix())
    {
        switch (formatProperties.getRadix())
        {
            case 8:
                if (valueString[0] != '0')
                {
                    valueString = String("0") + valueString;
                }
                break;
            case 16:
                valueString = String("0x") + valueString;
                break;
        }
    }
    //  Prepend sign if necessary
    if (valueIsNegative)
    {
        valueString = String("-") + valueString;
    }
    else if (formatProperties.getMandatorySign())
    {
        valueString = String("+") + valueString;
    }
    //  Pad to minimum width
    if (formatProperties.getMinimumWidth() != 0)
    {
        switch (formatProperties.getAlignment())
        {
            case IntegerFormatProperties::Alignment_Left:
                while (valueString.length() < formatProperties.getMinimumWidth())
                {
                    valueString += formatProperties.getFillerChar();
                }
                break;
            case IntegerFormatProperties::Alignment_Right:
                while (valueString.length() < formatProperties.getMinimumWidth())
                {
                    valueString = String(formatProperties.getFillerChar()) + valueString;
                }
                break;
            default:
                require(false);
        }
    }
    //  Make sure the required maximum width is not exceeded
    if (formatProperties.getMaximumWidth() != 0 &&
        valueString.length() > formatProperties.getMaximumWidth())
    {
        valueString = createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMaximumWidth());
    }
    //  Enforce letter case
    switch (formatProperties.getLetterCase())
    {
        case IntegerFormatProperties::LetterCase_Lower:
            valueString.lowerCase();
            break;
        case IntegerFormatProperties::LetterCase_Upper:
            valueString.upperCase();
            break;
        case IntegerFormatProperties::LetterCase_Title:
            valueString.titleCase();
            break;
        default:
            require(false);
    }
    //  Done
    return valueString;
}

String _impl::_DefaultNumberFormat::_postFormatRealValue(String valueString, const RealFormatProperties & formatProperties)
{
    //  Pad value string to minimum width
    if (formatProperties.getMinimumWidth() != 0)
    {
        switch (formatProperties.getAlignment())
        {
            case RealFormatProperties::Alignment_Left:
                while (valueString.length() < formatProperties.getMinimumWidth())
                {
                    valueString += formatProperties.getFillerChar();
                }
                break;
            case RealFormatProperties::Alignment_Right:
                while (valueString.length() < formatProperties.getMinimumWidth())
                {
                    valueString = String(formatProperties.getFillerChar()) + valueString;
                }
                break;
            default:
                require(false);
        }
    }
    //  Make sure the required maximum width is not exceeded
    if (formatProperties.getMaximumWidth() != 0 &&
        valueString.length() > formatProperties.getMaximumWidth())
    {
        valueString = createReplicatedString(formatProperties.getErrorChar(), formatProperties.getMaximumWidth());
    }
    //  Enforce letter case
    switch (formatProperties.getLetterCase())
    {
        case RealFormatProperties::LetterCase_Lower:
            valueString.lowerCase();
            break;
        case RealFormatProperties::LetterCase_Upper:
            valueString.upperCase();
            break;
        case RealFormatProperties::LetterCase_Title:
            valueString.titleCase();
            break;
        default:
            require(false);
    }
    //  done
    return valueString;
}

//  End of DefaultNumberFormat.cpp
