//
//  EmulatedSettings.cpp - Implementation of the carl::_impl::_EmulatedSettings class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:52 $
//  $Revision: 1.12 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "EmulatedSettingsImpl.hpp"

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
//////////
//  Singleton
_impl::_EmulatedSettings::_EmulatedSettings(const String & fileName)
    :   _rootNode(NULL, String()),
        _fileName((PathName(fileName)).toAbsolute().toString())
{
    try
    {
        FileSystem::createDirectories((PathName(_fileName)).getParent());
        _pFile = new RandomAccessFile(_fileName, RandomAccessFile::OpenMode_CreateOrOpen, RandomAccessFile::AccessMode_ReadWrite);
    }
    catch (IOException)
    {
        _pFile = NULL;
    }
}

_impl::_EmulatedSettings::~_EmulatedSettings()
{
    delete _pFile;
}

_impl::_EmulatedSettings * _impl::_EmulatedSettings::getSharedInstance()
{
    static _EmulatedSettings sharedInstance(_prepareSharedInstanceFileName());
    return &sharedInstance;
}

_impl::_EmulatedSettings * _impl::_EmulatedSettings::getPrivateInstance()
{
    static _EmulatedSettings privateInstance(_preparePrivateInstanceFileName());
    return &privateInstance;
}

//////////
//  Operations
bool _impl::_EmulatedSettings::nodeExists(const SettingsNode & node)
{
    _load();
    Array<String> nodePath = _parseNodePath(node.toString());
    if (!_isValidNodePath(nodePath))
    {
        _store();
        return false;
    }
    bool result = _findNode(nodePath, false) != NULL;
    _store();
    return result;
}

bool _impl::_EmulatedSettings::createNode(const SettingsNode & node)
{
    _load();
    Array<String> nodePath = _parseNodePath(node.toString());
    if (!_isValidNodePath(nodePath))
    {
        _store();
        return false;
    }
    bool result = _findNode(nodePath, true) != NULL;
    _modifiedSinceLoad = true;
    _store();
    return result;
}

bool _impl::_EmulatedSettings::destroyNode(const SettingsNode & node)
{
    _load();
    Array<String> nodePath = _parseNodePath(node.toString());
    if (!_isValidNodePath(nodePath))
    {
        _store();
        return true;
    }
    else
    {
        _EmulatedSettingsNode * pDeletedNode = _findNode(nodePath, false);
        if (pDeletedNode == NULL)
        {
            _store();
            return true;
        }
        else if (pDeletedNode->getParent() == NULL)
        {
            _store();
            return false;
        }
        else
        {
            pDeletedNode->getParent()->destroyChildNode(pDeletedNode->getName());
            _modifiedSinceLoad = true;
            _store();
            return true;
        }
    }
}

Array<SettingsNode> _impl::_EmulatedSettings::getChildNodes(const SettingsNode & node)
{
    _load();
    Array<String> nodePath = _parseNodePath(node.toString());
    if (!_isValidNodePath(nodePath))
    {
        _store();
        return Array<SettingsNode>();
    }
    else
    {
        _EmulatedSettingsNode * pNode = _findNode(nodePath, false);
        if (pNode == NULL)
        {
            _store();
            return Array<SettingsNode>();
        }
        else
        {
            Array<SettingsNode> result;
            for (UIntSize i = 0; i < pNode->getChildNodeCount(); i++)
            {
                result.append(SettingsNode(pNode->getChildNodeAt(i)->getPath()));
            }
            _store();
            return result;
        }
    }
}

bool _impl::_EmulatedSettings::valueExists(const SettingsNode & node, const String & valueName)
{
    _load();
    Array<String> nodePath = _parseNodePath(node.toString());
    if (!_isValidNodePath(nodePath))
    {
        _store();
        return false;
    }
    else
    {
        _EmulatedSettingsNode * pNode = _findNode(nodePath, false);
        if (pNode == NULL)
        {
            _store();
            return false;
        }
        else
        {
            bool result = pNode->findValue(valueName) != NULL;
            _store();
            return result;
        }
    }
}

Array<String> _impl::_EmulatedSettings::getValueNames(const SettingsNode & node)
{
    _load();
    Array<String> nodePath = _parseNodePath(node.toString());
    if (!_isValidNodePath(nodePath))
    {
        _store();
        return Array<String>();
    }
    else
    {
        _EmulatedSettingsNode * pNode = _findNode(nodePath, false);
        if (pNode == NULL)
        {
            _store();
            return Array<String>();
        }
        else
        {
            Array<String> result;
            for (UIntSize i = 0; i < pNode->getValueCount(); i++)
            {
                result.append(pNode->getValueAt(i)->getName());
            }
            _store();
            return result;
        }
    }
}

bool _impl::_EmulatedSettings::destroyValue(const SettingsNode & node, const String & valueName)
{
    _load();
    Array<String> nodePath = _parseNodePath(node.toString());
    if (!_isValidNodePath(nodePath))
    {
        _store();
        return false;
    }
    else
    {
        _EmulatedSettingsNode * pNode = _findNode(nodePath, false);
        if (pNode != NULL && pNode->findValue(valueName) != NULL)
        {
            pNode->destroyValue(valueName);
            _modifiedSinceLoad = true;
        }
        _store();
        return true;
    }
}

String _impl::_EmulatedSettings::getValue(const SettingsNode & node, const String & valueName, const String & defaultValue)
{
    _load();
    Array<String> nodePath = _parseNodePath(node.toString());
    if (!_isValidNodePath(nodePath))
    {
        _store();
        return defaultValue;
    }
    else
    {
        _EmulatedSettingsNode * pNode = _findNode(nodePath, false);
        if (pNode != NULL)
        {
            _EmulatedSettingsValue * pValue = pNode->findValue(valueName);
            if (pValue != NULL)
            {
                String result = pValue->getValue();
                _store();
                return result;
            }
        }
        _store();
        return defaultValue;
    }
}

bool _impl::_EmulatedSettings::setValue(const SettingsNode & node, const String & valueName, const String & value)
{
    _load();
    Array<String> nodePath = _parseNodePath(node.toString());
    if (!_isValidNodePath(nodePath))
    {
        _store();
        return false;
    }
    else
    {
        _EmulatedSettingsNode * pNode = _findNode(nodePath, true);
        if (pNode == NULL)
        {
            _store();
            return false;
        }
        else
        {
            _EmulatedSettingsValue * pValue = pNode->findValue(valueName);
            if (pValue != NULL)
            {
                pValue->setValue(value);
            }
            else
            {
                pNode->createValue(valueName, value);
            }
            _modifiedSinceLoad = true;
            _store();
            return true;
        }
    }
}

//////////
//  Implementation helpers
String _impl::_EmulatedSettings::_prepareSharedInstanceFileName()
{
    String result;
    if (Process::getCurrentProcess()->getEnvironment()->hasVariable("CARL_1_0_SHARED_SETTINGS"))
    {
        result = Process::getCurrentProcess()->getEnvironment()->getVariableValue("CARL_1_0_SHARED_SETTINGS", "").trimmed();
    }
    if (result.length() == 0)
    {
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        result = "C:\\CARL_1_0\\shared_settings.dat";
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        result = "/usr/CARL_1_0/shared_settings.dat";
#else
    #error Unsupported host
#endif
    }
    return result;
}

String _impl::_EmulatedSettings::_preparePrivateInstanceFileName()
{
    String result;
    if (Process::getCurrentProcess()->getEnvironment()->hasVariable("CARL_1_0_PRIVATE_SETTINGS"))
    {
        result = Process::getCurrentProcess()->getEnvironment()->getVariableValue("CARL_1_0_PRIVATE_SETTINGS", "").trimmed();
    }
    if (result.length() == 0)
    {
        result = PathName(PathName(FileSystem::getHomeDirectory(), "CARL_1_0"), "private_settings.dat").toString();
    }
    return result;
}

_impl::_EmulatedSettingsNode * _impl::_EmulatedSettings::_findNode(const Array<String> & path, bool createIfNotFound)
{
    _EmulatedSettingsNode * pNode = &_rootNode;
    for (UIntSize i = 0; i < path.size(); i++)
    {
        _EmulatedSettingsNode * pChildNode = pNode->findChildNode(path[i]);
        if (pChildNode == NULL)
        {
            if (createIfNotFound)
            {
                pChildNode = pNode->createChildNode(path[i]);
            }
            else
            {
                return NULL;
            }
        }
        pNode = pChildNode;
    }
    return pNode;
}

Array<String> _impl::_EmulatedSettings::_parseNodePath(const String & path)
{
    Array<String> result;

    if (path != '/')
    {
        for (UIntSize scan = 0; scan < path.length(); )
        {
            UIntSize nextSeparatorIndex;
            String pathComponent;
            if (path.findFirst('/', nextSeparatorIndex, scan + 1))
            {   //  There IS a next path component
                pathComponent = path.mid(scan + 1, nextSeparatorIndex - scan - 1);
                scan = nextSeparatorIndex;
            }
            else
            {   //  This is the last path component
                pathComponent = path.mid(scan + 1);
                scan = path.length();
            }
            if (pathComponent.length() != 0)
            {
                result.append(pathComponent);
            }
        }
    }
    return result;
}

bool _impl::_EmulatedSettings::_isValidNodePath(const Array<String> & path)
{
    for (UIntSize i = 0; i < path.size(); i++)
    {
        UIntSize findPosition;
        if (path[i].trimmed().length() != path[i].length() ||
            path[i].length() == 0 ||
            path[i].findFirst('/', findPosition) ||
            path[i].findFirst('\\', findPosition))
        {
            return false;
        }
    }
    return true;
}

void _impl::_EmulatedSettings::_load()
{
    if (_pFile == NULL)
    {
        return;
    }

    //  Lock settings file
    try
    {
        _pFile->lock(0, 1);
    }
    catch (IOException)
    {
    }

    //  Read settings
    try
    {
        _pFile->setCurrentPosition(RandomAccessFile::SeekOrigin_Start, 0);
        BufferedInputStream bis(_pFile);
        DataInputStream dis(&bis);
        _rootNode.read(dis);
        dis.close();
        bis.close();
    }
    catch (IOException)
    {
    }

    //  Done
    _modifiedSinceLoad = false;
}

void _impl::_EmulatedSettings::_store()
{
    if (_pFile == NULL)
    {
        return;
    }

    //  Write settings
    if (_modifiedSinceLoad)
    {
        try
        {
            _pFile->setCurrentPosition(RandomAccessFile::SeekOrigin_Start, 0);
            BufferedOutputStream bos(_pFile);
            DataOutputStream dos(&bos);
            _rootNode.write(dos);
            dos.close();
            bos.close();
        }
        catch (IOException)
        {
        }
    }

    //  Unlock settings file
    try
    {
        _pFile->unlock(0, 1);
    }
    catch (IOException)
    {
    }
}

#else
    #error Unsupported platform
#endif

//  End of EmulatedSettings.cpp
