//
//  EmulatedSettingsImpl.hpp - carl::_impl::_EmulatedSettingsImpl interface
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:52 $
//  $Revision: 1.7 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#ifndef __EmulatedSettingsImpl_hpp_included__
#define __EmulatedSettingsImpl_hpp_included__

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
namespace carl
{
    namespace _impl
    {
        //////////
        //  Settings tree emulated via file system persistemcy
        class CARL_PUBLIC _EmulatedSettingsValue
        {
            //////////
            //  Can't assign or copy construct
        private:
            _EmulatedSettingsValue(const _EmulatedSettingsValue &);
            void                    operator = (const _EmulatedSettingsValue &);

            //////////
            //  Construction/destruction
        public:
            _EmulatedSettingsValue(const String & name, const String & value);
            ~_EmulatedSettingsValue();

            //////////
            //  Operations
        public:
            String                          getName() const;
            String                          getValue() const;
            void                            setValue(const String & value);
            void                            write(IDataOutput & dos) const throw(IOException);
            void                            read(IDataInput & dis) throw(IOException);

            //////////
            //  Implementation
        private:
            String                          _name;
            String                          _value;
        };

        class CARL_PUBLIC _EmulatedSettingsNode
        {
            //////////
            //  Can't assign or copy construct
        private:
            _EmulatedSettingsNode(const _EmulatedSettingsNode &);
            void                    operator = (const _EmulatedSettingsNode &);

            //////////
            //  Construction/destruction
        public:
            _EmulatedSettingsNode(_EmulatedSettingsNode * pParent, const String & name);
            ~_EmulatedSettingsNode();

            //////////
            //  Operations
        public:
            _EmulatedSettingsNode *         getParent() const;
            String                          getName() const;
            String                          getPath() const;
            UIntSize                        getChildNodeCount() const;
            _EmulatedSettingsNode *         getChildNodeAt(UIntSize index) const;
            _EmulatedSettingsNode *         findChildNode(const String & name) const;
            _EmulatedSettingsNode *         createChildNode(const String & name);
            void                            destroyChildNode(const String & name);
            UIntSize                        getValueCount() const;
            _EmulatedSettingsValue *        getValueAt(UIntSize index) const;
            _EmulatedSettingsValue *        findValue(const String & name) const;
            _EmulatedSettingsValue *        createValue(const String & name, const String & value);
            void                            destroyValue(const String & name);
            void                            write(IDataOutput & dos) const throw(IOException);
            void                            read(IDataInput & dis) throw(IOException);

            //////////
            //  Implementation
        private:
            _EmulatedSettingsNode *         _pParent;   //  NULL == this is a root node
            String                          _name;
            Array<_EmulatedSettingsNode*>   _childNodes;
            Array<_EmulatedSettingsValue*>  _values;
        };

        class CARL_PUBLIC _EmulatedSettings
        {
            //////////
            //  Can't assign or copy construct
        private:
            _EmulatedSettings(const _EmulatedSettings &);
            void                            operator = (const _EmulatedSettings &);

            //////////
            //  Singleton
        private:
            explicit _EmulatedSettings(const String & fileName);
            ~_EmulatedSettings();
        public:
            static _EmulatedSettings *      getSharedInstance();
            static _EmulatedSettings *      getPrivateInstance();

            //////////
            //  Operations
        public:
            bool                            nodeExists(const SettingsNode & node);
            bool                            createNode(const SettingsNode & node);
            bool                            destroyNode(const SettingsNode & node);
            Array<SettingsNode>             getChildNodes(const SettingsNode & node);
            bool                            valueExists(const SettingsNode & node, const String & valueName);
            Array<String>                   getValueNames(const SettingsNode & node);
            bool                            destroyValue(const SettingsNode & node, const String & valueName);
            String                          getValue(const SettingsNode & node, const String & valueName, const String & defaultValue);
            bool                            setValue(const SettingsNode & node, const String & valueName, const String & value);

            //////////
            //  Implementation
        private:
            _EmulatedSettingsNode           _rootNode;
            String                          _fileName;
            RandomAccessFile *              _pFile;
            bool                            _modifiedSinceLoad;

            static String                   _prepareSharedInstanceFileName();
            static String                   _preparePrivateInstanceFileName();

            _EmulatedSettingsNode *         _findNode(const Array<String> & path, bool createIfNotFound);
            Array<String>                   _parseNodePath(const String & path);
            static bool                     _isValidNodePath(const Array<String> & path);
            void                            _load();
            void                            _store();
        };
    }
}
#else
    #error Unsupported platform
#endif

#endif  //  __EmulatedSettingsImpl_hpp_included__
//  End of EmulatedSettingsImpl.hpp
