//
//  EncryptOutputStream.cpp - Implementation of the carl::EncryptOutputStream class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:52 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

//////////
//  Construction/destruction
EncryptOutputStream::EncryptOutputStream(IByteOutput * pTarget, StreamEncryptor * pStreamEncryptor, bool autoDeleteStreamEncryptor /*= false*/)
    :   _pTarget(pTarget),
        _pStreamEncryptor(pStreamEncryptor),
        _autoDeleteStreamEncryptor(autoDeleteStreamEncryptor),
        _pBlockEncryptor(NULL),
        _pPaddingAlgorithm(NULL),
        _autoDeleteBlockEncryptor(false),
        _bitsReadyInInputBuffer(0),
        _bitsReadyInOutputBuffer(0),
        _largestSupportedInputBlockSize(0)
{
    require(pTarget != NULL);
    require(pStreamEncryptor != NULL);
}

EncryptOutputStream::EncryptOutputStream(IByteOutput * pTarget, BlockEncryptor * pBlockEncryptor, PaddingAlgorithm * pPaddingAlgorithm, bool autoDeleteBlockEncryptor /*= false*/)
    :   _pTarget(pTarget),
        _pStreamEncryptor(NULL),
        _autoDeleteStreamEncryptor(false),
        _pBlockEncryptor(pBlockEncryptor),
        _pPaddingAlgorithm(pPaddingAlgorithm),
        _autoDeleteBlockEncryptor(autoDeleteBlockEncryptor),
        _bitsReadyInInputBuffer(0),
        _bitsReadyInOutputBuffer(0),
        _largestSupportedInputBlockSize(0)
{
    require(pTarget != NULL);
    require(pBlockEncryptor != NULL);
    require(pPaddingAlgorithm != NULL);

    BlockCipher * pBlockCipher = dynamic_cast<BlockCipher*>(pBlockEncryptor->getCipher());
    Array<UIntSize> supportedBlockSizes = pBlockCipher->getSupportedBlockLenghts();
    for (UIntSize i = 0; i < supportedBlockSizes.size(); i++)
    {
        _largestSupportedInputBlockSize = Math::max(supportedBlockSizes[i], _largestSupportedInputBlockSize);
    }
    require(_largestSupportedInputBlockSize != 0);
}

//////////
//  Operations
void EncryptOutputStream::close()
{
    if (_pTarget != NULL)
    {
        if (_autoDeleteStreamEncryptor)
        {
            delete _pStreamEncryptor;
        }
        if (_pBlockEncryptor != NULL)
        {
            _padLastBlockAndFlush();
            if (_autoDeleteBlockEncryptor)
            {
                delete _pBlockEncryptor;
            }
        }
        _pTarget = NULL;
    }
}

//////////
//  IByteOutput
void EncryptOutputStream::flush() throw(IOException)
{
    require(_pTarget != NULL);

    _pTarget->flush();
}

void EncryptOutputStream::write(UInt8 byte) throw(IOException)
{
    require(_pTarget != NULL);

    if (_pStreamEncryptor != NULL)
    {   //  Stream encryption
        UInt8 encryptedByte = 0;
        for (UIntSize i = 0; i < System::BitsPerByte; i++)
        {
            int cleartextBit = (byte >> i) & 0x01;
            int ciphertextBit = _pStreamEncryptor->encryptBit(cleartextBit) & 0x01;
            encryptedByte = static_cast<UInt8>(encryptedByte | (ciphertextBit << i));
        }
        _pTarget->write(encryptedByte);
    }
    else if (_pBlockEncryptor != NULL)
    {   //  Block encryption
        for (UIntSize i = 0; i < System::BitsPerByte; i++)
        {
            _addBitToInputBuffer((byte >> (System::BitsPerByte - i - 1)) & 0x01);
        }
    }
    else
    {
        require(false);
    }
}

void EncryptOutputStream::write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException)
{
    require(_pTarget != NULL);

    const UInt8 * pBytes = static_cast<const UInt8 *>(pBuffer);
    for (UIntSize i = 0; i < bytesToWrite; i++)
    {
        write(pBytes[i]);
    }
}

//////////
//  Helpers
void EncryptOutputStream::_addBitToInputBuffer(int bit)
{
    UIntSize byteIndex = _bitsReadyInInputBuffer / System::BitsPerByte;
    UIntSize bitIndex = _bitsReadyInInputBuffer % System::BitsPerByte;
    while (_inputBuffer.size() < byteIndex + 1)
    {
        _inputBuffer.append(0);
    }
    _inputBuffer[byteIndex] = static_cast<UInt8>(_inputBuffer[byteIndex] | ((bit & 0x01) << (System::BitsPerByte - bitIndex - 1)));
    _bitsReadyInInputBuffer++;
    if (_bitsReadyInInputBuffer == _largestSupportedInputBlockSize)
    {   //  A block is ready for encryption
        UIntSize outputBlockSize = _pBlockEncryptor->getCiphertextSizeForCleartext(_inputBuffer.getData(), _bitsReadyInInputBuffer);
        _outputBuffer.setSize(outputBlockSize / System::BitsPerByte + 1);
        _bitsReadyInOutputBuffer = _pBlockEncryptor->encryptBlock(_inputBuffer.getData(), _bitsReadyInInputBuffer, _outputBuffer.getData());
        require(_bitsReadyInOutputBuffer == outputBlockSize);
        //  Write encrypted block
        _writeUleb(_bitsReadyInOutputBuffer);
        for (UIntSize i = 0; i < _bitsReadyInOutputBuffer; i += System::BitsPerByte)
        {
            _pTarget->write(_outputBuffer[i / System::BitsPerByte]);
        }
        _bitsReadyInInputBuffer = 0;
        _bitsReadyInOutputBuffer = 0;
        for (UIntSize i = 0; i < _inputBuffer.size(); i++)
        {
            _inputBuffer[i] = 0;
        }
    }
}

void EncryptOutputStream::_padLastBlockAndFlush()
{
    //  Calculate required padding
    UIntSize paddingBitsNeeded = _pPaddingAlgorithm->getPaddingBitsRequiredForData(_inputBuffer.getData(), _bitsReadyInInputBuffer, _largestSupportedInputBlockSize);
    require((_bitsReadyInInputBuffer + paddingBitsNeeded) % _largestSupportedInputBlockSize == 0);
    Array<UInt8> paddingBits;
    paddingBits.setSize(paddingBitsNeeded / System::BitsPerByte + 1);
    _pPaddingAlgorithm->createPaddingBitsForData(_inputBuffer.getData(), _bitsReadyInInputBuffer, _largestSupportedInputBlockSize, paddingBits.getData());
    //  Add padding to input bits
    for (UIntSize i = 0; i < paddingBitsNeeded; i++)
    {
        UIntSize byteIndex = i / System::BitsPerByte;
        UIntSize bitIndex = i % System::BitsPerByte;
        _addBitToInputBuffer((paddingBits[byteIndex] >> (System::BitsPerByte - bitIndex - 1)) & 0x01);
    }
    require(_bitsReadyInInputBuffer == 0);
    require(_bitsReadyInOutputBuffer == 0);
}

void EncryptOutputStream::_writeUleb(UIntSize value)
{
    if (value == 0)
    {
        _pTarget->write(0);
    }
    else
    {
        while (value != 0)
        {
            UInt8 lowBits = static_cast<UInt8>(value & 0x007F);
            value >>= 7;
            if (value != 0)
            {
                lowBits = static_cast<UInt8>(lowBits | 0x80);
            }
            _pTarget->write(lowBits);
        }
    }
}

//  End of EncryptOutputStream.cpp
