//
//  EncryptionManagerImpl.cpp - Implementation of the carl::_impl::_EncryptionManagerImpl class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:52 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "EncryptionImpl.hpp"

//////////
//  Singleton
_impl::_EncryptionManagerImpl::_EncryptionManagerImpl()
    :   _resourceFactory(new _EncryptionResourceMap_en(), carl::String("CARL_ENCRYPTION")),
        _mapNamesToCiphers(StandardHashFunctions::hashString),
        _mapNamesToPaddingAlgorithms(StandardHashFunctions::hashString)
{
    //  Ciphers
    registerCipher(_NullCipher::getInstance());
    registerCipher(_BlowfishCipher::getInstance());
    registerCipher(_RsaCipher::getInstance());

    //  Padding algorithms
    registerPaddingAlgorithm(_Pkcs5PaddingAlgorithm::getInstance());
}

_impl::_EncryptionManagerImpl::~_EncryptionManagerImpl()
{
}

_impl::_EncryptionManagerImpl * _impl::_EncryptionManagerImpl::getInstance()
{
    static _EncryptionManagerImpl instance;
    return &instance;
}

//////////
//  Operations
MapResourceFactory * _impl::_EncryptionManagerImpl::getResourceFactory() const
{
    return &_resourceFactory;
}

Array<Cipher*> _impl::_EncryptionManagerImpl::getAllCiphers()
{
    CriticalSectionLock lock(_guard);
    return _ciphers;
}

Cipher * _impl::_EncryptionManagerImpl::findCipher(const String & name)
{
    CriticalSectionLock lock(_guard);
    String nameUppercased = name.upperCased();
    if (_mapNamesToCiphers.containsKey(nameUppercased))
    {
        return _mapNamesToCiphers.get(nameUppercased);
    }
    return NULL;
}

Array<StreamCipher*> _impl::_EncryptionManagerImpl::getAllStreamCiphers()
{
    CriticalSectionLock lock(_guard);
    Array<StreamCipher*> result;
    for (UIntSize i = 0; i < _ciphers.size(); i++)
    {
        Cipher * pCipher = _ciphers[i];
        StreamCipher * pStreamCipher = dynamic_cast<StreamCipher*>(pCipher);
        if (pStreamCipher != NULL)
        {
            result.append(pStreamCipher);
        }
    }
    return result;
}

StreamCipher * _impl::_EncryptionManagerImpl::findStreamCipher(const String & name)
{
    CriticalSectionLock lock(_guard);
    Cipher * pCipher = findCipher(name);
    return dynamic_cast<StreamCipher*>(pCipher);
}

Array<BlockCipher*> _impl::_EncryptionManagerImpl::getAllBlockCiphers()
{
    CriticalSectionLock lock(_guard);
    Array<BlockCipher*> result;
    for (UIntSize i = 0; i < _ciphers.size(); i++)
    {
        Cipher * pCipher = _ciphers[i];
        BlockCipher * pBlockCipher = dynamic_cast<BlockCipher*>(pCipher);
        if (pBlockCipher != NULL)
        {
            result.append(pBlockCipher);
        }
    }
    return result;
}

BlockCipher * _impl::_EncryptionManagerImpl::findBlockCipher(const String & name)
{
    CriticalSectionLock lock(_guard);
    Cipher * pCipher = findCipher(name);
    return dynamic_cast<BlockCipher*>(pCipher);
}

Array<SymmetricCipher*> _impl::_EncryptionManagerImpl::getAllSymmetricCiphers()
{
    CriticalSectionLock lock(_guard);
    Array<SymmetricCipher*> result;
    for (UIntSize i = 0; i < _ciphers.size(); i++)
    {
        Cipher * pCipher = _ciphers[i];
        SymmetricCipher * pSymmetricCipher = dynamic_cast<SymmetricCipher*>(pCipher);
        if (pSymmetricCipher != NULL)
        {
            result.append(pSymmetricCipher);
        }
    }
    return result;
}

SymmetricCipher * _impl::_EncryptionManagerImpl::findSymmetricCipher(const String & name)
{
    CriticalSectionLock lock(_guard);
    Cipher * pCipher = findCipher(name);
    return dynamic_cast<SymmetricCipher*>(pCipher);
}

Array<AsymmetricCipher*> _impl::_EncryptionManagerImpl::getAllAsymmetricCiphers()
{
    CriticalSectionLock lock(_guard);
    Array<AsymmetricCipher*> result;
    for (UIntSize i = 0; i < _ciphers.size(); i++)
    {
        Cipher * pCipher = _ciphers[i];
        AsymmetricCipher * pAsymmetricCipher = dynamic_cast<AsymmetricCipher*>(pCipher);
        if (pAsymmetricCipher != NULL)
        {
            result.append(pAsymmetricCipher);
        }
    }
    return result;
}

AsymmetricCipher * _impl::_EncryptionManagerImpl::findAsymmetricCipher(const String & name)
{
    CriticalSectionLock lock(_guard);
    Cipher * pCipher = findCipher(name);
    return dynamic_cast<AsymmetricCipher*>(pCipher);
}

Array<PaddingAlgorithm*> _impl::_EncryptionManagerImpl::getAllPaddingAlgorithms()
{
    CriticalSectionLock lock(_guard);
    return _paddingAlgorithms;
}

PaddingAlgorithm * _impl::_EncryptionManagerImpl::findPaddingAlgorithm(const String & name)
{
    CriticalSectionLock lock(_guard);
    String nameUppercased = name.upperCased();
    if (_mapNamesToPaddingAlgorithms.containsKey(nameUppercased))
    {
        return _mapNamesToPaddingAlgorithms.get(nameUppercased);
    }
    return NULL;
}

void _impl::_EncryptionManagerImpl::registerCipher(Cipher * pCipher) throw (InconsistentCipherException, DuplicateCipherException)
{
    require(pCipher != NULL);

    CriticalSectionLock lock(_guard);
    Array<String> names = pCipher->getAllNames();
    for (UIntSize i = 0; i < names.size(); i++)
    {
        names[i].upperCase();
        if (_mapNamesToCiphers.containsKey(names[i]))
        {
            throw DuplicateCipherException();
        }
    }
    if (!names.contains(pCipher->getCanonicalName().upperCased()) ||
        !names.contains(pCipher->getPreferredName().upperCased()))
    {
        throw InconsistentCipherException();
    }
    //
    _ciphers.append(pCipher);
    for (UIntSize i = 0; i < names.size(); i++)
    {
        _mapNamesToCiphers.insert(names[i], pCipher);
    }
}

void _impl::_EncryptionManagerImpl::registerPaddingAlgorithm(PaddingAlgorithm * pPaddingAlgorithm) throw (InconsistentPaddingAlgorithmException, DuplicatePaddingAlgorithmException)
{
    require(pPaddingAlgorithm != NULL);

    CriticalSectionLock lock(_guard);
    Array<String> names = pPaddingAlgorithm->getAllNames();
    for (UIntSize i = 0; i < names.size(); i++)
    {
        names[i].upperCase();
        if (_mapNamesToPaddingAlgorithms.containsKey(names[i]))
        {
            throw DuplicatePaddingAlgorithmException();
        }
    }
    if (!names.contains(pPaddingAlgorithm->getCanonicalName().upperCased()) ||
        !names.contains(pPaddingAlgorithm->getPreferredName().upperCased()))
    {
        throw InconsistentPaddingAlgorithmException();
    }
    //
    _paddingAlgorithms.append(pPaddingAlgorithm);
    for (UIntSize i = 0; i < names.size(); i++)
    {
        _mapNamesToPaddingAlgorithms.insert(names[i], pPaddingAlgorithm);
    }
}

//  End of CompressionManagerImpl.cpp
