//
//  EventImpl.cpp - Implementation of the carl::_impl::_EventImpl class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:52 $
//  $Revision: 1.9 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "EventImpl.hpp"

//////////
//  Operations
void _impl::_EventImpl::initialize(bool manualReset, bool initiallySet)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _osEvent = CreateEvent(NULL, manualReset, initiallySet, NULL);
    if (_osEvent == NULL)
    {
        throw SystemError();
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _usesManualReset = manualReset;
    _isSet = initiallySet;
    _waitingThreadCount = 0;
    _pGuard = new Mutex();
#else
    #error Unsupported platform
#endif
}

void _impl::_EventImpl::deinitialize()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    CloseHandle(_osEvent);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    delete _pGuard;
#else
    #error Unsupported platform
#endif
}

void _impl::_EventImpl::waitFor()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    DWORD oldLastError = GetLastError();
    DWORD waitResult = WaitForSingleObject(_osEvent, INFINITE);
    SetLastError(oldLastError);
    switch (waitResult)
    {
        case WAIT_ABANDONED:
        case WAIT_OBJECT_0:
            return;
        default:
            throw SystemError();
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    {
        Lock lock(*_pGuard);
        _waitingThreadCount++;
    }

    for (; ; )
    {
        {
            Lock lock(*_pGuard);
            if (_isSet)
            {
                _waitingThreadCount--;
                if (!_usesManualReset)
                {
                    _isSet = false;
                }
                return;
            }
        }
        Thread::sleep(TimeSpan(0));
    }
#else
    #error Unsupported platform
#endif
}

bool _impl::_EventImpl::waitFor(const TimeSpan & timeout)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    TimeStamp waitStartTime = TimeStamp::now();
    TimeStamp waitEndTime = waitStartTime + timeout;
    for (; ; )
    {
        Int64 millisecondsToWait = timeout.getNanoseconds() / 1000000000;
        if (millisecondsToWait > 1000)
        {
            millisecondsToWait = 1000;
        }
        DWORD oldLastError = GetLastError();
        DWORD waitResult = WaitForSingleObject(_osEvent, static_cast<DWORD>(millisecondsToWait));
        SetLastError(oldLastError);
        switch (waitResult)
        {
            case WAIT_ABANDONED:
            case WAIT_OBJECT_0:
                return true;
            case WAIT_TIMEOUT:
                if (TimeStamp::now() >= waitEndTime)
                {   //  Waited for the whole timeout
                    return false;
                }
                //  Keep trying
                break;
            default:
                throw SystemError();
        }
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    TimeSpan giveUpTime = TimeSpan::getTimeSpanSinceReferencePoint() + timeout;
    {
        Lock lock(*_pGuard);
        _waitingThreadCount++;
    }

    for (; ; )
    {
        {
            Lock lock(*_pGuard);
            if (_isSet)
            {
                _waitingThreadCount--;
                if (!_usesManualReset)
                {
                    _isSet = false;
                }
                return true;
            }
        }
        if (TimeSpan::getTimeSpanSinceReferencePoint() > giveUpTime)
        {
            Lock lock(*_pGuard);
            _waitingThreadCount--;
            return false;
        }
        Thread::sleep(TimeSpan(0));
    }
#else
    #error Unsupported platform
#endif
}

void _impl::_EventImpl::release()
{
}

void _impl::_EventImpl::set()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    SetEvent(_osEvent);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    {
        Lock lock(*_pGuard);
        _isSet = true;
    }
    //  For auto-reset event we have to wait until one waiting thread is
    //  released. The release will automatically reset the event.
    if (!_usesManualReset)
    {
        for (; ; )
        {
            {
                Lock lock(*_pGuard);
                if (!_isSet)
                {
                    break;
                }
            }
            Thread::sleep(TimeSpan(0));
        }
    }
#else
    #error Unsupported platform
#endif
}

void _impl::_EventImpl::reset()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    ResetEvent(_osEvent);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Lock lock(*_pGuard);
    _isSet = false;
#else
    #error Unsupported platform
#endif
}

//  End of EventImpl.cpp
