//
//  ExtensionLoader.cpp - Implementation of the carl::ExtensionLoader class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:52 $
//  $Revision: 1.12 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include <string.h>

#if defined(CARL_LINKAGE_DYNAMIC)

//////////
//  Operations
bool ExtensionLoader::loadFromFile(const char * pFileName)
{
    require(pFileName != NULL);

    return loadFromFile(String(pFileName));
}

bool ExtensionLoader::loadFromFile(const Char * pFileName)
{
    require(pFileName != NULL);

    return loadFromFile(String(pFileName));
}

bool ExtensionLoader::loadFromFile(const String & fileName)
{
    CriticalSectionLock lock(*_getStaticGuard());
    String fullFileName = (PathName(fileName)).toAbsolute().toString();
    if (!FileSystem::CaseSensitiveFileNames)
    {
        fullFileName.upperCase();
    }
    if (getMapFileNamesToLibraries()->containsKey(fullFileName))
    {   //  Extensions from this library were already loaded
        return true;
    }
    //  Try loading the library
    Library * pLibrary = new Library();
    if (!pLibrary->load(fullFileName))
    {
        delete pLibrary;
        return false;
    }
    //  Locate extension provider
    ExtensionProviderProc extensionProviderProc = (ExtensionProviderProc)(pLibrary->resolve("CARL_GetExtensions"));
    if (extensionProviderProc == NULL)
    {
        delete pLibrary;
        return false;
    }
    //  Locate extensions
    Array<Extension *> allExtensions;
    (*extensionProviderProc)(allExtensions);
    //  Choose only extensions valid for the current CARL version
    Array<Extension *> validExtensions;
    for (UIntSize i = 0; i < allExtensions.size(); i++)
    {
        Extension * pExtension = allExtensions[i];
        if (pExtension != NULL &&
            pExtension->getBuildKey() != NULL &&
            strcmp(pExtension->getBuildKey(), CARL_BUILD_KEY) == 0)
        {
            validExtensions.append(pExtension);
        }
    }
    if (validExtensions.size() == 0)
    {
        delete pLibrary;
        return false;
    }
    //  Apply extensions
    for (UIntSize i = 0; i < validExtensions.size(); i++)
    {
        validExtensions[i]->apply();
    }
    //  Record the fact that extensions have been loaded so that extensions
    //  from the same DLL are never applied twice
    getMapFileNamesToLibraries()->insert(fullFileName, pLibrary);
    //  Done
    return true;
}

bool ExtensionLoader::loadFromFile(const PathName & fileName)
{
    return loadFromFile(fileName.toString());
}

bool ExtensionLoader::loadFromDirectory(const char * pDirectoryName, bool processSubdirectories /*= false*/)
{
    require(pDirectoryName != NULL);

    return loadFromDirectory(String(pDirectoryName), processSubdirectories);
}

bool ExtensionLoader::loadFromDirectory(const Char * pDirectoryName, bool processSubdirectories /*= false*/)
{
    require(pDirectoryName != NULL);

    return loadFromDirectory(String(pDirectoryName), processSubdirectories);
}

bool ExtensionLoader::loadFromDirectory(const String & directoryName, bool processSubdirectories /*= false*/)
{
    CriticalSectionLock lock(*_getStaticGuard());

    PathName directory(directoryName);
    String requiredExtensionFileNamePattern;
    UIntSize requiredExtensionFileNamePatternStart;

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    requiredExtensionFileNamePattern = "x86-win32-MSVC2003";
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)
    requiredExtensionFileNamePattern = "x86-win32-cygwin";
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC)
    requiredExtensionFileNamePattern = "x86-linux-gcc";
#else
    #error Unsupported platform
#endif

    Array<DirectoryEntry> directoryEntries;
    if (!FileSystem::getDirectoryContents(directory, directoryEntries))
    {
        return false;
    }
    bool someExtensionsLoaded = false;
    for (UIntSize i = 0; i < directoryEntries.size(); i++)
    {
        PathName entry(directoryEntries[i].getFullName());
        if (entry.isDirectory() && processSubdirectories)
        {
            if (loadFromDirectory(entry.toString(), true))
            {
                someExtensionsLoaded = true;
            }
        }
        else if (entry.hasExtension(Library::PreferredExtension) &&
                 entry.getFilename().findFirst(requiredExtensionFileNamePattern, requiredExtensionFileNamePatternStart) != 0)
        {
            if (loadFromFile(entry.toString()))
            {
                someExtensionsLoaded = true;
            }
        }
    }
    return someExtensionsLoaded;
}

bool ExtensionLoader::loadFromDirectory(const PathName & directoryName, bool processSubdirectories /*= false*/)
{
    return loadFromDirectory(directoryName.toString(), processSubdirectories);
}

bool ExtensionLoader::loadInstalledExtensions()
{
    CriticalSectionLock lock(*_getStaticGuard());

    bool someExtensionsLoaded = false;

    //  Load extensions defined in the application itself
    if (loadFromFile(Process::getCurrentProcess()->getApplicationFileName()))
    {
        someExtensionsLoaded = true;
    }

    //  Load from extension directories referred to by CARL_1_0_EXTENSIONS environment variable
    Array<PathName> extensionPathList = FileSystem::splitPathListString(Process::getCurrentProcess()->getEnvironment()->getVariableValue("CARL_1_0_EXTENSIONS", ""));
    for (UIntSize i = 0; i < extensionPathList.size(); i++)
    {
        if (extensionPathList[i].isDirectory())
        {
            if (loadFromDirectory(extensionPathList[i].toString()))
            {
                someExtensionsLoaded = true;
            }
        }
    }
    //  Done
    return someExtensionsLoaded;
}

//////////
//  Implementation helpers
HashMap<String, Library*> * ExtensionLoader::getMapFileNamesToLibraries()
{
    static HashMap<String, Library*> _mapFileNamesToLibraries(StandardHashFunctions::hashString);
    return &_mapFileNamesToLibraries;
}

CriticalSection * ExtensionLoader::_getStaticGuard()
{
    static CriticalSection guard;
    return &guard;
}

#endif  //  defined(CARL_LINKAGE_DYNAMIC)

//  End of Extension.cpp
