//
//  FileInputStream.cpp - Implementation of the carl::FileInputStream class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/07/28 10:40:28 $
//  $Revision: 1.14 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <io.h>
    #include <fcntl.h>
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <errno.h>
    #include <limits.h>
    #include <stdio.h>
    #include <share.h>
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)
    #include <fcntl.h>
    #include <errno.h>
    #include <unistd.h>
    #include <limits.h>
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC)
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <fcntl.h>
    #include <errno.h>
    #include <unistd.h>
    #include <limits.h>
#else
    #error Unsupported platform
#endif

//////////
//  Construction/destruction
FileInputStream::FileInputStream(const char * pFileName) throw(IOException)
{
    require(pFileName != NULL);

    _openFile(String(pFileName));
}

FileInputStream::FileInputStream(const Char * pFileName) throw(IOException)
{
    require(pFileName != NULL);

    _openFile(String(pFileName));
}

FileInputStream::FileInputStream(const String & fileName) throw(IOException)
{
    _openFile(fileName);
}

FileInputStream::FileInputStream(const PathName & pathName) throw(IOException)
{
    _openFile(pathName.toString());
}

//////////
//  Operations
void FileInputStream::close()
{
    if (_handle != -1)
    {
        int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        ::_close(_handle);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        ::close(_handle);
#else
    #error Unsupported platform
#endif
        errno = oldErrno;
        _handle = -1;
    }
}

//////////
//  IByteInput
UIntSize FileInputStream::available() throw(IOException)
{
    require(isOpen());

    int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Int64 currentPosition = _telli64(_handle);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Int64 currentPosition = lseek(_handle, 0, SEEK_CUR);
#else
    #error Unsupported platform
#endif
    if (currentPosition == -1)
    {
        errno = oldErrno;
        throw IOException();
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Int64 eofPosition = _lseeki64(_handle, 0, SEEK_END);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Int64 eofPosition = lseek(_handle, 0, SEEK_END);
#else
    #error Unsupported platform
#endif
    if (eofPosition == -1)
    {
        errno = oldErrno;
        throw IOException();
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    if (_lseeki64(_handle, currentPosition, SEEK_SET) != currentPosition)
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (lseek(_handle, currentPosition, SEEK_SET) != currentPosition)
#else
    #error Unsupported platform
#endif
    {
        errno = oldErrno;
        throw IOException();
    }
    UInt64 bytesUntilEof = eofPosition - currentPosition;
    return (bytesUntilEof > UINT_MAX) ? (UINT_MAX) : (static_cast<UIntSize>(bytesUntilEof));
}

int FileInputStream::read() throw(IOException)
{
    UInt8 b;
    if (read(&b, 1) == 0)
    {
        return -1;
    }
    return b;
}

UIntSize FileInputStream::read(void * pBuffer, UIntSize bytesToRead) throw(IOException)
{
    require(isOpen());

    //  Read in 32K chunks
    int oldErrno = errno;
    UInt8 * pBytes = static_cast<UInt8*>(pBuffer);
    UIntSize bytesRead = 0;
    while (bytesToRead != 0)
    {
        UIntSize chunkSize = (bytesToRead > 32768) ? (32768) : (bytesToRead);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        int chunkBytes = ::_read(_handle, pBytes, chunkSize);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        int chunkBytes = ::read(_handle, pBytes, chunkSize);
#else
    #error Unsupported platform
#endif
        if (chunkBytes == -1)
        {
            errno = oldErrno;
            throw IOException();
        }
        bytesToRead -= chunkSize;
        bytesRead += chunkBytes;
        pBytes += chunkBytes;
        if (static_cast<UIntSize>(chunkBytes) != chunkSize)
        {
            break;
        }
    }
    errno = oldErrno;
    return bytesRead;
}

UIntSize FileInputStream::skip(UIntSize bytesToSkip) throw(IOException)
{
    require(isOpen());

    UIntSize bytesSkipped = 0;
    while (bytesToSkip != 0)
    {
        UInt8 chunk[4096];
        UIntSize chunkSize = (bytesToSkip > sizeof(chunk)) ? (sizeof(chunk)) : (bytesToSkip);
        UIntSize chunkSkipped = read(chunk, chunkSize);
        bytesSkipped += chunkSkipped;
        bytesToSkip -= chunkSize;
        if (chunkSkipped != chunkSize)
        {
            break;
        }
    }
    return bytesSkipped;
}

//////////
//  Implementation helpers
void FileInputStream::_openFile(const String & fileName) throw(IOException)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> fileNameString;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> fileNameString;
#else
    #error Unsupported platform
#endif

    //  Store path name string
    fileNameString.setSize(fileName.length() + 1);
    if (!fileName._store(fileNameString.getData(), fileNameString.size()))
    {
        throw FileNotFoundException(fileName);
    }

    //  Open the file
    int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && (defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003))
    _handle = _wsopen(fileNameString.getData(), _O_BINARY | _O_RDONLY | _O_SEQUENTIAL, _SH_DENYWR, _S_IREAD);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    errno = _wsopen_s(&_handle, fileNameString.getData(), _O_BINARY | _O_RDONLY | _O_SEQUENTIAL, _SH_DENYWR, _S_IREAD);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC))
    _handle = open(fileNameString.getData(), _O_BINARY | _O_RDONLY);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _handle = open(fileNameString.getData(), O_RDONLY);
#else
    #error Unsupported platform
#endif
    if (_handle == -1)
    {
        switch (errno)
        {
            case ENOENT:
                errno = oldErrno;
                throw FileNotFoundException(fileName);
            default:
                errno = oldErrno;
                throw IOException();
        }
    }
    errno = oldErrno;
}

//  End of FileInputStream.cpp
