//
//  FileOutputStream.cpp - Implementation of the carl::FileOutputStream class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/07/28 10:40:28 $
//  $Revision: 1.14 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <io.h>
    #include <fcntl.h>
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <errno.h>
    #include <limits.h>
    #include <stdio.h>
    #include <share.h>
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)
    #include <fcntl.h>
    #include <errno.h>
    #include <unistd.h>
    #include <limits.h>
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC)
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <fcntl.h>
    #include <errno.h>
    #include <unistd.h>
    #include <limits.h>
#else
    #error Unsupported platform
#endif

//////////
//  Construction/destruction
FileOutputStream::FileOutputStream(const char * pFileName) throw(IOException)
{
    require(pFileName != NULL);

    _openFile(String(pFileName));
}

FileOutputStream::FileOutputStream(const Char * pFileName) throw(IOException)
{
    require(pFileName != NULL);

    _openFile(String(pFileName));
}

FileOutputStream::FileOutputStream(const String & fileName) throw(IOException)
{
    _openFile(fileName);
}

FileOutputStream::FileOutputStream(const PathName & pathName) throw(IOException)
{
    _openFile(pathName.toString());
}

//////////
//  Operations
void FileOutputStream::close()
{
    if (_handle != -1)
    {
        int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        _close(_handle);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        ::close(_handle);
#else
    #error Unsupported platform
#endif
        errno = oldErrno;
        _handle = -1;
    }
}

//////////
//  IByteOutput
void FileOutputStream::flush() throw(IOException)
{
    require(isOpen());

    int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    if (_commit(_handle) != 0)
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (fdatasync(_handle) != 0)
#else
    #error Unsupported platform
#endif
    {
        errno = oldErrno;
        throw IOException();
    }
    errno = oldErrno;
}

void FileOutputStream::write(UInt8 byte) throw(IOException)
{
    write(&byte, 1);
}

void FileOutputStream::write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException)
{
    require(isOpen());

    //  Write in 32K chunks
    int oldErrno = errno;
    const UInt8 * pBytes = static_cast<const UInt8*>(pBuffer);
    UIntSize bytesWritten = 0;
    while (bytesToWrite != 0)
    {
        UIntSize chunkSize = (bytesToWrite > 32768) ? (32768) : (bytesToWrite);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        int chunkBytes = _write(_handle, pBytes, chunkSize);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        int chunkBytes = ::write(_handle, pBytes, chunkSize);
#else
    #error Unsupported platform
#endif
        if (chunkBytes < 0 || static_cast<UIntSize>(chunkBytes) != chunkSize)
        {
            errno = oldErrno;
            throw IOException();
        }
        bytesToWrite -= chunkSize;
        bytesWritten += chunkBytes;
        pBytes += chunkSize;
    }
    errno = oldErrno;
}

//////////
//  Implementation helpers
void FileOutputStream::_openFile(const String & fileName) throw(IOException)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> fileNameString;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> fileNameString;
#else
    #error Unsupported platform
#endif

    //  Store path name string
    fileNameString.setSize(fileName.length() + 1);
    if (!fileName._store(fileNameString.getData(), fileNameString.size()))
    {
        throw FileNotFoundException(fileName);
    }

    //  Open the file
    int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && (defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003))
    _handle = _wopen(fileNameString.getData(), _O_BINARY | _O_WRONLY | _O_SEQUENTIAL | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    errno = _wsopen_s(&_handle, fileNameString.getData(), _O_BINARY | _O_WRONLY | _O_SEQUENTIAL | _O_CREAT | _O_TRUNC, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)
    _handle = open(fileNameString.getData(), _O_BINARY | _O_WRONLY | _O_CREAT | _O_TRUNC, 0666);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC)
    _handle = open(fileNameString.getData(), O_WRONLY | O_CREAT | O_TRUNC, 0666);
#else
    #error Unsupported platform
#endif
    if (_handle == -1)
    {
        errno = oldErrno;
        throw IOException();
    }
    errno = oldErrno;
}

//  End of FileOutputStream.cpp
