//
//  FileResourceFactory.cpp - Implementation of the carl::FileResourceFactory class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/08/06 16:15:54 $
//  $Revision: 1.12 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

//////////
//  _ResourceMap
FileResourceFactory::_ResourceMap::_ResourceMap(const HashMap<String, String> & mapKeysToStrings)
    :   _mapKeysToStrings(mapKeysToStrings)
{
}

String FileResourceFactory::_ResourceMap::getString(const String & key)
{
    if (_mapKeysToStrings.containsKey(key))
    {
        return _mapKeysToStrings.get(key);
    }
    throw MissingResourceError(key);
}

//////////
//  Construction/destruction
FileResourceFactory::FileResourceFactory(ResourceMap * pDefaultResourceMap, const String & resourceFileNameBase, const String & resourceFileNameExtension, const Array<String> & searchPathList)
    :   MapResourceFactory(pDefaultResourceMap),
        _resourcesPreloaded(false),
        _requiredResourceFileNamePrefix(resourceFileNameBase),
        _requiredResourceFileNameSuffix(resourceFileNameExtension),
        _searchPathList(searchPathList),
        _supportedLocales(StandardHashFunctions::hashLocale),
        _mapSupportedLocalesToDeferredResourceFileNames(StandardHashFunctions::hashLocale)
{
    require(resourceFileNameBase.length() != 0);

    if (_requiredResourceFileNameSuffix.length() != 0 &&
        _requiredResourceFileNameSuffix[0] != FileSystem::FilenameExtensionSeparator)
    {
        _requiredResourceFileNameSuffix = String(FileSystem::FilenameExtensionSeparator) + _requiredResourceFileNameSuffix;
    }
}

//////////
//  Operations
Array<Locale> FileResourceFactory::getSupportedLocales()
{
    _ensureResourcesPreloaded();
    Array<Locale> result = MapResourceFactory::getSupportedLocales();
    for (Iterator<Locale> it = _supportedLocales.enumerateElements(); !it.finished(); it.advance())
    {
        if (!result.contains(it.item()))
        {
            result.append(it.item());
        }
    }
    return result;
}

//////////
//  Overridables
MapResourceFactory::HandleMissingLocaleOutcome FileResourceFactory::handleMissingLocale(const Locale & forLocale)
{
    _ensureResourcesPreloaded();
    if (_mapSupportedLocalesToDeferredResourceFileNames.containsKey(forLocale))
    {   //  There exists a resource file for the required locale - it just
        //  has not yet been loaded
        _ResourceMap * pResourceMap = _loadResourceFile(_mapSupportedLocalesToDeferredResourceFileNames.get(forLocale));
        if (pResourceMap == NULL)
        {   //  OOPS!
            _supportedLocales.remove(forLocale);
            _mapSupportedLocalesToDeferredResourceFileNames.removeKey(forLocale);
            return HandleMissingLocaleOutcome_UseDefault;
        }
        addResourceMap(forLocale, pResourceMap);
        _mapSupportedLocalesToDeferredResourceFileNames.removeKey(forLocale);
        return HandleMissingLocaleOutcome_Retry;
    }
    return HandleMissingLocaleOutcome_UseDefault;
}

//////////
//  Helpers
void FileResourceFactory::_ensureResourcesPreloaded()
{
    if (_resourcesPreloaded)
    {
        return;
    }
    //  Find all resource files
    for (UIntSize i = 0; i < _searchPathList.size(); i++)
    {
        //  Get all entries in the current search directory...
        Array<DirectoryEntry> directoryEntries;
        if (!FileSystem::getDirectoryContents(_searchPathList[i], directoryEntries))
        {
            continue;
        }
        //  Ignore all entries that are not files or do not match resource file name templates
        for (UIntSize j = 0; j < directoryEntries.size(); j++)
        {
            if (directoryEntries[j].isDirectory())
            {
                continue;
            }
            String shortName = directoryEntries[j].getShortName();
            if (FileSystem::CaseSensitiveFileNames)
            {
                if (shortName.length() < _requiredResourceFileNamePrefix.length() + _requiredResourceFileNameSuffix.length() ||
                    !shortName.startsWith(_requiredResourceFileNamePrefix) ||
                    !shortName.endsWith(_requiredResourceFileNameSuffix))
                {
                    continue;
                }
            }
            else
            {
                if (shortName.length() < _requiredResourceFileNamePrefix.length() + _requiredResourceFileNameSuffix.length() ||
                    !shortName.startsWithIgnoreCase(_requiredResourceFileNamePrefix) ||
                    !shortName.endsWithIgnoreCase(_requiredResourceFileNameSuffix))
                {
                    continue;
                }
            }
            //  Determine the locale represented by this resource file name
            Locale locale;
            String localeString = shortName.mid(_requiredResourceFileNamePrefix.length(), shortName.length() - _requiredResourceFileNamePrefix.length() - _requiredResourceFileNameSuffix.length());
            if (localeString.length() == 0)
            {   //  Invariant locale is the default
            }
            else if (localeString.length() == 3 && localeString[0] == '_')
            {   //  "_<language>"
                locale = Locale(localeString.mid(1));
            }
            else if (localeString.length() == 6 && localeString[0] == '_' && localeString[3] == '_')
            {   //  "_<language>_<country>"
                locale = Locale(localeString.mid(1, 2), localeString.mid(4, 2));
            }
            else if (localeString.length() > 7 && localeString[0] == '_' && localeString[3] == '_' && localeString[6] == '_')
            {   //  "_<language>_<country>_<variant>"
                locale = Locale(localeString.mid(1, 2), localeString.mid(4, 2), localeString.mid(7));
            }
            else
            {   //  OOPS! Filename does not match the required pattern
                continue;
            }
            //  Don't load more than one resource file for the same locale
            _supportedLocales.insert(locale);
            if (!_mapSupportedLocalesToDeferredResourceFileNames.containsKey(locale))
            {
                _mapSupportedLocalesToDeferredResourceFileNames.insert(locale, directoryEntries[j].getFullName());
            }
        }
    }
    //
    _resourcesPreloaded = true;
}

FileResourceFactory::_ResourceMap * FileResourceFactory::_loadResourceFile(const String & fileName)
{
    try
    {
        HashMap<String, String> mapKeysToStrings(StandardHashFunctions::hashString);
        TextFileReader tfr(fileName, CharacterSet::getInstance("UTF-8"));
        String line;
        while (_readLine(tfr, line))
        {
            UIntSize scan = 0;
            //  Skip initial spaces
            _skipSpaces(line, scan);
            if (scan >= line.length() || line[scan] == ';')
            {   //  Empty line or comment
                continue;
            }
            //  Fetch key
            String key;
            if (!_fetchStringConstant(line, scan, key))
            {
                continue;
            }
            //  Skip '='
            _skipSpaces(line, scan);
            if (scan >= line.length() || line[scan] != '=')
            {
                continue;
            }
            scan++;
            _skipSpaces(line, scan);
            //  Fetch value
            String value;
            if (!_fetchStringConstant(line, scan, value))
            {
                continue;
            }
            //  Record key -> value mapping
            if (!mapKeysToStrings.containsKey(key))
            {
                mapKeysToStrings.insert(key, value);
            }
        }
        //  Done
        return new _ResourceMap(mapKeysToStrings);
    }
    catch (IOException)
    {
        return NULL;
    }
}

bool FileResourceFactory::_readLine(ITextInput & ti, String & line) throw(IOException)
{
    if (!ti.readLine(line, true))
    {
        return false;
    }
    //  Are there continuation lines ?
    for (; ; )
    {
        line.trimRight();
        if (!line.endsWith('\\'))
        {
            return true;
        }
        //  Read continuation line
        String nextLine;
        if (!ti.readLine(nextLine, true))
        {
            throw EOFException();
        }
        //  Discard everything until the leading '\\'
        nextLine.trimLeft();
        if (!nextLine.startsWith('\\'))
        {
            throw EOFException();
        }
        line = line.left(line.length() - 1) + nextLine.mid(1);
    }
}

void FileResourceFactory::_skipSpaces(const String & line, UIntSize & scan)
{
    require(scan <= line.length());

    for (; scan < line.length() && Iso10646::isSpace(line[scan]); scan++)
    {   //  All work is done in the loop header
    }
}

bool FileResourceFactory::_fetchStringConstant(const String & line, UIntSize & scan, String & value)
{
    require(scan <= line.length());

    //  Skip opening quote
    if (scan >= line.length() || line[scan] != '"')
    {
        return false;
    }
    scan++;
    //  Fetch value
    value.setLength(0);
    while (scan < line.length() && line[scan] != '"')
    {
        Char c;
        if (!_fetchCharacterLiteral(line, scan, c))
        {
            return false;
        }
        value += c;
    }
    //  Skip closing quote
    if (scan >= line.length() || line[scan] != '"')
    {
        return false;
    }
    scan++;
    //  Done
    return true;
}

bool FileResourceFactory::_fetchCharacterLiteral(const String & line, UIntSize & scan, Char & value)
{
    require(scan <= line.length());

    while (scan >= line.length())
    {   //  EOLN
        return false;
    }
    if (line[scan] != '\\')
    {   //  Character
        value = line[scan++];
        return true;
    }
    else
    {   //  Escape sequence
        //  Skip '\\'
        scan++;
        if (scan >= line.length())
        {
            return false;
        }
        //  Process escape sequence
        switch (line[scan])
        {
            case 'a':
                scan++;
                value = 7;
                return true;
            case 'b':
                scan++;
                value = 8;
                return true;
            case 'e':
                scan++;
                value = 27;
                return true;
            case 'f':
                scan++;
                value = 12;
                return true;
            case 'n':
                scan++;
                value = 10;
                return true;
            case 'r':
                scan++;
                value = 13;
                return true;
            case 't':
                scan++;
                value = 9;
                return true;
            case 'v':
                scan++;
                value = 11;
                return true;
            case '0':   case '1':   case '2':   case '3':
            case '4':   case '5':   case '6':   case '7':
                //  8-bit octal escape sequence
                {
                    //  Fetch character code
                    int numDigits;
                    value = 0;
                    for (numDigits = 0; numDigits < 3; numDigits++, scan++)
                    {
                        if (scan >= line.length() || !Iso10646::isOctalDigit(line[scan]))
                        {
                            break;
                        }
                        UInt32 digit = line[scan] - '0';
                        value = (value << 3) | digit;
                    }
                    //  Done
                    return (numDigits != 0) && (value & 0xFFFFFF00) == 0;
                }
            case 'x':   case 'X':
                //  8-bit hexadecimal escape sequence
                {
                    //  Skip 'x'
                    scan++;
                    if (scan >= line.length())
                    {
                        return false;
                    }
                    //  Fetch character code
                    int numDigits;
                    value = 0;
                    for (numDigits = 0; numDigits < 2; numDigits++, scan++)
                    {
                        if (scan >= line.length() || !Iso10646::isHexadecimalDigit(line[scan]))
                        {
                            break;
                        }
                        UInt32 digit = Iso10646::isDecimalDigit(line[scan]) ? (line[scan] - '0') : (Iso10646::toUpper(line[scan]) - 'A' + 10);
                        value = (value << 4) | digit;
                    }
                    //  Done
                    return numDigits != 0;
                }
            case 'u':   case 'U':
                {
                    //  Skip 'u'
                    scan++;
                    if (scan >= line.length())
                    {
                        return false;
                    }
                    //  Fetch character code
                    int numDigits;
                    value = 0;
                    for (numDigits = 0; numDigits < 4; numDigits++, scan++)
                    {
                        if (scan >= line.length() || !Iso10646::isHexadecimalDigit(line[scan]))
                        {
                            break;
                        }
                        UInt32 digit = Iso10646::isDecimalDigit(line[scan]) ? (line[scan] - '0') : (Iso10646::toUpper(line[scan]) - 'A' + 10);
                        value = (value << 4) | digit;
                    }
                    //  Done
                    return numDigits != 0;
                }
            case 'i':   case 'I':
                {
                    //  Skip 'i'
                    scan++;
                    if (scan >= line.length())
                    {
                        return false;
                    }
                    //  Fetch character code
                    int numDigits;
                    value = 0;
                    for (numDigits = 0; numDigits < 8; numDigits++, scan++)
                    {
                        if (scan >= line.length() || !Iso10646::isHexadecimalDigit(line[scan]))
                        {
                            break;
                        }
                        UInt32 digit = Iso10646::isDecimalDigit(line[scan]) ? (line[scan] - '0') : (Iso10646::toUpper(line[scan]) - 'A' + 10);
                        value = (value << 4) | digit;
                    }
                    //  Done
                    return numDigits != 0;
                }
            default:
                value = line[scan++];
                return true;
        }
    }
}

//  End of FileResourceFactory.cpp
