//
//  FileSystem.cpp - Implementation of the carl::FileSystem class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/07/28 10:40:28 $
//  $Revision: 1.18 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include <stdlib.h>
#include "DateTimeImpl.hpp"

//////////
//  File system properties
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <windows.h>
    #include <shlobj.h>
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <io.h>
    #include <stdio.h>
    #include <time.h>
    const char * FileSystem::ThisDirectoryName = ".";
    const char * FileSystem::ParentDirectoryName = "..";
    const char * FileSystem::LineSeparator = "\r\n";
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <errno.h>
    #include <unistd.h>
    #include <ctype.h>
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <sys/statvfs.h>
    #include <stdio.h>
    #include <utime.h>
    #include <dirent.h>
    #include <string.h>
    const char * FileSystem::ThisDirectoryName = ".";
    const char * FileSystem::ParentDirectoryName = "..";
    const char * FileSystem::LineSeparator = "\n";
#else
    #error Unsupported platform
#endif

bool FileSystem::isPathComponentSeparator(Char c)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    return c == '/' || c == '\\';
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)
    return c == '/' || c == '\\';
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC)
    return c == '/';
#else
    #error Unsupported platform
#endif
}

bool FileSystem::isPathListSeparator(Char c)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    return c == ';';
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return c == ':';
#else
    #error Unsupported platform
#endif
}

bool FileSystem::isFilenameExtensionSeparator(Char c)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    return c == '.';
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return c == '.';
#else
    #error Unsupported platform
#endif
}

bool FileSystem::isDeviceSeparator(Char c)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    return c == ':';
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return false;
#else
    #error Unsupported platform
#endif
}

//////////
//  File system services
PathName FileSystem::getCurrentDirectory()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<char> currentDirectory;
    DWORD oldLastError = GetLastError();
    currentDirectory.setSize(MaximumPathLength);
    DWORD dw = GetCurrentDirectory(currentDirectory.size(), currentDirectory.getData());
    SetLastError(oldLastError);
    if (dw == 0)
    {
        return PathName();
    }
    else if (dw > currentDirectory.size())
    {
        currentDirectory.setSize(dw);
        dw = GetCurrentDirectory(currentDirectory.size(), currentDirectory.getData());
        SetLastError(oldLastError);
        if (dw == 0 || dw > currentDirectory.size())
        {
            return PathName();
        }
    }
    return PathName(currentDirectory.getData());
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    char cwd[MaximumPathLength];
    int oldErrno = errno;
    if (getcwd(cwd, MaximumPathLength) == NULL)
    {
        errno = oldErrno;
        return PathName();
    }
    return PathName(cwd);
#else
    #error Unsupported platform
#endif
}

bool FileSystem::setCurrentDirectory(const char * pCurrentDirectory)
{
    require(pCurrentDirectory != NULL);

    return setCurrentDirectory(PathName(pCurrentDirectory));
}

bool FileSystem::setCurrentDirectory(const Char * pCurrentDirectory)
{
    require(pCurrentDirectory != NULL);

    return setCurrentDirectory(PathName(pCurrentDirectory));
}

bool FileSystem::setCurrentDirectory(const String & currentDirectory)
{
    return setCurrentDirectory(PathName(currentDirectory));
}

bool FileSystem::setCurrentDirectory(const PathName & currentDirectory)
{
    if (currentDirectory.isEmpty())
    {
        return false;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> currentDirectoryString;
    currentDirectoryString.setSize(currentDirectory.toString().length() + 1);
    if (!currentDirectory._store(currentDirectoryString.getData(), currentDirectoryString.size()))
    {
        return false;
    }
    DWORD oldLastError = GetLastError();
    BOOL b = SetCurrentDirectoryW(currentDirectoryString.getData());
    SetLastError(oldLastError);
    return b != 0;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> currentDirectoryString;
    currentDirectoryString.setSize(currentDirectory.toString().length() + 1);
    if (!currentDirectory._store(currentDirectoryString.getData(), currentDirectoryString.size()))
    {
        return false;
    }
    int oldErrno = errno;
    int b = chdir(currentDirectoryString.getData());
    errno = oldErrno;
    return b == 0;
#else
    #error Unsupported platform
#endif
}

PathName FileSystem::getTempDirectory()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<char> tempDirectory;
    DWORD oldLastError = GetLastError();
    tempDirectory.setSize(MaximumPathLength);
    DWORD dw = GetTempPath(tempDirectory.size(), tempDirectory.getData());
    SetLastError(oldLastError);
    if (dw == 0)
    {
        return PathName();
    }
    else if (dw > tempDirectory.size())
    {
        tempDirectory.setSize(dw);
        dw = GetTempPath(tempDirectory.size(), tempDirectory.getData());
        SetLastError(oldLastError);
        if (dw == 0 || dw > tempDirectory.size())
        {
            return PathName();
        }
    }
    return PathName(tempDirectory.getData());
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    char * pTmpdirEnvironmaneVariable = getenv("TMPDIR");
    char * pTmpEnvironmaneVariable = getenv("TMP");
    char * pTempEnvironmaneVariable = getenv("TEMP");
    if (pTmpdirEnvironmaneVariable != NULL && access(pTmpdirEnvironmaneVariable, W_OK) == 0)
    {
        return PathName(pTmpdirEnvironmaneVariable);
    }
    else if (pTmpEnvironmaneVariable != NULL && access(pTmpEnvironmaneVariable, W_OK) == 0)
    {
        return PathName(pTmpEnvironmaneVariable);
    }
    else if (pTempEnvironmaneVariable != NULL && access(pTempEnvironmaneVariable, W_OK) == 0)
    {
        return PathName(pTempEnvironmaneVariable);
    }
    else if (access("/tmp", W_OK) == 0)
    {
        return PathName("/tmp");
    }
    else
    {
        char cwd[MaximumPathLength];
        int oldErrno = errno;
        if (getcwd(cwd, MaximumPathLength) == NULL)
        {
            errno = oldErrno;
            return PathName(".");
        }
        return PathName(cwd);
    }
#else
    #error Unsupported platform
#endif
}

PathName FileSystem::getHomeDirectory()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    char homeDirectory[MaximumPathLength];

    if (SUCCEEDED(SHGetFolderPath(NULL,
                                  CSIDL_PERSONAL | CSIDL_FLAG_CREATE,
                                  NULL,
                                  SHGFP_TYPE_CURRENT,
                                  homeDirectory)))
    {
        return PathName(homeDirectory);
    }
    return PathName();
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    char * pHomeEnvironmaneVariable = getenv("HOME");
    char * pHomedriveEnvironmaneVariable = getenv("HOMEDRIVE");
    char * pHomepathEnvironmaneVariable = getenv("HOMEPATH");
    if (pHomeEnvironmaneVariable != NULL && access(pHomeEnvironmaneVariable, W_OK) == 0)
    {
        return PathName(pHomeEnvironmaneVariable);
    }
    else if (pHomedriveEnvironmaneVariable != NULL &&
             isalpha(pHomedriveEnvironmaneVariable[0]) &&
             pHomedriveEnvironmaneVariable[1] == ':' &&
             pHomedriveEnvironmaneVariable[2] == 0 &&
             pHomepathEnvironmaneVariable != NULL &&
             (pHomepathEnvironmaneVariable[0] == '/' || pHomepathEnvironmaneVariable[0] == '\\'))
    {
        return PathName(String(pHomedriveEnvironmaneVariable) + String(pHomepathEnvironmaneVariable));
    }
    else
    {
        char cwd[MaximumPathLength];
        int oldErrno = errno;
        if (getcwd(cwd, MaximumPathLength) == NULL)
        {
            errno = oldErrno;
            return PathName(".");
        }
        return PathName(cwd);
    }
#else
    #error Unsupported platform
#endif
}

bool FileSystem::createDirectory(const char * pDirectory)
{
    require(pDirectory != NULL);

    return createDirectory(PathName(pDirectory));
}

bool FileSystem::createDirectory(const Char * pDirectory)
{
    require(pDirectory != NULL);

    return createDirectory(PathName(pDirectory));
}

bool FileSystem::createDirectory(const String & directory)
{
    return createDirectory(PathName(directory));
}

bool FileSystem::createDirectory(const PathName & directory)
{
    if (directory.isEmpty())
    {
        return false;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> directoryString;
    directoryString.setSize(directory.toString().length() + 1);
    if (!directory._store(directoryString.getData(), directoryString.size()))
    {
        return false;
    }
    DWORD oldLastError = GetLastError();
    BOOL b = CreateDirectoryW(directoryString.getData(), NULL);
    SetLastError(oldLastError);
    return b != 0;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> directoryString;
    directoryString.setSize(directory.toString().length() + 1);
    if (!directory._store(directoryString.getData(), directoryString.size()))
    {
        return false;
    }
    int oldErrno = errno;
    int b = mkdir(directoryString.getData(), 0777);
    errno = oldErrno;
    return b == 0;
#else
    #error Unsupported platform
#endif
}

bool FileSystem::createDirectories(const char * pDirectory)
{
    require(pDirectory != NULL);

    return createDirectories(PathName(pDirectory));
}

bool FileSystem::createDirectories(const Char * pDirectory)
{
    require(pDirectory != NULL);

    return createDirectories(PathName(pDirectory));
}

bool FileSystem::createDirectories(const String & directory)
{
    return createDirectories(PathName(directory));
}

bool FileSystem::createDirectories(const PathName & directory)
{
    if (directory.isEmpty())
    {
        return false;
    }
    else if (directory.isDirectory())
    {   //  Directory already exists
        return true;
    }
    else if (directory.exists())
    {   //  The entity exists but is not a directory
        return false;
    }
    else if (createDirectory(directory))
    {   //  Succeeded creating the directory in one go
        return true;
    }
    else
    {   //  Create parent, then create the required directory
        PathName parent = directory.getParent();
        if (parent != directory && !createDirectories(parent))
        {
            return false;
        }
        return createDirectory(directory);
    }
}

bool FileSystem::deleteFile(const char * pFileName)
{
    require(pFileName != NULL);

    return deleteFile(PathName(pFileName));
}

bool FileSystem::deleteFile(const Char * pFileName)
{
    require(pFileName != NULL);

    return deleteFile(PathName(pFileName));
}

bool FileSystem::deleteFile(const String & fileName)
{
    return deleteFile(PathName(fileName));
}

bool FileSystem::deleteFile(const PathName & fileName)
{
    if (fileName.isEmpty())
    {
        return false;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> fileNameChars;
    fileNameChars.setSize(fileName.toString().length() + 1);
    if (!fileName._store(fileNameChars.getData(), fileNameChars.size()))
    {
        return false;
    }
    DWORD oldLastError = GetLastError();
    BOOL b = DeleteFileW(fileNameChars.getData());
    SetLastError(oldLastError);
    return b != 0;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> fileNameChars;
    fileNameChars.setSize(fileName.toString().length() + 1);
    if (!fileName._store(fileNameChars.getData(), fileNameChars.size()))
    {
        return false;
    }
    int oldErrno = errno;
    int b = unlink(fileNameChars.getData());
    errno = oldErrno;
    return b == 0;
#else
    #error Unsupported platform
#endif
}

bool FileSystem::deleteDirectory(const char * pDirectory)
{
    require(pDirectory != NULL);

    return deleteDirectory(PathName(pDirectory));
}

bool FileSystem::deleteDirectory(const Char * pDirectory)
{
    require(pDirectory != NULL);

    return deleteDirectory(PathName(pDirectory));
}

bool FileSystem::deleteDirectory(const String & directory)
{
    return deleteDirectory(PathName(directory));
}

bool FileSystem::deleteDirectory(const PathName & directory)
{
    if (directory.isEmpty())
    {
        return false;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> directoryString;
    directoryString.setSize(directory.toString().length() + 1);
    if (!directory._store(directoryString.getData(), directoryString.size()))
    {
        return false;
    }
    DWORD oldLastError = GetLastError();
    BOOL b = RemoveDirectoryW(directoryString.getData());
    SetLastError(oldLastError);
    return b != 0;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> directoryString;
    directoryString.setSize(directory.toString().length() + 1);
    if (!directory._store(directoryString.getData(), directoryString.size()))
    {
        return false;
    }
    int oldErrno = errno;
    int b = rmdir(directoryString.getData());
    errno = oldErrno;
    return b == 0;
#else
    #error Unsupported platform
#endif
}

bool FileSystem::renameFile(const char * pOldFileName, const char * pNewFileName)
{
    require(pOldFileName != NULL);
    require(pNewFileName != NULL);

    return renameFile(PathName(pOldFileName), PathName(pNewFileName));
}

bool FileSystem::renameFile(const Char * pOldFileName, const char * pNewFileName)
{
    require(pOldFileName != NULL);
    require(pNewFileName != NULL);

    return renameFile(PathName(pOldFileName), PathName(pNewFileName));
}

bool FileSystem::renameFile(const String & oldFileName, const char * pNewFileName)
{
    require(pNewFileName != NULL);

    return renameFile(PathName(oldFileName), PathName(pNewFileName));
}

bool FileSystem::renameFile(const PathName & oldFileName, const char * pNewFileName)
{
    require(pNewFileName != NULL);

    return renameFile(oldFileName, PathName(pNewFileName));
}

bool FileSystem::renameFile(const char * pOldFileName, const Char * pNewFileName)
{
    require(pOldFileName != NULL);
    require(pNewFileName != NULL);

    return renameFile(PathName(pOldFileName), PathName(pNewFileName));
}

bool FileSystem::renameFile(const Char * pOldFileName, const Char * pNewFileName)
{
    require(pOldFileName != NULL);
    require(pNewFileName != NULL);

    return renameFile(PathName(pOldFileName), PathName(pNewFileName));
}

bool FileSystem::renameFile(const String & oldFileName, const Char * pNewFileName)
{
    require(pNewFileName != NULL);

    return renameFile(PathName(oldFileName), PathName(pNewFileName));
}

bool FileSystem::renameFile(const PathName & oldFileName, const Char * pNewFileName)
{
    require(pNewFileName != NULL);

    return renameFile(oldFileName, PathName(pNewFileName));
}

bool FileSystem::renameFile(const char * pOldFileName, const String & newFileName)
{
    require(pOldFileName != NULL);

    return renameFile(PathName(pOldFileName), PathName(newFileName));
}

bool FileSystem::renameFile(const Char * pOldFileName, const String & newFileName)
{
    require(pOldFileName != NULL);

    return renameFile(PathName(pOldFileName), PathName(newFileName));
}

bool FileSystem::renameFile(const String & oldFileName, const String & newFileName)
{
    return renameFile(PathName(oldFileName), PathName(newFileName));
}

bool FileSystem::renameFile(const PathName & oldFileName, const String & newFileName)
{
    return renameFile(oldFileName, PathName(newFileName));
}

bool FileSystem::renameFile(const char * pOldFileName, const PathName & newFileName)
{
    require(pOldFileName != NULL);

    return renameFile(PathName(pOldFileName), newFileName);
}

bool FileSystem::renameFile(const Char * pOldFileName, const PathName & newFileName)
{
    require(pOldFileName != NULL);

    return renameFile(PathName(pOldFileName), newFileName);
}

bool FileSystem::renameFile(const String & oldFileName, const PathName & newFileName)
{
    return renameFile(PathName(oldFileName), newFileName);
}

bool FileSystem::renameFile(const PathName & oldFileName, const PathName & newFileName)
{
    if (oldFileName.isEmpty() || newFileName.isEmpty())
    {
        return false;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> oldFileNameString;
    Array<wchar_t> newFileNameString;
    oldFileNameString.setSize(oldFileName.toString().length() + 1);
    newFileNameString.setSize(newFileName.toString().length() + 1);
    if (!oldFileName._store(oldFileNameString.getData(), oldFileNameString.size()) ||
        !newFileName._store(newFileNameString.getData(), newFileNameString.size()))
    {
        return false;
    }
    DWORD oldLastError = GetLastError();
    BOOL b = MoveFileW(oldFileNameString.getData(), newFileNameString.getData());
    SetLastError(oldLastError);
    return b != 0;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> oldFileNameString;
    Array<char> newFileNameString;
    oldFileNameString.setSize(oldFileName.toString().length() + 1);
    newFileNameString.setSize(newFileName.toString().length() + 1);
    if (!oldFileName._store(oldFileNameString.getData(), oldFileNameString.size()) ||
        !newFileName._store(newFileNameString.getData(), newFileNameString.size()))
    {
        return false;
    }
    int oldErrno = errno;
    int b = rename(oldFileNameString.getData(), newFileNameString.getData());
    errno = oldErrno;
    return b != 0;
#else
    #error Unsupported platform
#endif
}

bool FileSystem::getFileTimes(const char * pFileOrDirectoryName,
                              TimeStamp & creationTime,
                              TimeStamp & lastAccessTime,
                              TimeStamp & lastWriteTime)
{
    require(pFileOrDirectoryName != NULL);

    return getFileTimes(PathName(pFileOrDirectoryName), creationTime, lastAccessTime, lastWriteTime);
}

bool FileSystem::getFileTimes(const Char * pFileOrDirectoryName,
                              TimeStamp & creationTime,
                              TimeStamp & lastAccessTime,
                              TimeStamp & lastWriteTime)
{
    require(pFileOrDirectoryName != NULL);

    return getFileTimes(PathName(pFileOrDirectoryName), creationTime, lastAccessTime, lastWriteTime);
}

bool FileSystem::getFileTimes(const String & fileOrDirectoryName,
                              TimeStamp & creationTime,
                              TimeStamp & lastAccessTime,
                              TimeStamp & lastWriteTime)
{
    return getFileTimes(PathName(fileOrDirectoryName), creationTime, lastAccessTime, lastWriteTime);
}

bool FileSystem::getFileTimes(const PathName & fileOrDirectoryName,
                              TimeStamp & creationTime,
                              TimeStamp & lastAccessTime,
                              TimeStamp & lastWriteTime)
{
    if (fileOrDirectoryName.isEmpty())
    {
        return false;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> fileOrDirectoryNameString;
    fileOrDirectoryNameString.setSize(fileOrDirectoryName.toString().length() + 1);
    if (!fileOrDirectoryName._store(fileOrDirectoryNameString.getData(), fileOrDirectoryNameString.size()))
    {
        return false;
    }
    DWORD oldLastError = GetLastError();
    HANDLE hFile = CreateFileW(fileOrDirectoryNameString.getData(),
                               GENERIC_READ,
                               FILE_SHARE_READ,
                               NULL,
                               OPEN_EXISTING,
                               FILE_FLAG_BACKUP_SEMANTICS,
                               NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {
        SetLastError(oldLastError);
        return false;
    }
    FILETIME creationTimeWin32, lastAccessTimeWin32, lastWriteTimeWin32;
    BOOL b = GetFileTime(hFile, &creationTimeWin32, &lastAccessTimeWin32, &lastWriteTimeWin32);
    CloseHandle(hFile);
    SetLastError(oldLastError);
    if (!b)
    {
        return false;
    }
    creationTime = _impl::_DateTimeManagerImpl::convertTimeStampFromWin32Format(creationTimeWin32);
    lastAccessTime = _impl::_DateTimeManagerImpl::convertTimeStampFromWin32Format(lastAccessTimeWin32);
    lastWriteTime = _impl::_DateTimeManagerImpl::convertTimeStampFromWin32Format(lastWriteTimeWin32);
    return true;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> fileOrDirectoryNameString;
    fileOrDirectoryNameString.setSize(fileOrDirectoryName.toString().length() + 1);
    if (!fileOrDirectoryName._store(fileOrDirectoryNameString.getData(), fileOrDirectoryNameString.size()))
    {
        return false;
    }
    int oldErrno = errno;
    struct stat s;
    int b = stat(fileOrDirectoryNameString.getData(), &s);
    errno = oldErrno;
    if (b != 0)
    {
        return false;
    }
    creationTime = _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(s.st_ctime);
    lastAccessTime = _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(s.st_atime);
    lastWriteTime = _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(s.st_mtime);
    return true;
#else
    #error Unsupported platform
#endif
}

bool FileSystem::setFileTimes(const char * pFileOrDirectoryName,
                              const TimeStamp & creationTime,
                              const TimeStamp & lastAccessTime,
                              const TimeStamp & lastWriteTime)
{
    require(pFileOrDirectoryName != NULL);

    return setFileTimes(PathName(pFileOrDirectoryName), creationTime, lastAccessTime, lastWriteTime);
}

bool FileSystem::setFileTimes(const Char * pFileOrDirectoryName,
                              const TimeStamp & creationTime,
                              const TimeStamp & lastAccessTime,
                              const TimeStamp & lastWriteTime)
{
    require(pFileOrDirectoryName != NULL);

    return setFileTimes(PathName(pFileOrDirectoryName), creationTime, lastAccessTime, lastWriteTime);
}

bool FileSystem::setFileTimes(const String & fileOrDirectoryName,
                              const TimeStamp & creationTime,
                              const TimeStamp & lastAccessTime,
                              const TimeStamp & lastWriteTime)
{
    return setFileTimes(PathName(fileOrDirectoryName), creationTime, lastAccessTime, lastWriteTime);
}

bool FileSystem::setFileTimes(const PathName & fileOrDirectoryName,
                              const TimeStamp & creationTime,
                              const TimeStamp & lastAccessTime,
                              const TimeStamp & lastWriteTime)
{
    if (fileOrDirectoryName.isEmpty())
    {
        return false;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> fileOrDirectoryNameString;
    fileOrDirectoryNameString.setSize(fileOrDirectoryName.toString().length() + 1);
    if (!fileOrDirectoryName._store(fileOrDirectoryNameString.getData(), fileOrDirectoryNameString.size()))
    {
        return false;
    }
    DWORD oldLastError = GetLastError();
    HANDLE hFile = CreateFileW(fileOrDirectoryNameString.getData(),
                               GENERIC_READ,
                               FILE_SHARE_READ,
                               NULL,
                               OPEN_EXISTING,
                               FILE_FLAG_BACKUP_SEMANTICS,
                               NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {
        SetLastError(oldLastError);
        return false;
    }
    FILETIME creationTimeWin32;
    FILETIME lastAccessTimeWin32;
    FILETIME lastWriteTimeWin32;

    creationTimeWin32 = _impl::_DateTimeManagerImpl::convertTimeStampToWin32Format(creationTime);
    lastAccessTimeWin32 = _impl::_DateTimeManagerImpl::convertTimeStampToWin32Format(lastAccessTime);
    lastWriteTimeWin32 = _impl::_DateTimeManagerImpl::convertTimeStampToWin32Format(lastWriteTime);
    BOOL b = SetFileTime(hFile,
                         (creationTimeWin32.dwHighDateTime == 0 && creationTimeWin32.dwLowDateTime == 0) ? (NULL) : (&creationTimeWin32),
                         (lastAccessTimeWin32.dwHighDateTime == 0 && lastAccessTimeWin32.dwLowDateTime == 0) ? (NULL) : (&lastAccessTimeWin32),
                         (lastWriteTimeWin32.dwHighDateTime == 0 && lastWriteTimeWin32.dwLowDateTime == 0) ? (NULL) : (&lastWriteTimeWin32));
    CloseHandle(hFile);
    SetLastError(oldLastError);
    return b != 0;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> fileOrDirectoryNameString;
    fileOrDirectoryNameString.setSize(fileOrDirectoryName.toString().length() + 1);
    if (!fileOrDirectoryName._store(fileOrDirectoryNameString.getData(), fileOrDirectoryNameString.size()))
    {
        return false;
    }
    struct utimbuf u;
    u.actime = _impl::_DateTimeManagerImpl::convertTimeStampToCrtFormat(lastAccessTime);
    u.modtime = _impl::_DateTimeManagerImpl::convertTimeStampToCrtFormat(lastWriteTime);
    int oldErrno = errno;
    int b = utime(fileOrDirectoryNameString.getData(), &u);
    errno = oldErrno;
    return b == 0;
#else
    #error Unsupported platform
#endif
}

bool FileSystem::getFileSize(const char * pFileOrDirectoryName, UInt64 & size)
{
    require(pFileOrDirectoryName != NULL);

    return getFileSize(PathName(pFileOrDirectoryName), size);
}

bool FileSystem::getFileSize(const Char * pFileOrDirectoryName, UInt64 & size)
{
    require(pFileOrDirectoryName != NULL);

    return getFileSize(PathName(pFileOrDirectoryName), size);
}

bool FileSystem::getFileSize(const String & fileOrDirectoryName, UInt64 & size)
{
    return getFileSize(PathName(fileOrDirectoryName), size);
}

bool FileSystem::getFileSize(const PathName & fileOrDirectoryName, UInt64 & size)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> fileOrDirectoryNameString;
    fileOrDirectoryNameString.setSize(fileOrDirectoryName.toString().length() + 1);
    if (!fileOrDirectoryName._store(fileOrDirectoryNameString.getData(), fileOrDirectoryNameString.size()))
    {
        return false;
    }
    struct __stat64 st;
    int oldErrno = errno;
    if (_wstat64(fileOrDirectoryNameString.getData(), &st) == 0)
    {   //  Success.
        size = st.st_size;
        return true;
    }
    else
    {   //  Failure
        errno = oldErrno;
        return false;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> fileOrDirectoryNameString;
    fileOrDirectoryNameString.setSize(fileOrDirectoryName.toString().length() + 1);
    if (!fileOrDirectoryName._store(fileOrDirectoryNameString.getData(), fileOrDirectoryNameString.size()))
    {
        return false;
    }
    struct stat s;
    int oldErrno = errno;
    int b = stat(fileOrDirectoryNameString.getData(), &s);
    errno = oldErrno;
    if (b == 0)
    {
        size = s.st_size;
        return true;
    }
    return false;
#else
    #error Unsupported platform
#endif
}

PathName FileSystem::getTempFileName()
{
    return getTempFileName(getTempDirectory());
}

PathName FileSystem::getTempFileName(const char * pDirectory)
{
    require(pDirectory != NULL);

    return getTempFileName(PathName(pDirectory));
}

PathName FileSystem::getTempFileName(const Char * pDirectory)
{
    require(pDirectory != NULL);

    return getTempFileName(PathName(pDirectory));
}

PathName FileSystem::getTempFileName(const String & directory)
{
    return getTempFileName(PathName(directory));
}

PathName FileSystem::getTempFileName(const PathName & directory)
{
    if (directory.isEmpty())
    {
        return PathName();
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    static UInt16 suffix1 = 0;
    UInt32 suffix2 = static_cast<UInt32>(time(NULL));
    Array<wchar_t> fileNameChars;
    suffix1++;
    for (UInt32 attempt = 0; attempt < 0x10000; attempt++)
    {
        wchar_t candidateFileName[32];
#if defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003)
        swprintf(candidateFileName, L"tmp%04x%08x%04x.tmp", suffix1, suffix2, attempt);
#else
        swprintf(candidateFileName, sizeof(candidateFileName) / sizeof(candidateFileName[0]), L"tmp%04x%08x%04x.tmp", suffix1, suffix2, attempt);
#endif
        PathName tempPathName(directory, String(candidateFileName));
        //  Prepare CRT filename
        String tempFileName = tempPathName.toAbsolute().toString();
        fileNameChars.setSize(tempFileName.length() + 1);
        if (!tempFileName._store(fileNameChars.getData(), fileNameChars.size()))
        {   //  The only reason the store could fail is non-ANSI characters in directory name
            return PathName();
        }
        //  Does the file exist ?
        int oldErrno = errno;
        if (_waccess(fileNameChars.getData(), 0) != 0)
        {   //  Does not exist!
            errno = oldErrno;
            return tempPathName;
        }
        //  Already exists
    }
    return PathName();
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    static UInt16 suffix1 = 0;
    UInt32 suffix2 = static_cast<UInt32>(time(NULL));
    Array<char> fileNameChars;
    suffix1++;
    for (UInt32 attempt = 0; attempt < 0x10000; attempt++)
    {
        char candidateFileName[32];
        sprintf(candidateFileName, "tmp%04x%08x%04x.tmp", suffix1, suffix2, attempt);
        PathName tempPathName(directory, candidateFileName);
        //  Prepare CRT filename
        String tempFileName = tempPathName.toAbsolute().toString();
        fileNameChars.setSize(tempFileName.length() + 1);
        if (!tempFileName._store(fileNameChars.getData(), fileNameChars.size()))
        {   //  The only reason the store could fail is non-ANSI characters in directory name
            return PathName();
        }
        //  Does the file exist ?
        int oldErrno = errno;
        if (access(fileNameChars.getData(), F_OK) != 0)
        {   //  Does not exist!
            errno = oldErrno;
            return tempPathName;
        }
        //  Already exists
    }
    return PathName();
#else
    #error Unsupported platform
#endif
}

Array<PathName> FileSystem::getFileSystemRoots()
{
    Array<PathName> result;

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    for (char drive = 'A'; drive <= 'Z'; drive++)
    {
        char driveRootPath[4] = {drive, ':', '\\', 0};
        switch (GetDriveType(driveRootPath))
        {
            case DRIVE_REMOVABLE:
            case DRIVE_FIXED:
            case DRIVE_REMOTE:
            case DRIVE_CDROM:
            case DRIVE_RAMDISK:
                result.append(PathName(driveRootPath));
                break;
        }
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    result.append(PathName("/"));
#else
    #error Unsupported platform
#endif
    return result;
}

bool FileSystem::getTotalSpace(const char * pDirectory, UInt64 & totalSpace)
{
    require(pDirectory != NULL);

    return getTotalSpace(PathName(pDirectory), totalSpace);
}

bool FileSystem::getTotalSpace(const Char * pDirectory, UInt64 & totalSpace)
{
    require(pDirectory != NULL);

    return getTotalSpace(PathName(pDirectory), totalSpace);
}

bool FileSystem::getTotalSpace(const String & directory, UInt64 & totalSpace)
{
    return getTotalSpace(PathName(directory), totalSpace);
}

bool FileSystem::getTotalSpace(const PathName & directory, UInt64 & totalSpace)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> directoryString;
    directoryString.setSize(directory.toString().length() + 1);
    if (!directory._store(directoryString.getData(), directoryString.size()))
    {
        return false;
    }
    //  Ensure tailing backslash
    UIntSize directoryStringLength = static_cast<UIntSize>(wcslen(directoryString.getData()));
    if (directoryStringLength != 0 &&
        !isPathComponentSeparator(directoryString[directoryStringLength - 1]))
    {
        directoryString[directoryStringLength++] = '\\';
        directoryString[directoryStringLength] = 0;
    }

    ULARGE_INTEGER freeBytesAvailable, totalNumberOfBytes, totalNumberOfFreeBytes;
    DWORD oldLastError = GetLastError();
    if (!GetDiskFreeSpaceExW(directoryString.getData(), &freeBytesAvailable, &totalNumberOfBytes, &totalNumberOfFreeBytes))
    {
        SetLastError(oldLastError);
        return false;
    }
    SetLastError(oldLastError);
    totalSpace = totalNumberOfBytes.QuadPart;
    return true;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> directoryString;
    directoryString.setSize(directory.toString().length() + 1);
    if (!directory._store(directoryString.getData(), directoryString.size()))
    {
        return false;
    }
    struct statvfs s;
    int oldErrno = errno;
    int b = statvfs(directoryString.getData(), &s);
    errno = oldErrno;
    if (b == 0)
    {
        totalSpace = static_cast<UInt64>(s.f_blocks) * static_cast<UInt64>(s.f_bsize);
        return true;
    }
    return false;
#else
    #error Unsupported platform
#endif
}

bool FileSystem::getFreeSpace(const char * pDirectory, UInt64 & freeSpace)
{
    require(pDirectory != NULL);

    return getFreeSpace(PathName(pDirectory), freeSpace);
}

bool FileSystem::getFreeSpace(const Char * pDirectory, UInt64 & freeSpace)
{
    require(pDirectory != NULL);

    return getFreeSpace(PathName(pDirectory), freeSpace);
}

bool FileSystem::getFreeSpace(const String & directory, UInt64 & freeSpace)
{
    return getFreeSpace(PathName(directory), freeSpace);
}

bool FileSystem::getFreeSpace(const PathName & directory, UInt64 & freeSpace)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> directoryString;
    directoryString.setSize(directory.toString().length() + 1);
    if (!directory._store(directoryString.getData(), directoryString.size()))
    {
        return false;
    }
    ULARGE_INTEGER freeBytesAvailable, totalNumberOfBytes, totalNumberOfFreeBytes;
    DWORD oldLastError = GetLastError();
    if (!GetDiskFreeSpaceExW(directoryString.getData(), &freeBytesAvailable, &totalNumberOfBytes, &totalNumberOfFreeBytes))
    {
        SetLastError(oldLastError);
        return false;
    }
    SetLastError(oldLastError);
    freeSpace = freeBytesAvailable.QuadPart;
    return true;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> directoryString;
    directoryString.setSize(directory.toString().length() + 1);
    if (!directory._store(directoryString.getData(), directoryString.size()))
    {
        return false;
    }
    struct statvfs s;
    int oldErrno = errno;
    int b = statvfs(directoryString.getData(), &s);
    errno = oldErrno;
    if (b == 0)
    {
        freeSpace = static_cast<UInt64>(s.f_bavail) * static_cast<UInt64>(s.f_bsize);
        return true;
    }
    return false;
#else
    #error Unsupported platform
#endif
}

bool FileSystem::getMediaType(const char * pDirectory, MediaType & mediaType)
{
    require(pDirectory != NULL);

    return getMediaType(PathName(pDirectory), mediaType);
}

bool FileSystem::getMediaType(const Char * pDirectory, MediaType & mediaType)
{
    require(pDirectory != NULL);

    return getMediaType(PathName(pDirectory), mediaType);
}

bool FileSystem::getMediaType(const String & directory, MediaType & mediaType)
{
    return getMediaType(PathName(directory), mediaType);
}

bool FileSystem::getMediaType(const PathName & directory, MediaType & mediaType)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    PathName fullPathName = directory.toAbsolute();
    String pathNameString = fullPathName.toString();
    if (pathNameString.length() >= 3 && pathNameString[1] == ':' && pathNameString[2] == '\\' &&
        (pathNameString[0] & 0xFFFFFF00) == 0)
    {
        char rootPath[4] = { static_cast<char>(pathNameString[0]), ':', '\\', 0 };
        UINT dt = GetDriveType(rootPath);
        switch (dt)
        {
            case DRIVE_UNKNOWN:
                mediaType = MediaType_Unknown;
                return false;
            case DRIVE_NO_ROOT_DIR:
                mediaType = MediaType_Unknown;
                return false;
            case DRIVE_REMOVABLE:
                mediaType = MediaType_Removable;
                return true;
            case DRIVE_FIXED:
                mediaType = MediaType_Fixed;
                return true;
            case DRIVE_REMOTE:
                mediaType = MediaType_Remote;
                return true;
            case DRIVE_CDROM:
                mediaType = MediaType_CdRom;
                return true;
            case DRIVE_RAMDISK:
                mediaType = MediaType_RamDisk;
                return true;
            default:
                mediaType = MediaType_Unknown;
                return false;
        }
    }
    mediaType = MediaType_Unknown;
    return false;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    PathName fullPathName = directory.toAbsolute();
    Array<char> directoryString;
    directoryString.setSize(fullPathName.toString().length() + 1);
    if (!fullPathName._store(directoryString.getData(), directoryString.size()))
    {
        return false;
    }
    int oldErrno = errno;
    if (access(directoryString.getData(), F_OK) == 0)
    {
        mediaType = MediaType_Fixed;
    }
    else
    {
        errno = oldErrno;
        mediaType = MediaType_Unknown;
    }
    return true;
#else
    #error Unsupported platform
#endif
}

bool FileSystem::getDirectoryContents(const char * pDirectory, Array<DirectoryEntry> & directoryEntries, IDirectoryEntryFilter * pFilter /*= NULL*/)
{
    return getDirectoryContents(PathName(pDirectory), directoryEntries, pFilter);
}

bool FileSystem::getDirectoryContents(const Char * pDirectory, Array<DirectoryEntry> & directoryEntries, IDirectoryEntryFilter * pFilter /*= NULL*/)
{
    return getDirectoryContents(PathName(pDirectory), directoryEntries, pFilter);
}

bool FileSystem::getDirectoryContents(const String & directory, Array<DirectoryEntry> & directoryEntries, IDirectoryEntryFilter * pFilter /*= NULL*/)
{
    return getDirectoryContents(PathName(directory), directoryEntries, pFilter);
}

bool FileSystem::getDirectoryContents(const PathName & directory, Array<DirectoryEntry> & directoryEntries, IDirectoryEntryFilter * pFilter /*= NULL*/)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    //  Store directory name
    Array<wchar_t> searchPattern;
    searchPattern.setSize(directory.toString().length() + 9);
    if (!directory._store(searchPattern.getData(), searchPattern.size()))
    {
        return false;
    }
    UIntSize directoryStringLength = static_cast<UIntSize>(wcslen(searchPattern.getData()));
    //  Prepare search pattern
    if (searchPattern[0] != 0 &&
        searchPattern[directoryStringLength - 1]  != '/' &&
        searchPattern[directoryStringLength - 1]  != '\\')
    {
#if defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003)
        wcscat(searchPattern.getData(), L"\\");
#else
        wcscat_s(searchPattern.getData(), searchPattern.size(), L"\\");
#endif
    }
#if defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003)
    wcscat(searchPattern.getData(), L"*.*");
#else
    wcscat_s(searchPattern.getData(), searchPattern.size(), L"*.*");
#endif
    //  Search!
    directoryEntries.removeAll();
    struct _wfinddatai64_t fd;
    intptr_t fh = _wfindfirsti64(searchPattern.getData(), &fd);
    if (fh != -1)
    {
        for (; ; )
        {
            //  Process one entry
            if (wcscmp(fd.name, L".") != 0 && wcscmp(fd.name, L"..") != 0)
            {
                DirectoryEntry entry((fd.attrib & _A_SUBDIR ) != 0,
                                     String(fd.name),
                                     PathName(directory, String(fd.name)).toString(),
                                     fd.size,
                                     _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(fd.time_create),
                                     _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(fd.time_access),
                                     _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(fd.time_write));
                if (pFilter == NULL || pFilter->match(entry))
                {
                    directoryEntries.append(entry);
                }
            }
            //  More ?
            if (_wfindnexti64(fh, &fd) != 0)
            {
                break;
            }
        }
        _findclose(fh);
    }
    return true;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    //  Store directory name
    Array<char> directoryString;
    directoryString.setSize(directory.toString().length() + 9);
    if (!directory._store(directoryString.getData(), directoryString.size()))
    {
        return false;
    }
    int oldErrno = errno;
    DIR * pDir = opendir(directoryString.getData());
    if (pDir == NULL)
    {
        errno = oldErrno;
        return false;
    }
    struct dirent *de;
    while ((de = readdir(pDir)) != NULL)
    {
        if (strcmp(de->d_name, ".") != 0 && strcmp(de->d_name, "..") != 0)
        {
            continue;
        }
        bool isDirectory = false;
        UInt64 size = 0;
        TimeStamp creationTime, lastAccessTime, lastWriteTime;
        String shortName(de->d_name);
        String fullName = PathName(directory, shortName).toString();
        Array<char> fullNameString;
        fullNameString.setSize(fullName.length() + 1);
        bool b = fullName._store(fullNameString.getData(), fullNameString.size());
        if (b)
        {
            struct stat s;
            if (stat(fullNameString.getData(), &s) == 0)
            {
                isDirectory = S_ISDIR(s.st_mode);
                size = s.st_size;
                creationTime = _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(s.st_ctime);
                lastAccessTime = _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(s.st_atime);
                lastWriteTime = _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(s.st_mtime);
            }
        }
        DirectoryEntry entry(isDirectory, shortName, fullName, size, creationTime, lastAccessTime, lastWriteTime);
        if (pFilter == NULL || pFilter->match(entry))
        {
            directoryEntries.append(entry);
        }
    }
    closedir(pDir);
    errno = oldErrno;
    return true;
#else
    #error Unsupported platform
#endif
}

Array<PathName> FileSystem::splitPathListString(const char * pPathListString)
{
    require(pPathListString != NULL);

    return splitPathListString(String(pPathListString));
}

Array<PathName> FileSystem::splitPathListString(const Char * pPathListString)
{
    require(pPathListString != NULL);

    return splitPathListString(String(pPathListString));
}

Array<PathName> FileSystem::splitPathListString(const String & pathListString)
{
    Array<PathName> result;
    for (UIntSize scan = 0; scan < pathListString.length(); )
    {
        UIntSize separatorIndex;
        String pathListElement;
        if (pathListString.findFirst(FileSystem::PathListSeparator, separatorIndex, scan))
        {
            pathListElement = pathListString.mid(scan, separatorIndex - scan);
            scan = separatorIndex + 1;
        }
        else
        {
            pathListElement = pathListString;
            scan = pathListString.length();
        }
        pathListElement.trim();
        if (pathListElement.length() != 0)
        {
            result.append(PathName(pathListElement));
        }
    }
    return result;
}

String FileSystem::buildPathListString(const Array<PathName> & pathNames)
{
    String result;
    for (UIntSize i = 0; i < pathNames.size(); i++)
    {
        if (!pathNames[i].isEmpty())
        {
            if (result.length() != 0)
            {
                result += PathListSeparator;
            }
            result += pathNames[i].toString();
        }
    }
    return result;
}

//  End of FileSystem.cpp
