//
//  HeapManager.cpp - Implementation of the carl::HeapManager class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:52 $
//  $Revision: 1.15 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include <stdlib.h>
#include <malloc.h>

//////////
//  Singleton
HeapManager::HeapManager()
{
    _pInstalledGarbageCollectors = NULL;
    _installedGarbageCollectorsAllocated = 0;
    _installedGarbageCollectorsUsed = 0;
    _allocateCalls = 0;
    _freeCalls = 0;
    _currentAllocation = 0;
    _maximumAllocation = 0;
    _currentHeapSize = 0;
    _maximumHeapSize = 0;
    _currentUsedBlocks = 0;
    _maximumUsedBlocks = 0;
#if defined(CARL_MODE_DEBUG)
    _pFirstBlockDescriptor = NULL;
    _pLastBlockDescriptor = NULL;
#endif  //  defined(CARL_MODE_DEBUG)
}

HeapManager::~HeapManager()
{
    if (_pInstalledGarbageCollectors != NULL)
    {
        ::free(_pInstalledGarbageCollectors);
    }
}

//  Normally, the HeapManager::getInstance() would be implemented as:
//
//  HeapManager * HeapManager::getInstance()
//  {
//      static HeapManager instance;
//      return &instance;
//  }
//
//  which has also been found to work, as C++ ensures that a static local
//  variable is constructed when the control passes over its declaration
//  for the first time. However, tests have been found to crash on at least
//  one Debian installation for an unknown reason, and the ugly implementation
//  below seems to make the symptoms go away.
//
//  In theory, the following HeapManager::getInstance() implementation is
//  a) not thread-safe and b) inconsistent, as the HeapManager destructor
//  is never called. In practice this is not a problem because:
//
//  *   A heap manager is guaranteed to be initialized before the control enters
//      Application::run(), to allocate command line argument "String"s.
//  *   As heap manager is constructed before all other objects that use
//      the heap, it is to be destroyed last, so its destructor would have
//      been called immediately before exiting the application.
//
//  Note that we have to #undef the "new", as it may have been #defined
//  to redirect to a debug allocator.
//
//  Investigation shows that the problem was due to the presence of a global
//  static String variable in OutOfMemoryError implementation. The String
//  constructor for that variable was called first, causing HeapManager
//  instantiation. At exit time, HeapManager would be destructed first just
//  before the global static String variable; the String's destructor was
//  trying to use the already-destroyed HeapManager.
//
//  Although the HeapManager could have been reimplemented as a "clean
//  singleton" for CARL needs, there is no guarantee that CARL clients will
//  not have global static variables whose constructors use "operator new".
//  Therefore, the current implementation, in which the HeapManager is never
//  destructed, is likely to remain.
#ifdef new
    #undef new
#endif

void * HeapManager::operator new(unsigned int size)
{
    require(size == sizeof(HeapManager));
    static char block[sizeof(HeapManager)];
    return block;
}

HeapManager * HeapManager::getInstance()
{
    static HeapManager * pHeapManager = NULL;
    if (pHeapManager == NULL)
    {
        pHeapManager = new HeapManager();
    }
    return pHeapManager;
}

//////////
//  Operations
bool HeapManager::installGarbageCollector(IGarbageCollector * pGarbageCollector)
{
    require(pGarbageCollector != NULL);

    CriticalSectionLock guardLock(_guard);

    //  Do not install the same garbage collector more than once
    for (UIntSize i = 0; i < _installedGarbageCollectorsUsed; i++)
    {
        if (_pInstalledGarbageCollectors[i] == pGarbageCollector)
        {
            return false;
        }
    }
    //  Make sure there is a free slot
    if (_installedGarbageCollectorsUsed == _installedGarbageCollectorsAllocated)
    {   //  Expand the garbage collecor table
        UIntSize newInstalledGarbageCollectorsAllocated = Math::max(static_cast<UIntSize>((3 * _installedGarbageCollectorsAllocated) / 2 + 1), static_cast<UIntSize>(17));
        IGarbageCollector ** pNewInstalledGarbageCollectors =
            static_cast<IGarbageCollector **>(::malloc(sizeof(IGarbageCollector *) * newInstalledGarbageCollectorsAllocated));
        if (pNewInstalledGarbageCollectors == NULL)
        {
            throw OutOfMemoryError(sizeof(IGarbageCollector *) * newInstalledGarbageCollectorsAllocated);
        }
        if (_pInstalledGarbageCollectors != NULL)
        {
            for (UIntSize i = 0; i < _installedGarbageCollectorsUsed; i++)
            {
                pNewInstalledGarbageCollectors[i] = _pInstalledGarbageCollectors[i];
            }
            ::free(_pInstalledGarbageCollectors);
        }
        _pInstalledGarbageCollectors = pNewInstalledGarbageCollectors;
        newInstalledGarbageCollectorsAllocated = _installedGarbageCollectorsAllocated;
    }
    //  Install the garbage collector
    _pInstalledGarbageCollectors[_installedGarbageCollectorsUsed++] = pGarbageCollector;
    //  Done
    return true;
}

bool HeapManager::uninstallGarbageCollector(IGarbageCollector * pGarbageCollector)
{
    require(pGarbageCollector != NULL);

    CriticalSectionLock guardLock(_guard);

    //  Uninstall the garbage collector
    for (UIntSize i = 0; i < _installedGarbageCollectorsUsed; i++)
    {
        if (_pInstalledGarbageCollectors[i] == pGarbageCollector)
        {
            for (; i + 1 < _installedGarbageCollectorsUsed; i++)
            {
                _pInstalledGarbageCollectors[i] = _pInstalledGarbageCollectors[i + 1];
            }
            _installedGarbageCollectorsUsed--;
            return true;
        }
    }
    //  The garbage collector is not installed
    return false;
}

void * HeapManager::allocateFrom(UIntSize size, const char * pFileName, UIntSize lineNumber)
{
    _guard.enter();
    _allocateCalls++;
    for (; ; )
    {
        //  Try allocating the memory block
        void * pBlock = _tryAllocateFrom(size, pFileName, lineNumber);
        if (pBlock != NULL)
        {
            _guard.leave();
            return pBlock;
        }
        //  Allocation failed. Try garbage collection
        _guard.leave();
        if (!_applyGarbageCollectors())
        {   //  Garbage collection failed to produce more memory
            throw OutOfMemoryError(size);
        }
        _guard.enter();
    }
}

void HeapManager::freeFrom(void * pBlock, const char * pFileName, UIntSize lineNumber)
{
    CriticalSectionLock guardLock(_guard);

    _freeCalls++;
    _tryFreeFrom(pBlock, pFileName, lineNumber);
}

UInt64 HeapManager::getAllocateCalls() const
{
    CriticalSectionLock guardLock(_guard);

    return _allocateCalls;
}

UInt64 HeapManager::getFreeCalls() const
{
    CriticalSectionLock guardLock(_guard);

    return _freeCalls;
}

UIntSize HeapManager::getCurrentAllocation() const
{
    CriticalSectionLock guardLock(_guard);

    return _currentAllocation;
}

UIntSize HeapManager::getMaximumAllocation() const
{
    CriticalSectionLock guardLock(_guard);

    return _maximumAllocation;
}

UIntSize HeapManager::getCurrentHeapSize() const
{
    CriticalSectionLock guardLock(_guard);

    return _currentHeapSize;
}

UIntSize HeapManager::getMaximumHeapSize() const
{
    CriticalSectionLock guardLock(_guard);

    return _maximumHeapSize;
}

UIntSize HeapManager::getCurrentUsedBlocks() const
{
    CriticalSectionLock guardLock(_guard);

    return _currentUsedBlocks;
}

UIntSize HeapManager::getMaximumUsedBlocks() const
{
    CriticalSectionLock guardLock(_guard);

    return _maximumUsedBlocks;
}

//////////
//  Implementation helpers
#if defined(CARL_MODE_DEBUG)
void * HeapManager::_tryAllocateFrom(UIntSize size, const char * pFileName, UIntSize lineNumber)
#else   //  defined(CARL_MODE_DEBUG)
void * HeapManager::_tryAllocateFrom(UIntSize size, const char * /*pFileName*/, UIntSize /*lineNumber*/)
#endif  //  defined(CARL_MODE_DEBUG)
{
#if defined(CARL_MODE_DEBUG)
    //  Allocate descriptor
    _HeapBlockDescriptor * pDescriptor = static_cast<_HeapBlockDescriptor*>(::malloc(sizeof(_HeapBlockDescriptor)));
    if (pDescriptor == NULL)
    {
        return NULL;
    }
    //  Allocate cell
    UInt8 * pCell = static_cast<UInt8*>(::malloc(size + 2 * _HeapCellPadding + sizeof(_HeapBlockDescriptor*)));
    if (pCell == NULL)
    {
        ::free(pDescriptor);
        return NULL;
    }
    //  Initialize the cell
    Memory::set(pCell + sizeof(_HeapBlockDescriptor*), _Filler, size + 2 * _HeapCellPadding);
    *(reinterpret_cast<_HeapBlockDescriptor**>(pCell)) = pDescriptor;
    //  Initialize the descriptor
    pDescriptor->_cellSize = size + 2 * _HeapCellPadding;
    pDescriptor->_dataSize = size;
    pDescriptor->_pAllocatedFromSourceFile = pFileName;
    pDescriptor->_allocatedFromSourceLine = lineNumber;
    pDescriptor->_pCell = pCell + sizeof(_HeapBlockDescriptor*);
    pDescriptor->_pData = pCell + sizeof(_HeapBlockDescriptor*) + _HeapCellPadding;
    //  Add block descriptor to the list of active block descriptors
    pDescriptor->_pNextBlockDescriptor = NULL;
    pDescriptor->_pPrevBlockDescriptor = _pLastBlockDescriptor;
    if (_pLastBlockDescriptor == NULL)
    {
        _pFirstBlockDescriptor = _pLastBlockDescriptor = pDescriptor;
    }
    else
    {
        _pLastBlockDescriptor->_pNextBlockDescriptor = pDescriptor;
        _pLastBlockDescriptor = pDescriptor;
    }
    //  Done
    _currentAllocation += size;
    _maximumAllocation = Math::max(_maximumAllocation, _currentAllocation);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _currentHeapSize += static_cast<UIntSize>(_msize(pDescriptor) + _msize(pCell));
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _currentHeapSize += static_cast<UIntSize>(sizeof(pDescriptor) + sizeof(pCell));
#else
    #error Unsupported platform
#endif
    _maximumHeapSize = Math::max(_maximumHeapSize, _currentHeapSize);
    _currentUsedBlocks++;
    _maximumUsedBlocks = Math::max(_maximumUsedBlocks, _currentUsedBlocks);
    return pDescriptor->_pData;
#else   //  defined(CARL_MODE_DEBUG)
    UInt8 * pBlock = static_cast<UInt8*>(malloc(size + sizeof(UIntSize)));
    if (pBlock == NULL)
    {
        return false;
    }
    *(reinterpret_cast<UIntSize*>(pBlock)) = size;
    _currentAllocation += size;
    _maximumAllocation = Math::max(_currentAllocation, _maximumAllocation);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _currentHeapSize += static_cast<UIntSize>(_msize(pBlock));
    _maximumHeapSize = Math::max(_currentHeapSize, _maximumHeapSize);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _currentHeapSize += size;
    _maximumHeapSize = Math::max(_currentHeapSize, _maximumHeapSize);
#else
    #error Unsupported platform
#endif
    _currentUsedBlocks++;
    _maximumUsedBlocks = Math::max(_maximumUsedBlocks, _currentUsedBlocks);
    return pBlock + sizeof(UIntSize);
#endif  //  defined(CARL_MODE_DEBUG)
}

#if defined(CARL_MODE_DEBUG)
void HeapManager::_tryFreeFrom(void * pBlock, const char * pFileName, UIntSize lineNumber)
#else   //  defined(CARL_MODE_DEBUG)
void HeapManager::_tryFreeFrom(void * pBlock, const char * /*pFileName*/, UIntSize /*lineNumber*/)
#endif  //  defined(CARL_MODE_DEBUG)
{
    if (pBlock == NULL)
    {
        return;
    }
#if defined(CARL_MODE_DEBUG)
    //  Find out where the cell starts
    UInt8 * pCell = static_cast<UInt8*>(pBlock) - _HeapCellPadding;
    //  Locate the block descriptor
    _HeapBlockDescriptor * pDescriptor = *(reinterpret_cast<_HeapBlockDescriptor**>(pCell - sizeof(_HeapBlockDescriptor*)));
    //  Handshake
    if (pDescriptor->_pCell != pCell ||
        pDescriptor->_pData != pCell + _HeapCellPadding ||
        pDescriptor->_cellSize != pDescriptor->_dataSize + 2 * _HeapCellPadding)
    {   //  OOPS! Deallocating a nonexistent block
        throw NoSuchHeapBlockError(pBlock, pFileName, lineNumber);
    }
    //  Check cell's guard areas
    for (UIntSize i = 0; i < _HeapCellPadding; i++)
    {
        if (pCell[i] != _Filler || pCell[_HeapCellPadding + pDescriptor->_dataSize + i] != _Filler)
        {   //  OOPS! Cell corrupt
            throw HeapBlockCorruptError(pBlock, pFileName, lineNumber);
        }
    }
    //  Update statistics
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _currentHeapSize -= static_cast<UIntSize>(_msize(pDescriptor) + _msize(pCell - sizeof(_HeapBlockDescriptor*)));
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _currentHeapSize -= static_cast<UIntSize>(sizeof(pDescriptor) + sizeof(pCell - sizeof(_HeapBlockDescriptor*)));
#else
    #error Unsupported platform
#endif
    _currentAllocation -= pDescriptor->_dataSize;
    //  Free the cell
    Memory::set(pCell - sizeof(_HeapBlockDescriptor*), _Filler, pDescriptor->_cellSize + sizeof(_HeapBlockDescriptor*));
    ::free(pCell - sizeof(_HeapBlockDescriptor*));
    //  Remove descriptor from the list of existing descriptors
    if (pDescriptor->_pPrevBlockDescriptor == NULL)
    {
        _pFirstBlockDescriptor = pDescriptor->_pNextBlockDescriptor;
    }
    else
    {
        pDescriptor->_pPrevBlockDescriptor->_pNextBlockDescriptor = pDescriptor->_pNextBlockDescriptor;
    }
    if (pDescriptor->_pNextBlockDescriptor == NULL)
    {
        _pLastBlockDescriptor = pDescriptor->_pPrevBlockDescriptor;
    }
    else
    {
        pDescriptor->_pNextBlockDescriptor->_pPrevBlockDescriptor = pDescriptor->_pPrevBlockDescriptor;
    }
    //  Free the descriptor
    Memory::set(pDescriptor, _Filler, sizeof(_HeapBlockDescriptor));
    ::free(pDescriptor);
    //  Done
    _currentUsedBlocks--;
#else   //  defined(CARL_MODE_DEBUG)
    UInt8 * pBlockStart = static_cast<UInt8*>(pBlock) - sizeof(UIntSize);
    UIntSize size = *(reinterpret_cast<UIntSize*>(pBlockStart));
    _currentAllocation -= size;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _currentHeapSize -= static_cast<UIntSize>(_msize(pBlockStart));
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _currentHeapSize -= size;
#else
    #error Unsupported platform
#endif
    ::free(pBlockStart);
    _currentUsedBlocks--;
#endif  //  defined(CARL_MODE_DEBUG)
}

bool HeapManager::_applyGarbageCollectors()
{
    bool someGarbageCollected = false;
    for (UIntSize i = 0; ; i++)
    {
        //  Get one installed garbage collector
        if (i >= _installedGarbageCollectorsUsed)
        {
            break;
        }
        IGarbageCollector * pGarbageCollector = _pInstalledGarbageCollectors[i];
        //  Run this garbage collector
        if (pGarbageCollector->run() != 0)
        {
            someGarbageCollected = true;
        }
    }
    return someGarbageCollected;
}

//  End of HeapManager.cpp
