//
//  IntegerFormatProperties.cpp - Implementation of the carl::IntegerFormatProperties class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/08/13 13:28:54 $
//  $Revision: 1.10 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

//////////
//  Construction/destruction/assignment
IntegerFormatProperties::IntegerFormatProperties()
{
    _setStandardProperties();
}

IntegerFormatProperties::IntegerFormatProperties(const char * pCrtFormatString)
{
    _parseCrtFormatString(String(pCrtFormatString));
}

IntegerFormatProperties::IntegerFormatProperties(const Char * pCrtFormatString)
{
    _parseCrtFormatString(String(pCrtFormatString));
}

IntegerFormatProperties::IntegerFormatProperties(const String & crtFormatString)
{
    _parseCrtFormatString(crtFormatString);
}

IntegerFormatProperties::IntegerFormatProperties(const IntegerFormatProperties & src)
{
    _radix = src._radix;
    _minimumDigits = src._minimumDigits;
    _maximumDigits = src._maximumDigits;
    _minimumWidth = src._minimumWidth;
    _maximumWidth = src._maximumWidth;
    _fillerChar = src._fillerChar;
    _errorChar = src._errorChar;
    _alignment = src._alignment;
    _letterCase = src._letterCase;
    _usePrefix = src._usePrefix;
    _useSeparator = src._useSeparator;
}

IntegerFormatProperties & IntegerFormatProperties::operator = (const IntegerFormatProperties & src)
{
    if (this != (&src))
    {
        _radix = src._radix;
        _minimumDigits = src._minimumDigits;
        _maximumDigits = src._maximumDigits;
        _minimumWidth = src._minimumWidth;
        _maximumWidth = src._maximumWidth;
        _fillerChar = src._fillerChar;
        _errorChar = src._errorChar;
        _alignment = src._alignment;
        _letterCase = src._letterCase;
        _usePrefix = src._usePrefix;
        _useSeparator = src._useSeparator;
    }
    return *this;
}

//////////
//  Operations
IntegerFormatProperties & IntegerFormatProperties::setRadix(UIntSize radix)
{
    require(radix >= 2 && radix <= 36);

    _radix = radix;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setMinimumDigits(UIntSize digits)
{
    _minimumDigits = digits;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setMaximumDigits(UIntSize digits)
{
    _maximumDigits = digits;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setMinimumWidth(UIntSize width)
{
    _minimumWidth = width;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setMaximumWidth(UIntSize width)
{
    _maximumWidth = width;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setFillerChar(Char ch)
{
    _fillerChar = ch;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setErrorChar(Char ch)
{
    _errorChar = ch;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setAlignment(Alignment alignment)
{
    _alignment = alignment;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setLetterCase(LetterCase letterCase)
{
    _letterCase = letterCase;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setUsePrefix(bool usePrefix)
{
    _usePrefix = usePrefix;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setUseSeparator(bool useSeparator)
{
    _useSeparator = useSeparator;
    return *this;
}

IntegerFormatProperties & IntegerFormatProperties::setMandatorySign(bool mandatorySign)
{
    _mandatorySign = mandatorySign;
    return *this;
}

//////////
//  Implementation helpers
void IntegerFormatProperties::_setStandardProperties()
{
    _radix = 10;
    _minimumDigits = 0;
    _maximumDigits = 0;
    _minimumWidth = 0;
    _maximumWidth = 0;
    _fillerChar = ' ';
    _errorChar = '*';
    _alignment = Alignment_Right;
    _letterCase = LetterCase_Lower;
    _usePrefix = false;
    _useSeparator = false;
    _mandatorySign = false;
}

void IntegerFormatProperties::_parseCrtFormatString(const carl::String & crtFormatString)
{
    _setStandardProperties();
    UIntSize scan = 0;
    //  Skip optional leading '%'
    if (scan < crtFormatString.length() && crtFormatString[scan] == '%')
    {
        scan++;
    }
    //  Process flags
    for (; scan < crtFormatString.length(); scan++)
    {
        switch (crtFormatString[scan])
        {
            case '-':
                _alignment = Alignment_Left;
                break;
            case '+':
                _mandatorySign = true;
                break;
            case '#':
                _usePrefix = true;
                break;
            default:
                goto FlagsProcessed;
        }
    }
FlagsProcessed:;
    //  Process width
    if (scan < crtFormatString.length() && Iso10646::isDecimalDigit(crtFormatString[scan]))
    {
        bool widthHasLeadingZero = (scan + 1 < crtFormatString.length()) &&
                                   (crtFormatString[scan] == '0') &&
                                   (Iso10646::isDecimalDigit(crtFormatString[scan + 1]));
        for (; scan < crtFormatString.length() && Iso10646::isDecimalDigit(crtFormatString[scan]); scan++)
        {
            _minimumWidth = 10 * _minimumWidth + (crtFormatString[scan] - '0');
        }
        if (widthHasLeadingZero)
        {
            _minimumDigits = _minimumWidth;
        }
    }
    //  Process precision
    if (scan + 1 < crtFormatString.length() && crtFormatString[scan] == '.' && Iso10646::isDecimalDigit(crtFormatString[scan + 1]))
    {
        UIntSize precision = 0;
        for (scan++; scan < crtFormatString.length() && Iso10646::isDecimalDigit(crtFormatString[scan]); scan++)
        {
            precision = 10 * precision + (crtFormatString[scan] - '0');
        }
        _minimumDigits = precision;
        if (_minimumDigits > _minimumWidth)
        {
            _minimumWidth = _minimumDigits;
        }
    }
    //  Process type
    require(scan < crtFormatString.length());
    switch (crtFormatString[scan])
    {
        case 'd':   case 'i':   case 'u':
            scan++;
            break;
        case 'o':
            scan++;
            _radix  = 8;
            break;
        case 'x':
            scan++;
            _radix  = 16;
            break;
        case 'X':
            scan++;
            _radix  = 16;
            _letterCase = LetterCase_Upper;
            break;
        default:
            require(false);
    }
    if (scan != crtFormatString.length())
    {
        require(false);
    }
}

//  End of IntegerFormatProperties.cpp
