//
//  IoOperators.cpp - Implementation of the CARL I/O operators
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:53 $
//  $Revision: 1.12 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include <stdlib.h>

IDataInput & carl::operator >> (IDataInput & dis, bool & value) throw(IOException)
{
    value = dis.readBool();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, UInt8 & value) throw(IOException)
{
    value = dis.readUInt8();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, UInt16 & value) throw(IOException)
{
    value = dis.readUInt16();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, UInt32 & value) throw(IOException)
{
    value = dis.readUInt32();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, UInt64 & value) throw(IOException)
{
    value = dis.readUInt64();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, Int8 & value) throw(IOException)
{
    value = dis.readInt8();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, Int16 & value) throw(IOException)
{
    value = dis.readInt16();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, Int32 & value) throw(IOException)
{
    value = dis.readInt32();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, Int64 & value) throw(IOException)
{
    value = dis.readInt64();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, Real32 & value) throw(IOException)
{
    value = dis.readReal32();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, Real64 & value) throw(IOException)
{
    value = dis.readReal64();
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, Date & value) throw(IOException)
{
    UInt16 year = dis.readUInt16();
    UInt8 month = dis.readUInt8();
    UInt8 day = dis.readUInt8();
    if (year == 0 && month == 0 && day == 0)
    {
        value = Date();
        return dis;
    }
    else if (year > 0 &&
             month >= 1 && month <= 12 &&
             day >= 1 && day <= Calendar::daysInMonth(year, month))
    {
        value = Date(year, month, day);
        return dis;
    }
    else
    {
        throw InvalidDataException();
    }
}

IDataInput & carl::operator >> (IDataInput & dis, DateTime & value) throw(IOException)
{
    UInt16 year = dis.readUInt16();
    UInt8 month = dis.readUInt8();
    UInt8 day = dis.readUInt8();
    UInt8 hour = dis.readUInt8();
    UInt8 minute = dis.readUInt8();
    UInt8 second = dis.readUInt8();
    UInt32 nanosecond = dis.readUInt8();
    if (year == 0 && month == 0 && day == 0 && hour == 0 && minute == 0 && second == 0 && nanosecond == 0)
    {
        value = DateTime();
        return dis;
    }
    else if (year > 0 &&
             month >= 1 && month <= 12 &&
             day >= 1 && day <= Calendar::daysInMonth(year, month) &&
             hour < 24 &&
             minute < 60 &&
             second < 60 &&
             nanosecond < 1000000000)
    {
        value = DateTime(year, month, day, hour, minute, second, nanosecond);
        return dis;
    }
    else
    {
        throw InvalidDataException();
    }
}

IDataInput & carl::operator >> (IDataInput & dis, Locale & value) throw(IOException)
{
    String language, country, variant;

    dis >> language;
    dis >> country;
    dis >> variant;
    if (language.length() == 2 && country.length() == 2 && variant.length() != 0)
    {
        value = Locale(language, country, variant);
        return dis;
    }
    else if (language.length() == 2 && country.length() == 2 && variant.length() == 0)
    {
        value = Locale(language, country);
        return dis;
    }
    else if (language.length() == 2 && country.length() == 0 && variant.length() == 0)
    {
        value = Locale(language);
        return dis;
    }
    else if (language.length() == 0 && country.length() == 0 && variant.length() == 0)
    {
        value = Locale();
        return dis;
    }
    else
    {
        throw InvalidDataException();
    }
}

IDataInput & carl::operator >> (IDataInput & dis, PathName & value) throw(IOException)
{
    String pathNameString;
    dis >> pathNameString;
    value = PathName(pathNameString);
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, String & value) throw(IOException)
{
    //  Read string length in uleb128 form
    UIntSize strLen = static_cast<UIntSize>(ULeb128::readUInt32(dis));
    //  Read string characters in UTF form
    static CharacterSet * pUtf8CharacterSet = NULL;
    if (pUtf8CharacterSet == NULL)
    {
        pUtf8CharacterSet = CharacterSet::getInstance(String("utf-8"));
        require(pUtf8CharacterSet != NULL);
    }
    value.setLength(strLen);
    for (UIntSize charIndex = 0; charIndex < strLen; charIndex++)
    {
        UInt8 bytes[16];
        bytes[0] = dis.readUInt8();
        int encodingLength = pUtf8CharacterSet->getEncodingLengthForLeadByte(bytes[0]);
        if (encodingLength == -1)
        {
            throw InvalidDataException();
        }
        for (int i = 1; i < encodingLength; i++)
        {
            bytes[i] = dis.readUInt8();
        }
        Char c;
        int decodedBytes = pUtf8CharacterSet->decodeCharacter(bytes, encodingLength, c);
        if (decodedBytes != encodingLength)
        {
            throw InvalidDataException();
        }
        value[charIndex] = c;
    }
    //  Done
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, Time & value) throw(IOException)
{
    UInt8 hour = dis.readUInt8();
    UInt8 minute = dis.readUInt8();
    UInt8 second = dis.readUInt8();
    UInt32 nanosecond = dis.readUInt8();
    if (hour == 0 && minute == 0 && second == 0 && nanosecond == 0)
    {
        value = Time();
        return dis;
    }
    else if (hour < 24 &&
             minute < 60 &&
             second < 60 &&
             nanosecond < 1000000000)
    {
        value = Time(hour, minute, second, nanosecond);
        return dis;
    }
    else
    {
        throw InvalidDataException();
    }
}

IDataInput & carl::operator >> (IDataInput & dis, TimeSpan & value) throw(IOException)
{
    Int64 nanoseconds;
    dis >> nanoseconds;
    value = TimeSpan(nanoseconds);
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, TimeStamp & value) throw(IOException)
{
    UInt64 nanosecondsSinceEpoch;
    dis >> nanosecondsSinceEpoch;
    value = TimeStamp(nanosecondsSinceEpoch);
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, BinaryInteger & value) throw(IOException)
{
    value = ULeb128::readBinaryInteger(dis);
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, BinaryFixed & value) throw(IOException)
{
    BinaryInteger magnitude = Leb128::readBinaryInteger(dis);
    UInt32 scale = ULeb128::readUInt32(dis);
    value = BinaryFixed(magnitude, scale);
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, BinaryFloat & value) throw(IOException)
{
    int category = dis.readUInt8();
    if (category == BinaryFloat::ValueCategory_PositiveInfinity ||
        category == BinaryFloat::ValueCategory_NegativeInfinity ||
        category == BinaryFloat::ValueCategory_SignalingNan ||
        category == BinaryFloat::ValueCategory_QuietNan)
    {
        value = BinaryFloat(static_cast<BinaryFloat::ValueCategory>(category));
    }
    else if (category == BinaryFloat::ValueCategory_Normal)
    {
        BinaryInteger mantissa = Leb128::readBinaryInteger(dis);
        BinaryInteger scale = Leb128::readBinaryInteger(dis);
        value = BinaryFloat(mantissa, scale);
    }
    else
    {
        throw InvalidDataException();
    }
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, DecimalInteger & value) throw(IOException)
{
    value = 0;
    Int8 sign = dis.readInt8();
    if (sign < -1 || sign > 1)
    {
        throw InvalidDataException();
    }
    if (sign != 0)
    {
        UIntSize significantBytes = ULeb128::readUIntSize(dis);
        for (UIntSize i = 0; i < significantBytes; i++)
        {
            UInt8 bcdByte = dis.readUInt8();
            int digitLo = (bcdByte & 0x0F);
            int digitHi = ((bcdByte >> 4) & 0x0F);
            if (digitLo > 9 || digitHi > 9)
            {
                throw InvalidDataException();
            }
            value.setDigit(2 * i, digitLo);
            value.setDigit(2 * i + 1, digitHi);
        }
        if (value == 0)
        {
            throw InvalidDataException();
        }
        if (sign == -1)
        {
            value = -value;
        }
    }
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, DecimalFixed & value) throw(IOException)
{
    DecimalInteger magnitude;
    UIntSize scale;
    dis >> magnitude;
    scale = ULeb128::readUIntSize(dis);
    value = DecimalFixed(magnitude, scale);
    return dis;
}

IDataInput & carl::operator >> (IDataInput & dis, DecimalFloat & value) throw(IOException)
{
    int category = dis.readUInt8();
    if (category == DecimalFloat::ValueCategory_PositiveInfinity ||
        category == DecimalFloat::ValueCategory_NegativeInfinity ||
        category == DecimalFloat::ValueCategory_SignalingNan ||
        category == DecimalFloat::ValueCategory_QuietNan)
    {
        value = DecimalFloat(static_cast<DecimalFloat::ValueCategory>(category));
    }
    else if (category == DecimalFloat::ValueCategory_Normal)
    {
        DecimalInteger mantissa, scale;
        dis >> mantissa;
        dis >> scale;
        value = DecimalFloat(mantissa, scale);
    }
    else
    {
        throw InvalidDataException();
    }
    return dis;
}

IDataOutput & carl::operator << (IDataOutput & dos, bool value) throw(IOException)
{
    dos.writeBool(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, UInt8 value) throw(IOException)
{
    dos.writeUInt8(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, UInt16 value) throw(IOException)
{
    dos.writeUInt16(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, UInt32 value) throw(IOException)
{
    dos.writeUInt32(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, UInt64 value) throw(IOException)
{
    dos.writeUInt64(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, Int8 value) throw(IOException)
{
    dos.writeInt8(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, Int16 value) throw(IOException)
{
    dos.writeInt16(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, Int32 value) throw(IOException)
{
    dos.writeInt32(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, Int64 value) throw(IOException)
{
    dos.writeInt64(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, Real32 value) throw(IOException)
{
    dos.writeReal32(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, Real64 value) throw(IOException)
{
    dos.writeReal64(value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const Date & value) throw(IOException)
{
    dos.writeUInt16(static_cast<UInt16>(value.getYear()));
    dos.writeUInt8(static_cast<UInt8>(value.getMonth()));
    dos.writeUInt8(static_cast<UInt8>(value.getDay()));
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const DateTime & value) throw(IOException)
{
    dos.writeUInt16(static_cast<UInt16>(value.getYear()));
    dos.writeUInt8(static_cast<UInt8>(value.getMonth()));
    dos.writeUInt8(static_cast<UInt8>(value.getDay()));
    dos.writeUInt8(static_cast<UInt8>(value.getHour()));
    dos.writeUInt8(static_cast<UInt8>(value.getMinute()));
    dos.writeUInt8(static_cast<UInt8>(value.getSecond()));
    dos.writeUInt32(value.getNanosecond());
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const Locale & value) throw(IOException)
{
    dos << value.getLanguage();
    dos << value.getCountry();
    dos << value.getVariant();
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const PathName & value) throw(IOException)
{
    dos << value.toString();
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const String & value) throw(IOException)
{
    require(static_cast<UInt32>(value.length()) == value.length());
    //  Write string length in uleb128 form
    ULeb128::writeUInt32(dos, static_cast<UInt32>(value.length()));
    //  Write string characters in UTF form
    static CharacterSet * pUtf8CharacterSet = NULL;
    if (pUtf8CharacterSet == NULL)
    {
        pUtf8CharacterSet = CharacterSet::getInstance(String("utf-8"));
        require(pUtf8CharacterSet != NULL);
    }
    for (UIntSize charIndex = 0; charIndex < value.length(); charIndex++)
    {
        Char c = value[charIndex];
        UInt8 bytes[16];
        int encodingLength = pUtf8CharacterSet->encodeCharacter(c, bytes);
        require(encodingLength != -1);
        for (int i = 0; i < encodingLength; i++)
        {
            dos.writeUInt8(bytes[i]);
        }
    }
    //  Done
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const Time & value) throw(IOException)
{
    dos.writeUInt8(static_cast<UInt8>(value.getHour()));
    dos.writeUInt8(static_cast<UInt8>(value.getMinute()));
    dos.writeUInt8(static_cast<UInt8>(value.getSecond()));
    dos.writeUInt32(value.getNanosecond());
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const TimeSpan & value) throw(IOException)
{
    dos.writeInt64(value.getNanoseconds());
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const TimeStamp & value) throw(IOException)
{
    dos.writeUInt64(value.getNanosecondsSinceEpoch());
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const BinaryInteger & value) throw(IOException)
{
    ULeb128::writeBinaryInteger(dos, value);
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const BinaryFixed & value) throw(IOException)
{
    Leb128::writeBinaryInteger(dos, (value.getSign() < 0) ? (-value.getMagnitude()) : (value.getMagnitude()));
    ULeb128::writeUInt32(dos, value.getScale());
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const BinaryFloat & value) throw(IOException)
{
    dos.writeUInt8(static_cast<UInt8>(value.getCategory()));
    if (value.isNormal())
    {
        Leb128::writeBinaryInteger(dos, (value.getSign() < 0) ? (-value.getMantissa()) : (value.getMantissa()));
        Leb128::writeBinaryInteger(dos, value.getScale());
    }
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const DecimalInteger & value) throw(IOException)
{
    Int8 signByte = static_cast<Int8>(value.getSign());
    dos.writeInt8(signByte);
    if (signByte != 0)
    {
        UIntSize significantDigits = value.getRequiredDigits();
        UIntSize significantBytes = (significantDigits + 1) / 2;
        ULeb128::writeUIntSize(dos, significantBytes);
        for (UIntSize i = 0; i < significantBytes; i++)
        {
            int digitLo = value.getDigit(2 * i);
            int digitHi = value.getDigit(2 * i + 1);
            int bcdByte = (digitLo << 4) | digitHi;
            dos.writeInt8(static_cast<Int8>(bcdByte));
        }
    }
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const DecimalFixed & value) throw(IOException)
{
    dos << ((value.getSign() == -1) ? (-value.getMagnitude()) : (value.getMagnitude()));
    ULeb128::writeUIntSize(dos, value.getScale());
    return dos;
}

IDataOutput & carl::operator << (IDataOutput & dos, const DecimalFloat & value) throw(IOException)
{
    dos.writeUInt8(static_cast<UInt8>(value.getCategory()));
    if (value.isNormal())
    {
        dos << ((value.getSign() == -1) ? (-value.getMantissa()) : (value.getMantissa()));
        dos << value.getScale();
    }
    return dos;
}

//  End of IoOperators.cpp
