//
//  Iso10646.cpp - Implementation of the carl::Iso10646 class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/29 12:14:39 $
//  $Revision: 1.9 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "UnicodeData.hpp"
#include <stdlib.h>

//////////
//  Implementation helpers
namespace 
{
    const _CharacterInfo * _findCharacterInfo(Char c)
    {
        UInt32 level1Index = (c >> 22);
        UInt32 level2Index = (c >> 11) & 0x000007FF;
        UInt32 level3Index = c & 0x000007FF;

        _CharacterInfo *** pTable1Entry = _table1[level1Index];
        if (pTable1Entry == NULL)
        {
            return NULL;
        }
        _CharacterInfo ** pTable2Entry = pTable1Entry[level2Index];
        if (pTable2Entry == NULL)
        {
            return NULL;
        }
        return pTable2Entry[level3Index];
    }
}

//////////
//  Iso10646::_CompositionTree
Iso10646::_CompositionTree::_CompositionTree()
    :   _hasCharacter(false),
        _character(0),
        _subtrees(StandardHashFunctions::hashChar)
{
}

Iso10646::_CompositionTree::_CompositionTree(Char c)
    :   _hasCharacter(true),
        _character(c),
        _subtrees(StandardHashFunctions::hashChar)
{
}

Iso10646::_CompositionTree::~_CompositionTree()
{
    for (Iterator<Pair<Char,_CompositionTree*> > it = _subtrees.enumerateElements(); !it.finished(); it.advance())
    {
        delete it.item().second();
    }
}

bool Iso10646::_CompositionTree::hasCharacter() const
{
    return _hasCharacter;
}

Char Iso10646::_CompositionTree::getCharacter() const
{
    require(_hasCharacter);

    return _character;
}

Iso10646::_CompositionTree * Iso10646::_CompositionTree::findSubtree(Char c)
{
    if (_subtrees.containsKey(c))
    {
        return _subtrees.get(c);
    }
    else
    {
        return NULL;
    }
}

Iso10646::_CompositionTree * Iso10646::_CompositionTree::createSubtree(Char c)
{
    if (_subtrees.containsKey(c))
    {
        return _subtrees.get(c);
    }
    else
    {
        _CompositionTree * pCompositionTree = new _CompositionTree();
        _subtrees.insert(c, pCompositionTree);
        return pCompositionTree;
    }
}

Iso10646::_CompositionTree * Iso10646::_CompositionTree::createSubtree(Char c, Char composition)
{
    if (_subtrees.containsKey(c))
    {
        return _subtrees.get(c);
    }
    else
    {
        _CompositionTree * pCompositionTree = new _CompositionTree(composition);
        _subtrees.insert(c, pCompositionTree);
        return pCompositionTree;
    }
}

//////////
//  Iso10646::_CompositionTreeRoot
Iso10646::_CompositionTreeRoot::_CompositionTreeRoot()
{
    _pRoot = new _CompositionTree();
    //  Build a composition tree
    for (UIntSize level1Index = 0; level1Index < 1024; level1Index++)
    {
        _CharacterInfo *** pTable1Entry = _table1[level1Index];
        if (pTable1Entry != NULL)
        {
            for (UIntSize level2Index = 0; level2Index < 2048; level2Index++)
            {
                _CharacterInfo ** pTable2Entry = pTable1Entry[level2Index];
                if (pTable2Entry != NULL)
                {
                    for (UIntSize level3Index = 0; level3Index < 2048; level3Index++)
                    {
                        _CharacterInfo * pTable3Entry = pTable2Entry[level3Index];
                        if (pTable3Entry != NULL && pTable3Entry->pDecompositionMapping != NULL)
                        {   //  This character has a decomposition
                            Char c = static_cast<Char>((level1Index << 22) | (level2Index << 11) | (level3Index));
                            _CompositionTree * pCompositionTree = _pRoot;
                            for (UIntSize offset = 0; offset < pTable3Entry->decompositionMappingSize; offset++)
                            {
                                _CompositionTree * pSubtree;
                                if (offset + 1 == pTable3Entry->decompositionMappingSize)
                                {
                                    pSubtree = pCompositionTree->createSubtree(pTable3Entry->pDecompositionMapping[offset], c);
                                }
                                else
                                {
                                    pSubtree = pCompositionTree->createSubtree(pTable3Entry->pDecompositionMapping[offset]);
                                }
                                pCompositionTree = pSubtree;
                            }
                        }
                    }
                }
            }
        }
    }
}

Iso10646::_CompositionTreeRoot::~_CompositionTreeRoot()
{
    delete _pRoot;
}

Iso10646::_CompositionTreeRoot * Iso10646::_CompositionTreeRoot::getInstance()
{
    static _CompositionTreeRoot instance;
    return &instance;
}

bool Iso10646::_CompositionTreeRoot::findComposition(const carl::String & s, UIntSize startPosition, Char & composition, UIntSize & compositionLength)
{
    _CompositionTree * pCurrentNode = _pRoot;
    _CompositionTree * pLastMatchedNode = NULL;
    UIntSize lastMatchedSize = 0;
    for (UIntSize offset = 0; ; offset++)
    {
        //  Do not match past the end of string
        UIntSize index = startPosition + offset;
        if (index >= s.length())
        {
            break;
        }
        //  Starters break matching
        Char c = s[index];
        UInt8 thisCharCombiningClass;
        if (offset != 0 && 
            (!getCanonicalCombiningClass(c, thisCharCombiningClass) || thisCharCombiningClass == 0))
        {
            break;
        }
        //  Find a subtree for the current character
        _CompositionTree * pSubnode = pCurrentNode->findSubtree(c);
        if (pSubnode == NULL)
        {
            break;
        }
        //  If this composition exists, record the fact
        if (pSubnode->hasCharacter())
        {
            pLastMatchedNode = pSubnode;
            lastMatchedSize = offset + 1;
        }
        pCurrentNode = pSubnode;
    }
    //  Did we find anything ?
    if (pLastMatchedNode != NULL && lastMatchedSize > 1)
    {
        composition = pLastMatchedNode->getCharacter();
        compositionLength = lastMatchedSize;
        return true;
    }
    else
    {
        return false;
    }
}

//////////
//  Character classification
bool Iso10646::isAlNum(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    return pCharacterInfo->generalCategory == GeneralCategory_Lu ||
           pCharacterInfo->generalCategory == GeneralCategory_Ll ||
           pCharacterInfo->generalCategory == GeneralCategory_Lt ||
           pCharacterInfo->generalCategory == GeneralCategory_Lm ||
           pCharacterInfo->generalCategory == GeneralCategory_Lo ||
           pCharacterInfo->generalCategory == GeneralCategory_Nd ||
           pCharacterInfo->generalCategory == GeneralCategory_Nl ||
           pCharacterInfo->generalCategory == GeneralCategory_No;
}

bool Iso10646::isAlpha(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    return pCharacterInfo->generalCategory == GeneralCategory_Lu ||
           pCharacterInfo->generalCategory == GeneralCategory_Ll ||
           pCharacterInfo->generalCategory == GeneralCategory_Lt ||
           pCharacterInfo->generalCategory == GeneralCategory_Lm ||
           pCharacterInfo->generalCategory == GeneralCategory_Lo;
}

bool Iso10646::isControl(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    return pCharacterInfo->generalCategory == GeneralCategory_Cc ||
           pCharacterInfo->generalCategory == GeneralCategory_Cf ||
           pCharacterInfo->generalCategory == GeneralCategory_Cs ||
           pCharacterInfo->generalCategory == GeneralCategory_Co ||
           pCharacterInfo->generalCategory == GeneralCategory_Cn;
}

bool Iso10646::isDigit(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    return pCharacterInfo->generalCategory == GeneralCategory_Nd ||
           pCharacterInfo->generalCategory == GeneralCategory_Nl ||
           pCharacterInfo->generalCategory == GeneralCategory_No;
}

bool Iso10646::isGraph(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    return pCharacterInfo->generalCategory != GeneralCategory_Cc &&
           pCharacterInfo->generalCategory != GeneralCategory_Cf &&
           pCharacterInfo->generalCategory != GeneralCategory_Cs &&
           pCharacterInfo->generalCategory != GeneralCategory_Co &&
           pCharacterInfo->generalCategory != GeneralCategory_Cn &&
           pCharacterInfo->generalCategory != GeneralCategory_Zs &&
           pCharacterInfo->generalCategory != GeneralCategory_Zl &&
           pCharacterInfo->generalCategory != GeneralCategory_Zp;
}

bool Iso10646::isIso10646(Char c)
{
    return _findCharacterInfo(c) != NULL;
}

bool Iso10646::isLower(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    return pCharacterInfo->generalCategory == GeneralCategory_Ll;
}

bool Iso10646::isPrint(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    return pCharacterInfo->generalCategory != GeneralCategory_Cc &&
           pCharacterInfo->generalCategory != GeneralCategory_Cf &&
           pCharacterInfo->generalCategory != GeneralCategory_Cs &&
           pCharacterInfo->generalCategory != GeneralCategory_Co &&
           pCharacterInfo->generalCategory != GeneralCategory_Cn;
}

bool Iso10646::isPunct(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    return pCharacterInfo->generalCategory == GeneralCategory_Pc ||
           pCharacterInfo->generalCategory == GeneralCategory_Pd ||
           pCharacterInfo->generalCategory == GeneralCategory_Ps ||
           pCharacterInfo->generalCategory == GeneralCategory_Pe ||
           pCharacterInfo->generalCategory == GeneralCategory_Pi ||
           pCharacterInfo->generalCategory == GeneralCategory_Pf ||
           pCharacterInfo->generalCategory == GeneralCategory_Po ||
           pCharacterInfo->generalCategory == GeneralCategory_Sm ||
           pCharacterInfo->generalCategory == GeneralCategory_Sc ||
           pCharacterInfo->generalCategory == GeneralCategory_Sk ||
           pCharacterInfo->generalCategory == GeneralCategory_So;
}

bool Iso10646::isSpace(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    else
    {
        switch (pCharacterInfo->generalCategory)
        {
            case GeneralCategory_Zs:
            case GeneralCategory_Zl:
            case GeneralCategory_Zp:
                return true;
            default:
                return (c >= 0x00000009 && c <= 0x0000000D) || (c == 0x00000020);
        }
    }
}

bool Iso10646::isTitle(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    return pCharacterInfo->generalCategory == GeneralCategory_Lt;
}

bool Iso10646::isUnicode(Char c)
{
    return (c < 0x00200000) && (_findCharacterInfo(c) != NULL);
}

bool Iso10646::isUpper(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    return pCharacterInfo->generalCategory == GeneralCategory_Lu;
}

//////////
//  Character case conversion
Char Iso10646::toLower(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return c;
    }
    else
    {
        return pCharacterInfo->lowercaseMapping;
    }
}

Char Iso10646::toUpper(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return c;
    }
    else
    {
        return pCharacterInfo->uppercaseMapping;
    }
}

Char Iso10646::toTitle(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return c;
    }
    else
    {
        return pCharacterInfo->titlecaseMapping;
    }
}

//////////
//  Direct access to ISO-10646 data
bool Iso10646::getGeneralCategory(Char c, GeneralCategory & generalCategory)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    else
    {
        generalCategory = pCharacterInfo->generalCategory;
        return true;
    }
}

bool Iso10646::getCanonicalCombiningClass(Char c, UInt8 & canonicalCombiningClass)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    else
    {
        canonicalCombiningClass = pCharacterInfo->canonicalCombiningClass;
        return true;
    }
}

bool Iso10646::getBidirectionalCategory(Char c, BidirectionalCategory & bidirectionalCategory)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    else
    {
        bidirectionalCategory = pCharacterInfo->bidirectionalCategory;
        return true;
    }
}

bool Iso10646::getDecompositionMappingTag(Char c, DecompositionMappingTag & decompositionMappingTag)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL)
    {
        return false;
    }
    else
    {
        decompositionMappingTag = pCharacterInfo->decompositionMappingTag;
        return true;
    }
}

bool Iso10646::getDecompositionMappingSize(Char c, UIntSize & decompositionMappingSize)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL || pCharacterInfo->decompositionMappingSize == 0)
    {
        return false;
    }
    else
    {
        decompositionMappingSize = pCharacterInfo->decompositionMappingSize;
        return true;
    }
}

bool Iso10646::getDecompositionMapping(Char c, Char * pDecompositionMapping)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL || pCharacterInfo->decompositionMappingSize == 0)
    {
        return false;
    }
    else
    {
        for (UInt8 i = 0; i < pCharacterInfo->decompositionMappingSize; i++)
        {
            pDecompositionMapping[i] = pCharacterInfo->pDecompositionMapping[i];
        }
        return true;
    }
}

bool Iso10646::getDecimalDigitValue(Char c, UInt8 & decimalDigitValue)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL || pCharacterInfo->hasDecimalDigitValue)
    {
        return false;
    }
    else
    {
        decimalDigitValue = pCharacterInfo->decimalDigitValue;
        return true;
    }
}

bool Iso10646::getDigitValue(Char c, Int64 & digitValue)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL || pCharacterInfo->hasDigitValue)
    {
        return false;
    }
    else
    {
        digitValue = pCharacterInfo->digitValue;
        return true;
    }
}

bool Iso10646::getNumericValue(Char c, Int64 & numerator, Int64 & denominator)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    if (pCharacterInfo == NULL || pCharacterInfo->hasNumericValue)
    {
        return false;
    }
    else
    {
        numerator = pCharacterInfo->numericValueNumerator;
        denominator = pCharacterInfo->numericValueDenominator;
        return true;
    }
}

bool Iso10646::isMirrored(Char c)
{
    const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
    return (pCharacterInfo != NULL) && (pCharacterInfo->mirrored);
}

//  Character <-> digit conversion
bool Iso10646::charToDigit(Char c, int radix, int & digit)
{
    require(radix >= 2 && radix <= 36);

    if (c >= '0' && c <= '9')
    {
        digit = c - '0';
    }
    else if (c >= 'A' && c <= 'Z')
    {
        digit = c - 'A' + 10;
    }
    else if (c >= 'a' && c <= 'z')
    {
        digit = c - 'a' + 10;
    }
    else
    {
        const _CharacterInfo * pCharacterInfo = _findCharacterInfo(c);
        if (pCharacterInfo == NULL || !pCharacterInfo->hasDigitValue || 
            pCharacterInfo->digitValue < 0  || pCharacterInfo->digitValue >= radix)
        {
            return false;
        }
        else
        {
            digit = static_cast<int>(pCharacterInfo->digitValue);
        }
    }
    return digit < radix;
}

bool Iso10646::digitToChar(int digit, int radix, Char & c)
{
    require(radix >= 2 && radix <= 36);

    if (digit < 0 || digit >= radix)
    {
        return false;
    }
    else
    {
        static char digitChars[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
        c = digitChars[digit];
        return true;
    }
}

String Iso10646::performCanonicalOrdering(const String & s)
{
    if (_isAscii7String(s))
    {
        return s;
    }

    //  Find the leftmost position where an exchangable pair occurs
    UIntSize firstExchangablePairIndex = s.length();
    for (UIntSize i = 0; i < s.length(); i++)
    {
        UInt8 thisCharCombiningClass, nextCharCombiningClass;
        if (i + 1 < s.length() &&
            getCanonicalCombiningClass(s[i], thisCharCombiningClass) &&
            getCanonicalCombiningClass(s[i + 1], nextCharCombiningClass) &&
            thisCharCombiningClass > nextCharCombiningClass && 
            nextCharCombiningClass > 0)
        {   //  Found an exchangable pair!
            firstExchangablePairIndex = i;
            break;
        }
    }
    if (firstExchangablePairIndex == s.length())
    {   //  There are no exchangable pairs in the string
        return s;
    }
    else
    {   //  The string has exchangable pairs
        String result = s.left(firstExchangablePairIndex);
        for (UIntSize i = firstExchangablePairIndex; i < s.length();)
        {
            UInt8 thisCharCombiningClass, nextCharCombiningClass;
            if (i + 1 == s.length() ||
                !getCanonicalCombiningClass(s[i], thisCharCombiningClass) ||
                !getCanonicalCombiningClass(s[i + 1], nextCharCombiningClass))
            {
                result += s[i++];
            }
            else
            {
                if (thisCharCombiningClass > nextCharCombiningClass && nextCharCombiningClass > 0)
                {
                    result += s[i + 1];
                    result += s[i];
                    i += 2;
                }
                else
                {
                    result += s[i++];
                }
            }
        }
        return result;
    }
}

String Iso10646::performCanonicalDecomposition(const String & s)
{
    if (_isAscii7String(s))
    {
        return s;
    }

    return performCanonicalOrdering(_performCanonicalDecompositionWithoutOrdering(s));
}

String Iso10646::performCompatibilityDecomposition(const String & s)
{
    if (_isAscii7String(s))
    {
        return s;
    }

    return performCanonicalOrdering(_performCompatibilityDecompositionWithoutOrdering(s));
}

String Iso10646::performCanonicalComposition(const String & s)
{
    if (_isAscii7String(s))
    {
        return s;
    }

    _CompositionTreeRoot * pCompositionTreeRoot = _CompositionTreeRoot::getInstance();
    String result;
    for (UIntSize i = 0; i < s.length(); )
    {
        Char composition;
        UIntSize compositionLength;
        if (pCompositionTreeRoot->findComposition(s, i, composition, compositionLength))
        {
            result += composition;
            i += compositionLength;
        }
        else
        {
            result += s[i++];
        }
    }
    return result;
}

//////////
//  Implementation helpers
String Iso10646::_performCanonicalDecompositionWithoutOrdering(const String & s)
{
    String result;
    for (UIntSize i = 0; i < s.length(); i++)
    {
        Char c = s[i];
        DecompositionMappingTag decompositionMappingTag;
        UIntSize decompositionMappingSize;
        if (getDecompositionMappingTag(c, decompositionMappingTag) &&
            decompositionMappingTag != DecompositionMappingTag_compat &&
            getDecompositionMappingSize(c, decompositionMappingSize))
        {
            Array<Char> decompositionData;
            decompositionData.setSize(decompositionMappingSize);
            getDecompositionMapping(c, decompositionData.getData());
            String decompositionString(decompositionData.getData(), decompositionData.size());
            result += _performCanonicalDecompositionWithoutOrdering(decompositionString);
        }
        else
        {
            result += c;
        }
    }
    return result;
}

String Iso10646::_performCompatibilityDecompositionWithoutOrdering(const String & s)
{
    String result;
    for (UIntSize i = 0; i < s.length(); i++)
    {
        Char c = s[i];
        UIntSize decompositionMappingSize;
        if (getDecompositionMappingSize(c, decompositionMappingSize))
        {
            Array<Char> decompositionData;
            decompositionData.setSize(decompositionMappingSize);
            getDecompositionMapping(c, decompositionData.getData());
            String decompositionString(decompositionData.getData(), decompositionData.size());
            result += _performCanonicalDecompositionWithoutOrdering(decompositionString);
        }
        else
        {
            result += c;
        }
    }
    return result;
}

bool Iso10646::_isAscii7String(const String & s)
{
    for (UIntSize i = 0; i < s.length(); i++)
    {
        if ((s[i] & 0xFFFFFF80) != 0)
        {
            return false;
        }
    }
    return true;
}

//  End of Iso10646.cpp
