//
//  Leb128.cpp - Implementation of the carl::Leb128 class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:53 $
//  $Revision: 1.6 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

//////////
//  Operations
void Leb128::writeInt64(IByteOutput & bos, Int64 value) throw(IOException)
{
    for (bool more = true; more; )
    {
        UInt8 lowByte = static_cast<UInt8>(value & 0x7F);
        value >>= 7;
        if ((value == 0 && (lowByte & 0x40) == 0) ||
            (value == -1 && (lowByte & 0x40) != 0))
        {
            more = false;
        }
        else
        {
            lowByte = static_cast<UInt8>(lowByte | 0x80);
        }
        bos.write(lowByte);
    }
}

void Leb128::writeBinaryInteger(IByteOutput & bos, const BinaryInteger & value) throw(IOException)
{
    BinaryInteger v = value;
    for (bool more = true; more; )
    {
        UInt8 lowByte = static_cast<UInt8>( v[0]       |
                                           (v[1] << 1) |
                                           (v[2] << 2) |
                                           (v[3] << 3) |
                                           (v[4] << 4) |
                                           (v[5] << 5) |
                                           (v[6] << 6));
        v >>= 7;
        if ((v == 0 && (lowByte & 0x40) == 0) ||
            (v == -1 && (lowByte & 0x40) != 0))
        {
            more = false;
        }
        else
        {
            lowByte = static_cast<UInt8>(lowByte | 0x80);
        }
        bos.write(lowByte);
    }
}

Int8 Leb128::readInt8(IByteInput & bis) throw(IOException)
{
    Int64 result = readInt64(bis);
    if (static_cast<Int8>(result) != result)
    {   //  Overflow
        throw InvalidDataException();
    }
    return static_cast<Int8>(result);
}

Int16 Leb128::readInt16(IByteInput & bis) throw(IOException)
{
    Int64 result = readInt64(bis);
    if (static_cast<Int16>(result) != result)
    {   //  Overflow
        throw InvalidDataException();
    }
    return static_cast<Int16>(result);
}

Int32 Leb128::readInt32(IByteInput & bis) throw(IOException)
{
    Int64 result = readInt64(bis);
    if (static_cast<Int32>(result) != result)
    {   //  Overflow
        throw InvalidDataException();
    }
    return static_cast<Int32>(result);
}

Int64 Leb128::readInt64(IByteInput & bis) throw(IOException)
{
    Int64 result = 0;
    UIntSize shift = 0;
    int b;
    bool zeroBitsLost = false, oneBitsLost = false;
    for (; ; )
    {
        b = bis.read();
        if (b == -1)
        {
            throw EOFException();
        }
        UInt64 newResult = result | (static_cast<UInt64>(b & 0x7F) << shift);
        UIntSize lostBits;  //  The number of higher bits of b that are lost
        if (shift > 63)
        {
            lostBits = 7;
        }
        else if (shift > 56)
        {
            lostBits = shift - 56;
        }
        else
        {
            lostBits = 0;
        }
        for (UIntSize i = 0; i < lostBits; i++)
        {
            if ((b & (0x40 >> i)) != 0)
            {
                oneBitsLost = true;
            }
            else
            {
                zeroBitsLost = true;
            }
        }
        result = newResult;
        shift += 7;
        if ((b & 0x80) == 0)
        {
            break;
        }
    }
    if ((b & 0x40) != 0)
    {   //  Negative
        if (shift < 64)
        {
            result |= -(CARL_INT64_CONSTANT(1) << shift);
        }
        if (zeroBitsLost)
        {
            throw InvalidDataException();
        }
    }
    else
    {   //  Positive
        if (oneBitsLost)
        {
            throw InvalidDataException();
        }
    }
    return result;
}

BinaryInteger Leb128::readBinaryInteger(IByteInput & bis) throw(IOException)
{
    BinaryInteger result = 0;
    UIntSize shift = 0;
    int b;
    for (; ; )
    {
        b = bis.read();
        if (b == -1)
        {
            throw EOFException();
        }
        result |= BinaryInteger(b & 0x7F) << shift;
        shift += 7;
        if ((b & 0x80) == 0)
        {
            break;
        }
    }
    if ((b & 0x40) != 0)
    {
        result |= -(BinaryInteger(1) << shift);
    }
    return result;
}

IntPtrDiff Leb128::readIntPtrDiff(IByteInput & bis) throw(IOException)
{
    Int64 result = readInt64(bis);
    if (static_cast<IntPtrDiff>(result) != result)
    {   //  Overflow
        throw InvalidDataException();
    }
    return static_cast<IntPtrDiff>(result);
}

//  End of Leb128.cpp
