//
//  Library.cpp - Implementation of the carl::Library class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/07/28 10:40:28 $
//  $Revision: 1.15 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include <stdlib.h>
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <windows.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <dlfcn.h>
    #include <string.h>
#else
    #error Unsupported platform
#endif

//////////
//  Library properties
const char * Library::PreferredExtension = ".dll";

//////////
//  Construction/destruction
Library::Library()
{
    _isLoaded = false;
    _pHandle = NULL;
}

Library::Library(const char * pFileName, bool load /*= true*/) throw(IOException)
{
    _isLoaded = false;
    _pHandle = NULL;
    if (load && !_load(String(pFileName)))
    {
        throw IOException();
    }
}

Library::Library(const Char * pFileName, bool load /*= true*/) throw(IOException)
{
    _isLoaded = false;
    _pHandle = NULL;
    if (load && !_load(String(pFileName)))
    {
        throw IOException();
    }
}

Library::Library(const String & fileName, bool load /*= true*/) throw(IOException)
{
    _isLoaded = false;
    _pHandle = NULL;
    if (load && !_load(String(fileName)))
    {
        throw IOException();
    }
}

Library::Library(const PathName & pathName, bool load /*= true*/) throw(IOException)
{
    _isLoaded = false;
    _pHandle = NULL;
    if (load && !_load(pathName.toString()))
    {
        throw IOException();
    }
}

Library::~Library()
{
    if (_isLoaded)
    {
        unload();
    }
}


//////////
//  Operations
String Library::getFileName() const
{
    return _fileName;
}

void Library::setFileName(const char * pFileName)
{
    require(!_isLoaded);

    _fileName = String(pFileName);
}

void Library::setFileName(const Char * pFileName)
{
    require(!_isLoaded);

    _fileName = String(pFileName);
}

void Library::setFileName(const String & fileName)
{
    require(!_isLoaded);

    _fileName = fileName;
}

void Library::setFileName(const PathName & pathName)
{
    require(!_isLoaded);

    _fileName = pathName.toString();
}

bool Library::load()
{
    require(!_isLoaded);

    return _load(_fileName);
}

bool Library::load(const char * pFileName)
{
    require(!_isLoaded);

    return _load(String(pFileName));
}

bool Library::load(const Char * pFileName)
{
    require(!_isLoaded);

    return _load(String(pFileName));
}

bool Library::load(const String & fileName)
{
    require(!_isLoaded);

    return _load(fileName);
}

bool Library::load(const PathName & pathName)
{
    require(!_isLoaded);

    return _load(pathName.toString());
}

bool Library::unload()
{
    require(_isLoaded);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    DWORD oldLastError = GetLastError();
    BOOL b = FreeLibrary(static_cast<HMODULE>(_pHandle));
    SetLastError(oldLastError);
    return b != 0;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return dlclose(_pHandle) == 0;
#else
    #error Unsupported platform
#endif
}

void * Library::resolve(const char * pSymbolName)
{
    require(_isLoaded);

    return resolve(String(pSymbolName));
}

void * Library::resolve(const Char * pSymbolName)
{
    require(_isLoaded);

    return resolve(String(pSymbolName));
}

void * Library::resolve(const String & symbolName)
{
    require(_isLoaded);

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<char> asciiName;
    asciiName.setSize(symbolName.length() + 1);
    if (!symbolName._store(asciiName.getData(), asciiName.size()))
    {
        return NULL;
    }
    DWORD oldLastError = GetLastError();
    FARPROC fp = GetProcAddress(static_cast<HMODULE>(_pHandle), asciiName.getData());
    SetLastError(oldLastError);
    return static_cast<void*>(fp);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> asciiName;
    asciiName.setSize(symbolName.length() + 1);
    if (!symbolName._store(asciiName.getData(), asciiName.size()))
    {
        return NULL;
    }
    return dlsym(_pHandle, asciiName.getData());
#else
    #error Unsupported platform
#endif
}

//  Implementation helpers
bool Library::_load(const String & fileName)
{
    require(!_isLoaded);

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> libName;
    //  Try loading "as is"
    libName.setSize(fileName.length() + 1);
    if (fileName._store(libName.getData(), libName.size()))
    {
        DWORD oldLastError = GetLastError();
        HMODULE hm = LoadLibraryW(libName.getData());
        SetLastError(oldLastError);
        if (hm != NULL)
        {
            _pHandle = static_cast<void*>(hm);
            _isLoaded = true;
            return true;
        }
    }
    //  Try adding standard platform extension
    libName.setSize(fileName.length() + strlen(PreferredExtension) + 1);
    if ((fileName + PreferredExtension)._store(libName.getData(), libName.size()))
    {
        DWORD oldLastError = GetLastError();
        HMODULE hm = LoadLibraryW(libName.getData());
        SetLastError(oldLastError);
        if (hm != NULL)
        {
            _pHandle = static_cast<void*>(hm);
            _isLoaded = true;
            return true;
        }
    }
    //  Can't load
    return false;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> libName;
    //  Try loading "as is"
    libName.setSize(fileName.length() + 1);
    if (fileName._store(libName.getData(), libName.size()))
    {
        _pHandle = dlopen(libName.getData(), RTLD_NOW | RTLD_GLOBAL);
        if (_pHandle != NULL)
        {
            _isLoaded = true;
            return true;
        }
    }
    //  Try adding standard platform extension
    libName.setSize(fileName.length() + strlen(PreferredExtension) + 1);
    if ((fileName + PreferredExtension)._store(libName.getData(), libName.size()))
    {
        _pHandle = dlopen(libName.getData(), RTLD_NOW | RTLD_GLOBAL);
        if (_pHandle != NULL)
        {
            _isLoaded = true;
            return true;
        }
    }
    //  Can't load
    return false;
#else
    #error Unsupported platform
#endif
}

//  End of Library.cpp
