//
//  MapResourceFactory.cpp - Implementation of the carl::MapResourceFactory class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:53 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include <stdlib.h>

//////////
//  Construction/destruction
MapResourceFactory::MapResourceFactory(ResourceMap * pDefaultResourceMap)
    :   _mapLocalesToResourceMaps(StandardHashFunctions::hashLocale)
{
    require(pDefaultResourceMap != NULL);

    _pDefaultResourceMap = pDefaultResourceMap;
    _mapLocalesToResourceMaps.insert(Locale(), pDefaultResourceMap);
}

MapResourceFactory::~MapResourceFactory()
{
    Array<ResourceMap*> resourceMapsToDelete;
    for (Iterator< Pair<Locale, ResourceMap*> > it = _mapLocalesToResourceMaps.enumerateElements();
         !it.finished();
         it.advance())
    {
        if (!resourceMapsToDelete.contains(it.item().second()))
        {
            resourceMapsToDelete.append(it.item().second());
        }
    }
    for (UIntSize i = 0; i < resourceMapsToDelete.size(); i++)
    {
        delete resourceMapsToDelete[i];
    }
}

//////////
//  Operations
Array<Locale> MapResourceFactory::getSupportedLocales()
{
    Array<Locale> result;
    for (Iterator<Locale> it = _mapLocalesToResourceMaps.enumerateKeys(); !it.finished(); it.advance())
    {
        result.append(it.item());
    }
    return result;
}

String MapResourceFactory::getString(const char * pKey)
{
    require(pKey != NULL);

    return getString(String(pKey), Locale::getDefault());
}

String MapResourceFactory::getString(const Char * pKey)
{
    require(pKey != NULL);

    return getString(String(pKey), Locale::getDefault());
}

String MapResourceFactory::getString(const String & key)
{
    return getString(key, Locale::getDefault());
}

String MapResourceFactory::getString(const char * pKey, const Locale & forLocale)
{
    require(pKey != NULL);

    return getString(String(pKey), forLocale);
}

String MapResourceFactory::getString(const Char * pKey, const Locale & forLocale)
{
    require(pKey != NULL);

    return getString(String(pKey), forLocale);
}

String MapResourceFactory::getString(const String & key, const Locale & forLocale)
{
    //  Try specified locale and its parents
    for (Locale locale = forLocale; !locale.isInvariant(); locale = locale.getParent())
    {
Retry1:
        if (_mapLocalesToResourceMaps.containsKey(locale))
        {   //  Resource map exists for this locale
            try
            {
                return _mapLocalesToResourceMaps.get(locale)->getString(key);
            }
            catch (MissingResourceError)
            {
            }
        }
        else
        {   //  There is no resource map for this locale
            switch (handleMissingLocale(locale))
            {
                case HandleMissingLocaleOutcome_Retry:
                    goto Retry1;
                case HandleMissingLocaleOutcome_Abort:
                    throw MissingResourceError(key);
                case HandleMissingLocaleOutcome_UseDefault:
                    break;
                default:
                    require(false);
                    break;
            }
        }
    }
    //  Use default resource map
    return _pDefaultResourceMap->getString(key);
}

void MapResourceFactory::addResourceMap(const Locale & forLocale, ResourceMap * pResourceMap)
{
    require(pResourceMap != NULL);

    if (_mapLocalesToResourceMaps.containsKey(forLocale))
    {   //  The resource map for this locale is already aggregated. Note
        //  that even default resource map can be replaced.
        //  Delete the old resource map if that was the last reference
        ResourceMap * pOldResourceMap = _mapLocalesToResourceMaps.get(forLocale);
        UIntSize oldResourceMapReferenceCount = 0;
        for (Iterator< Pair<Locale, ResourceMap*> > it = _mapLocalesToResourceMaps.enumerateElements();
             !it.finished();
             it.advance())
        {
            if (it.item().second() == pOldResourceMap)
            {
                oldResourceMapReferenceCount++;
            }
        }
        require(oldResourceMapReferenceCount != 0);
        bool replaceDefaultResourceMap = (pOldResourceMap == _pDefaultResourceMap);
        if (oldResourceMapReferenceCount == 1)
        {   //  Yes, that was the last reference
            delete pOldResourceMap;
        }
        //  Set up a new association
        if (replaceDefaultResourceMap)
        {
            _pDefaultResourceMap = pResourceMap;
        }
        _mapLocalesToResourceMaps.insert(forLocale, pResourceMap);
    }
    else
    {   //  The resource map for this locale is not yet aggregated
        _mapLocalesToResourceMaps.insert(forLocale, pResourceMap);
    }
}

//////////
//  Overridables
MapResourceFactory::HandleMissingLocaleOutcome MapResourceFactory::handleMissingLocale(const Locale & /*forLocale*/)
{
    return HandleMissingLocaleOutcome_UseDefault;
}

//  End of MapResourceFactory.cpp
