//
//  MultiByte2CharacterSet.cpp - Implementation of the carl::_impl::_MultiByte2CharacterSet class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:53 $
//  $Revision: 1.5 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "CharsetImpl.hpp"

//////////
//  Construction/destruction - from derived classes only
_impl::_MultiByte2CharacterSet::_MultiByte2CharacterSet(const Char * pSingleByteCharTable256, const Char *const *ppDoubleByteCharTables256)
    :   _pSingleByteCharTable256(pSingleByteCharTable256),
        _ppDoubleByteCharTables256(ppDoubleByteCharTables256),
        _mapSingleByteCharsToBytes(StandardHashFunctions::hashChar),
        _mapDoubleByteCharsToBytePairs(StandardHashFunctions::hashChar)
{
    require(pSingleByteCharTable256 != NULL);
    require(ppDoubleByteCharTables256 != NULL);

    _initialized = false;
}

_impl::_MultiByte2CharacterSet::~_MultiByte2CharacterSet()
{
}

//////////
//  Operations
int _impl::_MultiByte2CharacterSet::getMaximumEncodingLength() const
{
    return 2;
}

int _impl::_MultiByte2CharacterSet::getEncodingLengthForCharacter(Char c) const
{
    _initialize();
    if (_mapSingleByteCharsToBytes.containsKey(c))
    {
        return 1;
    }
    if (_mapDoubleByteCharsToBytePairs.containsKey(c))
    {
        return 2;
    }
    return -1;
}

int _impl::_MultiByte2CharacterSet::encodeCharacter(Char c, void * buffer) const
{
    _initialize();
    if (_mapSingleByteCharsToBytes.containsKey(c))
    {
        UInt8 * pBytes = static_cast<UInt8*>(buffer);
        pBytes[0] = _mapSingleByteCharsToBytes.get(c);
        return 1;
    }
    if (_mapDoubleByteCharsToBytePairs.containsKey(c))
    {
        UInt8 * pBytes = static_cast<UInt8*>(buffer);
        UInt16 bytes = _mapDoubleByteCharsToBytePairs.get(c);
        pBytes[0] = static_cast<UInt8>(bytes >> 8);
        pBytes[1] = static_cast<UInt8>(bytes);
        return 2;
    }
    return -1;
}

int _impl::_MultiByte2CharacterSet::getEncodingLengthForLeadByte(UInt8 leadByte) const
{
    if (_pSingleByteCharTable256[leadByte] != 0xFFFFFFFF)
    {
        return 1;
    }
    if (_ppDoubleByteCharTables256[leadByte] != NULL)
    {
        return 2;
    }
    return -1;
}

int _impl::_MultiByte2CharacterSet::decodeCharacter(const void * buffer, UIntSize bytesAvailable, Char & c) const
{
    const UInt8 * pBytes = static_cast<const UInt8*>(buffer);
    if (bytesAvailable >= 1 && _pSingleByteCharTable256[pBytes[0]] != 0xFFFFFFFF)
    {
        c = _pSingleByteCharTable256[pBytes[0]];
        return 1;
    }
    if (bytesAvailable >= 2 && _ppDoubleByteCharTables256[pBytes[0]] != NULL && _ppDoubleByteCharTables256[pBytes[0]][pBytes[1]] != 0xFFFFFFFF)
    {
        c = _ppDoubleByteCharTables256[pBytes[0]][pBytes[1]];
        return 2;
    }
    return -1;
}

//////////
//  Implementation helpers
void _impl::_MultiByte2CharacterSet::_initialize() const
{
    if (!_initialized)
    {
        for (int i = 0; i < 256; i++)
        {
            if (_pSingleByteCharTable256[i] != 0xFFFFFFFF)
            {
                if(_mapSingleByteCharsToBytes.containsKey(_pSingleByteCharTable256[i]))
                {
                    throw InconsistentCharacterSetException();
                }
                _mapSingleByteCharsToBytes.insert(_pSingleByteCharTable256[i], static_cast<UInt8>(i));
            }
        }
        for (int i = 0; i < 256; i++)
        {
            if (_ppDoubleByteCharTables256[i] != NULL)
            {
                if (_pSingleByteCharTable256[i] != 0xFFFFFFFF)
                {
                    throw InconsistentCharacterSetException();
                }
                for (int j = 0; j < 256; j++)
                {
                    if (_ppDoubleByteCharTables256[i][j] != 0xFFFFFFFF)
                    {
                        _mapDoubleByteCharsToBytePairs.insert(_ppDoubleByteCharTables256[i][j], static_cast<UInt16>((i << 8) | j));
                    }
                }
            }
        }

        _initialized = true;
    }
}

//  End of MultiByte2CharacterSet.cpp
