//
//  MutexImpl.cpp - Implementation of the carl::_impl::_MutexImpl class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:53 $
//  $Revision: 1.9 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "MutexImpl.hpp"

//////////
//  Operations
void _impl::_MutexImpl::initialize()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _osMutex = CreateMutex(NULL, FALSE, NULL);
    if (_osMutex == NULL)
    {
        throw SystemError();
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    pthread_mutexattr_t mutexAttr;
    switch (pthread_mutexattr_init(&mutexAttr))
    {
        case 0:
            break;
        case ENOMEM:
            throw OutOfMemoryError(0);
        default:
            throw SystemError();
    }
    if (pthread_mutexattr_settype(&mutexAttr, PTHREAD_MUTEX_RECURSIVE) != 0)
    {
        throw SystemError();
    }
    switch (pthread_mutex_init(&_mutexData, &mutexAttr))
    {
        case 0:
            break;
        case ENOMEM:
            throw OutOfMemoryError(0);
        default:
            throw SystemError();
    }
#else
    #error Unsupported platform
#endif
}

void _impl::_MutexImpl::deinitialize()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    CloseHandle(_osMutex);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    switch (pthread_mutex_destroy(&_mutexData))
    {
        case 0:
            break;
        case EBUSY:
            throw ProgramError();
        default:
            throw SystemError();
    }
#else
    #error Unsupported platform
#endif
}

void _impl::_MutexImpl::waitFor()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    DWORD oldLastError = GetLastError();
    DWORD waitResult = WaitForSingleObject(_osMutex, INFINITE);
    SetLastError(oldLastError);
    switch (waitResult)
    {
        case WAIT_ABANDONED:
        case WAIT_OBJECT_0:
            return;
        default:
            throw SystemError();
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    switch (pthread_mutex_lock(&_mutexData))
    {
        case 0:
            return;
        case EINVAL:
        case EAGAIN:
        case EDEADLK:
            throw ProgramError();
        default:
            throw SystemError();
    }
#else
    #error Unsupported platform
#endif
}

bool _impl::_MutexImpl::waitFor(const TimeSpan & timeout)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    TimeStamp waitStartTime = TimeStamp::now();
    TimeStamp waitEndTime = waitStartTime + timeout;
    for (; ; )
    {
        Int64 millisecondsToWait = timeout.getNanoseconds() / 1000000000;
        if (millisecondsToWait > 1000)
        {
            millisecondsToWait = 1000;
        }
        DWORD oldLastError = GetLastError();
        DWORD waitResult = WaitForSingleObject(_osMutex, static_cast<DWORD>(millisecondsToWait));
        SetLastError(oldLastError);
        switch (waitResult)
        {
            case WAIT_ABANDONED:
            case WAIT_OBJECT_0:
                return true;
            case WAIT_TIMEOUT:
                if (TimeStamp::now() >= waitEndTime)
                {   //  Waited for the whole timeout
                    return false;
                }
                //  Keep trying
                break;
            default:
                throw SystemError();
        }
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    TimeStamp waitStartTime = TimeStamp::now();
    TimeStamp waitEndTime = waitStartTime + timeout;
    for (; ; )
    {
        switch (pthread_mutex_trylock(&_mutexData))
        {
            case 0:
                return true;
            case EBUSY:
                if (TimeStamp::now() > waitEndTime)
                {
                    return false;
                }
                {
                    int oldErrno = errno;
                    if (sched_yield() != 0)
                    {
                        errno = oldErrno;
                        throw SystemError();
                    }
                }
                break;
            case EINVAL:
            case EAGAIN:
                throw ProgramError();
            default:
                throw SystemError();
        }
    }
    return false;
#else
    #error Unsupported platform
#endif
}

void _impl::_MutexImpl::release()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    ReleaseMutex(_osMutex);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    switch (pthread_mutex_unlock(&_mutexData))
    {
        case 0:
            break;
        case EPERM:
            throw ProgramError();
        default:
            throw SystemError();
    }
#else
    #error Unsupported platform
#endif
}

//  End of MutexImpl.cpp
