//
//  NumberParseImpl.hpp - carl::NumberParse implementation helpers
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:53 $
//  $Revision: 1.13 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#ifndef __NumberParseImpl_hpp_included__
#define __NumberParseImpl_hpp_included__

namespace carl
{
    namespace _impl
    {
        //////////
        //  Number parse manager
        class CARL_PUBLIC _NumberParseManager
        {
            //////////
            //  Can't assign or copy construct
        private:
            _NumberParseManager(const _NumberParseManager &);
            void                            operator = (const _NumberParseManager &);

            //////////
            //  Singleton
        private:
            _NumberParseManager();
            ~_NumberParseManager();
        public:
            static _NumberParseManager *    getInstance();

            //////////
            //  Operations
        public:
            void                            registerNumberParse(const Locale & forLocale, NumberParse * pNumberParse);
            NumberParse *                   getNumberParse(const Locale & forLocale);

            //////////
            //  Implementation
        private:
            HashMap<Locale, NumberParse*>   _mapLocalesToNumberParses;
            NumberParse *                   _pDefaultNumberParse;
            CriticalSection                 _guard;
        };

        //////////
        //  Helper classes used by concrete number parsers for input unification
        class CARL_PUBLIC _IParsedString
        {
            //////////
            //  Construction/destruction
        public:
            _IParsedString() {}
            virtual ~_IParsedString() {}

            //////////
            //  Operations
        public:
            virtual bool                    atEnd(UIntSize position) const = 0;
            virtual Char                    charAt(UIntSize position) const = 0;
            virtual String                  toString() const = 0;
        };

        class CARL_PUBLIC _ParsedAsciiString : public _IParsedString
        {
            //////////
            //  Can't assign or copy construct
        private:
            _ParsedAsciiString(const _ParsedAsciiString &);
            void                    operator = (const _ParsedAsciiString &);

            //////////
            //  Construction/destruction
        public:
            _ParsedAsciiString(const char * pString) : _pString(pString) {}
            virtual ~_ParsedAsciiString() {}

            //////////
            //  _IParsedString
        public:
            virtual bool                    atEnd(UIntSize position) const { return _pString[position] == 0; }
            virtual Char                    charAt(UIntSize position) const { return static_cast<unsigned char>(_pString[position]); }
            virtual String                  toString() const { return String(_pString); }

            //////////
            //  Implementation
        private:
            const char *                    _pString;
        };
        
        class CARL_PUBLIC _ParsedIso10646String : public _IParsedString
        {
            //////////
            //  Can't assign or copy construct
        private:
            _ParsedIso10646String(const _ParsedIso10646String &);
            void                    operator = (const _ParsedIso10646String &);

            //////////
            //  Construction/destruction
        public:
            _ParsedIso10646String(const Char * pString) : _pString(pString) {}
            virtual ~_ParsedIso10646String() {}

            //////////
            //  _IParsedString
        public:
            virtual bool                    atEnd(UIntSize position) const { return _pString[position] == 0; }
            virtual Char                    charAt(UIntSize position) const { return _pString[position]; }
            virtual String                  toString() const { return String(_pString); }

            //////////
            //  Implementation
        private:
            const Char *                    _pString;
        };

        class CARL_PUBLIC _ParsedCARLString : public _IParsedString
        {
            //////////
            //  Can't assign or copy construct
        private:
            _ParsedCARLString(const _ParsedCARLString &);
            void                    operator = (const _ParsedCARLString &);

            //////////
            //  Construction/destruction
        public:
            _ParsedCARLString(const String & string) : _string(string) {}
            virtual ~_ParsedCARLString() {}

            //////////
            //  _IParsedString
        public:
            virtual bool                    atEnd(UIntSize position) const { return position >= _string.length(); }
            virtual Char                    charAt(UIntSize position) const { return _string[position]; }
            virtual String                  toString() const { return _string; }

            //////////
            //  Implementation
        private:
            const String &                  _string;
        };

        //////////
        //  Default (invariant) number parse
        class CARL_PUBLIC _DefaultNumberParse : public NumberParse
        {
            //////////
            //  Can't assign or copy construct
        private:
            _DefaultNumberParse(const _DefaultNumberParse &);
            void                    operator = (const _DefaultNumberParse &);

            //////////
            //  Construction/destruction
        public:
            _DefaultNumberParse(Char decimalPointChar, Char separatorChar);
            virtual ~_DefaultNumberParse();

            //////////
            //  Operations
        public:
            //  Custom parsing
            virtual Int8            parseInt8(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Int16           parseInt16(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Int32           parseInt32(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Int64           parseInt64(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt8           parseUInt8(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt16          parseUInt16(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt32          parseUInt32(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt64          parseUInt64(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual BinaryInteger   parseBinaryInteger(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException);
            virtual DecimalInteger  parseDecimalInteger(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException);
            virtual Real32          parseReal32(const char * pString, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Real64          parseReal64(const char * pString, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual BinaryFixed     parseBinaryFixed(const char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual BinaryFloat     parseBinaryFloat(const char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual DecimalFixed    parseDecimalFixed(const char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual DecimalFloat    parseDecimalFloat(const char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual Int8            parseInt8(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Int16           parseInt16(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Int32           parseInt32(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Int64           parseInt64(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt8           parseUInt8(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt16          parseUInt16(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt32          parseUInt32(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt64          parseUInt64(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual BinaryInteger   parseBinaryInteger(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException);
            virtual DecimalInteger  parseDecimalInteger(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException);
            virtual Real32          parseReal32(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Real64          parseReal64(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual BinaryFixed     parseBinaryFixed(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual BinaryFloat     parseBinaryFloat(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual DecimalFixed    parseDecimalFixed(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual DecimalFloat    parseDecimalFloat(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual Int8            parseInt8(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Int16           parseInt16(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Int32           parseInt32(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Int64           parseInt64(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt8           parseUInt8(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt16          parseUInt16(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt32          parseUInt32(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual UInt64          parseUInt64(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual BinaryInteger   parseBinaryInteger(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException);
            virtual DecimalInteger  parseDecimalInteger(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException);
            virtual BinaryFixed     parseBinaryFixed(const String & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual BinaryFloat     parseBinaryFloat(const String & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual DecimalFixed    parseDecimalFixed(const String & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual DecimalFloat    parseDecimalFloat(const String & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            virtual Real32          parseReal32(const String & string, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);
            virtual Real64          parseReal64(const String & string, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

            //////////
            //  Implementation
        private:
            Char                    _decimalPointChar;
            Char                    _separatorChar;

            //  Helpers
            Int64                   _parseInt64(const _IParsedString & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow) const throw(ParseException);
            UInt64                  _parseUInt64(const _IParsedString & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow) const throw(ParseException);
            BinaryInteger           _parseBinaryInteger(const _IParsedString & string, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException);
            DecimalInteger          _parseDecimalInteger(const _IParsedString & string, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException);
            Real64                  _parseReal64(const _IParsedString & string, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow) const throw(ParseException);
            BinaryFixed             _parseBinaryFixed(const _IParsedString & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            BinaryFloat             _parseBinaryFloat(const _IParsedString & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            DecimalFixed            _parseDecimalFixed(const _IParsedString & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
            DecimalFloat            _parseDecimalFloat(const _IParsedString & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException);
        };
    }
}

#endif  //  __NumberParseImpl_hpp_included__
//  End of NumberParseImpl.hpp
