//
//  ParameterizedDstAwareTimeZone.cpp - Implementation of the carl::_impl::_ParameterizedDstAwareTimeZone class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:53 $
//  $Revision: 1.5 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "DateTimeImpl.hpp"

//////////
//  Construction/destruction
_impl::_ParameterizedDstAwareTimeZone::_ParameterizedDstAwareTimeZone(Int64 bias)
{
    _bias = bias;
    _usesDst = false;
}

_impl::_ParameterizedDstAwareTimeZone::_ParameterizedDstAwareTimeZone(Int64 bias, 
                                                                      Int64 standardBias, const _DaylightSwitchpoint & standardSwitchpoint, 
                                                                      Int64 daylightBias, const _DaylightSwitchpoint & daylightSwitchpoint)
{
    _bias = bias;
    _usesDst = true;
    _standardBias = standardBias;
    _standardSwitchpoint = standardSwitchpoint;
    _daylightBias = daylightBias;
    _daylightSwitchpoint = daylightSwitchpoint;
}

_impl::_ParameterizedDstAwareTimeZone::~_ParameterizedDstAwareTimeZone()
{
}

//////////
//  Operations
DateTime _impl::_ParameterizedDstAwareTimeZone::timeToLocal(const TimeStamp & timeStamp) const
{
    //  Apply the bias; local time = UTC - bias
    TimeStamp localTimeStamp = _applyBias(timeStamp, -_bias);
    //  Calculate the local date+time
    DateTime localDateTime = _convertTimeStampToDateTime(localTimeStamp);
    if (!localDateTime.isValid())
    {
        return DateTime();
    }
    //  If DST is used, apply that as well
    if (_usesDst)
    {
        //  Calculate extra bias
        Int64 extraBias = _isDst(localDateTime) ? _daylightBias : _standardBias;
        //  Apply extra bias
        if (extraBias != 0)
        {
            TimeStamp localBiasedTimeStamp = _applyBias(localTimeStamp, -extraBias);
            localDateTime = _convertTimeStampToDateTime(localBiasedTimeStamp);
        }
    }
    //  Done
    return localDateTime;
}

TimeStamp _impl::_ParameterizedDstAwareTimeZone::timeFromLocal(const DateTime & dateTime) const
{
    //  Calculate the local timestamp
    TimeStamp localTimeStamp = _convertDateTimeToTimeStamp(dateTime);
    if (!localTimeStamp.isValid())
    {
        return TimeStamp();
    }
    //  Apply the bias; UTC = local time + bias
    TimeStamp utcTimeStamp = _applyBias(localTimeStamp, _bias);
    //  If DST is used, apply that as well
    if (_usesDst)
    {
        //  Calculate extra bias
        Int64 extraBias = _isDst(dateTime) ? _daylightBias : _standardBias;
        //  Apply extra bias
        if (extraBias != 0)
        {
            utcTimeStamp += TimeSpan(extraBias);
        }
    }
    //  Done
    return utcTimeStamp;
}

Int64 _impl::_ParameterizedDstAwareTimeZone::getBias() const
{
    return _bias;
}

bool _impl::_ParameterizedDstAwareTimeZone::usesDst() const
{
    return _usesDst;
}

Int64 _impl::_ParameterizedDstAwareTimeZone::getStandardBias() const
{
    return _standardBias;
}

_impl::_DaylightSwitchpoint _impl::_ParameterizedDstAwareTimeZone::getStandardSwitchpoint() const
{
    return _standardSwitchpoint;
}

Int64 _impl::_ParameterizedDstAwareTimeZone::getDaylightBias() const
{
    return _daylightBias;
}

_impl::_DaylightSwitchpoint _impl::_ParameterizedDstAwareTimeZone::getDaylightSwitchpoint() const
{
    return _daylightSwitchpoint;
}

//////////
//  Implementation helpers
DateTime _impl::_ParameterizedDstAwareTimeZone::_convertTimeStampToDateTime(const TimeStamp & timeStamp)
{
    //  Invalid values require special treatment
    if (!timeStamp.isValid())
    {
        return DateTime();
    }
    //  Prepare to convert
    int year = 1970;
    int month = 1;
    int day = 1;
    int hour = 0;
    int minute = 0;
    int second = 0;
    UInt64 nanosecondsSinceEpoch = timeStamp.getNanosecondsSinceEpoch();
    //  Process full years
    for (; ; )
    {
        int daysInYear = Calendar::daysInYear(year);
        UInt64 nanosecondsInYear = static_cast<UInt64>(daysInYear) * _NanosecondsInDay;
        if (nanosecondsInYear <= nanosecondsSinceEpoch)
        {
            year++;
            nanosecondsSinceEpoch -= nanosecondsInYear;
        }
        else
        {
            break;
        }
    }
    //  Process full months
    for (; ; )
    {
        int daysInMonth = Calendar::daysInMonth(year, month);
        UInt64 nanosecondsInMonth = static_cast<UInt64>(daysInMonth) * _NanosecondsInDay;
        if (nanosecondsInMonth <= nanosecondsSinceEpoch)
        {
            month++;
            nanosecondsSinceEpoch -= nanosecondsInMonth;
        }
        else
        {
            break;
        }
    }
    //  Process full days
    for (; ; )
    {
        if (_NanosecondsInDay <= nanosecondsSinceEpoch)
        {
            day++;
            nanosecondsSinceEpoch -= _NanosecondsInDay;
        }
        else
        {
            break;
        }
    }
    //  Process full hours
    for (; ; )
    {
        if (_NanosecondsInHour <= nanosecondsSinceEpoch)
        {
            hour++;
            nanosecondsSinceEpoch -= _NanosecondsInHour;
        }
        else
        {
            break;
        }
    }
    //  Process full minutes
    for (; ; )
    {
        if (_NanosecondsInMinute <= nanosecondsSinceEpoch)
        {
            minute++;
            nanosecondsSinceEpoch -= _NanosecondsInMinute;
        }
        else
        {
            break;
        }
    }
    //  Process full seconds
    for (; ; )
    {
        if (_NanosecondsInSecond <= nanosecondsSinceEpoch)
        {
            second++;
            nanosecondsSinceEpoch -= _NanosecondsInSecond;
        }
        else
        {
            break;
        }
    }
    //  Build & return date+time
    return DateTime(year, month, day, hour, minute, second, static_cast<UInt32>(nanosecondsSinceEpoch));
}

TimeStamp _impl::_ParameterizedDstAwareTimeZone::_convertDateTimeToTimeStamp(const DateTime & dateTime)
{
    //  Invalid values require special treatment
    if (!dateTime.isValid() || dateTime.getYear() < 1970)
    {
        return TimeStamp();
    }
    //  Prepare to convert
    UInt64 nanosecondsSinceEpoch = 0;
    int year = 1970;
    int month = 1;
    int day = 1;
    int hour = 0;
    int minute = 0;
    int second = 0;
    //  Process full years
    for (; year < dateTime.getYear(); year++)
    {
        UInt64 nanosecondsInYear = Calendar::daysInYear(year) * _NanosecondsInDay;
        UInt64 newNanosecondsSinceEpoch = nanosecondsSinceEpoch + nanosecondsInYear;
        if (newNanosecondsSinceEpoch < nanosecondsSinceEpoch)
        {
            return TimeStamp();
        }
        nanosecondsSinceEpoch = newNanosecondsSinceEpoch;
    }
    //  Process full months
    for (; month < dateTime.getMonth(); month++)
    {
        UInt64 nanosecondsInMonth = Calendar::daysInMonth(year, month) * _NanosecondsInDay;
        UInt64 newNanosecondsSinceEpoch = nanosecondsSinceEpoch + nanosecondsInMonth;
        if (newNanosecondsSinceEpoch < nanosecondsSinceEpoch)
        {
            return TimeStamp();
        }
        nanosecondsSinceEpoch = newNanosecondsSinceEpoch;
    }
    //  Process full days
    for (; day < dateTime.getDay(); day++)
    {
        UInt64 newNanosecondsSinceEpoch = nanosecondsSinceEpoch + _NanosecondsInDay;
        if (newNanosecondsSinceEpoch < nanosecondsSinceEpoch)
        {
            return TimeStamp();
        }
        nanosecondsSinceEpoch = newNanosecondsSinceEpoch;
    }
    //  Process full hours
    for (; hour < dateTime.getHour(); hour++)
    {
        UInt64 newNanosecondsSinceEpoch = nanosecondsSinceEpoch + _NanosecondsInHour;
        if (newNanosecondsSinceEpoch < nanosecondsSinceEpoch)
        {
            return TimeStamp();
        }
        nanosecondsSinceEpoch = newNanosecondsSinceEpoch;
    }
    //  Process full minutes
    for (; minute < dateTime.getMinute(); minute++)
    {
        UInt64 newNanosecondsSinceEpoch = nanosecondsSinceEpoch + _NanosecondsInMinute;
        if (newNanosecondsSinceEpoch < nanosecondsSinceEpoch)
        {
            return TimeStamp();
        }
        nanosecondsSinceEpoch = newNanosecondsSinceEpoch;
    }
    //  Process full seconds
    for (; second < dateTime.getSecond(); second++)
    {
        UInt64 newNanosecondsSinceEpoch = nanosecondsSinceEpoch + _NanosecondsInSecond;
        if (newNanosecondsSinceEpoch < nanosecondsSinceEpoch)
        {
            return TimeStamp();
        }
        nanosecondsSinceEpoch = newNanosecondsSinceEpoch;
    }
    //  Process nanoseconds
    {
        UInt64 newNanosecondsSinceEpoch = nanosecondsSinceEpoch + dateTime.getNanosecond();
        if (newNanosecondsSinceEpoch < nanosecondsSinceEpoch)
        {
            return TimeStamp();
        }
        nanosecondsSinceEpoch = newNanosecondsSinceEpoch;
    }
    //  Done
    return TimeStamp(nanosecondsSinceEpoch);
}

TimeStamp _impl::_ParameterizedDstAwareTimeZone::_applyBias(const TimeStamp & timeStamp, Int64 bias)
{
    return timeStamp + TimeSpan(bias);
}

_impl::_AbsoluteDaylightSwitchpoint _impl::_ParameterizedDstAwareTimeZone::_calculateAbsoluteDaylightSwitchpoint(const _DaylightSwitchpoint & daylightSwitchpoint, int year)
{
    if (daylightSwitchpoint.type == _DaylightSwitchpointType_Absolute)
    {
        return daylightSwitchpoint.absoluteDaylightSwitchpoint;
    }
    else
    {
        _AbsoluteDaylightSwitchpoint absoluteDaylightSwitchpoint;
        absoluteDaylightSwitchpoint.month = daylightSwitchpoint.dayOfWeekDaylightSwitchpoint.month;
        absoluteDaylightSwitchpoint.hour = daylightSwitchpoint.dayOfWeekDaylightSwitchpoint.hour;
        absoluteDaylightSwitchpoint.minute = daylightSwitchpoint.dayOfWeekDaylightSwitchpoint.minute;
        absoluteDaylightSwitchpoint.second = daylightSwitchpoint.dayOfWeekDaylightSwitchpoint.second;
        absoluteDaylightSwitchpoint.nanosecond = daylightSwitchpoint.dayOfWeekDaylightSwitchpoint.nanosecond;
        int daysInMonth = Calendar::daysInMonth(year, absoluteDaylightSwitchpoint.month);
        //  Determine what day of week was the 1st of the required month
        int firstOfMonthDayOfWeek = Calendar::getDayOfWeek(year, daylightSwitchpoint.dayOfWeekDaylightSwitchpoint.month, 1);
        //  Deternine the day of the required day of week 1st occurrence
        int firstRequiredDayOfWeekOccurrenceDay = (daylightSwitchpoint.dayOfWeekDaylightSwitchpoint.dayOfWeek + 7 - firstOfMonthDayOfWeek) % 7 + 1;
        //  Calculate the day of the required occurrence
        switch (daylightSwitchpoint.dayOfWeekDaylightSwitchpoint.occurence)
        {
            case 1: case 2: case 3: case 4:
                absoluteDaylightSwitchpoint.day = static_cast<UInt8>(firstRequiredDayOfWeekOccurrenceDay + 7 * (daylightSwitchpoint.dayOfWeekDaylightSwitchpoint.occurence - 1));
                break;
            case 5:
                absoluteDaylightSwitchpoint.day = static_cast<UInt8>(firstRequiredDayOfWeekOccurrenceDay + 7 * 5);
                while (absoluteDaylightSwitchpoint.day > daysInMonth)
                {
                    absoluteDaylightSwitchpoint.day -= 7;
                }
                break;
            default:
                require(false);
        }
        return absoluteDaylightSwitchpoint;
    }
}

int _impl::_ParameterizedDstAwareTimeZone::_compare(const _AbsoluteDaylightSwitchpoint & switchpoint1, const _AbsoluteDaylightSwitchpoint & switchpoint2)
{
    if (switchpoint1.month < switchpoint2.month)
    {
        return -1;
    }
    if (switchpoint1.month > switchpoint2.month)
    {
        return 1;
    }
    if (switchpoint1.day < switchpoint2.day)
    {
        return -1;
    }
    if (switchpoint1.day > switchpoint2.day)
    {
        return 1;
    }
    if (switchpoint1.hour < switchpoint2.hour)
    {
        return -1;
    }
    if (switchpoint1.hour > switchpoint2.hour)
    {
        return 1;
    }
    if (switchpoint1.minute < switchpoint2.minute)
    {
        return -1;
    }
    if (switchpoint1.minute > switchpoint2.minute)
    {
        return 1;
    }
    if (switchpoint1.second < switchpoint2.second)
    {
        return -1;
    }
    if (switchpoint1.second > switchpoint2.second)
    {
        return 1;
    }
    if (switchpoint1.nanosecond < switchpoint2.nanosecond)
    {
        return -1;
    }
    if (switchpoint1.nanosecond > switchpoint2.nanosecond)
    {
        return 1;
    }
    return 0;
}

int _impl::_ParameterizedDstAwareTimeZone::_compare(const _AbsoluteDaylightSwitchpoint & switchpoint, const DateTime & dateTime)
{
    if (switchpoint.month < dateTime.getMonth())
    {
        return -1;
    }
    if (switchpoint.month > dateTime.getMonth())
    {
        return 1;
    }
    if (switchpoint.day < dateTime.getDay())
    {
        return -1;
    }
    if (switchpoint.day > dateTime.getDay())
    {
        return 1;
    }
    if (switchpoint.hour < dateTime.getHour())
    {
        return -1;
    }
    if (switchpoint.hour > dateTime.getHour())
    {
        return 1;
    }
    if (switchpoint.minute < dateTime.getMinute())
    {
        return -1;
    }
    if (switchpoint.minute > dateTime.getMinute())
    {
        return 1;
    }
    if (switchpoint.second < dateTime.getSecond())
    {
        return -1;
    }
    if (switchpoint.second > dateTime.getSecond())
    {
        return 1;
    }
    if (switchpoint.nanosecond < dateTime.getNanosecond())
    {
        return -1;
    }
    if (switchpoint.nanosecond > dateTime.getNanosecond())
    {
        return 1;
    }
    return 0;
}

bool _impl::_ParameterizedDstAwareTimeZone::_isDst(const DateTime & dateTime) const
{
    if (!_usesDst || !dateTime.isValid())
    {
        return false;
    }
    //  Calculate absolute switch points for the required
    _impl::_AbsoluteDaylightSwitchpoint standardSwitchpoint = _calculateAbsoluteDaylightSwitchpoint(_standardSwitchpoint, dateTime.getYear());
    _impl::_AbsoluteDaylightSwitchpoint daylightSwitchpoint = _calculateAbsoluteDaylightSwitchpoint(_daylightSwitchpoint, dateTime.getYear());
    //  Calculate extra bias
    if (_compare(daylightSwitchpoint, standardSwitchpoint) < 0)
    {   //  Year is standard ... daylight .. standard
        return _compare(daylightSwitchpoint, dateTime) <= 0 && _compare(standardSwitchpoint, dateTime) >= 0;
    }
    else
    {   //  Year is daylight .. standard ... daylight
        return _compare(standardSwitchpoint, dateTime) <= 0 && _compare(daylightSwitchpoint, dateTime) >= 0;
    }
}

//  End of ParameterizedDstAwareTimeZone.cpp
