//
//  PathName.cpp - Implementation of the carl::PathName class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/07/28 10:40:28 $
//  $Revision: 1.22 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <windows.h>
    #include <io.h>
    #include <sys/stat.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <errno.h>
    #include <limits.h>
    #include <stdlib.h>
    #include <unistd.h>
    #include <sys/types.h>
    #include <sys/stat.h>
#else
    #error Unsupported platform
#endif

//////////
//  Construction/destruction/assignment
PathName::PathName(const char * pPathNameString)
    :   _pathNameString(pPathNameString)
{
    _adjust();
}

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
PathName::PathName(const wchar_t * pPathNameString)
    :   _pathNameString(pPathNameString)
{
    _adjust();
}
#endif

PathName::PathName(const Char * pPathNameString)
    :   _pathNameString(pPathNameString)
{
    _adjust();
}

PathName::PathName(const String & pathNameString)
    :   _pathNameString(pathNameString)
{
    _adjust();
}

PathName::PathName(const char * pDirectoryString, const char * pFilenameString)
    :   _pathNameString(_combineDirectoryAndFilename(String(pDirectoryString), String(pFilenameString)))
{
    _adjust();
}

PathName::PathName(const char * pDirectoryString, const Char * pFilenameString)
    :   _pathNameString(_combineDirectoryAndFilename(String(pDirectoryString), String(pFilenameString)))
{
    _adjust();
}

PathName::PathName(const char * pDirectoryString, const String & filenameString)
    :   _pathNameString(_combineDirectoryAndFilename(String(pDirectoryString), filenameString))
{
    _adjust();
}

PathName::PathName(const Char * pDirectoryString, const char * pFilenameString)
    :   _pathNameString(_combineDirectoryAndFilename(String(pDirectoryString), String(pFilenameString)))
{
    _adjust();
}

PathName::PathName(const Char * pDirectoryString, const Char * pFilenameString)
    :   _pathNameString(_combineDirectoryAndFilename(String(pDirectoryString), String(pFilenameString)))
{
    _adjust();
}

PathName::PathName(const Char * pDirectoryString, const String & filenameString)
    :   _pathNameString(_combineDirectoryAndFilename(String(pDirectoryString), filenameString))
{
    _adjust();
}

PathName::PathName(const String & directoryString, const char * pFilenameString)
    :   _pathNameString(_combineDirectoryAndFilename(directoryString, String(pFilenameString)))
{
    _adjust();
}

PathName::PathName(const String & directoryString, const Char * pFilenameString)
    :   _pathNameString(_combineDirectoryAndFilename(directoryString, String(pFilenameString)))
{
    _adjust();
}

PathName::PathName(const String & directoryString, const String & filenameString)
    :   _pathNameString(_combineDirectoryAndFilename(directoryString, filenameString))
{
    _adjust();
}

PathName::PathName(const PathName & directory, const char * pFilenameString)
    :   _pathNameString(_combineDirectoryAndFilename(directory._pathNameString, String(pFilenameString)))
{
    _adjust();
}

PathName::PathName(const PathName & directory, const Char * pFilenameString)
    :   _pathNameString(_combineDirectoryAndFilename(directory._pathNameString, String(pFilenameString)))
{
    _adjust();
}

PathName::PathName(const PathName & directory, const String & filenameString)
    :   _pathNameString(_combineDirectoryAndFilename(directory._pathNameString, filenameString))
{
    _adjust();
}

PathName::PathName(const PathName & pathName)
    :   _pathNameString(pathName._pathNameString)
{
}

PathName & PathName::operator = (const PathName & src)
{
    if (this != (&src))
    {
        _pathNameString = src._pathNameString;
    }
    return *this;
}

PathName & PathName::operator = (const char * pPathNameString)
{
    _pathNameString = String(pPathNameString);
    _adjust();
    return *this;
}

PathName & PathName::operator = (const Char * pPathNameString)
{
    _pathNameString = String(pPathNameString);
    _adjust();
    return *this;
}

PathName & PathName::operator = (const String & pathNameString)
{
    _pathNameString = pathNameString;
    _adjust();
    return *this;
}

//////////
//  Operations
bool PathName::isAbsolute() const
{
    //  A path name is absolute if any of the following is true:
    //  1.  It starts with a path component separator
    //  2.  It has a device separator followed by a path component separator anywhere in it
    //  The above rules cover:
    //  1.  Unix file names.
    //  2.  Win32 local file names.
    //  3.  Win32 UNC file names.

    if (_pathNameString.length() == 0)
    {
        return false;
    }
    else if (FileSystem::isPathComponentSeparator(_pathNameString[0]))
    {
        return true;
    }
    for (UIntSize i = 0; i + 1 < _pathNameString.length(); i++)
    {
        if (FileSystem::isDeviceSeparator(_pathNameString[i]) &&
            FileSystem::isPathComponentSeparator(_pathNameString[i + 1]))
        {
            return true;
        }
    }
    return false;
}

bool PathName::hasDirectory() const
{
    return _findRightmostPathComponentSeparatorIndex(_pathNameString) != -1;
}

bool PathName::hasExtension() const
{
    //  Find where the filename starts
    int filenameStartIndex = _findRightmostPathComponentSeparatorIndex(_pathNameString) + 1;

    //  Find where the extension starts
    int rightmostExtensionSeparatorIndex = _findRightmostExtensionSeparatorIndex(_pathNameString,
                                                                                 filenameStartIndex);

    //  If a filename starts with an extension separator, the whole thing is NOT an extension.
    //  This can happen for Unix hidden files.
    return rightmostExtensionSeparatorIndex > filenameStartIndex;
}

bool PathName::hasExtension(const char * pExtension) const
{
    require(pExtension != NULL);

    return hasExtension(String(pExtension));
}

bool PathName::hasExtension(const Char * pExtension) const
{
    require(pExtension != NULL);

    return hasExtension(String(pExtension));
}

bool PathName::hasExtension(const String & extension) const
{
    String ext = extension;
    if (ext.length() != 0 && !FileSystem::isFilenameExtensionSeparator(ext[0]))
    {
        ext = FileSystem::FilenameExtensionSeparator + ext;
    }
    if (FileSystem::CaseSensitiveFileNames)
    {
        return getExtension().compare(ext) == 0;
    }
    else
    {
        return getExtension().compareIgnoreCase(ext) == 0;
    }
}

String PathName::getDirectory() const
{
    //  Find the rightmost path component separator index
    int rightmostPathComponentSeparatorIndex = _findRightmostPathComponentSeparatorIndex(_pathNameString);
    if (rightmostPathComponentSeparatorIndex == -1)
    {   //  This path name does not contain directory, only a filename
        return String();
    }

    //  Do we need to store the tailing path component separator ?
    bool tailingPathComponentSeparatorNeeded =
        (rightmostPathComponentSeparatorIndex == 0) ||
        (FileSystem::isPathComponentSeparator(_pathNameString[rightmostPathComponentSeparatorIndex - 1])) ||
        (FileSystem::UsesNamedDevices &&
         FileSystem::isDeviceSeparator(_pathNameString[rightmostPathComponentSeparatorIndex - 1]));
    UIntSize charactersToStore = (tailingPathComponentSeparatorNeeded) ?
                                 (static_cast<UIntSize>(rightmostPathComponentSeparatorIndex) + 1) :
                                 (static_cast<UIntSize>(rightmostPathComponentSeparatorIndex));

    //  Return the directory
    return _pathNameString.left(charactersToStore);
}

String PathName::getFilename() const
{
    //  Find where the filename starts
    int filenameStartIndex = _findRightmostPathComponentSeparatorIndex(_pathNameString) + 1;

    //  Return filename
    return _pathNameString.mid(filenameStartIndex);
}

String PathName::getName() const
{
    //  Find where the filename starts
    int filenameStartIndex = _findRightmostPathComponentSeparatorIndex(_pathNameString) + 1;

    //  Find where the extension starts.
    int rightmostExtensionSeparatorIndex = _findRightmostExtensionSeparatorIndex(_pathNameString,
                                                                                 filenameStartIndex);

    //  If a filename starts with an extension separator, the whole thing is NOT an extension.
    //  This can happen for Unix hidden files.
    if (rightmostExtensionSeparatorIndex == filenameStartIndex)
    {
        rightmostExtensionSeparatorIndex = -1;
    }

    //  Determine how many characters to store
    UIntSize charactersToStore = (rightmostExtensionSeparatorIndex == -1) ?
                                 (_pathNameString.length() - filenameStartIndex):
                                 (rightmostExtensionSeparatorIndex - filenameStartIndex);

    //  Return name
    return _pathNameString.mid(filenameStartIndex, charactersToStore);
}

String PathName::getExtension() const
{
    //  Find where the filename starts
    int filenameStartIndex = _findRightmostPathComponentSeparatorIndex(_pathNameString) + 1;

    //  Find where the extension starts
    int rightmostExtensionSeparatorIndex = _findRightmostExtensionSeparatorIndex(_pathNameString,
                                                                                 filenameStartIndex);

    //  If a filename starts with an extension separator, the whole thing is NOT an extension.
    //  This can happen for Unix hidden files.
    if (rightmostExtensionSeparatorIndex == filenameStartIndex)
    {
        rightmostExtensionSeparatorIndex = -1;
    }

    //  If there is no extension, a special treatment is needed
    if (rightmostExtensionSeparatorIndex == -1)
    {
        return String();
    }

    //  Return extension
    return _pathNameString.mid(rightmostExtensionSeparatorIndex);
}

PathName PathName::toCanonical() const
{
    PathName result(*this);
    result.makeCanonical();
    return result;
}

PathName PathName::toAbsolute() const
{
    PathName result(*this);
    result.makeAbsolute();
    return result;
}

PathName PathName::withExtension(const char * pExtension)
{
    require(pExtension != NULL);

    return withExtension(String(pExtension));
}

PathName PathName::withExtension(const Char * pExtension)
{
    require(pExtension != NULL);

    return withExtension(String(pExtension));
}

PathName PathName::withExtension(const String & extension)
{
    PathName result(*this);
    result.ensureExtension(extension);
    return result;
}

PathName PathName::getParent() const
{
    return PathName(getDirectory());
}

void PathName::makeCanonical()
{
    //  Replace all separators with their canonical counterparts
    for (UIntSize i = 0; i < _pathNameString.length(); i++)
    {
        Char c = _pathNameString[i];
        if (FileSystem::isPathComponentSeparator(c) && c != FileSystem::PathComponentSeparator)
        {
            _pathNameString[i] = FileSystem::PathComponentSeparator;
        }
        if (FileSystem::isFilenameExtensionSeparator(c) && c != FileSystem::FilenameExtensionSeparator)
        {
            _pathNameString[i] = FileSystem::FilenameExtensionSeparator;
        }
        if (FileSystem::UsesNamedDevices &&
            FileSystem::isDeviceSeparator(c) && c != FileSystem::DeviceSeparator)
        {
            _pathNameString[i] = FileSystem::DeviceSeparator;
        }
    }
}

void PathName::makeAbsolute()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> relativePathName;
    Array<wchar_t> absolutePathName;

    //  Empty path names require special treatment
    if (_pathNameString.length() == 0)
    {
        return;
    }

    //  Store relative path name
    relativePathName.setSize(_pathNameString.length() + 1);
    if (!_store(relativePathName.getData(), relativePathName.size()))
    {
        return;
    }

    //  Convert to absolute path name
    DWORD lastError = GetLastError();
    wchar_t * fileNamePortion;
    absolutePathName.setSize(relativePathName.size() + FileSystem::MaximumPathLength);
    DWORD dw = GetFullPathNameW(relativePathName.getData(), absolutePathName.size(), absolutePathName.getData(), &fileNamePortion);
    if (dw > absolutePathName.size() || dw == 0)
    {   //  Full path name will be too long OR something went wrong
        SetLastError(lastError);
        return;
    }

    //  Store absolute path name
    _pathNameString = String(absolutePathName.getData());
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> relativePathName;
    Array<char> absolutePathName;

    //  Empty path names require special treatment
    if (_pathNameString.length() == 0)
    {
        return;
    }

    //  Store relative path name
    relativePathName.setSize(_pathNameString.length() + 1);
    if (!_store(relativePathName.getData(), relativePathName.size()))
    {
        return;
    }

    //  Convert to absolute path name
    int oldErrno = errno;
    absolutePathName.setSize(relativePathName.size() + FileSystem::MaximumPathLength);
    if (realpath(relativePathName.getData(), absolutePathName.getData()) != NULL)
    {   //  Absolute path obtained successfully
        _pathNameString = String(absolutePathName.getData());
    }
    errno = oldErrno;
#else
    #error Unsupported platform
#endif
}

bool PathName::ensureExtension(const char * pExtension)
{
    require(pExtension != NULL);

    return ensureExtension(String(pExtension));
}

bool PathName::ensureExtension(const Char * pExtension)
{
    require(pExtension != NULL);

    return ensureExtension(String(pExtension));
}

bool PathName::ensureExtension(const String & extension)
{
    String ext = extension;
    if (ext.length() != 0 && !FileSystem::isFilenameExtensionSeparator(ext[0]))
    {
        ext = FileSystem::FilenameExtensionSeparator + ext;
    }
    if (!isEmpty())
    {
        //  Find where the filename starts
        int filenameStartIndex = _findRightmostPathComponentSeparatorIndex(_pathNameString) + 1;

        //  Find where the extension starts
        int rightmostExtensionSeparatorIndex = _findRightmostExtensionSeparatorIndex(_pathNameString, filenameStartIndex);

        //  Enforce the extension
        if (rightmostExtensionSeparatorIndex > filenameStartIndex)
        {   //  Extension already exists, replace
            _pathNameString = _pathNameString.left(rightmostExtensionSeparatorIndex) + ext;
        }
        else
        {   //  No extension yet; add
            _pathNameString += ext;
        }
        return true;
    }
    else
    {
        return false;
    }
}

bool PathName::exists() const
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check access rights
    int oldErrno = errno;
    if (_waccess(pathNameString.getData(), 0) != 0)
    {
        errno = oldErrno;
        return false;
    }
    return true;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check access rights
    int oldErrno = errno;
    if (access(pathNameString.getData(), F_OK) != 0)
    {
        errno = oldErrno;
        return false;
    }
    return true;
#else
    #error Unsupported platform
#endif
}

bool PathName::isFile() const
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check file attributes
    DWORD lastError = GetLastError();
    DWORD dw = GetFileAttributesW(pathNameString.getData());
    SetLastError(lastError);
    if (dw == INVALID_FILE_ATTRIBUTES)
    {
        return false;
    }
    else
    {
        return (dw & FILE_ATTRIBUTE_DIRECTORY) == 0;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check file attributes
    struct stat s;
    int lastErrno = errno;
    if (stat(pathNameString.getData(), &s) == -1)
    {
        errno = lastErrno;
        return false;
    }
    return S_ISREG(s.st_mode);
#else
    #error Unsupported platform
#endif
}

bool PathName::isDirectory() const
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check file attributes
    DWORD lastError = GetLastError();
    DWORD dw = GetFileAttributesW(pathNameString.getData());
    SetLastError(lastError);
    if (dw == INVALID_FILE_ATTRIBUTES)
    {
        return false;
    }
    else
    {
        return (dw & FILE_ATTRIBUTE_DIRECTORY) != 0;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check file attributes
    struct stat s;
    int lastErrno = errno;
    if (stat(pathNameString.getData(), &s) == -1)
    {
        errno = lastErrno;
        return false;
    }
    return S_ISDIR(s.st_mode);
#else
    #error Unsupported platform
#endif
}

bool PathName::canRead() const
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check access rights
    int oldErrno = errno;
    if (_waccess(pathNameString.getData(), 04) != 0)
    {
        errno = oldErrno;
        return false;
    }
    return true;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check access rights
    int oldErrno = errno;
    if (access(pathNameString.getData(), R_OK | F_OK) != 0)
    {
        errno = oldErrno;
        return false;
    }
    return true;
#else
    #error Unsupported platform
#endif
}

bool PathName::setReadable(bool yes /*= true*/)
{
    int permissions = 0;
    if (yes)
    {
        return getPermissions(permissions) &&
               setPermissions(permissions | (FileSystemPermissions_OwnerRead | FileSystemPermissions_GroupRead | FileSystemPermissions_OtherRead));
    }
    else
    {
        return getPermissions(permissions) &&
               setPermissions(permissions & ~(FileSystemPermissions_OwnerRead | FileSystemPermissions_GroupRead | FileSystemPermissions_OtherRead));
    }
}

bool PathName::canWrite() const
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check access rights
    int oldErrno = errno;
    if (_waccess(pathNameString.getData(), 02) != 0)
    {
        errno = oldErrno;
        return false;
    }
    return true;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check access rights
    int oldErrno = errno;
    if (access(pathNameString.getData(), W_OK | F_OK) != 0)
    {
        errno = oldErrno;
        return false;
    }
    return true;
#else
    #error Unsupported platform
#endif
}

bool PathName::setWriteable(bool yes /*= true*/)
{
    int permissions = 0;
    if (yes)
    {
        return getPermissions(permissions) &&
               setPermissions(permissions | (FileSystemPermissions_OwnerWrite | FileSystemPermissions_GroupWrite | FileSystemPermissions_OtherWrite));
    }
    else
    {
        return getPermissions(permissions) &&
               setPermissions(permissions & ~(FileSystemPermissions_OwnerWrite | FileSystemPermissions_GroupWrite | FileSystemPermissions_OtherWrite));
    }
}

bool PathName::canExecute() const
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    //  On Win32, "read permission" == "execute permission"
    return canRead();
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check access rights
    int oldErrno = errno;
    if (access(pathNameString.getData(), X_OK | F_OK) != 0)
    {
        errno = oldErrno;
        return false;
    }
    return true;
#else
    #error Unsupported platform
#endif
}

bool PathName::setExecutable(bool yes /*= true*/)
{
    int permissions = 0;
    if (yes)
    {
        return getPermissions(permissions) &&
               setPermissions(permissions | (FileSystemPermissions_OwnerExecute | FileSystemPermissions_GroupExecute | FileSystemPermissions_OtherExecute));
    }
    else
    {
        return getPermissions(permissions) &&
               setPermissions(permissions & ~(FileSystemPermissions_OwnerExecute | FileSystemPermissions_GroupExecute | FileSystemPermissions_OtherExecute));
    }
}

bool PathName::isHidden() const
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check file attributes
    DWORD lastError = GetLastError();
    DWORD dw = GetFileAttributesW(pathNameString.getData());
    SetLastError(lastError);
    if (dw == INVALID_FILE_ATTRIBUTES)
    {
        return false;
    }
    else
    {
        return (dw & FILE_ATTRIBUTE_HIDDEN) != 0;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Check access rights
    int oldErrno = errno;
    if (access(pathNameString.getData(), F_OK) != 0)
    {
        errno = oldErrno;
        return false;
    }

    //  In Unix environments a file is considered hidden iff its name starts
    //  with a dot '.'. We have already checked that the file exists.
    int lastPathComponentSeparatorIndex = -1;
    for (int i = 0; pathNameString[i] != 0; i++)
    {
        if (FileSystem::isPathComponentSeparator(pathNameString[i]))
        {
            lastPathComponentSeparatorIndex = i;
        }
    }

    return pathNameString[lastPathComponentSeparatorIndex + 1] == '.';
#else
    #error Unsupported platform
#endif
}

bool PathName::operator == (const PathName & pathName) const
{
    PathName op1 = toAbsolute();
    PathName op2 = pathName.toAbsolute();
    if (FileSystem::CaseSensitiveFileNames)
    {
        return op1._pathNameString.compare(op2._pathNameString) == 0;
    }
    else
    {
        return op1._pathNameString.compareIgnoreCase(op2._pathNameString) == 0;
    }
}

bool PathName::operator != (const PathName & pathName) const
{
    return !((*this) == pathName);
}

bool PathName::getPermissions(int & permissions)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Query permissions
    int oldErrno = errno;
    struct _stat s;
    if (_wstat(pathNameString.getData(), &s) != 0)
    {
        errno = oldErrno;
        return false;
    }

    //  Translate permissions
    permissions = 0;
    if ((s.st_mode & _S_IREAD) != 0)
    {
        permissions |= FileSystemPermissions_OwnerRead |
                       FileSystemPermissions_GroupRead |
                       FileSystemPermissions_OtherRead |
                       FileSystemPermissions_OwnerExecute |
                       FileSystemPermissions_GroupExecute |
                       FileSystemPermissions_OtherExecute;
    }
    if ((s.st_mode & _S_IWRITE) != 0)
    {
        permissions |= FileSystemPermissions_OwnerWrite |
                       FileSystemPermissions_GroupWrite |
                       FileSystemPermissions_OtherWrite;
    }

    //  Done
    return true;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Query permissions
    int oldErrno = errno;
    struct stat s;
    if (stat(pathNameString.getData(), &s) != 0)
    {
        errno = oldErrno;
        return false;
    }

    //  Translate permissions
    permissions = 0;
    if ((s.st_mode & S_IRUSR) != 0)
    {
        permissions |= FileSystemPermissions_OwnerRead;
    }
    if ((s.st_mode & S_IWUSR) != 0)
    {
        permissions |= FileSystemPermissions_OwnerWrite;
    }
    if ((s.st_mode & S_IXUSR) != 0)
    {
        permissions |= FileSystemPermissions_OwnerExecute;
    }
    if ((s.st_mode & S_IRGRP) != 0)
    {
        permissions |= FileSystemPermissions_GroupRead;
    }
    if ((s.st_mode & S_IWGRP) != 0)
    {
        permissions |= FileSystemPermissions_GroupWrite;
    }
    if ((s.st_mode & S_IXGRP) != 0)
    {
        permissions |= FileSystemPermissions_GroupExecute;
    }
    if ((s.st_mode & S_IROTH) != 0)
    {
        permissions |= FileSystemPermissions_OtherRead;
    }
    if ((s.st_mode & S_IWOTH) != 0)
    {
        permissions |= FileSystemPermissions_OtherWrite;
    }
    if ((s.st_mode & S_IXOTH) != 0)
    {
        permissions |= FileSystemPermissions_OtherExecute;
    }

    //  Done
    return true;
#else
    #error Unsupported platform
#endif
}

bool PathName::setPermissions(int permissions)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Translate permissions
    int crtPermissions = 0;
    if ((permissions & (FileSystemPermissions_OwnerRead |
                        FileSystemPermissions_GroupRead |
                        FileSystemPermissions_OtherRead |
                        FileSystemPermissions_OwnerExecute |
                        FileSystemPermissions_GroupExecute |
                        FileSystemPermissions_OtherExecute)) != 0)
    {
        crtPermissions |= _S_IREAD;
    }
    if ((permissions & (FileSystemPermissions_OwnerWrite |
                        FileSystemPermissions_GroupWrite |
                        FileSystemPermissions_OtherWrite)) != 0)
    {
        crtPermissions |= _S_IWRITE;
    }

    //  Set permissions
    int oldErrno = errno;
    if (_wchmod(pathNameString.getData(),crtPermissions) != 0)
    {
        errno = oldErrno;
        return false;
    }

    //  Done
    return true;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> pathNameString;

    //  Store path name string
    pathNameString.setSize(_pathNameString.length() + 1);
    if (!_store(pathNameString.getData(), pathNameString.size()))
    {
        return false;
    }

    //  Query old permissions
    int oldErrno = errno;
    struct stat s;
    if (stat(pathNameString.getData(), &s) != 0)
    {
        errno = oldErrno;
        return false;
    }

    //  Translate permissions
    int crtPermissions = s.st_mode & (~(S_IRWXU | S_IRWXG | S_IRWXO));
    if ((permissions & (FileSystemPermissions_OwnerRead)) != 0)
    {
        crtPermissions |= S_IRUSR;
    }
    if ((permissions & (FileSystemPermissions_OwnerWrite)) != 0)
    {
        crtPermissions |= S_IWUSR;
    }
    if ((permissions & (FileSystemPermissions_OwnerExecute)) != 0)
    {
        crtPermissions |= S_IXUSR;
    }
    if ((permissions & (FileSystemPermissions_GroupRead)) != 0)
    {
        crtPermissions |= S_IRGRP;
    }
    if ((permissions & (FileSystemPermissions_GroupWrite)) != 0)
    {
        crtPermissions |= S_IWGRP;
    }
    if ((permissions & (FileSystemPermissions_GroupExecute)) != 0)
    {
        crtPermissions |= S_IXGRP;
    }
    if ((permissions & (FileSystemPermissions_OtherRead)) != 0)
    {
        crtPermissions |= S_IROTH;
    }
    if ((permissions & (FileSystemPermissions_OtherWrite)) != 0)
    {
        crtPermissions |= S_IWOTH;
    }
    if ((permissions & (FileSystemPermissions_OtherExecute)) != 0)
    {
        crtPermissions |= S_IXOTH;
    }

    //  Set permissions
    if (chmod(pathNameString.getData(),crtPermissions) != 0)
    {
        errno = oldErrno;
        return false;
    }

    //  Done
    return true;
#else
    #error Unsupported platform
#endif
}

//////////
//  Implementation helpers
String PathName::_combineDirectoryAndFilename(const String & directory, const String & filename)
{
    if (directory.length() == 0)
    {
        return filename;
    }
    bool directoryEndsWithPathComponentSeparator =
        directory.length() != 0 && FileSystem::isPathComponentSeparator(directory[directory.length() - 1]);
    bool filenameStartsWithPathComponentSeparator =
        filename.length() != 0 && FileSystem::isPathComponentSeparator(filename[0]);
    if (directoryEndsWithPathComponentSeparator)
    {
        if (filenameStartsWithPathComponentSeparator)
        {
            return directory + filename.mid(1);
        }
        else
        {
            return directory + filename;
        }
    }
    else
    {
        if (filenameStartsWithPathComponentSeparator)
        {
            return directory + filename;
        }
        else
        {
            return directory + FileSystem::PathComponentSeparator + filename;
        }
    }
}

void PathName::_adjust()
{
    if (_pathNameString.length() == 0)
    {
        return;
    }
    //
    if (FileSystem::isPathComponentSeparator(_pathNameString[_pathNameString.length() - 1]))
    {   //  Path name string ends with a path component separator - this path component
        //  separator must be removed unless it's preceeded by a device separator or
        //  if the only character in path name string
        if (_pathNameString.length() == 1 ||
            (FileSystem::UsesNamedDevices &&
             FileSystem::isDeviceSeparator(_pathNameString[_pathNameString.length() - 2])))
        {   //  Keep tailing separator
            return;
        }
        _pathNameString.setLength(_pathNameString.length() - 1);
    }
    else if (FileSystem::UsesNamedDevices &&
             FileSystem::isDeviceSeparator(_pathNameString[_pathNameString.length() - 1]))
    {   //  Path name string ends with a device separator
        _pathNameString += FileSystem::PathComponentSeparator;
    }
}

int PathName::_findRightmostPathComponentSeparatorIndex(const String & pathNameString)
{
    int result = -1;
    for (UIntSize i = 0; i < pathNameString.length(); i++)
    {
        if (FileSystem::isPathComponentSeparator(pathNameString[i]))
        {
            result = i;
        }
    }
    return result;
}

int PathName::_findRightmostExtensionSeparatorIndex(const String & pathNameString, UIntSize startAtIndex)
{
    int result = -1;
    for (UIntSize i = startAtIndex; i < pathNameString.length(); i++)
    {
        if (FileSystem::isFilenameExtensionSeparator(pathNameString[i]))
        {
            result = i;
        }
    }
    return result;
}

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
bool PathName::_store(wchar_t * pBuffer, UIntSize bufferSize) const
{
    return _pathNameString._store(pBuffer, bufferSize);
}
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
bool PathName::_store(char * pBuffer, UIntSize bufferSize) const
{
    return _pathNameString._store(pBuffer, bufferSize);
}
#endif

//////////
//  Hashing path names
UIntSize CARL_CALLBACK StandardHashFunctions::hashPathName(const PathName & key)
{
    String fullPathNameString = key.toAbsolute().toString();
    if (!FileSystem::CaseSensitiveFileNames)
    {
        fullPathNameString.upperCase();
    }
    return hashString(fullPathNameString);
}

//  End of PathName.cpp
