//
//  PluginLoader.cpp - Implementation of the carl::PluginLoader class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:53 $
//  $Revision: 1.7 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include <string.h>
#include <stdio.h>

#if defined(CARL_LINKAGE_DYNAMIC)

//////////
//  Operations
Array<Plugin*> PluginLoader::loadFromFile(const char * pFileName, const char * pApplicationKey)
{
    require(pFileName != NULL);
    require(pApplicationKey != NULL);

    return loadFromFile(String(pFileName), pApplicationKey);
}

Array<Plugin*> PluginLoader::loadFromFile(const Char * pFileName, const char * pApplicationKey)
{
    require(pFileName != NULL);
    require(pApplicationKey != NULL);

    return loadFromFile(String(pFileName), pApplicationKey);
}

Array<Plugin*> PluginLoader::loadFromFile(const String & fileName, const char * pApplicationKey)
{
    require(pApplicationKey != NULL);

    CriticalSectionLock lock(*_getStaticGuard());

    //  Obtain the library where plugin is defined
    String fullFileName = (PathName(fileName)).toAbsolute().toString();
    if (!FileSystem::CaseSensitiveFileNames)
    {
        fullFileName.upperCase();
    }
    Library * pLibrary;
    bool newLibraryCreated;
    if (getMapFileNamesToLibraries()->containsKey(fullFileName))
    {
        pLibrary = getMapFileNamesToLibraries()->get(fullFileName);
        newLibraryCreated = false;
    }
    else
    {
        try
        {
            pLibrary = new Library(fullFileName);
            newLibraryCreated = true;
        }
        catch (IOException)
        {
            return Array<Plugin*>();
        }
    }
    //  Find plugin provider
    PluginProviderProc pluginProviderProc = (PluginProviderProc)(pLibrary->resolve("CARL_GetPlugins"));
    if (pluginProviderProc == NULL)
    {
        if (newLibraryCreated)
        {
            delete pLibrary;
        }
        return Array<Plugin*>();
    }
    //  Find valid plugins for this version of CARL & application key
    Array<Plugin *> allPlugins;
    (*pluginProviderProc)(allPlugins);
    Array<Plugin *> validPlugins;
    for (UIntSize i = 0; i < allPlugins.size(); i++)
    {
        Plugin * pPlugin = allPlugins[i];
        if (pPlugin != NULL &&
            pPlugin->getBuildKey() != NULL &&
            strcmp(pPlugin->getBuildKey(), CARL_BUILD_KEY) == 0 &&
            strcmp(pPlugin->getApplicationKey(), pApplicationKey) == 0)
        {
            validPlugins.append(pPlugin);
        }
    }
    //  Register library if necessary & return plugins
    if (validPlugins.size() == 0)
    {
        if (newLibraryCreated)
        {
            delete pLibrary;
        }
    }
    else
    {
        getMapFileNamesToLibraries()->insert(fullFileName, pLibrary);
    }
    return validPlugins;
}

Array<Plugin*> PluginLoader::loadFromFile(const PathName & fileName, const char * pApplicationKey)
{
    require(pApplicationKey != NULL);

    return loadFromFile(fileName.toString(), pApplicationKey);
}

Array<Plugin*> PluginLoader::loadFromDirectory(const char * pDirectoryName, const char * pApplicationKey, bool processSubdirectories /*= false*/)
{
    require(pDirectoryName != NULL);
    require(pApplicationKey != NULL);

    return loadFromDirectory(String(pDirectoryName), pApplicationKey, processSubdirectories);
}

Array<Plugin*> PluginLoader::loadFromDirectory(const Char * pDirectoryName, const char * pApplicationKey, bool processSubdirectories /*= false*/)
{
    require(pDirectoryName != NULL);
    require(pApplicationKey != NULL);

    return loadFromDirectory(String(pDirectoryName), pApplicationKey, processSubdirectories);
}

Array<Plugin*> PluginLoader::loadFromDirectory(const String & directoryName, const char * pApplicationKey, bool processSubdirectories /*= false*/)
{
    require(pApplicationKey != NULL);

    CriticalSectionLock lock(*_getStaticGuard());

    Array<Plugin*> result;
    PathName directory(directoryName);
    Array<DirectoryEntry> directoryEntries;
    if (!FileSystem::getDirectoryContents(directory, directoryEntries))
    {
        return result;
    }
    for (UIntSize i = 0; i < directoryEntries.size(); i++)
    {
        PathName entry(directoryEntries[i].getFullName());
        if (entry.isDirectory() && processSubdirectories)
        {
            result.append(loadFromDirectory(entry.toString(), pApplicationKey, true));
        }
        else if (entry.hasExtension(Library::PreferredExtension))
        {
            result.append(loadFromFile(entry.toString(), pApplicationKey));
        }
    }
    return result;
}

Array<Plugin*> PluginLoader::loadFromDirectory(const PathName & directoryName, const char * pApplicationKey, bool processSubdirectories /*= false*/)
{
    require(pApplicationKey != NULL);

    return loadFromDirectory(directoryName.toString(), pApplicationKey, processSubdirectories);
}

Array<Plugin*> PluginLoader::loadInstalledPlugins(const char * pApplicationKey)
{
    require(pApplicationKey != NULL);

    CriticalSectionLock lock(*_getStaticGuard());

    String applicationFile = Process::getCurrentProcess()->getApplicationFileName();
    String applicationDirectory = (PathName(applicationFile)).getParent().toString();
    Array<Plugin*> result;
    //  First come plugins defined by application itself...
    result.append(loadFromFile(applicationFile, pApplicationKey));
    //  ...then plugins from "pluging" subdirectory of application's directory...
    result.append(loadFromDirectory((PathName(applicationDirectory, "plugins")).toString(), pApplicationKey, false));
    //  ...then plugins from application directory
    result.append(loadFromDirectory(applicationDirectory, pApplicationKey, false));
    //  Done
    return result;
}

//////////
//  Implementation helpers
HashMap<String, Library*> * PluginLoader::getMapFileNamesToLibraries()
{
    static HashMap<String, Library*> _mapFileNamesToLibraries(StandardHashFunctions::hashString);
    return &_mapFileNamesToLibraries;
}

CriticalSection * PluginLoader::_getStaticGuard()
{
    static CriticalSection guard;
    return &guard;
}

#endif  //  defined(CARL_LINKAGE_DYNAMIC)

//  End of Extension.cpp
