//
//  Process.cpp - Implementation of the carl::Process class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:53 $
//  $Revision: 1.31 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <windows.h>
    #include <psapi.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <unistd.h>
    #include <stdlib.h>
    #include <signal.h>
    #include <sys/wait.h>
    #include <sys/times.h>
    #include <errno.h>
    #include <fcntl.h>
    extern const char * _CARL_pArgv0;
#else
    #error Unsupported platform
#endif

Process * Process::_pCurrentProcess = NULL;

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
namespace
{
    //////////
    //  Binary process streams
    class CARL_PUBLIC _ProcessByteOutputStream : public IByteInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _ProcessByteOutputStream(const _ProcessByteOutputStream &);
        void                    operator = (const _ProcessByteOutputStream &);

        //////////
        //  Construction/destruction
    public:
        _ProcessByteOutputStream(HANDLE hProcess, HANDLE hPipe) 
        { 
            _hProcess = hProcess;
            _hPipe = hPipe; 
        }

        virtual ~_ProcessByteOutputStream() 
        {
            CloseHandle(_hPipe);
        }

        //////////
        //  IByteInput
    public:
        virtual UIntSize        available() throw(IOException) 
        { 
            UInt8 byte;
            DWORD bytesRead, totalBytesAvail, bytesLeftThisMessage;
            DWORD lastError = GetLastError();
            if (!PeekNamedPipe(_hPipe, &byte, 1, &bytesRead, &totalBytesAvail, &bytesLeftThisMessage))
            {
                SetLastError(lastError);
                throw IOException();
            }
            return totalBytesAvail;
        }

        virtual int             read() throw(IOException) 
        { 
            UInt8 byte;
            DWORD bytesRead, totalBytesAvail, bytesLeftThisMessage;

            for (; ; )
            {
                DWORD lastError = GetLastError();
                if (!PeekNamedPipe(_hPipe, &byte, 1, &bytesRead, &totalBytesAvail, &bytesLeftThisMessage))
                {
                    SetLastError(lastError);
                    throw IOException();
                }
                if (totalBytesAvail == 0)
                {   //  No bytes available in pipe
                    DWORD exitCodeProcess;
                    if (!GetExitCodeProcess(_hProcess, &exitCodeProcess))
                    {
                        SetLastError(lastError);
                        throw IOException();
                    }
                    if (exitCodeProcess == STILL_ACTIVE)
                    {   //  Process is still active, i.e. can potentially write
                        //  something to the pipe
                        Sleep(0);
                        continue;
                    }
                    else
                    {   //  Process has ended, no more writing to the pipe will be done
                        SetLastError(lastError);
                        return -1;
                    }
                }
                else
                {   //  Some bytes available in pipe
                    if (!ReadFile(_hPipe, &byte, 1, &bytesRead, NULL))
                    {
                        SetLastError(lastError);
                        throw IOException();
                    }
                    if (bytesRead == 0)
                    {
                        SetLastError(lastError);
                        return -1;
                    }
                    SetLastError(lastError);
                    return byte;
                }
            }
        }

        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException)
        {
            for (UIntSize i = 0; i < bytesToRead; i++)
            {
                int bb = read();
                if (bb == -1)
                {
                    return i;
                }
                (static_cast<UInt8*>(pBuffer))[i] = static_cast<UInt8>(bb);
            }
            return bytesToRead;
        }

        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException)
        {
            for (UIntSize i = 0; i < bytesToSkip; i++)
            {
                int bb = read();
                if (bb == -1)
                {
                    return i;
                }
            }
            return bytesToSkip;
        }

        //////////
        //  Implementation
    private:
        HANDLE                  _hProcess;
        HANDLE                  _hPipe;
    };

    class CARL_PUBLIC _ProcessByteInputStream : public IByteOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _ProcessByteInputStream(const _ProcessByteInputStream &);
        void                    operator = (const _ProcessByteInputStream &);

        //////////
        //  Construction/destruction
    public:
        _ProcessByteInputStream(HANDLE hProcess, HANDLE hPipe) 
        { 
            _hProcess = hProcess;
            _hPipe = hPipe; 
        }

        virtual ~_ProcessByteInputStream() 
        {
            CloseHandle(_hPipe);
        }

        //////////
        //  IByteOutput
    public:
        virtual void            flush() throw(IOException) 
        { 
            //FlushFileBuffers(_hPipe); 
        }

        virtual void            write(UInt8 byte) throw(IOException) 
        { 
            DWORD bytesWritten;
            DWORD lastError = GetLastError();
            if (!WriteFile(_hPipe, &byte, 1, &bytesWritten, NULL) || bytesWritten != 1)
            {
                SetLastError(lastError);
                throw IOException();
            }
            //FlushFileBuffers(_hPipe); 
        }

        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException) 
        { 
            DWORD bytesWritten;
            DWORD lastError = GetLastError();
            if (!WriteFile(_hPipe, pBuffer, bytesToWrite, &bytesWritten, NULL) || bytesWritten != bytesToWrite)
            {
                SetLastError(lastError);
                throw IOException();
            }
            //FlushFileBuffers(_hPipe); 
        }

        //////////
        //  Implementation
    private:
        HANDLE                  _hProcess;
        HANDLE                  _hPipe;
    };

    //////////
    //  Text process streams
    class CARL_PUBLIC _ProcessOutputStream : public IProcessOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _ProcessOutputStream(const _ProcessOutputStream &);
        void                    operator = (const _ProcessOutputStream &);

        //////////
        //  Construction/destruction
    public:
        _ProcessOutputStream(HANDLE hProcess, HANDLE hPipe) 
            :   _pbis(hProcess, hPipe), _tis(&_pbis) 
        {
        }

        virtual ~_ProcessOutputStream() 
        {
        }

        //////////
        //  IProcessOutput
    public:
        virtual CharacterSet *  getCharacterSet() const 
        { 
            return _tis.getCharacterSet(); 
        }

        virtual void            setCharacterSet(CharacterSet * pCharacterSet) 
        { 
            _tis.setCharacterSet(pCharacterSet); 
        }

        //////////
        //  ITextInput
    public:
        virtual bool            ready() throw(IOException) 
        { 
            return _tis.ready(); 
        }

        virtual bool            read(Char & c) throw(IOException) 
        { 
            return _tis.read(c); 
        }

        virtual UIntSize        read(Char * pBuffer, UIntSize charsToRead) throw(IOException) 
        { 
            return _tis.read(pBuffer, charsToRead); 
        }

        virtual bool            readLine(String & line, bool discardEoln = false) throw(IOException) 
        { 
            return _tis.readLine(line, discardEoln); 
        }

        virtual UIntSize        skip(UIntSize charsToSkip) throw(IOException) 
        { 
            return _tis.skip(charsToSkip); 
        }

        virtual bool            unread(Char c) throw(IOException) 
        { 
            return _tis.unread(c); 
        }

        //////////
        //  Implementation
    private:
        _ProcessByteOutputStream _pbis;
        TextInputStream         _tis;
    };

    class CARL_PUBLIC _ProcessInputStream : public IProcessInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _ProcessInputStream(const _ProcessInputStream &);
        void                    operator = (const _ProcessInputStream &);

        //////////
        //  Construction/destruction
    public:
        _ProcessInputStream(HANDLE hProcess, HANDLE hPipe) 
            :   _pbos(hProcess, hPipe), _tos(&_pbos) 
        {
        }

        virtual ~_ProcessInputStream() 
        {
        }

        //////////
        //  IProcessInput
    public:
        virtual CharacterSet *  getCharacterSet() const 
        { 
            return _tos.getCharacterSet(); 
        }

        virtual void            setCharacterSet(CharacterSet * pCharacterSet) 
        { 
            _tos.setCharacterSet(pCharacterSet); 
        }

        //////////
        //  ITextOutput
    public:
        virtual void            flush() throw(IOException) 
        { 
            _tos.flush(); 
        }

        virtual void            write(Char c) throw(IOException) 
        { 
            _tos.write(c); 
        }

        virtual void            write(const char * pBuffer) throw(IOException) 
        { 
            _tos.write(pBuffer); 
        }

        virtual void            write(const char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.write(pBuffer, charsToWrite); 
        }

        virtual void            write(const Char * pBuffer) throw(IOException) 
        { 
            _tos.write(pBuffer); 
        }

        virtual void            write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.write(pBuffer, charsToWrite); 
        }

        virtual void            write(const String & s) throw(IOException) 
        { 
            _tos.write(s); 
        }

        virtual void            writeLn() throw(IOException) 
        { 
            _tos.writeLn(); 
        }

        virtual void            writeLn(Char c) throw(IOException) 
        { 
            _tos.writeLn(c); 
        }

        virtual void            writeLn(const char * pBuffer) throw(IOException) 
        { 
            _tos.writeLn(pBuffer); 
        }

        virtual void            writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.writeLn(pBuffer, charsToWrite); 
        }

        virtual void            writeLn(const Char * pBuffer) throw(IOException) 
        { 
            _tos.writeLn(pBuffer); 
        }

        virtual void            writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.writeLn(pBuffer, charsToWrite); 
        }

        virtual void            writeLn(const String & s) throw(IOException) 
        { 
            _tos.writeLn(s); 
        }

        //////////
        //  Implementation
    private:
        _ProcessByteInputStream _pbos;
        TextOutputStream        _tos;
    };
}
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
namespace
{
    //////////
    //  Binary process streams
    class CARL_PUBLIC _ProcessByteOutputStream : public IByteInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _ProcessByteOutputStream(const _ProcessByteOutputStream &);
        void                    operator = (const _ProcessByteOutputStream &);

        //////////
        //  Construction/destruction
    public:
        explicit _ProcessByteOutputStream(pid_t childPid, int hPipe) 
        { 
            _childPid = childPid;
            _hPipe = hPipe;
            _readaheadExists = false;
            //  Switch pipe to nonblocking mode
            if (fcntl(hPipe, F_SETFL, fcntl(hPipe, F_GETFL) | O_NONBLOCK) != 0)
            {
                throw SystemError();
            }
        }

        virtual ~_ProcessByteOutputStream() 
        {
            ::close(_hPipe);
        }

        //////////
        //  IByteInput
    public:
        virtual UIntSize        available() throw(IOException) 
        { 
            if (_readaheadExists)
            {
                return 1;
            }
            else
            {
                int b = _read();
                if (b == -1)
                {
                    return 0;
                }
                else
                {
                    _readaheadExists = true;
                    _readahead = static_cast<UInt8>(b);
                    return 1;
                }
            }
        }

        virtual int             read() throw(IOException) 
        { 
            if (_readaheadExists)
            {
                _readaheadExists = false;
                return _readahead;
            }
            else
            {
                return _read();
            }
        }

        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException)
        {
            for (UIntSize i = 0; i < bytesToRead; i++)
            {
                int bb = read();
                if (bb == -1)
                {
                    return i;
                }
                (static_cast<UInt8*>(pBuffer))[i] = static_cast<UInt8>(bb);
            }
            return bytesToRead;
        }

        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException)
        {
            for (UIntSize i = 0; i < bytesToSkip; i++)
            {
                int bb = read();
                if (bb == -1)
                {
                    return i;
                }
            }
            return bytesToSkip;
        }

        //////////
        //  Implementation
    private:
        pid_t                   _childPid;
        int                     _hPipe;
        bool                    _readaheadExists;
        UInt8                   _readahead;

        int                     _read() throw(IOException) 
        {
            require(!_readaheadExists);

            for (; ; )
            {
                UInt8 b;
                switch (::read(_hPipe, &b, 1))
                {
                    case 1: //  1 byte read
                        return b;
                    case 0: //  no data available from the child
                        {
                            int status;
                            waitpid(_childPid, &status, WNOHANG | WUNTRACED);
                            if (WIFEXITED(status))
                            {   //  The child has finished - there will be no more input from it
                                return -1;
                            }
                        }
                        continue;
                    case -1:
                        if (errno == EAGAIN)
                        {
                            int status;
                            waitpid(_childPid, &status, WNOHANG | WUNTRACED);
                            if (WIFEXITED(status))
                            {   //  The child has finished - there will be no more input from it
                                return -1;
                            }
                            continue;
                        }
                        throw IOException();
                }
            }
        }
    };

    class CARL_PUBLIC _ProcessByteInputStream : public IByteOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _ProcessByteInputStream(const _ProcessByteInputStream &);
        void                    operator = (const _ProcessByteInputStream &);

        //////////
        //  Construction/destruction
    public:
        _ProcessByteInputStream(int hPipe) 
        { 
            _hPipe = hPipe; 
        }

        virtual ~_ProcessByteInputStream() 
        {
            ::close(_hPipe);
        }

        //////////
        //  IByteOutput
    public:
        virtual void            flush() throw(IOException) 
        { 
        }

        virtual void            write(UInt8 byte) throw(IOException) 
        { 
            int oldErrno = errno;
            if (::write(_hPipe, &byte, 1) != 1)
            {
                errno = oldErrno;
                throw IOException();
            }
            flush();
        }

        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException) 
        { 
            int oldErrno = errno;
            if (static_cast<UIntSize>(::write(_hPipe, pBuffer, bytesToWrite)) != bytesToWrite)
            {
                errno = oldErrno;
                throw IOException();
            }
            flush();
        }

        //////////
        //  Implementation
    private:
        int                     _hPipe;
    };

    //////////
    //  Text process streams
    class CARL_PUBLIC _ProcessOutputStream : public IProcessOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _ProcessOutputStream(const _ProcessOutputStream &);
        void                    operator = (const _ProcessOutputStream &);

        //////////
        //  Construction/destruction
    public:
        _ProcessOutputStream(pid_t childPid, int hPipe) 
            :   _pbis(childPid, hPipe), _tis(&_pbis) 
        {
        }

        virtual ~_ProcessOutputStream() 
        {
        }

        //////////
        //  IProcessOutput
    public:
        virtual CharacterSet *  getCharacterSet() const 
        { 
            return _tis.getCharacterSet(); 
        }

        virtual void            setCharacterSet(CharacterSet * pCharacterSet) 
        { 
            _tis.setCharacterSet(pCharacterSet); 
        }

        //////////
        //  ITextInput
    public:
        virtual bool            ready() throw(IOException) 
        { 
            return _tis.ready(); 
        }

        virtual bool            read(Char & c) throw(IOException) 
        { 
            return _tis.read(c); 
        }

        virtual UIntSize        read(Char * pBuffer, UIntSize charsToRead) throw(IOException) 
        { 
            return _tis.read(pBuffer, charsToRead); 
        }

        virtual bool            readLine(String & line, bool discardEoln = false) throw(IOException) 
        { 
            return _tis.readLine(line, discardEoln); 
        }

        virtual UIntSize        skip(UIntSize charsToSkip) throw(IOException) 
        { 
            return _tis.skip(charsToSkip); 
        }

        virtual bool            unread(Char c) throw(IOException) 
        { 
            return _tis.unread(c); 
        }

        //////////
        //  Implementation
    private:
        _ProcessByteOutputStream _pbis;
        TextInputStream         _tis;
    };

    class CARL_PUBLIC _ProcessInputStream : public IProcessInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _ProcessInputStream(const _ProcessInputStream &);
        void                    operator = (const _ProcessInputStream &);

        //////////
        //  Construction/destruction
    public:
        _ProcessInputStream(int hPipe) 
            :   _pbos(hPipe), _tos(&_pbos) 
        {
        }

        virtual ~_ProcessInputStream() 
        {
        }

        //////////
        //  IProcessInput
    public:
        virtual CharacterSet *  getCharacterSet() const 
        { 
            return _tos.getCharacterSet(); 
        }

        virtual void            setCharacterSet(CharacterSet * pCharacterSet) 
        { 
            _tos.setCharacterSet(pCharacterSet); 
        }

        //////////
        //  ITextOutput
    public:
        virtual void            flush() throw(IOException) 
        { 
            _tos.flush(); 
        }

        virtual void            write(Char c) throw(IOException) 
        { 
            _tos.write(c); 
        }

        virtual void            write(const char * pBuffer) throw(IOException) 
        { 
            _tos.write(pBuffer); 
        }

        virtual void            write(const char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.write(pBuffer, charsToWrite); 
        }

        virtual void            write(const Char * pBuffer) throw(IOException) 
        { 
            _tos.write(pBuffer); 
        }

        virtual void            write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.write(pBuffer, charsToWrite); 
        }

        virtual void            write(const String & s) throw(IOException) 
        { 
            _tos.write(s); 
        }

        virtual void            writeLn() throw(IOException) 
        { 
            _tos.writeLn(); 
        }

        virtual void            writeLn(Char c) throw(IOException) 
        { 
            _tos.writeLn(c); 
        }

        virtual void            writeLn(const char * pBuffer) throw(IOException) 
        { 
            _tos.writeLn(pBuffer); 
        }

        virtual void            writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.writeLn(pBuffer, charsToWrite); 
        }

        virtual void            writeLn(const Char * pBuffer) throw(IOException) 
        { 
            _tos.writeLn(pBuffer); 
        }

        virtual void            writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.writeLn(pBuffer, charsToWrite); 
        }

        virtual void            writeLn(const String & s) throw(IOException) 
        { 
            _tos.writeLn(s); 
        }

        //////////
        //  Implementation
    private:
        _ProcessByteInputStream _pbos;
        TextOutputStream        _tos;
    };
}
#else
    #error Unsupported platform
#endif

//////////
//  Construction/destruction
Process::Process(bool)
{
    _state = State_Active;  //  We're attaching to a current process
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _hProcess = GetCurrentProcess();
    _isCurrent = true;
    _processId = GetCurrentProcessId();
    _pEnvironment = new Environment();
    char * pEnvironmentStrings = static_cast<char*>(GetEnvironmentStrings());
    UIntSize scan = 0;
    for (; ; )
    {
        String name, value;
        while (pEnvironmentStrings[scan] != 0 && (pEnvironmentStrings[scan] != '=' || name.length() == 0))
        {
            name += pEnvironmentStrings[scan++];
        }
        if (pEnvironmentStrings[scan] == '=')
        {
            scan++;
            while (pEnvironmentStrings[scan] != 0)
            {
                value += pEnvironmentStrings[scan++];
            }
        }
        scan++;
        if (!name.startsWith('='))
        {
            _pEnvironment->setVariableValue(name, value);
        }
        if (pEnvironmentStrings[scan] == 0)
        {
            break;
        }
    }
    FreeEnvironmentStrings(pEnvironmentStrings);
    _pStdin = NULL;
    _pStdout = NULL;
    _pStderr = NULL;
    _hStdinRead = NULL;
    _hStdoutWrite = NULL;
    _hStderrWrite = NULL;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _pid = (int)getpid();
    _isCurrent = true;
    _exitCode = -1;
    _exitState = GetExitCodeOutcome_NotAvailable;

    _pEnvironment = new Environment();
    for (int i = 0; environ[i] != NULL; i++)
    {
        String nameAndValue = String(environ[i]);
        UIntSize eqIndex;
        if (nameAndValue.findFirst('=', eqIndex))
        {
            String name = nameAndValue.left(eqIndex).trimmed();
            String value = nameAndValue.mid(eqIndex + 1).trimmed();
            _pEnvironment->setVariableValue(name, value);
        }
    }
    _pStdin = NULL;
    _pStdout = NULL;
    _pStderr = NULL;
    _hStdinRead = -1;
    _hStdoutWrite = -1;
    _hStderrWrite = -1;
#else
    #error Unsupported platform
#endif
}

Process::Process()
{
    _state = State_Initializing;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _hProcess = NULL;
    _isCurrent = false;
    _pEnvironment = new Environment();
    _pEnvironment->_mapNamesToValues = getCurrentProcess()->_pEnvironment->_mapNamesToValues;
    _processId = 0;
    _pStdin = NULL;
    _pStdout = NULL;
    _pStderr = NULL;
    _hStdinRead = NULL;
    _hStdoutWrite = NULL;
    _hStderrWrite = NULL;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _pid = -1;
    _isCurrent = false;
    _exitCode = -1;
    _exitState = GetExitCodeOutcome_NotAvailable;
    _pEnvironment = new Environment();
    _pEnvironment->_mapNamesToValues = getCurrentProcess()->_pEnvironment->_mapNamesToValues;
    _pStdin = NULL;
    _pStdout = NULL;
    _pStderr = NULL;
    _hStdinRead = -1;
    _hStdoutWrite = -1;
    _hStderrWrite = -1;
    _elapsedTimeStart = _elapsedTimeEnd = TimeSpan::getTimeSpanSinceReferencePoint();
#else
    #error Unsupported platform
#endif
}

Process::~Process()
{
    delete _pEnvironment;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    if (_hProcess != NULL && !_isCurrent)
    {
        CloseHandle(_hProcess);
    }
    if (_hStdinRead != 0) 
    {
        CloseHandle(_hStdinRead);
    }
    if (_hStdoutWrite != 0) 
    {
        CloseHandle(_hStdoutWrite);
    }
    if (_hStderrWrite != 0) 
    {
        CloseHandle(_hStderrWrite);
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (_hStdinRead != -1) 
    {
        ::close(_hStdinRead);
    }
    if (_hStdoutWrite != -1) 
    {
        ::close(_hStdoutWrite);
    }
    if (_hStderrWrite != -1) 
    {
        ::close(_hStderrWrite);
    }
#else
    #error Unsupported platform
#endif
    delete _pStdin;
    delete _pStdout;
    delete _pStderr;
}

//////////
//  Operations
Process * Process::getCurrentProcess()
{
    _getCurrentProcessGuard()->enter();
    if (_pCurrentProcess == NULL)
    {
        _pCurrentProcess = new Process(true);
    }
    Process * pCurrentProcess = _pCurrentProcess;
    _getCurrentProcessGuard()->leave();
    return pCurrentProcess;
}

Process::State Process::getState() const
{
    if (_state != State_Active)
    {   //  A last known state of this process was not 'Active',
        //  therefore it could not have changed
        return _state;
    }
    else
    {   //  The process may have finished by now
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        DWORD exitCode;
        if (!GetExitCodeProcess(reinterpret_cast<HANDLE>(_hProcess), &exitCode))
        {   //  Could not find out if a process is still active.
            //  The best we can do is to report the last known state
            return _state;
        }
        else if (exitCode != STILL_ACTIVE)
        {   //  The process it has finished
            _state = State_Finished;
        }
        //  The process state has been updated as necessary
        return _state;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        int status;
        struct rusage ru;
        pid_t waitResult = wait4(_pid, &status, WNOHANG, &ru);
        if (waitResult == _pid)
        {   //  The child has exited
            _systemTime = TimeSpan((((ru.ru_stime.tv_sec) * CARL_UINT64_CONSTANT(1000000)) +
                                    ru.ru_stime.tv_usec) * CARL_UINT64_CONSTANT(1000));
            _userTime = TimeSpan((((ru.ru_utime.tv_sec) * CARL_UINT64_CONSTANT(1000000)) +
                                  ru.ru_utime.tv_usec) * CARL_UINT64_CONSTANT(1000));
            _elapsedTimeEnd = TimeSpan::getTimeSpanSinceReferencePoint();
            _state = State_Finished;
            if (WIFEXITED(status))
            {
                _exitCode = WEXITSTATUS(status);
                _exitState = GetExitCodeOutcome_Success;
            }
            else if (WIFSIGNALED(status))
            {
                _exitCode = ExitCode_Failure;
                _exitState = GetExitCodeOutcome_Success;
            }
        }
        return _state;
#else
    #error Unsupported platform
#endif
    }
}

void Process::exit(ExitCode exitCode)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    require(_hProcess != NULL);

    if (_isCurrent)
    {
        ::exit(exitCode);
    }
    else
    {
        TerminateProcess(_hProcess, exitCode);
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    require(_pid != -1);

    if (_isCurrent)
    {
        ::exit(exitCode);
    }
    else
    {
         kill((pid_t)_pid, SIGKILL);
    }
#else
    #error Unsupported platform
#endif
}

String Process::getApplicationFileName() const
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    require(_hProcess != NULL);

    //  Can't use GetProcessImageFileName, because that returns the internal
    //  process file name 'Device\Harddisk1\...' instead of Win32 file name 'C:\...'
    Array<HMODULE> modules;
    modules.setSize(64);
    DWORD modulesNeeded;
    BOOL b = EnumProcessModules(_hProcess, modules.getData(), modules.size() * sizeof(HMODULE), &modulesNeeded);
    if (modulesNeeded > modules.size() * sizeof(HMODULE))
    {
        modules.setSize(modulesNeeded / sizeof(HMODULE));
        b = EnumProcessModules(_hProcess, modules.getData(), modules.size() * sizeof(HMODULE), &modulesNeeded);
    }
    if (!b)
    {
        return String("");
    }
    modules.setSize(modulesNeeded / sizeof(HMODULE));
    //  Try finding the .exe filename
    for (UIntSize i = 0; i < modules.size(); i++)
    {
        wchar_t fileName[_MAX_PATH];
        DWORD dw = GetModuleFileNameExW(_hProcess, modules[i], fileName, sizeof(fileName));
        if (dw != 0)
        {
            String result = String(fileName);
            if (result.endsWithIgnoreCase(".exe"))
            {
                return result;
            }
        }
    }
    //  Otherwise, use the 1st known file name
    for (UIntSize i = 0; i < modules.size(); i++)
    {
        wchar_t fileName[_MAX_PATH];
        DWORD dw = GetModuleFileNameExW(_hProcess, modules[i], fileName, sizeof(fileName));
        if (dw != 0)
        {
            return String(fileName);
        }
    }
    return String("");
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (_CARL_pArgv0 != NULL)
    {
        return String(_CARL_pArgv0);
    }
    return String("");
#else
    #error Unsupported platform
#endif
}

Process::StartOutcome Process::start(const Array<String> & commandNameAndArguments, int captureStandardStreams)
{
    require(commandNameAndArguments.size() != 0 && commandNameAndArguments[0].length() != 0);

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    require(_hProcess == NULL && !_isCurrent);

    //  Prepare the Unicode command line for CreateProcess
    UIntSize requiredCommandLineLength = 0;
    for (UIntSize i = 0; i < commandNameAndArguments.size(); i++)
    {
        String argument = commandNameAndArguments[i];
        requiredCommandLineLength += 3 + argument.length();
        for (UIntSize j = 0; j < argument.length(); j++)
        {
            if (argument[j] == '"')
            {
                requiredCommandLineLength++;
            }
            else if (argument[j] < 32 || argument[j] > 0x0000FFFF)
            {   //  Not a Win32 Unicode character
                return StartOutcome_InvalidCommandLine;
            }
        }
    }
    requiredCommandLineLength++;
    wchar_t * pCommandLine = new wchar_t[requiredCommandLineLength];
    UIntSize pCommandLinePosition = 0;
    for (UIntSize i = 0; i < commandNameAndArguments.size(); i++)
    {
        String argument = commandNameAndArguments[i];
        bool argumentNeedsQuoting = false;
        for (UIntSize j = 0; j < argument.length(); j++)
        {
            if (argument[j] == '"' || Iso10646::isSpace(argument[j]))
            {
                argumentNeedsQuoting = true;
                break;
            }
        }
        if (argumentNeedsQuoting)
        {
            pCommandLine[pCommandLinePosition++] = '"';
            for (UIntSize j = 0; j < argument.length(); j++)
            {
                if (argument[j] == '"')
                {
                    pCommandLine[pCommandLinePosition++] = '"';
                }
                pCommandLine[pCommandLinePosition++] = (wchar_t)argument[j];
            }
            pCommandLine[pCommandLinePosition++] = '"';
        }
        else
        {
            for (UIntSize j = 0; j < argument.length(); j++)
            {
                pCommandLine[pCommandLinePosition++] = (wchar_t)argument[j];
            }
        }
        if (i + 1 < commandNameAndArguments.size())
        {
            pCommandLine[pCommandLinePosition++] = ' ';
        }
    }
    pCommandLine[pCommandLinePosition++] = 0;
    require(pCommandLinePosition <= requiredCommandLineLength);

    //  Prepare the environment for the new process
    UIntSize requiredEnvironmentLength = 0;
    Array<String> environmentVariableNames = _pEnvironment->getAllVariableNames();
    for (bool sorted = false; !sorted; )
    {
        sorted = true;
        for (UIntSize i = 0; i + 1 < environmentVariableNames.size(); i++)
        {
            if (environmentVariableNames[i].upperCased() > environmentVariableNames[i + 1].upperCased())
            {
                String temp = environmentVariableNames[i];
                environmentVariableNames[i] = environmentVariableNames[i + 1];
                environmentVariableNames[i + 1] = temp;
                sorted = false;
            }
        }
    }
    for (UIntSize i = 0; i < environmentVariableNames.size(); i++)
    {
        String name = environmentVariableNames[i];
        String value = _pEnvironment->getVariableValue(name, "");
        requiredEnvironmentLength += name.length() + value.length() + 2;
    }
    requiredEnvironmentLength++;
    wchar_t * pEnvironmentBlock = new wchar_t[requiredEnvironmentLength];
    UIntSize pEnvironmentBlockPosition = 0;
    for (UIntSize i = 0; i < environmentVariableNames.size(); i++)
    {
        String name = environmentVariableNames[i];
        String value = _pEnvironment->getVariableValue(name, "");
        for (UIntSize j = 0; j < name.length(); j++)
        {
            Char c = name[j];
            if ((c & 0xFFFF0000) != 0)
            {
                delete [] pCommandLine;
                return StartOutcome_InvalidEnvironment;
            }
            pEnvironmentBlock[pEnvironmentBlockPosition++] = (wchar_t)c;
        }
        pEnvironmentBlock[pEnvironmentBlockPosition++] = '=';
        for (UIntSize j = 0; j < value.length(); j++)
        {
            Char c = value[j];
            if ((c & 0xFFFF0000) != 0)
            {
                delete [] pCommandLine;
                return StartOutcome_InvalidEnvironment;
            }
            pEnvironmentBlock[pEnvironmentBlockPosition++] = (wchar_t)c;
        }
        pEnvironmentBlock[pEnvironmentBlockPosition++] = 0;
    }
    pEnvironmentBlock[pEnvironmentBlockPosition++] = 0;

    //  Prepare stdio streams for the process
    SECURITY_ATTRIBUTES saAttr; 
    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES); 
    saAttr.bInheritHandle = TRUE; 
    saAttr.lpSecurityDescriptor = NULL; 

    HANDLE hStdinRead = 0, hStdinWrite = 0;
    HANDLE hStdoutRead = 0, hStdoutWrite = 0;
    HANDLE hStderrRead = 0, hStderrWrite = 0;
    if ((captureStandardStreams & CaptureStandardInputStream) != 0)
    {
        DWORD oldLastError = GetLastError();
        //  Create pipe to send input from parent to child process
        if (!CreatePipe(&hStdinRead, &hStdinWrite, &saAttr, 0))
        {
            delete [] pCommandLine;
            delete [] pEnvironmentBlock;
            SetLastError(oldLastError);
            return StartOutcome_InsufficientMemory;
        }
        //  The write end of the pipe must not be inheritable
        HANDLE hDupStdinWrite;
        if (!DuplicateHandle(GetCurrentProcess(), hStdinWrite, GetCurrentProcess(), &hDupStdinWrite, 0, FALSE, DUPLICATE_SAME_ACCESS))
        {   //  OOPS! Cannot duplicate the handle
            CloseHandle(hStdinRead);
            CloseHandle(hStdinWrite);
            delete [] pCommandLine;
            delete [] pEnvironmentBlock;
            SetLastError(oldLastError);
            return StartOutcome_InsufficientMemory;
        }
        CloseHandle(hStdinWrite);
        hStdinWrite = hDupStdinWrite;
        SetLastError(oldLastError);
    }
    if ((captureStandardStreams & CaptureStandardOutputStream) != 0)
    {
        DWORD oldLastError = GetLastError();
        //  Create pipe to receive output from child to parent process
        if (!CreatePipe(&hStdoutRead, &hStdoutWrite, &saAttr, 0))
        {
            if (hStdinRead != 0) 
            {
                CloseHandle(hStdinRead);
                CloseHandle(hStdinWrite);
            }
            delete [] pCommandLine;
            delete [] pEnvironmentBlock;
            SetLastError(oldLastError);
            return StartOutcome_InsufficientMemory;
        }
        //  The read end of the pipe must not be inheritable
        HANDLE hDupStdoutRead;
        if (!DuplicateHandle(GetCurrentProcess(), hStdoutRead, GetCurrentProcess(), &hDupStdoutRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
        {   //  OOPS! Cannot duplicate the handle
            if (hStdinRead != 0) 
            {
                CloseHandle(hStdinRead);
                CloseHandle(hStdinWrite);
            }
            CloseHandle(hStdoutRead);
            CloseHandle(hStdoutWrite);
            delete [] pCommandLine;
            delete [] pEnvironmentBlock;
            SetLastError(oldLastError);
            return StartOutcome_InsufficientMemory;
        }
        CloseHandle(hStdoutRead);
        hStdoutRead = hDupStdoutRead;
        SetLastError(oldLastError);
    }
    if ((captureStandardStreams & CaptureStandardErrorStream) != 0)
    {
        DWORD oldLastError = GetLastError();
        //  Create pipe to receive error stream from child to parent process
        if (!CreatePipe(&hStderrRead, &hStderrWrite, &saAttr, 0))
        {
            if (hStdinRead != 0) 
            {
                CloseHandle(hStdinRead);
                CloseHandle(hStdinWrite);
            }
            if (hStdoutRead != 0) 
            {
                CloseHandle(hStdoutRead);
                CloseHandle(hStdoutWrite);
            }
            delete [] pCommandLine;
            delete [] pEnvironmentBlock;
            SetLastError(oldLastError);
            return StartOutcome_InsufficientMemory;
        }
        //  The read end of the pipe must not be inheritable
        HANDLE hDupStderrRead;
        if (!DuplicateHandle(GetCurrentProcess(), hStderrRead, GetCurrentProcess(), &hDupStderrRead, 0, FALSE, DUPLICATE_SAME_ACCESS))
        {   //  OOPS! Cannot duplicate the handle
            if (hStdinRead != 0) 
            {
                CloseHandle(hStdinRead);
                CloseHandle(hStdinWrite);
            }
            if (hStdoutRead != 0) 
            {
                CloseHandle(hStdoutRead);
                CloseHandle(hStdoutWrite);
            }
            CloseHandle(hStderrRead);
            CloseHandle(hStderrWrite);
            delete [] pCommandLine;
            delete [] pEnvironmentBlock;
            SetLastError(oldLastError);
            return StartOutcome_InsufficientMemory;
        }
        CloseHandle(hStderrRead);
        hStderrRead = hDupStderrRead;
        SetLastError(oldLastError);
    }

    //  Start the process!
    DWORD oldLastError = GetLastError();

    STARTUPINFOW startupInfo;
    startupInfo.cb = sizeof(startupInfo);
    startupInfo.lpReserved = NULL;
    startupInfo.lpDesktop = NULL;
    startupInfo.lpTitle = NULL;
    startupInfo.dwX = 0;
    startupInfo.dwY = 0;
    startupInfo.dwXSize = 0;
    startupInfo.dwYSize = 0;
    startupInfo.dwXCountChars = 0;
    startupInfo.dwYCountChars = 0;
    startupInfo.dwFillAttribute = 0;
    startupInfo.dwFlags = STARTF_USESTDHANDLES;
    startupInfo.wShowWindow = 0;
    startupInfo.cbReserved2 = 0;
    startupInfo.lpReserved2 = 0;
    startupInfo.hStdInput = (hStdinRead != 0) ? (hStdinRead) : (GetStdHandle(STD_INPUT_HANDLE));
    startupInfo.hStdOutput = (hStdoutWrite != 0) ? (hStdoutWrite) : (GetStdHandle(STD_OUTPUT_HANDLE));
    startupInfo.hStdError = (hStderrWrite != 0) ? (hStderrWrite) : (GetStdHandle(STD_ERROR_HANDLE));

    PROCESS_INFORMATION processInformation;

    if (CreateProcessW(NULL, pCommandLine, NULL, NULL, TRUE, CREATE_UNICODE_ENVIRONMENT, pEnvironmentBlock, NULL, &startupInfo, &processInformation))
    {   //  Process created successfully
        _hStdinRead = hStdinRead;
        _hStdoutWrite = hStdoutWrite;
        _hStderrWrite = hStderrWrite;
        if ((captureStandardStreams & CaptureStandardInputStream) != 0)
        {
            _pStdin = new _ProcessInputStream(processInformation.hProcess, hStdinWrite);
        }
        if ((captureStandardStreams & CaptureStandardOutputStream) != 0)
        {
            _pStdout = new _ProcessOutputStream(processInformation.hProcess, hStdoutRead);
        }
        if ((captureStandardStreams & CaptureStandardErrorStream) != 0)
        {
            _pStderr = new _ProcessOutputStream(processInformation.hProcess, hStderrRead);
        }
        delete [] pCommandLine;
        delete [] pEnvironmentBlock;
        _hProcess = processInformation.hProcess;
        _processId = processInformation.dwProcessId;
        CloseHandle(processInformation.hThread);
        SetLastError(oldLastError);
        _state = State_Active;
        return StartOutcome_Success;
    }
    else
    {   //  Could not create a process
        delete [] pCommandLine;
        delete [] pEnvironmentBlock;
        if (hStdinRead != 0) 
        {
            CloseHandle(hStdinRead);
            CloseHandle(hStdinWrite);
        }
        if (hStdoutRead != 0) 
        {
            CloseHandle(hStdoutRead);
            CloseHandle(hStdoutWrite);
        }
        if (hStderrRead != 0) 
        {
            CloseHandle(hStderrRead);
            CloseHandle(hStderrWrite);
        }
        DWORD createProcessError = GetLastError();
        SetLastError(oldLastError);
        //  Implement fine-grain diagnostics of the reason WHY the process did not start
        switch (createProcessError)
        {
            case ERROR_FILE_NOT_FOUND:
            case ERROR_PATH_NOT_FOUND:
            case ERROR_INVALID_DRIVE:
                return StartOutcome_ImageNotFound;
            case ERROR_TOO_MANY_OPEN_FILES:
            case ERROR_NOT_ENOUGH_MEMORY:
            case ERROR_OUTOFMEMORY:
                return StartOutcome_InsufficientMemory;
            case ERROR_ACCESS_DENIED:
                return StartOutcome_AccessDenied;
            default:
                return StartOutcome_UnknownError;
        }
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    require(_pid == -1 && !_isCurrent);

    //  Prepare the ANSI command line for execve
    UIntSize requiredCommandLineLength = 0;
    for (UIntSize i = 0; i < commandNameAndArguments.size(); i++)
    {
        String argument = commandNameAndArguments[i];
        requiredCommandLineLength += 3 + argument.length();
        for (UIntSize j = 0; j < argument.length(); j++)
        {
            if (argument[j] < 32 || argument[j] > 0x000000FF)
            {   //  Not an ANSI character
                return StartOutcome_InvalidCommandLine;
            }
        }
    }
    requiredCommandLineLength++;
    char * pCommandLine = new char[requiredCommandLineLength];
    Array<char*> argv;
    UIntSize pCommandLinePosition = 0;
    for (UIntSize i = 0; i < commandNameAndArguments.size(); i++)
    {
        String argument = commandNameAndArguments[i];
        argv.append(pCommandLine + pCommandLinePosition);
        for (UIntSize j = 0; j < argument.length(); j++)
        {
            pCommandLine[pCommandLinePosition++] = (wchar_t)argument[j];
        }
        pCommandLine[pCommandLinePosition++] = 0;
    }
    argv.append(NULL);
    require(pCommandLinePosition <= requiredCommandLineLength);

    //  Prepare the environment for the new process
    Array<char*> envp;
    UIntSize requiredEnvironmentLength = 0;
    Array<String> environmentVariableNames = _pEnvironment->getAllVariableNames();
    for (bool sorted = false; !sorted; )
    {
        sorted = true;
        for (UIntSize i = 0; i + 1 < environmentVariableNames.size(); i++)
        {
            if (environmentVariableNames[i].upperCased() > environmentVariableNames[i + 1].upperCased())
            {
                String temp = environmentVariableNames[i];
                environmentVariableNames[i] = environmentVariableNames[i + 1];
                environmentVariableNames[i + 1] = temp;
                sorted = false;
            }
        }
    }
    for (UIntSize i = 0; i < environmentVariableNames.size(); i++)
    {
        String name = environmentVariableNames[i];
        String value = _pEnvironment->getVariableValue(name, "");
        requiredEnvironmentLength += name.length() + value.length() + 2;
    }
    char * pEnvironmentBlock = new char[requiredEnvironmentLength];
    UIntSize pEnvironmentBlockPosition = 0;
    for (UIntSize i = 0; i < environmentVariableNames.size(); i++)
    {
        envp.append(pEnvironmentBlock + pEnvironmentBlockPosition);
        String name = environmentVariableNames[i];
        String value = _pEnvironment->getVariableValue(name, "");
        for (UIntSize j = 0; j < name.length(); j++)
        {
            Char c = name[j];
            if ((c & 0xFFFFFF00) != 0)
            {
                delete [] pCommandLine;
                return StartOutcome_InvalidEnvironment;
            }
            pEnvironmentBlock[pEnvironmentBlockPosition++] = (wchar_t)c;
        }
        pEnvironmentBlock[pEnvironmentBlockPosition++] = '=';
        for (UIntSize j = 0; j < value.length(); j++)
        {
            Char c = value[j];
            if ((c & 0xFFFFFF00) != 0)
            {
                delete [] pCommandLine;
                return StartOutcome_InvalidEnvironment;
            }
            pEnvironmentBlock[pEnvironmentBlockPosition++] = (wchar_t)c;
        }
        pEnvironmentBlock[pEnvironmentBlockPosition++] = 0;
    }
    envp.append(NULL);

    //  Create pipes for redirected child process standard streams
    int stdinPipe[2];
    int stdoutPipe[2];
    int stderrPipe[2];
    if ((captureStandardStreams & CaptureStandardInputStream) != 0)
    {
        int oldErrno = errno;
        if (pipe(stdinPipe) != 0)
        {
            switch (errno)
            {
                case EMFILE:
                case ENFILE:
                    delete [] pCommandLine;
                    delete [] pEnvironmentBlock;
                    errno = oldErrno;
                    throw OutOfMemoryError(0);
                default:
                    throw SystemError();
            }
        }
    }
    if ((captureStandardStreams & CaptureStandardOutputStream) != 0)
    {
        int oldErrno = errno;
        if (pipe(stdoutPipe) != 0)
        {
            switch (errno)
            {
                case EMFILE:
                case ENFILE:
                    delete [] pCommandLine;
                    delete [] pEnvironmentBlock;
                    errno = oldErrno;
                    throw OutOfMemoryError(0);
                default:
                    throw SystemError();
            }
        }
    }
    if ((captureStandardStreams & CaptureStandardErrorStream) != 0)
    {
        int oldErrno = errno;
        if (pipe(stderrPipe) != 0)
        {
            switch (errno)
            {
                case EMFILE:
                case ENFILE:
                    delete [] pCommandLine;
                    delete [] pEnvironmentBlock;
                    errno = oldErrno;
                    throw OutOfMemoryError(0);
                default:
                    throw SystemError();
            }
        }
    }

    //  Check that command to be exec()'d can be executed:
    //  (a) User has execute permission
    //  (b) Is a file with size > 0 or a symbolic link
    struct stat sbuf;
    if (::access(argv[0], X_OK) != 0 || ::stat(argv[0], &sbuf)  != 0)
    {
        switch (errno)
        {
            case EACCES:
                return StartOutcome_AccessDenied;
            case ENOENT:
            case ENOTDIR:
            case ELOOP:
            case ENAMETOOLONG:
                return StartOutcome_ImageNotFound;
            case ENOMEM:
                return StartOutcome_InsufficientMemory;
            default:
                return StartOutcome_UnknownError;
        }
    }

    //  Here iff both access() & stat() succeeded,
    //  do some more sanity checks based on sbuf
    if (!(S_ISREG(sbuf.st_mode) && sbuf.st_size > 0))
    {
        return StartOutcome_InvalidCommandLine;
    }

    //  Fork a child process
    int oldErrno = errno;
    pid_t childPid = fork();
    if (childPid == -1)
    {   //  fork() error!
        switch (errno)
        {
            case EAGAIN:
            case ENOMEM:
                delete [] pCommandLine;
                delete [] pEnvironmentBlock;
                errno = oldErrno;
                throw OutOfMemoryError(0);
            default:
                throw SystemError();
        }
    }
    else if (childPid == 0)
    {   //  This is a child process - execute the specified command
        //  The child may have to redirect its standard streams to pipes created by parent
        if ((captureStandardStreams & CaptureStandardInputStream) != 0)
        {
            if (dup2(stdinPipe[0], 0) == -1)
            {
                goto HandleError;
            }
            //  Close unused end of pipe
            if (close(stdinPipe[1]) == -1)
            {
                if (errno != EINTR)
                {
                    goto HandleError;
                }
            }
        }
        if ((captureStandardStreams & CaptureStandardOutputStream) != 0)
        {
            if (dup2(stdoutPipe[1], 1) == -1)
            {
                goto HandleError;
            }
            //  Close unused end of pipe
            if (close(stdoutPipe[0]) == -1)
            {
                if (errno != EINTR)
                {
                    goto HandleError;
                }
            }
        }
        if ((captureStandardStreams & CaptureStandardErrorStream) != 0)
        {
            if (dup2(stderrPipe[1], 2) == -1)
            {
                goto HandleError;
            }
            //  Close unused end of pipe
            if (close(stderrPipe[0]) == -1)
            {
                if (errno != EINTR)
                {
                    goto HandleError;
                }
            }
        }
        execve(argv[0], argv.getData(), envp.getData());
        //  If we are here, an error has occurred. Return it to the parent process
HandleError:
        ::exit(ExitCode_Panic);
    }
    else
    {   //  This is a parent process.
        delete [] pCommandLine;
        delete [] pEnvironmentBlock;
        _pid = childPid;
        //  Create necessary streams for talking to the child
        if ((captureStandardStreams & CaptureStandardInputStream) != 0)
        {
            _pStdin = new _ProcessInputStream(stdinPipe[1]);
            //  Close unused end of pipe
            if (close(stdinPipe[0]) == -1)
            {
                switch (errno)
                {
                    case EBADF:
                        throw ProgramError();
                    case EIO:
                        throw SystemError();
                    case EINTR: //  Maybe should retry
                        break;
                    default:
                        throw SystemError();
                }
            }
        }
        if ((captureStandardStreams & CaptureStandardOutputStream) != 0)
        {
            _pStdout = new _ProcessOutputStream(childPid, stdoutPipe[0]);
            //  Close unused end of pipe
            if (close(stdoutPipe[1]) == -1)
            {
                switch (errno)
                {
                    case EBADF:
                        throw ProgramError();
                    case EIO:
                        throw SystemError();
                    case EINTR: //  Maybe should retry
                        break;
                    default:
                        throw SystemError();
                }
            }
        }
        if ((captureStandardStreams & CaptureStandardErrorStream) != 0)
        {
            _pStderr = new _ProcessOutputStream(childPid, stderrPipe[0]);
            // close unused end of pipe
            if (close(stderrPipe[1]) == -1)
            {
                switch (errno)
                {
                    case EBADF:
                        throw ProgramError();
                    case EIO:
                        throw SystemError();
                    case EINTR: //  Maybe should retry
                        break;
                    default:
                        throw SystemError();
                }
            }
        }
        //  Record the moment when child process was started
        _elapsedTimeStart = _elapsedTimeEnd = TimeSpan::getTimeSpanSinceReferencePoint();
        //  Done
        _state = State_Active;
        return StartOutcome_Success;
    }
#else
    #error Unsupported platform
#endif
}

Process::GetExitCodeOutcome Process::getExitCode(int & exitCode)
{
    if (_isCurrent)
    {
        return GetExitCodeOutcome_StillRunning;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    if (_hProcess == NULL)
    {
        return GetExitCodeOutcome_InvalidState;
    }
    DWORD nativeExitCode;
    if (!GetExitCodeProcess((HANDLE)_hProcess, &nativeExitCode))
    {
        return GetExitCodeOutcome_NotAvailable;
    }
    if (nativeExitCode == STILL_ACTIVE)
    {
        return GetExitCodeOutcome_StillRunning;
    }
    else
    {
        exitCode = static_cast<int>(nativeExitCode);
        return GetExitCodeOutcome_Success;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (_pid == -1)
    {
        return GetExitCodeOutcome_InvalidState;
    }
    if (_state == State_Finished)
    {   //  here if process ended in getState()
        exitCode = _exitCode;
        return _exitState;
    }
    //  Update process' state - we need getExitCode() to be nonblocking
    switch (Process::getState())
    {
        case State_Initializing:
            return GetExitCodeOutcome_InvalidState;
        case State_Active:
            return GetExitCodeOutcome_StillRunning;
        case State_Finished:
            break;
        default:
            require(0);
    }
    //  If process running, wait until it ends
    int status;
    struct rusage ru;
    pid_t pt = wait4(_pid, &status, 0, &ru);
    if (pt == 0)
    {   //  state unchanged
        require(0);     //  can't happen without WNOHANG option!
        return GetExitCodeOutcome_StillRunning;
    }
    else if (pt < 0)
    {   //  error 
        switch (errno)
        {
            case ECHILD:    //  child no longer exists
                exitCode = ExitCode_Failure;
                return GetExitCodeOutcome_InvalidState;
            default:
                throw SystemError();
        }
    } 
    else
    {
        require(pt == _pid);
    }

    _systemTime = TimeSpan((((ru.ru_stime.tv_sec) * CARL_UINT64_CONSTANT(1000000)) +
                            ru.ru_stime.tv_usec) * CARL_UINT64_CONSTANT(1000));
    _userTime = TimeSpan((((ru.ru_utime.tv_sec) * CARL_UINT64_CONSTANT(1000000)) +
                          ru.ru_utime.tv_usec) * CARL_UINT64_CONSTANT(1000));
    _elapsedTimeEnd = TimeSpan::getTimeSpanSinceReferencePoint();

    if (WIFEXITED(status))
    {
        exitCode = WEXITSTATUS(status);
        return GetExitCodeOutcome_Success;
    }
    else if (WIFSIGNALED(status))
    {
        exitCode = ExitCode_Failure;
        return GetExitCodeOutcome_Success;
    }
    else
    {
        return GetExitCodeOutcome_StillRunning;
    }
#else
    #error Unsupported platform
#endif
}

IProcessInput * Process::getStdinStream()
{
    return _pStdin;
}

IProcessOutput * Process::getStdoutStream()
{
    return _pStdout;
}

IProcessOutput * Process::getStderrStream()
{
    return _pStderr;
}

TimeSpan Process::getElapsedTime() const
{
    //  Validate state
    if (getState() == State_Initializing)
    {   //  Process is just preparing to run - no time taken yet
        return TimeSpan(0);
    }

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    //  Win32 implementation

    //  Get process timing data
    if (getState() == State_Finished)
    {   //  The process has finished.
        FILETIME creationTime;
        FILETIME exitTime;
        FILETIME systemTime;
        FILETIME userTime;
        ULARGE_INTEGER creationTimeConverted;
        ULARGE_INTEGER exitTimeConverted;

        if (GetProcessTimes(reinterpret_cast<HANDLE>(_hProcess), &creationTime, &exitTime, &systemTime, &userTime))
        {
            creationTimeConverted.HighPart = creationTime.dwHighDateTime;
            creationTimeConverted.LowPart = creationTime.dwLowDateTime;
            exitTimeConverted.HighPart = exitTime.dwHighDateTime;
            exitTimeConverted.LowPart = exitTime.dwLowDateTime;
            return TimeSpan((exitTimeConverted.QuadPart - creationTimeConverted.QuadPart) * 100);
        }
    }
    else
    {   //  The process is still active
        FILETIME currentTime;
        FILETIME creationTime;
        FILETIME exitTime;
        FILETIME systemTime;
        FILETIME userTime;
        ULARGE_INTEGER creationTimeConverted;
        ULARGE_INTEGER currentTimeConverted;

        if (GetProcessTimes(reinterpret_cast<HANDLE>(_hProcess), &creationTime, &exitTime, &systemTime, &userTime))
        {
            GetSystemTimeAsFileTime(&currentTime);
            creationTimeConverted.HighPart = creationTime.dwHighDateTime;
            creationTimeConverted.LowPart = creationTime.dwLowDateTime;
            currentTimeConverted.HighPart = currentTime.dwHighDateTime;
            currentTimeConverted.LowPart = currentTime.dwLowDateTime;
            return TimeSpan((currentTimeConverted.QuadPart - creationTimeConverted.QuadPart) * 100);
        }
    }
    //  Getting process times has failed
    return TimeSpan();
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (_pid == getpid())
    {   //  Querying current process
        return TimeSpan::getTimeSpanSinceReferencePoint() - TimeSpan(Application::_startTimeNanosecondsFromReferencePoint);
    }
    else
    {   //  Querying child process. If it's still active, elapsed time still ticks
        if (getState() == State_Active)
        {
            _elapsedTimeEnd = TimeSpan::getTimeSpanSinceReferencePoint();
        }
        return _elapsedTimeEnd - _elapsedTimeStart;
    }
#else
    #error Unsupported host platform
#endif
}

TimeSpan Process::getSystemTime() const
{
    if (getState() == State_Initializing)
    {   //  Process is just preparing to run - no time taken yet
        return TimeSpan(0);
    }

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    //  Win32 implementation

    //  Get process timing data
    bool processTimesAvailable;
    FILETIME creationTime;
    FILETIME exitTime;
    FILETIME systemTime;
    FILETIME userTime;

    processTimesAvailable = GetProcessTimes(reinterpret_cast<HANDLE>(_hProcess), &creationTime, &exitTime, &systemTime, &userTime) != 0;

    //  Calculate system time
    if (processTimesAvailable)
    {
        ULARGE_INTEGER systemTimeConverted;

        systemTimeConverted.HighPart = systemTime.dwHighDateTime;
        systemTimeConverted.LowPart = systemTime.dwLowDateTime;
        return TimeSpan(systemTimeConverted.QuadPart * 100);
    }
    else
    {   //  No way to tell
        return TimeSpan();
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (_pid == getpid())
    {   //  Querying current process
        struct tms t;

        if (times(&t) == (clock_t)(-1))
        {   //  Error
            return TimeSpan();
        }
        else
        {   //  Process times retrieved
            return TimeSpan(static_cast<Int64>((static_cast<double>(t.tms_stime) * 1.0E9) / (double)sysconf(_SC_CLK_TCK)));
        }
    }
    else
    {   //  Querying child process
        getState();
        return _systemTime;
    }
#else   //  Host selection
    #error Unsupported host platform
#endif  //  Host selection
}

TimeSpan Process::getUserTime() const
{
    if (getState() == State_Initializing)
    {   //  Process is just preparing to run - no time taken yet
        return TimeSpan(0);
    }

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    //  Win32 implementation

    //  Get process timing data
    bool processTimesAvailable;
    FILETIME creationTime;
    FILETIME exitTime;
    FILETIME systemTime;
    FILETIME userTime;

    processTimesAvailable = GetProcessTimes(reinterpret_cast<HANDLE>(_hProcess), &creationTime, &exitTime, &systemTime, &userTime) != 0;

    //  Calculate user time
    if (processTimesAvailable)
    {   //  That's the best we can really do
        ULARGE_INTEGER userTimeConverted;

        userTimeConverted.HighPart = userTime.dwHighDateTime;
        userTimeConverted.LowPart = userTime.dwLowDateTime;
        return TimeSpan(userTimeConverted.QuadPart * 100);
    }
    else
    {   //  No way to tell
        return TimeSpan();
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (_pid == getpid())
    {   //  Querying current process
        struct tms t;

        if (times(&t) == (clock_t)(-1))
        {   //  Error
            return TimeSpan();
        }
        else
        {   //  Process times retrieved
            return TimeSpan(static_cast<Int64>((static_cast<double>(t.tms_utime) * 1.0E9) / (double)sysconf(_SC_CLK_TCK)));
        }
    }
    else
    {   //  Querying child process
        getState();
        return _userTime;
    }
#else   //  Host selection
    #error Unsupported host platform
#endif  //  Host selection
}

//////////
//  Implementation helpers
CriticalSection * Process::_getCurrentProcessGuard()
{
    static CriticalSection _currentProcessGuard;
    return &_currentProcessGuard;
}

//  End of Process.cpp
