//
//  RandomAccessFile.cpp - Implementation of the carl::RandomAccessFile class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/07/28 10:40:28 $
//  $Revision: 1.20 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <io.h>
    #include <fcntl.h>
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <sys/locking.h>
    #include <errno.h>
    #include <limits.h>
    #include <stdio.h>
    #include <share.h>
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)
    #include <fcntl.h>
    #include <errno.h>
    #include <unistd.h>
    #include <limits.h>
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC)
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <fcntl.h>
    #include <errno.h>
    #include <unistd.h>
    #include <limits.h>
#else
    #error Unsupported platform
#endif

//////////
//  Construction/destruction
RandomAccessFile::RandomAccessFile(const char * pFileName, OpenMode openMode, AccessMode accessMode) throw(IOException)
{
    require(pFileName != NULL);
    _openFile(String(pFileName), openMode, accessMode);
}

RandomAccessFile::RandomAccessFile(const Char * pFileName, OpenMode openMode, AccessMode accessMode) throw(IOException)
{
    require(pFileName != NULL);

    _openFile(String(pFileName), openMode, accessMode);
}

RandomAccessFile::RandomAccessFile(const String & fileName, OpenMode openMode, AccessMode accessMode) throw(IOException)
{
    _openFile(fileName, openMode, accessMode);
}

RandomAccessFile::RandomAccessFile(const PathName & pathName, OpenMode openMode, AccessMode accessMode) throw(IOException)
{
    _openFile(pathName.toString(), openMode, accessMode);
}

RandomAccessFile::RandomAccessFile(PathName & fileName) throw(IOException)
{
    PathName tempDirectory = FileSystem::getTempDirectory();
    if (tempDirectory.isEmpty())
    {
        throw IOException();
    }
    for (int attempt = 0; attempt < 100; attempt++)
    {
        //  Try to generate temp file name & open the file
        PathName tempFileName = FileSystem::getTempFileName(tempDirectory);
        if (!tempFileName.isEmpty())
        {   //  Temp file name generated - try opening the file
            try
            {
                _openFile(tempFileName.toString(), OpenMode_CreateNew, AccessMode_ReadWrite);
                fileName = tempFileName;
                return;
            }
            catch (IOException)
            {   //  Failed to create a temporary file
            }
        }
    }
    throw IOException();
}

//////////
//  Operations
void RandomAccessFile::close()
{
    if (_handle != -1)
    {
        int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        _close(_handle);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        ::close(_handle);
#else
    #error Unsupported platform
#endif
        errno = oldErrno;
    }
}

UInt64 RandomAccessFile::getSize() const throw(IOException)
{
    require(isOpen());

    int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Int64 curPos = _telli64(_handle);
    Int64 endPos = _lseeki64(_handle, 0, SEEK_END);
    Int64 newPos = _lseeki64(_handle, curPos, SEEK_SET);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    long curPos = lseek(_handle, 0, SEEK_CUR);
    Int64 endPos = lseek(_handle, 0, SEEK_END);
    Int64 newPos = lseek(_handle, curPos, SEEK_SET);
#else
    #error Unsupported platform
#endif
    errno = oldErrno;
    if (curPos == -1 || endPos == -1 || newPos == -1)
    {
        throw IOException();
    }
    return static_cast<UInt64>(endPos);
}

UInt64 RandomAccessFile::getCurrentPosition() const throw(IOException)
{
    require(isOpen());

    int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Int64 curPos = _telli64(_handle);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    long curPos = lseek(_handle, 0, SEEK_CUR);
#else
    #error Unsupported platform
#endif
    errno = oldErrno;
    if (curPos == -1)
    {
        throw IOException();
    }
    return static_cast<UInt64>(curPos);
}

void RandomAccessFile::setCurrentPosition(SeekOrigin seekOrigin, Int64 offset) throw(IOException)
{
    require(isOpen());

#if (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
    (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if ((long)offset != offset)
    {
        throw IOException();
    }
#endif
    int oldErrno = errno;
    Int64 newPos = -1;
    switch (seekOrigin)
    {
        case SeekOrigin_Start:
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
            newPos = _lseeki64(_handle, offset, SEEK_SET);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
            newPos = lseek(_handle, (long)offset, SEEK_SET);
#else
    #error Unsupported platform
#endif
            break;
        case SeekOrigin_End:
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
            newPos = _lseeki64(_handle, offset, SEEK_END);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
            newPos = lseek(_handle, (long)offset, SEEK_END);
#else
    #error Unsupported platform
#endif
            break;
        case SeekOrigin_Current:
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
            newPos = _lseeki64(_handle, offset, SEEK_CUR);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
            newPos = lseek(_handle, (long)offset, SEEK_CUR);
#else
    #error Unsupported platform
#endif
            break;
        default:
            require(false);
    }
    errno = oldErrno;
    if (newPos == -1)
    {
        throw IOException();
    }
}

void RandomAccessFile::lock(UInt64 start, UInt64 length) throw(IOException)
{
    require(isOpen());
    require(length <= LONG_MAX);

    int oldErrno = errno;
    UInt64 currentPosition = getCurrentPosition();
    setCurrentPosition(SeekOrigin_Start, start);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    int lockingResult = _locking(_handle, _LK_LOCK, static_cast<long>(length));
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    struct flock lockData;
    lockData.l_type = F_WRLCK;
    lockData.l_whence = SEEK_CUR;
    lockData.l_start = 0;
    lockData.l_len = static_cast<long>(length);
    lockData.l_pid = getpid();
    int lockingResult = fcntl(_handle, F_SETLK, &lockData);
#else
    #error Unsupported platform
#endif
    errno = oldErrno;
    setCurrentPosition(SeekOrigin_Start, currentPosition);
    if (lockingResult != 0)
    {
        throw IOException();
    }
}

void RandomAccessFile::unlock(UInt64 start, UInt64 length) throw(IOException)
{
    require(isOpen());
    require(length <= LONG_MAX);

    int oldErrno = errno;
    UInt64 currentPosition = getCurrentPosition();
    setCurrentPosition(SeekOrigin_Start, start);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    int lockingResult = _locking(_handle, _LK_UNLCK , static_cast<long>(length));
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    struct flock lockData;
    lockData.l_type = F_UNLCK;
    lockData.l_whence = SEEK_CUR;
    lockData.l_start = 0;
    lockData.l_len = static_cast<long>(length);
    lockData.l_pid = getpid();
    int lockingResult = fcntl(_handle, F_SETLK, &lockData);
#else
    #error Unsupported platform
#endif
    errno = oldErrno;
    setCurrentPosition(SeekOrigin_Start, currentPosition);
    if (lockingResult != 0)
    {
        throw IOException();
    }
}

//////////
//  IByteInput
UIntSize RandomAccessFile::available() throw(IOException)
{
    require(isOpen());

    int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Int64 currentPosition = _telli64(_handle);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Int64 currentPosition = lseek(_handle, 0, SEEK_CUR);
#else
    #error Unsupported platform
#endif
    if (currentPosition == -1)
    {
        errno = oldErrno;
        throw IOException();
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Int64 eofPosition = _lseeki64(_handle, 0, SEEK_END);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Int64 eofPosition = lseek(_handle, 0, SEEK_END);
#else
    #error Unsupported platform
#endif
    if (eofPosition == -1)
    {
        errno = oldErrno;
        throw IOException();
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    if (_lseeki64(_handle, currentPosition, SEEK_SET) != currentPosition)
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (lseek(_handle, currentPosition, SEEK_SET) != currentPosition)
#else
    #error Unsupported platform
#endif
    {
        errno = oldErrno;
        throw IOException();
    }
    UInt64 bytesUntilEof = eofPosition - currentPosition;
    return (bytesUntilEof > UINT_MAX) ? (UINT_MAX) : (static_cast<UIntSize>(bytesUntilEof));
}

int RandomAccessFile::read() throw(IOException)
{
    UInt8 b;
    if (read(&b, 1) == 0)
    {
        return -1;
    }
    return b;
}

UIntSize RandomAccessFile::read(void * pBuffer, UIntSize bytesToRead) throw(IOException)
{
    require(isOpen());

    //  Read in 32K chunks
    int oldErrno = errno;
    UInt8 * pBytes = static_cast<UInt8*>(pBuffer);
    UIntSize bytesRead = 0;
    while (bytesToRead != 0)
    {
        UIntSize chunkSize = (bytesToRead > 32768) ? (32768) : (bytesToRead);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        int chunkBytes = ::_read(_handle, pBytes, chunkSize);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        int chunkBytes = ::read(_handle, pBytes, chunkSize);
#else
    #error Unsupported platform
#endif
        if (chunkBytes == -1)
        {
            errno = oldErrno;
            throw IOException();
        }
        bytesToRead -= chunkSize;
        bytesRead += chunkBytes;
        pBytes += chunkBytes;
        if (static_cast<UIntSize>(chunkBytes) != chunkSize)
        {
            break;
        }
    }
    errno = oldErrno;
    return bytesRead;
}

UIntSize RandomAccessFile::skip(UIntSize bytesToSkip) throw(IOException)
{
    require(isOpen());

    UIntSize bytesSkipped = 0;
    while (bytesToSkip != 0)
    {
        UInt8 chunk[4096];
        UIntSize chunkSize = (bytesToSkip > sizeof(chunk)) ? (sizeof(chunk)) : (bytesToSkip);
        UIntSize chunkSkipped = read(chunk, chunkSize);
        bytesSkipped += chunkSkipped;
        bytesToSkip -= chunkSize;
        if (chunkSkipped != chunkSize)
        {
            break;
        }
    }
    return bytesSkipped;
}

//////////
//  IByteOutput
void RandomAccessFile::flush() throw(IOException)
{
    require(isOpen());

    int oldErrno = errno;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    if (_commit(_handle) != 0)
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (fdatasync(_handle) != 0)
#else
    #error Unsupported platform
#endif
    {
        errno = oldErrno;
        throw IOException();
    }
    errno = oldErrno;
}

void RandomAccessFile::write(UInt8 byte) throw(IOException)
{
    write(&byte, 1);
}

void RandomAccessFile::write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException)
{
    require(isOpen());

    //  Write in 32K chunks
    int oldErrno = errno;
    const UInt8 * pBytes = static_cast<const UInt8*>(pBuffer);
    UIntSize bytesWritten = 0;
    while (bytesToWrite != 0)
    {
        UIntSize chunkSize = (bytesToWrite > 32768) ? (32768) : (bytesToWrite);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        int chunkBytes = _write(_handle, pBytes, chunkSize);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        int chunkBytes = ::write(_handle, pBytes, chunkSize);
#else
    #error Unsupported platform
#endif
        if (chunkBytes < 0 || static_cast<UIntSize>(chunkBytes) != chunkSize)
        {
            errno = oldErrno;
            throw IOException();
        }
        bytesToWrite -= chunkSize;
        bytesWritten += chunkBytes;
        pBytes += chunkBytes;
    }
    errno = oldErrno;
}

//////////
//  IDataInput
bool RandomAccessFile::readBool() throw(IOException)
{
    require(isOpen());

    UInt8 b;
    if (read(&b, 1) != 1)
    {
        throw EOFException();
    }
    switch (b)
    {
        case 0:
            return false;
        case 1:
            return true;
        default:
            throw InvalidDataException();
    }
}

UInt8 RandomAccessFile::readUInt8() throw(IOException)
{
    require(isOpen());

    UInt8 b;
    if (read(&b, 1) != 1)
    {
        throw EOFException();
    }
    return b;
}

UInt16 RandomAccessFile::readUInt16() throw(IOException)
{
    require(isOpen());

    UInt8 b[2];
    if (read(&b, 2) != 2)
    {
        throw EOFException();
    }
    return Memory::loadUnalignedUInt16(b, ByteOrder_BigEndian);
}

UInt32 RandomAccessFile::readUInt32() throw(IOException)
{
    require(isOpen());

    UInt8 b[4];
    if (read(&b, 4) != 4)
    {
        throw EOFException();
    }
    return Memory::loadUnalignedUInt32(b, ByteOrder_BigEndian);
}

UInt64 RandomAccessFile::readUInt64() throw(IOException)
{
    require(isOpen());

    UInt8 b[8];
    if (read(&b, 8) != 8)
    {
        throw EOFException();
    }
    return Memory::loadUnalignedUInt64(b, ByteOrder_BigEndian);
}

Int8 RandomAccessFile::readInt8() throw(IOException)
{
    require(isOpen());

    Int8 b;
    if (read(&b, 1) != 1)
    {
        throw EOFException();
    }
    return b;
}

Int16 RandomAccessFile::readInt16() throw(IOException)
{
    require(isOpen());

    UInt8 b[2];
    if (read(&b, 2) != 2)
    {
        throw EOFException();
    }
    return Memory::loadUnalignedInt16(b, ByteOrder_BigEndian);
}

Int32 RandomAccessFile::readInt32() throw(IOException)
{
    require(isOpen());

    UInt8 b[4];
    if (read(&b, 4) != 4)
    {
        throw EOFException();
    }
    return Memory::loadUnalignedInt32(b, ByteOrder_BigEndian);
}

Int64 RandomAccessFile::readInt64() throw(IOException)
{
    require(isOpen());

    UInt8 b[8];
    if (read(&b, 8) != 8)
    {
        throw EOFException();
    }
    return Memory::loadUnalignedInt64(b, ByteOrder_BigEndian);
}

Real32 RandomAccessFile::readReal32() throw(IOException)
{
    require(isOpen());

    UInt8 b[4];
    if (read(&b, 4) != 4)
    {
        throw EOFException();
    }
    UInt32 valueBits = Memory::loadUnalignedUInt32(b, ByteOrder_BigEndian);
    return *(reinterpret_cast<Real32*>(&valueBits));
}

Real64 RandomAccessFile::readReal64() throw(IOException)
{
    require(isOpen());

    UInt8 b[8];
    if (read(&b, 8) != 8)
    {
        throw EOFException();
    }
    UInt64 valueBits = Memory::loadUnalignedUInt64(b, ByteOrder_BigEndian);
    return *(reinterpret_cast<Real64*>(&valueBits));
}

Char RandomAccessFile::readChar() throw(IOException)
{
    require(isOpen());

    UInt8 b[4];
    if (read(&b, 4) != 4)
    {
        throw EOFException();
    }
    return Memory::loadUnalignedUInt32(b, ByteOrder_BigEndian);
}

//////////
//  IDataOutput
void RandomAccessFile::writeBool(bool value) throw(IOException)
{
    require(isOpen());

    UInt8 b = (value) ? (1) : (0);
    write(&b, 1);
}

void RandomAccessFile::writeUInt8(UInt8 value) throw(IOException)
{
    require(isOpen());

    write(&value, 1);
}

void RandomAccessFile::writeUInt16(UInt16 value) throw(IOException)
{
    require(isOpen());

    UInt8 b[2];
    Memory::storeUnalignedUInt16(value, &b, ByteOrder_BigEndian);
    write(b, 2);
}

void RandomAccessFile::writeUInt32(UInt32 value) throw(IOException)
{
    require(isOpen());

    UInt8 b[4];
    Memory::storeUnalignedUInt32(value, &b, ByteOrder_BigEndian);
    write(b, 4);
}

void RandomAccessFile::writeUInt64(UInt64 value) throw(IOException)
{
    require(isOpen());

    UInt8 b[8];
    Memory::storeUnalignedUInt64(value, &b, ByteOrder_BigEndian);
    write(b, 8);
}

void RandomAccessFile::writeInt8(Int8 value) throw(IOException)
{
    require(isOpen());

    write(&value, 1);
}

void RandomAccessFile::writeInt16(Int16 value) throw(IOException)
{
    require(isOpen());

    UInt8 b[2];
    Memory::storeUnalignedInt16(value, &b, ByteOrder_BigEndian);
    write(b, 2);
}

void RandomAccessFile::writeInt32(Int32 value) throw(IOException)
{
    require(isOpen());

    UInt8 b[4];
    Memory::storeUnalignedInt32(value, &b, ByteOrder_BigEndian);
    write(b, 4);
}

void RandomAccessFile::writeInt64(Int64 value) throw(IOException)
{
    require(isOpen());

    UInt8 b[8];
    Memory::storeUnalignedInt64(value, &b, ByteOrder_BigEndian);
    write(b, 8);
}

void RandomAccessFile::writeReal32(Real32 value) throw(IOException)
{
    require(isOpen());

    UInt8 b[4];
    Memory::storeUnalignedUInt32(*(reinterpret_cast<UInt32*>(&value)), &b, ByteOrder_BigEndian);
    write(b, 4);
}

void RandomAccessFile::writeReal64(Real64 value) throw(IOException)
{
    require(isOpen());

    UInt8 b[8];
    Memory::storeUnalignedUInt64(*(reinterpret_cast<UInt64*>(&value)), &b, ByteOrder_BigEndian);
    write(b, 8);
}

void RandomAccessFile::writeChar(Char value) throw(IOException)
{
    require(isOpen());

    UInt8 b[4];
    Memory::storeUnalignedUInt32(value, &b, ByteOrder_BigEndian);
    write(b, 4);
}

//////////
//  Implementation
void RandomAccessFile::_openFile(const String & fileName, OpenMode openMode, AccessMode accessMode) throw(IOException)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<wchar_t> fileNameString;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<char> fileNameString;
#else
    #error Unsupported platform
#endif

    //  Store path name string
    fileNameString.setSize(fileName.length() + 1);
    if (!fileName._store(fileNameString.getData(), fileNameString.size()))
    {
        throw FileNotFoundException(fileName);
    }

    //  Open the file
    int oldErrno = errno;
    switch (openMode)
    {
        case OpenMode_OpenExisting:
            if (accessMode == AccessMode_ReadOnly)
            {
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && (defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003))
                _handle = _wsopen(fileNameString.getData(), _O_BINARY | _O_RDONLY | _O_RANDOM, _SH_DENYWR, _S_IREAD);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
                errno = _wsopen_s(&_handle, fileNameString.getData(), _O_BINARY | _O_RDONLY | _O_RANDOM, _SH_DENYWR, _S_IREAD);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
                _handle = open(fileNameString.getData(), O_RDONLY);
#else
    #error Unsupported platform
#endif
            }
            else
            {
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && (defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003))
                _handle = _wsopen(fileNameString.getData(), _O_BINARY | _O_RDWR | _O_RANDOM, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
                errno = _wsopen_s(&_handle, fileNameString.getData(), _O_BINARY | _O_RDWR | _O_RANDOM, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
                _handle = open(fileNameString.getData(), O_RDWR);
#else
    #error Unsupported platform
#endif
            }
            break;
        case OpenMode_CreateNew:
            require(accessMode == AccessMode_ReadWrite);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && (defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003))
            _handle = _wsopen(fileNameString.getData(), _O_BINARY | _O_RDWR | _O_RANDOM | _O_CREAT | _O_EXCL, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
            errno = _wsopen_s(&_handle, fileNameString.getData(), _O_BINARY | _O_RDWR | _O_RANDOM | _O_CREAT | _O_EXCL, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
            _handle = open(fileNameString.getData(), O_RDWR | O_CREAT | O_EXCL, 0666);
#else
    #error Unsupported platform
#endif
            break;
        case OpenMode_TruncateExisting:
            require(accessMode == AccessMode_ReadWrite);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && (defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003))
            _handle = _wsopen(fileNameString.getData(), _O_BINARY | _O_RDWR | _O_RANDOM | _O_TRUNC, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
            errno = _wsopen_s(&_handle, fileNameString.getData(), _O_BINARY | _O_RDWR | _O_RANDOM | _O_TRUNC, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
            if (access(fileNameString.getData(), F_OK) != 0)
            {   //  File does not currently exist
                _handle = -1;
                errno = ENOENT;
            }
            else
            {
                _handle = open(fileNameString.getData(), O_RDWR | O_TRUNC);
            }
#else
    #error Unsupported platform
#endif
            break;
        case OpenMode_CreateOrOpen:
            require(accessMode == AccessMode_ReadWrite);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && (defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003))
            _handle = _wsopen(fileNameString.getData(), _O_BINARY | _O_RDWR | _O_RANDOM | _O_CREAT, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
            errno = _wsopen_s(&_handle, fileNameString.getData(), _O_BINARY | _O_RDWR | _O_RANDOM | _O_CREAT, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
            _handle = open(fileNameString.getData(), O_RDWR | O_CREAT, 0666);
#else
    #error Unsupported platform
#endif
            break;
        case OpenMode_CreateAlways:
            require(accessMode == AccessMode_ReadWrite);
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && (defined(CARL_TOOLCHAIN_MSVC6) || defined(CARL_TOOLCHAIN_MSVCNET2003))
            _handle = _wsopen(fileNameString.getData(), _O_BINARY | _O_RDWR | _O_RANDOM | _O_CREAT | _O_TRUNC, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
            errno = _wsopen_s(&_handle, fileNameString.getData(), _O_BINARY | _O_RDWR | _O_RANDOM | _O_CREAT | _O_TRUNC, _SH_DENYRW, _S_IREAD | _S_IWRITE);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
            _handle = open(fileNameString.getData(), O_RDWR | O_CREAT | O_TRUNC, 0666);
#else
    #error Unsupported platform
#endif
            break;
        default:
            require(false);
    }
    if (_handle == -1)
    {
        switch (errno)
        {
            case ENOENT:
                errno = oldErrno;
                throw FileNotFoundException(fileName);
            default:
                errno = oldErrno;
                throw IOException();
        }
    }
    errno = oldErrno;
}

//  End of RandomAccessFile.cpp
