//
//  RsaCipher.cpp - Implementation of the carl::_impl::_RsaCipher class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:54 $
//  $Revision: 1.4 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "EncryptionImpl.hpp"

//////////
//  Singleton
_impl::_RsaCipher::_RsaCipher()
{
}

_impl::_RsaCipher::~_RsaCipher()
{
}

_impl::_RsaCipher * _impl::_RsaCipher::getInstance()
{
    static _RsaCipher instance;
    return &instance;
}

//////////
//  BlockCipher
Array<UIntSize> _impl::_RsaCipher::getSupportedBlockLenghts() const
{
    Array<UIntSize> result;
    result.append(1024);
    return result;
}

//////////
//  AsymmetricCipher
Array<Pair<UIntSize,UIntSize> > _impl::_RsaCipher::getSupportedKeyPairLenghts() const
{
    Array<Pair<UIntSize,UIntSize> > result;
    for (UIntSize i = 128; i <= 4096; i *= 2)
    {
        result.append(Pair<UIntSize,UIntSize>(i, i));
    }
    return result;
}

KeyPair _impl::_RsaCipher::generateKeyPair(UInt64 seed, UIntSize publicKeyLengthBits, UIntSize privateKeyLengthBits) const
{
    //  Validate key lengths
    Array<Pair<UIntSize,UIntSize> > supportedKeyPairLengths = getSupportedKeyPairLenghts();
    bool keyLengthsValid = false;
    for (UIntSize i = 0; i < supportedKeyPairLengths.size(); i++)
    {
        if (supportedKeyPairLengths[i].first() == publicKeyLengthBits &&
            supportedKeyPairLengths[i].second() == privateKeyLengthBits)
        {
            keyLengthsValid = true;
            break;
        }
    }
    if (!keyLengthsValid)
    {
        throw InvalidKeyLengthError();
    }
    //  Generate key components
    Random r(static_cast<UInt32>(seed) ^ static_cast<UInt32>(seed >> 32));
    BinaryInteger p, q;
    for (; ; )
    {
        p = BinaryInteger::generateProbablePrime(publicKeyLengthBits / 2, 10, r);
        q = BinaryInteger::generateProbablePrime(privateKeyLengthBits / 2, 10, r);
        if (p.gcd(q) == 1)
        {
            break;
        }
    }
    BinaryInteger n = p * q;
    BinaryInteger fn = (p - 1) * (q - 1);
    BinaryInteger e;
    for (; ; )
    {
        e = r.nextBinaryInteger(2, fn - 1);
        if (e.gcd(fn) == 1)
        {
            break;
        }
    }
    BinaryInteger d = e.modInverse(fn);
    require((d * e) % fn == 1);
    //  Create & return key pair
    Array<UInt8> publicKeyBits = _assembleKeyBits(n, e, publicKeyLengthBits);
    Array<UInt8> privateKeyBits = _assembleKeyBits(n, d, privateKeyLengthBits);
    PublicKey publicKey(publicKeyBits.getData(), publicKeyBits.size() * 8);
    PrivateKey privateKey(privateKeyBits.getData(), privateKeyBits.size() * 8);
    return KeyPair(publicKey, privateKey);
}

KeyPair _impl::_RsaCipher::generateRandomKeyPair(UIntSize publicKeyLengthBits, UIntSize privateKeyLengthBits) const
{
    Random r;
    return generateKeyPair(r.nextUInt64(), publicKeyLengthBits, privateKeyLengthBits);
}

//////////
//  Cipher
Array<String> _impl::_RsaCipher::getAllNames() const
{
    Array<String> result;
    result.append(getCanonicalName());
    return result;
}

String _impl::_RsaCipher::getCanonicalName() const
{
    return String("RSA");
}

String _impl::_RsaCipher::getPreferredName() const
{
    return getCanonicalName();
}

String _impl::_RsaCipher::getDisplayName() const
{
    return _EncryptionManagerImpl::getInstance()->getResourceFactory()->getString(getCanonicalName());
}

String _impl::_RsaCipher::getDisplayName(const Locale & forLocale) const
{
    return _EncryptionManagerImpl::getInstance()->getResourceFactory()->getString(getCanonicalName(), forLocale);
}

Encryptor * _impl::_RsaCipher::createEncryptor(const CryptoKey & key) const throw(KeyCorruptException)
{
    //  Break key into components
    BinaryInteger modulus, exponent;
    UIntSize keyLength;
    if (!_disassembleKeyBits(static_cast<const UInt8*>(key.getKeyBits()), key.getKeyLengthBits() / 8, modulus, exponent, keyLength))
    {
        throw KeyCorruptException();
    }
    //  Validate key length
    Array<Pair<UIntSize,UIntSize> > supportedKeyPairLengths = getSupportedKeyPairLenghts();
    bool keyLengthValid = false;
    for (UIntSize i = 0; i < supportedKeyPairLengths.size(); i++)
    {
        if (supportedKeyPairLengths[i].first() == keyLength ||
            supportedKeyPairLengths[i].second() == keyLength)
        {
            keyLengthValid = true;
            break;
        }
    }
    if (!keyLengthValid)
    {
        throw InvalidKeyLengthError();
    }
    //  Create & return encryptor
    return new _RsaEncryptor(modulus, exponent, ((modulus.getRequiredBits() - 2) / 8));
}

Decryptor * _impl::_RsaCipher::createDecryptor(const CryptoKey & key) const throw(KeyCorruptException)
{
    //  Break key into components
    BinaryInteger modulus, exponent;
    UIntSize keyLength;
    if (!_disassembleKeyBits(static_cast<const UInt8*>(key.getKeyBits()), key.getKeyLengthBits() / 8, modulus, exponent, keyLength))
    {
        throw KeyCorruptException();
    }
    //  Validate key length
    Array<Pair<UIntSize,UIntSize> > supportedKeyPairLengths = getSupportedKeyPairLenghts();
    bool keyLengthValid = false;
    for (UIntSize i = 0; i < supportedKeyPairLengths.size(); i++)
    {
        if (supportedKeyPairLengths[i].first() == keyLength ||
            supportedKeyPairLengths[i].second() == keyLength)
        {
            keyLengthValid = true;
            break;
        }
    }
    if (!keyLengthValid)
    {
        throw InvalidKeyLengthError();
    }
    //  Create & return encryptor
    return new _RsaDecryptor(modulus, exponent, ((modulus.getRequiredBits() - 2) / 8));
}

//////////
//  Implementation helpers
Array<UInt8> _impl::_RsaCipher::_assembleKeyBits(const BinaryInteger & modulus, const BinaryInteger & exponent, UIntSize keyLength)
{
    MemoryOutputStream mos;

    Leb128::writeBinaryInteger(mos, modulus);
    Leb128::writeBinaryInteger(mos, exponent);
    ULeb128::writeUIntSize(mos, keyLength);
    return mos.getContents();
}

bool _impl::_RsaCipher::_disassembleKeyBits(const UInt8 * pKeyBytes, UIntSize keyBytesCount, BinaryInteger & modulus, BinaryInteger & exponent, UIntSize & keyLength)
{
    require(pKeyBytes != NULL);

    MemoryInputStream mis(pKeyBytes, keyBytesCount, false);
    try
    {
        modulus = Leb128::readBinaryInteger(mis);
        exponent = Leb128::readBinaryInteger(mis);
        keyLength = ULeb128::readUIntSize(mis);
        return mis.available() == 0;
    }
    catch (IOException)
    {
        return false;
    }
}

//  End of RsaCipher.cpp
