//
//  SemaphoreImpl.cpp - Implementation of the carl::_impl::_SemaphoreImpl class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:54 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "SemaphoreImpl.hpp"

//////////
//  Operations
void _impl::_SemaphoreImpl::initialize(UIntSize initialCount)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _osSemaphore = CreateSemaphore(NULL, initialCount, initialCount, NULL);
    if (_osSemaphore == NULL)
    {
        throw SystemError();
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _currentCount = initialCount;
    _maximumCount = initialCount;
    _pGuard = new Mutex();
#else
    #error Unsupported platform
#endif
}

void _impl::_SemaphoreImpl::deinitialize()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    CloseHandle(_osSemaphore);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    delete _pGuard;
#else
    #error Unsupported platform
#endif
}

void _impl::_SemaphoreImpl::waitFor()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    DWORD oldLastError = GetLastError();
    DWORD waitResult = WaitForSingleObject(_osSemaphore, INFINITE);
    SetLastError(oldLastError);
    switch (waitResult)
    {
        case WAIT_ABANDONED:
        case WAIT_OBJECT_0:
            return;
        default:
            throw SystemError();
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    for (; ; )
    {
        {
            Lock lock(*_pGuard);
            if (_currentCount != 0)
            {
                _currentCount--;
                return;
            }
        }
        Thread::sleep(TimeSpan(0));
    }
#else
    #error Unsupported platform
#endif
}

bool _impl::_SemaphoreImpl::waitFor(const TimeSpan & timeout)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    TimeStamp waitStartTime = TimeStamp::now();
    TimeStamp waitEndTime = waitStartTime + timeout;
    for (; ; )
    {
        Int64 millisecondsToWait = timeout.getNanoseconds() / 1000000000;
        if (millisecondsToWait > 1000)
        {
            millisecondsToWait = 1000;
        }
        DWORD oldLastError = GetLastError();
        DWORD waitResult = WaitForSingleObject(_osSemaphore, static_cast<DWORD>(millisecondsToWait));
        SetLastError(oldLastError);
        switch (waitResult)
        {
            case WAIT_ABANDONED:
            case WAIT_OBJECT_0:
                return true;
            case WAIT_TIMEOUT:
                if (TimeStamp::now() >= waitEndTime)
                {   //  Waited for the whole timeout
                    return false;
                }
                //  Keep trying
                break;
            default:
                throw SystemError();
        }
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    TimeSpan giveUpTime = TimeSpan::getTimeSpanSinceReferencePoint() + timeout;
    for (; ; )
    {
        {
            Lock lock(*_pGuard);
            if (_currentCount != 0)
            {
                _currentCount--;
                return true;
            }
        }
        if (TimeSpan::getTimeSpanSinceReferencePoint() > giveUpTime)
        {
            return false;
        }
        Thread::sleep(TimeSpan(0));
    }
#else
    #error Unsupported platform
#endif
}

void _impl::_SemaphoreImpl::release()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    ReleaseSemaphore(_osSemaphore, 1, NULL);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Lock lock(*_pGuard);
    if (_currentCount < _maximumCount)
    {
        _currentCount++;
    }
#else
    #error Unsupported platform
#endif
}

//  End of SemaphoreImpl.cpp
