//
//  Settings.cpp - Implementation of the carl::Settings class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:55 $
//  $Revision: 1.12 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <windows.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include "EmulatedSettingsImpl.hpp"
#else
    #error Unsupported platform
#endif

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
namespace
{
    class _RegistryException {};

    HKEY _openRootKey(Settings * pSettings)
    {
        //  Choose either HKEY_CURRENT_USER or HKEY_LOCAL_MACHINE as an ultimate parent
        HKEY hUltimateParentKey = HKEY_LOCAL_MACHINE;
        if (pSettings == Settings::getShared())
        {
            hUltimateParentKey = HKEY_LOCAL_MACHINE;
        }
        else if (pSettings == Settings::getPrivate())
        {
            hUltimateParentKey = HKEY_CURRENT_USER;
        }
        else
        {
            require(false);
        }
        //  .../Software
        HKEY hSoftwareKey;
        DWORD softwareKeyDisposition;
        if (RegCreateKeyEx(hUltimateParentKey, "Software", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hSoftwareKey, &softwareKeyDisposition) != ERROR_SUCCESS)
        {
            throw _RegistryException();
        }
        //  .../Cybernetic Intelligence GmbH
        HKEY hCiKey;
        DWORD ciKeyDisposition;
        if (RegCreateKeyEx(hSoftwareKey, "Cybernetic Intelligence GmbH", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hCiKey, &ciKeyDisposition) != ERROR_SUCCESS)
        {
            RegCloseKey(hSoftwareKey);
            throw _RegistryException();
        }
        RegCloseKey(hSoftwareKey);
        //  .../CARL
        HKEY hCarlKey;
        DWORD carlKeyDisposition;
        if (RegCreateKeyEx(hCiKey, "CARL", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hCarlKey, &carlKeyDisposition) != ERROR_SUCCESS)
        {
            RegCloseKey(hCiKey);
            throw _RegistryException();
        }
        RegCloseKey(hCiKey);
        //  .../Settings
        HKEY hSettingsKey;
        DWORD settingsKeyDisposition;
        if (RegCreateKeyEx(hCarlKey, "Settings", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hSettingsKey, &settingsKeyDisposition) != ERROR_SUCCESS)
        {
            RegCloseKey(hCarlKey);
            throw _RegistryException();
        }
        RegCloseKey(hCarlKey);
        //  Done
        return hSettingsKey;
    }

    bool _deleteSubKey(HKEY hKey, const char * pSubKeyName)
    {
        //  Open the subkey that is to be deleted
        HKEY hSubKey;
        if (RegOpenKeyEx(hKey, pSubKeyName, 0, KEY_ALL_ACCESS, &hSubKey) != ERROR_SUCCESS)
        {   //  The subkey ALREADY does not exist!
            return true;
        }
        //  Delete all sub-subkeys
        for (; ; )
        {
            char subkeyName[256];
            DWORD subkeyNameSize = 255;
            FILETIME lastWriteTime;
            LONG enumResult = RegEnumKeyEx(hSubKey, 0, subkeyName, &subkeyNameSize, NULL, NULL, NULL, &lastWriteTime);
            if (enumResult == ERROR_NO_MORE_ITEMS)
            {
                break;
            }
            else if (enumResult != ERROR_SUCCESS)
            {
                RegCloseKey(hSubKey);
                throw _RegistryException();
            }
            if (!_deleteSubKey(hSubKey, subkeyName))
            {
                RegCloseKey(hSubKey);
                return false;
            }
        }
        RegCloseKey(hSubKey);
        //  Delete this subkey
        return RegDeleteKey(hKey, pSubKeyName) == ERROR_SUCCESS;
    }

    Array<String> _splitNodePath(const String & path)
    {
        Array<String> result;

        if (path != '/')
        {
            for (UIntSize scan = 0; scan < path.length(); )
            {
                UIntSize nextSeparatorIndex;
                String pathComponent;
                if (path.findFirst('/', nextSeparatorIndex, scan + 1))
                {   //  There IS a next path component
                    pathComponent = path.mid(scan + 1, nextSeparatorIndex - scan - 1);
                    scan = nextSeparatorIndex;
                }
                else
                {   //  This is the last path component
                    pathComponent = path.mid(scan + 1);
                    scan = path.length();
                }
                if (pathComponent.length() != 0)
                {
                    result.append(pathComponent);
                }
            }
        }
        return result;
    }

    //  CARL names of nodes and values are case-sensitive and can contain
    //  any ISO-10646 characters. Windows Registry is neither. Therefore, each
    //  CARL name must be translated before being sent to a Windows Registry service.
    //  Similarly, any string returned by the Windows Registry service must be
    //  translated back.
    char * _toAnsi(const String & s)
    {
        String result;
        for (UIntSize i = 0; i < s.length(); i++)
        {
            Char c = s[i];
            if (c >= 32 && c <= 127 && c != '/' && c != '\\')
            {   //  Printable character
                if (c == '_')
                {
                    result += "__";
                }
                else if (c >= 'A' && c <= 'Z')
                {
                    result += '_';
                    result += Iso10646::toLower(c);
                }
                else
                {
                    result += c;
                }
            }
            else
            {
                result += '_';
                result += NumberFormat::getInstance(Locale())->format(static_cast<UInt32>(c), IntegerFormatProperties("%08X"));
            }
        }
        //
        char * pChars = new char[result.length() + 1];
        for (UIntSize i = 0; i < result.length(); i++)
        {
            pChars[i] = (char)result[i];
        }
        pChars[result.length()] = 0;
        return pChars;
    }

    String _fromAnsi(const char * s)
    {
        UIntSize slength = static_cast<UIntSize>(strlen(s));
        String result;

        for (UIntSize i = 0; i < slength; )
        {
            if (i + 1 < slength && s[i] == '_' && s[i + 1] == '_')
            {
                i += 2;
                result += '_';
            }
            else if (i + 1 < slength && s[i] == '_' && s[i + 1] >= 'a' && s[i + 1] <= 'z')
            {
                result += Iso10646::toUpper(s[i + 1]);
                i += 2;
            }
            else if (i + 8 < slength && 
                     s[i] == '_' &&
                     Iso10646::isHexadecimalDigit(s[i + 1]) &&
                     Iso10646::isHexadecimalDigit(s[i + 2]) &&
                     Iso10646::isHexadecimalDigit(s[i + 3]) &&
                     Iso10646::isHexadecimalDigit(s[i + 4]) &&
                     Iso10646::isHexadecimalDigit(s[i + 5]) &&
                     Iso10646::isHexadecimalDigit(s[i + 6]) &&
                     Iso10646::isHexadecimalDigit(s[i + 7]) &&
                     Iso10646::isHexadecimalDigit(s[i + 8]))
            {
                UInt32 charCode = 0;
                i++;
                for (int numDigits = 0; numDigits < 8; numDigits++, i++)
                {
                    UInt32 digit = (Iso10646::isDecimalDigit(s[i])) ? (s[i] - '0') : (Iso10646::toUpper(s[i]) - 'A' + 10);
                    charCode = (charCode << 4) | digit;
                }
                result += static_cast<Char>(charCode);
            }
            else if (s[i] >= 'A' && s[i] <= 'Z')
            {
                result += Iso10646::toLower(s[i++]);
            }
            else
            {
                result += s[i++];
            }

        }
        return result;
    }
}
#endif

//////////
//  Construction/destruction - from CARL only
Settings::Settings()
{
    _virtualRoot = SettingsNode("//");
}

Settings::~Settings()
{
}

//////////
//  Operations
Settings * Settings::getShared()
{
    static Settings instance;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    instance._pEmulatedSettings = _impl::_EmulatedSettings::getSharedInstance();
#else
    #error Unsupported platform
#endif
    return &instance;
}

Settings * Settings::getPrivate()
{
    static Settings instance;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    instance._pEmulatedSettings = _impl::_EmulatedSettings::getPrivateInstance();
#else
    #error Unsupported platform
#endif
    return &instance;
}

bool Settings::nodeExists(const SettingsNode & node)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    try
    {
        HKEY hKey = _openRootKey(this);
        Array<String> nodePath = _splitNodePath(_constructAbsolutePath(node).toString());
        for (UIntSize i = 0; i < nodePath.size(); i++)
        {
            HKEY hSubKey;
            char * pSubKeyName = _toAnsi(nodePath[i]);
            if (RegOpenKeyEx(hKey, pSubKeyName, 0, KEY_ALL_ACCESS, &hSubKey) != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                delete [] pSubKeyName;
                return false;
            }
            delete [] pSubKeyName;
            RegCloseKey(hKey);
            hKey = hSubKey;
        }
        RegCloseKey(hKey);
        return true;
    }
    catch (_RegistryException)
    {
        return false;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return _pEmulatedSettings->nodeExists(_constructAbsolutePath(node));
#else
    #error Unsupported platform
#endif
}

bool Settings::createNode(const SettingsNode & node)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    try
    {
        HKEY hKey = _openRootKey(this);
        Array<String> nodePath = _splitNodePath(_constructAbsolutePath(node).toString());
        for (UIntSize i = 0; i < nodePath.size(); i++)
        {
            HKEY hSubKey;
            DWORD subKeyDisposition;
            char * pSubKeyName = _toAnsi(nodePath[i]);
            if (RegCreateKeyEx(hKey, pSubKeyName, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hSubKey, &subKeyDisposition) != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                delete [] pSubKeyName;
                return false;
            }
            delete [] pSubKeyName;
            RegCloseKey(hKey);
            hKey = hSubKey;
        }
        RegCloseKey(hKey);
        return true;
    }
    catch (_RegistryException)
    {
        return false;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return _pEmulatedSettings->createNode(_constructAbsolutePath(node));
#else
    #error Unsupported platform
#endif
}

bool Settings::destroyNode(const SettingsNode & node)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    try
    {
        Array<String> nodePath = _splitNodePath(_constructAbsolutePath(node).toString());
        if (nodePath.size() == 0)
        {   //  The root node can never be destroyed
            return false;
        }
        HKEY hKey = _openRootKey(this);
        for (UIntSize i = 0; i < nodePath.size() - 1; i++)
        {
            HKEY hSubKey;
            char * pSubKeyName = _toAnsi(nodePath[i]);
            if (RegOpenKeyEx(hKey, pSubKeyName, 0, KEY_ALL_ACCESS, &hSubKey) != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                delete [] pSubKeyName;
                return false;
            }
            delete [] pSubKeyName;
            RegCloseKey(hKey);
            hKey = hSubKey;
        }
        char * pSubKeyName = _toAnsi(nodePath[nodePath.size() - 1]);
        bool result = _deleteSubKey(hKey, pSubKeyName);
        delete [] pSubKeyName;
        RegCloseKey(hKey);
        return result;
    }
    catch (_RegistryException)
    {
        return false;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return _pEmulatedSettings->destroyNode(_constructAbsolutePath(node));
#else
    #error Unsupported platform
#endif
}

Array<SettingsNode> Settings::getChildNodes(const SettingsNode & node)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    try
    {
        Array<SettingsNode> result;
        HKEY hKey = _openRootKey(this);
        Array<String> nodePath = _splitNodePath(_constructAbsolutePath(node).toString());
        for (UIntSize i = 0; i < nodePath.size(); i++)
        {
            HKEY hSubKey;
            DWORD subKeyDisposition;
            char * pSubKeyName = _toAnsi(nodePath[i]);
            if (RegCreateKeyEx(hKey, pSubKeyName, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hSubKey, &subKeyDisposition) != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                delete [] pSubKeyName;
                return Array<SettingsNode>();
            }
            delete [] pSubKeyName;
            RegCloseKey(hKey);
            hKey = hSubKey;
        }
        for (DWORD index = 0; ; index++)
        {
            char subkeyName[256];
            DWORD subkeyNameSize = 255;
            FILETIME lastWriteTime;
            LONG enumResult = RegEnumKeyEx(hKey, index, subkeyName, &subkeyNameSize, NULL, NULL, NULL, &lastWriteTime);
            if (enumResult == ERROR_NO_MORE_ITEMS)
            {
                break;
            }
            else if (enumResult != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                throw _RegistryException();
            }
            else
            {
                result.append(SettingsNode(node, _fromAnsi(subkeyName)));
            }
        }
        RegCloseKey(hKey);
        return result;
    }
    catch (_RegistryException)
    {
        return Array<SettingsNode>();
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return _pEmulatedSettings->getChildNodes(_constructAbsolutePath(node));
#else
    #error Unsupported platform
#endif
}

bool Settings::valueExists(const SettingsNode & node, const String & valueName)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    try
    {
        HKEY hKey = _openRootKey(this);
        Array<String> nodePath = _splitNodePath(_constructAbsolutePath(node).toString());
        for (UIntSize i = 0; i < nodePath.size(); i++)
        {
            HKEY hSubKey;
            char * pSubKeyName = _toAnsi(nodePath[i]);
            if (RegOpenKeyEx(hKey, pSubKeyName, 0, KEY_ALL_ACCESS, &hSubKey) != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                delete [] pSubKeyName;
                return false;
            }
            delete [] pSubKeyName;
            RegCloseKey(hKey);
            hKey = hSubKey;
        }
        for (DWORD index = 0; ; index++)
        {
            char valueNameAnsi[16384];
            DWORD valueNameSize = 16383;
            LONG enumResult = RegEnumValue(hKey, index, valueNameAnsi, &valueNameSize, NULL, NULL, NULL, NULL);
            if (enumResult == ERROR_NO_MORE_ITEMS)
            {
                RegCloseKey(hKey);
                return false;
            }
            else if (enumResult != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                return false;
            }
            else if (_fromAnsi(valueNameAnsi) == valueName)
            {
                RegCloseKey(hKey);
                return true;
            }
        }
    }
    catch (_RegistryException)
    {
        return false;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return _pEmulatedSettings->valueExists(_constructAbsolutePath(node), valueName);
#else
    #error Unsupported platform
#endif
}

Array<String> Settings::getValueNames(const SettingsNode & node)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    try
    {
        Array<String> result;
        HKEY hKey = _openRootKey(this);
        Array<String> nodePath = _splitNodePath(_constructAbsolutePath(node).toString());
        for (UIntSize i = 0; i < nodePath.size(); i++)
        {
            HKEY hSubKey;
            char * pSubKeyName = _toAnsi(nodePath[i]);
            if (RegOpenKeyEx(hKey, pSubKeyName, 0, KEY_ALL_ACCESS, &hSubKey) != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                delete [] pSubKeyName;
                return Array<String>();
            }
            delete [] pSubKeyName;
            RegCloseKey(hKey);
            hKey = hSubKey;
        }
        for (DWORD index = 0; ; index++)
        {
            char valueName[16384];
            DWORD valueNameSize = 16383;
            LONG enumResult = RegEnumValue(hKey, index, valueName, &valueNameSize, NULL, NULL, NULL, NULL);
            if (enumResult == ERROR_NO_MORE_ITEMS)
            {
                RegCloseKey(hKey);
                return result;
            }
            else if (enumResult != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                return Array<String>();
            }
            else
            {
                result.append(_fromAnsi(valueName));
            }
        }
    }
    catch (_RegistryException)
    {
        return Array<String>();
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return _pEmulatedSettings->getValueNames(_constructAbsolutePath(node));
#else
    #error Unsupported platform
#endif
}

bool Settings::destroyValue(const SettingsNode & node, const String & valueName)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    if (!valueExists(node, valueName))
    {
        return true;
    }
    try
    {
        HKEY hKey = _openRootKey(this);
        Array<String> nodePath = _splitNodePath(_constructAbsolutePath(node).toString());
        for (UIntSize i = 0; i < nodePath.size(); i++)
        {
            HKEY hSubKey;
            char * pSubKeyName = _toAnsi(nodePath[i]);
            if (RegOpenKeyEx(hKey, pSubKeyName, 0, KEY_ALL_ACCESS, &hSubKey) != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                delete [] pSubKeyName;
                return true;
            }
            delete [] pSubKeyName;
            RegCloseKey(hKey);
            hKey = hSubKey;
        }
        LONG deleteResult;
        if (valueName.length() == 0)
        {
            deleteResult = RegDeleteValue(hKey, NULL);
        }
        else
        {
            char * pValueName = _toAnsi(valueName);
            deleteResult = RegDeleteValue(hKey, pValueName);
            delete [] pValueName;
        }
        RegCloseKey(hKey);
        return deleteResult == ERROR_SUCCESS;
    }
    catch (_RegistryException)
    {
        return false;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return _pEmulatedSettings->destroyValue(_constructAbsolutePath(node), valueName);
#else
    #error Unsupported platform
#endif
}

String Settings::getValue(const SettingsNode & node, const String & valueName, const String & defaultValue)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    try
    {
        HKEY hKey = _openRootKey(this);
        Array<String> nodePath = _splitNodePath(_constructAbsolutePath(node).toString());
        for (UIntSize i = 0; i < nodePath.size(); i++)
        {
            HKEY hSubKey;
            char * pSubKeyName = _toAnsi(nodePath[i]);
            if (RegOpenKeyEx(hKey, pSubKeyName, 0, KEY_ALL_ACCESS, &hSubKey) != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                delete [] pSubKeyName;
                return defaultValue;
            }
            delete [] pSubKeyName;
            RegCloseKey(hKey);
            hKey = hSubKey;
        }
        LONG queryValueResult;
        DWORD valueType;
        char * pValue = new char[1024];
        DWORD valueSize = 1024;
        if (valueName.length() == 0)
        {
            queryValueResult = RegQueryValueEx(hKey, NULL, NULL, &valueType, (LPBYTE)pValue, &valueSize);
            if (queryValueResult == ERROR_MORE_DATA)
            {
                delete [] pValue;
                pValue = new char[valueSize];
                queryValueResult = RegQueryValueEx(hKey, NULL, NULL, &valueType, (LPBYTE)pValue, &valueSize);
            }
        }
        else
        {
            char * pValueName = _toAnsi(valueName);
            queryValueResult = RegQueryValueEx(hKey, pValueName, NULL, &valueType, (LPBYTE)pValue, &valueSize);
            if (queryValueResult == ERROR_MORE_DATA)
            {
                delete [] pValue;
                pValue = new char[valueSize];
                queryValueResult = RegQueryValueEx(hKey, pValueName, NULL, &valueType, (LPBYTE)pValue, &valueSize);
            }
            delete [] pValueName;
        }
        RegCloseKey(hKey);
        if (queryValueResult != ERROR_SUCCESS)
        {
            return defaultValue;
        }
        if (valueType != REG_SZ)
        {
            return defaultValue;
        }
        String result = _fromAnsi(pValue);
        delete [] pValue;
        return result;
    }
    catch (_RegistryException)
    {
        return defaultValue;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return _pEmulatedSettings->getValue(_constructAbsolutePath(node), valueName, defaultValue);
#else
    #error Unsupported platform
#endif
}

bool Settings::setValue(const SettingsNode & node, const String & valueName, const String & value)
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    try
    {
        HKEY hKey = _openRootKey(this);
        Array<String> nodePath = _splitNodePath(_constructAbsolutePath(node).toString());
        for (UIntSize i = 0; i < nodePath.size(); i++)
        {
            HKEY hSubKey;
            DWORD subKeyDisposition;
            char * pSubKeyName = _toAnsi(nodePath[i]);
            if (RegCreateKeyEx(hKey, pSubKeyName, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hSubKey, &subKeyDisposition) != ERROR_SUCCESS)
            {
                RegCloseKey(hKey);
                delete [] pSubKeyName;
                return false;
            }
            delete [] pSubKeyName;
            RegCloseKey(hKey);
            hKey = hSubKey;
        }
        LONG setValueResult;
        const char * pValue = _toAnsi(value);
        if (valueName.length() == 0)
        {
            setValueResult = RegSetValueEx(hKey, NULL, 0, REG_SZ, (const BYTE*)pValue, (DWORD)strlen(pValue) + 1);
        }
        else
        {
            char * pValueName = _toAnsi(valueName);
            setValueResult = RegSetValueEx(hKey, pValueName, 0, REG_SZ, (const BYTE*)pValue, (DWORD)strlen(pValue) + 1);
            delete [] pValueName;
        }
        RegCloseKey(hKey);
        return setValueResult == ERROR_SUCCESS;
    }
    catch (_RegistryException)
    {
        return false;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return _pEmulatedSettings->setValue(_constructAbsolutePath(node), valueName, value);
#else
    #error Unsupported platform
#endif
}

//////////
//  Implementation helpers
SettingsNode Settings::_constructAbsolutePath(const SettingsNode & relativePath)
{
    if (relativePath.isAbsolute())
    {
        return relativePath;
    }
    else
    {
        return SettingsNode(_virtualRoot, relativePath.toString());
    }
}

SettingsNode Settings::_constructRelativePath(const SettingsNode & absolutePath)
{
    if (absolutePath.toString() == _virtualRoot.toString())
    {
        return SettingsNode("/");
    }
    else if (absolutePath.toString().startsWith(_virtualRoot.toString() + "/"))
    {
        return SettingsNode(absolutePath.toString().mid(_virtualRoot.toString().length()));
    }
    else
    {
        return absolutePath;
    }
}

//  End of Settings.cpp
