//
//  StandardErrorStream.cpp - Implementation of the carl::StandardErrorStream
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:55 $
//  $Revision: 1.3 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

#include <stdio.h>
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <windows.h>
    #include <io.h>
    #include <fcntl.h>
#endif 

namespace
{
    //////////
    //  Binary console stream
    class CARL_PUBLIC _StandardByteErrorStream : public IByteOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _StandardByteErrorStream(const _StandardByteErrorStream &);
        void                    operator = (const _StandardByteErrorStream &);

        //////////
        //  Construction/destruction
    public:
        _StandardByteErrorStream() 
        { 
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
            _setmode(_fileno(stderr), _O_BINARY);
#endif
        }

        virtual ~_StandardByteErrorStream() 
        {
        }

        //////////
        //  IByteOutput
    public:
        virtual void            flush() throw(IOException) 
        { 
            fflush(stderr); 
        }

        virtual void            write(UInt8 byte) throw(IOException) 
        { 
            if (fputc(byte, stderr) == EOF)
            {
                throw IOException();
            }
            flush();
        }

        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException) 
        {
            for (UIntSize i = 0; i < bytesToWrite; i++)
            {
                write((static_cast<const UInt8*>(pBuffer))[i]);
            }
        }
    };

    //////////
    //  Text console stream
    class CARL_PUBLIC _StandardTextErrorStream : public ITextOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _StandardTextErrorStream(const _StandardTextErrorStream &);
        void                    operator = (const _StandardTextErrorStream &);

        //////////
        //  Construction/destruction
    public:
        _StandardTextErrorStream() 
            :   _sbos(), _tos(&_sbos) 
        {
        }

        virtual ~_StandardTextErrorStream() 
        {
        }

        //////////
        //  ITextOutput
    public:
        virtual void            flush() throw(IOException) 
        { 
            _tos.flush(); 
        }

        virtual void            write(Char c) throw(IOException) 
        { 
            _tos.write(c); 
        }

        virtual void            write(const char * pBuffer) throw(IOException) 
        { 
            _tos.write(pBuffer); 
        }

        virtual void            write(const char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.write(pBuffer, charsToWrite); 
        }

        virtual void            write(const Char * pBuffer) throw(IOException) 
        { 
            _tos.write(pBuffer); 
        }

        virtual void            write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.write(pBuffer, charsToWrite); 
        }

        virtual void            write(const String & s) throw(IOException) 
        { 
            _tos.write(s); 
        }

        virtual void            writeLn() throw(IOException) 
        { 
            _tos.writeLn(); 
        }

        virtual void            writeLn(Char c) throw(IOException) 
        { 
            _tos.writeLn(c); 
        }

        virtual void            writeLn(const char * pBuffer) throw(IOException) 
        { 
            _tos.writeLn(pBuffer); 
        }

        virtual void            writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.writeLn(pBuffer, charsToWrite); 
        }

        virtual void            writeLn(const Char * pBuffer) throw(IOException) 
        { 
            _tos.writeLn(pBuffer); 
        }

        virtual void            writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.writeLn(pBuffer, charsToWrite); 
        }

        virtual void            writeLn(const String & s) throw(IOException) 
        { 
            _tos.writeLn(s); 
        }

        //////////
        //  Operations
    public:
        virtual CharacterSet *  getCharacterSet() const 
        { 
            return _tos.getCharacterSet(); 
        }

        virtual void            setCharacterSet(CharacterSet * pCharacterSet) 
        { 
            _tos.setCharacterSet(pCharacterSet); 
        }

        //////////
        //  Implementation
    private:
        _StandardByteErrorStream    _sbos;
        TextOutputStream            _tos;
    };
}

//////////
//  StandardErrorStream
CharacterSet * StandardErrorStream::getCharacterSet() const
{
    return (static_cast<_StandardTextErrorStream*>(_getDestination()))->getCharacterSet();
}

void StandardErrorStream::setCharacterSet(CharacterSet * pCharacterSet)
{
    (static_cast<_StandardTextErrorStream*>(_getDestination()))->setCharacterSet(pCharacterSet);
}

ITextOutput * StandardErrorStream::_getDestination()
{
    static _StandardTextErrorStream standardErrorStream;
    return &standardErrorStream;
}


//  End of StandardErrorStream.cpp
