//
//  StandardInputStream.cpp - Implementation of the carl::StandardInputStream
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:55 $
//  $Revision: 1.3 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

#include <stdio.h>
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <windows.h>
    #include <io.h>
    #include <fcntl.h>
#endif 

namespace
{
    //////////
    //  Binary console stream
    class _StandardByteInputStream : public IByteInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _StandardByteInputStream(const _StandardByteInputStream &);
        void                    operator = (const _StandardByteInputStream &);

        //////////
        //  Construction/destruction
    public:
        _StandardByteInputStream() 
        { 
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
            _setmode(_fileno(stdin), _O_BINARY);
#endif
        }

        virtual ~_StandardByteInputStream() 
        {
        }

        //////////
        //  IByteInput
    public:
        virtual UIntSize        available() throw(IOException) 
        { 
            return true; 
        }

        virtual int             read() throw(IOException) 
        { 
            return fgetc(stdin); 
        }

        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException)
        {
            for (UIntSize i = 0; i < bytesToRead; i++)
            {
                int bb = read();
                if (bb == -1)
                {
                    return i;
                }
                (static_cast<UInt8*>(pBuffer))[i] = static_cast<UInt8>(bb);
            }
            return bytesToRead;
        }

        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException)
        {
            for (UIntSize i = 0; i < bytesToSkip; i++)
            {
                int bb = read();
                if (bb == -1)
                {
                    return i;
                }
            }
            return bytesToSkip;
        }
    };

    //////////
    //  Text console stream
    class _StandardTextInputStream : public ITextInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _StandardTextInputStream(const _StandardTextInputStream &);
        void                    operator = (const _StandardTextInputStream &);

        //////////
        //  Construction/destruction
    public:
        _StandardTextInputStream() 
            :   _sbis(), _tis(&_sbis) 
        {
        }

        virtual ~_StandardTextInputStream() 
        {
        }

        //////////
        //  ITextInput
    public:
        virtual bool            ready() throw(IOException) 
        { 
            return _tis.ready(); 
        }

        virtual bool            read(Char & c) throw(IOException) 
        { 
            return _tis.read(c); 
        }

        virtual UIntSize        read(Char * pBuffer, UIntSize charsToRead) throw(IOException) 
        { 
            return _tis.read(pBuffer, charsToRead); 
        }

        virtual bool            readLine(String & line, bool discardEoln = false) throw(IOException) 
        { 
            return _tis.readLine(line, discardEoln); 
        }

        virtual UIntSize        skip(UIntSize charsToSkip) throw(IOException) 
        { 
            return _tis.skip(charsToSkip); 
        }

        virtual bool            unread(Char c) throw(IOException) 
        { 
            return _tis.unread(c); 
        }

        //////////
        //  Operations
    public:
        virtual CharacterSet *  getCharacterSet() const 
        { 
            return _tis.getCharacterSet(); 
        }

        virtual void            setCharacterSet(CharacterSet * pCharacterSet) 
        { 
            _tis.setCharacterSet(pCharacterSet); 
        }

        //////////
        //  Implementation
    private:
        _StandardByteInputStream    _sbis;
        TextInputStream             _tis;
    };
}

//////////
//  StandardInputStream
CharacterSet * StandardInputStream::getCharacterSet() const
{
    return (static_cast<_StandardTextInputStream*>(_getSource()))->getCharacterSet();
}

void StandardInputStream::setCharacterSet(CharacterSet * pCharacterSet)
{
    (static_cast<_StandardTextInputStream*>(_getSource()))->setCharacterSet(pCharacterSet);
}

ITextInput * StandardInputStream::_getSource()
{
    static _StandardTextInputStream standardInputStream;
    return &standardInputStream;
}

//  End of StandardInputStream.cpp
