//
//  StandardOutputStream.cpp - Implementation of the carl::StandardOutputStream
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:55 $
//  $Revision: 1.3 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

#include <stdio.h>
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <windows.h>
    #include <io.h>
    #include <fcntl.h>
#endif 

namespace
{
    //////////
    //  Binary console stream
    class CARL_PUBLIC _StandardByteOutputStream : public IByteOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _StandardByteOutputStream(const _StandardByteOutputStream &);
        void                    operator = (const _StandardByteOutputStream &);

        //////////
        //  Construction/destruction
    public:
        _StandardByteOutputStream() 
        { 
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
            _setmode(_fileno(stdout), _O_BINARY);
#endif
        }

        virtual ~_StandardByteOutputStream() 
        {
        }

        //////////
        //  IByteOutput
    public:
        virtual void            flush() throw(IOException) 
        { 
            fflush(stdout); 
        }

        virtual void            write(UInt8 byte) throw(IOException) 
        { 
            if (fputc(byte, stdout) == EOF)
            {
                throw IOException();
            }
            flush();
        }

        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException) 
        {
            for (UIntSize i = 0; i < bytesToWrite; i++)
            {
                write((static_cast<const UInt8*>(pBuffer))[i]);
            }
        }
    };

    //////////
    //  Text console stream
    class CARL_PUBLIC _StandardTextOutputStream : public ITextOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        _StandardTextOutputStream(const _StandardTextOutputStream &);
        void                    operator = (const _StandardTextOutputStream &);

        //////////
        //  Construction/destruction
    public:
        _StandardTextOutputStream() 
            :   _sbos(), _tos(&_sbos) 
        {
        }

        virtual ~_StandardTextOutputStream() 
        {
        }

        //////////
        //  ITextOutput
    public:
        virtual void            flush() throw(IOException) 
        { 
            _tos.flush(); 
        }

        virtual void            write(Char c) throw(IOException) 
        { 
            _tos.write(c); 
        }

        virtual void            write(const char * pBuffer) throw(IOException) 
        { 
            _tos.write(pBuffer); 
        }

        virtual void            write(const char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.write(pBuffer, charsToWrite); 
        }

        virtual void            write(const Char * pBuffer) throw(IOException) 
        { 
            _tos.write(pBuffer); 
        }

        virtual void            write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.write(pBuffer, charsToWrite); 
        }

        virtual void            write(const String & s) throw(IOException) 
        { 
            _tos.write(s); 
        }

        virtual void            writeLn() throw(IOException) 
        { 
            _tos.writeLn(); 
        }

        virtual void            writeLn(Char c) throw(IOException) 
        { 
            _tos.writeLn(c); 
        }

        virtual void            writeLn(const char * pBuffer) throw(IOException) 
        { 
            _tos.writeLn(pBuffer); 
        }

        virtual void            writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.writeLn(pBuffer, charsToWrite); 
        }

        virtual void            writeLn(const Char * pBuffer) throw(IOException) 
        { 
            _tos.writeLn(pBuffer); 
        }

        virtual void            writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) 
        { 
            _tos.writeLn(pBuffer, charsToWrite); 
        }

        virtual void            writeLn(const String & s) throw(IOException) 
        { 
            _tos.writeLn(s); 
        }

        //////////
        //  Operations
    public:
        virtual CharacterSet *  getCharacterSet() const 
        { 
            return _tos.getCharacterSet(); 
        }

        virtual void            setCharacterSet(CharacterSet * pCharacterSet) 
        { 
            _tos.setCharacterSet(pCharacterSet); 
        }

        //////////
        //  Implementation
    private:
        _StandardByteOutputStream    _sbos;
        TextOutputStream            _tos;
    };
}

//////////
//  StandardOutputStream
CharacterSet * StandardOutputStream::getCharacterSet() const
{
    return (static_cast<_StandardTextOutputStream*>(_getDestination()))->getCharacterSet();
}

void StandardOutputStream::setCharacterSet(CharacterSet * pCharacterSet)
{
    (static_cast<_StandardTextOutputStream*>(_getDestination()))->setCharacterSet(pCharacterSet);
}

ITextOutput * StandardOutputStream::_getDestination()
{
    static _StandardTextOutputStream standardOutputStream;
    return &standardOutputStream;
}


//  End of StandardOutputStream.cpp
