//
//  SynchronizationObject.cpp - Implementation of the carl::SynchronizationObject class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:55 $
//  $Revision: 1.9 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "SynchronizationObjectImpl.hpp"
#include "MutexImpl.hpp"
#include "SemaphoreImpl.hpp"
#include "EventImpl.hpp"

//////////
//  Helpers
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
void _impl::_SynchronizationObjectImpl::storeSynchronizationObjectHandles(SynchronizationObject * const * pSynchronizationObjects,
                                                                          UIntSize synchronizationObjectCount,
                                                                          Array<HANDLE> & handles)
{
    require(pSynchronizationObjects != NULL);
    require(synchronizationObjectCount != 0);
    require(handles.size() == 0);

    for (UIntSize i = 0; i < synchronizationObjectCount; i++)
    {
        const SynchronizationObject * pSynchronizationObject = pSynchronizationObjects[i];
        require(pSynchronizationObject != NULL);
        //  Determine the synchronization object type & store the handle
        const Mutex * pMutex = dynamic_cast<const Mutex*>(pSynchronizationObject);
        const Semaphore * pSemaphore = dynamic_cast<const Semaphore*>(pSynchronizationObject);
        const Event * pEvent = dynamic_cast<const Event*>(pSynchronizationObject);
        if (pMutex != NULL)
        {
            handles.append(pMutex->_pImpl->_osMutex);
        }
        else if (pSemaphore != NULL)
        {
            handles.append(pSemaphore->_pImpl->_osSemaphore);
        }
        else if (pEvent != NULL)
        {
            handles.append(pEvent->_pImpl->_osEvent);
        }
        else
        {
            throw SystemError();
        }
    }
}

void _impl::_SynchronizationObjectImpl::waitForAll(const Array<HANDLE> & handles)
{
    require(handles.size() != 0);

    if (handles.size() <= MAXIMUM_WAIT_OBJECTS)
    {   //  Can wait all in one go
        DWORD oldLastError = GetLastError();
        DWORD waitResult = WaitForMultipleObjects(handles.size(), handles.getData(), TRUE, INFINITE);
        SetLastError(oldLastError);
        if ((waitResult >= WAIT_OBJECT_0 && waitResult < WAIT_OBJECT_0 + handles.size()) ||
            (waitResult >= WAIT_ABANDONED_0 && waitResult < WAIT_ABANDONED_0 + handles.size()))
        {   //  All the handles have been acquired
            return;
        }
        else
        {
            throw SystemError();
        }
    }
    else
    {   //  Must break the wait into chunks
        UIntSize handlesToWaitFor = handles.size();
        UIntSize startHandleIndex = 0;
        while (handlesToWaitFor != 0)
        {
            //  Determine how many handles to wait for this time
            UIntSize chunkSize = handlesToWaitFor;
            if (chunkSize > MAXIMUM_WAIT_OBJECTS)
            {
                chunkSize = MAXIMUM_WAIT_OBJECTS;
            }
            //  Wait for a chunk of handles
            DWORD oldLastError = GetLastError();
            DWORD waitResult = WaitForMultipleObjects(chunkSize, handles.getData() + startHandleIndex, TRUE, INFINITE);
            SetLastError(oldLastError);
            if ((waitResult >= WAIT_OBJECT_0 && waitResult < WAIT_OBJECT_0 + chunkSize) ||
                (waitResult >= WAIT_ABANDONED_0 && waitResult < WAIT_ABANDONED_0 + chunkSize))
            {   //  All the handles in the chunk has been acquired
                handlesToWaitFor -= chunkSize;
                startHandleIndex += chunkSize;
                if (handlesToWaitFor == 0)
                {
                    return;
                }
            }
            else
            {
                throw SystemError();
            }
        }
    }
}

bool _impl::_SynchronizationObjectImpl::waitForAll(const Array<HANDLE> & handles, const TimeSpan & timeout)
{
    require(handles.size() != 0);

    if (handles.size() <= MAXIMUM_WAIT_OBJECTS)
    {   //  Can wait all in one go
        TimeStamp waitStartTime = TimeStamp::now();
        TimeStamp waitEndTime = waitStartTime + timeout;
        for (; ; )
        {
            //  Calculate how long to wait for this time
            Int64 millisecondsToWait = (waitEndTime - TimeStamp::now()).getNanoseconds();
            if (millisecondsToWait < 0)
            {
                millisecondsToWait = 0;
            }
            if (millisecondsToWait > 1000000)
            {
                millisecondsToWait = 1000000;
            }
            //  Wait for a bit
            DWORD oldLastError = GetLastError();
            DWORD waitResult = WaitForMultipleObjects(handles.size(), handles.getData(), TRUE, static_cast<DWORD>(millisecondsToWait));
            SetLastError(oldLastError);
            if ((waitResult >= WAIT_OBJECT_0 && waitResult < WAIT_OBJECT_0 + handles.size()) ||
                (waitResult >= WAIT_ABANDONED_0 && waitResult < WAIT_ABANDONED_0 + handles.size()))
            {   //  All the handles have been acquired
                return true;
            }
            else if (waitResult == WAIT_TIMEOUT)
            {   //  Could not acquire all objects during this round
                if (TimeStamp::now() >= waitEndTime)
                {   //  Waited for the whole timeout
                    return false;
                }
            }
            else
            {
                throw SystemError();
            }
        }
    }
    else
    {   //  Must break the wait into chunks
        TimeStamp waitStartTime = TimeStamp::now();
        TimeStamp waitEndTime = waitStartTime + timeout;
        UIntSize handlesToWaitFor = handles.size();
        UIntSize startHandleIndex = 0;
        for (; ; )
        {
            //  Determine how many handles to wait for this time
            UIntSize chunkSize = handlesToWaitFor;
            if (chunkSize > MAXIMUM_WAIT_OBJECTS)
            {
                chunkSize = MAXIMUM_WAIT_OBJECTS;
            }
            //  Wait for a chunk of handles
            Int64 millisecondsToWait = (waitEndTime - TimeStamp::now()).getNanoseconds();
            if (millisecondsToWait < 0)
            {
                millisecondsToWait = 0;
            }
            if (millisecondsToWait > 1000)
            {
                millisecondsToWait = 1000;
            }
            DWORD oldLastError = GetLastError();
            DWORD waitResult = WaitForMultipleObjects(chunkSize, handles.getData() + startHandleIndex, TRUE, static_cast<DWORD>(millisecondsToWait));
            SetLastError(oldLastError);
            if ((waitResult >= WAIT_OBJECT_0 && waitResult < WAIT_OBJECT_0 + chunkSize) ||
                (waitResult >= WAIT_ABANDONED_0 && waitResult < WAIT_ABANDONED_0 + chunkSize))
            {
                handlesToWaitFor -= chunkSize;
                startHandleIndex += chunkSize;
                if (handlesToWaitFor == 0)
                {
                    return true;
                }
            }
            else if (waitResult == WAIT_TIMEOUT)
            {
                if (TimeStamp::now() >= waitEndTime)
                {   //  Waited for the whole timeout
                    return false;
                }
            }
            else
            {
                throw SystemError();
            }
        }
    }
}

bool _impl::_SynchronizationObjectImpl::waitForAny(const Array<HANDLE> & handles, UIntSize & obtainedHandleIndex)
{
    require(handles.size() != 0);

    if (handles.size() <= MAXIMUM_WAIT_OBJECTS)
    {   //  Can wait all in one go
        DWORD oldLastError = GetLastError();
        DWORD waitResult = WaitForMultipleObjects(handles.size(), handles.getData(), FALSE, INFINITE);
        SetLastError(oldLastError);
        if (waitResult >= WAIT_OBJECT_0 && waitResult < WAIT_OBJECT_0 + handles.size())
        {
            obtainedHandleIndex = waitResult - WAIT_OBJECT_0;
            return true;
        }
        else if (waitResult >= WAIT_ABANDONED_0 && waitResult < WAIT_ABANDONED_0 + handles.size())
        {
            obtainedHandleIndex = waitResult - WAIT_ABANDONED_0;
            return true;
        }
        else
        {
            throw SystemError();
        }
    }
    else
    {   //  Must break the wait into chunks
        for (; ; )
        {
            UIntSize handlesToWaitFor = handles.size();
            UIntSize startHandleIndex = 0;
            while (handlesToWaitFor != 0)
            {
                //  Determine how many handles to wait for this time
                UIntSize chunkSize = handlesToWaitFor;
                if (chunkSize > MAXIMUM_WAIT_OBJECTS)
                {
                    chunkSize = MAXIMUM_WAIT_OBJECTS;
                }
                //  Wait for a chunk of handles
                DWORD oldLastError = GetLastError();
                DWORD waitResult = WaitForMultipleObjects(chunkSize, handles.getData() + startHandleIndex, FALSE, 100);
                SetLastError(oldLastError);
                if (waitResult >= WAIT_OBJECT_0 && waitResult < WAIT_OBJECT_0 + chunkSize)
                {
                    obtainedHandleIndex  = startHandleIndex + waitResult - WAIT_OBJECT_0;
                    return true;
                }
                else if (waitResult >= WAIT_ABANDONED_0 && waitResult < WAIT_ABANDONED_0 + chunkSize)
                {
                    obtainedHandleIndex  = startHandleIndex + waitResult - WAIT_ABANDONED_0;
                    return true;
                }
                else if (waitResult == WAIT_TIMEOUT)
                {   //  Wait timed out - try the next chunk
                    handlesToWaitFor -= chunkSize;
                    startHandleIndex += chunkSize;
                    if (handlesToWaitFor == 0)
                    {   //  We have waited for ALL of the handles - and still none is signalled.
                        //  Start another iteration of waiting
                        break;
                    }
                }
                else
                {
                    throw SystemError();
                }
            }
        }
    }
}

bool _impl::_SynchronizationObjectImpl::waitForAny(const Array<HANDLE> & handles, const TimeSpan & timeout, UIntSize & obtainedHandleIndex)
{
    require(handles.size() != 0);

    if (handles.size() <= MAXIMUM_WAIT_OBJECTS)
    {   //  Can wait all in one go
        TimeStamp waitStartTime = TimeStamp::now();
        TimeStamp waitEndTime = waitStartTime + timeout;
        for (; ; )
        {
            //  Calculate how long to wait for this time
            Int64 millisecondsToWait = (waitEndTime - TimeStamp::now()).getNanoseconds();
            if (millisecondsToWait < 0)
            {
                millisecondsToWait = 0;
            }
            if (millisecondsToWait > 1000000)
            {
                millisecondsToWait = 1000000;
            }
            //  Wait for a bit
            DWORD oldLastError = GetLastError();
            DWORD waitResult = WaitForMultipleObjects(handles.size(), handles.getData(), FALSE, static_cast<DWORD>(millisecondsToWait));
            SetLastError(oldLastError);
            if (waitResult >= WAIT_OBJECT_0 && waitResult < WAIT_OBJECT_0 + handles.size())
            {
                obtainedHandleIndex = waitResult - WAIT_OBJECT_0;
                return true;
            }
            else if (waitResult >= WAIT_ABANDONED_0 && waitResult < WAIT_ABANDONED_0 + handles.size())
            {
                obtainedHandleIndex = waitResult - WAIT_ABANDONED_0;
                return true;
            }
            else if (waitResult == WAIT_TIMEOUT)
            {   //  Could not acquire all objects during this round
                if (TimeStamp::now() >= waitEndTime)
                {   //  Waited for the whole timeout
                    return false;
                }
            }
            else
            {
                throw SystemError();
            }
        }
    }
    else
    {   //  Must break the wait into chunks
        TimeStamp waitStartTime = TimeStamp::now();
        TimeStamp waitEndTime = waitStartTime + timeout;
        for (; ; )
        {
            UIntSize handlesToWaitFor = handles.size();
            UIntSize startHandleIndex = 0;
            while (handlesToWaitFor != 0)
            {
                //  Determine how many handles to wait for this time
                UIntSize chunkSize = handlesToWaitFor;
                if (chunkSize > MAXIMUM_WAIT_OBJECTS)
                {
                    chunkSize = MAXIMUM_WAIT_OBJECTS;
                }
                //  Wait for a chunk of handles
                Int64 millisecondsToWait = (waitEndTime - TimeStamp::now()).getNanoseconds();
                if (millisecondsToWait < 0)
                {
                    millisecondsToWait = 0;
                }
                if (millisecondsToWait > 100)
                {
                    millisecondsToWait = 100;
                }
                DWORD oldLastError = GetLastError();
                DWORD waitResult = WaitForMultipleObjects(chunkSize, handles.getData() + startHandleIndex, FALSE, static_cast<DWORD>(millisecondsToWait));
                SetLastError(oldLastError);
                if (waitResult >= WAIT_OBJECT_0 && waitResult < WAIT_OBJECT_0 + chunkSize)
                {
                    obtainedHandleIndex = startHandleIndex + waitResult - WAIT_OBJECT_0;
                    return true;
                }
                else if (waitResult >= WAIT_ABANDONED_0 && waitResult < WAIT_ABANDONED_0 + chunkSize)
                {
                    obtainedHandleIndex  = startHandleIndex + waitResult - WAIT_ABANDONED_0;
                    return true;
                }
                else if (waitResult == WAIT_TIMEOUT)
                {   //  Wait timed out - try the next chunk
                    handlesToWaitFor -= chunkSize;
                    startHandleIndex += chunkSize;
                    if (handlesToWaitFor == 0)
                    {   //  We have waited for ALL of the handles - and still none is signalled.
                        //  Start another iteration of waiting
                        break;
                    }
                }
                else
                {
                    throw SystemError();
                }
            }
            if (TimeStamp::now() >= waitEndTime)
            {   //  Waited for the whole timeout
                return NULL;
            }
        }
    }
}
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
#else
    #error Unsupported platform
#endif

//////////
//  Construction
SynchronizationObject::SynchronizationObject()
{
}

SynchronizationObject::~SynchronizationObject()
{
}

//////////
//  Generic wait services
void SynchronizationObject::waitForAll(SynchronizationObject * const pSynchronizationObjects[], UIntSize synchronizationObjectCount)
{
    require(pSynchronizationObjects != NULL || synchronizationObjectCount == 0);

    if (synchronizationObjectCount == 0)
    {
        return;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<HANDLE> handles(synchronizationObjectCount);
    _impl::_SynchronizationObjectImpl::storeSynchronizationObjectHandles(pSynchronizationObjects,
                                                                         synchronizationObjectCount,
                                                                         handles);
    _impl::_SynchronizationObjectImpl::waitForAll(handles);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    Array<bool> acquired;
    acquired.setSize(synchronizationObjectCount);
    for (UIntSize i = 0; i < synchronizationObjectCount; i++)
    {
        acquired[i] = false;
    }
    for (; ; )
    {
        //  Try to acquire
        UIntSize acquiredCount = 0;
        for (UIntSize i = 0; i < synchronizationObjectCount; i++)
        {
            if (pSynchronizationObjects[i]->waitFor(TimeSpan(0)))
            {
                acquiredCount++;
                acquired[i] = true;
            }
        }
        //  All acquired ?
        if (acquiredCount == synchronizationObjectCount)
        {
            return;
        }
        //  Release the ones that we DID acquire & keep waiting
        for (UIntSize i = 0; i < synchronizationObjectCount; i++)
        {
            if (acquired[i])
            {
                pSynchronizationObjects[i]->release();
            }
        }
        Thread::sleep(TimeSpan(0));
    }
#else
    #error Unsupported platform
#endif
}

void SynchronizationObject::waitForAll(const Array<SynchronizationObject*> & synchronizationObjects)
{
    waitForAll(synchronizationObjects.getData(), synchronizationObjects.size());
}

bool SynchronizationObject::waitForAll(SynchronizationObject * const pSynchronizationObjects[], UIntSize synchronizationObjectCount, const TimeSpan & timeout)
{
    require(pSynchronizationObjects != NULL || synchronizationObjectCount == 0);

    if (synchronizationObjectCount == 0)
    {
        return true;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<HANDLE> handles(synchronizationObjectCount);
    _impl::_SynchronizationObjectImpl::storeSynchronizationObjectHandles(pSynchronizationObjects,
                                                                         synchronizationObjectCount,
                                                                         handles);
    return _impl::_SynchronizationObjectImpl::waitForAll(handles, timeout);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    TimeSpan giveUpTime = TimeSpan::getTimeSpanSinceReferencePoint() + timeout;
    Array<bool> acquired;
    acquired.setSize(synchronizationObjectCount);
    for (UIntSize i = 0; i < synchronizationObjectCount; i++)
    {
        acquired[i] = false;
    }
    for (; ; )
    {
        //  Try to acquire
        UIntSize acquiredCount = 0;
        for (UIntSize i = 0; i < synchronizationObjectCount; i++)
        {
            if (pSynchronizationObjects[i]->waitFor(TimeSpan(0)))
            {
                acquiredCount++;
                acquired[i] = true;
            }
        }
        //  All acquired ?
        if (acquiredCount == synchronizationObjectCount)
        {
            return true;
        }
        //  Release the ones that we DID acquire & keep waiting
        for (UIntSize i = 0; i < synchronizationObjectCount; i++)
        {
            if (acquired[i])
            {
                pSynchronizationObjects[i]->release();
            }
        }
        if (TimeSpan::getTimeSpanSinceReferencePoint() > giveUpTime)
        {
            return false;
        }
        Thread::sleep(TimeSpan(0));
    }
#else
    #error Unsupported platform
#endif
}

bool SynchronizationObject::waitForAll(const Array<SynchronizationObject*> & synchronizationObjects, const TimeSpan & timeout)
{
    return waitForAll(synchronizationObjects.getData(), synchronizationObjects.size(), timeout);
}

SynchronizationObject * SynchronizationObject::waitForAny(SynchronizationObject * const pSynchronizationObjects[], UIntSize synchronizationObjectCount)
{
    require(pSynchronizationObjects != NULL || synchronizationObjectCount == 0);

    if (synchronizationObjectCount == 0)
    {
        return NULL;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<HANDLE> handles(synchronizationObjectCount);
    _impl::_SynchronizationObjectImpl::storeSynchronizationObjectHandles(pSynchronizationObjects,
                                                                         synchronizationObjectCount,
                                                                         handles);
    UIntSize obtainedHandleIndex;
    if (_impl::_SynchronizationObjectImpl::waitForAny(handles, obtainedHandleIndex))
    {
        return pSynchronizationObjects[obtainedHandleIndex];
    }
    return NULL;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    for (; ; )
    {
        //  Try to acquire
        for (UIntSize i = 0; i < synchronizationObjectCount; i++)
        {
            if (pSynchronizationObjects[i]->waitFor(TimeSpan(0)))
            {
                return pSynchronizationObjects[i];
            }
        }
        //  Keep waiting
        Thread::sleep(TimeSpan(0));
    }
#else
    #error Unsupported platform
#endif
}

SynchronizationObject * SynchronizationObject::waitForAny(const Array<SynchronizationObject*> & synchronizationObjects)
{
    return waitForAny(synchronizationObjects.getData(), synchronizationObjects.size());
}

SynchronizationObject * SynchronizationObject::waitForAny(SynchronizationObject * const pSynchronizationObjects[], UIntSize synchronizationObjectCount, const TimeSpan & timeout)
{
    require(pSynchronizationObjects != NULL || synchronizationObjectCount == 0);

    if (synchronizationObjectCount == 0)
    {
        return NULL;
    }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    Array<HANDLE> handles(synchronizationObjectCount);
    _impl::_SynchronizationObjectImpl::storeSynchronizationObjectHandles(pSynchronizationObjects,
                                                                         synchronizationObjectCount,
                                                                         handles);
    UIntSize obtainedHandleIndex;
    if (_impl::_SynchronizationObjectImpl::waitForAny(handles, timeout, obtainedHandleIndex))
    {
        return pSynchronizationObjects[obtainedHandleIndex];
    }
    return NULL;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    TimeSpan giveUpTime = TimeSpan::getTimeSpanSinceReferencePoint() + timeout;
    for (; ; )
    {
        //  Try to acquire
        for (UIntSize i = 0; i < synchronizationObjectCount; i++)
        {
            if (pSynchronizationObjects[i]->waitFor(TimeSpan(0)))
            {
                return pSynchronizationObjects[i];
            }
        }
        if (TimeSpan::getTimeSpanSinceReferencePoint() > giveUpTime)
        {
            return false;
        }
        //  Keep waiting
        Thread::sleep(TimeSpan(0));
    }
#else
    #error Unsupported platform
#endif
}

SynchronizationObject * SynchronizationObject::waitForAny(const Array<SynchronizationObject*> & synchronizationObjects, const TimeSpan & timeout)
{
    return waitForAny(synchronizationObjects.getData(), synchronizationObjects.size(), timeout);
}

//  End of SynchronizationObject.cpp
