//
//  TextInputStream.cpp - Implementation of the carl::TextInputStream class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:55 $
//  $Revision: 1.8 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include <stdlib.h>

//////////
//  Construction/destruction
TextInputStream::TextInputStream(IByteInput * pSource)
{
    require(pSource != NULL);

    _pSource = pSource;
    _pCharacterSet = CharacterSet::getDefault();
}

TextInputStream::TextInputStream(IByteInput * pSource, CharacterSet * pCharacterSet)
{
    require(pSource != NULL);
    require(pCharacterSet != NULL);

    _pSource = pSource;
    _pCharacterSet = pCharacterSet;
}

//////////
//  Operations
CharacterSet * TextInputStream::getCharacterSet() const
{
    require(isOpen());

    return _pCharacterSet;
}

void TextInputStream::setCharacterSet(CharacterSet * pCharacterSet)
{
    require(isOpen());
    require(pCharacterSet != NULL);

    _pCharacterSet = pCharacterSet;
}

//////////
//  Operations
bool TextInputStream::ready() throw(IOException)
{
    require(isOpen());

    return !_charsPushedBack.isEmpty() || _pSource->available() != 0;
}

bool TextInputStream::read(Char & c) throw(IOException)
{
    require(isOpen());

    if (!_charsPushedBack.isEmpty())
    {   //  There are characters puched back - read them first. Note that characters
        //  that can be pushed back already went through EOLN translation
        c = _charsPushedBack.pop();
        return true;
    }

    //  Read next character
    if (!_read(c))
    {
        return false;
    }

    //  Is this the 1st character of the EOLN sequence ?
    if (c != static_cast<unsigned char>(FileSystem::LineSeparator[0]))
    {   //  No, just a character
        return true;
    }

    //  Read enough characters to match against EOLN sequence
    _eolnSequence.setLength(0);
    _eolnSequence += c;
    for (UIntSize i = 1; FileSystem::LineSeparator[i] != 0; i++)
    {
        Char cc;
        if (!_read(cc))
        {   //  EOF before character sequence is read - push everything back except
            //  the 1st character
            while (_eolnSequence.length() != 1)
            {
                _untranslatedCharsPushedBack.push(_eolnSequence[_eolnSequence.length() - 1]);
                _eolnSequence.setLength(_eolnSequence.length() - 1);
            }
            return true;
        }
        _eolnSequence += cc;
    }

    //  EOLN ?
    if (_eolnSequence == FileSystem::LineSeparator)
    {   //  EOLN !
        c = '\n';
        return true;
    }

    //  Not EOLN - push back everything except the 1st character
    while (_eolnSequence.length() != 1)
    {
        _untranslatedCharsPushedBack.push(_eolnSequence[_eolnSequence.length() - 1]);
        _eolnSequence.setLength(_eolnSequence.length() - 1);
    }
    return true;
}

UIntSize TextInputStream::read(Char * pBuffer, UIntSize charsToRead) throw(IOException)
{
    require(isOpen());
    require(pBuffer != NULL);

    for (UIntSize i = 0; i < charsToRead; i++)
    {
        if (!read(pBuffer[i]))
        {
            return i;
        }
    }
    return charsToRead;
}

bool TextInputStream::readLine(String & line, bool discardEoln /*= false*/) throw(IOException)
{
    line.setLength(0);
    for (; ; )
    {
        Char c;

        if (!read(c))
        {
            return (line.length() != 0);
        }
        else if (c == '\n')
        {
            if (!discardEoln)
            {
                line += c;
            }
            return true;
        }
        else
        {
            line += c;
        }
    }
}

UIntSize TextInputStream::skip(UIntSize charsToSkip) throw(IOException)
{
    require(isOpen());

    Char c;
    for (UIntSize i = 0; i < charsToSkip; i++)
    {
        if (!read(c))
        {
            return i;
        }
    }
    return charsToSkip;
}

bool TextInputStream::unread(Char c) throw(IOException)
{
    require(isOpen());

    _charsPushedBack.push(c);
    return true;
}

//////////
//  Implementation helpers
bool TextInputStream::_read(Char & c) throw(IOException)
{
    if (_untranslatedCharsPushedBack.size() != 0)
    {   //  There are characters puched back - read them first. Note that THESE characters
        //  have not yet went through EOLN translation
        c = _untranslatedCharsPushedBack.pop();
        return true;
    }

    //  Read lead byte
    UInt8 leadByte;
    if (!_readByte(leadByte))
    {   //  EOF
        return false;
    }
    _encoding.setSize(1);
    _encoding[0] = static_cast<UInt8>(leadByte);
    //  Read the rest of the encoding.
    int encodingLength = _pCharacterSet->getEncodingLengthForLeadByte(_encoding[0]);
    if (encodingLength == -1)
    {   //  OOPS! Skip this unrecognized lead byte & return default character
        c = _pCharacterSet->getDefaultCharacter();
        return true;
    }
    for (int i = 1; i < encodingLength; i++)
    {
        UInt8 b;
        if (!_readByte(b))
        {   //  EOF in the middle of an encoding
            throw InvalidDataException();
        }
        _encoding.append(static_cast<UInt8>(b));
    }
    //  Decode the character
    int bytesDecoded = _pCharacterSet->decodeCharacter(_encoding.getData(), encodingLength, c);
    if (bytesDecoded != encodingLength)
    {   //  It may be that the encoding length has not been correctly predicted
        //  from the lead byte. Read up to maximum encoding length & try again
        //  (example: in UTF-16 LE it's the TAILING pair of bytes that specifies
        //  the overall encoding sequence). So read more bytes...
        int maximumEncodingLength = _pCharacterSet->getMaximumEncodingLength();
        while (static_cast<int>(_encoding.size()) < maximumEncodingLength)
        {
            UInt8 b;
            if (!_readByte(b))
            {   //  EOF in the middle of an encoding
                throw InvalidDataException();
            }
            _encoding.append(static_cast<UInt8>(b));
        }
        //  ... and try decoding again
        bytesDecoded = _pCharacterSet->decodeCharacter(_encoding.getData(), maximumEncodingLength, c);
        if (bytesDecoded == -1)
        {   //  Still can't!
            throw InvalidDataException();
        }
        //  Decoded OK, but we may have read too many bytes
        for (int i = maximumEncodingLength; i > bytesDecoded; i--)
        {
            _bytesPushedBack.push(_encoding[i - 1]);
        }
    }
    //  Done
    return true;
}

bool TextInputStream::_readByte(UInt8 & b) throw(IOException)
{
    if (!_bytesPushedBack.isEmpty())
    {
        b = _bytesPushedBack.pop();
        return true;
    }
    int ib = _pSource->read();
    if (ib == -1)
    {
        return false;
    }
    b = static_cast<UInt8>(ib);
    return true;
}

//  End of TextInputStream.cpp
