//
//  TextOutputStream.cpp - Implementation of the carl::TextOutputStream class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:55 $
//  $Revision: 1.9 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include <stdlib.h>

//////////
//  Construction/destruction
TextOutputStream::TextOutputStream(IByteOutput * pTarget)
{
    require(pTarget != NULL);

    _pTarget = pTarget;
    _pCharacterSet = CharacterSet::getDefault();
}

TextOutputStream::TextOutputStream(IByteOutput * pTarget, CharacterSet * pCharacterSet)
{
    require(pTarget != NULL);
    require(pCharacterSet != NULL);

    _pTarget = pTarget;
    _pCharacterSet = pCharacterSet;
}

//////////
//  Operations
CharacterSet * TextOutputStream::getCharacterSet() const
{
    require(isOpen());

    return _pCharacterSet;
}

void TextOutputStream::setCharacterSet(CharacterSet * pCharacterSet)
{
    require(isOpen());
    require(pCharacterSet != NULL);

    _pCharacterSet = pCharacterSet;
}

//////////
//  Operations
void TextOutputStream::flush() throw(IOException)
{
    require(isOpen());

    _pTarget->flush();
}

void TextOutputStream::write(Char c) throw(IOException)
{
    require(isOpen());

    if (c == '\n')
    {
        writeLn();
    }
    else
    {
        _write(c);
    }
}

void TextOutputStream::write(const char * pBuffer) throw(IOException)
{
    require(isOpen());
    require(pBuffer != NULL);

    for (UIntSize i = 0; pBuffer[i] != 0; i++)
    {
        write(static_cast<unsigned char>(pBuffer[i]));
    }
}

void TextOutputStream::write(const char * pBuffer, UIntSize charsToWrite) throw(IOException)
{
    require(isOpen());
    require(pBuffer != NULL);

    for (UIntSize i = 0; i < charsToWrite; i++)
    {
        write(static_cast<unsigned char>(pBuffer[i]));
    }
}

void TextOutputStream::write(const Char * pBuffer) throw(IOException)
{
    require(isOpen());
    require(pBuffer != NULL);

    for (UIntSize i = 0; pBuffer[i] != 0; i++)
    {
        write(pBuffer[i]);
    }
}

void TextOutputStream::write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException)
{
    require(isOpen());
    require(pBuffer != NULL);

    for (UIntSize i = 0; i < charsToWrite; i++)
    {
        write(pBuffer[i]);
    }
}

void TextOutputStream::write(const String & s) throw(IOException)
{
    require(isOpen());

    for (UIntSize i = 0; i < s.length(); i++)
    {
        write(s[i]);
    }
}

void TextOutputStream::writeLn() throw(IOException)
{
    require(isOpen());

    for (UIntSize i = 0; FileSystem::LineSeparator[i] != 0; i++)
    {
        _write(FileSystem::LineSeparator[i]);
    }
}

void TextOutputStream::writeLn(Char c) throw(IOException)
{
    write(c);
    writeLn();
}

void TextOutputStream::writeLn(const char * pBuffer) throw(IOException)
{
    write(pBuffer);
    writeLn();
}

void TextOutputStream::writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException)
{
    write(pBuffer, charsToWrite);
    writeLn();
}

void TextOutputStream::writeLn(const Char * pBuffer) throw(IOException)
{
    write(pBuffer);
    writeLn();
}

void TextOutputStream::writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException)
{
    write(pBuffer, charsToWrite);
    writeLn();
}

void TextOutputStream::writeLn(const String & s) throw(IOException)
{
    write(s);
    writeLn();
}

//////////
//  Implementation helpers
void TextOutputStream::_write(Char c) throw(IOException)
{
    int encodingLength = _pCharacterSet->getEncodingLengthForCharacter(c);
    if (encodingLength != -1)
    {   //  Character can be encoded "as is"
        _encoding.setSize(encodingLength);
        int encodedBytes = _pCharacterSet->encodeCharacter(c, _encoding.getData());
        if (encodedBytes == encodingLength)
        {
            _pTarget->write(_encoding.getData(), encodedBytes);
        }
        else
        {
            throw InconsistentCharacterSetException();
        }
    }
    else
    {   //  Character cannot be encoded; use default character instead
        _write(_pCharacterSet->getDefaultCharacter());
    }
}

//  End of TextOutputStream.cpp
