//
//  Thread.cpp - Implementation of the carl::Thread class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 14:34:24 $
//  $Revision: 1.11 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <process.h>
    #include <windows.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <pthread.h>
    #include <time.h>
    #include <errno.h>
#else
    #error Unsupported platform
#endif

//////////
//  Construction/destruction
Thread::Thread()
{
    _getMapThreadIdsToThreadsGuard()->enter();
    _guard.enter();
    _state = State_Initialized;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _hcrtThread = static_cast<UIntSize>(_beginthreadex(NULL, 0, _threadProc, this, CREATE_SUSPENDED, &_threadId));
    if (_hcrtThread == 0)
    {
        throw SystemError();
    }
    _getMapThreadIdsToThreads()->insert(_threadId, this);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _currentPriority = Priority_Normal;
#else
    #error Unsupported platform
#endif
    _guard.leave();
    _getMapThreadIdsToThreadsGuard()->leave();
}

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
Thread::Thread(UIntSize hcrtThread, UIntSize threadId)
{
    _getMapThreadIdsToThreadsGuard()->enter();
    _guard.enter();
    _state = State_Running;
    _hcrtThread = hcrtThread;
    _threadId = threadId;
    _guard.leave();
    _getMapThreadIdsToThreads()->insert(_threadId, this);
    _getMapThreadIdsToThreadsGuard()->leave();
}
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
Thread::Thread(UInt32 pthreadId)
{
    _getMapThreadIdsToThreadsGuard()->enter();
    _guard.enter();
    _state = State_Running;
    _pthreadId = pthreadId;
    _guard.leave();
    _getMapThreadIdsToThreads()->insert(_pthreadId, this);
    _getMapThreadIdsToThreadsGuard()->leave();
}
#else
    #error Unsupported platform
#endif

Thread::~Thread()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    _getMapThreadIdsToThreadsGuard()->enter();
    _guard.enter();
    TerminateThread(reinterpret_cast<HANDLE>(static_cast<size_t>(_hcrtThread)), 0);
    CloseHandle(reinterpret_cast<HANDLE>(static_cast<size_t>(_hcrtThread)));
    _guard.leave();
    _getMapThreadIdsToThreads()->removeKey(_threadId);
    _getMapThreadIdsToThreadsGuard()->leave();
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    _getMapThreadIdsToThreadsGuard()->enter();
    _guard.enter();
    if (_state == State_Running)
    {
        pthread_cancel((pthread_t)_pthreadId);
        _getMapThreadIdsToThreads()->removeKey(_pthreadId);
    }
    _guard.leave();
    _getMapThreadIdsToThreadsGuard()->leave();
#else
    #error Unsupported platform
#endif
}

//////////
//  Operations
Thread * Thread::getCurrentThread()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    UIntSize currentThreadId = GetCurrentThreadId();
    _getMapThreadIdsToThreadsGuard()->enter();
    if (!_getMapThreadIdsToThreads()->containsKey(currentThreadId))
    {   //  There is no Thread object that corresponds to the calling thread
        HANDLE hThread = OpenThread(THREAD_ALL_ACCESS, true, currentThreadId);
        if (hThread == NULL)
        {
            throw SystemError();
        }
        Thread * pThread = new Thread(static_cast<UIntSize>(reinterpret_cast<size_t>(hThread)), currentThreadId);
        _getMapThreadIdsToThreads()->insert(currentThreadId, pThread);
    }
    Thread * pThread = _getMapThreadIdsToThreads()->get(currentThreadId);
    _getMapThreadIdsToThreadsGuard()->leave();
    return pThread;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    UInt32 currentThreadId = (UInt32)pthread_self();
    _getMapThreadIdsToThreadsGuard()->enter();
    if (!_getMapThreadIdsToThreads()->containsKey(currentThreadId))
    {   //  There is no Thread object that corresponds to the calling thread
        Thread * pThread = new Thread(currentThreadId);
        _getMapThreadIdsToThreads()->insert(currentThreadId, pThread);
    }
    Thread * pThread = _getMapThreadIdsToThreads()->get(currentThreadId);
    _getMapThreadIdsToThreadsGuard()->leave();
    return pThread;
#else
    #error Unsupported platform
#endif
}

Thread::State Thread::getState() const
{
    _guard.enter();
    State result = _state;
    _guard.leave();
    return result;
}

Thread::Priority Thread::getPriority() const
{
    _guard.enter();
    Thread::Priority result;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    switch (GetThreadPriority(reinterpret_cast<HANDLE>(static_cast<size_t>(_hcrtThread))))
    {
        case THREAD_PRIORITY_IDLE:
            result = Priority_Idle;
            break;
        case THREAD_PRIORITY_LOWEST:
            result = Priority_Lowest;
            break;
        case THREAD_PRIORITY_BELOW_NORMAL:
            result = Priority_BelowNormal;
            break;
        case THREAD_PRIORITY_ABOVE_NORMAL:
            result = Priority_AboveNormal;
            break;
        case THREAD_PRIORITY_HIGHEST:
            result = Priority_Highest;
            break;
        case THREAD_PRIORITY_TIME_CRITICAL :
            result = Priority_TimeCritical;
            break;
        default:
            result = Priority_Normal;
            break;
    }
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    return _currentPriority;
#else
    #error Unsupported platform
#endif
    _guard.leave();
    return result;
}

void Thread::setPriority(Priority priority)
{
    _guard.enter();
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    int hostPriority;
    switch (priority)
    {
        case Priority_Idle:
            hostPriority = THREAD_PRIORITY_IDLE;
            break;
        case Priority_Lowest:
            hostPriority = THREAD_PRIORITY_LOWEST;
            break;
        case Priority_BelowNormal:
            hostPriority = THREAD_PRIORITY_BELOW_NORMAL;
            break;
        case Priority_AboveNormal:
            hostPriority = THREAD_PRIORITY_ABOVE_NORMAL;
            break;
        case Priority_Highest:
            hostPriority = THREAD_PRIORITY_HIGHEST;
            break;
        case Priority_TimeCritical:
            hostPriority = THREAD_PRIORITY_TIME_CRITICAL;
            break;
        default:
            hostPriority = THREAD_PRIORITY_NORMAL;
            break;
    }
    SetThreadPriority(reinterpret_cast<HANDLE>(static_cast<size_t>(_hcrtThread)), hostPriority);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    if (_state == State_Running)
    {
        int policy;
        struct sched_param schedParam;
        if (pthread_getschedparam((pthread_t)_pthreadId, &policy, &schedParam) != 0)
        {
            throw SystemError();
        }
        int minPriority = sched_get_priority_min(policy);
        int maxPriority = sched_get_priority_max(policy);
        int hostPriority = minPriority + ((maxPriority - minPriority) * (priority - 1) + 3) / 6;
        hostPriority = Math::max(hostPriority, minPriority);
        hostPriority = Math::min(hostPriority, maxPriority);
        schedParam.sched_priority = hostPriority;
        if (pthread_setschedparam((pthread_t)_pthreadId, policy, &schedParam) != 0)
        {
            throw SystemError();
        }
    }
    _currentPriority = priority;
#else
    #error Unsupported platform
#endif
    _guard.leave();
}

void Thread::start()
{
    _guard.enter();
    require(_state == State_Initialized);

#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    ResumeThread(reinterpret_cast<HANDLE>(static_cast<size_t>(_hcrtThread)));
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    //  Prepare thread attributes
    pthread_attr_t threadAttributes;
    switch (pthread_attr_init(&threadAttributes))
    {
        case 0:
            break;
        case ENOMEM:
            throw OutOfMemoryError(0);
        default:
            throw SystemError();
    }
    if (pthread_attr_setinheritsched(&threadAttributes, PTHREAD_INHERIT_SCHED) != 0)
    {
        throw SystemError();
    }
    switch (pthread_attr_setschedpolicy(&threadAttributes, SCHED_RR))
    {
        case 0:
            break;
        case ENOTSUP:
            //  RR scheduling not supported - but we can still run with
            //  whatever scheduling policy is used by default
            break;
        default:
            throw SystemError();
    }
    //  Create a new thread
    pthread_t ptt;
    int errorCode = pthread_create(&ptt, &threadAttributes, _threadProc, this);
    _pthreadId = (UInt32)ptt;
    _getMapThreadIdsToThreadsGuard()->enter();
    _getMapThreadIdsToThreads()->insert(_pthreadId, this);
    _getMapThreadIdsToThreadsGuard()->leave();
    pthread_attr_destroy(&threadAttributes);
    switch (errorCode)
    {
        case 0:
            break;
        case EAGAIN:
            throw OutOfMemoryError(0);
        default:
            throw SystemError();
    }
    //  Set new thread's initial priority
    setPriority(_currentPriority);
#else
    #error Unsupported platform
#endif

    _guard.leave();
}

void Thread::run()
{
}

void Thread::sleep(const TimeSpan & timeSpan)
{
    require(timeSpan.isValid());

    Int64 nanosecondsToSleep = timeSpan.getNanoseconds();
    bool forceSleep = true;
    if (nanosecondsToSleep < 0)
    {
        nanosecondsToSleep = 0;
    }
    while (nanosecondsToSleep > 0 || forceSleep)
    {
        Int64 nanosecondsInSleepChunk = nanosecondsToSleep;
        if (nanosecondsInSleepChunk > 1000000000)
        {
            nanosecondsInSleepChunk = 1000000000;
        }
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
        DWORD millisecondsToSleep = static_cast<DWORD>(nanosecondsInSleepChunk / 1000000);
        Sleep(millisecondsToSleep);
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
        struct timespec rqtp;
        rqtp.tv_sec = static_cast<time_t>(nanosecondsToSleep / 1000000000);
        rqtp.tv_nsec = static_cast<long>(nanosecondsToSleep % 1000000000);
        nanosleep(&rqtp, NULL);
#else
    #error Unsupported platform
#endif
        nanosecondsToSleep -= nanosecondsInSleepChunk;
        forceSleep = false;
    }
}

//////////
//  Implementation helpers
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
UIntSize __stdcall Thread::_threadProc(void * pParam)
{
    Thread * pThread = static_cast<Thread*>(pParam);
    //  Make sure run() is not called until start() returns
    pThread->_guard.enter();
    pThread->_state = State_Running;
    pThread->_guard.leave();
    //  Run the thread
    pThread->run();
    //  Done running
    pThread->_guard.enter();
    pThread->_state = State_Finished;
    pThread->_guard.leave();
    return 0;
}

HashMap<UIntSize, Thread*> * Thread::_getMapThreadIdsToThreads()
{
    static HashMap<UIntSize, Thread*> _mapThreadIdsToThreads(StandardHashFunctions::hashUIntSize);
    return &_mapThreadIdsToThreads;
}

CriticalSection * Thread::_getMapThreadIdsToThreadsGuard()
{
    static CriticalSection _mapThreadIdsToThreadsGuard;
    return &_mapThreadIdsToThreadsGuard;
}
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
void * Thread::_threadProc(void * pParam)
{
    Thread * pThread = static_cast<Thread*>(pParam);
    //  Make sure run() is not called until start() returns
    pThread->_guard.enter();
    pThread->_state = State_Running;
    int oldCancelState, oldCancelType;
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldCancelState);
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldCancelType);
    pThread->_guard.leave();
    //  Run the thread
    pThread->run();
    //  Done running
    _getMapThreadIdsToThreadsGuard()->enter();
    pThread->_guard.enter();
    pThread->_state = State_Finished;
    _getMapThreadIdsToThreads()->removeKey(pThread->_pthreadId);
    pThread->_guard.leave();
    _getMapThreadIdsToThreadsGuard()->leave();
    return 0;
}

HashMap<UInt32, Thread*> * Thread::_getMapThreadIdsToThreads()
{
    static HashMap<UInt32, Thread*> _mapThreadIdsToThreads(StandardHashFunctions::hashUInt32);
    return &_mapThreadIdsToThreads;
}

CriticalSection * Thread::_getMapThreadIdsToThreadsGuard()
{
    static CriticalSection _mapThreadIdsToThreadsGuard;
    return &_mapThreadIdsToThreadsGuard;
}
#else
    #error Unsupported platform
#endif

//  End of Thread.cpp
