//
//  TimeSpan.cpp - Implementation of the carl::TimeSpan class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:56 $
//  $Revision: 1.13 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    #include <windows.h>
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    #include <time.h>
#else
    #error Unsupported platform
#endif


//////////
//  Construction/destruction/assignment
TimeSpan & TimeSpan::operator = (const TimeSpan & src)
{
    if (this != (&src))
    {
        _nanoseconds = src._nanoseconds;
    }
    return *this;
}

//////////
//  Operations
TimeSpan TimeSpan::getTimeSpanSinceReferencePoint()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    LARGE_INTEGER performanceCounter;
    static LARGE_INTEGER performanceFrequency;
    static bool performanceFrequencyObtained = false;
    BOOL qpc = QueryPerformanceCounter(&performanceCounter);
    BOOL qpf;
    if (performanceFrequencyObtained)
    {
        qpf = true;
    }
    else
    {
        qpf = QueryPerformanceFrequency(&performanceFrequency);
        performanceFrequencyObtained = (qpf != 0);
    }
    Int64 nanosecondsSinceReferencePoint;
    if (qpc && qpf)
    {   //  High-performance counter is operational
        nanosecondsSinceReferencePoint = static_cast<Int64>(static_cast<double>(performanceCounter.QuadPart) * 1000000000.0 / static_cast<double>(performanceFrequency.QuadPart));
    }
    else
    {   //  Use standard date/time facilities
        FILETIME systemTimeAsFileTime;
        GetSystemTimeAsFileTime(&systemTimeAsFileTime);
        LARGE_INTEGER li;
        li.HighPart = systemTimeAsFileTime.dwHighDateTime;
        li.LowPart = systemTimeAsFileTime.dwLowDateTime;
        nanosecondsSinceReferencePoint = li.QuadPart;
    }
    return TimeSpan(nanosecondsSinceReferencePoint);
#else
    //  Use default implementation
    TimeStamp currentTime = TimeStamp::now();
    if (currentTime.isValid())
    {
        return TimeSpan(currentTime.getNanosecondsSinceEpoch());
    }
    else
    {
        return TimeSpan();
    }
#endif
}

int TimeSpan::compare(const TimeSpan & timeSpan) const
{
    if (isValid())
    {
        if (timeSpan.isValid())
        {   //  <valid time span> <compare> <valid time span>
            if (_nanoseconds < timeSpan._nanoseconds)
            {
                return -1;
            }
            if (_nanoseconds > timeSpan._nanoseconds)
            {
                return 1;
            }
            return 0;
        }
        else
        {   //  <valid time span> <compare> <invalid time span>
            return 1;
        }
    }
    else
    {
        if (timeSpan.isValid())
        {   //  <invalid time span> <compare> <valid time span>
            return -1;
        }
        else
        {   //  <invalid time span> <compare> <invalid time span>
            return 0;
        }
    }
}

TimeSpan TimeSpan::operator - () const
{
    if (!isValid())
    {
        return TimeSpan();
    }
    return TimeSpan(-_nanoseconds);
}

TimeSpan TimeSpan::operator + (Int64 nanoseconds) const
{
    if (!isValid())
    {
        return TimeSpan();
    }
    TimeSpan result(*this);
    result += nanoseconds;
    return result;
}

TimeSpan TimeSpan::operator + (const TimeSpan & timeSpan) const
{
    if (!isValid() || !timeSpan.isValid())
    {
        return TimeSpan();
    }
    TimeSpan result(*this);
    result += timeSpan;
    return result;
}

TimeSpan TimeSpan::operator - (Int64 nanoseconds) const
{
    if (!isValid())
    {
        return TimeSpan();
    }
    TimeSpan result(*this);
    result -= nanoseconds;
    return result;
}

TimeSpan TimeSpan::operator - (const TimeSpan & timeSpan) const
{
    if (!isValid() || !timeSpan.isValid())
    {
        return TimeSpan();
    }
    TimeSpan result(*this);
    result -= timeSpan;
    return result;
}

TimeSpan TimeSpan::operator * (Int64 factor) const
{
    if (!isValid())
    {
        return TimeSpan();
    }
    TimeSpan result(*this);
    result *= factor;
    return result;
}

TimeSpan TimeSpan::operator / (Int64 factor) const
{
    if (!isValid())
    {
        return TimeSpan();
    }
    TimeSpan result(*this);
    result /= factor;
    return result;
}

TimeSpan & TimeSpan::operator += (Int64 nanoseconds)
{
    if (!isValid())
    {
        return *this;
    }
    if (_nanoseconds >= 0 && nanoseconds >= 0)
    {
        _nanoseconds += nanoseconds;
        if (_nanoseconds < 0)
        {
            _nanoseconds = _InvalidValue;
        }
    }
    else if (_nanoseconds < 0 && nanoseconds < 0)
    {
        _nanoseconds += nanoseconds;
        if (_nanoseconds > 0)
        {
            _nanoseconds = _InvalidValue;
        }
    }
    else
    {
        _nanoseconds += nanoseconds;
    }
    return *this;
}

TimeSpan & TimeSpan::operator += (const TimeSpan & timeSpan)
{
    if (!isValid() || !timeSpan.isValid())
    {
        (*this) = TimeSpan();
        return *this;
    }
    return (*this) += timeSpan._nanoseconds;
}

TimeSpan & TimeSpan::operator -= (Int64 nanoseconds)
{
    if (!isValid())
    {
        return *this;
    }
    if (_nanoseconds >= 0 && nanoseconds < 0)
    {
        _nanoseconds -= nanoseconds;
        if (_nanoseconds < 0)
        {
            _nanoseconds = _InvalidValue;
        }
    }
    else if (_nanoseconds < 0 && nanoseconds >= 0)
    {
        _nanoseconds -= nanoseconds;
        if (_nanoseconds > 0)
        {
            _nanoseconds = _InvalidValue;
        }
    }
    else
    {
        _nanoseconds -= nanoseconds;
    }
    return *this;
}

TimeSpan & TimeSpan::operator -= (const TimeSpan & timeSpan)
{
    if (!isValid() || !timeSpan.isValid())
    {
        (*this) = TimeSpan();
        return *this;
    }
    return (*this) -= timeSpan._nanoseconds;
}

TimeSpan & TimeSpan::operator *= (Int64 factor)
{
    if (!isValid())
    {
        return *this;
    }
    Int64 newNanoseconds = _nanoseconds * factor;
    if (factor != 0 && newNanoseconds / factor != _nanoseconds)
    {
        _nanoseconds = _InvalidValue;
    }
    else
    {
        _nanoseconds = newNanoseconds;
    }
    return *this;
}

TimeSpan & TimeSpan::operator /= (Int64 factor)
{
    require (factor != 0);

    if (!isValid())
    {
        return *this;
    }
    _nanoseconds /= factor;
    return *this;
}

//////////
//  Hashing
UIntSize CARL_CALLBACK StandardHashFunctions::hashTimeSpan(const TimeSpan & key)
{
    return static_cast<UIntSize>(key.getNanoseconds() / 1000000000);
}

//  End of TimeSpan.cpp
