//
//  TimeStamp.cpp - Implementation of the carl::TimeStamp class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:56 $
//  $Revision: 1.10 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;
#include "DateTimeImpl.hpp"

//////////
//  Construction/destruction/assignment
TimeStamp & TimeStamp::operator = (const TimeStamp & src)
{
    if (this != (&src))
    {
        _nanosecondsSinceEpoch = src._nanosecondsSinceEpoch;
    }
    return *this;
}

//////////
//  Operations
TimeStamp TimeStamp::now()
{
#if defined(CARL_CPU_IA32) && defined(CARL_OS_WIN32) && defined(CARL_TOOLCHAIN_MSVC)
    //  Win32/MSVC++ implementation

    //  Get current time
    FILETIME currentTime;
    GetSystemTimeAsFileTime(&currentTime);

    //  Convert it into platform-neutral format
    TimeStamp timeStamp = _impl::_DateTimeManagerImpl::convertTimeStampFromWin32Format(currentTime);

    //  Return
    if (!timeStamp.isValid())
    {   //  Something wrong - try default implementation
        return _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(time(NULL));
    }
    return timeStamp;
#elif (defined(CARL_CPU_IA32) && defined(CARL_OS_CYGWIN32) && defined(CARL_TOOLCHAIN_GCC)) || \
      (defined(CARL_CPU_IA32) && defined(CARL_OS_LINUX32) && defined(CARL_TOOLCHAIN_GCC))
    time_t t = time(NULL);
    if (t == (time_t)(-1))
    {   //  Can't obtain low-precision time
        return TimeStamp();
    }
    struct timeval tv;
    if (gettimeofday(&tv, 0) == 0)
    {   //  Time obtained successfully
        return TimeStamp(static_cast<UInt64>(tv.tv_sec) * 1000000000 + static_cast<UInt64>(tv.tv_usec) * 1000);
    }
    else
    {   //  Failed to obtain time - use default implementation
        return _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(t);
    }
#else
    //  Unknown host - default implementation
    return _impl::_DateTimeManagerImpl::convertTimeStampFromCrtFormat(time(NULL));
#endif  //  Host selection
}

int TimeStamp::compare(const TimeStamp & timeStamp) const
{
    if (isValid())
    {
        if (timeStamp.isValid())
        {   //  <valid timestamp> <compare> <valid timestamp>
            if (_nanosecondsSinceEpoch < timeStamp._nanosecondsSinceEpoch)
            {
                return -1;
            }
            if (_nanosecondsSinceEpoch > timeStamp._nanosecondsSinceEpoch)
            {
                return 1;
            }
            return 0;
        }
        else
        {   //  <valid timestamp> <compare> <invalid timestamp>
            return 1;
        }
    }
    else
    {
        if (timeStamp.isValid())
        {   //  <invalid timestamp> <compare> <valid timestamp>
            return -1;
        }
        else
        {   //  <invalid timestamp> <compare> <invalid timestamp>
            return 0;
        }
    }
}

TimeStamp & TimeStamp::operator += (const TimeSpan & timeSpan)
{
    if (!isValid() || !timeSpan.isValid())
    {
        (*this) = TimeStamp();
        return *this;
    }
    Int64 ns = timeSpan.getNanoseconds();
    UInt64 newNanosecondsSinceEpoch = _nanosecondsSinceEpoch + ns;
    if (ns >= 0)
    {
        if (newNanosecondsSinceEpoch < _nanosecondsSinceEpoch)
        {
            newNanosecondsSinceEpoch = _InvalidValue;
        }
    }
    else
    {
        if (newNanosecondsSinceEpoch > _nanosecondsSinceEpoch)
        {
            newNanosecondsSinceEpoch = _InvalidValue;
        }
    }
    _nanosecondsSinceEpoch = newNanosecondsSinceEpoch;
    return *this;
}

TimeStamp & TimeStamp::operator -= (const TimeSpan & timeSpan)
{
    if (!isValid() || !timeSpan.isValid())
    {
        (*this) = TimeStamp();
        return *this;
    }
    Int64 ns = timeSpan.getNanoseconds();
    UInt64 newNanosecondsSinceEpoch = _nanosecondsSinceEpoch - ns;
    if (ns >= 0)
    {
        if (newNanosecondsSinceEpoch > _nanosecondsSinceEpoch)
        {
            newNanosecondsSinceEpoch = _InvalidValue;
        }
    }
    else
    {
        if (newNanosecondsSinceEpoch < _nanosecondsSinceEpoch)
        {
            newNanosecondsSinceEpoch = _InvalidValue;
        }
    }
    _nanosecondsSinceEpoch = newNanosecondsSinceEpoch;
    return *this;
}

TimeStamp TimeStamp::operator + (const TimeSpan & timeSpan) const
{
    TimeStamp result(*this);
    result += timeSpan;
    return result;
}

TimeStamp TimeStamp::operator - (const TimeSpan & timeSpan) const
{
    TimeStamp result(*this);
    result -= timeSpan;
    return result;
}

TimeSpan TimeStamp::operator - (const TimeStamp & timeStamp) const
{
    Int64 deltaNs;
    deltaNs = _nanosecondsSinceEpoch - timeStamp._nanosecondsSinceEpoch;
    if (_nanosecondsSinceEpoch >= timeStamp._nanosecondsSinceEpoch)
    {
        if (deltaNs < 0)
        {
            return TimeSpan();
        }
    }
    else
    {
        if (deltaNs > 0)
        {
            return TimeSpan();
        }
    }
    return TimeSpan(deltaNs);
}

//////////
//  Hashing
UIntSize CARL_CALLBACK StandardHashFunctions::hashTimeStamp(const TimeStamp & key)
{
    return static_cast<UIntSize>(key.getNanosecondsSinceEpoch() / 1000000000);
}

//  End of TimeStamp.cpp
