//
//  ULeb128.cpp - Implementation of the carl::ULeb128 class
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:56 $
//  $Revision: 1.5 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
#include "CARL.hpp"
using namespace carl;

//////////
//  Operations
void ULeb128::writeUInt64(IByteOutput & bos, UInt64 value) throw(IOException)
{
    for (; ; )
    {
        UInt8 lowByte = static_cast<UInt8>(value & 0x7F);
        value >>= 7;
        if (value != 0)
        {
            lowByte = static_cast<UInt8>(lowByte | 0x80);
        }
        bos.write(lowByte);
        if (value == 0)
        {
            break;
        }
    }
}

void ULeb128::writeBinaryInteger(IByteOutput & bos, const BinaryInteger & value) throw(IOException)
{
    require(value >= 0);

    BinaryInteger v = value;
    for (; ; )
    {
        UInt8 lowByte = static_cast<UInt8>( v[0]       |
                                           (v[1] << 1) |
                                           (v[2] << 2) |
                                           (v[3] << 3) |
                                           (v[4] << 4) |
                                           (v[5] << 5) |
                                           (v[6] << 6));
        v >>= 7;
        if (value != 0)
        {
            lowByte = static_cast<UInt8>(lowByte | 0x80);
        }
        bos.write(lowByte);
        if (v == 0)
        {
            break;
        }
    }
}

UInt8 ULeb128::readUInt8(IByteInput & bis) throw(IOException)
{
    UInt64 result = readUInt64(bis);
    if (static_cast<UInt8>(result) != result)
    {   //  Overflow
        throw InvalidDataException();
    }
    return static_cast<UInt8>(result);
}

UInt16 ULeb128::readUInt16(IByteInput & bis) throw(IOException)
{
    UInt64 result = readUInt64(bis);
    if (static_cast<UInt16>(result) != result)
    {   //  Overflow
        throw InvalidDataException();
    }
    return static_cast<UInt16>(result);
}

UInt32 ULeb128::readUInt32(IByteInput & bis) throw(IOException)
{
    UInt64 result = readUInt64(bis);
    if (static_cast<UInt32>(result) != result)
    {   //  Overflow
        throw InvalidDataException();
    }
    return static_cast<UInt32>(result);
}

UInt64 ULeb128::readUInt64(IByteInput & bis) throw(IOException)
{
    UInt64 result = 0;
    UIntSize shift = 0;
    for (; ; )
    {
        //  Get the next byte of the ULEB-128 encoding
        int b = bis.read();
        if (b == -1)
        {
            throw EOFException();
        }
        //  Process the byte just read
        UInt64 newResult = result | (static_cast<UInt64>(b & 0x7F) << shift);
        if (((static_cast<UInt64>(b & 0x7F) << shift) >> shift) != static_cast<UInt64>(b & 0x7F))
        {   //  Overflow
            throw InvalidDataException();
        }
        result = newResult;
        if ((b & 0x80) == 0)
        {
            break;
        }
        shift += 7;
    }
    return result;
}

BinaryInteger ULeb128::readBinaryInteger(IByteInput & bis) throw(IOException)
{
    BinaryInteger result = 0;
    UIntSize shift = 0;
    for (; ; )
    {
        //  Get the next byte of the ULEB-128 encoding
        int b = bis.read();
        if (b == -1)
        {
            throw EOFException();
        }
        //  Process the byte just read
        result |= BinaryInteger(b & 0x7F) << shift;
        if ((b & 0x80) == 0)
        {
            break;
        }
        shift += 7;
    }
    return result;
}

UIntSize ULeb128::readUIntSize(IByteInput & bis) throw(IOException)
{
    UInt64 result = readUInt64(bis);
    if (static_cast<UIntSize>(result) != result)
    {   //  Overflow
        throw InvalidDataException();
    }
    return static_cast<UIntSize>(result);
}

UIntPtr ULeb128::readUIntPtr(IByteInput & bis) throw(IOException)
{
    UInt64 result = readUInt64(bis);
    if (static_cast<UIntPtr>(result) != result)
    {   //  Overflow
        throw InvalidDataException();
    }
    return static_cast<UIntPtr>(result);
}

//  End of ULeb128.cpp
