//
//  Main.cpp
//
//  The CARL compression-aware streams test.
//
//  $Date: 2008/02/22 12:09:20 $
//  $Revision: 1.6 $
//  $Author: akapusti $
//
//  Copyright (C) 2006-2008 Cybernetic Intelligence GmbH
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////
#include "CARL.hpp"
using namespace carl;

class MyApp : public Application
{
    //////////
    //  Can't assign or copy construct
private:
    MyApp(const MyApp &);
    void                        operator = (const MyApp &);

    //////////
    //  Construction/destruction
public:
    MyApp() {}
    virtual ~MyApp() {}

    //////////
    //  Operations
public:
    virtual int run(const Array<String> & args);
};

MyApp theApp;

int MyApp::run(const Array<String> & /*args*/)
{
    Process::ExitCode exitCode = Process::ExitCode_Success; 

    //  Prepare random data
    const UIntSize RandomDataSize = 1024 * 1024;    //  1MB
    const UIntSize RandomDataChunkSize = 16;
    require(RandomDataSize % RandomDataChunkSize == 0);
    Array<UInt8> randomData;
    Random r;
    for (UIntSize i = 0; i < RandomDataSize / RandomDataChunkSize; i++)
    {
        UInt8 b = r.nextInt8();
        for (UIntSize j = 0; j < RandomDataChunkSize; j++)
        {
            randomData.append(b);
        }
    }

    //  Exercise all compression algorithms
    Array<CompressionAlgorithm*> allCompressionAlgorithms = CompressionAlgorithm::getAllInstances();
    for (UIntSize i = 0; i < allCompressionAlgorithms.size(); i++)
    {
        CompressionAlgorithm * pCompressionAlgorithm = allCompressionAlgorithms[i];
        StandardStreams::out.write("Testing " + pCompressionAlgorithm->getDisplayName() + "...");
        //  Prepare output streams
        MemoryOutputStream mos;
        CompressOutputStream cos(&mos, pCompressionAlgorithm->getCompressor());
        //  Write!
        TimeSpan beginWriteTime = TimeSpan::getTimeSpanSinceReferencePoint();
        cos.write(randomData.getData(), randomData.size());
        TimeSpan endWriteTime = TimeSpan::getTimeSpanSinceReferencePoint();
        cos.close();
        Array<UInt8> compressedData = mos.getContents();
        //  Prepare input streams
        MemoryInputStream mis(compressedData.getData(), compressedData.size(), false);
        DecompressInputStream dis(&mis, pCompressionAlgorithm->getDecompressor());
        //  Read!
        Array<UInt8> decompressedData;
        TimeSpan beginReadTime = TimeSpan::getTimeSpanSinceReferencePoint();
        for (; ; )
        {
            int b = dis.read();
            if (b == -1)
            {
                break;
            }
            decompressedData.append(static_cast<UInt8>(b));
        }
        TimeSpan endReadTime = TimeSpan::getTimeSpanSinceReferencePoint();
        //  Compare
        bool dataMismatch = false;
        if (decompressedData.size() != randomData.size())
        {
            dataMismatch = true;
        }
        for (UIntSize j = 0; j < Math::min(decompressedData.size(), randomData.size()); j++)
        {
            if (decompressedData[j] != randomData[j])
            {
                dataMismatch = true;
            }
        }
        if (dataMismatch)
        {
            StandardStreams::out.writeLn("Failed");
            exitCode = Process::ExitCode_Failure;
        }
        else
        {
            TimeSpan readTime = endReadTime - beginReadTime;
            if (readTime.getNanoseconds() == 0)
            {
                readTime = TimeSpan(1);
            }
            TimeSpan writeTime = endWriteTime - beginWriteTime;
            if (writeTime.getNanoseconds() == 0)
            {
                writeTime = TimeSpan(1);
            }
            StandardStreams::out.writeLn("OK, compression speed " +
                                         NumberFormat::getInstance()->format((RandomDataSize * CARL_UINT64_CONSTANT(1000000000)) / (writeTime.getNanoseconds())) +
                                         " bytes/sec, decompression speed " +
                                         NumberFormat::getInstance()->format((RandomDataSize * CARL_UINT64_CONSTANT(1000000000)) / (readTime.getNanoseconds())) +
                                         " bytes/sec, compression ratio " +
                                         NumberFormat::getInstance()->format(compressedData.size() * 100 / decompressedData.size()) +
                                         "%");
        }
    }

    return exitCode;
}

//  End of Main.cpp
