//
//  Main.cpp
//
//  The CARL encryption-aware streams test.
//
//  $Date: 2008/02/22 12:09:20 $
//  $Revision: 1.7 $
//  $Author: akapusti $
//
//  Copyright (C) 2006-2008 Cybernetic Intelligence GmbH
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////
#include "CARL.hpp"
using namespace carl;

class MyApp : public Application
{
    //////////
    //  Can't assign or copy construct
private:
    MyApp(const MyApp &);
    void                        operator = (const MyApp &);

    //////////
    //  Construction/destruction
public:
    MyApp() {}
    virtual ~MyApp() {}

    //////////
    //  Operations
public:
    virtual int run(const Array<String> & args);
};

MyApp theApp;

int MyApp::run(const Array<String> & /*args*/)
{
    Process::ExitCode exitCode = Process::ExitCode_Success; 

    //  Prepare random data
    const UIntSize RandomDataSize = 1024;    //  1KB
    Array<UInt8> randomData;
    Random r;
    for (UIntSize i = 0; i < RandomDataSize; i++)
    {
        randomData.append(r.nextInt8());
    }

    //  Exercise all ciphers
    Array<Cipher*> allCiphers = Cipher::getAllInstances();
    PaddingAlgorithm * pPaddingAlgorithm = PaddingAlgorithm::getAllInstances()[0];
    for (UIntSize i = 0; i < allCiphers.size(); i++)
    {
        Cipher * pCipher = allCiphers[i];
        StandardStreams::out.write("Testing " + pCipher->getDisplayName() + "...");
        //  Generate key(s) & prepare algorithms
        Encryptor * pEncryptor = NULL;
        Decryptor * pDecryptor = NULL;
        if (dynamic_cast<SymmetricCipher*>(pCipher) != NULL)
        {
            SymmetricCipher * pSymmetricCipher = dynamic_cast<SymmetricCipher*>(pCipher);
            UIntSize keyLength = pSymmetricCipher->getSupportedKeyLenghts()[0];
            SecretKey key = pSymmetricCipher->generateRandomKey(keyLength);
            pEncryptor = pCipher->createEncryptor(key);
            pDecryptor = pCipher->createDecryptor(key);
        }
        else if (dynamic_cast<AsymmetricCipher*>(pCipher) != NULL)
        {
            AsymmetricCipher * pAsymmetricCipher = dynamic_cast<AsymmetricCipher*>(pCipher);
            Pair<UIntSize, UIntSize> keyPairLength = pAsymmetricCipher->getSupportedKeyPairLenghts()[0];
            KeyPair keys = pAsymmetricCipher->generateRandomKeyPair(keyPairLength.first(), keyPairLength.second());
            pEncryptor = pCipher->createEncryptor(keys.getPublicKey());
            pDecryptor = pCipher->createDecryptor(keys.getPrivateKey());
        }
        else
        {
            require(false);
        }
        //  Prepare output streams
        MemoryOutputStream mos;
        EncryptOutputStream * pEos = NULL;
        if (dynamic_cast<StreamCipher*>(pCipher) != NULL)
        {
            pEos = new EncryptOutputStream(&mos, dynamic_cast<StreamEncryptor*>(pEncryptor), true);
        }
        else if (dynamic_cast<BlockCipher*>(pCipher) != NULL)
        {
            pEos = new EncryptOutputStream(&mos, dynamic_cast<BlockEncryptor*>(pEncryptor), pPaddingAlgorithm, true);
        }
        else
        {
            require(false);
        }
        //  Write!
        TimeSpan beginWriteTime = TimeSpan::getTimeSpanSinceReferencePoint();
        pEos->write(randomData.getData(), randomData.size());
        TimeSpan endWriteTime = TimeSpan::getTimeSpanSinceReferencePoint();
        delete pEos;
        Array<UInt8> encryptedData = mos.getContents();
        //  Prepare input streams
        MemoryInputStream mis(encryptedData.getData(), encryptedData.size(), false);
        DecryptInputStream * pEis = NULL;
        if (dynamic_cast<StreamCipher*>(pCipher) != NULL)
        {
            pEis = new DecryptInputStream(&mis, dynamic_cast<StreamDecryptor*>(pDecryptor), true);
        }
        else if (dynamic_cast<BlockCipher*>(pCipher) != NULL)
        {
            pEis = new DecryptInputStream(&mis, dynamic_cast<BlockDecryptor*>(pDecryptor), pPaddingAlgorithm, true);
        }
        else
        {
            require(false);
        }
        //  Read!
        Array<UInt8> decryptedData;
        TimeSpan beginReadTime = TimeSpan::getTimeSpanSinceReferencePoint();
        for (; ; )
        {
            int b = pEis->read();
            if (b == -1)
            {
                break;
            }
            decryptedData.append(static_cast<UInt8>(b));
        }
        TimeSpan endReadTime = TimeSpan::getTimeSpanSinceReferencePoint();
        delete pEis;
        //  Compare
        bool dataMismatch = false;
        if (decryptedData.size() != randomData.size())
        {
            dataMismatch = true;
        }
        for (UIntSize j = 0; j < Math::min(decryptedData.size(), randomData.size()); j++)
        {
            if (decryptedData[j] != randomData[j])
            {
                dataMismatch = true;
            }
        }
        if (dataMismatch)
        {
            StandardStreams::out.writeLn("Failed");
            exitCode = Process::ExitCode_Failure;
        }
        else
        {
            TimeSpan readTime = endReadTime - beginReadTime;
            if (readTime.getNanoseconds() == 0)
            {
                readTime = TimeSpan(1);
            }
            TimeSpan writeTime = endWriteTime - beginWriteTime;
            if (writeTime.getNanoseconds() == 0)
            {
                writeTime = TimeSpan(1);
            }
            StandardStreams::out.writeLn("OK, read speed " +
                                         NumberFormat::getInstance()->format((RandomDataSize * CARL_UINT64_CONSTANT(1000000000)) / (readTime.getNanoseconds())) +
                                         " bytes/sec, write speed " +
                                         NumberFormat::getInstance()->format((RandomDataSize * CARL_UINT64_CONSTANT(1000000000)) / (writeTime.getNanoseconds())) +
                                         " bytes/sec");
        }
    }

    return exitCode;
}

//  End of Main.cpp
