//
//  Main.cpp
//
//  The real NumberFormat/NumberParse test.
//
//  $Date: 2008/09/12 11:36:17 $
//  $Revision: 1.1 $
//  $Author: akapusti $
//
//  Copyright (C) 2008 Cybernetic Intelligence GmbH
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////
#include "CARL.hpp"
using namespace carl;

class MyApp : public Application
{
    //////////
    //  Can't assign or copy construct
private:
    MyApp(const MyApp &);
    void                        operator = (const MyApp &);

    //////////
    //  Construction/destruction
public:
    MyApp() {}
    virtual ~MyApp() {}

    //////////
    //  Operations
public:
    virtual int run(const Array<String> & args);
};

MyApp theApp;

#define CHECK(x) if (!(x)) { exitCode = Process::ExitCode_Failure; }
#define CLOSE(x,y) CHECK(Math::abs((x - y) / x) < 1e-4)

int MyApp::run(const Array<String> & /*args*/)
{
    Process::ExitCode exitCode = Process::ExitCode_Success; 

    Random r;
    NumberFormat * pNumberFormat = NumberFormat::getInstance(Locale());
    NumberParse * pNumberParse = NumberParse::getInstance(Locale());

    //  Default formatting (decimal)
    {
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(123.456)).startsWith("123.456"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(-123.456)).startsWith("-123.456"))

        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(123.456)).startsWith("123.456"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(-123.456)).startsWith("-123.456"))
    }

    //  Fixed-point formatting (decimal)
    {
        RealFormatProperties rfp;
        rfp.setFormat(RealFormatProperties::Format_FixedPoint);
        
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(1.23456), rfp).startsWith("1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(0.000123456), rfp).startsWith("0.000123456"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(12345600000), rfp).startsWith("12345600000"))

        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(1.23456), rfp).startsWith("1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(0.000123456), rfp).startsWith("0.000123456"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(12345600000), rfp).startsWith("12345600000"))
    }

    //  Fixed-point formatting with specified fractional digits
    {
        RealFormatProperties rfp;
        rfp.setFormat(RealFormatProperties::Format_FixedPoint)
           .setMinimumFractionalDigits(3)
           .setMaximumFractionalDigits(3);

        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(1.23456), rfp) == "1.235")
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(0.000123456), rfp) == "0.000")
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(12345600000), rfp) == "12345600000.000")

        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(1.23456), rfp) == "1.235")
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(0.000123456), rfp) == "0.000")
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(12345600000), rfp) == "12345600000.000")

        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(-1.23456), rfp) == "-1.235")
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(-0.000123456), rfp) == "-0.000")
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(-12345600000), rfp) == "-12345600000.000")

        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(-1.23456), rfp) == "-1.235")
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(-0.000123456), rfp) == "-0.000")
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(-12345600000), rfp) == "-12345600000.000")
    }

    //  Fixed-point formatting with mandatory sign
    {
        RealFormatProperties rfp;
        rfp.setFormat(RealFormatProperties::Format_FixedPoint)
           .setMandatorySign(true);
        
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(1.23456), rfp).startsWith("+1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(0.000123456), rfp).startsWith("+0.000123456"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(12345600000), rfp).startsWith("+12345600000"))

        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(1.23456), rfp).startsWith("+1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(0.000123456), rfp).startsWith("+0.000123456"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(12345600000), rfp).startsWith("+12345600000"))

        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(-1.23456), rfp).startsWith("-1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(-0.000123456), rfp).startsWith("-0.000123456"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(-12345600000), rfp).startsWith("-12345600000"))

        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(-1.23456), rfp).startsWith("-1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(-0.000123456), rfp).startsWith("-0.000123456"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(-12345600000), rfp).startsWith("-12345600000"))
    }

    //  Floating-point formatting (decimal)
    {
        RealFormatProperties rfp;
        rfp.setFormat(RealFormatProperties::Format_FloatingPoint)
           .setMinimumExponentDigits(3)
           .setMaximumExponentDigits(3);

        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(1.23456), rfp).startsWith("1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(1.23456), rfp).endsWith("e+000"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(0.000123456), rfp).startsWith("1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(0.000123456), rfp).endsWith("e-004"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(12345600000), rfp).startsWith("1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(12345600000), rfp).endsWith("e+010"))

        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(1.23456), rfp).startsWith("1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(1.23456), rfp).endsWith("e+000"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(0.000123456), rfp).startsWith("1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(0.000123456), rfp).endsWith("e-004"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(12345600000), rfp).startsWith("1.23456"))
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(12345600000), rfp).endsWith("e+010"))
    }

    //  Fixed-point formatting with specified precision
    {
        RealFormatProperties rfp;
        rfp.setFormat(RealFormatProperties::Format_FloatingPoint)
           .setMinimumDigits(5)
           .setMaximumDigits(5)
           .setMinimumExponentDigits(3)
           .setMaximumExponentDigits(3);

        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(1.23456), rfp) == "1.2346e+000")
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(0.000123456), rfp) == "1.2346e-004")
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(12345600000), rfp) == "1.2346e+010")

        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(1.23456), rfp) == "1.2346e+000")
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(0.000123456), rfp) == "1.2346e-004")
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(12345600000), rfp) == "1.2346e+010")

        rfp.setLetterCase(RealFormatProperties::LetterCase_Upper);
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(-1.23456), rfp) == "-1.2346E+000")
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(-0.000123456), rfp) == "-1.2346E-004")
        CHECK(pNumberFormat->format(CARL_REAL32_CONSTANT(-12345600000), rfp) == "-1.2346E+010")

        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(-1.23456), rfp) == "-1.2346E+000")
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(-0.000123456), rfp) == "-1.2346E-004")
        CHECK(pNumberFormat->format(CARL_REAL64_CONSTANT(-12345600000), rfp) == "-1.2346E+010")
    }

    //  Default parsing
    for (Real64 x = 1.2345e-20; x <= 1e+20; x *= 8)
    {
        Real64 y = pNumberParse->parseReal64(pNumberFormat->format(x));
        CLOSE(x, y)
    }

    //  Parsing with delimiters
    for (Real64 x = 1.2345e-20; x <= 1e+20; x *= 8)
    {
        RealFormatProperties rfp;
        rfp.setUseSeparator(true);
        RealParseProperties rpp;
        rpp.setAllowSeparator(true);

        Real64 y = pNumberParse->parseReal64(pNumberFormat->format(x, rfp), rpp);
        CLOSE(x, y)
    }

    //  Hexadecimal parsing
    for (Real64 x = 1.2345e-20; x <= 1e+20; x *= 8)
    {
        RealFormatProperties rfp;
        rfp.setRadix(16);
        RealParseProperties rpp;
        rpp.setRadix(16);

        Real64 y = pNumberParse->parseReal64(pNumberFormat->format(x, rfp), rpp);
        CLOSE(x, y)
    }

    //  Arbitrary precision formatting & parsing
    for (DecimalFixed x = 1.2345e-20; x <= 1e+20; x *= 8)
    {
        RealFormatProperties rfp;
        rfp.setFormat(RealFormatProperties::Format_FixedPoint);

        DecimalFixed y = pNumberParse->parseDecimalFixed(pNumberFormat->format(x, rfp));
        CLOSE(x, y)
    }

    for (BinaryFixed x = 1.2345e-20; x <= 1e+20; x *= 8)
    {
        RealFormatProperties rfp;
        rfp.setFormat(RealFormatProperties::Format_FixedPoint);

        BinaryFixed y = pNumberParse->parseBinaryFixed(pNumberFormat->format(x, rfp));
        CLOSE(x, y)
    }

    DecimalFloat::setDefaultDivisionPrecision(16);
    for (DecimalFloat x = 1.2345e-20; x <= 1e+20; x *= 8)
    {
        DecimalFloat y = pNumberParse->parseDecimalFloat(pNumberFormat->format(x));
        CLOSE(x, y)
    }

    BinaryFloat::setDefaultDivisionPrecision(16);
    for (BinaryFloat x = 1.2345e-20; x <= 1e+20; x *= 8)
    {
        BinaryFloat y = pNumberParse->parseBinaryFloat(pNumberFormat->format(x));
        CLOSE(x, y)
    }

    //  Done
    return exitCode;
}

//  End of Main.cpp
