//
//  Main.cpp
//
//  The TimeFormat/TimeParse test.
//
//  $Date: 2008/08/09 14:13:20 $
//  $Revision: 1.1 $
//  $Author: akapusti $
//
//  Copyright (C) 2008 Cybernetic Intelligence GmbH
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////
#include "CARL.hpp"
using namespace carl;

class MyApp : public Application
{
    //////////
    //  Can't assign or copy construct
private:
    MyApp(const MyApp &);
    void                        operator = (const MyApp &);

    //////////
    //  Construction/destruction
public:
    MyApp() {}
    virtual ~MyApp() {}

    //////////
    //  Operations
public:
    virtual int run(const Array<String> & args);
};

MyApp theApp;

namespace
{
    bool match(Time t1, Time t2, TimeFormat::Precision precision)
    {
        switch (precision)
        {
            case TimeFormat::Precision_Minute:
                return t1.getHour() == t2.getHour() &&
                       t1.getMinute() == t2.getMinute();
            case TimeFormat::Precision_Second:
                return t1.getHour() == t2.getHour() &&
                       t1.getMinute() == t2.getMinute() &&
                       t1.getSecond() == t2.getSecond();
            case TimeFormat::Precision_Millisecond:
                return t1.getHour() == t2.getHour() &&
                       t1.getMinute() == t2.getMinute() &&
                       t1.getSecond() == t2.getSecond() &&
                       (t1.getNanosecond() / 1000000) == (t2.getNanosecond() / 1000000);
            case TimeFormat::Precision_Microsecond:
                return t1.getHour() == t2.getHour() &&
                       t1.getMinute() == t2.getMinute() &&
                       t1.getSecond() == t2.getSecond() &&
                       (t1.getNanosecond() / 1000) == (t2.getNanosecond() / 1000);
            case TimeFormat::Precision_Nanosecond:
                return t1.getHour() == t2.getHour() &&
                       t1.getMinute() == t2.getMinute() &&
                       t1.getSecond() == t2.getSecond() &&
                       t1.getNanosecond() == t2.getNanosecond();
            default:
                require(false);
                return false;
        }
    }
}

#define CHECK(x) if (!(x)) { exitCode = Process::ExitCode_Failure; }

int MyApp::run(const Array<String> & /*args*/)
{
    Process::ExitCode exitCode = Process::ExitCode_Success; 

    Random r;
    TimeFormat * pTimeFormat = TimeFormat::getInstance(Locale());
    TimeParse * pTimeParse = TimeParse::getInstance(Locale());

    //  Standard formats
    for (int hour = 0; hour < 24; hour += r.nextInt32(1, 7))
    {
        for (int minute = 0; minute < 60; minute += r.nextInt32(1, 60))
        {
            StandardStreams::out.writeLn("Testing standard formatting for " + 
                                         NumberFormat::getInstance()->format(hour) + ":" +
                                         NumberFormat::getInstance()->format(minute));
            for (int second = 0; second < 60; second += r.nextInt32(1, 60))
            {
                for (int nanosecond = 0; nanosecond < 1000000000; nanosecond += r.nextInt32(1, 1000000000))
                {
                    Time t(hour, minute, second, nanosecond);
                    CHECK(match(pTimeParse->parse(pTimeFormat->format(t, TimeFormat::Precision_Minute)), t, TimeFormat::Precision_Minute))
                    CHECK(match(pTimeParse->parse(pTimeFormat->format(t, TimeFormat::Precision_Second)), t, TimeFormat::Precision_Second))
                    CHECK(match(pTimeParse->parse(pTimeFormat->format(t, TimeFormat::Precision_Millisecond)), t, TimeFormat::Precision_Millisecond))
                    CHECK(match(pTimeParse->parse(pTimeFormat->format(t, TimeFormat::Precision_Microsecond)), t, TimeFormat::Precision_Microsecond))
                    CHECK(match(pTimeParse->parse(pTimeFormat->format(t, TimeFormat::Precision_Nanosecond)), t, TimeFormat::Precision_Nanosecond))
                }
            }
        }
    }

    //  Custom formats
    for (int hour = 0; hour < 24; hour += r.nextInt32(1, 7))
    {
        for (int minute = 0; minute < 60; minute += r.nextInt32(1, 60))
        {
            StandardStreams::out.writeLn("Testing custom formatting for " + 
                                         NumberFormat::getInstance()->format(hour) + ":" +
                                         NumberFormat::getInstance()->format(minute));
            for (int second = 0; second < 60; second += r.nextInt32(1, 60))
            {
                for (int nanosecond = 0; nanosecond < 1000000000; nanosecond += r.nextInt32(1, 1000000000))
                {
                    Time t(hour, minute, second, nanosecond);
                    String formatString1("{hour24:full}{minute:full}{second:full}{nanosecond:full}");
                    String formatString2("{hour24:full} {minute:full} {second:full} {nanosecond:full}");
                    String formatString3("{hour24:full}:{minute:full}:{second:full}:{nanosecond:full}");
                    String formatString4("[{hour24:full},{minute:full},{second:full},{nanosecond:full}]");
                    String formatString5("{hour12:full}{minute:full}{second:full}{nanosecond:full}{ampm}");
                    String formatString6("{hour12:full} {minute:full} {second:full} {nanosecond:full}{ampm}");
                    String formatString7("{hour12:full}:{minute:full}:{second:full}:{nanosecond:full}{ampm}");
                    String formatString8("[{hour12:full},{minute:full},{second:full},{nanosecond:full}]{ampm}");
                    CHECK(pTimeParse->parse(pTimeFormat->format(t, formatString1), formatString1) == t)
                    CHECK(pTimeParse->parse(pTimeFormat->format(t, formatString2), formatString2) == t)
                    CHECK(pTimeParse->parse(pTimeFormat->format(t, formatString3), formatString3) == t)
                    CHECK(pTimeParse->parse(pTimeFormat->format(t, formatString4), formatString4) == t)
                    CHECK(pTimeParse->parse(pTimeFormat->format(t, formatString5), formatString5) == t)
                    CHECK(pTimeParse->parse(pTimeFormat->format(t, formatString6), formatString6) == t)
                    CHECK(pTimeParse->parse(pTimeFormat->format(t, formatString7), formatString7) == t)
                    CHECK(pTimeParse->parse(pTimeFormat->format(t, formatString8), formatString8) == t)
                }
            }
        }
    }

    //  Done
    return exitCode;
}

//  End of Main.cpp
