//
//  Main.cpp
//
//  The CARL resource framework test.
//
//  $Date: 2008/02/26 11:08:27 $
//  $Revision: 1.1 $
//  $Author: akapusti $
//
//  Copyright (C) 2008 Cybernetic Intelligence GmbH
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////
#include "CARL.hpp"
using namespace carl;

class MyApp : public Application
{
    //////////
    //  Can't assign or copy construct
private:
    MyApp(const MyApp &);
    void                        operator = (const MyApp &);

    //////////
    //  Construction/destruction
public:
    MyApp() {}
    virtual ~MyApp() {}

    //////////
    //  Operations
public:
    virtual int run(const Array<String> & args);
};

MyApp theApp;

#define CHECK(condition) if (!(condition)) { exitCode = Process::ExitCode_Failure; }

class MyResourceKeys
{
    //////////
    //  Can't instantiate
private:
    MyResourceKeys();
    MyResourceKeys(const MyResourceKeys &);
    void                        operator = (const MyResourceKeys &);
    ~MyResourceKeys();

    //////////
    //  Resource keys
public:
    static const char *         StringKey;
    static const char *         StringWithPlaceholdersKey;
    static const char *         BadKey;
};

const char * MyResourceKeys::StringKey = "MyResourceKeys::StringKey";
const char * MyResourceKeys::StringWithPlaceholdersKey = "MyResourceKeys::StringWithPlaceholdersKey";
const char * MyResourceKeys::BadKey = "MyResourceKeys::BadKey";

class MyResourceMap_en : public ResourceMap
{
    //////////
    //  Can't assign or copy construct
private:
    MyResourceMap_en(const MyResourceMap_en &);
    void                        operator = (const MyResourceMap_en &);

    //////////
    //  Construction/destruction
public:
    MyResourceMap_en()
        :   _mapKeysToValues(StandardHashFunctions::hashString)
    {
        _mapKeysToValues.insert(String(MyResourceKeys::StringKey), String("Init"));
        _mapKeysToValues.insert(String(MyResourceKeys::StringWithPlaceholdersKey), String("Init $0 with $1"));
    }
    virtual ~MyResourceMap_en() {}

    //////////
    //  Operations
public:
    virtual String              getString(const String & key)
    {
        if (_mapKeysToValues.containsKey(key))
        {
            return _mapKeysToValues.get(key);
        }
        throw MissingResourceError(key);
    }

    //////////
    //  Implementation
private:
    HashMap<String, String>     _mapKeysToValues;
};

class MyResourceMap_en_GB : public ResourceMap
{
    //////////
    //  Can't assign or copy construct
private:
    MyResourceMap_en_GB(const MyResourceMap_en_GB &);
    void                        operator = (const MyResourceMap_en_GB &);

    //////////
    //  Construction/destruction
public:
    MyResourceMap_en_GB()
        :   _mapKeysToValues(StandardHashFunctions::hashString)
    {
        _mapKeysToValues.insert(String(MyResourceKeys::StringKey), String("Initialise"));
        _mapKeysToValues.insert(String(MyResourceKeys::StringWithPlaceholdersKey), String("Initialise $0 with $1"));
    }
    virtual ~MyResourceMap_en_GB() {}

    //////////
    //  Operations
public:
    virtual String              getString(const String & key)
    {
        if (_mapKeysToValues.containsKey(key))
        {
            return _mapKeysToValues.get(key);
        }
        throw MissingResourceError(key);
    }

    //////////
    //  Implementation
private:
    HashMap<String, String>     _mapKeysToValues;
};

class MyResourceMap_en_US : public ResourceMap
{
    //////////
    //  Can't assign or copy construct
private:
    MyResourceMap_en_US(const MyResourceMap_en_US &);
    void                        operator = (const MyResourceMap_en_US &);

    //////////
    //  Construction/destruction
public:
    MyResourceMap_en_US()
        :   _mapKeysToValues(StandardHashFunctions::hashString)
    {
        _mapKeysToValues.insert(String(MyResourceKeys::StringKey), String("Initialize"));
        _mapKeysToValues.insert(String(MyResourceKeys::StringWithPlaceholdersKey), String("Initialize $0 with $1"));
    }
    virtual ~MyResourceMap_en_US() {}

    //////////
    //  Operations
public:
    virtual String              getString(const String & key)
    {
        if (_mapKeysToValues.containsKey(key))
        {
            return _mapKeysToValues.get(key);
        }
        throw MissingResourceError(key);
    }

    //////////
    //  Implementation
private:
    HashMap<String, String>     _mapKeysToValues;
};

int MyApp::run(const Array<String> & /*args*/)
{
    Process::ExitCode exitCode = Process::ExitCode_Success;

    MapResourceFactory rf(new MyResourceMap_en());
    rf.addResourceMap(Locale("en", "GB"), new MyResourceMap_en_GB());
    rf.addResourceMap(Locale("en", "US"), new MyResourceMap_en_US());

    //  Test locale-seisitivity of existing resources
    CHECK(rf.getString(MyResourceKeys::StringKey, Locale()) == "Init")
    CHECK(rf.getString(MyResourceKeys::StringKey, Locale("en")) == "Init")
    CHECK(rf.getString(MyResourceKeys::StringKey, Locale("en", "GB")) == "Initialise")
    CHECK(rf.getString(MyResourceKeys::StringKey, Locale("en", "US")) == "Initialize")
    CHECK(rf.getString(MyResourceKeys::StringKey, Locale("de")) == "Init")

    //  Test placeholder handling
    StandardStreams::out.writeLn("'" + rf.getParameterizedString(MyResourceKeys::StringWithPlaceholdersKey, String("X"), String("Y"), Locale()) + "'");
    CHECK(rf.getParameterizedString(MyResourceKeys::StringWithPlaceholdersKey, String("X"), String("Y"), Locale()) == "Init X with Y")
    CHECK(rf.getParameterizedString(MyResourceKeys::StringWithPlaceholdersKey, String("X"), String("Y"), Locale("en")) == "Init X with Y")
    CHECK(rf.getParameterizedString(MyResourceKeys::StringWithPlaceholdersKey, String("X"), String("Y"), Locale("en", "GB")) == "Initialise X with Y")
    CHECK(rf.getParameterizedString(MyResourceKeys::StringWithPlaceholdersKey, String("X"), String("Y"), Locale("en", "US")) == "Initialize X with Y")
    CHECK(rf.getParameterizedString(MyResourceKeys::StringWithPlaceholdersKey, String("X"), String("Y"), Locale("de")) == "Init X with Y")
    
    //  Test treatment of nonexistent resources
    try
    {
        String str = rf.getString(MyResourceKeys::BadKey);
        CHECK(false)
    }
    catch (MissingResourceError)
    {
    }

    return exitCode;
}

//  End of Main.cpp
