//
//  Main.cpp
//
//  The CARL I/O performance test.
//
//  $Date: 2007/06/24 15:45:15 $
//  $Revision: 1.3 $
//  $Author: akapusti $
//
//  Copyright (C) 2006 Cybernetic Intelligence GmbH
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////
#include "CARL.hpp"
using namespace carl;
#include <stdio.h>

class MyApp : public Application
{
    //////////
    //  Can't assign or copy construct
private:
    MyApp(const MyApp &);
    void                        operator = (const MyApp &);

    //////////
    //  Construction/destruction
public:
    MyApp() {}
    virtual ~MyApp() {}

    //////////
    //  Operations
public:
    virtual int run(const Array<String> & args);
};

MyApp theApp;

int MyApp::run(const Array<String> & args)
{
    const int Iterations = 5;
    const int Attempts = 2;
    const int Bytes = 1000000;
    const int Blocks = 100000;
    const int BlockSize = 100;

    //  Character...
    StandardStreams::out.writeLn("*** Testing character I/O speed...");
    for (int i = 0; i < Iterations; i++)
    {
        TimeSpan startTime, endTime;
        //  CRT I/O
        for (int j = 0; j < Attempts; j++)
        {
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            FILE * f = fopen("crt_chars.dat", "w");
            for (int j = 0; j < Bytes; j++)
            {
                fputc(j, f);
            }
            fclose(f);
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CRT I/O (character write) : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
            //
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            f = fopen("crt_chars.dat", "r");
            for (int j = 0; j < Bytes; j++)
            {
                fgetc(f);
            }
            fclose(f);
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CRT I/O (character read)  : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
        }
        //  CARL I/O
        for (int j = 0; j < Attempts; j++)
        {
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            TextFileWriter w("carl_chars.dat");
            w.setCharacterSet(CharacterSet::getInstance("iso-8859-1"));
            for (int j = 0; j < Bytes; j++)
            {
                w.write(static_cast<char>(j));
            }
            w.close();
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CARL I/O (character write): " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
            //
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            TextFileReader r("carl_chars.dat");
            for (int j = 0; j < Bytes; j++)
            {
                Char c;
                r.read(c);
            }
            r.close();
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CARL I/O (character read) : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
        }
    }
    StandardStreams::out.writeLn("*** Done");
    StandardStreams::out.writeLn();
    FileSystem::deleteFile("crt_chars.dat");
    FileSystem::deleteFile("carl_chars.dat");

    //  Byte...
    StandardStreams::out.writeLn("*** Testing byte I/O speed...");
    for (int i = 0; i < Iterations; i++)
    {
        TimeSpan startTime, endTime;
        //  CRT I/O
        for (int j = 0; j < Attempts; j++)
        {
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            FILE * f = fopen("crt_bytes.dat", "wb");
            for (int j = 0; j < Bytes; j++)
            {
                fputc(j, f);
            }
            fclose(f);
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CRT I/O (byte write)      : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
            //
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            f = fopen("crt_bytes.dat", "rb");
            for (int j = 0; j < Bytes; j++)
            {
                fgetc(f);
            }
            fclose(f);
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CRT I/O (byte read)       : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
        }
        //  CARL I/O
        for (int j = 0; j < Attempts; j++)
        {
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            BinaryFileWriter w("carl_bytes.dat");
            for (int j = 0; j < Bytes; j++)
            {
                w.write(static_cast<UInt8>(j));
            }
            w.close();
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CARL I/O (byte write)     : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
            //
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            BinaryFileReader r("carl_bytes.dat");
            for (int j = 0; j < Bytes; j++)
            {
                r.read();
            }
            r.close();
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CARL I/O (byte read)      : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
        }
    }
    StandardStreams::out.writeLn("*** Done");
    StandardStreams::out.writeLn();
    FileSystem::deleteFile("crt_bytes.dat");
    FileSystem::deleteFile("carl_bytes.dat");

    //  Block...
    StandardStreams::out.writeLn("*** Testing block I/O speed...");
    for (int i = 0; i < Iterations; i++)
    {
        TimeSpan startTime, endTime;
        static UInt8 data[BlockSize];

        //  CRT I/O
        for (int j = 0; j < Attempts; j++)
        {
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            FILE * f = fopen("crt_blocks.dat", "wb");
            for (int j = 0; j < Blocks; j++)
            {
                fwrite(data, BlockSize, 1, f);
            }
            fclose(f);
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CRT I/O (block write)     : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
            //
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            f = fopen("crt_blocks.dat", "rb");
            for (int j = 0; j < Blocks; j++)
            {
                fread(data, BlockSize, 1, f);
            }
            fclose(f);
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CRT I/O (block read)      : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
        }
        //  CARL I/O
        for (int j = 0; j < Attempts; j++)
        {
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            BinaryFileWriter w("carl_blocks.dat");
            for (int j = 0; j < Blocks; j++)
            {
                w.write(data, BlockSize);
            }
            w.close();
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CARL I/O (block write)    : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
            //
            startTime = TimeSpan::getTimeSpanSinceReferencePoint();
            BinaryFileReader r("carl_blocks.dat");
            for (int j = 0; j < Blocks; j++)
            {
                r.read(data, BlockSize);
            }
            r.close();
            endTime = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("CARL I/O (block read)     : " + NumberFormat::getInstance()->format((endTime - startTime).getNanoseconds(), IntegerFormatProperties("%12u")) + "ns");
        }
    }
    StandardStreams::out.writeLn("*** Done");
    StandardStreams::out.writeLn();
    FileSystem::deleteFile("crt_blocks.dat");
    FileSystem::deleteFile("carl_blocks.dat");

    return 0;
}

//  End of Main.cpp
