//
//  Main.cpp
//
//  The BinaryFixed arithmetics test.
//
//  $Date: 2008/08/09 13:18:18 $
//  $Revision: 1.3 $
//  $Author: akapusti $
//
//  Copyright (C) 2008 Cybernetic Intelligence GmbH
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////
#include "CARL.hpp"
using namespace carl;

class MyApp : public Application
{
    //////////
    //  Can't assign or copy construct
private:
    MyApp(const MyApp &);
    void                        operator = (const MyApp &);

    //////////
    //  Construction/destruction
public:
    MyApp() {}
    virtual ~MyApp() {}

    //////////
    //  Operations
public:
    virtual int run(const Array<String> & args);
};

MyApp theApp;

namespace
{
    Real64 generateRandomRealValue(Random & r)
    {
        Int64 magnitude = r.nextInt64();
        Int32 exponent = r.nextInt32(-32, 16);
        Real64 result = static_cast<Real64>(magnitude);
        while (exponent < 0)
        {
            result /= 2;
            exponent++;
        }
        while (exponent > 0)
        {
            result *= 2;
            exponent--;
        }
        return result;
    }

    bool closeValues(Real64 r1, const BinaryFixed & bf)
    {
        Real64 r2 = bf.toReal64();
        Real64 delta = Math::abs(r1 - r2);
        return (delta / Math::abs(r1)) < 1e-5;
    }
}

#define CHECK(x) if (!(x)) { exitCode = Process::ExitCode_Failure; }

int MyApp::run(const Array<String> & /*args*/)
{
    Process::ExitCode exitCode = Process::ExitCode_Success; 

    Random r;

    //  Consistency test
    for (int i = 0; i < 1000; i++)
    {
        StandardStreams::out.writeLn("Processing test vector " + NumberFormat::getInstance()->format(i));
        Real64 op1r = generateRandomRealValue(r);
        Real64 op2r = generateRandomRealValue(r);
        Real64 op1rPlusOp2r = op1r + op2r;
        Real64 op1rMinusOp2r = op1r - op2r;
        Real64 op1rMulOp2r = op1r * op2r;
        Real64 op1rDivOp2r = op1r / op2r;
        //
        BinaryFixed op1(op1r);
        BinaryFixed op2(op2r);
        BinaryFixed op1PlusOp2 = op1 + op2;
        BinaryFixed op1MinusOp2 = op1 - op2;
        BinaryFixed op1MulOp2 = op1 * op2;
        BinaryFixed op1DivOp2 = op1 / op2;
        //
        CHECK(closeValues(op1r, op1))
        CHECK(closeValues(op2r, op2))
        CHECK(closeValues(op1rPlusOp2r, op1PlusOp2))
        CHECK(closeValues(op1rMinusOp2r, op1MinusOp2))
        CHECK(closeValues(op1rMulOp2r, op1MulOp2))
        CHECK(closeValues(op1rDivOp2r, op1DivOp2))
    }

    //  Speed test
    for (UIntSize bits = 32; bits <= 4096; bits *= 2)
    {
        //  Addition
        {
            const int Iterations = 10000;
            Array<UInt8> value1, value2;
            value1.setSize(bits / 8);
            value2.setSize(bits / 8);
            r.nextBytes(value1.getData(), bits / 8);
            r.nextBytes(value2.getData(), bits / 8);
            BinaryInteger op1m(value1, ByteOrder_LittleEndian);
            BinaryInteger op2m(value2, ByteOrder_LittleEndian);
            BinaryFixed op1(op1m, r.nextInt32(0, bits));
            BinaryFixed op2(op2m, r.nextInt32(0, bits));
            TimeSpan start = TimeSpan::getTimeSpanSinceReferencePoint();
            for (int i = 0; i < Iterations; i++)
            {
                BinaryFixed result = op1 + op2;
            }
            TimeSpan end = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("Adding 2 " +
                                         NumberFormat::getInstance()->format(bits) +
                                         "-bit values : " +
                                         NumberFormat::getInstance()->format((end.getNanoseconds() - start.getNanoseconds()) / Iterations) +
                                         "ns");
        }

        //  Subtraction
        {
            const int Iterations = 10000;
            Array<UInt8> value1, value2;
            value1.setSize(bits / 8);
            value2.setSize(bits / 8);
            r.nextBytes(value1.getData(), bits / 8);
            r.nextBytes(value2.getData(), bits / 8);
            BinaryInteger op1m(value1, ByteOrder_LittleEndian);
            BinaryInteger op2m(value2, ByteOrder_LittleEndian);
            BinaryFixed op1(op1m, r.nextInt32(0, bits));
            BinaryFixed op2(op2m, r.nextInt32(0, bits));
            TimeSpan start = TimeSpan::getTimeSpanSinceReferencePoint();
            for (int i = 0; i < Iterations; i++)
            {
                BinaryFixed result = op1 - op2;
            }
            TimeSpan end = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("Subtracting 2 " +
                                         NumberFormat::getInstance()->format(bits) +
                                         "-bit values : " +
                                         NumberFormat::getInstance()->format((end.getNanoseconds() - start.getNanoseconds()) / Iterations) +
                                         "ns");
        }

        //  Multiplication
        {
            const int Iterations = 1000;
            Array<UInt8> value1, value2;
            value1.setSize(bits / 8);
            value2.setSize(bits / 8);
            r.nextBytes(value1.getData(), bits / 8);
            r.nextBytes(value2.getData(), bits / 8);
            BinaryInteger op1m(value1, ByteOrder_LittleEndian);
            BinaryInteger op2m(value2, ByteOrder_LittleEndian);
            BinaryFixed op1(op1m, r.nextInt32(0, bits));
            BinaryFixed op2(op2m, r.nextInt32(0, bits));
            TimeSpan start = TimeSpan::getTimeSpanSinceReferencePoint();
            for (int i = 0; i < Iterations; i++)
            {
                BinaryFixed result = op1 * op2;
            }
            TimeSpan end = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("Multiplying 2 " +
                                         NumberFormat::getInstance()->format(bits) +
                                         "-bit values : " +
                                         NumberFormat::getInstance()->format((end.getNanoseconds() - start.getNanoseconds()) / Iterations) +
                                         "ns");
        }

        //  Division
        {
            const int Iterations = 100;
            Array<UInt8> value1, value2;
            value1.setSize(bits / 8);
            value2.setSize(bits / 8);
            r.nextBytes(value1.getData(), bits / 8);
            r.nextBytes(value2.getData(), bits / 8);
            BinaryInteger op1m(value1, ByteOrder_LittleEndian);
            BinaryInteger op2m(value2, ByteOrder_LittleEndian);
            BinaryFixed op1(op1m, r.nextInt32(0, bits));
            BinaryFixed op2(op2m, r.nextInt32(0, bits));
            BinaryFixed op1mulOp2 = op1 * op2;
            TimeSpan start = TimeSpan::getTimeSpanSinceReferencePoint();
            for (int i = 0; i < Iterations; i++)
            {
                BinaryFixed result = op1mulOp2 / op2;
            }
            TimeSpan end = TimeSpan::getTimeSpanSinceReferencePoint();
            StandardStreams::out.writeLn("Dividing " +
                                         NumberFormat::getInstance()->format(2 * bits) +
                                         "-bit value by " +
                                         NumberFormat::getInstance()->format(bits) +
                                         "-bit value : " +
                                         NumberFormat::getInstance()->format((end.getNanoseconds() - start.getNanoseconds()) / Iterations) +
                                         "ns");
        }
    }

    //  Done
    return exitCode;
}

//  End of Main.cpp
