#
#   runtest.py
#
#   CARL builder.
#
#   $Date: 2008/07/28 10:40:28 $
#   $Revision: 1.5 $
#   $Author: akapusti $
#
#   Copyright (C) 2007-2008 Cybernetic Intelligence GmbH
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##########
import sys, getopt, subprocess, shutil, os, time

#   Supported platforms
class Platform(object):
    X86Win32Msvc2003 = "x86-win32-msvc2003"
    X86Win32Msvc2005 = "x86-win32-msvc2005"
    X86Win32Cygwin   = "x86-win32-cygwin"
    X86LinuxGcc      = "x86-linux-gcc"

#   Supported configurations
class Configuration(object):
    Debug   = "debug"
    Release = "release"
    All     = "all"

#   Supported linkages                          
class Linkage(object):
    Static  = "static"
    Dynamic = "dynamic"
    All     = "all"

#   Command line help
def printUsage():
    print("Usage:")
    print("    python build.py --platform=<platform> --configuration=<configuration>")
    print("                    --linkage=<linkage> <test or test set name>...")
    print("")
    print("Where:")
    print("    --platform=x86-win32-msvc2003- Test CARL for MSVC2003 on x86/Win32")
    print("    --platform=x86-win32-msvc2005- Test CARL for MSVC2005 on x86/Win32")
    print("    --platform=x86-win32-cygwin  - Test CARL for cygwin on x86/Win32")
    print("    --platform=x86-linux-gcc     - Test CARL for gcc on x86/Linux")
    print("")
    print("    --configuration=debug        - Test Debug version of CARL")
    print("    --configuration=release      - Test Release version of CARL")
    print("    --configuration=all          - Test both Debug and Release versions")
    print("")
    print("    --linkage=static             - Test CARL static library")
    print("    --linkage=dynamic            - Test CARL dynamic library")
    print("    --linkage=all                - Test both static and dynamic libraries")

#   The helper to spawn a subprocess and capture its stdout, stderr & exit code
def spawnProcess(commandLineArguments, environment):
    try:
        #   Start the process
        childProcess = subprocess.Popen(commandLineArguments, bufsize = 1024 * 8, stdout = subprocess.PIPE, stderr = subprocess.PIPE, universal_newlines = True, env = environment)
        redirectedStdout = childProcess.stdout
        redirectedStderr = childProcess.stderr
        #   Now wait for the process to terminate
        exitCode = None
        stdoutContents = ""
        stderrContents = ""
        while True:
            stdoutContents = stdoutContents + redirectedStdout.read()
            stderrContents = stderrContents + redirectedStderr.read()
            exitCode = childProcess.poll()
            if exitCode is not None:
                break
            time.sleep(1)
        #   Break output into lines & return
        if len(stdoutContents.strip(" \t\r\n")) == 0:
            stdoutLines = []
        else:
            stdoutLines = stdoutContents.split("\n")
        if len(stderrContents.strip(" \t\r\n")) == 0:
            stderrLines = []
        else:
            stderrLines = stderrContents.split("\n")
        return (stdoutLines, stderrLines, exitCode)
    except OSError:
        return ([], [], 1)

#   The helper service to safely casulate test pass rate
def passRate(testsExecuted, testsFailed):
    if testsExecuted == 0:
        return "100%"
    else:
        return str((testsExecuted - testsFailed) * 100 / testsExecuted) + "%"

#   The runner for a single test
def run(platform, configuration, linkage, test):
    if not os.path.isdir(test):
        print "'" + test + "' is not a valid test"
        return False
    print "Running " + test,
    sys.stdout.flush()
    #   Find all header (*.hpp) and source (*.cpp) files in the test directory
    headerFiles = []
    sourceFiles = []
    for (dirpath, dirnames, filenames) in os.walk(test):
        if dirpath == test:
            for filename in filenames:
                if filename.endswith(".hpp"):
                    headerFiles.append(filename)
                if filename.endswith(".cpp"):
                    sourceFiles.append(filename)
    #   Determine what project template to use
    projectBaseFileName = "test-" + platform + "-" + configuration + "-" + linkage + ".spp"
    projectTemplateFileName = os.path.join("ProjectFileTemplates", projectBaseFileName)
    projectFileName = os.path.join(test, projectBaseFileName)
    #   Now we need to copy the project file template into the
    #   test's directory and append source & header file names
    try:
        projectTemplateFile = open(projectTemplateFileName, "r")
    except Exception:
        print "\nCan't read project template file '" + projectTemplateFileName + "'"
        return False
    try:
        projectFile = open(projectFileName, "w")
    except Exception:
        print "\nCan't write project file '" + projectFileName + "'"
        return False
    projectFile.writelines(projectTemplateFile.readlines())
    projectTemplateFile.close()
    for headerFile in headerFiles:
        projectFile.write("header:                 " + headerFile + "\n")
    for sourceFile in sourceFiles:
        projectFile.write("source:                 " + sourceFile + "\n")
    #   Add dependency on the CARL library
    projectFile.write("dependency: " + os.path.abspath("../CARL-1_0-" + platform + "-" + configuration + "-" + linkage + ".spp") + "\n")
    projectFile.close()
    #   Now build the project
    print ".",
    sys.stdout.flush()
    commandLineArguments = ["python", "../Builder/splbuild.py", projectFileName]
    (stdoutContents, stderrContents, exitCode) = spawnProcess(commandLineArguments, os.environ)
    buildLogFile.write(" ".join(commandLineArguments) + "\n")
    buildLogFile.write("\n".join(stdoutContents) + "\n")
    buildLogFile.write("\n".join(stderrContents) + "\n\n")
    if exitCode != 0:
        commandLine = " ".join(commandLineArguments)
        print "\nError executing command '" + commandLine + "'"
        print "\n".join(stdoutContents)
        print "\n".join(stderrContents)
        return False
    #   If the test uses dynamic linkage, we must copy the CARL DLL
    #   into the test's run directory. To avoid guessing at the
    #   CARL DLL's name we can just copy everything from the CARL/Bin.
    tempDllNames = []
    if linkage == Linkage.Dynamic:
        for (dirpath, dirnames, filenames) in os.walk("../Bin"):
            for filename in filenames:
                tempDllName = os.path.join(test, "Bin/", filename)
                tempDllNames.append(tempDllName)
                try:
                    shutil.copyfile(os.path.join(dirpath, filename), tempDllName)
                except Exception:
                    print "\nCan't copy file '" + os.path.join(dirpath, filename) + "' to '" + tempDllName + "'"
                    return False
    #   Now run it! Note that we must add the test run directory to the LD_LIBRARY_PATH,
    #   otherwise tests will not run on Unix
    testEnvironment = os.environ
    testEnvironment['LD_LIBRARY_PATH'] = os.path.join(test, "Bin/")
    print ".",
    sys.stdout.flush()
    commandLineArguments = [os.path.join(test, "Bin/test-" + platform + "-" + configuration + "-" + linkage)]
    (stdoutContents, stderrContents, exitCode) = spawnProcess(commandLineArguments, testEnvironment)
    runLogFile.write(" ".join(commandLineArguments) + "\n")
    runLogFile.write("\n".join(stdoutContents) + "\n")
    runLogFile.write("\n".join(stderrContents) + "\n\n")
    if exitCode == 0:
        print "OK"
    else:
        commandLine = " ".join(commandLineArguments)
        print "Failed"
        return False
    #   Perform the cleanup
    for tempDllName in tempDllNames:
        os.remove(tempDllName)
    os.remove(projectFileName)
    #   Done
    return True

#   The helper for reading test set files
def readTestSetFile(fileName):
    try:
        #   Read all lines
        f = open(fileName, "r")
        lines = f.readlines();
        f.close();
        #   Remove all lines that are empty or start with a '#'
        return [line.strip(" \t\n\r") for line in lines 
                if len(line.strip(" \t\n\r")) <> 0 and not line.strip(" \t\n\r").startswith("#")]
    except Exception:
        print "Error reading test set file '" + fileName + "'"
        sys.exit(1)

#   The tester function
def runAllTests(platform, configuration, linkage, tests):
    if configuration == Configuration.All:
        runAllTests(platform, Configuration.Debug, linkage, tests)
        runAllTests(platform, Configuration.Release, linkage, tests)
    elif linkage == Linkage.All:
        runAllTests(platform, configuration, Linkage.Static, tests)
        runAllTests(platform, configuration, Linkage.Dynamic, tests)
    else:
        print "Testing " + configuration + " " + linkage + " CARL library for " + platform
        buildLogFileName = "build-log-" + platform + "-" + configuration + "-" + linkage + ".log"
        runLogFileName = "run-log-" + platform + "-" + configuration + "-" + linkage + ".log"
        global buildLogFile, runLogFile
        buildLogFile = open(buildLogFileName, "w")
        runLogFile = open(runLogFileName, "w")
        #   Go through all tests
        testsExecuted = 0
        testsFailed = 0
        for test in tests:
            if os.path.isdir(test):
                #   This is a test
                testsExecuted = testsExecuted + 1
                if not run(platform, configuration, linkage, test):
                    testsFailed = testsFailed + 1
            elif os.path.isfile(test):
                #   This is a test set file
                testList = readTestSetFile(test)
                for t in testList:
                    testsExecuted = testsExecuted + 1
                    if not run(platform, configuration, linkage, t):
                        testsFailed = testsFailed + 1
            else:
                print test + " is not a valid test or test set"
                testsExecuted = testsExecuted + 1
                testsFailed = testsFailed + 1
        print str(testsExecuted) + " tests executed, " + str(testsFailed) + " tests failed, pass rate " + passRate(testsExecuted, testsFailed)
        buildLogFile.close()
        runLogFile.close()
        print
        global totalTestsExecuted, totalTestsFailed
        totalTestsExecuted = totalTestsExecuted + testsExecuted
        totalTestsFailed = totalTestsFailed + testsFailed

#   Program's entry point
if __name__ == "__main__":
    global buildLogFile, runLogFile, totalTestsExecuted, totalTestsFailed
    buildLogFile = None
    runLogFile = None
    totalTestsExecuted = 0
    totalTestsFailed = 0
    #   Fetch command line options
    opts, args = getopt.getopt(sys.argv[1:], "", ["platform=", "configuration=", "linkage="])
    platform = None
    configuration = None
    linkage = None
    for name, value in opts:
        if name == "--platform":
            platform = value
        elif name == "--configuration":
            configuration = value
        elif name == "--linkage":
            linkage = value
    if (platform is None) or (configuration is None) or (linkage is None):
        printUsage()
        sys.exit(0)
    #   Validate command line options
    if (platform != Platform.X86Win32Msvc2003) and (platform != Platform.X86Win32Msvc2005) and (platform != Platform.X86Win32Cygwin) and (platform != Platform.X86LinuxGcc):
        print("Invalid platform " + platform)
        printUsage()
        sys.exit(1)
    if (configuration != Configuration.Debug) and (configuration != Configuration.Release) and (configuration != Configuration.All):
        print("Invalid configuration " + configuration)
        printUsage();
        sys.exit(1)
    if (linkage != Linkage.Static) and (linkage != Linkage.Dynamic) and (linkage != Linkage.All):
        print("Invalid linkage " + linkage)
        printUsage();
        sys.exit(1)
    if len(args) == 0:
        print("No tests specified")
        printUsage();
        sys.exit(1)
    #   Test!
    runAllTests(platform, configuration, linkage, args)
    print "\nTotal: " + str(totalTestsExecuted) + " tests executed, " + str(totalTestsFailed) + " tests failed, pass rate " + passRate(totalTestsExecuted, totalTestsFailed)
    if totalTestsFailed <> 0:
        sys.exit(1)
    else:
        sys.exit(0)

#   End of runtest.py
