//
//  CARL_IO.hpp - I/O facilities
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.32 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL I/O framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_IO.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_IO_hpp_included__
#define __CARL_IO_hpp_included__

namespace carl
{
    /**
     *  @class IOException CARL.hpp
     *  A common base for all CARL I/O exceptions.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IOException : public Exception
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        IOException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        IOException(const IOException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~IOException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        IOException &           operator = (const IOException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class EOFException CARL.hpp
     *  An exception thrown when an unexpected enf-of-file is detected 
     *  during a read operation.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC EOFException : public IOException
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        EOFException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        EOFException(const EOFException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~EOFException();

        /**
         *  The class destructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        EOFException &          operator = (const EOFException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class InvalidDataException CARL.hpp
     *  An exception thrown when data read from a stream does not obey
     *  the integrity constraints it is expected to obey.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC InvalidDataException : public IOException
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        InvalidDataException();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        InvalidDataException(const InvalidDataException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~InvalidDataException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        InvalidDataException &  operator = (const InvalidDataException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class FileNotFoundException CARL.hpp
     *  An exception thrown when an attempt is made to read from
     *  a nonexistent file.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC FileNotFoundException : public IOException
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The class constructor.
         *
         *  @param pFileName    A pointer to nul-terminated ASCII string
         *                      representing the name of the nonexistent file.
         *  @since 1.0
         */
        explicit FileNotFoundException(const char * pFileName);

        /**
         *  The class constructor.
         *
         *  @param pFileName    A pointer to nul-terminated ISo-10646 string
         *                      representing the name of the nonexistent file.
         *  @since 1.0
         */
        explicit FileNotFoundException(const Char * pFileName);

        /**
         *  The class constructor.
         *
         *  @param fileName     A string representing the name of the nonexistent file.
         *  @since 1.0
         */
        explicit FileNotFoundException(const String & fileName);

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        FileNotFoundException(const FileNotFoundException & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~FileNotFoundException();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        FileNotFoundException & operator = (const FileNotFoundException & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the name ofthe nonexistent file.
         *
         *  @return             The name ofthe nonexistent file.
         *  @since 1.0
         */
        String                  getFileName() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;

        //////////
        //  Implementation
    private:
        String                  _fileName;
    };

    //////////
    //  Byte stream interfaces
    /**
     *  @interface IByteInput CARL.hpp
     *  An abstract interface to a byte input stream.
     *
     *  A byte input stream allows reading raw bytes individually or
     *  in consecutive blocks.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IByteInput
    {
        //////////
        //  Construction/destruction
    public:
        IByteInput() {}
        virtual ~IByteInput() {}

        //////////
        //  Operations
    public:
        /**
         *  Returns the number of bytes that can be read from this input 
         *  stream without blocking.
         *
         *  @return                 The number of bytes that can be read from this
         *                          input stream without blocking.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        available() throw(IOException) = 0;

        /**
         *  Reads and returns the next byte or -1 if an end of stream is reached.
         *
         *  @return                 The next input byte, or -1 if an end of stream is reached.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual int             read() throw(IOException) = 0;

        /**
         *  Reads a sequence of bytes into the specified buffer.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is read, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  @param pBuffer          The buffer where bytes read from the stream
         *                          shall be placed.
         *  @param bytesToRead      The maximum number of bytes to read.
         *  @return                 The number of bytes actually read. This can
         *                          be smaller than \a bytesToRead if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException) = 0;

        /**
         *  Skips a sequence of input bytes.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is skipped, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  @param bytesToSkip      The maximum number of bytes to skip.
         *  @return                 The number of bytes actually skipped. This can
         *                          be smaller than \a bytesToSkip if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException) = 0;
    };

    /**
     *  @interface IByteOutput CARL.hpp
     *  An abstract interface to a byte output stream.
     *
     *  A byte input stream allows writing raw bytes individually or
     *  in consecutive blocks.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IByteOutput
    {
        //////////
        //  Construction/destruction
    public:
        IByteOutput() {}
        virtual ~IByteOutput() {}

        //////////
        //  Operations
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  If there is any data that has been written to this output stream
         *  and not yet forwarded to the stream's destination, this data is 
         *  written to the stream's destination immediately.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException) = 0;

        /**
         *  Writes a single byte to the stream.
         *
         *  @param byte             The byte to write to this stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(UInt8 byte) throw(IOException) = 0;

        /**
         *  Writes a sequence of bytes to the stream.
         *
         *  @param pBuffer          The address of the first byte of the sequence
         *                          that shall be written to the stream. The
         *                          sequence must be at least \a bytesToWrite
         *                          bytes long; if it is longer then only the
         *                          forst \a bytesToWrite bytes are written.
         *  @param bytesToWrite     The number of bytes to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException) = 0;
    };

    //////////
    //  Data stream interfaces
    /**
     *  @interface IDataInput CARL.hpp
     *  An abstract interface to a data input stream.
     *
     *  A data input stream allows reading values of primitive data types.
     *  In addition, the ">>" operator overloading for data input streams
     *  allows reading values of composite types from data input streams.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IDataInput : public IByteInput
    {
        //////////
        //  Construction/destruction
    public:
        IDataInput() {}
        virtual ~IDataInput() {}

        //////////
        //  Operations
    public:
        /**
         *  Reads a boolean value from this stream.
         *
         *  A boolean value is expected to be represented as a single byte
         *  with value 0 (false) or 1 (true). Any other value is an error.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            readBool() throw(IOException) = 0;

        /**
         *  Reads an UInt8 value from this stream.
         *
         *  An UInt8 value is expected to be represented as a single byte.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt8           readUInt8() throw(IOException) = 0;

        /**
         *  Reads an UInt16 value from this stream.
         *
         *  An UInt16 value is expected to be represented as 2 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt16          readUInt16() throw(IOException) = 0;

        /**
         *  Reads an UInt32 value from this stream.
         *
         *  An UInt32 value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt32          readUInt32() throw(IOException) = 0;

        /**
         *  Reads an UInt64 value from this stream.
         *
         *  An UInt64 value is expected to be represented as 8 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt64          readUInt64() throw(IOException) = 0;

        /**
         *  Reads an Int8 value from this stream.
         *
         *  An Int8 value is expected to be represented as a single byte.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int8            readInt8() throw(IOException) = 0;

        /**
         *  Reads an Int16 value from this stream.
         *
         *  An Int16 value is expected to be represented as 2 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int16           readInt16() throw(IOException) = 0;

        /**
         *  Reads an Int32 value from this stream.
         *
         *  An Int32 value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int32           readInt32() throw(IOException) = 0;

        /**
         *  Reads an Int64 value from this stream.
         *
         *  An Int64 value is expected to be represented as 8 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int64           readInt64() throw(IOException) = 0;

        /**
         *  Reads a Real32 value from this stream.
         *
         *  A Real32 value is expected to be represented as 4 consecutive IEEE754
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Real32          readReal32() throw(IOException) = 0;

        /**
         *  Reads a Real64 value from this stream.
         *
         *  A Real64 value is expected to be represented as 8 consecutive IEEE754
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Real64          readReal64() throw(IOException) = 0;

        /**
         *  Reads a Char value from this stream.
         *
         *  An Char value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Char            readChar() throw(IOException) = 0;
    };

    /**
     *  @interface IDataOutput CARL.hpp
     *  An abstract interface to a data output stream.
     *
     *  A data output stream allows writing values of primitive data types.
     *  In addition, the "<<" operator overloading for data output streams
     *  allows writing values of composite types to data output streams.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IDataOutput : public IByteOutput
    {
        //////////
        //  Construction/destruction
    public:
        IDataOutput() {}
        virtual ~IDataOutput() {}

        //////////
        //  Operations
    public:
        /**
         *  Writes a boolean value to this stream.
         *
         *  A boolean value is represented as a single byte with value 0 
         *  (false) or 1 (true).
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeBool(bool value) throw(IOException) = 0;

        /**
         *  Writes an UInt8 value to this stream.
         *
         *  An UInt8 value is represented as a single byte.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt8(UInt8 value) throw(IOException) = 0;

        /**
         *  Writes an UInt16 value to this stream.
         *
         *  An UInt16 value is represented as 2 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt16(UInt16 value) throw(IOException) = 0;

        /**
         *  Writes an UInt32 value to this stream.
         *
         *  An UInt32 value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt32(UInt32 value) throw(IOException) = 0;

        /**
         *  Writes an UInt64 value to this stream.
         *
         *  An UInt64 value is represented as 8 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt64(UInt64 value) throw(IOException) = 0;

        /**
         *  Writes an Int8 value to this stream.
         *
         *  An Int8 value is represented as a single byte.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt8(Int8 value) throw(IOException) = 0;

        /**
         *  Writes an Int16 value to this stream.
         *
         *  An Int16 value is represented as 2 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt16(Int16 value) throw(IOException) = 0;

        /**
         *  Writes an Int32 value to this stream.
         *
         *  An Int32 value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt32(Int32 value) throw(IOException) = 0;

        /**
         *  Writes an Int64 value to this stream.
         *
         *  An Int64 value is represented as 8 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt64(Int64 value) throw(IOException) = 0;

        /**
         *  Writes a Real32 value to this stream.
         *
         *  A Real32 value is represented as 4 consecutive IEEE754 bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeReal32(Real32 value) throw(IOException) = 0;

        /**
         *  Writes a Real64 value to this stream.
         *
         *  A Real64 value is represented as 8 consecutive IEEE754 bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeReal64(Real64 value) throw(IOException) = 0;

        /**
         *  Writes a Char value to this stream.
         *
         *  A Char value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeChar(Char value) throw(IOException) = 0;
    };

    //////////
    //  Text stream interfaces
    /**
     *  @interface ITextInput CARL.hpp
     *  An abstract interface to a text input stream.
     *
     *  A text input stream allows reading ISO-10646 characters individually
     *  or line-by-line. In addition, textg input streams allow "un-reading"
     *  at least one most recently read character (or more).
     *
     *  @since 1.0
     */
    class CARL_PUBLIC ITextInput
    {
        //////////
        //  Construction/destruction
    public:
        ITextInput() {}
        virtual ~ITextInput() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if a character can be read from this text input 
         *  stream without blocking.
         *
         *  @return             True if at least one character can be read from
         *                      this stream without blocking, false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            ready() throw(IOException) = 0;

        /**
         *  Reads a single character from this stream.
         *
         *  @param c            If a character is read successfully, it is
         *                      assigned to this parameter. Otherwise this
         *                      parameter remains unchanged.
         *  @return             True if the character has been read
         *                      successfully; false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            read(Char & c) throw(IOException) = 0;

        /**
         *  Reads a sequence of characters from this stream.
         *
         *  The actual number of character read can be smaller than the
         *  requested number of characters  if an end-of-stream is reached 
         *  before a specified number of character has been read.
         *
         *  @param pBuffer      A pointer to a buffer where characters read
         *                      from the stream are placed. The buffer is
         *                      assumed to be at least \a charsToRead
         *                      characters long.
         *  @param charsToRead  A maximum number of characters to read.
         *  @return             The actual number of characters read.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(Char * pBuffer, UIntSize charsToRead) throw(IOException) = 0;

        /**
         *  Reads a line of characters from this stream.
         *
         *  Characters are read until either end-of-line or end-of-stream
         *  is detected, whichever happens first.
         *
         *  @param line         Upon successful read, this parameter is
         *                      assigned the characters of the line just 
         *                      read.
         *  @param discardEoln  If the line was terminated by an end-of-line
         *                      marker and this parameter is false, the
         *                      end-of-line marker is included into the
         *                      \a line. If the line was terminated by an
         *                      end-of-line marker and this parameter is true,
         *                      the end-of-line marker is not included into
         *                      the \a line. If the line was terminated by
         *                      an end-of-file marker, the end-of-line marker
         *                      is not included into the \a line regardless
         *                      of whether \a discardEoln is true or false.
         *  @return             True if the line has been read successfully; 
         *                      false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            readLine(String & line, bool discardEoln = false) throw(IOException) = 0;

        /**
         *  Skips a sequence of characters from this stream.
         *
         *  The actual number of character skipped can be smaller than the
         *  requested number of characters  if an end-of-stream is reached 
         *  before a specified number of character has been skipped.
         *
         *  @param charsToSkip  A maximum number of characters to skip.
         *  @return             The actual number of characters skipped.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize charsToSkip) throw(IOException) = 0;

        /**
         *  Un-reads the character to the stream.
         *
         *  Upon successful un-read, the next character read from the
         *  stream is guaranteed to be the last character un-read to
         *  that stream. Text streams always guarantee that at least one
         *  character can be un-read, although some concrete text input
         *  streams may allow un-reading mul;tiple characters.
         *
         *  @param c            The character to push back into the 
         *                      input stream.
         *  @return             True if the character has been successfully
         *                      pushed back into the input stream; false
         *                      otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            unread(Char c) throw(IOException) = 0;
    };

    /**
     *  @interface ITextOutput CARL.hpp
     *  An abstract interface to a text output stream.
     *
     *  A text output stream allows writing ISO-10646 characters individually
     *  or line-by-line.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC ITextOutput
    {
        //////////
        //  Construction/destruction
    public:
        ITextOutput() {}
        virtual ~ITextOutput() {}

        //////////
        //  Operations
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  If there is any data that has been written to this output stream
         *  and not yet forwarded to the stream's destination, this data is 
         *  written to the stream's destination immediately.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException) = 0;

        /**
         *  Writes a character to this output stream.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(Char c) throw(IOException) = 0;

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer) throw(IOException) = 0;

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer, UIntSize charsToWrite) throw(IOException) = 0;

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ISO-10646 characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer) throw(IOException) = 0;

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the sequence of ISo-10646 characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) = 0;

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the stream. All characters from the
         *                      string are written, in order, to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const String & s) throw(IOException) = 0;

        /**
         *  Wrtes an end-of-line marker to the stream.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn() throw(IOException) = 0;

        /**
         *  Wrtes a character followed by end-of-line marker to the stream.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(Char c) throw(IOException) = 0;

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer) throw(IOException) = 0;

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException) = 0;

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ISO-10646 characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const Char * pBuffer) throw(IOException) = 0;

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the sequence of ISo-10646 characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) = 0;

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the stream. All characters from the
         *                      string are written, in order, to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const String & s) throw(IOException) = 0;
    };

    //////////
    //  Basic file I/O
    /**
     *  @class FileInputStream CARL.hpp
     *  A byte input stream for reading from a binary file.
     *
     *  File input streams allow reading existing files in raw binary
     *  mode. No internal buffering is performed. The file must already
     *  exist to be readable via a file input stream.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC FileInputStream : public IByteInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        FileInputStream(const FileInputStream &);
        void                    operator = (const FileInputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created file input stream to read
         *  from a specified file, which must already exist.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        FileInputStream(const char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created file input stream to read
         *  from a specified file, which must already exist.
         *
         *  @param pFileName    A pointer to a nul-terminated ISo-10646 string
         *                      representing a name of the file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        FileInputStream(const Char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created file input stream to read
         *  from a specified file, which must already exist.
         *
         *  @param fileName     A string representing a name of the 
         *                      file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        FileInputStream(const String & fileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created file input stream to read
         *  from a specified file, which must already exist.
         *
         *  @param pathName     A path to the file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        FileInputStream(const PathName & pathName) throw(IOException);

        /**
         *  The class destructor.
         *
         *  Closes the input stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~FileInputStream() throw(IOException) { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this input stream is still open.
         *
         *  @return                 True if this input stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _handle != -1; }

        /**
         *  Closes this input stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteInput
    public:
        /**
         *  Returns the number of bytes that can be read from this input 
         *  stream without blocking.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The number of bytes that can be read from this
         *                          input stream without blocking.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        available() throw(IOException);

        /**
         *  Reads and returns the next byte or -1 if an end of stream is reached.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The next input byte, or -1 if an end of stream is reached.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual int             read() throw(IOException);

        /**
         *  Reads a sequence of bytes into the specified buffer.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is read, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The buffer where bytes read from the stream
         *                          shall be placed.
         *  @param bytesToRead      The maximum number of bytes to read.
         *  @return                 The number of bytes actually read. This can
         *                          be smaller than \a bytesToRead if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException);

        /**
         *  Skips a sequence of input bytes.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is skipped, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  This method can only be called on open streams.
         *
         *  @param bytesToSkip      The maximum number of bytes to skip.
         *  @return                 The number of bytes actually skipped. This can
         *                          be smaller than \a bytesToSkip if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException);

        //////////
        //  Implementation
    private:
        int                     _handle;
        void                    _openFile(const String & fileName) throw(IOException);
    };

    /**
     *  @class FileOutputStream CARL.hpp
     *  A byte input stream for writing to a binary file.
     *
     *  File output streams allow writing to files in raw binary
     *  mode. No internal buffering is performed. The file is created
     *  if it does not exist, or overwritten it it does exist.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC FileOutputStream : public IByteOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        FileOutputStream(const FileOutputStream &);
        void                    operator = (const FileOutputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created file output stream to write
         *  to a specified file. The file is created if it does not
         *  exist, or overwritten it it does exist.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        FileOutputStream(const char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created file output stream to write
         *  to a specified file. The file is created if it does not
         *  exist, or overwritten it it does exist.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing a name of the file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        FileOutputStream(const Char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created file output stream to write
         *  to a specified file. The file is created if it does not
         *  exist, or overwritten it it does exist.
         *
         *  @param fileName     A string representing a name of the 
         *                      file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        FileOutputStream(const String & fileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created file output stream to write
         *  to a specified file. The file is created if it does not
         *  exist, or overwritten it it does exist.
         *
         *  @param pathName     A path to the file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        FileOutputStream(const PathName & pathName) throw(IOException);

        /**
         *  The class destructor.
         *
         *  Closes the output stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~FileOutputStream() throw(IOException) { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this output stream is still open.
         *
         *  @return                 True if this output stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _handle != -1; }

        /**
         *  Closes this output stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteOutput
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  Causes OS file buffers for this file to be flushed to disk.
         *
         *  This method can only be called on open streams.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException);

        /**
         *  Writes a single byte to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param byte             The byte to write to this stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(UInt8 byte) throw(IOException);

        /**
         *  Writes a sequence of bytes to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The address of the first byte of the sequence
         *                          that shall be written to the stream. The
         *                          sequence must be at least \a bytesToWrite
         *                          bytes long; if it is longer then only the
         *                          forst \a bytesToWrite bytes are written.
         *  @param bytesToWrite     The number of bytes to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException);

        //////////
        //  Implementation
    private:
        int                     _handle;
        void                    _openFile(const String & fileName) throw(IOException);
    };

    /**
     *  @class RandomAccessFile CARL.hpp
     *  A random-access binary file.
     *
     *  Random-access files can be both read from and written to in
     *  an arbitrary manner. No buffering is performed.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC RandomAccessFile : public IDataInput, public IDataOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        RandomAccessFile(const RandomAccessFile &);
        void                    operator = (const RandomAccessFile &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  Represents the supported file open behaviours.
         *
         *  @since 1.0
         */
        enum OpenMode
        {
            /**
             *  Open an existing file.
             *
             *  If the file does not exist, an error occurs. If the file exists, 
             *  it is opened without destroying its contents.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            OpenMode_OpenExisting       = 1,

            /**
             *  Create a new file.
             *
             *  If the file already exists, an error occurs.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            OpenMode_CreateNew          = 2,

            /**
             *  Open an existing file and truncate.
             *
             *  If the file does not exist, an error occurs. If the file exists, 
             *  its contents is destroyed when it is opened.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            OpenMode_TruncateExisting   = 3,

            /**
             *  Open an existing file or create a new one.
             *
             *  If the file does not exist, it is created empty. If the file exists, 
             *  it is opened without destroying its contents.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            OpenMode_CreateOrOpen       = 4,

            /**
             *  Create an empty file.
             *
             *  If the file does not exist, it is created. If the file exists, 
             *  its contents is destroyed when it is opened.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            OpenMode_CreateAlways       = 5
        };

        /**
         *  Represents a mode in which a random access file can be accessed.
         *
         *  @since 1.0
         */
        enum AccessMode
        {
            /**
             *  Read-only access.
             *
             *  Reading from a random-access file is permitted, but writing is not.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            AccessMode_ReadOnly         = 1,

            /**
             *  Read-write access.
             *
             *  Both reading from and writing to a random-access file is permitted.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            AccessMode_ReadWrite        = 2
        };

        /**
         *  The class constructor.
         *
         *  Initializes a newly created random access file to perform
         *  I/O from/to a specified file. The file may be created (provided
         *  it does not exist) or truncated (provided it exists) depending
         *  on the specified open mode.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to open.
         *  @param openMode     Specifies how to open a file (i.e. whether
         *                      to create it if not there, etc.)
         *  @param accessMode   Specifies the required access (read/write
         *                      vs read-only).
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        RandomAccessFile(const char * pFileName, OpenMode openMode, AccessMode accessMode) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created random access file to perform
         *  I/O from/to a specified file. The file may be created (provided
         *  it does not exist) or truncated (provided it exists) depending
         *  on the specified open mode.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing a name of the file to open.
         *  @param openMode     Specifies how to open a file (i.e. whether
         *                      to create it if not there, etc.)
         *  @param accessMode   Specifies the required access (read/write
         *                      vs read-only).
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        RandomAccessFile(const Char * pFileName, OpenMode openMode, AccessMode accessMode) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created random access file to perform
         *  I/O from/to a specified file. The file may be created (provided
         *  it does not exist) or truncated (provided it exists) depending
         *  on the specified open mode.
         *
         *  @param fileName     A string representing a name of the file to open.
         *  @param openMode     Specifies how to open a file (i.e. whether
         *                      to create it if not there, etc.)
         *  @param accessMode   Specifies the required access (read/write
         *                      vs read-only).
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        RandomAccessFile(const String & fileName, OpenMode openMode, AccessMode accessMode) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created random access file to perform
         *  I/O from/to a specified file. The file may be created (provided
         *  it does not exist) or truncated (provided it exists) depending
         *  on the specified open mode.
         *
         *  @param pathName     A path to the file to open.
         *  @param openMode     Specifies how to open a file (i.e. whether
         *                      to create it if not there, etc.)
         *  @param accessMode   Specifies the required access (read/write
         *                      vs read-only).
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        RandomAccessFile(const PathName & pathName, OpenMode openMode, AccessMode accessMode) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Creates a new, unique, temporary binary random access file
         *  with read+write access mode. The newly created file is initially
         *  empty. Application is responsible for destroying this file
         *  after it is no longer needed.
         *
         *  @param fileName     If a new temporary file is created successfully,
         *                      this parameter is assigned the full canonical
         *                      path to the temporary file. This name is the
         *                      one that shall be used when the temporary file
         *                      must be deleted.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        explicit RandomAccessFile(PathName & fileName) throw(IOException);

        /**
         *  The class destructor.
         *
         *  Closes the file if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~RandomAccessFile() throw(IOException) { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Represents an origin for a seek operation.
         *
         *  @since 1.0
         */
        enum SeekOrigin
        {
            /**
             *  Seek from the start of a file.
             *
             *  The offset specified in a seek operation is relative to the
             *  start of the file.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            SeekOrigin_Start    = 1,

            /**
             *  Seek from the end of a file.
             *
             *  The offset specified in a seek operation is relative to the
             *  end of the file.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            SeekOrigin_End      = 2,

            /**
             *  Seek from the current position.
             *
             *  The offset specified in a seek operation is relative to the
             *  current position of a file pointer.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            SeekOrigin_Current  = 3
        };

        /**
         *  Checks if this random access file is still open.
         *
         *  @return                 True if this random access file is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _handle != -1; }

        /**
         *  Closes this random access file.
         *
         *  An attempt to close a random access file that is already 
         *  closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        /**
         *  Returns the size, in bytes, of this file.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The size of this file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        UInt64                  getSize() const throw(IOException);

        /**
         *  Returns the current file pointer position.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The current file pointer position.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        UInt64                  getCurrentPosition() const throw(IOException);

        /**
         *  Sets the current file pointer position.
         *
         *  This method can only be called on open streams.
         *
         *  @param seekOrigin       The origin from which the \a offset is specified.
         *  @param offset           The offset, in bytes, from the specified origin.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        void                    setCurrentPosition(SeekOrigin seekOrigin, Int64 offset) throw(IOException);

        /**
         *  Locks a region of file.
         *
         *  Each byte in a file can be subject to at most one lock.
         *  Once some process establishes this lock, other processes
         *  cannot lock the same byte, read it or write it until
         *  the lock is released.
         *
         *  Locked file regions must be operated upon atomically (i.e.
         *  it is not possible to lock two consecutive regions of a
         *  file and then unlock them as one region which is twice as
         *  large).
         *
         *  @param start            The offset, in bytes, of the locked
         *                          region from the beginning of the file.
         *  @param length           The size, in bytes, of the region to lock.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        void                    lock(UInt64 start, UInt64 length) throw(IOException);

        /**
         *  Unlocks a region of file.
         *
         *  Each byte in a file can be subject to at most one lock.
         *  Once some process establishes this lock, other processes
         *  cannot lock the same byte, read it or write it until
         *  the lock is released.
         *
         *  Locked file regions must be operated upon atomically (i.e.
         *  it is not possible to lock two consecutive regions of a
         *  file and then unlock them as one region which is twice as
         *  large).
         *
         *  @param start            The offset, in bytes, of the locked
         *                          region from the beginning of the file.
         *  @param length           The size, in bytes, of the region to unlock.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        void                    unlock(UInt64 start, UInt64 length) throw(IOException);

        //////////
        //  IByteInput
    public:
        /**
         *  Returns the number of bytes that can be read from this random
         *  access file without blocking.
         *
         *  This method can only be called on open files.
         *
         *  @return                 The number of bytes that can be read from this
         *                          random access file without blocking.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        available() throw(IOException);

        /**
         *  Reads and returns the next byte or -1 if an end of file is reached.
         *
         *  Upon successful read, the file pointer is advanced by 1.
         *
         *  This method can only be called on open files.
         *
         *  @return                 The next input byte, or -1 if an end of file is reached.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual int             read() throw(IOException);

        /**
         *  Reads a sequence of bytes into the specified buffer.
         *
         *  If an end of the file is reached before the required number
         *  of bytes is read, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  Upon successful read, the file pointer is advanced by the number
         *  of bytes read.
         *
         *  This method can only be called on open files.
         *
         *  @param pBuffer          The buffer where bytes read from the file
         *                          shall be placed.
         *  @param bytesToRead      The maximum number of bytes to read.
         *  @return                 The number of bytes actually read. This can
         *                          be smaller than \a bytesToRead if an end
         *                          of file has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException);

        /**
         *  Skips a sequence of input bytes.
         *
         *  If an end of the file is reached before the required number
         *  of bytes is skipped, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  Upon successful skip, the file pointer is advanced by the number
         *  of bytes skipped.
         *
         *  This method can only be called on open files.
         *
         *  @param bytesToSkip      The maximum number of bytes to skip.
         *  @return                 The number of bytes actually skipped. This can
         *                          be smaller than \a bytesToSkip if an end
         *                          of file has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException);

        //////////
        //  IByteOutput
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  Causes OS file buffers for this file to be flushed to disk.
         *
         *  This method can only be called on open files.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException);

        /**
         *  Writes a single byte to the file.
         *
         *  Upon successful write, the file pointer is advanced by 1.
         *  If an attempt is made to write beyond the end-of-file, the 
         *  file grows to accomodate the written data.
         *
         *  This method can only be called on open files.
         *
         *  @param byte             The byte to write to this file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(UInt8 byte) throw(IOException);

        /**
         *  Writes a sequence of bytes to the file.
         *
         *  Upon successful write, the file pointer is advanced by the
         *  number of bytes written. If an attempt is made to write 
         *  beyond the end-of-file, the file grows to accomodate the 
         *  written data.
         *
         *  This method can only be called on open files.
         *
         *  @param pBuffer          The address of the first byte of the sequence
         *                          that shall be written to the stream. The
         *                          sequence must be at least \a bytesToWrite
         *                          bytes long; if it is longer then only the
         *                          forst \a bytesToWrite bytes are written.
         *  @param bytesToWrite     The number of bytes to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException);

        //////////
        //  IDataInput
    public:
        /**
         *  Reads a boolean value from this file.
         *
         *  A boolean value is expected to be represented as a single byte
         *  with value 0 (false) or 1 (true). Any other value is an error.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            readBool() throw(IOException);

        /**
         *  Reads an UInt8 value from this file.
         *
         *  An UInt8 value is expected to be represented as a single byte.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt8           readUInt8() throw(IOException);

        /**
         *  Reads an UInt16 value from this file.
         *
         *  An UInt16 value is expected to be represented as 2 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt16          readUInt16() throw(IOException);

        /**
         *  Reads an UInt32 value from this file.
         *
         *  An UInt32 value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt32          readUInt32() throw(IOException);

        /**
         *  Reads an UInt64 value from this file.
         *
         *  An UInt64 value is expected to be represented as 8 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt64          readUInt64() throw(IOException);

        /**
         *  Reads an Int8 value from this file.
         *
         *  An Int8 value is expected to be represented as a single byte.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int8            readInt8() throw(IOException);

        /**
         *  Reads an Int16 value from this file.
         *
         *  An Int16 value is expected to be represented as 2 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int16           readInt16() throw(IOException);

        /**
         *  Reads an Int32 value from this file.
         *
         *  An Int32 value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int32           readInt32() throw(IOException);

        /**
         *  Reads an Int64 value from this file.
         *
         *  An Int64 value is expected to be represented as 8 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int64           readInt64() throw(IOException);

        /**
         *  Reads a Real32 value from this file.
         *
         *  A Real32 value is expected to be represented as 4 consecutive IEEE754
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Real32          readReal32() throw(IOException);

        /**
         *  Reads a Real64 value from this file.
         *
         *  A Real64 value is expected to be represented as 8 consecutive IEEE754
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Real64          readReal64() throw(IOException);

        /**
         *  Reads a Char value from this file.
         *
         *  An Char value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @return             The value read from the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Char            readChar() throw(IOException);

        //////////
        //  IDataOutput
    public:
        /**
         *  Writes a boolean value to this file.
         *
         *  A boolean value is represented as a single byte with value 0 
         *  (false) or 1 (true).
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeBool(bool value) throw(IOException);

        /**
         *  Writes an UInt8 value to this file.
         *
         *  An UInt8 value is represented as a single byte.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt8(UInt8 value) throw(IOException);

        /**
         *  Writes an UInt16 value to this file.
         *
         *  An UInt16 value is represented as 2 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt16(UInt16 value) throw(IOException);

        /**
         *  Writes an UInt32 value to this file.
         *
         *  An UInt32 value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt32(UInt32 value) throw(IOException);

        /**
         *  Writes an UInt64 value to this file.
         *
         *  An UInt64 value is represented as 8 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt64(UInt64 value) throw(IOException);

        /**
         *  Writes an Int8 value to this file.
         *
         *  An Int8 value is represented as a single byte.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt8(Int8 value) throw(IOException);

        /**
         *  Writes an Int16 value to this file.
         *
         *  An Int16 value is represented as 2 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt16(Int16 value) throw(IOException);

        /**
         *  Writes an Int32 value to this file.
         *
         *  An Int32 value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt32(Int32 value) throw(IOException);

        /**
         *  Writes an Int64 value to this file.
         *
         *  An Int64 value is represented as 8 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt64(Int64 value) throw(IOException);

        /**
         *  Writes a Real32 value to this file.
         *
         *  A Real32 value is represented as 4 consecutive IEEE754 bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeReal32(Real32 value) throw(IOException);

        /**
         *  Writes a Real64 value to this file.
         *
         *  A Real64 value is represented as 8 consecutive IEEE754 bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeReal64(Real64 value) throw(IOException);

        /**
         *  Writes a Char value to this file.
         *
         *  A Char value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open files.
         *
         *  @param value        The value to write to the file.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeChar(Char value) throw(IOException);

        //////////
        //  Implementation
    private:
        int                     _handle;
        void                    _openFile(const String & fileName, OpenMode openMode, AccessMode accessMode) throw(IOException);
    };

    //////////
    //  I/O from/to memory buffers
    /**
     *  @class MemoryInputStream CARL.hpp
     *  A byte input stream for reading from an in-memory buffer.
     *
     *  Memory input streams allow treating memory areas as streams.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC MemoryInputStream : public IByteInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        MemoryInputStream(const MemoryInputStream &);
        void                    operator = (const MemoryInputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created memory input stream to read
         *  from a specified buffer, making a private copy of data
         *  if necessary.
         *
         *  @param pBuffer      A pointer to the start of the buffer
         *                      to read from. The buffer must be at
         *                      least \a bufferSize bytes long; if it 
         *                      is longer then only the first \a bufferSize
         *                      bytes are taken into consideration.
         *  @param bufferSize   The size, in bytes, of the data buffer.
         *  @param makePrivateCopy  If true, the memory input stream makes a
         *                      private copy of the buffer's data, which allows
         *                      the buffer contents to be changed without
         *                      affecting the contents of the stream. If false,
         *                      then reading is done directly from the specified
         *                      buffer.
         *  @since 1.0
         */
        MemoryInputStream(const void * pBuffer, UIntSize bufferSize, bool makePrivateCopy);

        /**
         *  The class destructor.
         *
         *  Closes the input stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~MemoryInputStream() { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this input stream is still open.
         *
         *  @return                 True if this input stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _isOpen; }

        /**
         *  Closes this input stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteInput
    public:
        /**
         *  Returns the number of bytes that can be read from this input 
         *  stream without blocking.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The number of bytes that can be read from this
         *                          input stream without blocking.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        available() throw(IOException);

        /**
         *  Reads and returns the next byte or -1 if an end of stream is reached.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The next input byte, or -1 if an end of stream is reached.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual int             read() throw(IOException);

        /**
         *  Reads a sequence of bytes into the specified buffer.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is read, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The buffer where bytes read from the stream
         *                          shall be placed.
         *  @param bytesToRead      The maximum number of bytes to read.
         *  @return                 The number of bytes actually read. This can
         *                          be smaller than \a bytesToRead if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException);

        /**
         *  Skips a sequence of input bytes.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is skipped, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  This method can only be called on open streams.
         *
         *  @param bytesToSkip      The maximum number of bytes to skip.
         *  @return                 The number of bytes actually skipped. This can
         *                          be smaller than \a bytesToSkip if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException);

        //////////
        //  Implementation
    private:
        bool                    _isOpen;
        const UInt8 *           _pData;
        UIntSize                _dataSize;
        UIntSize                _currentPosition;
        bool                    _isPrivateData;
    };

    /**
     *  @class MemoryOutputStream CARL.hpp
     *  A byte output stream for writing to an in-memory buffer.
     *
     *  Memory output streams allow treating memory areas as streams.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC MemoryOutputStream : public IByteOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        MemoryOutputStream(const MemoryOutputStream &);
        void                    operator = (const MemoryOutputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  A newly created memory output stream is empty.
         *  @since 1.0
         */
        MemoryOutputStream() { _isOpen = true; }

        /**
         *  The class destructor.
         *
         *  Closes the input stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~MemoryOutputStream() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if this input stream is still open.
         *
         *  @return                 True if this input stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _isOpen; }

        /**
         *  Closes this input stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close() { _isOpen = false; }

        /**
         *  Returns the total number of bytes written to this memory
         *  stream so far.
         *
         *  The method can be called even after the memory output stream
         *  is closed, returning the correct result.
         *
         *  @return             The total number of bytes written to 
         *                      this memory stream so far.
         *  @since 1.0
         */
        UIntSize                getDataSize() const { return _data.size(); }

        /**
         *  Returns a pointer to memory stream's data.
         *
         *  The returned pointer points to a memory area managed internally
         *  by the memory output stream. This area contains getDataSize() bytes
         *  of data and can be read, but not written to, by the application.
         *
         *  The method can be called even after the memory output stream
         *  is closed, returning the correct result.
         *
         *  @return             A pointer to memory stream's data.
         *  @since 1.0
         */
        const UInt8 *           getData() const { return _data.getData(); }

        /**
         *  Returns a copy of the current contents of this output stream.
         *
         *  @return             The copy of the current contents of this output stream.
         *  @since 1.0
         */
        Array<UInt8>            getContents() const { return _data; }

        //////////
        //  IByteOutput
    public:
        /**
         *  This call has no effect on memory output streams.
         *
         *  This method can only be called on open streams.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException) { require(_isOpen); }

        /**
         *  Writes a single byte to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param byte             The byte to write to this stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(UInt8 byte) throw(IOException);

        /**
         *  Writes a sequence of bytes to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The address of the first byte of the sequence
         *                          that shall be written to the stream. The
         *                          sequence must be at least \a bytesToWrite
         *                          bytes long; if it is longer then only the
         *                          forst \a bytesToWrite bytes are written.
         *  @param bytesToWrite     The number of bytes to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException);

        //////////
        //  Implementation
    private:
        bool                    _isOpen;
        Array<UInt8>            _data;
    };

    //////////
    //  I/O buffering
    /**
     *  @class BufferedInputStream CARL.hpp
     *  A byte input stream that provides buffering facilities.
     *
     *  A buffered input stream acts as a buffer for an existing input
     *  stream, which greatly increases performance when reading the
     *  input in small chunks.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC BufferedInputStream : public IByteInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        BufferedInputStream(const BufferedInputStream &);
        void                    operator = (const BufferedInputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created buffered input stream to read
         *  raw bytes from the specified stream and perform buffering.
         *  The internal buffer of a default size is used.
         *
         *  @param pSource      A pointer to a byte input stream where
         *                      the actual data comes from.
         *  @since 1.0
         */
        explicit BufferedInputStream(IByteInput * pSource);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created buffered input stream to read
         *  raw bytes from the specified stream and perform buffering.
         *  The internal buffer of a specified size is used.
         *
         *  @param pSource      A pointer to a byte input stream where
         *                      the actual data comes from.
         *  @param bufferSize   The size of the buffer to use.
         *  @since 1.0
         */
        BufferedInputStream(IByteInput * pSource, UIntSize bufferSize);

        /**
         *  The class destructor.
         *
         *  Closes the input stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~BufferedInputStream() { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this input stream is still open.
         *
         *  @return                 True if this input stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _pBuffer != NULL; }

        /**
         *  Closes this input stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteInput
    public:
        /**
         *  Returns the number of bytes that can be read from this input 
         *  stream without blocking.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The number of bytes that can be read from this
         *                          input stream without blocking.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        available() throw(IOException);

        /**
         *  Reads and returns the next byte or -1 if an end of stream is reached.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The next input byte, or -1 if an end of stream is reached.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual int             read() throw(IOException);

        /**
         *  Reads a sequence of bytes into the specified buffer.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is read, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The buffer where bytes read from the stream
         *                          shall be placed.
         *  @param bytesToRead      The maximum number of bytes to read.
         *  @return                 The number of bytes actually read. This can
         *                          be smaller than \a bytesToRead if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException);

        /**
         *  Skips a sequence of input bytes.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is skipped, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  This method can only be called on open streams.
         *
         *  @param bytesToSkip      The maximum number of bytes to skip.
         *  @return                 The number of bytes actually skipped. This can
         *                          be smaller than \a bytesToSkip if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException);

        //////////
        //  Implementation
    private:
        static const UIntSize   _DefaultBufferSize = 4096;
        IByteInput *            _pSource;
        UIntSize                _bufferSize;
        UInt8 *                 _pBuffer;
        UIntSize                _bytesInBuffer;
        UIntSize                _currentPosition;

        //  Helpers
        void                    _refillBuffer();
    };

    /**
     *  @class BufferedOutputStream CARL.hpp
     *  A byte output stream that provides buffering facilities.
     *
     *  A buffered output stream acts as a buffer for an existing output
     *  stream, which greatly increases performance when writing the
     *  output in small chunks.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC BufferedOutputStream : public IByteOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        BufferedOutputStream(const BufferedOutputStream &);
        void                    operator = (const BufferedOutputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created buffered output stream to write
         *  raw bytes to the specified stream and perform buffering.
         *  The internal buffer of a default size is used.
         *
         *  @param pTarget      A pointer to a byte output stream where
         *                      the actual data if forwarded to.
         *  @since 1.0
         */
        explicit BufferedOutputStream(IByteOutput * pTarget);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created buffered output stream to write
         *  raw bytes to the specified stream and perform buffering.
         *  The internal buffer of a specified size is used.
         *
         *  @param pTarget      A pointer to a byte output stream where
         *                      the actual data if forwarded to.
         *  @param bufferSize   The size of the buffer to use.
         *  @since 1.0
         */
        BufferedOutputStream(IByteOutput * pTarget, UIntSize bufferSize);

        /**
         *  The class destructor.
         *
         *  Closes the output stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~BufferedOutputStream() { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this output stream is still open.
         *
         *  @return                 True if this output stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _pBuffer != NULL; }

        /**
         *  Closes this output stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteOutput
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  Causes the data currently buffered within this stream to be
         *  forwarded to the target stream, then flushes the target stream.
         *
         *  This method can only be called on open streams.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException);

        /**
         *  Writes a single byte to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param byte             The byte to write to this stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(UInt8 byte) throw(IOException);

        /**
         *  Writes a sequence of bytes to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The address of the first byte of the sequence
         *                          that shall be written to the stream. The
         *                          sequence must be at least \a bytesToWrite
         *                          bytes long; if it is longer then only the
         *                          forst \a bytesToWrite bytes are written.
         *  @param bytesToWrite     The number of bytes to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException);

        //////////
        //  Implementation
    private:
        static const UIntSize   _DefaultBufferSize = 4096;
        IByteOutput *           _pTarget;
        UIntSize                _bufferSize;
        UInt8 *                 _pBuffer;
        UIntSize                _bytesInBuffer;

        //  Helpers
        void                    _flushBuffer();
    };

    //////////
    //  Data I/O
    /**
     *  @class DataInputStream CARL.hpp
     *  A data input stream that reads raw bytes from an existing
     *  byte input stream.
     *
     *  A data input stream reads raw bytes from an existing byte input
     *  stream and translates these bytes into values of basic types
     *  as necessary.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DataInputStream : public IDataInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        DataInputStream(const DataInputStream &);
        void                    operator = (const DataInputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created data input stream to read
         *  raw bytes from the specified stream.
         *
         *  @param pSource      A pointer to a byte input stream where
         *                      the actual data comes from.
         *  @since 1.0
         */
        explicit DataInputStream(IByteInput * pSource);

        /**
         *  The class destructor.
         *
         *  Closes the input stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~DataInputStream() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if this input stream is still open.
         *
         *  @return                 True if this input stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _pSource != NULL; }

        /**
         *  Closes this input stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close() { _pSource = NULL; }

        //////////
        //  IByteInput
    public:
        /**
         *  Returns the number of bytes that can be read from this input 
         *  stream without blocking.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The number of bytes that can be read from this
         *                          input stream without blocking.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        available() throw(IOException) { require(isOpen()); return _pSource->available(); }

        /**
         *  Reads and returns the next byte or -1 if an end of stream is reached.
         *
         *  This method can only be called on open streams.
         *
         *  @return                 The next input byte, or -1 if an end of stream is reached.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual int             read() throw(IOException) { require(isOpen()); return _pSource->read(); }

        /**
         *  Reads a sequence of bytes into the specified buffer.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is read, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The buffer where bytes read from the stream
         *                          shall be placed.
         *  @param bytesToRead      The maximum number of bytes to read.
         *  @return                 The number of bytes actually read. This can
         *                          be smaller than \a bytesToRead if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException) { require(isOpen()); return _pSource->read(pBuffer, bytesToRead); }

        /**
         *  Skips a sequence of input bytes.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is skipped, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  This method can only be called on open streams.
         *
         *  @param bytesToSkip      The maximum number of bytes to skip.
         *  @return                 The number of bytes actually skipped. This can
         *                          be smaller than \a bytesToSkip if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException) { require(isOpen()); return _pSource->skip(bytesToSkip); }

        //////////
        //  IDataInput
    public:
        /**
         *  Reads a boolean value from this stream.
         *
         *  A boolean value is expected to be represented as a single byte
         *  with value 0 (false) or 1 (true). Any other value is an error.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            readBool() throw(IOException);

        /**
         *  Reads an UInt8 value from this stream.
         *
         *  An UInt8 value is expected to be represented as a single byte.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt8           readUInt8() throw(IOException);

        /**
         *  Reads an UInt16 value from this stream.
         *
         *  An UInt16 value is expected to be represented as 2 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt16          readUInt16() throw(IOException);

        /**
         *  Reads an UInt32 value from this stream.
         *
         *  An UInt32 value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt32          readUInt32() throw(IOException);

        /**
         *  Reads an UInt64 value from this stream.
         *
         *  An UInt64 value is expected to be represented as 8 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt64          readUInt64() throw(IOException);

        /**
         *  Reads an Int8 value from this stream.
         *
         *  An Int8 value is expected to be represented as a single byte.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int8            readInt8() throw(IOException);

        /**
         *  Reads an Int16 value from this stream.
         *
         *  An Int16 value is expected to be represented as 2 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int16           readInt16() throw(IOException);

        /**
         *  Reads an Int32 value from this stream.
         *
         *  An Int32 value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int32           readInt32() throw(IOException);

        /**
         *  Reads an Int64 value from this stream.
         *
         *  An Int64 value is expected to be represented as 8 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int64           readInt64() throw(IOException);

        /**
         *  Reads a Real32 value from this stream.
         *
         *  A Real32 value is expected to be represented as 4 consecutive IEEE754
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Real32          readReal32() throw(IOException);

        /**
         *  Reads a Real64 value from this stream.
         *
         *  A Real64 value is expected to be represented as 8 consecutive IEEE754
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Real64          readReal64() throw(IOException);

        /**
         *  Reads a Char value from this stream.
         *
         *  An Char value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Char            readChar() throw(IOException);

        //////////
        //  Implementation
    private:
        IByteInput *            _pSource;
    };

    /**
     *  @class DataOutputStream CARL.hpp
     *  A data output stream that writes raw bytes to an existing
     *  byte output stream.
     *
     *  A data output stream translates values of basic types into raw byte
     *  sequences and forwards these sequences to an existing byte output
     *  stream.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DataOutputStream : public IDataOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        DataOutputStream(const DataOutputStream &);
        void                    operator = (const DataOutputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created data output stream to write
         *  raw bytes to the specified stream.
         *
         *  @param pTarget      A pointer to a byte output stream where
         *                      the actual data if forwarded to.
         *  @since 1.0
         */
        explicit DataOutputStream(IByteOutput * pTarget);

        /**
         *  The class destructor.
         *
         *  Closes the output stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~DataOutputStream() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if this output stream is still open.
         *
         *  @return                 True if this output stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _pTarget != NULL; }

        /**
         *  Closes this output stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close() { _pTarget = NULL; }

        //////////
        //  IByteOutput
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  Flushes the target stream.
         *
         *  This method can only be called on open streams.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException) { require(isOpen()); _pTarget->flush(); }

        /**
         *  Writes a single byte to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param byte             The byte to write to this stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(UInt8 byte) throw(IOException) { require(isOpen()); _pTarget->write(byte); }

        /**
         *  Writes a sequence of bytes to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer          The address of the first byte of the sequence
         *                          that shall be written to the stream. The
         *                          sequence must be at least \a bytesToWrite
         *                          bytes long; if it is longer then only the
         *                          forst \a bytesToWrite bytes are written.
         *  @param bytesToWrite     The number of bytes to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException) { require(isOpen()); _pTarget->write(pBuffer, bytesToWrite); }

        //////////
        //  IDataOutput
    public:
        /**
         *  Writes a boolean value to this stream.
         *
         *  A boolean value is represented as a single byte with value 0 
         *  (false) or 1 (true).
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeBool(bool value) throw(IOException);

        /**
         *  Writes an UInt8 value to this stream.
         *
         *  An UInt8 value is represented as a single byte.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt8(UInt8 value) throw(IOException);

        /**
         *  Writes an UInt16 value to this stream.
         *
         *  An UInt16 value is represented as 2 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt16(UInt16 value) throw(IOException);

        /**
         *  Writes an UInt32 value to this stream.
         *
         *  An UInt32 value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt32(UInt32 value) throw(IOException);

        /**
         *  Writes an UInt64 value to this stream.
         *
         *  An UInt64 value is represented as 8 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt64(UInt64 value) throw(IOException);

        /**
         *  Writes an Int8 value to this stream.
         *
         *  An Int8 value is represented as a single byte.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt8(Int8 value) throw(IOException);

        /**
         *  Writes an Int16 value to this stream.
         *
         *  An Int16 value is represented as 2 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt16(Int16 value) throw(IOException);

        /**
         *  Writes an Int32 value to this stream.
         *
         *  An Int32 value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt32(Int32 value) throw(IOException);

        /**
         *  Writes an Int64 value to this stream.
         *
         *  An Int64 value is represented as 8 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt64(Int64 value) throw(IOException);

        /**
         *  Writes a Real32 value to this stream.
         *
         *  A Real32 value is represented as 4 consecutive IEEE754 bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeReal32(Real32 value) throw(IOException);

        /**
         *  Writes a Real64 value to this stream.
         *
         *  A Real64 value is represented as 8 consecutive IEEE754 bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeReal64(Real64 value) throw(IOException);

        /**
         *  Writes a Char value to this stream.
         *
         *  A Char value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  This method can only be called on open streams.
         *
         *  @param value        The value to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeChar(Char value) throw(IOException);

        //////////
        //  Implementation
    private:
        IByteOutput *           _pTarget;
    };

    //////////
    //  Basic text I/O
    /**
     *  @class TextInputStream CARL.hpp
     *  A text input stream that reads raw bytes from an existing
     *  byte input stream.
     *
     *  A text input stream reads raw bytes from an existing byte input
     *  stream and translates these bytes into characters using a specified 
     *  character set.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC TextInputStream : public ITextInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        TextInputStream(const TextInputStream &);
        void                    operator = (const TextInputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created buffered data stream to read
         *  raw bytes from the specified stream. The current default
         *  character set is used to translate bytes into characters.
         *
         *  @param pSource      A pointer to a byte input stream where
         *                      the raw data bytes come from.
         *  @since 1.0
         */
        explicit TextInputStream(IByteInput * pSource);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created buffered data stream to read
         *  raw bytes from the specified stream. The specified
         *  character set is used to translate bytes into characters.
         *
         *  @param pSource      A pointer to a byte input stream where
         *                      the raw data bytes come from.
         *  @param pCharacterSet    The character set to use for
         *                      byte-to-character translation.
         *  @since 1.0
         */
        TextInputStream(IByteInput * pSource, CharacterSet * pCharacterSet);

        /**
         *  The class destructor.
         *
         *  Closes the input stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~TextInputStream() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if this input stream is still open.
         *
         *  @return                 True if this input stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _pSource != NULL; }

        /**
         *  Closes this input stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close() { _pSource = NULL; }

        /**
         *  Returns the character set used by this input stream.
         *
         *  This method can only be called on open streams.
         *
         *  @return             The character set used by this input stream.
         *  @since 1.0
         */
        CharacterSet *          getCharacterSet() const;

        /**
         *  Sets the character set used by this input stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pCharacterSet    The character set to use for 
         *                      byte-to-character translation.
         *  @since 1.0
         */
        void                    setCharacterSet(CharacterSet * pCharacterSet);

        //////////
        //  Operations
    public:
        /**
         *  Checks if a character can be read from this text input 
         *  stream without blocking.
         *
         *  This method can only be called on open streams.
         *
         *  @return             True if at least one character can be read from
         *                      this stream without blocking, false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            ready() throw(IOException);

        /**
         *  Reads a single character from this stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param c            If a character is read successfully, it is
         *                      assigned to this parameter. Otherwise this
         *                      parameter remains unchanged.
         *  @return             True if the character has been read
         *                      successfully; false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            read(Char & c) throw(IOException);

        /**
         *  Reads a sequence of characters from this stream.
         *
         *  The actual number of character read can be smaller than the
         *  requested number of characters  if an end-of-stream is reached 
         *  before a specified number of character has been read.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer      A pointer to a buffer where characters read
         *                      from the stream are placed. The buffer is
         *                      assumed to be at least \a charsToRead
         *                      characters long.
         *  @param charsToRead  A maximum number of characters to read.
         *  @return             The actual number of characters read.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(Char * pBuffer, UIntSize charsToRead) throw(IOException);

        /**
         *  Reads a line of characters from this stream.
         *
         *  Characters are read until either end-of-line or end-of-stream
         *  is detected, whichever happens first.
         *
         *  This method can only be called on open streams.
         *
         *  @param line         Upon successful read, this parameter is
         *                      assigned the characters of the line just 
         *                      read.
         *  @param discardEoln  If the line was terminated by an end-of-line
         *                      marker and this parameter is false, the
         *                      end-of-line marker is included into the
         *                      \a line. If the line was terminated by an
         *                      end-of-line marker and this parameter is true,
         *                      the end-of-line marker is not included into
         *                      the \a line. If the line was terminated by
         *                      an end-of-file marker, the end-of-line marker
         *                      is not included into the \a line regardless
         *                      of whether \a discardEoln is true or false.
         *  @return             True if the line has been read successfully; 
         *                      false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            readLine(String & line, bool discardEoln = false) throw(IOException);

        /**
         *  Skips a sequence of characters from this stream.
         *
         *  The actual number of character skipped can be smaller than the
         *  requested number of characters  if an end-of-stream is reached 
         *  before a specified number of character has been skipped.
         *
         *  This method can only be called on open streams.
         *
         *  @param charsToSkip  A maximum number of characters to skip.
         *  @return             The actual number of characters skipped.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize charsToSkip) throw(IOException);

        /**
         *  Un-reads the character to the stream.
         *
         *  Upon successful un-read, the next character read from the
         *  stream is guaranteed to be the last character un-read to
         *  that stream. Text input streams support an arbitrary
         *  un-read depth.
         *
         *  This method can only be called on open streams.
         *
         *  @param c            The character to push back into the 
         *                      input stream.
         *  @return             True if the character has been successfully
         *                      pushed back into the input stream; false
         *                      otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            unread(Char c) throw(IOException);

        //////////
        //  Implementation
    private:
        IByteInput *            _pSource;
        CharacterSet *          _pCharacterSet;
        Stack<Char>             _charsPushedBack;
        Array<UInt8>            _encoding;
        String                  _eolnSequence;
        Stack<Char>             _untranslatedCharsPushedBack;
        Stack<UInt8>            _bytesPushedBack;

        //  Helpers
        bool                    _read(Char & c) throw(IOException);
        bool                    _readByte(UInt8 & b) throw(IOException);
    };

    /**
     *  @class TextOutputStream CARL.hpp
     *  A text output stream that writes raw bytes to an existing
     *  byte output stream.
     *
     *  A text output stream translates characters into bytes using a
     *  specified character set and  writes these raw bytes to an 
     *  existing byte output stream.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC TextOutputStream : public ITextOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        TextOutputStream(const TextOutputStream &);
        void                    operator = (const TextOutputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created buffered data stream to write
         *  raw bytes to the specified stream. The current default
         *  character set is used to translate characters into bytes.
         *
         *  @param pTarget      A pointer to a byte output stream where
         *                      the raw data bytes are forwarded.
         *  @since 1.0
         */
        explicit TextOutputStream(IByteOutput * pTarget);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created buffered data stream to write
         *  raw bytes to the specified stream. The current default
         *  character set is used to translate characters into bytes.
         *
         *  @param pTarget      A pointer to a byte output stream where
         *                      the raw data bytes are forwarded.
         *  @param pCharacterSet    The character set to use for
         *                      character-to-byte translation.
         *  @since 1.0
         */
        TextOutputStream(IByteOutput * pTarget, CharacterSet * pCharacterSet);

        /**
         *  The class destructor.
         *
         *  Closes the input stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~TextOutputStream() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if this output stream is still open.
         *
         *  @return                 True if this output stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _pTarget != NULL;}

        /**
         *  Closes this output stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close() { _pTarget = NULL; }

        /**
         *  Returns the character set used by this output stream.
         *
         *  This method can only be called on open streams.
         *
         *  @return             The character set used by this output stream.
         *  @since 1.0
         */
        CharacterSet *          getCharacterSet() const;

        /**
         *  Sets the character set used by this output stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pCharacterSet    The character set to use for 
         *                      character-to-byte translation.
         *  @since 1.0
         */
        void                    setCharacterSet(CharacterSet * pCharacterSet);

        //////////
        //  Operations
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  Flushes the target stream.
         *
         *  This method can only be called on open streams.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException);

        /**
         *  Writes a character to this output stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(Char c) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer, UIntSize charsToWrite) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer      A pointer to the sequence of ISo-10646 characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  This method can only be called on open streams.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the stream. All characters from the
         *                      string are written, in order, to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const String & s) throw(IOException);

        /**
         *  Wrtes an end-of-line marker to the stream.
         *
         *  This method can only be called on open streams.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn() throw(IOException);

        /**
         *  Wrtes a character followed by end-of-line marker to the stream.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(Char c) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ISO-10646 characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const Char * pBuffer) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  This method can only be called on open streams.
         *
         *  @param pBuffer      A pointer to the sequence of ISo-10646 characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  This method can only be called on open streams.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the stream. All characters from the
         *                      string are written, in order, to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const String & s) throw(IOException);

        //////////
        //  Implementation
    private:
        IByteOutput *           _pTarget;
        CharacterSet *          _pCharacterSet;
        Array<UInt8>            _encoding;

        //  Helpers
        void                    _write(Char c) throw(IOException);
    };

    //////////
    //  String I/O
    /**
     *  @class StringInputStream CARL.hpp
     *  A text input stream for reading from an in-memory string.
     *
     *  String input streams allow treating strings as text streams.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC StringInputStream : public ITextInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        StringInputStream(const StringInputStream &);
        void                    operator = (const StringInputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created string input stream to read
         *  from a specified string. A private copy of the string contents
         *  is always made.
         *
         *  @param str          A string to read from.
         *  @since 1.0
         */
        StringInputStream(const String & str);

        /**
         *  The class destructor.
         *
         *  Closes the input stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~StringInputStream() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if this input stream is still open.
         *
         *  @return                 True if this input stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _isOpen; }

        /**
         *  Closes this input stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close() { _isOpen = false; }

        //////////
        //  ITextInput
    public:
        /**
         *  Checks if a character can be read from this text input 
         *  stream without blocking.
         *
         *  @return             True if at least one character can be read from
         *                      this stream without blocking, false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            ready() throw(IOException);

        /**
         *  Reads a single character from this stream.
         *
         *  @param c            If a character is read successfully, it is
         *                      assigned to this parameter. Otherwise this
         *                      parameter remains unchanged.
         *  @return             True if the character has been read
         *                      successfully; false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            read(Char & c) throw(IOException);

        /**
         *  Reads a sequence of characters from this stream.
         *
         *  The actual number of character read can be smaller than the
         *  requested number of characters  if an end-of-stream is reached 
         *  before a specified number of character has been read.
         *
         *  @param pBuffer      A pointer to a buffer where characters read
         *                      from the stream are placed. The buffer is
         *                      assumed to be at least \a charsToRead
         *                      characters long.
         *  @param charsToRead  A maximum number of characters to read.
         *  @return             The actual number of characters read.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(Char * pBuffer, UIntSize charsToRead) throw(IOException);

        /**
         *  Reads a line of characters from this stream.
         *
         *  Characters are read until either end-of-line or end-of-stream
         *  is detected, whichever happens first.
         *
         *  @param line         Upon successful read, this parameter is
         *                      assigned the characters of the line just 
         *                      read.
         *  @param discardEoln  If the line was terminated by an end-of-line
         *                      marker and this parameter is false, the
         *                      end-of-line marker is included into the
         *                      \a line. If the line was terminated by an
         *                      end-of-line marker and this parameter is true,
         *                      the end-of-line marker is not included into
         *                      the \a line. If the line was terminated by
         *                      an end-of-file marker, the end-of-line marker
         *                      is not included into the \a line regardless
         *                      of whether \a discardEoln is true or false.
         *  @return             True if the line has been read successfully; 
         *                      false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            readLine(String & line, bool discardEoln = false) throw(IOException);

        /**
         *  Skips a sequence of characters from this stream.
         *
         *  The actual number of character skipped can be smaller than the
         *  requested number of characters  if an end-of-stream is reached 
         *  before a specified number of character has been skipped.
         *
         *  @param charsToSkip  A maximum number of characters to skip.
         *  @return             The actual number of characters skipped.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize charsToSkip) throw(IOException);

        /**
         *  Un-reads the character to the stream.
         *
         *  Upon successful un-read, the next character read from the
         *  stream is guaranteed to be the last character un-read to
         *  that stream. Text streams always guarantee that at least one
         *  character can be un-read, although some concrete text input
         *  streams may allow un-reading mul;tiple characters.
         *
         *  @param c            The character to push back into the 
         *                      input stream.
         *  @return             True if the character has been successfully
         *                      pushed back into the input stream; false
         *                      otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            unread(Char c) throw(IOException);

        //////////
        //  Implementation
    private:
        bool                    _isOpen;
        String                  _str;
        UIntSize                _currentPosition;
        Stack<Char>             _charsPushedBack;
    };

    /**
     *  @class StringOutputStream CARL.hpp
     *  A text output stream for writing to an in-memory string.
     *
     *  String output streams allow treating strings as text streams.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC StringOutputStream : public ITextOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        StringOutputStream(const StringOutputStream &);
        void                    operator = (const StringOutputStream &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The default constructor.
         *
         *  Initializes a newly created string output stream to wrote
         *  to an initially empty internally maintained string.
         *
         *  @since 1.0
         */
        StringOutputStream();

        /**
         *  The class destructor.
         *
         *  Closes the output stream if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~StringOutputStream() {}

        //////////
        //  Operations
    public:
        /**
         *  Checks if this output stream is still open.
         *
         *  @return                 True if this output stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _isOpen; }

        /**
         *  Closes this output stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close() { _isOpen = false; }

        /**
         *  Returns the current contents of this output stream.
         *
         *  The call is valid even after the stream has been closed.
         *
         *  @return                 The current contents of this output stream.
         *  @since 1.0
         */
        String                  getContents() const { return _str; }

        //////////
        //  ITextOutput
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  If there is any data that has been written to this output stream
         *  and not yet forwarded to the stream's destination, this data is 
         *  written to the stream's destination immediately.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException) { require(_isOpen); }

        /**
         *  Writes a character to this output stream.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(Char c) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer, UIntSize charsToWrite) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the sequence of ISo-10646 characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the stream. All characters from the
         *                      string are written, in order, to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const String & s) throw(IOException);

        /**
         *  Wrtes an end-of-line marker to the stream.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn() throw(IOException);

        /**
         *  Wrtes a character followed by end-of-line marker to the stream.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(Char c) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ISO-10646 characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const Char * pBuffer) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the sequence of ISo-10646 characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException);

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the stream. All characters from the
         *                      string are written, in order, to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const String & s) throw(IOException);

        //////////
        //  Implementation
    private:
        bool                    _isOpen;
        String                  _str;
    };

    //////////
    //  Binary file I/O
    /**
     *  @class BinaryFileReader CARL.hpp
     *  A helper class for buffered reading of binary files.
     *
     *  A binary file reader combines a file input stream, a buffering
     *  facilities and a data input stream, all working together to
     *  allow buffered reading of binary files.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC BinaryFileReader : public IDataInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        BinaryFileReader(const BinaryFileReader &);
        void                    operator = (const BinaryFileReader &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file reader to read
         *  from an existing binary file. The binary file reader uses
         *  default buffer size.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileReader(const char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file reader to read
         *  from an existing binary file. The binary file reader uses
         *  default buffer size.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing a name of the file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileReader(const Char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file reader to read
         *  from an existing binary file. The binary file reader uses
         *  default buffer size.
         *
         *  @param fileName     A string representing a name of the 
         *                      file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileReader(const String & fileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file reader to read
         *  from an existing binary file. The binary file reader uses
         *  default buffer size.
         *
         *  @param pathName     A path to the file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileReader(const PathName & pathName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file reader to read
         *  from an existing binary file. The binary file reader uses
         *  the specified buffer size.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to read from.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileReader(const char * pFileName, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file reader to read
         *  from an existing binary file. The binary file reader uses
         *  the specified buffer size.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing a name of the file to read from.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileReader(const Char * pFileName, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file reader to read
         *  from an existing binary file. The binary file reader uses
         *  the specified buffer size.
         *
         *  @param fileName     A string representing a name of the 
         *                      file to read from.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileReader(const String & fileName, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file reader to read
         *  from an existing binary file. The binary file reader uses
         *  the specified buffer size.
         *
         *  @param pathName     A path to the file to read from.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileReader(const PathName & pathName, UIntSize bufferSize) throw(IOException);

        /**
         *  The class destructor.
         *
         *  Closes the reader if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~BinaryFileReader() throw(IOException) { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this reader is still open.
         *
         *  @return                 True if this reader is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _fis.isOpen(); }

        /**
         *  Closes this reader.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteInput
        /**
         *  Returns the number of bytes that can be read from this reader
         *  without blocking.
         *
         *  @return                 The number of bytes that can be read from this
         *                          reader without blocking.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        available() throw(IOException) { require(isOpen()); return _dis.available(); }

        /**
         *  Reads and returns the next byte or -1 if an end of stream is reached.
         *
         *  @return                 The next input byte, or -1 if an end of stream is reached.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual int             read() throw(IOException) { require(isOpen()); return _dis.read(); }

        /**
         *  Reads a sequence of bytes into the specified buffer.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is read, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  @param pBuffer          The buffer where bytes read from the stream
         *                          shall be placed.
         *  @param bytesToRead      The maximum number of bytes to read.
         *  @return                 The number of bytes actually read. This can
         *                          be smaller than \a bytesToRead if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(void * pBuffer, UIntSize bytesToRead) throw(IOException) { require(isOpen()); return _dis.read(pBuffer, bytesToRead); }

        /**
         *  Skips a sequence of input bytes.
         *
         *  If an end of the stream is reached before the required number
         *  of bytes is skipped, the method returns immediately, signaling
         *  the fact to the caller.
         *
         *  @param bytesToSkip      The maximum number of bytes to skip.
         *  @return                 The number of bytes actually skipped. This can
         *                          be smaller than \a bytesToSkip if an end
         *                          of stream has been detected.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize bytesToSkip) throw(IOException) { require(isOpen()); return _dis.skip(bytesToSkip); }

        //////////
        //  IDataInput
        /**
         *  Reads a boolean value from this reader.
         *
         *  A boolean value is expected to be represented as a single byte
         *  with value 0 (false) or 1 (true). Any other value is an error.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            readBool() throw(IOException) { require(isOpen()); return _dis.readBool(); }

        /**
         *  Reads an UInt8 value from this reader.
         *
         *  An UInt8 value is expected to be represented as a single byte.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt8           readUInt8() throw(IOException) { require(isOpen()); return _dis.readUInt8(); }

        /**
         *  Reads an UInt16 value from this reader.
         *
         *  An UInt16 value is expected to be represented as 2 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt16          readUInt16() throw(IOException) { require(isOpen()); return _dis.readUInt16(); }

        /**
         *  Reads an UInt32 value from this reader.
         *
         *  An UInt32 value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt32          readUInt32() throw(IOException) { require(isOpen()); return _dis.readUInt32(); }

        /**
         *  Reads an UInt64 value from this reader.
         *
         *  An UInt64 value is expected to be represented as 8 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UInt64          readUInt64() throw(IOException) { require(isOpen()); return _dis.readUInt64(); }

        /**
         *  Reads an Int8 value from this reader.
         *
         *  An Int8 value is expected to be represented as a single byte.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int8            readInt8() throw(IOException) { require(isOpen()); return _dis.readInt8(); }

        /**
         *  Reads an Int16 value from this reader.
         *
         *  An Int16 value is expected to be represented as 2 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int16           readInt16() throw(IOException) { require(isOpen()); return _dis.readInt16(); }

        /**
         *  Reads an Int32 value from this reader.
         *
         *  An Int32 value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int32           readInt32() throw(IOException) { require(isOpen()); return _dis.readInt32(); }

        /**
         *  Reads an Int64 value from this reader.
         *
         *  An Int64 value is expected to be represented as 8 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Int64           readInt64() throw(IOException) { require(isOpen()); return _dis.readInt64(); }

        /**
         *  Reads a Real32 value from this reader.
         *
         *  A Real32 value is expected to be represented as 4 consecutive IEEE754
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Real32          readReal32() throw(IOException) { require(isOpen()); return _dis.readReal32(); }

        /**
         *  Reads a Real64 value from this reader.
         *
         *  A Real64 value is expected to be represented as 8 consecutive IEEE754
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Real64          readReal64() throw(IOException) { require(isOpen()); return _dis.readReal64(); }

        /**
         *  Reads a Char value from this reader.
         *
         *  An Char value is expected to be represented as 4 consecutive
         *  bytes stored in big-endian (network) order regardless of the byte
         *  order used by the host platform.
         *
         *  @return             The value read from the reader.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual Char            readChar() throw(IOException) { require(isOpen()); return _dis.readChar(); }

        //////////
        //  Implementation
    private:
        FileInputStream         _fis;
        BufferedInputStream     _bis;
        DataInputStream         _dis;
    };

    /**
     *  @class BinaryFileWriter CARL.hpp
     *  A helper class for buffered writing of binary files.
     *
     *  A binary file writer combines a file output stream, a buffering
     *  facilities and a data output stream, all working together to
     *  allow buffered writing of binary files.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC BinaryFileWriter : public IDataOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        BinaryFileWriter(const BinaryFileWriter &);
        void                    operator = (const BinaryFileWriter &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file writer to write
         *  to an existing binary file. The binary file writer uses
         *  default buffer size.
         *
         *  The specified file is created if it does not already exist, or
         *  truncated to zero length if is already exists.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileWriter(const char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file writer to write
         *  to an existing binary file. The binary file writer uses
         *  default buffer size.
         *
         *  The specified file is created if it does not already exist, or
         *  truncated to zero length if is already exists.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing a name of the file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileWriter(const Char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file writer to write
         *  to an existing binary file. The binary file writer uses
         *  default buffer size.
         *
         *  The specified file is created if it does not already exist, or
         *  truncated to zero length if is already exists.
         *
         *  @param fileName     A string representing a name of the 
         *                      file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileWriter(const String & fileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file writer to write
         *  to an existing binary file. The binary file writer uses
         *  default buffer size.
         *
         *  The specified file is created if it does not already exist, or
         *  truncated to zero length if is already exists.
         *
         *  @param pathName     A path to the file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileWriter(const PathName & pathName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file writer to write
         *  to an existing binary file. The binary file writer uses
         *  the specified buffer size.
         *
         *  The specified file is created if it does not already exist, or
         *  truncated to zero length if is already exists.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to write to.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileWriter(const char * pFileName, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file writer to write
         *  to an existing binary file. The binary file writer uses
         *  the specified buffer size.
         *
         *  The specified file is created if it does not already exist, or
         *  truncated to zero length if is already exists.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing a name of the file to write to.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileWriter(const Char * pFileName, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file writer to write
         *  to an existing binary file. The binary file writer uses
         *  the specified buffer size.
         *
         *  The specified file is created if it does not already exist, or
         *  truncated to zero length if is already exists.
         *
         *  @param fileName     A string representing a name of the 
         *                      file to write to.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileWriter(const String & fileName, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created binary file writer to write
         *  to an existing binary file. The binary file writer uses
         *  the specified buffer size.
         *
         *  The specified file is created if it does not already exist, or
         *  truncated to zero length if is already exists.
         *
         *  @param pathName     A path to the file to write to.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        BinaryFileWriter(const PathName & pathName, UIntSize bufferSize) throw(IOException);

        /**
         *  The class destructor.
         *
         *  Closes the writer if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~BinaryFileWriter() throw(IOException) { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this input stream is still open.
         *
         *  @return                 True if this input stream is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _fos.isOpen(); }

        /**
         *  Closes this input stream.
         *
         *  An attempt to close a stream that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        //////////
        //  IByteOutput
        /**
         *  Flushes any unwritten data.
         *
         *  If there is any data that has been written to this writer
         *  and not yet forwarded to the stream's destination, this data is 
         *  written to the stream's destination immediately.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException) { require(isOpen()); _dos.flush(); }

        /**
         *  Writes a single byte to the writer.
         *
         *  @param byte             The byte to write to this writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(UInt8 byte) throw(IOException) { require(isOpen()); _dos.write(byte); }

        /**
         *  Writes a sequence of bytes to the writer.
         *
         *  @param pBuffer          The address of the first byte of the sequence
         *                          that shall be written to the writer. The
         *                          sequence must be at least \a bytesToWrite
         *                          bytes long; if it is longer then only the
         *                          forst \a bytesToWrite bytes are written.
         *  @param bytesToWrite     The number of bytes to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const void * pBuffer, UIntSize bytesToWrite) throw(IOException) { require(isOpen()); _dos.write(pBuffer, bytesToWrite); }

        //////////
        //  IDataOutput
        /**
         *  Writes a boolean value to this writer.
         *
         *  A boolean value is represented as a single byte with value 0 
         *  (false) or 1 (true).
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeBool(bool value) throw(IOException) { require(isOpen()); _dos.writeBool(value); }

        /**
         *  Writes an UInt8 value to this writer.
         *
         *  An UInt8 value is represented as a single byte.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt8(UInt8 value) throw(IOException) { require(isOpen()); _dos.writeUInt8(value); }

        /**
         *  Writes an UInt16 value to this writer.
         *
         *  An UInt16 value is represented as 2 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt16(UInt16 value) throw(IOException) { require(isOpen()); _dos.writeUInt16(value); }

        /**
         *  Writes an UInt32 value to this writer.
         *
         *  An UInt32 value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt32(UInt32 value) throw(IOException) { require(isOpen()); _dos.writeUInt32(value); }

        /**
         *  Writes an UInt64 value to this writer.
         *
         *  An UInt64 value is represented as 8 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeUInt64(UInt64 value) throw(IOException) { require(isOpen()); _dos.writeUInt64(value); }

        /**
         *  Writes an Int8 value to this writer.
         *
         *  An Int8 value is represented as a single byte.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt8(Int8 value) throw(IOException) { require(isOpen()); _dos.writeInt8(value); }

        /**
         *  Writes an Int16 value to this writer.
         *
         *  An Int16 value is represented as 2 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt16(Int16 value) throw(IOException) { require(isOpen()); _dos.writeInt16(value); }

        /**
         *  Writes an Int32 value to this writer.
         *
         *  An Int32 value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt32(Int32 value) throw(IOException) { require(isOpen()); _dos.writeInt32(value); }

        /**
         *  Writes an Int64 value to this writer.
         *
         *  An Int64 value is represented as 8 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeInt64(Int64 value) throw(IOException) { require(isOpen()); _dos.writeInt64(value); }

        /**
         *  Writes a Real32 value to this writer.
         *
         *  A Real32 value is represented as 4 consecutive IEEE754 bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeReal32(Real32 value) throw(IOException) { require(isOpen()); _dos.writeReal32(value); }

        /**
         *  Writes a Real64 value to this writer.
         *
         *  A Real64 value is represented as 8 consecutive IEEE754 bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeReal64(Real64 value) throw(IOException) { require(isOpen()); _dos.writeReal64(value); }

        /**
         *  Writes a Char value to this writer.
         *
         *  A Char value is represented as 4 consecutive bytes stored in 
         *  big-endian (network) order regardless of the byte order used by 
         *  the host platform.
         *
         *  @param value        The value to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeChar(Char value) throw(IOException) { require(isOpen()); _dos.writeChar(value); }

        //////////
        //  Implementation
    private:
        FileOutputStream        _fos;
        BufferedOutputStream    _bos;
        DataOutputStream        _dos;
    };

    //////////
    //  Text file I/O
    /**
     *  @class TextFileReader CARL.hpp
     *  A helper class for buffered reading of text files.
     *
     *  A text file reader combines a file input stream, a buffering
     *  facilities and a text input stream, all working together to
     *  allow buffered reading of text files.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC TextFileReader : public ITextInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        TextFileReader(const TextFileReader &);
        void                    operator = (const TextFileReader &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  default buffer size and current default character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  default buffer size and current default character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing a name of the file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const Char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  default buffer size and current default character set.
         *
         *  @param fileName     A string representing a name of 
         *                      the file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const String & fileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  default buffer size and current default character set.
         *
         *  @param pathName     A path to the file to read from.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const PathName & pathName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  default buffer size and the specified character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to read from.
         *  @param pCharacterSet    A character set to use for 
         *                      byte-to-character translation.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const char * pFileName, CharacterSet * pCharacterSet) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  default buffer size and the specified character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing a name of the file to read from.
         *  @param pCharacterSet    A character set to use for 
         *                      byte-to-character translation.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const Char * pFileName, CharacterSet * pCharacterSet) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  default buffer size and the specified character set.
         *
         *  @param fileName     A string representing a name of 
         *                      the file to read from.
         *  @param pCharacterSet    A character set to use for 
         *                      byte-to-character translation.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const String & fileName, CharacterSet * pCharacterSet) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  default buffer size and the specified character set.
         *
         *  @param pathName     A path to the file to read from.
         *  @param pCharacterSet    A character set to use for 
         *                      byte-to-character translation.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const PathName & pathName, CharacterSet * pCharacterSet) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  specified buffer size and character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to read from.
         *  @param pCharacterSet    A character set to use for 
         *                      byte-to-character translation.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const char * pFileName, CharacterSet * pCharacterSet, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  specified buffer size and character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ISo-10646 string
         *                      representing a name of the file to read from.
         *  @param pCharacterSet    A character set to use for 
         *                      byte-to-character translation.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const Char * pFileName, CharacterSet * pCharacterSet, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  specified buffer size and character set.
         *
         *  @param fileName     A string representing a name of 
         *                      the file to read from.
         *  @param pCharacterSet    A character set to use for 
         *                      byte-to-character translation.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const String & fileName, CharacterSet * pCharacterSet, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file reader to read
         *  from an existing text file. The text file reader uses
         *  specified buffer size and character set.
         *
         *  @param pathName     A path to the file to read from.
         *  @param pCharacterSet    A character set to use for 
         *                      byte-to-character translation.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileReader(const PathName & pathName, CharacterSet * pCharacterSet, UIntSize bufferSize) throw(IOException);

        /**
         *  The class destructor.
         *
         *  Closes the reader if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~TextFileReader() throw(IOException) { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this reader is still open.
         *
         *  @return                 True if this reader is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _fis.isOpen(); }

        /**
         *  Closes this reader.
         *
         *  An attempt to close a reader that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        /**
         *  Returns the character set used by this reader.
         *
         *  This method can only be called on open readers.
         *
         *  @return             The character set used by this reader.
         *  @since 1.0
         */
        CharacterSet *          getCharacterSet() const { return _tis.getCharacterSet();}

        /**
         *  Sets the character set used by this reader.
         *
         *  This method can only be called on open readers.
         *
         *  @param pCharacterSet    The character set to use for 
         *                      byte-to-character translation.
         *  @since 1.0
         */
        void                    setCharacterSet(CharacterSet * pCharacterSet) { _tis.setCharacterSet(pCharacterSet); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if a character can be read from this reader
         *  without blocking.
         *
         *  This method can only be called on open readers.
         *
         *  @return             True if at least one character can be read from
         *                      this reader without blocking, false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            ready() throw(IOException) { require(isOpen()); return _tis.ready(); }

        /**
         *  Reads a single character from this reader.
         *
         *  This method can only be called on open readers.
         *
         *  @param c            If a character is read successfully, it is
         *                      assigned to this parameter. Otherwise this
         *                      parameter remains unchanged.
         *  @return             True if the character has been read
         *                      successfully; false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            read(Char & c) throw(IOException) { require(isOpen()); return _tis.read(c); }

        /**
         *  Reads a sequence of characters from this reader.
         *
         *  The actual number of character read can be smaller than the
         *  requested number of characters  if an end-of-stream is reached 
         *  before a specified number of character has been read.
         *
         *  This method can only be called on open readers.
         *
         *  @param pBuffer      A pointer to a buffer where characters read
         *                      from the reader are placed. The buffer is
         *                      assumed to be at least \a charsToRead
         *                      characters long.
         *  @param charsToRead  A maximum number of characters to read.
         *  @return             The actual number of characters read.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(Char * pBuffer, UIntSize charsToRead) throw(IOException) { require(isOpen()); return _tis.read(pBuffer, charsToRead); }

        /**
         *  Reads a line of characters from this reader.
         *
         *  Characters are read until either end-of-line or end-of-stream
         *  is detected, whichever happens first.
         *
         *  This method can only be called on open readers.
         *
         *  @param line         Upon successful read, this parameter is
         *                      assigned the characters of the line just 
         *                      read.
         *  @param discardEoln  If the line was terminated by an end-of-line
         *                      marker and this parameter is false, the
         *                      end-of-line marker is included into the
         *                      \a line. If the line was terminated by an
         *                      end-of-line marker and this parameter is true,
         *                      the end-of-line marker is not included into
         *                      the \a line. If the line was terminated by
         *                      an end-of-file marker, the end-of-line marker
         *                      is not included into the \a line regardless
         *                      of whether \a discardEoln is true or false.
         *  @return             True if the line has been read successfully; 
         *                      false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            readLine(String & line, bool discardEoln = false) throw(IOException) { require(isOpen()); return _tis.readLine(line, discardEoln); }

        /**
         *  Skips a sequence of characters from this reader.
         *
         *  The actual number of character skipped can be smaller than the
         *  requested number of characters  if an end-of-stream is reached 
         *  before a specified number of character has been skipped.
         *
         *  This method can only be called on open readers.
         *
         *  @param charsToSkip  A maximum number of characters to skip.
         *  @return             The actual number of characters skipped.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize charsToSkip) throw(IOException) { require(isOpen()); return _tis.skip(charsToSkip); }

        /**
         *  Un-reads the character to the reader.
         *
         *  Upon successful un-read, the next character read from the
         *  reader is guaranteed to be the last character un-read to
         *  that reader. Text file readers support an arbitrary
         *  un-read depth.
         *
         *  This method can only be called on open readers.
         *
         *  @param c            The character to push back into the 
         *                      reader.
         *  @return             True if the character has been successfully
         *                      pushed back into the reader; false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            unread(Char c) throw(IOException) { require(isOpen()); return _tis.unread(c); }

        //////////
        //  Implementation
    private:
        FileInputStream         _fis;
        BufferedInputStream     _bis;
        TextInputStream         _tis;
    };

    /**
     *  @class TextFileWriter CARL.hpp
     *  A helper class for buffered writing of text files.
     *
     *  A text file writer combines a file output stream, a buffering
     *  facilities and a text output stream, all working together to
     *  allow buffered writing of text files.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC TextFileWriter : public ITextOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        TextFileWriter(const TextFileWriter &);
        void                    operator = (const TextFileWriter &);

        //////////
        //  Construction/destruction
    public:
        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses default buffer size and current default character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses default buffer size and current default character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ISo-10646 string
         *                      representing a name of the file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const Char * pFileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses default buffer size and current default character set.
         *
         *  @param fileName     A string representing a name of the 
         *                      file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const String & fileName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses default buffer size and current default character set.
         *
         *  @param pathName     A path to the file to write to.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const PathName & pathName) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses default buffer size and specified character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to write to.
         *  @param pCharacterSet    A character set to use for 
         *                      character-to-byte translation.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const char * pFileName, CharacterSet * pCharacterSet) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses default buffer size and specified character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ISO-10646 string
         *                      representing a name of the file to write to.
         *  @param pCharacterSet    A character set to use for 
         *                      character-to-byte translation.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const Char * pFileName, CharacterSet * pCharacterSet) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses default buffer size and specified character set.
         *
         *  @param fileName     A string representing a name of the 
         *                      file to write to.
         *  @param pCharacterSet    A character set to use for 
         *                      character-to-byte translation.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const String & fileName, CharacterSet * pCharacterSet) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses default buffer size and specified character set.
         *
         *  @param pathName     A path to the file to write to.
         *  @param pCharacterSet    A character set to use for 
         *                      character-to-byte translation.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const PathName & pathName, CharacterSet * pCharacterSet) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses specified buffer size character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ASCII string
         *                      representing a name of the file to write to.
         *  @param pCharacterSet    A character set to use for 
         *                      character-to-byte translation.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const char * pFileName, CharacterSet * pCharacterSet, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses specified buffer size character set.
         *
         *  @param pFileName    A pointer to a nul-terminated ISo-10646 string
         *                      representing a name of the file to write to.
         *  @param pCharacterSet    A character set to use for 
         *                      character-to-byte translation.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const Char * pFileName, CharacterSet * pCharacterSet, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses specified buffer size character set.
         *
         *  @param fileName     A string representing a name of 
         *                      the file to write to.
         *  @param pCharacterSet    A character set to use for 
         *                      character-to-byte translation.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const String & fileName, CharacterSet * pCharacterSet, UIntSize bufferSize) throw(IOException);

        /**
         *  The class constructor.
         *
         *  Initializes a newly created text file writer to write
         *  to the specified text file, which is created if it does
         *  not yet exist or truncated if it does. The text file writer
         *  uses specified buffer size character set.
         *
         *  @param pathName     A path to the file to write to.
         *  @param pCharacterSet    A character set to use for 
         *                      character-to-byte translation.
         *  @param bufferSize   The buffer size.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        TextFileWriter(const PathName & pathName, CharacterSet * pCharacterSet, UIntSize bufferSize) throw(IOException);

        /**
         *  The class destructor.
         *
         *  Closes the writer if not yet closed.
         *
         *  @since 1.0
         */
        virtual ~TextFileWriter() throw(IOException) { close(); }

        //////////
        //  Operations
    public:
        /**
         *  Checks if this writer is still open.
         *
         *  @return                 True if this writer is currently open, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                    isOpen() const { return _fos.isOpen(); }

        /**
         *  Closes this writer.
         *
         *  An attempt to close a writer that is already closed has no effect.
         *
         *  @since 1.0
         */
        void                    close();

        /**
         *  Returns the character set used by this writer.
         *
         *  This method can only be called on open writers.
         *
         *  @return             The character set used by this writer.
         *  @since 1.0
         */
        CharacterSet *          getCharacterSet() const { return _tos.getCharacterSet(); }

        /**
         *  Sets the character set used by this writer.
         *
         *  This method can only be called on open writers.
         *
         *  @param pCharacterSet    The character set to use for 
         *                      character-to-byte translation.
         *  @since 1.0
         */
        void                    setCharacterSet(CharacterSet * pCharacterSet) { _tos.setCharacterSet(pCharacterSet); }

        //////////
        //  Operations
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  Causes all buffered contents to be written to an underlying
         *  file, then flushes OS file buffers for this file to disk.
         *
         *  This method can only be called on open readers.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException) { require(isOpen()); _tos.flush(); }

        /**
         *  Writes a character to this writer.
         *
         *  This method can only be called on open writers.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(Char c) throw(IOException) { require(isOpen()); _tos.write(c); }

        /**
         *  Writes a sequence of characters to this writer.
         *
         *  This method can only be called on open writers.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer) throw(IOException) { require(isOpen()); _tos.write(pBuffer); }

        /**
         *  Writes a sequence of characters to this writer.
         *
         *  This method can only be called on open writers.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the writer. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer, UIntSize charsToWrite) throw(IOException) { require(isOpen()); _tos.write(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this writer.
         *
         *  This method can only be called on open writers.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer) throw(IOException) { require(isOpen()); _tos.write(pBuffer); }

        /**
         *  Writes a sequence of characters to this writer.
         *
         *  This method can only be called on open writers.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the writer. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) { require(isOpen()); _tos.write(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this writer.
         *
         *  This method can only be called on open writers.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the writer. All characters from the
         *                      string are written, in order, to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const String & s) throw(IOException) { require(isOpen()); _tos.write(s); }

        /**
         *  Wrtes an end-of-line marker to the writer.
         *
         *  This method can only be called on open writers.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn() throw(IOException) { require(isOpen()); _tos.writeLn(); }

        /**
         *  Wrtes a character followed by end-of-line marker to the stream.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(Char c) throw(IOException) { require(isOpen()); _tos.writeLn(c); }

        /**
         *  Writes a sequence of characters to this writer, then
         *  writes an end-of-line marker.
         *
         *  This method can only be called on open writers.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer) throw(IOException) { require(isOpen()); _tos.writeLn(pBuffer); }

        /**
         *  Writes a sequence of characters to this writer, then writes
         *  an end-of-line marker.
         *
         *  This method can only be called on open writers.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the writer. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException) { require(isOpen()); _tos.writeLn(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ISO-10646 characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const Char * pBuffer) throw(IOException) { require(isOpen()); _tos.writeLn(pBuffer); }

        /**
         *  Writes a sequence of characters to this writer, then writes
         *  an end-of-line marker.
         *
         *  This method can only be called on open writers.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the writer. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) { require(isOpen()); _tos.writeLn(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this writer, then writes
         *  an end-of-line marker.
         *
         *  This method can only be called on open writers.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the writer. All characters from the
         *                      string are written, in order, to the writer.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const String & s) throw(IOException) { require(isOpen()); _tos.writeLn(s); }

        //////////
        //  Implementation
    private:
        FileOutputStream        _fos;
        BufferedOutputStream    _bos;
        TextOutputStream        _tos;
    };

    //////////
    //  Standard streams
    /**
     *  @class StandardInputStream CARL.hpp
     *  The standard input stream type.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC StandardInputStream : public ITextInput
    {
        //////////
        //  Can't assign or copy construct
    private:
        StandardInputStream(const StandardInputStream &);
        void                        operator = (const StandardInputStream &);

        //////////
        //  Construction/destruction
    public:
        StandardInputStream() {}
        virtual ~StandardInputStream() {}

        //////////
        //  ITextInput
    public:
        /**
         *  Checks if a character can be read from this text input 
         *  stream without blocking.
         *
         *  @return             True if at least one character can be read from
         *                      this stream without blocking, false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            ready() throw(IOException) { return _getSource()->ready(); }

        /**
         *  Reads a single character from this stream.
         *
         *  @param c            If a character is read successfully, it is
         *                      assigned to this parameter. Otherwise this
         *                      parameter remains unchanged.
         *  @return             True if the character has been read
         *                      successfully; false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            read(Char & c) throw(IOException) { return _getSource()->read(c); }

        /**
         *  Reads a sequence of characters from this stream.
         *
         *  The actual number of character read can be smaller than the
         *  requested number of characters  if an end-of-stream is reached 
         *  before a specified number of character has been read.
         *
         *  @param pBuffer      A pointer to a buffer where characters read
         *                      from the stream are placed. The buffer is
         *                      assumed to be at least \a charsToRead
         *                      characters long.
         *  @param charsToRead  A maximum number of characters to read.
         *  @return             The actual number of characters read.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        read(Char * pBuffer, UIntSize charsToRead) throw(IOException) { return _getSource()->read(pBuffer, charsToRead); }

        /**
         *  Reads a line of characters from this stream.
         *
         *  Characters are read until either end-of-line or end-of-stream
         *  is detected, whichever happens first.
         *
         *  @param line         Upon successful read, this parameter is
         *                      assigned the characters of the line just 
         *                      read.
         *  @param discardEoln  If the line was terminated by an end-of-line
         *                      marker and this parameter is false, the
         *                      end-of-line marker is included into the
         *                      \a line. If the line was terminated by an
         *                      end-of-line marker and this parameter is true,
         *                      the end-of-line marker is not included into
         *                      the \a line. If the line was terminated by
         *                      an end-of-file marker, the end-of-line marker
         *                      is not included into the \a line regardless
         *                      of whether \a discardEoln is true or false.
         *  @return             True if the line has been read successfully; 
         *                      false otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            readLine(String & line, bool discardEoln = false) throw(IOException) { return _getSource()->readLine(line, discardEoln); }

        /**
         *  Skips a sequence of characters from this stream.
         *
         *  The actual number of character skipped can be smaller than the
         *  requested number of characters  if an end-of-stream is reached 
         *  before a specified number of character has been skipped.
         *
         *  @param charsToSkip  A maximum number of characters to skip.
         *  @return             The actual number of characters skipped.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual UIntSize        skip(UIntSize charsToSkip) throw(IOException) { return _getSource()->skip(charsToSkip); }

        /**
         *  Un-reads the character to the stream.
         *
         *  Upon successful un-read, the next character read from the
         *  stream is guaranteed to be the last character un-read to
         *  that stream. Text streams always guarantee that at least one
         *  character can be un-read, although some concrete text input
         *  streams may allow un-reading mul;tiple characters.
         *
         *  @param c            The character to push back into the 
         *                      input stream.
         *  @return             True if the character has been successfully
         *                      pushed back into the input stream; false
         *                      otherwise.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual bool            unread(Char c) throw(IOException) { return _getSource()->unread(c); }

        //////////
        //  Operations
    public:
        /**
         *  Returns the character set used by this input stream.
         *
         *  @return             The character set used by this input stream.
         *  @since 1.0
         */
        CharacterSet *          getCharacterSet() const;

        /**
         *  Sets the character set used by this input stream.
         *
         *  @param pCharacterSet    The character set to use for 
         *                      byte-to-character translation.
         *  @since 1.0
         */
        void                    setCharacterSet(CharacterSet * pCharacterSet);

        //////////
        //  Implementation
    private:
        static ITextInput *     _getSource();
    };

    /**
     *  @class StandardOutputStream CARL.hpp
     *  The standard out stream type.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC StandardOutputStream : public ITextOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        StandardOutputStream(const StandardOutputStream &);
        void                        operator = (const StandardOutputStream &);

        //////////
        //  Construction/destruction
    public:
        StandardOutputStream() {}
        virtual ~StandardOutputStream() {}

        //////////
        //  ITextOutput
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  If there is any data that has been written to this output stream
         *  and not yet forwarded to the stream's destination, this data is 
         *  written to the stream's destination immediately.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException) { _getDestination()->flush(); }

        /**
         *  Writes a character to this output stream.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(Char c) throw(IOException) { _getDestination()->write(c); }

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer) throw(IOException) { _getDestination()->write(pBuffer); }

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer, UIntSize charsToWrite) throw(IOException) { _getDestination()->write(pBuffer, charsToWrite);write(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ISO-10646 characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer) throw(IOException) { _getDestination()->write(pBuffer); }

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the sequence of ISo-10646 characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) { _getDestination()->write(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the stream. All characters from the
         *                      string are written, in order, to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const String & s) throw(IOException) { _getDestination()->write(s); }

        /**
         *  Wrtes an end-of-line marker to the stream.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn() throw(IOException) { _getDestination()->writeLn(); }

        /**
         *  Wrtes a character followed by end-of-line marker to the stream.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(Char c) throw(IOException) { _getDestination()->writeLn(c); }

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer) throw(IOException) { _getDestination()->writeLn(pBuffer); }

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException) { _getDestination()->writeLn(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ISO-10646 characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const Char * pBuffer) throw(IOException) { _getDestination()->writeLn(pBuffer); }

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the sequence of ISo-10646 characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
         virtual void            writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) { _getDestination()->writeLn(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the stream. All characters from the
         *                      string are written, in order, to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const String & s) throw(IOException) { _getDestination()->writeLn(s); }

        //////////
        //  Operations
    public:
        /**
         *  Returns the character set used by this output stream.
         *
         *  @return             The character set used by this output stream.
         *  @since 1.0
         */
        CharacterSet *          getCharacterSet() const;

        /**
         *  Sets the character set used by this output stream.
         *
         *  @param pCharacterSet    The character set to use for 
         *                      byte-to-character translation.
         *  @since 1.0
         */
        void                    setCharacterSet(CharacterSet * pCharacterSet);

        //////////
        //  Implementation
    private:
        static ITextOutput *    _getDestination();
    };

    /**
     *  @class StandardErrorStream CARL.hpp
     *  The standard error stream type.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC StandardErrorStream : public ITextOutput
    {
        //////////
        //  Can't assign or copy construct
    private:
        StandardErrorStream(const StandardErrorStream &);
        void                        operator = (const StandardErrorStream &);

        //////////
        //  Construction/destruction
    public:
        StandardErrorStream() {}
        virtual ~StandardErrorStream() {}

        //////////
        //  ITextOutput
    public:
        /**
         *  Flushes any unwritten data.
         *
         *  If there is any data that has been written to this output stream
         *  and not yet forwarded to the stream's destination, this data is 
         *  written to the stream's destination immediately.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            flush() throw(IOException) { _getDestination()->flush(); }

        /**
         *  Writes a character to this output stream.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(Char c) throw(IOException) { _getDestination()->write(c); }

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer) throw(IOException) { _getDestination()->write(pBuffer); }

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const char * pBuffer, UIntSize charsToWrite) throw(IOException) { _getDestination()->write(pBuffer, charsToWrite);write(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ISO-10646 characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer) throw(IOException) { _getDestination()->write(pBuffer); }

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param pBuffer      A pointer to the sequence of ISo-10646 characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) { _getDestination()->write(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this output stream.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the stream. All characters from the
         *                      string are written, in order, to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            write(const String & s) throw(IOException) { _getDestination()->write(s); }

        /**
         *  Wrtes an end-of-line marker to the stream.
         *
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn() throw(IOException) { _getDestination()->writeLn(); }

        /**
         *  Wrtes a character followed by end-of-line marker to the stream.
         *
         *  @param c            The character to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(Char c) throw(IOException) { _getDestination()->writeLn(c); }

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ASCII characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer) throw(IOException) { _getDestination()->writeLn(pBuffer); }

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the sequence of ASCII characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const char * pBuffer, UIntSize charsToWrite) throw(IOException) { _getDestination()->writeLn(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this output stream, then
         *  writes an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the nul-terminated sequence
         *                      of ISO-10646 characters to write to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const Char * pBuffer) throw(IOException) { _getDestination()->writeLn(pBuffer); }

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param pBuffer      A pointer to the sequence of ISo-10646 characters 
         *                      to write to the stream. The sequence must be
         *                      at least \a charsToWrite characters long; if
         *                      it is longer than only the first \a charsToWrite
         *                      characters are written.
         *  @param charsToWrite The number of characters to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
         virtual void            writeLn(const Char * pBuffer, UIntSize charsToWrite) throw(IOException) { _getDestination()->writeLn(pBuffer, charsToWrite); }

        /**
         *  Writes a sequence of characters to this output stream, then writes
         *  an end-of-line marker.
         *
         *  @param s            A string specifying the sequence of characters
         *                      to write to the stream. All characters from the
         *                      string are written, in order, to the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        virtual void            writeLn(const String & s) throw(IOException) { _getDestination()->writeLn(s); }

        //////////
        //  Operations
    public:
        /**
         *  Returns the character set used by this output stream.
         *
         *  @return             The character set used by this output stream.
         *  @since 1.0
         */
        CharacterSet *          getCharacterSet() const;

        /**
         *  Sets the character set used by this output stream.
         *
         *  @param pCharacterSet    The character set to use for 
         *                      byte-to-character translation.
         *  @since 1.0
         */
        void                    setCharacterSet(CharacterSet * pCharacterSet);

        //////////
        //  Implementation
    private:
        static ITextOutput *    _getDestination();
    };

    /**
     *  @class StandardStreams CARL.hpp
     *  Encapsulates standard streams.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC StandardStreams
    {
        //////////
        //  This is a utility class that cannot be instantiated
    private:
        StandardStreams();
        StandardStreams (const StandardStreams &);
        ~StandardStreams();
        void                    operator = (const StandardStreams & src);

        //////////
        //  Console streams
    public:
        /**
         *  The standard input stream.
         *
         *  @since 1.0
         */
        static StandardInputStream  in;

        /**
         *  The standard output stream.
         *
         *  @since 1.0
         */
        static StandardOutputStream  out;

        /**
         *  The standard error stream.
         *
         *  @since 1.0
         */
        static StandardErrorStream  err;
    };

    //////////
    //  Binary I/O operators for CARL types
    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, bool & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, UInt8 & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, UInt16 & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, UInt32 & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, UInt64 & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, Int8 & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, Int16 & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, Int32 & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, Int64 & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, Real32 & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, Real64 & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, Date & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, DateTime & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, Locale & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, PathName & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, String & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, Time & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, TimeSpan & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, TimeStamp & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, BinaryInteger & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, BinaryFixed & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, BinaryFloat & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, DecimalInteger & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, DecimalFixed & value) throw(IOException);

    /**
     *  Reads a value from the specified data input stream.
     *
     *  @param dis              The data input stream, to read a value from.
     *  @param value            Upon successful read, the value read from
     *                          the stream is assigned to this parameter.
     *  @return                 The data input stream that has been read from.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataInput & operator >> (IDataInput & dis, DecimalFloat & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, bool value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, UInt8 value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, UInt16 value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, UInt32 value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, UInt64 value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, Int8 value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, Int16 value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, Int32 value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, Int64 value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, Real32 value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, Real64 value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const Date & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const DateTime & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const Locale & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const PathName & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const String & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const Time & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const TimeSpan & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const TimeStamp & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const BinaryInteger & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const BinaryFixed & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const BinaryFloat & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const DecimalInteger & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const DecimalFixed & value) throw(IOException);

    /**
     *  Writes a value to the specified data output stream.
     *
     *  @param dos              The data output stream, to write a value to.
     *  @param value            The value to write.
     *  @return                 The data output stream that has been written to.
     *  @exception IOException  If an I/O error occurs.
     *  @since 1.0
     */
    CARL_PUBLIC IDataOutput & operator << (IDataOutput & dos, const DecimalFloat & value) throw(IOException);

    //////////
    //  LEB128 I/O
    /**
     *  @class Leb128 CARL.hpp
     *  Provides LEB128 encoding and decoding for I/O of integer values.
     *
     *  LEB128 is an encoding scheme that allows representing signed integer
     *  values as sequences of bytes that generally take less space than the
     *  values themselves.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Leb128
    {
        //////////
        //  This is an utility class
    private:
        Leb128();
        Leb128(const Leb128 &);
        ~Leb128();
        void                    operator = (const Leb128 &);

        //////////
        //  Operations
    public:
        /**
         *  Writes a LEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      LEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeInt8(IByteOutput & bos, Int8 value) throw(IOException) { return writeInt64(bos, value); }

        /**
         *  Writes a LEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      LEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeInt16(IByteOutput & bos, Int16 value) throw(IOException) { return writeInt64(bos, value); }

        /**
         *  Writes a LEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      LEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeInt32(IByteOutput & bos, Int32 value) throw(IOException) { return writeInt64(bos, value); }

        /**
         *  Writes a LEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      LEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeInt64(IByteOutput & bos, Int64 value) throw(IOException);

        /**
         *  Writes a LEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      LEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeBinaryInteger(IByteOutput & bos, const BinaryInteger & value) throw(IOException);

        /**
         *  Writes a LEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      LEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeIntPtrDiff(IByteOutput & bos, IntPtrDiff value) throw(IOException) { return writeInt64(bos, value); }

        /**
         *  Reads a LEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      LEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static Int8             readInt8(IByteInput & bis) throw(IOException);

        /**
         *  Reads a LEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      LEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static Int16            readInt16(IByteInput & bis) throw(IOException);

        /**
         *  Reads a LEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      LEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static Int32            readInt32(IByteInput & bis) throw(IOException);

        /**
         *  Reads a LEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      LEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static Int64            readInt64(IByteInput & bis) throw(IOException);

        /**
         *  Reads a LEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      LEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static BinaryInteger    readBinaryInteger(IByteInput & bis) throw(IOException);

        /**
         *  Reads a LEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      LEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static IntPtrDiff       readIntPtrDiff(IByteInput & bis) throw(IOException);
    };

    //////////
    //  ULEB128 I/O
    /**
     *  @class ULeb128 CARL.hpp
     *  Provides ULEB128 encoding and decoding for I/O of integer values.
     *
     *  ULEB128 is an encoding scheme that allows representing unsigned integer
     *  values as sequences of bytes that generally take less space than the
     *  values themselves.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC ULeb128
    {
        //////////
        //  This is an utility class
    private:
        ULeb128();
        ULeb128(const ULeb128 &);
        ~ULeb128();
        void                    operator = (const ULeb128 &);

        //////////
        //  Operations
    public:
        /**
         *  Writes an ULEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      ULEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeUInt8(IByteOutput & bos, UInt8 value) throw(IOException) {     return writeUInt64(bos, value); }

        /**
         *  Writes an ULEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      ULEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeUInt16(IByteOutput & bos, UInt16 value) throw(IOException) {     return writeUInt64(bos, value); }

        /**
         *  Writes an ULEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      ULEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeUInt32(IByteOutput & bos, UInt32 value) throw(IOException) {     return writeUInt64(bos, value); }

        /**
         *  Writes an ULEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      ULEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeUInt64(IByteOutput & bos, UInt64 value) throw(IOException);

        /**
         *  Writes an ULEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      ULEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeBinaryInteger(IByteOutput & bos, const BinaryInteger & value) throw(IOException);

        /**
         *  Writes an ULEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      ULEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeUIntSize(IByteOutput & bos, UIntSize value) throw(IOException) {     return writeUInt64(bos, value);}

        /**
         *  Writes an ULEB128 - encoded value to the specified byte
         *  output stream.
         *
         *  @param bos          The byte output stream to write the
         *                      ULEB128 - encoded value to.
         *  @param value        The value to write.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static void             writeUIntPtr(IByteOutput & bos, UIntPtr value) throw(IOException) {     return writeUInt64(bos, value); }

        /**
         *  Reads an ULEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      ULEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static UInt8            readUInt8(IByteInput & bis) throw(IOException);

        /**
         *  Reads an ULEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      ULEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static UInt16           readUInt16(IByteInput & bis) throw(IOException);

        /**
         *  Reads an ULEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      ULEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static UInt32           readUInt32(IByteInput & bis) throw(IOException);

        /**
         *  Reads an ULEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      ULEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static UInt64           readUInt64(IByteInput & bis) throw(IOException);

        /**
         *  Reads an ULEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      ULEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static BinaryInteger    readBinaryInteger(IByteInput & bis) throw(IOException);

        /**
         *  Reads an ULEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      ULEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static UIntSize         readUIntSize(IByteInput & bis) throw(IOException);

        /**
         *  Reads an ULEB128 - encoded value from the specified byte
         *  output stream.
         *
         *  @param bis          The byte input stream to read the
         *                      ULEB128 - encoded value from.
         *  @return             The value reads from the stream.
         *  @exception IOException  If an I/O error occurs.
         *  @since 1.0
         */
        static UIntPtr          readUIntPtr(IByteInput & bis) throw(IOException);
    };

    //////////
    //  Hashing path names
    namespace StandardHashFunctions
    {
        /**
         *  Calculates a hash code for a PathName instance.
         *
         *  @param key              The PathName instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashPathName(const PathName & key);
    }
}

#endif  //  __CARL_IO_hpp_included__
//  End of CARL_IO.hpp
