//
//  CARL_Math.hpp - Math routines.
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/08/08 14:42:28 $
//  $Revision: 1.34 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL math services.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Math.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Math_hpp_included__
#define __CARL_Math_hpp_included__

namespace carl
{
    class Random;

    /**
     *  @class FixedPointDivisionByZeroError CARL.hpp
     *  An exception thrown when a fixed-point division by zero occurs.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC FixedPointDivisionByZeroError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        FixedPointDivisionByZeroError();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        FixedPointDivisionByZeroError(const FixedPointDivisionByZeroError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~FixedPointDivisionByZeroError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        FixedPointDivisionByZeroError & operator = (const FixedPointDivisionByZeroError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class SignalingNanOperandError CARL.hpp
     *  An exception thrown when a SNaN is used as an operand in an arithmetic expression.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC SignalingNanOperandError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        SignalingNanOperandError();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        SignalingNanOperandError(const SignalingNanOperandError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~SignalingNanOperandError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        SignalingNanOperandError &  operator = (const SignalingNanOperandError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  @class ValueNotRepresentableAsFixedPointError CARL.hpp
     *  An exception thrown when an attempt is made to translate a floating-point
     *  value to a fixed-point value and the floating-point value is an
     *  infinity or NaN.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC ValueNotRepresentableAsFixedPointError : public ProgramError
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  @since 1.0
         */
        ValueNotRepresentableAsFixedPointError();

        /**
         *  The copy constructor.
         *
         *  @param src          The exception to copy properties from.
         *  @since 1.0
         */
        ValueNotRepresentableAsFixedPointError(const ValueNotRepresentableAsFixedPointError & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        virtual ~ValueNotRepresentableAsFixedPointError();

        /**
         *  The assignment operator.
         *
         *  @param src          The exception to copy properties from.
         *  @return             A reference to the assigned-to object.
         *  @since 1.0
         */
        ValueNotRepresentableAsFixedPointError &    operator = (const ValueNotRepresentableAsFixedPointError & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns a user-readable error message describing this exception
         *  for the current default locale.
         *
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString() const;

        /**
         *  Returns a user-readable error message describing this exception
         *  for the specified locale.
         *
         *  @param forLocale    The locale for which an error message shall
         *                      be tailored.
         *  @return             The user-readable error message describing
         *                      this exception.
         *  @since 1.0
         */
        virtual String          toString(const Locale & forLocale) const;
    };

    /**
     *  Supported rounding modes for real arithmetics.
     *
     *  @since 1.0
     */
    enum RoundingMode
    {
        /**
         *  Round towards nearest representable value.
         *
         *  If a value being rounded is exactly halfway between two
         *  representable values, the rounding is performed towards
         *  the even nearest value.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        RoundingMode_TowardsNearest             = 1,

        /**
         *  Round towards zero.
         *
         *  Positive values are rounded down; negative values are rounded up.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        RoundingMode_TowardsZero                = 2,

        /**
         *  Round towards infinity.
         *
         *  Positive values are rounded up; negative values are rounded down.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        RoundingMode_TowardsInfinity            = 3,

        /**
         *  Round towards positive infinity.
         *
         *  All values are rounded up.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        RoundingMode_TowardsPositiveInfinity    = 4,

        /**
         *  Round towards negative infinity.
         *
         *  All values are rounded down.
         *
         *  @hideinitializer
         *  @since 1.0
         */
        RoundingMode_TowardsNegativeInfinity    = 5
    };

    /**
     *  @class BinaryInteger CARL.hpp
     *  An arbitrary-precision binary integer number.
     *
     *  Conceptually, a BinaryInteger instance represents a signed binary
     *  integer value of an infinite precision stored in 2's complement
     *  format. Naturally, the binary integer values manipulated by the
     *  program will all be finite, so starting at some bit position all
     *  higher bits will all be 0s (for positive values) or 1s (for negative
     *  values); these bits are implied instead of being actually represented
     *  in memory.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC BinaryInteger
    {
        //////////
        //  Construction/destruction/assignment
    private:
        BinaryInteger(UInt32 * pBits, UIntSize wordsAllocated, UIntSize wordsUsed, bool takeOwnershipOfBits);
    public:
        /**
         *  The default constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to 0.
         *
         *  @since 1.0
         */
        BinaryInteger();

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryInteger(Int8 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryInteger(Int16 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryInteger(Int32 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryInteger(Int64 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryInteger(UInt8 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryInteger(UInt16 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryInteger(UInt32 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryInteger(UInt64 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified sequence of bits. If the sequence of bits has 0
         *  as its highest bit, the resulting BinaryInteger instance is
         *  positive; otherwise it is negative.
         *
         *  @param value            The bits to initialize with.
         *  @param valueByteOrder   Specifies whether bytes in the \a value
         *                          array are stored in big- or little-endian order.
         *  @since 1.0
         */
        BinaryInteger(const Array<UInt8> & value, ByteOrder valueByteOrder);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified sequence of bits. If the sequence of bits has 0
         *  as its highest bit, the resulting BinaryInteger instance is
         *  positive; otherwise it is negative.
         *
         *  @param pValue           The pointer to an array of bits to initialize
         *                          with. The array must be at least \a valueSize
         *                          bytes long; if it is longer than only the
         *                          first \a valueSize bytes are used.
         *  @param valueSize        The number of bytes in the initializer.
         *  @param valueByteOrder   Specifies whether bytes in the \a value
         *                          array are stored in big- or little-endian order.
         *  @since 1.0
         */
        BinaryInteger(const UInt8 * pValue, UIntSize valueSize, ByteOrder valueByteOrder);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified sequence of bits, which is always assumed to
         *  be unsigned. The \a sign parameter is then examined and the
         *  constructed BinaryInteger instance negated if sign < 0.
         *
         *  @param sign             The sign of this value determines the
         *                          sign of the resulting BinaryInteger.
         *  @param magnitude        The magnitude bits to initialize with.
         *  @param magnitudeByteOrder   Specifies whether bytes in the \a value
         *                          array are stored in big- or little-endian order.
         *  @since 1.0
         */
        BinaryInteger(int sign, const Array<UInt8> & magnitude, ByteOrder magnitudeByteOrder);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryInteger instance is initialized to
         *  the specified sequence of bits, which is always assumed to
         *  be unsigned. The \a sign parameter is then examined and the
         *  constructed BinaryInteger instance negated if sign < 0.
         *
         *  @param sign             The sign of this value determines the
         *                          sign of the resulting BinaryInteger.
         *  @param pMagnitude       The pointer to an array of bits to initialize
         *                          with. The array must be at least \a valueSize
         *                          bytes long; if it is longer than only the
         *                          first \a valueSize bytes are used.
         *  @param magnitudeSize    The number of bytes in the initializer.
         *  @param magnitudeByteOrder   Specifies whether bytes in the \a value
         *                          array are stored in big- or little-endian order.
         *  @since 1.0
         */
        BinaryInteger(int sign, const UInt8 * pMagnitude, UIntSize magnitudeSize, ByteOrder magnitudeByteOrder);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryInteger instance is initialized by parsing
         *  a string representation of an integer value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryInteger()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a radix 10 string representation
         *                          of an integer value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          an integer value.
         *  @since 1.0
         */
        explicit BinaryInteger(const char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryInteger instance is initialized by parsing
         *  a string representation of an integer value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryInteger()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISO-10646 string
         *                          containing a radix 10 string representation
         *                          of an integer value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          an integer value.
         *  @since 1.0
         */
        explicit BinaryInteger(const Char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryInteger instance is initialized by parsing
         *  a string representation of an integer value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryInteger()
         *  for the invariant locale.
         *
         *  @param valueString      A string containing a radix 10 string representation
         *                          of an integer value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          an integer value.
         *  @since 1.0
         */
        explicit BinaryInteger(const String & valueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryInteger instance is initialized by parsing
         *  a string representation of an integer value using specified radix. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryInteger()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a string representation of an 
         *                          integer value.
         *  @param radix            The radix used by the string representation
         *                          of an integer value (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of an integer 
         *                          value.
         *  @since 1.0
         */
        BinaryInteger(const char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryInteger instance is initialized by parsing
         *  a string representation of an integer value using specified radix. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryInteger()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISo-10646 string
         *                          containing a string representation of an 
         *                          integer value.
         *  @param radix            The radix used by the string representation
         *                          of an integer value (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of an integer 
         *                          value.
         *  @since 1.0
         */
        BinaryInteger(const Char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryInteger instance is initialized by parsing
         *  a string representation of an integer value using specified radix. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryInteger()
         *  for the invariant locale.
         *
         *  @param valueString      A string representation of an  integer value.
         *  @param radix            The radix used by the string representation
         *                          of an integer value (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of an integer 
         *                          value.
         *  @since 1.0
         */
        BinaryInteger(const String & valueString, UIntSize radix) throw (ParseException);

        /**
         *  The copy constructor.
         *
         *  @param src              The binary integer to copy value from.
         *  @since 1.0
         */
        BinaryInteger(const BinaryInteger & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~BinaryInteger();

        /**
         *  The assignment operator.
         *
         *  @param src              The binary integer to copy value from.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        BinaryInteger &             operator = (const BinaryInteger & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the value unchanged.
         *
         *  @return                 A value equal to the operand.
         *  @since 1.0
         */
        BinaryInteger               operator + () const;

        /**
         *  Returns the negated operand value.
         *
         *  @return                 The negated operand value.
         *  @since 1.0
         */
        BinaryInteger               operator - () const;

        /**
         *  Returns the 1's complement of an operand value.
         *
         *  @return                 The 1's complement of an operand value.
         *  @since 1.0
         */
        BinaryInteger               operator ~ () const;

        /**
         *  Returns the sum of two values.
         *
         *  @param op2              The 2nd operand of an addition operation.
         *  @return                 The sum of two values.
         *  @since 1.0
         */
        BinaryInteger               operator + (const BinaryInteger & op2) const;

        /**
         *  Returns the difference of two values.
         *
         *  @param op2              The 2nd operand of a subtraction operation.
         *  @return                 The difference of two values.
         *  @since 1.0
         */
        BinaryInteger               operator - (const BinaryInteger & op2) const;

        /**
         *  Returns the product of two values.
         *
         *  @param op2              The 2nd operand of a multiplication operation.
         *  @return                 The product of two values.
         *  @since 1.0
         */
        BinaryInteger               operator * (const BinaryInteger & op2) const;

        /**
         *  Returns the quotient of two values.
         *
         *  @param op2              The 2nd operand of a division operation.
         *  @return                 The quotient of two values.
         *  @since 1.0
         */
        BinaryInteger               operator / (const BinaryInteger & op2) const;

        /**
         *  Returns the remainder from division of two values.
         *
         *  @param op2              The 2nd operand of a remainder operation.
         *  @return                 The remainder from division of two values.
         *  @since 1.0
         */
        BinaryInteger               operator % (const BinaryInteger & op2) const;

        /**
         *  Returns the bitwise logical conjunction of two values.
         *
         *  @param op2              The 2nd operand of a conjunction operation.
         *  @return                 The bitwise logical conjunction of two values.
         *  @since 1.0
         */
        BinaryInteger               operator & (const BinaryInteger & op2) const;

        /**
         *  Returns the bitwise logical disjunction of two values.
         *
         *  @param op2              The 2nd operand of a disjunction operation.
         *  @return                 The bitwise logical disjunction of two values.
         *  @since 1.0
         */
        BinaryInteger               operator | (const BinaryInteger & op2) const;

        /**
         *  Returns the bitwise logical exclusive disjunction of two values.
         *
         *  @param op2              The 2nd operand of a exclusive disjunction operation.
         *  @return                 The bitwise logical exclusive disjunction of two values.
         *  @since 1.0
         */
        BinaryInteger               operator ^ (const BinaryInteger & op2) const;

        /**
         *  Returns an operand shifted left by the specified number of bits.
         *
         *  All bits in the shift counter are significant.
         *
         *  @param op2              The number of bits to shift the 1st operand by.
         *  @return                 An operand shifted left by the specified number of bits.
         *  @since 1.0
         */
        BinaryInteger               operator << (UIntSize op2) const;

        /**
         *  Returns an operand shifted right by the specified number of bits.
         *
         *  All bits in the shift counter are significant.
         *
         *  @param op2              The number of bits to shift the 1st operand by.
         *  @return                 An operand shifted right by the specified number of bits.
         *  @since 1.0
         */
        BinaryInteger               operator >> (UIntSize op2) const;

        /**
         *  Calculates the sum of two values and assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of an addition operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        BinaryInteger &             operator += (const BinaryInteger & op2);

        /**
         *  Calculates the difference of two values and assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a subtraction operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        BinaryInteger &             operator -= (const BinaryInteger & op2);

        /**
         *  Calculates the product of two values and assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a multiplication operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        BinaryInteger &             operator *= (const BinaryInteger & op2);

        /**
         *  Calculates the quotient of two values and assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a division operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        BinaryInteger &             operator /= (const BinaryInteger & op2);

        /**
         *  Calculates the remainder from division of two values and 
         *  assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a remainder operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        BinaryInteger &             operator %= (const BinaryInteger & op2);

        /**
         *  Calculates the bitwise logical conjunction of two values and 
         *  assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a conjunction operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        BinaryInteger &             operator &= (const BinaryInteger & op2);

        /**
         *  Calculates the bitwise logical disjunction of two values and 
         *  assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a disjunction operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        BinaryInteger &             operator |= (const BinaryInteger & op2);

        /**
         *  Calculates the bitwise logical exclusive disjunction of two 
         *  values and assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a disjunction operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        BinaryInteger &             operator ^= (const BinaryInteger & op2);

        /**
         *  Shifts the 1st operand left by the specified number of bits.
         *
         *  All bits in the shift counter are significant.
         *
         *  @param op2              The number of bits to shift the 1st operand by.
         *  @return                 A reference to the shifted object.
         *  @since 1.0
         */
        BinaryInteger &             operator <<= (UIntSize op2);

        /**
         *  Shifts the 1st operand right by the specified number of bits.
         *
         *  All bits in the shift counter are significant.
         *
         *  @param op2              The number of bits to shift the 1st operand by.
         *  @return                 A reference to the shifted object.
         *  @since 1.0
         */
        BinaryInteger &             operator >>= (UIntSize op2);

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value after increment.
         *  @since 1.0
         */
        BinaryInteger               operator ++();

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value after decrement.
         *  @since 1.0
         */
        BinaryInteger               operator --();

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value before increment.
         *  @since 1.0
         */
        BinaryInteger               operator ++(int);

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value before decrement.
         *  @since 1.0
         */
        BinaryInteger               operator --(int);

        /**
         *  Calculates ((*this) ** \a exponent) % \a modulus.
         *
         *  @param exponent         The exponent (must be >= 0).
         *  @param modulus          The modulus (must be > 0).
         *  @return                 The value of ((*this) ** exponent) % modulus.
         *  @since 1.0
         */
        BinaryInteger               modPow(const BinaryInteger & exponent, const BinaryInteger & modulus) const;

        /**
         *  Calculates a modular inverse of this value.
         *
         *  A modular inverse of the value X modulo M is a value Y
         *  such that X*Y == 1 (mod M).
         *
         *  @param modulus          The modulus (must be > 0).
         *  @return                 The modular inverse of this value.
         *  @since 1.0
         */
        BinaryInteger               modInverse(const BinaryInteger & modulus) const;

        /**
         *  Calculates the greatest common denominator of two values.
         *
         *  @param op2              The 2nd operand.
         *  @return                 the greatest common denominator of (*\a this) and \a op2.
         *  @since 1.0
         */
        BinaryInteger               gcd(const BinaryInteger & op2) const;

        /**
         *  Compares two values.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return             
         *                          -   0 if the two values are equal.
         *                          -   <0 if this value is smaller than \a op2.
         *                          -   >0 if this value is greater than \a op2.
         *  @since 1.0
         */
        int                         compare(const BinaryInteger & op2) const;

        /**
         *  Compares two values for equality.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator == (const BinaryInteger & op2) const;

        /**
         *  Compares two values for inequality.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is not equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator != (const BinaryInteger & op2) const;

        /**
         *  Checks if one value is smaller than another value.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is smaller than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator < (const BinaryInteger & op2) const;

        /**
         *  Checks if one value is smaller than or equal to another value.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is smaller than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator <= (const BinaryInteger & op2) const;

        /**
         *  Checks if one value is greater than another value.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is greater than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator > (const BinaryInteger & op2) const;

        /**
         *  Checks if one value is greater than or equal to another value.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is greater than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator >= (const BinaryInteger & op2) const;

        /**
         *  Returns the sign of this value.
         *
         *  @return
         *                          -   0 of \a (*this) = 0.
         *                          -   1 of \a (*this) > 0.
         *                          -   -1 of \a (*this) < 0.
         *  @since 1.0
         */
        int                         getSign() const;

        /**
         *  Checks if a value is odd.
         *
         *  @return                 True if this value is odd; false otherwise.
         *  @since 1.0
         */
        bool                        isOdd() const;

        /**
         *  Checks if a value is even.
         *
         *  @return                 True if this value is even; false otherwise.
         *  @since 1.0
         */
        bool                        isEven() const;

        /**
         *  Returns the number of significant bits in the binary representation
         *  of this value.
         *
         *  The "required number of bits" is defined as a smallest number of bits
         *  that can be used to represent this value without loss of significance.
         *  This specifically includes the sign bit.
         *
         *  @return                 the number of significant bits in the binary 
         *                          representation of this value.
         *  @since 1.0
         */
        UIntSize                    getRequiredBits() const;

        /**
         *  Returns the bit at the specified position.
         *
         *  Position 0 denotes the lowest bit of the value; position 1
         *  denotes the next bit, and so on.
         *
         *  @param position         The bit position.
         *  @return                 Either 0 or 1, depending on whether the
         *                          bit at the specified position is 0 or 1.
         *  @since 1.0
         */
        int                         operator [] (UIntSize position) const;

        /**
         *  Returns the bit at the specified position.
         *
         *  Position 0 denotes the lowest bit of the value; position 1
         *  denotes the next bit, and so on.
         *
         *  @param position         The bit position.
         *  @return                 Either 0 or 1, depending on whether the
         *                          bit at the specified position is 0 or 1.
         *  @since 1.0
         */
        int                         getBit(UIntSize position) const;

        /**
         *  Sets the bit at the specified position to 1.
         *
         *  Position 0 denotes the lowest bit of the value; position 1
         *  denotes the next bit, and so on.
         *
         *  @param position         The bit position.
         *  @since 1.0
         */
        void                        setBit(UIntSize position);

        /**
         *  Sets the bit at the specified position to 0.
         *
         *  Position 0 denotes the lowest bit of the value; position 1
         *  denotes the next bit, and so on.
         *
         *  @param position         The bit position.
         *  @since 1.0
         */
        void                        resetBit(UIntSize position);

        /**
         *  Flips the bit at the specified position.
         *
         *  Position 0 denotes the lowest bit of the value; position 1
         *  denotes the next bit, and so on.
         *
         *  Flipping a bit causes it to take an opposite value (i.e. 0
         *  becomes 1 and 1 becomes 0).
         *
         *  @param position         The bit position.
         *  @since 1.0
         */
        void                        flipBit(UIntSize position);

        /**
         *  Converts this value to Int8 and returns the result.
         *
         *  If the value is too large to fit into the Int8 type,
         *  only the lowest 8 bits of the value are used.
         *
         *  @return                 The result of converting this value to Int8.
         *  @since 1.0
         */
        Int8                        toInt8() const;

        /**
         *  Converts this value to Int16 and returns the result.
         *
         *  If the value is too large to fit into the Int16 type,
         *  only the lowest 16 bits of the value are used.
         *
         *  @return                 The result of converting this value to Int16.
         *  @since 1.0
         */
        Int16                       toInt16() const;

        /**
         *  Converts this value to Int32 and returns the result.
         *
         *  If the value is too large to fit into the Int32 type,
         *  only the lowest 32 bits of the value are used.
         *
         *  @return                 The result of converting this value to Int32.
         *  @since 1.0
         */
        Int32                       toInt32() const;

        /**
         *  Converts this value to Int64 and returns the result.
         *
         *  If the value is too large to fit into the Int64 type,
         *  only the lowest 64 bits of the value are used.
         *
         *  @return                 The result of converting this value to Int64.
         *  @since 1.0
         */
        Int64                       toInt64() const;

        /**
         *  Converts this value to UInt8 and returns the result.
         *
         *  If the value is too large to fit into the UInt8 type,
         *  only the lowest 8 bits of the value are used.
         *
         *  @return                 The result of converting this value to UInt8.
         *  @since 1.0
         */
        UInt8                       toUInt8() const;

        /**
         *  Converts this value to UInt16 and returns the result.
         *
         *  If the value is too large to fit into the UInt16 type,
         *  only the lowest 16 bits of the value are used.
         *
         *  @return                 The result of converting this value to UInt16.
         *  @since 1.0
         */
        UInt16                      toUInt16() const;

        /**
         *  Converts this value to UInt32 and returns the result.
         *
         *  If the value is too large to fit into the UInt32 type,
         *  only the lowest 32 bits of the value are used.
         *
         *  @return                 The result of converting this value to UInt32.
         *  @since 1.0
         */
        UInt32                      toUInt32() const;

        /**
         *  Converts this value to UInt64 and returns the result.
         *
         *  If the value is too large to fit into the UInt64 type,
         *  only the lowest 64 bits of the value are used.
         *
         *  @return                 The result of converting this value to UInt64.
         *  @since 1.0
         */
        UInt64                      toUInt64() const;

        /**
         *  Converts this value to UIntSize and returns the result.
         *
         *  If the value is too large to fit into the UIntSize type,
         *  only the lowest sizeof(UIntSize)*System::BitsPerByte bits 
         *  of the value are used.
         *
         *  @return                 The result of converting this value to UIntSize.
         *  @since 1.0
         */
        UIntSize                    toUIntSize() const;

        /**
         *  Converts this value to UIntPtr and returns the result.
         *
         *  If the value is too large to fit into the UIntPtr type,
         *  only the lowest sizeof(UIntPtr)*System::BitsPerByte bits 
         *  of the value are used.
         *
         *  @return                 The result of converting this value to UIntPtr.
         *  @since 1.0
         */
        UIntPtr                     toUIntPtr() const;

        /**
         *  Converts this value to IntPtrDiff and returns the result.
         *
         *  If the value is too large to fit into the IntPtrDiff type,
         *  only the lowest sizeof(IntPtrDiff)*System::BitsPerByte bits 
         *  of the value are used.
         *
         *  @return                 The result of converting this value to IntPtrDiff.
         *  @since 1.0
         */
        IntPtrDiff                  toIntPtrDiff() const;

        /**
         *  Converts this value to string form.
         *
         *  The conversion is always performed in the same way as implemented
         *  by NumberFormat::format() for an invariant locale.
         *
         *  @param radix            The conversion radix (2 <= \a radix <= 36).
         *  @result                 The string representation of this value.
         *  @since 1.0
         */
        String                      toString(UIntSize radix = 10) const;

        /**
         *  Checks if a value is a probable prime.
         *
         *  @param certainty        The required certainty that the number
         *                          is a probable prime.
         *  @return                 True if the value is a probable prime with
         *                          the specified certainty; false otherwise.
         *  @since 1.0
         */
        bool                        isProbablePrime(UIntSize certainty) const;

        /**
         *  Generates a probable prime of the specified size.
         *
         *  @param numberOfBits     The required number of bits in the generated
         *                          probable prime exclusing the sign bit.
         *  @param certainty        The required certainty that the number
         *                          is a probable prime.
         *  @param random           The random number generator to use.
         *  @return                 A probable prime.
         *  @since 1.0
         */
        static BinaryInteger        generateProbablePrime(UIntSize numberOfBits, UIntSize certainty, Random & random);

        //////////
        //  Implementation
    private:
        UInt32 *                    _pBits;     //  [0] == lower 32 bits, etc.
        UIntSize                    _wordsAllocated;
        UIntSize                    _wordsUsed; //  _pBits[_bitsUsed - 1] is the highest significant word

        //  Helpers
        void                        _ensureAllocation(UIntSize allocation);
        void                        _ensureUsage(UIntSize usage);
        bool                        _isNegative() const;
        void                        _normalize();
        void                        _negate();
        static int                  _compareWords(const UInt32 * pOp1, UIntSize op1Length, UInt32 op1Pad, const UInt32 * pOp2, UIntSize op2Length, UInt32 op2Pad);
        static void                 _clearWords(UInt32 * pDest, UIntSize destLength);
        static void                 _copyWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pSrc, UIntSize srcLength, UInt32 srcPad);
        static UIntSize             _countSignificantWords(const UInt32 * pSrc, UIntSize srcLength);
        static void                 _andWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, UInt32 op1Pad, const UInt32 * pOp2, UIntSize op2Length, UInt32 op2Pad);
        static void                 _orWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, UInt32 op1Pad, const UInt32 * pOp2, UIntSize op2Length, UInt32 op2Pad);
        static void                 _xorWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, UInt32 op1Pad, const UInt32 * pOp2, UIntSize op2Length, UInt32 op2Pad);
        static void                 _addWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, UInt32 op1Pad, const UInt32 * pOp2, UIntSize op2Length, UInt32 op2Pad);
        static void                 _addWordsInPlace(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp2, UIntSize op2Length, UInt32 op2Pad);
        static void                 _subtractWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, UInt32 op1Pad, const UInt32 * pOp2, UIntSize op2Length, UInt32 op2Pad);
        static void                 _subtractWordsInPlace(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp2, UIntSize op2Length, UInt32 op2Pad);
        static void                 _multiplyByWord(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, UInt32 op2);
        static void                 _multiplyWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, const UInt32 * pOp2, UIntSize op2Length);
        static void                 _multiplyWordsSimple(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, const UInt32 * pOp2, UIntSize op2Length);
        static void                 _multiplyWordsDivideAndConquer(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, const UInt32 * pOp2, UIntSize op2Length);
        static void                 _shiftWordsLeftInPlace(UInt32 * pDest, UIntSize destLength, UIntSize shiftCount, UInt32 pad);
        static void                 _shiftWordsRightInPlace(UInt32 * pDest, UIntSize destLength, UIntSize shiftCount, UInt32 pad);
        static carl::UInt32         _guessPartialDivisionResult(carl::UInt32 op1hi, carl::UInt32 op1mid, carl::UInt32 op1lo, carl::UInt32 op2hi, carl::UInt32 op2mid);
        static void                 _divideAndModulo(const UInt32 * pOp1, UIntSize op1Length, const UInt32 * pOp2, UIntSize op2Length, UInt32 * pQuotient, UIntSize quotientLength, UInt32 * pRemainder, UIntSize remainderLength);
        void                        _fromString(const String & s, UIntSize radix) throw (ParseException);
        bool                        _isRabinMillerProbablePrime(UIntSize certainty) const;
        static void                 _gcde(BinaryInteger a, BinaryInteger b, BinaryInteger & u, BinaryInteger & v, BinaryInteger & g);
    };

    /**
     *  @class BinaryFixed CARL.hpp
     *  An arbitrary-precision binary fixed-point number.
     *
     *  Conceptually, a BinaryFixed instance represents a signed binary
     *  integer magnitude with a scale; the value of such a BinaryFixed
     *  instance is magnitude * (2 ** -scale) [i.e. a BinaryFixed value
     *  has exactly scale digits after the decimal point].
     *
     *  @since 1.0
     */
    class CARL_PUBLIC BinaryFixed
    {
        //////////
        //  Construction/destruction/assignment
    private:
        BinaryFixed(bool isNegative, const BinaryInteger & magnitude, UInt32 scale);
    public:
        /**
         *  The default constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to 0.
         *
         *  @since 1.0
         */
        BinaryFixed();

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(Int8 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(Int16 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(Int32 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(Int64 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(UInt8 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(UInt16 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(UInt32 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(UInt64 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(Real32 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(Real64 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed BinaryFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFixed(const BinaryInteger & value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFixed instance to the specified
         *  magnitude value and scale.
         *
         *  @param magnitude        The magnitude to initialize with.
         *  @param scale            The scale to use.
         *  @since 1.0
         */
        BinaryFixed(const BinaryInteger & magnitude, UInt32 scale);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryFixed()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a radix 10 string representation
         *                          of a fixed-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a fixed-point value.
         *  @since 1.0
         */
        explicit BinaryFixed(const char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryFixed()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISO-10646 string
         *                          containing a radix 10 string representation
         *                          of a fixed-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a fixed-point value.
         *  @since 1.0
         */
        explicit BinaryFixed(const Char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryFixed()
         *  for the invariant locale.
         *
         *  @param valueString      A string containing a radix 10 string representation
         *                          of a fixed-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a fixed-point value.
         *  @since 1.0
         */
        explicit BinaryFixed(const String & valueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseBinaryFixed() for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a string representation of a 
         *                          fixed-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a fixed-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          fixed-point value for the specified radix.
         *  @since 1.0
         */
        BinaryFixed(const char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseBinaryFixed() for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISO-10646 string
         *                          containing a string representation of a 
         *                          fixed-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a fixed-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          fixed-point value for the specified radix.
         *  @since 1.0
         */
        BinaryFixed(const Char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseBinaryFixed() for the invariant locale.
         *
         *  @param valueString      A string containing a string representation of a 
         *                          fixed-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a fixed-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          fixed-point value for the specified radix.
         *  @since 1.0
         */
        BinaryFixed(const String & valueString, UIntSize radix) throw (ParseException);

        /**
         *  The copy constructor.
         *
         *  Initializes the newly created BinaryFixed instance to the same
         *  value as an existing BinaryFixed instance.
         *
         *  @param src              The instance to copy value from.
         *  @since 1.0
         */
        BinaryFixed(const BinaryFixed & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~BinaryFixed();

        /**
         *  The assignment operator.
         *
         *  Makes the BinaryFixed instance to have the same value as another
         *  BinaryFixed instance.
         *
         *  @param src              The instance to copy value from.
         *  @since 1.0
         */
        BinaryFixed &               operator = (const BinaryFixed & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the default division precision.
         *
         *  By default, when one BinaryFixed value is divided by another, the scale
         *  of the result is the greater of the operand scales. However, it is sometimes
         *  necessary to perform division with a greater precision to minimize
         *  rounding errors. By setting the default division precision property, the
         *  client ensures that at least that many significant digits will be
         *  available in a division result.
         *
         *  The default division precision is a global setting, shared by all
         *  threads of an application.
         *
         *  @return                 The default division precision.
         *  @since 1.0
         */
        static UIntSize             getDefaultDivisionPrecision();

        /**
         *  Sets the default division precision.
         *
         *  By default, when one BinaryFixed value is divided by another, the scale
         *  of the result is the greater of the operand scales. However, it is sometimes
         *  necessary to perform division with a greater precision to minimize
         *  rounding errors. By setting the default division precision property, the
         *  client ensures that at least that many significant digits will be
         *  available in a division result.
         *
         *  The default division precision is a global setting, shared by all
         *  threads of an application.
         *
         *  @param precision        The new default division precision.
         *  @since 1.0
         */
        static void                 setDefaultDivisionPrecision(UIntSize precision);

        /**
         *  Returns the value unchanged.
         *
         *  @return                 A copy of an argument.
         *  @since 1.0
         */
        BinaryFixed                 operator + () const;

        /**
         *  Returns the value negated.
         *
         *  @return                 A negated argument.
         *  @since 1.0
         */
        BinaryFixed                 operator - () const;

        /**
         *  Returns the sum of two values.
         *
         *  @param op2              The 2nd operand of an addition.
         *  @return                 The addition result.
         *  @since 1.0
         */
        BinaryFixed                 operator + (const BinaryFixed & op2) const;

        /**
         *  Returns the difference of two values.
         *
         *  @param op2              The 2nd operand of a subtraction.
         *  @return                 The subtraction result.
         *  @since 1.0
         */
        BinaryFixed                 operator - (const BinaryFixed & op2) const;

        /**
         *  Returns the product of two values.
         *
         *  @param op2              The 2nd operand of a multiplication.
         *  @return                 The multiplication result.
         *  @since 1.0
         */
        BinaryFixed                 operator * (const BinaryFixed & op2) const;

        /**
         *  Returns the quotient of two values.
         *
         *  The division result is represented with at least as many fractional digits
         *  as both operands have. This precision can be further increased by setting 
         *  the default division precision for BinaryFixed values.
         *
         *  @param op2              The 2nd operand of a division.
         *  @return                 The division result.
         *  @since 1.0
         */
        BinaryFixed                 operator / (const BinaryFixed & op2) const;

        /**
         *  Assigns the sum of two values to this value.
         *
         *  @param op2              The 2nd operand of an addition.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        BinaryFixed &               operator += (const BinaryFixed & op2);

        /**
         *  Assigns the difference of two values to this value.
         *
         *  @param op2              The 2nd operand of a subtraction.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        BinaryFixed &               operator -= (const BinaryFixed & op2);

        /**
         *  Assigns the product of two values to this value.
         *
         *  @param op2              The 2nd operand of a multiplication.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        BinaryFixed &               operator *= (const BinaryFixed & op2);

        /**
         *  Assigns the quotient of two values to this value.
         *
         *  The division result is represented with at least as many fractional digits
         *  as both operands have. This precision can be further increased by setting 
         *  the default division precision for BinaryFixed values.
         *
         *  @param op2              The 2nd operand of a division.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        BinaryFixed &               operator /= (const BinaryFixed & op2);

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value after increment.
         *  @since 1.0
         */
        BinaryFixed                 operator ++();

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value after decrement.
         *  @since 1.0
         */
        BinaryFixed                 operator --();

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value before increment.
         *  @since 1.0
         */
        BinaryFixed                 operator ++(int);

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value before decrement.
         *  @since 1.0
         */
        BinaryFixed                 operator --(int);

        /**
         *  Compares two values.
         *
         *  @param op2              The value to compare this binary fixed with.
         *  @return             
         *                          -   0 if the two values are equal.
         *                          -   <0 if this value is smaller than \a op2.
         *                          -   >0 if this value is greater than \a op2.
         *  @since 1.0
         */
        int                         compare(const BinaryFixed & op2) const;

        /**
         *  Compares two values for equality.
         *
         *  @param op2              The value to compare this binary fixed with.
         *  @return                 True if this value is equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator == (const BinaryFixed & op2) const;

        /**
         *  Compares two values for inequality.
         *
         *  @param op2              The value to compare this binary fixed with.
         *  @return                 True if this value is not equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator != (const BinaryFixed & op2) const;

        /**
         *  Checks if one value is smaller than another value.
         *
         *  @param op2              The value to compare this binary fixed with.
         *  @return                 True if this value is smaller than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator < (const BinaryFixed & op2) const;

        /**
         *  Checks if one value is smaller than or equal to another value.
         *
         *  @param op2              The value to compare this binary fixed with.
         *  @return                 True if this value is smaller than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator <= (const BinaryFixed & op2) const;

        /**
         *  Checks if one value is greater than another value.
         *
         *  @param op2              The value to compare this binary fixed with.
         *  @return                 True if this value is greater than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator > (const BinaryFixed & op2) const;

        /**
         *  Checks if one value is greater than or equal to another value.
         *
         *  @param op2              The value to compare this binary fixed with.
         *  @return                 True if this value is greater than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator >= (const BinaryFixed & op2) const;

        /**
         *  Converts this value to Int8 and returns the result.
         *
         *  If the value is too large to fit into the Int8 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int8.
         *  @since 1.0
         */
        Int8                        toInt8(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int16 and returns the result.
         *
         *  If the value is too large to fit into the Int16 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int16.
         *  @since 1.0
         */
        Int16                       toInt16(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int32 and returns the result.
         *
         *  If the value is too large to fit into the Int32 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int32.
         *  @since 1.0
         */
        Int32                       toInt32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int64 and returns the result.
         *
         *  If the value is too large to fit into the Int64 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int64.
         *  @since 1.0
         */
        Int64                       toInt64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt8 and returns the result.
         *
         *  If the value is too large to fit into the UInt8 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt8.
         *  @since 1.0
         */
        UInt8                       toUInt8(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt16 and returns the result.
         *
         *  If the value is too large to fit into the UInt16 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt16.
         *  @since 1.0
         */
        UInt16                      toUInt16(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt32 and returns the result.
         *
         *  If the value is too large to fit into the UInt32 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt32.
         *  @since 1.0
         */
        UInt32                      toUInt32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt64 and returns the result.
         *
         *  If the value is too large to fit into the UInt64 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt64.
         *  @since 1.0
         */
        UInt64                      toUInt64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UIntSize and returns the result.
         *
         *  If the value is too large to fit into the UIntSize type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UIntSize.
         *  @since 1.0
         */
        UIntSize                    toUIntSize(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UIntPtr and returns the result.
         *
         *  If the value is too large to fit into the UIntPtr type,
         *  only the lowest sizeof(UIntPtr)*System::BitsPerByte bits 
         *  of the value are used.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UIntPtr.
         *  @since 1.0
         */
        UIntPtr                     toUIntPtr(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to IntPtrDiff and returns the result.
         *
         *  If the value is too large to fit into the IntPtrDiff type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to IntPtrDiff.
         *  @since 1.0
         */
        IntPtrDiff                  toIntPtrDiff(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Real32 and returns the result.
         *
         *  If the value is too large to fit into the Real32 type,
         *  the result of conversion is a positive or negative infinity,
         *  depending on the sign of an argument.
         *
         *  @param roundingMode     The rounding mode to use.
         *  @return                 The result of converting this value to Real32.
         *  @since 1.0
         */
        Real32                      toReal32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Real64 and returns the result.
         *
         *  If the value is too large to fit into the Real64 type,
         *  the result of conversion is a positive or negative infinity,
         *  depending on the sign of an argument.
         *
         *  @param roundingMode     The rounding mode to use.
         *  @return                 The result of converting this value to Real64.
         *  @since 1.0
         */
        Real64                      toReal64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to BinaryInteger and returns the result.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to BinaryInteger.
         *  @since 1.0
         */
        BinaryInteger               toBinaryInteger(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to string form.
         *
         *  The conversion is always performed in the same way as implemented
         *  by NumberFormat::format() for an invariant locale.
         *
         *  @param radix            The conversion radix (2 <= \a radix <= 36).
         *  @result                 The string representation of this value.
         *  @since 1.0
         */
        String                      toString(UIntSize radix = 10) const;

        /**
         *  Returns the sign of this fixed point value.
         *
         *  @return
         *                          -   0 of \a (*this) = 0.
         *                          -   1 of \a (*this) > 0.
         *                          -   -1 of \a (*this) < 0.
         *  @since 1.0
         */
        int                         getSign() const;

        /**
         *  Returns the absolute value of the magnitude portion of this 
         *  binary fixed value.
         *
         *  @return                 The magnitude portion of this binary fixed value.
         *  @since 1.0
         */
        BinaryInteger               getMagnitude() const;

        /**
         *  Returns the scale portion of this binary fixed value.
         *
         *  @return                 The scale portion of this binary fixed value.
         *  @since 1.0
         */
        UInt32                      getScale() const;

        /**
         *  Changes the scale of this binary fixed value.
         *
         *  If the new scale is smaller than the old precision and
         *  some significant digits have to be discarded, the rounding is
         *  performed according to the specified rounding mode.
         *
         *  @param scale            The new scale.
         *  @param roundingMode     The rounding mode to use if rounsing is necessary.
         *  @since 1.0
         */
        void                        setScale(UInt32 scale, RoundingMode roundingMode = RoundingMode_TowardsZero);

        //////////
        //  Implementation
    private:
        //  Default division precision
        static UIntSize             _defaultDivisionPrecision;
        static CriticalSection *    _getStaticGuard();

        //  Effective value == sign * value * 2 ** (-_scale)
        bool                        _isNegative;    //  true == value < 0
        BinaryInteger               _magnitude;     //  Magnitude; always >= 0
        UInt32                      _scale;         //  always >= 0; <true value> = (sign) * magnitude * (2 ** -scale))

        //  Helpers
        void                        _normalize();
        void                        _fromReal(Real64 value);
        void                        _fromString(const String & s, UIntSize radix) throw (ParseException);
        static void                 _multiplyBinaryIntegerByRadix(BinaryInteger & bi, UInt8 radix, UIntSize repeatCount);
    };

    /**
     *  @class BinaryFloat CARL.hpp
     *  An arbitrary-precision binary floating-point number.
     *
     *  Conceptually, a BinaryFloat instance represents a signed binary
     *  mantissa integer value with a scale; the value of such a BinaryFloat
     *  instance is mantissa * (2 ** scale). In addition, a BinaryFloat
     *  instance can also have one of the 4 special values (positive
     *  infility, negative infinity, quiet NaN and signaling NaN).
     *
     *  @since 1.0
     */
    class CARL_PUBLIC BinaryFloat
    {
        //////////
        //  Construction/destruction/assignment
    private:
        BinaryFloat(bool isNegative, const BinaryInteger & mantissa, const BinaryInteger & scale);
    public:
        /**
         *  Different categories of BinaryFloat values.
         *
         *  @since 1.0
         */
        enum ValueCategory
        {
            /**
             *  The BinaryFloat value is a "normal" (finite numeric) value.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ValueCategory_Normal            = 1,

            /**
             *  The BinaryFloat value is a positive infinity.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ValueCategory_PositiveInfinity  = 2,

            /**
             *  The BinaryFloat value is a negative infinity.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ValueCategory_NegativeInfinity  = 3,

            /**
             *  The BinaryFloat value is a signaling NaN.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ValueCategory_SignalingNan      = 4,

            /**
             *  The BinaryFloat value is a quiet NaN.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ValueCategory_QuietNan          = 5
        };

        /**
         *  The default constructor.
         *
         *  Initializes the newly created BinaryFloat instance to 0.
         *
         *  @since 1.0
         */
        BinaryFloat();

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to a special
         *  value (infinity or NaN).
         *
         *  @param category         The special value category (cannot be Normal).
         *  @since 1.0
         */
        BinaryFloat(ValueCategory category);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(Int8 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(Int16 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(Int32 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(Int64 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(UInt8 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(UInt16 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(UInt32 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(UInt64 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(Real32 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(Real64 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        BinaryFloat(const BinaryInteger & value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the specified
         *  mantissa value and scale.
         *
         *  @param mantissa         The mantissa to initialize with.
         *  @param scale            The scale to use.
         *  @since 1.0
         */
        BinaryFloat(const BinaryInteger & mantissa, const BinaryInteger & scale);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFloat instance is initialized by parsing
         *  a string representation of a floating-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryFloat()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a radix 10 string representation
         *                          of a floating-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a floating-point value.
         *  @since 1.0
         */
        explicit BinaryFloat(const char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFloat instance is initialized by parsing
         *  a string representation of a floating-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryFloat()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISO-10646 string
         *                          containing a radix 10 string representation
         *                          of a floating-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a floating-point value.
         *  @since 1.0
         */
        explicit BinaryFloat(const Char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFloat instance is initialized by parsing
         *  a string representation of a floating-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseBinaryFloat()
         *  for the invariant locale.
         *
         *  @param valueString      A string containing a radix 10 string representation
         *                          of a floating-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a floating-point value.
         *  @since 1.0
         */
        explicit BinaryFloat(const String & valueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFloat instance is initialized by parsing
         *  a string representation of a floating-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseBinaryFloat() for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a string representation of a 
         *                          floating-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a floating-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          floating-point value for the specified radix.
         *  @since 1.0
         */
        BinaryFloat(const char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFloat instance is initialized by parsing
         *  a string representation of a floating-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseBinaryFloat() for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISO-10646 string
         *                          containing a string representation of a 
         *                          floating-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a floating-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          floating-point value for the specified radix.
         *  @since 1.0
         */
        BinaryFloat(const Char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created BinaryFloat instance is initialized by parsing
         *  a string representation of a floating-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseBinaryFloat() for the invariant locale.
         *
         *  @param valueString      A string containing a string representation of a 
         *                          floating-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a floating-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          floating-point value for the specified radix.
         *  @since 1.0
         */
        BinaryFloat(const String & valueString, UIntSize radix) throw (ParseException);

        /**
         *  The copy constructor.
         *
         *  Initializes the newly created BinaryFloat instance to the same
         *  value as an existing BinaryFloat instance.
         *
         *  @param src              The instance to copy value from.
         *  @since 1.0
         */
        BinaryFloat(const BinaryFloat & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~BinaryFloat();

        /**
         *  The assignment operator.
         *
         *  Makes the BinaryFloat instance to have the same value as another
         *  BinaryFloat instance.
         *
         *  @param src              The instance to copy value from.
         *  @since 1.0
         */
        BinaryFloat &               operator = (const BinaryFloat & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the default division precision.
         *
         *  By default, when one BinaryFloat value is divided by another, the number
         *  of significant digits in the result is the same as the number of significant
         *  digits in the longer of the two division operands. However, it is sometimes
         *  necessary to perform division with a greater precision to minimize
         *  rounding errors. By setting the default division precision property, the
         *  client ensures that at least that many significant digits will be
         *  available in a division result.
         *
         *  The default division precision is a global setting, shared by all
         *  threads of an application.
         *
         *  @return                 The default division precision.
         *  @since 1.0
         */
        static UIntSize             getDefaultDivisionPrecision();

        /**
         *  Sets the default division precision.
         *
         *  By default, when one BinaryFloat value is divided by another, the number
         *  of significant digits in the result is the same as the number of significant
         *  digits in the longer of the two division operands. However, it is sometimes
         *  necessary to perform division with a greater precision to minimize
         *  rounding errors. By setting the default division precision property, the
         *  client ensures that at least that many significant digits will be
         *  available in a division result.
         *
         *  The default division precision is a global setting, shared by all
         *  threads of an application.
         *
         *  @param precision        The new default division precision.
         *  @since 1.0
         */
        static void                 setDefaultDivisionPrecision(UIntSize precision);

        /**
         *  Returns the value unchanged.
         *
         *  @return                 A copy of an argument.
         *  @since 1.0
         */
        BinaryFloat                 operator + () const;

        /**
         *  Returns the value negated.
         *
         *  @return                 A negated argument.
         *  @since 1.0
         */
        BinaryFloat                 operator - () const;

        /**
         *  Returns the sum of two values.
         *
         *  @param op2              The 2nd operand of an addition.
         *  @return                 The addition result.
         *  @since 1.0
         */
        BinaryFloat                 operator + (const BinaryFloat & op2) const;

        /**
         *  Returns the difference of two values.
         *
         *  @param op2              The 2nd operand of a subtraction.
         *  @return                 The subtraction result.
         *  @since 1.0
         */
        BinaryFloat                 operator - (const BinaryFloat & op2) const;

        /**
         *  Returns the product of two values.
         *
         *  @param op2              The 2nd operand of a multiplication.
         *  @return                 The multiplication result.
         *  @since 1.0
         */
        BinaryFloat                 operator * (const BinaryFloat & op2) const;

        /**
         *  Returns the quotient of two values.
         *
         *  The division result is represented with at least as many binary digits
         *  as the longets of the division operands has. This precision can be further
         *  increased by setting the default division precision for BinaryFloat values.
         *
         *  @param op2              The 2nd operand of a division.
         *  @return                 The division result.
         *  @since 1.0
         */
        BinaryFloat                 operator / (const BinaryFloat & op2) const;

        /**
         *  Assigns the sum of two values to this value.
         *
         *  @param op2              The 2nd operand of an addition.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        BinaryFloat &               operator += (const BinaryFloat & op2);

        /**
         *  Assigns the difference of two values to this value.
         *
         *  @param op2              The 2nd operand of a subtraction.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        BinaryFloat &               operator -= (const BinaryFloat & op2);

        /**
         *  Assigns the product of two values to this value.
         *
         *  @param op2              The 2nd operand of a multiplication.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        BinaryFloat &               operator *= (const BinaryFloat & op2);

        /**
         *  Assigns the quotient of two values to this value.
         *
         *  The division result is represented with at least as many binary digits
         *  as the longets of the division operands has. This precision can be further
         *  increased by setting the default division precision for BinaryFloat values.
         *
         *  @param op2              The 2nd operand of a division.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        BinaryFloat &               operator /= (const BinaryFloat & op2);

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value after increment.
         *  @since 1.0
         */
        BinaryFloat                 operator ++();

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value after decrement.
         *  @since 1.0
         */
        BinaryFloat                 operator --();

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value before increment.
         *  @since 1.0
         */
        BinaryFloat                 operator ++(int);

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value before decrement.
         *  @since 1.0
         */
        BinaryFloat                 operator --(int);

        /**
         *  Compares two values.
         *
         *  @param op2              The value to compare this binary float with.
         *  @return             
         *                          -   0 if the two values are equal.
         *                          -   <0 if this value is smaller than \a op2.
         *                          -   >0 if this value is greater than \a op2.
         *  @since 1.0
         */
        int                         compare(const BinaryFloat & op2) const;

        /**
         *  Compares two values for equality.
         *
         *  @param op2              The value to compare this binary float with.
         *  @return                 True if this value is equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator == (const BinaryFloat & op2) const;

        /**
         *  Compares two values for inequality.
         *
         *  @param op2              The value to compare this binary float with.
         *  @return                 True if this value is not equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator != (const BinaryFloat & op2) const;

        /**
         *  Checks if one value is smaller than another value.
         *
         *  @param op2              The value to compare this binary float with.
         *  @return                 True if this value is smaller than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator < (const BinaryFloat & op2) const;

        /**
         *  Checks if one value is smaller than or equal to another value.
         *
         *  @param op2              The value to compare this binary float with.
         *  @return                 True if this value is smaller than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator <= (const BinaryFloat & op2) const;

        /**
         *  Checks if one value is greater than another value.
         *
         *  @param op2              The value to compare this binary float with.
         *  @return                 True if this value is greater than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator > (const BinaryFloat & op2) const;

        /**
         *  Checks if one value is greater than or equal to another value.
         *
         *  @param op2              The value to compare this binary float with.
         *  @return                 True if this value is greater than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator >= (const BinaryFloat & op2) const;

        /**
         *  Converts this value to Int8 and returns the result.
         *
         *  If the value is too large to fit into the Int8 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int8.
         *  @since 1.0
         */
        Int8                        toInt8(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int16 and returns the result.
         *
         *  If the value is too large to fit into the Int16 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int16.
         *  @since 1.0
         */
        Int16                       toInt16(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int32 and returns the result.
         *
         *  If the value is too large to fit into the Int32 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int32.
         *  @since 1.0
         */
        Int32                       toInt32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int64 and returns the result.
         *
         *  If the value is too large to fit into the Int64 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int64.
         *  @since 1.0
         */
        Int64                       toInt64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt8 and returns the result.
         *
         *  If the value is too large to fit into the UInt8 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt8.
         *  @since 1.0
         */
        UInt8                       toUInt8(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt16 and returns the result.
         *
         *  If the value is too large to fit into the UInt16 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt16.
         *  @since 1.0
         */
        UInt16                      toUInt16(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt32 and returns the result.
         *
         *  If the value is too large to fit into the UInt32 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt32.
         *  @since 1.0
         */
        UInt32                      toUInt32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt64 and returns the result.
         *
         *  If the value is too large to fit into the UInt64 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt64.
         *  @since 1.0
         */
        UInt64                      toUInt64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UIntSize and returns the result.
         *
         *  If the value is too large to fit into the UIntSize type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UIntSize.
         *  @since 1.0
         */
        UIntSize                    toUIntSize(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UIntPtr and returns the result.
         *
         *  If the value is too large to fit into the UIntPtr type,
         *  only the lowest sizeof(UIntPtr)*System::BitsPerByte bits 
         *  of the value are used.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UIntPtr.
         *  @since 1.0
         */
        UIntPtr                     toUIntPtr(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to IntPtrDiff and returns the result.
         *
         *  If the value is too large to fit into the IntPtrDiff type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to IntPtrDiff.
         *  @since 1.0
         */
        IntPtrDiff                  toIntPtrDiff(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Real32 and returns the result.
         *
         *  If the value is too large to fit into the Real32 type,
         *  the result of conversion is a positive or negative infinity,
         *  depending on the sign of an argument.
         *
         *  @param roundingMode     The rounding mode to use.
         *  @return                 The result of converting this value to Real32.
         *  @since 1.0
         */
        Real32                      toReal32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Real64 and returns the result.
         *
         *  If the value is too large to fit into the Real64 type,
         *  the result of conversion is a positive or negative infinity,
         *  depending on the sign of an argument.
         *
         *  @param roundingMode     The rounding mode to use.
         *  @return                 The result of converting this value to Real64.
         *  @since 1.0
         */
        Real64                      toReal64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to BinaryInteger and returns the result.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to BinaryInteger.
         *  @since 1.0
         */
        BinaryInteger               toBinaryInteger(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to string form.
         *
         *  The conversion is always performed in the same way as implemented
         *  by NumberFormat::format() for an invariant locale.
         *
         *  @param radix            The conversion radix (2 <= \a radix <= 36).
         *  @result                 The string representation of this value.
         *  @since 1.0
         */
        String                      toString(UIntSize radix = 10) const;

        /**
         *  Returns the sign of this floating point value.
         *
         *  The call is an error if this floating point value is an infinity
         *  or a NaN.
         *
         *  @return
         *                          -   0 of \a (*this) = 0.
         *                          -   1 of \a (*this) > 0.
         *                          -   -1 of \a (*this) < 0.
         *  @since 1.0
         */
        int                         getSign() const;

        /**
         *  Returns the absolute value of the mantissa portion of this 
         *  binary float value.
         *
         *  The call is an error if this floating point value is an infinity
         *  or a NaN.
         *
         *  @return                 The mantissa portion of this binary float value.
         *  @since 1.0
         */
        BinaryInteger               getMantissa() const;

        /**
         *  Returns the scale portion of this binary float value.
         *
         *  The call is an error if this floating point value is an infinity
         *  or a NaN.
         *
         *  @return                 The scale portion of this binary float value.
         *  @since 1.0
         */
        BinaryInteger               getScale() const;

        /**
         *  Changes the precision of this binary float value.
         *
         *  If the new precision is smaller than the old precision and
         *  some significant digits have to be discarded, the rounding is
         *  performed according to the specified rounding mode.
         *
         *  @param precision        The new precision.
         *  @param roundingMode     The rounding mode to use if rounsing is necessary.
         *  @since 1.0
         */
        void                        setPrecision(UIntSize precision, RoundingMode roundingMode = RoundingMode_TowardsZero);

        /**
         *  Returns the category of this value.
         *
         *  @return                 The category of this value.
         *  @since 1.0
         */
        ValueCategory               getCategory() const;

        /**
         *  Checks if this value is a "normal" (finite numeric) value.
         *
         *  @return                 True iff this is a "normal" value; false otherwise.
         *  @since 1.0
         */
        bool                        isNormal() const;

        /**
         *  Checks if this value is a positive infinity.
         *
         *  @return                 True iff this is a positive infinity; false otherwise.
         *  @since 1.0
         */
        bool                        isPositiveInfinity() const;

        /**
         *  Checks if this value is a negative infinity.
         *
         *  @return                 True iff this is a negative infinity; false otherwise.
         *  @since 1.0
         */
        bool                        isNegativeInfinity() const;

        /**
         *  Checks if this value is a signaling NaN.
         *
         *  @return                 True iff this is a signaling NaN; false otherwise.
         *  @since 1.0
         */
        bool                        isSignalingNan() const;

        /**
         *  Checks if this value is a quiet NaN.
         *
         *  @return                 True iff this is a quiet NaN; false otherwise.
         *  @since 1.0
         */
        bool                        isQuietNan() const;

        //////////
        //  Implementation
    private:
        //  Default division precision
        static UIntSize             _defaultDivisionPrecision;
        static CriticalSection *    _getStaticGuard();

        //  Effective value == value * (2 ** _scale) UNLESS a special case is in effect
        ValueCategory               _category;
        bool                        _isNegative;    //  true == value < 0
        BinaryInteger               _mantissa;      //  Mantissa; always >= 0
        BinaryInteger               _scale;         //  <true value> = (sign) * (mantissa * (2 ** scale))

        //  Helpers
        void                        _normalize();
        void                        _fromReal(Real64 value);
        void                        _fromString(const String & s, UIntSize radix) throw (ParseException);
    };

    /**
     *  @class DecimalInteger CARL.hpp
     *  An arbitrary-precision decimal integer number.
     *
     *  Conceptually, a DecimalInteger instance represents a pairing of a sign
     *  and an infinite-precision magnitude. The sign of a DecimalInteger value
     *  can be -1 (for negative values), 0 (for zero) and 1 (for positive values).
     *  The magnitude of a DecimalInteger value is represented as a sequence of
     *  decimal digits. Naturally, the decimal integer values manipulated by the
     *  program will all be finite, so starting at some position all higher digits 
     *  will all be 0s; these digits are implied instead of being actually represented
     *  in memory.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DecimalInteger
    {
        //////////
        //  Construction/destruction/assignment
    private:
        DecimalInteger(bool isNegative, UInt32 * pDigits, UIntSize wordsAllocated, UIntSize wordsUsed, bool takeOwnershipOfDigits);
    public:
        /**
         *  The default constructor.
         *
         *  The newly constructed DecimalInteger instance is initialized to 0.
         *
         *  @since 1.0
         */
        DecimalInteger();

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalInteger(Int8 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalInteger(Int16 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalInteger(Int32 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalInteger(Int64 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalInteger(UInt8 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalInteger(UInt16 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalInteger(UInt32 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalInteger instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalInteger(UInt64 value);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalInteger instance is initialized by parsing
         *  a string representation of an integer value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalInteger()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a radix 10 string representation
         *                          of an integer value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          an integer value.
         *  @since 1.0
         */
        explicit DecimalInteger(const char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalInteger instance is initialized by parsing
         *  a string representation of an integer value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalInteger()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISO-10646 string
         *                          containing a radix 10 string representation
         *                          of an integer value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          an integer value.
         *  @since 1.0
         */
        explicit DecimalInteger(const Char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalInteger instance is initialized by parsing
         *  a string representation of an integer value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalInteger()
         *  for the invariant locale.
         *
         *  @param valueString      A string containing a radix 10 string representation
         *                          of an integer value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          an integer value.
         *  @since 1.0
         */
        explicit DecimalInteger(const String & valueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalInteger instance is initialized by parsing
         *  a string representation of an integer value using specified radix. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalInteger()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a string representation of an 
         *                          integer value.
         *  @param radix            The radix used by the string representation
         *                          of an integer value (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of an integer 
         *                          value.
         *  @since 1.0
         */
        DecimalInteger(const char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalInteger instance is initialized by parsing
         *  a string representation of an integer value using specified radix. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalInteger()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISo-10646 string
         *                          containing a string representation of an 
         *                          integer value.
         *  @param radix            The radix used by the string representation
         *                          of an integer value (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of an integer 
         *                          value.
         *  @since 1.0
         */
        DecimalInteger(const Char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalInteger instance is initialized by parsing
         *  a string representation of an integer value using specified radix. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalInteger()
         *  for the invariant locale.
         *
         *  @param valueString      A string representation of an  integer value.
         *  @param radix            The radix used by the string representation
         *                          of an integer value (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of an integer 
         *                          value.
         *  @since 1.0
         */
        DecimalInteger(const String & valueString, UIntSize radix) throw (ParseException);

        /**
         *  The copy constructor.
         *
         *  @param src              The binary integer to copy value from.
         *  @since 1.0
         */
        DecimalInteger(const DecimalInteger & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~DecimalInteger();

        /**
         *  The assignment operator.
         *
         *  @param src              The binary integer to copy value from.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        DecimalInteger &            operator = (const DecimalInteger & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the value unchanged.
         *
         *  @return                 A value equal to the operand.
         *  @since 1.0
         */
        DecimalInteger              operator + () const;

        /**
         *  Returns the negated operand value.
         *
         *  @return                 The negated operand value.
         *  @since 1.0
         */
        DecimalInteger              operator - () const;

        /**
         *  Returns the sum of two values.
         *
         *  @param op2              The 2nd operand of an addition operation.
         *  @return                 The sum of two values.
         *  @since 1.0
         */
        DecimalInteger              operator + (const DecimalInteger & op2) const;

        /**
         *  Returns the difference of two values.
         *
         *  @param op2              The 2nd operand of a subtraction operation.
         *  @return                 The difference of two values.
         *  @since 1.0
         */
        DecimalInteger              operator - (const DecimalInteger & op2) const;

        /**
         *  Returns the product of two values.
         *
         *  @param op2              The 2nd operand of a multiplication operation.
         *  @return                 The product of two values.
         *  @since 1.0
         */
        DecimalInteger              operator * (const DecimalInteger & op2) const;

        /**
         *  Returns the quotient of two values.
         *
         *  @param op2              The 2nd operand of a division operation.
         *  @return                 The quotient of two values.
         *  @since 1.0
         */
        DecimalInteger              operator / (const DecimalInteger & op2) const;

        /**
         *  Returns the remainder from division of two values.
         *
         *  @param op2              The 2nd operand of a remainder operation.
         *  @return                 The remainder from division of two values.
         *  @since 1.0
         */
        DecimalInteger              operator % (const DecimalInteger & op2) const;

        /**
         *  Calculates the sum of two values and assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of an addition operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        DecimalInteger &            operator += (const DecimalInteger & op2);

        /**
         *  Calculates the difference of two values and assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a subtraction operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        DecimalInteger &            operator -= (const DecimalInteger & op2);

        /**
         *  Calculates the product of two values and assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a multiplication operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        DecimalInteger &            operator *= (const DecimalInteger & op2);

        /**
         *  Calculates the quotient of two values and assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a division operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        DecimalInteger &            operator /= (const DecimalInteger & op2);

        /**
         *  Calculates the remainder from division of two values and 
         *  assigns it to the 1st operand.
         *
         *  @param op2              The 2nd operand of a remainder operation.
         *  @return                 A reference to the assigned-to object.
         *  @since 1.0
         */
        DecimalInteger &            operator %= (const DecimalInteger & op2);

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value after increment.
         *  @since 1.0
         */
        DecimalInteger              operator ++();

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value after decrement.
         *  @since 1.0
         */
        DecimalInteger              operator --();

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value before increment.
         *  @since 1.0
         */
        DecimalInteger              operator ++(int);

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value before decrement.
         *  @since 1.0
         */
        DecimalInteger              operator --(int);

        /**
         *  Compares two values.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return             
         *                          -   0 if the two values are equal.
         *                          -   <0 if this value is smaller than \a op2.
         *                          -   >0 if this value is greater than \a op2.
         *  @since 1.0
         */
        int                         compare(const DecimalInteger & op2) const;

        /**
         *  Compares two values for equality.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator == (const DecimalInteger & op2) const;

        /**
         *  Compares two values for inequality.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is not equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator != (const DecimalInteger & op2) const;

        /**
         *  Checks if one value is smaller than another value.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is smaller than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator < (const DecimalInteger & op2) const;

        /**
         *  Checks if one value is smaller than or equal to another value.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is smaller than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator <= (const DecimalInteger & op2) const;

        /**
         *  Checks if one value is greater than another value.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is greater than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator > (const DecimalInteger & op2) const;

        /**
         *  Checks if one value is greater than or equal to another value.
         *
         *  @param op2              The value to compare this binary integer with.
         *  @return                 True if this value is greater than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator >= (const DecimalInteger & op2) const;

        /**
         *  Returns the sign of this value.
         *
         *  @return
         *                          -   0 of \a (*this) = 0.
         *                          -   1 of \a (*this) > 0.
         *                          -   -1 of \a (*this) < 0.
         *  @since 1.0
         */
        int                         getSign() const;

        /**
         *  Returns the number of significant digit in the decimal representation
         *  of this value.
         *
         *  The "required number of digits" is defined as a smallest number of decimal
         *  digitys that can be used to represent the magnitude of this value without 
         *  loss of significance.
         *
         *  @return                 the number of significant digits in the decimal
         *                          representation of this value.
         *  @since 1.0
         */
        UIntSize                    getRequiredDigits() const;

        /**
         *  Returns the digit at the specified position.
         *
         *  Position 0 denotes the lowest digit of the value; position 1
         *  denotes the next digit, and so on.
         *
         *  @param position         The digit position.
         *  @return                 The digit value 0..9 at the specified position.
         *  @since 1.0
         */
        int                         operator [] (UIntSize position) const;

        /**
         *  Returns the digit at the specified position.
         *
         *  Position 0 denotes the lowest digit of the value; position 1
         *  denotes the next digit, and so on.
         *
         *  @param position         The digit position.
         *  @return                 The digit value 0..9 at the specified position.
         *  @since 1.0
         */
        int                         getDigit(UIntSize position) const;

        /**
         *  Sets the digit at the specified position.
         *
         *  Position 0 denotes the lowest digit of the value; position 1
         *  denotes the next digit, and so on.
         *
         *  @param position         The bit position.
         *  @param digit            The digit 0..9 to assign to the specified position.
         *  @since 1.0
         */
        void                        setDigit(UIntSize position, int digit);

        /**
         *  Converts this value to Int8 and returns the result.
         *
         *  If the value is too large to fit into the Int8 type,
         *  only the lowest 8 bits of the value are used.
         *
         *  @return                 The result of converting this value to Int8.
         *  @since 1.0
         */
        Int8                        toInt8() const;

        /**
         *  Converts this value to Int16 and returns the result.
         *
         *  If the value is too large to fit into the Int16 type,
         *  only the lowest 16 bits of the value are used.
         *
         *  @return                 The result of converting this value to Int16.
         *  @since 1.0
         */
        Int16                       toInt16() const;

        /**
         *  Converts this value to Int32 and returns the result.
         *
         *  If the value is too large to fit into the Int32 type,
         *  only the lowest 32 bits of the value are used.
         *
         *  @return                 The result of converting this value to Int32.
         *  @since 1.0
         */
        Int32                       toInt32() const;

        /**
         *  Converts this value to Int64 and returns the result.
         *
         *  If the value is too large to fit into the Int64 type,
         *  only the lowest 64 bits of the value are used.
         *
         *  @return                 The result of converting this value to Int64.
         *  @since 1.0
         */
        Int64                       toInt64() const;

        /**
         *  Converts this value to UInt8 and returns the result.
         *
         *  If the value is too large to fit into the UInt8 type,
         *  only the lowest 8 bits of the value are used.
         *
         *  @return                 The result of converting this value to UInt8.
         *  @since 1.0
         */
        UInt8                       toUInt8() const;

        /**
         *  Converts this value to UInt16 and returns the result.
         *
         *  If the value is too large to fit into the UInt16 type,
         *  only the lowest 16 bits of the value are used.
         *
         *  @return                 The result of converting this value to UInt16.
         *  @since 1.0
         */
        UInt16                      toUInt16() const;

        /**
         *  Converts this value to UInt32 and returns the result.
         *
         *  If the value is too large to fit into the UInt32 type,
         *  only the lowest 32 bits of the value are used.
         *
         *  @return                 The result of converting this value to UInt32.
         *  @since 1.0
         */
        UInt32                      toUInt32() const;

        /**
         *  Converts this value to UInt64 and returns the result.
         *
         *  If the value is too large to fit into the UInt64 type,
         *  only the lowest 64 bits of the value are used.
         *
         *  @return                 The result of converting this value to UInt64.
         *  @since 1.0
         */
        UInt64                      toUInt64() const;

        /**
         *  Converts this value to UIntSize and returns the result.
         *
         *  If the value is too large to fit into the UIntSize type,
         *  only the lowest sizeof(UIntSize)*System::BitsPerByte bits 
         *  of the value are used.
         *
         *  @return                 The result of converting this value to UIntSize.
         *  @since 1.0
         */
        UIntSize                    toUIntSize() const;

        /**
         *  Converts this value to UIntPtr and returns the result.
         *
         *  If the value is too large to fit into the UIntPtr type,
         *  only the lowest sizeof(UIntPtr)*System::BitsPerByte bits 
         *  of the value are used.
         *
         *  @return                 The result of converting this value to UIntPtr.
         *  @since 1.0
         */
        UIntPtr                     toUIntPtr() const;

        /**
         *  Converts this value to IntPtrDiff and returns the result.
         *
         *  If the value is too large to fit into the IntPtrDiff type,
         *  only the lowest sizeof(IntPtrDiff)*System::BitsPerByte bits 
         *  of the value are used.
         *
         *  @return                 The result of converting this value to IntPtrDiff.
         *  @since 1.0
         */
        IntPtrDiff                  toIntPtrDiff() const;

        /**
         *  Converts this value to string form.
         *
         *  The conversion is always performed in the same way as implemented
         *  by NumberFormat::format() for an invariant locale.
         *
         *  @param radix            The conversion radix (2 <= \a radix <= 36).
         *  @result                 The string representation of this value.
         *  @since 1.0
         */
        String                      toString(UIntSize radix = 10) const;

        /**
         *  Returns this value multiplied by a power of 10.
         *
         *  This method provides a decimal equivalent of the shift-left operator,
         *  only shifting is performed by decimal digits instead of bits.
         *
         *  @param powerOf10        The power of 10 to multiply by.
         *  @result                 (*this) * (10 ** \a powerOf10).
         *  @since 1.0
         */
        DecimalInteger              multipliedByPowerOf10(UIntSize powerOf10) const;

        /**
         *  Returns this value divided by a power of 10.
         *
         *  This method provides a decimal equivalent of the shift-right operator,
         *  only shifting is performed by decimal digits instead of bits.
         *
         *  @param powerOf10        The power of 10 to divide by.
         *  @result                 (*this) / (10 ** \a powerOf10).
         *  @since 1.0
         */
        DecimalInteger              dividedByPowerOf10(UIntSize powerOf10) const;

        /**
         *  Multiplies this value by a power of 10.
         *
         *  This method provides a decimal equivalent of the shift-left-and-assign
         *  operator, only shifting is performed by decimal digits instead of bits.
         *
         *  @param powerOf10        The power of 10 to multiply by.
         *  @result                 The reference to the assigned-to value.
         *  @since 1.0
         */
        DecimalInteger &            multiplyByPowerOf10(UIntSize powerOf10);

        /**
         *  Divides this value by a power of 10.
         *
         *  This method provides a decimal equivalent of the shift-right-and-assign
         *  operator, only shifting is performed by decimal digits instead of bits.
         *
         *  @param powerOf10        The power of 10 to divide by.
         *  @result                 The reference to the assigned-to value.
         *  @since 1.0
         */
        DecimalInteger &            divideByPowerOf10(UIntSize powerOf10);

        //////////
        //  Implementation
    private:
        bool                        _isNegative;
        UInt32 *                    _pDigits;   //  [0] == lower 8 digits, etc. Each digit is represented as a BCD using 4 bits.
        UIntSize                    _wordsAllocated;
        UIntSize                    _wordsUsed; //  _pBits[_bitsUsed - 1] is the highest significant word

        //  Helpers
        void                        _ensureAllocation(UIntSize allocation);
        void                        _ensureUsage(UIntSize usage);
        void                        _normalize();
        void                        _fromUInt64(UInt64 value);
        void                        _fromString(const String & s, UIntSize radix) throw (ParseException);
        BinaryInteger               _toBinaryInteger() const;
        static int                  _compareWords(const UInt32 * pOp1, UIntSize op1Length, const UInt32 * pOp2, UIntSize op2Length);
        static UInt32               _addBcdWords(UInt32 w1, UInt32 w2, UInt32 & carry);
        static void                 _addWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, const UInt32 * pOp2, UIntSize op2Length);
        static UInt32               _subtractBcdWords(UInt32 w1, UInt32 w2, UInt32 & carry);
        static void                 _subtractWords(UInt32 * pDest, UIntSize destLength, const UInt32 * pOp1, UIntSize op1Length, const UInt32 * pOp2, UIntSize op2Length);
        static void                 _clearWords(UInt32 * pDest, UIntSize destLength);
        static void                 _addDigitInPlace(UInt32 * pDest, UIntSize destLength, int digit, UIntSize digitPosition);
    };

    /**
     *  @class DecimalFixed CARL.hpp
     *  An arbitrary-precision defimal fixed-point number.
     *
     *  Conceptually, a DecimalFixed instance represents a signed decimal
     *  integer magnitude with a scale; the value of such a DecimalFixed
     *  instance is magnitude * (10 ** -scale) [i.e. a DecimalFixed value
     *  has exactly scale digits after the decimal point].
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DecimalFixed
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to 0.
         *
         *  @since 1.0
         */
        DecimalFixed();

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(Int8 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(Int16 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(Int32 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(Int64 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(UInt8 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(UInt16 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(UInt32 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(UInt64 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(Real32 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(Real64 value);

        /**
         *  The class constructor.
         *
         *  The newly constructed DecimalFixed instance is initialized to
         *  the specified \a value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFixed(const DecimalInteger & value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFixed instance to the specified
         *  magnitude value and scale.
         *
         *  @param magnitude        The magnitude to initialize with.
         *  @param scale            The scale to use.
         *  @since 1.0
         */
        DecimalFixed(const DecimalInteger & magnitude, UInt32 scale);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalFixed()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a radix 10 string representation
         *                          of a fixed-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a fixed-point value.
         *  @since 1.0
         */
        explicit DecimalFixed(const char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalFixed()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISO-10646 string
         *                          containing a radix 10 string representation
         *                          of a fixed-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a fixed-point value.
         *  @since 1.0
         */
        explicit DecimalFixed(const Char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalFixed()
         *  for the invariant locale.
         *
         *  @param valueString      A string containing a radix 10 string representation
         *                          of a fixed-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a fixed-point value.
         *  @since 1.0
         */
        explicit DecimalFixed(const String & valueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseDecimalFloat() for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a string representation of a 
         *                          fixed-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a fixed-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          fixed-point value for the specified radix.
         *  @since 1.0
         */
        DecimalFixed(const char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseDecimalFloat() for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISO-10646 string
         *                          containing a string representation of a 
         *                          fixed-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a fixed-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          fixed-point value for the specified radix.
         *  @since 1.0
         */
        DecimalFixed(const Char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFixed instance is initialized by parsing
         *  a string representation of a fixed-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseDecimalFixed() for the invariant locale.
         *
         *  @param valueString      A string containing a string representation of a 
         *                          fixed-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a fixed-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          fixed-point value for the specified radix.
         *  @since 1.0
         */
        DecimalFixed(const String & valueString, UIntSize radix) throw (ParseException);

        /**
         *  The copy constructor.
         *
         *  Initializes the newly created DecimalFixed instance to the same
         *  value as an existing DecimalFixed instance.
         *
         *  @param src              The instance to copy value from.
         *  @since 1.0
         */
        DecimalFixed(const DecimalFixed & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~DecimalFixed();

        /**
         *  The assignment operator.
         *
         *  Makes the DecimalFixed instance to have the same value as another
         *  DecimalFixed instance.
         *
         *  @param src              The instance to copy value from.
         *  @since 1.0
         */
        DecimalFixed &              operator = (const DecimalFixed & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the default division precision.
         *
         *  By default, when one DecimalFixed value is divided by another, the scale
         *  of the result is the greater of the operand scales. However, it is sometimes
         *  necessary to perform division with a greater precision to minimize
         *  rounding errors. By setting the default division precision property, the
         *  client ensures that at least that many significant digits will be
         *  available in a division result.
         *
         *  The default division precision is a global setting, shared by all
         *  threads of an application.
         *
         *  @return                 The default division precision.
         *  @since 1.0
         */
        static UIntSize             getDefaultDivisionPrecision();

        /**
         *  Sets the default division precision.
         *
         *  By default, when one DecimalFixed value is divided by another, the scale
         *  of the result is the greater of the operand scales. However, it is sometimes
         *  necessary to perform division with a greater precision to minimize
         *  rounding errors. By setting the default division precision property, the
         *  client ensures that at least that many significant digits will be
         *  available in a division result.
         *
         *  The default division precision is a global setting, shared by all
         *  threads of an application.
         *
         *  @param precision        The new default division precision.
         *  @since 1.0
         */
        static void                 setDefaultDivisionPrecision(UIntSize precision);

        /**
         *  Returns the value unchanged.
         *
         *  @return                 A copy of an argument.
         *  @since 1.0
         */
        DecimalFixed                operator + () const;

        /**
         *  Returns the value negated.
         *
         *  @return                 A negated argument.
         *  @since 1.0
         */
        DecimalFixed                operator - () const;

        /**
         *  Returns the sum of two values.
         *
         *  @param op2              The 2nd operand of an addition.
         *  @return                 The addition result.
         *  @since 1.0
         */
        DecimalFixed                operator + (const DecimalFixed & op2) const;

        /**
         *  Returns the difference of two values.
         *
         *  @param op2              The 2nd operand of a subtraction.
         *  @return                 The subtraction result.
         *  @since 1.0
         */
        DecimalFixed                operator - (const DecimalFixed & op2) const;

        /**
         *  Returns the product of two values.
         *
         *  @param op2              The 2nd operand of a multiplication.
         *  @return                 The multiplication result.
         *  @since 1.0
         */
        DecimalFixed                operator * (const DecimalFixed & op2) const;

        /**
         *  Returns the quotient of two values.
         *
         *  The division result is represented with at least as many fractional digits
         *  as both operands have. This precision can be further increased by setting 
         *  the default division precision for DecimalFixed values.
         *
         *  @param op2              The 2nd operand of a division.
         *  @return                 The division result.
         *  @since 1.0
         */
        DecimalFixed                operator / (const DecimalFixed & op2) const;

        /**
         *  Assigns the sum of two values to this value.
         *
         *  @param op2              The 2nd operand of an addition.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        DecimalFixed &              operator += (const DecimalFixed & op2);

        /**
         *  Assigns the difference of two values to this value.
         *
         *  @param op2              The 2nd operand of a subtraction.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        DecimalFixed &              operator -= (const DecimalFixed & op2);

        /**
         *  Assigns the product of two values to this value.
         *
         *  @param op2              The 2nd operand of a multiplication.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        DecimalFixed &              operator *= (const DecimalFixed & op2);

        /**
         *  Assigns the quotient of two values to this value.
         *
         *  The division result is represented with at least as many fractional digits
         *  as both operands have. This precision can be further increased by setting 
         *  the default division precision for DecimalFixed values.
         *
         *  @param op2              The 2nd operand of a division.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        DecimalFixed &              operator /= (const DecimalFixed & op2);

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value after increment.
         *  @since 1.0
         */
        DecimalFixed                operator ++();

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value after decrement.
         *  @since 1.0
         */
        DecimalFixed                operator --();

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value before increment.
         *  @since 1.0
         */
        DecimalFixed                operator ++(int);

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value before decrement.
         *  @since 1.0
         */
        DecimalFixed                operator --(int);

        /**
         *  Compares two values.
         *
         *  @param op2              The value to compare this decimal fixed with.
         *  @return             
         *                          -   0 if the two values are equal.
         *                          -   <0 if this value is smaller than \a op2.
         *                          -   >0 if this value is greater than \a op2.
         *  @since 1.0
         */
        int                         compare(const DecimalFixed & op2) const;

        /**
         *  Compares two values for equality.
         *
         *  @param op2              The value to compare this decimal fixed with.
         *  @return                 True if this value is equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator == (const DecimalFixed & op2) const;

        /**
         *  Compares two values for inequality.
         *
         *  @param op2              The value to compare this decimal fixed with.
         *  @return                 True if this value is not equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator != (const DecimalFixed & op2) const;

        /**
         *  Checks if one value is smaller than another value.
         *
         *  @param op2              The value to compare this decimal fixed with.
         *  @return                 True if this value is smaller than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator < (const DecimalFixed & op2) const;

        /**
         *  Checks if one value is smaller than or equal to another value.
         *
         *  @param op2              The value to compare this decimal fixed with.
         *  @return                 True if this value is smaller than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator <= (const DecimalFixed & op2) const;

        /**
         *  Checks if one value is greater than another value.
         *
         *  @param op2              The value to compare this decimal fixed with.
         *  @return                 True if this value is greater than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator > (const DecimalFixed & op2) const;

        /**
         *  Checks if one value is greater than or equal to another value.
         *
         *  @param op2              The value to compare this decimal fixed with.
         *  @return                 True if this value is greater than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator >= (const DecimalFixed & op2) const;

        /**
         *  Converts this value to Int8 and returns the result.
         *
         *  If the value is too large to fit into the Int8 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int8.
         *  @since 1.0
         */
        Int8                        toInt8(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int16 and returns the result.
         *
         *  If the value is too large to fit into the Int16 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int16.
         *  @since 1.0
         */
        Int16                       toInt16(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int32 and returns the result.
         *
         *  If the value is too large to fit into the Int32 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int32.
         *  @since 1.0
         */
        Int32                       toInt32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int64 and returns the result.
         *
         *  If the value is too large to fit into the Int64 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int64.
         *  @since 1.0
         */
        Int64                       toInt64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt8 and returns the result.
         *
         *  If the value is too large to fit into the UInt8 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt8.
         *  @since 1.0
         */
        UInt8                       toUInt8(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt16 and returns the result.
         *
         *  If the value is too large to fit into the UInt16 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt16.
         *  @since 1.0
         */
        UInt16                      toUInt16(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt32 and returns the result.
         *
         *  If the value is too large to fit into the UInt32 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt32.
         *  @since 1.0
         */
        UInt32                      toUInt32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt64 and returns the result.
         *
         *  If the value is too large to fit into the UInt64 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt64.
         *  @since 1.0
         */
        UInt64                      toUInt64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UIntSize and returns the result.
         *
         *  If the value is too large to fit into the UIntSize type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UIntSize.
         *  @since 1.0
         */
        UIntSize                    toUIntSize(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UIntPtr and returns the result.
         *
         *  If the value is too large to fit into the UIntPtr type,
         *  only the lowest sizeof(UIntPtr)*System::BitsPerByte bits 
         *  of the value are used.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UIntPtr.
         *  @since 1.0
         */
        UIntPtr                     toUIntPtr(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to IntPtrDiff and returns the result.
         *
         *  If the value is too large to fit into the IntPtrDiff type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to IntPtrDiff.
         *  @since 1.0
         */
        IntPtrDiff                  toIntPtrDiff(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Real32 and returns the result.
         *
         *  If the value is too large to fit into the Real32 type,
         *  the result of conversion is a positive or negative infinity,
         *  depending on the sign of an argument.
         *
         *  @param roundingMode     The rounding mode to use.
         *  @return                 The result of converting this value to Real32.
         *  @since 1.0
         */
        Real32                      toReal32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Real64 and returns the result.
         *
         *  If the value is too large to fit into the Real64 type,
         *  the result of conversion is a positive or negative infinity,
         *  depending on the sign of an argument.
         *
         *  @param roundingMode     The rounding mode to use.
         *  @return                 The result of converting this value to Real64.
         *  @since 1.0
         */
        Real64                      toReal64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to DecimalInteger and returns the result.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to DecimalInteger.
         *  @since 1.0
         */
        DecimalInteger              toDecimalInteger(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to string form.
         *
         *  The conversion is always performed in the same way as implemented
         *  by NumberFormat::format() for an invariant locale.
         *
         *  @param radix            The conversion radix (2 <= \a radix <= 36).
         *  @result                 The string representation of this value.
         *  @since 1.0
         */
        String                      toString(UIntSize radix = 10) const;

        /**
         *  Returns the sign of this fixed point value.
         *
         *  @return
         *                          -   0 of \a (*this) = 0.
         *                          -   1 of \a (*this) > 0.
         *                          -   -1 of \a (*this) < 0.
         *  @since 1.0
         */
        int                         getSign() const;

        /**
         *  Returns the absolute value of the magnitude portion of this 
         *  decimal fixed value.
         *
         *  @return                 The magnitude portion of this decimal fixed value.
         *  @since 1.0
         */
        DecimalInteger              getMagnitude() const;

        /**
         *  Returns the scale portion of this decimal fixed value.
         *
         *  @return                 The scale portion of this Decimal fixed value.
         *  @since 1.0
         */
        UInt32                      getScale() const;

        /**
         *  Changes the scale of this decimal fixed value.
         *
         *  If the new scale is smaller than the old precision and
         *  some significant digits have to be discarded, the rounding is
         *  performed according to the specified rounding mode.
         *
         *  @param scale            The new scale.
         *  @param roundingMode     The rounding mode to use if rounsing is necessary.
         *  @since 1.0
         */
        void                        setScale(UInt32 scale, RoundingMode roundingMode = RoundingMode_TowardsZero);

        //////////
        //  Implementation
    private:
        //  Default division precision
        static UIntSize             _defaultDivisionPrecision;
        static CriticalSection *    _getStaticGuard();

        //  Effective value == value * 10 ** (-_scale)
        DecimalInteger              _magnitude;     //  Magnitude, also includes sign
        UInt32                      _scale;         //  always >= 0; <true value> = magnitude * (10 ** -scale))

        //  Helpers
        void                        _normalize();
        void                        _fromReal(Real64 value);
        void                        _fromString(const String & s, UIntSize radix) throw (ParseException);
        static void                 _multiplyDecimalIntegerByRadix(DecimalInteger & bi, UInt8 radix, UIntSize repeatCount);
    };

    /**
     *  @class DecimalFloat CARL.hpp
     *  An arbitrary-precision decimal floating-point number.
     *
     *  Conceptually, a DecimalFloat instance represents a signed binary
     *  mantissa integer value with a scale; the value of such a DecimalFloat
     *  instance is mantissa * (10 ** scale). In addition, a DecimalFloat
     *  instance can also have one of the 4 special values (positive
     *  infility, negative infinity, quiet NaN and signaling NaN).
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DecimalFloat
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  Different categories of DecimalFloat values.
         *
         *  @since 1.0
         */
        enum ValueCategory
        {
            /**
             *  The DecimalFloat value is a "normal" (finite numeric) value.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ValueCategory_Normal            = 1,

            /**
             *  The DecimalFloat value is a positive infinity.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ValueCategory_PositiveInfinity  = 2,

            /**
             *  The DecimalFloat value is a negative infinity.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ValueCategory_NegativeInfinity  = 3,

            /**
             *  The DecimalFloat value is a signaling NaN.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ValueCategory_SignalingNan      = 4,

            /**
             *  The DecimalFloat value is a quiet NaN.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            ValueCategory_QuietNan          = 5
        };

        /**
         *  The default constructor.
         *
         *  Initializes the newly created DecimalFloat instance to 0.
         *
         *  @since 1.0
         */
        DecimalFloat();

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to a special
         *  value (infinity or NaN).
         *
         *  @param category         The special value category (cannot be Normal).
         *  @since 1.0
         */
        DecimalFloat(ValueCategory category);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(Int8 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(Int16 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(Int32 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(Int64 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(UInt8 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(UInt16 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(UInt32 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(UInt64 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(Real32 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(Real64 value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified value.
         *
         *  @param value            The value to initialize with.
         *  @since 1.0
         */
        DecimalFloat(const DecimalInteger & value);

        /**
         *  The class constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the specified
         *  mantissa value and scale.
         *
         *  @param mantissa         The mantissa to initialize with.
         *  @param scale            The scale to use.
         *  @since 1.0
         */
        DecimalFloat(const DecimalInteger & mantissa, const DecimalInteger & scale);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFloat instance is initialized by parsing
         *  a string representation of a floating-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalFloat()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a radix 10 string representation
         *                          of a floating-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a floating-point value.
         *  @since 1.0
         */
        explicit DecimalFloat(const char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFloat instance is initialized by parsing
         *  a string representation of a floating-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalFloat()
         *  for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISO-10646 string
         *                          containing a radix 10 string representation
         *                          of a floating-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a floating-point value.
         *  @since 1.0
         */
        explicit DecimalFloat(const Char * pValueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFloat instance is initialized by parsing
         *  a string representation of a floating-point value using radix 10. The format
         *  of this string must be the same as accepted by NumberParse::parseDecimalFloat()
         *  for the invariant locale.
         *
         *  @param valueString      A string containing a radix 10 string representation
         *                          of a floating-point value.
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid radix 10 representation of
         *                          a floating-point value.
         *  @since 1.0
         */
        explicit DecimalFloat(const String & valueString) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFloat instance is initialized by parsing
         *  a string representation of a floating-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseDecimalFloat() for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ASCII string
         *                          containing a string representation of a 
         *                          floating-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a floating-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          floating-point value for the specified radix.
         *  @since 1.0
         */
        DecimalFloat(const char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFloat instance is initialized by parsing
         *  a string representation of a floating-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseDecimalFloat() for the invariant locale.
         *
         *  @param pValueString     A pointer to a nul-terminated ISO-10646 string
         *                          containing a string representation of a 
         *                          floating-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a floating-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          floating-point value for the specified radix.
         *  @since 1.0
         */
        DecimalFloat(const Char * pValueString, UIntSize radix) throw (ParseException);

        /**
         *  The class constructor.
         *
         *  The newly created DecimalFloat instance is initialized by parsing
         *  a string representation of a floating-point value using the specified 
         *  radix. The format of this string must be the same as accepted by 
         *  NumberParse::parseDecimalFloat() for the invariant locale.
         *
         *  @param valueString      A string containing a string representation of a 
         *                          floating-point value for the specified radix.
         *  @param radix            The radix used by the string representation
         *                          of a floating-point (2 <= \a radix <= 36).
         *  @exception ParseException   Thrown if the specified string does not
         *                          contain a valid representation of a 
         *                          floating-point value for the specified radix.
         *  @since 1.0
         */
        DecimalFloat(const String & valueString, UIntSize radix) throw (ParseException);

        /**
         *  The copy constructor.
         *
         *  Initializes the newly created DecimalFloat instance to the same
         *  value as an existing DecimalFloat instance.
         *
         *  @param src              The instance to copy value from.
         *  @since 1.0
         */
        DecimalFloat(const DecimalFloat & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~DecimalFloat();

        /**
         *  The assignment operator.
         *
         *  Makes the DecimalFloat instance to have the same value as another
         *  DecimalFloat instance.
         *
         *  @param src              The instance to copy value from.
         *  @since 1.0
         */
        DecimalFloat &              operator = (const DecimalFloat & src);

        //////////
        //  Operations
    public:
        /**
         *  Returns the default division precision.
         *
         *  By default, when one DecimalFloat value is divided by another, the number
         *  of significant digits in the result is the same as the number of significant
         *  digits in the longer of the two division operands. However, it is sometimes
         *  necessary to perform division with a greater precision to minimize
         *  rounding errors. By setting the default division precision property, the
         *  client ensures that at least that many significant digits will be
         *  available in a division result.
         *
         *  The default division precision is a global setting, shared by all
         *  threads of an application.
         *
         *  @return                 The default division precision.
         *  @since 1.0
         */
        static UIntSize             getDefaultDivisionPrecision();

        /**
         *  Sets the default division precision.
         *
         *  By default, when one DecimalFloat value is divided by another, the number
         *  of significant digits in the result is the same as the number of significant
         *  digits in the longer of the two division operands. However, it is sometimes
         *  necessary to perform division with a greater precision to minimize
         *  rounding errors. By setting the default division precision property, the
         *  client ensures that at least that many significant digits will be
         *  available in a division result.
         *
         *  The default division precision is a global setting, shared by all
         *  threads of an application.
         *
         *  @param precision        The new default division precision.
         *  @since 1.0
         */
        static void                 setDefaultDivisionPrecision(UIntSize precision);

        /**
         *  Returns the value unchanged.
         *
         *  @return                 A copy of an argument.
         *  @since 1.0
         */
        DecimalFloat                operator + () const;

        /**
         *  Returns the value negated.
         *
         *  @return                 A negated argument.
         *  @since 1.0
         */
        DecimalFloat                operator - () const;

        /**
         *  Returns the sum of two values.
         *
         *  @param op2              The 2nd operand of an addition.
         *  @return                 The addition result.
         *  @since 1.0
         */
        DecimalFloat                operator + (const DecimalFloat & op2) const;

        /**
         *  Returns the difference of two values.
         *
         *  @param op2              The 2nd operand of a subtraction.
         *  @return                 The subtraction result.
         *  @since 1.0
         */
        DecimalFloat                operator - (const DecimalFloat & op2) const;

        /**
         *  Returns the product of two values.
         *
         *  @param op2              The 2nd operand of a multiplication.
         *  @return                 The multiplication result.
         *  @since 1.0
         */
        DecimalFloat                operator * (const DecimalFloat & op2) const;

        /**
         *  Returns the quotient of two values.
         *
         *  The division result is represented with at least as many decimal digits
         *  as the longets of the division operands has. This precision can be further
         *  increased by setting the default division precision for DecimalFloat values.
         *
         *  @param op2              The 2nd operand of a division.
         *  @return                 The division result.
         *  @since 1.0
         */
        DecimalFloat                operator / (const DecimalFloat & op2) const;

        /**
         *  Assigns the sum of two values to this value.
         *
         *  @param op2              The 2nd operand of an addition.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        DecimalFloat &              operator += (const DecimalFloat & op2);

        /**
         *  Assigns the difference of two values to this value.
         *
         *  @param op2              The 2nd operand of a subtraction.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        DecimalFloat &              operator -= (const DecimalFloat & op2);

        /**
         *  Assigns the product of two values to this value.
         *
         *  @param op2              The 2nd operand of a multiplication.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        DecimalFloat &              operator *= (const DecimalFloat & op2);

        /**
         *  Assigns the quotient of two values to this value.
         *
         *  The division result is represented with at least as many decimal digits
         *  as the longets of the division operands has. This precision can be further
         *  increased by setting the default division precision for DecimalFloat values.
         *
         *  @param op2              The 2nd operand of a division.
         *  @return                 The assigned-to value.
         *  @since 1.0
         */
        DecimalFloat &              operator /= (const DecimalFloat & op2);

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value after increment.
         *  @since 1.0
         */
        DecimalFloat                operator ++();

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value after decrement.
         *  @since 1.0
         */
        DecimalFloat                operator --();

        /**
         *  Increments this value by 1.
         *
         *  @return                 The value before increment.
         *  @since 1.0
         */
        DecimalFloat                operator ++(int);

        /**
         *  Decrements this value by 1.
         *
         *  @return                 The value before decrement.
         *  @since 1.0
         */
        DecimalFloat                operator --(int);

        /**
         *  Compares two values.
         *
         *  @param op2              The value to compare this decimal float with.
         *  @return             
         *                          -   0 if the two values are equal.
         *                          -   <0 if this value is smaller than \a op2.
         *                          -   >0 if this value is greater than \a op2.
         *  @since 1.0
         */
        int                         compare(const DecimalFloat & op2) const;

        /**
         *  Compares two values for equality.
         *
         *  @param op2              The value to compare this decimal float with.
         *  @return                 True if this value is equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator == (const DecimalFloat & op2) const;

        /**
         *  Compares two values for inequality.
         *
         *  @param op2              The value to compare this decimal float with.
         *  @return                 True if this value is not equal to \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator != (const DecimalFloat & op2) const;

        /**
         *  Checks if one value is smaller than another value.
         *
         *  @param op2              The value to compare this decimal float with.
         *  @return                 True if this value is smaller than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator < (const DecimalFloat & op2) const;

        /**
         *  Checks if one value is smaller than or equal to another value.
         *
         *  @param op2              The value to compare this decimal float with.
         *  @return                 True if this value is smaller than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator <= (const DecimalFloat & op2) const;

        /**
         *  Checks if one value is greater than another value.
         *
         *  @param op2              The value to compare this decimal float with.
         *  @return                 True if this value is greater than \a op2, 
         *                          false otherwise.
         *  @since 1.0
         */
        bool                        operator > (const DecimalFloat & op2) const;

        /**
         *  Checks if one value is greater than or equal to another value.
         *
         *  @param op2              The value to compare this decimal float with.
         *  @return                 True if this value is greater than or equal to
         *                          \a op2, false otherwise.
         *  @since 1.0
         */
        bool                        operator >= (const DecimalFloat & op2) const;

        /**
         *  Converts this value to Int8 and returns the result.
         *
         *  If the value is too large to fit into the Int8 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int8.
         *  @since 1.0
         */
        Int8                        toInt8(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int16 and returns the result.
         *
         *  If the value is too large to fit into the Int16 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int16.
         *  @since 1.0
         */
        Int16                       toInt16(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int32 and returns the result.
         *
         *  If the value is too large to fit into the Int32 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int32.
         *  @since 1.0
         */
        Int32                       toInt32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Int64 and returns the result.
         *
         *  If the value is too large to fit into the Int64 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to Int64.
         *  @since 1.0
         */
        Int64                       toInt64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt8 and returns the result.
         *
         *  If the value is too large to fit into the UInt8 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt8.
         *  @since 1.0
         */
        UInt8                       toUInt8(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt16 and returns the result.
         *
         *  If the value is too large to fit into the UInt16 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt16.
         *  @since 1.0
         */
        UInt16                      toUInt16(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt32 and returns the result.
         *
         *  If the value is too large to fit into the UInt32 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt32.
         *  @since 1.0
         */
        UInt32                      toUInt32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UInt64 and returns the result.
         *
         *  If the value is too large to fit into the UInt64 type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UInt64.
         *  @since 1.0
         */
        UInt64                      toUInt64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UIntSize and returns the result.
         *
         *  If the value is too large to fit into the UIntSize type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UIntSize.
         *  @since 1.0
         */
        UIntSize                    toUIntSize(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to UIntPtr and returns the result.
         *
         *  If the value is too large to fit into the UIntPtr type,
         *  only the lowest sizeof(UIntPtr)*System::BitsPerByte bits 
         *  of the value are used.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to UIntPtr.
         *  @since 1.0
         */
        UIntPtr                     toUIntPtr(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to IntPtrDiff and returns the result.
         *
         *  If the value is too large to fit into the IntPtrDiff type,
         *  the closest representable value is returned.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to IntPtrDiff.
         *  @since 1.0
         */
        IntPtrDiff                  toIntPtrDiff(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Real32 and returns the result.
         *
         *  If the value is too large to fit into the Real32 type,
         *  the result of conversion is a positive or negative infinity,
         *  depending on the sign of an argument.
         *
         *  @param roundingMode     The rounding mode to use.
         *  @return                 The result of converting this value to Real32.
         *  @since 1.0
         */
        Real32                      toReal32(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to Real64 and returns the result.
         *
         *  If the value is too large to fit into the Real64 type,
         *  the result of conversion is a positive or negative infinity,
         *  depending on the sign of an argument.
         *
         *  @param roundingMode     The rounding mode to use.
         *  @return                 The result of converting this value to Real64.
         *  @since 1.0
         */
        Real64                      toReal64(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to DecimalInteger and returns the result.
         *
         *  If this value has fractional part, the specified rounding mode is
         *  used to round it to the nearest integer.
         *
         *  @param roundingMode     The rounding mode to use when necessary.
         *  @return                 The result of converting this value to DecimalInteger.
         *  @since 1.0
         */
        DecimalInteger              toDecimalInteger(RoundingMode roundingMode = RoundingMode_TowardsZero) const;

        /**
         *  Converts this value to string form.
         *
         *  The conversion is always performed in the same way as implemented
         *  by NumberFormat::format() for an invariant locale.
         *
         *  @param radix            The conversion radix (2 <= \a radix <= 36).
         *  @result                 The string representation of this value.
         *  @since 1.0
         */
        String                      toString(UIntSize radix = 10) const;

        /**
         *  Returns the sign of this floating point value.
         *
         *  The call is an error if this floating point value is an infinity
         *  or a NaN.
         *
         *  @return
         *                          -   0 of \a (*this) = 0.
         *                          -   1 of \a (*this) > 0.
         *                          -   -1 of \a (*this) < 0.
         *  @since 1.0
         */
        int                         getSign() const;

        /**
         *  Returns the absolute value of the mantissa portion of this 
         *  decimal float value.
         *
         *  The call is an error if this floating point value is an infinity
         *  or a NaN.
         *
         *  @return                 The mantissa portion of this decimal float value.
         *  @since 1.0
         */
        DecimalInteger              getMantissa() const;

        /**
         *  Returns the scale portion of this decimal float value.
         *
         *  The call is an error if this floating point value is an infinity
         *  or a NaN.
         *
         *  @return                 The scale portion of this decimal float value.
         *  @since 1.0
         */
        DecimalInteger              getScale() const;

        /**
         *  Changes the precision of this decimal float value.
         *
         *  If the new precision is smaller than the old precision and
         *  some significant digits have to be discarded, the rounding is
         *  performed according to the specified rounding mode.
         *
         *  @param precision        The new precision.
         *  @param roundingMode     The rounding mode to use if rounsing is necessary.
         *  @since 1.0
         */
        void                        setPrecision(UIntSize precision, RoundingMode roundingMode = RoundingMode_TowardsZero);

        /**
         *  Returns the category of this value.
         *
         *  @return                 The category of this value.
         *  @since 1.0
         */
        ValueCategory               getCategory() const;

        /**
         *  Checks if this value is a "normal" (finite numeric) value.
         *
         *  @return                 True iff this is a "normal" value; false otherwise.
         *  @since 1.0
         */
        bool                        isNormal() const;

        /**
         *  Checks if this value is a positive infinity.
         *
         *  @return                 True iff this is a positive infinity; false otherwise.
         *  @since 1.0
         */
        bool                        isPositiveInfinity() const;

        /**
         *  Checks if this value is a negative infinity.
         *
         *  @return                 True iff this is a negative infinity; false otherwise.
         *  @since 1.0
         */
        bool                        isNegativeInfinity() const;

        /**
         *  Checks if this value is a signaling NaN.
         *
         *  @return                 True iff this is a signaling NaN; false otherwise.
         *  @since 1.0
         */
        bool                        isSignalingNan() const;

        /**
         *  Checks if this value is a quiet NaN.
         *
         *  @return                 True iff this is a quiet NaN; false otherwise.
         *  @since 1.0
         */
        bool                        isQuietNan() const;

        //////////
        //  Implementation
    private:
        //  Default division precision
        static UIntSize             _defaultDivisionPrecision;
        static CriticalSection *    _getStaticGuard();

        //  Effective value == value * (10 ** _scale) UNLESS a special case is in effect
        ValueCategory               _category;
        DecimalInteger              _mantissa;      //  Mantissa + sign
        DecimalInteger              _scale;         //  <true value> = (mantissa * (10 ** scale))

        //  Helpers
        void                        _normalize();
        void                        _fromReal(Real64 value);
        void                        _fromString(const String & s, UIntSize radix) throw (ParseException);
    };

    /**
     *  @class Math CARL.hpp
     *  Miscellaneous math services.
     *
     *  This utility class encapsulates common mathematical operations.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC Math
    {
        //////////
        //  This is a utility class
    private:
        Math();
        Math(const Math &);
        ~Math();
        void                        operator = (const Math &);

        //////////
        //  Operations
    public:
        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static Int8                 abs(Int8 x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static Int16                abs(Int16 x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static Int32                abs(Int32 x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static Int64                abs(Int64 x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static Real32               abs(Real32 x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static Real64               abs(Real64 x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static BinaryInteger        abs(const BinaryInteger & x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static BinaryFixed          abs(const BinaryFixed & x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static BinaryFloat          abs(const BinaryFloat & x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static DecimalInteger       abs(const DecimalInteger & x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static DecimalFixed         abs(const DecimalFixed & x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an absolute value of an argument.
         *
         *  @param x                An argument.
         *  @return                 \a x if \a x >= 0; otherwise -\a x.
         *  @since 1.0
         */
        static DecimalFloat         abs(const DecimalFloat & x) { return (x >= 0) ? (x) : (-x); }

        /**
         *  Returns an arccosine of the specified argument.
         *
         *  The argument must be in range [-1.0 .. 1.0].
         *
         *  @param x                An argument.
         *  @return                 arccos(\a x).
         *  @since 1.0
         */
        static Real32               acos(Real32 x);

        /**
         *  Returns an arccosine of the specified argument.
         *
         *  The argument must be in range [-1.0 .. 1.0].
         *
         *  @param x                An argument.
         *  @return                 arccos(\a x).
         *  @since 1.0
         */
        static Real64               acos(Real64 x);

        /**
         *  Returns an arcsine of the specified argument.
         *
         *  The argument must be in range [-1.0 .. 1.0].
         *
         *  @param x                An argument.
         *  @return                 arcsin(\a x).
         *  @since 1.0
         */
        static Real32               asin(Real32 x);

        /**
         *  Returns an arcsine of the specified argument.
         *
         *  The argument must be in range [-1.0 .. 1.0].
         *
         *  @param x                An argument.
         *  @return                 arcsin(\a x).
         *  @since 1.0
         */
        static Real64               asin(Real64 x);

        /**
         *  Returns an arctangent of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 arctan(\a x).
         *  @since 1.0
         */
        static Real32               atan(Real32 x);

        /**
         *  Returns an arctangent of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 arctan(\a x).
         *  @since 1.0
         */
        static Real64               atan(Real64 x);

        /**
         *  Returns an arctangent of the specified argument pair.
         *
         *  If both parameters are 0, the method returns 0. 
         *
         *  @param y                The 1st argument.
         *  @param x                The 2nd argument.
         *  @return                 arctan(\a y / \a x).
         *  @since 1.0
         */
        static Real32               atan2(Real32 y, Real32 x);

        /**
         *  Returns an arctangent of the specified argument pair.
         *
         *  If both parameters are 0, the method returns 0. 
         *
         *  @param y                The 1st argument.
         *  @param x                The 2nd argument.
         *  @return                 arctan(\a y / \a x).
         *  @since 1.0
         */
        static Real64               atan2(Real64 y, Real64 x);

        /**
         *  Returns a floating-point value representing the smallest integer 
         *  that is greater than or equal to an argument.
         *
         *  @param x                An argument.
         *  @return                 A floating-point value representing the smallest 
         *                          integer that is greater than or equal to an \a x.
         *  @since 1.0
         */
        static Real32               ceiling(Real32 x);

        /**
         *  Returns a floating-point value representing the smallest integer 
         *  that is greater than or equal to an argument.
         *
         *  @param x                An argument.
         *  @return                 A floating-point value representing the smallest 
         *                          integer that is greater than or equal to an \a x.
         *  @since 1.0
         */
        static Real64               ceiling(Real64 x);

        /**
         *  Returns a cosine of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 cos(\a x).
         *  @since 1.0
         */
        static Real32               cos(Real32 x);

        /**
         *  Returns a cosine of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 cos(\a x).
         *  @since 1.0
         */
        static Real64               cos(Real64 x);

        /**
         *  Returns a hyperbolic cosine of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 cosh(\a x).
         *  @since 1.0
         */
        static Real32               cosh(Real32 x);

        /**
         *  Returns a hyperbolic cosine of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 cosh(\a x).
         *  @since 1.0
         */
        static Real64               cosh(Real64 x);

        /**
         *  Returns an exponential of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 e**\a x.
         *  @since 1.0
         */
        static Real32               exp(Real32 x);

        /**
         *  Returns an exponential of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 e**\a x.
         *  @since 1.0
         */
        static Real64               exp(Real64 x);

        /**
         *  Returns a floating-point value representing the largest integer 
         *  that is smaller than or equal to an argument.
         *
         *  @param x                An argument.
         *  @return                 A floating-point value representing the largest
         *                          integer  that is smaller than or equal to \a x.
         *  @since 1.0
         */
        static Real32               floor(Real32 x);

        /**
         *  Returns a floating-point value representing the largest integer 
         *  that is smaller than or equal to an argument.
         *
         *  @param x                An argument.
         *  @return                 A floating-point value representing the largest
         *                          integer  that is smaller than or equal to \a x.
         *  @since 1.0
         */
        static Real64               floor(Real64 x);

        /**
         *  Returns a natuiral logarithm of an argument.
         *
         *  An argument must not be negative. If it is zero,the return
         *  value is infinity.
         *
         *  @param x                An argument.
         *  @return                 ln(\a x).
         *  @since 1.0
         */
        static Real32               log(Real32 x);

        /**
         *  Returns a natuiral logarithm of an argument.
         *
         *  An argument must not be negative. If it is zero,the return
         *  value is infinity.
         *
         *  @param x                An argument.
         *  @return                 ln(\a x).
         *  @since 1.0
         */
        static Real64               log(Real64 x);

        /**
         *  Returns a decimal logarithm of an argument.
         *
         *  An argument must not be negative. If it is zero,the return
         *  value is infinity.
         *
         *  @param x                An argument.
         *  @return                 lg(\a x).
         *  @since 1.0
         */
        static Real32               log10(Real32 x);

        /**
         *  Returns a decimal logarithm of an argument.
         *
         *  An argument must not be negative. If it is zero,the return
         *  value is infinity.
         *
         *  @param x                An argument.
         *  @return                 lg(\a x).
         *  @since 1.0
         */
        static Real64               log10(Real64 x);

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Int8                 max(Int8 x, Int8 y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Int16                max(Int16 x, Int16 y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Int32                max(Int32 x, Int32 y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Int64                max(Int64 x, Int64 y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static UInt8                max(UInt8 x, UInt8 y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static UInt16               max(UInt16 x, UInt16 y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static UInt32               max(UInt32 x, UInt32 y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static UInt64               max(UInt64 x, UInt64 y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Real32               max(Real32 x, Real32 y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Real64               max(Real64 x, Real64 y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static BinaryInteger        max(const BinaryInteger & x, const BinaryInteger & y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static BinaryFixed          max(const BinaryFixed & x, const BinaryFixed & y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static BinaryFloat          max(const BinaryFloat & x, const BinaryFloat & y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static DecimalInteger       max(const DecimalInteger & x, const DecimalInteger & y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static DecimalFixed         max(const DecimalFixed & x, const DecimalFixed & y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the larger of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x >= \a y; otherwise \a y.
         *  @since 1.0
         */
        static DecimalFloat         max(const DecimalFloat & x, const DecimalFloat & y) { return (x > y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Int8                 min(Int8 x, Int8 y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Int16                min(Int16 x, Int16 y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Int32                min(Int32 x, Int32 y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Int64                min(Int64 x, Int64 y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static UInt8                min(UInt8 x, UInt8 y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static UInt16               min(UInt16 x, UInt16 y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static UInt32               min(UInt32 x, UInt32 y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static UInt64               min(UInt64 x, UInt64 y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Real32               min(Real32 x, Real32 y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static Real64               min(Real64 x, Real64 y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static BinaryInteger        min(const BinaryInteger & x, const BinaryInteger & y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static BinaryFixed          min(const BinaryFixed & x, const BinaryFixed & y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static BinaryFloat          min(const BinaryFloat & x, const BinaryFloat & y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static DecimalInteger       min(const DecimalInteger & x, const DecimalInteger & y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static DecimalFixed         min(const DecimalFixed & x, const DecimalFixed & y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns the smaller of two values.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x iff \a x <= \a y; otherwise \a y.
         *  @since 1.0
         */
        static DecimalFloat         min(const DecimalFloat & x, const DecimalFloat & y) { return (x < y) ? (x) : (y); }

        /**
         *  Returns \a x raised to the power of \a y.
         *
         *  If \a x = 0 and \a y < 0, the returned value is an infinity.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x raised to the power of \a y.
         *  @since 1.0
         */
        static Real32               pow(Real32 x, Real32 y);

        /**
         *  Returns \a x raised to the power of \a y.
         *
         *  If \a x = 0 and \a y < 0, the returned value is an infinity.
         *
         *  @param x                The 1st argument.
         *  @param y                The 2nd argument.
         *  @return                 \a x raised to the power of \a y.
         *  @since 1.0
         */
        static Real64               pow(Real64 x, Real64 y);

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(Int8 x) { return (x < 0) ? (-1) : ((x > 0) ? (1) : 0); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(Int16 x) { return (x < 0) ? (-1) : ((x > 0) ? (1) : 0); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(Int32 x) { return (x < 0) ? (-1) : ((x > 0) ? (1) : 0); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(Int64 x) { return (x < 0) ? (-1) : ((x > 0) ? (1) : 0); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(Real32 x) { return (x < 0) ? (-1) : ((x > 0) ? (1) : 0); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(Real64 x) { return (x < 0) ? (-1) : ((x > 0) ? (1) : 0); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(const BinaryInteger & x) { return x.getSign(); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(const BinaryFixed & x) { return x.getSign(); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(const BinaryFloat & x) { return x.getSign(); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(const DecimalInteger & x) { return x.getSign(); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(const DecimalFixed & x) { return x.getSign(); }

        /**
         *  Returns the sign of an argument.
         *
         *  @param x                An argument.
         *  @return
         *                          -   0 of \a x = 0.
         *                          -   1 of \a x > 0.
         *                          -   -1 of \a x < 0.
         *  @since 1.0
         */
        static int                  sign(const DecimalFloat & x) { return x.getSign(); }

        /**
         *  Returns a sine of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 sin(\a x).
         *  @since 1.0
         */
        static Real32               sin(Real32 x);

        /**
         *  Returns a sine of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 sin(\a x).
         *  @since 1.0
         */
        static Real64               sin(Real64 x);

        /**
         *  Returns a hyperbolic sine of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 sinh(\a x).
         *  @since 1.0
         */
        static Real32               sinh(Real32 x);

        /**
         *  Returns a hyperbolic sine of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 sinh(\a x).
         *  @since 1.0
         */
        static Real64               sinh(Real64 x);

        /**
         *  Returns a square root of the specified argument.
         *
         *  An argument must not be negative.
         *
         *  @param x                An argument.
         *  @return                 sqrt(\a x).
         *  @since 1.0
         */
        static Real32               sqrt(Real32 x);

        /**
         *  Returns a square root of the specified argument.
         *
         *  An argument must not be negative.
         *
         *  @param x                An argument.
         *  @return                 sqrt(\a x).
         *  @since 1.0
         */
        static Real64               sqrt(Real64 x);

        /**
         *  Returns a tangent of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 tg(\a x).
         *  @since 1.0
         */
        static Real32               tan(Real32 x);

        /**
         *  Returns a tangent of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 tg(\a x).
         *  @since 1.0
         */
        static Real64               tan(Real64 x);

        /**
         *  Returns a hyperbolic tangent of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 th(\a x).
         *  @since 1.0
         */
        static Real32               tanh(Real32 x);

        /**
         *  Returns a hyperbolic tangent of the specified argument.
         *
         *  @param x                An argument.
         *  @return                 th(\a x).
         *  @since 1.0
         */
        static Real64               tanh(Real64 x);

        /**
         *  Checks if a floating-point value is a NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isNan(Real32 x);

        /**
         *  Checks if a floating-point value is a NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isNan(Real64 x);

        /**
         *  Checks if a floating-point value is a NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isNan(const BinaryFloat & x);

        /**
         *  Checks if a floating-point value is a NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isNan(const DecimalFloat & x);

        /**
         *  Checks if a floating-point value is a signaling NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a signaling NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isSignalingNan(Real32 x);

        /**
         *  Checks if a floating-point value is a signaling NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a signaling NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isSignalingNan(Real64 x);

        /**
         *  Checks if a floating-point value is a signaling NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a signaling NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isSignalingNan(const BinaryFloat & x);

        /**
         *  Checks if a floating-point value is a signaling NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a signaling NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isSignalingNan(const DecimalFloat & x);

        /**
         *  Checks if a floating-point value is a quiet NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a quiet NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isQuietNan(Real32 x);

        /**
         *  Checks if a floating-point value is a quiet NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a quiet NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isQuietNan(Real64 x);

        /**
         *  Checks if a floating-point value is a quiet NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a quiet NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isQuietNan(const BinaryFloat & x);

        /**
         *  Checks if a floating-point value is a quiet NaN.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a quiet NaN, false otherwise.
         *  @since 1.0
         */
        static bool                 isQuietNan(const DecimalFloat & x);

        /**
         *  Checks if a floating-point value is an infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is an infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isInfinite(Real32 x);

        /**
         *  Checks if a floating-point value is an infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is an infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isInfinite(Real64 x);

        /**
         *  Checks if a floating-point value is an infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is an infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isInfinite(const BinaryFloat & x);

        /**
         *  Checks if a floating-point value is an infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is an infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isInfinite(const DecimalFloat & x);

        /**
         *  Checks if a floating-point value is a positive infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a positive infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isPositiveInfinite(Real32 x);

        /**
         *  Checks if a floating-point value is a positive infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a positive infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isPositiveInfinite(Real64 x);

        /**
         *  Checks if a floating-point value is a positive infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a positive infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isPositiveInfinite(const BinaryFloat & x);

        /**
         *  Checks if a floating-point value is a positive infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a positive infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isPositiveInfinite(const DecimalFloat & x);

        /**
         *  Checks if a floating-point value is a negative infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a negative  infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isNegativeInfinite(Real32 x);

        /**
         *  Checks if a floating-point value is a negative infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a negative  infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isNegativeInfinite(Real64 x);

        /**
         *  Checks if a floating-point value is a negative infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a negative  infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isNegativeInfinite(const BinaryFloat & x);

        /**
         *  Checks if a floating-point value is a negative infinity.
         *
         *  @param x                The value to check.
         *  @return                 True if \a x is a negative  infinity, false otherwise.
         *  @since 1.0
         */
        static bool                 isNegativeInfinite(const DecimalFloat & x);
    };

    //////////
    //  Additional BinaryInteger operators
    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator + (Int8 op1, const BinaryInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator + (Int16 op1, const BinaryInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator + (Int32 op1, const BinaryInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator + (Int64 op1, const BinaryInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator + (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator + (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator + (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator + (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator - (Int8 op1, const BinaryInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator - (Int16 op1, const BinaryInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator - (Int32 op1, const BinaryInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator - (Int64 op1, const BinaryInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator - (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator - (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator - (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator - (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator * (Int8 op1, const BinaryInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator * (Int16 op1, const BinaryInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator * (Int32 op1, const BinaryInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator * (Int64 op1, const BinaryInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator * (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator * (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator * (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator * (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator / (Int8 op1, const BinaryInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator / (Int16 op1, const BinaryInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator / (Int32 op1, const BinaryInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator / (Int64 op1, const BinaryInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator / (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator / (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator / (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator / (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator % (Int8 op1, const BinaryInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator % (Int16 op1, const BinaryInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator % (Int32 op1, const BinaryInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator % (Int64 op1, const BinaryInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator % (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator % (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator % (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator % (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator & (Int8 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator & (Int16 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator & (Int32 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator & (Int64 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator & (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator & (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator & (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator & (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator | (Int8 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator | (Int16 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator | (Int32 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator | (Int64 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator | (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator | (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator | (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator | (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator ^ (Int8 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator ^ (Int16 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator ^ (Int32 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator ^ (Int64 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator ^ (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator ^ (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator ^ (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryInteger CARL_API operator ^ (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int64 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt8 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt16 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt32 op1, const BinaryInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt64 op1, const BinaryInteger & op2);

    //////////
    //  Additional BinaryFixed operators
    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator + (Int8 op1, const BinaryFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator + (Int16 op1, const BinaryFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator + (Int32 op1, const BinaryFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator + (Int64 op1, const BinaryFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator + (UInt8 op1, const BinaryFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator + (UInt16 op1, const BinaryFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator + (UInt32 op1, const BinaryFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator + (UInt64 op1, const BinaryFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator + (Real32 op1, const BinaryFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator + (Real64 op1, const BinaryFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator - (Int8 op1, const BinaryFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator - (Int16 op1, const BinaryFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator - (Int32 op1, const BinaryFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator - (Int64 op1, const BinaryFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator - (UInt8 op1, const BinaryFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator - (UInt16 op1, const BinaryFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator - (UInt32 op1, const BinaryFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator - (UInt64 op1, const BinaryFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator - (Real32 op1, const BinaryFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator - (Real64 op1, const BinaryFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator * (Int8 op1, const BinaryFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator * (Int16 op1, const BinaryFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator * (Int32 op1, const BinaryFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator * (Int64 op1, const BinaryFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator * (UInt8 op1, const BinaryFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator * (UInt16 op1, const BinaryFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator * (UInt32 op1, const BinaryFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator * (UInt64 op1, const BinaryFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator * (Real32 op1, const BinaryFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator * (Real64 op1, const BinaryFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator / (Int8 op1, const BinaryFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator / (Int16 op1, const BinaryFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator / (Int32 op1, const BinaryFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator / (Int64 op1, const BinaryFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator / (UInt8 op1, const BinaryFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator / (UInt16 op1, const BinaryFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator / (UInt32 op1, const BinaryFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator / (UInt64 op1, const BinaryFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator / (Real32 op1, const BinaryFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFixed CARL_API operator / (Real64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Real32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Real64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Real32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Real64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Real32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Real64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Real32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Real64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Real32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Real64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt8 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt16 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt64 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Real32 op1, const BinaryFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Real64 op1, const BinaryFixed & op2);

    //////////
    //  Additional BinaryFloat operators
    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator + (Int8 op1, const BinaryFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator + (Int16 op1, const BinaryFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator + (Int32 op1, const BinaryFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator + (Int64 op1, const BinaryFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator + (UInt8 op1, const BinaryFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator + (UInt16 op1, const BinaryFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator + (UInt32 op1, const BinaryFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator + (UInt64 op1, const BinaryFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator + (Real32 op1, const BinaryFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator + (Real64 op1, const BinaryFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator - (Int8 op1, const BinaryFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator - (Int16 op1, const BinaryFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator - (Int32 op1, const BinaryFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator - (Int64 op1, const BinaryFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator - (UInt8 op1, const BinaryFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator - (UInt16 op1, const BinaryFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator - (UInt32 op1, const BinaryFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator - (UInt64 op1, const BinaryFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator - (Real32 op1, const BinaryFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator - (Real64 op1, const BinaryFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator * (Int8 op1, const BinaryFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator * (Int16 op1, const BinaryFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator * (Int32 op1, const BinaryFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator * (Int64 op1, const BinaryFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator * (UInt8 op1, const BinaryFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator * (UInt16 op1, const BinaryFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator * (UInt32 op1, const BinaryFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator * (UInt64 op1, const BinaryFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator * (Real32 op1, const BinaryFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator * (Real64 op1, const BinaryFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator / (Int8 op1, const BinaryFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator / (Int16 op1, const BinaryFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator / (Int32 op1, const BinaryFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator / (Int64 op1, const BinaryFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator / (UInt8 op1, const BinaryFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator / (UInt16 op1, const BinaryFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator / (UInt32 op1, const BinaryFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator / (UInt64 op1, const BinaryFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator / (Real32 op1, const BinaryFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC BinaryFloat CARL_API operator / (Real64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Real32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Real64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Real32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Real64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Real32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Real64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Real32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Real64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Real32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Real64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt8 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt16 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt64 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Real32 op1, const BinaryFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Real64 op1, const BinaryFloat & op2);

    //////////
    //  Additional DecimalInteger operators
    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator + (Int8 op1, const DecimalInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator + (Int16 op1, const DecimalInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator + (Int32 op1, const DecimalInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator + (Int64 op1, const DecimalInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator + (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator + (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator + (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator + (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator - (Int8 op1, const DecimalInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator - (Int16 op1, const DecimalInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator - (Int32 op1, const DecimalInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator - (Int64 op1, const DecimalInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator - (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator - (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator - (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator - (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator * (Int8 op1, const DecimalInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator * (Int16 op1, const DecimalInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator * (Int32 op1, const DecimalInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator * (Int64 op1, const DecimalInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator * (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator * (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator * (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator * (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator / (Int8 op1, const DecimalInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator / (Int16 op1, const DecimalInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator / (Int32 op1, const DecimalInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator / (Int64 op1, const DecimalInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator / (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator / (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator / (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator / (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator % (Int8 op1, const DecimalInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator % (Int16 op1, const DecimalInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator % (Int32 op1, const DecimalInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator % (Int64 op1, const DecimalInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator % (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator % (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator % (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Returns the remainder from division of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The remainder from division of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator % (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator & (Int8 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator & (Int16 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator & (Int32 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator & (Int64 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator & (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator & (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator & (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical conjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical conjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator & (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator | (Int8 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator | (Int16 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator | (Int32 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator | (Int64 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator | (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator | (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator | (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator | (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator ^ (Int8 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator ^ (Int16 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator ^ (Int32 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator ^ (Int64 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator ^ (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator ^ (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator ^ (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Returns the bitwise logical exclusive disjunction of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The bitwise logical exclusive disjunction of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalInteger CARL_API operator ^ (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int64 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt8 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt16 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt32 op1, const DecimalInteger & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt64 op1, const DecimalInteger & op2);

    //////////
    //  Additional DecimalFixed operators
    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator + (Int8 op1, const DecimalFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator + (Int16 op1, const DecimalFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator + (Int32 op1, const DecimalFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator + (Int64 op1, const DecimalFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator + (UInt8 op1, const DecimalFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator + (UInt16 op1, const DecimalFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator + (UInt32 op1, const DecimalFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator + (UInt64 op1, const DecimalFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator + (Real32 op1, const DecimalFixed & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator + (Real64 op1, const DecimalFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator - (Int8 op1, const DecimalFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator - (Int16 op1, const DecimalFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator - (Int32 op1, const DecimalFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator - (Int64 op1, const DecimalFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator - (UInt8 op1, const DecimalFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator - (UInt16 op1, const DecimalFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator - (UInt32 op1, const DecimalFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator - (UInt64 op1, const DecimalFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator - (Real32 op1, const DecimalFixed & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator - (Real64 op1, const DecimalFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator * (Int8 op1, const DecimalFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator * (Int16 op1, const DecimalFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator * (Int32 op1, const DecimalFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator * (Int64 op1, const DecimalFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator * (UInt8 op1, const DecimalFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator * (UInt16 op1, const DecimalFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator * (UInt32 op1, const DecimalFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator * (UInt64 op1, const DecimalFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator * (Real32 op1, const DecimalFixed & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator * (Real64 op1, const DecimalFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator / (Int8 op1, const DecimalFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator / (Int16 op1, const DecimalFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator / (Int32 op1, const DecimalFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator / (Int64 op1, const DecimalFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator / (UInt8 op1, const DecimalFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator / (UInt16 op1, const DecimalFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator / (UInt32 op1, const DecimalFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator / (UInt64 op1, const DecimalFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator / (Real32 op1, const DecimalFixed & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFixed CARL_API operator / (Real64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Real32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Real64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Real32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Real64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Real32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Real64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Real32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Real64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Real32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Real64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt8 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt16 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt64 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Real32 op1, const DecimalFixed & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Real64 op1, const DecimalFixed & op2);

    //////////
    //  Additional DecimalFloat operators
    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator + (Int8 op1, const DecimalFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator + (Int16 op1, const DecimalFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator + (Int32 op1, const DecimalFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator + (Int64 op1, const DecimalFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator + (UInt8 op1, const DecimalFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator + (UInt16 op1, const DecimalFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator + (UInt32 op1, const DecimalFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator + (UInt64 op1, const DecimalFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator + (Real32 op1, const DecimalFloat & op2);

    /**
     *  Returns the sum of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The sum of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator + (Real64 op1, const DecimalFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator - (Int8 op1, const DecimalFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator - (Int16 op1, const DecimalFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator - (Int32 op1, const DecimalFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator - (Int64 op1, const DecimalFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator - (UInt8 op1, const DecimalFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator - (UInt16 op1, const DecimalFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator - (UInt32 op1, const DecimalFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator - (UInt64 op1, const DecimalFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator - (Real32 op1, const DecimalFloat & op2);

    /**
     *  Returns the difference of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The difference of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator - (Real64 op1, const DecimalFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator * (Int8 op1, const DecimalFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator * (Int16 op1, const DecimalFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator * (Int32 op1, const DecimalFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator * (Int64 op1, const DecimalFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator * (UInt8 op1, const DecimalFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator * (UInt16 op1, const DecimalFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator * (UInt32 op1, const DecimalFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator * (UInt64 op1, const DecimalFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator * (Real32 op1, const DecimalFloat & op2);

    /**
     *  Returns the product of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The product of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator * (Real64 op1, const DecimalFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator / (Int8 op1, const DecimalFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator / (Int16 op1, const DecimalFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator / (Int32 op1, const DecimalFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator / (Int64 op1, const DecimalFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator / (UInt8 op1, const DecimalFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator / (UInt16 op1, const DecimalFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator / (UInt32 op1, const DecimalFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator / (UInt64 op1, const DecimalFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator / (Real32 op1, const DecimalFloat & op2);

    /**
     *  Returns the quotient of two operands.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     The quotient of two operands.
     *  @since 1.0
     */
    CARL_PUBLIC DecimalFloat CARL_API operator / (Real64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Int64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (UInt64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Real32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are equal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 == op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator == (Real64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Int64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (UInt64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Real32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the two operands are unequal.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 != op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator != (Real64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Int64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (UInt64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Real32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 < op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator < (Real64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Int64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (UInt64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Real32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is smaller than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 <= op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator <= (Real64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Int64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (UInt64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Real32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator > (Real64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Int64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt8 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt16 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (UInt64 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Real32 op1, const DecimalFloat & op2);

    /**
     *  Checks if the 1st operand is greater than or equal to the 2nd operand.
     *
     *  @param op1                  The 1st operand.
     *  @param op2                  The 2nd operand.
     *  @return                     True iff op1 > op2; false otherwise.
     *  @since 1.0
     */
    CARL_PUBLIC bool CARL_API operator >= (Real64 op1, const DecimalFloat & op2);

    //////////
    //  Hashing BinaryInteger values
    namespace StandardHashFunctions
    {
        /**
         *  Calculates a hash code for a BinaryInteger instance.
         *
         *  @param key              The BinaryInteger instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashBinaryInteger(const BinaryInteger & key);

        /**
         *  Calculates a hash code for a BinaryFixed instance.
         *
         *  @param key              The BinaryFixed instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashBinaryFixed(const BinaryFixed & key);

        /**
         *  Calculates a hash code for a BinaryFloat instance.
         *
         *  @param key              The BinaryFloat instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashBinaryFloat(const BinaryFloat & key);

        /**
         *  Calculates a hash code for a DecimalInteger instance.
         *
         *  @param key              The DecimalInteger instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashDecimalInteger(const DecimalInteger & key);

        /**
         *  Calculates a hash code for a DecimalFixed instance.
         *
         *  @param key              The DecimalFixed instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashDecimalFixed(const DecimalFixed & key);

        /**
         *  Calculates a hash code for a DecimalFloat instance.
         *
         *  @param key              The DecimalFixed instance to hash.
         *  @return                 The hash code.
         *  @since 1.0
         */
        CARL_PUBLIC UIntSize CARL_CALLBACK hashDecimalFloat(const DecimalFloat & key);
    }
}

#endif  //  __CARL_Math_hpp_included__
//  End of CARL_Math.hpp
