//
//  CARL_Parse.hpp - Value parsing
//
//  Copyright (C) 2006-2008, Cybernetic Intelligence GmbH
//
//  $Date: 2008/02/22 11:54:48 $
//  $Revision: 1.20 $
//  $Author: akapusti $
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//////////
/**
 *  @file
 *
 *  Defines CARL string parsing framework.
 *
 *  Do not include this file directly. Use the master header CARL.hpp instead.
 *
 *  @since 1.0
 */
#ifndef __CARL_hpp_included__
    #error The file CARL_Parse.hpp shall not be included directly. Use CARL.hpp instead
#endif  //  __CARL_hpp_included__

#ifndef __CARL_Parse_hpp_included__
#define __CARL_Parse_hpp_included__

namespace carl
{
    namespace _impl
    {
        //////////
        //  Parse managers
        class CARL_PUBLIC _NumberParseManager;
    }

    /**
     *  @class IntegerParseProperties CARL.hpp
     *  Format properties governing parsing of integer values.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC IntegerParseProperties
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  Iniitializes a newly created integer format properties to default
         *  values:
         *
         *  -   Radix = 10.
         *  -   No minimum or maximum width limits.
         *  -   Do not allow prefix or thousand separator.
         *
         *  @since 1.0
         */
        IntegerParseProperties() { _setStandardProperties(); }

        /**
         *  The class constructor.
         *
         *  Initializes a newly created integer parse properties instance to
         *  specify the same parsing rules as would be effected by the
         *  specified string in CRT scanf() family of functions.
         *
         *  @param pCrtFormatString A pointer to nul-terminated ASCII string
         *                          representing the CRT scanf()-style
         *                          integer format specification.
         *  @since 1.0
         */
        explicit IntegerParseProperties(const char * pCrtFormatString) { _parseCrtFormatString(String(pCrtFormatString)); }

        /**
         *  The class constructor.
         *
         *  Initializes a newly created integer parse properties instance to
         *  specify the same parsing rules as would be effected by the
         *  specified string in CRT scanf() family of functions.
         *
         *  @param pCrtFormatString A pointer to nul-terminated ISO-10646 string
         *                          representing the CRT scanf()-style
         *                          integer format specification.
         *  @since 1.0
         */
        explicit IntegerParseProperties(const Char * pCrtFormatString) { _parseCrtFormatString(String(pCrtFormatString)); }

        /**
         *  The class constructor.
         *
         *  Initializes a newly created integer parse properties instance to
         *  specify the same parsing rules as would be effected by the
         *  specified string in CRT scanf() family of functions.
         *
         *  @param crtFormatString  A string representing the CRT scanf()-style
         *                          integer format specification.
         *  @since 1.0
         */
        explicit IntegerParseProperties(const String & crtFormatString) { _parseCrtFormatString(crtFormatString); }

        /**
         *  The copy constructor.
         *
         *  The newly created integer parse properties instance is initialized to 
         *  be an exact copy of an existing integer parse properties.
         *
         *  @param src              The integer parse properties name to copy 
         *                          properties from.
         *  @since 1.0
         */
        IntegerParseProperties(const IntegerParseProperties & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~IntegerParseProperties() {}

        /**
         *  The assigment operator.
         *
         *  Makes an integer parse properties an exact copy of an existing 
         *  integer parse properties.
         *
         *  @param src              The integer parse properties to copy properties from.
         *  @return                 A reference to an assigned-to object.
         *  @since 1.0
         */
        IntegerParseProperties &    operator = (const IntegerParseProperties & src);

        //////////
        //  Operations
    public:
        /**
         *  Changes the radix of the integer format specification.
         *
         *  The new radix must be in range 2..36.
         *
         *  @param radix            The new radix.
         *  @return                 A reference to the IntegerParseProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerParseProperties &    setRadix(UIntSize radix);

        /**
         *  Changes the minimum width of an integer value being parsed.
         *
         *  At least the specified number of characters will be processed
         *  when parsing an integer value. If the minimum width is 0, then
         *  there is no minimum limit on the width of an integer value
         *  being parsed.
         *
         *  @param width            The new minimum width.
         *  @return                 A reference to the RealParseProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerParseProperties &    setMinimumWidth(UIntSize width);

        /**
         *  Changes the maximum width of an integer value being parsed.
         *
         *  At most the specified number of characters will be processed
         *  when parsing an integer value. If the maximum width is 0, then
         *  there is no maximum limit on the width of an integer value
         *  being parsed.
         *
         *  @param width            The new maximum width.
         *  @return                 A reference to the IntegerFormatProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerParseProperties &    setMaximumWidth(UIntSize width);

        /**
         *  Allows or disallows parsed values to have radix-specific prefixes.
         *
         *  When prefixes are enabled:
         *
         *  -   All values parsed with radix 8 can be prefixed with '0'.
         *  -   All values parsed with radix 16 can be prefixed with '0x' or '0X'.
         *
         *  @param allowPrefix      Allows (true) or disallows (false) the presence
         *                          of prefixes in the parsed value.
         *  @return                 A reference to the IntegerParseProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerParseProperties &    setAllowPrefix(bool allowPrefix);

        /**
         *  Allows or disallows thousand separators within parsed values.
         *
         *  When thousand separators are allowed, they are ignored by the parser.
         *
         *  @param allowSeparator   Allows (true) or disallows (false) the presence
         *                          of thousand separators in the parsed value.
         *  @return                 A reference to the IntegerParseProperties
         *                          instance being modified.
         *  @since 1.0
         */
        IntegerParseProperties &    setAllowSeparator(bool allowSeparator);

        /**
         *  Returns the radix of the integer format specification.
         *
         *  @return                 The radix of the integer format specification.
         *  @since 1.0
         */
        UIntSize                    getRadix() const { return _radix; }

        /**
         *  Returns the minimum width of an integer value being parsed.
         *
         *  At least the specified number of characters will be processed
         *  when parsing an integer value. If the minimum width is 0, then
         *  there is no minimum limit on the width of an integer value
         *  being parsed.
         *
         *  @return                 The minimum width of an integer value being parsed.
         *  @since 1.0
         */
        UIntSize                    getMinimumWidth() const { return _minimumWidth; }

        /**
         *  Returns the maximum width of an integer value being parsed.
         *
         *  At most the specified number of characters will be processed
         *  when parsing an integer value. If the maximum width is 0, then
         *  there is no maximum limit on the width of an integer value
         *  being parsed.
         *
         *  @return                 The maximum width of an integer value being parsed.
         *  @since 1.0
         */
        UIntSize                    getMaximumWidth() const { return _maximumWidth; }

        /**
         *  Checks if parsed values can have radix-specific prefixes.
         *
         *  When prefixes are enabled:
         *
         *  -   All values parsed with radix 8 can be prefixed with '0'.
         *  -   All values parsed with radix 16 can be prefixed with '0x' or '0X'.
         *
         *  @return                 True if parsed values can have radix-specific
         *                          prefixes; false otherwise.
         *  @since 1.0
         */
        bool                        getAllowPrefix() const { return _allowPrefix; }

        /**
         *  Checks if thousand separators are allowed within parsed values.
         *
         *  When thousand separators are allowed, they are ignored by the parser.
         *
         *  @return                 True if thousand separators are allowed within
         *                          parsed values; false otherwise.
         *  @since 1.0
         */
        bool                        getAllowSeparator() const { return _allowSeparator; }

        //////////
        //  Implementation
    private:
        UIntSize                    _radix;         //  2..36
        UIntSize                    _minimumWidth;  //  0 == unspecified
        UIntSize                    _maximumWidth;  //  0 == unspecified
        bool                        _allowPrefix;
        bool                        _allowSeparator;

        //  Helpers
        void                        _setStandardProperties();
        void                        _parseCrtFormatString(const String & crtFormatString);
    };

    /**
     *  @class RealParseProperties CARL.hpp
     *  Format properties governing parsing of real values.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC RealParseProperties
    {
        //////////
        //  Construction/destruction/assignment
    public:
        /**
         *  The default constructor.
         *
         *  Iniitializes a newly created ream format properties to default
         *  values:
         *
         *  -   No minimum or maximum width limits.
         *  -   Do not allow thousand separator.
         *
         *  @since 1.0
         */
        RealParseProperties() { _setStandardProperties(); }

        /**
         *  The class constructor.
         *
         *  Initializes a newly created real parse properties instance to
         *  specify the same parsing rules as would be effected by the
         *  specified string in CRT scanf() family of functions.
         *
         *  @param pCrtFormatString A pointer to nul-terminated ASCII string
         *                          representing the CRT scanf()-style
         *                          real format specification.
         *  @since 1.0
         */
        explicit RealParseProperties(const char * pCrtFormatString) { _parseCrtFormatString(String(pCrtFormatString)); }

        /**
         *  The class constructor.
         *
         *  Initializes a newly created real parse properties instance to
         *  specify the same parsing rules as would be effected by the
         *  specified string in CRT scanf() family of functions.
         *
         *  @param pCrtFormatString A pointer to nul-terminated ISO-10646 string
         *                          representing the CRT scanf()-style
         *                          real format specification.
         *  @since 1.0
         */
        explicit RealParseProperties(const Char * pCrtFormatString) { _parseCrtFormatString(String(pCrtFormatString)); }

        /**
         *  The class constructor.
         *
         *  Initializes a newly created real parse properties instance to
         *  specify the same parsing rules as would be effected by the
         *  specified string in CRT scanf() family of functions.
         *
         *  @param crtFormatString  A string representing the CRT scanf()-style
         *                          real format specification.
         *  @since 1.0
         */
        explicit RealParseProperties(const String & crtFormatString) { _parseCrtFormatString(crtFormatString); }

        /**
         *  The copy constructor.
         *
         *  The newly created real parse properties instance is initialized to 
         *  be an exact copy of an existing real parse properties.
         *
         *  @param src              The real parse properties name to copy 
         *                          properties from.
         *  @since 1.0
         */
        RealParseProperties(const RealParseProperties & src);

        /**
         *  The class destructor.
         *
         *  @since 1.0
         */
        ~RealParseProperties() {}

        /**
         *  The assigment operator.
         *
         *  Makes an real parse properties an exact copy of an existing 
         *  real parse properties.
         *
         *  @param src              The real parse properties to copy properties from.
         *  @return                 A reference to an assigned-to object.
         *  @since 1.0
         */
        RealParseProperties &       operator = (const RealParseProperties & src);

        //////////
        //  Operations
    public:
        /**
         *  Changes the radix of the real format specification.
         *
         *  The new radix must be in range 2..36.
         *
         *  @param radix            The new radix.
         *  @return                 A reference to the RealParseProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealParseProperties &       setRadix(UIntSize radix);

        /**
         *  Changes the minimum width of an real value being parsed.
         *
         *  At least the specified number of characters will be processed
         *  when parsing an real value. If the minimum width is 0, then
         *  there is no minimum limit on the width of an real value
         *  being parsed.
         *
         *  @param width            The new minimum width.
         *  @return                 A reference to the RealParseProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealParseProperties &       setMinimumWidth(UIntSize width);

        /**
         *  Changes the maximum width of an real value being parsed.
         *
         *  At most the specified number of characters will be processed
         *  when parsing an real value. If the maximum width is 0, then
         *  there is no maximum limit on the width of an real value
         *  being parsed.
         *
         *  @param width           The new maximum width.
         *  @return                 A reference to the RealParseProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealParseProperties &       setMaximumWidth(UIntSize width);

        /**
         *  Allows or disallows thousand separators within parsed values.
         *
         *  When thousand separators are allowed, they are ignored by the parser.
         *
         *  @param allowSeparator   Allows (true) or disallows (false) the presence
         *                          of thousand separators in the parsed value.
         *  @return                 A reference to the IntegerParseProperties
         *                          instance being modified.
         *  @since 1.0
         */
        RealParseProperties &       setAllowSeparator(bool allowSeparator);

        /**
         *  Returns the radix of the real format specification.
         *
         *  @return                 The radix of the real format specification.
         *  @since 1.0
         */
        UIntSize                    getRadix() const { return _radix; }

        /**
         *  Returns the minimum width of an real value being parsed.
         *
         *  At least the specified number of characters will be processed
         *  when parsing an real value. If the minimum width is 0, then
         *  there is no minimum limit on the width of an real value
         *  being parsed.
         *
         *  @return                 The minimum width of an real value being parsed.
         *  @since 1.0
         */
        UIntSize                    getMinimumWidth() const { return _minimumWidth; }

        /**
         *  Returns the maximum width of an real value being parsed.
         *
         *  At most the specified number of characters will be processed
         *  when parsing an real value. If the maximum width is 0, then
         *  there is no maximum limit on the width of an real value
         *  being parsed.
         *
         *  @return                 The maximum width of an real value being parsed.
         *  @since 1.0
         */
        UIntSize                    getMaximumWidth() const { return _maximumWidth; }

        /**
         *  Checks if thousand separators are allowed within parsed values.
         *
         *  When thousand separators are allowed, they are ignored by the parser.
         *
         *  @return                 True if thousand separators are allowed within
         *                          parsed values; false otherwise.
         *  @since 1.0
         */
        bool                        getAllowSeparator() const { return _allowSeparator; }

        //////////
        //  Implementation
    private:
        UIntSize                    _radix;         //  2..36
        UIntSize                    _minimumWidth;  //  0 == unspecified
        UIntSize                    _maximumWidth;  //  0 == unspecified
        bool                        _allowSeparator;

        //  Helpers
        void                        _setStandardProperties();
        void                        _parseCrtFormatString(const String & crtFormatString);
    };

    /**
     *  @class NumberParse CARL.hpp
     *  A generic parser for numeric values.
     *
     *  Parsing numeric values (i.e. converting them to string representation)
     *  is performed by locale-specific instances of concrete NumberParse-derived
     *  classes. A NumberParse subclass for a specific locale "knows" about rules
     *  for representing numbers that are specific to the appropriate locale. The
     *  constraints on the required representation of the parsed value can be varied 
     *  to a certain degree by specifying custom "parse properties" when parsing a value.
     *
     *  When obtaining a number parser for a specified locale, it may well
     *  be the case that there is no number parser registered for that locale.
     *  In this case the immediate parent locale is tried next, and the parent of 
     *  that if necessary, until either an invariant locale is reached or a number
     *  parser is found. In the former case, a default number parser is
     *  used. This behaviour allows specifying locale-generic number parsers
     *  (e.g. if an "German neutral" number parser is specified, then it will
     *  be used for both "German/Germany" and "German/Switzerland" locales provided
     *  no separate number parsers are registered for either of those).
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC NumberParse
    {
        friend class _impl::_NumberParseManager;
        //////////
        //  Can't assign or copy construct
    private:
        NumberParse(const NumberParse &);
        void                    operator = (const NumberParse &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        NumberParse() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~NumberParse() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified number parse with the CARL framework.
         *
         *  A registration of a number parse informs the CARL framework that
         *  a specific number parse exists. From this point on, the application
         *  can use the newly registers number parse in exactly the same manner
         *  as a predefined number parse. Once registered, a number parse
         *  cannot be unregistered.
         *
         *  A number parse is always registered for a specific locale. Any attempt
         *  to get a number parse instance for that locale will always yield the
         *  number parse registered for that locale. It is possible to register
         *  the same number parse for more than one locale.
         *
         *  @param forLocale        The locale for which the newly registered
         *                          number parse shall be used.
         *  @param pNumberParse     The number parse to register.
         *  @since 1.0
         */
        static void             registerInstance(const Locale & forLocale, NumberParse * pNumberParse);

        /**
         *  Returns a number parser for the current default locale.
         *
         *  @return                 A number parser for the current default locale.
         *  @since 1.0
         */
        static NumberParse *    getInstance();

        /**
         *  Returns a number parser for the specified locale.
         *
         *  @param forLocale        The locale for which a number parser is sought.
         *  @return                 A number parser for the specified locale.
         *  @since 1.0
         */
        static NumberParse *    getInstance(const Locale & forLocale);

        //  Default parsing
        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int8                    parseInt8(const char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt8(pString, position, IntegerParseProperties(), pOverflow);}

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int16                   parseInt16(const char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt16(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int32                   parseInt32(const char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt32(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int64                   parseInt64(const char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt64(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt8                   parseUInt8(const char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt8(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt16                  parseUInt16(const char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt16(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt32                  parseUInt32(const char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt32(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt64                  parseUInt64(const char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt64(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryInteger           parseBinaryInteger(const char * pString, UIntSize & position) const throw(ParseException) { return parseBinaryInteger(pString, position, IntegerParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalInteger          parseDecimalInteger(const char * pString, UIntSize & position) const throw(ParseException) { return parseDecimalInteger(pString, position, IntegerParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real32                  parseReal32(const char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseReal32(pString, position, RealParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real64                  parseReal64(const char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseReal64(pString, position, RealParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFixed             parseBinaryFixed(const char * pString, UIntSize & position) const throw(ParseException) { return parseBinaryFixed(pString, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFloat             parseBinaryFloat(const char * pString, UIntSize & position) const throw(ParseException) { return parseBinaryFloat(pString, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFixed            parseDecimalFixed(const char * pString, UIntSize & position) const throw(ParseException) { return parseDecimalFixed(pString, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFloat            parseDecimalFloat(const char * pString, UIntSize & position) const throw(ParseException) { return parseDecimalFloat(pString, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int8                    parseInt8(const Char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt8(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int16                   parseInt16(const Char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt16(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int32                   parseInt32(const Char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt32(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int64                   parseInt64(const Char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt64(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt8                   parseUInt8(const Char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt8(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt16                  parseUInt16(const Char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt16(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt32                  parseUInt32(const Char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt32(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt64                  parseUInt64(const Char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt64(pString, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryInteger           parseBinaryInteger(const Char * pString, UIntSize & position) const throw(ParseException) { return parseBinaryInteger(pString, position, IntegerParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalInteger          parseDecimalInteger(const Char * pString, UIntSize & position) const throw(ParseException) { return parseDecimalInteger(pString, position, IntegerParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real32                  parseReal32(const Char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseReal32(pString, position, RealParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real64                  parseReal64(const Char * pString, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseReal64(pString, position, RealParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFixed             parseBinaryFixed(const Char * pString, UIntSize & position) const throw(ParseException) { return parseBinaryFixed(pString, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFloat             parseBinaryFloat(const Char * pString, UIntSize & position) const throw(ParseException) { return parseBinaryFloat(pString, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFixed            parseDecimalFixed(const Char * pString, UIntSize & position) const throw(ParseException) { return parseDecimalFixed(pString, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFloat            parseDecimalFloat(const Char * pString, UIntSize & position) const throw(ParseException) { return parseDecimalFloat(pString, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int8                    parseInt8(const String & string, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt8(string, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int16                   parseInt16(const String & string, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt16(string, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int32                   parseInt32(const String & string, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt32(string, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int64                   parseInt64(const String & string, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseInt64(string, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt8                   parseUInt8(const String & string, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt8(string, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt16                  parseUInt16(const String & string, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt16(string, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt32                  parseUInt32(const String & string, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt32(string, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt64                  parseUInt64(const String & string, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseUInt64(string, position, IntegerParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryInteger           parseBinaryInteger(const String & string, UIntSize & position) const throw(ParseException) { return parseBinaryInteger(string, position, IntegerParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalInteger          parseDecimalInteger(const String & string, UIntSize & position) const throw(ParseException) { return parseDecimalInteger(string, position, IntegerParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real32                  parseReal32(const String & string, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseReal32(string, position, RealParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real64                  parseReal64(const String & string, UIntSize & position, bool * pOverflow = NULL) const throw(ParseException) { return parseReal64(string, position, RealParseProperties(), pOverflow); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFixed             parseBinaryFixed(const String & string, UIntSize & position) const throw(ParseException) { return parseBinaryFixed(string, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFloat             parseBinaryFloat(const String & string, UIntSize & position) const throw(ParseException) { return parseBinaryFloat(string, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFixed            parseDecimalFixed(const String & string, UIntSize & position) const throw(ParseException) { return parseDecimalFixed(string, position, RealParseProperties()); }

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFloat            parseDecimalFloat(const String & string, UIntSize & position) const throw(ParseException) { return parseDecimalFloat(string, position, RealParseProperties()); }

        //  Custom parsing
        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int8            parseInt8(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int16           parseInt16(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int32           parseInt32(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int64           parseInt64(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt8           parseUInt8(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt16          parseUInt16(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt32          parseUInt32(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt64          parseUInt64(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual BinaryInteger   parseBinaryInteger(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DecimalInteger  parseDecimalInteger(const char * pString, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Real32          parseReal32(const char * pString, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Real64          parseReal64(const char * pString, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual BinaryFixed     parseBinaryFixed(const char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual BinaryFloat     parseBinaryFloat(const char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DecimalFixed    parseDecimalFixed(const char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DecimalFloat    parseDecimalFloat(const char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int8            parseInt8(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int16           parseInt16(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int32           parseInt32(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int64           parseInt64(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt8           parseUInt8(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt16          parseUInt16(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt32          parseUInt32(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt64          parseUInt64(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual BinaryInteger   parseBinaryInteger(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DecimalInteger  parseDecimalInteger(const Char * pString, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Real32          parseReal32(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Real64          parseReal64(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual BinaryFixed     parseBinaryFixed(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual BinaryFloat     parseBinaryFloat(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DecimalFixed    parseDecimalFixed(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DecimalFloat    parseDecimalFloat(const Char * pString, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int8            parseInt8(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int16           parseInt16(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int32           parseInt32(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Int64           parseInt64(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt8           parseUInt8(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt16          parseUInt16(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt32          parseUInt32(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual UInt64          parseUInt64(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual BinaryInteger   parseBinaryInteger(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DecimalInteger  parseDecimalInteger(const String & string, UIntSize & position, const IntegerParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Real32          parseReal32(const String & string, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Real64          parseReal64(const String & string, UIntSize & position, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual BinaryFixed     parseBinaryFixed(const String & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual BinaryFloat     parseBinaryFloat(const String & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DecimalFixed    parseDecimalFixed(const String & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        /**
         *  Parses a value from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DecimalFloat    parseDecimalFloat(const String & string, UIntSize & position, const RealParseProperties & parseProperties) const throw(ParseException) = 0;

        //  Helpers for default parsing of full strings
        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int8                    parseInt8(const char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int16                   parseInt16(const char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int32                   parseInt32(const char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int64                   parseInt64(const char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt8                   parseUInt8(const char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt16                  parseUInt16(const char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt32                  parseUInt32(const char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt64                  parseUInt64(const char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryInteger           parseBinaryInteger(const char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalInteger          parseDecimalInteger(const char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real32                  parseReal32(const char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real64                  parseReal64(const char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFixed             parseBinaryFixed(const char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFloat             parseBinaryFloat(const char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFixed            parseDecimalFixed(const char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFloat          parseDecimalFloat(const char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int8                    parseInt8(const Char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int16                   parseInt16(const Char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int32                   parseInt32(const Char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int64                   parseInt64(const Char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt8                   parseUInt8(const Char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt16                  parseUInt16(const Char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt32                  parseUInt32(const Char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt64                  parseUInt64(const Char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryInteger           parseBinaryInteger(const Char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalInteger          parseDecimalInteger(const Char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real32                  parseReal32(const Char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real64                  parseReal64(const Char * pString, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFixed             parseBinaryFixed(const Char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFloat             parseBinaryFloat(const Char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFixed            parseDecimalFixed(const Char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFloat            parseDecimalFloat(const Char * pString) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int8                    parseInt8(const String & string, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int16                   parseInt16(const String & string, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int32                   parseInt32(const String & string, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int64                   parseInt64(const String & string, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt8                   parseUInt8(const String & string, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt16                  parseUInt16(const String & string, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt32                  parseUInt32(const String & string, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt64                  parseUInt64(const String & string, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryInteger           parseBinaryInteger(const String & string) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default integer parse properties.
         *
         *  @param string           The string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalInteger          parseDecimalInteger(const String & string) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real32                  parseReal32(const String & string, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real64                  parseReal64(const String & string, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFixed             parseBinaryFixed(const String & string) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFloat             parseBinaryFloat(const String & string) const throw(ParseException);

        //  Helpers for custom parsing of full strings
        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFixed            parseDecimalFixed(const String & string) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by default real parse properties.
         *
         *  @param string           The string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFloat            parseDecimalFloat(const String & string) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int8                    parseInt8(const char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int16                   parseInt16(const char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int32                   parseInt32(const char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int64                   parseInt64(const char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt8                   parseUInt8(const char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt16                  parseUInt16(const char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt32                  parseUInt32(const char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt64                  parseUInt64(const char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryInteger           parseBinaryInteger(const char * pString, const IntegerParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalInteger          parseDecimalInteger(const char * pString, const IntegerParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real32                  parseReal32(const char * pString, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real64                  parseReal64(const char * pString, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFixed             parseBinaryFixed(const char * pString, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFloat             parseBinaryFloat(const char * pString, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFixed            parseDecimalFixed(const char * pString, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFloat            parseDecimalFloat(const char * pString, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int8                    parseInt8(const Char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int16                   parseInt16(const Char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int32                   parseInt32(const Char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int64                   parseInt64(const Char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt8                   parseUInt8(const Char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt16                  parseUInt16(const Char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt32                  parseUInt32(const Char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt64                  parseUInt64(const Char * pString, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryInteger           parseBinaryInteger(const Char * pString, const IntegerParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalInteger          parseDecimalInteger(const Char * pString, const IntegerParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real32                  parseReal32(const Char * pString, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real64                  parseReal64(const Char * pString, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFixed             parseBinaryFixed(const Char * pString, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFloat             parseBinaryFloat(const Char * pString, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFixed            parseDecimalFixed(const Char * pString, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param pString          The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFloat            parseDecimalFloat(const Char * pString, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int8                    parseInt8(const String & string, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int16                   parseInt16(const String & string, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int32                   parseInt32(const String & string, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Int64                   parseInt64(const String & string, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt8                   parseUInt8(const String & string, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt16                  parseUInt16(const String & string, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt32                  parseUInt32(const String & string, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        UInt64                  parseUInt64(const String & string, const IntegerParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryInteger           parseBinaryInteger(const String & string, const IntegerParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified integer parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalInteger          parseDecimalInteger(const String & string, const IntegerParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real32                  parseReal32(const String & string, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @param pOverflow        If not NULL and an overflow occurs (i.e. the
         *                          parsed value is too large to be represented
         *                          using the required parse result type), a variable
         *                          pointer to by this parameter is set to true.
         *                          If not NULL and no overflow occurs, a variable
         *                          pointer to by this parameter is set to false.
         *                          If NULL, overflow is not reported.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        Real64                  parseReal64(const String & string, const RealParseProperties & parseProperties, bool * pOverflow = NULL) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFixed             parseBinaryFixed(const String & string, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        BinaryFloat             parseBinaryFloat(const String & string, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFixed            parseDecimalFixed(const String & string, const RealParseProperties & parseProperties) const throw(ParseException);

        /**
         *  Parses the specified string for a value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The parsing is governed by the specified real parse properties.
         *
         *  @param string           The string to parse.
         *  @param parseProperties  The custom rules specifying how to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        DecimalFloat            parseDecimalFloat(const String & string, const RealParseProperties & parseProperties) const throw(ParseException);

        //////////
        //  Implementation helpers (also used in derived classes)
    protected:
        static void             _skipSpaces(const char * pString, UIntSize & position);
        static void             _skipSpaces(const Char * pString, UIntSize & position);
        static void             _skipSpaces(const String & string, UIntSize & position);
        static void             _ensureEos(const char * pString, UIntSize position);
        static void             _ensureEos(const Char * pString, UIntSize position);
        static void             _ensureEos(const String & string, UIntSize position);

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };

    /**
     *  @class DateTimeParse CARL.hpp
     *  A generic parser for DateTime values.
     *
     *  Parsing date+time values (i.e. converting them to string representation)
     *  is performed by locale-specific instances of concrete DateTimeParse-derived
     *  classes. A DateTimeParse subclass for a specific locale "knows" about rules
     *  for representing dates+times that are specific to the appropriate locale. The
     *  constraints on the required representation of the parsed value can be varied 
     *  to a certain degree by specifying style, precision and/or custom formst 
     *  when parsing a value.
     *
     *  When obtaining a date+time parser for a specified locale, it may well
     *  be the case that there is no date+time parser registered for that locale.
     *  In this case the immediate parent locale is tried next, and the parent of 
     *  that if necessary, until either an invariant locale is reached or a date+time
     *  parser is found. In the former case, a default date+time parser is
     *  used. This behaviour allows specifying locale-generic date+time parsers
     *  (e.g. if an "German neutral" date+time parser is specified, then it will
     *  be used for both "German/Germany" and "German/Switzerland" locales provided
     *  no separate date+time parsers are registered for either of those).
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DateTimeParse
    {
        friend class _impl::_DateTimeFormatManager;
        //////////
        //  Can't assign or copy construct
    private:
        DateTimeParse(const DateTimeParse &);
        void                        operator = (const DateTimeParse &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        DateTimeParse() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~DateTimeParse() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified date+time parse with the CARL framework.
         *
         *  A registration of a date+time parse informs the CARL framework that
         *  a specific date+time parse exists. From this point on, the application
         *  can use the newly registers date+time parse in exactly the same manner
         *  as a predefined date+time parse. Once registered, a date+time parse
         *  cannot be unregistered.
         *
         *  A date+time parse is always registered for a specific locale. Any attempt
         *  to get a date+time parse instance for that locale will always yield the
         *  date+time parse registered for that locale. It is possible to register
         *  the same date+time parse for more than one locale.
         *
         *  @param forLocale        The locale for which the newly registered
         *                          date+time parse shall be used.
         *  @param pDateTimeParse   The date+time parse to register.
         *  @since 1.0
         */
        static void                 registerInstance(const Locale & forLocale, DateTimeParse * pDateTimeParse);

        /**
         *  Returns a date+time parser for the current default locale.
         *
         *  @return                 A date+time parser for the current default locale.
         *  @since 1.0
         */
        static DateTimeParse *      getInstance();

        /**
         *  Returns a date+time parser for the specified locale.
         *
         *  @param forLocale        The locale for which a date+time parser is sought.
         *  @return                 A date+time parser for the specified locale.
         *  @since 1.0
         */
        static DateTimeParse *      getInstance(const Locale & forLocale);

        //  Parsing
        /**
         *  Represents various date+time representation styles.
         *
         *  @since 1.0
         */
        enum Style
        {
            /**
             *  The shortest possible representation for formatted date+time.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Short             = 1,

            /**
             *  The somewhat longer representation for formatted date+time.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Medium            = 2,

            /**
             *  The longer yet representation for formatted date+time.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Long              = 3,

            /**
             *  The longest possible representation for formatted date+time.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Full              = 4
        };

        /**
         *  Represents a precision with which the time portion of date+time
         *  is represented.
         *
         *  @since 1.0
         */
        enum Precision
        {
            /**
             *  The time is represented with a minute precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Minute        = 1,

            /**
             *  The time is represented with a second precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Second        = 2,

            /**
             *  The time is represented with a millisecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Millisecond   = 3,

            /**
             *  The time is represented with a microsecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Microsecond   = 4,

            /**
             *  The time is represented with a nanosecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Nanosecond    = 5
        };

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported style and/or precision. If more than one style and
         *  precision can be used to successfully parse the date+time value, the
         *  style/precision combination is selected that matches the longest
         *  date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const char * pDateTimeString, UIntSize & position) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported precision. If more than one precision can be used 
         *  to successfully parse the date+time value, the precision is selected 
         *  that matches the longest date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const char * pDateTimeString, UIntSize & position, Style style) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date+time parser allows the date+time value to be represented
         *  with any supported style. If more than one style can be used 
         *  to successfully parse the date+time value, the style is selected 
         *  that matches the longest date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const char * pDateTimeString, UIntSize & position, Precision precision) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const char * pDateTimeString, UIntSize & position, Style style, Precision precision) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The format in which a date+time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date or time must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param format           The expected format for date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const char * pDateTimeString, UIntSize & position, const String & format) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported style and/or precision. If more than one style and
         *  precision can be used to successfully parse the date+time value, the
         *  style/precision combination is selected that matches the longest
         *  date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const Char * pDateTimeString, UIntSize & position) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported precision. If more than one precision can be used 
         *  to successfully parse the date+time value, the precision is selected 
         *  that matches the longest date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const Char * pDateTimeString, UIntSize & position, Style style) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date+time parser allows the date+time value to be represented
         *  with any supported style. If more than one style can be used 
         *  to successfully parse the date+time value, the style is selected 
         *  that matches the longest date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const Char * pDateTimeString, UIntSize & position, Precision precision) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const Char * pDateTimeString, UIntSize & position, Style style, Precision precision) const throw(ParseException) = 0;
        
        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The format in which a date+time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date or time must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param format           The expected format for date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const Char * pDateTimeString, UIntSize & position, const String & format) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported style and/or precision. If more than one style and
         *  precision can be used to successfully parse the date+time value, the
         *  style/precision combination is selected that matches the longest
         *  date+time substring.
         *
         *  @param dateTimeString   The string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const String & dateTimeString, UIntSize & position) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported precision. If more than one precision can be used 
         *  to successfully parse the date+time value, the precision is selected 
         *  that matches the longest date+time substring.
         *
         *  @param dateTimeString   The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const String & dateTimeString, UIntSize & position, Style style) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date+time parser allows the date+time value to be represented
         *  with any supported style. If more than one style can be used 
         *  to successfully parse the date+time value, the style is selected 
         *  that matches the longest date+time substring.
         *
         *  @param dateTimeString   The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const String & dateTimeString, UIntSize & position, Precision precision) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  @param dateTimeString   The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const String & dateTimeString, UIntSize & position, Style style, Precision precision) const throw(ParseException) = 0;

        /**
         *  Parses a date+time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The format in which a date+time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date or time must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param dateTimeString   The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param format           The expected format for date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const String & dateTimeString, UIntSize & position, const String & format) const throw(ParseException) = 0;

        //  Parsing of full strings
        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported style and/or precision. If more than one style and
         *  precision can be used to successfully parse the date+time value, the
         *  style/precision combination is selected that matches the longest
         *  date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const char * pDateTimeString) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported precision. If more than one precision can be used 
         *  to successfully parse the date+time value, the precision is selected 
         *  that matches the longest date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const char * pDateTimeString, Style style) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date+time parser allows the date+time value to be represented
         *  with any supported style. If more than one style can be used 
         *  to successfully parse the date+time value, the style is selected 
         *  that matches the longest date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const char * pDateTimeString, Precision precision) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const char * pDateTimeString, Style style, Precision precision) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The format in which a date+time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date or time must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param format           The expected format for date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const char * pDateTimeString, const String & format) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported style and/or precision. If more than one style and
         *  precision can be used to successfully parse the date+time value, the
         *  style/precision combination is selected that matches the longest
         *  date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const Char * pDateTimeString) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported precision. If more than one precision can be used 
         *  to successfully parse the date+time value, the precision is selected 
         *  that matches the longest date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const Char * pDateTimeString, Style style) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date+time parser allows the date+time value to be represented
         *  with any supported style. If more than one style can be used 
         *  to successfully parse the date+time value, the style is selected 
         *  that matches the longest date+time substring.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const Char * pDateTimeString, Precision precision) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const Char * pDateTimeString, Style style, Precision precision) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The format in which a date+time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date or time must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param pDateTimeString  The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param format           The expected format for date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const Char * pDateTimeString, const String & format) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported style and/or precision. If more than one style and
         *  precision can be used to successfully parse the date+time value, the
         *  style/precision combination is selected that matches the longest
         *  date+time substring.
         *
         *  @param dateTimeString   The string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const String & dateTimeString) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date+time parser allows the date+time value to be represented
         *  in any supported precision. If more than one precision can be used 
         *  to successfully parse the date+time value, the precision is selected 
         *  that matches the longest date+time substring.
         *
         *  @param dateTimeString   The string to parse.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const String & dateTimeString, Style style) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date+time parser allows the date+time value to be represented
         *  with any supported style. If more than one style can be used 
         *  to successfully parse the date+time value, the style is selected 
         *  that matches the longest date+time substring.
         *
         *  @param dateTimeString   The string to parse.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const String & dateTimeString, Precision precision) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  @param dateTimeString   The string to parse.
         *  @param style            The expected style of the date+time string
         *                          representation.
         *  @param precision        The expected precision of the date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const String & dateTimeString, Style style, Precision precision) const throw(ParseException);

        /**
         *  Parses the specified string for a date+time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The format in which a date+time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date or time must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param dateTimeString   The string to parse.
         *  @param format           The expected format for date+time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual DateTime            parse(const String & dateTimeString, const String & format) const throw(ParseException);

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };

    /**
     *  @class DateParse CARL.hpp
     *  A generic parser for Date values.
     *
     *  Parsing date values (i.e. converting them to string representation)
     *  is performed by locale-specific instances of concrete DateParse-derived
     *  classes. A DateParse subclass for a specific locale "knows" about rules
     *  for representing dates that are specific to the appropriate locale. The
     *  constraints on the required representation of the parsed value can be varied 
     *  to a certain degree by specifying style, precision and/or custom formst 
     *  when parsing a value.
     *
     *  When obtaining a date parser for a specified locale, it may well
     *  be the case that there is no date parser registered for that locale.
     *  In this case the immediate parent locale is tried next, and the parent of 
     *  that if necessary, until either an invariant locale is reached or a date
     *  parser is found. In the former case, a default date parser is
     *  used. This behaviour allows specifying locale-generic date parsers
     *  (e.g. if an "German neutral" date parser is specified, then it will
     *  be used for both "German/Germany" and "German/Switzerland" locales provided
     *  no separate date parsers are registered for either of those).
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC DateParse
    {
        friend class _impl::_DateTimeFormatManager;
        //////////
        //  Can't assign or copy construct
    private:
        DateParse(const DateParse &);
        void                        operator = (const DateParse &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        DateParse() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~DateParse() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified date parse with the CARL framework.
         *
         *  A registration of a date parse informs the CARL framework that
         *  a specific date parse exists. From this point on, the application
         *  can use the newly registers date parse in exactly the same manner
         *  as a predefined date parse. Once registered, a date parse
         *  cannot be unregistered.
         *
         *  A date parse is always registered for a specific locale. Any attempt
         *  to get a date parse instance for that locale will always yield the
         *  date parse registered for that locale. It is possible to register
         *  the same date parse for more than one locale.
         *
         *  @param forLocale        The locale for which the newly registered
         *                          date parse shall be used.
         *  @param pDateParse       The date parse to register.
         *  @since 1.0
         */
        static void                 registerInstance(const Locale & forLocale, DateParse * pDateParse);

        /**
         *  Returns a date parser for the current default locale.
         *
         *  @return                 A date parser for the current default locale.
         *  @since 1.0
         */
        static DateParse *          getInstance();

        /**
         *  Returns a date parser for the specified locale.
         *
         *  @param forLocale        The locale for which a date parser is sought.
         *  @return                 A date parser for the specified locale.
         *  @since 1.0
         */
        static DateParse *          getInstance(const Locale & forLocale);

        //  Parsing
        /**
         *  Represents various date representation styles.
         *
         *  @since 1.0
         */
        enum Style
        {
            /**
             *  The shortest possible representation for formatted date.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Short             = 1,

            /**
             *  The somewhat longer representation for formatted date.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Medium            = 2,

            /**
             *  The longer yet representation for formatted date.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Long              = 3,

            /**
             *  The longest possible representation for formatted date.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Style_Full              = 4
        };

        /**
         *  Parses a date from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date parser allows the date value to be represented
         *  in any supported style. If more than one style can be used to 
         *  successfully parse the date value, the style is selected that 
         *  matches the longest date substring.
         *
         *  @param pDateString      The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const char * pDateString, UIntSize & position) const throw(ParseException) = 0;

        /**
         *  Parses a date from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  @param pDateString      The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param style            The expected style of the date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const char * pDateString, UIntSize & position, Style style) const throw(ParseException) = 0;

        /**
         *  Parses a date from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The format in which a dat value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *
         *  @param pDateString      The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param format           The expected format for date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const char * pDateString, UIntSize & position, const String & format) const throw(ParseException) = 0;

        /**
         *  Parses a date from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date parser allows the date value to be represented
         *  in any supported style. If more than one style can be used to 
         *  successfully parse the date value, the style is selected that 
         *  matches the longest date substring.
         *
         *  @param pDateString      The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const Char * pDateString, UIntSize & position) const throw(ParseException) = 0;

        /**
         *  Parses a date from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  @param pDateString      The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param style            The expected style of the date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const Char * pDateString, UIntSize & position, Style style) const throw(ParseException) = 0;

        /**
         *  Parses a date from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The format in which a dat value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *
         *  @param pDateString      The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param format           The expected format for date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const Char * pDateString, UIntSize & position, const String & format) const throw(ParseException) = 0;

        /**
         *  Parses a date from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The date parser allows the date value to be represented
         *  in any supported style. If more than one style can be used to 
         *  successfully parse the date value, the style is selected that 
         *  matches the longest date substring.
         *
         *  @param dateString       The string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const String & dateString, UIntSize & position) const throw(ParseException) = 0;

        /**
         *  Parses a date from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  @param dateString       The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param style            The expected style of the date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const String & dateString, UIntSize & position, Style style) const throw(ParseException) = 0;

        /**
         *  Parses a date from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The format in which a dat value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *
         *  @param dateString       The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param format           The expected format for date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const String & dateString, UIntSize & position, const String & format) const throw(ParseException) = 0;

        //  Parsing of full strings
        /**
         *  Parses the specified string for a date value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date parser allows the date value to be represented
         *  in any supported style. If more than one style can be used to 
         *  successfully parse the date value, the style is selected that 
         *  matches the longest date substring.
         *
         *  @param pDateString      The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const char * pDateString) const throw(ParseException);

        /**
         *  Parses the specified string for a date value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  @param pDateString      The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param style            The expected style of the date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const char * pDateString, Style style) const throw(ParseException);

        /**
         *  Parses the specified string for a date value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The format in which a dat value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *
         *  @param pDateString      The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param format           The expected format for date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const char * pDateString, const String & format) const throw(ParseException);

        /**
         *  Parses the specified string for a date value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date parser allows the date value to be represented
         *  in any supported style. If more than one style can be used to 
         *  successfully parse the date value, the style is selected that 
         *  matches the longest date substring.
         *
         *  @param pDateString      The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const Char * pDateString) const throw(ParseException);

        /**
         *  Parses the specified string for a date value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  @param pDateString      The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param style            The expected style of the date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const Char * pDateString, Style style) const throw(ParseException);

        /**
         *  Parses the specified string for a date value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The format in which a dat value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *
         *  @param pDateString      The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param format           The expected format for date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const Char * pDateString, const String & format) const throw(ParseException);

        /**
         *  Parses the specified string for a date value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The date parser allows the date value to be represented
         *  in any supported style. If more than one style can be used to 
         *  successfully parse the date value, the style is selected that 
         *  matches the longest date substring.
         *
         *  @param dateString       The string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const String & dateString) const throw(ParseException);

        /**
         *  Parses the specified string for a date value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  @param dateString       The string to parse.
         *  @param style            The expected style of the date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const String & dateString, Style style) const throw(ParseException);

        /**
         *  Parses the specified string for a date value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The format in which a dat value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of date must occur. The following placeholders are defined:
         *      -   {year:full} - Expects a full (4-digit) year (0001..9999).
         *      -   {year:short} - Expects a short (2-digit) year (00..99).
         *      -   {month:full} - Expects a full (2-digit) month number (01..12).
         *      -   {month:short} - Expects a short (1- or 2-digit) month number (1..12).
         *      -   {month_name:full} - Expects a full month name.
         *      -   {month_name:short} - Expects an abbreviated month name.
         *      -   {day:full} - Expects a full (2-digit) day (01..31).
         *      -   {day:short} - Expects a short (1- or 2-digit) day (1..31).
         *      -   {day_of_week_name:full} - Expects a full day-of-week name.
         *      -   {day_of_week_name:short} - Expects an abbreviated day-of-week name.
         *
         *  @param dateString       The string to parse.
         *  @param format           The expected format for date string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Date                parse(const String & dateString, const String & format) const throw(ParseException);

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };

    /**
     *  @class TimeParse CARL.hpp
     *  A generic parser for Time values.
     *
     *  Parsing time values (i.e. converting them to string representation)
     *  is performed by locale-specific instances of concrete TimeParse-derived
     *  classes. A TimeParse subclass for a specific locale "knows" about rules
     *  for representing times that are specific to the appropriate locale. The
     *  constraints on the required representation of the parsed value can be varied 
     *  to a certain degree by specifying style, precision and/or custom formst 
     *  when parsing a value.
     *
     *  When obtaining a time parser for a specified locale, it may well
     *  be the case that there is no time parser registered for that locale.
     *  In this case the immediate parent locale is tried next, and the parent of 
     *  that if necessary, until either an invariant locale is reached or a time
     *  parser is found. In the former case, a default time parser is
     *  used. This behaviour allows specifying locale-generic time parsers
     *  (e.g. if an "German neutral" time parser is specified, then it will
     *  be used for both "German/Germany" and "German/Switzerland" locales provided
     *  no separate time parsers are registered for either of those).
     *
     *  It is recommended that concrete subclasses of this class shall be
     *  singletons.
     *
     *  @since 1.0
     */
    class CARL_PUBLIC TimeParse
    {
        friend class _impl::_DateTimeFormatManager;
        //////////
        //  Can't assign or copy construct
    private:
        TimeParse(const TimeParse &);
        void                        operator = (const TimeParse &);

        //////////
        //  Construction/destruction - from derived classes only
    protected:
        /**
         *  The default constructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        TimeParse() {}

        /**
         *  The class destructor.
         *
         *  Can be used from derived classes only.
         *
         *  @since 1.0
         */
        virtual ~TimeParse() {}

        //////////
        //  Operations
    public:
        /**
         *  Registers the specified time parse with the CARL framework.
         *
         *  A registration of a time parse informs the CARL framework that
         *  a specific time parse exists. From this point on, the application
         *  can use the newly registers time parse in exactly the same manner
         *  as a predefined time parse. Once registered, a time parse
         *  cannot be unregistered.
         *
         *  A time parse is always registered for a specific locale. Any attempt
         *  to get a time parse instance for that locale will always yield the
         *  time parse registered for that locale. It is possible to register
         *  the same time parse for more than one locale.
         *
         *  @param forLocale        The locale for which the newly registered
         *                          time parse shall be used.
         *  @param pTimeParse       The time parse to register.
         *  @since 1.0
         */
        static void                 registerInstance(const Locale & forLocale, TimeParse * pTimeParse);

        /**
         *  Returns a time parser for the current default locale.
         *
         *  @return                 A time parser for the current default locale.
         *  @since 1.0
         */
        static TimeParse *          getInstance();

        /**
         *  Returns a time parser for the specified locale.
         *
         *  @param forLocale        The locale for which a time parser is sought.
         *  @return                 A time parser for the specified locale.
         *  @since 1.0
         */
        static TimeParse *          getInstance(const Locale & forLocale);

        //  Parsing
        /**
         *  Represents a precision with which the time is represented.
         *
         *  @since 1.0
         */
        enum Precision
        {
            /**
             *  The time is represented with a minute precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Minute        = 1,

            /**
             *  The time is represented with a second precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Second        = 2,

            /**
             *  The time is represented with a millisecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Millisecond   = 3,

            /**
             *  The time is represented with a microsecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Microsecond   = 4,

            /**
             *  The time is represented with a nanosecond precision.
             *
             *  @hideinitializer
             *  @since 1.0
             */
            Precision_Nanosecond    = 5
        };

        /**
         *  Parses a time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The time parser allows the time value to be represented
         *  with any supported precision. If more than one precision can be 
         *  used to successfully parse the time value, the precision 
         *  is selected that matches the longest time substring.
         *
         *  @param pTimeString      The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const char * pTimeString, UIntSize & position) const throw(ParseException) = 0;

        /**
         *  Parses a time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  @param pTimeString      The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param precision        The expected precision of the time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const char * pTimeString, UIntSize & position, Precision precision) const throw(ParseException) = 0;

        /**
         *  Parses a time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The format in which a time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of time must occur. The following placeholders are defined:
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param pTimeString      The pointer to a nul-terminated ASCII
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param format           The expected format for time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const char * pTimeString, UIntSize & position, const String & format) const throw(ParseException) = 0;

        /**
         *  Parses a time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The time parser allows the time value to be represented
         *  with any supported precision. If more than one precision can be 
         *  used to successfully parse the time value, the precision 
         *  is selected that matches the longest time substring.
         *
         *  @param pTimeString      The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const Char * pTimeString, UIntSize & position) const throw(ParseException) = 0;

        /**
         *  Parses a time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  @param pTimeString      The pointer to a nul-terminated ISo-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param precision        The expected precision of the time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const Char * pTimeString, UIntSize & position, Precision precision) const throw(ParseException) = 0;

        /**
         *  Parses a time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The format in which a time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of time must occur. The following placeholders are defined:
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param pTimeString      The pointer to a nul-terminated ISO-10646
         *                          string to parse from.
         *  @param position         The current parse position within the string.
         *  @param format           The expected format for time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const Char * pTimeString, UIntSize & position, const String & format) const throw(ParseException) = 0;

        /**
         *  Parses a time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The time parser allows the time value to be represented
         *  with any supported precision. If more than one precision can be 
         *  used to successfully parse the time value, the precision 
         *  is selected that matches the longest time substring.
         *
         *  @param timeString       The string to parse from.
         *  @param position         The current parse position within the string.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const String & timeString, UIntSize & position) const throw(ParseException) = 0;

        /**
         *  Parses a time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  @param timeString       The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param precision        The expected precision of the time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const String & timeString, UIntSize & position, Precision precision) const throw(ParseException) = 0;

        /**
         *  Parses a time from within the specified string, starting at 
         *  the specified position.
         *
         *  Upon successful parse, the position is advanced to just after 
         *  the value that has been parsed. If the parsing fails, the
         *  position is advanced to the first character where a parse
         *  error was detected.
         *
         *  The format in which a time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of time must occur. The following placeholders are defined:
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param timeString       The string to parse from.
         *  @param position         The current parse position within the string.
         *  @param format           The expected format for time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const String & timeString, UIntSize & position, const String & format) const throw(ParseException) = 0;

        //  Parsing of full strings
        /**
         *  Parses the specified string for a time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The time parser allows the time value to be represented
         *  with any supported precision. If more than one precision can be 
         *  used to successfully parse the time value, the precision 
         *  is selected that matches the longest time substring.
         *
         *  @param pTimeString      The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const char * pTimeString) const throw(ParseException);

        /**
         *  Parses the specified string for a time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  @param pTimeString      The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param precision        The expected precision of the time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const char * pTimeString, Precision precision) const throw(ParseException);

        /**
         *  Parses the specified string for a time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The format in which a time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of time must occur. The following placeholders are defined:
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param pTimeString      The pointer to a nul-terminated ASCII
         *                          string to parse.
         *  @param format           The expected format for time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const char * pTimeString, const String & format) const throw(ParseException);

        /**
         *  Parses the specified string for a time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The time parser allows the time value to be represented
         *  with any supported precision. If more than one precision can be 
         *  used to successfully parse the time value, the precision 
         *  is selected that matches the longest time substring.
         *
         *  @param pTimeString      The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const Char * pTimeString) const throw(ParseException);

        /**
         *  Parses the specified string for a time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  @param pTimeString      The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param precision        The expected precision of the time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const Char * pTimeString, Precision precision) const throw(ParseException);

        /**
         *  Parses the specified string for a time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The format in which a time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of time must occur. The following placeholders are defined:
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param pTimeString      The pointer to a nul-terminated ISO-10646
         *                          string to parse.
         *  @param format           The expected format for time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const Char * pTimeString, const String & format) const throw(ParseException);

        /**
         *  Parses the specified string for a time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The time parser allows the time value to be represented
         *  with any supported precision. If more than one precision can be 
         *  used to successfully parse the time value, the precision 
         *  is selected that matches the longest time substring.
         *
         *  @param timeString       The string to parse.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const String & timeString) const throw(ParseException);

        /**
         *  Parses the specified string for a time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  @param timeString       The string to parse.
         *  @param precision        The expected precision of the time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const String & timeString, Precision precision) const throw(ParseException);

        /**
         *  Parses the specified string for a time value.
         *
         *  The entire string is assumed to be a string representation of
         *  a value to parse (i.e. parsing must consume the entire string).
         *
         *  The format in which a time value is expected to occur is specified
         *  by a format string, which is interpreted according to the following rules:
         *
         *  -   Any character other than whitespace must match the correcponsing
         *      character in the parsed string. Matching is case-insensitive.
         *  -   Any whitespace character matches zero or more consecutive
         *      whitespaces in the parsed string.
         *  -   Any character preceeded by a backslash is treated literally even
         *      it it would otherwise have a special meaning. In particular, to
         *      match a single backslash in a parsed string the format string must
         *      specify two consecutive backslashed.
         *  -   A placeholder marks the place in the parsed string where some portion
         *      of time must occur. The following placeholders are defined:
         *      -   {hour12:full} - Expects a full (2-digit) hour on a 12-hour scale (01..12).
         *      -   {hour12:short} - Expects a short (1- or 2-digit) hour on a 12-hour scale (1..12).
         *      -   {hour24:full} - Expects a full (2-digit) hour on a 24-hour scale (00..23).
         *      -   {hour24:short} - Expects a short (1- or 2-digit) hour on a 24-hour scale (0..23).
         *      -   {minute:full} - Expects a full (2-digit) minutes (00..59).
         *      -   {minute:short} - Expects a short (1- or 2-digit) minutes (0..59).
         *      -   {second:full} - Expects a full (2-digit) seconds (00..59).
         *      -   {second:short} - Expects a short (1- or 2-digit) seconds (0..59).
         *      -   {millisecond:full} - Expects a full (3-digit) milliseconds (000..999).
         *      -   {millisecond:short} - Expects a full (1- to 3-digit) milliseconds (0..999).
         *      -   {microsecond:full} - Expects a full (6-digit) microseconds (000000..999999).
         *      -   {microsecond:short} - Expects a short (1- to 6-digit) microseconds (0..999999).
         *      -   {nanosecond:full} - Expects a full (9-digit) nanoseconds (000000000..999999999).
         *      -   {nanosecond:short} - Expects a short (1- to 9-digit) nanoseconds (0..999999999).
         *      -   {ampm} - Expects a locale-specific AM/PM indicator.
         *
         *  @param timeString       The string to parse.
         *  @param format           The expected format for time string
         *                          representation.
         *  @return                 The parsed value.
         *  @exception ParseException   Thrown if parsing could not be completed
         *                          successfully for any reason.
         *  @since 1.0
         */
        virtual Time                parse(const String & timeString, const String & format) const throw(ParseException);

        //////////
        //  Implementation
    private:
        static CriticalSection *    _getStaticGuard();
    };
}

#endif  //  __CARL_Parse_hpp_included__
//  End of CARL_Parse.hpp
